; ===========================================================================
; ---------------------------------------------------------------------------
; The Checksum routine portion of the ROM
; ---------------------------------------------------------------------------
; NHL 94's checksum system is quite an interesting one.  It seems the routine
; does NOT count itself as part of the checksum.  Meaning, the addition stops
; JUST BEFORE this checksum routine, only the rest of the ROM is checked.
;
; There is no way of knowing where the checksum routine is in the ROM
; especially if it is edit-able, making it unfamiliar.  So to get around this
; the main NHL game and this checksum routine are assembled seperately, so
; that a checksum count can be done properly with "CheckFix.exe".
;
; After this, "PatchRom.exe" will patch the checksum routine ROM onto the
; "main" ROM, thus, ensuring the ROM is assembled 1:1 100% perfectly against
; the original.
; ---------------------------------------------------------------------------

		include	"EquMain.asm"				; main 68k source labels/equates/etc
		include	"CheckValue.asm"			; checksum value (generated by CheckFix)

; ===========================================================================
; ---------------------------------------------------------------------------
; Start of the portion (this is technically starting at "CalcChecksum"'s address)
; ---------------------------------------------------------------------------

StartCheck:	; a.k.a. "CalcChecksum:"
		moveq	#$00,d0					; clear d0
		suba.l	a0,a0					; clear a0 (address is now the beginning of the ROM)
		move.l	#(CalcChecksum/$04),d1			; set number of long-words to add (from the start til this checksum routine)

SC_NextLong:
		cmp.w	#Checksum,a0				; is the next long-word to add at the "proper" checksum space?
		bne.s	SC_NoProper				; if not, branch
		addq.w	#$04,a0					; skip over it (this is useless since the "proper" checksum space isn't used for varification
		bra.s	SC_YesProper				; continue to decrement counter

SC_NoProper:
		add.l	(a0)+,d0				; add long-word to sum

SC_YesProper:
		subq.l	#$01,d1					; decrease long-word counter
		bgt.s	SC_NextLong				; if it's still above 0 (still counting), branch
		cmpi.l	#CheckValue,d0				; does the sum match the expected value?
		bne.s	SC_Error				; if not, branch (there was a problem, incorrectly flushed/downloaded)
		rts						; return back to main ROM (and main game, checksum is fine).

	; --- When an error is found ---

SC_Error:
		move.l	#$C00004,a4				; load VDP control port address
		move.w	#$8F00|$02,(a4)				; 7654 3210 - Auto Increament: By 2
		move.w	#$8000|%00000100,(a4)			; 00LH 01CD - Leftover SMS bar (0N|1Y) | H-Interrupt (0N|1Y) | H,V Counter (0N|1Y) | Disable video signal (0N|1Y)
		move.w	#$8700|$00,(a4)				; 00PP CCCC - Backdrop Colour: Palette Line 0/Colour ID 0
		move.w	#$8100|%01000100,(a4)			; SDVM P100 - SMS mode (0N|1Y) | Display (0N|1Y) | V-Interrupt (0N|1Y) | DMA (0N|1Y) | V-resolution (0-1C|1-1E)
		move.w	#$C000,(a4)				; set VDP to CRAM write mode (this is done incorrectly...)
	;	move.w	#$0000,(a4)		; bug-fix	; uncomment this line to fix the incorrect VDP setting
		move.w	#($0080/2)-1,d1				; set number of colours to write (the entire of CRAM)

SC_SetRedCRAM:
		move.w	#$000E,($C00000).l			; write RED colour to CRAM
		dbf	d1,SC_SetRedCRAM			; repeat for entire of CRAM

		bra.s	*					; trap the CPU in an endless branching loop

; ===========================================================================