; ===========================================================================
; ---------------------------------------------------------------------------
; NHL Hockey 94 (16-bit) Mega Drive source
; ---------------------------------------------------------------------------
; Disassembled December 2016 by MarkeyJester~
; Merry Christmas!
; ---------------------------------------------------------------------------

		include	"Equates.asm"
		include	"Macros.asm"

; ===========================================================================
; ---------------------------------------------------------------------------
; Start of ROM
; ---------------------------------------------------------------------------

StartROM:	dc.l	R_StackSP, SetupGame, ErAddress, ErAddress
		dc.l	ErIllegal, ErrDivide, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	ExpReturn, ExpReturn, ExpReturn, ExpReturn
		dc.l	ExpReturn, $00000000, VBlankInt, ExpReturn
		dc.l	$00000000, $00000000, $00000000, $00000000
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF
		dc.l	$FFFFFFFF, $FFFFFFFF, $FFFFFFFF, $FFFFFFFF

ConsoleName:	dc.b	"SEGA GENESIS    "
ProductDate:	dc.b	"(C)T-50 1993.JUL"
TitleLocal:	dc.b	"NHL Hockey '94                                  "
TitleInter:	dc.b	"NHL Hockey '94                                  "
SerialNumber:	dc.b	"GM T-50656 -"
Checksum:	dc.l	$30305512			; This checksum is technically unused (but also uncounted)
IOSupport:	dc.b	"J               "
ROMStart:	dc.l	StartROM
ROMFinish:	dc.l	FinishROM-1
RAMStart:	dc.l	$00FF0000
RAMFinish:	dc.l	$00FFFFFF
		;	;;;;;%1E1TT000,%00100000	; E = Erase SRAM on start (0 Yes | 1 No) 
SupportRAM:	dc.b	'RA',%11111000,%00100000	; TT = Type (00 Even & Odd | 10 Even | 11 Odd)
SRAMStart:	dc.l	$200001
SRAMFinish:	dc.l	$203FFF
ModemInfo:	dc.b	"                                                    "
ProductRegion:	dc.b	"UE              "

; ===========================================================================
; ---------------------------------------------------------------------------
; Starting routine (setting up the Mega Drive's individual components)
; ---------------------------------------------------------------------------

SetupGame:
		tst.l	($A10008).l				; have controller ports 1 (A) or 2 (B) been initialised?
		bne.s	SG_Reset				; if so, branch
		tst.w	($A1000C).l				; check controller port 3 (ext)

SG_Reset:
		bne.s	SG_SkipSetup				; if any port has been initialise, branch (reset button was pressed).
		lea	SG_Values(pc),a5			; load address of setup values
		movem.w	(a5)+,d5-d7				; load VDP value, 68k RAM size and Z80 port values to d5, d6 and d7 respectively
		movem.l	(a5)+,a0-a4				; load Z80 and VDP access addresses to a0, a1, a2, a3 and a4
		move.b	-$10FF(a1),d0				; load Mega Drive hardware version (offset A10001)
		andi.b	#$0F,d0					; get only the version nybble
		beq.s	SG_NoTMSS				; if it is 0, branch (version 0 does not have TMSS security)
		move.l	#'SEGA',$2F00(a1)			; copy "SEGA" security code to TMSS chip (offset A14000)

SG_NoTMSS:
		move.w	(a4),d0					; release latch
		moveq	#$00,d0					; clear d0
		movea.l	d0,a6					; clear a6
		move.l	a6,usp					; clear user stack pointer (unused)
		moveq	#(SG_RegVDP_End-SG_RegVDP)-1,d1		; set repeat times

SG_SetupVDP:
		move.b	(a5)+,d5				; load value next to register ID
		move.w	d5,(a4)					; save to VDP (setting the register)
		add.w	d7,d5					; increase register ID
		dbf	d1,SG_SetupVDP				; repeat for all VDP registers
		move.l	(a5)+,(a4)				; write DMA destination to VDP (VRAM 0000)
		move.w	d0,(a3)					; set DMA fill value to 00 (DMA starts here, clears entire VRAM)
		move.w	d7,(a1)					; set Z80 stop on
		move.w	d7,(a2)					; set Z80 reset off

SG_WaitZ80:
		btst.b	d0,(a1)					; has the Z80 stopped yet?
		bne.s	SG_WaitZ80				; if not, branch and loop until it has
		moveq	#(SG_Z80_End-SG_Z80)-1,d2		; set repeat times

SG_LoadZ80:
		move.b	(a5)+,(a0)+				; copy Z80 instructions to Z80 RAM
		dbf	d2,SG_LoadZ80				; repeat until all bytes are copied over
		move.w	d0,(a2)					; set Z80 reset on
		move.w	d0,(a1)					; set Z80 stop off
		move.w	d7,(a2)					; set Z80 reset off

SG_ClearRAM:
		move.l	d0,-(a6)				; clear long-word of 68k RAM and move back
		dbf	d6,SG_ClearRAM				; repeat until all 68k RAM from 0000 to FFFF is clear
		move.l	(a5)+,(a4)				; turn VDP DMA off and set VDP auto increment to 2
		move.l	(a5)+,(a4)				; set VDP to CRAM write mode (address 0000)
		moveq	#$20-1,d3				; set repeat times

SG_ClearCRAM:
		move.l	d0,(a3)					; clear colours
		dbf	d3,SG_ClearCRAM				; repeat until the entire palette is clear (black)
		move.l	(a5)+,(a4)				; set VDP to VSRAM write mode (address 0000)
		moveq	#$14-1,d4				; set repeat times

SG_ClearVSRAM:
		move.l	d0,(a3)					; clear vertical scroll positions
		dbf	d4,SG_ClearVSRAM			; repeat until all scroll positions are set to 0000
		moveq	#(SG_PSG_End-SG_PSG)-1,d5		; set repeat times

SG_SetupPSG:
		move.b	(a5)+,$11(a3)				; write PSG volume values to PSG port (offset C00011)
		dbf	d5,SG_SetupPSG				; repeat for all PSG channels
		move.w	d0,(a2)					; set Z80 reset on
		movem.l	(a6),d0-a6				; clear all 68k registers
		move	#$2700,sr				; disable all interrupts

SG_SkipSetup:
		bra.s	StartGame				; continue to actual game code

; ---------------------------------------------------------------------------
; Setup values
; ---------------------------------------------------------------------------

SG_Values:	dc.w	$8000					; VDP register value (80??)
		dc.w	$FFFF/4					; size of 68k RAM
		dc.w	$0100					; Z80 bus/reset value
		dc.l	$A00000					; address of Z80 RAM
		dc.l	$A11100					; Z80 bus request port
		dc.l	$A11200					; Z80 reset port
		dc.l	$C00000					; VDP data port
		dc.l	$C00004					; VDP control port

	; --- VDP register values ---

SG_RegVDP:	dc.b	%00000100				; 00LH 01CD - Leftover SMS bar (0N|1Y) | H-Interrupt (0N|1Y) | H,V Counter (0N|1Y) | Disable video signal (0N|1Y)
		dc.b	%00010100				; SDVM P100 - SMS mode (0N|1Y) | Display (0N|1Y) | V-Interrupt (0N|1Y) | DMA (0N|1Y) | V-resolution (0-1C|1-1E)
		dc.b	(($C000)>>$0A)&$FF			; 00FE DCBA - Scroll Plane A Map Table VRam address
		dc.b	(($F000)>>$0A)&$FF			; 00FE DCB0 / 00FE DC00 (20 H-resol) - Window Plane A Map Table VRam address
		dc.b	(($E000)>>$0D)&$FF			; 0000 0FED - Scroll Plane B Map Table VRam address
		dc.b	(($D800)>>$09)&$FF			; 0FED CBA9 / 0FED CBA0 (20 H-resol) - Sprite Plane Map Table VRam address
		dc.b	%00000000				; 0000 0000 - Unknown/Unused Register
		dc.b	$00					; 00PP CCCC - Backdrop Colour: Palette Line | Colour ID
		dc.b	%00000000				; 0000 0000 - Unknown/Unused Register
		dc.b	%00000000				; 0000 0000 - Unknown/Unused Register
		dc.b	$FF					; 7654 3210 - H-Interrupt Register
		dc.b	%00000000				; 0000 EVHH - External Interrupt (0N|1Y) | V-Scroll (0-Full|1-2Celled) | H-Scroll: (00-Full|10-Celled|11-Sliced)
		dc.b	%10000001				; APHE SNNB - H-resol (0N|1Y) | Pixel int (0N|1Y) | H-sync (0N|1Y) | Extern-pix (0N|1Y) | S/H (0N|1Y) | Interlace (00N|01Y|11-Split) | H-resol (0-20|1-28)
		dc.b	(($DC00)>>$0A)&$FF			; 00FE DCBA - Horizontal Scroll Table VRam address
		dc.b	%00000000				; 0000 0000 - Unknown/Unused Register
		dc.b	$01					; 7654 3210 - Auto Increament
		dc.b	%00000001				; 00VV 00HH - Plane Y Size (00-20|01-40|11-80) | Plane X size (00-20|01-40|11-80)
		dc.b	$00					; 7654 3210 - Window Horizontal Position
		dc.b	$00					; 7654 3210 - Window Vertical Position
		dc.b	$FF					; DMA fill Size (Lower Byte --FF)
		dc.b	$FF					; DMA fill Size (Upper Byte FF--)
		dc.b	$00					; DMA fill Source (Low Byte ----00)
		dc.b	$00					; DMA fill Source (Mid Byte --00--)
		dc.b	$80					; DMA fill Source (High Byte 80----)
SG_RegVDP_End:	dc.l	$40000080				; DMA fill destination (VRAM 0000)

	; --- Z80 instructions ---

SG_Z80:		dc.b	$AF,$01,$D9,$1F,$11,$27,$00,$21		; ...these are just instructions for the Z80
		dc.b	$26,$00,$F9,$77,$ED,$B0,$DD,$E1		; to clear its own RAM, its registers, and to
		dc.b	$FD,$E1,$ED,$47,$ED,$4F,$D1,$E1		; trap itself in an endless loop.
		dc.b	$F1,$08,$D9,$C1,$D1,$E1,$F1,$F9
		dc.b	$F3,$ED,$56,$36,$E9,$E9
SG_Z80_End:	even

	; --- VDP final settings ---

		dc.w	$8104					; Display Off; V-Interrupt Off; DMA Off; V 28 Cell PAL/NTSC Mode
		dc.w	$8F02					; Auto Increament: By 2
		dc.l	$C0000000				; CRAM Write 0000
		dc.l	$40000010				; VSRAM Write 0000

	; --- PSG volume settings ---
	; PSG Volume control format: "PZZF VVVV"
	;	P = Part (1 writing first byte | 0 writing second byte)
	;	ZZ = Channel
	;		00 = PSG 1 tone
	;		01 = PSG 2 tone
	;		10 = PSG 3 tone
	;		11 = PSG 4 noise
	;	F = Function (1 writing volume | 0 writing frequency)
	;	VVVV = Volume (0 = loud | F = quiet)
	; ---------------------------

SG_PSG:		dc.b	%10010000|$0F				; PSG 1 volume (Mute)
		dc.b	%10110000|$0F				; PSG 2 volume (Mute)
		dc.b	%11010000|$0F				; PSG 3 volume (Mute)
		dc.b	%11110000|$0F				; PSG 4 volume (Mute)
SG_PSG_End:	even

; ===========================================================================
; ---------------------------------------------------------------------------
; The start of the actual game's setup
; ---------------------------------------------------------------------------

StartGame:
		tst.w	($C00004).l				; release latch
		jsr	CalcChecksum
		bra.w	RunGame

; ---------------------------------------------------------------------------
dword_30A:	dc.l	0					; DATA XREF: ROM:off_BFD66o
					; ROM:000BFD6Ao ...
		dc.l	unk_5330
		dc.l	unk_966
		dc.l	unk_C4C
		dc.l	unk_F3A
		dc.l	unk_1234
off_322:	dc.l	unk_214E				; DATA XREF: ROM:off_D1AEEo
		dc.l	unk_152E
off_32A:	dc.l	unk_1848				; DATA XREF: ROM:off_C142Ao
					; ROM:off_C1FECo ...
		dc.l	unk_50E6
		dc.l	unk_1B44
		dc.l	unk_1E52
		dc.l	unk_243E
		dc.l	unk_2740
		dc.l	unk_2A58
		dc.l	unk_2D5C
off_34A:	dc.l	unk_3054				; DATA XREF: ROM:off_E9ED6o
		dc.l	unk_3348
		dc.l	unk_3646
		dc.l	unk_393C
		dc.l	unk_3C42
		dc.l	unk_3F30
		dc.l	unk_4216
		dc.l	unk_4518
off_36A:	dc.l	unk_481C				; DATA XREF: ROM:off_C3750o
					; ROM:off_C63F8o
		dc.l	unk_4DFC
		dc.l	unk_4B0A
		dc.l	unk_37E
		dc.l	word_674
unk_37E:	dc.b	0					; DATA XREF: ROM:00000376o
		dc.b	$92					; 		; Palette data
		dc.w	$C
		dc.w	$2C6
		dc.w	$52
		dc.w	$4C
		dc.w	$50
word_38A:	dc.w	$EE8					; DATA XREF: ROM:off_C4608o
					; ROM:off_C5C4Co ...
		dc.w	$222
		dc.w	0
		dc.w	0
		dc.w	$222
		dc.w	$420
		dc.w	$88C
		dc.w	$66A
		dc.w	$2A
		dc.w	$24C
		dc.w	$420
		dc.w	$CCC
		dc.w	$EEE
		dc.w	$AAA
		dc.w	$48C
		dc.w	$A84
word_3AA:	dc.w	$EE8					; DATA XREF: ROM:off_4DEEEo
					; ROM:off_C1034o
		dc.w	$222
		dc.w	0
		dc.w	0
		dc.w	$222
		dc.w	$420
		dc.w	$88C
		dc.w	$66A
		dc.w	$2A
		dc.w	$24C
		dc.w	$CCC
		dc.w	$222
		dc.w	$420
		dc.w	0
		dc.w	$48C
		dc.w	$A84					; Palette data end?
unk_3CA:	dc.b	$77					; w		; DATA XREF: ROM:off_C333Ao
					; ROM:off_C49DEo ...
		dc.b	$20
		dc.b	0
		dc.b	$E8					; 
		dc.b	$CB					; 
		dc.b	$C0					; 
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	4
		dc.b	$10
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	4
		dc.b	$10
		dc.b	$A
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$14
		dc.b	$B
		dc.b	$A
		dc.b	$11
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$17
unk_3EA:	dc.b	$18					; DATA XREF: ROM:off_C08A8o
					; ROM:off_C41D2o ...
		dc.b	5
		dc.b	$C
		dc.b	$E
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	7
		dc.b	4
		dc.b	$10
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	8
		dc.b	9
		dc.b	$11
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	$D
		dc.b	5
		dc.b	6
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$18
unk_40A:	dc.b	$19					; DATA XREF: ROM:off_C1B96o
		dc.b	$F
		dc.b	6
		dc.b	5
		dc.b	$D
		dc.b	0
		dc.w	$E
aPatrickRoy:	dc.b	'Patrick Roy',0				; DATA XREF: ROM:off_54E24o
		dc.b	$33
		dc.b	$66
		dc.b	$44
		dc.b	$46
		dc.b	1
		dc.b	0
		dc.b	$55
		dc.b	$66
		dc.w	$C
aGrantFuhr:	dc.b	'Grant Fuhr'				; DATA XREF: ROM:off_C57D6o
		dc.b	$31
		dc.b	$75
		dc.b	$45
		dc.b	$55
		dc.b	0
		dc.b	0
		dc.b	$44
		dc.b	$55
		dc.w	$E
aTomBarrasso:	dc.b	'Tom Barrasso'
		dc.b	$35					; 5
		dc.b	$A4					; 
unk_44A:	dc.b	$45					; E		; DATA XREF: ROM:off_BF8D0o
					; ROM:off_C0412o
		dc.b	$54					; T
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$10
aMarioLemieux:	dc.b	'Mario Lemieux',0
		dc.b	$66					; f
		dc.b	$A5					; 
		dc.b	$46					; F
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$66					; f
		dc.b	$60					; `
		dc.b	$62					; b
		dc.w	$E
aMarkMessier:	dc.b	'Mark Messier'				; DATA XREF: ROM:off_C29AEo
		dc.b	$11
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$54					; T
		dc.b	$53					; S
		dc.b	$40					; @
		dc.b	$53					; S
		dc.w	$E
aKirkMuller:	dc.b	'Kirk Muller',0				; DATA XREF: ROM:off_C2E64o
					; ROM:off_EAB10o
		dc.b	$10
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$C
aAdamOates:	dc.b	'Adam Oates'
		dc.b	$12
		dc.b	$75					; u
		dc.b	$46					; F
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$54					; T
		dc.b	$50					; P
		dc.b	$61					; a
		dc.w	$C
aJoeSakic:	dc.b	'Joe Sakic',0
		dc.b	$19
		dc.b	$64					; d
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$20
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$52					; R
		dc.w	$10
aPierreTurgeon:	dc.b	'Pierre Turgeon'			; DATA XREF: ROM:off_B389Co
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$51					; Q
		dc.b	$41					; A
		dc.w	$10
aPatLafontaine:	dc.b	'Pat LaFontaine'
		dc.b	$16
		dc.b	$55					; U
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$54					; T
unk_4EA:	dc.b	$60					; `		; DATA XREF: ROM:off_5338Co
		dc.b	$52					; R
		dc.w	$10
aKevinStevens:	dc.b	'Kevin Stevens',0
		dc.b	$25					; %
		dc.b	$B3					; 
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	0
		dc.b	$E
aJaromirJagrh:	dc.b	'Jaromir Jagrh'				; DATA XREF: ROM:off_52DAAo
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$E
aPeterBondra:	dc.b	'Peter Bondra'
		dc.b	$13
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.w	$E
aMikeGartner:	dc.b	'Mike Gartner'
		dc.b	$22					; "
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$52					; R
		dc.w	$E
aRickTocchet:	dc.b	'Rick Tocchet'
		dc.b	$23					; #
		dc.b	$92					; 
		dc.b	$25					; %
		dc.b	$44					; D
		dc.b	$37					; 7
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.w	$12
aAlexnderMogilny: dc.b	'Alexnder Mogilny'			; DATA XREF: ROM:off_EC284o
		dc.b	$89					; 
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$52					; R
		dc.w	$E
aMarkRecchi:	dc.b	'Mark Recchi',0
		dc.b	8
		dc.b	$65					; e
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$43					; C
		dc.b	0
		dc.b	$E
aLarryMurphyu:	dc.b	'Larry MurphyU'
		dc.b	$A4					; 
		dc.b	$34					; 4
		dc.b	$54					; T
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$51					; Q
		dc.b	$43					; C
		dc.w	$E
aRayBourque:	dc.b	'Ray Bourque',0
		dc.b	$77					; w
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$62					; b
		dc.b	$52					; R
		dc.w	$E
aBrianLeetch:	dc.b	'Brian Leetch'
		dc.b	2
		dc.b	$66					; f
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$26					; &
		dc.b	$51					; Q
		dc.b	$52					; R
		dc.b	$52					; R
		dc.w	$10
aSteveDuchesne:	dc.b	'Steve Duchesne'
		dc.b	$28					; (
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$44					; D
unk_5E2:	dc.b	$24					; $		; DATA XREF: ROM:off_B425Ao
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$C
aAlIafrate:	dc.b	'Al Iafrate'
		dc.b	$34					; 4
		dc.b	$B4					; 
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$10
aScottStevens:	dc.b	'Scott Stevens',0
		dc.b	4
		dc.b	$B4					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$56					; V
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$43					; C
		dc.w	$12
aZarleyZalapski: dc.b	'Zarley Zalapski',0
		dc.b	3
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$51					; Q
		dc.b	$33					; 3
		dc.w	$E
aGlenWesley:	dc.b	'Glen Wesley',0
		dc.b	$26					; &
		dc.b	$85					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.b	$41					; A
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	0
		dc.b	2
		dc.w	$10
aAllStarsEast:	dc.b	'All Stars East'			; DATA XREF: ROM:off_EB816o
					; ROM:off_ED1B0o
		dc.w	6
aAse:		dc.b	'ASE',0
		dc.w	2					; This is a blank string
		dc.w	$18
aMadisonSquareGarden: dc.b 'Madison Square Garden',0
word_674:	dc.w	$92					; DATA XREF: ROM:0000037Ao
		dc.w	$C					; More palette data
		dc.w	$2C2
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$20
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	2
		dc.b	$4C					; L
		dc.b	4
		dc.b	$20
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$20
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	2
		dc.b	$4C					; L
		dc.b	$C
		dc.b	$CC					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$20
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$77					; w
		dc.b	$20
		dc.b	0
		dc.b	$E8					; 
		dc.b	$ED					; 
		dc.b	$90					; 
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	$B
		dc.b	4
		dc.b	$F
		dc.b	$E
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$16
		dc.b	$B
		dc.b	4
		dc.b	$F
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$17
		dc.b	$E
		dc.b	6
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	$A
		dc.b	8
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	$B
		dc.b	9
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$16
		dc.b	6
		dc.b	4
		dc.b	$F
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$19
		dc.b	$15
		dc.b	8
		dc.b	$C
		dc.b	9
		dc.b	$11
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$18
		dc.b	9
		dc.b	$11
		dc.b	8
		dc.b	$C
		dc.b	0
		dc.w	$C
aEdBelfour:	dc.b	'Ed Belfour'
		dc.b	$30					; 0
		dc.b	$66					; f
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	0
		dc.b	0
		dc.b	$66					; f
		dc.b	$55					; U
		dc.w	$10
aTimCheveldae:	dc.b	'Tim Cheveldae',0
		dc.b	$32					; 2
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$43					; C
		dc.b	$44					; D
		dc.w	$E
aFelixPotvin:	dc.b	'Felix Potvin'
		dc.b	$29					; )
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$10
aSteveYzerman:	dc.b	'Steve Yzerman',0
		dc.b	$19
		dc.b	$66					; f
		dc.b	$56					; V
		dc.b	$44					; D
		dc.b	$11
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$52					; R
		dc.w	$E
aMikeModano:	dc.b	'Mike Modano',0
		dc.b	9
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$24					; $
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$10
aJeremyRoenick:	dc.b	'Jeremy Roenick'
		dc.b	$27					; '
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$53					; S
		dc.w	$10
aBrianBradley:	dc.b	'Brian Bradley',0
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aDougGilmour:	dc.b	'Doug Gilmour'
		dc.b	$93					; 
		dc.b	$45					; E
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$60					; `
		dc.b	$43					; C
		dc.w	$10
aWayneGretzky:	dc.b	'Wayne Gretzky',0
		dc.b	$99					; 
		dc.b	$46					; F
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$62					; b
		dc.b	$60					; `
		dc.b	$60					; `
		dc.w	$E
aGaryRoberts:	dc.b	'Gary Roberts'
		dc.b	$11
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$51					; Q
		dc.b	$34					; 4
		dc.w	$10
aLucRobitaille:	dc.b	'Luc Robitaille'
		dc.b	$20
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$10
aTheorenFleury:	dc.b	'Theoren Fleury'
		dc.b	$14
		dc.b	$35					; 5
		dc.b	$54					; T
		dc.b	$54					; T
		dc.b	$47					; G
		dc.b	$43					; C
		dc.b	$61					; a
		dc.b	$33					; 3
		dc.w	$C
aBrettHull:	dc.b	'Brett Hull'
		dc.b	$16
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$36					; 6
		dc.b	$33					; 3
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.w	$C
aPavelBure:	dc.b	'Pavel Bure'
		dc.b	$10
		dc.b	$55					; U
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$20
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$42					; B
		dc.w	$10
aTeemuSelanne:	dc.b	'Teemu Selanne',0
		dc.b	$13
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$42					; B
		dc.w	$E
aPatFalloon:	dc.b	'Pat Falloon',0
		dc.b	$18
		dc.b	$74					; t
		dc.b	$43					; C
		dc.b	$23					; #
		dc.b	$13
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.w	$C
aJariKurri:	dc.b	'Jari Kurri'
		dc.b	$17
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$C
aGarySuter:	dc.b	'Gary Suter'
		dc.b	$22					; "
		dc.b	$75					; u
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$C
aJeffBrown:	dc.b	'Jeff Brown'
		dc.b	$21					; !
		dc.b	$93					; 
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$51					; Q
		dc.b	$42					; B
		dc.w	$E
aPaulCoffey:	dc.b	'Paul Coffey',0
		dc.b	$77					; w
		dc.b	$96					; 
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$61					; a
		dc.b	$51					; Q
		dc.b	$53					; S
		dc.w	$10
aChrisChelios:	dc.b	'Chris Chelios',0
		dc.b	7
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$47					; G
		dc.b	$51					; Q
		dc.b	$62					; b
		dc.b	$45					; E
		dc.w	$E
aPhilHousley:	dc.b	'Phil Housley'
		dc.b	6
		dc.b	$66					; f
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$62					; b
		dc.b	$40					; @
		dc.b	$62					; b
		dc.w	$E
aDaveManson:	dc.b	'Dave Manson',0
		dc.b	$24					; $
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.w	$E
aSteveSmith:	dc.b	'Steve Smith',0
		dc.b	5
		dc.b	$B4					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$44					; D
		dc.w	$10
aSteveChiasson:	dc.b	'Steve Chiasson'
		dc.b	3
		dc.b	$94					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.w	2					; Blank	string
		dc.w	$10
aAllStarsWest:	dc.b	'All Stars West',0
		dc.b	6
		dc.b	$41					; A
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	0
		dc.b	0
		dc.b	2
		dc.w	$18
aMadisonSquareGarden_0: dc.b 'Madison Square Garden',0
unk_966:	dc.b	0					; DATA XREF: ROM:00000312o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C0					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	$24					; $
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$8C					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$A
		dc.b	$AA					; 
unk_98A:	dc.b	$C					; DATA XREF: ROM:off_BB4EEo
		dc.b	$CC					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	$24					; $
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$8C					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	$24					; $
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$41					; A
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$F8					; 
		dc.b	$B4					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	8
		dc.b	3
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	7
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$17
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$12
		dc.b	8
		dc.b	3
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$15
		dc.b	7
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	9
		dc.b	4
		dc.b	3
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	3
		dc.b	6
		dc.b	9
		dc.b	4
		dc.b	0
		dc.w	$C
aAndyMoog:	dc.b	'Andy Moog',0
		dc.b	$35					; 5
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$24					; $
		dc.b	0
		dc.b	0
		dc.b	$43					; C
		dc.b	$43					; C
		dc.w	$C
aJohnBlue:	dc.b	'John Blue',0
		dc.b	$39					; 9
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	0
		dc.b	0
		dc.b	$43					; C
		dc.b	$43					; C
		dc.w	$C
aAdamOates_0:	dc.b	'Adam Oates'
		dc.b	$12
		dc.b	$75					; u
		dc.b	$46					; F
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$54					; T
		dc.b	$50					; P
		dc.b	$61					; a
		dc.w	$E
aDavePoulin:	dc.b	'Dave Poulin',0
		dc.b	$19
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$12
aVladimirRuzicka: dc.b	'Vladimir Ruzicka'
		dc.b	$38					; 8
		dc.b	$A4					; 
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.w	$C
aTedDonato:	dc.b	'Ted Donato'
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.w	$C
aJoeJuneau:	dc.b	'Joe Juneau'
		dc.b	$49					; I
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$41					; A
		dc.w	$14
aDmitriKvartalnov: dc.b	'Dmitri Kvartalnov',0
		dc.b	$10
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$23					; #
		dc.b	$20
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$41					; A
		dc.w	$C
aDaveReid:	dc.b	'Dave Reid',0
		dc.b	$17
		dc.b	$92					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$14
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$20
		dc.w	$14
aGregoriPantaleyev: dc.b 'Gregori Pantaleyev'
		dc.b	$13
		dc.b	$64					; d
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.w	$E
aBrentHughes:	dc.b	'Brent Hughes'
		dc.b	$42					; B
		dc.b	$63					; c
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$34					; 4
		dc.w	$C
aCamNeely:	dc.b	'Cam Neely',0
		dc.b	8
		dc.b	$A4					; 
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$49					; I
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	$44					; D
		dc.w	$10
aStephenLeach:	dc.b	'Stephen Leach',0
		dc.b	$27					; '
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$37					; 7
		dc.b	$32					; 2
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.w	$10
aStephenHeinze:	dc.b	'Stephen Heinze'
		dc.b	$23					; #
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.w	$C
aC_j_Young:	dc.b	'C.J. Young'
		dc.b	$18
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$22					; "
		dc.w	$E
aDarinKimble:	dc.b	'Darin Kimble'
		dc.b	$29					; )
		dc.b	$91					; 
		dc.b	$12
		dc.b	$11
		dc.b	$29					; )
		dc.b	$24					; $
		dc.b	$13
		dc.b	$15
		dc.w	$E
aPeterDouris:	dc.b	'Peter Douris'
		dc.b	$16
		dc.b	$83					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$25					; %
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$21					; !
		dc.w	$E
aRayBourque_0:	dc.b	'Ray Bourque',0
		dc.b	$77					; w
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$62					; b
		dc.b	$52					; R
		dc.w	$E
aDonSweeney:	dc.b	'Don Sweeney',0
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$E
aGlenWesley_0:	dc.b	'Glen Wesley',0
		dc.b	$26					; &
		dc.b	$85					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.b	$41					; A
		dc.b	$44					; D
		dc.b	$42					; B
		dc.w	$C
aDavidShaw:	dc.b	'David Shaw'
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$E
aGordMurphy:	dc.b	'Gord Murphy',0				; DATA XREF: ROM:off_BE26Ao
		dc.b	$28					; (
		dc.b	$85					; 
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$49					; I
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$10
aGordieRoberts:	dc.b	'Gordie Roberts'
		dc.b	$14
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.w	$12
aGlenFeathrston: dc.b	'Glen Feathrston',0
		dc.b	6
		dc.b	$B3					; 
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$48					; H
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$24					; $
		dc.w	$C
aJimWiemer:	dc.b	'Jim Wiemer'
		dc.b	$36					; 6
		dc.b	$A2					; 
		dc.b	$12
		dc.b	$21					; !
		dc.b	$28					; (
		dc.b	$20
		dc.b	$23					; #
		dc.b	$24					; $
		dc.w	2
		dc.w	8
aBoston:	dc.b	'Boston'
		dc.w	6
aBos:		dc.b	'BOS',0
		dc.w	8
aBruins:	dc.b	'Bruins'
		dc.w	$10
aBostonGarden:	dc.b	'Boston Garden',0
unk_C4C:	dc.b	0					; DATA XREF: ROM:00000316o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C0					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$6C					; l
		dc.b	0
		dc.b	$8E					; 
		dc.b	8
		dc.b	$20
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$6C					; l
		dc.b	0
		dc.b	$8E					; 
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	4
		dc.b	$20
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$10
		dc.b	$21					; !
		dc.b	$E8					; 
		dc.b	$B2					; 
		dc.b	$50					; P
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	5
		dc.b	4
		dc.b	$D
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	9
		dc.b	4
		dc.b	$D
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	$B
		dc.b	5
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	8
		dc.b	6
		dc.b	$E
		dc.b	4
unk_CBD:	dc.b	0					; DATA XREF: ROM:off_5DE84o
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	5
		dc.b	4
		dc.b	$D
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$16
		dc.b	8
		dc.b	6
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	9
		dc.b	6
		dc.b	8
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	8
		dc.b	4
		dc.b	9
		dc.b	6
		dc.b	0
		dc.w	$C
aGrantFuhr_0:	dc.b	'Grant Fuhr'
		dc.b	$31					; 1
		dc.b	$75					; u
		dc.b	$45					; E
		dc.b	$55					; U
		dc.b	1
		dc.b	0
		dc.b	$44					; D
		dc.b	$55					; U
		dc.w	$C
aTomDraper:	dc.b	'Tom Draper'
		dc.b	$35					; 5
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aDominikHasek:	dc.b	'Dominik Hasek',0
		dc.b	$39					; 9
		dc.b	$43					; C
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aPatLafontaine_0: dc.b	'Pat LaFontaine'
		dc.b	$16
		dc.b	$55					; U
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$54					; T
		dc.b	$60					; `
		dc.b	$52					; R
		dc.w	$10
aDaleHawerchuk:	dc.b	'Dale Hawerchuk'
		dc.b	$10
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$20
		dc.b	$51					; Q
		dc.b	$30					; 0
		dc.b	$52					; R
		dc.w	$E
aBobSweeney:	dc.b	'Bob Sweeney',0
		dc.b	$20
		dc.b	$94					; 
		dc.b	$23					; #
		dc.b	$53					; S
		dc.b	$47					; G
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aDaveHannan:	dc.b	'Dave Hannan',0
		dc.b	$14
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$20
		dc.b	$22					; "
		dc.w	$C
aRandyWood:	dc.b	'Randy Wood'
		dc.b	$19
		dc.b	$81					; 
		dc.b	$53					; S
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.w	$E
aYuriKhmylev:	dc.b	'Yuri Khmylev'
		dc.b	$13
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.w	$A
aBradMay:	dc.b	'Brad May'
		dc.b	$27					; '
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$3A					; :
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$14
		dc.w	$C
aBobErrey:	dc.b	'Bob Errey',0
		dc.b	$12
		dc.b	$64					; d
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$A
aRobRay:	dc.b	'Rob Ray',0
		dc.b	$32					; 2
		dc.b	$A3					; 
		dc.b	$41					; A
		dc.b	$11
		dc.b	$3A					; :
		dc.b	$22					; "
		dc.b	$15
		dc.b	$14
		dc.w	$12
aAlexnderMogilny_0: dc.b 'Alexnder Mogilny'
		dc.b	$89					; 
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$52					; R
		dc.w	$10
aWaynePresley:	dc.b	'Wayne Presley',0
		dc.b	$18
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$27					; '
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$10
aDonaldAudette:	dc.b	'Donald Audette'
		dc.b	$28					; (
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$29					; )
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$C
aBobCorkum:	dc.b	'Bob Corkum'
		dc.b	$29					; )
		dc.b	$A2					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.w	$12
aColinPatterson: dc.b	'Colin Patterson',0
		dc.b	$17
		dc.b	$84					; 
		dc.b	$42					; B
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$E
aDougBodger:	dc.b	'Doug Bodger',0
		dc.b	8
		dc.b	$A2					; 
		dc.b	$33					; 3
		dc.b	$42					; B
unk_E66:	dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$E
aPetrSvoboda:	dc.b	'Petr Svoboda'
		dc.b	7
		dc.b	$54					; T
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$36					; 6
		dc.b	$40					; @
		dc.b	$40					; @
		dc.b	$33					; 3
		dc.w	$12
aRichardSmehlik: dc.b	'Richard Smehlik',0
		dc.b	$42					; B
		dc.b	$A3					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$30					; 0
		dc.b	$32					; 2
		dc.w	$C
aKenSutton:	dc.b	'Ken Sutton'
		dc.b	$41					; A
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$14
		dc.b	$12
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$10
aGrantLedyard:	dc.b	'Grant Ledyard',0
		dc.b	3
		dc.b	$94					; 
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$3C					; <
		dc.b	$10
		dc.b	$23					; #
		dc.b	$23					; #
		dc.w	$10
aGordDonnelly:	dc.b	'Gord Donnelly',0			; DATA XREF: ROM:off_A4B54o
		dc.b	$34					; 4
		dc.b	$91					; 
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$3B					; ;
		dc.b	$11
		dc.b	$22					; "
		dc.b	$25					; %
		dc.w	$E
aRandyMoller:	dc.b	'Randy Moller'
		dc.b	$24					; $
		dc.b	$A3					; 
		dc.b	$22					; "
		dc.b	$12
		dc.b	$39					; 9
		dc.b	$21					; !
		dc.b	$11
		dc.b	$24					; $
		dc.w	$E
aKeithCarney:	dc.b	'Keith Carney'
		dc.b	6
		dc.b	$92					; 
		dc.b	$32					; 2
		dc.b	$12
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$14
		dc.b	$24					; $
		dc.w	2
		dc.w	$A
aBuffalo:	dc.b	'Buffalo',0
		dc.w	6
aBuf:		dc.b	'BUF',0
		dc.w	8
aSabres:	dc.b	'Sabres'
		dc.w	$16
aMemorialAuditorium: dc.b 'Memorial Auditorium',0
unk_F3A:	dc.b	0					; DATA XREF: ROM:0000031Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CE					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$8C					; 
		dc.b	0
		dc.b	$AE					; 
		dc.b	0
		dc.b	$A
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$8C					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	6
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$42					; B
		dc.b	0
		dc.b	0
		dc.b	$DA					; 
		dc.b	$C5					; 
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	$11
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$11
		dc.b	7
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	$16
		dc.b	8
		dc.b	4
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$17
		dc.b	$A
		dc.b	5
		dc.b	$F
unk_FAA:	dc.b	3					; DATA XREF: ROM:off_A8922o
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	$11
		dc.b	7
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$16
		dc.b	4
		dc.b	5
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	$16
		dc.b	8
		dc.b	$C
		dc.b	7
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$13
		dc.b	7
		dc.b	5
		dc.b	8
		dc.b	$C
		dc.b	0
		dc.w	$E
aMikeVernon:	dc.b	'Mike Vernon',0
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$C
aJeffReese:	dc.b	'Jeff Reese'
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.w	$10
aJoeNieuwendyk:	dc.b	'Joe Nieuwendyk'
		dc.b	$25					; %
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$22					; "
		dc.w	$10
aRobertReichel:	dc.b	'Robert Reichel'
		dc.b	$26					; &
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.w	$C
aJoelOtto:	dc.b	'Joel Otto',0
		dc.b	$29					; )
		dc.b	$B4					; 
		dc.b	$33					; 3
		dc.b	$52					; R
		dc.b	$57					; W
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$24					; $
		dc.w	$12
aBrianSkrudland: dc.b	'Brian Skrudland',0
		dc.b	$39					; 9
		dc.b	$84					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.w	$E
aGaryRoberts_0:	dc.b	'Gary Roberts'
		dc.b	$10
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$51					; Q
		dc.b	$34					; 4
		dc.w	$E
aPaulRanheim:	dc.b	'Paul Ranheim'
		dc.b	$28					; (
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$13
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.w	$E
aBrentAshton:	dc.b	'Brent Ashton'				; DATA XREF: ROM:off_A9A10o
		dc.b	$15
		dc.b	$A3					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$10
aChrisLindberg:	dc.b	'Chris Lindberg'
		dc.b	$11
		dc.b	$73					; s
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$21					; !
		dc.w	$E
aCraigBerube:	dc.b	'Craig Berube'
		dc.b	$16
		dc.b	$92					; 
		dc.b	$31					; 1
		dc.b	$12
		dc.b	$2A					; *
		dc.b	$11
		dc.b	$14
		dc.b	$14
		dc.w	$10
aTheorenFleury_0: dc.b	'Theoren Fleury'
		dc.b	$14
		dc.b	$35					; 5
		dc.b	$54					; T
		dc.b	$54					; T
		dc.b	$47					; G
		dc.b	$43					; C
		dc.b	$61					; a
		dc.b	$33					; 3
		dc.w	$10
aSergeiMakarov:	dc.b	'Sergei Makarov'
		dc.b	$42					; B
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$22					; "
		dc.b	$10
		dc.b	$54					; T
		dc.b	$30					; 0
		dc.b	$52					; R
		dc.w	$10
aGregPaslawski:	dc.b	'Greg Paslawski'
		dc.b	$23					; #
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$13
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$20
		dc.w	$E
aRonnieStern:	dc.b	'Ronnie Stern'
		dc.b	$22					; "
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$12
		dc.b	$37					; 7
		dc.b	$23					; #
		dc.b	$13
		dc.b	$24					; $
		dc.w	$C
aGarySuter_0:	dc.b	'Gary Suter'
		dc.b	$20
		dc.b	$75					; u
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$E
aAlMacinnis:	dc.b	'Al MacInnis',0
		dc.b	2
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$12
aRogerJohansson: dc.b	'Roger Johansson',0
		dc.b	$34					; 4
		dc.b	$74					; t
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.w	$E
aTrentYawney:	dc.b	'Trent Yawney'
		dc.b	$18
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.w	$E
aFrankMusil:	dc.b	'Frank Musil',0
		dc.b	3
		dc.b	$94					; 
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$E
aMichelPetit:	dc.b	'Michel Petit'
		dc.b	7
		dc.b	$93					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$39					; 9
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$C
aKevinDahl:	dc.b	'Kevin Dahl'
		dc.b	4
		dc.b	$72					; r
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$12
aChrisDahlquist: dc.b	'Chris Dahlquist',0
		dc.b	5
		dc.b	$83					; 
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$14
aAlexnderGodynyuk: dc.b	'Alexnder Godynyuk',0
		dc.b	$21					; !
		dc.b	$A2					; 
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$38					; 8
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$12
		dc.w	$C
aGregSmyth:	dc.b	'Greg Smyth'
		dc.b	6
		dc.b	$A0					; 
		dc.b	1
		dc.b	0
		dc.b	$19
		dc.b	1
		dc.b	$14
		dc.b	4
		dc.b	0
		dc.b	2
		dc.w	$A
aCalgary:	dc.b	'Calgary',0
		dc.w	6
aCgy:		dc.b	'CGY',0
		dc.w	8
aFlames:	dc.b	'Flames'
		dc.w	$14
aOlympicSaddledome: dc.b 'Olympic Saddledome'
unk_1234:	dc.b	0					; DATA XREF: ROM:0000031Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	8
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$2A					; *
		dc.b	0
		dc.b	4
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$20
		dc.b	$10
		dc.b	$21					; !
		dc.b	$F8					; 
		dc.b	$E4					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$18
		dc.b	$A
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	4
		dc.b	5
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	5
		dc.b	$C
		dc.b	3
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	3
		dc.b	$D
		dc.b	5
		dc.b	$C
		dc.b	0
		dc.w	$C
aEdBelfour_0:	dc.b	'Ed Belfour'
		dc.b	$30					; 0
		dc.b	$66					; f
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	0
		dc.b	0
		dc.b	$66					; f
		dc.b	$55					; U
		dc.w	$C
aJimWaite:	dc.b	'Jim Waite',0
		dc.b	$29					; )
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$32					; 2
off_12EA:	dc.l	unk_2233				; DATA XREF: ROM:off_5394Eo
		dc.w	$10
aJeremyRoenick_0: dc.b	'Jeremy Roenick'
		dc.b	$27					; '
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$53					; S
		dc.w	$12
aChristanRuuttu: dc.b	'Christan Ruuttu',0
		dc.b	$22					; "
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$E
aBrentSutter:	dc.b	'Brent Sutter'
		dc.b	$12
		dc.b	$63					; c
		dc.b	$24					; $
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$51					; Q
		dc.b	$33					; 3
		dc.w	$E
aTroyMurray:	dc.b	'Troy Murray',0
		dc.b	$19
		dc.b	$84					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.w	$10
aMichelGoulet:	dc.b	'Michel Goulet',0
		dc.b	$16
		dc.b	$84					; 
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.b	$12
		dc.w	$12
aStephaneMatteau: dc.b	'Stephane Matteau'
		dc.b	$32					; 2
		dc.b	$82					; 
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$36					; 6
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$13
		dc.w	$E
aGregGilbert:	dc.b	'Greg Gilbert'
		dc.b	$14
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$12
		dc.w	$12
aJocelynLemieux: dc.b	'Jocelyn Lemieux',0
		dc.b	$26					; &
		dc.b	$93					; 
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$13
		dc.w	$E
aStuGrimson:	dc.b	'Stu Grimson',0
		dc.b	$23					; #
		dc.b	$B1					; 
		dc.b	$10
		dc.b	$21					; !
		dc.b	$3A					; :
		dc.b	$11
		dc.b	$25					; %
		dc.b	$14
		dc.w	$E
aSteveLarmer:	dc.b	'Steve Larmer'
		dc.b	$28					; (
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$63					; c
		dc.b	$42					; B
		dc.w	$E
aDirkGraham:	dc.b	'Dirk Graham',0
		dc.b	$33					; 3
		dc.b	$84					; 
		dc.b	$33					; 3
		dc.b	$53					; S
		dc.b	$47					; G
		dc.b	$32					; 2
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.w	$E
aBrianNoonan:	dc.b	'Brian Noonan'
		dc.b	$10
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.w	$10
aDaveChristian:	dc.b	'Dave Christian'
		dc.b	$25					; %
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.w	$C
aJoeMurphy:	dc.b	'Joe Murphy'
		dc.b	$17
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$C
aRobBrown:	dc.b	'Rob Brown',0
		dc.b	$44					; D
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$28					; (
		dc.b	$41					; A
		dc.b	$20
		dc.b	$24					; $
		dc.w	$10
aChrisChelios_0: dc.b	'Chris Chelios',0
		dc.b	7
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$47					; G
		dc.b	$51					; Q
		dc.b	$62					; b
		dc.b	$45					; E
		dc.w	$E
aSteveSmith_0:	dc.b	'Steve Smith',0
		dc.b	5
		dc.b	$B4					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$44					; D
		dc.w	$12
aBryanMarchment: dc.b	'Bryan Marchment'
		dc.b	0
		dc.b	2
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$15
		dc.w	$12
aFrantsekKucera: dc.b	'Frantsek Kucera',0
		dc.b	6
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$13
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.w	$C
aCraigMuni:	dc.b	'Craig Muni'
		dc.b	3
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$20
		dc.b	$42					; B
		dc.b	$23					; #
		dc.w	$E
aKeithBrown:	dc.b	'Keith Brown',0
		dc.b	4
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$39					; 9
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.w	$E
aCamRussell:	dc.b	'Cam Russell',0
		dc.b	8
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$24					; $
		dc.w	$E
aAdamBennett:	dc.b	'Adam Bennett'
		dc.b	$47					; G
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$11
		dc.b	$25					; %
		dc.b	$20
		dc.b	$14
		dc.b	$12
		dc.b	0
		dc.b	2
		dc.w	$A
aChicago:	dc.b	'Chicago',0
		dc.w	6
aChi:		dc.b	'CHI',0
		dc.w	$C
aBlackhawks:	dc.b	'Blackhawks'
		dc.w	$12
aChicagoStadium: dc.b	'Chicago Stadium',0
unk_152E:	dc.b	0					; DATA XREF: ROM:00000326o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$E6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$C
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	6
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$21					; !
		dc.b	0
		dc.b	$F8					; 
		dc.b	$D4					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	4
		dc.b	3
		dc.b	$D
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$19
		dc.b	$B
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$17
		dc.b	8
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	9
		dc.b	5
		dc.b	$F
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	8
		dc.b	3
		dc.b	$D
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	$F
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	8
		dc.b	4
		dc.b	3
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$19
		dc.b	5
		dc.b	3
		dc.b	8
		dc.b	4
		dc.b	0
		dc.w	$10
aTimCheveldae_0: dc.b	'Tim Cheveldae',0
		dc.b	$32					; 2
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$43					; C
		dc.b	$44					; D
		dc.w	$12
aVincentRiendeau: dc.b	'Vincent Riendeau'
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.w	$10
aSteveYzerman_0: dc.b	'Steve Yzerman',0
		dc.b	$19
		dc.b	$66					; f
		dc.b	$56					; V
		dc.b	$44					; D
		dc.b	$11
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$52					; R
		dc.w	$10
aSergeiFedorov:	dc.b	'Sergei Fedorov'
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$54					; T
		dc.b	$41					; A
		dc.b	$53					; S
		dc.w	$E
aDallasDrake:	dc.b	'Dallas Drake'
		dc.b	$28					; (
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$10
aMikeSillinger:	dc.b	'Mike Sillinger'
		dc.b	$23					; #
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$40					; @
		dc.b	$21					; !
		dc.w	$12
aVachslavKozlov: dc.b	'Vachslav Kozlov',0
		dc.b	$13
		dc.b	$54					; T
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.w	$10
aPaulYsebaert:	dc.b	'Paul Ysebaert',0
		dc.b	$21					; !
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$26					; &
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.w	$C
aShawnBurr:	dc.b	'Shawn Burr'
		dc.b	$11
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.w	$10
aKeithPrimeau:	dc.b	'Keith Primeau',0
		dc.b	$55					; U
		dc.b	$B3					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$16
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.w	$10
aGerardGallant:	dc.b	'Gerard Gallant'
		dc.b	$17
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.w	$10
aJohnOgrodnick:	dc.b	'John Ogrodnick'
		dc.b	$25					; %
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$21					; !
		dc.b	$30					; 0
		dc.w	$12
aDinoCiccarelli: dc.b	'Dino Ciccarelli',0
		dc.b	$22					; "
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$25					; %
		dc.b	$23					; #
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$E
aRaySheppard:	dc.b	'Ray Sheppard'
		dc.b	$26					; &
		dc.b	$62					; b
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$E
aBobProbert:	dc.b	'Bob Probert',0
		dc.b	$24					; $
		dc.b	$B4					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$4A					; J
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$35					; 5
		dc.w	$12
aSheldonKennedy: dc.b	'Sheldon Kennedy',0
		dc.b	$15
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$C
aJimHiller:	dc.b	'Jim Hiller'
		dc.b	$14
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$12
		dc.b	$37					; 7
		dc.b	$22					; "
		dc.b	$14
		dc.b	$24					; $
		dc.w	$E
aPaulCoffey_0:	dc.b	'Paul Coffey',0
		dc.b	$77					; w
		dc.b	$96					; 
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$61					; a
		dc.b	$51					; Q
		dc.b	$53					; S
		dc.w	$10
aSteveChiasson_0: dc.b	'Steve Chiasson'
		dc.b	3
		dc.b	$94					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.w	$E
aYvesRacine:	dc.b	'Yves Racine',0
		dc.b	$33					; 3
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$12
aNicklasLidstrom: dc.b	'Nicklas Lidstrom'
		dc.b	5
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$41					; A
		dc.w	$C
aMarkHowe:	dc.b	'Mark Howe',0
		dc.b	4
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.w	$16
aVladimirKonstantov: dc.b 'Vladimir Konstantov',0
		dc.b	$16
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$10
aSteveKonroyd:	dc.b	'Steve Konroyd',0
		dc.b	8
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$10
aBradMccrimmon:	dc.b	'Brad McCrimmon'
		dc.b	2
		dc.b	$83					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$10
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	0
		dc.b	2
		dc.w	$A
aDetroit:	dc.b	'Detroit',0
		dc.w	6
aDet:		dc.b	'DET',0
		dc.w	$C
aRedWings:	dc.b	'Red Wings',0
		dc.w	$18
aJoeLouisSportsArena: dc.b 'Joe Louis Sports Arena'
unk_1848:	dc.b	0					; DATA XREF: ROM:off_32Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CE					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2C					; ,
		dc.b	6
		dc.b	0
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$2C					; ,
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$16
		dc.b	2
		dc.b	$12
		dc.b	$F8					; 
		dc.b	$D5					; 
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	9
		dc.b	3
		dc.b	$F
		dc.b	$A
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	$A
		dc.b	3
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$14
		dc.b	9
		dc.b	4
		dc.b	$10
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$17
		dc.b	$B
		dc.b	$C
		dc.b	5
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	$A
		dc.b	9
		dc.b	$F
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$14
		dc.b	$B
		dc.b	7
		dc.b	$10
		dc.b	$A
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	9
		dc.b	5
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	$C
		dc.b	3
		dc.b	9
		dc.b	5
		dc.b	0
		dc.w	$E
aBillRanford:	dc.b	'Bill Ranford'
		dc.b	$30					; 0
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$54					; T
		dc.w	$E
aRonTugnutt:	dc.b	'Ron Tugnutt',0
		dc.b	1
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	$12
		dc.w	$E
aDougWeight:	dc.b	'Doug Weight',0
		dc.b	$39					; 9
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$C
aToddElik:	dc.b	'Todd Elik',0
		dc.b	$34					; 4
		dc.b	$73					; s
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.w	$12
aCraigMactavish: dc.b	'Craig MacTavish',0
		dc.b	$14
		dc.b	$84					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$52					; R
		dc.b	$23					; #
		dc.w	$C
aKevinTodd:	dc.b	'Kevin Todd'
		dc.b	$15
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.w	$E
aShjonPodein:	dc.b	'Shjon Podein'
		dc.b	$26					; &
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$24					; $
		dc.b	$25					; %
		dc.b	$12
		dc.w	$E
aMikeHudson:	dc.b	'Mike Hudson',0
		dc.b	$20
		dc.b	$93					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$48					; H
		dc.b	$20
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$10
aShayneCorson:	dc.b	'Shayne Corson',0
		dc.b	9
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.w	$10
aCraigSimpson:	dc.b	'Craig Simpson',0
		dc.b	$18
		dc.b	$83					; 
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$E
aZdenoCiger:	dc.b	'Zdeno Ciger',0
		dc.b	8
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$30					; 0
		dc.w	$12
aKellyBuchberger: dc.b	'Kelly Buchberger'
		dc.b	$16
		dc.b	$A2					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$4A					; J
		dc.b	$13
		dc.b	$32					; 2
		dc.b	$13
		dc.w	$10
aMartinGelinas:	dc.b	'Martin Gelinas'
		dc.b	7
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$22					; "
		dc.w	$10
aLouieDebrusk:	dc.b	'Louie DeBrusk',0
		dc.b	$29					; )
		dc.b	$C1					; 
		dc.b	$12
		dc.b	$12
		dc.b	$3C					; <
		dc.b	$14
		dc.b	$15
		dc.b	$15
		dc.w	$C
aPetrKlima:	dc.b	'Petr Klima'
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$24					; $
		dc.b	$13
		dc.b	$55					; U
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.w	$10
aScottMellanby:	dc.b	'Scott Mellanby'
		dc.b	$27					; '
		dc.b	$91					; 
		dc.b	$13
		dc.b	$33					; 3
		dc.b	$47					; G
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	$E
aStevenRice:	dc.b	'Steven Rice',0
		dc.b	$12
		dc.b	$B2					; 
		dc.b	$22					; "
		dc.b	$12
		dc.b	$13
		dc.b	$21					; !
		dc.b	$12
		dc.b	$23					; #
		dc.w	$E
aDaveManson_0:	dc.b	'Dave Manson',0
		dc.b	$24					; $
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.w	$10
aIgorKravchuk:	dc.b	'Igor Kravchuk',0
		dc.b	$21					; !
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$32					; 2
word_1A8A:	dc.w	$10					; DATA XREF: ROM:off_BC05Co
aBrianBenning:	dc.b	'Brian Benning',0
		dc.b	$19
		dc.b	$82					; 
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.w	$E
aGeoffSmith:	dc.b	'Geoff Smith',0
		dc.b	$25					; %
		dc.b	$94					; 
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$14
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.w	$E
aBrianGlynn:	dc.b	'Brian Glynn',0
		dc.b	6
		dc.b	$B3					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$23					; #
		dc.w	$12
aLukeRichardson: dc.b	'Luke Richardson',0
		dc.b	$22					; "
		dc.b	$B3					; 
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$30					; 0
		dc.b	$24					; $
		dc.b	$34					; 4
		dc.w	$E
aChrisJoseph:	dc.b	'Chris Joseph'
		dc.b	2
		dc.b	$A3					; 
		dc.b	$33					; 3
		dc.b	$13
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.b	$12
		dc.b	$33					; 3
		dc.w	$E
aBradWerenka:	dc.b	'Brad Werenka'
		dc.b	$36					; 6
		dc.b	$93					; 
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$26					; &
		dc.b	$33					; 3
		dc.b	$25					; %
		dc.b	$23					; #
		dc.b	0
		dc.b	2
		dc.w	$A
aEdmonton:	dc.b	'Edmonton'
		dc.w	6
aEdm:		dc.b	'EDM',0
		dc.w	8
aOilers:	dc.b	'Oilers'
		dc.w	$16
aNorthlandsColiseum: dc.b 'Northlands Coliseum',0
unk_1B44:	dc.b	0					; DATA XREF: ROM:00000332o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$DC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	6
		dc.b	0
		dc.b	0
		dc.b	$42					; B
		dc.b	$A
		dc.b	0
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$40					; @
		dc.b	2
		dc.b	$84					; 
		dc.b	$C
		dc.b	$20
		dc.b	8
		dc.b	0
		dc.b	$C
		dc.b	$20
		dc.b	6
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$17
		dc.b	2
		dc.b	0
		dc.b	$F7					; 
		dc.b	$A3					; 
		dc.b	$60					; `
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$19
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$14
		dc.b	$B
		dc.b	5
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	$A
		dc.b	6
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$16
		dc.b	$A
		dc.b	5
		dc.b	6
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$14
		dc.b	$10
		dc.b	$F
		dc.b	9
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$19
		dc.b	9
		dc.b	4
		dc.b	$10
		dc.b	$F
		dc.b	0
		dc.w	$C
aSeanBurke:	dc.b	'Sean Burke'
		dc.b	1
		dc.b	$A4					; 
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$10
aMarioGosselin:	dc.b	'Mario Gosselin'
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	$22					; "
		dc.w	$12
aFrankPietrngelo: dc.b	'Frank Pietrngelo'
		dc.b	$40					; @
		dc.b	$63					; c
		dc.b	$31					; 1
		dc.b	$13
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aAndrewCassels:	dc.b	'Andrew Cassels'
		dc.b	$21					; !
		dc.b	$73					; s
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$20
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$C
aTerryYake:	dc.b	'Terry Yake'
		dc.b	$25					; %
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$12
aMikaelNylander: dc.b	'Mikael Nylander',0
		dc.b	$36					; 6
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$E
aRobertKron:	dc.b	'Robert Kron',0
		dc.b	$38					; 8
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.w	$14
aRobertPetrovicky: dc.b	'Robert Petrovicky',0
		dc.b	$39					; 9
		dc.b	$54					; T
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.w	$12
aGeoffSanderson: dc.b	'Geoff Sanderson',0
		dc.b	8
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.w	$10
aPatrickPoulin:	dc.b	'Patrick Poulin'
		dc.b	$24					; $
		dc.b	$A3					; 
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$10
aYvonCorriveau:	dc.b	'Yvon Corriveau'
		dc.b	$11
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$38					; 8
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$11
		dc.w	$E
aJimMckenzie:	dc.b	'Jim McKenzie'
		dc.b	$33					; 3
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$12
		dc.b	$24					; $
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$14
		dc.w	$12
aRandyCunnyworth: dc.b	'Randy Cunnyworth'
		dc.b	7
		dc.b	$62					; b
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.w	$E
aPatVerbeek:	dc.b	'Pat Verbeek',0
		dc.b	$16
		dc.b	$73					; s
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$44					; D
		dc.w	$10
aMarkJanssens:	dc.b	'Mark Janssens',0
		dc.b	$22					; "
		dc.b	$B3					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$24					; $
		dc.w	$E
aNickKypreos:	dc.b	'Nick Kypreos'
		dc.b	$20
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$36					; 6
		dc.b	$24					; $
		dc.b	$34					; 4
		dc.b	$15
		dc.w	$C
aMarkGreig:	dc.b	'Mark Greig'
		dc.b	$17
		dc.b	$71					; q
		dc.b	$12
		dc.b	$10
		dc.b	$15
		dc.b	$11
		dc.b	$10
		dc.b	$13
		dc.w	$E
aJamieLeach:	dc.b	'Jamie Leach',0
		dc.b	$34					; 4
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$12
		dc.b	$39					; 9
		dc.b	$23					; #
		dc.b	$15
		dc.b	$20
		dc.w	$12
aZarleyZalapski_0: dc.b	'Zarley Zalapski',0
		dc.b	3
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$51					; Q
		dc.b	$33					; 3
		dc.w	$10
aEricWeinrich:	dc.b	'Eric Weinrich',0
		dc.b	4
		dc.b	$A3					; 
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$C
aAdamBurt:	dc.b	'Adam Burt',0
		dc.b	6
		dc.b	$74					; t
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.w	$E
aDanKeczmer:	dc.b	'Dan Keczmer',0
		dc.b	$37					; 7
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$16
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$13
		dc.w	$C
aDougHouda:	dc.b	'Doug Houda'
		dc.b	$27					; '
		dc.b	$72					; r
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$29					; )
		dc.b	$21					; !
		dc.b	$24					; $
		dc.b	$14
		dc.w	$12
aRandyLadouceur: dc.b	'Randy Ladouceur',0
		dc.b	$29					; )
		dc.b	$B3					; 
		dc.b	$31					; 1
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$14
		dc.w	$10
aAllenPedersen:	dc.b	'Allen Pedersen'
		dc.b	$41					; A
		dc.b	$A2					; 
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	0
		dc.b	2
		dc.w	$A
aHartford:	dc.b	'Hartford'
		dc.w	6
aHfd:		dc.b	'HFD',0
		dc.w	$A
aWhalers:	dc.b	'Whalers',0
		dc.w	$18
aHartfordCivicCenter: dc.b 'Hartford Civic Center',0
unk_1E52:	dc.b	0					; DATA XREF: ROM:00000336o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$66					; f
		dc.b	$12
		dc.b	$11
		dc.b	$D9					; 
		dc.b	$96					; 
		dc.b	$20
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	8
		dc.b	4
		dc.b	$D
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$13
		dc.b	8
		dc.b	4
		dc.b	$D
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	9
		dc.b	5
		dc.b	$E
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	$A
		dc.b	6
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$11
		dc.b	8
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$12
		dc.b	9
		dc.b	5
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$12
		dc.b	9
		dc.b	$F
		dc.b	4
		dc.b	$E
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	4
		dc.b	$E
		dc.b	9
		dc.b	$F
		dc.b	0
		dc.w	$E
aKellyHrudey:	dc.b	'Kelly Hrudey'
		dc.b	$32					; 2
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$23					; #
		dc.w	$E
aRobbStauber:	dc.b	'Robb Stauber'
		dc.b	$35					; 5
		dc.b	$62					; b
		dc.b	$35					; 5
		dc.b	$53					; S
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aRickKnickle:	dc.b	'Rick Knickle'
		dc.b	1
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$21					; !
		dc.w	$10
aWayneGretzky_0: dc.b	'Wayne Gretzky',0
		dc.b	$99					; 
		dc.b	$46					; F
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$62					; b
		dc.b	$60					; `
		dc.b	$60					; `
		dc.w	$E
aJimmyCarson:	dc.b	'Jimmy Carson'
		dc.b	$12
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$41					; A
		dc.w	$E
aCoreyMillen:	dc.b	'Corey Millen'
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$15
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$E
aGaryShuchuk:	dc.b	'Gary Shuchuk'
		dc.b	$14
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	1
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	3
		dc.b	$22					; "
		dc.w	$10
aLucRobitaille_0: dc.b	'Luc Robitaille'
		dc.b	$20
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$E
aTonyGranato:	dc.b	'Tony Granato'
		dc.b	$21					; !
		dc.b	$64					; d
		dc.b	$54					; T
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.w	$E
aLonnieLoach:	dc.b	'Lonnie Loach'
		dc.b	$29					; )
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aPatConacher:	dc.b	'Pat Conacher'
		dc.b	$15
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.w	$10
aWarrenRychel:	dc.b	'Warren Rychel',0
		dc.b	$10
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$11
		dc.b	$36					; 6
		dc.b	$12
		dc.b	$15
		dc.b	$15
		dc.w	$12
aTomasSandstrom: dc.b	'Tomas Sandstrom',0
		dc.b	7
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$35					; 5
		dc.b	$38					; 8
		dc.b	$45					; E
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$C
aJariKurri_0:	dc.b	'Jari Kurri'
		dc.b	$17
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$10
aMikeDonnelly:	dc.b	'Mike Donnelly',0
		dc.b	$11
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.w	$E
aDaveTaylor:	dc.b	'Dave Taylor',0
		dc.b	$18
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$C
aRobBlake:	dc.b	'Rob Blake',0
		dc.b	4
		dc.b	$B4					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$43					; C
		dc.b	$44					; D
		dc.w	$10
aMartyMcsorley:	dc.b	'Marty McSorley'
		dc.b	$33					; 3
		dc.b	$E3					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$39					; 9
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$36					; 6
		dc.w	$10
aAlexeiZhitnik:	dc.b	'Alexei Zhitnik'
		dc.b	2
		dc.b	$54					; T
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$E
aDarrylSydor:	dc.b	'Darryl Sydor'
		dc.b	$25					; %
		dc.b	$93					; 
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$10
aCharlieHuddy:	dc.b	'Charlie Huddy',0
		dc.b	$22					; "
		dc.b	$A3					; 
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$C
aMarkHardy:	dc.b	'Mark Hardy'
		dc.b	$24					; $
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$36					; 6
		dc.b	$20
		dc.b	$21					; !
		dc.b	$23					; #
		dc.w	$10
aBrentThompson:	dc.b	'Brent Thompson'
		dc.b	3
		dc.b	$52					; R
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$28					; (
		dc.b	$20
		dc.b	$21					; !
		dc.b	$14
		dc.w	$E
aTimWatters:	dc.b	'Tim Watters',0
		dc.b	5
		dc.b	$62					; b
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$26					; &
		dc.b	$20
		dc.b	$21					; !
		dc.b	$22					; "
		dc.w	$12
aReneChapdlaine: dc.b	'Rene Chapdlaine',0
		dc.b	8
		dc.b	$81					; 
		dc.b	$10
		dc.b	$10
		dc.b	$15
		dc.b	$10
		dc.b	$13
		dc.b	$13
		dc.w	2
		dc.w	$E
aLosAngeles:	dc.b	'Los Angeles',0
		dc.w	4
aLa:		dc.b	'LA'
		dc.w	8
aKings:		dc.b	'Kings',0
		dc.w	$16
aGreatWesternForum: dc.b 'Great Western Forum',0
unk_214E:	dc.b	0					; DATA XREF: ROM:off_322o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CA					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$20
		dc.b	$A
		dc.b	$CA					; 
		dc.b	0
		dc.b	$60					; `
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	6
		dc.b	$66					; f
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$20
		dc.b	$A
		dc.b	$CA					; 
		dc.b	0
		dc.b	$60					; `
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$14
		dc.b	$10
		dc.b	0
		dc.b	$E9					; 
		dc.b	$B5					; 
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	4
		dc.b	3
		dc.b	$B
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$14
		dc.b	3
		dc.b	$C
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	8
		dc.b	4
		dc.b	$E
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$12
		dc.b	9
		dc.b	5
		dc.b	$B
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	$B
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$13
		dc.b	8
		dc.b	4
		dc.b	$D
		dc.b	$B
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$14
		dc.b	5
		dc.b	$B
		dc.b	8
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$15
		dc.b	8
		dc.b	9
		dc.b	5
		dc.b	$B
		dc.b	0
		dc.w	$C
aJonCasey:	dc.b	'Jon Casey',0
		dc.b	$30					; 0
		dc.b	$24					; $
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.w	$10
aDarcyWakaluk:	dc.b	'Darcy Wakaluk',0			; DATA XREF: ROM:off_8000Ao
		dc.b	$35					; 5
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.w	$E
aMikeModano_0:	dc.b	'Mike Modano',0
		dc.b	9
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$24					; $
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$43					; C
		dc.w	$E
aDaveGagner:	dc.b	'Dave Gagner',0
		dc.b	$15
		dc.b	$64					; d
		dc.b	$44					; D
unk_2233:	dc.b	$43					; C		; DATA XREF: ROM:off_12EAo
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$44					; D
		dc.w	$E
aNealBroten:	dc.b	'Neal Broten',0
		dc.b	7
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$53					; S
		dc.b	$30					; 0
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$41					; A
		dc.w	$12
aBrentGilchrist: dc.b	'Brent Gilchrist',0
		dc.b	$41					; A
		dc.b	$64					; d
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$20
		dc.b	$42					; B
		dc.b	$25					; %
		dc.b	$32					; 2
		dc.w	$E
aBobbySmith:	dc.b	'Bobby Smith',0
		dc.b	$18
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.w	$E
aMikeMcphee:	dc.b	'Mike McPhee',0
		dc.b	$17
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$22					; "
		dc.b	$44					; D
		dc.b	$22					; "
		dc.w	$12
aGaetanDuchesne: dc.b	'Gaetan Duchesne',0
		dc.b	$10
		dc.b	$93					; 
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$11
		dc.w	$E
aBrianPropp:	dc.b	'Brian Propp',0
		dc.b	$16
		dc.b	$83					; 
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$30					; 0
		dc.w	$10
aRussCourtnall:	dc.b	'Russ Courtnall'
		dc.b	$26					; &
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.w	$C
aUlfDahlen:	dc.b	'Ulf Dahlen'
		dc.b	$22					; "
		dc.b	$83					; 
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.w	$C
aMikeCraig:	dc.b	'Mike Craig'
		dc.b	$20
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$37					; 7
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.w	$E
aTrentKlatt:	dc.b	'Trent Klatt',0
		dc.b	$29					; )
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$12
		dc.b	$15
		dc.b	$21					; !
		dc.b	$11
		dc.b	$22					; "
		dc.w	$E
aShaneChurla:	dc.b	'Shane Churla'
		dc.b	$27					; '
		dc.b	$91					; 
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$25					; %
		dc.w	$10
aStewartGavin:	dc.b	'Stewart Gavin',0
		dc.b	$12
		dc.b	$73					; s
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$20
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$13
		dc.w	$E
aMarkTinordi:	dc.b	'Mark Tinordi'
		dc.b	$24					; $
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$24					; $
		dc.w	$E
aTommySjodin:	dc.b	'Tommy Sjodin'
		dc.b	$33					; 3
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$21					; !
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.w	$E
aJimJohnson:	dc.b	'Jim Johnson',0
		dc.b	6
		dc.b	$72					; r
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$10
aDerianHatcher:	dc.b	'Derian Hatcher'
		dc.b	2
		dc.b	$93					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$42					; B
		dc.b	$24					; $
		dc.w	$E
aCraigLudwig:	dc.b	'Craig Ludwig'
		dc.b	3
		dc.b	$C2					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.w	$14
aRichardMatvichuk: dc.b	'Richard Matvichuk',0
		dc.b	4
		dc.b	$72					; r
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$30					; 0
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.w	$E
aMarkOsiecki:	dc.b	'Mark Osiecki'
		dc.b	$23					; #
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$20
		dc.b	$13
		dc.b	$12
		dc.w	$C
aBradBerry:	dc.b	'Brad Berry'
		dc.b	5
		dc.b	$72					; r
		dc.b	$20
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$35					; 5
		dc.b	$14
		dc.w	$10
aEnricoCiccone:	dc.b	'Enrico Ciccone'
		dc.b	$39					; 9
		dc.b	$91					; 
		dc.b	$10
		dc.b	$22					; "
		dc.b	$28					; (
		dc.b	$20
		dc.b	$35					; 5
		dc.b	$25					; %
		dc.b	0
		dc.b	2
		dc.w	8
aDallas:	dc.b	'Dallas'
		dc.w	6
aDal:		dc.b	'DAL',0
		dc.w	8
aStars:		dc.b	'Stars',0
		dc.w	$10
aReunionArena:	dc.b	'Reunion Arena',0
unk_243E:	dc.b	0					; DATA XREF: ROM:0000033Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$D6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$22					; "
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$42					; B
		dc.b	0
		dc.b	8
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$22					; "
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	6
		dc.b	$22					; "
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$A
		dc.b	0
		dc.b	4
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$52					; R
		dc.b	$11
		dc.b	$21					; !
		dc.b	$E9					; 
		dc.b	$C5					; 
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$11
		dc.b	8
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	9
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	8
		dc.b	4
		dc.b	5
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$17
		dc.b	7
		dc.b	6
		dc.b	3
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$13
		dc.b	3
		dc.b	$D
		dc.b	7
		dc.b	6
		dc.b	0
		dc.w	$E
aPatrickRoy_0:	dc.b	'Patrick Roy',0
		dc.b	$33					; 3
		dc.b	$66					; f
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	0
		dc.b	0
		dc.b	$55					; U
		dc.b	$66					; f
		dc.w	$10
aAndreRacicot:	dc.b	'Andre Racicot',0
		dc.b	$37					; 7
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aKirkMuller_0:	dc.b	'Kirk Muller',0
		dc.b	$11
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$10
aStephanLebeau:	dc.b	'Stephan Lebeau'
		dc.b	$47					; G
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$21					; !
		dc.b	$45					; E
		dc.b	$40					; @
		dc.b	$31					; 1
		dc.w	$E
aDenisSavard:	dc.b	'Denis Savard'
		dc.b	$18
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$53					; S
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.w	$10
aGuyCarbonneau:	dc.b	'Guy Carbonneau'
		dc.b	$21					; !
		dc.b	$64					; d
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.w	$14
aVincentDamphousse: dc.b 'Vincent Damphousse'
		dc.b	$25					; %
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$53					; S
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$10
aGilbertDionne:	dc.b	'Gilbert Dionne'
		dc.b	$45					; E
		dc.b	$83					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$E
aJohnLeclair:	dc.b	'John Leclair'
		dc.b	$17
		dc.b	$B3					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$10
aBenoitBrunet:	dc.b	'Benoit Brunet',0
		dc.b	$22					; "
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aMarioRoberge:	dc.b	'Mario Roberge',0
		dc.b	$32					; 2
		dc.b	$61					; a
		dc.b	$11
		dc.b	$12
		dc.b	$38					; 8
		dc.b	$13
		dc.b	$13
		dc.b	$14
		dc.w	$10
aBrianBellows:	dc.b	'Brian Bellows',0
		dc.b	$23					; #
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$42					; B
		dc.w	$C
aMikeKeane:	dc.b	'Mike Keane'
		dc.b	$12
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.w	$E
aGaryLeeman:	dc.b	'Gary Leeman',0
		dc.b	$26					; &
		dc.b	$54					; T
		dc.b	$43					; C
		dc.b	$23					; #
		dc.b	$29					; )
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$C
aToddEwen:	dc.b	'Todd Ewen',0
		dc.b	$36					; 6
		dc.b	$B1					; 
		dc.b	$12
		dc.b	$11
		dc.b	$37					; 7
		dc.b	$12
		dc.b	$23					; #
		dc.b	$14
		dc.w	$A
aEdRonan:	dc.b	'Ed Ronan'
		dc.b	$31					; 1
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$14
		dc.b	$21					; !
		dc.w	$12
aEricDesjardins: dc.b	'Eric Desjardins',0
		dc.b	$28					; (
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$10
aMattSchneider:	dc.b	'Matt Schneider'
		dc.b	$27					; '
		dc.b	$73					; s
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$14
aPatriceBrisebois: dc.b	'Patrice Brisebois',0
		dc.b	$43					; C
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$37					; 7
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$23					; #
		dc.w	$E
aKevinHaller:	dc.b	'Kevin Haller'
		dc.b	$14
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$26					; &
		dc.b	$12
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$C
aRobRamage:	dc.b	'Rob Ramage'
		dc.b	5
		dc.b	$93					; 
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$34					; 4
		dc.w	$12
aJ_j_Daigneault: dc.b	'J.J. Daigneault',0
		dc.b	$48					; H
		dc.b	$64					; d
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$22					; "
		dc.w	$E
aLyleOdelein:	dc.b	'Lyle Odelein'
		dc.b	$24					; $
		dc.b	$93					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	$C
aSeanHill:	dc.b	'Sean Hill',0
		dc.b	$38					; 8
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$14
		dc.w	$12
aDonaldDufresne: dc.b	'Donald Dufresne',0
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$39					; 9
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$13
		dc.b	0
		dc.b	2
		dc.w	$A
aMontreal:	dc.b	'Montreal'
		dc.w	6
aMtl:		dc.b	'MTL',0
		dc.w	$C
aCanadiens:	dc.b	'Canadiens',0
		dc.w	$10
aMontrealForum:	dc.b	'Montreal Forum'
unk_2740:	dc.b	0					; DATA XREF: ROM:0000033Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$E6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$44					; D
		dc.b	0
		dc.b	8
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$11
		dc.b	$12
		dc.b	$F8					; 
		dc.b	$98					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	$C
		dc.b	3
		dc.b	$D
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	8
		dc.b	4
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	7
		dc.b	3
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$17
		dc.b	9
		dc.b	5
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	$C
		dc.b	5
		dc.b	$D
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$15
		dc.b	8
		dc.b	9
		dc.b	7
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$17
		dc.b	7
		dc.b	$11
		dc.b	8
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	8
		dc.b	3
		dc.b	7
		dc.b	$11
		dc.b	0
		dc.w	$10
aChrisTerreri:	dc.b	'Chris Terreri',0
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$12
aCraigBillington: dc.b	'Craig Billington'
		dc.b	1
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$21					; !
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aAlexnderSemak:	dc.b	'Alexnder Semak'
		dc.b	$20
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.w	$12
aBernieNicholls: dc.b	'Bernie Nicholls',0
		dc.b	$19
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$13
		dc.b	$42					; B
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.w	$10
aPeterStastny:	dc.b	'Peter Stastny',0
		dc.b	$26					; &
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.w	$E
aJanneOjanen:	dc.b	'Janne Ojanen'
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.w	$12
aValeriZelepukin: dc.b	'Valeri Zelepukin'
		dc.b	$25					; %
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$10
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.w	$E
aJohnMaclean:	dc.b	'John MacLean'
		dc.b	$15
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$23					; #
		dc.w	$E
aBobbyHolik:	dc.b	'Bobby Holik',0
		dc.b	$16
		dc.b	$A3					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.w	$E
aTomChorske:	dc.b	'Tom Chorske',0
		dc.b	9
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.w	$10
aTroyMallette:	dc.b	'Troy Mallette',0
		dc.b	8
		dc.b	$A2					; 
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$18
		dc.b	$24					; $
		dc.b	$13
		dc.b	$13
		dc.w	$12
aStephaneRicher: dc.b	'Stephane Richer',0
		dc.b	$44					; D
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$35					; 5
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.w	$10
aClaudeLemieux:	dc.b	'Claude Lemieux'
		dc.b	$22					; "
		dc.b	$B4					; 
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.w	$E
aBillGuerin:	dc.b	'Bill Guerin',0
		dc.b	$12
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$23					; #
		dc.w	$E
aRandyMckay:	dc.b	'Randy McKay',0
		dc.b	$21					; !
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.w	$10
aScottPellerin:	dc.b	'Scott Pellerin'
		dc.b	$18
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$23					; #
		dc.w	$C
aDaveBarr:	dc.b	'Dave Barr',0
		dc.b	$11
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$10
aScottStevens_0: dc.b	'Scott Stevens',0
		dc.b	4
		dc.b	$B4					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$56					; V
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$43					; C
		dc.w	$12
aVachslavFetisov: dc.b	'Vachslav Fetisov'
		dc.b	2
		dc.b	$B4					; 
		dc.b	$22					; "
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$30					; 0
		dc.b	$44					; D
		dc.w	$E
aBruceDriver:	dc.b	'Bruce Driver'
		dc.b	$23					; #
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.w	$12
aScottNiedrmayer: dc.b	'Scott Niedrmayer'
		dc.b	$27					; '
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$12
aAlexeiKasatonov: dc.b	'Alexei Kasatonov'
		dc.b	7
		dc.b	$B4					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.w	$E
aKenDaneyko:	dc.b	'Ken Daneyko',0
		dc.b	3
		dc.b	$A2					; 
		dc.b	$21					; !
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$20
		dc.b	$44					; D
		dc.b	$14
		dc.w	$10
aTommyAlbelin:	dc.b	'Tommy Albelin',0
		dc.b	6
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$10
aMylesOConnor:	dc.b	'Myles O',$27,'Connor'
		dc.b	5
		dc.b	$71					; q
		dc.b	$10
		dc.b	$11
		dc.b	$12
		dc.b	$10
		dc.b	$13
		dc.b	3
		dc.b	0
		dc.b	2
		dc.w	$C
aNewJersey:	dc.b	'New Jersey'
		dc.w	4
aNj:		dc.b	'NJ'
		dc.w	8
aDevils:	dc.b	'Devils'
		dc.w	$1A
aByrneMeadowlandsArena: dc.b 'Byrne Meadowlands Arena',0
unk_2A58:	dc.b	0					; DATA XREF: ROM:00000342o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$D6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	2
		dc.b	$C
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	2
		dc.b	$C
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$53					; S
		dc.b	$21					; !
		dc.b	0
		dc.b	$F8					; 
		dc.b	$97					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	9
		dc.b	3
		dc.b	4
		dc.b	$A
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$14
		dc.b	9
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	$A
		dc.b	4
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$16
		dc.b	$B
		dc.b	7
		dc.b	$F
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$16
		dc.b	9
		dc.b	3
		dc.b	$A
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$14
		dc.b	7
		dc.b	4
		dc.b	$C
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$15
		dc.b	$D
		dc.b	$C
		dc.b	4
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$19
		dc.b	4
		dc.b	$F
		dc.b	$C
		dc.b	$D
		dc.b	0
		dc.w	$E
aGlennHealy:	dc.b	'Glenn Healy',0
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$12
aMarkFitzpatrik: dc.b	'Mark Fitzpatrik',0			; DATA XREF: ROM:off_E6632o
		dc.b	$30					; 0
		dc.b	$72					; r
		dc.b	$42					; B
		dc.b	$23					; #
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aPierreTurgeon_0: dc.b	'Pierre Turgeon'
		dc.b	$77					; w
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$51					; Q
		dc.b	$41					; A
		dc.w	$E
aBenoitHogue:	dc.b	'Benoit Hogue'
		dc.b	$33					; 3
		dc.b	$74					; t
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$10
aMartyMcinnis:	dc.b	'Marty McInnis',0
		dc.b	$18
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.w	$E
aTravisGreen:	dc.b	'Travis Green'
		dc.b	$39					; 9
		dc.b	$82					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$22					; "
		dc.w	$E
aRayFerraro:	dc.b	'Ray Ferraro',0
		dc.b	$20
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.w	$12
aClaudeLoiselle: dc.b	'Claude Loiselle',0
		dc.b	$10
		dc.b	$83					; 
		dc.b	$22					; "
		dc.b	$11
		dc.b	$38					; 8
		dc.b	$33					; 3
		dc.b	$15
		dc.b	$24					; $
		dc.w	$E
aSteveThomas:	dc.b	'Steve Thomas'
		dc.b	$32					; 2
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$C
aDerekKing:	dc.b	'Derek King'
		dc.b	$27					; '
		dc.b	$93					; 
		dc.b	$24					; $
		dc.b	$34					; 4
		dc.b	$14
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$22					; "
		dc.w	$C
aDaveVolek:	dc.b	'Dave Volek'
		dc.b	$25					; %
		dc.b	$64					; d
		dc.b	$43					; C
		dc.b	$22					; "
		dc.b	$12
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$32					; 2
		dc.w	$12
aPatrickFlatley: dc.b	'Patrick Flatley',0
		dc.b	$26					; &
		dc.b	$83					; 
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$40					; @
		dc.b	$42					; B
		dc.w	$E
aBrianMullen:	dc.b	'Brian Mullen'
		dc.b	$16
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.w	$10
aBradDalgarno:	dc.b	'Brad Dalgarno',0
		dc.b	$15
		dc.b	$B1					; 
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$15
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$23					; #
		dc.w	$10
aTomFitzgerald:	dc.b	'Tom Fitzgerald'
		dc.b	$14
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$C
aDanMarois:	dc.b	'Dan Marois'
		dc.b	$17
		dc.b	$73					; s
		dc.b	$32					; 2
		dc.b	$13
		dc.b	$15
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.b	$23					; #
		dc.w	$E
aMickVukota:	dc.b	'Mick Vukota',0
		dc.b	$12
		dc.b	$82					; 
		dc.b	$31					; 1
		dc.b	$11
		dc.b	$19
		dc.b	$11
		dc.b	$24					; $
		dc.b	$14
		dc.w	$14
aVladimirMalakhov: dc.b	'Vladimir Malakhov',0
		dc.b	$23					; #
		dc.b	$A3					; 
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$14
aDariusKasparitis: dc.b	'Darius Kasparitis',0
		dc.b	$11
		dc.b	$74					; t
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.w	$E
aJeffNorton:	dc.b	'Jeff Norton',0
		dc.b	8
		dc.b	$83					; 
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$C
aUweKrupp:	dc.b	'Uwe Krupp',0
		dc.b	4
		dc.b	$E2					; 
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$E
aTomKurvers:	dc.b	'Tom Kurvers',0
		dc.b	$28					; (
		dc.b	$83					; 
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$10
aScottLachance:	dc.b	'Scott Lachance'
		dc.b	7
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$E
aDennisVaske:	dc.b	'Dennis Vaske'
		dc.b	$37					; 7
		dc.b	$A1					; 
		dc.b	$12
		dc.b	$12
		dc.b	$26					; &
		dc.b	$11
		dc.b	$11
		dc.b	$13
		dc.w	$10
aRichardPilon:	dc.b	'Richard Pilon',0
		dc.b	$47					; G
		dc.b	$93					; 
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.b	$38					; 8
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$15
		dc.w	2
		dc.w	$A
aNewYork:	dc.b	'New York'
		dc.w	6
aNyi:		dc.b	'NYI',0
		dc.w	$C
aIslanders:	dc.b	'Islanders',0
		dc.w	$12
aNassauColiseum: dc.b	'Nassau Coliseum',0
unk_2D5C:	dc.b	0					; DATA XREF: ROM:00000346o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	8
		dc.b	2
		dc.b	$A
		dc.b	6
		dc.b	0
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	2
		dc.b	$A
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$11
		dc.b	$11
		dc.b	$F8					; 
		dc.b	$96					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$C
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	8
		dc.b	3
		dc.b	$D
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	7
		dc.b	6
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$19
		dc.b	9
		dc.b	4
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	$D
		dc.b	3
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$17
		dc.b	8
		dc.b	5
		dc.b	$E
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	8
		dc.b	3
		dc.b	5
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$19
		dc.b	5
		dc.b	4
		dc.b	8
		dc.b	3
		dc.b	0
		dc.w	$12
aJohnVanbiesbrk: dc.b	'John Vanbiesbrk',0
		dc.b	$34					; 4
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	0
		dc.b	0
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.w	$E
aMikeRichter:	dc.b	'Mike Richter'
		dc.b	$35					; 5
		dc.b	$73					; s
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$E
aMarkMessier_0:	dc.b	'Mark Messier'
		dc.b	$11
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$54					; T
		dc.b	$53					; S
		dc.b	$40					; @
		dc.b	$53					; S
		dc.w	$12
aSergeiNemchinov: dc.b	'Sergei Nemchinov'
		dc.b	$13
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$53					; S
		dc.b	$40					; @
		dc.b	$34					; 4
		dc.b	$52					; R
		dc.b	$32					; 2
		dc.w	$12
aDarrenTurcotte: dc.b	'Darren Turcotte',0
		dc.b	8
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.w	$C
aEdOlczyk:	dc.b	'Ed Olczyk',0
		dc.b	$12
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.w	$E
aEsaTikkanen:	dc.b	'Esa Tikkanen'
		dc.b	$10
		dc.b	$95					; 
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$43					; C
		dc.w	$E
aAdamGraves:	dc.b	'Adam Graves',0
		dc.b	9
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.w	$E
aPhilBourque:	dc.b	'Phil Bourque'
		dc.b	$29					; )
		dc.b	$82					; 
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.b	$48					; H
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$C
aJanErixon:	dc.b	'Jan Erixon'
		dc.b	$20
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.w	$E
aStevenKing:	dc.b	'Steven King',0
		dc.b	$25					; %
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$13
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.w	$E
aMikeGartner_0:	dc.b	'Mike Gartner'
		dc.b	$22					; "
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$52					; R
		dc.w	$E
aTonyAmonte:	dc.b	'Tony Amonte',0
		dc.b	$33					; 3
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$13
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.w	$10
aAlexeiKovalev:	dc.b	'Alexei Kovalev'
		dc.b	$27					; '
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$10
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.w	$E
aPaulBroten:	dc.b	'Paul Broten',0
		dc.b	$37					; 7
		dc.b	$71					; q
		dc.b	$12
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.b	$12
		dc.b	$23					; #
		dc.b	$12
		dc.w	$C
aJoeyKocur:	dc.b	'Joey Kocur'
		dc.b	$26					; &
		dc.b	$82					; 
		dc.b	$11
		dc.b	$32					; 2
		dc.b	$39					; 9
		dc.b	$21					; !
		dc.b	$24					; $
		dc.b	$24					; $
		dc.w	$E
aMikeHartman:	dc.b	'Mike Hartman'
		dc.b	$18
		dc.b	$72					; r
		dc.b	$31					; 1
		dc.b	1
		dc.b	$36					; 6
		dc.b	$11
		dc.b	$15
		dc.b	$14
		dc.w	$E
aBrianLeetch_0:	dc.b	'Brian Leetch'
		dc.b	2
		dc.b	$66					; f
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$26					; &
		dc.b	$51					; Q
		dc.b	$52					; R
		dc.b	$52					; R
		dc.w	$10
aJamesPatrick:	dc.b	'James Patrick',0
		dc.b	3
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$E
aSergeiZubov:	dc.b	'Sergei Zubov'
		dc.b	$21					; !
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$30					; 0
		dc.w	$10
aJeffBeukeboom:	dc.b	'Jeff Beukeboom'
		dc.b	$23					; #
		dc.b	$B2					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$37					; 7
		dc.b	$20
		dc.b	$30					; 0
		dc.b	$24					; $
		dc.w	$C
aKevinLowe:	dc.b	'Kevin Lowe'
		dc.b	4
		dc.b	$84					; 
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.w	$12
aPeterAndersson: dc.b	'Peter Andersson',0
		dc.b	5
		dc.b	$92					; 
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.w	$C
aJayWells:	dc.b	'Jay Wells',0
		dc.b	$24					; $
		dc.b	$A2					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$20
		dc.b	$21					; !
		dc.b	$24					; $
		dc.w	$E
aJoeCirella:	dc.b	'Joe Cirella',0
		dc.b	6
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	2
		dc.w	$A
aNewYork_0:	dc.b	'New York'
		dc.w	6
aNyr:		dc.b	'NYR',0
		dc.w	$A
aRangers:	dc.b	'Rangers',0
		dc.w	$18
aMadisonSquareGarden_1: dc.b 'Madison Square Garden',0
unk_3054:	dc.b	0					; DATA XREF: ROM:off_34Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$2E					; .
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$44					; D
		dc.b	0
		dc.b	2
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	7
		dc.b	2
		dc.b	$13
		dc.b	$F8					; 
		dc.b	$C5					; 
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	$A
		dc.b	3
		dc.b	$F
		dc.b	$B
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	$A
		dc.b	3
		dc.b	$F
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$19
		dc.b	$C
		dc.b	7
		dc.b	$11
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	$B
		dc.b	4
		dc.b	$10
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	$F
		dc.b	3
		dc.b	$10
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$16
		dc.b	$A
		dc.b	5
		dc.b	8
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	6
		dc.b	7
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$19
		dc.b	$C
		dc.b	4
		dc.b	7
		dc.b	6
		dc.b	0
		dc.w	$12
aPeterSidorkwicz1: dc.b	'Peter Sidorkwicz1'
		dc.b	$62					; b
		dc.b	$31					; 1
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$14
aDanielBerthiaume: dc.b	'Daniel Berthiaume',0
		dc.b	$32					; 2
		dc.b	$13
		dc.b	$31					; 1
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aJamieBaker:	dc.b	'Jamie Baker',0
		dc.b	$13
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.w	$C
aMarkLamb:	dc.b	'Mark Lamb',0
		dc.b	7
		dc.b	$64					; d
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$C
aNeilBrady:	dc.b	'Neil Brady'
		dc.b	$12
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.w	$C
aMarkFreer:	dc.b	'Mark Freer'
		dc.b	$11
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$22					; "
		dc.w	$12
aLaurieBoschman: dc.b	'Laurie Boschman',0
		dc.b	$16
		dc.b	$64					; d
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$12
aDavidArchibald: dc.b	'David Archibald',0
		dc.b	$15
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$16
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.w	$C
aRobMurphy:	dc.b	'Rob Murphy'
		dc.b	$18
		dc.b	$91					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$24					; $
		dc.b	$12
		dc.w	$12
aSylvainTurgeon: dc.b	'Sylvain Turgeon',0
		dc.b	$61					; a
		dc.b	$94					; 
unk_31CA:	dc.b	$43					; C		; DATA XREF: ROM:off_AFE12o
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$E
aMikePeluso:	dc.b	'Mike Peluso',0
		dc.b	$44					; D
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$3A					; :
		dc.b	$13
		dc.b	$35					; 5
		dc.b	$25					; %
		dc.w	$C
aDougSmail:	dc.b	'Doug Smail'
		dc.b	9
		dc.b	$53					; S
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$21					; !
		dc.b	$24					; $
		dc.b	$23					; #
		dc.w	$E
aJeffLazaro:	dc.b	'Jeff Lazaro',0
		dc.b	$28					; (
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$12
		dc.b	$28					; (
		dc.b	$33					; 3
		dc.b	$15
		dc.b	$22					; "
		dc.w	$E
aDarcyLoewen:	dc.b	'Darcy Loewen'
		dc.b	$10
		dc.b	$62					; b
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$16
		dc.b	$12
		dc.b	$24					; $
		dc.b	$24					; $
		dc.w	$E
aBobKudelski:	dc.b	'Bob Kudelski'
		dc.b	$26					; &
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.w	$C
aJodyHull:	dc.b	'Jody Hull',0
		dc.b	$17
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$21					; !
		dc.w	$10
aAndrewMcbain:	dc.b	'Andrew McBain',0
		dc.b	$20
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$22					; "
		dc.w	$10
aTomasJelinek:	dc.b	'Tomas Jelinek',0
		dc.b	$25					; %
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$16
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$13
		dc.w	$E
aNormMaciver:	dc.b	'Norm Maciver'
		dc.b	$22					; "
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.w	$C
aBradShaw:	dc.b	'Brad Shaw',0
		dc.b	4
		dc.b	$73					; s
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$10
aDarrenRumble:	dc.b	'Darren Rumble',0
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.w	$E
aChrisLuongo:	dc.b	'Chris Luongo'
		dc.b	$23					; #
		dc.b	$62					; b
		dc.b	$21					; !
		dc.b	$31					; 1
		dc.b	$13
		dc.b	$20
		dc.b	$24					; $
		dc.b	$23					; #
		dc.w	$E
aKenHammond:	dc.b	'Ken Hammond',0
		dc.b	5
		dc.b	$73					; s
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$13
		dc.w	$E
aGordDineen:	dc.b	'Gord Dineen',0
		dc.b	6
		dc.b	$81					; 
		dc.b	$12
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$11
		dc.b	$25					; %
		dc.b	$13
		dc.w	$C
aBradMarsh:	dc.b	'Brad Marsh'
		dc.b	$14
		dc.b	$B2					; 
		dc.b	$20
		dc.b	$31					; 1
		dc.b	$36					; 6
		dc.b	$10
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	2
		dc.w	8
aOttawa:	dc.b	'Ottawa'
		dc.w	6
aOtw:		dc.b	'OTW',0
		dc.w	$A
aSenators:	dc.b	'Senators'
		dc.w	$16
aOttawaCivicCenter: dc.b 'Ottawa Civic Center',0
unk_3348:	dc.b	0					; DATA XREF: ROM:0000034Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$D8					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$2E					; .
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$28					; (
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$36					; 6
		dc.b	2
		dc.b	$22					; "
		dc.b	$E8					; 
		dc.b	$86					; 
		dc.b	$30					; 0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	5
		dc.b	4
		dc.b	$F
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	$A
		dc.b	4
		dc.b	$F
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$17
		dc.b	$C
		dc.b	5
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	$B
		dc.b	6
		dc.b	$11
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	5
		dc.b	4
		dc.b	$F
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$13
		dc.b	$A
		dc.b	6
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$17
		dc.b	8
		dc.b	$F
		dc.b	$A
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$13
		dc.b	$A
		dc.b	5
		dc.b	8
		dc.b	$F
		dc.b	0
		dc.w	$12
aTommySoderstrom: dc.b	'Tommy Soderstrom'
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	0
		dc.b	0
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.w	$12
aDominicRoussel: dc.b	'Dominic Roussel',0
		dc.b	$33					; 3
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$12
aStephBeauregard: dc.b	'Steph Beauregard'
		dc.b	$35					; 5
		dc.b	$72					; r
		dc.b	$30					; 0
		dc.b	1
		dc.b	1
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aEricLindros:	dc.b	'Eric Lindros'
		dc.b	$88					; 
		dc.b	$C4					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$44					; D
		dc.w	$10
aRodBrindamour:	dc.b	'Rod BrindAmour'
		dc.b	$17
		dc.b	$94					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$43					; C
		dc.w	$E
aPelleEklund:	dc.b	'Pelle Eklund'
		dc.b	9
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$41					; A
		dc.w	$10
aJosefBeranek:	dc.b	'Josef Beranek',0
		dc.b	$42					; B
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$10
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.w	$E
aKeithActon:	dc.b	'Keith Acton',0
		dc.b	$25					; %
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.w	$14
aVachslavButsayev: dc.b	'Vachslav Butsayev',0
		dc.b	$22					; "
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$20
		dc.b	$21					; !
		dc.b	$23					; #
		dc.w	$E
aBrentFedyk:	dc.b	'Brent Fedyk',0
		dc.b	$18
		dc.b	$82					; 
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$C
aDougEvans:	dc.b	'Doug Evans'
		dc.b	$15
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$23					; #
		dc.w	$10
aAndreiLomakin:	dc.b	'Andrei Lomakin'
		dc.b	$23					; #
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.w	$10
aDaveSnuggerud:	dc.b	'Dave Snuggerud'
		dc.b	$14
		dc.b	$73					; s
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.w	$10
aClaudeBoivin:	dc.b	'Claude Boivin',0
		dc.b	$10
		dc.b	$91					; 
		dc.b	$12
		dc.b	$12
		dc.b	$28					; (
		dc.b	$14
		dc.b	$12
		dc.b	$14
		dc.w	$E
aMarkRecchi_0:	dc.b	'Mark Recchi',0
		dc.b	8
		dc.b	$65					; e
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$E
aKevinDineen:	dc.b	'Kevin Dineen'
		dc.b	$11
		dc.b	$74					; t
		dc.b	$53					; S
		dc.b	$42					; B
		dc.b	$27					; '
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.w	$C
aDaveBrown:	dc.b	'Dave Brown'
		dc.b	$21					; !
		dc.b	$91					; 
		dc.b	$10
		dc.b	$21					; !
		dc.b	$29					; )
		dc.b	$10
		dc.b	$25					; %
		dc.b	$13
		dc.w	$E
aGarryGalley:	dc.b	'Garry Galley'
		dc.b	3
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$14
aDimitriYushkevich: dc.b 'Dimitri Yushkevich'
		dc.b	2
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.w	$E
aGregHawgood:	dc.b	'Greg Hawgood'
		dc.b	$20
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.w	$10
aTerryCarkner:	dc.b	'Terry Carkner',0
		dc.b	$29					; )
		dc.b	$A2					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$30					; 0
		dc.b	$24					; $
		dc.w	$E
aRicNattress:	dc.b	'Ric Nattress'
		dc.b	5
		dc.b	$A2					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$E
aRyanMcgill:	dc.b	'Ryan McGill',0
		dc.b	$27					; '
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$27					; '
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$14
		dc.w	$C
aGordHynes:	dc.b	'Gord Hynes'
		dc.b	$26					; &
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$E
aShawnCronin:	dc.b	'Shawn Cronin'
		dc.b	$44					; D
		dc.b	$A1					; 
		dc.b	$11
		dc.b	$11
		dc.b	$2C					; ,
		dc.b	$13
		dc.b	$15
		dc.b	$13
		dc.w	2
		dc.w	$E
aPhiladelphia:	dc.b	'Philadelphia'
		dc.w	6
aPhi:		dc.b	'PHI',0
		dc.w	8
aFlyers:	dc.b	'Flyers'
		dc.w	$A
aSpectrum:	dc.b	'Spectrum'
unk_3646:	dc.b	0					; DATA XREF: ROM:00000352o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$AE					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$AE					; 
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$20
		dc.b	$21					; !
		dc.b	$E9					; 
		dc.b	$C5					; 
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$B
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$B
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$13
		dc.b	9
		dc.b	4
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$14
		dc.b	8
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$14
		dc.b	7
		dc.b	3
		dc.b	$B
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$12
		dc.b	4
		dc.b	$D
		dc.b	$C
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	$C
		dc.b	3
		dc.b	4
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$17
		dc.b	$16
		dc.b	4
		dc.b	$D
		dc.b	3
		dc.b	$C
		dc.b	0
		dc.w	$E
aTomBarrasso_0:	dc.b	'Tom Barrasso'
		dc.b	$35					; 5
		dc.b	$A4					; 
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	1
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$E
aKenWregget:	dc.b	'Ken Wregget',0
		dc.b	$31					; 1
		dc.b	$82					; 
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	0
		dc.b	0
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$10
aMarioLemieux_0: dc.b	'Mario Lemieux',0
		dc.b	$66					; f
		dc.b	$A5					; 
		dc.b	$46					; F
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$66					; f
		dc.b	$60					; `
		dc.b	$62					; b
		dc.w	$E
aRonFrancis:	dc.b	'Ron Francis',0
		dc.b	$10
		dc.b	$94					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$42					; B
		dc.w	$12
aShawnMceachern: dc.b	'Shawn McEachern',0
		dc.b	$15
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.w	$10
aMikeStapleton:	dc.b	'Mike Stapleton'
		dc.b	$26					; &
		dc.b	$63					; c
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$30					; 0
		dc.w	$10
aKevinStevens_0: dc.b	'Kevin Stevens',0
		dc.b	$25					; %
		dc.b	$B3					; 
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$44					; D
		dc.w	$E
aDaveTippett:	dc.b	'Dave Tippett'
		dc.b	$14
		dc.b	$62					; b
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$C
aTroyLoney:	dc.b	'Troy Loney'
		dc.b	$24					; $
		dc.b	$A2					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$13
		dc.w	$E
aJeffDaniels:	dc.b	'Jeff Daniels'
		dc.b	$20
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$21					; !
		dc.w	$E
aJaromirJagr:	dc.b	'Jaromir Jagr'
		dc.b	$68					; h
		dc.b	$A5					; 
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$E
aRickTocchet_0:	dc.b	'Rick Tocchet'
		dc.b	$22					; "
		dc.b	$92					; 
		dc.b	$25					; %
		dc.b	$44					; D
		dc.b	$37					; 7
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.w	$C
aJoeMullen:	dc.b	'Joe Mullen'
		dc.b	7
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$30					; 0
		dc.w	$10
aMartinStraka:	dc.b	'Martin Straka',0
		dc.b	$82					; 
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$20
		dc.b	$32					; 2
		dc.w	$E
aMikeNeedham:	dc.b	'Mike Needham'
		dc.b	$39					; 9
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$21					; !
		dc.w	$E
aJayCaufield:	dc.b	'Jay Caufield'
		dc.b	$16
		dc.b	$E1					; 
		dc.b	$20
		dc.b	$11
		dc.b	$3B					; ;
		dc.b	$10
		dc.b	$13
		dc.b	$14
		dc.w	$E
aLarryMurphy:	dc.b	'Larry Murphy'
		dc.b	$55					; U
		dc.b	$A4					; 
		dc.b	$34					; 4
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$51					; Q
		dc.b	$43					; C
		dc.w	$10
aUlfSamuelsson:	dc.b	'Ulf Samuelsson'
		dc.b	5
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$30					; 0
		dc.b	$41					; A
		dc.b	$35					; 5
		dc.w	$A
aJimPaek:	dc.b	'Jim Paek'
		dc.b	2
		dc.b	$82					; 
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$E
aPaulStanton:	dc.b	'Paul Stanton'				; DATA XREF: ROM:off_B5180o
		dc.b	$23					; #
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$43					; C
		dc.w	$12
aPeterTaglianeti: dc.b	'Peter Taglianeti'
		dc.b	$32					; 2
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.b	$24					; $
		dc.w	$E
aMikeRamsey:	dc.b	'Mike Ramsey',0
		dc.b	6
		dc.b	$83					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$12
aKjellSamuelsson: dc.b	'Kjell Samuelsson'			; DATA XREF: ROM:off_5605Ao
		dc.b	$28					; (
		dc.b	$E2					; 
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$10
aGrantJennings:	dc.b	'Grant Jennings'
		dc.b	3
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$38					; 8
		dc.b	$20
		dc.b	$23					; #
		dc.b	$13
		dc.w	$10
aBryanFogarty:	dc.b	'Bryan Fogarty',0
		dc.b	$33					; 3
		dc.b	$81					; 
		dc.b	$12
		dc.b	$11
		dc.b	$22					; "
		dc.b	$20
		dc.b	$10
		dc.b	$11
		dc.w	2
		dc.w	$C
aPittsburgh:	dc.b	'Pittsburgh'
		dc.w	6
aPit:		dc.b	'PIT',0
		dc.w	$A
aPenguins:	dc.b	'Penguins'
		dc.w	$E
aCivicArena:	dc.b	'Civic Arena',0
unk_393C:	dc.b	0					; DATA XREF: ROM:00000356o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$D8					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	6
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$21					; !
		dc.b	0
		dc.b	$F8					; 
		dc.b	$A6					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	3
		dc.b	$E
		dc.b	$F
		dc.b	$11
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$E
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	8
		dc.b	4
		dc.b	$F
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$17
		dc.b	6
		dc.b	5
		dc.b	$10
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	4
		dc.b	3
		dc.b	$F
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	7
		dc.b	$E
		dc.b	$10
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	3
		dc.b	$11
		dc.b	5
		dc.b	$E
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	5
		dc.b	$E
		dc.b	3
		dc.b	$11
		dc.b	0
		dc.w	$E
aRonHextall:	dc.b	'Ron Hextall',0
		dc.b	$27					; '
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$10
aStephaneFiset:	dc.b	'Stephane Fiset'
		dc.b	$35					; 5
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$C
aJoeSakic_0:	dc.b	'Joe Sakic',0
		dc.b	$19
		dc.b	$64					; d
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$20
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$52					; R
		dc.w	$C
aMikeRicci:	dc.b	'Mike Ricci'
		dc.b	9
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$35					; 5
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.w	$12
aClaudeLapointe: dc.b	'Claude Lapointe',0
		dc.b	$47					; G
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.w	$12
aMartinRucinsky: dc.b	'Martin Rucinsky',0
		dc.b	$25					; %
		dc.b	$53					; S
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$14
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$12
aValeriKamensky: dc.b	'Valeri Kamensky',0
		dc.b	$31					; 1
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$35					; 5
		dc.b	$25					; %
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$C
aMikeHough:	dc.b	'Mike Hough'
		dc.b	$18
		dc.b	$73					; s
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$10
aGinoCavallini:	dc.b	'Gino Cavallini'
		dc.b	$44					; D
		dc.b	$B1					; 
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aScottPearson:	dc.b	'Scott Pearson',0
		dc.b	$22					; "
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$24					; $
		dc.b	$35					; 5
		dc.b	$24					; $
		dc.w	$E
aBillLindsay:	dc.b	'Bill Lindsay'
		dc.b	$20
		dc.b	$61					; a
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$21					; !
		dc.w	$E
aChrisSimon:	dc.b	'Chris Simon',0
		dc.b	$12
		dc.b	$D1					; 
		dc.b	$11
		dc.b	1
		dc.b	$28					; (
		dc.b	$11
		dc.b	5
		dc.b	$15
		dc.w	$C
aTonyTwist:	dc.b	'Tony Twist'
		dc.b	$15
		dc.b	$A1					; 
		dc.b	$11
		dc.b	1
		dc.b	$1C
		dc.b	$10
		dc.b	$14
		dc.b	$14
		dc.w	$E
aMatsSundin:	dc.b	'Mats Sundin',0
		dc.b	$13
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$13
		dc.b	$45					; E
		dc.b	$40					; @
		dc.b	$43					; C
		dc.w	$C
aOwenNolan:	dc.b	'Owen Nolan'
		dc.b	$11
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$17
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.w	$12
aAndreiKovalenko: dc.b	'Andrei Kovalenko'
		dc.b	$51					; Q
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$E
aScottYoung:	dc.b	'Scott Young',0
		dc.b	$48					; H
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$21					; !
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.w	$10
aSteveDuchesne_0: dc.b	'Steve Duchesne'
		dc.b	$28					; (
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$24					; $
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$14
aCurtisLeschyshyn: dc.b	'Curtis Leschyshyn',0
		dc.b	7
		dc.b	$93					; 
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$10
aAlexeiGusarov:	dc.b	'Alexei Gusarov'
		dc.b	5
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$10
aKerryHuffman:	dc.b	'Kerry Huffman',0
		dc.b	2
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.w	$C
aAdamFoote:	dc.b	'Adam Foote'
		dc.b	$52					; R
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.w	$E
aStevenFinn:	dc.b	'Steven Finn',0
		dc.b	$29					; )
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.w	$14
aMikhailTatarinov: dc.b	'Mikhail Tatarinov',0
		dc.b	4
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$24					; $
		dc.b	$43					; C
		dc.w	$10
aCraigWolanin:	dc.b	'Craig Wolanin',0
		dc.b	6
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$28					; (
		dc.b	$11
		dc.b	$21					; !
		dc.b	$14
		dc.w	2
		dc.w	8
aQuebec:	dc.b	'Quebec'
		dc.w	6
aQue:		dc.b	'QUE',0
		dc.w	$C
aNordiques:	dc.b	'Nordiques',0
		dc.w	$14
aColiseeDeQuebec: dc.b	'Colisee de Quebec',0
unk_3C42:	dc.b	0					; DATA XREF: ROM:0000035Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	8
		dc.b	$82					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	8
		dc.b	$88					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	2
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	8
		dc.b	$82					; 
		dc.b	6
		dc.b	$60					; `
		dc.b	8
		dc.b	$82					; 
		dc.b	4
		dc.b	$40					; @
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	7
		dc.b	2
		dc.b	$12
		dc.b	$D9					; 
		dc.b	$66					; f
		dc.b	$30					; 0
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	$A
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	$A
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	$C
		dc.b	5
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$19
		dc.b	$B
		dc.b	6
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$13
		dc.b	$A
		dc.b	4
		dc.b	$E
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$15
		dc.b	$B
		dc.b	5
		dc.b	6
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$18
		dc.b	4
		dc.b	$E
		dc.b	5
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$17
		dc.b	7
		dc.b	5
		dc.b	4
		dc.b	$E
		dc.b	0
		dc.w	$E
aArtursIrbe:	dc.b	'Arturs Irbe',0
		dc.b	$32					; 2
		dc.b	$52					; R
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$E
aJeffHackett:	dc.b	'Jeff Hackett'
		dc.b	$30					; 0
		dc.b	$52					; R
		dc.b	$31					; 1
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aBrianHayward:	dc.b	'Brian Hayward',0
		dc.b	1
		dc.b	$62					; b
		dc.b	$30					; 0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aKellyKisio:	dc.b	'Kelly Kisio',0
		dc.b	$11
		dc.b	$62					; b
		dc.b	$24					; $
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.b	$33					; 3
		dc.w	$E
aRobGaudreau:	dc.b	'Rob Gaudreau'
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.w	$E
aDeanEvason:	dc.b	'Dean Evason',0
		dc.b	$12
		dc.b	$61					; a
		dc.b	$12
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$10
aMikeSullivan:	dc.b	'Mike Sullivan',0
		dc.b	$47					; G
		dc.b	$62					; b
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$10
		dc.b	$21					; !
		dc.b	$45					; E
		dc.b	$21					; !
		dc.w	$10
aPerryBerezan:	dc.b	'Perry Berezan',0
		dc.b	$16
		dc.b	$74					; t
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$33					; 3
		dc.w	$C
aRobinBawa:	dc.b	'Robin Bawa'
		dc.b	$26					; &
		dc.b	$B2					; 
		dc.b	$21					; !
		dc.b	$12
		dc.b	$15
		dc.b	$14
		dc.b	$13
		dc.b	$23					; #
		dc.w	$12
aJohanGarpenlov: dc.b	'Johan Garpenlov',0
		dc.b	$10
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$10
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$E
aJeffOdgers:	dc.b	'Jeff Odgers',0
		dc.b	$36					; 6
		dc.b	$91					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$15
		dc.w	$E
aJohnCarter:	dc.b	'John Carter',0
		dc.b	$20
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$28					; (
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$23					; #
		dc.w	$E
aDavidMaley:	dc.b	'David Maley',0
		dc.b	$25					; %
		dc.b	$81					; 
		dc.b	$11
		dc.b	$12
		dc.b	$38					; 8
		dc.b	$20
		dc.b	$14
		dc.b	$14
		dc.w	$E
aPatFalloon_0:	dc.b	'Pat Falloon',0
		dc.b	$17
		dc.b	$74					; t
		dc.b	$43					; C
		dc.b	$23					; #
		dc.b	$13
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.w	$E
aEdCourtenay:	dc.b	'Ed Courtenay'
		dc.b	$39					; 9
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$13
		dc.b	$13
		dc.b	$21					; !
		dc.b	$21					; !
		dc.w	$10
aMarkPederson:	dc.b	'Mark Pederson',0
		dc.b	$18
		dc.b	$82					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$E
aDougWilson:	dc.b	'Doug Wilson',0
		dc.b	$24					; $
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$36					; 6
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.w	$10
aNeilWilkinson:	dc.b	'Neil Wilkinson'
		dc.b	5
		dc.b	$73					; s
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$30					; 0
		dc.b	$44					; D
		dc.b	$43					; C
		dc.w	$12
aSandisOzolinsh: dc.b	'Sandis Ozolinsh',0
		dc.b	6
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aTomPederson:	dc.b	'Tom Pederson'
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.w	$E
aDougZmolek:	dc.b	'Doug Zmolek',0
unk_3EAA:	dc.b	$19					; DATA XREF: ROM:off_ABA14o
		dc.b	$C2					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$24					; $
		dc.w	$10
aDavidWilliams:	dc.b	'David Williams'
		dc.b	3
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$20
		dc.b	$22					; "
		dc.b	$13
		dc.w	$A
aJayMore:	dc.b	'Jay More'
		dc.b	4
		dc.b	$73					; s
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$34					; 4
		dc.w	$E
aPeterAhola:	dc.b	'Peter Ahola',0
		dc.b	$21					; !
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.w	$E
aRobZettler:	dc.b	'Rob Zettler',0
		dc.b	2
		dc.b	$72					; r
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.w	2
		dc.w	$A
aSanJose:	dc.b	'San Jose'
		dc.w	4
aSj:		dc.b	'SJ'
		dc.w	8
aSharks:	dc.b	'Sharks'
		dc.w	$10
aSanJoseArena:	dc.b	'San Jose Arena'
unk_3F30:	dc.b	0					; DATA XREF: ROM:0000035Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$BA					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$6E					; n
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$2C					; ,
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$22					; "
		dc.b	$20
		dc.b	$21					; !
		dc.b	$E9					; 
		dc.b	$D4					; 
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	$C
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$12
		dc.b	$B
		dc.b	6
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$18
		dc.b	$A
		dc.b	3
		dc.b	$C
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	9
		dc.b	4
		dc.b	$F
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$13
		dc.b	$C
		dc.b	3
		dc.b	$D
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$16
		dc.b	4
		dc.b	5
		dc.b	$E
		dc.b	$C
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$18
		dc.b	6
		dc.b	5
		dc.b	$A
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$12
		dc.b	$A
		dc.b	$F
		dc.b	5
		dc.b	6
		dc.b	0
		dc.w	$10
aCurtisJoseph:	dc.b	'Curtis Joseph',0
		dc.b	$31					; 1
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.w	$C
aGuyHebert:	dc.b	'Guy Hebert'
		dc.b	$29					; )
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aCraigJanney:	dc.b	'Craig Janney'
		dc.b	$15
		dc.b	$74					; t
		dc.b	$35					; 5
		dc.b	$43					; C
		dc.b	$20
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$40					; @
		dc.w	$10
aNelsonEmerson:	dc.b	'Nelson Emerson'
		dc.b	7
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$C
aRonSutter:	dc.b	'Ron Sutter'
		dc.b	$22					; "
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$43					; C
		dc.w	$C
aRonWilson:	dc.b	'Ron Wilson'
		dc.b	$18
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.b	$20
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$22					; "
		dc.w	$C
aBobBassen:	dc.b	'Bob Bassen'
		dc.b	$28					; (
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$48					; H
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$23					; #
		dc.w	$10
aPhilippeBozon:	dc.b	'Philippe Bozon'
		dc.b	$36					; 6
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$16
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$13
		dc.w	$E
aIgorKorolev:	dc.b	'Igor Korolev'
		dc.b	$38					; 8
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$20
		dc.b	$21					; !
		dc.b	$30					; 0
		dc.b	$21					; !
		dc.w	$C
aDaveLowry:	dc.b	'Dave Lowry'
		dc.b	$10
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.w	$E
aBasilMcrae:	dc.b	'Basil McRae',0
		dc.b	$17
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$2C					; ,
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$15
		dc.w	$12
aBrendanShanahan: dc.b	'Brendan Shanahan'
		dc.b	$19
		dc.b	$A3					; 
		dc.b	$35					; 5
		dc.b	$44					; D
		dc.b	$27					; '
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$44					; D
		dc.w	$C
aBrettHull_0:	dc.b	'Brett Hull'
		dc.b	$16
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$36					; 6
		dc.b	$33					; 3
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.w	$E
aKevinMiller:	dc.b	'Kevin Miller'
		dc.b	$14
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.w	$E
aRichSutter:	dc.b	'Rich Sutter',0
		dc.b	$23					; #
		dc.b	$73					; s
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$E
aKellyChase:	dc.b	'Kelly Chase',0
		dc.b	$39					; 9
		dc.b	$81					; 
		dc.b	$11
		dc.b	$21					; !
		dc.b	$19
		dc.b	$11
		dc.b	$23					; #
		dc.b	$15
		dc.w	$C
aJeffBrown_0:	dc.b	'Jeff Brown'
		dc.b	$21					; !
		dc.b	$93					; 
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$51					; Q
		dc.b	$42					; B
		dc.w	$10
aGarthButcher:	dc.b	'Garth Butcher',0
		dc.b	5
		dc.b	$92					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.w	$10
aDougCrossman:	dc.b	'Doug Crossman',0
		dc.b	6
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$C
aRickZombo:	dc.b	'Rick Zombo'
		dc.b	4
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$20
		dc.b	$40					; @
		dc.b	$23					; #
		dc.w	$12
aStephaneQuintal: dc.b	'Stephane Quintal'
		dc.b	$33					; 3
		dc.b	$B2					; 
		dc.b	$21					; !
		dc.b	$42					; B
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$E
aLeeNorwood:	dc.b	'Lee Norwood',0
		dc.b	$20
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$38					; 8
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.w	$E
aBretHedican:	dc.b	'Bret Hedican'
		dc.b	$44					; D
		dc.b	$82					; 
		dc.b	$21					; !
		dc.b	$12
		dc.b	$36					; 6
		dc.b	$20
		dc.b	$13
		dc.b	$22					; "
		dc.w	$C
aCurtGiles:	dc.b	'Curt Giles'
unk_41CA:	dc.b	2					; DATA XREF: ROM:off_4E45Co
		dc.b	$53					; S
		dc.b	$21					; !
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$12
		dc.w	$E
aMurrayBaron:	dc.b	'Murray Baron'
		dc.b	$34					; 4
		dc.b	$A3					; 
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$13
		dc.w	2
		dc.w	$C
aSt_Louis:	dc.b	'St. Louis',0
		dc.w	6
aStl:		dc.b	'STL',0
		dc.w	8
aBlues:		dc.b	'Blues',0
		dc.w	$12
aSt_LouisArena:	dc.b	'St. Louis Arena',0
unk_4216:	dc.b	0					; DATA XREF: ROM:00000362o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CE					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	$20
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$A
		dc.b	$22					; "
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	$C
		dc.b	$A6					; 
		dc.b	$E
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EC					; 
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	5
		dc.b	2
		dc.b	0
		dc.b	$F7					; 
		dc.b	$68					; h
		dc.b	$40					; @
		dc.b	1
		dc.b	$14
		dc.b	$13
		dc.b	$C
		dc.b	4
		dc.b	5
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$17
		dc.b	$D
		dc.b	4
		dc.b	$F
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$18
		dc.b	$C
		dc.b	6
		dc.b	$11
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	$E
		dc.b	5
		dc.b	$10
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	$D
		dc.b	4
		dc.b	$F
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	$C
		dc.b	6
		dc.b	5
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$13
		dc.b	$C
		dc.b	$11
		dc.b	$F
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$15
		dc.b	$10
		dc.b	$F
		dc.b	$C
		dc.b	$11
		dc.b	0
		dc.w	$10
aWendellYoung:	dc.b	'Wendell Young',0
		dc.b	1
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	0
		dc.b	0
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$10
aPatJablonski:	dc.b	'Pat Jablonski',0
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	1
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aJ_c_Bergeron:	dc.b	'J.C. Bergeron',0
		dc.b	$30					; 0
		dc.b	$72					; r
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	$11
		dc.w	$10
aBrianBradley_0: dc.b	'Brian Bradley',0
		dc.b	$19
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aChrisKontos:	dc.b	'Chris Kontos'
		dc.b	$16
		dc.b	$83					; 
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$35					; 5
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.w	$10
aAdamCreighton:	dc.b	'Adam Creighton'
		dc.b	$10
		dc.b	$A3					; 
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.w	$E
aMarcBureau:	dc.b	'Marc Bureau',0
		dc.b	$28					; (
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	$C
aRobDimaio:	dc.b	'Rob DiMaio'
		dc.b	$18
		dc.b	$52					; R
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$25					; %
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.w	$E
aSteveKasper:	dc.b	'Steve Kasper'
		dc.b	$11
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.w	$12
aJasonLafreniere: dc.b	'Jason Lafreniere'
		dc.b	$17
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$13
		dc.b	$24					; $
		dc.b	$22					; "
		dc.b	$12
		dc.w	$E
aRandyGilhen:	dc.b	'Randy Gilhen'
		dc.b	$20
		dc.b	$72					; r
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$21					; !
		dc.w	$12
aMikaelAndersson: dc.b	'Mikael Andersson'
		dc.b	$34					; 4
		dc.b	$62					; b
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.b	$10
		dc.b	$32					; 2
		dc.b	$45					; E
		dc.b	$30					; 0
		dc.w	$E
aRobZamuner:	dc.b	'Rob Zamuner',0
		dc.b	7
		dc.b	$92					; 
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$10
aSteveMaltais:	dc.b	'Steve Maltais',0
		dc.b	$37					; 7
		dc.b	$A2					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$30					; 0
		dc.b	$21					; !
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.w	$E
aJohnTucker:	dc.b	'John Tucker',0
		dc.b	$14
		dc.b	$92					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$E
aDantonCole:	dc.b	'Danton Cole',0
		dc.b	$24					; $
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$21					; !
		dc.w	$E
aTimBergland:	dc.b	'Tim Bergland'
		dc.b	$21					; !
		dc.b	$82					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.w	$E
aStanDrulia:	dc.b	'Stan Drulia',0
		dc.b	$27					; '
		dc.b	$71					; q
		dc.b	$21					; !
		dc.b	2
		dc.b	$23					; #
		dc.b	$12
		dc.b	5
		dc.b	$12
		dc.w	$C
aBobBeers:	dc.b	'Bob Beers',0
		dc.b	2
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$10
aRomanHamrlik:	dc.b	'Roman Hamrlik',0
		dc.b	$44					; D
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.w	$10
aShawnChambers:	dc.b	'Shawn Chambers'
		dc.b	$22					; "
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$10
aMarcBergevin:	dc.b	'Marc Bergevin',0
		dc.b	$25					; %
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$C
aJoeReekie:	dc.b	'Joe Reekie'
		dc.b	$29					; )
		dc.b	$B3					; 
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$26					; &
		dc.b	$30					; 0
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.w	$E
aChrisLipuma:	dc.b	'Chris Lipuma'
		dc.b	$40					; @
		dc.b	$61					; a
		dc.b	$12
		dc.b	$11
		dc.b	2
		dc.b	$10
		dc.b	$11
		dc.b	$14
		dc.w	$E
aMattHervey:	dc.b	'Matt Hervey',0
		dc.b	$26					; &
		dc.b	$91					; 
		dc.b	$12
		dc.b	$11
		dc.b	$29					; )
		dc.b	$10
		dc.b	$13
		dc.b	4
		dc.w	2
		dc.w	$C
aTampaBay:	dc.b	'Tampa Bay',0
		dc.w	4
aTb:		dc.b	'TB'
		dc.w	$C
aLightning:	dc.b	'Lightning',0
		dc.w	$18
aFairgroundsExpoHall: dc.b 'Fairgrounds Expo Hall',0
unk_4518:	dc.b	0					; DATA XREF: ROM:00000366o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$D2					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$20
		dc.b	$20
		dc.b	$11
		dc.b	$E8					; 
		dc.b	$96					; 
		dc.b	$10
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	9
		dc.b	4
		dc.b	$D
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$12
		dc.b	9
		dc.b	4
		dc.b	$D
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$17
		dc.b	$B
		dc.b	6
		dc.b	$E
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	$A
		dc.b	5
		dc.b	$F
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	9
		dc.b	4
		dc.b	$D
		dc.b	$F
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$16
		dc.b	$F
		dc.b	5
		dc.b	$E
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$18
		dc.b	6
		dc.b	4
		dc.b	$C
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$12
		dc.b	$C
		dc.b	7
		dc.b	4
		dc.b	6
		dc.b	0
		dc.w	$E
aFelixPotvin_0:	dc.b	'Felix Potvin'
		dc.b	$29					; )
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$E
aDarenPuppa:	dc.b	'Daren Puppa',0
		dc.b	1
		dc.b	$93					; 
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	1
		dc.b	0
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.w	$E
aRickWamsley:	dc.b	'Rick Wamsley'
		dc.b	$30					; 0
		dc.b	$62					; b
		dc.b	$30					; 0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aDougGilmour_0:	dc.b	'Doug Gilmour'
		dc.b	$93					; 
		dc.b	$45					; E
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$60					; `
		dc.b	$43					; C
		dc.w	$E
aJohnCullen:	dc.b	'John Cullen',0
		dc.b	$19
		dc.b	$74					; t
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$12
aMikeKrushelski: dc.b	'Mike Krushelski',0
		dc.b	$26					; &
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$22					; "
		dc.w	$E
aPeterZezel:	dc.b	'Peter Zezel',0
		dc.b	$25					; %
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$40					; @
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$41					; A
		dc.w	$10
aDaveMcllwain:	dc.b	'Dave McLlwain',0
		dc.b	7
		dc.b	$73					; s
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$32					; 2
		dc.w	$12
aDaveAndreychuk: dc.b	'Dave Andreychuk',0
		dc.b	$14
		dc.b	$B3					; 
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$54					; T
		dc.b	$32					; 2
		dc.w	$E
aWendelClark:	dc.b	'Wendel Clark'
		dc.b	$17
		dc.b	$83					; 
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$44					; D
		dc.w	$E
aMarkOsborne:	dc.b	'Mark Osborne'
		dc.b	$21					; !
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$44					; D
		dc.b	$13
		dc.w	$C
aBillBerg:	dc.b	'Bill Berg',0
		dc.b	$10
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$14
aNikolaiBorshevsky: dc.b 'Nikolai Borshevsky'
		dc.b	$16
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$23					; #
		dc.b	$20
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$41					; A
		dc.w	$10
aGlennAnderson:	dc.b	'Glenn Anderson'
		dc.b	9
		dc.b	$74					; t
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$26					; &
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$43					; C
		dc.w	$E
aRobPearson:	dc.b	'Rob Pearson',0
		dc.b	$12
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$11
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$24					; $
		dc.w	$E
aMikeFoligno:	dc.b	'Mike Foligno'
		dc.b	$71					; q
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$55					; U
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.w	$10
aKenBaumgartnr:	dc.b	'Ken Baumgartnr'
		dc.b	$22					; "
		dc.b	$92					; 
		dc.b	$20
		dc.b	$21					; !
		dc.b	$2A					; *
		dc.b	$10
		dc.b	$23					; #
		dc.b	$24					; $
		dc.w	$C
aToddGill:	dc.b	'Todd Gill',0
		dc.b	$23					; #
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$23					; #
		dc.w	$E
aDaveEllett:	dc.b	'Dave Ellett',0
		dc.b	4
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	$45					; E
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$42					; B
		dc.w	$12
aDimitriMironov: dc.b	'Dimitri Mironov',0
		dc.b	$15
		dc.b	$73					; s
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$12
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.w	$12
aDrakeBerehowsky: dc.b	'Drake Berehowsky'
		dc.b	$55					; U
		dc.b	$A2					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.w	$E
aJamieMacoun:	dc.b	'Jamie Macoun'
		dc.b	$34					; 4
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$44					; D
		dc.b	$42					; B
		dc.w	$C
aBobRouse:	dc.b	'Bob Rouse',0
		dc.b	3
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$20
		dc.b	$44					; D
		dc.b	$23					; #
		dc.w	$12
aSylvainLefebvre: dc.b	'Sylvain Lefebvre'
		dc.b	2
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$20
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$C
aBobMcgill:	dc.b	'Bob McGill'
		dc.b	8
		dc.b	$82					; 
		dc.b	$20
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$24					; $
		dc.w	2
		dc.w	$A
aToronto:	dc.b	'Toronto',0
		dc.w	6
aTor:		dc.b	'TOR',0
		dc.w	$E
aMapleLeafs:	dc.b	'Maple Leafs',0
		dc.w	$14
aMapleLeafGardens: dc.b	'Maple Leaf Gardens'
unk_481C:	dc.b	0					; DATA XREF: ROM:off_36Ao
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$C0					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$AE					; 
		dc.b	0
		dc.b	0
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$8C					; 
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$61					; a
		dc.b	2
		dc.b	$11
		dc.b	$F8					; 
		dc.b	$A6					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	7
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$16
		dc.b	$A
		dc.b	3
		dc.b	$E
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$12
		dc.b	9
		dc.b	5
		dc.b	$D
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$19
		dc.b	7
		dc.b	4
		dc.b	$10
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$14
		dc.b	$D
		dc.b	4
		dc.b	$E
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	7
		dc.b	3
		dc.b	9
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$12
		dc.b	$D
		dc.b	5
		dc.b	4
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$17
		dc.b	4
		dc.b	$10
		dc.b	$D
		dc.b	5
		dc.b	0
		dc.w	$E
aKirkMclean:	dc.b	'Kirk McLean',0
		dc.b	1
		dc.b	$84					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.w	$E
aKayWhitmore:	dc.b	'Kay Whitmore'
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	0
		dc.b	0
		dc.b	$23					; #
		dc.b	$23					; #
		dc.w	$10
aCliffRonning:	dc.b	'Cliff Ronning',0
		dc.b	7
		dc.b	$55					; U
		dc.b	$54					; T
		dc.b	$32					; 2
		dc.b	$20
		dc.b	$43					; C
		dc.b	$51					; Q
		dc.b	$41					; A
		dc.w	$E
aPetrNedved:	dc.b	'Petr Nedved',0
		dc.b	$19
		dc.b	$52					; R
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$10
		dc.b	$36					; 6
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$12
aAnatoliSemenov: dc.b	'Anatoli Semenov',0
		dc.b	$20
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$C
aTomFergus:	dc.b	'Tom Fergus'
		dc.b	$15
		dc.b	$A2					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.w	$12
aGeoffCourtnall: dc.b	'Geoff Courtnall',0
		dc.b	$14
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$14
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.w	$10
aMurrayCraven:	dc.b	'Murray Craven',0
		dc.b	$32					; 2
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.w	$C
aGregAdams:	dc.b	'Greg Adams'
		dc.b	8
		dc.b	$73					; s
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$31					; 1
		dc.w	$10
aSergioMomesso:	dc.b	'Sergio Momesso'
		dc.b	$27					; '
		dc.b	$B4					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$24					; $
		dc.w	$E
aGinoOdjick:	dc.b	'Gino Odjick',0
		dc.b	$29					; )
		dc.b	$B2					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$3A					; :
		dc.b	$21					; !
		dc.b	$33					; 3
		dc.b	$26					; &
		dc.w	$C
aGarryValk:	dc.b	'Garry Valk'
		dc.b	$23					; #
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$C
aPavelBure_0:	dc.b	'Pavel Bure'
		dc.b	$10
		dc.b	$55					; U
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$20
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$42					; B
		dc.w	$10
aTrevorLinden:	dc.b	'Trevor Linden',0
		dc.b	$16
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$43					; C
		dc.b	$42					; B
		dc.w	$C
aDixonWard:	dc.b	'Dixon Ward'
		dc.b	$17
		dc.b	$92					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.w	$E
aJimSandlak:	dc.b	'Jim Sandlak',0
		dc.b	$25					; %
		dc.b	$B1					; 
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	$C
aTimHunter:	dc.b	'Tim Hunter'
		dc.b	$26					; &
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$39					; 9
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$24					; $
		dc.w	$E
aJyrkiLumme:	dc.b	'Jyrki Lumme',0
		dc.b	$21					; !
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$E
aDougLidster:	dc.b	'Doug Lidster'
		dc.b	3
		dc.b	$93					; 
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$10
aAdrienPlavsic:	dc.b	'Adrien Plavsic'
		dc.b	6
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.w	$C
aJiriSlegr:	dc.b	'Jiri Slegr'
		dc.b	$24					; $
		dc.b	$A3					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.w	$10
aGeraldDiduck:	dc.b	'Gerald Diduck',0
		dc.b	4
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.w	$E
aDaveBabych:	dc.b	'Dave Babych',0
		dc.b	$44					; D
		dc.b	$B2					; 
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$26					; &
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aDanaMurzyn:	dc.b	'Dana Murzyn',0
		dc.b	5
		dc.b	$92					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$44					; D
		dc.b	$24					; $
		dc.w	$E
aRobertDirk:	dc.b	'Robert Dirk',0
		dc.b	$22					; "
		dc.b	$B2					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	2
		dc.w	$C
aVancouver:	dc.b	'Vancouver',0
		dc.w	6
aVan:		dc.b	'VAN',0
		dc.w	$A
aCanucks:	dc.b	'Canucks',0
		dc.w	$12
aPacificColiseum: dc.b	'Pacific Coliseum'
unk_4B0A:	dc.b	0					; DATA XREF: ROM:00000372o
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$CC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	4
		dc.b	$C
		dc.b	4
		dc.b	$2C					; ,
		dc.b	$A
		dc.b	$20
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	4
		dc.b	$C
		dc.b	4
		dc.b	$2C					; ,
		dc.b	$C
		dc.b	$CC					; 
		dc.b	8
		dc.b	$20
		dc.b	8
		dc.b	$42					; B
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$45					; E
		dc.b	$11
		dc.b	$11
		dc.b	$F8					; 
		dc.b	$D6					; 
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$12
		dc.b	4
		dc.b	3
		dc.b	$D
		dc.b	9
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$12
		dc.b	8
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$18
		dc.b	$A
		dc.b	6
		dc.b	$11
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$12
		dc.b	8
		dc.b	3
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	9
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$15
		dc.b	$A
		dc.b	6
		dc.b	5
		dc.b	$10
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	$16
		dc.b	5
		dc.b	$10
		dc.b	$A
		dc.b	6
		dc.b	0
		dc.w	$E
aBobEssensa:	dc.b	'Bob Essensa',0
		dc.b	$35					; 5
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	0
		dc.b	0
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$E
aJimHrivnak:	dc.b	'Jim Hrivnak',0
		dc.b	$30					; 0
		dc.b	$62					; b
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aAlexeiZhamnov:	dc.b	'Alexei Zhamnov'
		dc.b	$10
		dc.b	$75					; u
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$54					; T
		dc.b	$41					; A
		dc.b	$52					; R
		dc.w	$E
aThomasSteen:	dc.b	'Thomas Steen'
		dc.b	$25					; %
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$53					; S
		dc.w	$12
aLucianoBorsato: dc.b	'Luciano Borsato',0
		dc.b	$38					; 8
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.w	$E
aMikeEagles:	dc.b	'Mike Eagles',0
		dc.b	$36					; 6
		dc.b	$73					; s
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$54					; T
		dc.b	$22					; "
		dc.b	$51					; Q
		dc.b	$23					; #
		dc.w	$C
aStuBarnes:	dc.b	'Stu Barnes'
		dc.b	$14
		dc.b	$52					; R
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$21					; !
		dc.w	$10
aDarrinShannon:	dc.b	'Darrin Shannon'
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.b	$33					; 3
		dc.w	$10
aKeithTkachuk:	dc.b	'Keith Tkachuk',0
		dc.b	7
		dc.b	$93					; 
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$36					; 6
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$24					; $
		dc.w	$C
aKrisKing:	dc.b	'Kris King',0
		dc.b	$17
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$24					; $
		dc.w	$10
aRussRomaniuk:	dc.b	'Russ Romaniuk',0
		dc.b	$21					; !
		dc.b	$63					; c
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$10
aAndyBrickley:	dc.b	'Andy Brickley',0
		dc.b	$23					; #
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$20
		dc.b	$20
		dc.b	$30					; 0
		dc.w	$10
aTeemuSelanne_0: dc.b	'Teemu Selanne',0
		dc.b	$13
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$42					; B
		dc.w	$10
aEvgenyDavydov:	dc.b	'Evgeny Davydov'
		dc.b	$11
		dc.b	$64					; d
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$13
		dc.b	$34					; 4
		dc.b	$35					; 5
		dc.b	$32					; 2
		dc.w	$C
aJohnDruce:	dc.b	'John Druce'
		dc.b	$15
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$10
aBryanErickson:	dc.b	'Bryan Erickson'
		dc.b	$18
		dc.b	$52					; R
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$21					; !
		dc.w	$A
aTieDomi:	dc.b	'Tie Domi'
		dc.b	$20
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$39					; 9
		dc.b	$23					; #
		dc.b	$41					; A
		dc.b	$26					; &
		dc.w	$E
aPhilHousley_0:	dc.b	'Phil Housley'
		dc.b	6
		dc.b	$66					; f
		dc.b	$54					; T
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$62					; b
		dc.b	$40					; @
		dc.b	$62					; b
		dc.w	$10
aTeppoNumminen:	dc.b	'Teppo Numminen'
		dc.b	$27					; '
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$32					; 2
		dc.w	$12
aFredrikOlausson: dc.b	'Fredrik Olausson'
		dc.b	4
		dc.b	$94					; 
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$11
		dc.b	$42					; B
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.w	$10
aSergeiBautin:	dc.b	'Sergei Bautin',0
		dc.b	3
		dc.b	$63					; c
		dc.b	$32					; 2
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.w	$E
aIgorUlanov:	dc.b	'Igor Ulanov',0
		dc.b	5
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$40					; @
		dc.b	$24					; $
		dc.w	$C
aMikeLalor:	dc.b	'Mike Lalor'
		dc.b	$22					; "
		dc.b	$93					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$20
		dc.b	$45					; E
		dc.b	$23					; #
		dc.w	$E
aDeanKennedy:	dc.b	'Dean Kennedy'
		dc.b	$26					; &
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$44					; D
		dc.b	$23					; #
		dc.w	$10
aRandyCarlyle:	dc.b	'Randy Carlyle',0
		dc.b	8
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	2
		dc.w	$A
aWinnipeg:	dc.b	'Winnipeg'
		dc.w	6
aWpg:		dc.b	'WPG',0
		dc.w	6
aJets:		dc.b	'Jets'
		dc.w	$10
aWinnipegArena:	dc.b	'Winnipeg Arena'
unk_4DFC:	dc.b	0					; DATA XREF: ROM:0000036Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$BE					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2C					; ,
		dc.b	$C
		dc.b	$CC					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	$22					; "
		dc.b	8
		dc.b	$44					; D
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	8
		dc.b	$88					; 
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2A					; *
		dc.b	0
		dc.b	8
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$52					; R
		dc.b	$20
		dc.b	$11
		dc.b	$E8					; 
		dc.b	$B5					; 
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	4
		dc.b	3
		dc.b	$C
		dc.b	$D
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	$A
		dc.b	4
		dc.b	$E
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	9
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$16
		dc.b	$B
		dc.b	5
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$14
		dc.b	$11
		dc.b	4
		dc.b	3
		dc.b	$C
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$12
		dc.b	$13
		dc.b	6
		dc.b	5
		dc.b	$D
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$15
		dc.b	$11
		dc.b	3
		dc.b	$E
		dc.b	$A
		dc.b	6
		dc.b	0
		dc.b	1
		dc.b	$13
		dc.b	$14
		dc.b	$A
		dc.b	6
		dc.b	3
		dc.b	$E
		dc.b	0
		dc.w	$E
aDonBeaupre:	dc.b	'Don Beaupre',0
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$10
aRickTabaracci:	dc.b	'Rick Tabaracci'
		dc.b	$31					; 1
		dc.b	$62					; b
		dc.b	$31					; 1
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	$E
aMikeRidley:	dc.b	'Mike Ridley',0
		dc.b	$17
		dc.b	$94					; 
		dc.b	$44					; D
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$42					; B
		dc.w	$14
aDimitriKhristich: dc.b	'Dimitri Khristich',0
		dc.b	8
		dc.b	$73					; s
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$E
aDaleHunter:	dc.b	'Dale Hunter',0
		dc.b	$32					; 2
		dc.b	$82					; 
		dc.b	$24					; $
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$24					; $
		dc.b	$40					; @
		dc.b	$44					; D
		dc.w	$10
aMichalPivonka:	dc.b	'Michal Pivonka'
		dc.b	$20
		dc.b	$83					; 
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$43					; C
		dc.w	$12
aSteveKonowlchuk: dc.b	'Steve Konowlchuk'
		dc.b	$22					; "
		dc.b	$62					; b
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aReggieSavage:	dc.b	'Reggie Savage',0
		dc.b	$15
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$12
		dc.b	$26					; &
		dc.b	$22					; "
		dc.b	$13
		dc.b	$22					; "
		dc.w	$10
aBobCarpenter:	dc.b	'Bob Carpenter',0
		dc.b	$11
		dc.b	$74					; t
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$43					; C
		dc.w	$E
aToddKrygier:	dc.b	'Todd Krygier'
		dc.b	$21					; !
		dc.b	$62					; b
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$24					; $
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$12
		dc.w	$A
aAlanMay:	dc.b	'Alan May'
		dc.b	$16
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$37					; 7
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$25					; %
		dc.w	$E
aPeterBondra_0:	dc.b	'Peter Bondra'
		dc.b	$12
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$32					; 2
		dc.w	$E
aPatElynuik:	dc.b	'Pat Elynuik',0
		dc.b	$19
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.w	$E
aKellyMiller:	dc.b	'Kelly Miller'
		dc.b	$10
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$20
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$41					; A
		dc.w	$E
aKeithJones:	dc.b	'Keith Jones',0
		dc.b	$26					; &
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.w	$10
aPaulMacdermid:	dc.b	'Paul MacDermid'
		dc.b	$23					; #
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$47					; G
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$10
aKevinHatcher:	dc.b	'Kevin Hatcher',0
		dc.b	4
		dc.b	$C3					; 
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$43					; C
		dc.w	$C
aAlIafrate_0:	dc.b	'Al Iafrate'
		dc.b	$34					; 4
		dc.b	$B4					; 
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$44					; D
		dc.w	$E
aSylvainCote:	dc.b	'Sylvain Cote'
		dc.b	3
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$44					; D
		dc.b	$32					; 2
		dc.w	$12
aCalleJohansson: dc.b	'Calle Johansson',0
		dc.b	6
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$42					; B
		dc.w	$10
aPaulCavallini:	dc.b	'Paul Cavallini'
		dc.b	$14
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$32					; 2
		dc.w	$10
aShawnAnderson:	dc.b	'Shawn Anderson'
		dc.b	$36					; 6
		dc.b	$92					; 
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$36					; 6
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.w	$10
aJasonWoolley:	dc.b	'Jason Woolley',0
		dc.b	$25					; %
		dc.b	$62					; b
		dc.b	$21					; !
		dc.b	$11
		dc.b	$22					; "
		dc.b	$20
		dc.b	$12
		dc.b	$12
		dc.w	$E
aRodLangway:	dc.b	'Rod Langway',0
		dc.b	5
		dc.b	$B3					; 
		dc.b	$20
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.b	$23					; #
		dc.w	2
		dc.w	$C
aWashington:	dc.b	'Washington'
		dc.w	6
aWsh:		dc.b	'WSH',0
		dc.w	$A
aCapitals:	dc.b	'Capitals'
		dc.w	$10
aCapitalCentre:	dc.b	'Capital Centre'
unk_50E6:	dc.b	0					; DATA XREF: ROM:0000032Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	8
		dc.b	8
		dc.b	$88					; 
		dc.b	$C
		dc.b	$CC					; 
		dc.b	6
		dc.b	$66					; f
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	$6C					; l
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	2
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	7
		dc.b	2
		dc.b	$13
		dc.b	$96					; 
		dc.b	$D6					; 
		dc.b	0
		dc.b	1
		dc.b	$E
		dc.b	$C
		dc.b	8
		dc.b	5
		dc.b	4
		dc.b	7
		dc.b	0
		dc.b	1
		dc.b	$E
		dc.b	$11
		dc.b	8
		dc.b	5
		dc.b	3
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	$C
		dc.b	6
		dc.b	$A
		dc.b	$B
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$F
		dc.b	7
		dc.b	9
		dc.b	4
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$E
		dc.b	$C
		dc.b	8
		dc.b	5
		dc.b	4
		dc.b	$A
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$11
		dc.b	6
		dc.b	$A
		dc.b	3
		dc.b	5
		dc.b	0
		dc.b	1
		dc.b	$E
		dc.b	$C
		dc.b	9
		dc.b	5
		dc.b	7
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$11
		dc.b	7
		dc.b	8
		dc.b	9
		dc.b	5
		dc.b	0
		dc.w	$12
aJohnVanbiesbrk_0: dc.b	'John Vanbiesbrk',0
		dc.b	$34					; 4
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	0
		dc.b	0
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.w	$12
aMarkFitzpatrik_0: dc.b	'Mark Fitzpatrik',0
		dc.b	$30					; 0
		dc.b	$72					; r
		dc.b	$42					; B
		dc.b	$23					; #
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$10
aTomFitzgerald_0: dc.b	'Tom Fitzgerald'
		dc.b	$14
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.w	$10
aScottMellanby_0: dc.b	'Scott Mellanby'
		dc.b	$27					; '
		dc.b	$91					; 
		dc.b	$13
		dc.b	$33					; 3
		dc.b	$47					; G
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.w	$12
aBrianSkrudland_0: dc.b	'Brian Skrudland',0
		dc.b	$39					; 9
		dc.b	$84					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$44					; D
		dc.b	$34					; 4
		dc.w	$C
aMikeHough_0:	dc.b	'Mike Hough'
		dc.b	$18
		dc.b	$73					; s
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$24					; $
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.w	$C
aDaveLowry_0:	dc.b	'Dave Lowry'
		dc.b	$10
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$22					; "
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.w	$10
aAndreiLomakin_0: dc.b	'Andrei Lomakin'
		dc.b	$23					; #
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$12
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.w	$E
aRandyGilhen_0:	dc.b	'Randy Gilhen'
		dc.b	$20
		dc.b	$72					; r
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$21					; !
		dc.w	$10
aJesseBelanger:	dc.b	'Jesse Belanger'
		dc.b	$29					; )
		dc.b	$41					; A
		dc.b	$23					; #
		dc.b	$12
		dc.b	$20
		dc.b	$22					; "
		dc.b	$14
		dc.b	$21					; !
		dc.w	$E
aBillLindsay_0:	dc.b	'Bill Lindsay'
		dc.b	$22					; "
		dc.b	$71					; q
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$20
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$21					; !
		dc.w	$E
aJoeCirella_0:	dc.b	'Joe Cirella',0
		dc.b	6
		dc.b	$A3					; 
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.w	$14
aAlexnderGodynyuk_0: dc.b 'Alexnder Godynyuk',0
		dc.b	$21					; !
		dc.b	$A2					; 
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$38					; 8
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$12
		dc.w	$E
aGordMurphy_0:	dc.b	'Gord Murphy',0
		dc.b	$28					; (
		dc.b	$85					; 
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$49					; I
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.w	$C
aGordHynes_0:	dc.b	'Gord Hynes'
		dc.b	$26					; &
		dc.b	$42					; B
		dc.b	$22					; "
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$22					; "
		dc.w	$E
aMilanTichy:	dc.b	'Milan Tichy',0
		dc.b	$43					; C
		dc.b	$83					; 
		dc.b	$21					; !
		dc.b	$12
		dc.b	$38					; 8
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$24					; $
		dc.w	$12
aStephaneRicher_0: dc.b	'Stephane Richer',0
		dc.b	$25					; %
		dc.b	$93					; 
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.w	2
		dc.w	$A
aFlorida:	dc.b	'Florida',0
		dc.w	6
aFla:		dc.b	'FLA',0
		dc.w	$A
aPanthers:	dc.b	'Panthers'
		dc.w	$E
aMiamiArena:	dc.b	'Miami Arena',0
unk_5330:	dc.b	0					; DATA XREF: ROM:0000030Eo
		dc.b	$92					; 
		dc.b	0
		dc.b	$C
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$4C					; L
		dc.b	0
		dc.b	$50					; P
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	0
		dc.b	$40					; @
		dc.b	2
		dc.b	$60					; `
		dc.b	2
		dc.b	$40					; @
		dc.b	$A
		dc.b	$AA					; 
		dc.b	$E
		dc.b	$EE					; 
		dc.b	4
		dc.b	$44					; D
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	$E
		dc.b	$E8					; 
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	$22					; "
		dc.b	8
		dc.b	$8C					; 
		dc.b	6
		dc.b	$6A					; j
		dc.b	2
		dc.b	$A6					; 
		dc.b	$A
		dc.b	$CA					; 
		dc.b	0
		dc.b	$40					; @
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$22					; "
		dc.b	2
		dc.b	$22					; "
		dc.b	4
		dc.b	$8C					; 
		dc.b	$A
		dc.b	$84					; 
		dc.b	7
		dc.b	2
		dc.b	$13
		dc.b	$97					; 
		dc.b	$D6					; 
		dc.b	0
		dc.b	1
		dc.b	$C
		dc.b	$E
		dc.b	4
		dc.b	6
		dc.b	8
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$C
		dc.b	$10
		dc.b	3
		dc.b	6
		dc.b	$B
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$E
		dc.b	$F
		dc.b	9
		dc.b	8
		dc.b	7
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$12
		dc.b	4
		dc.b	$A
		dc.b	5
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$C
		dc.b	$10
		dc.b	4
		dc.b	6
		dc.b	8
		dc.b	4
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$F
		dc.b	3
		dc.b	9
		dc.b	$B
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	$C
		dc.b	$E
		dc.b	6
		dc.b	7
		dc.b	4
		dc.b	8
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	$10
		dc.b	4
		dc.b	8
		dc.b	7
		dc.b	6
		dc.b	0
		dc.w	$C
aGuyHebert_0:	dc.b	'Guy Hebert'
		dc.b	$29					; )
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$E
aRonTugnutt_0:	dc.b	'Ron Tugnutt',0
		dc.b	1
		dc.b	$22					; "
		dc.b	$33					; 3
		dc.b	$31					; 1
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	$12
		dc.w	$E
aStevenKing_0:	dc.b	'Steven King',0
		dc.b	$27					; '
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$13
		dc.b	$23					; #
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.w	$C
aTroyLoney_0:	dc.b	'Troy Loney'
		dc.b	$24					; $
		dc.b	$A2					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$13
		dc.w	$E
aStuGrimson_0:	dc.b	'Stu Grimson',0
		dc.b	$23					; #
		dc.b	$B1					; 
		dc.b	$10
		dc.b	$21					; !
		dc.b	$3A					; :
		dc.b	$11
		dc.b	$25					; %
		dc.b	$14
		dc.w	$C
aTerryYake_0:	dc.b	'Terry Yake'
		dc.b	$25					; %
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$C
aBobCorkum_0:	dc.b	'Bob Corkum'
		dc.b	$30					; 0
		dc.b	$A2					; 
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.w	$12
aAnatoliSemenov_0: dc.b	'Anatoli Semenov',0
		dc.b	$20
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$43					; C
		dc.b	$20
		dc.b	$33					; 3
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.w	$E
aLonnieLoach_0:	dc.b	'Lonnie Loach'
		dc.b	$28					; (
		dc.b	$62					; b
		dc.b	$33					; 3
		dc.b	$22					; "
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$22					; "
		dc.w	$C
aRobinBawa_0:	dc.b	'Robin Bawa'
		dc.b	$26					; &
		dc.b	$B2					; 
		dc.b	$21					; !
		dc.b	$12
		dc.b	$15
		dc.b	$14
		dc.b	$13
		dc.b	$23					; #
		dc.w	$E
aTimSweeney:	dc.b	'Tim Sweeney',0
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$23					; #
		dc.b	$22					; "
		dc.b	$10
		dc.b	$22					; "
		dc.b	$14
		dc.b	$11
		dc.w	$12
aAlexeiKasatonov_0: dc.b 'Alexei Kasatonov'
		dc.b	7
		dc.b	$B4					; 
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$44					; D
		dc.b	$41					; A
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.w	$C
aSeanHill_0:	dc.b	'Sean Hill',0
		dc.b	$38					; 8
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$14
		dc.w	$12
aRandyLadouceur_0: dc.b	'Randy Ladouceur',0
		dc.b	$39					; 9
		dc.b	$B3					; 
		dc.b	$31					; 1
		dc.b	$31					; 1
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$35					; 5
		dc.b	$14
		dc.w	$10
aDavidWilliams_0: dc.b	'David Williams'
		dc.b	3
		dc.b	$82					; 
		dc.b	$22					; "
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$20
		dc.b	$22					; "
		dc.b	$13
		dc.w	$E
aBillHoulder:	dc.b	'Bill Houlder'
		dc.b	$33					; 3
		dc.b	$B3					; 
		dc.b	$23					; #
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$24					; $
		dc.b	$21					; !
		dc.w	$E
aBobbyDollas:	dc.b	'Bobby Dollas'
		dc.b	$32					; 2
		dc.b	$B2					; 
		dc.b	$11
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$10
		dc.b	$21					; !
		dc.b	$11
		dc.w	$E
aDennisVial:	dc.b	'Dennis Vial',0
		dc.b	$17
		dc.b	$B1					; 
		dc.b	$11
		dc.b	$21					; !
		dc.b	$38					; 8
		dc.b	$10
		dc.b	$21					; !
		dc.b	$13
		dc.w	2
		dc.w	$A
aAnaheim:	dc.b	'Anaheim',0
		dc.w	6
aAnh:		dc.b	'ANH',0
		dc.w	$E
aMightyDucks:	dc.b	'Mighty Ducks'
		dc.w	6
aPond:		dc.b	'Pond'
		dc.b	2					; Unkown data.
		dc.b	1
		dc.b	$B
		dc.b	$12
		dc.b	$D
		dc.b	$18
		dc.b	$C
		dc.b	$11
		dc.b	$16
		dc.b	6
		dc.b	$14
		dc.b	4
		dc.b	$A
		dc.b	3
		dc.b	$19
		dc.b	$17
		dc.b	$D
		dc.b	2
		dc.b	$15
		dc.b	$B
		dc.b	$12
		dc.b	$11
		dc.b	$E
		dc.b	1
		dc.b	3
		dc.b	$16
		dc.b	$A
		dc.b	$17
		dc.b	$14
		dc.b	6
		dc.b	5
		dc.b	4
		dc.b	$C
		dc.b	$11
		dc.b	$10
		dc.b	2
		dc.b	9
		dc.b	$B
		dc.b	$18
		dc.b	1
		dc.b	5
		dc.b	$16
		dc.b	7
		dc.b	4
		dc.b	$13
		dc.b	$17
		dc.b	$A
		dc.b	6
		dc.b	9
		dc.b	$11
		dc.b	$12
		dc.b	$B
		dc.b	$18
		dc.b	1
		dc.b	$C
		dc.b	2
		dc.b	$A
		dc.b	$16
		dc.b	7
		dc.b	6
		dc.b	5
		dc.b	4
		dc.b	3
		dc.b	$17
		dc.b	$12
		dc.b	2
		dc.b	$C
		dc.b	$11
		dc.b	9
		dc.b	$B
		dc.b	$10
		dc.b	1
		dc.b	3
		dc.b	6
		dc.b	$13
		dc.b	$16
		dc.b	5
		dc.b	$17
		dc.b	7
		dc.b	4
		dc.b	$F
		dc.b	1
		dc.b	9
		dc.b	$B
		dc.b	$D
		dc.b	$11
		dc.b	$E
		dc.b	2
		dc.b	$13
		dc.b	$17
		dc.b	3
		dc.b	6
		dc.b	5
		dc.b	4
		dc.b	7
		dc.b	$16
		dc.b	$10
		dc.b	$B
		dc.b	$F
		dc.b	$11
		dc.b	$15
		dc.b	2
		dc.b	$D
		dc.b	1
		dc.b	$14
		dc.b	6
		dc.b	$19
		dc.b	4
		dc.b	0
		dc.b	$17
		dc.b	3
		dc.b	$16
		dc.b	$C
		dc.b	1
		dc.b	$12
		dc.b	$11
		dc.b	$10
		dc.b	$B
		dc.b	$D
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	$17
		dc.b	$A
		dc.b	$16
		dc.b	$F
		dc.b	$11
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	$B
		dc.b	$18
		dc.b	2
		dc.b	3
		dc.b	$16
		dc.b	5
		dc.b	$17
		dc.b	$19
		dc.b	4
		dc.b	$A
		dc.b	6
		dc.b	$E
		dc.b	2
		dc.b	$10
		dc.b	1
		dc.b	$12
		dc.b	$B
		dc.b	9
		dc.b	$11
		dc.b	$19
		dc.b	4
		dc.b	$13
		dc.b	6
		dc.b	$14
		dc.b	$17
		dc.b	3
		dc.b	$16
		dc.b	$10
		dc.b	$B
		dc.b	$12
		dc.b	1
		dc.b	$C
		dc.b	$11
		dc.b	$D
		dc.b	2
		dc.b	$13
		dc.b	6
		dc.b	$19
		dc.b	$16
		dc.b	$A
		dc.b	$17
		dc.b	3
		dc.b	4
		dc.b	$D
		dc.b	$11
		dc.b	$F
		dc.b	$B
		dc.b	$E
		dc.b	2
		dc.b	$C
		dc.b	1
		dc.b	5
		dc.b	6
		dc.b	$A
		dc.b	4
		dc.b	$13
		dc.b	$16
		dc.b	$14
		dc.b	$17
		dc.b	$12
		dc.b	1
		dc.b	$E
		dc.b	2
		dc.b	$F
		dc.b	$B
		dc.b	$10
		dc.b	$11
		dc.b	5
		dc.b	$17
		dc.b	$13
		dc.b	$16
		dc.b	7
		dc.b	4
		dc.b	3
		dc.b	6
		dc.b	$C
		dc.b	$B
		dc.b	$E
		dc.b	$11
		dc.b	$F
		dc.b	1
		dc.b	9
		dc.b	2
		dc.b	$14
		dc.b	$16
		dc.b	$13
		dc.b	4
		dc.b	5
		dc.b	6
		dc.b	7
		dc.b	$17
		dc.b	$18
		dc.b	$D
		dc.b	$E
		dc.b	$B
		dc.b	$10
		dc.b	2
		dc.b	$15
		dc.b	$11
		dc.b	$16
		dc.b	$14
		dc.b	0
		dc.b	$A
		dc.b	$19
		dc.b	4
		dc.b	$13
		dc.b	$17
		dc.b	$12
		dc.b	2
		dc.b	$15
		dc.b	$D
		dc.b	$C
		dc.b	$B
		dc.b	9
		dc.b	$11
		dc.b	7
		dc.b	$14
		dc.b	5
		dc.b	$A
		dc.b	$13
		dc.b	$17
		dc.b	6
		dc.b	4
		dc.b	1
		dc.b	$15
		dc.b	$F
		dc.b	2
		dc.b	$C
		dc.b	$B
		dc.b	$12
		dc.b	$11
		dc.b	$16
		dc.b	$14
		dc.b	$13
		dc.b	$A
		dc.b	5
		dc.b	$17
		dc.b	$19
		dc.b	4
		dc.b	9
		dc.b	$D
		dc.b	$C
		dc.b	2
		dc.b	$F
		dc.b	$B
		dc.b	1
		dc.b	$11
		dc.b	3
		dc.b	$17
		dc.b	5
		dc.b	$A
		dc.b	$16
		dc.b	4
		dc.b	7
		dc.b	$14
		dc.b	$F
		dc.b	$11
		dc.b	9
		dc.b	$B
		dc.b	8
		dc.b	2
		dc.b	$E
		dc.b	$D
		dc.b	$16
		dc.b	$14
		dc.b	$19
		dc.b	4
		dc.b	6
		dc.b	$17
		dc.b	7
		dc.b	$A
		dc.b	$18
		dc.b	$D
		dc.b	$C
		dc.b	$11
		dc.b	$E
		dc.b	2
		dc.b	1
		dc.b	$B
		dc.b	6
		dc.b	$17
		dc.b	5
		dc.b	$A
		dc.b	7
		dc.b	4
		dc.b	$19
		dc.b	$14
		dc.b	$F
		dc.b	$B
		dc.b	$12
		dc.b	2
		dc.b	1
		dc.b	$D
		dc.b	$18
		dc.b	$11
		dc.b	7
		dc.b	$14
		dc.b	3
		dc.b	$A
		dc.b	$19
		dc.b	$17
		dc.b	6
		dc.b	4
		dc.b	$F
		dc.b	$11
		dc.b	$15
		dc.b	$B
		dc.b	$12
		dc.b	$D
		dc.b	$C
		dc.b	2
		dc.b	$13
		dc.b	$A
		dc.b	3
		dc.b	4
		dc.b	$19
		dc.b	$14
		dc.b	7
		dc.b	$17
		dc.b	$F
		dc.b	$B
		dc.b	$10
		dc.b	2
		dc.b	$C
		dc.b	8
		dc.b	$12
		dc.b	$11
		dc.b	5
		dc.b	4
		dc.b	$16
		dc.b	$A
		dc.b	$19
		dc.b	$14
		dc.b	$13
		dc.b	$17
		dc.b	$C
		dc.b	$B
		dc.b	$12
		dc.b	$11
		dc.b	$F
		dc.b	$D
		dc.b	$10
		dc.b	2
		dc.b	6
		dc.b	$17
		dc.b	$13
		dc.b	$A
		dc.b	$16
		dc.b	$14
		dc.b	7
		dc.b	4
		dc.b	9
		dc.b	$11
		dc.b	$12
		dc.b	$D
		dc.b	$F
		dc.b	$B
		dc.b	1
		dc.b	2
		dc.b	6
		dc.b	4
		dc.b	7
		dc.b	$14
		dc.b	3
		dc.b	$A
		dc.b	$16
		dc.b	$17
		dc.b	$E
		dc.b	$D
		dc.b	1
		dc.b	2
		dc.b	$18
		dc.b	$B
		dc.b	$F
		dc.b	$11
		dc.b	6
		dc.b	4
		dc.b	5
		dc.b	$17
		dc.b	7
		dc.b	$A
		dc.b	$19
		dc.b	$14
		dc.b	9
		dc.b	2
		dc.b	$F
		dc.b	$D
		dc.b	$C
		dc.b	$11
		dc.b	1
		dc.b	$B
		dc.b	$16
		dc.b	$A
		dc.b	3
		dc.b	$14
		dc.b	6
		dc.b	$17
		dc.b	$13
		dc.b	4
		dc.b	$12
		dc.b	$D
		dc.b	$15
		dc.b	2
		dc.b	1
		dc.b	$B
		dc.b	8
		dc.b	$11
		dc.b	7
		dc.b	$A
		dc.b	$19
		dc.b	$14
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	$17
		dc.b	9
		dc.b	$B
		dc.b	$C
		dc.b	$D
		dc.b	$18
		dc.b	2
		dc.b	$12
		dc.b	$11
		dc.b	$13
		dc.b	$14
		dc.b	0
		dc.b	$17
		dc.b	3
		dc.b	4
		dc.b	6
		dc.b	$A
		dc.b	$E
		dc.b	2
		dc.b	8
		dc.b	$15
		dc.b	$18
		dc.b	$11
		dc.b	1
		dc.b	$B
		dc.b	$16
		dc.b	4
		dc.b	$19
		dc.b	$17
		dc.b	5
		dc.b	$A
		dc.b	0
		dc.b	$14
		dc.b	$C
		dc.b	$B
		dc.b	$F
		dc.b	2
		dc.b	$15
		dc.b	$11
		dc.b	$E
		dc.b	$D
		dc.b	7
		dc.b	$A
		dc.b	$13
		dc.b	4
		dc.b	0
		dc.b	$17
		dc.b	6
		dc.b	$14
		dc.b	$18
		dc.b	8
		dc.b	$C
		dc.b	$11
		dc.b	1
		dc.b	$B
		dc.b	$15
		dc.b	2
		dc.b	$13
		dc.b	$14
		dc.b	6
		dc.b	4
		dc.b	0
		dc.b	$A
		dc.b	$19
		dc.b	$17
		dc.w	$18
a1993ElectronicArts: dc.b '$ 1993 Electronic Arts'
		dc.w	$E
aLicensedBy:	dc.b	'Licensed by',0
		dc.w	$18
aSegaEnterprisesLtd_: dc.b 'Sega Enterprises Ltd.',0
		dc.w	4					; The 00 04 FF 00 are likely to	signify	a new paragraph
		dc.b	$FF
		dc.b	0
		dc.w	$14
aDesignAdaptedBy: dc.b	'Design adapted by',0
		dc.w	$10
aMichaelBrook_0: dc.b	'Michael Brook',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$10
aProgrammedBy:	dc.b	'Programmed by',0
		dc.w	$E
aMarkLesser:	dc.b	'Mark Lesser',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$E
aGraphicsBy:	dc.b	'Graphics by',0
		dc.w	$C
aDougWike:	dc.b	'Doug Wike',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$16
aBasedOnNhlpa93By: dc.b	'Based on NHLPA 93 by'
		dc.w	$E
aJimSimmons:	dc.b	'Jim Simmons',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aMusicAndSoundBy: dc.b	'Music and Sound by'
		dc.w	$E
aRobHubbard:	dc.b	'Rob Hubbard',0
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.w	$10
aOrganMusicBy:	dc.b	'Organ Music by'
		dc.w	$10
aDieterRuehle:	dc.b	'Dieter Ruehle',0
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aExecutiveProducer: dc.b 'Executive Producer'
		dc.w	$C
aScottOrr:	dc.b	'Scott Orr',0
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.w	$E
aProducedBy:	dc.b	'Produced by',0
		dc.w	$10
aMichaelBrook:	dc.b	'Michael Brook',0
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aAssistantProducer: dc.b 'Assistant Producer'
		dc.w	$E
aKevinHogan:	dc.b	'Kevin Hogan',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$16
aTechnicalDirectors: dc.b 'Technical Directors',0
		dc.w	$C
aRobHarris:	dc.b	'Rob Harris'
		dc.w	$E
aLonMeinecke:	dc.b	'Lon Meinecke'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$C
aTestingBy:	dc.b	'Testing by'
		dc.w	$A
aYunShin:	dc.b	'Yun Shin'
		dc.w	$C
aKenRogers:	dc.b	'Ken Rogers'
		dc.w	$C
aCraigWike:	dc.b	'Craig Wike'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$C
aTestingBy_0:	dc.b	'Testing by'
		dc.w	$E
aJohnBoerio:	dc.b	'John Boerio',0
		dc.w	$C
aJedGarvey:	dc.b	'Jed Garvey'
		dc.w	$C
aGabeBoys:	dc.b	'Gabe Boys',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aPlayerRatingsBy: dc.b	'Player Ratings by',0
		dc.w	$10
aIgorKuperman:	dc.b	'Igor Kuperman',0
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$18
aPlayerCardPhotosBy: dc.b 'Player Card Photos by',0
		dc.w	$10
aSteveBabineau:	dc.b	'Steve Babineau'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aSpecialThanksTo: dc.b	'Special Thanks to',0
		dc.w	$C
aBobBorgen:	dc.b	'Bob Borgen'
		dc.w	$C
aDanBrook:	dc.b	'Dan Brook',0
		dc.w	$E
aJulieCressa:	dc.b	'Julie Cressa'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aSpecialThanksTo_0: dc.b 'Special Thanks to',0
		dc.w	$10
aRandyDelucchi:	dc.b	'Randy Delucchi'
		dc.w	$E
aJeffFennel:	dc.b	'Jeff Fennel',0
		dc.w	$C
aChipLange:	dc.b	'Chip Lange'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aSpecialThanksTo_1: dc.b 'Special Thanks to',0
		dc.w	$10
aBarryMelrose:	dc.b	'Barry Melrose',0
		dc.w	$E
aScottProbin:	dc.b	'Scott Probin'
		dc.w	$C
aIanPulver:	dc.b	'Ian Pulver'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$14
aSpecialThanksTo_2: dc.b 'Special Thanks to',0
		dc.w	$10
aMikeRubinelli:	dc.b	'Mike Rubinelli'
		dc.w	$C
aKyraWoody:	dc.b	'Kyra Woody'
		dc.w	4
		dc.b	$FF
		dc.b	0
		dc.w	$1C
aEaHockeyLeagueChampio: dc.b 'EA Hockey League Champion',0
		dc.w	$E
aKevinHogan_0:	dc.b	'Kevin Hogan',0
		dc.b	0					; Unknown data following...
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$92					; 
		dc.b	0
		dc.b	$B2					; 
		dc.b	0
		dc.b	$D2					; 
		dc.b	0
		dc.b	$F2					; 
		dc.b	0
		dc.b	1
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$1B
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$1C
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$1D
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$1B
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$1E
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$1F
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$20
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$1E
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$21					; !
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$22					; "
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$23					; #
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$21					; !
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$24					; $
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$25					; %
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$26					; &
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$24					; $
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$27					; '
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$28					; (
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$29					; )
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$27					; '
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$2A					; *
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2B					; +
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2A					; *
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$2D					; -
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2E					; .
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$2F					; /
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$2D					; -
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	$B4					; 
		dc.b	2
		dc.b	$30					; 0
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$31					; 1
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	2
		dc.b	$32					; 2
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$30					; 0
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$97					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$9A					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$9D					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$A0					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$A3					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$A6					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$A9					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$AC					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$98					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$9B					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$9E					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A1					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A4					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A7					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AA					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AD					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$99					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$9C					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$9F					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A5					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AE					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$17
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$18
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$17
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$9C					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$9F					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$A2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$19
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$1A
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$19
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$A8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$AE					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$D7					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$D9					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DB					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DD					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DF					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E1					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E3					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E5					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$D8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DA					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DC					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$DE					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E0					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E4					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$E6					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$BF					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C0					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C1					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C1					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C3					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C4					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C5					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C6					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C7					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C7					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$C8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$C9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$CA					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$CB					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$CC					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$CD					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$CD					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$CF					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D0					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$D1					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D2					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$D3					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D4					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$D3					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D4					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	1
		dc.b	$D5					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D6					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	7
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	8
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	7
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	9
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$A
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	9
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$B
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$B
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$D
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$E
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$D
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$F
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$10
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$F
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$11
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$12
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$11
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$13
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$14
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$13
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$15
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$16
		dc.b	0
		dc.b	$30					; 0
		dc.b	2
		dc.b	$15
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$AF					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$B0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$AF					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$B1					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$B2					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B1					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$B3					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$B4					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B3					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$B5					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$B6					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B5					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$B7					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$B8					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B7					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$B9					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$BA					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B9					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$BB					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$BC					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$BB					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	1
		dc.b	$BD					; 
		dc.b	0
		dc.b	5
		dc.b	1
		dc.b	$BE					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$BD					; 
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$E7					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$E8					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$E9					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$EA					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$EB					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$EC					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$ED					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$EE					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$EF					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F0					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F1					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F2					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$F3					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F4					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F5					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F6					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$F7					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F8					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$F9					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$FA					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$FB					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$FC					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$FD					; 
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	1
		dc.b	$FF
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	1
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	2
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	2
		dc.b	3
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	4
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	5
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	6
		dc.b	$FF
		dc.b	$F1					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$92					; 
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	$9A					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	1
		dc.b	1
		dc.b	$8C					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$90					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$91					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8B					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$93					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8B					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$92					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8F					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$94					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8F					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$92					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8B					; 
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	1
		dc.b	$8E					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8D					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8C					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8B					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$91					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8D					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$90					; 
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$8F					; 
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	1
		dc.b	$92					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8B					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$93					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8B					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$92					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$94					; 
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$8F					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$8B					; 
		dc.b	$F0					; 
		dc.b	0
		dc.b	1
		dc.b	$8B					; 
		dc.b	$F0					; 
		dc.b	0
		dc.b	1
		dc.b	$8F					; 
		dc.b	$F0					; 
		dc.b	0
		dc.b	1
		dc.b	$8F					; 
		dc.b	$F0					; 
		dc.b	0
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	6
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$B
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$10
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$15
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$1A
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$1F
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$24					; $
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	5
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	9
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$D
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$E
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$F
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$11
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$13
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$14
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$17
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$18
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$19
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$1B
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$1C
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$1D
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$1E
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$20
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$21					; !
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$23					; #
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$25					; %
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$27					; '
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$28					; (
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$2B					; +
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$2D					; -
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$2F					; /
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$30					; 0
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$31					; 1
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$32					; 2
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$34					; 4
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$35					; 5
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$37					; 7
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$39					; 9
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$3B					; ;
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$3C					; <
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$3F					; ?
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$40					; @
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$41					; A
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$43					; C
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$44					; D
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$45					; E
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$46					; F
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$48					; H
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$49					; I
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$4B					; K
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$4D					; M
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$4F					; O
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$50					; P
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$54					; T
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$56					; V
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$58					; X
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$59					; Y
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5A					; Z
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5B					; [
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5C					; \
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5D					; ]
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5E					; ^
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$5F					; _
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$60					; `
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$91					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$92					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$93					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$94					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$95					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$96					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$97					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$98					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$99					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$9A					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$9B					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$9C					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$9D					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$9E					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$9F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$A0					; 
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$65					; e
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$66					; f
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6B					; k
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6C					; l
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$71					; q
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$72					; r
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$77					; w
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$78					; x
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7D					; }
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$83					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$84					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$89					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8A					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$90					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$63					; c
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$62					; b
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$69					; i
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$68					; h
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6F					; o
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6E					; n
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$75					; u
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$74					; t
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7B					; {
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7A					; z
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$81					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$87					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$86					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8C					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$7E					; ~
		dc.b	0
		dc.b	$A2					; 
		dc.b	0
		dc.b	$C6					; 
		dc.b	0
		dc.b	$EA					; 
		dc.b	1
		dc.b	$E
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$63					; c
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$61					; a
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$63					; c
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$65					; e
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$66					; f
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$69					; i
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$68					; h
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$67					; g
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$68					; h
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$69					; i
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6B					; k
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6C					; l
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6F					; o
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6E					; n
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6D					; m
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6E					; n
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6F					; o
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$71					; q
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$72					; r
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$75					; u
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$74					; t
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$73					; s
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$74					; t
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$75					; u
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$77					; w
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$78					; x
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7B					; {
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7A					; z
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$79					; y
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7A					; z
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7B					; {
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7D					; }
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$81					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$80					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$80					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$81					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$83					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$84					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$87					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$86					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$85					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$86					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$87					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$89					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8A					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8C					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8B					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8C					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$90					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$7E					; ~
		dc.b	0
		dc.b	$A2					; 
		dc.b	0
		dc.b	$C6					; 
		dc.b	0
		dc.b	$EA					; 
		dc.b	1
		dc.b	$E
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$65					; e
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$65					; e
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$63					; c
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$62					; b
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6B					; k
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6C					; l
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6C					; l
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6C					; l
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6B					; k
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$69					; i
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$68					; h
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$71					; q
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$71					; q
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$70					; p
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6F					; o
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$6E					; n
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$77					; w
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$78					; x
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$78					; x
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$78					; x
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$77					; w
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$75					; u
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$74					; t
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7D					; }
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7E					; ~
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7E					; ~
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7E					; ~
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7D					; }
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7C					; |
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7B					; {
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$7A					; z
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$83					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$84					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$84					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$84					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$83					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$81					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$89					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$89					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$88					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$87					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$86					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$90					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$90					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$90					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8F					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8E					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$8C					; 
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$A1					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$A4					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$A7					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$AA					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$AD					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$B0					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$B3					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$B6					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$A1					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A2					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A1					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A3					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$A4					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A5					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A4					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A6					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$A7					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A8					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A7					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$A9					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$AA					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AB					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AA					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AC					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$AD					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AE					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AD					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$AF					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$B0					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B1					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B0					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B2					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$B3					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B4					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B3					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B5					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$B6					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B7					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B6					; 
		dc.b	0
		dc.b	$A
		dc.b	0
		dc.b	$B8					; 
		dc.b	$FF
		dc.b	$F6					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	1
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	1
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	3
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	4
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	3
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	5
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	6
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	5
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	7
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	8
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	7
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	9
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$A
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$B
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$C
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$B
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$D
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$E
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$D
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	1
		dc.b	$F
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$10
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$F
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$F1					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F2					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F3					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F4					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F5					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F6					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F7					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F8					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$F9					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FA					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FC					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FD					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	1
		dc.b	0
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$E1					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E2					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E3					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E4					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E5					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E6					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E7					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$E8					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$E9					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$EA					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$EB					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$EC					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$ED					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$EE					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$EF					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$F0					; 
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$29					; )
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$2E					; .
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$33					; 3
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$3D					; =
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$42					; B
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$47					; G
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$4C					; L
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$59					; Y
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5A					; Z
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5B					; [
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5C					; \
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5D					; ]
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$5F					; _
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	1
		dc.b	$60					; `
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$59					; Y
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5A					; Z
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5B					; [
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5C					; \
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5D					; ]
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$5F					; _
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	1
		dc.b	$60					; `
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$29					; )
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$2E					; .
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$33					; 3
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$3D					; =
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$42					; B
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$47					; G
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$4C					; L
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$31					; 1
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$32					; 2
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$33					; 3
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$34					; 4
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$35					; 5
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$36					; 6
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$37					; 7
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$38					; 8
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$52					; R
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$39					; 9
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$3A					; :
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$3B					; ;
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$3C					; <
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$3D					; =
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$3E					; >
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$3F					; ?
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$40					; @
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$54					; T
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$41					; A
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$42					; B
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$43					; C
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$44					; D
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$45					; E
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$46					; F
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$47					; G
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$48					; H
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$56					; V
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$49					; I
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$4A					; J
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$4B					; K
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$4C					; L
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$4D					; M
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$4E					; N
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$4F					; O
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$50					; P
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$58					; X
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$12
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$13
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$15
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$16
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$17
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$18
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$52					; R
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$19
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1A
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1B
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$1D
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1E
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1F
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$20
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$54					; T
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$21					; !
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$22					; "
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$23					; #
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$24					; $
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$25					; %
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$26					; &
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$27					; '
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$28					; (
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$56					; V
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$29					; )
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2A					; *
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2B					; +
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$2D					; -
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2E					; .
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2F					; /
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$30					; 0
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$58					; X
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$B9					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$BA					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$B9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$BB					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$BC					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$BB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$BD					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$BE					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$BD					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$BF					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$C0					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$BF					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$C1					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$C2					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$C1					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$C3					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$C4					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$C3					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$C5					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$C6					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$C7					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$C8					; 
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	$C7					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$C9					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CA					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$CB					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CC					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CB					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$CD					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CE					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CD					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$CF					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D0					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$CF					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$D1					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D2					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D1					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$D3					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D4					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D3					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$D5					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D6					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D5					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$D7					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D8					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$D7					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$7D					; }
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7E					; ~
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7F					; 
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$79					; y
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7A					; z
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7B					; {
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7C					; |
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$7C					; |
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7B					; {
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7A					; z
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$79					; y
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$7F					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7E					; ~
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7D					; }
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$7C					; |
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7B					; {
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7A					; z
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$79					; y
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	1
		dc.b	$79					; y
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7A					; z
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7B					; {
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$7C					; |
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	1
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	0
		dc.b	$15
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	$A
		dc.b	0
		dc.b	$A
		dc.b	1
		dc.b	9
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$15
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	$33					; 3
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$34					; 4
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$35					; 5
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$36					; 6
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$37					; 7
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$38					; 8
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$39					; 9
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3A					; :
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3B					; ;
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3C					; <
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3D					; =
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3E					; >
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$3F					; ?
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$FD					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$8D					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8E					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8F					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$90					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$91					; 
		dc.b	$FC					; 
		dc.b	$18
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$D9					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DA					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DB					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DC					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DD					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DE					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$DF					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$E0					; 
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$41					; A
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$42					; B
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$43					; C
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$44					; D
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$45					; E
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$46					; F
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$47					; G
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$48					; H
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$49					; I
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$4A					; J
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$4B					; K
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$4C					; L
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$4D					; M
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$4E					; N
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$4F					; O
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$50					; P
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$59					; Y
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$5B					; [
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$5D					; ]
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	2
		dc.b	$5F					; _
		dc.b	$FF
		dc.b	$E2					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$51					; Q
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$52					; R
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$51					; Q
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$52					; R
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$53					; S
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$54					; T
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$53					; S
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$54					; T
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$55					; U
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$56					; V
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$55					; U
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$56					; V
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$57					; W
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$58					; X
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$57					; W
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$58					; X
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$59					; Y
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5A					; Z
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$59					; Y
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5A					; Z
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$5B					; [
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5C					; \
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5B					; [
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5C					; \
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$5D					; ]
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5E					; ^
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5D					; ]
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5E					; ^
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	2
		dc.b	$5F					; _
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$60					; `
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$5F					; _
		dc.b	0
		dc.b	9
		dc.b	2
		dc.b	$60					; `
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$4E					; N
		dc.b	0
		dc.b	$62					; b
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$61					; a
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$62					; b
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$63					; c
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$62					; b
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$61					; a
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$64					; d
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$65					; e
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$66					; f
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$65					; e
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$64					; d
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$67					; g
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$68					; h
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$69					; i
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$68					; h
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$67					; g
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$6A					; j
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6B					; k
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6C					; l
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$6B					; k
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6A					; j
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$6D					; m
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6E					; n
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6F					; o
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$6E					; n
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$6D					; m
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$70					; p
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$71					; q
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$72					; r
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$71					; q
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$70					; p
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$73					; s
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$74					; t
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$75					; u
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$74					; t
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$73					; s
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	2
		dc.b	$76					; v
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$77					; w
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$78					; x
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$77					; w
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$76					; v
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	$A2					; 
		dc.b	0
		dc.b	$BA					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$51					; Q
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$58					; X
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$D
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$60					; `
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$59					; Y
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5A					; Z
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$51					; Q
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$F
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$59					; Y
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5B					; [
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$53					; S
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$52					; R
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5B					; [
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5C					; \
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$54					; T
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$53					; S
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	3
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5B					; [
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5C					; \
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5D					; ]
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$55					; U
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$54					; T
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5C					; \
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5D					; ]
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5E					; ^
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$56					; V
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$55					; U
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	7
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5D					; ]
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5E					; ^
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5F					; _
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$57					; W
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$56					; V
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	9
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5E					; ^
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$5F					; _
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$60					; `
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$58					; X
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$57					; W
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	$B
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$5F					; _
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$60					; `
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$59					; Y
		dc.b	$FF
		dc.b	$FC					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	$66					; f
		dc.b	0
		dc.b	$82					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	$BA					; 
		dc.b	0
		dc.b	$D6					; 
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$79					; y
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7A					; z
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7B					; {
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7C					; |
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$13
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$79					; y
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7A					; z
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7B					; {
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7C					; |
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$17
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$18
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$52					; R
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$79					; y
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7A					; z
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7B					; {
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7C					; |
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1B
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$7D					; }
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7E					; ~
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7F					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$80					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1F
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$20
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$54					; T
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$7D					; }
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7E					; ~
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7F					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$80					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$23					; #
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$24					; $
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$7D					; }
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7E					; ~
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7F					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$80					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$27					; '
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$28					; (
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$56					; V
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$7D					; }
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7E					; ~
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7F					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$80					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2B					; +
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$79					; y
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7A					; z
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7B					; {
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$7C					; |
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2F					; /
		dc.b	0
		dc.b	$64					; d
		dc.b	1
		dc.b	$30					; 0
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$58					; X
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$12
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$13
		dc.b	0
		dc.b	$1E
		dc.b	1
		dc.b	$13
		dc.b	$FC					; 
		dc.b	0
		dc.b	1
		dc.b	$21					; !
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$22					; "
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$23					; #
		dc.b	0
		dc.b	$1E
		dc.b	1
		dc.b	$23					; #
		dc.b	$FC					; 
		dc.b	0
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$92					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$94					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$96					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$98					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9A					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9C					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9E					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$A0					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$93					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$95					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$97					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$99					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9B					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9D					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$9F					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$A1					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$A2					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$A3					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$A6					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$A7					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$AA					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$AE					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$AF					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$B2					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$B3					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$B6					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$B7					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$BA					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$BB					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$BE					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$BF					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$32					; 2
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$4A					; J
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$A4					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$A5					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$A8					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$A9					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$AC					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$AD					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$B0					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$B1					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$B4					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$B5					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$B8					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$B9					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$BC					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$BD					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	2
		dc.b	$C0					; 
		dc.b	0
		dc.b	4
		dc.b	2
		dc.b	$C1					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$56					; V
		dc.b	0
		dc.b	$76					; v
		dc.b	0
		dc.b	$9A					; 
		dc.b	0
		dc.b	$BA					; 
		dc.b	0
		dc.b	$D6					; 
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$AF					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C2					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C3					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C4					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C5					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$56					; V
		dc.b	0
		dc.b	$72					; r
		dc.b	0
		dc.b	$8A					; 
		dc.b	0
		dc.b	$9E					; 
		dc.b	0
		dc.b	$B6					; 
		dc.b	0
		dc.b	$D2					; 
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$97					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$9A					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$9D					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A0					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A3					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A6					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	1
		dc.b	$AC					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$A9					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$C6					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C7					; 
		dc.b	0
		dc.b	5
		dc.b	2
		dc.b	$C8					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$C9					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$E2					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E3					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E4					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E5					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$E6					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E7					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E8					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$E9					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$3C					; <
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$EA					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$EB					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$EC					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$ED					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$44					; D
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$EE					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$EF					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F0					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F1					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$4C					; L
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$F2					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F3					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F4					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F5					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$34					; 4
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$F6					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F7					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F8					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F9					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$FA					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$FB					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$FC					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$FD					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$44					; D
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$FE					; 
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$FF
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	0
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	2
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	3
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	4
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	5
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$34					; 4
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	6
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	7
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	8
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	9
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$3C					; <
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$A
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$B
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$C
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$D
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$24					; $
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$E
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$F
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$10
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$11
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$4C					; L
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$42					; B
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	$12
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$13
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$14
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$15
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$34					; 4
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$51					; Q
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$16
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$17
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$18
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$19
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$1A
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$1B
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$1C
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$1D
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$44					; D
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$55					; U
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$1E
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$1F
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$20
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$21					; !
		dc.b	0
		dc.b	$3C					; <
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	8
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	$22					; "
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$23					; #
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$24					; $
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$25					; %
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$26					; &
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$27					; '
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$28					; (
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$20
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$16
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$29					; )
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2A					; *
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2B					; +
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2C					; ,
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2D					; -
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$29					; )
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$18
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F9					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$2E					; .
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2F					; /
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$24					; $
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$25					; %
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$26					; &
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$27					; '
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$28					; (
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$20
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$1E
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$2F					; /
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$24					; $
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$25					; %
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$26					; &
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$27					; '
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$28					; (
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$20
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$3E					; >
		dc.b	0
		dc.b	$6A					; j
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	$96					; 
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	$30					; 0
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$31					; 1
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$32					; 2
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$33					; 3
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$34					; 4
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$35					; 5
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$36					; 6
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$18
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F9					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$F6					; 
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$37					; 7
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$32					; 2
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$33					; 3
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$34					; 4
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$35					; 5
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$36					; 6
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$18
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F9					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	3
		dc.b	$38					; 8
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$37					; 7
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$32					; 2
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$33					; 3
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$34					; 4
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$35					; 5
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$36					; 6
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$18
		dc.b	0
		dc.b	6
		dc.b	2
		dc.b	$F9					; 
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$1C
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$53					; S
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	2
		dc.b	$FE					; 
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$39					; 9
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$3A					; :
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$3B					; ;
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$3C					; <
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$3D					; =
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$39					; 9
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	$20
		dc.b	0
		dc.b	6
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$2C					; ,
		dc.b	0
		dc.b	$50					; P
		dc.b	1
		dc.b	$57					; W
		dc.b	$FF
		dc.b	$F8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$16
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$1E
		dc.b	0
		dc.b	$22					; "
		dc.b	0
		dc.b	$26					; &
		dc.b	0
		dc.b	$2A					; *
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	$3E					; >
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$3F					; ?
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$41					; A
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$42					; B
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$43					; C
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$44					; D
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	3
		dc.b	$45					; E
		dc.b	$FF
		dc.b	$E8					; 
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$5A					; Z
		dc.b	0
		dc.b	$12
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$11
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$12
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$13
		dc.b	0
		dc.b	$1E
		dc.b	2
		dc.b	$81					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$82					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$83					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$84					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$85					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$86					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$84					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$85					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$86					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$84					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$85					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$86					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$84					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$85					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$86					; 
		dc.b	$FC					; 
		dc.b	$7C					; |
		dc.b	1
		dc.b	$21					; !
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$22					; "
		dc.b	0
		dc.b	6
		dc.b	1
		dc.b	$23					; #
		dc.b	0
		dc.b	$1E
		dc.b	2
		dc.b	$87					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$88					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$89					; 
		dc.b	0
		dc.b	$A
		dc.b	2
		dc.b	$8A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8B					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8C					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8B					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8C					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8B					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8C					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8A					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8B					; 
		dc.b	0
		dc.b	8
		dc.b	2
		dc.b	$8C					; 
		dc.b	$FC					; 
		dc.b	$7C					; |
; ---------------------------------------------------------------------------

VBlankInt:							; DATA XREF: ROM:00000078o
		move.l	($FFFFB03A).w,-(sp)			; copy V-blank routine to the stack
		rts						; return (will jump to the V-blank routine in question)
; ---------------------------------------------------------------------------

RunGame:
		move	#$2700,sr				; disable interrupts
		move.w	#$FFFE,sp				; set stack pointer address to FFFFFFFE (RAM FFFE)
		move.w	#$B000,a0

loc_76C4:
		clr.l	(a0)+					; Clear	RAM B000 - DEF4
		cmp.w	#$DEF4,a0
		blt.s	loc_76C4
		clr.w	($FFFFDEF2).w				; Clear	DMA flag
		move.w	($C00004).l,d0
		and.w	#1,d0					; DMA busy bit
		move.w	d0,($FFFFD06E).w
		beq.w	loc_76E8				; if the VDP is	not busy with DMA fill/copy, branch
		bset	#0,($FFFFDEF2).w			; set DMA flag

loc_76E8:							; CODE XREF: ROM:000076DEj
		jsr	Z80_LoadROM
		jsr	Z80_MuteAllFM
		jsr	sub_1A50A
		jsr	PSG_SetupNoise
		jsr	Unk_ControlsSetRelated
		move.w	#$FFFF,($FFFFBE86).w
		jsr	sub_17A18
		jsr	sub_1A050
		jsr	sub_FE660
		jsr	sub_FED70
		jsr	sub_17C72
		move.w	($FFFFD058).w,($FFFFD072).w
		move.w	($FFFFD048).w,($FFFFD074).w
		jsr	sub_112BC
		jmp	loc_172F0
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_7CB0

loc_7742:							; CODE XREF: sub_7CB0:loc_17332j
		st	($FFFFD6B4).w
		clr.l	($FFFFDEA4).w
		clr.l	($FFFFDEAC).w
		clr.l	($FFFFDEA8).w
		clr.l	($FFFFDEB0).w
		jsr	sub_11340
		cmpnb	 $E0,d3
		bne.w	loc_776A
		move.w	#3,($FFFFD050).w

loc_776A:							; CODE XREF: sub_7CB0-550j
		clr.b	($FFFFC2EA).w
		cmp.w	#1,($FFFFD056).w
		bne.w	loc_777E
		bset	#5,($FFFFC2EA).w

loc_777E:							; CODE XREF: sub_7CB0-53Cj
		cmp.w	#1,($FFFFD048).w
		ble.w	loc_778C
		bsr.w	sub_77E4

loc_778C:							; CODE XREF: sub_7CB0-52Cj
		jsr	sub_17102
		btst	#0,($FFFFC2FA).w
		beq.w	loc_77B8
		move.l	a0,-(sp)
		move.l	#$FFFFC6CE,a0
		jsr	sub_F70A2
		move.l	#$FFFFCA32,a0
		jsr	sub_F70A2
		move.l	(sp)+,a0

loc_77B8:							; CODE XREF: sub_7CB0-518j
		clr.w	($FFFFC472).w
		clr.w	($FFFFC5DC).w
		clr.w	($FFFFC466).w
		clr.w	($FFFFC46E).w
		bsr.w	sub_77F4
		jsr	sub_131F4
		jsr	loc_169FA
		jsr	sub_F6E8A
		jmp	loc_17278
; END OF FUNCTION CHUNK	FOR sub_7CB0

; =============== S U B	R O U T	I N E =======================================


sub_77E4:							; CODE XREF: sub_7CB0-528p
		move.w	#$E3,d0					; ''
		move.w	#$D092,a0

loc_77EC:							; CODE XREF: sub_77E4+Aj
		clr.w	(a0)+
		dbf	d0,loc_77EC
		rts
; End of function sub_77E4


; =============== S U B	R O U T	I N E =======================================


sub_77F4:							; CODE XREF: sub_7CB0-4E8p
					; ROM:000135D8p
		move.w	#$C6CE,a2
		bsr.w	sub_7800
		add.w	#$364,a2
; End of function sub_77F4


; =============== S U B	R O U T	I N E =======================================


sub_7800:							; CODE XREF: sub_77F4+4p
		move.w	#6,$24(a2)
		moveq	#$32,d0					; '2'

loc_7808:							; CODE XREF: sub_7800+10j
		move.w	#$FFFE,$66(a2,d0.w)
		subq.w	#2,d0
		bpl.s	loc_7808
		rts
; End of function sub_7800


; =============== S U B	R O U T	I N E =======================================


sub_7814:							; CODE XREF: sub_7CB0-426p
					; sub_7CB0:loc_17278p
		bsr.w	sub_784E
		cmp.w	#3,($FFFFC466).w
		blt.w	loc_782E
		tst.w	($FFFFD048).w
		bne.w	loc_782E
		move.w	#$258,d0

loc_782E:							; CODE XREF: sub_7814+Aj sub_7814+12j
		move.w	d0,($FFFFC468).w
		move.w	d0,($FFFFC46C).w
		move.w	d0,($FFFFB048).w
		asr.w	#1,d0
		jsr	sub_11086
		sub.w	d0,($FFFFB048).w
		bset	#0,($FFFFC2EA).w
		rts
; End of function sub_7814


; =============== S U B	R O U T	I N E =======================================


sub_784E:							; CODE XREF: sub_7814p	sub_15AA4+C2p ...
		move.w	($FFFFD050).w,d0
		asl.w	#1,d0
		lea	word_785E(pc),a0
		move.w	(a0,d0.w),d0
		rts
; End of function sub_784E

; ---------------------------------------------------------------------------
word_785E:	dc.w	$12C					; DATA XREF: sub_784E+6o
		dc.w	$258
		dc.w	$4B0
		dc.w	$1E
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_7CB0

loc_7866:							; CODE XREF: sub_7CB0+F614j
		cmp.w	#3,($FFFFC466).w
		bne.w	loc_7876
		bset	#1,($FFFFC2FC).w

loc_7876:							; CODE XREF: sub_7CB0-444j
		st	($FFFFD6BE).w
		move.w	#$FFFE,sp
		jsr	Z80_MuteAllFM
		jsr	loc_169FA
		bsr.s	sub_7814
		or.w	#$F000,($FFFFBE78).w
		st	($FFFFC320).w
		st	($FFFFC322).w
		move.w	#$B74A,a3
		clr.w	($FFFFBF94).w
		clr.w	($FFFFBF96).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_78C4
		move.w	#$1E,d0
		move.w	#8,($FFFFD40A).w
		move.w	#$B,($FFFFD408).w
		bra.w	loc_78CA
; ---------------------------------------------------------------------------

loc_78C4:							; CODE XREF: sub_7CB0-404j
		move.l	#$1B,d0

loc_78CA:							; CODE XREF: sub_7CB0-3F0j
		jsr	sub_10662
		bset	#2,($FFFFC2EE).w
		bclr	#4,($FFFFC2EC).w
		move.w	#$FFFF,($FFFFC016).w
		move.l	#$FFFF0000,($FFFFB036).w
		move.w	($FFFFB03E).w,($FFFFB040).w
		bsr.w	sub_794E
		move.w	#$FFFF,($FFFFBD26).w
		bsr.w	sub_794E
		move.w	($FFFFCEEC).w,d0
		asl.w	#4,d0
		move.w	d0,($FFFFB8A2).w
		clr.w	($FFFFC30E).w
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#0,($FFFFD6CC).w
		jsr	sub_FE556
		move.w	($FFFFD6C8).w,-(sp)
		jsr	sub_11156
		cmp.w	#2,($FFFFC466).w
		bge.w	loc_7938
		bset	#7,($FFFFC2F0).w

loc_7938:							; CODE XREF: sub_7CB0-382j
					; sub_7CB0-36Aj ...
		bsr.w	sub_794E
		bsr.w	sub_7C58
		btst	#0,($FFFFC2EC).w
		beq.s	loc_7938
		bsr.w	sub_7CF8
		bra.s	loc_7938
; END OF FUNCTION CHUNK	FOR sub_7CB0

; =============== S U B	R O U T	I N E =======================================


sub_794E:							; CODE XREF: sub_7CB0-3C0p
					; sub_7CB0-3B6p ...
		move.w	($FFFFB03E).w,d7
		sub.w	($FFFFB040).w,d7
		beq.s	sub_794E
		move.w	($FFFFB03E).w,($FFFFB040).w
		bsr.w	sub_799E
		bsr.w	sub_A9D6
		tst.w	($FFFFDECC).w
		beq.w	loc_798A
		bmi.w	loc_798A
		subq.w	#1,($FFFFDECC).w
		bne.w	loc_798A
		jsr	sub_1A304
		move.w	($FFFFDECE).w,-(sp)
		jsr	sub_11156

loc_798A:							; CODE XREF: sub_794E+1Cj sub_794E+20j ...
		jsr	sub_F8BA0
		bsr.w	sub_AFCA
		bsr.w	sub_A8CA
		jmp	sub_15EC0
; End of function sub_794E


; =============== S U B	R O U T	I N E =======================================


sub_799E:							; CODE XREF: sub_794E+10p

; FUNCTION CHUNK AT 00012A7C SIZE 000000B6 BYTES

		jsr	sub_11FF2
		bsr.w	sub_7A9E
		jsr	sub_FE2C8
		jsr	sub_16976
		bsr.w	sub_7B30
		btst	#7,($FFFFC2EC).w
		bne.w	locret_A9D4
		sub.w	d7,($FFFFB046).w
		bpl.w	locret_A9D4
		add.w	#$18,($FFFFB046).w
		bsr.w	sub_F6AA
		bsr.w	sub_7A76
		bsr.w	sub_7A34
		bsr.w	sub_79F8
		bsr.w	sub_79EA
		jmp	loc_12A7C
; End of function sub_799E


; =============== S U B	R O U T	I N E =======================================


sub_79EA:							; CODE XREF: sub_799E+42p

; FUNCTION CHUNK AT 0001871C SIZE 00000033 BYTES
; FUNCTION CHUNK AT 00018768 SIZE 00000013 BYTES

		subq.w	#1,($FFFFC3EC).w
		bne.w	locret_A9D4
		jmp	loc_1871C
; End of function sub_79EA


; =============== S U B	R O U T	I N E =======================================


sub_79F8:							; CODE XREF: sub_799E+3Ep
		tst.w	($FFFFD058).w
		bne.w	locret_7A32
		move.w	#$C6CE,a2
		bsr.w	sub_7A0C
		lea	$364(a2),a2
; End of function sub_79F8


; =============== S U B	R O U T	I N E =======================================


sub_7A0C:							; CODE XREF: sub_79F8+Cp
		moveq	#$32,d0					; '2'

loc_7A0E:							; CODE XREF: sub_7A0C+24j
		cmp.w	#$FFFE,$66(a2,d0.w)
		bne.w	loc_7A2E
		add.w	#9,$32(a2,d0.w)
		cmp.w	#$1000,$32(a2,d0.w)
		blt.w	loc_7A2E
		move.w	#$1000,$32(a2,d0.w)

loc_7A2E:							; CODE XREF: sub_7A0C+8j sub_7A0C+18j
		subq.w	#2,d0
		bpl.s	loc_7A0E

locret_7A32:							; CODE XREF: sub_79F8+4j
		rts
; End of function sub_7A0C


; =============== S U B	R O U T	I N E =======================================


sub_7A34:							; CODE XREF: sub_799E+3Ap
		cmp.w	#2,($FFFFC466).w
		bne.w	locret_7A74
		btst	#4,($FFFFC2EA).w
		bne.w	locret_7A74
		move.w	($FFFFC468).w,d0
		cmp.w	($FFFFB048).w,d0
		bgt.w	locret_7A74
		st	($FFFFB048).w
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#5,($FFFFD6CC).w
		jsr	sub_FE556
		move.w	($FFFFD6C8).w,-(sp)
		jsr	sub_11156

locret_7A74:							; CODE XREF: sub_7A34+6j sub_7A34+10j	...
		rts
; End of function sub_7A34


; =============== S U B	R O U T	I N E =======================================


sub_7A76:							; CODE XREF: sub_799E+36p
		move.w	($FFFFB8A2).w,d0
		cmp.w	($FFFFB8A4).w,d0
		bls.w	loc_7A86
		move.w	d0,($FFFFB8A4).w

loc_7A86:							; CODE XREF: sub_7A76+8j
		ext.l	d0
		add.l	d0,($FFFFB8A8).w
		addq.w	#1,($FFFFB8A6).w
		subq.w	#1,($FFFFB8A2).w
		bpl.w	locret_7A9C
		clr.w	($FFFFB8A2).w

locret_7A9C:							; CODE XREF: sub_7A76+1Ej
		rts
; End of function sub_7A76


; =============== S U B	R O U T	I N E =======================================


sub_7A9E:							; CODE XREF: sub_799E+6p sub_111D0+1Ap
		cmp.w	#$15E,($FFFFB89C).w
		blt.w	loc_7AAC
		subq.w	#3,($FFFFB89C).w

loc_7AAC:							; CODE XREF: sub_7A9E+6j
		sub.w	d7,($FFFFB89C).w
		bpl.w	loc_7AB8
		clr.w	($FFFFB89C).w

loc_7AB8:							; CODE XREF: sub_7A9E+12j
		sub.w	d7,($FFFFB8A0).w
		bpl.w	loc_7B02
		move.w	($FFFFB89C).w,d0
		lsr.w	#1,d0
		cmpnw	 $7F,d0					; ''
		bls.w	loc_7AD0
		moveq	#$7F,d0					; ''

loc_7AD0:							; CODE XREF: sub_7A9E+2Cj
		and.w	#$60,d0					; '`'
		addq.w	#2,($FFFFB89E).w
		and.w	#$1E,($FFFFB89E).w
		add.w	($FFFFB89E).w,d0
		move.l	#unk_18CFC,a0
		move.b	(a0,d0.w),($FFFFB89B).w
		clr.w	d1
		move.b	1(a0,d0.w),d1
		move.w	($C00008).l,d2
		and.w	d1,d2
		add.w	d2,d1
		move.w	d1,($FFFFB8A0).w

loc_7B02:							; CODE XREF: sub_7A9E+1Ej
		clr.b	($FFFFB89A).w
		cmp.w	#$118,($FFFFB89C).w
		bls.w	locret_A9D4
		move.w	($C00008).l,d0
		and.w	#$7F,d0					; ''
		cmpnw	 $13,d0
		blt.w	locret_A9D4
		cmpnw	 $19,d0
		bgt.w	locret_A9D4
		move.b	d0,($FFFFB89A).w
		rts
; End of function sub_7A9E


; =============== S U B	R O U T	I N E =======================================


sub_7B30:							; CODE XREF: sub_799E+16p
		btst	#0,($FFFFC2EA).w
		bne.w	locret_A9D4
		tst.w	($FFFFC468).w
		bne.w	locret_A9D4
		bset	#3,($FFFFBF78).w
		jsr	sub_1A304
		move.w	#4,-(sp)
		jsr	sub_11132
		bsr.w	sub_AFB6

loc_7B5C:							; CODE XREF: sub_F3E4+B2j
		move.w	#$B74A,a3
		move.l	#$18,d0
		jsr	sub_10658
		cmp.w	#2,($FFFFC466).w
		blt.w	loc_7C3E
		move.l	#7,d0
		move.w	#$B04A,a3
		cmp.w	#3,($FFFFCEEC).w
		bne.w	loc_7BE4
		cmp.w	#7,($FFFFCEEA).w
		beq.w	loc_7BCA
		moveq	#$10,d3
		mulu.w	($FFFFCEE6).w,d3
		move.w	#$CE66,a0
		add.w	d3,a0
		clr.w	d3
		btst	#0,$E(a0)
		beq.w	loc_7BB0
		eor.w	#2,d3

loc_7BB0:							; CODE XREF: sub_7B30+78j
		move.w	($FFFFC6DA).w,d1
		sub.w	($FFFFCA3E).w,d1
		bpl.w	loc_7BC0
		eor.w	#2,d3

loc_7BC0:							; CODE XREF: sub_7B30+88j
		cmp.w	#3,4(a0,d3.w)
		bne.w	loc_7BE4

loc_7BCA:							; CODE XREF: sub_7B30+60j
		move.l	#8,d0
		moveq	#$B,d2

loc_7BD2:							; CODE XREF: sub_7B30+ACj
		bclr	#3,$62(a3)
		add.w	#$80,a3					; ''
		dbf	d2,loc_7BD2
		move.w	#$B04A,a3

loc_7BE4:							; CODE XREF: sub_7B30+56j sub_7B30+96j
		moveq	#5,d2
		move.w	($FFFFC6DA).w,d1
		sub.w	($FFFFCA3E).w,d1
		beq.w	loc_7C3E
		bpl.w	loc_7BFA
		add.w	#$300,a3

loc_7BFA:							; CODE XREF: sub_7B30+C2j sub_7B30+E2j
		tst.w	$34(a3)
		ble.w	loc_7C0E
		jsr	sub_10658
		move.l	#7,d0

loc_7C0E:							; CODE XREF: sub_7B30+CEj
		add.w	#$80,a3					; ''
		dbf	d2,loc_7BFA

loc_7C16:							; CODE XREF: sub_7B30+11Cj
		jsr	sub_128A4
		add.w	#$3E8,($FFFFB89C).w
		bset	#0,($FFFFC2EA).w
		bset	#6,($FFFFC2EA).w
		jsr	sub_FF88E
		move.w	#4,d0
		jmp	sub_11F62
; ---------------------------------------------------------------------------

loc_7C3E:							; CODE XREF: sub_7B30+42j sub_7B30+BEj
		cmp.w	#3,($FFFFC466).w
		bne.w	loc_7C4E
		tst.w	($FFFFD048).w
		beq.s	loc_7C16

loc_7C4E:							; CODE XREF: sub_7B30+114j
		move.w	#2,d0
		jmp	sub_11F62
; End of function sub_7B30


; =============== S U B	R O U T	I N E =======================================


sub_7C58:							; CODE XREF: sub_7CB0-374p

; FUNCTION CHUNK AT 00007CBC SIZE 0000003C BYTES

		tst.w	($FFFFC328).w
		bne.w	locret_A9D4
		tst.w	($FFFFC32A).w
		bne.w	locret_A9D4
		jsr	sub_11340
		btst	#7,d1
		bne.w	loc_7CBC
		bsr.w	sub_7CB0
		jsr	sub_11358
		btst	#7,d1
		bne.w	loc_7CCA
		tst.w	($FFFFD046).w
		beq.w	sub_7CB0
		jsr	sub_11370
		btst	#7,d1
		bne.w	loc_7CDC
		bsr.w	sub_7CB0
		jsr	sub_11388
		btst	#7,d1
		bne.w	loc_7CEA
; End of function sub_7C58


; =============== S U B	R O U T	I N E =======================================


sub_7CB0:							; CODE XREF: sub_7C58+1Ep sub_7C58+34j ...

; FUNCTION CHUNK AT 00007742 SIZE 000000A2 BYTES
; FUNCTION CHUNK AT 00007866 SIZE 000000E8 BYTES
; FUNCTION CHUNK AT 00017278 SIZE 00000078 BYTES
; FUNCTION CHUNK AT 000172FA SIZE 0000003C BYTES

		tst.w	d1
		beq.w	locret_A9D4
		jmp	loc_172E4
; End of function sub_7CB0

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_7C58

loc_7CBC:							; CODE XREF: sub_7C58+1Aj sub_B0E8+90j
		clr.w	($FFFFC316).w
		bclr	#1,($FFFFC2EC).w
		bra.w	loc_7CD4
; ---------------------------------------------------------------------------

loc_7CCA:							; CODE XREF: sub_7C58+2Cj sub_B0E8+94j
		clr.w	($FFFFC316).w
		bset	#1,($FFFFC2EC).w

loc_7CD4:							; CODE XREF: sub_7C58+6Ej sub_7C58+90j ...
		bset	#0,($FFFFC2EC).w
		rts
; ---------------------------------------------------------------------------

loc_7CDC:							; CODE XREF: sub_7C58+42j sub_B0E8+A4j
		move.w	#3,($FFFFC316).w
		bclr	#1,($FFFFC2EC).w
		bra.s	loc_7CD4
; ---------------------------------------------------------------------------

loc_7CEA:							; CODE XREF: sub_7C58+54j sub_B0E8+A8j
		move.w	#4,($FFFFC316).w
		bset	#1,($FFFFC2EC).w
		bra.s	loc_7CD4
; END OF FUNCTION CHUNK	FOR sub_7C58

; =============== S U B	R O U T	I N E =======================================


sub_7CF8:							; CODE XREF: sub_7CB0-368p
		jsr	sub_10F32
		move.w	d0,-(sp)
		move.w	($FFFFB03E).w,d0

loc_7D04:							; CODE XREF: sub_7CF8+10j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_7D04
		move.w	(sp)+,d0
		jsr	Z80_MuteAllFM
		move.w	($FFFFC2EC).w,-(sp)
		bsr.w	sub_7E0E
		move.l	#off_19700,a0
		lea	loc_7DCE(pc),a1
		btst	#0,($FFFFC2FA).w
		beq.w	loc_7D38
		move.l	#word_19664,a0
		bra.w	loc_7D48
; ---------------------------------------------------------------------------

loc_7D38:							; CODE XREF: sub_7CF8+32j
		btst	#2,$30(a2)
		beq.w	loc_7D48
		move.l	#off_1988C,a0

loc_7D48:							; CODE XREF: sub_7CF8+3Cj sub_7CF8+46j
		bsr.w	sub_7E36

loc_7D4C:							; CODE XREF: sub_7CF8+6Cj
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		jsr	sub_162FE
		jsr	sub_11318
		bsr.w	loc_7E88
		bne.s	loc_7D4C
		jsr	sub_10F32
		move.w	(sp)+,($FFFFC2EC).w
		btst	#7,($FFFFC2EC).w
		beq.w	loc_7D8A
		jsr	sub_16CE0
		jsr	sub_12B94
		bra.w	loc_7D90
; ---------------------------------------------------------------------------

loc_7D8A:							; CODE XREF: sub_7CF8+7Ej
		jsr	sub_12B40

loc_7D90:							; CODE XREF: sub_7CF8+8Ej
		move.l	#$C00000,a0
		move.w	#$9100,4(a0)
		move.w	#$9200,4(a0)
		bset	#3,($FFFFBF78).w
		bclr	#0,($FFFFC2EC).w
		jsr	loc_12C04
		jsr	sub_15EC0
		move.w	#$18,($FFFFBD26).w

loc_7DC0:							; CODE XREF: sub_7CF8+CCj
		tst.w	($FFFFBD26).w
		bpl.s	loc_7DC0
		move.w	($FFFFB03E).w,($FFFFB040).w
		rts
; End of function sub_7CF8

; ---------------------------------------------------------------------------

loc_7DCE:							; CODE XREF: ROM:00013780p
					; DATA XREF: sub_7CF8+28o
		move.l	#$C00000,a0
		move.w	#$9100,4(a0)
		move.w	#$921C,4(a0)
		jsr	sub_12B94
		jsr	sub_15EC0
		jsr	PSG_SetupNoise
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FF
		dc.b	3
		dc.b	$FD
		dc.b	0
		dc.b	$FC
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#$1C,d1
		move.w	#$7FF,d2
		jmp	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_7E0E:							; CODE XREF: sub_7CF8+1Ep
					; ROM:00007EB8p
		move.w	#$C6CE,a2
		btst	#1,($FFFFC2EC).w
		beq.w	loc_7E26
		cmp.w	#1,($FFFFC32A).w
		bra.w	loc_7E2C
; ---------------------------------------------------------------------------

loc_7E26:							; CODE XREF: sub_7E0E+Aj
		cmp.w	#1,($FFFFC328).w

loc_7E2C:							; CODE XREF: sub_7E0E+14j
		beq.w	locret_7E34
		add.w	#$364,a2

locret_7E34:							; CODE XREF: sub_7E0E:loc_7E2Cj
		rts
; End of function sub_7E0E


; =============== S U B	R O U T	I N E =======================================


sub_7E36:							; CODE XREF: sub_7CF8:loc_7D48p
					; ROM:00008854p ...
		move.l	a0,($FFFFCF24).w
		move.l	a1,($FFFFCF28).w
		clr.w	($FFFFCF20).w
		clr.w	($FFFFCF22).w
; End of function sub_7E36


loc_7E46:							; CODE XREF: ROM:00007EE2p
					; ROM:00013784p
		move.l	($FFFFCF28).w,a0
		jsr	(a0)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	4
		dc.b	$FC
		dc.b	$C
; ---------------------------------------------------------------------------
		bsr.w	sub_7E72
		moveq	#$16,d0
		moveq	#6,d1
		jsr	sub_119B8
		bsr.w	loc_7EF2
		move.w	#$18,($FFFFBD26).w
		rts

; =============== S U B	R O U T	I N E =======================================


sub_7E72:							; CODE XREF: ROM:00007E58p
					; ROM:loc_7F3Cp ...
		move.w	#5,($FFFFB028).w
		btst	#1,($FFFFBF78).w
		bne.w	locret_A9D4
		addq.w	#4,($FFFFB028).w
		rts
; End of function sub_7E72

; ---------------------------------------------------------------------------

loc_7E88:							; CODE XREF: sub_7CF8+68p
					; ROM:00008868p ...
		btst	#7,d1
		bne.w	loc_7EEC
		btst	#1,d1
		beq.w	loc_7EA0
		addq.w	#1,($FFFFCF20).w
		bra.w	loc_7EF2
; ---------------------------------------------------------------------------

loc_7EA0:							; CODE XREF: ROM:00007E94j
		btst	#0,d1
		beq.w	loc_7EB0
		subq.w	#1,($FFFFCF20).w
		bra.w	loc_7EF2
; ---------------------------------------------------------------------------

loc_7EB0:							; CODE XREF: ROM:00007EA4j
		btst	#5,d1
		beq.w	loc_7EEC
		bsr.w	sub_7E0E
		move.w	($FFFFCF20).w,d0
		move.l	($FFFFCF24).w,a0
		add.w	(a0),a0
		add.w	(a0),a0
		bra.w	loc_7ECE
; ---------------------------------------------------------------------------

loc_7ECC:							; CODE XREF: ROM:00007ED0j
		addq.w	#4,a0

loc_7ECE:							; CODE XREF: ROM:00007EC8j
		add.w	(a0),a0
		dbf	d0,loc_7ECC
		move.l	(a0),a0
		jsr	(a0)
		bclr	#1,($FFFFC2F6).w
		bne.w	loc_7EE6
		bsr.w	loc_7E46

loc_7EE6:							; CODE XREF: ROM:00007EDEj
		tst.w	($FFFFCF20).w
		rts
; ---------------------------------------------------------------------------

loc_7EEC:							; CODE XREF: ROM:00007E8Cj
					; ROM:00007EB4j
		eor	#4,ccr
		rts
; ---------------------------------------------------------------------------

loc_7EF2:							; CODE XREF: ROM:00007E66p
					; ROM:00007E9Cj ...
		move.w	($FFFFCF20).w,d0
		bpl.w	loc_7F00
		clr.w	($FFFFCF20).w
		clr.w	d0

loc_7F00:							; CODE XREF: ROM:00007EF6j
		move.l	($FFFFCF24).w,a0
		add.w	(a0),a0
		add.w	(a0),a0
		bra.w	loc_7F14
; ---------------------------------------------------------------------------

loc_7F0C:							; CODE XREF: ROM:loc_7F14j
		add.w	(a0),a0
		addq.w	#4,a0
		tst.w	2(a0)

loc_7F14:							; CODE XREF: ROM:00007F08j
		dbmi	d0,loc_7F0C
		addq.w	#1,d0
		sub.w	d0,($FFFFCF20).w
		move.w	($FFFFCF20).w,d0
		cmp.w	($FFFFCF22).w,d0
		bge.w	loc_7F2E
		move.w	d0,($FFFFCF22).w

loc_7F2E:							; CODE XREF: ROM:00007F26j
		subq.w	#3,d0
		cmp.w	($FFFFCF22).w,d0
		ble.w	loc_7F3C
		move.w	d0,($FFFFCF22).w

loc_7F3C:							; CODE XREF: ROM:00007F34j
		bsr.w	sub_7E72
		move.w	#$D,($FFFFB02A).w
		move.l	($FFFFCF24).w,a1
		jsr	sub_11A48
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$26
		dc.b	$FB
		dc.b	1
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	1
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	1
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	1
		dc.b	$20
		dc.b	$FB
		dc.b	$12
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$FF
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$FF
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$FF
		dc.b	$20
; ---------------------------------------------------------------------------
		add.w	(a1),a1
		move.w	($FFFFCF22).w,d0
		bra.w	loc_7F8A
; ---------------------------------------------------------------------------

loc_7F86:							; CODE XREF: ROM:loc_7F8Aj
		add.w	(a1),a1
		addq.w	#4,a1

loc_7F8A:							; CODE XREF: ROM:00007F82j
		dbf	d0,loc_7F86
		moveq	#3,d1
		move.w	#$C,($FFFFB02A).w
		bsr.w	sub_7E72
		move.w	($FFFFCF22).w,d0
		beq.w	loc_7FB4
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
		dc.b	$FE
		dc.b	5
		dc.b	$FB
		dc.b	1
		dc.b	$FA
		dc.b	1
		dc.b	$7B
		dc.b	$FA
		dc.b	$FF
		dc.b	0
; ---------------------------------------------------------------------------

loc_7FB4:							; CODE XREF: ROM:00007F9Ej
					; ROM:00007FECj
		bsr.w	sub_7E72
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FB
		dc.b	2
		dc.b	$FA
		dc.b	1
		dc.b	$2F
		dc.b	9
; ---------------------------------------------------------------------------
		move.l	($FFFFCF24).w,a1
		jsr	sub_11A48
		cmp.w	($FFFFCF20).w,d0
		bne.w	loc_7FDE
		jsr	sub_11A48

loc_7FDE:							; CODE XREF: ROM:00007FD4j
		move.l	(sp)+,a1
		bsr.w	sub_8008
		addq.w	#1,d0
		addq.w	#4,a1
		tst.w	2(a1)
		dbmi	d1,loc_7FB4
		bmi.w	locret_8006
		bsr.w	sub_7E72
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FE
		dc.b	5
		dc.b	$FB
		dc.b	1
		dc.b	$7D
		dc.b	0
; ---------------------------------------------------------------------------

locret_8006:							; CODE XREF: ROM:00007FF0j
		rts

; =============== S U B	R O U T	I N E =======================================


sub_8008:							; CODE XREF: ROM:00007FE0p
		cmp.b	#$78,2(a1)				; 'x'
		bne.w	loc_8048
		move.l	a1,-(sp)
		move.l	#$8050,a1
		btst	#1,($FFFFC2EC).w
		beq.w	loc_802C
		tst.w	($FFFFD05C).w
		bra.w	loc_8030
; ---------------------------------------------------------------------------

loc_802C:							; CODE XREF: sub_8008+18j
		tst.w	($FFFFD05A).w

loc_8030:							; CODE XREF: sub_8008+20j
		beq.w	loc_803A
		move.l	#$8064,a1

loc_803A:							; CODE XREF: sub_8008:loc_8030j
		jsr	sub_11A48
		move.l	(sp)+,a1
		add.w	(a1),a1
		bra.w	locret_804E
; ---------------------------------------------------------------------------

loc_8048:							; CODE XREF: sub_8008+6j
		jsr	sub_11A48

locret_804E:							; CODE XREF: sub_8008+3Cj
		rts
; End of function sub_8008

; ---------------------------------------------------------------------------
		dc.w	$14
aManualGoalie:	dc.b	'  Manual Goalie   '
		dc.w	$14
aAutoGoalie:	dc.b	'   Auto Goalie    '

; =============== S U B	R O U T	I N E =======================================


sub_8078:							; CODE XREF: ROM:loc_8A00p
					; ROM:00008EF0p ...
		movem.l	d0-d2/a0-a1,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	#$C334,a1
		add.w	d0,a1
		moveq	#1,d2

loc_808E:							; CODE XREF: sub_8078+2Aj
		jsr	sub_11952
		move.w	#$A,d1

loc_8098:							; CODE XREF: sub_8078+22j
		move.w	(a1)+,(a0)
		dbf	d1,loc_8098
		addq.w	#1,($FFFFB02A).w
		dbf	d2,loc_808E
		add.w	#$B,($FFFFB028).w
		subq.w	#2,($FFFFB02A).w
		move.w	(sp)+,($FFFFBF78).w
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_8078


; =============== S U B	R O U T	I N E =======================================


sub_80BA:							; CODE XREF: sub_7CF8:loc_7D4Cp
					; ROM:loc_816Ap ...
		move.w	d0,-(sp)
		move.w	($FFFFB040).w,d0
		move.w	($FFFFB040).w,d0

loc_80C4:							; CODE XREF: sub_80BA+Ej
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_80C4
		move.w	($FFFFB040).w,($FFFFB03E).w
		move.w	(sp)+,d0
		rts
; End of function sub_80BA

; ---------------------------------------------------------------------------

loc_80D4:							; DATA XREF: ROM:000197F4o
					; ROM:00019980o ...
		moveq	#6,d0
		moveq	#$1A,d1
		bsr.w	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	5
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1E,d0
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	$BD					; 
		dc.b	$E
		dc.b	3
aScores:	dc.b	'Scores'
		dc.b	$BD					; 
		dc.b	$B
		dc.b	1
; ---------------------------------------------------------------------------
		move.l	#unk_B3E74,a1
		lea	8(a1),a2
		add.l	4(a1),a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		clr.w	d5
		jsr	sub_1169A
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		clr.w	($FFFFD5AE).w
		bsr.w	sub_81CA
		jsr	sub_1828A
		move.w	d1,d3
		ble.w	loc_8150

loc_8148:							; CODE XREF: ROM:0000814Cj
		bsr.w	sub_81F2
		dbf	d3,loc_8148

loc_8150:							; CODE XREF: ROM:00008144j
		clr.w	($FFFFD5B2).w
		move.w	($FFFFB02A).w,d0
		sub.w	#$15,d0
		bmi.w	loc_816A
		asl.w	#3,d0
		move.w	d0,($FFFFD5B2).w
		bsr.w	loc_8280

loc_816A:							; CODE XREF: ROM:0000815Cj
					; ROM:00008196j
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		btst	#7,d3
		bne.w	loc_9CD8
		moveq	#$FFFFFFFE,d0
		btst	#0,d3
		bne.w	loc_818E
		neg.w	d0
		btst	#1,d3
		beq.w	loc_8192

loc_818E:							; CODE XREF: ROM:00008180j
		move.w	d0,($FFFFD5B0).w

loc_8192:							; CODE XREF: ROM:0000818Aj
		bsr.w	sub_8198
		bra.s	loc_816A

; =============== S U B	R O U T	I N E =======================================


sub_8198:							; CODE XREF: ROM:loc_8192p
		move.w	($FFFFD5B0).w,d0
		beq.w	locret_A9D4
		add.w	($FFFFD5AE).w,d0
		bmi.w	locret_A9D4
		cmp.w	($FFFFD5B2).w,d0
		bgt.w	locret_A9D4
		move.w	d0,($FFFFD5AE).w
		ext.l	d0
		divu.w	#$18,d0
		swap	d0
		tst.w	d0
		bne.w	sub_81CA
		bsr.w	loc_8280
		clr.w	($FFFFD5B0).w
; End of function sub_8198


; =============== S U B	R O U T	I N E =======================================


sub_81CA:							; CODE XREF: ROM:00008138p
					; sub_8198+26j
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		move.w	#$FFD0,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_81CA


; =============== S U B	R O U T	I N E =======================================


sub_81F2:							; CODE XREF: ROM:loc_8148p
		cmp.w	($FFFFCEE6).w,d3
		beq.w	locret_A9D4
		moveq	#$10,d0
		mulu.w	d3,d0
		move.w	#$CE66,a0
		add.w	d0,a0
		btst	#1,$E(a0)
		bne.w	locret_A9D4
		btst	#2,$E(a0)
		bne.w	locret_A9D4
		move.w	2(a0),d0
		move.w	$C(a0),d1
		bsr.w	sub_8250
		addq.w	#1,($FFFFB02A).w
		move.w	(a0),d0
		move.w	$A(a0),d1
		bsr.w	sub_8250
		move.w	8(a0),d0
		subq.w	#1,d0
		move.l	#word_191E4,a1
		move.w	#$1C,($FFFFB028).w
		jsr	sub_13508
		addq.w	#2,($FFFFB02A).w
		rts
; End of function sub_81F2


; =============== S U B	R O U T	I N E =======================================


sub_8250:							; CODE XREF: sub_81F2+2Ep sub_81F2+3Cp
		move.w	#$30E,a1
		asl.w	#2,d0
		move.l	(a1,d0.w),a1
		add.w	4(a1),a1
		move.w	#6,($FFFFB028).w
		jsr	sub_11BA4
		move.w	#$18,($FFFFB028).w
		move.w	d1,d0
		moveq	#2,d1
		jsr	sub_11D3A
		jmp	sub_11BA4
; End of function sub_8250

; ---------------------------------------------------------------------------

loc_8280:							; CODE XREF: ROM:00008166p
					; sub_8198+2Ap
		movem.l	d0-d1/a1,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	4
		dc.b	6
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		clr.w	d0
		move.w	($FFFFD5AE).w,d1
		tst.w	d1
		sgt	d0
		neg.b	d0
		cmp.w	($FFFFD5B2).w,d1
		slt	d1
		neg.b	d1
		add.b	d1,d0
		add.b	d1,d0
		lea	word_82BC(pc),a1
		jsr	sub_13508
		movem.l	(sp)+,d0-d1/a1
		rts
; ---------------------------------------------------------------------------
word_82BC:	dc.w	$A					; DATA XREF: ROM:000082ACo
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$13
		dc.b	$20
		dc.b	$F9
		dc.b	0
		dc.w	$A
		dc.b	$7B
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$13
		dc.b	$20
		dc.b	$F9
		dc.b	0
		dc.w	$A
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$13
		dc.b	$7D
		dc.b	$F9
		dc.b	0
; ---------------------------------------------------------------------------

loc_82DA:							; DATA XREF: ROM:00019764o
					; ROM:000198F0o ...
		bset	#0,$30(a2)
		moveq	#0,d0
		moveq	#$1C,d1
		bsr.w	loc_9BD8
		clr.w	($FFFFD5AC).w
		clr.w	($FFFFD5B0).w
		move.w	#1,($FFFFD5B8).w

loc_82F6:							; CODE XREF: ROM:00008890j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FF
		dc.b	2
		dc.b	$FD
		dc.b	0
		dc.b	$FC
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$1C,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		st	($FFFFC012).w
		bsr.w	sub_878E

loc_831A:							; CODE XREF: ROM:00008442j
					; ROM:000084CAj
		bsr.w	loc_86F8
		bsr.w	loc_8566

loc_8322:							; CODE XREF: ROM:00008364j
					; ROM:00008376j ...
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		jsr	sub_11318
		btst	#7,d1
		bne.w	loc_8812
		btst	#5,d1
		bne.w	loc_838E
		moveq	#1,d0
		btst	#1,d1
		bne.w	loc_8366
		moveq	#$FFFFFFFF,d0
		btst	#0,d1
		bne.w	loc_8366
		moveq	#8,d0
		btst	#3,d1
		bne.w	loc_8366
		moveq	#$FFFFFFF8,d0
		btst	#2,d1
		beq.s	loc_8322

loc_8366:							; CODE XREF: ROM:00008346j
					; ROM:00008350j ...
		add.w	($FFFFD5B8).w,d0
		tst.w	($FFFFD058).w
		beq.w	loc_837E
		cmpnw	 1,d0
		blt.s	loc_8322
		cmpnw	 5,d0
		bgt.s	loc_8322

loc_837E:							; CODE XREF: ROM:0000836Ej
		lea	$87CA(pc),a0
		move.b	byte_87D2-unk_87CA(a0,d0.w),($FFFFD5B9).w
		bsr.w	loc_8566
		bra.s	loc_8322
; ---------------------------------------------------------------------------

loc_838E:							; CODE XREF: ROM:0000833Cj
		bsr.w	sub_9F40
		move.w	d0,d1
		bsr.w	sub_9F7E
		move.w	($FFFFD5B8).w,d2
		and.w	#7,d2
		cmpnw	 2,d2
		bgt.w	loc_83B0
		add.w	d0,d1
		bsr.w	sub_9F9A
		sub.w	d1,d0

loc_83B0:							; CODE XREF: ROM:000083A4j
		subq.w	#1,d0
		move.w	d0,($FFFFD5B6).w
		clr.w	($FFFFD5B0).w
		clr.w	($FFFFD5AE).w
		move.w	#$C018,a0
		clr.w	d2

loc_83C4:							; CODE XREF: ROM:000083D6j
		move.b	d1,(a0,d2.w)
		cmp.w	d1,d6
		bne.w	loc_83D2
		move.w	d2,($FFFFD5AE).w

loc_83D2:							; CODE XREF: ROM:000083CAj
		addq.w	#1,d1
		addq.w	#1,d2
		dbf	d0,loc_83C4
		bsr.w	sub_8774
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	0
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$12,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	$15
		dc.b	$FC
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$12,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	2
		dc.b	1
aSelectPlayer:	dc.b	'{Select  Player}',0
; ---------------------------------------------------------------------------
		clr.w	d0
		bra.w	loc_8484
; ---------------------------------------------------------------------------

loc_8430:							; CODE XREF: ROM:00008472j
					; ROM:00008478j ...
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		jsr	sub_11318
		btst	#7,d1
		bne.w	loc_831A
		btst	#5,d1
		bne.w	loc_84B6
		moveq	#1,d0
		btst	#1,d1
		bne.w	loc_8484
		btst	#3,d1
		bne.w	loc_8474
		moveq	#$FFFFFFFF,d0
		btst	#0,d1
		bne.w	loc_8484
		btst	#2,d1
		bne.w	loc_8474
		bra.s	loc_8430
; ---------------------------------------------------------------------------

loc_8474:							; CODE XREF: ROM:0000845Cj
					; ROM:0000846Ej
		add.w	($FFFFD5AC).w,d0
		bmi.s	loc_8430
		move.w	d0,($FFFFD5AC).w
		bsr.w	loc_84CE

loc_8482:
		bra.s	loc_8430
; ---------------------------------------------------------------------------

loc_8484:							; CODE XREF: ROM:0000842Cj
					; ROM:00008454j ...
		add.w	($FFFFD5AE).w,d0
		bmi.s	loc_8430
		cmp.w	($FFFFD5B6).w,d0
		bgt.s	loc_8430
		move.w	d0,($FFFFD5AE).w
		cmp.w	($FFFFD5B0).w,d0

loc_8498:
		bgt.w	loc_84A0
		move.w	d0,($FFFFD5B0).w

loc_84A0:							; CODE XREF: ROM:loc_8498j
		subq.w	#5,d0
		cmp.w	($FFFFD5B0).w,d0
		ble.w	loc_84AE
		move.w	d0,($FFFFD5B0).w

loc_84AE:							; CODE XREF: ROM:000084A6j
		bsr.w	loc_84CE
		bra.w	loc_8430
; ---------------------------------------------------------------------------

loc_84B6:							; CODE XREF: ROM:0000844Aj
		move.w	#$C018,a3
		add.w	($FFFFD5AE).w,a3
		move.b	(a3),d0
		addq.b	#1,d0
		move.w	($FFFFD5B8).w,d2
		bsr.w	sub_8898
		bra.w	loc_831A
; ---------------------------------------------------------------------------

loc_84CE:							; CODE XREF: ROM:0000847Ep
					; ROM:loc_84AEp
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$16
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------

loc_84DA:							; CODE XREF: ROM:000084FAj
		move.l	#word_193F8,a1
		move.w	($FFFFD5AC).w,d0
		bra.w	loc_84EC
; ---------------------------------------------------------------------------

loc_84E8:							; CODE XREF: ROM:000084EEj
		add.w	(a1),a1
		addq.w	#4,a1

loc_84EC:							; CODE XREF: ROM:000084E4j
		tst.w	(a1)
		dbmi	d0,loc_84E8
		bpl.w	loc_84FC
		subq.w	#1,($FFFFD5AC).w
		bra.s	loc_84DA
; ---------------------------------------------------------------------------

loc_84FC:							; CODE XREF: ROM:000084F2j
		jsr	sub_11BA4
		move.l	(a1),d4
		move.w	#$C018,a3
		move.w	($FFFFD5B0).w,d2
		move.w	($FFFFD5B6).w,d1
		sub.w	d2,d1
		cmpnw	 5,d1
		bls.w	loc_851C
		moveq	#5,d1

loc_851C:							; CODE XREF: ROM:00008516j
		move.w	#2,($FFFFB02A).w

loc_8522:							; CODE XREF: ROM:00008560j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$20
		dc.b	$FE
		dc.b	4
		dc.b	$FD
		dc.b	5
		dc.b	$FA
		dc.b	1
		dc.b	'                      '
		dc.b	$FD
		dc.b	5
; ---------------------------------------------------------------------------
		cmp.w	($FFFFD5AE).w,d2
		bne.w	loc_8554
		move.w	d7,($FFFFB02C).w

loc_8554:							; CODE XREF: ROM:0000854Cj
		clr.w	d0
		move.b	(a3,d2.w),d0
		bsr.w	sub_8D4E
		addq.w	#1,d2
		dbf	d1,loc_8522
		rts
; ---------------------------------------------------------------------------

loc_8566:							; CODE XREF: ROM:0000831Ep
					; ROM:00008388p
		moveq	#6,d5
		lea	unk_85E7(pc),a0
		move.w	($FFFFD5B8).w,d0
		lsr.w	#3,d0
		add.w	d0,a0
		tst.w	($FFFFD058).w
		beq.w	loc_857E
		subq.w	#1,a0

loc_857E:							; CODE XREF: ROM:00008578j
		move.b	(a0),d0
		cmp.b	($FFFFC012).w,d0
		beq.w	loc_8590
		move.b	d0,($FFFFC012).w
		bsr.w	loc_868C

loc_8590:							; CODE XREF: ROM:00008584j
					; ROM:loc_859Cj
		btst	d5,($FFFFC012).w
		beq.w	loc_859C
		bsr.w	loc_85EE

loc_859C:							; CODE XREF: ROM:00008594j
		dbf	d5,loc_8590
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	8
		dc.b	7
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$18,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	$15
		dc.b	$FC
		dc.b	8
		dc.b	$30
		dc.b	6
; ---------------------------------------------------------------------------
		jsr	sub_18A90
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		move.w	d7,($FFFFB02C).w
		jsr	sub_11A48
		clr.w	($FFFFB030).w
		rts
; ---------------------------------------------------------------------------
		dc.b	1
unk_85E7:	dc.b	7					; DATA XREF: ROM:00008568o
		dc.b	7
		dc.b	7
		dc.b	$18
		dc.b	$18
		dc.b	$60					; `
		dc.b	$60					; `
; ---------------------------------------------------------------------------

loc_85EE:							; CODE XREF: ROM:00008598p
		moveq	#6,d0
		mulu.w	d5,d0
		lea	word_879E(pc),a0
		add.w	d0,a0
		tst.w	($FFFFD058).w
		beq.w	loc_8602
		subq.w	#6,a0

loc_8602:							; CODE XREF: ROM:000085FCj
		move.w	(a0),($FFFFB028).w
		move.w	2(a0),($FFFFB02A).w
		move.w	d5,d0
		move.l	#word_191A6,a1
		move.w	#$8000,($FFFFB02C).w
		jsr	sub_13508
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
aLine:		dc.b	' Line',0
; ---------------------------------------------------------------------------
		move.w	(a0),($FFFFB028).w
		move.w	d5,d4
		asl.w	#3,d4
		addq.w	#1,d4
		move.w	2(a0),($FFFFB02A).w
		move.w	4(a0),d3
		lea	$16A(a2),a3

loc_8646:							; CODE XREF: ROM:00008686j
		move.w	(a0),($FFFFB028).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	2
		dc.b	$FE
		dc.b	6
; ---------------------------------------------------------------------------
		clr.w	d0
		move.b	(a3,d4.w),d0
		subq.w	#1,d0
		jsr	sub_18B6E
		cmp.w	($FFFFD5B8).w,d4
		bne.w	loc_867A
		move.w	d0,d6
		move.w	($FFFFB02C).w,d7
		move.w	#2,($FFFFB030).w

loc_867A:							; CODE XREF: ROM:0000866Aj
		jsr	sub_11A48
		clr.w	($FFFFB030).w
		addq.w	#1,d4
		dbf	d3,loc_8646
		rts
; ---------------------------------------------------------------------------

loc_868C:							; CODE XREF: ROM:0000858Cp
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FF
		dc.b	2
		dc.b	$FD
		dc.b	0
		dc.b	$FC
		dc.b	$A
		dc.b	$70					; p
		dc.b	$28					; (
		dc.b	$72					; r
		dc.b	$12
		dc.b	$34					; 4
		dc.b	$3C					; <
		dc.b	7
		dc.b	$FF
		dc.b	$4E					; N
		dc.b	$B9					; 
		dc.b	0
		dc.b	1
		dc.b	$19
		dc.b	$7E					; ~
		dc.b	$33					; 3
		dc.b	$F9					; 
		dc.b	0
		dc.b	0
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	$B0					; 
		dc.b	$2A					; *
		dc.b	$33					; 3
		dc.b	$F9					; 
		dc.b	0
		dc.b	0
		dc.b	$87					; 
		dc.b	$9E					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	$B0					; 
		dc.b	$28					; (
		dc.b	$4A					; J
		dc.b	$78					; x
		dc.b	$D0					; 
		dc.b	$58					; X
		dc.b	$67					; g
		dc.b	0
		dc.b	0
		dc.b	$C
		dc.b	$33					; 3
		dc.b	$F9					; 
		dc.b	0
		dc.b	0
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	$B0					; 
		dc.b	$28					; (
		dc.b	$4E					; N
		dc.b	$B9					; 
		dc.b	0
		dc.b	1
		dc.b	$1A
		dc.b	$36					; 6
		dc.w	$22
		dc.b	$FE
		dc.b	4
		dc.b	$FB
		dc.b	$FD
		dc.b	$FA
		dc.b	2
		dc.b	$4C
		dc.b	$44
		dc.b	$FB
		dc.b	$FE
		dc.b	$FA
		dc.b	2
		dc.b	$52
		dc.b	$44
		dc.b	$FB
		dc.b	$FE
		dc.b	$FA
		dc.b	2
		dc.b	$4C
		dc.b	$57
		dc.b	$FB
		dc.b	$FE
		dc.b	$FA
		dc.b	2
		dc.b	$43
		dc.b	$20
		dc.b	$FB
		dc.b	$FE
		dc.b	$FA
		dc.b	2
		dc.b	$52
		dc.b	$57
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_86F8:							; CODE XREF: ROM:loc_831Ap
		bsr.w	sub_8774
		movem.l	d0-d5/a0-a2,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#1,d4
		moveq	#0,d5
		jsr	sub_1169A
		movem.l	(sp)+,d0-d5/a0-a2
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	7
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1A,d0
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$BE
		dc.b	$A
		dc.b	4
		dc.b	$4C
		dc.b	$69
		dc.b	$6E
		dc.b	$65
		dc.b	$20
		dc.b	$20
		dc.b	$45
		dc.b	$64
		dc.b	$69
		dc.b	$74
		dc.b	$6F
		dc.b	$72
		dc.b	$BE
		dc.b	$E
		dc.b	1
		dc.b	$42					; New data starts here,	donno...
		dc.b	$40
		dc.b	$B4
		dc.b	$FC
		dc.b	$C6
		dc.b	$CE
		dc.b	$67
		dc.b	0
		dc.b	$F9
		dc.b	$E
		dc.b	$30
		dc.b	$3C
		dc.b	0
		dc.b	$2C
		dc.b	$60
		dc.b	0
		dc.b	$F9
		dc.b	6

; =============== S U B	R O U T	I N E =======================================


sub_8774:							; CODE XREF: ROM:000083DAp
					; ROM:loc_86F8p
		jsr	sub_11B92
; End of function sub_8774

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$A,d1
		move.w	#$7FF,d2
		jmp	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_878E:							; CODE XREF: ROM:00008316p
					; ROM:loc_8812p
		clr.w	($FFFFBD82).w
		clr.w	($FFFFBDA2).w
		rts
; End of function sub_878E

; ---------------------------------------------------------------------------
		dc.w	$10
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	4
word_879E:	dc.w	4					; DATA XREF: ROM:000085F2o
		dc.w	$C
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$1C
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	3
		dc.b	$FF
		dc.b	$FF
unk_87CA:	dc.b	1
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	$19
		dc.b	1
byte_87D2:	dc.b	1					; DATA XREF: ROM:00008382r
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	$19
		dc.b	1
		dc.b	9
		dc.b	9
		dc.b	$A
		dc.b	$B
		dc.b	$C
		dc.b	$D
		dc.b	$21					; !
		dc.b	1
		dc.b	$11
		dc.b	$11
		dc.b	$12
		dc.b	$13
		dc.b	$14
		dc.b	$15
		dc.b	$21					; !
		dc.b	1
		dc.b	5
		dc.b	$19
		dc.b	$1A
		dc.b	$1B
		dc.b	$1C
		dc.b	$1D
		dc.b	$29					; )
		dc.b	1
		dc.b	$D
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	1
		dc.b	$1D
		dc.b	$29					; )
		dc.b	$2A					; *
		dc.b	$2B					; +
		dc.b	$2C					; ,
		dc.b	$2C					; ,
		dc.b	1
		dc.b	1
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	1
		dc.b	1
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$34					; 4
		dc.b	1
		dc.b	1
; ---------------------------------------------------------------------------

loc_8812:							; CODE XREF: ROM:00008334j
		bsr.w	sub_878E
		move.w	#$18,($FFFFBD26).w
		move.l	($FFFFCF20).w,-(sp)
		move.l	($FFFFCF24).w,-(sp)
		move.l	($FFFFCF28).w,-(sp)
		move.l	#locret_15464,a1
		move.l	#word_19F88,a0
		move.w	#$D076,a3
		move.w	$28(a2),d0
		addq.w	#1,d0
		cmp.b	(a3),d0
		beq.w	loc_884A
		move.l	#word_19FF8,a0

loc_884A:							; CODE XREF: ROM:00008840j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	4
		dc.w	$FF02
; ---------------------------------------------------------------------------
		bsr.w	sub_7E36

loc_8858:							; CODE XREF: ROM:00008872j
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		jsr	sub_11318
		move.w	d1,-(sp)
		bsr.w	loc_7E88
		move.w	(sp)+,d1
		and.w	#$A0,d1					; ''
		beq.s	loc_8858
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	4
		dc.w	$F900
; ---------------------------------------------------------------------------
		move.w	($FFFFCF20).w,d0
		move.l	(sp)+,($FFFFCF28).w
		move.l	(sp)+,($FFFFCF24).w
		move.l	(sp)+,($FFFFCF20).w
		tst.w	d0
		bne.w	loc_82F6
		bra.w	loc_9CD8

; =============== S U B	R O U T	I N E =======================================


sub_8898:							; CODE XREF: ROM:000084C6p
					; sub_88C8+54p
		movem.l	d0-d2/a0-a1,-(sp)
		lea	$16A(a2),a0
		move.w	d2,d1
		and.w	#$FFF8,d1
		lea	(a0,d1.w),a1
		moveq	#5,d1

loc_88AC:							; CODE XREF: sub_8898+18j
		cmp.b	1(a1,d1.w),d0
		dbeq	d1,loc_88AC
		bne.w	loc_88BE
		move.b	(a0,d2.w),1(a1,d1.w)

loc_88BE:							; CODE XREF: sub_8898+1Cj
		move.b	d0,(a0,d2.w)
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_8898


; =============== S U B	R O U T	I N E =======================================


sub_88C8:							; DATA XREF: ROM:00019FF0o
		movem.l	d0-d4/a0/a3,-(sp)
		move.w	#$D076,a0
		addq.w	#1,a0
		bsr.w	sub_9F40
		move.w	d0,d1
		bsr.w	sub_9F7E
		move.w	d0,d3
		move.l	#$898A,a3
		clr.w	d4
		clr.w	d2

loc_88E8:							; CODE XREF: sub_88C8+58j
		move.b	(a3),d2
		bmi.w	loc_8922
		bchg	#0,d4
		bne.w	loc_88FC
		move.b	(a0),d0
		bra.w	loc_8900
; ---------------------------------------------------------------------------

loc_88FC:							; CODE XREF: sub_88C8+2Aj
		move.b	(a0)+,d0
		lsr.w	#4,d0

loc_8900:							; CODE XREF: sub_88C8+30j
		and.w	#$F,d0
		add.b	d1,d0
		and.w	#7,d2
		cmpnw	 2,d2
		bgt.w	loc_8914
		add.b	d3,d0

loc_8914:							; CODE XREF: sub_88C8+46j
		addq.b	#1,d0
		and.w	#$FF,d0
		move.b	(a3)+,d2
		bsr.w	sub_8898
		bra.s	loc_88E8
; ---------------------------------------------------------------------------

loc_8922:							; CODE XREF: sub_88C8+22j
		movem.l	(sp)+,d0-d4/a0/a3
		rts
; End of function sub_88C8


; =============== S U B	R O U T	I N E =======================================


sub_8928:							; DATA XREF: ROM:00019FD8o
					; ROM:0001A048o
		movem.l	d0-d4/a0-a3,-(sp)
		move.w	#$D076,a0
		move.w	$28(a2),d0
		addq.w	#1,d0
		move.b	d0,(a0)+
		lea	$16A(a2),a1
		bsr.w	sub_9F40
		move.w	d0,d1
		bsr.w	sub_9F7E
		move.l	#$898A,a3
		clr.w	d4
		clr.w	d2

loc_8950:							; CODE XREF: sub_8928+4Ej sub_8928+54j
		move.b	(a3)+,d2
		bmi.w	loc_897E
		move.b	(a1,d2.w),d3
		subq.b	#1,d3
		sub.b	d1,d3
		and.w	#7,d2
		cmpnw	 2,d2
		bgt.w	loc_896C
		sub.b	d0,d3

loc_896C:							; CODE XREF: sub_8928+3Ej
		bchg	#0,d4
		bne.w	loc_8978
		move.b	d3,(a0)
		bra.s	loc_8950
; ---------------------------------------------------------------------------

loc_8978:							; CODE XREF: sub_8928+48j
		asl.b	#4,d3
		or.b	d3,(a0)+
		bra.s	loc_8950
; ---------------------------------------------------------------------------

loc_897E:							; CODE XREF: sub_8928+2Aj
		jsr	sub_FE696
		movem.l	(sp)+,d0-d4/a0-a3
		rts
; End of function sub_8928

; ---------------------------------------------------------------------------
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	9
		dc.b	$A
		dc.b	$B
		dc.b	$C
		dc.b	$D
		dc.b	$11
		dc.b	$12
		dc.b	$13
		dc.b	$14
		dc.b	$15
		dc.b	$19
		dc.b	$1A
		dc.b	$1B
		dc.b	$1C
		dc.b	$1D
		dc.b	$21					; !
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$24					; $
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$2A					; *
		dc.b	$2B					; +
		dc.b	$2C					; ,
		dc.b	$31					; 1
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$FF
; ---------------------------------------------------------------------------

loc_89AC:							; DATA XREF: ROM:000196B0o
					; ROM:000197DCo ...
		moveq	#$D,d0
		moveq	#$17,d1
		bsr.w	loc_9BD8
		moveq	#1,d0
		add.w	$16(a2),d0
		move.w	d0,($FFFFD5B2).w
		clr.w	($FFFFD5AC).w

loc_89C2:							; CODE XREF: ROM:00008AF8j
					; ROM:00008B00j
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	7
		dc.b	1
		dc.b	0
		dc.b	$70
		dc.b	$1A
; ---------------------------------------------------------------------------
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$BD
		dc.b	9
		dc.b	4
		dc.b	$54
		dc.b	$65
		dc.b	$61
		dc.b	$6D
aRoster:	dc.b	'  Roster'
		dc.b	$BD
		dc.b	$E
		dc.b	1
; ---------------------------------------------------------------------------
		clr.w	d0
		cmp.w	#$C6CE,a2
		beq.w	loc_8A00
		move.w	#$2C,d0					; ','

loc_8A00:							; CODE XREF: ROM:000089F8j
		bsr.w	sub_8078
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$34
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	2
		dc.b	$C
		dc.b	$F9
		dc.b	1
aPos_Player:	dc.b	'Pos.^Player'
		dc.b	$FD
		dc.b	$1E
		dc.b	$FC
		dc.b	$C
aRating:	dc.b	'Rating'
		dc.b	$FD
		dc.b	$C
		dc.b	$FC
		dc.b	$1A
		dc.b	$41					; A
		dc.b	$5E					; ^
		dc.b	$2D					; -
		dc.b	$5E					; ^
aSwitchTeams_0:	dc.b	'Switch^Teams'
		dc.b	$F9					; 
		dc.b	0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	1
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$12,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	$15
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$12,d0
		moveq	#3,d1
		jsr	sub_119B8
		bsr.w	loc_8BA6
		move.w	($FFFFD5B2).w,d0
		mulu.w	#$80,d0					; ''
		move.w	d0,($FFFFD5AE).w
		bsr.w	sub_8B1C
		clr.w	($FFFFD5B0).w

loc_8A82:							; CODE XREF: ROM:00008ADEj
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		jsr	sub_11318
		btst	#7,d3
		bne.w	loc_9CD8
		btst	#6,d1
		bne.w	loc_8AF0
		jsr	sub_112F0
		moveq	#1,d0
		btst	#3,d1
		bne.w	loc_8AE0
		neg.w	d0
		btst	#2,d1
		bne.w	loc_8AE0
		tst.w	($FFFFD5B0).w
		bne.w	loc_8ADA
		moveq	#$FFFFFFFE,d0
		btst	#0,d3
		bne.w	loc_8AD6
		neg.w	d0
		btst	#1,d3
		beq.w	loc_8ADA

loc_8AD6:							; CODE XREF: ROM:00008AC8j
		move.w	d0,($FFFFD5B0).w

loc_8ADA:							; CODE XREF: ROM:00008ABEj
					; ROM:00008AD2j ...
		bsr.w	sub_8B04
		bra.s	loc_8A82
; ---------------------------------------------------------------------------

loc_8AE0:							; CODE XREF: ROM:00008AACj
					; ROM:00008AB6j
		add.w	($FFFFD5AC).w,d0
		bmi.s	loc_8ADA
		move.w	d0,($FFFFD5AC).w
		bsr.w	loc_8BA6
		bra.s	loc_8ADA
; ---------------------------------------------------------------------------

loc_8AF0:							; CODE XREF: ROM:00008A9Cj
		lea	$364(a2),a2
		cmp.w	#$CA32,a2
		beq.w	loc_89C2
		move.w	#$C6CE,a2
		bra.w	loc_89C2

; =============== S U B	R O U T	I N E =======================================


sub_8B04:							; CODE XREF: ROM:loc_8ADAp

; FUNCTION CHUNK AT 00008B7E SIZE 00000006 BYTES

		move.w	($FFFFD5B0).w,d0
		beq.w	locret_A9D4
		add.w	($FFFFD5AE).w,d0
		bmi.w	loc_8B7E
		cmpnw	 $380,d0
		bgt.w	loc_8B7E
; End of function sub_8B04


; =============== S U B	R O U T	I N E =======================================


sub_8B1C:							; CODE XREF: ROM:00008A7Ap
		move.w	($FFFFD5AE).w,d1
		move.w	d0,($FFFFD5AE).w
		ext.l	d0
		divs.w	#$80,d0					; ''
		swap	d0
		tst.w	d0
		bne.w	loc_8B36
		clr.w	($FFFFD5B0).w

loc_8B36:							; CODE XREF: sub_8B1C+12j
		and.w	#$7F,d1					; ''
		bne.w	loc_8B56
		cmpnw	 $7E,d0					; '~'
		bne.w	loc_8B4A
		bsr.w	sub_8B84

loc_8B4A:							; CODE XREF: sub_8B1C+26j
		cmpnw	 2,d0
		bne.w	loc_8B56
		bsr.w	sub_8B94

loc_8B56:							; CODE XREF: sub_8B1C+1Ej sub_8B1C+32j
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		move.w	#$FF90,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_8B1C

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_8B04

loc_8B7E:							; CODE XREF: sub_8B04+Cj sub_8B04+14j
		clr.w	($FFFFD5B0).w
		rts
; END OF FUNCTION CHUNK	FOR sub_8B04

; =============== S U B	R O U T	I N E =======================================


sub_8B84:							; CODE XREF: sub_8B1C+2Ap
		move.l	d0,-(sp)
		swap	d0
		move.w	d0,($FFFFD5B2).w
		bsr.w	loc_8BA6
		move.l	(sp)+,d0
		rts
; End of function sub_8B84


; =============== S U B	R O U T	I N E =======================================


sub_8B94:							; CODE XREF: sub_8B1C+36p
		move.l	d0,-(sp)
		swap	d0
		addq.w	#1,d0
		move.w	d0,($FFFFD5B2).w
		bsr.w	loc_8BA6
		move.l	(sp)+,d0
		rts
; End of function sub_8B94

; ---------------------------------------------------------------------------

loc_8BA6:							; CODE XREF: ROM:00008A6Ap
					; ROM:00008AEAp ...
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	7
		dc.b	3
		dc.b	2
		dc.b	9
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5B2).w,d0
		lea	word_8CBE(pc),a1
		jsr	sub_13510
		jsr	sub_11A48
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	$16
		dc.b	$FC
		dc.b	9
; ---------------------------------------------------------------------------

loc_8BD6:							; CODE XREF: ROM:00008C04j
		move.l	#word_193F8,a1
		move.w	($FFFFD5AC).w,d0
		tst.w	($FFFFD5B2).w
		bne.w	loc_8BF6
		move.l	#word_1955A,a1
		bra.w	loc_8BF6
; ---------------------------------------------------------------------------

loc_8BF2:							; CODE XREF: ROM:00008BF8j
		add.w	(a1),a1
		addq.w	#4,a1

loc_8BF6:							; CODE XREF: ROM:00008BE4j
					; ROM:00008BEEj
		tst.w	(a1)
		dbmi	d0,loc_8BF2
		bpl.w	loc_8C06
		subq.w	#1,($FFFFD5AC).w
		bra.s	loc_8BD6
; ---------------------------------------------------------------------------

loc_8C06:							; CODE XREF: ROM:00008BFCj
		jsr	sub_11A48
		move.l	(a1),d4
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	$F9
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		btst	#0,($FFFFD5B3).w
		beq.w	loc_8C2E
		add.w	#$10,($FFFFB02A).w

loc_8C2E:							; CODE XREF: ROM:00008C24j
		move.w	($FFFFB02A).w,-(sp)
		moveq	#$28,d0					; '('
		moveq	#$10,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w
		bsr.w	sub_9F40
		move.w	d0,d1
		subq.w	#1,d1
		move.l	#$8CAE,a5
		move.w	#$C018,a3
		move.l	#$1020304,(a3)
		move.w	($FFFFD5B2).w,d0
		subq.w	#1,d0
		bmi.w	loc_8C84
		asl.w	#3,d0
		lea	$16A(a2),a3
		lea	1(a3,d0.w),a3
		move.l	#word_191D0,a5
		moveq	#4,d1
		cmp.w	#5,($FFFFD5B2).w
		ble.w	loc_8C84
		subq.w	#1,d1

loc_8C84:							; CODE XREF: ROM:00008C62j
					; ROM:00008C7Ej ...
		move.w	#2,($FFFFB028).w
		move.l	a5,a1
		jsr	sub_11BA4
		move.l	a1,a5
		move.w	#7,($FFFFB028).w
		clr.w	d0
		move.b	(a3)+,d0
		subq.w	#1,d0
		bsr.w	sub_8D4E
		addq.w	#2,($FFFFB02A).w
		dbf	d1,loc_8C84
		rts
; ---------------------------------------------------------------------------
		dc.l	unk_44700
		dc.l	unk_44700
		dc.l	unk_44700
		dc.l	unk_44700
word_8CBE:	dc.w	$12					; DATA XREF: ROM:00008BBAo
aGoalies:	dc.b	'    Goalies    }'
		dc.w	$12
aScoring1:	dc.b	'{  Scoring  1  }'
		dc.w	$12
aScoring2:	dc.b	'{  Scoring  2  }'
		dc.w	$12
aChecking:	dc.b	'{   Checking   }'
		dc.w	$12
aPowerPlay1:	dc.b	'{ Power Play 1 }'
		dc.w	$12
aPowerPlay2:	dc.b	'{ Power Play 2 }'
		dc.w	$12
aPenaltyKill1:	dc.b	'{Penalty Kill 1}'
		dc.w	$12
aPenaltyKill2:	dc.b	'{Penalty Kill 2 '

; =============== S U B	R O U T	I N E =======================================


sub_8D4E:							; CODE XREF: ROM:0000855Ap
					; ROM:00008CA0p ...
		movem.l	d0-d4/a0-a1/a4-a5,-(sp)
		pea	sub_8D82(pc)
		jsr	sub_18A90
		jsr	sub_11BA4
		move.w	#$1E,($FFFFB028).w
		cmpnw	 2,d4
		bls.w	loc_8D78
		jsr	sub_FA9F8
		swap	d4

loc_8D78:							; CODE XREF: sub_8D4E+1Ej
		lea	word_8D88(pc),a0
		add.w	(a0,d4.w),a0
		jmp	(a0)
; End of function sub_8D4E


; =============== S U B	R O U T	I N E =======================================


sub_8D82:							; DATA XREF: sub_8D4E+4o
		movem.l	(sp)+,d0-d4/a0-a1/a4-a5
		rts
; End of function sub_8D82

; ---------------------------------------------------------------------------
word_8D88:	dc.w	$C					; DATA XREF: sub_8D4E:loc_8D78o
		dc.b	0
		dc.b	$98
		dc.b	0
		dc.b	$E0
		dc.b	1
		dc.b	6
		dc.b	0
		dc.b	$B2
		dc.b	0
		dc.b	$B8
; ---------------------------------------------------------------------------
		add.w	d0,d0
		move.w	$66(a2,d0.w),d0
		bpl.w	loc_8DBA
		not.w	d0
		cmpnw	 3,d0
		bls.w	loc_8DB0
		moveq	#1,d0
		bra.w	loc_8DB0
; ---------------------------------------------------------------------------

loc_8DAE:							; CODE XREF: ROM:00008DBEj
		moveq	#3,d0

loc_8DB0:							; CODE XREF: ROM:00008DA4j
					; ROM:00008DAAj
		lea	word_8DEC(pc),a1
		jmp	sub_13508
; ---------------------------------------------------------------------------

loc_8DBA:							; CODE XREF: ROM:00008D9Aj
		btst	#$C,d0
		bne.s	loc_8DAE
		subq.w	#1,($FFFFB028).w
		move.w	d0,d1
		and.w	#$FFF,d0
		jsr	sub_11CA2
		jsr	sub_11BA4
		moveq	#4,d0
		bclr	#$E,d1
		beq.w	loc_8DE2
		moveq	#5,d0

loc_8DE2:							; CODE XREF: ROM:00008DDCj
		lea	word_8DEC(pc),a1
		jmp	sub_13508
; ---------------------------------------------------------------------------
word_8DEC:	dc.w	$A					; DATA XREF: ROM:loc_8DB0o
					; ROM:loc_8DE2o
aIce:		dc.b	'Ice     '
		dc.w	$A
aBench:		dc.b	'Bench   '
		dc.w	$A
aInjuryP:	dc.b	'Injury P'
		dc.w	$A
aInjuryG:	dc.b	'Injury G'
		dc.w	6
		dc.b	'    '
		dc.w	6
		dc.b	' C  '
; ---------------------------------------------------------------------------
		add.w	d0,d0
		move.w	$32(a2,d0.w),d0
		ext.l	d0
		divu.w	#$28,d0					; '('
		cmpnw	 $64,d0					; 'd'
		ble.w	loc_8E4C
		moveq	#$64,d0					; 'd'
		bra.w	loc_8E4C
; ---------------------------------------------------------------------------
		and.w	#$E,d0
		subq.w	#1,d1
		mulu.w	#$64,d0					; 'd'
		divu.w	d1,d0
		jsr	sub_FEF7C

loc_8E4C:							; CODE XREF: ROM:00008E30j
					; ROM:00008E36j
		moveq	#4,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	'    '
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------
		and.w	#1,d0
		eor.w	#1,d0
		lea	word_8E7A(pc),a1
		jmp	sub_13508
; ---------------------------------------------------------------------------
word_8E7A:	dc.w	$A					; DATA XREF: ROM:00008E70o
aRighty:	dc.b	'Righty  '
		dc.w	$A
aLefty:		dc.b	'Lefty   '
; ---------------------------------------------------------------------------
		asl.w	#3,d0
		add.w	#$8C,d0					; ''
		jsr	sub_11D06
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	8
aLb:		dc.b	' lb  ',0
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_8EB0:							; DATA XREF: ROM:000197ACo
					; ROM:00019938o ...
		moveq	#$A,d0
		moveq	#$1A,d1
		bsr.w	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	4
		dc.b	1
		dc.b	0
		dc.b	$70
		dc.b	$20
; ---------------------------------------------------------------------------
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	6
		dc.b	4
aScoringSummary: dc.b	'Scoring Summary'
		dc.b	$BD
		dc.b	7
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$2C,d0					; ','
		bsr.w	sub_8078
		addq.w	#4,($FFFFB028).w
		clr.w	d0
		bsr.w	sub_8078
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$34
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	0
		dc.b	8
		dc.b	$F9
		dc.b	1
aPerTimeTmGoalAssistPS: dc.b '^Per^^Time^^Tm^^Goal/Assist^^^^^^^^P/S^^'
		dc.b	$F9
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		clr.l	d0
		move.w	($FFFFC472).w,d0
		divu.w	#6,d0
		asl.w	#5,d0
		move.w	d0,($FFFFD5AE).w
		sub.w	#$80,d0					; ''
		bpl.w	loc_8F52
		clr.w	d0

loc_8F52:							; CODE XREF: ROM:00008F4Cj
		move.w	d0,($FFFFD5B2).w
		move.w	($FFFFD5AE).w,d0

loc_8F5A:							; CODE XREF: ROM:00008F6Cj
		bsr.w	sub_80BA
		bsr.w	sub_8FB8
		move.w	($FFFFD5AE).w,d0
		subq.w	#2,d0
		cmp.w	($FFFFD5B2).w,d0
		bge.s	loc_8F5A
		clr.w	($FFFFD5B0).w

loc_8F72:							; CODE XREF: ROM:00008F9Ej
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		btst	#7,d3
		bne.w	loc_9CD8
		moveq	#$FFFFFFFE,d0
		btst	#0,d3
		bne.w	loc_8F96
		neg.w	d0
		btst	#1,d3
		beq.w	loc_8F9A

loc_8F96:							; CODE XREF: ROM:00008F88j
		move.w	d0,($FFFFD5B0).w

loc_8F9A:							; CODE XREF: ROM:00008F92j
		bsr.w	sub_8FA0
		bra.s	loc_8F72

; =============== S U B	R O U T	I N E =======================================


sub_8FA0:							; CODE XREF: ROM:loc_8F9Ap
		move.w	($FFFFD5B0).w,d0
		beq.w	locret_A9D4
		add.w	($FFFFD5AE).w,d0
		bmi.w	locret_A9D4
		cmp.w	($FFFFD5B2).w,d0
		bgt.w	locret_A9D4
; End of function sub_8FA0


; =============== S U B	R O U T	I N E =======================================


sub_8FB8:							; CODE XREF: ROM:00008F5Ep
		move.w	($FFFFD5AE).w,d1
		move.w	d0,($FFFFD5AE).w
		ext.l	d0
		divs.w	#$20,d0					; ' '
		swap	d0
		tst.w	d0
		bne.w	loc_8FD6
		bsr.w	sub_93C4
		clr.w	($FFFFD5B0).w

loc_8FD6:							; CODE XREF: sub_8FB8+12j
		and.w	#$1F,d1
		bne.w	loc_8FFA
		move.l	d0,-(sp)
		cmpnw	 $1E,d0
		bne.w	loc_8FEC
		bsr.w	sub_9020

loc_8FEC:							; CODE XREF: sub_8FB8+2Cj
		move.l	(sp)+,d0
		cmpnw	 2,d0
		bne.w	loc_8FFA
		bsr.w	sub_904C

loc_8FFA:							; CODE XREF: sub_8FB8+22j sub_8FB8+3Aj
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		moveq	#$FFFFFFB0,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_8FB8


; =============== S U B	R O U T	I N E =======================================


sub_9020:							; CODE XREF: sub_8FB8+30p
		move.l	d0,-(sp)
		swap	d0
		moveq	#6,d3
		mulu.w	d0,d3
		jsr	sub_11B92
; End of function sub_9020

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5AE).w,d0
		lsr.w	#3,d0
		subq.w	#3,d0
		and.w	#$1F,d0
		move.w	d0,($FFFFB02A).w
		bsr.w	sub_907C
		move.l	(sp)+,d0
		rts

; =============== S U B	R O U T	I N E =======================================


sub_904C:							; CODE XREF: sub_8FB8+3Ep
		move.l	d0,-(sp)
		swap	d0
		addq.w	#4,d0
		moveq	#6,d3
		mulu.w	d0,d3
		jsr	sub_11B92
; End of function sub_904C

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5AE).w,d0
		lsr.w	#3,d0
		add.w	#$10,d0
		and.w	#$1F,d0
		move.w	d0,($FFFFB02A).w
		bsr.w	sub_907C
		move.l	(sp)+,d0
		rts

; =============== S U B	R O U T	I N E =======================================


sub_907C:							; CODE XREF: ROM:00009044p
					; ROM:00009074p
		move.w	($FFFFB02A).w,-(sp)
		moveq	#$28,d0					; '('
		moveq	#4,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w
		move.w	#$C474,a0
		move.w	(a0,d3.w),d0
		move.w	#1,($FFFFB028).w
		jsr	sub_11C72
		move.w	#$C6CE,a2
		btst	#7,2(a0,d3.w)
		beq.w	loc_90B8
		add.w	#$364,a2

loc_90B8:							; CODE XREF: sub_907C+34j
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		move.w	#$C,($FFFFB028).w
		jsr	sub_11BA4
		move.w	#$23,($FFFFB028).w			; '#'
		lea	word_912A(pc),a1
		move.b	2(a0,d3.w),d0
		and.w	#$7F,d0					; ''
		jsr	sub_13508
		move.w	#$10,($FFFFB028).w
		move.b	3(a0,d3.w),d0
		bsr.w	sub_9112
		move.w	#$E000,($FFFFB02C).w
		move.w	#$12,($FFFFB028).w
		move.b	4(a0,d3.w),d0
		bsr.w	sub_9112
		move.w	#$12,($FFFFB028).w
		move.b	5(a0,d3.w),d0
; End of function sub_907C


; =============== S U B	R O U T	I N E =======================================


sub_9112:							; CODE XREF: sub_907C+74p sub_907C+88p
		ext.w	d0
		bmi.w	loc_9124
		jsr	sub_18AE8
		jsr	sub_11BA4

loc_9124:							; CODE XREF: sub_9112+2j
		addq.w	#1,($FFFFB02A).w
		rts
; End of function sub_9112

; ---------------------------------------------------------------------------
word_912A:	dc.w	6					; DATA XREF: sub_907C+58o
aSh2:		dc.b	'SH2',0
		dc.b	0
		dc.b	4
aSh:		dc.b	'SH',0
		dc.b	4
		dc.b	$20
		dc.b	0
		dc.b	0
		dc.b	4
aPp:		dc.b	'PP',0
		dc.b	6
aPp2:		dc.b	'PP2',0
; ---------------------------------------------------------------------------

loc_9142:							; DATA XREF: ROM:000197C4o
					; ROM:00019950o ...
		moveq	#$A,d0
		moveq	#$19,d1
		bsr.w	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	4
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	5
		dc.b	4
aPenaltySummary: dc.b	'Penalty  Summary'
		dc.b	$BD
		dc.b	7
		dc.b	1
; ---------------------------------------------------------------------------
		move.w	#$2C,d0					; ','
		bsr.w	sub_8078
		addq.w	#4,($FFFFB028).w
		clr.w	d0
		bsr.w	sub_8078
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$34
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	0
		dc.b	8
		dc.b	$F9
		dc.b	1
aPerTimeTmPlayerPenalt: dc.b '^Per^^Time^^Tm^^Player/Penalty^^^^min^^^',0
		dc.b	0
; ---------------------------------------------------------------------------
		clr.l	d0
		move.w	($FFFFC5DC).w,d0
		lsr.w	#2,d0
		mulu.w	#$18,d0
		move.w	d0,($FFFFD5AE).w
		sub.w	#$78,d0					; 'x'
		bpl.w	loc_91E4
		clr.w	d0

loc_91E4:							; CODE XREF: ROM:000091DEj
		move.w	d0,($FFFFD5B2).w
		move.w	($FFFFD5AE).w,d0

loc_91EC:							; CODE XREF: ROM:000091FEj
		bsr.w	sub_80BA
		bsr.w	sub_924A
		move.w	($FFFFD5AE).w,d0
		subq.w	#2,d0
		cmp.w	($FFFFD5B2).w,d0
		bge.s	loc_91EC
		clr.w	($FFFFD5B0).w

loc_9204:							; CODE XREF: ROM:00009230j
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		btst	#7,d3
		bne.w	loc_9CD8
		moveq	#$FFFFFFFE,d0
		btst	#0,d3
		bne.w	loc_9228
		neg.w	d0
		btst	#1,d3
		beq.w	loc_922C

loc_9228:							; CODE XREF: ROM:0000921Aj
		move.w	d0,($FFFFD5B0).w

loc_922C:							; CODE XREF: ROM:00009224j
		bsr.w	sub_9232
		bra.s	loc_9204

; =============== S U B	R O U T	I N E =======================================


sub_9232:							; CODE XREF: ROM:loc_922Cp
		move.w	($FFFFD5B0).w,d0
		beq.w	locret_A9D4
		add.w	($FFFFD5AE).w,d0
		bmi.w	locret_A9D4
		cmp.w	($FFFFD5B2).w,d0
		bgt.w	locret_A9D4
; End of function sub_9232


; =============== S U B	R O U T	I N E =======================================


sub_924A:							; CODE XREF: ROM:000091F0p
		move.w	($FFFFD5AE).w,d1
		move.w	d0,($FFFFD5AE).w
		ext.l	d0
		divs.w	#$18,d0
		swap	d0
		tst.w	d0
		bne.w	loc_9268
		bsr.w	sub_93C4
		clr.w	($FFFFD5B0).w

loc_9268:							; CODE XREF: sub_924A+12j
		ext.l	d1
		divs.w	#$18,d1
		swap	d1
		tst.w	d1
		bne.w	loc_9292
		move.l	d0,-(sp)
		cmpnw	 $16,d0
		bne.w	loc_9284
		bsr.w	sub_92BA

loc_9284:							; CODE XREF: sub_924A+32j
		move.l	(sp)+,d0
		cmpnw	 2,d0
		bne.w	loc_9292
		bsr.w	sub_92E6

loc_9292:							; CODE XREF: sub_924A+28j sub_924A+40j
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		move.w	#$FFB0,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_924A


; =============== S U B	R O U T	I N E =======================================


sub_92BA:							; CODE XREF: sub_924A+36p
		move.l	d0,-(sp)
		swap	d0
		move.w	d0,d3
		asl.w	#2,d3
		jsr	sub_11B92
; End of function sub_92BA

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5AE).w,d0
		lsr.w	#3,d0
		subq.w	#2,d0
		and.w	#$1F,d0
		move.w	d0,($FFFFB02A).w
		bsr.w	sub_9316
		move.l	(sp)+,d0
		rts

; =============== S U B	R O U T	I N E =======================================


sub_92E6:							; CODE XREF: sub_924A+44p
		move.l	d0,-(sp)
		swap	d0
		addq.w	#5,d0
		move.w	d0,d3
		asl.w	#2,d3
		jsr	sub_11B92
; End of function sub_92E6

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5AE).w,d0
		lsr.w	#3,d0
		add.w	#$F,d0
		and.w	#$1F,d0
		move.w	d0,($FFFFB02A).w
		bsr.w	sub_9316
		move.l	(sp)+,d0
		rts

; =============== S U B	R O U T	I N E =======================================


sub_9316:							; CODE XREF: ROM:000092DEp
					; ROM:0000930Ep
		move.w	($FFFFB02A).w,-(sp)
		moveq	#$28,d0					; '('
		moveq	#3,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w
		move.w	#$C5DE,a0
		move.w	(a0,d3.w),d0
		move.w	#1,($FFFFB028).w
		jsr	sub_11C72
		move.w	#$C6CE,a2
		btst	#7,2(a0,d3.w)
		beq.w	loc_9352
		add.w	#$364,a2

loc_9352:							; CODE XREF: sub_9316+34j
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		move.w	#$C,($FFFFB028).w
		jsr	sub_11BA4
		move.b	2(a0,d3.w),d0
		and.w	#$7F,d0					; ''
		move.l	#word_18E0C,a3
		add.w	(a3,d0.w),a3
		clr.w	d0
		move.b	1(a3),d0
		jsr	sub_11D06
		move.w	#$23,($FFFFB028).w			; '#'
		jsr	sub_11BA4
		move.w	#$10,($FFFFB028).w
		clr.w	d0
		move.b	3(a0,d3.w),d0
		jsr	sub_18AE8
		jsr	sub_11BA4
		addq.w	#1,($FFFFB02A).w
		move.w	#$13,($FFFFB028).w
		move.w	#$E000,($FFFFB02C).w
		lea	2(a3),a1
		jmp	sub_11BA4
; End of function sub_9316


; =============== S U B	R O U T	I N E =======================================


sub_93C4:							; CODE XREF: sub_8FB8+16p sub_924A+16p
		movem.l	d0-d1/a1,-(sp)
		jsr	sub_11A36
; End of function sub_93C4

; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	4
		dc.b	9
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		clr.w	d0
		move.w	($FFFFD5AE).w,d1
		tst.w	d1
		sgt	d0
		neg.b	d0
		cmp.w	($FFFFD5B2).w,d1
		slt	d1
		neg.b	d1
		add.b	d1,d0
		add.b	d1,d0
		lea	word_9400(pc),a1
		jsr	sub_13508
		movem.l	(sp)+,d0-d1/a1
		rts
; ---------------------------------------------------------------------------
word_9400:	dc.w	$A					; DATA XREF: ROM:000093F0o
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$11
		dc.b	$20
		dc.b	$F9
		dc.b	0
		dc.w	$A
		dc.b	$7B
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$11
		dc.b	$20
		dc.b	$F9
		dc.b	0
		dc.w	$A
		dc.b	$20
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$11
		dc.b	$7D
		dc.b	$F9
		dc.b	0
		dc.w	$A
		dc.b	$7B
		dc.b	$FB
		dc.b	$FF
		dc.b	$FA
		dc.b	$11
		dc.b	$7D
		dc.b	$F9
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_9428:							; CODE XREF: sub_7CB0+F62Ap
					; DATA XREF: ROM:00019BB4o
		movem.l	a2,-(sp)
		jsr	sub_1833A
		move.w	#$CEF4,a0
		move.w	($FFFFCEEE).w,d0
		move.b	(a0,d0.w),d0
		move.w	#$C6CE,a2
		cmp.w	$28(a2),d0
		beq.w	loc_944E
		add.w	#$364,a2

loc_944E:							; CODE XREF: sub_9428+1Ej
		moveq	#1,d7
		bsr.w	loc_945C
		movem.l	(sp)+,a2
		rts
; End of function sub_9428

; ---------------------------------------------------------------------------

loc_945A:							; DATA XREF: ROM:00019794o
					; ROM:00019920o ...
		clr.w	d7

loc_945C:							; CODE XREF: sub_9428+28p
		moveq	#$D,d0
		moveq	#$16,d1
		bsr.w	loc_9BD8
		clr.w	($FFFFD5AC).w
		clr.w	($FFFFD5AE).w

loc_946C:							; CODE XREF: ROM:0000957Cj
					; ROM:00009584j
		clr.w	($FFFFD5B0).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	5
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1E,d0
		moveq	#6,d1
		jsr	sub_119B8
		tst.w	d7
		beq.w	loc_94AC
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$BD
		dc.b	7
		dc.b	4
aPlayoffStats:	dc.b	'Playoff  Stats'
		dc.b	$BD
		dc.b	$E
		dc.b	1
; ---------------------------------------------------------------------------
		bra.w	loc_94EA
; ---------------------------------------------------------------------------

loc_94AC:							; CODE XREF: ROM:00009488j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$1C
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	$C
		dc.b	$1A
		dc.b	$F9
		dc.b	1
		dc.b	$41
		dc.b	$5E
		dc.b	$2D
		dc.b	$5E
aSwitchTeams:	dc.b	'Switch^Teams'
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$BD
		dc.b	8
		dc.b	4
aPlayerStats:	dc.b	'Player  Stats'
		dc.b	$BD
		dc.b	$E
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------

loc_94EA:							; CODE XREF: ROM:000094A8j
		clr.w	d0
		cmp.w	#$C6CE,a2
		beq.w	loc_94F8
		move.w	#$2C,d0					; ','

loc_94F8:							; CODE XREF: ROM:000094F0j
		bsr.w	sub_8078
		bsr.w	loc_961A
		bsr.w	sub_99C6

loc_9504:							; CODE XREF: ROM:00009550j
		bsr.w	sub_80BA
		bsr.w	sub_A41E
		btst	#7,d3
		bne.w	loc_9CD8
		btst	#6,d1
		bne.w	loc_9570
		bsr.w	sub_112F0
		moveq	#1,d0
		btst	#3,d1
		bne.w	loc_9552
		neg.w	d0
		btst	#2,d1
		bne.w	loc_9552
		moveq	#$FFFFFFFE,d0
		btst	#0,d3
		bne.w	loc_9548
		neg.w	d0
		btst	#1,d3
		beq.w	loc_954C

loc_9548:							; CODE XREF: ROM:0000953Aj
		move.w	d0,($FFFFD5B0).w

loc_954C:							; CODE XREF: ROM:00009544j
					; ROM:0000955Aj ...
		bsr.w	sub_9588
		bra.s	loc_9504
; ---------------------------------------------------------------------------

loc_9552:							; CODE XREF: ROM:00009526j
					; ROM:00009530j
		add.w	($FFFFD5AC).w,d0
		cmpnw	 $FFFF,d0
		blt.s	loc_954C
		cmpnw	 4,d0
		bgt.s	loc_954C
		move.w	d0,($FFFFD5AC).w
		bsr.w	loc_961A
		bsr.w	sub_99C6
		bra.s	loc_954C
; ---------------------------------------------------------------------------

loc_9570:							; CODE XREF: ROM:00009518j
		tst.w	d7
		bne.s	loc_954C
		lea	$364(a2),a2
		cmp.w	#$CA32,a2
		beq.w	loc_946C
		move.w	#$C6CE,a2
		bra.w	loc_946C

; =============== S U B	R O U T	I N E =======================================


sub_9588:							; CODE XREF: ROM:loc_954Cp
		move.w	($FFFFD5B0).w,d0
		beq.w	locret_A9D4
		add.w	($FFFFD5AE).w,d0
		bmi.w	locret_A9D4
		cmp.w	($FFFFD5B2).w,d0
		bgt.w	locret_A9D4
		move.w	($FFFFD5AE).w,d1
		move.w	d0,($FFFFD5AE).w
		and.w	#$F,d0
		bne.w	loc_95B8
		bsr.w	sub_99C6
		clr.w	($FFFFD5B0).w

loc_95B8:							; CODE XREF: sub_9588+24j
		and.w	#$F,d1
		bne.w	loc_95DC
		move.w	d0,-(sp)
		cmpnw	 $E,d0
		bne.w	loc_95CE
		bsr.w	sub_9604

loc_95CE:							; CODE XREF: sub_9588+3Ej
		move.w	(sp)+,d0
		cmpnw	 2,d0
		bne.w	loc_95DC
		bsr.w	sub_960E

loc_95DC:							; CODE XREF: sub_9588+34j sub_9588+4Cj ...
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		move.w	#$FF98,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_9588


; =============== S U B	R O U T	I N E =======================================


sub_9604:							; CODE XREF: sub_9588+42p
		move.w	($FFFFD5AE).w,d3
		lsr.w	#4,d3
		bra.w	sub_9744
; End of function sub_9604


; =============== S U B	R O U T	I N E =======================================


sub_960E:							; CODE XREF: sub_9588+50p
		move.w	($FFFFD5AE).w,d3
		lsr.w	#4,d3
		addq.w	#6,d3
		bra.w	sub_9744
; End of function sub_960E

; ---------------------------------------------------------------------------

loc_961A:							; CODE XREF: ROM:000094FCp
					; ROM:00009566p
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$12
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	4
		dc.b	$B
		dc.b	$F9
		dc.b	1
aPlayer:	dc.b	'Player'
		dc.b	$FB
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		lea	word_990E(pc),a1
		moveq	#5,d3
		tst.w	($FFFFD5AC).w
		bmi.w	loc_9644
		add.w	(a1),a1
		clr.w	d3

loc_9644:							; CODE XREF: ROM:0000963Cj
					; ROM:00009664j
		move.w	#$8000,($FFFFB02C).w
		cmp.w	($FFFFD5AC).w,d3
		bne.w	loc_9658
		move.w	#$C000,($FFFFB02C).w

loc_9658:							; CODE XREF: ROM:0000964Ej
		jsr	sub_11A48
		addq.w	#1,d3
		cmpnw	 5,d3
		blt.s	loc_9644
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	9
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$16,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$8D
		dc.b	$A
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		lea	word_9942(pc),a1
		moveq	#1,d0
		add.w	($FFFFD5AC).w,d0
		jsr	sub_13508
		clr.w	($FFFFB030).w
		move.w	#$C018,a3
		clr.l	d6
		tst.w	($FFFFD5AC).w
		bpl.w	loc_96B6
		bsr.w	sub_9F40
		bset	d0,d6
		subq.w	#1,d6
		not.l	d6

loc_96B6:							; CODE XREF: ROM:000096A8j
					; ROM:00009704j
		moveq	#$FFFFFFFF,d4
		st	d5
		bsr.w	sub_9F9A
		bra.w	loc_96FC
; ---------------------------------------------------------------------------

loc_96C2:							; CODE XREF: ROM:loc_96FCj
		btst	d0,d6
		bne.w	loc_96FC
		clr.w	d2
		move.w	($FFFFD5AC).w,d1
		bmi.w	loc_96D6
		bsr.w	loc_9872

loc_96D6:							; CODE XREF: ROM:000096CEj
		ext.l	d2
		asl.l	#8,d2
		move.l	$1E(a2),a1
		add.w	(a1),a1
		move.w	d0,d1
		subq.w	#8,a1

loc_96E4:							; CODE XREF: ROM:000096E8j
		addq.w	#8,a1
		add.w	(a1),a1
		dbf	d1,loc_96E4
		move.b	#$FF,d2
		sub.b	(a1),d2
		cmp.l	d4,d2
		ble.w	loc_96FC
		move.l	d2,d4
		move.w	d0,d5

loc_96FC:							; CODE XREF: ROM:000096BEj
					; ROM:000096C4j ...
		dbf	d0,loc_96C2
		bset	d5,d6
		move.b	d5,(a3)+
		bpl.s	loc_96B6
		moveq	#5,d0

loc_9708:							; CODE XREF: ROM:0000970Cj
		st	(a3,d0.w)
		dbf	d0,loc_9708
		move.w	a3,d0
		sub.w	#$C01E,d0
		bpl.w	loc_971C
		clr.w	d0

loc_971C:							; CODE XREF: ROM:00009716j
		asl.w	#4,d0
		move.w	d0,($FFFFD5B2).w
		cmp.w	($FFFFD5AE).w,d0
		bcc.w	loc_972E
		move.w	d0,($FFFFD5AE).w

loc_972E:							; CODE XREF: ROM:00009726j
		moveq	#5,d4

loc_9730:							; CODE XREF: ROM:0000973Cj
		move.w	($FFFFD5AE).w,d3
		lsr.w	#4,d3
		add.w	d4,d3
		bsr.w	sub_9744
		dbf	d4,loc_9730
		bra.w	loc_95DC

; =============== S U B	R O U T	I N E =======================================


sub_9744:							; CODE XREF: sub_9604+6j sub_960E+8j ...
		move.w	#$C018,a4
		add.w	d3,a4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	d3,d3
		and.w	#$1F,d3
		move.w	d3,($FFFFB02A).w
		moveq	#$28,d0					; '('
		moveq	#1,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		tst.b	(a4)
		bmi.w	locret_A9D4
		subq.w	#1,($FFFFB02A).w
		move.w	#1,($FFFFB028).w
		move.w	a4,d0
		sub.w	#$C017,d0
		moveq	#2,d1
		move.w	#$E000,($FFFFB02C).w
		jsr	sub_11D3A
		jsr	sub_11BA4
		clr.w	d0
		move.b	(a4),d0
		jsr	sub_18B26
		addq.w	#1,($FFFFB028).w
		move.w	#$8000,($FFFFB02C).w
		jsr	sub_11BA4
		move.w	#$11,($FFFFB028).w
		tst.w	($FFFFD5AC).w
		bmi.w	loc_97FC
		clr.w	d5

loc_97C2:							; CODE XREF: sub_9744+B2j
		clr.w	d0
		move.b	(a4),d0
		move.w	d5,d1
		bsr.w	loc_9872
		move.w	d2,d0
		moveq	#4,d1
		jsr	sub_11D3A
		move.w	#$8000,($FFFFB02C).w
		cmp.w	($FFFFD5AC).w,d5
		bne.w	loc_97EA
		move.w	#$C000,($FFFFB02C).w

loc_97EA:							; CODE XREF: sub_9744+9Cj
		jsr	sub_11BA4
		addq.w	#1,d5
		cmpnw	 5,d5
		bne.s	loc_97C2
		bra.w	locret_9870
; ---------------------------------------------------------------------------

loc_97FC:							; CODE XREF: sub_9744+78j
		clr.w	d0
		move.b	(a4),d0
		moveq	#3,d1
		bsr.w	sub_988E
		clr.w	d0
		move.b	(a4),d0
		moveq	#0,d1
		move.w	d2,-(sp)
		bsr.w	sub_988E
		move.w	(sp)+,d0
		move.w	d0,d3
		sub.w	d2,d0
		bpl.w	loc_981E
		clr.w	d0

loc_981E:							; CODE XREF: sub_9744+D4j
		move.w	d0,d2
		moveq	#4,d1
		jsr	sub_11D3A
		addq.w	#1,($FFFFB028).w
		jsr	sub_11BA4
		move.w	d3,d0
		moveq	#4,d1
		jsr	sub_11D3A
		addq.w	#2,($FFFFB028).w
		jsr	sub_11BA4
		tst.w	d3
		beq.w	loc_9852
		mulu.w	#$64,d2					; 'd'
		divu.w	d3,d2

loc_9852:							; CODE XREF: sub_9744+104j
		move.w	d2,d0
		moveq	#4,d1
		jsr	sub_11D3A
		addq.w	#2,($FFFFB028).w
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'%',0
; ---------------------------------------------------------------------------

locret_9870:							; CODE XREF: sub_9744+B4j
		rts
; End of function sub_9744

; ---------------------------------------------------------------------------

loc_9872:							; CODE XREF: ROM:000096D2p
					; sub_9744+84p
		move.w	d0,-(sp)
		bsr.w	sub_9F40
		move.w	d0,d2
		move.w	(sp)+,d0
		cmp.w	d2,d0
		bge.w	sub_988E
		clr.w	d2
; ---------------------------------------------------------------------------
		dc.b	3					; Not sure, the	data start's somewhere above here...
		dc.b	$3C					; <
		dc.b	0
		dc.b	6
		dc.b	$67
		dc.b	0
		dc.b	$11
		dc.b	$4A
		dc.b	$72
		dc.b	1

; =============== S U B	R O U T	I N E =======================================


sub_988E:							; CODE XREF: sub_9744+BEp sub_9744+CAp ...
		tst.w	d7
		bne.w	loc_98BA
		lea	unk_98FA(pc),a1
		asl.w	#2,d1
		add.w	d1,a1
		move.w	(a1),d1
		add.w	d0,d1
		clr.w	d2
		move.b	(a2,d1.w),d2
		move.w	2(a1),d1
		bmi.w	locret_A9D4
		add.w	d0,d1
		clr.w	d3
		move.b	(a2,d1.w),d3
		add.w	d3,d2
		rts
; ---------------------------------------------------------------------------

loc_98BA:							; CODE XREF: sub_988E+2j
		asl.w	#2,d1
		add.w	d0,d0
		clr.w	d2
		bsr.w	sub_98CE
		addq.w	#2,d1
		bsr.w	sub_98CE
		lsr.w	#1,d0
		rts
; End of function sub_988E


; =============== S U B	R O U T	I N E =======================================


sub_98CE:							; CODE XREF: sub_988E+32p sub_988E+38p
		move.l	#$98E6,a1
		tst.w	(a1,d1.w)
		beq.w	locret_A9D4
		move.w	(a1,d1.w),a1
		add.w	(a1,d0.w),d2
		rts
; End of function sub_98CE

; ---------------------------------------------------------------------------
		dc.b	$CD					; 
		dc.b	$96					; 
		dc.b	0
		dc.b	0
		dc.b	$CD					; 
		dc.b	$CA					; 
		dc.b	0
		dc.b	0
		dc.b	$CD					; 
		dc.b	$96					; 
		dc.b	$CD					; 
		dc.b	$CA					; 
		dc.b	$CD					; 
		dc.b	$FE					; 
		dc.b	0
		dc.b	0
		dc.b	$CE					; 
		dc.b	$32					; 2
		dc.b	0
		dc.b	0
unk_98FA:	dc.b	0					; DATA XREF: sub_988E+6o
		dc.b	$B4					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	$B4					; 
		dc.b	0
		dc.b	$CE					; 
		dc.b	0
		dc.b	$E8					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	2
		dc.b	$FF
		dc.b	$FF
word_990E:	dc.w	$16					; DATA XREF: ROM:00009632o
aSavesShotsSave: dc.b	'Saves Shots Save %  '
		dc.w	6
aG:		dc.b	'   G'
		dc.w	6
aA:		dc.b	'   A'
		dc.w	6
aPts:		dc.b	' Pts'
		dc.w	6
aSog:		dc.b	' SOG'
		dc.w	6
aPim:		dc.b	' PIM'
word_9942:	dc.w	$16					; DATA XREF: ROM:0000968Ao
aGoalieSaves:	dc.b	'    Goalie Saves   ]'
		dc.w	$16
aGoals:		dc.b	'[      Goals       ]'
		dc.w	$16
aAssists:	dc.b	'[     Assists      ]'
		dc.w	$16
aPoints:	dc.b	'[      Points      ]'
		dc.w	$16
aShotsOnGoal:	dc.b	'[  Shots On Goal   ]'
		dc.w	$16
aPenaltyMinutes: dc.b	'[ Penalty Minutes   '

; =============== S U B	R O U T	I N E =======================================


sub_99C6:							; CODE XREF: ROM:00009500p
					; ROM:0000956Ap ...
		movem.l	d0-d1/a1,-(sp)
		jsr	sub_11A36
; End of function sub_99C6

; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	2
		dc.b	$B
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		clr.w	d0
		move.w	($FFFFD5AE).w,d1
		tst.w	d1
		sgt	d0
		neg.b	d0
		cmp.w	($FFFFD5B2).w,d1
		slt	d1
		neg.b	d1
		add.b	d1,d0
		add.b	d1,d0
		lea	word_9A02(pc),a1
		jsr	sub_13508
		movem.l	(sp)+,d0-d1/a1
		rts
; ---------------------------------------------------------------------------
word_9A02:	dc.w	$A					; DATA XREF: ROM:000099F2o
		dc.b	$20
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	$C
		dc.b	$20
		dc.b	$F9					; 
		dc.b	0
		dc.w	$A
		dc.b	$7B					; {
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	$C
		dc.b	$20
		dc.b	$F9					; 
		dc.b	0
		dc.w	$A
		dc.b	$20
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	$C
		dc.b	$7D					; }
		dc.b	$F9					; 
		dc.b	0
		dc.w	$A
		dc.b	$7B					; {
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$FA					; 
		dc.b	$C
		dc.b	$7D					; }
		dc.b	$F9					; 
		dc.b	0
; ---------------------------------------------------------------------------

loc_9A2A:							; DATA XREF: ROM:0001980Co
					; ROM:00019998o ...
		moveq	#9,d0
		moveq	#$1C,d1
		bsr.w	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	6
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1C,d0
		moveq	#6,d1
		bsr.w	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$BD
		dc.b	9
		dc.b	4
aCrowdMeter:	dc.b	'Crowd  Meter'
		dc.b	$BD
		dc.b	7
		dc.b	1
; ---------------------------------------------------------------------------
		move.w	#$2C,d0					; ','
		bsr.w	sub_8078
		addq.w	#4,($FFFFB028).w
		clr.w	d0
		bsr.w	sub_8078
		bsr.w	loc_9A84

loc_9A76:							; CODE XREF: ROM:00009A82j
		bsr.w	sub_9FB8
		btst	#7,d1
		bne.w	loc_9CD8
		bra.s	loc_9A76
; ---------------------------------------------------------------------------

loc_9A84:							; CODE XREF: ROM:00009A72p
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	$C
aCurrentLevel:	dc.b	'Current Level'
		dc.b	$FD
		dc.b	$1C
; ---------------------------------------------------------------------------
		move.w	($FFFFB8A2).w,d0
		bsr.w	sub_9BB4
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$FD
		dc.b	4
		dc.b	$FA
		dc.b	2
aAverageLevel:	dc.b	'Average Level'
		dc.b	$FD
		dc.b	$1C
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	($FFFFB8A8).w,d0
		divu.w	($FFFFB8A6).w,d0
		bsr.w	sub_9BB4
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$12
		dc.b	$FD
		dc.b	4
		dc.b	$FA
		dc.b	2
aPeakLevel:	dc.b	'Peak Level'
		dc.b	$FD
		dc.b	$1C
; ---------------------------------------------------------------------------
		move.w	($FFFFB8A4).w,d0
		bsr.w	sub_9BB4
		movem.l	d0-a6,-(sp)
		tst.w	($FFFFD458).w
		bmi.w	loc_9BAE
		move.w	($FFFFC330).w,d1
		ext.l	d1
		move.l	#$FFFFCF36,a0
		jsr	sub_F9BE2
		move.b	8(a0),d0
		bne.w	loc_9B14
		move.b	#$50,d0					; 'P'

loc_9B14:							; CODE XREF: ROM:00009B0Cj
		and.w	#$FF,d0
		move.w	d0,-(sp)
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$FD
		dc.b	4
		dc.b	$FA
		dc.b	2
aArenaRecord:	dc.b	'Arena Record'
		dc.b	$FD
		dc.b	$1C
; ---------------------------------------------------------------------------
		move.w	(sp)+,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$20
		dc.b	$64
		dc.b	$42
		dc.b	0
; ---------------------------------------------------------------------------
		jsr	sub_FEC5E
		move.w	d0,d1
		ext.l	d1
		move.l	#$FFFFCF36,a0
		jsr	sub_F9BE2
		move.b	8(a0),d0
		bne.w	loc_9B72
		move.w	#$50,d0					; 'P'

loc_9B72:							; CODE XREF: ROM:00009B6Aj
		and.w	#$FF,d0
		move.w	d0,-(sp)
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$FD
		dc.b	4
		dc.b	$FA
		dc.b	2
aLeagueRecord:	dc.b	'League Record'
		dc.b	$FD
		dc.b	$1C
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	(sp)+,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
aDb:		dc.b	' dB',0
; ---------------------------------------------------------------------------

loc_9BAE:							; CODE XREF: ROM:00009AF2j
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_9BB4:							; CODE XREF: ROM:00009AA2p
					; ROM:00009AC8p ...
		ext.l	d0
		asl.w	#2,d0
		bsr.w	sub_110BE
		add.w	#$41,d0					; 'A'
		moveq	#3,d1
		jsr	sub_11D3A
		bsr.w	sub_11BA4
		bsr.w	sub_11B92
; End of function sub_9BB4

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$20
		dc.b	$64
		dc.b	$42
		dc.b	0
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_9BD8:							; CODE XREF: ROM:000080D8p
					; ROM:000082E4p ...
		movem.l	d0-d1/a2,-(sp)
		bsr.w	sub_10F32
		bclr	#0,($FFFFBF78).w
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	($FFFFB002).w,d0
		bsr.w	VDP_SetAddress
		move.l	#0,(a0)
		move.w	#$8C81,4(a0)
		move.w	#6,($FFFFB00E).w
		move.w	#$8D00,4(a0)
		clr.w	d0
		bsr.w	VDP_SetAddress
		move.l	#0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		bclr	#1,($FFFFBF78).w
		move.w	($FFFFB01E).w,d4
		move.l	#unk_55B86,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$9123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	($FFFFB012).w,d4
		jsr	sub_11F04
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#1,d4
		moveq	#0,d5
		bsr.w	sub_1169A
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		movem.l	(sp),d0-d1/a2
		add.w	d0,($FFFFB02A).w
		move.l	#off_54E24,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		move.w	d1,d3
		sub.w	d0,d3
		move.w	d0,d1
		clr.w	d0
		moveq	#$28,d2					; '('
		moveq	#$D,d5
		bsr.w	sub_1169A
		move.w	d4,($FFFFB014).w
		move.l	#unk_AAC5A,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$D123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$40,d0					; '@'
		moveq	#$20,d1					; ' '
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		move.w	#$18,($FFFFBD26).w
		movem.l	(sp)+,d0-d1/a2
		rts
; ---------------------------------------------------------------------------

loc_9CD8:							; CODE XREF: ROM:00008176j
					; ROM:00008894j ...
		bsr.w	sub_10F32

loc_9CDC:							; CODE XREF: sub_9FD0+406p
					; sub_A528+2Cp
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	($FFFFB000).w,d0
		bsr.w	VDP_SetAddress
		move.l	#0,(a0)
		move.w	#$8D00,d0
		move.b	($FFFFB000).w,d0
		lsr.b	#2,d0
		move.w	d0,4(a0)
		move.w	#$8C00,4(a0)
		move.w	#5,($FFFFB00E).w
		move.w	(sp)+,($FFFFBF78).w
		bset	#1,($FFFFBF78).w
		move.w	($FFFFB018).w,d4
		move.l	#unk_56062,a2
		btst	#4,($FFFFC2F4).w
		beq.w	loc_9D32
		move.l	#unk_B5188,a2

loc_9D32:							; CODE XREF: ROM:00009D28j
		bsr.w	sub_11738
		jsr	sub_FEA52
		move.w	($FFFFB012).w,d4
		move.l	#unk_AAC5A,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$4343
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	($FFFFB01E).w,d4
		jsr	sub_11F12
		jsr	loc_16CAC
		jsr	sub_16CC4
		jsr	sub_16CD2
		jsr	sub_16C96
		jmp	sub_1720C
; ---------------------------------------------------------------------------

loc_9D7A:							; DATA XREF: ROM:00019824o
		subq.w	#1,($FFFFCF20).w
		subq.w	#1,($FFFFCF22).w
		move.l	#off_1988C,($FFFFCF24).l
		bset	#2,$30(a2)
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	5
		dc.b	$C
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$16,d0
		moveq	#6,d1
		bsr.w	sub_119B8
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$10
		dc.b	$BD
		dc.b	$C
		dc.b	$E
aTimeout:	dc.b	'Timeout'
		dc.b	$BD
		dc.b	$11
		dc.b	$F
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bsr.w	sub_11BA4
		move.w	#$C6CE,a2
		jsr	sub_13098
		add.w	#$364,a2
		jsr	sub_13098
		moveq	#$78,d0					; 'x'
		bra.w	sub_11176
; ---------------------------------------------------------------------------

loc_9DE6:							; DATA XREF: ROM:0001974Co
					; ROM:000198D8o ...
		move.w	($FFFFCF20).w,-(sp)
		move.w	($FFFFCF22).w,-(sp)
		bsr.w	sub_9F40
		move.w	d0,($FFFFCF22).w
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	4
		dc.b	$C
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$18,d0
		moveq	#3,d1
		add.w	($FFFFCF22).w,d1
		bsr.w	sub_119B8
		move.w	$26(a2),d0
		bpl.w	loc_9E16
		moveq	#$FFFFFFFF,d0

loc_9E16:							; CODE XREF: ROM:00009E10j
		addq.w	#1,d0
		move.w	d0,($FFFFCF20).w

loc_9E1C:							; CODE XREF: ROM:00009E46j
					; ROM:00009E50j ...
		bsr.w	loc_9E86
		bsr.w	sub_9FB8
		btst	#7,d1
		bne.w	loc_9E5E
		btst	#5,d1
		bne.w	loc_9E5E
		btst	#1,d1
		beq.w	loc_9E4C
		move.w	($FFFFCF20).w,d0
		addq.w	#1,d0
		cmp.w	($FFFFCF22).w,d0
		bgt.s	loc_9E1C
		move.w	d0,($FFFFCF20).w

loc_9E4C:							; CODE XREF: ROM:00009E38j
		btst	#0,d1
		beq.s	loc_9E1C
		subq.w	#1,($FFFFCF20).w
		bpl.s	loc_9E1C
		clr.w	($FFFFCF20).w
		bra.s	loc_9E1C
; ---------------------------------------------------------------------------

loc_9E5E:							; CODE XREF: ROM:00009E28j
					; ROM:00009E30j
		move.w	($FFFFCF20).w,d0
		subq.w	#1,d0
		bpl.w	loc_9E72
		cmp.w	#$FFFF,$26(a2)
		blt.w	loc_9E7C

loc_9E72:							; CODE XREF: ROM:00009E64j
		move.w	d0,$26(a2)
		jsr	sub_15788

loc_9E7C:							; CODE XREF: ROM:00009E6Ej
		move.w	(sp)+,($FFFFCF22).w
		move.w	(sp)+,($FFFFCF20).w
		rts
; ---------------------------------------------------------------------------

loc_9E86:							; CODE XREF: ROM:loc_9E1Cp
		move.w	#$D,($FFFFB02A).w
		move.w	($FFFFCF22).w,d1
		moveq	#0,d0

loc_9E92:							; CODE XREF: ROM:00009F3Aj
		move.w	#5,($FFFFB028).w
		move.w	#$A000,($FFFFB02C).w
		cmp.w	($FFFFCF20).w,d0
		bne.w	loc_9EAC
		move.w	#$8000,($FFFFB02C).w

loc_9EAC:							; CODE XREF: ROM:00009EA2j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	'                      '
; ---------------------------------------------------------------------------
		tst.w	d0
		bne.w	loc_9EE8
		move.w	#$B,($FFFFB028).w
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$C
aNoGoalie:	dc.b	'no goalie',0
; ---------------------------------------------------------------------------
		bra.w	loc_9F34
; ---------------------------------------------------------------------------

loc_9EE8:							; CODE XREF: ROM:00009ECAj
		move.l	$1E(a2),a1
		add.w	(a1),a1
		move.w	d0,d2
		subq.w	#1,d2
		bra.w	loc_9EFA
; ---------------------------------------------------------------------------

loc_9EF6:							; CODE XREF: ROM:loc_9EFAj
		add.w	(a1),a1
		addq.w	#8,a1

loc_9EFA:							; CODE XREF: ROM:00009EF2j
		dbf	d2,loc_9EF6
		move.w	#8,($FFFFB028).w
		bsr.w	sub_11BA4
		move.b	(a1),d2
		lsr.b	#4,d2
		add.b	#$30,d2					; '0'
		move.b	d2,($FFFFBFA6).w
		move.b	(a1),d2
		and.b	#$F,d2
		add.b	#$30,d2					; '0'
		move.b	d2,($FFFFBFA7).w
		move.w	#$BFA4,a1
		move.w	#4,(a1)
		move.w	#5,($FFFFB028).w
		bsr.w	sub_11BA4

loc_9F34:							; CODE XREF: ROM:00009EE4j
		addq.w	#1,($FFFFB02A).w
		addq.w	#1,d0
		dbf	d1,loc_9E92
		rts

; =============== S U B	R O U T	I N E =======================================


sub_9F40:							; CODE XREF: ROM:loc_838Ep sub_88C8+Ap ...
		movem.l	d1/a0,-(sp)
		move.l	$1E(a2),a0

loc_9F48:
		add.w	$A(a0),a0
		move.w	(a0),d1
		clr.w	d0

loc_9F50:							; CODE XREF: sub_9F40+14j
		addq.w	#1,d0
		asl.w	#4,d1
		bne.s	loc_9F50
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_9F40


; =============== S U B	R O U T	I N E =======================================


sub_9F5C:							; CODE XREF: ROM:000FAB7Ep
		movem.l	a0,-(sp)
		bsr.s	sub_9F40
		move.w	d0,-(sp)
		move.l	$1E(a2),a0
		add.w	8(a0),a0
		move.b	3(a0),d0
		lsr.w	#4,d0
		and.w	#$F,d0
		add.w	(sp)+,d0
		movem.l	(sp)+,a0
		rts
; End of function sub_9F5C


; =============== S U B	R O U T	I N E =======================================


sub_9F7E:							; CODE XREF: ROM:00008394p
					; sub_88C8+10p	...
		movem.l	a0,-(sp)
		move.l	$1E(a2),a0
		add.w	8(a0),a0
		move.b	3(a0),d0
		lsr.w	#4,d0
		and.w	#$F,d0
		movem.l	(sp)+,a0
		rts
; End of function sub_9F7E


; =============== S U B	R O U T	I N E =======================================


sub_9F9A:							; CODE XREF: ROM:000083AAp
					; ROM:000096BAp ...
		movem.l	a0,-(sp)
		move.l	$1E(a2),a0
		add.w	(a0),a0
		clr.w	d0

loc_9FA6:							; CODE XREF: sub_9F9A+16j
		addq.w	#1,d0
		add.w	(a0),a0
		addq.w	#8,a0
		cmp.w	#2,(a0)
		bne.s	loc_9FA6
		movem.l	(sp)+,a0
		rts
; End of function sub_9F9A


; =============== S U B	R O U T	I N E =======================================


sub_9FB8:							; CODE XREF: ROM:loc_9A76p
					; ROM:00009E20p ...
		move.w	($FFFFB03E).w,d1

loc_9FBC:							; CODE XREF: sub_9FB8+8j
		cmp.w	($FFFFB03E).w,d1
		beq.s	loc_9FBC
		bsr.w	sub_A41E
		bsr.w	sub_112F0
		tst.b	d1
		beq.s	sub_9FB8
		rts
; End of function sub_9FB8


; =============== S U B	R O U T	I N E =======================================


sub_9FD0:							; DATA XREF: ROM:00019698o
					; ROM:00019734o ...
		bclr	#0,($FFFFC2F6).w
		bsr.w	sub_10F32
		move.w	($FFFFBF78).w,-(sp)
		bset	#3,($FFFFC2EE).w
		bclr	#4,($FFFFC2F4).w
		jsr	sub_12B40
		move.w	#$400,d0
		move.w	($FFFFB00C).w,d1
		move.w	#$7FF,d2
		bsr.w	sub_11544
		move.w	($FFFFB026).w,d4
		move.l	#unk_BB4F6,a2
		jsr	sub_11738
		jsr	sub_A448
		bclr	#5,($FFFFC2EC).w
		bclr	#5,($FFFFC2F0).w
		st	($FFFFBE1E).w
		move.l	($FFFFB036).w,a4

loc_A02A:							; CODE XREF: sub_9FD0+60j
		bsr.w	sub_A4F6
		tst.w	d7
		bne.s	loc_A02A
		jsr	sub_1702E
		jsr	sub_15EC0
		bclr	#1,($FFFFC2F0).w
		clr.l	($FFFFBE7A).w
		clr.l	($FFFFBE7E).w
		clr.l	($FFFFBE82).w
		move.w	#$18,($FFFFBD26).w
		moveq	#1,d7

loc_A058:							; CODE XREF: sub_9FD0+92j
					; sub_9FD0+206j ...
		move.w	($FFFFB03E).w,d0
		sub.w	($FFFFB040).w,d0
		cmp.w	d0,d7
		bhi.s	loc_A058
		move.w	($FFFFB03E).w,($FFFFB040).w
		tst.w	($FFFFDEB4).w
		bmi.w	loc_A084
		sub.w	d7,($FFFFDEB4).w
		bpl.w	loc_A084
		clr.w	($FFFFDEB4).w
		jsr	loc_A48A

loc_A084:							; CODE XREF: sub_9FD0+9Ej sub_9FD0+A6j
		movem.l	d0/a3,-(sp)
		move.l	#$FFFFB64A,a3
		move.w	#1,8(a3)
		move.l	#$FFFFB6CA,a3
		move.w	#1,8(a3)
		movem.l	(sp)+,d0/a3
		moveq	#1,d7
		bsr.w	sub_A41E
		btst	#0,($FFFFC2F6).w
		beq.w	loc_A0C4
		and.w	#$FFF0,d0
		or.w	#8,d0
		and.w	#$FFF0,d1
		and.w	#$FFF0,d3

loc_A0C4:							; CODE XREF: sub_9FD0+E0j
		move.w	#$BE08,a5
		btst	#5,($FFFFC2F0).w
		beq.w	loc_A0DC
		move.w	d1,d5
		and.w	#$F,d5
		beq.w	loc_A25E

loc_A0DC:							; CODE XREF: sub_9FD0+FEj
		btst	#3,d0
		bne.w	loc_A25E
		cmp.w	#$FF60,(a5)
		bgt.w	loc_A0F4
		move.w	#0,(a5)
		bra.w	*+4

loc_A0F4:							; CODE XREF: sub_9FD0+118j
		cmp.w	#$FED0,$14(a5)
		bgt.w	loc_A104
		move.w	#0,$14(a5)

loc_A104:							; CODE XREF: sub_9FD0+12Aj
		bset	#5,($FFFFC2EC).w
		bne.w	loc_A118
		move.w	($FFFFBD1C).w,(a5)
		move.w	($FFFFBD18).w,$14(a5)

loc_A118:							; CODE XREF: sub_9FD0+13Aj
		move.w	d0,d5
		and.w	#7,d5
		eor.w	#4,d5
		move.w	#$B04A,a1
		st	d3
		bclr	#5,($FFFFC2F0).w
		beq.w	loc_A136
		move.w	$16(a5),d3

loc_A136:							; CODE XREF: sub_9FD0+15Ej
		move.w	(a5),d0
		move.w	$14(a5),d1
		moveq	#$B,d2
		move.l	#$100,d4
		movem.w	d0-d1,-(sp)

loc_A148:							; CODE XREF: sub_9FD0+1DAj
		cmp.w	$52(a1),d3
		beq.w	loc_A1A6
		movem.w	(sp),d0-d1
		sub.w	(a1),d0
		sub.w	$14(a1),d1
		jsr	sub_10676
		btst	#3,d0
		bne.w	loc_A1A6
		sub.w	d5,d0
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bhi.w	loc_A1A6
		movem.w	(sp),d0-d1
		sub.w	(a1),d0
		sub.w	$14(a1),d1
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmpnl	 4,d0
		bls.w	loc_A1A6
		cmp.l	d4,d0
		bhi.w	loc_A1A6
		move.l	d0,d4
		move.w	$52(a1),$16(a5)
		bset	#5,($FFFFC2F0).w

loc_A1A6:							; CODE XREF: sub_9FD0+17Cj
					; sub_9FD0+194j ...
		add.w	#$80,a1					; ''
		dbf	d2,loc_A148
		addq.w	#4,sp
		btst	#5,($FFFFC2F0).w
		beq.w	loc_A1DA
		move.w	$16(a5),d0
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		moveq	#4,d4
		bsr.w	sub_C656
		bsr.w	sub_A616
		jsr	sub_15EC0
		bra.w	loc_A058
; ---------------------------------------------------------------------------

loc_A1DA:							; CODE XREF: sub_9FD0+1E6j
		bset	#5,($FFFFC2EC).w
		eor.w	#4,d5
		asl.w	#2,d5
		lea	unk_A23E(pc),a0
		move.w	(a0,d5.w),d0
		add.w	(a5),d0
		cmpnw	 $88,d0					; ''
		bgt.w	loc_A202
		cmpnw	 $FF78,d0
		blt.w	loc_A202
		move.w	d0,(a5)

loc_A202:							; CODE XREF: sub_9FD0+224j
					; sub_9FD0+22Cj
		move.w	2(a0,d5.w),d1
		add.w	$14(a5),d1
		cmpnw	 $126,d1
		bgt.w	loc_A21E
		cmpnw	 $FEDA,d1
		blt.w	loc_A21E
		move.w	d1,$14(a5)

loc_A21E:							; CODE XREF: sub_9FD0+23Ej
					; sub_9FD0+246j
		move.w	a5,a3
		clr.w	$18(a5)
		and.w	#$FFF,($FFFFBE78).w
		or.w	#$E000,($FFFFBE78).w
		bsr.w	sub_A616
		jsr	sub_15EC0
		bra.w	loc_A058
; ---------------------------------------------------------------------------
unk_A23E:	dc.b	0					; DATA XREF: sub_9FD0+216o
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FE					; 
		dc.b	0
		dc.b	2
; ---------------------------------------------------------------------------

loc_A25E:							; CODE XREF: sub_9FD0+108j
					; sub_9FD0+110j
		st	$18(a5)
		btst	#5,d1
		beq.w	loc_A27A
		btst	#0,($FFFFC2F6).w
		bne.w	loc_A27A
		bchg	#1,($FFFFC2F0).w

loc_A27A:							; CODE XREF: sub_9FD0+296j
					; sub_9FD0+2A0j
		btst	#6,d3
		beq.w	loc_A2B0
		btst	#0,($FFFFC2F6).w
		bne.w	loc_A2B0
		bsr.w	sub_A4F6
		bsr.w	sub_A4F6
		bsr.w	sub_A4F6
		bsr.w	sub_A4F6
		jsr	sub_1702E
		jsr	sub_15EC0
		moveq	#1,d7
		bclr	#1,($FFFFC2F0).w

loc_A2B0:							; CODE XREF: sub_9FD0+2AEj
					; sub_9FD0+2B8j
		btst	#0,($FFFFC2F6).w
		bne.w	loc_A306
		btst	#4,d3
		beq.w	loc_A306
		btst	#6,d3
		beq.w	loc_A2EE
		bclr	#5,($FFFFC2F0).w
		bclr	#5,($FFFFC2EC).w
		move.w	d0,-(sp)
		move.w	($FFFFBE78).w,d0
		and.w	#$FFF,d0
		or.w	#$E000,d0
		move.w	d0,($FFFFBE78).w
		move.w	(sp)+,d0
		bra.w	loc_A058
; ---------------------------------------------------------------------------

loc_A2EE:							; CODE XREF: sub_9FD0+2F6j
		bsr.w	sub_A528
		jsr	sub_1702E
		jsr	sub_15EC0
		asl.w	#1,d7
		bclr	#1,($FFFFC2F0).w

loc_A306:							; CODE XREF: sub_9FD0+2E6j
					; sub_9FD0+2EEj
		btst	#1,($FFFFC2F0).w
		beq.w	loc_A326
		st	($FFFFC016).w
		bsr.w	sub_A528
		move.w	($FFFFC016).w,-(sp)
		bsr.w	sub_11132
		jsr	sub_1702E

loc_A326:							; CODE XREF: sub_9FD0+33Cj
		jsr	sub_15EC0
		btst	#0,($FFFFC2F6).w
		bne.w	loc_A358
		btst	#7,d1
		beq.w	loc_A058
		jsr	sub_A4A8
		bclr	#1,($FFFFC2F0).w
		bset	#0,($FFFFC2F6).w
		bne.w	loc_A358
		bra.w	loc_A058
; ---------------------------------------------------------------------------

loc_A358:							; CODE XREF: sub_9FD0+362j
					; sub_9FD0+380j
		btst	#7,d1
		bne.w	loc_A39E
		btst	#6,d1
		bne.w	loc_A398
		btst	#4,d1
		bne.w	loc_A3A8
		btst	#5,d1
		beq.w	loc_A058
		bset	#6,($FFFFC2F4).w
		bclr	#0,($FFFFC2F6).w
		jsr	loc_A4D8
		bsr.w	sub_A528
		jsr	sub_1702E
		bra.w	loc_A058
; ---------------------------------------------------------------------------

loc_A398:							; CODE XREF: sub_9FD0+394j
		jmp	loc_A058
; ---------------------------------------------------------------------------

loc_A39E:							; CODE XREF: sub_9FD0+38Cj
		clr.w	($FFFFCF20).w
		bset	#1,($FFFFC2F6).w

loc_A3A8:							; CODE XREF: sub_9FD0+39Cj
		bsr.w	sub_10F32
		or.w	#$F000,($FFFFBE78).w
		bclr	#5,($FFFFC2EC).w
		move.l	($FFFFB036).w,a4
		bclr	#4,($FFFFC2F4).w
		bsr.w	sub_A68C
		jsr	sub_1702E
		bclr	#3,($FFFFC2EE).w
		move.w	(sp)+,($FFFFBF78).w
		jsr	loc_9CDC
		jsr	loc_16CAC
		jsr	sub_16CC4
		jsr	sub_16CD2
		jsr	sub_16CEE
		btst	#4,($FFFFBF78).w
		beq.w	loc_A40A
		jsr	sub_16D0A
		jsr	sub_16CFC

loc_A40A:							; CODE XREF: sub_9FD0+42Aj
		jsr	sub_12B94
		jsr	sub_15EC0
		move.w	#$18,($FFFFBD26).w
		rts
; End of function sub_9FD0


; =============== S U B	R O U T	I N E =======================================


sub_A41E:							; CODE XREF: sub_7CF8+58p
					; ROM:0000816Ep ...
		tst.w	($FFFFD046).w
		bne.w	loc_A434

loc_A426:							; CODE XREF: sub_A41E+1Aj
		btst	#1,($FFFFC2EC).w
		bne.w	sub_11358
		bra.w	sub_11340
; ---------------------------------------------------------------------------

loc_A434:							; CODE XREF: sub_A41E+4j
		tst.w	($FFFFC316).w
		beq.s	loc_A426
		cmp.w	#3,($FFFFC316).w
		beq.w	sub_11370
		bra.w	sub_11388
; End of function sub_A41E


; =============== S U B	R O U T	I N E =======================================


sub_A448:							; CODE XREF: sub_9FD0+40p sub_A528+9Cp
		move.w	#$F0,($FFFFDEB4).w			; ''
		bsr.w	sub_11B92
; End of function sub_A448

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	2
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#$55,d1					; 'U'
		moveq	#8,d2
		moveq	#4,d3
		move.w	($FFFFB018).w,d4
		move.w	#0,d5
		move.l	#off_5605A,a1
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A47E
		move.l	#off_B5180,a1

loc_A47E:							; CODE XREF: ROM:0000A474j
		add.l	4(a1),a1
		move.w	#$30A,a2
		bra.w	sub_1169A
; ---------------------------------------------------------------------------

loc_A48A:							; CODE XREF: sub_9FD0+AEp
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	2
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#4,d1
		move.w	#$7FF,d2
		jmp	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_A4A8:							; CODE XREF: sub_9FD0+36Ep
		st	($FFFFDEB4).w
		bsr.w	sub_11B92
; End of function sub_A4A8

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	2
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#0,d0
		moveq	#0,d1
		moveq	#$10,d2
		moveq	#$B,d3
		move.w	($FFFFB026).w,d4
		move.w	#0,d5
		move.l	#off_BB4EE,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		bra.w	sub_1169A
; ---------------------------------------------------------------------------

loc_A4D8:							; CODE XREF: sub_9FD0+3B4p
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	2
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$10,d0
		move.w	#$B,d1
		move.w	#$7FF,d2
		jmp	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_A4F6:							; CODE XREF: sub_9FD0:loc_A02Ap
					; sub_9FD0+2BCp ...
		cmp.l	#$FFFF0000,a4
		bne.w	loc_A510
		btst	#4,($FFFFC2EC).w
		beq.w	loc_A520
		move.l	#$FFFFAF54,a4

loc_A510:							; CODE XREF: sub_A4F6+6j
		suba.w	#$62,a4					; 'b'
		cmp.l	($FFFFB036).w,a4
		bne.w	sub_A68C
		add.w	#$62,a4					; 'b'

loc_A520:							; CODE XREF: sub_A4F6+10j
		bsr.w	sub_A68C
		clr.w	d7
		rts
; End of function sub_A4F6


; =============== S U B	R O U T	I N E =======================================


sub_A528:							; CODE XREF: sub_9FD0:loc_A2EEp
					; sub_9FD0+344p ...
		bclr	#6,($FFFFC2F4).w
		beq.w	loc_A5D4
		movem.l	d0-a6,-(sp)
		bchg	#4,($FFFFC2F4).w
		move.w	($FFFFBD1C).w,-(sp)
		move.w	($FFFFBD18).w,-(sp)
		bsr.w	sub_10F5C
		jsr	sub_80BA
		jsr	sub_80BA
		jsr	loc_9CDC
		clr.w	($FFFFBD1C).w
		clr.w	($FFFFBD18).w
		move.w	#$7D0,($FFFFBD1A).w
		move.w	#1,d0

loc_A56C:							; CODE XREF: sub_A528+50j
		jsr	sub_15EC0
		jsr	sub_80BA
		dbf	d0,loc_A56C
		move.w	(sp)+,($FFFFBD18).w
		move.w	(sp)+,($FFFFBD1C).w
		move.w	#1,d0
		jsr	sub_15EC0

loc_A58E:							; CODE XREF: sub_A528+6Cj
		jsr	sub_80BA
		dbf	d0,loc_A58E
		movem.l	a0,-(sp)
		move.l	#$FFFFB64A,a0
		move.w	#1,8(a0)
		add.w	#$80,a0					; ''
		move.w	#1,8(a0)
		movem.l	(sp)+,a0
		bsr.w	sub_A67C
		btst	#5,($FFFFC2EC).w
		beq.w	*+4
		jsr	sub_A448
		move.w	#$64,($FFFFBD26).w			; 'd'
		movem.l	(sp)+,d0-a6

loc_A5D4:							; CODE XREF: sub_A528+6j
		clr.w	d7
		btst	#2,($FFFFC2EE).w
		beq.w	loc_A5F4
		movem.l	a4,-(sp)
		bsr.w	sub_A600
		cmp.l	($FFFFB036).w,a4
		movem.l	(sp)+,a4
		beq.w	locret_A9D4

loc_A5F4:							; CODE XREF: sub_A528+B4j
		cmp.l	($FFFFB036).w,a4
		beq.w	locret_A9D4
		bsr.w	sub_A68C
; End of function sub_A528


; =============== S U B	R O U T	I N E =======================================


sub_A600:							; CODE XREF: sub_A528+BCp
		add.w	#$62,a4					; 'b'
		cmp.l	#$FFFFAF54,a4
		bne.w	locret_A9D4
		move.l	#$FFFF0000,a4
		rts
; End of function sub_A600


; =============== S U B	R O U T	I N E =======================================


sub_A616:							; CODE XREF: sub_9FD0+1FCp
					; sub_9FD0+260p ...
		movem.l	d0-d2,-(sp)
		move.w	(a3),d0
		btst	#7,($FFFFC2F4).w
		beq.w	loc_A628
		neg.w	d0

loc_A628:							; CODE XREF: sub_A616+Cj
		move.w	d0,(a5)
		move.w	$14(a3),d1
		btst	#7,($FFFFC2F4).w
		beq.w	loc_A63A
		neg.w	d1

loc_A63A:							; CODE XREF: sub_A616+1Ej
		move.w	d1,$14(a5)
		cmpnw	 $3C,d0					; '<'
		blt.w	loc_A64A
		move.w	#$3C,d0					; '<'

loc_A64A:							; CODE XREF: sub_A616+2Cj
		cmpnw	 $FFC4,d0
		bgt.w	loc_A656
		move.w	#$FFC4,d0

loc_A656:							; CODE XREF: sub_A616+38j
		move.w	d0,($FFFFBD1C).w
		cmpnw	 $100,d1
		blt.w	loc_A666
		move.w	#$100,d1

loc_A666:							; CODE XREF: sub_A616+48j
		cmpnw	 $FF38,d1
		bgt.w	loc_A672
		move.w	#$FF38,d1

loc_A672:							; CODE XREF: sub_A616+54j
		move.w	d1,($FFFFBD18).w
		movem.l	(sp)+,d0-d2
		rts
; End of function sub_A616


; =============== S U B	R O U T	I N E =======================================


sub_A67C:							; CODE XREF: sub_A528+8Ep
		btst	#4,($FFFFC2F4).w
		beq.w	sub_A68C
		bset	#7,($FFFFC2F4).w
; End of function sub_A67C


; =============== S U B	R O U T	I N E =======================================


sub_A68C:							; CODE XREF: sub_9FD0+3F2p
					; sub_A4F6+22j	...
		movem.l	d0-d2/a0/a6,-(sp)
		move.l	#unk_F600E,a6
		move.l	a4,a0
		moveq	#$F,d1
		move.w	#$B04A,a3

loc_A69E:							; CODE XREF: sub_A68C+10Cj
		move.w	2(a0),d2
		and.w	#$3FF,d2
		btst	#9,d2
		beq.w	loc_A6B2
		or.w	#$FC00,d2

loc_A6B2:							; CODE XREF: sub_A68C+1Ej
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A6BE
		neg.w	d2

loc_A6BE:							; CODE XREF: sub_A68C+2Cj
		move.w	d2,(a3)
		move.l	(a0),d2
		asr.l	#4,d2
		asr.w	#6,d2
		btst	#9,d2
		beq.w	loc_A6D2
		or.w	#$FC00,d2

loc_A6D2:							; CODE XREF: sub_A68C+3Ej
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A6E8
		neg.w	d2
		cmpnw	 0,d1
		bne.w	loc_A6E8
		addq.w	#2,d2

loc_A6E8:							; CODE XREF: sub_A68C+4Cj sub_A68C+56j
		move.w	d2,$14(a3)
		move.w	(a0),d2
		asr.w	#4,d2
		and.w	#$3FF,d2
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A738
		asl.w	#1,d2
		move.w	(a6,d2.w),d2
		cmp.w	#$F,$52(a3)
		bne.w	loc_A71C
		cmpnw	 $18A,d2
		beq.w	loc_A71C
		move.w	#$FC00,$14(a3)

loc_A71C:							; CODE XREF: sub_A68C+7Ej sub_A68C+86j
		cmpnw	 1,d2
		blt.w	loc_A730
		cmpnw	 $34E,d2
		bge.w	loc_A730
		bra.w	loc_A738
; ---------------------------------------------------------------------------

loc_A730:							; CODE XREF: sub_A68C+94j sub_A68C+9Cj
		move.w	(a0),d2
		asr.w	#4,d2
		and.w	#$3FF,d2

loc_A738:							; CODE XREF: sub_A68C+6Ej sub_A68C+A0j
		move.w	d2,6(a3)
		cmpnw	 $28D,d2
		blt.w	loc_A75C
		cmpnw	 $292,d2
		bge.w	loc_A75C
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A75C
		move.w	#$190,$14(a3)

loc_A75C:							; CODE XREF: sub_A68C+B4j sub_A68C+BCj ...
		move.w	(a0),d2
		asr.w	#3,d2
		and.w	#$1800,d2
		and.w	#$E7FF,4(a3)
		or.w	d2,4(a3)
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A792
		move.w	6(a3),d2
		cmpnw	 $162,d2
		blt.w	loc_A792
		cmpnw	 $179,d2
		bge.w	loc_A792
		bchg	#3,4(a3)

loc_A792:							; CODE XREF: sub_A68C+E8j sub_A68C+F4j ...
		addq.w	#4,a0
		add.w	#$80,a3					; ''
		dbf	d1,loc_A69E
		moveq	#5,d2
		move.w	#$B04A,a3

loc_A7A2:							; CODE XREF: sub_A68C+15Cj
		move.b	(a0)+,$6F(a3)
		move.b	(a0),d0
		and.w	#$F,d0
		cmpnw	 $F,d0
		bne.w	loc_A7B6
		moveq	#$FFFFFFFF,d0

loc_A7B6:							; CODE XREF: sub_A68C+124j
		move.w	d0,$34(a3)
		add.w	#$80,a3					; ''
		move.b	(a0)+,d0
		lsr.b	#4,d0
		move.b	d0,$6F(a3)
		move.b	(a0),d0
		asl.b	#4,d0
		or.b	d0,$6F(a3)
		move.b	(a0)+,d0
		lsr.b	#4,d0
		and.w	#$F,d0
		cmpnw	 $F,d0
		bne.w	loc_A7E0
		moveq	#$FFFFFFFF,d0

loc_A7E0:							; CODE XREF: sub_A68C+14Ej
		move.w	d0,$34(a3)
		add.w	#$80,a3					; ''
		dbf	d2,loc_A7A2
		move.b	(a0)+,d0
		ext.w	d0
		move.w	d0,($FFFFB762).w
		move.b	(a0)+,d0
		ext.w	d0
		move.w	d0,($FFFFB7E2).w
		move.b	(a0)+,d0
		ext.w	d0
		move.w	d0,($FFFFC016).w
		clr.w	d7
		move.b	(a0)+,d7
		move.w	(a0)+,d0
		and.w	#$FFF,d0
		and.w	#$F000,($FFFFBE78).w
		or.w	d0,($FFFFBE78).w
		move.w	(a0)+,($FFFFB89A).w
		move.w	(a0)+,($FFFFBEE2).w
		move.b	(a0)+,($FFFFC3EA).w
		clr.w	d0
		move.b	(a0)+,d0
		lsl.w	#4,d0
		or.w	#$F00F,d0
		move.w	d0,($FFFFBE86).w
		btst	#5,($FFFFC2EC).w
		beq.w	loc_A860
		move.w	a5,a3
		btst	#5,($FFFFC2F0).w
		beq.w	loc_A858
		clr.w	$18(a5)
		move.w	$16(a5),d0
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3

loc_A858:							; CODE XREF: sub_A68C+1B8j
		bsr.w	sub_A616
		bra.w	loc_A880
; ---------------------------------------------------------------------------

loc_A860:							; CODE XREF: sub_A68C+1ACj
		move.w	(a0)+,($FFFFBD1C).w
		move.w	(a0)+,($FFFFBD18).w
		btst	#4,($FFFFC2F4).w
		beq.w	loc_A880
		neg.w	($FFFFBD1C).w
		neg.w	($FFFFBD18).w
		jsr	sub_A88C

loc_A880:							; CODE XREF: sub_A68C+1D0j
					; sub_A68C+1E2j
		bclr	#7,($FFFFC2F4).w
		movem.l	(sp)+,d0-d2/a0/a6
		rts
; End of function sub_A68C


; =============== S U B	R O U T	I N E =======================================


sub_A88C:							; CODE XREF: sub_A68C+1EEp
		move.w	d0,-(sp)
		move.w	#$3C,d0					; '<'
		cmp.w	($FFFFBD1C).w,d0
		blt.w	loc_A8A6
		move.w	#$FFC4,d0
		cmp.w	($FFFFBD1C).w,d0
		ble.w	loc_A8AA

loc_A8A6:							; CODE XREF: sub_A88C+Aj
		move.w	d0,($FFFFBD1C).w

loc_A8AA:							; CODE XREF: sub_A88C+16j
		move.w	#$100,d0
		cmp.w	($FFFFBD18).w,d0
		blt.w	loc_A8C2
		move.w	#$FF38,d0
		cmp.w	($FFFFBD18).w,d0
		ble.w	loc_A8C6

loc_A8C2:							; CODE XREF: sub_A88C+26j
		move.w	d0,($FFFFBD18).w

loc_A8C6:							; CODE XREF: sub_A88C+32j
		move.w	(sp)+,d0
		rts
; End of function sub_A88C


; =============== S U B	R O U T	I N E =======================================


sub_A8CA:							; CODE XREF: sub_794E+46p
		btst	#4,($FFFFC2EA).w
		bne.w	locret_A9D4
		btst	#2,($FFFFC2EE).w
		beq.w	loc_A8F6
		tst.w	($FFFFDEF0).w
		beq.w	loc_A918
		subq.w	#1,($FFFFDEF0).w
		bpl.w	loc_A8F6
		clr.w	($FFFFDEF0).w
		bra.w	loc_A918
; ---------------------------------------------------------------------------

loc_A8F6:							; CODE XREF: sub_A8CA+10j sub_A8CA+20j
		add.l	#$62,($FFFFB036).w			; 'b'
		cmp.l	#$FFFFAF54,($FFFFB036).w
		bne.w	loc_A918
		bset	#4,($FFFFC2EC).w
		move.l	#$FFFF0000,($FFFFB036).w

loc_A918:							; CODE XREF: sub_A8CA+18j sub_A8CA+28j ...
		move.l	($FFFFB036).w,a0
		moveq	#$F,d2
		move.w	#$B04A,a3

loc_A922:							; CODE XREF: sub_A8CA+8Aj
		clr.l	(a0)
		move.w	(a3),d1
		and.w	#$3FF,d1
		move.w	d1,2(a0)
		clr.l	d1
		move.w	$14(a3),d1
		asl.w	#6,d1
		asl.l	#4,d1
		or.l	d1,(a0)
		move.w	6(a3),d1
		asl.w	#4,d1
		or.w	d1,(a0)
		move.w	4(a3),d1
		and.w	#$1800,d1
		asl.w	#3,d1
		or.w	d1,(a0)
		addq.w	#4,a0
		add.w	#$80,a3					; ''
		dbf	d2,loc_A922
		moveq	#5,d2
		move.w	#$B04A,a3

loc_A95E:							; CODE XREF: sub_A8CA+CEj
		move.b	$6F(a3),(a0)+
		move.w	$34(a3),d0
		bpl.w	loc_A96C
		moveq	#$F,d0

loc_A96C:							; CODE XREF: sub_A8CA+9Cj
		and.w	#$F,d0
		move.b	d0,(a0)
		add.w	#$80,a3					; ''
		move.b	$6F(a3),d0
		asl.w	#4,d0
		or.b	d0,(a0)+
		move.b	$6F(a3),d0
		lsr.b	#4,d0
		move.b	d0,(a0)
		move.w	$34(a3),d0
		bpl.w	loc_A990
		moveq	#$F,d0

loc_A990:							; CODE XREF: sub_A8CA+C0j
		asl.w	#4,d0
		or.b	d0,(a0)+
		add.w	#$80,a3					; ''
		dbf	d2,loc_A95E
		move.b	($FFFFB763).w,(a0)+
		move.b	($FFFFB7E3).w,(a0)+
		move.b	($FFFFC017).w,(a0)+
		bset	#7,($FFFFC017).w
		move.b	d7,(a0)+
		move.w	($FFFFBE78).w,(a0)+
		move.w	($FFFFB89A).w,(a0)+
		move.w	($FFFFBEE2).w,(a0)+
		move.b	($FFFFC3EA).w,(a0)+
		move.w	d0,-(sp)
		move.w	($FFFFBE86).w,d0
		lsr.w	#4,d0
		move.b	d0,(a0)+
		move.w	(sp)+,d0
		move.w	($FFFFBD1C).w,(a0)+
		move.w	($FFFFBD18).w,(a0)+

locret_A9D4:							; CODE XREF: sub_799E+20j sub_799E+28j ...
		rts
; End of function sub_A8CA


; =============== S U B	R O U T	I N E =======================================


sub_A9D6:							; CODE XREF: sub_794E+14p
		btst	#7,($FFFFC2EC).w
		bne.w	loc_AA22
		or.w	#$F,($FFFFBE78).w
		cmp.w	#5,($FFFFB762).w
		bgt.w	loc_A9F8

loc_A9F0:							; CODE XREF: sub_A9D6+34j
		clr.w	($FFFFDED2).w
		bra.w	loc_AA22
; ---------------------------------------------------------------------------

loc_A9F8:							; CODE XREF: sub_A9D6+16j
		move.w	($FFFFDED6).w,d0
		cmp.w	($FFFFB762).w,d0
		beq.w	loc_AA0C
		move.w	($FFFFB762).w,($FFFFDED6).w
		bra.s	loc_A9F0
; ---------------------------------------------------------------------------

loc_AA0C:							; CODE XREF: sub_A9D6+2Aj
		addq.w	#1,($FFFFDED2).w
		cmp.w	#$10,($FFFFDED2).w
		blt.w	loc_AA22
		clr.w	($FFFFDED2).w
		subq.w	#1,($FFFFB762).w

loc_AA22:							; CODE XREF: sub_A9D6+6j sub_A9D6+1Ej	...
		bclr	#6,($FFFFC2FC).w
		move.w	($FFFFC30E).w,d0
		cmp.w	($FFFFC310).w,d0
		blt.w	loc_AA3A
		bset	#6,($FFFFC2FC).w

loc_AA3A:							; CODE XREF: sub_A9D6+5Aj
		jsr	sub_FEF66
		tst.w	($FFFFB7AA).w
		bmi.w	loc_AA72
		bclr	#7,($FFFFC2FE).w
		bclr	#1,($FFFFC2F8).w
		move.w	($FFFFB7AA).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a3
		add.w	d0,a3
		btst	#1,$64(a3)
		beq.w	loc_AA72
		bclr	#4,($FFFFC2FA).w

loc_AA72:							; CODE XREF: sub_A9D6+6Ej sub_A9D6+92j
		move.w	#$B04A,a3

loc_AA76:							; CODE XREF: sub_A9D6+508j
		move.l	(a3),$1C(a3)
		move.l	$14(a3),$20(a3)
		move.l	$18(a3),$24(a3)
		btst	#5,$64(a3)
		beq.w	loc_AB86
		clr.w	$28(a3)
		clr.w	$2A(a3)
		tst.w	$5A(a3)
		bne.w	loc_AAAA
		jsr	sub_FEB54
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AAAA:							; CODE XREF: sub_A9D6+C6j
		cmp.w	#$193E,$58(a3)
		beq.w	loc_AB74
		cmp.w	#$1A00,$58(a3)
		beq.w	loc_AB48
		cmp.w	#$18CC,$58(a3)
		beq.w	loc_AB5E
		cmp.w	#$17E8,$58(a3)
		beq.w	loc_AB32
		cmp.w	#$1776,$58(a3)
		beq.w	loc_AAEA
		cmp.w	#$185A,$58(a3)
		beq.w	loc_AB0E
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AAEA:							; CODE XREF: sub_A9D6+102j
		move.w	#$124,d0
		cmp.w	#$56,($FFFFD41C).w			; 'V'
		bgt.w	loc_AB02
		cmp.w	#$FFAA,($FFFFD41C).w
		bgt.w	loc_AB06

loc_AB02:							; CODE XREF: sub_A9D6+11Ej
		move.w	#$116,d0

loc_AB06:							; CODE XREF: sub_A9D6+128j
		move.w	d0,$14(a3)
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AB0E:							; CODE XREF: sub_A9D6+10Cj
		move.w	#$FEDC,d0
		cmp.w	#$56,($FFFFD41C).w			; 'V'
		bgt.w	loc_AB26
		cmp.w	#$FFAA,($FFFFD41C).w
		bgt.w	loc_AB2A

loc_AB26:							; CODE XREF: sub_A9D6+142j
		move.w	#$FEEA,d0

loc_AB2A:							; CODE XREF: sub_A9D6+14Cj
		move.w	d0,$14(a3)
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AB32:							; CODE XREF: sub_A9D6+F8j
		move.w	#$82,(a3)				; ''
		btst	#3,4(a3)
		beq.w	loc_AB86
		move.w	#$FF7E,(a3)
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AB48:							; CODE XREF: sub_A9D6+E4j
		move.w	#$88,(a3)				; ''
		btst	#3,4(a3)
		beq.w	loc_AB86
		move.w	#$FF78,(a3)
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AB5E:							; CODE XREF: sub_A9D6+EEj
		move.w	#$FF7E,(a3)
		btst	#3,4(a3)
		beq.w	loc_AB86
		move.w	#$82,(a3)				; ''
		bra.w	loc_AB86
; ---------------------------------------------------------------------------

loc_AB74:							; CODE XREF: sub_A9D6+DAj
		move.w	#$FF78,(a3)
		btst	#3,4(a3)
		beq.w	loc_AB86
		move.w	#$88,(a3)				; ''

loc_AB86:							; CODE XREF: sub_A9D6+B6j sub_A9D6+D0j ...
		tst.w	$34(a3)
		bmi.w	loc_AECE
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_ABA6
		bclr	#1,$64(a3)
		bclr	#0,$64(a3)

loc_ABA6:							; CODE XREF: sub_A9D6+1C0j
		bsr.w	sub_AEE4
		sub.b	d7,$5E(a3)
		bpl.w	loc_ABB6
		clr.b	$5E(a3)

loc_ABB6:							; CODE XREF: sub_A9D6+1D8j
		sub.b	d7,$5F(a3)
		bpl.w	loc_ABC2
		clr.b	$5F(a3)

loc_ABC2:							; CODE XREF: sub_A9D6+1E4j
		btst	#6,($FFFFC2F0).w
		beq.w	loc_ABE6
		clr.w	d0
		move.b	$66(a3),d0
		bmi.w	loc_ABE6
		add.w	d0,d0
		add.w	#$136,d0
		jsr	sub_13040
		addq.w	#1,(a2,d0.w)

loc_ABE6:							; CODE XREF: sub_A9D6+1F2j
					; sub_A9D6+1FCj
		cmp.w	#$145C,$58(a3)
		beq.w	loc_ACBA
		cmp.w	#$1AF4,$58(a3)
		beq.w	loc_ACBA
		moveq	#$11,d4
		tst.w	($FFFFD06E).w
		beq.w	loc_AC06
		moveq	#$16,d4

loc_AC06:							; CODE XREF: sub_A9D6+22Aj
		mulu.w	d7,d4
		tst.w	$18(a3)
		bne.w	loc_AC46
		moveq	#6,d2
		btst	#0,$62(a3)
		beq.w	loc_AC1E
		moveq	#9,d2

loc_AC1E:							; CODE XREF: sub_A9D6+242j
		move.w	$28(a3),d0
		beq.w	loc_AC32
		asr.w	d2,d0
		bne.w	loc_AC2E
		moveq	#1,d0

loc_AC2E:							; CODE XREF: sub_A9D6+252j
		sub.w	d0,$28(a3)

loc_AC32:							; CODE XREF: sub_A9D6+24Cj
		move.w	$2A(a3),d0
		beq.w	loc_AC46
		asr.w	d2,d0
		bne.w	loc_AC42
		moveq	#1,d0

loc_AC42:							; CODE XREF: sub_A9D6+266j
		sub.w	d0,$2A(a3)

loc_AC46:							; CODE XREF: sub_A9D6+236j
					; sub_A9D6+260j
		move.w	$28(a3),d0
		beq.w	loc_AC52
		muls.w	d4,d0
		add.l	d0,(a3)

loc_AC52:							; CODE XREF: sub_A9D6+274j
		move.w	$2A(a3),d0
		beq.w	loc_AC60
		muls.w	d4,d0
		add.l	d0,$14(a3)

loc_AC60:							; CODE XREF: sub_A9D6+280j
		tst.w	$18(a3)
		bmi.w	loc_ACBA
		bne.w	loc_AC74
		tst.w	$2C(a3)
		beq.w	loc_ACBA

loc_AC74:							; CODE XREF: sub_A9D6+292j
		asl.w	#1,d4
		sub.w	d4,$2C(a3)
		asl.w	#1,d4
		sub.w	d4,$2C(a3)
		lsr.w	#2,d4
		move.w	$2C(a3),d0
		muls.w	d4,d0
		add.l	d0,$18(a3)
		bpl.w	loc_ACBA
		clr.l	$18(a3)
		neg.w	$2C(a3)
		asr	$2C(a3)
		moveq	#5,d0
		sub.b	$2C(a3),d0
		bpl.w	loc_ACA8
		clr.w	d0

loc_ACA8:							; CODE XREF: sub_A9D6+2CCj
		cmpnw	 3,d0
		bhi.w	loc_ACBA
		add.w	#$2C,d0					; ','
		move.w	d0,-(sp)
		bsr.w	sub_11132

loc_ACBA:							; CODE XREF: sub_A9D6+216j
					; sub_A9D6+220j ...
		move.w	$52(a3),d6
		cmp.w	($FFFFB7AA).w,d6
		bne.w	loc_ACE0
		btst	#0,$63(a3)
		bne.w	loc_ACE0
		btst	#7,($FFFFC2EC).w
		bne.w	loc_ACE0
		moveq	#$FFFFFFFE,d4
		bsr.w	sub_C656

loc_ACE0:							; CODE XREF: sub_A9D6+2ECj
					; sub_A9D6+2F6j ...
		btst	#0,($FFFFC2EE).w
		bne.w	loc_AD0E
		tst.w	($FFFFD412).w
		beq.w	loc_AD0E
		subq.w	#1,($FFFFD412).w
		beq.w	loc_ACFE
		bpl.w	loc_AD0E

loc_ACFE:							; CODE XREF: sub_A9D6+320j
		clr.w	($FFFFD412).w
		move.w	#$1111,($FFFFBF06).w
		move.w	#$1111,($FFFFBF08).w

loc_AD0E:							; CODE XREF: sub_A9D6+310j
					; sub_A9D6+318j ...
		cmp.w	($FFFFC320).w,d6
		bne.w	loc_AD2A
		bsr.w	sub_11340
		clr.w	d4
		move.w	#$FFFF,($FFFFC318).w
		bsr.w	sub_B0E8
		bra.w	loc_AD42
; ---------------------------------------------------------------------------

loc_AD2A:							; CODE XREF: sub_A9D6+33Cj
		cmp.w	($FFFFC322).w,d6
		bne.w	loc_AD42
		bsr.w	sub_11358
		moveq	#2,d4
		move.w	#$FFFF,($FFFFC318).w
		bsr.w	sub_B0E8

loc_AD42:							; CODE XREF: sub_A9D6+350j
					; sub_A9D6+358j
		cmpnw	 $B,d6
		bgt.w	loc_AE7C
		tst.w	($FFFFC32C).w
		beq.w	loc_ADE2
		cmp.w	($FFFFC324).w,d6
		bne.w	loc_ADE2
		jsr	sub_11370
		clr.l	d4
		clr.w	($FFFFC318).w
		move.w	($FFFFC328).w,-(sp)
		move.w	($FFFFC32C).w,($FFFFC328).w
		move.w	($FFFFC320).w,-(sp)
		move.w	($FFFFC324).w,($FFFFC320).w
		move.w	($FFFFBF06).w,-(sp)
		move.w	($FFFFBF08).w,($FFFFBF06).w
		move.w	($FFFFBF0E).w,-(sp)
		move.w	($FFFFBF0A).w,($FFFFBF0E).w
		move.w	($FFFFD41E).w,-(sp)
		move.w	($FFFFD420).w,($FFFFD41E).w
		move.w	($FFFFBE78).w,-(sp)
		move.w	($FFFFBE86).w,($FFFFBE78).w
		bsr.w	sub_B0E8
		move.w	($FFFFBE78).w,($FFFFBE86).w
		move.w	(sp)+,($FFFFBE78).w
		move.w	($FFFFD41E).w,($FFFFD420).w
		move.w	(sp)+,($FFFFD41E).w
		move.w	($FFFFBF0E).w,($FFFFBF0A).w
		move.w	(sp)+,($FFFFBF0E).w
		move.w	($FFFFBF06).w,($FFFFBF08).w
		move.w	(sp)+,($FFFFBF06).w
		move.w	($FFFFC320).w,($FFFFC324).w
		move.w	(sp)+,($FFFFC320).w
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	(sp)+,($FFFFC328).w

loc_ADE2:							; CODE XREF: sub_A9D6+378j
					; sub_A9D6+380j
		tst.w	($FFFFC32E).w
		beq.w	loc_AE7C
		cmp.w	($FFFFC326).w,d6
		bne.w	loc_AE7C
		jsr	sub_11388
		moveq	#2,d4
		move.w	#1,($FFFFC318).w
		move.w	($FFFFC32A).w,-(sp)
		move.w	($FFFFC32E).w,($FFFFC32A).w
		move.w	($FFFFC322).w,-(sp)
		move.w	($FFFFC326).w,($FFFFC322).w
		move.w	($FFFFBF06).w,-(sp)
		move.w	($FFFFBF08).w,($FFFFBF06).w
		move.w	($FFFFBF10).w,-(sp)
		move.w	($FFFFBF0C).w,($FFFFBF10).w
		move.w	($FFFFD41E).w,-(sp)
		move.w	($FFFFD420).w,($FFFFD41E).w
		move.w	($FFFFBE78).w,-(sp)
		move.w	($FFFFBE86).w,($FFFFBE78).w
		bsr.w	sub_B0E8
		move.w	($FFFFBE78).w,($FFFFBE86).w
		move.w	(sp)+,($FFFFBE78).w
		move.w	($FFFFD41E).w,($FFFFD420).w
		move.w	(sp)+,($FFFFD41E).w
		move.w	($FFFFBF10).w,($FFFFBF0C).w
		move.w	(sp)+,($FFFFBF10).w
		move.w	($FFFFBF06).w,($FFFFBF08).w
		move.w	(sp)+,($FFFFBF06).w
		move.w	($FFFFC322).w,($FFFFC326).w
		move.w	(sp)+,($FFFFC322).w
		move.w	($FFFFC32A).w,($FFFFC32E).w
		move.w	(sp)+,($FFFFC32A).w

loc_AE7C:							; CODE XREF: sub_A9D6+370j
					; sub_A9D6+410j ...
		move.w	$36(a3),d0
		clr.w	d1
		move.b	$38(a3,d0.w),d1
		asl.w	#2,d1
		move.l	#ObjListMaybe,a0
		move.l	(a0,d1.w),a0
		jsr	sub_13040
		jsr	(a0)
		clr.w	$4E(a3)
		clr.w	$50(a3)
		move.w	(a3),d2
		move.w	$14(a3),d3
		cmp.w	$1C(a3),d2
		bne.w	loc_AEB8
		cmp.w	$20(a3),d3
		beq.w	loc_AEBE

loc_AEB8:							; CODE XREF: sub_A9D6+4D6j
		jsr	sub_138AC

loc_AEBE:							; CODE XREF: sub_A9D6+4DEj
		move.w	d7,d0
		asl.w	#1,d0
		sub.w	d0,$32(a3)
		bpl.w	loc_AECE
		clr.w	$32(a3)

loc_AECE:							; CODE XREF: sub_A9D6+1B4j
					; sub_A9D6+4F0j
		move.w	$32(a3),$30(a3)
		add.w	#$80,a3					; ''
		cmp.w	#$F,-$2E(a3)
		blt.w	loc_AA76
		rts
; End of function sub_A9D6


; =============== S U B	R O U T	I N E =======================================


sub_AEE4:							; CODE XREF: sub_A9D6:loc_ABA6p
		tst.w	$58(a3)
		bne.w	loc_AEFA
		bclr	#5,$62(a3)
		bclr	#1,$63(a3)
		rts
; ---------------------------------------------------------------------------

loc_AEFA:							; CODE XREF: sub_AEE4+4j
		move.l	#$5B1C,a0
		add.w	$58(a3),a0
		move.w	$10(a0),d1
		move.w	$54(a3),d0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_AF26
		cmp.w	#$C,$52(a3)
		bge.w	loc_AF26
		subq.w	#2,d0
		and.w	#7,d0

loc_AF26:							; CODE XREF: sub_AEE4+2Ej sub_AEE4+38j
		btst	#3,4(a3)
		beq.w	loc_AF38
		neg.w	d0
		addq.w	#8,d0
		and.w	#7,d0

loc_AF38:							; CODE XREF: sub_AEE4+48j
		asl.w	#1,d0
		add.w	(a0,d0.w),a0
		move.w	$5A(a3),d0
		move.w	(a0,d0.w),d2
		tst.w	$5C(a3)
		bmi.w	loc_AF88
		sub.w	d7,$5C(a3)
		bpl.w	loc_AF96
		addq.w	#4,$5A(a3)
		addq.w	#4,d0
		tst.w	-2(a0,d0.w)
		bpl.w	loc_AF88
		clr.w	d0
		clr.w	$5A(a3)
		bclr	#5,$62(a3)
		bclr	#1,$63(a3)
		bclr	#5,$64(a3)
		btst	#0,d1
		bne.w	loc_AF88
		clr.w	$58(a3)

loc_AF88:							; CODE XREF: sub_AEE4+66j sub_AEE4+7Cj ...
		move.w	2(a0,d0.w),d0
		bpl.w	loc_AF92
		neg.w	d0

loc_AF92:							; CODE XREF: sub_AEE4+A8j
		move.w	d0,$5C(a3)

loc_AF96:							; CODE XREF: sub_AEE4+6Ej
		sub.b	d7,$65(a3)
		bpl.w	locret_AFB4
		clr.b	$65(a3)
		cmp.w	6(a3),d2
		beq.w	locret_AFB4
		move.w	d2,6(a3)
		move.b	#4,$65(a3)

locret_AFB4:							; CODE XREF: sub_AEE4+B6j sub_AEE4+C2j
		rts
; End of function sub_AEE4


; =============== S U B	R O U T	I N E =======================================


sub_AFB6:							; CODE XREF: sub_7B30+28p sub_F37Cp ...
		move.w	($FFFFBD18).w,($FFFFBF90).w
		move.w	($FFFFBD1C).w,($FFFFBF8E).w
		bset	#6,($FFFFC2EC).w
		rts
; End of function sub_AFB6


; =============== S U B	R O U T	I N E =======================================


sub_AFCA:							; CODE XREF: sub_794E+42p
					; sub_EF92:loc_F06Cp ...
		move.w	($FFFFBF90).w,d2
		move.w	($FFFFBF8E).w,d3
		btst	#6,($FFFFC2EC).w
		bne.w	loc_B042
		move.w	#$B74A,a3
		move.w	($FFFFB7AA).w,d0
		bmi.w	loc_B02A
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		move.w	d7,d0
		add.w	d0,d0
		btst	#7,$62(a3)
		beq.w	loc_B016
		add.w	d0,($FFFFBF92).w
		cmp.w	#$32,($FFFFBF92).w			; '2'
		blt.w	loc_B02A
		move.w	#$32,($FFFFBF92).w			; '2'
		bra.w	loc_B02A
; ---------------------------------------------------------------------------

loc_B016:							; CODE XREF: sub_AFCA+30j
		sub.w	d0,($FFFFBF92).w
		cmp.w	#$FFCE,($FFFFBF92).w

loc_B020:
		bgt.w	loc_B02A
		move.w	#$FFCE,($FFFFBF92).w

loc_B02A:							; CODE XREF: sub_AFCA+1Aj sub_AFCA+3Ej ...
		move.w	$2A(a3),d2
		asr.w	#7,d2
		add.w	$14(a3),d2
		add.w	($FFFFBF92).w,d2
		move.w	(a3),d3
		move.w	d2,($FFFFBF90).w
		move.w	d3,($FFFFBF8E).w

loc_B042:							; CODE XREF: sub_AFCA+Ej
		move.w	d2,d0
		sub.w	($FFFFBD18).w,d0
		cmpnw	 $FFF6,d0
		bge.w	loc_B066
		move.w	d2,d1
		sub.w	#$FFF6,d1
		cmpnw	 $FF38,d1
		bgt.w	loc_B080
		move.w	#$FF38,d1
		bra.w	loc_B080
; ---------------------------------------------------------------------------

loc_B066:							; CODE XREF: sub_AFCA+82j
		cmpnw	 $A,d0
		ble.w	loc_B094
		move.w	d2,d1
		sub.w	#$A,d1
		cmpnw	 $100,d1
		blt.w	loc_B080
		move.w	#$100,d1

loc_B080:							; CODE XREF: sub_AFCA+90j sub_AFCA+98j ...
		sub.w	($FFFFBD18).w,d1
		beq.w	loc_B094
		asr.w	#4,d1
		bne.w	loc_B090
		addq.w	#1,d1

loc_B090:							; CODE XREF: sub_AFCA+C0j
		add.w	d1,($FFFFBD18).w

loc_B094:							; CODE XREF: sub_AFCA+A0j sub_AFCA+BAj
		move.w	d3,d0
		sub.w	($FFFFBD1C).w,d0
		cmpnw	 $FFD8,d0
		bge.w	loc_B0B8
		move.w	d3,d1
		sub.w	#$FFD8,d1
		cmpnw	 $FFC4,d1
		bge.w	loc_B0D2
		move.w	#$FFC4,d1
		bra.w	loc_B0D2
; ---------------------------------------------------------------------------

loc_B0B8:							; CODE XREF: sub_AFCA+D4j
		cmpnw	 $28,d0					; '('
		ble.w	locret_B0E6
		move.w	d3,d1
		sub.w	#$28,d1					; '('
		cmpnw	 $3C,d1					; '<'
		ble.w	loc_B0D2
		move.w	#$3C,d1					; '<'

loc_B0D2:							; CODE XREF: sub_AFCA+E2j sub_AFCA+EAj ...
		sub.w	($FFFFBD1C).w,d1
		beq.w	locret_B0E6
		asr.w	#4,d1
		bne.w	loc_B0E2
		addq.w	#1,d1

loc_B0E2:							; CODE XREF: sub_AFCA+112j
		add.w	d1,($FFFFBD1C).w

locret_B0E6:							; CODE XREF: sub_AFCA+F2j
					; sub_AFCA+10Cj
		rts
; End of function sub_AFCA


; =============== S U B	R O U T	I N E =======================================


sub_B0E8:							; CODE XREF: sub_A9D6+34Cp
					; sub_A9D6+368p ...

var_3E		=	-$3E
var_3C		=	-$3C
var_3A		=	-$3A

; FUNCTION CHUNK AT 0000B616 SIZE 00000254 BYTES
; FUNCTION CHUNK AT 0000B89E SIZE 00000054 BYTES
; FUNCTION CHUNK AT 0000BA88 SIZE 0000007E BYTES
; FUNCTION CHUNK AT 0000BB62 SIZE 000000A4 BYTES
; FUNCTION CHUNK AT 0000C1B8 SIZE 0000013A BYTES
; FUNCTION CHUNK AT 00010750 SIZE 000001AA BYTES
; FUNCTION CHUNK AT 0001090A SIZE 0000046C BYTES
; FUNCTION CHUNK AT 00010DB6 SIZE 00000036 BYTES
; FUNCTION CHUNK AT 00010DF0 SIZE 0000002A BYTES

		btst	#4,d1
		beq.w	loc_B106
		btst	#5,d1
		beq.w	loc_B106
		bclr	#4,d1
		bset	#6,$64(a3)
		bra.w	loc_B140
; ---------------------------------------------------------------------------

loc_B106:							; CODE XREF: sub_B0E8+4j sub_B0E8+Cj
		btst	#4,d1
		beq.w	loc_B114
		bclr	#6,$64(a3)

loc_B114:							; CODE XREF: sub_B0E8+22j
		btst	#4,d2
		beq.w	loc_B140
		bclr	#6,$64(a3)
		beq.w	loc_B140
		tst.w	d4
		bne.w	loc_B136
		move.b	#$11,($FFFFBF06).w
		bra.w	loc_B13C
; ---------------------------------------------------------------------------

loc_B136:							; CODE XREF: sub_B0E8+40j
		move.b	#$11,($FFFFBF07).w

loc_B13C:							; CODE XREF: sub_B0E8+4Aj
		bclr	#4,d2

loc_B140:							; CODE XREF: sub_B0E8+1Aj sub_B0E8+30j ...
		btst	#7,($FFFFC2FA).w
		beq.w	loc_B158
		cmpnb	 8,d0
		beq.w	loc_B158
		jsr	sub_FE1AA

loc_B158:							; CODE XREF: sub_B0E8+5Ej sub_B0E8+66j
		move.w	d0,($FFFFBF12).w
		and.w	#$F,($FFFFBF12).w
		bsr.w	sub_C656
		btst	#7,d1
		beq.w	loc_B194
		tst.w	($FFFFC318).w
		bpl.w	loc_B180
		tst.w	d4
		beq.w	loc_7CBC
		bra.w	loc_7CCA
; ---------------------------------------------------------------------------

loc_B180:							; CODE XREF: sub_B0E8+8Aj
		btst	#7,d1
		beq.w	loc_B194
		tst.w	($FFFFC318).w
		beq.w	loc_7CDC
		bra.w	loc_7CEA
; ---------------------------------------------------------------------------

loc_B194:							; CODE XREF: sub_B0E8+82j sub_B0E8+9Cj
		btst	#7,($FFFFC2EC).w
		beq.w	loc_B1AC
		btst	#3,d0
		bne.w	loc_B1AC
		addq.w	#2,d0
		and.w	#7,d0

loc_B1AC:							; CODE XREF: sub_B0E8+B2j sub_B0E8+BAj
		btst	#0,($FFFFC2EE).w
		bne.w	loc_B89E
		btst	#3,$63(a3)
		bne.w	loc_BA88
		btst	#3,$62(a3)
		beq.w	locret_B46E
		movem.l	d0-d2/a0/a3,-(sp)
		move.w	($FFFFBEDA).w,d0
		cmp.w	$52(a3),d0
		bne.w	loc_B240
		tst.w	($FFFFBEE0).w
		bmi.w	loc_B240
		tst.w	($FFFFBF6C).w
		bpl.w	loc_B240
		btst	#5,d1
		beq.w	loc_B240
		move.w	($FFFFBEE0).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a3
		add.w	d0,a3
		tst.w	$34(a3)
		beq.w	loc_B240
		jsr	sub_F6C44
		beq.w	loc_B240
		btst	#3,$62(a3)
		bne.w	loc_B240
		btst	#3,$64(a3)
		bne.w	loc_B240
		move.w	d4,($FFFFDED8).w
		move.w	($FFFFC318).w,($FFFFDEDA).w
		move.w	#$23,d0					; '#'
		jsr	sub_10662
		movem.l	(sp)+,d0-d2/a0/a3
		rts
; ---------------------------------------------------------------------------

loc_B240:							; CODE XREF: sub_B0E8+EEj sub_B0E8+F6j ...
		movem.l	(sp)+,d0-d2/a0/a3
		btst	#3,$64(a3)
		bne.w	loc_B258
		btst	#5,$62(a3)
		bne.w	loc_B81A

loc_B258:							; CODE XREF: sub_B0E8+162j
		btst	#0,$63(a3)
		bne.w	locret_B8F0
		move.w	($FFFFB7AA).w,d5
		cmp.w	$52(a3),d5
		beq.w	loc_B72E
		tst.w	$34(a3)
		beq.w	loc_B28C
		btst	#3,$64(a3)
		beq.w	loc_B284
		bra.w	loc_B28C
; ---------------------------------------------------------------------------

loc_B284:							; CODE XREF: sub_B0E8+194j
		btst	#6,d1
		bne.w	loc_BBB6

loc_B28C:							; CODE XREF: sub_B0E8+18Aj
					; sub_B0E8+198j
		tst.w	$34(a3)
		beq.w	loc_B2A2
		btst	#2,($FFFFC2F2).w
		bne.w	loc_B42E
		bra.w	loc_B2DA
; ---------------------------------------------------------------------------

loc_B2A2:							; CODE XREF: sub_B0E8+1A8j
		btst	#6,($FFFFC2F6).w
		bne.w	loc_B2DA
		tst.w	d4
		beq.w	loc_B2BA
		tst.w	($FFFFD05C).w
		bra.w	loc_B2BE
; ---------------------------------------------------------------------------

loc_B2BA:							; CODE XREF: sub_B0E8+1C6j
		tst.w	($FFFFD05A).w

loc_B2BE:							; CODE XREF: sub_B0E8+1CEj
		beq.w	loc_B2DA
		movem.w	d0,-(sp)
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_B2DA
		bra.w	sub_BFBC
; ---------------------------------------------------------------------------

loc_B2DA:							; CODE XREF: sub_B0E8+1B6j
					; sub_B0E8+1C0j ...
		btst	#4,d3
		beq.w	loc_B35C
		tst.w	($FFFFD412).w
		bne.w	loc_B35C
		tst.w	d4
		beq.w	loc_B328
		tst.b	($FFFFBF07).w
		beq.w	loc_B42E
		subq.b	#1,($FFFFBF07).w
		bpl.w	loc_B306
		move.b	#0,($FFFFBF07).w

loc_B306:							; CODE XREF: sub_B0E8+214j
		tst.b	($FFFFBF07).w
		bne.w	loc_B35C
		tst.w	d4
		beq.w	loc_B31C
		tst.w	($FFFFD05C).w
		bra.w	loc_B320
; ---------------------------------------------------------------------------

loc_B31C:							; CODE XREF: sub_B0E8+228j
		tst.w	($FFFFD05A).w

loc_B320:							; CODE XREF: sub_B0E8+230j
		bne.w	loc_B35C
		bra.w	loc_B3E4
; ---------------------------------------------------------------------------

loc_B328:							; CODE XREF: sub_B0E8+204j
		tst.b	($FFFFBF06).w
		beq.w	loc_B42E
		subq.b	#1,($FFFFBF06).w
		bpl.w	loc_B33E
		move.b	#0,($FFFFBF06).w

loc_B33E:							; CODE XREF: sub_B0E8+24Cj
		tst.w	d4
		beq.w	loc_B34C
		tst.w	($FFFFD05C).w
		bra.w	loc_B350
; ---------------------------------------------------------------------------

loc_B34C:							; CODE XREF: sub_B0E8+258j
		tst.w	($FFFFD05A).w

loc_B350:							; CODE XREF: sub_B0E8+260j
		bne.w	loc_B35C
		tst.b	($FFFFBF06).w
		beq.w	loc_B3E4

loc_B35C:							; CODE XREF: sub_B0E8+1F6j
					; sub_B0E8+1FEj ...
		btst	#4,d1
		beq.w	loc_B37E
		tst.w	d4
		bne.w	loc_B374
		move.b	#$11,($FFFFBF06).w
		bra.w	loc_B42E
; ---------------------------------------------------------------------------

loc_B374:							; CODE XREF: sub_B0E8+27Ej
		move.b	#$11,($FFFFBF07).w
		bra.w	loc_B42E
; ---------------------------------------------------------------------------

loc_B37E:							; CODE XREF: sub_B0E8+278j
		btst	#4,d2
		beq.w	loc_B42E
		btst	#4,d3
		bne.w	loc_B42E
		move.w	($FFFFBF06).w,d0
		tst.w	d4
		beq.w	loc_B3C0
		move.b	#$11,($FFFFBF07).w
		and.w	#$FF,d0
		bne.w	sub_BFBC
		tst.w	d4
		beq.w	loc_B3B4
		tst.w	($FFFFD05C).w
		bra.w	loc_B3B8
; ---------------------------------------------------------------------------

loc_B3B4:							; CODE XREF: sub_B0E8+2C0j
		tst.w	($FFFFD05A).w

loc_B3B8:							; CODE XREF: sub_B0E8+2C8j
		bne.w	sub_BFBC
		bra.w	loc_B3E4
; ---------------------------------------------------------------------------

loc_B3C0:							; CODE XREF: sub_B0E8+2ACj
		move.b	#$11,($FFFFBF06).w
		and.w	#$FF00,d0
		bne.w	sub_BFBC
		tst.w	d4
		beq.w	loc_B3DC
		tst.w	($FFFFD05C).w
		bra.w	loc_B3E0
; ---------------------------------------------------------------------------

loc_B3DC:							; CODE XREF: sub_B0E8+2E8j
		tst.w	($FFFFD05A).w

loc_B3E0:							; CODE XREF: sub_B0E8+2F0j
		bne.w	sub_BFBC

loc_B3E4:							; CODE XREF: sub_B0E8+23Cj
					; sub_B0E8+270j ...
		tst.w	($FFFFD412).w
		bne.w	sub_BFBC
		move.w	#5,d0
		cmpnw	 5,d6
		ble.w	loc_B3FC
		move.w	#$B,d0

loc_B3FC:							; CODE XREF: sub_B0E8+30Cj
		jsr	sub_B86A
		tst.w	d0
		bmi.w	locret_B46E
		movem.l	d0/a3,-(sp)
		move.l	#$FFFFB04A,a3
		asl.w	#7,d0
		add.w	d0,a3
		btst	#3,$62(a3)
		movem.l	(sp)+,d0/a3
		bne.w	locret_B46E
		tst.w	d4
		beq.w	sub_C0BC
		bra.w	sub_C0DA
; ---------------------------------------------------------------------------

loc_B42E:							; CODE XREF: sub_B0E8+1B2j
					; sub_B0E8+20Cj ...
		tst.w	$34(a3)
		bne.w	loc_B6BA
		btst	#6,d1
		beq.w	loc_B470
		move.w	($FFFFBF12).w,d0
		cmpnb	 8,d0
		beq.w	loc_B470
		move.w	d0,$54(a3)
		move.b	#8,$5E(a3)
		move.w	#$2F4,d1
		bsr.w	sub_1073A
		bset	#1,$63(a3)
		bset	#5,$62(a3)
		add.w	#$96,($FFFFB89C).w			; ''

locret_B46E:							; CODE XREF: sub_B0E8+DEj
					; sub_B0E8+31Cj ...
		rts
; ---------------------------------------------------------------------------

loc_B470:							; CODE XREF: sub_B0E8+352j
					; sub_B0E8+35Ej
		btst	#5,d1
		bne.w	loc_B4D0
		btst	#5,d3
		bne.w	loc_B48A
		bclr	#7,$63(a3)
		bra.w	loc_B616
; ---------------------------------------------------------------------------

loc_B48A:							; CODE XREF: sub_B0E8+394j
		btst	#7,$63(a3)
		beq.w	loc_B616
		movem.l	d0-d3/a0-a3,-(sp)
		move.l	#$5B1C,a0
		add.w	$58(a3),a0
		move.w	$54(a3),d0
		btst	#3,4(a3)
		beq.w	loc_B4B8
		neg.w	d0
		addq.w	#8,d0
		and.w	#7,d0

loc_B4B8:							; CODE XREF: sub_B0E8+3C4j
		asl.w	#1,d0
		add.w	(a0,d0.w),a0
		tst.b	$5B(a3)
		movem.l	(sp)+,d0-d3/a0-a3
		bpl.s	locret_B46E
		move.w	#$A,$5C(a3)
		rts
; ---------------------------------------------------------------------------

loc_B4D0:							; CODE XREF: sub_B0E8+38Cj
		btst	#5,$62(a3)
		bne.w	loc_B616
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		move.w	($FFFFB75E).w,d1
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		move.w	d0,$54(a3)
		movem.w	(sp)+,d0-d1
		move.w	($FFFFBF12).w,d0
		bclr	#0,($FFFFC2F2).w
		move.w	($FFFFC468).w,d0
		and.w	#7,d0
		asl.w	#4,d0
		add.w	#$A0,d0					; ''
		cmp.w	#$DB,($FFFFB75E).w			; ''
		bgt.w	loc_B524
		cmp.w	#$FF25,($FFFFB75E).w
		bgt.w	loc_B528

loc_B524:							; CODE XREF: sub_B0E8+42Ej
		sub.w	#$40,d0					; '@'

loc_B528:							; CODE XREF: sub_B0E8+438j
		move.w	d0,d1
		muls.w	($FFFFB772).w,d0
		swap	d0
		add.w	($FFFFB74A).w,d0
		muls.w	($FFFFB774).w,d1
		swap	d1
		add.w	($FFFFB75E).w,d1
		bsr.w	sub_B5D8
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmpnl	 $384,d0
		bhi.w	loc_B55E
		movem.w	(sp)+,d0-d1
		bra.w	loc_B582
; ---------------------------------------------------------------------------

loc_B55E:							; CODE XREF: sub_B0E8+46Aj
		bsr.w	sub_110BE
		moveq	#1,d2
		add.w	d0,d2
		moveq	#$12,d4
		btst	#3,($FFFFC2EC).w
		beq.w	loc_B574
		addq.w	#8,d4

loc_B574:							; CODE XREF: sub_B0E8+486j
		movem.w	(sp)+,d0-d1
		muls.w	d4,d1
		addq.w	#8,d4
		muls.w	d4,d0
		divs.w	d2,d0
		divs.w	d2,d1

loc_B582:							; CODE XREF: sub_B0E8+472j
		add.w	d3,d1
		move.w	d1,d2
		cmp.w	#$22,2(a0)				; '"'
		cmp.w	#$18,(a0)
		cmp.w	#$FFE8,(a0)
		cmp.w	#$C,2(a0)
		cmp.w	#$108,($FFFFB75E).w
		cmp.w	#$FEF8,($FFFFB75E).w
		bset	#1,$63(a3)
		bne.w	loc_B616
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFBEE6,a0
		move.w	#$108,d3
		btst	#7,$62(a3)
		beq.w	loc_B5CC
		neg.w	d3
		addq.w	#4,a0

loc_B5CC:							; CODE XREF: sub_B0E8+4DCj
		jsr	sub_DBBA
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_B0E8


; =============== S U B	R O U T	I N E =======================================


sub_B5D8:							; CODE XREF: sub_B0E8+456p
		move.w	($FFFFB7AA).w,d2
		cmp.w	$52(a3),d2
		bne.w	loc_B5E6
		clr.w	d0

loc_B5E6:							; CODE XREF: sub_B5D8+8j
		cmpnw	 $103,d1
		blt.w	loc_B5F2
		move.w	#$103,d1

loc_B5F2:							; CODE XREF: sub_B5D8+12j
		cmpnw	 $FEFD,d1
		bgt.w	loc_B5FE
		move.w	#$FEFD,d1

loc_B5FE:							; CODE XREF: sub_B5D8+1Ej
		sub.w	d3,d1
		rts
; End of function sub_B5D8

; ---------------------------------------------------------------------------
		dc.w	$146
		dc.w	$178
		dc.w	$250
		dc.w	$2A2
		dc.w	$1EC
		dc.w	$21E
		dc.w	$148E
		dc.w	$14C0
		dc.w	$14F2
		dc.w	$1544
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_B616:							; CODE XREF: sub_B0E8+39Ej
					; sub_B0E8+3A8j ...
		btst	#1,$63(a3)
		beq.w	loc_B622
		rts
; ---------------------------------------------------------------------------

loc_B622:							; CODE XREF: sub_B0E8+534j
		tst.w	$34(a3)
		bne.w	loc_10750
		bclr	#1,($FFFFC2F2).w
		cmp.w	#$24,(a3)				; '$'
		ble.w	loc_B64E
		tst.w	$28(a3)
		bmi.w	loc_B64E
		beq.w	loc_B64E
		bset	#1,($FFFFC2F2).w
		clr.w	$28(a3)

loc_B64E:							; CODE XREF: sub_B0E8+54Cj
					; sub_B0E8+554j ...
		cmp.w	#$FFDC,(a3)
		bge.w	loc_B668
		tst.w	$28(a3)
		bpl.w	loc_B668
		bset	#1,($FFFFC2F2).w
		clr.w	$28(a3)

loc_B668:							; CODE XREF: sub_B0E8+56Aj
					; sub_B0E8+572j
		cmp.w	#$E7,$14(a3)				; ''
		bgt.w	loc_B6A8
		cmp.w	#$FF19,$14(a3)
		ble.w	loc_B6A8
		bra.w	*+4
		movem.w	d0-d1,-(sp)
		move.w	$14(a3),d0
		move.w	$2A(a3),d1
		eor.w	d1,d0
		movem.w	(sp)+,d0-d1
		bpl.w	loc_B6A8
		tst.w	$2A(a3)
		beq.w	loc_B6A8
		bset	#1,($FFFFC2F2).w
		clr.w	$2A(a3)

loc_B6A8:							; CODE XREF: sub_B0E8+586j
					; sub_B0E8+590j ...
		btst	#1,($FFFFC2F2).w
		bne.w	locret_B46E
		move.w	($FFFFBF12).w,d0
		bra.w	loc_10750
; ---------------------------------------------------------------------------

loc_B6BA:							; CODE XREF: sub_B0E8+34Aj
		move.w	($FFFFBF12).w,d0
		btst	#3,$64(a3)
		bne.w	locret_B868
		btst	#5,d1
		beq.w	loc_10750
		movem.l	d7-a0,-(sp)
		move.w	($FFFFBEDA).w,d7
		asl.w	#7,d7
		move.l	#$FFFFB04A,a0
		add.w	d7,a0
		tst.w	$34(a0)
		movem.l	(sp)+,d7-a0
		beq.w	loc_BB62
		movem.w	d7,-(sp)
		move.w	$52(a3),d7
		cmp.w	($FFFFBEE0).w,d7
		movem.w	(sp)+,d7
		bne.w	loc_BB62
		tst.w	($FFFFB7AA).w
		bpl.w	loc_BB62
		jsr	sub_F6C44
		beq.w	loc_BB62
		move.w	d0,-(sp)
		move.w	d4,($FFFFDED8).w
		move.w	($FFFFC318).w,($FFFFDEDA).w
		move.w	#$23,d0					; '#'
		jsr	sub_10662
		move.w	(sp)+,d0
		rts
; ---------------------------------------------------------------------------

loc_B72E:							; CODE XREF: sub_B0E8+182j
		bsr.w	sub_E62E
		tst.w	$34(a3)
		bne.w	loc_B744
		btst	#1,$63(a3)
		bne.w	locret_B46E

loc_B744:							; CODE XREF: sub_B0E8+64Ej
		btst	#2,($FFFFC2EC).w
		bne.w	sub_BC40
		btst	#3,($FFFFC2EC).w
		bne.w	loc_C224
		btst	#4,d1
		beq.w	loc_B76A
		jsr	sub_FEE60
		bra.w	sub_BC06
; ---------------------------------------------------------------------------

loc_B76A:							; CODE XREF: sub_B0E8+674j
		btst	#6,d1
		beq.w	loc_B788
		tst.w	d4
		beq.w	loc_B782
		move.b	#$F,($FFFFD41F).w
		bra.w	loc_B788
; ---------------------------------------------------------------------------

loc_B782:							; CODE XREF: sub_B0E8+68Cj
		move.b	#$F,($FFFFD41E).w

loc_B788:							; CODE XREF: sub_B0E8+686j
					; sub_B0E8+696j
		btst	#6,d3
		beq.w	loc_B7E2
		tst.w	d4
		beq.w	loc_B7BC
		subq.b	#1,($FFFFD41F).w
		bpl.w	loc_B7A4
		move.b	#0,($FFFFD41F).w

loc_B7A4:							; CODE XREF: sub_B0E8+6B2j
		tst.b	($FFFFD41F).w
		bne.w	loc_B7E2
		bset	#3,($FFFFC2F6).w
		jsr	sub_BC06
		bra.w	sub_B8F2
; ---------------------------------------------------------------------------

loc_B7BC:							; CODE XREF: sub_B0E8+6AAj
		subq.b	#1,($FFFFD41E).w
		bpl.w	loc_B7CA
		move.b	#0,($FFFFD41E).w

loc_B7CA:							; CODE XREF: sub_B0E8+6D8j
		tst.b	($FFFFD41E).w
		bne.w	loc_B7E2
		bset	#3,($FFFFC2F6).w
		jsr	sub_BC06
		bra.w	sub_B8F2
; ---------------------------------------------------------------------------

loc_B7E2:							; CODE XREF: sub_B0E8+6A4j
					; sub_B0E8+6C0j ...
		btst	#6,d3
		bne.w	loc_B7FC
		btst	#6,d2
		beq.w	loc_B7FC
		bset	#3,($FFFFC2F6).w
		bra.w	sub_BC06
; ---------------------------------------------------------------------------

loc_B7FC:							; CODE XREF: sub_B0E8+6FEj
					; sub_B0E8+706j
		tst.w	$34(a3)
		bne.w	loc_B808
		bra.w	loc_10750
; ---------------------------------------------------------------------------

loc_B808:							; CODE XREF: sub_B0E8+718j
		btst	#5,d1
		beq.w	loc_10750
		jsr	sub_FEE60
		bra.w	loc_C1B8
; ---------------------------------------------------------------------------

loc_B81A:							; CODE XREF: sub_B0E8+16Cj
		move.w	($FFFFB7AA).w,d5
		cmp.w	$52(a3),d5
		beq.w	locret_B46E
		btst	#4,d1
		beq.w	locret_B868
		btst	#6,($FFFFC2F6).w
		bne.w	locret_B868
		bsr.w	sub_BFBC
		movem.l	d0/a0,-(sp)
		move.w	($FFFFC320).w,d0
		tst.w	d4
		beq.w	loc_B84E
		move.w	($FFFFC322).w,d0

loc_B84E:							; CODE XREF: sub_B0E8+75Ej
		tst.w	d0
		bmi.w	loc_B864
		asl.w	#7,d0
		move.l	#$FFFFB04A,a0
		add.w	d0,a0
		bset	#6,$64(a0)

loc_B864:							; CODE XREF: sub_B0E8+768j
		movem.l	(sp)+,d0/a0

locret_B868:							; CODE XREF: sub_B0E8+5DCj
					; sub_B0E8+742j ...
		rts
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_B86A:							; CODE XREF: sub_B0E8:loc_B3FCp
					; sub_B0E8:loc_C28Cp ...
		movem.l	d1/a0,-(sp)
		move.l	#$FFFFB04A,a0
		asl.w	#7,d0
		add.w	d0,a0
		move.w	#5,d1

loc_B87C:							; CODE XREF: sub_B86A+1Ej
		tst.w	$34(a0)
		beq.w	loc_B894
		suba.w	#$80,a0					; ''
		dbf	d1,loc_B87C
		move.w	#$FFFF,d0
		bra.w	loc_B898
; ---------------------------------------------------------------------------

loc_B894:							; CODE XREF: sub_B86A+16j
		move.w	$52(a0),d0

loc_B898:							; CODE XREF: sub_B86A+26j
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_B86A

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_B89E:							; CODE XREF: sub_B0E8+CAj
		btst	#6,($FFFFC2F6).w
		beq.w	loc_B8AA
		rts
; ---------------------------------------------------------------------------

loc_B8AA:							; CODE XREF: sub_B0E8+7BCj
		move.w	$36(a3),d4
		cmp.b	#$17,$38(a3,d4.w)
		bne.s	locret_B868
		move.w	#$BF98,a0
		btst	#7,$62(a3)
		beq.w	loc_B8C8
		move.w	#$BF9A,a0

loc_B8C8:							; CODE XREF: sub_B0E8+7D8j
		move.w	d0,(a0)
		btst	#1,$63(a3)
		bne.s	locret_B868
		btst	#4,d1
		beq.w	loc_B8E8
		move.w	#$FEA,d1
		bset	#1,$63(a3)
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_B8E8:							; CODE XREF: sub_B0E8+7EEj
		move.w	#$1014,d1
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

locret_B8F0:							; CODE XREF: sub_B0E8+176j
		rts
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_B8F2:							; CODE XREF: sub_B0E8+6D0j
					; sub_B0E8+6F6j ...
		tst.w	($FFFFD058).w
		bne.w	locret_B868
		btst	#4,($FFFFC2FC).w
		bne.w	locret_B868
		bsr.w	sub_13040
		bset	#1,$30(a2)
		bne.w	locret_B868
		btst	#3,($FFFFC2F6).w
		bne.w	loc_B922
		bclr	#2,($FFFFC2EC).w

loc_B922:							; CODE XREF: sub_B8F2+26j
		bclr	#3,($FFFFC2EC).w
		bset	#3,$63(a3)
; End of function sub_B8F2


; =============== S U B	R O U T	I N E =======================================


sub_B92E:							; CODE XREF: sub_B0E8+9DAp
					; sub_F64E+32j	...
		bsr.w	sub_B9C2
		cmp.w	#$F,($FFFFB02A).w
		blt.w	loc_B946
		bset	#0,($FFFFC2F0).w
		bra.w	loc_B950
; ---------------------------------------------------------------------------

loc_B946:							; CODE XREF: sub_B92E+Aj
		jsr	sub_18A56
		bsr.w	sub_B9C2

loc_B950:							; CODE XREF: sub_B92E+14j
		bsr.w	sub_119B8
		subq.w	#2,($FFFFB02A).w
		addq.w	#1,($FFFFB028).w
		moveq	#2,d4

loc_B95E:							; CODE XREF: sub_B92E:loc_B9ACj
		move.w	d4,d0
		bsr.w	sub_BA04
		tst.w	d0
		bmi.w	loc_B9AC
		btst	#1,$30(a2)
		bne.w	loc_B980
		cmp.w	$2E(a2),d4
		bne.w	loc_B9A8
		move.w	$16(a2),d0

loc_B980:							; CODE XREF: sub_B92E+42j
		move.w	#$BFA4,a1
		move.l	#unk_44120,(a1)
		add.b	d4,2(a1)
		bsr.w	sub_11BA4
		move.w	d0,-(sp)
		move.l	#word_191A6,a1
		bsr.w	sub_13508
		move.w	(sp)+,d0
		bsr.w	sub_12E66
		subq.w	#5,($FFFFB028).w

loc_B9A8:							; CODE XREF: sub_B92E+4Aj
		subq.w	#1,($FFFFB02A).w

loc_B9AC:							; CODE XREF: sub_B92E+38j
		dbf	d4,loc_B95E
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		addq.w	#2,($FFFFB028).w
		bra.w	sub_11BA4
; End of function sub_B92E


; =============== S U B	R O U T	I N E =======================================


sub_B9C2:							; CODE XREF: sub_B92Ep	sub_B92E+1Ep ...
		clr.w	d0
		cmp.w	#$C6CE,a2
		bne.w	loc_B9D0
		eor.w	#$16,d0

loc_B9D0:							; CODE XREF: sub_B9C2+6j
		btst	#1,($FFFFC2EA).w
		beq.w	loc_B9DE
		eor.w	#$16,d0

loc_B9DE:							; CODE XREF: sub_B9C2+14j
		bsr.w	sub_11B92
; End of function sub_B9C2

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$16
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	d0,($FFFFB02A).w
		moveq	#2,d0
		bsr.w	sub_BA04
		moveq	#6,d1
		tst.w	d0
		bpl.w	loc_BA00
		subq.w	#1,d1
		addq.w	#1,($FFFFB02A).w

loc_BA00:							; CODE XREF: ROM:0000B9F6j
		moveq	#9,d0
		rts

; =============== S U B	R O U T	I N E =======================================


sub_BA04:							; CODE XREF: sub_B92E+32p
					; ROM:0000B9EEp ...
		movem.l	d1-d2,-(sp)
		move.w	$388(a2),d2
		cmp.w	#$C6CE,a2
		beq.w	loc_BA18
		move.w	-$340(a2),d2

loc_BA18:							; CODE XREF: sub_BA04+Cj
		sub.w	$24(a2),d2
		beq.w	loc_BA2E
		add.w	#$15,d0
		tst.w	d2
		bmi.w	loc_BA2E
		add.w	#$15,d0

loc_BA2E:							; CODE XREF: sub_BA04+18j sub_BA04+22j
		move.w	$16(a2),d1
		add.w	d1,d0
		add.w	d1,d0
		add.w	d1,d0
		lea	off_BA48(pc),a0
		move.b	(a0,d0.w),d0
		ext.w	d0
		movem.l	(sp)+,d1-d2
		rts
; End of function sub_BA04

; ---------------------------------------------------------------------------
off_BA48:	dc.l	loc_10200+1				; DATA XREF: sub_BA04+34o
		dc.b	2
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	4
		dc.b	3
		dc.b	$FF
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	3
		dc.b	4
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	5
		dc.b	6
		dc.b	$FF
		dc.b	6
		dc.b	5
		dc.b	$FF
		dc.b	$FF
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_BA88:							; CODE XREF: sub_B0E8+D4j
		btst	#3,($FFFFC2F6).w
		beq.w	loc_BAA0
		movem.l	d0-a6,-(sp)
		jsr	sub_BC40
		movem.l	(sp)+,d0-a6

loc_BAA0:							; CODE XREF: sub_B0E8+9A6j
		move.w	d1,-(sp)
		move.w	#$C6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_BAB4
		add.w	#$364,a2

loc_BAB4:							; CODE XREF: sub_B0E8+9C4j
		bclr	#0,$30(a2)
		beq.w	loc_BAC6
		bsr.w	sub_BB36
		bsr.w	sub_B92E

loc_BAC6:							; CODE XREF: sub_B0E8+9D2j
		move.w	(sp)+,d1
		clr.w	d2
		btst	#6,d1
		bne.w	sub_BB06
		addq.w	#1,d2
		btst	#4,d1
		bne.w	sub_BB06
		addq.w	#1,d2
		btst	#5,d1
		bne.w	sub_BB06
		btst	#3,$62(a3)
		beq.w	locret_BB04
		btst	#5,$62(a3)
		bne.w	locret_BB04
		btst	#0,$63(a3)
		beq.w	loc_10750

locret_BB04:							; CODE XREF: sub_B0E8+A04j
					; sub_B0E8+A0Ej
		rts
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_BB06:							; CODE XREF: sub_B0E8+9E6j
					; sub_B0E8+9F0j ...
		move.w	d2,d0
		move.w	d2,$2E(a2)
		bsr.w	sub_BA04
		tst.w	d0
		bmi.w	locret_A9D4
		bclr	#3,$63(a3)
		bset	#3,$62(a3)
		bsr.w	sub_13040
		bclr	#1,$30(a2)
		move.w	d0,$16(a2)
		jsr	sub_15788
; End of function sub_BB06


; =============== S U B	R O U T	I N E =======================================


sub_BB36:							; CODE XREF: sub_B0E8+9D6p
					; ROM:000188B2p ...
		btst	#7,($FFFFC2EC).w
		bne.w	locret_BC3E
		bsr.w	sub_B9C2
		cmp.w	#$F,($FFFFB02A).w
		blt.w	loc_BB54
		bclr	#0,($FFFFC2F0).w

loc_BB54:							; CODE XREF: sub_BB36+14j
		addq.w	#1,d1
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		bra.w	loc_12C04
; End of function sub_BB36

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_BB62:							; CODE XREF: sub_B0E8+602j
					; sub_B0E8+616j ...
		jsr	sub_1575A
		tst.w	($FFFFD058).w
		bne.w	loc_BB88
		sub.w	#$CC,d0					; ''
		jsr	sub_1577A
		btst	#4,($FFFFC2FC).w
		beq.w	loc_BB88
		move.w	#$1000,d0

loc_BB88:							; CODE XREF: sub_B0E8+A84j
					; sub_B0E8+A98j
		lsr.w	#7,d0
		move.w	d0,d1
		move.w	$54(a3),d2
		asl.w	#2,d2
		move.l	#unk_10E64,a0
		muls.w	(a0,d2.w),d0
		muls.w	2(a0,d2.w),d1
		add.w	d0,$28(a3)
		add.w	d1,$2A(a3)
		bset	#5,$62(a3)
		move.w	#$C5E,d1
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_BBB6:							; CODE XREF: sub_B0E8+1A0j
		bset	#5,$62(a3)
		move.w	#$1122,d1
		tst.w	$32(a3)
		beq.w	sub_1073A
		move.w	#$B04A,a0
		move.w	$2E(a3),d0
		asl.w	#7,d0
		add.w	d0,a0

loc_BBD4:							; CODE XREF: sub_EACE+96j sub_EACE+A8j
		bset	#5,$62(a3)
		move.w	#$1122,d1
		tst.w	$32(a3)
		beq.w	loc_BC02
		move.w	$14(a3),d0
		sub.w	$14(a0),d0
		btst	#7,$62(a3)
		beq.w	loc_BBFA
		neg.w	d0

loc_BBFA:							; CODE XREF: sub_B0E8+B0Cj
		bmi.w	loc_BC02
		move.w	#$C90,d1

loc_BC02:							; CODE XREF: sub_B0E8+AFAj
					; sub_B0E8:loc_BBFAj
		bra.w	sub_1073A
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_BC06:							; CODE XREF: sub_B0E8+67Ej
					; sub_B0E8+6CAp ...
		move.w	$54(a3),($FFFFBEDC).w
		and.w	#7,($FFFFBEDC).w
		btst	#2,($FFFFC2F2).w
		beq.w	loc_BC38
		bclr	#2,($FFFFC2FA).w
		bset	#5,($FFFFC2F2).w
		bne.w	locret_B868
		bclr	#5,($FFFFC2FA).w
		move.w	#$64,($FFFFC31C).w			; 'd'

loc_BC38:							; CODE XREF: sub_BC06+12j
		bset	#2,($FFFFC2EC).w

locret_BC3E:							; CODE XREF: sub_BB36+6j sub_BC40+16j
		rts
; End of function sub_BC06


; =============== S U B	R O U T	I N E =======================================


sub_BC40:							; CODE XREF: sub_B0E8+662j
					; sub_B0E8+9AEp
		btst	#4,d2
		bne.w	loc_BC64
		btst	#3,($FFFFC2F6).w
		bne.w	loc_BC64
		btst	#3,d0
		bne.s	locret_BC3E
		and.w	#7,d0
		move.w	d0,($FFFFBEDC).w
		bset	#3,d0

loc_BC64:							; CODE XREF: sub_BC40+4j sub_BC40+Ej ...
		movem.l	d0-d5/a0-a1,-(sp)
		bclr	#2,($FFFFC2EC).w
		st	($FFFFB7AA).w
		move.b	#$10,$5E(a3)
		move.w	$52(a3),($FFFFBEDA).w
		bclr	#3,($FFFFC2F6).w
		beq.w	loc_BCAE
		jsr	sub_F67E4
		move.w	#$12,($FFFFD418).w
		btst	#2,($FFFFC2F8).w
		beq.w	loc_BCA4
		move.w	#$3A,($FFFFD418).w			; ':'

loc_BCA4:							; CODE XREF: sub_BC40+5Aj
		jsr	sub_F6778
		bra.w	loc_BDB2
; ---------------------------------------------------------------------------

loc_BCAE:							; CODE XREF: sub_BC40+44j
		moveq	#8,d0
		tst.w	$34(a3)
		beq.w	loc_BCBC
		move.b	$6E(a3),d0

loc_BCBC:							; CODE XREF: sub_BC40+74j
		asl.w	#2,d0
		asr.w	#1,d0
		add.w	#$A0,d0					; ''
		move.w	d0,($FFFFBEDE).w
		btst	#0,$6E(a3)
		beq.w	loc_BCD8
		asr.w	#4,d0
		add.w	d0,($FFFFBEDE).w

loc_BCD8:							; CODE XREF: sub_BC40+8Ej
		moveq	#$FFFFFFFF,d4
		moveq	#5,d3
		move.w	#$B04A,a1
		cmp.w	#6,$52(a3)
		blt.w	loc_BCF8
		add.w	#$300,a1
		btst	#2,($FFFFC2F2).w
		bne.w	loc_BD6C

loc_BCF8:							; CODE XREF: sub_BC40+A6j
					; sub_BC40+11Aj
		cmp.l	a1,a3
		beq.w	loc_BD56
		tst.w	$34(a1)
		beq.w	loc_BD56
		btst	#2,$63(a1)
		bne.w	loc_BD56
		move.w	(a1),d0
		sub.w	($FFFFB74A).w,d0
		move.w	$14(a1),d1
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		bsr.w	sub_10676
		movem.w	(sp)+,d1-d2
		sub.w	($FFFFBEDC).w,d0
		and.w	#7,d0
		asl.b	#5,d0
		ext.w	d0
		asl.w	#3,d0
		muls.w	d0,d0
		cmpnl	 (loc_FFFE+2),d0
		bhi.w	loc_BD56
		muls.w	d1,d1
		muls.w	d2,d2
		add.l	d1,d2
		add.l	d0,d2
		cmp.l	d4,d2
		bhi.w	loc_BD56
		move.l	d2,d4
		move.l	a1,a0

loc_BD56:							; CODE XREF: sub_BC40+BAj sub_BC40+C2j ...
		add.w	#$80,a1					; ''
		dbf	d3,loc_BCF8
		tst.l	d4
		bmi.w	loc_BD6C
		bsr.w	sub_BE26
		bra.w	loc_BDB2
; ---------------------------------------------------------------------------

loc_BD6C:							; CODE XREF: sub_BC40+B4j
					; sub_BC40+120j
		move.w	($FFFFBEDC).w,d0
		asl.w	#2,d0
		move.l	#unk_10E64,a0
		move.w	2(a0,d0.w),d1
		muls.w	($FFFFBEDE).w,d1
		moveq	#$A,d2
		asl.l	d2,d1
		divs.w	#$BB8,d1
		add.w	$2A(a3),d1
		move.w	d1,($FFFFB774).w
		move.w	(a0,d0.w),d1
		muls.w	($FFFFBEDE).w,d1
		asl.l	d2,d1
		divs.w	#$BB8,d1
		add.w	$28(a3),d1
		move.w	d1,($FFFFB772).w
		move.w	#$1000,d0
		bsr.w	sub_11086
		move.w	d0,($FFFFB776).w

loc_BDB2:							; CODE XREF: sub_BC40+6Aj
					; sub_BC40+128j
		tst.w	$34(a3)
		bne.w	loc_BDD8
		tst.w	($FFFFB774).w
		btst	#7,$62(a3)
		beq.w	loc_BDD0
		bmi.w	loc_BDD4
		bra.w	loc_BDD8
; ---------------------------------------------------------------------------

loc_BDD0:							; CODE XREF: sub_BC40+184j
		bmi.w	loc_BDD8

loc_BDD4:							; CODE XREF: sub_BC40+188j
		neg.w	($FFFFB774).w

loc_BDD8:							; CODE XREF: sub_BC40+176j
					; sub_BC40+18Cj ...
		move.w	($FFFFB772).w,d0
		move.w	($FFFFB774).w,d1
		bsr.w	sub_10676
		move.w	#$366,d1
		tst.w	$34(a3)
		beq.w	loc_BE00
		move.w	#$718,d1
		bsr.w	loc_C198
		beq.w	loc_BE00
		move.w	#$78A,d1

loc_BE00:							; CODE XREF: sub_BC40+1ACj
					; sub_BC40+1B8j
		bsr.w	sub_1073A
		bset	#5,$62(a3)
		moveq	#$C,d0
		sub.b	($FFFFB776).w,d0
		lsr.w	#2,d0
		and.w	#3,d0
		add.w	#$10,d0
		move.w	d0,-(sp)
		bsr.w	sub_11132
		movem.l	(sp)+,d0-d5/a0-a1
		rts
; End of function sub_BC40


; =============== S U B	R O U T	I N E =======================================


sub_BE26:							; CODE XREF: sub_BC40+124p
		btst	#3,$62(a3)
		beq.w	loc_BEAE
		tst.w	$34(a3)
		bne.w	loc_BEAE
		movem.l	d0-d1,-(sp)
		btst	#3,$62(a0)
		bne.w	loc_BEAA
		bset	#6,$64(a0)
		move.w	($FFFFC328).w,d1
		cmp.w	($FFFFC32A).w,d1
		bne.w	loc_BE7C
		move.w	$52(a0),d0
		move.w	$52(a3),d1
		cmp.w	($FFFFC320).w,d1
		bne.w	loc_BE72
		jsr	sub_C0BC
		bra.w	loc_BEAA
; ---------------------------------------------------------------------------

loc_BE72:							; CODE XREF: sub_BE26+3Ej
		jsr	sub_C0DA
		bra.w	loc_BEAA
; ---------------------------------------------------------------------------

loc_BE7C:							; CODE XREF: sub_BE26+2Ej
		move.w	#1,d1
		btst	#6,$62(a3)
		beq.w	loc_BE8E
		move.w	#2,d1

loc_BE8E:							; CODE XREF: sub_BE26+60j
		move.w	$52(a0),d0
		cmp.w	($FFFFC328).w,d1
		bne.w	loc_BEA4
		jsr	sub_C0BC
		bra.w	loc_BEAA
; ---------------------------------------------------------------------------

loc_BEA4:							; CODE XREF: sub_BE26+70j
		jsr	sub_C0DA

loc_BEAA:							; CODE XREF: sub_BE26+1Cj sub_BE26+48j ...
		movem.l	(sp)+,d0-d1

loc_BEAE:							; CODE XREF: sub_BE26+6j sub_BE26+Ej
		bsr.w	sub_13040
		addq.w	#1,$12(a2)
		move.w	$52(a0),($FFFFBEE0).w
		move.w	($FFFFBEDE).w,d5
		asr.w	#2,d5
		exg	a0,a3
		move.l	#$13,d0
		bsr.w	sub_10658
		exg	a0,a3
		move.l	a0,-(sp)
		bsr.w	sub_106E0
		add.w	(a0),d0
		sub.w	($FFFFB74A).w,d0
		add.w	$14(a0),d1
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		movem.w	(sp),d2-d3
		asr.w	#2,d2
		asr.w	#2,d3
		move.w	$28(a0),d0
		muls.w	#$F0,d0					; ''
		swap	d0
		move.w	$2A(a0),d1
		muls.w	#$F0,d1					; ''
		swap	d1
		movem.w	d0-d1,-(sp)
		muls.w	d2,d0
		muls.w	d3,d1
		add.w	d1,d0
		asl.w	#1,d0
		move.w	d0,d4
		movem.w	(sp),d0-d1
		muls.w	d0,d0
		muls.w	d1,d1
		muls.w	d5,d5
		neg.l	d5
		add.l	d0,d5
		add.l	d1,d5
		muls.w	d2,d2
		muls.w	d3,d3
		add.l	d2,d3
		muls.w	d5,d3
		asl.l	#2,d3
		move.w	d4,d0
		muls.w	d0,d0
		sub.l	d3,d0
		bsr.w	sub_110BE
		moveq	#1,d3
		asr.w	#2,d5
		bne.w	loc_BF40
		moveq	#1,d5

loc_BF40:							; CODE XREF: sub_BE26+114j
					; sub_BE26+124j
		move.w	d0,d2
		neg.w	d0
		sub.w	d4,d2
		ext.l	d2
		divs.w	d5,d2
		dbpl	d3,loc_BF40
		bne.w	loc_BF54
		addq.w	#1,d2

loc_BF54:							; CODE XREF: sub_BE26+128j
		cmpnw	 $18,d2
		bls.w	loc_BF5E
		moveq	#$18,d2

loc_BF5E:							; CODE XREF: sub_BE26+132j
		move.b	d2,($FFFFB776).w
		cmpnw	 $C,d2
		blt.w	loc_BF70
		move.b	#$C,($FFFFB776).w

loc_BF70:							; CODE XREF: sub_BE26+140j
		move.w	d2,d0
		asl.w	#3,d0
		sub.w	#$A,d0
		move.b	d0,$40(a0)
		subq.w	#6,d0
		move.b	d0,($FFFFB7A8).w
		movem.w	(sp)+,d0-d1
		muls.w	d2,d0
		asr.l	#1,d0
		add.w	(sp)+,d0
		move.w	($FFFFB74A).w,$44(a0)
		add.w	d0,$44(a0)
		muls.w	d2,d1
		asr.l	#1,d1
		add.w	(sp)+,d1
		move.w	($FFFFB75E).w,$46(a0)
		add.w	d1,$46(a0)
		mulu.w	#$78,d2					; 'x'
		swap	d0
		divs.w	d2,d0
		move.w	d0,($FFFFB772).w
		swap	d1
		divs.w	d2,d1
		move.w	d1,($FFFFB774).w
		rts
; End of function sub_BE26


; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_BFBC:							; CODE XREF: sub_B0E8+1EEj
					; sub_B0E8+2BAj ...
		jmp	sub_FF9A8
; End of function sub_BFBC

; ---------------------------------------------------------------------------
		bne.w	locret_C0AC
		movem.l	d0-d6/a0-a1,-(sp)
		move.w	($FFFFB772).w,d0
		asr.w	#8,d0
		add.w	($FFFFB74A).w,d0
		move.w	($FFFFB774).w,d1
		asr.w	#8,d1
		add.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		moveq	#5,d2
		move.w	d4,d3
		eor.w	#2,d3
		moveq	#$FFFFFFFF,d5
		move.w	#$B04A,a0
		move.w	#$C328,a1
		cmp.w	#1,(a1,d4.w)
		beq.w	loc_C002
		add.w	#$300,a0

loc_C002:							; CODE XREF: ROM:0000BFFAj
		move.w	#$C320,a1

loc_C006:							; CODE XREF: ROM:0000C088j
		tst.w	$34(a0)
		ble.w	loc_C084
		btst	#2,$63(a0)
		bne.w	loc_C084
		btst	#3,$62(a0)
		bne.w	loc_C084
		btst	#2,($FFFFC2F2).w
		beq.w	loc_C054
		movem.l	d0,-(sp)
		move.w	($FFFFD406).w,d0
		cmp.w	$52(a0),d0
		movem.l	(sp)+,d0
		beq.w	loc_C054
		movem.l	d0,-(sp)
		move.w	($FFFFD408).w,d0
		cmp.w	$52(a0),d0
		movem.l	(sp)+,d0
		bne.w	loc_C084

loc_C054:							; CODE XREF: ROM:0000C028j
					; ROM:0000C03Cj
		btst	#5,$62(a0)
		bne.w	loc_C084
		movem.w	(sp),d0-d1
		sub.w	(a0),d0
		muls.w	d0,d0
		sub.w	$14(a0),d1
		muls.w	d1,d1
		add.l	d1,d0
		cmp.l	d5,d0
		bhi.w	loc_C084
		move.w	$52(a0),d1
		cmp.w	(a1,d3.w),d1
		beq.w	loc_C084
		move.l	d0,d5
		move.w	d1,d6

loc_C084:							; CODE XREF: ROM:0000C00Aj
					; ROM:0000C014j ...
		add.w	#$80,a0					; ''
		dbf	d2,loc_C006
		addq.w	#4,sp
		pea	(loc_C0A8).l
		cmp.w	(a1,d4.w),d6
		beq.w	loc_C0AE
		move.w	d6,d0
		tst.w	d4
		beq.w	sub_C0BC
		bra.w	sub_C0DA
; ---------------------------------------------------------------------------

loc_C0A8:							; DATA XREF: ROM:0000C08Eo
		movem.l	(sp)+,d0-d6/a0-a1

locret_C0AC:							; CODE XREF: ROM:0000BFC2j
		rts
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_FF9A8

loc_C0AE:							; CODE XREF: ROM:0000C098j
					; sub_105A2+98j ...
		bset	#5,$62(a3)
		move.w	#$B24,d1
		bra.w	sub_1073A
; END OF FUNCTION CHUNK	FOR sub_FF9A8

; =============== S U B	R O U T	I N E =======================================


sub_C0BC:							; CODE XREF: sub_B0E8+33Ej
					; sub_BE26+42p	...
		cmp.w	($FFFFC320).w,d0
		beq.w	locret_C0D8
		movem.l	d1/a0-a1,-(sp)
		move.w	($FFFFC320).w,d1
		bsr.w	sub_C0F8
		move.w	d0,($FFFFC320).w
		movem.l	(sp)+,d1/a0-a1

locret_C0D8:							; CODE XREF: sub_C0BC+4j
		rts
; End of function sub_C0BC


; =============== S U B	R O U T	I N E =======================================


sub_C0DA:							; CODE XREF: sub_B0E8+342j
					; sub_BE26:loc_BE72p ...
		cmp.w	($FFFFC322).w,d0
		beq.w	locret_C0F6
		movem.l	d1/a0-a1,-(sp)
		move.w	($FFFFC322).w,d1
		bsr.w	sub_C0F8
		move.w	d0,($FFFFC322).w
		movem.l	(sp)+,d1/a0-a1

locret_C0F6:							; CODE XREF: sub_C0DA+4j
		rts
; End of function sub_C0DA


; =============== S U B	R O U T	I N E =======================================


sub_C0F8:							; CODE XREF: sub_C0BC+10p sub_C0DA+10p
		move.w	#$B04A,a0
		tst.w	d1
		blt.w	loc_C132
		cmpnw	 $B,d1
		bgt.w	loc_C132
		asl.w	#7,d1
		btst	#3,$63(a0,d1.w)
		beq.w	loc_C11C
		lsr.w	#7,d1
		move.w	d1,d0
		rts
; ---------------------------------------------------------------------------

loc_C11C:							; CODE XREF: sub_C0F8+1Aj
		bclr	#3,$62(a0,d1.w)
		btst	#3,$64(a0,d1.w)
		bne.w	loc_C132
		bset	#1,$62(a0,d1.w)

loc_C132:							; CODE XREF: sub_C0F8+6j sub_C0F8+Ej ...
		tst.w	d0
		blt.w	locret_C196
		cmpnw	 $B,d0
		bgt.w	locret_C196
		move.w	d0,d1
		asl.w	#7,d1
		btst	#2,($FFFFC2F2).w
		bne.w	loc_C158
		btst	#0,($FFFFC2FA).w
		beq.w	loc_C190

loc_C158:							; CODE XREF: sub_C0F8+52j
		tst.w	$34(a0,d1.w)
		bne.w	loc_C190
		btst	#6,$62(a0,d1.w)
		beq.w	loc_C172
		tst.w	($FFFFD05C).w
		bra.w	loc_C176
; ---------------------------------------------------------------------------

loc_C172:							; CODE XREF: sub_C0F8+6Ej
		tst.w	($FFFFD05A).w

loc_C176:							; CODE XREF: sub_C0F8+76j
		beq.w	loc_C190
		move.w	#0,d0
		btst	#6,$62(a0,d1.w)
		beq.w	loc_C18C
		move.w	#6,d0

loc_C18C:							; CODE XREF: sub_C0F8+8Cj
		bra.w	locret_C196
; ---------------------------------------------------------------------------

loc_C190:							; CODE XREF: sub_C0F8+5Cj sub_C0F8+64j ...
		bset	#3,$62(a0,d1.w)

locret_C196:							; CODE XREF: sub_C0F8+3Cj sub_C0F8+44j ...
		rts
; End of function sub_C0F8

; ---------------------------------------------------------------------------

loc_C198:							; CODE XREF: sub_BC40+1B4p
					; sub_B0E8+1128p
		neg.w	d0
		add.w	$54(a3),d0
		and.w	#7,d0
		btst	#3,4(a3)
		beq.w	byte_C1B2
; ---------------------------------------------------------------------------
		dc.b	1
		dc.b	$3C
		dc.b	0
		dc.b	$F0
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------
byte_C1B2:	dc.b	1					; CODE XREF: ROM:0000C1A8j
		dc.b	$3C
		dc.b	0
		dc.b	$1E
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_C1B8:							; CODE XREF: sub_B0E8+72Ej
					; ROM:0000EC96j
		btst	#2,($FFFFC2F2).w
		beq.w	loc_C1DE
		bclr	#2,($FFFFC2FA).w
		bset	#5,($FFFFC2F2).w
		bne.w	locret_B868
		bclr	#5,($FFFFC2FA).w
		move.w	#$64,($FFFFC31C).w			; 'd'

loc_C1DE:							; CODE XREF: sub_B0E8+10D6j
		move.w	#8,($FFFFBEDC).w
		bset	#3,($FFFFC2EC).w
		clr.w	d0
		move.w	#$128,d1
		btst	#7,$62(a3)
		bne.w	loc_C1FC
		neg.w	d1

loc_C1FC:							; CODE XREF: sub_B0E8+110Ej
		sub.w	(a3),d0
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		move.w	#$F,($FFFFBEDE).w
		move.w	#$7FC,d1
		bsr.s	loc_C198
		beq.w	loc_C220
		move.w	#$92E,d1
		move.w	#$F,($FFFFBEDE).w

loc_C220:							; CODE XREF: sub_B0E8+112Aj
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_C224:							; CODE XREF: sub_B0E8+66Cj
					; ROM:0000ECAAj ...
		cmp.w	#$1C,$5A(a3)
		bge.w	loc_C274
		btst	#3,d0
		bne.w	loc_C23E
		and.w	#7,d0
		move.w	d0,($FFFFBEDC).w

loc_C23E:							; CODE XREF: sub_B0E8+114Aj
		cmp.w	#$10,$5A(a3)
		bge.w	locret_C272
		add.w	d7,($FFFFBEDE).w
		cmp.b	#$14,$6C(a3)
		bge.w	loc_C260
		cmp.w	#8,$5A(a3)
		bgt.w	loc_C268

loc_C260:							; CODE XREF: sub_B0E8+116Aj
		btst	#5,d2

loc_C264:
		beq.w	locret_C272

loc_C268:							; CODE XREF: sub_B0E8+1174j
		neg.w	$5A(a3)
		add.w	#$1C,$5A(a3)

locret_C272:							; CODE XREF: sub_B0E8+115Cj
					; sub_B0E8:loc_C264j
		rts
; ---------------------------------------------------------------------------

loc_C274:							; CODE XREF: sub_B0E8+1142j
		bclr	#4,($FFFFC2F6).w
		move.w	#$B,d0
		btst	#6,$62(a3)
		beq.w	loc_C28C
		move.w	#5,d0

loc_C28C:							; CODE XREF: sub_B0E8+119Cj
		jsr	sub_B86A
		tst.w	d0
		bmi.w	loc_C2BC
		movem.l	a0,-(sp)
		asl.w	#7,d0
		move.l	#$FFFFB04A,a0
		cmp.w	#$250,$58(a0,d0.w)
		beq.w	loc_C2B4
		cmp.w	#$2A2,$58(a0,d0.w)

loc_C2B4:							; CODE XREF: sub_B0E8+11C2j
		movem.l	(sp)+,a0
		bne.w	loc_C2EE

loc_C2BC:							; CODE XREF: sub_B0E8+11ACj
		cmp.w	#1,($FFFFBEDC).w
		ble.w	loc_C2D0
		cmp.w	#7,($FFFFBEDC).w
		bne.w	loc_C2EE

loc_C2D0:							; CODE XREF: sub_B0E8+11DAj
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_C2E0
		neg.w	d0

loc_C2E0:							; CODE XREF: sub_B0E8+11F2j
		cmpnw	 $D8,d0					; ''
		blt.w	loc_C2EE
		bset	#4,($FFFFC2F6).w

loc_C2EE:							; CODE XREF: sub_B0E8+11D0j
					; sub_B0E8+11E4j ...
		bra.w	*+4
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_C2F2:							; CODE XREF: sub_F6CBA+6p

var_3E		=	-$3E
var_3C		=	-$3C
var_3A		=	-$3A

		movem.l	d0-a3,-(sp)
		bclr	#4,($FFFFC2FA).w
		btst	#1,$64(a3)
		beq.w	loc_C30C
		bset	#4,($FFFFC2FA).w

loc_C30C:							; CODE XREF: sub_C2F2+10j
		bsr.w	sub_C566
		move.w	#5,-(sp)
		move.w	$52(a3),($FFFFBED8).w
		bclr	#3,($FFFFC2EC).w
		bset	#5,$62(a3)
		btst	#3,$64(a3)
		bne.w	loc_C33C
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		bne.w	loc_C538

loc_C33C:							; CODE XREF: sub_C2F2+3Aj
		move.w	#$18,(sp)
		bset	#4,($FFFFC2EE).w
		cmp.w	#$92E,$58(a3)
		bne.w	loc_C35E
		move.w	#$14,(sp)
		move.w	($FFFFBEDE).w,d0
		lsr.w	#2,d0
		sub.w	d0,($FFFFBEDE).w

loc_C35E:							; CODE XREF: sub_C2F2+5Aj
		btst	#3,$64(a3)
		beq.w	loc_C378
		movem.l	d0-d1,-(sp)
		move.w	#$1F,d0
		move.w	d0,($FFFFBEDE).w
		movem.l	(sp)+,d0-d1

loc_C378:							; CODE XREF: sub_C2F2+72j
		clr.w	d0
		move.b	$6C(a3),d0
		lsr.b	#1,d0
		move.l	a3,a0
		jsr	sub_15730
		add.w	#$14,d0
		mulu.w	($FFFFBEDE).w,d0
		mulu.w	#$5249,d0
		swap	d0
		move.w	d0,($FFFFBEDE).w
		btst	#0,$6C(a3)
		beq.w	loc_C3AA
		asr.w	#4,d0
		add.w	($FFFFBEDE).w,d0

loc_C3AA:							; CODE XREF: sub_C2F2+AEj
		lsr.w	#4,d0
		neg.w	d0
		addq.w	#3,d0
		bpl.w	loc_C3B6
		clr.w	d0

loc_C3B6:							; CODE XREF: sub_C2F2+BEj
		add.w	d0,(sp)
		st	($FFFFB7AA).w
		move.b	#$10,$5E(a3)
		move.w	$52(a3),($FFFFBEDA).w
		move.w	#$108,d1
		btst	#7,$62(a3)
		bne.w	loc_C3D8
		neg.w	d1

loc_C3D8:							; CODE XREF: sub_C2F2+E0j
		move.w	($FFFFBEDC).w,d2
		asl.w	#2,d2
		lea	unk_C542(pc),a0
		move.w	(a0,d2.w),d0
		move.w	2(a0,d2.w),d2
		sub.w	($FFFFB74A).w,d0
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d2,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		bsr.w	sub_110BE
		tst.w	d0
		bne.w	loc_C408
		addq.w	#1,d0

loc_C408:							; CODE XREF: sub_C2F2+110j
		move.w	d0,d3
		btst	#4,($FFFFC2EA).w
		bne.w	loc_C4A8
		cmpnw	 $C8,d3					; ''
		bhi.w	loc_C442
		jsr	sub_FEFCC
		bmi.w	loc_C4A8
		btst	#0,($FFFFC2FA).w
		bne.w	loc_C4A8
		moveq	#$10,d0
		add.b	$6D(a3),d0
		bsr.w	sub_11086
		cmpnw	 $E,d0
		bgt.w	loc_C4A8

loc_C442:							; CODE XREF: sub_C2F2+126j
		clr.w	d0
		move.b	$6D(a3),d0
		lsr.w	#1,d0
		move.b	d0,-(sp)
		move.w	($FFFFBEDE).w,d0
		lsr.w	#4,d0
		sub.b	(sp)+,d0
		add.b	#$10,d0
		mulu.w	d3,d0
		lsr.w	#6,d0
		cmpnw	 $FA,d3					; ''
		bhi.w	loc_C466
		lsr.w	#1,d0

loc_C466:							; CODE XREF: sub_C2F2+16Ej
		cmpnw	 $88,d0					; ''
		blt.w	loc_C472
		move.w	#$88,d0					; ''

loc_C472:							; CODE XREF: sub_C2F2+178j
		move.w	d0,-(sp)
		bsr.w	sub_1107A
		add.w	d0,$40+var_3E(sp)
		move.w	(sp),d0
		cmpnw	 $3C,d0					; '<'
		bls.w	loc_C48A
		moveq	#$3C,d0					; '<'
		move.w	d0,(sp)

loc_C48A:							; CODE XREF: sub_C2F2+190j
		tst.w	$40+var_3C(sp)
		bpl.w	loc_C494
		lsr.w	#1,d0

loc_C494:							; CODE XREF: sub_C2F2+19Cj
		bsr.w	sub_1107A
		add.w	d0,$40+var_3C(sp)
		move.w	(sp)+,d0
		lsr.w	#1,d0
		bsr.w	sub_11086
		add.w	d0,$3E+var_3A(sp)

loc_C4A8:							; CODE XREF: sub_C2F2+11Ej
					; sub_C2F2+130j ...
		move.w	($FFFFBEDE).w,d2
		muls.w	#$44,d2					; 'D'
		muls.w	(sp)+,d2
		divs.w	d3,d2
		move.w	d2,($FFFFB772).w
		move.w	($FFFFBEDE).w,d2
		muls.w	#$44,d2					; 'D'
		muls.w	(sp)+,d2
		divs.w	d3,d2
		move.w	d2,($FFFFB774).w
		move.w	#$8000,d1
		btst	#7,$62(a3)
		beq.w	loc_C4D8
		clr.w	d1

loc_C4D8:							; CODE XREF: sub_C2F2+1E0j
		eor.w	d2,d1
		bpl.w	loc_C4F4
		move.w	#$3810,($FFFFB774).w
		btst	#7,$62(a3)
		bne.w	loc_C4F4
		move.w	#$C7F0,($FFFFB774).w

loc_C4F4:							; CODE XREF: sub_C2F2+1E8j
					; sub_C2F2+1F8j
		move.w	(sp)+,d1
		beq.w	loc_C538
		mulu.w	($FFFFBEDE).w,d1
		mulu.w	#$44,d1					; 'D'
		divu.w	d3,d1
		mulu.w	#$B33,d3
		divu.w	($FFFFBEDE).w,d3
		add.w	d1,d3
		cmpnw	 $1800,d3
		bls.w	loc_C51A
		move.w	#$1800,d3

loc_C51A:							; CODE XREF: sub_C2F2+220j
		move.w	d3,($FFFFB776).w
		bclr	#4,($FFFFC2F6).w
		beq.w	loc_C538
		btst	#3,$64(a3)
		bne.w	loc_C538
		jsr	sub_F66EE

loc_C538:							; CODE XREF: sub_C2F2+46j
					; sub_C2F2+204j ...
		bsr.w	sub_11132
		movem.l	(sp)+,d0-a3
		rts
; End of function sub_C2F2

; ---------------------------------------------------------------------------
unk_C542:	dc.b	0					; DATA XREF: sub_C2F2+ECo
		dc.b	0
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	0
		dc.b	6
		dc.b	$FF
		dc.b	$F0					; 
		dc.b	0
		dc.b	$C
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	6

; =============== S U B	R O U T	I N E =======================================


sub_C566:							; CODE XREF: sub_C2F2:loc_C30Cp
		btst	#3,$64(a3)
		bne.w	loc_C57A
		btst	#3,$62(a3)
		bne.w	locret_C642

loc_C57A:							; CODE XREF: sub_C566+6j
		moveq	#8,d0
		moveq	#5,d1
		move.w	#$AFCA,a0
		btst	#6,$62(a3)
		bne.w	loc_C590
		add.w	#$300,a0

loc_C590:							; CODE XREF: sub_C566+22j sub_C566+32j
		add.w	#$80,a0					; ''
		tst.w	$34(a0)
		dbeq	d1,loc_C590
		bne.w	loc_C624
		move.b	$28(a0),d0
		ext.w	d0
		asr.w	#1,d0
		add.w	(a0),d0
		sub.w	($FFFFB74A).w,d0
		move.b	$2A(a0),d1
		ext.w	d1
		asr.w	#1,d1
		add.w	$14(a0),d1
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		addq.l	#1,d0
		bsr.w	sub_110BE
		move.w	d0,d2
		movem.w	(sp)+,d0-d1
		moveq	#$12,d3
		move.w	#$108,d4
		btst	#7,$62(a3)
		bne.w	loc_C5E6
		neg.w	d4

loc_C5E6:							; CODE XREF: sub_C566+7Aj
		movem.w	d3-d4,-(sp)
		bsr.w	sub_C644
		move.w	d4,d5
		movem.w	(sp)+,d3-d4
		neg.w	d3
		bsr.w	sub_C644
		add.w	d5,d4
		clr.w	d0
		cmpnw	 $2C,d4					; ','
		bgt.w	loc_C624
		cmpnw	 $FFD4,d4
		blt.w	loc_C624
		btst	#7,$62(a3)
		beq.w	loc_C61A
		neg.w	d4

loc_C61A:							; CODE XREF: sub_C566+AEj
		moveq	#2,d0
		tst.w	d4
		bpl.w	loc_C624
		moveq	#6,d0

loc_C624:							; CODE XREF: sub_C566+36j sub_C566+9Cj ...
		move.w	d0,($FFFFBEDC).w
		btst	#2,($FFFFC2F2).w
		bne.w	loc_C63C
		btst	#0,($FFFFC2FA).w
		beq.w	locret_C642

loc_C63C:							; CODE XREF: sub_C566+C8j
		jsr	sub_FE71C

locret_C642:							; CODE XREF: sub_C566+10j sub_C566+D2j
		rts
; End of function sub_C566


; =============== S U B	R O U T	I N E =======================================


sub_C644:							; CODE XREF: sub_C566+84p sub_C566+90p
		sub.w	($FFFFB74A).w,d3
		sub.w	($FFFFB75E).w,d4
		muls.w	d0,d4
		muls.w	d1,d3
		sub.l	d3,d4
		divs.w	d2,d4
		rts
; End of function sub_C644


; =============== S U B	R O U T	I N E =======================================


sub_C656:							; CODE XREF: sub_9FD0+1F8p
					; sub_A9D6+306p ...
		movem.l	d0-d1,-(sp)
		moveq	#2,d0
		add.w	d4,d0
		add.w	d0,d0
		move.w	#$FFF0,d1
		rol.w	d0,d1
		and.w	d1,($FFFFBE78).w
		move.w	$52(a3),d1
		asl.w	d0,d1
		or.w	d1,($FFFFBE78).w
		movem.l	(sp)+,d0-d1

locret_C678:							; CODE XREF: sub_C67A+6j sub_C67A+Ej ...
		rts
; End of function sub_C656


; =============== S U B	R O U T	I N E =======================================


sub_C67A:							; CODE XREF: sub_C710+14p
					; ROM:0000CD12p ...
		btst	#2,($FFFFC2F2).w
		bne.s	locret_C678
		btst	#3,$62(a3)
		bne.s	locret_C678
		btst	#4,$63(a3)
		bne.s	locret_C678
		tst.b	$60(a3)
		bpl.w	loc_C6A0
		tst.b	$61(a3)
		bmi.s	locret_C678

loc_C6A0:							; CODE XREF: sub_C67A+1Cj
		move.b	$61(a3),d0
		cmp.b	$66(a3),d0
		beq.w	loc_C6D8
		move.w	$36(a3),d0
		cmp.b	#$B,$38(a3,d0.w)
		beq.s	locret_C678
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.s	locret_C678
		addq.w	#4,sp
		bset	#2,$63(a3)
		clr.w	$40(a3)
		move.l	#$B,d0
		bra.w	sub_10662
; ---------------------------------------------------------------------------

loc_C6D8:							; CODE XREF: sub_C67A+2Ej
		addq.w	#4,sp
		bclr	#2,$63(a3)
		bclr	#2,$62(a3)
		st	$61(a3)
		st	$60(a3)
		move.w	$34(a3),d0
		tst.b	$60(a3)
		bpl.w	loc_C700
		jmp	sub_15A88
; ---------------------------------------------------------------------------

loc_C700:							; CODE XREF: sub_C67A+7Cj
		move.b	$60(a3),d0
		ext.w	d0
		move.w	d0,$34(a3)
		jmp	sub_15A88
; End of function sub_C67A


; =============== S U B	R O U T	I N E =======================================


sub_C710:							; DATA XREF: ROM:00018DA8o
		btst	#5,$62(a3)
		bne.w	locret_C678
		cmp.w	#$64,$40(a3)				; 'd'
		beq.w	loc_C814
		bsr.w	sub_C67A
		btst	#4,$63(a3)
		bne.w	sub_10646
		bclr	#1,$62(a3)
		beq.w	loc_C766
		move.w	#8,$42(a3)
		moveq	#$50,d0					; 'P'
		btst	#6,$62(a3)
		bne.w	loc_C750
		neg.w	d0

loc_C750:							; CODE XREF: sub_C710+3Aj
		move.w	d0,$46(a3)
		move.w	#$88,$44(a3)				; ''
		neg.w	$44(a3)
		subq.w	#8,$44(a3)
		clr.w	$40(a3)

loc_C766:							; CODE XREF: sub_C710+28j
		move.b	$61(a3),d0
		cmp.b	$66(a3),d0
		beq.w	loc_C862
		sub.w	d7,$40(a3)
		bpl.w	loc_C848
		addq.w	#8,$40(a3)
		move.w	$14(a3),d0
		sub.w	$46(a3),d0
		cmpnw	 $28,d0					; '('
		bgt.w	loc_C848
		cmpnw	 $FFD8,d0
		blt.w	loc_C848
		move.w	(a3),d0
		sub.w	$44(a3),d0
		cmpnw	 $20,d0					; ' '
		bgt.w	loc_C848
		move.w	#$50C,d1
		tst.w	$34(a3)
		bne.w	loc_C7B4
		move.w	#2,d1

loc_C7B4:							; CODE XREF: sub_C710+9Cj
		bsr.w	sub_1073A
		bset	#2,$62(a3)
		cmp.w	#4,$54(a3)
		beq.w	loc_C7D2
		addq.w	#1,$54(a3)
		and.w	#7,$54(a3)

loc_C7D2:							; CODE XREF: sub_C710+B4j
		clr.w	$2A(a3)
		move.w	#$F800,$28(a3)
		cmpnw	 $10,d0
		bgt.w	locret_C812
		clr.w	$28(a3)
		cmp.w	#4,$54(a3)
		bne.w	locret_C812
		move.w	#$F800,$28(a3)
		move.w	#2,$54(a3)
		move.w	#$FAC,d1
		bsr.w	sub_1073A
		bset	#5,$62(a3)
		move.w	#$64,$40(a3)				; 'd'

locret_C812:							; CODE XREF: sub_C710+D0j sub_C710+DEj ...
		rts
; ---------------------------------------------------------------------------

loc_C814:							; CODE XREF: sub_C710+10j
		clr.w	6(a3)
		clr.w	d0
		move.b	$66(a3),d0
		add.w	d0,d0
		move.l	#$FFFFC6CE,a0
		btst	#6,$62(a3)
		beq.w	loc_C834
		add.w	#$364,a0

loc_C834:							; CODE XREF: sub_C710+11Cj
		move.w	#$FFFE,$66(a0,d0.w)
		move.b	$61(a3),d3
		bsr.w	sub_C868
		jmp	sub_15AA4
; ---------------------------------------------------------------------------

loc_C848:							; CODE XREF: sub_C710+66j sub_C710+7Aj ...
		btst	#2,$62(a3)
		bne.s	locret_C812
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		move.l	#$E594,a0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_C862:							; CODE XREF: sub_C710+5Ej
		bclr	#2,$63(a3)
; End of function sub_C710


; =============== S U B	R O U T	I N E =======================================


sub_C868:							; CODE XREF: sub_C710+12Ep
		move.b	$60(a3),d0
		ext.w	d0
		move.w	d0,$34(a3)
		jsr	sub_15A88
		st	$61(a3)
		st	$60(a3)

locret_C880:							; CODE XREF: sub_C882+6j sub_C8F0+6j ...
		rts
; End of function sub_C868


; =============== S U B	R O U T	I N E =======================================


sub_C882:							; DATA XREF: ROM:00018DA0o
		btst	#5,$62(a3)
		bne.s	locret_C880
		bclr	#1,$62(a3)
		beq.w	loc_C8CA
		clr.w	$28(a3)
		clr.w	$2A(a3)
		move.w	$52(a3),d0
		subq.w	#6,d0
		bmi.w	loc_C8A8
		addq.w	#1,d0

loc_C8A8:							; CODE XREF: sub_C882+20j
		muls.w	#$E,d0
		move.w	d0,$14(a3)
		move.w	#$88,(a3)				; ''
		neg.w	(a3)
		move.w	#2,$54(a3)
		bset	#5,$62(a3)
		move.w	#$F6E,d1
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_C8CA:							; CODE XREF: sub_C882+Ej
		move.w	#4,$54(a3)
		bclr	#2,$62(a3)
		bclr	#5,$63(a3)
		bclr	#2,$63(a3)
		clr.w	$58(a3)
		move.w	#$1000,$28(a3)
		bra.w	sub_10646
; End of function sub_C882


; =============== S U B	R O U T	I N E =======================================


sub_C8F0:							; DATA XREF: ROM:00018DACo
		btst	#5,$62(a3)
		bne.s	locret_C880
		bclr	#1,$62(a3)
		beq.w	loc_C95A
		bset	#2,$63(a3)
		bsr.w	sub_CA0C
		moveq	#$FFFFFFFE,d4
		bsr.w	sub_C656
		move.w	#8,$42(a3)
		moveq	#$B,d0
		move.b	($FFFFC3EA).w,d1
		btst	#6,$62(a3)
		bne.w	loc_C92C
		lsr.w	#4,d1
		neg.w	d0

loc_C92C:							; CODE XREF: sub_C8F0+34j
		and.w	#$F,d1
		cmpnw	 2,d1
		bls.w	loc_C93A
		moveq	#2,d1

loc_C93A:							; CODE XREF: sub_C8F0+44j
		addq.w	#3,d1
		muls.w	d0,d1
		move.w	d1,$46(a3)
		move.w	#$88,$44(a3)				; ''
		clr.w	$40(a3)
		bset	#5,$63(a3)
		clr.w	$4E(a3)
		clr.w	$50(a3)

loc_C95A:							; CODE XREF: sub_C8F0+Ej
		move.w	$14(a3),d0
		sub.w	$46(a3),d0
		cmpnw	 $C,d0
		bgt.w	loc_C9FA
		cmpnw	 $FFF4,d0
		blt.w	loc_C9FA
		move.w	(a3),d0
		sub.w	$44(a3),d0
		cmpnw	 $FFE8,d0
		blt.w	loc_C9FA
		sub.w	d7,$40(a3)
		bpl.w	locret_C880
		addq.w	#8,$40(a3)
		bset	#2,$62(a3)
		move.w	#$50C,d1
		bsr.w	sub_1073A
		moveq	#6,d2
		tst.b	$76(a3)
		beq.w	loc_C9A6
		moveq	#2,d2

loc_C9A6:							; CODE XREF: sub_C8F0+B0j
		cmp.w	$54(a3),d2
		beq.w	loc_C9B8
		addq.w	#1,$54(a3)
		and.w	#7,$54(a3)

loc_C9B8:							; CODE XREF: sub_C8F0+BAj
		clr.w	$2A(a3)
		move.w	#$1000,$28(a3)
		cmpnw	 $FFF8,d0
		blt.w	locret_C880
		clr.w	$28(a3)
		cmp.w	$54(a3),d2
		bne.w	locret_C880
		bset	#5,$62(a3)
		move.w	#2,$54(a3)
		move.w	#$F6E,d1
		bsr.w	sub_1073A
		bclr	#4,$63(a3)
		move.l	#$D,d0
		bra.w	sub_10662
; ---------------------------------------------------------------------------

loc_C9FA:							; CODE XREF: sub_C8F0+76j sub_C8F0+7Ej ...
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		move.l	#locret_15464,a0
		bra.w	sub_103C8
; End of function sub_C8F0


; =============== S U B	R O U T	I N E =======================================


sub_CA0C:							; CODE XREF: sub_C8F0+18p
		btst	#3,$62(a3)
		beq.w	locret_C880
		clr.w	d4
		move.w	$52(a3),d0
		cmp.w	($FFFFC320).w,d0
		beq.w	sub_BFBC
		moveq	#2,d4
		bra.w	sub_BFBC
; End of function sub_CA0C


; =============== S U B	R O U T	I N E =======================================


sub_CA2A:							; DATA XREF: ROM:00018DB0o
		btst	#5,$62(a3)
		bne.w	locret_C880
		moveq	#$10,d0
		btst	#6,$62(a3)
		beq.w	loc_CA42
		moveq	#1,d0

loc_CA42:							; CODE XREF: sub_CA2A+12j
		add.b	d0,($FFFFC3EA).w
		st	$34(a3)
		clr.w	6(a3)
		rts
; End of function sub_CA2A


; =============== S U B	R O U T	I N E =======================================


sub_CA50:							; DATA XREF: ROM:00018DA4o
		btst	#5,$62(a3)
		bne.w	locret_C880
		bclr	#1,$62(a3)
		beq.w	loc_CAB0
		bset	#2,$62(a3)
		bset	#2,$63(a3)
		moveq	#$10,d0
		moveq	#$FFFFFFC4,d1
		btst	#6,$62(a3)
		beq.w	loc_CA82
		moveq	#1,d0
		neg.w	d1

loc_CA82:							; CODE XREF: sub_CA50+2Aj
		sub.b	d0,($FFFFC3EA).w
		move.w	d1,$14(a3)
		clr.w	$28(a3)
		clr.w	$2A(a3)
		move.w	#$86,(a3)				; ''
		move.w	#2,$54(a3)
		bset	#5,$62(a3)
		move.w	#$FAC,d1
		bsr.w	sub_1073A
		jmp	sub_1702E
; ---------------------------------------------------------------------------

loc_CAB0:							; CODE XREF: sub_CA50+10j
		move.w	#4,$54(a3)
		st	$61(a3)
		st	$60(a3)
		bclr	#3,$62(a3)
		bclr	#2,$62(a3)
		bclr	#5,$63(a3)
		bclr	#2,$63(a3)
		move.w	#$F000,$28(a3)
		bra.w	sub_10646
; End of function sub_CA50


; =============== S U B	R O U T	I N E =======================================


sub_CAE0:							; DATA XREF: ROM:00018DFCo
		btst	#5,$62(a3)
		bne.w	locret_CB38
		btst	#2,($FFFFC2F2).w
		bne.w	loc_CB04
		btst	#0,($FFFFC2FA).w
		bne.w	loc_CB04
		nop
		bra.w	sub_10646
; ---------------------------------------------------------------------------

loc_CB04:							; CODE XREF: sub_CAE0+10j sub_CAE0+1Aj
		clr.w	$28(a3)
		clr.w	$2A(a3)
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFB04A,a0
		move.w	($FFFFD408).w,d0
		asl.w	#7,d0
		add.w	d0,a0
		move.w	#$191,$14(a3)
		btst	#7,$62(a0)
		bne.w	loc_CB34
		move.w	#$FE6F,$14(a3)

loc_CB34:							; CODE XREF: sub_CAE0+4Aj
		movem.l	(sp)+,d0/a0

locret_CB38:							; CODE XREF: sub_CAE0+6j
		rts
; End of function sub_CAE0


; =============== S U B	R O U T	I N E =======================================


sub_CB3A:							; DATA XREF: ROM:00018DD4o
		btst	#5,$62(a3)
		bne.w	locret_CB4E
		btst	#0,($FFFFC2EE).w
		beq.w	sub_10646

locret_CB4E:							; CODE XREF: sub_CB3A+6j
		rts
; End of function sub_CB3A


; =============== S U B	R O U T	I N E =======================================


sub_CB50:							; DATA XREF: ROM:00018DD8o
		btst	#5,$62(a3)
		bne.w	locret_CBDE
		btst	#0,($FFFFC2EE).w
		beq.w	loc_CBD4
		bclr	#1,$62(a3)
		beq.w	loc_CB72
		clr.w	$40(a3)

loc_CB72:							; CODE XREF: sub_CB50+1Aj
		move.l	#$FFFFBDA8,a0
		btst	#6,$62(a3)
		beq.w	loc_CB84
		addq.w	#4,a0

loc_CB84:							; CODE XREF: sub_CB50+2Ej
		move.w	$5A(a3),d0
		lsr.w	#2,d0
		addq.w	#1,d0
		tst.b	$76(a3)
		bne.w	loc_CB96
		addq.w	#3,d0

loc_CB96:							; CODE XREF: sub_CB50+40j
		move.w	d0,(a0)
		btst	#3,$62(a3)
		bne.w	locret_CBDE
		bset	#1,$63(a3)
		bne.w	locret_CBDE
		move.w	#$FEA,d1
		cmp.w	#$10,($FFFFB78A).w
		bls.w	loc_CBCA
		moveq	#8,d0
		bsr.w	sub_11086
		tst.w	d0
		beq.w	loc_CBCA
		move.w	#$1014,d1

loc_CBCA:							; CODE XREF: sub_CB50+66j sub_CB50+72j
		bset	#1,$63(a3)
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_CBD4:							; CODE XREF: sub_CB50+10j
		move.b	#$14,$5E(a3)
		bra.w	sub_10646
; ---------------------------------------------------------------------------

locret_CBDE:							; CODE XREF: sub_CB50+6j sub_CB50+4Ej	...
		rts
; End of function sub_CB50


; =============== S U B	R O U T	I N E =======================================


nullsub_4:							; DATA XREF: ROM:00018DCCo
		rts
; End of function nullsub_4


; =============== S U B	R O U T	I N E =======================================


nullsub_5:							; CODE XREF: sub_EB84-1F9Aj
					; sub_EB84-1F92j ...
		rts
; End of function nullsub_5

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_EB84

loc_CBE4:							; CODE XREF: ROM:0000CD0Ej
					; ROM:0000CDF0j ...
		btst	#5,$62(a3)
		bne.s	nullsub_5
		btst	#3,$62(a3)
		bne.s	nullsub_5
		moveq	#8,d0
		bra.w	loc_10750
; END OF FUNCTION CHUNK	FOR sub_EB84
; ---------------------------------------------------------------------------

loc_CBFA:							; DATA XREF: ROM:00018D9Co
		btst	#5,$62(a3)
		bne.s	nullsub_5
		bclr	#1,$62(a3)
		beq.w	loc_CC2C
		move.w	#$5A,$44(a3)				; 'Z'
		tst.w	($FFFFBD1C).w
		bpl.w	loc_CC1E
		neg.w	$44(a3)

loc_CC1E:							; CODE XREF: ROM:0000CC16j
		move.w	($FFFFBD18).w,$46(a3)
		move.w	#$109E,d1
		bsr.w	sub_1073A

loc_CC2C:							; CODE XREF: ROM:0000CC08j
		move.w	$44(a3),d0
		sub.w	(a3),d0
		move.w	$46(a3),d1
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		cmpnw	 7,d0
		bgt.s	nullsub_5
		move.w	d0,d2
		move.w	d2,$54(a3)
		bra.w	loc_10BC2
; ---------------------------------------------------------------------------

loc_CC4E:							; DATA XREF: ROM:00018D98o
		btst	#5,$62(a3)
		bne.s	nullsub_5
		bclr	#1,$62(a3)
		beq.w	loc_CC82
		bsr.w	sub_CCE8
		move.w	#8,$42(a3)
		move.w	#$5A,$44(a3)				; 'Z'
		tst.w	($FFFFBD1C).w
		bpl.w	loc_CC7C
		neg.w	$44(a3)

loc_CC7C:							; CODE XREF: ROM:0000CC74j
		move.w	($FFFFBD18).w,$46(a3)

loc_CC82:							; CODE XREF: ROM:0000CC5Cj
		move.l	#locret_15464,a0
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		btst	#0,($FFFFC2FA).w
		beq.w	loc_CCC2
		tst.w	($FFFFDED0).w
		beq.w	loc_CCB0
		subq.w	#1,($FFFFDED0).w
		bne.w	loc_CCB0
		bset	#2,($FFFFC2EE).w

loc_CCB0:							; CODE XREF: ROM:0000CC9Ej
					; ROM:0000CCA6j
		cmp.w	#$88,(a3)				; ''
		bgt.w	loc_CCC0
		cmp.w	#$FF78,(a3)
		bgt.w	loc_CCC2

loc_CCC0:							; CODE XREF: ROM:0000CCB4j
		clr.w	d1

loc_CCC2:							; CODE XREF: ROM:0000CC96j
					; ROM:0000CCBCj
		sub.w	d7,$40(a3)
		bpl.w	loc_CCF4
		bset	#5,$62(a3)
		move.w	#$E8A,d1
		move.w	($FFFFBED8).w,d0
		cmp.w	$52(a3),d0
		bne.w	loc_CCE4
		move.w	#$EFC,d1

loc_CCE4:							; CODE XREF: ROM:0000CCDCj
		bsr.w	sub_1073A

; =============== S U B	R O U T	I N E =======================================


sub_CCE8:							; CODE XREF: ROM:0000CC60p
		moveq	#$78,d0					; 'x'
		bsr.w	sub_11086
		move.w	d0,$40(a3)
		rts
; End of function sub_CCE8

; ---------------------------------------------------------------------------

loc_CCF4:							; CODE XREF: ROM:0000CCC6j
		btst	#3,$62(a3)
		beq.w	sub_103C8

locret_CCFE:							; CODE XREF: ROM:0000CD06j
					; ROM:0000CD1Cj ...
		rts
; ---------------------------------------------------------------------------

loc_CD00:							; DATA XREF: ROM:00018D80o
		btst	#5,$62(a3)
		bne.s	locret_CCFE
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.s	locret_CCFE
		bclr	#1,$62(a3)
		beq.w	loc_CD32
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_CD32:							; CODE XREF: ROM:0000CD24j
		sub.b	d7,$40(a3)
		bpl.w	loc_CD9E
		move.b	$6A(a3),$40(a3)
		jsr	sub_FEFCC
		bmi.w	loc_CD5C
		btst	#6,($FFFFC2FC).w
		beq.w	loc_CD60
		tst.b	$40(a3)
		beq.w	loc_CD60

loc_CD5C:							; CODE XREF: ROM:0000CD46j
		subq.b	#1,$40(a3)

loc_CD60:							; CODE XREF: ROM:0000CD50j
					; ROM:0000CD58j
		move.l	#2,d0
		btst	#4,$30(a2)
		bne.w	sub_10662
		move.w	($FFFFB75E).w,d1
		btst	#7,$62(a3)
		bne.w	loc_CD80
		neg.w	d1

loc_CD80:							; CODE XREF: ROM:0000CD7Aj
		cmpnw	 $5D,d1					; ']'
		blt.w	sub_10662
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_CD9E
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bmi.w	sub_10662

loc_CD9E:							; CODE XREF: ROM:0000CD36j
					; ROM:0000CD8Cj
		lea	sub_E594(pc),a0
		moveq	#$50,d0					; 'P'
		cmp.w	#2,$34(a3)
		beq.w	loc_CDB0
		neg.w	d0

loc_CDB0:							; CODE XREF: ROM:0000CDAAj
		move.w	#$62,d1					; 'b'
		btst	#7,$62(a3)
		bne.w	loc_CDC2
		neg.w	d0
		neg.w	d1

loc_CDC2:							; CODE XREF: ROM:0000CDBAj
		move.w	($FFFFB74A).w,d2
		eor.w	d0,d2
		bmi.w	loc_CDD4
		move.w	($FFFFB74A).w,d0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_CDD4:							; CODE XREF: ROM:0000CDC8j
		move.w	($FFFFB74A).w,d2
		asr.w	#1,d2
		add.w	d2,d0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_CDE0:							; DATA XREF: ROM:00018D84o
		btst	#5,$62(a3)
		bne.w	locret_CCFE
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.w	locret_CCFE
		bclr	#1,$62(a3)
		beq.w	loc_CE16
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_CE16:							; CODE XREF: ROM:0000CE08j
		sub.b	d7,$40(a3)
		bpl.w	loc_CF82
		move.b	$6B(a3),$40(a3)
		btst	#6,($FFFFC2FC).w
		beq.w	loc_CE3A
		tst.b	$40(a3)
		beq.w	loc_CE3A
		subq.b	#1,$40(a3)

loc_CE3A:							; CODE XREF: ROM:0000CE2Aj
					; ROM:0000CE32j
		move.w	($FFFFB75E).w,d1
		move.w	($FFFFB774).w,d3
		asr.w	#6,d3
		btst	#7,$62(a3)
		bne.w	loc_CE52
		neg.w	d1
		neg.w	d3

loc_CE52:							; CODE XREF: ROM:0000CE4Aj
		tst.w	d3
		bpl.w	loc_CE5A
		clr.w	d3

loc_CE5A:							; CODE XREF: ROM:0000CE54j
		add.w	d1,d3
		cmpnw	 $58,d3					; 'X'
		blt.w	loc_CE92
		btst	#4,$30(a2)
		bne.w	loc_CE92
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_CE92
		bsr.w	sub_E264
		beq.w	loc_CE92
		move.l	#1,d0
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bpl.w	sub_10662

loc_CE92:							; CODE XREF: ROM:0000CE60j
					; ROM:0000CE6Aj ...
		move.w	#$41,$44(a3)				; 'A'
		cmp.w	#2,$34(a3)
		beq.w	loc_CEA6
		neg.w	$44(a3)

loc_CEA6:							; CODE XREF: ROM:0000CE9Ej
		move.w	#$B04A,a0
		cmp.w	#6,$52(a3)
		bge.w	loc_CEB8
		add.w	#$300,a0

loc_CEB8:							; CODE XREF: ROM:0000CEB0j
		moveq	#5,d2
		move.w	#$FC18,d0
		btst	#7,$62(a3)
		beq.w	loc_CF26
		neg.w	d0

loc_CECA:							; CODE XREF: ROM:0000CEF8j
		btst	#2,$63(a0)
		bne.w	loc_CEF4
		tst.w	$34(a0)
		bmi.w	loc_CEF4
		move.w	$2A(a0),d1
		bmi.w	loc_CEE6
		clr.w	d1

loc_CEE6:							; CODE XREF: ROM:0000CEE0j
		asr.w	#4,d1
		add.w	$14(a0),d1
		cmp.w	d0,d1
		bgt.w	loc_CEF4
		move.w	d1,d0

loc_CEF4:							; CODE XREF: ROM:0000CED0j
					; ROM:0000CED8j ...
		add.w	#$80,a0					; ''
		dbf	d2,loc_CECA
		sub.w	#$32,d0					; '2'
		cmpnw	 $FF42,d0
		bgt.w	loc_CF0C
		move.w	#$FF24,d0

loc_CF0C:							; CODE XREF: ROM:0000CF04j
		move.w	d0,$46(a3)
		move.w	($FFFFB74A).w,d0
		move.w	$44(a3),d1
		eor.w	d0,d1
		bpl.w	loc_CF82
		clr.w	$44(a3)
		bra.w	loc_CF82
; ---------------------------------------------------------------------------

loc_CF26:							; CODE XREF: ROM:0000CEC4j
					; ROM:0000CF54j
		btst	#2,$63(a0)
		bne.w	loc_CF50
		tst.w	$34(a0)
		bmi.w	loc_CF50
		move.w	$2A(a0),d1
		bpl.w	loc_CF42
		clr.w	d1

loc_CF42:							; CODE XREF: ROM:0000CF3Cj
		asr.w	#4,d1
		add.w	$14(a0),d1
		cmp.w	d0,d1
		blt.w	loc_CF50
		move.w	d1,d0

loc_CF50:							; CODE XREF: ROM:0000CF2Cj
					; ROM:0000CF34j ...
		add.w	#$80,a0					; ''
		dbf	d2,loc_CF26
		add.w	#$32,d0					; '2'
		cmpnw	 $BE,d0					; ''
		blt.w	loc_CF68
		move.w	#$DC,d0					; ''

loc_CF68:							; CODE XREF: ROM:0000CF60j
		move.w	d0,$46(a3)
		neg.w	$44(a3)
		move.w	($FFFFB74A).w,d0
		move.w	$44(a3),d1
		eor.w	d0,d1
		bpl.w	loc_CF82
		clr.w	$44(a3)

loc_CF82:							; CODE XREF: ROM:0000CE1Aj
					; ROM:0000CF1Aj ...
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		move.l	#$E594,a0
		bsr.w	sub_103C8
		bra.w	sub_EACE
; ---------------------------------------------------------------------------

locret_CF98:							; CODE XREF: ROM:0000CFA0j
					; ROM:0000CFB6j ...
		rts
; ---------------------------------------------------------------------------

loc_CF9A:							; DATA XREF: ROM:00018D88o
		btst	#5,$62(a3)
		bne.s	locret_CF98
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.s	locret_CF98
		bclr	#1,$62(a3)
		beq.w	loc_CFCC
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_CFCC:							; CODE XREF: ROM:0000CFBEj
		sub.b	d7,$40(a3)
		bpl.w	loc_D00C
		move.b	$6B(a3),$40(a3)
		btst	#6,($FFFFC2FC).w
		beq.w	loc_CFF0
		tst.b	$40(a3)
		beq.w	loc_CFF0
		subq.b	#1,$40(a3)

loc_CFF0:							; CODE XREF: ROM:0000CFE0j
					; ROM:0000CFE8j
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_D00C
		move.l	#4,d0
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bpl.w	sub_10662

loc_D00C:							; CODE XREF: ROM:0000CFD0j
					; ROM:0000CFF4j
		moveq	#$64,d0					; 'd'
		cmp.w	#5,$34(a3)
		bne.w	loc_D01A
		neg.w	d0

loc_D01A:							; CODE XREF: ROM:0000D014j
		move.l	#$9E,d1					; ''
		btst	#7,$62(a3)
		beq.w	loc_D05C
		neg.w	d0
		neg.w	d1
		moveq	#$FFFFFFB2,d2
		btst	#4,$30(a2)
		bne.w	loc_D084
		cmp.w	($FFFFB75E).w,d2
		bgt.w	loc_D04A
		move.w	($FFFFB75E).w,d1
		bra.w	loc_D086
; ---------------------------------------------------------------------------

loc_D04A:							; CODE XREF: ROM:0000D03Ej
		move.w	#$FEF8,d2
		cmp.w	($FFFFB75E).w,d2
		blt.w	loc_D086
		move.w	d2,d1
		bra.w	loc_D086
; ---------------------------------------------------------------------------

loc_D05C:							; CODE XREF: ROM:0000D026j
		moveq	#$4E,d2					; 'N'
		btst	#4,$30(a2)
		bne.w	loc_D084
		cmp.w	($FFFFB75E).w,d2
		blt.w	loc_D078
		move.w	($FFFFB75E).w,d1
		bra.w	loc_D086
; ---------------------------------------------------------------------------

loc_D078:							; CODE XREF: ROM:0000D06Cj
		move.w	#$108,d2
		cmp.w	($FFFFB75E).w,d2
		bgt.w	loc_D086

loc_D084:							; CODE XREF: ROM:0000D036j
					; ROM:0000D064j
		move.w	d2,d1

loc_D086:							; CODE XREF: ROM:0000D046j
					; ROM:0000D052j ...
		lea	locret_CF98(pc),a0
		move.w	($FFFFB74A).w,d2
		eor.w	d0,d2
		bmi.w	sub_103C8
		move.w	($FFFFB74A).w,d0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_D09C:							; DATA XREF: ROM:00018D8Co
		btst	#5,$62(a3)
		bne.w	locret_CF98
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.w	locret_CF98
		bclr	#1,$62(a3)
		beq.w	loc_D0D6
		st	$48(a3)
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_D0D6:							; CODE XREF: ROM:0000D0C4j
		sub.b	d7,$40(a3)
		bpl.w	loc_D1BC
		move.b	$6B(a3),$40(a3)
		btst	#6,($FFFFC2FC).w
		beq.w	loc_D0FA
		tst.b	$40(a3)
		beq.w	loc_D0FA
		subq.b	#1,$40(a3)

loc_D0FA:							; CODE XREF: ROM:0000D0EAj
					; ROM:0000D0F2j
		move.l	#3,d0
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_D116
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bmi.w	sub_10662

loc_D116:							; CODE XREF: ROM:0000D104j
		move.w	($FFFFB75E).w,d0
		move.w	($FFFFB774).w,d3
		asr.w	#6,d3
		add.w	d0,d3
		btst	#7,$62(a3)
		bne.w	loc_D130
		neg.w	d0
		neg.w	d3

loc_D130:							; CODE XREF: ROM:0000D128j
		clr.w	d2
		cmpnw	 $FFA8,d3
		blt.w	loc_D15A
		addq.w	#8,d2
		cmpnw	 $58,d0					; 'X'
		blt.w	loc_D15A
		btst	#4,$30(a2)
		bne.w	loc_D15A
		addq.w	#8,d2
		cmpnw	 $108,d3
		blt.w	loc_D15A
		addq.w	#8,d2

loc_D15A:							; CODE XREF: ROM:0000D136j
					; ROM:0000D140j ...
		cmp.w	$48(a3),d2
		bne.w	loc_D170
		move.w	($C00008).l,d0
		and.w	#$7F,d0					; ''
		bne.w	loc_D1BC

loc_D170:							; CODE XREF: ROM:0000D15Ej
		move.w	d2,$48(a3)
		lea	$D19C(pc),a0
		move.w	word_D19E-word_D19C(a0,d2.w),d0
		bsr.w	sub_1107A
		add.w	(a0,d2.w),d0
		move.w	d0,$44(a3)
		move.w	6(a0,d2.w),d0
		bsr.w	sub_1107A
		add.w	4(a0,d2.w),d0
		move.w	d0,$46(a3)
		bra.w	loc_D1BC
; ---------------------------------------------------------------------------
word_D19C:	dc.w	$50
word_D19E:	dc.w	$14					; DATA XREF: ROM:0000D178r
		dc.w	$FFBA
		dc.w	$A
		dc.w	$64
		dc.w	$14
		dc.w	$3A
		dc.w	5
		dc.w	$3C
		dc.w	$32
		dc.w	$E6
		dc.w	$14
		dc.w	$50
		dc.w	$1E
		dc.w	$FA
		dc.w	$14
; ---------------------------------------------------------------------------

loc_D1BC:							; CODE XREF: ROM:0000D0DAj
					; ROM:0000D16Cj ...
		move.w	$44(a3),d0
		cmp.w	#5,$34(a3)
		beq.w	loc_D1CC
		neg.w	d0

loc_D1CC:							; CODE XREF: ROM:0000D1C6j
		move.w	$46(a3),d1
		btst	#7,$62(a3)
		bne.w	loc_D1DE
		neg.w	d0
		neg.w	d1

loc_D1DE:							; CODE XREF: ROM:0000D1D6j
		lea	sub_E594(pc),a0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

locret_D1E6:							; CODE XREF: ROM:0000D1EEj
					; ROM:0000D204j ...
		rts
; ---------------------------------------------------------------------------

loc_D1E8:							; DATA XREF: ROM:00018D90o
		btst	#5,$62(a3)
		bne.s	locret_D1E6
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.s	locret_D1E6
		bclr	#1,$62(a3)
		beq.w	loc_D21A
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_D21A:							; CODE XREF: ROM:0000D20Cj
		sub.b	d7,$40(a3)
		bpl.w	loc_D25A
		move.b	$6B(a3),$40(a3)
		btst	#6,($FFFFC2FC).w
		beq.w	loc_D23E
		tst.b	$40(a3)
		beq.w	loc_D23E
		subq.b	#1,$40(a3)

loc_D23E:							; CODE XREF: ROM:0000D22Ej
					; ROM:0000D236j
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_D25A
		move.l	#6,d0
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bpl.w	sub_10662

loc_D25A:							; CODE XREF: ROM:0000D21Ej
					; ROM:0000D242j
		move.w	($FFFFB74A).w,d0
		asr.w	#1,d0
		move.w	($FFFFB75E).w,d2
		btst	#7,$62(a3)
		bne.w	loc_D270
		neg.w	d2

loc_D270:							; CODE XREF: ROM:0000D26Aj
		moveq	#$FFFFFF80,d1
		cmpnw	 $FFA8,d2
		blt.w	loc_D280
		add.w	($FFFFB75E).w,d1
		asr.w	#1,d1

loc_D280:							; CODE XREF: ROM:0000D276j
		btst	#7,$62(a3)
		bne.w	loc_D28C
		neg.w	d1

loc_D28C:							; CODE XREF: ROM:0000D286j
		lea	locret_D1E6(pc),a0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_D294:							; DATA XREF: ROM:00018D94o
		btst	#5,$62(a3)
		bne.w	locret_D1E6
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		bsr.w	sub_C67A
		btst	#3,$62(a3)
		bne.w	locret_D1E6
		bclr	#1,$62(a3)
		beq.w	loc_D2CE
		st	$48(a3)
		clr.w	$40(a3)
		move.w	#8,$42(a3)

loc_D2CE:							; CODE XREF: ROM:0000D2BCj
		sub.b	d7,$40(a3)
		bpl.w	loc_D3AA
		move.b	$6B(a3),$40(a3)
		btst	#6,($FFFFC2FC).w
		beq.w	loc_D2F2
		tst.b	$40(a3)
		beq.w	loc_D2F2
		subq.b	#1,$40(a3)

loc_D2F2:							; CODE XREF: ROM:0000D2E2j
					; ROM:0000D2EAj
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_D30E
		move.l	#5,d0
		subq.w	#6,d1
		move.w	$52(a3),d2
		subq.w	#6,d2
		eor.w	d2,d1
		bmi.w	sub_10662

loc_D30E:							; CODE XREF: ROM:0000D2F6j
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_D31E
		neg.w	d0

loc_D31E:							; CODE XREF: ROM:0000D318j
		clr.w	d2
		cmpnw	 $FFA8,d0
		blt.w	loc_D348
		addq.w	#8,d2
		cmpnw	 $58,d0					; 'X'
		blt.w	loc_D348
		btst	#4,$30(a2)
		bne.w	loc_D348
		addq.w	#8,d2
		cmpnw	 $108,d0
		blt.w	loc_D348
		addq.w	#8,d2

loc_D348:							; CODE XREF: ROM:0000D324j
					; ROM:0000D32Ej ...
		cmp.w	$48(a3),d2
		bne.w	loc_D35E
		move.w	($C00008).l,d0
		and.w	#$7F,d0					; ''
		bne.w	loc_D3AA

loc_D35E:							; CODE XREF: ROM:0000D34Cj
		move.w	d2,$48(a3)
		lea	$D38A(pc),a0
		move.w	word_D38C-word_D38A(a0,d2.w),d0
		bsr.w	sub_1107A
		add.w	(a0,d2.w),d0
		move.w	d0,$44(a3)
		move.w	6(a0,d2.w),d0
		bsr.w	sub_1107A
		add.w	4(a0,d2.w),d0
		move.w	d0,$46(a3)
		bra.w	loc_D3AA
; ---------------------------------------------------------------------------
word_D38A:	dc.w	0
word_D38C:	dc.w	$3C					; DATA XREF: ROM:0000D366r
		dc.w	$FFBA
		dc.w	$A
		dc.w	0
		dc.w	$3C
		dc.w	$3C
		dc.w	$A
		dc.w	0
		dc.w	$28
		dc.w	$AA
		dc.w	$1E
		dc.w	0
		dc.w	$50
		dc.w	$AA
		dc.w	$14
; ---------------------------------------------------------------------------

loc_D3AA:							; CODE XREF: ROM:0000D2D2j
					; ROM:0000D35Aj ...
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		btst	#7,$62(a3)
		bne.w	loc_D3BE
		neg.w	d1

loc_D3BE:							; CODE XREF: ROM:0000D3B8j
		lea	sub_E594(pc),a0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_D51C

loc_D3C6:							; CODE XREF: sub_D51C+6j
					; DATA XREF: ROM:00018DF0o
		btst	#3,$62(a3)
		bne.w	loc_D3D4
		bra.w	sub_10646
; ---------------------------------------------------------------------------

loc_D3D4:							; CODE XREF: sub_D51C-150j
		btst	#1,$62(a3)
		bne.w	loc_D470
		movem.w	d0-d1,-(sp)
		move.w	(a3),d0
		move.w	$14(a3),d1
		cmpnw	 $AA,d1					; ''
		bgt.w	loc_D410
		cmpnw	 $FF56,d1
		blt.w	loc_D410
		btst	#0,($FFFFC2EA).w
		bne.w	loc_D410
		move.w	d0,-(sp)
		move.w	#8,d0
		jsr	sub_11F62
		move.w	(sp)+,d0

loc_D410:							; CODE XREF: sub_D51C-130j
					; sub_D51C-128j ...
		sub.w	($FFFFBD1C).w,d0
		sub.w	($FFFFBD18).w,d1
		bclr	#2,$64(a3)
		cmpnw	 $74,d0					; 't'
		blt.w	loc_D430
		bset	#2,$64(a3)
		bra.w	loc_D462
; ---------------------------------------------------------------------------

loc_D430:							; CODE XREF: sub_D51C-FAj
		cmpnw	 $FF8C,d0
		bgt.w	loc_D442
		bset	#2,$64(a3)
		bra.w	loc_D462
; ---------------------------------------------------------------------------

loc_D442:							; CODE XREF: sub_D51C-E8j
		cmpnw	 $64,d1					; 'd'
		blt.w	loc_D454
		bset	#2,$64(a3)
		bra.w	loc_D462
; ---------------------------------------------------------------------------

loc_D454:							; CODE XREF: sub_D51C-D6j
		cmpnw	 $FF9C,d1
		bgt.w	loc_D462
		bset	#2,$64(a3)

loc_D462:							; CODE XREF: sub_D51C-F0j sub_D51C-DEj ...
		movem.w	(sp)+,d0-d1
		btst	#2,$64(a3)
		bne.w	loc_D526

loc_D470:							; CODE XREF: sub_D51C-142j
		btst	#5,$62(a3)
		bne.w	locret_15464
		bsr.w	sub_C67A
		bclr	#1,$62(a3)
		beq.w	loc_D496
		clr.w	$40(a3)
		move.w	#8,$42(a3)
		st	$46(a3)

loc_D496:							; CODE XREF: sub_D51C-98j
		btst	#1,$63(a3)
		bne.w	locret_15464
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		tst.w	$48(a3)
		bmi.w	loc_D4EC
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_D4C6
		st	$48(a3)
		bra.w	loc_D4EC
; ---------------------------------------------------------------------------

loc_D4C6:							; CODE XREF: sub_D51C-62j
		sub.w	d7,$48(a3)
		bpl.w	loc_D4EC
		btst	#2,($FFFFC2F2).w
		beq.w	loc_D4E2
		bset	#4,($FFFFC2F2).w
		bra.w	loc_D4EC
; ---------------------------------------------------------------------------

loc_D4E2:							; CODE XREF: sub_D51C-48j
		move.l	#8,d0
		bsr.w	sub_11F62

loc_D4EC:							; CODE XREF: sub_D51C-6Ej sub_D51C-5Aj ...
		sub.b	d7,$40(a3)
		bpl.w	locret_D51A
		move.b	$6B(a3),d0
		beq.w	loc_D508
		btst	#6,($FFFFC2FC).w
		beq.w	loc_D508
		subq.b	#1,d0

loc_D508:							; CODE XREF: sub_D51C-24j sub_D51C-1Aj
		lsr.b	#2,d0
		move.b	d0,$40(a3)
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	*+4

locret_D51A:							; CODE XREF: sub_D51C-2Cj
		rts
; END OF FUNCTION CHUNK	FOR sub_D51C

; =============== S U B	R O U T	I N E =======================================


sub_D51C:							; CODE XREF: sub_E6CE+E0j
					; DATA XREF: ROM:00018DB4o

; FUNCTION CHUNK AT 0000D3C6 SIZE 00000156 BYTES

		btst	#3,$62(a3)
		bne.w	loc_D3C6

loc_D526:							; CODE XREF: sub_D51C-B0j
		move.w	($FFFFB74A).w,($FFFFDEB6).w
		btst	#0,($FFFFC2F4).w
		beq.w	loc_D5C8
		btst	#1,$63(a3)
		bne.w	loc_D5C8
		btst	#1,($FFFFC2F4).w
		beq.w	loc_D57A
		btst	#3,($FFFFC2F4).w
		bne.w	loc_D5C8
		cmp.w	#$2C4,6(a3)
		bne.w	loc_D5C8
		cmp.w	#$2C8,6(a3)
		bne.w	loc_D5C8
		bset	#3,($FFFFC2F4).w
		move.w	#$1C,-(sp)
		bsr.w	sub_11132
		bra.w	loc_D5C8
; ---------------------------------------------------------------------------

loc_D57A:							; CODE XREF: sub_D51C+2Aj
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB75E).w,d0
		move.w	$14(a3),d1
		eor.w	d1,d0
		movem.w	(sp)+,d0-d1
		bmi.w	loc_D5C8
		cmp.w	#$14,(a3)
		bgt.w	loc_D5C8
		cmp.w	#$FFEC,(a3)
		ble.w	loc_D5C8
		bset	#1,($FFFFC2F4).w
		bclr	#3,($FFFFC2F4).w
		move.w	#$1596,d1
		btst	#7,$62(a3)
		beq.w	loc_D5BE
		move.w	#$1684,d1

loc_D5BE:							; CODE XREF: sub_D51C+9Aj
		bsr.w	sub_1073A
		bset	#1,$63(a3)

loc_D5C8:							; CODE XREF: sub_D51C+16j sub_D51C+20j ...
		btst	#5,$62(a3)
		bne.w	locret_15464
		bsr.w	sub_C67A
		bclr	#1,$62(a3)
		beq.w	loc_D5EE
		clr.w	$40(a3)
		move.w	#8,$42(a3)
		st	$46(a3)

loc_D5EE:							; CODE XREF: sub_D51C+C0j
		cmp.w	#$34,(a3)				; '4'
		bgt.w	loc_D626
		cmp.w	#$FFCC,(a3)
		blt.w	loc_D626
		cmp.w	#$10E,$14(a3)
		bgt.w	loc_D626
		cmp.w	#$FEF2,$14(a3)
		blt.w	loc_D626
		cmp.w	#$D2,$14(a3)				; ''
		bgt.w	loc_D648
		cmp.w	#$FF2E,$14(a3)
		blt.w	loc_D648

loc_D626:							; CODE XREF: sub_D51C+D6j sub_D51C+DEj ...
		lea	locret_15464(pc),a0
		moveq	#4,d0
		tst.w	(a3)
		bpl.w	loc_D634
		neg.w	d0

loc_D634:							; CODE XREF: sub_D51C+112j
		move.w	#$E4,d1					; ''
		btst	#7,$62(a3)
		beq.w	sub_103C8
		neg.w	d1
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_D648:							; CODE XREF: sub_D51C+FCj
					; sub_D51C+106j
		btst	#1,$63(a3)
		bne.w	locret_15464
		btst	#0,($FFFFC2F4).w
		beq.w	loc_D680
		cmp.w	#$2C5,6(a3)
		beq.w	loc_D67A
		cmp.w	#$2C9,6(a3)
		bne.w	loc_D680
		move.w	#4,$54(a3)
		bra.w	loc_D680
; ---------------------------------------------------------------------------

loc_D67A:							; CODE XREF: sub_D51C+146j
		move.w	#0,$54(a3)

loc_D680:							; CODE XREF: sub_D51C+13Cj
					; sub_D51C+150j ...
		move.w	#2,d1
		bsr.w	sub_1073A
		btst	#2,($FFFFC2F2).w
		bne.w	loc_D69C
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464

loc_D69C:							; CODE XREF: sub_D51C+172j
		tst.w	$48(a3)
		bmi.w	loc_D6DE
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_D6B8
		st	$48(a3)
		bra.w	loc_D6DE
; ---------------------------------------------------------------------------

loc_D6B8:							; CODE XREF: sub_D51C+190j
		sub.w	d7,$48(a3)
		bpl.w	loc_D6DE
		btst	#2,($FFFFC2F2).w
		beq.w	loc_D6D4
		bset	#4,($FFFFC2F2).w
		bra.w	loc_D6DE
; ---------------------------------------------------------------------------

loc_D6D4:							; CODE XREF: sub_D51C+1AAj
		move.l	#8,d0
		bsr.w	sub_11F62

loc_D6DE:							; CODE XREF: sub_D51C+184j
					; sub_D51C+198j ...
		sub.b	d7,$40(a3)
		bpl.w	loc_DAA4
		move.b	$6B(a3),d0
		beq.w	loc_D6FA
		btst	#6,($FFFFC2FC).w
		beq.w	loc_D6FA
		subq.b	#1,d0

loc_D6FA:							; CODE XREF: sub_D51C+1CEj
					; sub_D51C+1D8j
		lsr.b	#2,d0
		move.b	d0,$40(a3)
		btst	#3,$62(a3)
		beq.w	loc_D71E
		btst	#2,$64(a3)
		bne.w	loc_D71E
		move.l	#$1D,d0
		bra.w	sub_10658
; ---------------------------------------------------------------------------

loc_D71E:							; CODE XREF: sub_D51C+1EAj
					; sub_D51C+1F4j
		move.w	($FFFFB75E).w,d0
		move.w	$14(a3),d1
		eor.w	d0,d1
		bpl.w	loc_D742
		clr.w	d0
		move.w	#$F4,d2					; ''
		btst	#7,$62(a3)
		beq.w	loc_DA58
		neg.w	d2
		bra.w	loc_DA58
; ---------------------------------------------------------------------------

loc_D742:							; CODE XREF: sub_D51C+20Cj
		subq.w	#1,$46(a3)
		bpl.w	loc_D750
		move.w	#$FFFF,$46(a3)

loc_D750:							; CODE XREF: sub_D51C+22Aj
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		bne.w	loc_D7E8
		tst.w	$48(a3)
		bpl.w	loc_D76A
		move.w	#$5A,$48(a3)				; 'Z'

loc_D76A:							; CODE XREF: sub_D51C+244j
		st	$46(a3)
		cmp.w	#$5A,$48(a3)				; 'Z'
		bgt.w	loc_D880
		move.w	($C00008).l,d0
		and.w	#3,d0
		bne.w	loc_D880
		moveq	#5,d0
		move.w	#$B04A,a0
		btst	#6,$62(a3)
		bne.w	loc_D79A
		add.w	#$300,a0

loc_D79A:							; CODE XREF: sub_D51C+276j
					; sub_D51C+2BAj
		btst	#2,$63(a0)
		bne.w	loc_D7D2
		move.w	($FFFFB75E).w,d1
		sub.w	$14(a0),d1
		cmpnw	 $1C,d1
		bgt.w	loc_D7D2
		cmpnw	 $FFE4,d1
		blt.w	loc_D7D2
		move.w	($FFFFB74A).w,d1
		sub.w	(a0),d1
		cmpnw	 $19,d1
		bgt.w	loc_D7D2
		cmpnw	 $FFE7,d1
		bgt.w	loc_D880

loc_D7D2:							; CODE XREF: sub_D51C+284j
					; sub_D51C+294j ...
		add.w	#$80,a0					; ''
		dbf	d0,loc_D79A
		move.w	#1,($FFFFBEE4).w
		bsr.w	sub_E444
		bra.w	loc_D880
; ---------------------------------------------------------------------------

loc_D7E8:							; CODE XREF: sub_D51C+23Cj
		tst.w	$46(a3)
		bne.w	loc_D880
		tst.w	($FFFFB7AA).w
		bpl.w	loc_D880
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		cmpnw	 $14,d0
		bgt.w	loc_D880
		cmpnw	 $FFEC,d0
		blt.w	loc_D880
		move.w	($FFFFB75E).w,d1
		cmpnw	 $108,d1
		bgt.w	loc_D880
		cmpnw	 $FEF8,d1
		blt.w	loc_D880
		sub.w	$14(a3),d1
		cmpnw	 $1E,d1
		bgt.w	loc_D880
		cmpnw	 $FFE2,d1
		blt.w	loc_D880
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB772).w,d0
		bpl.w	loc_D844
		neg.w	d0

loc_D844:							; CODE XREF: sub_D51C+322j
		move.w	($FFFFB774).w,d1
		bpl.w	loc_D84E
		neg.w	d1

loc_D84E:							; CODE XREF: sub_D51C+32Cj
		add.w	d1,d0
		cmpnw	 $1000,d0
		movem.w	(sp)+,d0-d1
		bgt.w	loc_D880
		bsr.w	sub_10676
		move.w	d0,$54(a3)
		move.b	#8,$5E(a3)
		move.w	#$2F4,d1
		bsr.w	sub_1073A
		bset	#1,$63(a3)
		add.w	#$96,($FFFFB89C).w			; ''
		rts
; ---------------------------------------------------------------------------

loc_D880:							; CODE XREF: sub_D51C+258j
					; sub_D51C+266j ...
		move.w	#$BEE6,a0
		move.w	#$104,d3
		btst	#7,$62(a3)
		beq.w	loc_D896
		addq.w	#4,a0
		neg.w	d3

loc_D896:							; CODE XREF: sub_D51C+372j
		cmp.w	#$104,$14(a3)
		bgt.w	loc_D8AA
		cmp.w	#$FEFC,$14(a3)
		bgt.w	loc_D8CE

loc_D8AA:							; CODE XREF: sub_D51C+380j
		clr.w	d2
		clr.w	d0
		cmp.w	#$2C,(a3)				; ','
		bgt.w	loc_DA58
		cmp.w	#$FFD4,(a3)
		blt.w	loc_DA58
		move.w	#$88,d0					; ''
		tst.w	(a3)
		bpl.w	loc_DA58
		neg.w	d0
		bra.w	loc_DA58
; ---------------------------------------------------------------------------

loc_D8CE:							; CODE XREF: sub_D51C+38Aj
		tst.w	($FFFFB7AA).w
		bmi.w	loc_D902
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_D902
		move.l	a0,-(sp)
		move.l	#$FFFFB04A,a0
		move.w	($FFFFB7AA).w,d0
		asl.w	#7,d0
		add.w	d0,a0
		move.w	($FFFFB74A).w,d0
		sub.w	(a0),d0
		asr.w	#1,d0
		add.w	(a0),d0
		move.w	d0,($FFFFDEB6).w
		move.l	(sp)+,a0

loc_D902:							; CODE XREF: sub_D51C+3B6j
					; sub_D51C+3C2j
		move.w	($FFFFDEB6).w,d0
		move.w	($FFFFB75E).w,d1
		bsr.w	sub_DB3E
		bsr.w	sub_10676
		bsr.w	loc_DB68
		move.w	($FFFFC468).w,d0
		and.w	#7,d0
		asl.w	#4,d0
		add.w	#$A0,d0					; ''
		cmp.w	#$DB,($FFFFB75E).w			; ''
		bgt.w	loc_D938
		cmp.w	#$FF25,($FFFFB75E).w
		bgt.w	loc_D93C

loc_D938:							; CODE XREF: sub_D51C+40Ej
		sub.w	#$40,d0					; '@'

loc_D93C:							; CODE XREF: sub_D51C+418j
		move.w	d0,d1
		muls.w	($FFFFB772).w,d0
		swap	d0
		add.w	($FFFFDEB6).w,d0
		muls.w	($FFFFB774).w,d1
		swap	d1
		add.w	($FFFFB75E).w,d1
		bsr.w	sub_DB3E
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmpnl	 $384,d0
		bhi.w	loc_D972
		movem.w	(sp)+,d0-d1
		bra.w	loc_D996
; ---------------------------------------------------------------------------

loc_D972:							; CODE XREF: sub_D51C+44Aj
		bsr.w	sub_110BE
		moveq	#1,d2
		add.w	d0,d2
		moveq	#$12,d4
		btst	#3,($FFFFC2EC).w
		beq.w	loc_D988
		addq.w	#8,d4

loc_D988:							; CODE XREF: sub_D51C+466j
		movem.w	(sp)+,d0-d1
		muls.w	d4,d1
		addq.w	#8,d4
		muls.w	d4,d0
		divs.w	d2,d0
		divs.w	d2,d1

loc_D996:							; CODE XREF: sub_D51C+452j
		add.w	d3,d1
		move.w	d1,d2
		btst	#0,($FFFFC2FA).w
		bne.w	loc_D9EE
		btst	#2,($FFFFC2F2).w
		bne.w	loc_D9EE
		btst	#3,($FFFFC2EC).w
		beq.w	loc_D9EE
		cmp.w	#$D0,($FFFFB75E).w			; ''
		bgt.w	loc_D9D0
		cmp.w	#$FF30,($FFFFB75E).w
		blt.w	loc_D9D0
		bra.w	loc_D9EE
; ---------------------------------------------------------------------------

loc_D9D0:							; CODE XREF: sub_D51C+4A2j
					; sub_D51C+4ACj
		cmp.w	#$14,(a3)
		bgt.w	loc_D9EE
		cmp.w	#$FFEC,(a3)
		blt.w	loc_D9EE
		btst	#1,($FFFFC469).w
		bne.w	loc_D9EE
		bra.w	loc_DA12
; ---------------------------------------------------------------------------

loc_D9EE:							; CODE XREF: sub_D51C+484j
					; sub_D51C+48Ej ...
		cmp.w	#$22,2(a0)				; '"'
		bhi.w	loc_DA58
		cmp.w	#$18,(a0)
		bgt.w	loc_DAB6
		cmp.w	#$FFE8,(a0)
		blt.w	loc_DAB6
		cmp.w	#$C,2(a0)
		bhi.w	loc_DA36

loc_DA12:							; CODE XREF: sub_D51C+4CEj
		cmp.w	#$108,($FFFFB75E).w
		bgt.w	loc_DA36
		cmp.w	#$FEF8,($FFFFB75E).w
		blt.w	loc_DA36
		bset	#1,$63(a3)
		bne.w	loc_DA36
		jsr	sub_DBBA

loc_DA36:							; CODE XREF: sub_D51C+4F2j
					; sub_D51C+4FCj ...
		move.w	(a0),d0
		cmp.w	#$FC,($FFFFB75E).w			; ''
		bgt.w	loc_DA4C
		cmp.w	#$FF04,($FFFFB75E).w
		bgt.w	loc_DA58

loc_DA4C:							; CODE XREF: sub_D51C+522j
		moveq	#$18,d0
		tst.w	($FFFFDEB6).w
		bpl.w	loc_DA58
		neg.w	d0

loc_DA58:							; CODE XREF: sub_D51C+21Cj
					; sub_D51C+222j ...
		move.w	d2,d1
		movem.w	d0-d1,-(sp)
		move.b	$28(a3),d0
		ext.w	d0
		neg.w	d0
		add.w	(sp)+,d0
		sub.w	(a3),d0
		move.b	$2A(a3),d1
		ext.w	d1
		neg.w	d1
		add.w	(sp)+,d1
		sub.w	$14(a3),d1
		cmpnw	 4,d0
		bgt.w	loc_DA9C
		cmpnw	 $FFFC,d0
		blt.w	loc_DA9C
		cmpnw	 4,d1
		bgt.w	loc_DA9C
		cmpnw	 $FFFC,d1
		blt.w	loc_DA9C
		clr.w	d0
		clr.w	d1

loc_DA9C:							; CODE XREF: sub_D51C+560j
					; sub_D51C+568j ...
		bsr.w	sub_10676
		move.b	d0,$43(a3)

loc_DAA4:							; CODE XREF: sub_D51C+1C6j
		move.b	$43(a3),d2
		ext.w	d2
		cmpnw	 7,d2
		ble.w	loc_10BC2
		bra.w	sub_10E1A
; ---------------------------------------------------------------------------

loc_DAB6:							; CODE XREF: sub_D51C+4E0j
					; sub_D51C+4E8j
		tst.w	($FFFFB7AA).w
		bpl.s	loc_DA58
		btst	#2,($FFFFBF80).w
		bne.s	loc_DA58
		move.w	#$C6CE,a1
		lea	$364(a1),a2
		btst	#6,$62(a3)
		beq.w	loc_DAD8
		exg	a1,a2

loc_DAD8:							; CODE XREF: sub_D51C+5B6j
		cmp.l	#$1324,$2A(a1)
		blt.w	loc_DA58
		cmp.l	#$9C4,$2A(a2)
		blt.w	loc_DA58
		cmp.w	#$E0,($FFFFB75E).w			; ''
		bgt.w	loc_DB04
		cmp.w	#$FF20,($FFFFB75E).w
		bgt.w	loc_DA58

loc_DB04:							; CODE XREF: sub_D51C+5DAj
		tst.w	($FFFFB774).w
		btst	#7,$62(a3)
		beq.w	loc_DB16
		eor	#8,ccr

loc_DB16:							; CODE XREF: sub_D51C+5F2j
		bmi.w	loc_DA58
		move.w	($FFFFB772).w,d0
		bpl.w	loc_DB24
		neg.w	d0

loc_DB24:							; CODE XREF: sub_D51C+602j
		move.w	($FFFFB774).w,d1
		bpl.w	loc_DB2E
		neg.w	d1

loc_DB2E:							; CODE XREF: sub_D51C+60Cj
		cmp.w	d0,d1
		blt.w	loc_DA58
		move.l	#$F,d0
		bra.w	sub_10658
; End of function sub_D51C


; =============== S U B	R O U T	I N E =======================================


sub_DB3E:							; CODE XREF: sub_D51C+3EEp
					; sub_D51C+436p
		move.w	($FFFFB7AA).w,d2
		cmp.w	$52(a3),d2
		bne.w	loc_DB4C
		clr.w	d0

loc_DB4C:							; CODE XREF: sub_DB3E+8j
		cmpnw	 $103,d1
		blt.w	loc_DB58
		move.w	#$103,d1

loc_DB58:							; CODE XREF: sub_DB3E+12j
		cmpnw	 $FEFD,d1
		bgt.w	loc_DB64
		move.w	#$FEFD,d1

loc_DB64:							; CODE XREF: sub_DB3E+1Ej
		sub.w	d3,d1
		rts
; End of function sub_DB3E

; ---------------------------------------------------------------------------

loc_DB68:							; CODE XREF: sub_D51C+3F6p
					; sub_B0E8+58C4p ...
		move.w	$54(a3),d1
		sub.w	d1,d0
		beq.w	locret_15464
		neg.w	d0
		and.w	#4,d0
		lsr.w	#1,d0
		subq.w	#1,d0
		btst	#3,$62(a3)
		bne.w	loc_DBAE
; ---------------------------------------------------------------------------
		dc.w	$33C
		dc.w	$42
		dc.w	$6700
		dc.w	$22
		dc.w	$D240
		dc.w	$82B
		dc.w	7
		dc.w	$62
		dc.w	$6600
		dc.w	$A
		dc.w	$33C
		dc.w	$83
		dc.w	$6000
		dc.w	6
		dc.w	$33C
		dc.w	$38
		dc.w	$6700
		dc.w	8
; ---------------------------------------------------------------------------
		neg.w	d0
		add.w	d0,d1

loc_DBAE:							; CODE XREF: ROM:0000DB82j
		add.w	d0,d1
		and.w	#7,d1
		move.w	d1,$54(a3)
		rts

; =============== S U B	R O U T	I N E =======================================


sub_DBBA:							; CODE XREF: sub_B0E8:loc_B5CCp
					; sub_D51C+514p
		move.w	(a0),d0
		sub.w	(a3),d0
		move.w	d3,d1
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		sub.w	$54(a3),d0
		and.w	#7,d0
		move.w	d0,d3
		lsr.w	#2,d0
		btst	#3,4(a3)
		beq.w	loc_DBE2
		eor.w	#1,d0

loc_DBE2:							; CODE XREF: sub_DBBA+20j
		cmp.w	#8,($FFFFB762).w
		bgt.w	loc_DBFA
		cmp.w	#$800,($FFFFB776).w
		bgt.w	loc_DBFA
		bra.w	loc_DC24
; ---------------------------------------------------------------------------

loc_DBFA:							; CODE XREF: sub_DBBA+2Ej sub_DBBA+38j
		movem.l	d0,-(sp)
		move.w	(a3),d0
		sub.w	(a0),d0
		cmpnw	 $10,d0
		bgt.w	loc_DC1C
		cmpnw	 $FFF0,d0
		blt.w	loc_DC1C
		movem.l	(sp)+,d0
		addq.w	#6,d0
		bra.w	loc_DCD0
; ---------------------------------------------------------------------------

loc_DC1C:							; CODE XREF: sub_DBBA+4Cj sub_DBBA+54j
		movem.l	(sp)+,d0
		bra.w	loc_DCD0
; ---------------------------------------------------------------------------

loc_DC24:							; CODE XREF: sub_DBBA+3Cj
		addq.w	#4,d0
		cmp.w	#8,2(a0)
		bls.w	loc_DCAA
		cmp.w	#2,$54(a3)
		beq.w	loc_DCD0
		cmp.w	#6,$54(a3)
		beq.w	loc_DCD0
		tst.w	($FFFFB7AA).w
		bmi.w	loc_DCD0
		subq.w	#2,d0
		move.w	d1,-(sp)
		move.w	#$1000,d1
		move.w	d1,$28(a3)
		or.w	#1,d0
		btst	#7,$62(a3)
		bne.w	loc_DC6A
		eor.w	#1,d0

loc_DC6A:							; CODE XREF: sub_DBBA+A8j
		move.w	#$FFFA,d1
		tst.w	$28(a3)
		bmi.w	loc_DC7A
		neg.w	$28(a3)

loc_DC7A:							; CODE XREF: sub_DBBA+B8j
		tst.w	(a3)
		bpl.w	loc_DC94
		tst.w	$28(a3)
		bpl.w	loc_DC8C
		neg.w	$28(a3)

loc_DC8C:							; CODE XREF: sub_DBBA+CAj
		move.w	#6,d1
		eor.w	#1,d0

loc_DC94:							; CODE XREF: sub_DBBA+C2j
		add.w	d1,(a3)
		btst	#0,$76(a3)
		beq.w	loc_DCA4
		eor.w	#1,d0

loc_DCA4:							; CODE XREF: sub_DBBA+E2j
		move.w	(sp)+,d1
		bra.w	loc_DCD0
; ---------------------------------------------------------------------------

loc_DCAA:							; CODE XREF: sub_DBBA+72j
		movem.l	d0,-(sp)
		move.w	(a3),d0
		sub.w	(a0),d0
		cmpnw	 $10,d0
		ble.w	loc_DCC2
		movem.l	(sp)+,d0
		bra.w	loc_DCCE
; ---------------------------------------------------------------------------

loc_DCC2:							; CODE XREF: sub_DBBA+FCj
		cmpnw	 $FFF0,d0
		movem.l	(sp)+,d0
		bgt.w	loc_DCD0

loc_DCCE:							; CODE XREF: sub_DBBA+104j
		addq.w	#4,d0

loc_DCD0:							; CODE XREF: sub_DBBA+5Ej sub_DBBA+66j ...
		add.w	d0,d0
		lea	word_DD1A(pc),a1
		move.w	(a1,d0.w),d1
		cmpnw	 $178,d1
		bne.w	loc_DCF8
		and.w	#3,d3
		beq.w	loc_DCF8
		cmp.b	#$B,$73(a3)
		blt.w	loc_DCF8
		move.w	#$1AA,d1

loc_DCF8:							; CODE XREF: sub_DBBA+124j
					; sub_DBBA+12Cj ...
		bsr.w	sub_1073A
		add.w	#$96,($FFFFB89C).w			; ''
		add.w	#$A,($FFFFB8A2).w
		asr	$28(a3)
		asr	$28(a3)
		asr	$2A(a3)
		asr	$2A(a3)
		rts
; End of function sub_DBBA

; ---------------------------------------------------------------------------
word_DD1A:	dc.w	$146					; DATA XREF: sub_DBBA+118o
		dc.w	$178
		dc.w	$250
		dc.w	$2A2
		dc.w	$1EC
		dc.w	$21E
		dc.w	$148E
		dc.w	$14C0
		dc.w	$14F2
		dc.w	$1544

; =============== S U B	R O U T	I N E =======================================


sub_DD2E:							; DATA XREF: ROM:00018DB8o
		btst	#3,$62(a3)
		bne.w	sub_10646
		btst	#0,($FFFFC2EA).w
		bne.w	sub_10646
		bsr.w	sub_C67A
		bclr	#1,$62(a3)
		beq.w	loc_DD54
		clr.w	$40(a3)

loc_DD54:							; CODE XREF: sub_DD2E+1Ej
		sub.b	d7,$40(a3)
		bpl.w	loc_DDC0
		move.b	$6B(a3),d0
		beq.w	loc_DD70
		btst	#6,($FFFFC2FC).w
		beq.w	loc_DD70
		subq.b	#1,d0

loc_DD70:							; CODE XREF: sub_DD2E+32j sub_DD2E+3Cj
		lsr.b	#2,d0
		move.b	d0,$40(a3)
		tst.w	($FFFFB7AA).w
		bpl.w	sub_10646
		tst.w	($FFFFB774).w
		btst	#7,$62(a3)
		beq.w	loc_DD90
		eor	#8,ccr

loc_DD90:							; CODE XREF: sub_DD2E+5Aj
		bmi.w	sub_10646
		move.w	#$C6CE,a1
		lea	$364(a1),a2
		btst	#6,$62(a3)
		beq.w	loc_DDA8
		exg	a1,a2

loc_DDA8:							; CODE XREF: sub_DD2E+74j
		cmp.l	#$E10,$2A(a1)
		blt.w	sub_10646
		cmp.l	#$640,$2A(a2)
		blt.w	sub_10646

loc_DDC0:							; CODE XREF: sub_DD2E+2Aj
		bra.w	sub_105A2
; End of function sub_DD2E


; =============== S U B	R O U T	I N E =======================================


sub_DDC4:							; CODE XREF: sub_DEEE+14p
					; sub_E6CE:loc_E6ECp
		bset	#2,($FFFFC2F6).w
		bclr	#1,$62(a3)
		beq.w	loc_DE0C
		move.w	#1,-(sp)
		jsr	sub_11132
		movem.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_DDF8
		move.l	#$FFFFCA32,a2

loc_DDF8:							; CODE XREF: sub_DDC4+2Aj
		addq.w	#1,$358(a2)
		add.w	#$14,($FFFFB8A2).w
		add.w	#$C8,($FFFFB89C).w			; ''
		movem.l	(sp)+,a2

loc_DE0C:							; CODE XREF: sub_DDC4+Cj
		movem.w	d1,-(sp)
		move.w	$14(a3),d0
		move.w	$2A(a3),d1
		beq.w	loc_DE24
		eor.w	d1,d0

loc_DE1E:							; CODE XREF: sub_DDC4+64j
		movem.w	(sp)+,d1
		rts
; ---------------------------------------------------------------------------

loc_DE24:							; CODE XREF: sub_DDC4+54j
		move.w	#$FFFF,d1
		bra.s	loc_DE1E
; End of function sub_DDC4


; =============== S U B	R O U T	I N E =======================================


sub_DE2A:							; CODE XREF: sub_DEEE+6Ep sub_E6CE+56p
		btst	#1,$62(a3)
		beq.w	loc_DE40
		bclr	#0,$64(a3)
		bclr	#1,$64(a3)

loc_DE40:							; CODE XREF: sub_DE2A+6j
		movem.w	d0-d1/a0,-(sp)
		move.w	#$108,d0
		move.w	$14(a3),d1
		bpl.w	loc_DE52
		neg.w	d1

loc_DE52:							; CODE XREF: sub_DE2A+22j
		cmp.w	d0,d1
		bge.w	loc_DECC
		move.w	#$58,d0					; 'X'
		move.w	$14(a3),d1
		btst	#7,$62(a3)
		bne.w	loc_DE76
		neg.w	d0
		cmp.w	d1,d0
		blt.w	loc_DE7C
		bra.w	loc_DE86
; ---------------------------------------------------------------------------

loc_DE76:							; CODE XREF: sub_DE2A+3Cj
		cmp.w	d1,d0
		blt.w	loc_DE86

loc_DE7C:							; CODE XREF: sub_DE2A+44j
		bset	#0,$64(a3)
		bra.w	loc_DECC
; ---------------------------------------------------------------------------

loc_DE86:							; CODE XREF: sub_DE2A+48j sub_DE2A+4Ej
		bclr	#0,$64(a3)
		beq.w	loc_DECC
		move.l	#$FFFFB5CA,a0
		move.w	#$B,d0
		tst.w	d1
		bmi.w	loc_DED2

loc_DEA0:							; CODE XREF: sub_DE2A+8Ej
		cmp.w	$14(a0),d1
		bge.w	loc_DEB4
		tst.w	$34(a0)
		beq.w	loc_DEB4
		bra.w	loc_DECC
; ---------------------------------------------------------------------------

loc_DEB4:							; CODE XREF: sub_DE2A+7Aj sub_DE2A+82j
		suba.w	#$80,a0					; ''
		dbf	d0,loc_DEA0

loc_DEBC:							; CODE XREF: sub_DE2A+C2j
		bset	#1,$64(a3)
		move.w	#1,d0

loc_DEC6:							; CODE XREF: sub_DE2A+A6j
		movem.w	(sp)+,d0-d1/a0
		rts
; ---------------------------------------------------------------------------

loc_DECC:							; CODE XREF: sub_DE2A+2Aj sub_DE2A+58j ...
		move.w	#0,d0
		bra.s	loc_DEC6
; ---------------------------------------------------------------------------

loc_DED2:							; CODE XREF: sub_DE2A+72j sub_DE2A+BEj
		cmp.w	$14(a0),d1
		ble.w	loc_DEE4
		tst.w	$34(a0)
		beq.w	loc_DEE4
		bra.s	loc_DECC
; ---------------------------------------------------------------------------

loc_DEE4:							; CODE XREF: sub_DE2A+ACj sub_DE2A+B4j
		suba.w	#$80,a0					; ''
		dbf	d0,loc_DED2
		bra.s	loc_DEBC
; End of function sub_DE2A


; =============== S U B	R O U T	I N E =======================================


sub_DEEE:							; DATA XREF: ROM:00018E00o
		btst	#0,($FFFFC2FA).w
		bne.w	loc_DF0A
		btst	#2,($FFFFC2F2).w
		bne.w	loc_DF0A
		bsr.w	sub_DDC4
		bpl.w	loc_DF18

loc_DF0A:							; CODE XREF: sub_DEEE+6j sub_DEEE+10j
		bclr	#1,$64(a3)
		move.w	#$10,d0
		bsr.w	sub_10662

loc_DF18:							; CODE XREF: sub_DEEE+18j
		bclr	#2,($FFFFC2F6).w
		bne.w	loc_DF28
		bclr	#1,$64(a3)

loc_DF28:							; CODE XREF: sub_DEEE+30j
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		bne.w	sub_10646
		btst	#2,($FFFFC2F2).w
		beq.w	loc_DF48
		cmp.w	#1,($FFFFC31A).w
		bgt.w	locret_E1CA

loc_DF48:							; CODE XREF: sub_DEEE+4Cj
		btst	#1,$64(a3)
		bne.w	loc_DF72
		btst	#2,($FFFFC2F2).w
		bne.w	loc_DF72
		jsr	sub_DE2A
		beq.w	loc_DF72
		move.w	#$21,d0					; '!'
		bsr.w	sub_10662
		bra.w	*+4

loc_DF72:							; CODE XREF: sub_DEEE+60j sub_DEEE+6Aj ...
		btst	#5,$62(a3)
		bne.w	locret_15464
		btst	#2,($FFFFC2F2).w
		bne.w	loc_DF90
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4

loc_DF90:							; CODE XREF: sub_DEEE+94j
		btst	#3,$62(a3)
		bne.w	sub_10646
		bclr	#1,$62(a3)
		beq.w	loc_E008
		btst	#1,$64(a3)
		beq.w	loc_DFB8
		move.w	#1,-(sp)
		jsr	sub_11132

loc_DFB8:							; CODE XREF: sub_DEEE+BCj
		btst	#1,$64(a3)
		beq.w	loc_DFF0
		movem.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_DFDC
		move.l	#$FFFFCA32,a2

loc_DFDC:							; CODE XREF: sub_DEEE+E4j
		addq.w	#1,$358(a2)
		add.w	#$14,($FFFFB8A2).w
		add.w	#$C8,($FFFFB89C).w			; ''
		movem.l	(sp)+,a2

loc_DFF0:							; CODE XREF: sub_DEEE+D0j
		clr.w	$40(a3)
		move.w	#8,$42(a3)
		move.w	($C00008).l,d0
		and.w	#3,d0
		move.w	d0,$44(a3)

loc_E008:							; CODE XREF: sub_DEEE+B2j
		sub.b	d7,$40(a3)
		bpl.w	loc_E0B8
		move.b	$6A(a3),$40(a3)
		jsr	sub_FEFCC
		bmi.w	loc_E032
		btst	#6,($FFFFC2FC).w
		beq.w	loc_E036
		tst.b	$40(a3)
		beq.w	loc_E036

loc_E032:							; CODE XREF: sub_DEEE+12Ej
		subq.b	#1,$40(a3)

loc_E036:							; CODE XREF: sub_DEEE+138j
					; sub_DEEE+140j
		btst	#2,($FFFFC2F2).w
		bne.w	loc_E0A4
		btst	#0,($FFFFC2FA).w
		bne.w	loc_E0A4
		bsr.w	sub_E62E
		bsr.w	sub_E1F4
		btst	#0,($FFFFC2FA).w
		bne.w	loc_E0A4
		btst	#2,($FFFFC2F2).w
		bne.w	loc_E0A4
		btst	#1,$64(a3)
		bne.w	loc_E09C
		move.w	#1,d0
		btst	#6,$62(a3)
		beq.w	loc_E082
		move.w	#2,d0

loc_E082:							; CODE XREF: sub_DEEE+18Cj
		cmp.w	($FFFFC328).w,d0
		beq.w	loc_E09C
		cmp.w	($FFFFC32A).w,d0
		beq.w	loc_E09C
		btst	#1,($FFFFB03F).w
		bne.w	loc_E0B4

loc_E09C:							; CODE XREF: sub_DEEE+17Ej
					; sub_DEEE+198j ...
		bsr.w	sub_E292
		bra.w	loc_E0B4
; ---------------------------------------------------------------------------

loc_E0A4:							; CODE XREF: sub_DEEE+14Ej
					; sub_DEEE+158j ...
		jsr	sub_FE8EC
		bne.w	loc_E0B8
		jsr	sub_E412

loc_E0B4:							; CODE XREF: sub_DEEE+1AAj
					; sub_DEEE+1B2j
		bsr.w	sub_E444

loc_E0B8:							; CODE XREF: sub_DEEE+11Ej
					; sub_DEEE+1BCj
		moveq	#6,d0
		add.w	$44(a3),d0
		lea	word_E1CC(pc),a0
		btst	#7,($FFFFC2EE).w
		beq.w	loc_E0D0
		move.w	$34(a3),d0

loc_E0D0:							; CODE XREF: sub_DEEE+1DAj
		asl.w	#2,d0
		move.w	2(a0,d0.w),d1
		move.w	(a0,d0.w),d0
		btst	#2,($FFFFC2F2).w
		bne.w	loc_E0EE
		btst	#0,($FFFFC2FA).w
		beq.w	loc_E0FE

loc_E0EE:							; CODE XREF: sub_DEEE+1F2j
		jsr	sub_FE864
		cmp.b	#$80,($FFFFDA16).w
		beq.w	locret_E1CA

loc_E0FE:							; CODE XREF: sub_DEEE+1FCj
		btst	#7,$62(a3)
		bne.w	loc_E10C
		neg.w	d0
		neg.w	d1

loc_E10C:							; CODE XREF: sub_DEEE+216j
		lea	loc_E12C(pc),a0
		btst	#2,($FFFFC2F2).w
		bne.w	loc_E124
		btst	#0,($FFFFC2FA).w
		beq.w	loc_E128

loc_E124:							; CODE XREF: sub_DEEE+228j
		lea	locret_E1CA(pc),a0

loc_E128:							; CODE XREF: sub_DEEE+232j
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_E12C:							; DATA XREF: sub_DEEE:loc_E10Co
		ext.w	d0
		move.b	$28(a3),d2
		ext.w	d2
		add.w	($FFFFB74A).w,d2
		move.b	$2A(a3),d3
		ext.w	d3
		add.w	($FFFFB75E).w,d3
		clr.w	($FFFFBEE4).w
		moveq	#5,d4
		move.w	#$B04A,a0
		cmp.w	#6,$52(a3)
		bge.w	loc_E15A
		add.w	#$300,a0

loc_E15A:							; CODE XREF: sub_DEEE+264j
					; sub_DEEE+2D8j
		move.b	$28(a0),d1
		ext.w	d1
		add.w	(a0),d1
		sub.w	d2,d1
		cmpnw	 $14,d1
		bgt.w	loc_E1C2
		cmpnw	 $FFEC,d1
		blt.w	loc_E1C2
		move.b	$2A(a0),d1
		ext.w	d1
		add.w	$14(a0),d1
		sub.w	d3,d1
		cmpnw	 $14,d1
		bgt.w	loc_E1C2
		cmpnw	 $FFEC,d1
		blt.w	loc_E1C2
		addq.w	#1,($FFFFBEE4).w
		move.w	(a3),d0
		sub.w	(a0),d0
		move.w	$14(a3),d1
		sub.w	$14(a0),d1
		bsr.w	sub_10676
		move.w	$54(a3),d1
		eor.w	#4,d1
		cmp.w	d0,d1
		bne.w	loc_E1C2
		move.w	($C00008).l,d1
		and.w	#1,d1
		add.w	d1,d0
		and.w	#7,d0

loc_E1C2:							; CODE XREF: sub_DEEE+27Aj
					; sub_DEEE+282j ...
		add.w	#$80,a0					; ''
		dbf	d4,loc_E15A

locret_E1CA:							; CODE XREF: sub_DEEE+56j
					; sub_DEEE+20Cj
					; DATA XREF: ...
		rts
; End of function sub_DEEE

; ---------------------------------------------------------------------------
word_E1CC:	dc.w	$FF9C					; DATA XREF: sub_DEEE+1D0o
		dc.w	$FFEC
		dc.w	$64
		dc.w	$FFEC
		dc.w	$FF88
		dc.w	$28
		dc.w	$14
		dc.w	$28
		dc.w	$78
		dc.w	$28
		dc.w	$FFEC
		dc.w	$28
		dc.w	$FFD8
		dc.w	$F0
		dc.w	0
		dc.w	$DC
		dc.w	$14
		dc.w	$E6
		dc.w	$1E
		dc.w	$E6

; =============== S U B	R O U T	I N E =======================================


sub_E1F4:							; CODE XREF: sub_DEEE+160p
		btst	#4,($FFFFC2FC).w
		bne.w	locret_15464
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_E20E
		neg.w	d0

loc_E20E:							; CODE XREF: sub_E1F4+14j
		tst.w	d0
		bmi.w	locret_15464
		cmpnw	 $58,d0					; 'X'
		bgt.w	locret_15464
		move.w	($C00008).l,d0
		and.w	#3,d0
		bne.w	locret_15464
		btst	#3,$63(a3)
		bne.w	locret_15464
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		btst	#6,$62(a3)
		beq.w	loc_E248
		dc.w	$C549					; exg.l	 a2,a1

loc_E248:							; CODE XREF: sub_E1F4+4Ej
		bsr.w	sub_12EF6
		cmpnw	 $C00,d0
		bhi.w	locret_15464
		bsr.w	sub_F790
		bsr.w	sub_15788
		bsr.w	loc_12C04
		bra.w	sub_E412
; End of function sub_E1F4


; =============== S U B	R O U T	I N E =======================================


sub_E264:							; CODE XREF: ROM:0000CE76p sub_E292p
		btst	#5,($FFFFC2EE).w
		bne.w	sub_E274
		eor	#4,ccr
		rts
; End of function sub_E264


; =============== S U B	R O U T	I N E =======================================


sub_E274:							; CODE XREF: sub_E264+6j sub_E6CE+26Ap ...
		movem.l	d0-d1,-(sp)
		btst	#6,($FFFFC2EE).w
		move	sr,d0
		btst	#6,$62(a3)
		move	sr,d1
		eor.w	d1,d0
		move	d0,ccr
		movem.l	(sp)+,d0-d1
		rts
; End of function sub_E274


; =============== S U B	R O U T	I N E =======================================


sub_E292:							; CODE XREF: sub_DEEE:loc_E09Cp
		bsr.s	sub_E264
		bne.w	loc_E2E4
		tst.w	($FFFFBEE4).w
		beq.w	loc_E2E4
		move.w	($C00008).l,d0
		and.w	#3,d0
		bne.w	loc_E2E4
		move.w	#$58,d0					; 'X'
		btst	#7,$62(a3)
		beq.w	loc_E2CC
		move.w	#$FFA8,d0
		cmp.w	$14(a3),d0
		blt.w	sub_E412
		bra.w	loc_E2D4
; ---------------------------------------------------------------------------

loc_E2CC:							; CODE XREF: sub_E292+26j
		cmp.w	$14(a3),d0
		bgt.w	sub_E412

loc_E2D4:							; CODE XREF: sub_E292+36j
		bset	#3,($FFFFC2F6).w
		jmp	loc_BC64
; ---------------------------------------------------------------------------
		bra.w	sub_E412
; ---------------------------------------------------------------------------

loc_E2E4:							; CODE XREF: sub_E292+2j sub_E292+Aj ...
		btst	#1,$64(a3)
		beq.w	loc_E32C
		move.w	#$1E,d0
		jsr	sub_11086
		add.w	#$82,d0					; ''
		cmp.w	($FFFFB75E).w,d0
		blt.w	loc_E312
		neg.w	d0
		cmp.w	($FFFFB75E).w,d0
		bgt.w	loc_E312
		bra.w	loc_E32C
; ---------------------------------------------------------------------------

loc_E312:							; CODE XREF: sub_E292+6Ej sub_E292+78j
		move.w	#2,d0
		jsr	sub_1107A
		add.w	$54(a3),d0
		and.w	#7,d0
		move.w	d0,$54(a3)
		bra.w	sub_E412
; ---------------------------------------------------------------------------

loc_E32C:							; CODE XREF: sub_E292+58j sub_E292+7Cj
		moveq	#$20,d4					; ' '
		clr.w	d1
		move.b	$70(a3),d1
		lsr.w	#1,d1
		sub.b	d1,d4
		asl.w	#4,d4
		move.w	#$108,d1
		btst	#7,$62(a3)
		bne.w	loc_E34A
		neg.w	d1

loc_E34A:							; CODE XREF: sub_E292+B2j
		sub.w	($FFFFB75E).w,d1
		move.w	($FFFFB74A).w,d0
		neg.w	d0
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d0,d1
		cmpnl	 $2710,d1
		movem.w	(sp)+,d0-d1
		bhi.w	loc_E3C4
		lsr.w	#4,d4
		bsr.w	sub_10676
		move.w	d0,d5
		moveq	#5,d3
		move.w	#$B04A,a1
		cmp.w	#6,$52(a3)
		bge.w	loc_E388
		add.w	#$300,a1

loc_E388:							; CODE XREF: sub_E292+EEj
					; sub_E292+12Ej
		tst.w	$34(a1)
		beq.w	loc_E3AE
		move.w	(a1),d0
		sub.w	($FFFFB74A).w,d0
		move.w	$14(a1),d1
		sub.w	($FFFFB75E).w,d1
		bsr.w	sub_10676
		cmp.w	d5,d0
		bne.w	loc_E3BC
		asl.w	#1,d4
		bra.w	loc_E3BC
; ---------------------------------------------------------------------------

loc_E3AE:							; CODE XREF: sub_E292+FAj
		btst	#1,$63(a1)
		beq.w	loc_E3BC
		clr.w	d3
		moveq	#1,d4

loc_E3BC:							; CODE XREF: sub_E292+112j
					; sub_E292+118j ...
		add.w	#$80,a1					; ''
		dbf	d3,loc_E388

loc_E3C4:							; CODE XREF: sub_E292+D6j
		move.w	d4,d0
		bsr.w	sub_11086
		btst	#0,$70(a3)
		beq.w	loc_E3E0
		cmpnw	 7,d0
		bgt.w	locret_15464
		bra.w	loc_E3E8
; ---------------------------------------------------------------------------

loc_E3E0:							; CODE XREF: sub_E292+13Ej
		cmpnw	 8,d0
		bgt.w	locret_15464

loc_E3E8:							; CODE XREF: sub_E292+14Aj
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_E3F8
		neg.w	d0

loc_E3F8:							; CODE XREF: sub_E292+160j
		tst.w	d0
		bmi.w	locret_15464
		move.w	#$108,d1
		sub.w	d0,d1
		bmi.w	locret_15464
		btst	#7,($FFFFC2EE).w
		bne.w	locret_15464
; End of function sub_E292


; =============== S U B	R O U T	I N E =======================================


sub_E412:							; CODE XREF: sub_DEEE+1C0p
					; sub_E1F4+6Cj	...
		addq.w	#4,sp
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_E424
		neg.w	d0

loc_E424:							; CODE XREF: sub_E412+Cj
		move.w	#$108,d1
		sub.w	d0,d1
		lsr.w	#3,d1
		cmpnw	 $14,d1
		blt.w	loc_E436
		moveq	#$14,d1

loc_E436:							; CODE XREF: sub_E412+1Ej
		move.w	d1,$42(a3)
		move.l	#$12,d0
		bra.w	sub_10662
; End of function sub_E412


; =============== S U B	R O U T	I N E =======================================


sub_E444:							; CODE XREF: sub_D51C+2C4p
					; sub_DEEE:loc_E0B4p
		tst.w	($FFFFBEE4).w
		bne.w	loc_E45E
		moveq	#$10,d0
		add.b	$70(a3),d0
		bsr.w	sub_11086
		cmpnw	 $C,d0
		bgt.w	locret_15464

loc_E45E:							; CODE XREF: sub_E444+4j
		moveq	#6,d0
		bsr.w	sub_11086
		cmp.w	#6,$52(a3)
		blt.w	loc_E470
		addq.w	#6,d0

loc_E470:							; CODE XREF: sub_E444+26j
		tst.w	$34(a3)
		beq.w	loc_E4A0
		cmp.w	#$28,($FFFFB74A).w			; '('
		bgt.w	loc_E4A0
		cmp.w	#$FFD8,($FFFFB74A).w
		blt.w	loc_E4A0
		cmp.w	#$CC,($FFFFB75E).w			; ''
		bgt.w	locret_15464
		cmp.w	#$FF34,($FFFFB75E).w
		blt.w	locret_15464

loc_E4A0:							; CODE XREF: sub_E444+30j sub_E444+3Aj ...
		cmp.w	$52(a3),d0
		beq.w	locret_15464
		asl.w	#7,d0
		move.w	#$B04A,a0
		add.w	d0,a0
		tst.w	$34(a0)
		ble.w	locret_15464
		btst	#2,$63(a0)
		bne.w	locret_15464
		btst	#5,$62(a0)
		bne.w	locret_15464
		move.w	$14(a0),d0
		move.w	$14(a3),d1
		btst	#7,$62(a3)
		bne.w	loc_E4E2
		neg.w	d0
		neg.w	d1

loc_E4E2:							; CODE XREF: sub_E444+96j
		btst	#5,($FFFFC2EA).w
		beq.w	loc_E502
		movem.w	d0-d1,-(sp)
		sub.w	#$58,d0					; 'X'
		sub.w	#$58,d1					; 'X'
		eor.w	d0,d1
		movem.w	(sp)+,d0-d1
		bmi.w	locret_15464

loc_E502:							; CODE XREF: sub_E444+A4j
		cmpnw	 $58,d0					; 'X'
		bgt.w	loc_E514
		sub.w	d1,d0
		cmpnw	 $FFF1,d0
		blt.w	locret_15464

loc_E514:							; CODE XREF: sub_E444+C2j
		move.w	(a0),d0
		sub.w	($FFFFB74A).w,d0
		move.w	$14(a0),d1
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		bsr.w	sub_10676
		move.w	d0,($FFFFBEDC).w
		movem.w	(sp)+,d1-d2
		muls.w	d1,d1
		muls.w	d2,d2
		add.l	d1,d2
		moveq	#5,d3
		move.w	#$B04A,a1
		cmp.w	#6,$52(a3)
		bge.w	loc_E54C
		add.w	#$300,a1

loc_E54C:							; CODE XREF: sub_E444+100j
					; sub_E444+13Aj
		move.w	(a1),d0
		sub.w	($FFFFB74A).w,d0
		move.w	$14(a1),d1
		sub.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d0,d1
		cmp.l	d2,d1
		movem.w	(sp)+,d0-d1
		bhi.w	loc_E57A
		bsr.w	sub_10676
		cmp.w	($FFFFBEDC).w,d0
		beq.w	locret_15464

loc_E57A:							; CODE XREF: sub_E444+126j
		add.w	#$80,a1					; ''
		dbf	d3,loc_E54C
		bsr.w	loc_BC64
		tst.w	$34(a3)
		beq.w	locret_15464
		addq.w	#4,sp
		bra.w	sub_10646
; End of function sub_E444


; =============== S U B	R O U T	I N E =======================================


sub_E594:							; DATA XREF: ROM:loc_CD9Eo
					; ROM:loc_D1DEo ...
		tst.w	($FFFFB7AA).w
		bmi.w	locret_15464
		move.b	$28(a3),d2
		sub.b	($FFFFB772).w,d2
		ext.w	d2
		add.w	(a3),d2
		sub.w	($FFFFB74A).w,d2
		cmpnw	 $28,d2					; '('
		bgt.w	locret_15464
		cmpnw	 $FFD8,d2
		blt.w	locret_15464
		move.b	$2A(a3),d1
		sub.b	($FFFFB774).w,d1
		ext.w	d1
		add.w	$14(a3),d1
		sub.w	($FFFFB75E).w,d1
		cmpnw	 $28,d1					; '('
		bgt.w	locret_15464
		cmpnw	 $FFD8,d1
		blt.w	locret_15464
		move.w	(a3),d0
		sub.w	($FFFFB74A).w,d0
		move.w	$14(a3),d1
		sub.w	($FFFFB75E).w,d1
		bsr.w	sub_10676
		btst	#5,($FFFFC2EA).w
		beq.w	locret_E62C
		move.w	$14(a3),d1
		btst	#7,$62(a3)
		bne.w	loc_E60A
		neg.w	d1

loc_E60A:							; CODE XREF: sub_E594+70j
		sub.w	#$58,d1					; 'X'
		cmpnw	 $A,d1
		bgt.w	locret_E62C
		cmpnw	 $FFCE,d1
		blt.w	locret_E62C
		moveq	#2,d0
		move.w	(a3),d1
		cmp.w	($FFFFB74A).w,d1
		bgt.w	locret_E62C
		moveq	#6,d0

locret_E62C:							; CODE XREF: sub_E594+62j sub_E594+7Ej ...
		rts
; End of function sub_E594


; =============== S U B	R O U T	I N E =======================================


sub_E62E:							; CODE XREF: sub_B0E8:loc_B72Ep
					; sub_DEEE+15Cp
		btst	#5,($FFFFC2EA).w
		beq.w	locret_15464
		btst	#2,($FFFFC2F2).w
		bne.w	locret_15464
		movem.l	d0-d1/a0,-(sp)
		moveq	#5,d0
		move.w	#$B04A,a0
		cmp.w	#6,$52(a3)
		blt.w	loc_E65A
		add.w	#$300,a0

loc_E65A:							; CODE XREF: sub_E62E+24j
		moveq	#$5C,d1					; '\'
		btst	#7,$62(a3)
		bne.w	loc_E6B0
		neg.w	d1
		cmp.w	($FFFFB75E).w,d1
		bgt.w	loc_E688

loc_E670:							; CODE XREF: sub_E62E+56j
		tst.w	$34(a0)
		bmi.w	loc_E680
		cmp.w	$14(a0),d1
		bgt.w	loc_E69A

loc_E680:							; CODE XREF: sub_E62E+46j
		add.w	#$80,a0					; ''
		dbf	d0,loc_E670

loc_E688:							; CODE XREF: sub_E62E+3Ej sub_E62E+86j ...
		moveq	#$40,d0					; '@'
		bclr	#7,($FFFFC2EE).w
		bne.w	loc_E6A4

loc_E694:							; CODE XREF: sub_E62E+74j sub_E62E+7Aj ...
		movem.l	(sp)+,d0-d1/a0
		rts
; ---------------------------------------------------------------------------

loc_E69A:							; CODE XREF: sub_E62E+4Ej sub_E62E+94j
		moveq	#6,d0
		bset	#7,($FFFFC2EE).w
		bne.s	loc_E694

loc_E6A4:							; CODE XREF: sub_E62E+62j
		tst.w	($FFFFC3F0).w
		bpl.s	loc_E694
		bsr.w	sub_126F8
		bra.s	loc_E694
; ---------------------------------------------------------------------------

loc_E6B0:							; CODE XREF: sub_E62E+34j
		cmp.w	($FFFFB75E).w,d1
		blt.s	loc_E688

loc_E6B6:							; CODE XREF: sub_E62E+9Aj
		tst.w	$34(a0)
		bmi.w	loc_E6C4
		cmp.w	$14(a0),d1
		blt.s	loc_E69A

loc_E6C4:							; CODE XREF: sub_E62E+8Cj
		add.w	#$80,a0					; ''
		dbf	d0,loc_E6B6
		bra.s	loc_E688
; End of function sub_E62E


; =============== S U B	R O U T	I N E =======================================


sub_E6CE:							; DATA XREF: ROM:00018E04o
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_E6EC

loc_E6DA:							; CODE XREF: sub_E6CE+24j
		bclr	#1,$64(a3)
		move.w	#$11,d0
		bsr.w	sub_10662
		bra.w	loc_E6F4
; ---------------------------------------------------------------------------

loc_E6EC:							; CODE XREF: sub_E6CE+8j
		jsr	sub_DDC4
		bmi.s	loc_E6DA

loc_E6F4:							; CODE XREF: sub_E6CE+1Aj
		bclr	#2,($FFFFC2F6).w
		bne.w	loc_E704
		bclr	#1,$64(a3)

loc_E704:							; CODE XREF: sub_E6CE+2Cj
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		bne.w	loc_E73A
		btst	#1,$64(a3)
		bne.w	loc_E73A
		btst	#2,($FFFFC2F2).w
		bne.w	loc_E73A
		jsr	sub_DE2A
		beq.w	loc_E73A
		move.w	#$22,d0					; '"'
		bsr.w	sub_10662
		bra.w	*+4

loc_E73A:							; CODE XREF: sub_E6CE+3Ej sub_E6CE+48j ...
		bclr	#1,$62(a3)
		beq.w	loc_E79E
		btst	#1,$64(a3)
		beq.w	loc_E758
		move.w	#1,-(sp)
		jsr	sub_11132

loc_E758:							; CODE XREF: sub_E6CE+7Cj
		btst	#1,$64(a3)
		beq.w	loc_E790
		movem.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_E77C
		move.l	#$FFFFCA32,a2

loc_E77C:							; CODE XREF: sub_E6CE+A4j
		addq.w	#1,$358(a2)
		add.w	#$14,($FFFFB8A2).w
		add.w	#$C8,($FFFFB89C).w			; ''
		movem.l	(sp)+,a2

loc_E790:							; CODE XREF: sub_E6CE+90j
		clr.w	$44(a3)
		move.w	#8,$42(a3)
		clr.w	$40(a3)

loc_E79E:							; CODE XREF: sub_E6CE+72j
		move.w	$52(a3),d1
		cmp.w	($FFFFB7AA).w,d1
		bne.w	loc_E7C4
		tst.w	$34(a3)
		beq.w	sub_D51C
		move.l	#$10,d0
		btst	#3,$62(a3)
		beq.w	sub_10658
		rts
; ---------------------------------------------------------------------------

loc_E7C4:							; CODE XREF: sub_E6CE+D8j
		sub.b	d7,$40(a3)
		bpl.w	loc_E95E
		move.b	$6A(a3),$40(a3)
		jsr	sub_FEFCC
		bmi.w	loc_E7EE
		btst	#6,($FFFFC2FC).w
		beq.w	loc_E7F2
		tst.b	$40(a3)
		beq.w	loc_E7F2

loc_E7EE:							; CODE XREF: sub_E6CE+10Aj
		subq.b	#1,$40(a3)

loc_E7F2:							; CODE XREF: sub_E6CE+114j
					; sub_E6CE+11Cj
		btst	#5,($FFFFC2F8).w
		beq.w	loc_E834
		move.w	($FFFFB7AA).w,d1
		cmpnw	 5,d1
		bgt.w	loc_E82C
		btst	#6,$62(a3)
		beq.w	loc_E834

loc_E812:							; CODE XREF: sub_E6CE+164j
		move.b	$40(a3),d1
		ext.w	d1
		asr.w	#1,d1
		move.b	d1,$40(a3)
		bne.w	loc_E834
		move.b	#1,$40(a3)
		bra.w	loc_E834
; ---------------------------------------------------------------------------

loc_E82C:							; CODE XREF: sub_E6CE+136j
		btst	#6,$62(a3)
		beq.s	loc_E812

loc_E834:							; CODE XREF: sub_E6CE+12Aj
					; sub_E6CE+140j ...
		clr.l	$2A(a2)
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_E85A
		asl.w	#7,d1
		move.w	#$B04A,a1
		add.w	d1,a1
		move.b	$62(a1),d0
		move.b	$62(a3),d1
		eor.b	d0,d1
		btst	#6,d1
		beq.w	loc_E8D0

loc_E85A:							; CODE XREF: sub_E6CE+16Ej
		moveq	#$FFFFFFFF,d2
		moveq	#5,d4
		move.w	$22(a2),a0

loc_E862:							; CODE XREF: sub_E6CE+1F4j
		tst.w	$34(a0)
		ble.w	loc_E8BE
		tst.b	$5E(a0)
		bne.w	loc_E8BE
		btst	#2,$63(a0)
		bne.w	loc_E8BE
		move.w	$36(a0),d0
		cmp.b	#$13,$38(a0,d0.w)
		beq.w	loc_E8BE
		move.l	a0,-(sp)
		bsr.w	sub_106E0
		add.w	(a0),d0
		sub.w	($FFFFB74A).w,d0
		add.w	$14(a0),d1
		sub.w	($FFFFB75E).w,d1
		move.w	($FFFFB772).w,d3
		asr.w	#6,d3
		sub.w	d3,d0
		move.w	($FFFFB774).w,d3
		asr.w	#6,d3
		sub.w	d3,d1
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmp.l	d2,d0
		bhi.w	loc_E8BE
		move.l	d0,d2
		move.w	a0,a1

loc_E8BE:							; CODE XREF: sub_E6CE+198j
					; sub_E6CE+1A0j ...
		add.w	#$80,a0					; ''
		dbf	d4,loc_E862
		tst.l	d2
		bmi.w	loc_E90C
		move.l	d2,$2A(a2)

loc_E8D0:							; CODE XREF: sub_E6CE+188j
		cmp.w	a1,a3
		beq.w	loc_E90C
		btst	#0,($FFFFC2EA).w
		bne.w	loc_E90C
		btst	#2,$63(a1)
		bne.w	loc_E90C
		btst	#3,$64(a1)
		bne.w	loc_E90C
		exg	a1,a3
		bclr	#0,$62(a3)
		move.l	#$11,d0
		bsr.w	sub_10658
		exg	a1,a3
		bra.w	sub_10646
; ---------------------------------------------------------------------------

loc_E90C:							; CODE XREF: sub_E6CE+1FAj
					; sub_E6CE+204j ...
		btst	#5,$62(a3)
		bne.w	locret_15464
		moveq	#2,d1
		cmpnw	 $190,d2
		bhi.w	loc_E92C
		subq.w	#2,d1
		cmp.w	#2,$34(a3)
		bls.w	loc_E95E

loc_E92C:							; CODE XREF: sub_E6CE+24Ej
		btst	#5,($FFFFC2EE).w
		beq.w	loc_E942
		subq.w	#2,d1
		bsr.w	sub_E274
		bne.w	loc_E942
		addq.w	#4,d1

loc_E942:							; CODE XREF: sub_E6CE+264j
					; sub_E6CE+26Ej
		move.w	#$28,d0					; '('
		sub.b	$75(a3),d0
		asl.w	d1,d0
		bsr.w	sub_11086
		cmpnw	 2,d0
		bhi.w	loc_E95E
		move.w	#$F0,$44(a3)				; ''

loc_E95E:							; CODE XREF: sub_E6CE+FAj
					; sub_E6CE+25Aj ...
		btst	#5,$62(a3)
		bne.w	locret_15464
		btst	#2,($FFFFC2F2).w
		beq.w	loc_E97C
		btst	#5,($FFFFC2F2).w
		beq.w	loc_E986

loc_E97C:							; CODE XREF: sub_E6CE+2A0j
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4

loc_E986:							; CODE XREF: sub_E6CE+2AAj
		btst	#3,$62(a3)
		bne.w	locret_15464
		btst	#2,$30(a2)
		bne.w	loc_E9AE
		tst.w	($FFFFB7AA).w
		bmi.w	loc_EACA
		sub.w	d7,$44(a3)
		bpl.w	loc_EACA
		clr.w	$44(a3)

loc_E9AE:							; CODE XREF: sub_E6CE+2C8j
		bsr.w	sub_10588
		movem.w	d0-d1,-(sp)
		neg.w	d0
		neg.w	d1
		add.w	#$F4,d1					; ''
		btst	#7,$62(a3)
		beq.w	loc_E9CC
		sub.w	#$1E8,d1

loc_E9CC:							; CODE XREF: sub_E6CE+2F6j
		asr.w	#1,d0
		asr.w	#1,d1
		add.w	(sp)+,d0
		add.w	(sp)+,d1
		btst	#2,$30(a2)
		beq.w	loc_EA04
		btst	#7,$62(a3)
		beq.w	loc_E9F8
		cmpnw	 $53,d1					; 'S'
		blt.w	loc_EA04
		move.w	#$44,d1					; 'D'
		bra.w	loc_EA04
; ---------------------------------------------------------------------------

loc_E9F8:							; CODE XREF: sub_E6CE+316j
		cmpnw	 $FFAD,d1
		bgt.w	loc_EA04
		move.w	#$FFBC,d1

loc_EA04:							; CODE XREF: sub_E6CE+30Cj
					; sub_E6CE+31Ej ...
		btst	#5,($FFFFC2F8).w
		beq.w	loc_EA52
		move.b	$69(a3),($FFFFBF1E).w
		addq.b	#6,$69(a3)
		btst	#1,($FFFFC2FE).w
		bne.w	loc_EA2C
		btst	#6,($FFFFC2FC).w
		beq.w	loc_EA30

loc_EA2C:							; CODE XREF: sub_E6CE+350j
		addq.b	#2,$69(a3)

loc_EA30:							; CODE XREF: sub_E6CE+35Aj
		cmp.b	#$1E,$69(a3)
		ble.w	loc_EA40
		move.b	#$1E,$69(a3)

loc_EA40:							; CODE XREF: sub_E6CE+368j
		lea	locret_15464(pc),a0
		bsr.w	sub_103C8
		move.b	($FFFFBF1E).w,$69(a3)
		bra.w	loc_EA5A
; ---------------------------------------------------------------------------

loc_EA52:							; CODE XREF: sub_E6CE+33Cj
		lea	locret_15464(pc),a0
		bsr.w	sub_103C8

loc_EA5A:							; CODE XREF: sub_E6CE+380j
		btst	#5,($FFFFC2F8).w
		beq.w	loc_EA96
		move.w	($FFFFB75E).w,d0
		move.w	$14(a3),d2
		tst.w	d0
		bpl.w	loc_EA76
		neg.w	d0
		neg.w	d2

loc_EA76:							; CODE XREF: sub_E6CE+3A0j
		sub.w	d0,d2
		cmpnw	 $A,d2
		blt.w	locret_EA94
		move.w	(a3),d0
		sub.w	($FFFFB74A).w,d0
		bpl.w	loc_EA8C
		neg.w	d0

loc_EA8C:							; CODE XREF: sub_E6CE+3B8j
		cmpnw	 $F,d0
		blt.w	loc_EA96

locret_EA94:							; CODE XREF: sub_E6CE+3AEj
		rts
; ---------------------------------------------------------------------------

loc_EA96:							; CODE XREF: sub_E6CE+392j
					; sub_E6CE+3C2j
		btst	#5,($FFFFC2F8).w
		beq.w	sub_EACE
		move.b	$75(a3),d0
		ext.w	d0
		movem.l	d0/a3,-(sp)
		add.b	d0,d0
		cmpnb	 $1E,d0
		blt.w	loc_EAB8
		move.b	#$1E,d0

loc_EAB8:							; CODE XREF: sub_E6CE+3E2j
		move.b	d0,$75(a3)
		bsr.w	sub_EACE
		movem.l	(sp)+,d0/a3
		move.b	d0,$75(a3)
		rts
; ---------------------------------------------------------------------------

loc_EACA:							; CODE XREF: sub_E6CE+2D0j
					; sub_E6CE+2D8j
		bsr.w	sub_105A2
; End of function sub_E6CE


; =============== S U B	R O U T	I N E =======================================


sub_EACE:							; CODE XREF: ROM:0000CF94j
					; sub_E6CE+3CEj ...
		tst.w	$34(a3)
		bne.w	loc_EAD8
		rts
; ---------------------------------------------------------------------------

loc_EAD8:							; CODE XREF: sub_EACE+4j
		move.w	#$28,d0					; '('
		sub.b	$75(a3),d0
		tst.w	($FFFFD056).w
		beq.w	loc_EAEA
		asl.w	#1,d0

loc_EAEA:							; CODE XREF: sub_EACE+16j
		bsr.w	sub_11086
		cmpnw	 6,d0
		bhi.w	locret_15464
		moveq	#5,d2
		move.w	#$B04A,a0
		cmp.w	#6,$52(a3)
		bge.w	loc_EB0A
		add.w	#$300,a0

loc_EB0A:							; CODE XREF: sub_EACE+34j sub_EACE+B0j
		tst.w	$34(a0)
		beq.w	loc_EB7A
		btst	#5,$62(a0)
		bne.w	loc_EB7A
		btst	#0,$63(a0)
		bne.w	loc_EB7A
		move.w	(a0),d0
		sub.w	(a3),d0
		cmpnw	 $1E,d0
		bgt.w	loc_EB7A
		cmpnw	 $FFE2,d0
		blt.w	loc_EB7A
		move.w	$14(a0),d1
		sub.w	$14(a3),d1
		cmpnw	 $1E,d1
		bgt.w	loc_EB7A
		cmpnw	 $FFE2,d1
		blt.w	loc_EB7A
		bsr.w	sub_10676
		cmp.w	$54(a3),d0
		bne.w	loc_EB7A
		btst	#5,($FFFFC2F8).w
		bne.w	loc_BBD4
		move.w	($C00008).l,d0
		and.w	#3,d0
		bne.w	loc_BB62
		bra.w	loc_BBD4
; ---------------------------------------------------------------------------

loc_EB7A:							; CODE XREF: sub_EACE+40j sub_EACE+4Aj ...
		add.w	#$80,a0					; ''
		dbf	d2,loc_EB0A
		rts
; End of function sub_EACE


; =============== S U B	R O U T	I N E =======================================


sub_EB84:							; DATA XREF: ROM:00018DC8o

; FUNCTION CHUNK AT 0000CBE4 SIZE 00000016 BYTES

		btst	#5,$62(a3)
		bne.w	locret_15464
		btst	#0,($FFFFC2EA).w
		bne.w	loc_CBE4
		btst	#3,$62(a3)
		bne.w	sub_10646
		bclr	#1,$62(a3)
		beq.w	loc_EBC2
		bset	#0,$62(a3)
		move.b	#8,$43(a3)
		clr.b	$42(a3)
		move.w	#$FFFE,$46(a3)

loc_EBC2:							; CODE XREF: sub_EB84+24j
		tst.w	($FFFFB7AA).w
		bpl.w	loc_EC68
		addq.w	#1,$46(a3)
		beq.w	loc_EBD6
		bpl.w	loc_EC60

loc_EBD6:							; CODE XREF: sub_EB84+4Aj
		tst.w	($FFFFBF6C).w
		bpl.w	loc_EC60
		tst.w	$34(a3)
		beq.w	loc_EC60
		move.w	d0,-(sp)
		clr.w	$46(a3)
		move.w	#1,d0
		btst	#6,$62(a3)
		beq.w	loc_EBFE
		move.w	#2,d0

loc_EBFE:							; CODE XREF: sub_EB84+72j
		cmp.w	($FFFFC328).w,d0
		beq.w	loc_EC5E
		cmp.w	($FFFFC32A).w,d0
		beq.w	loc_EC5E
		jsr	sub_F6C44
		beq.w	loc_EC5E
		move.w	$14(a3),d0
		btst	#7,$62(a3)
		bne.w	loc_EC28
		neg.w	d0

loc_EC28:							; CODE XREF: sub_EB84+9Ej
		cmpnw	 $58,d0					; 'X'
		blt.w	loc_EC3C
		cmpnw	 $108,d0
		bgt.w	loc_EC3C
		bra.w	loc_EC4C
; ---------------------------------------------------------------------------

loc_EC3C:							; CODE XREF: sub_EB84+A8j sub_EB84+B0j
		move.w	#8,d0
		jsr	sub_11086
		tst.w	d0
		bne.w	loc_EC5E

loc_EC4C:							; CODE XREF: sub_EB84+B4j
		move.w	(sp)+,d0
		move.w	#$FFFF,($FFFFDED8).w
		move.w	#$23,d0					; '#'
		jmp	sub_10662
; ---------------------------------------------------------------------------

loc_EC5E:							; CODE XREF: sub_EB84+7Ej sub_EB84+86j ...
		move.w	(sp)+,d0

loc_EC60:							; CODE XREF: sub_EB84+4Ej sub_EB84+56j ...
		sub.b	d7,$40(a3)
		bpl.w	locret_15464

loc_EC68:							; CODE XREF: sub_EB84+42j
		bclr	#0,$62(a3)
		bra.w	sub_10646
; End of function sub_EB84

; ---------------------------------------------------------------------------
		move.w	$44(a3),d0
		move.w	$46(a3),d1
		lea	locret_15464(pc),a0
		bra.w	sub_103C8
; ---------------------------------------------------------------------------

loc_EC82:							; DATA XREF: ROM:00018DC4o
		btst	#5,$62(a3)
		bne.w	locret_15464
		bclr	#1,$62(a3)
		beq.w	loc_EC9A
		bra.w	loc_C1B8
; ---------------------------------------------------------------------------

loc_EC9A:							; CODE XREF: ROM:0000EC92j
		btst	#3,($FFFFC2EC).w
		beq.w	sub_10646
		clr.w	d2
		sub.w	d7,$42(a3)
		bpl.w	loc_C224
		bset	#5,d2
		bra.w	loc_C224
; ---------------------------------------------------------------------------

loc_ECB6:							; DATA XREF: ROM:00018DF4o
		bclr	#1,$62(a3)
		beq.w	loc_EDB0
		bset	#7,($FFFFC2FC).w
		clr.l	($FFFFBE7A).w
		clr.l	($FFFFBE7E).w
		clr.l	($FFFFBE82).w
		bclr	#7,($FFFFC2FA).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_ECEC
		jsr	sub_FC4C0
		bra.w	loc_ED0E
; ---------------------------------------------------------------------------

loc_ECEC:							; CODE XREF: ROM:0000ECDEj
		jsr	sub_FE756
		move.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD404).w
		beq.w	loc_ED08
		move.l	#$FFFFCA32,a2

loc_ED08:							; CODE XREF: ROM:0000ECFEj
		addq.w	#1,$360(a2)
		move.l	(sp)+,a2

loc_ED0E:							; CODE XREF: ROM:0000ECE8j
		bclr	#2,($FFFFC2F2).w
		bclr	#4,($FFFFC2F2).w
		bclr	#5,($FFFFC2F2).w
		bclr	#6,($FFFFC2F2).w
		bset	#1,($FFFFC2F8).w
		btst	#3,($FFFFC2EA).w
		bne.w	sub_124A6
		bclr	#1,$62(a3)
		btst	#0,($FFFFC2FA).w
		beq.w	loc_ED70
		tst.w	($FFFFD594).w
		bne.w	loc_ED8A
		tst.w	($FFFFD586).w
		bne.w	loc_ED8A
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#2,($FFFFD6CC).w
		jsr	sub_FE556
		move.w	($FFFFD6C8).w,-(sp)
		bra.w	loc_ED86
; ---------------------------------------------------------------------------

loc_ED70:							; CODE XREF: ROM:0000ED42j
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#2,($FFFFD6CC).w
		jsr	sub_FE556
		move.w	($FFFFD6C8).w,-(sp)

loc_ED86:							; CODE XREF: ROM:0000ED6Cj
		bsr.w	sub_11156

loc_ED8A:							; CODE XREF: ROM:0000ED4Aj
					; ROM:0000ED52j
		bset	#0,($FFFFC2EA).w
		bset	#1,($FFFFC2FA).w
		clr.w	($FFFFB772).w
		clr.w	($FFFFB774).w
		move.w	#$19,($FFFFD454).w
		bsr.w	loc_F712
		st	$40(a3)
		st	$42(a3)

loc_EDB0:							; CODE XREF: ROM:0000ECBCj
		bset	#2,($FFFFC2F2).w
		movem.w	d1-d2,-(sp)
		jsr	sub_EE58
		bmi.w	loc_EDCC
		movem.w	(sp)+,d1-d2
		bra.w	loc_EDE0
; ---------------------------------------------------------------------------

loc_EDCC:							; CODE XREF: ROM:0000EDC0j
					; ROM:0000EE0Ej
		movem.w	(sp)+,d1-d2
		bclr	#2,($FFFFC2F2).w
		move.w	#$1B,d0
		bsr.w	sub_10662
		rts
; ---------------------------------------------------------------------------

loc_EDE0:							; CODE XREF: ROM:0000EDC8j
		movem.w	d1-d2,-(sp)
		move.w	($FFFFD408).w,d0
		move.l	#$FFFFB04A,a2
		asl.w	#7,d0
		add.w	d0,a2
		tst.w	$34(a2)
		bne.w	loc_EE04
		btst	#2,$63(a2)
		beq.w	loc_EE48

loc_EE04:							; CODE XREF: ROM:0000EDF6j
		move.w	($FFFFD408).w,d0
		move.w	#6,d2

loc_EE0C:							; CODE XREF: ROM:0000EE3Aj
					; ROM:0000EE42j
		subq.w	#1,d2
		bmi.s	loc_EDCC
		subq.w	#1,d0
		bpl.w	loc_EE1E
		move.w	#5,d0
		bra.w	loc_EE2A
; ---------------------------------------------------------------------------

loc_EE1E:							; CODE XREF: ROM:0000EE12j
		cmpnw	 5,d0
		bne.w	loc_EE2A
		move.w	#$B,d0

loc_EE2A:							; CODE XREF: ROM:0000EE1Aj
					; ROM:0000EE22j
		move.l	#$FFFFB04A,a2
		move.w	d0,d1
		asl.w	#7,d1
		add.w	d1,a2
		tst.w	$34(a2)
		bne.s	loc_EE0C
		btst	#2,$63(a2)
		bne.s	loc_EE0C
		move.w	d0,($FFFFD408).w

loc_EE48:							; CODE XREF: ROM:0000EE00j
		movem.w	(sp)+,d1-d2
		move.w	#$C6CE,a2
		move.w	#$1F,d0
		bra.w	sub_10662

; =============== S U B	R O U T	I N E =======================================


sub_EE58:							; CODE XREF: ROM:0000EDBAp
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFC6CE,a0
		tst.w	($FFFFD404).w
		beq.w	loc_EE70
		move.l	#$FFFFCA32,a0

loc_EE70:							; CODE XREF: sub_EE58+Ej
		move.w	#0,d1
		move.w	#$FFFF,d6
		move.w	#$FFFF,d5
		move.l	$1E(a0),a2
		add.w	(a2),a2

loc_EE82:							; CODE XREF: sub_EE58+F0j
		cmp.w	#2,(a2)
		beq.w	loc_EF4C
		add.w	(a2),a2
		move.w	d1,d7
		asl.w	#1,d7
		move.b	5(a2),d0
		and.w	#$F,d0
		cmpnb	 1,d0
		ble.w	loc_EF40
		btst	#0,($FFFFC2FA).w
		bne.w	loc_EF56
		cmp.w	#$FFFE,$66(a0,d7.w)
		beq.w	loc_EEBE
		cmp.w	#$FFFF,$66(a0,d7.w)
		bne.w	loc_EF40

loc_EEBE:							; CODE XREF: sub_EE58+58j
		clr.w	d3
		clr.w	d0
		move.b	1(a2),d0
		and.w	#$F,d0
		add.w	d0,d3
		move.b	2(a2),d0
		and.w	#$F,d0
		add.w	d0,d3
		move.b	2(a2),d0
		and.w	#$F0,d0					; ''
		lsr.w	#4,d0
		add.w	d0,d3
		move.b	3(a2),d0
		and.w	#$F,d0
		add.w	d0,d3
		move.b	3(a2),d0
		and.w	#$F0,d0					; ''
		lsr.w	#4,d0
		add.w	d0,d3
		move.b	5(a2),d0
		and.w	#$F,d0
		add.w	d0,d3
		move.b	5(a2),d0
		and.w	#$F0,d0					; ''
		lsr.w	#4,d0
		add.w	d0,d3
		move.b	6(a2),d0
		and.w	#$F0,d0					; ''
		lsr.w	#4,d0
		add.w	d0,d3
		move.b	7(a2),d0
		and.w	#$F0,d0					; ''
		lsr.w	#4,d0
		add.w	d0,d3
		cmp.w	($FFFFD40A).w,d1
		bne.w	loc_EF32
		move.w	#$7FFF,d3

loc_EF32:							; CODE XREF: sub_EE58+D2j
		cmp.w	d6,d3
		blt.w	loc_EF40
		move.w	d3,d6
		move.w	d1,d5
		bra.w	*+4

loc_EF40:							; CODE XREF: sub_EE58+44j sub_EE58+62j ...
		addq.l	#8,a2
		addq.w	#1,d1
		cmpnw	 $1A,d1
		blt.w	loc_EE82

loc_EF4C:							; CODE XREF: sub_EE58+2Ej
		tst.w	d5
		bmi.w	loc_EF80
		move.w	d5,($FFFFD40A).w

loc_EF56:							; CODE XREF: sub_EE58+4Ej
		move.w	#0,($FFFFD406).w
		tst.w	($FFFFD404).w
		beq.w	loc_EF6A
		move.w	#6,($FFFFD406).w

loc_EF6A:							; CODE XREF: sub_EE58+108j
		move.w	($FFFFD406).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a3
		add.w	d0,a3
		move.w	($FFFFD40A).w,d3
		bra.w	loc_EF88
; ---------------------------------------------------------------------------

loc_EF80:							; CODE XREF: sub_EE58+F6j
		move.w	#$FFFF,d0
		bra.w	loc_EF8C
; ---------------------------------------------------------------------------

loc_EF88:							; CODE XREF: sub_EE58+124j
		move.w	#1,d0

loc_EF8C:							; CODE XREF: sub_EE58+12Cj
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_EE58


; =============== S U B	R O U T	I N E =======================================


sub_EF92:							; DATA XREF: ROM:00018DF8o
		bclr	#1,$62(a3)
		beq.w	loc_F2A8
		movem.l	d0-a6,-(sp)
		bsr.w	sub_10F32

loc_EFA4:							; CODE XREF: sub_EF92+18j
		btst	#0,($FFFFBF78).w
		bne.s	loc_EFA4
		cmp.w	#$258,($FFFFB89C).w
		bls.w	loc_EFC2
		move.w	#$258,($FFFFB89C).w
		add.w	#$14,($FFFFB8A2).w

loc_EFC2:							; CODE XREF: sub_EF92+20j
		move.w	($FFFFB026).w,d4
		move.l	#unk_5C410,a2
		jsr	sub_11738
		bset	#3,($FFFFBF78).w
		bclr	#0,($FFFFC2EC).w
		bclr	#0,($FFFFC2F0).w
		clr.w	($FFFFBEE2).w
		clr.b	($FFFFBF80).w
		st	($FFFFC3F0).w
		bclr	#1,($FFFFC2EE).w
		st	($FFFFBEE0).w
		bsr.w	sub_12B40
		clr.w	($FFFFBD18).w
		clr.w	($FFFFBD1C).w
		move.w	#0,($FFFFB74A).w
		move.w	#0,($FFFFB75E).w
		clr.w	($FFFFB762).w
		clr.w	($FFFFB772).w
		clr.w	($FFFFB774).w
		clr.w	($FFFFB776).w
		st	($FFFFB7AA).w
		move.w	#$B64A,a0
		clr.w	$28(a0)
		clr.w	$2A(a0)
		clr.w	(a0)
		move.w	#$10C,$14(a0)
		add.w	#$80,a0					; ''
		clr.w	$28(a0)
		clr.w	$2A(a0)
		clr.w	(a0)
		move.w	#$FEF4,$14(a0)
		move.w	#$B7CA,a0
		move.w	#$18A,6(a0)
		clr.w	$58(a0)
		clr.w	4(a0)
		clr.w	($FFFFB74E).w
		bclr	#6,($FFFFC2EC).w
		moveq	#$64,d4					; 'd'

loc_F06C:							; CODE XREF: sub_EF92+DEj
		bsr.w	sub_AFCA
		dbf	d4,loc_F06C
		move.w	#$3C,($FFFFBF92).w			; '<'
		bsr.w	sub_15A24
		move.w	#$C6CE,a2
		bsr.w	sub_15788
		bsr.w	sub_1592C
		add.w	#$364,a2
		bsr.w	sub_15788
		bsr.w	sub_1592C
		jsr	sub_170B0
		move.l	a3,-(sp)
		move.l	#$FFFFB5CA,a3
		move.w	#$B,d2

loc_F0A8:							; CODE XREF: sub_EF92+270j
		cmp.w	($FFFFD406).w,d2
		beq.w	loc_F0D0
		cmp.w	($FFFFD408).w,d2
		beq.w	loc_F150
		move.w	#$FF10,(a3)
		clr.w	$14(a3)
		clr.w	6(a3)
		move.w	#$20,d0					; ' '
		bsr.w	sub_10658
		bra.w	loc_F1FC
; ---------------------------------------------------------------------------

loc_F0D0:							; CODE XREF: sub_EF92+11Aj
		tst.w	$34(a3)
		bpl.w	loc_F104
		bclr	#2,$63(a3)
		beq.w	loc_F104
		movem.l	d0-a6,-(sp)
		clr.w	d3
		move.b	$66(a3),d3
		jsr	sub_15AA4
		movem.l	(sp)+,d0-a6
		tst.w	$34(a3)
		beq.w	loc_F102
		bpl.w	loc_F104

loc_F102:							; CODE XREF: sub_EF92+168j
		nop

loc_F104:							; CODE XREF: sub_EF92+142j
					; sub_EF92+14Cj ...
		move.w	($FFFFB74A).w,d0
		sub.w	#0,d0
		move.w	d0,(a3)
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_F12A
		add.w	#$10,d0
		move.w	#4,$54(a3)
		bra.w	loc_F134
; ---------------------------------------------------------------------------

loc_F12A:							; CODE XREF: sub_EF92+186j
		add.w	#-$10,d0
		move.w	#0,$54(a3)

loc_F134:							; CODE XREF: sub_EF92+194j
		move.w	d0,$14(a3)
		clr.w	$28(a3)
		clr.w	$2A(a3)
		clr.w	$2C(a3)
		move.w	#$11,d0
		bsr.w	sub_10658
		bra.w	loc_F1FC
; ---------------------------------------------------------------------------

loc_F150:							; CODE XREF: sub_EF92+122j
		btst	#0,($FFFFC2FA).w
		beq.w	loc_F172
		move.b	($FFFFD593).w,$66(a3)
		tst.w	($FFFFD594).w
		beq.w	loc_F184
		move.b	($FFFFD585).w,$66(a3)
		bra.w	loc_F184
; ---------------------------------------------------------------------------

loc_F172:							; CODE XREF: sub_EF92+1C4j
		tst.w	$34(a3)
		bpl.w	loc_F1A6
		bclr	#2,$63(a3)
		beq.w	loc_F1A6

loc_F184:							; CODE XREF: sub_EF92+1D2j
					; sub_EF92+1DCj
		movem.l	d0-a6,-(sp)
		clr.w	d3
		move.b	$66(a3),d3
		jsr	sub_15AA4
		movem.l	(sp)+,d0-a6
		tst.w	$34(a3)
		bne.w	loc_F1A4
		bpl.w	loc_F1A6

loc_F1A4:							; CODE XREF: sub_EF92+20Aj
		nop

loc_F1A6:							; CODE XREF: sub_EF92+1E4j
					; sub_EF92+1EEj ...
		bclr	#2,$63(a3)
		move.w	#0,d0
		move.w	#$E5,d1					; ''
		btst	#7,$62(a3)
		bne.w	loc_F1C2
		move.w	#$FF1B,d1

loc_F1C2:							; CODE XREF: sub_EF92+228j
		move.w	d0,(a3)
		move.w	d1,$14(a3)
		clr.w	$28(a3)
		clr.w	$2A(a3)
		clr.w	$18(a3)
		sub.w	($FFFFB74A).w,d0
		sub.w	($FFFFB75E).w,d1
		neg.w	d0
		neg.w	d1
		bsr.w	sub_10676
		move.w	d0,$54(a3)
		bclr	#2,$63(a3)
		bclr	#5,$62(a3)
		move.w	#$50C,d1
		bsr.w	sub_1073A

loc_F1FC:							; CODE XREF: sub_EF92+13Aj
					; sub_EF92+1BAj
		suba.l	#$80,a3					; ''
		dbf	d2,loc_F0A8
		bsr.w	sub_1702E
		move.l	(sp)+,a3
		bsr.w	sub_FD22
		move.w	($FFFFD408).w,d0
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		move.w	#0,(a3)
		move.w	#$FF06,$14(a3)
		btst	#7,$62(a3)
		bne.w	loc_F236
		move.w	#$FA,$14(a3)				; ''

loc_F236:							; CODE XREF: sub_EF92+29Aj
		move.w	($FFFFD408).w,d0
		move.w	#1,d1
		btst	#6,$62(a3)
		beq.w	loc_F24C
		move.w	#2,d1

loc_F24C:							; CODE XREF: sub_EF92+2B2j
		cmp.w	($FFFFC328).w,d1
		bne.w	loc_F25E
		jsr	sub_C0BC
		bra.w	loc_F26C
; ---------------------------------------------------------------------------

loc_F25E:							; CODE XREF: sub_EF92+2BEj
		cmp.w	($FFFFC32A).w,d1
		bne.w	loc_F26C
		jsr	sub_C0DA

loc_F26C:							; CODE XREF: sub_EF92+2C8j
					; sub_EF92+2D0j
		move.w	#$E,d0
		bsr.w	sub_10662
		bset	#7,($FFFFC2F2).w
		bset	#2,($FFFFC2EE).w
		move.w	#$190,($FFFFC31A).w
		tst.w	($FFFFC328).w
		bne.w	loc_F29C
		tst.w	($FFFFC32A).w
		bne.w	loc_F29C
		move.w	#$64,($FFFFC31A).w			; 'd'

loc_F29C:							; CODE XREF: sub_EF92+2F8j
					; sub_EF92+300j
		move.w	#$18,($FFFFBD26).w
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_F2A8:							; CODE XREF: sub_EF92+6j
		bset	#2,($FFFFC2FA).w
		move.w	#$18,d0
		bsr.w	sub_10662
		clr.w	$28(a3)
		clr.w	$2A(a3)
		rts
; End of function sub_EF92

; ---------------------------------------------------------------------------
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	6
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	5
		dc.b	3
		dc.b	4
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.w	$FF06
		dc.w	$FFDD
		dc.w	$FFCE
		dc.w	$23
		dc.w	$FFCE
		dc.w	$FFCE
		dc.w	$FFF6
		dc.w	0
		dc.w	$FFF1
		dc.w	$32
		dc.w	$FFF6
		dc.w	0
		dc.w	$FFC4

; =============== S U B	R O U T	I N E =======================================


sub_F2F4:							; CODE XREF: sub_FF0C+Ap
		movem.w	d0,-(sp)
		move.w	($FFFFB7AA).w,d0
		cmp.w	($FFFFD408).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_F342
		btst	#5,($FFFFC2F2).w
		beq.w	loc_F348
		btst	#5,($FFFFC2FA).w
		bne.w	loc_F32A
		tst.w	($FFFFB7AA).w
		bpl.w	loc_F332
		bset	#5,($FFFFC2FA).w

loc_F32A:							; CODE XREF: sub_F2F4+24j
		tst.w	($FFFFB7AA).w
		bpl.w	loc_F342

loc_F332:							; CODE XREF: sub_F2F4+2Cj
		tst.w	($FFFFC31C).w
		bmi.w	loc_F342
		subq.w	#1,($FFFFC31C).w
		bra.w	loc_F348
; ---------------------------------------------------------------------------

loc_F342:							; CODE XREF: sub_F2F4+10j sub_F2F4+3Aj ...
		bset	#4,($FFFFC2F2).w

loc_F348:							; CODE XREF: sub_F2F4+1Aj sub_F2F4+4Aj
		tst.w	($FFFFD454).w
		bne.w	loc_F356
		bset	#4,($FFFFC2F2).w

loc_F356:							; CODE XREF: sub_F2F4+58j
		btst	#4,($FFFFC2F2).w
		bne.w	loc_F362
		rts
; ---------------------------------------------------------------------------

loc_F362:							; CODE XREF: sub_F2F4+68j
		btst	#6,($FFFFC2F2).w
		bne.w	locret_F3E2
		bset	#6,($FFFFC2F2).w
		move.w	#$A,d0
		jsr	sub_11F62
; End of function sub_F2F4


; =============== S U B	R O U T	I N E =======================================


sub_F37C:							; CODE XREF: sub_14620+1F4p
		jsr	sub_AFB6
		btst	#0,($FFFFC2FA).w
		bne.w	loc_F398
		move.w	#$A,($FFFFDEF0).w
		bset	#2,($FFFFC2EE).w

loc_F398:							; CODE XREF: sub_F37C+Cj
		bset	#0,($FFFFC2EA).w
		bclr	#2,($FFFFC2FA).w
		bclr	#2,($FFFFC2F2).w
		bclr	#3,($FFFFC2F2).w
		bclr	#5,($FFFFC2F2).w
		bclr	#4,($FFFFC2F2).w
		bclr	#6,($FFFFC2F2).w
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFB04A,a0
		move.w	($FFFFD406).w,d0
		asl.w	#7,d0
		move.b	($FFFFC31E).w,$61(a0,d0.w)
		movem.l	(sp)+,d0/a0
		jsr	sub_FC516

locret_F3E2:							; CODE XREF: sub_F2F4+74j
		rts
; End of function sub_F37C


; =============== S U B	R O U T	I N E =======================================


sub_F3E4:							; DATA XREF: ROM:00018DE8o

; FUNCTION CHUNK AT 0000F612 SIZE 0000003C BYTES
; FUNCTION CHUNK AT 00017236 SIZE 00000042 BYTES

		bclr	#4,($FFFFC2FE).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_F404
		clr.w	($FFFFB772).w
		clr.w	($FFFFB774).w
		move.w	#$1E,d0
		bra.w	sub_10662
; ---------------------------------------------------------------------------

loc_F404:							; CODE XREF: sub_F3E4+Cj
		bclr	#1,$62(a3)
		beq.w	loc_F612
		bclr	#0,($FFFFC2FE).w
		beq.w	loc_F420
		clr.w	($FFFFBF94).w
		clr.w	($FFFFBF96).w

loc_F420:							; CODE XREF: sub_F3E4+30j
		bclr	#7,($FFFFC2FC).w
		clr.l	($FFFFBE7A).w
		clr.l	($FFFFBE7E).w
		clr.l	($FFFFBE82).w
		bclr	#6,($FFFFC2FC).w
		bclr	#2,($FFFFC2FC).w
		jsr	sub_FECF8
		tst.w	($FFFFD6BE).w
		bmi.w	loc_F454
		move.w	($FFFFD6BE).w,d0
		bra.w	loc_F460
; ---------------------------------------------------------------------------

loc_F454:							; CODE XREF: sub_F3E4+64j
		jsr	sub_FECAA
		tst.w	d0
		bmi.w	loc_F466

loc_F460:							; CODE XREF: sub_F3E4+6Cj
		jsr	sub_FE53C

loc_F466:							; CODE XREF: sub_F3E4+78j
		bclr	#1,($FFFFC2F8).w
		bclr	#1,($FFFFC2FA).w
		tst.w	($FFFFC468).w
		beq.w	loc_17236
		btst	#6,($FFFFC2EA).w
		bne.w	loc_17236
		cmp.w	#3,($FFFFC466).w
		bne.w	loc_F49A
		move.w	($FFFFC6DA).w,d0
		cmp.w	($FFFFCA3E).w,d0
		bne.w	loc_7B5C

loc_F49A:							; CODE XREF: sub_F3E4+A6j
		btst	#3,($FFFFC2EA).w
		bne.w	sub_124A6
		move.w	($FFFFC46C).w,d0
		asr.w	#1,d0
		cmp.w	($FFFFC468).w,d0
		bls.w	loc_F4F2
		cmp.w	#$3C,($FFFFC468).w			; '<'
		blt.w	loc_F4F2
		bclr	#7,($FFFFC2F0).w
		beq.w	loc_F4F2
		btst	#0,($FFFFC2EA).w
		beq.w	loc_F4F2
		btst	#6,($FFFFC2FE).w
		bne.w	loc_F4F2
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#2,($FFFFD6CC).w
		jsr	sub_FE556
		bset	#4,($FFFFC2FE).w

loc_F4F2:							; CODE XREF: sub_F3E4+CAj sub_F3E4+D4j ...
		bsr.w	loc_F712
		st	$40(a3)
		st	$42(a3)
		tst.w	($FFFFD058).w
		bne.w	loc_F612
		bclr	#1,($FFFFC6FE).w
		bclr	#1,($FFFFCA62).w
		move.w	#$B04A,a0
		moveq	#$B,d0

loc_F518:							; CODE XREF: sub_F3E4+13Ej
		bclr	#3,$63(a0)
		add.w	#$80,a0					; ''
		dbf	d0,loc_F518
		move.w	($FFFFC328).w,d0
		or.w	($FFFFC32A).w,d0
		beq.w	loc_F566
		btst	#7,($FFFFC2EC).w
		beq.w	loc_F566
		bsr.w	sub_10F32
		bset	#6,($FFFFC2EC).w
		bsr.w	sub_12B40
		move.w	#$18,($FFFFBD26).w
		tst.w	($FFFFD058).w
		bne.w	loc_F566
		btst	#4,($FFFFC2FC).w
		beq.w	loc_F566
		clr.w	($FFFFBD26).w

loc_F566:							; CODE XREF: sub_F3E4+14Aj
					; sub_F3E4+154j ...
		move.w	($FFFFC320).w,d0
		bmi.w	loc_F572
		bsr.w	sub_F5C6

loc_F572:							; CODE XREF: sub_F3E4+186j
		move.w	($FFFFC322).w,d0
		bmi.w	loc_F57E
		bsr.w	sub_F5C6

loc_F57E:							; CODE XREF: sub_F3E4+192j
		move.w	#$C6CE,a1
		lea	$364(a1),a2
		moveq	#2,d0
		bsr.w	sub_F590
		bra.w	loc_F612
; End of function sub_F3E4


; =============== S U B	R O U T	I N E =======================================


sub_F590:							; CODE XREF: sub_F3E4+1A4p
					; sub_F3E4+258p
		tst.w	($FFFFD058).w
		bne.w	locret_15464
		btst	#4,($FFFFC2FC).w
		bne.w	loc_F5B2
		cmp.w	($FFFFC328).w,d0
		beq.w	locret_15464
		cmp.w	($FFFFC32A).w,d0
		beq.w	locret_15464

loc_F5B2:							; CODE XREF: sub_F590+Ej
		bsr.w	sub_F790
		bsr.w	sub_15788
		bsr.w	loc_12C04
		move.w	#$2710,($FFFFC304).w
		rts
; End of function sub_F590


; =============== S U B	R O U T	I N E =======================================


sub_F5C6:							; CODE XREF: sub_F3E4+18Ap
					; sub_F3E4+196p
		exg	a2,a3
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		bclr	#3,$63(a3)
		move.l	a2,-(sp)
		bsr.w	sub_B8F2
		move.l	(sp)+,a2
		btst	#3,$63(a3)
		beq.w	loc_F60E
		btst	#6,$62(a3)
		beq.w	loc_F602
		move.w	#$168,$42(a2)
		move.w	$52(a3),$46(a2)
		bra.w	loc_F60E
; ---------------------------------------------------------------------------

loc_F602:							; CODE XREF: sub_F5C6+28j
		move.w	#$258,$40(a2)
		move.w	$52(a3),$44(a2)

loc_F60E:							; CODE XREF: sub_F5C6+1Ej sub_F5C6+38j
		exg	a2,a3
		rts
; End of function sub_F5C6

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_F3E4

loc_F612:							; CODE XREF: sub_F3E4+26j
					; sub_F3E4+11Ej ...
		move.w	#$40,d0					; '@'
		bsr.w	sub_F64E
		move.w	#$42,d0					; 'B'
		bsr.w	sub_F64E
		tst.w	$40(a3)
		bpl.w	locret_15464
		tst.w	$42(a3)
		bpl.w	locret_15464
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		moveq	#1,d0
		bsr.w	sub_F590
		move.w	#$FFFF,($FFFFD6C6).w
		move.w	#$1C,d0
		bra.w	sub_10662
; END OF FUNCTION CHUNK	FOR sub_F3E4

; =============== S U B	R O U T	I N E =======================================


sub_F64E:							; CODE XREF: sub_F3E4+232p
					; sub_F3E4+23Ap
		tst.w	(a3,d0.w)
		bmi.w	locret_15464
		move.w	4(a3,d0.w),d1
		asl.w	#7,d1
		move.w	#$B04A,a0
		move.w	#$C6CE,a2
		btst	#6,$62(a0,d1.w)
		beq.w	loc_F672
		add.w	#$364,a2

loc_F672:							; CODE XREF: sub_F64E+1Cj
		btst	#3,$63(a0,d1.w)
		bne.w	loc_F684
		st	(a3,d0.w)
		bra.w	sub_B92E
; ---------------------------------------------------------------------------

loc_F684:							; CODE XREF: sub_F64E+2Aj
		sub.w	d7,(a3,d0.w)
		bpl.w	locret_15464
		move.l	a3,-(sp)
		lea	(a0,d1.w),a3
		btst	#3,$63(a3)
		beq.w	loc_F6A6
		clr.w	d2
		bsr.w	sub_BB06
		bsr.w	sub_B92E

loc_F6A6:							; CODE XREF: sub_F64E+4Aj
		move.l	(sp)+,a3
		rts
; End of function sub_F64E


; =============== S U B	R O U T	I N E =======================================


sub_F6AA:							; CODE XREF: sub_799E+32p
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		moveq	#1,d0
		bsr.w	sub_F6C6
		moveq	#2,d0
		exg	a1,a2
; End of function sub_F6AA


; =============== S U B	R O U T	I N E =======================================


sub_F6C6:							; CODE XREF: sub_F6AA+14p
		tst.w	$26(a2)
		bmi.w	locret_15464
		move.w	($FFFFB7AA).w,d1
		bmi.w	locret_15464
		subq.w	#6,d1
		cmp.w	#$C6CE,a2
		beq.w	loc_F6E2
		not.w	d1

loc_F6E2:							; CODE XREF: sub_F6C6+16j
		tst.w	d1
		bpl.w	locret_15464
		btst	#3,($FFFFC2EA).w
		beq.w	loc_F6FA
		st	$26(a2)
		bra.w	sub_15788
; ---------------------------------------------------------------------------

loc_F6FA:							; CODE XREF: sub_F6C6+28j
		cmp.w	($FFFFC328).w,d0
		beq.w	locret_15464
		cmp.w	($FFFFC32A).w,d0
		beq.w	locret_15464
		move.w	($FFFFB75E).w,d1
		bra.w	loc_F746
; ---------------------------------------------------------------------------

loc_F712:							; CODE XREF: ROM:0000EDA4p
					; sub_F3E4:loc_F4F2p
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		moveq	#1,d0
		bsr.w	loc_F724
		moveq	#2,d0
		exg	a1,a2

loc_F724:							; CODE XREF: sub_F6C6+56p
		cmp.w	#$FFFF,$26(a2)
		beq.w	locret_15464
		clr.b	$26(a2)
		cmp.w	($FFFFC328).w,d0
		beq.w	locret_15464
		cmp.w	($FFFFC32A).w,d0
		beq.w	locret_15464
		move.w	($FFFFBF96).w,d1

loc_F746:							; CODE XREF: sub_F6C6+48j
		cmp.w	#2,($FFFFC466).w
		bne.w	locret_15464
		move.w	$C(a1),d0
		sub.w	$C(a2),d0
		bmi.w	locret_15464
		cmpnw	 2,d0
		bne.w	locret_15464
		cmp.w	#$3C,($FFFFC468).w			; '<'
		bgt.w	locret_15464
		move.l	a0,-(sp)
		move.w	$22(a2),a0
		btst	#7,$62(a0)
		move.l	(sp)+,a0
		bne.w	loc_F782
		neg.w	d1

loc_F782:							; CODE XREF: sub_F6C6+B6j
		tst.w	d1
		bmi.w	locret_15464
		st	$26(a2)
		bra.w	sub_15788
; End of function sub_F6C6


; =============== S U B	R O U T	I N E =======================================


sub_F790:							; CODE XREF: sub_E1F4+60p
					; sub_F590:loc_F5B2p
		movem.l	d0-d2/a0,-(sp)
		moveq	#3,d0
		move.w	$24(a2),d1
		sub.w	$24(a1),d1
		beq.w	loc_F7CA
		bpl.w	loc_F7A8
		addq.w	#2,d0

loc_F7A8:							; CODE XREF: sub_F790+12j
		move.w	d0,-(sp)
		bsr.w	sub_12EB4
		move.w	d0,d1
		move.w	(sp),d0
		addq.w	#1,d0
		bsr.w	sub_12EB4
		cmp.w	d0,d1
		bge.w	loc_F7C0
		addq.w	#1,(sp)

loc_F7C0:							; CODE XREF: sub_F790+2Aj
		move.w	(sp)+,$16(a2)

loc_F7C4:							; CODE XREF: sub_F790+86j sub_F790+E8j
		movem.l	(sp)+,d0-d2/a0
		rts
; ---------------------------------------------------------------------------

loc_F7CA:							; CODE XREF: sub_F790+Ej
		cmp.w	#$C6CE,a2
		bne.w	loc_F842
		moveq	#2,d1
		lea	word_F818(pc),a0
		cmp.w	#2,($FFFFC466).w
		bne.w	loc_F7FA
		move.w	$C(a2),d2
		cmp.w	$C(a1),d2
		beq.w	loc_F7FA
		add.w	#$E,a0
		bgt.w	loc_F7FA
		add.w	#$E,a0

loc_F7FA:							; CODE XREF: sub_F790+4Ej sub_F790+5Aj ...
		move.w	$16(a1),d1
		asl.w	#1,d1
		move.w	(a0,d1.w),d0
		bsr.w	sub_12EB4
		cmpnw	 $C00,d0
		bls.w	loc_F842
		move.w	(a0,d1.w),$16(a2)
		bra.s	loc_F7C4
; ---------------------------------------------------------------------------
word_F818:	dc.w	0					; DATA XREF: sub_F790+44o
		dc.w	1
		dc.w	2
		dc.w	0
		dc.w	1
		dc.w	0
		dc.w	1
		dc.w	2
		dc.w	0
		dc.w	1
		dc.w	2
		dc.w	0
		dc.w	2
		dc.w	0
		dc.w	0
		dc.w	1
		dc.w	0
		dc.w	0
		dc.w	1
		dc.w	0
		dc.w	1
; ---------------------------------------------------------------------------

loc_F842:							; CODE XREF: sub_F790+3Ej sub_F790+7Cj
		moveq	#2,d1
		lea	word_F87C(pc),a0
		cmp.w	#2,($FFFFC466).w
		bne.w	loc_F866
		move.w	$C(a2),d2
		cmp.w	$C(a1),d2
		beq.w	loc_F866
		addq.w	#6,a0
		bgt.w	loc_F866
		addq.w	#6,a0

loc_F866:							; CODE XREF: sub_F790+BEj sub_F790+CAj ...
		move.w	(a0)+,d0
		bsr.w	sub_12EB4
		cmpnw	 $C00,d0
		dbhi	d1,loc_F866
		move.w	-(a0),$16(a2)
		bra.w	loc_F7C4
; End of function sub_F790

; ---------------------------------------------------------------------------
word_F87C:	dc.w	0					; DATA XREF: sub_F790+B4o
		dc.w	1
		dc.w	2
		dc.w	0
		dc.w	2
		dc.w	1
		dc.w	0
		dc.w	1
		dc.w	0
; ---------------------------------------------------------------------------

loc_F88E:							; DATA XREF: ROM:00018DECo
		bclr	#1,$62(a3)
		beq.w	loc_FCDC
		bclr	#2,($FFFFC2F2).w
		bclr	#5,($FFFFC2F2).w
		bclr	#0,($FFFFC2F4).w
		bclr	#1,($FFFFC2F4).w
		bclr	#0,($FFFFC300).w
		bclr	#1,($FFFFC300).w
		bclr	#2,($FFFFC300).w
		movem.l	d0-a6,-(sp)
		bsr.w	sub_10F32

loc_F8CA:							; CODE XREF: ROM:0000F8D0j
		btst	#0,($FFFFBF78).w
		bne.s	loc_F8CA
		move	sr,-(sp)
		move.w	#$3C,($FFFFD412).w			; '<'
		tst.w	($FFFFBF94).w
		bne.w	loc_F8F4
		tst.w	($FFFFBF96).w
		bne.w	loc_F8F4
		move.w	#$FFFF,($FFFFD6BE).w
		bra.w	loc_F8F8
; ---------------------------------------------------------------------------

loc_F8F4:							; CODE XREF: ROM:0000F8DEj
					; ROM:0000F8E6j
		move	#$2700,sr

loc_F8F8:							; CODE XREF: ROM:0000F8F0j
		cmp.w	#$258,($FFFFB89C).w
		bls.w	loc_F908
		move.w	#$258,($FFFFB89C).w

loc_F908:							; CODE XREF: ROM:0000F8FEj
		bset	#3,($FFFFBF78).w
		bclr	#0,($FFFFC2EC).w
		bclr	#0,($FFFFC2F0).w
		clr.w	($FFFFBEE2).w
		clr.b	($FFFFBF80).w
		st	($FFFFC3F0).w
		st	($FFFFBEE8).w
		st	($FFFFBEEC).w
		bclr	#1,($FFFFC2EE).w
		bset	#0,($FFFFC2EE).w
		bset	#2,($FFFFC2EE).w
		st	($FFFFBEE0).w
		st	($FFFFBF6C).w
		bset	#4,($FFFFBF78).w
		bsr.w	sub_12B40
		move.w	#$2710,($FFFFC304).w
		clr.w	($FFFFBD18).w
		clr.w	($FFFFBD1C).w
		move.w	($FFFFBF94).w,($FFFFB74A).w
		move.w	($FFFFBF96).w,($FFFFB75E).w
		st	($FFFFB762).w
		clr.w	($FFFFB772).w
		clr.w	($FFFFB774).w
		clr.w	($FFFFB776).w
		st	($FFFFB7AA).w
		move.w	#$B64A,a0
		clr.w	$28(a0)
		clr.w	$2A(a0)
		clr.w	(a0)
		move.w	#$10C,$14(a0)
		add.w	#$80,a0					; ''
		clr.w	$28(a0)
		clr.w	$2A(a0)
		clr.w	(a0)
		move.w	#$FEF4,$14(a0)
		move.w	#$B7CA,a0
		move.w	#$18A,6(a0)
		clr.w	$58(a0)
		clr.w	4(a0)
		clr.w	($FFFFB74E).w
		bclr	#6,($FFFFC2EC).w
		moveq	#$64,d4					; 'd'

loc_F9C6:							; CODE XREF: ROM:0000F9CAj
		bsr.w	sub_AFCA
		dbf	d4,loc_F9C6
		move.w	#$3C,($FFFFBF92).w			; '<'
		bsr.w	sub_15A24
		move.w	#$C6CE,a2
		bsr.w	sub_15788
		bsr.w	sub_159A8
		add.w	#$364,a2
		bsr.w	sub_15788
		bsr.w	sub_159A8
		bsr.w	sub_170B0
		move.l	a3,-(sp)
		move.w	#$B04A,a3
		moveq	#$B,d2

loc_F9FC:							; CODE XREF: ROM:0000FB0Aj
		move.w	#$FF10,(a3)
		clr.w	$14(a3)
		clr.w	6(a3)
		move.w	$34(a3),d1
		bmi.w	loc_FB06
		beq.w	loc_FA5C
		move.l	#$16,d0
		cmpnw	 4,d1
		bne.w	loc_FA58
		move.w	#$C6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_FA34
		add.w	#$364,a2

loc_FA34:							; CODE XREF: ROM:0000FA2Cj
		clr.w	$18(a2)
		move.b	$66(a3),$19(a2)
		bclr	#7,($FFFFC2FE).w
		st	$1A(a2)
		st	$1C(a2)
		bclr	#3,$30(a2)
		move.l	#$17,d0

loc_FA58:							; CODE XREF: ROM:0000FA1Ej
		bsr.w	sub_10658

loc_FA5C:							; CODE XREF: ROM:0000FA10j
		move.w	($FFFFC6F2).w,d4
		btst	#6,$62(a3)
		beq.w	loc_FA6E
		move.w	($FFFFCA56).w,d4

loc_FA6E:							; CODE XREF: ROM:0000FA66j
		neg.w	d4
		addq.w	#6,d4
		asl.w	#3,d4
		move.l	#$FCEE,a1
		add.w	d4,a1
		move.b	(a1,d1.w),d4
		asl.w	#2,d4
		move.l	#$FD06,a1
		move.w	(a1,d4.w),d0
		move.w	2(a1,d4.w),d1
		btst	#7,$62(a3)
		bne.w	loc_FA9E
		neg.w	d0
		neg.w	d1

loc_FA9E:							; CODE XREF: ROM:0000FA96j
		tst.w	$34(a3)
		beq.w	loc_FAD0
		cmpnw	 8,d4
		bgt.w	loc_FAC8
		move.w	($FFFFBF94).w,d3
		eor.w	d0,d3
		bpl.w	loc_FAC0
		move.w	($FFFFBF96).w,d3
		asr.w	#3,d3
		sub.w	d3,d1

loc_FAC0:							; CODE XREF: ROM:0000FAB4j
		move.w	($FFFFBF94).w,d3
		asr.w	#2,d3
		sub.w	d3,d0

loc_FAC8:							; CODE XREF: ROM:0000FAAAj
		add.w	($FFFFBF94).w,d0
		add.w	($FFFFBF96).w,d1

loc_FAD0:							; CODE XREF: ROM:0000FAA2j
		move.w	d0,(a3)
		move.w	d1,$14(a3)
		clr.w	$28(a3)
		clr.w	$2A(a3)
		sub.w	($FFFFB74A).w,d0
		sub.w	($FFFFB75E).w,d1
		neg.w	d0
		neg.w	d1
		bsr.w	sub_10676
		move.w	d0,$54(a3)
		bclr	#2,$63(a3)
		bclr	#5,$62(a3)
		move.w	#$50C,d1
		bsr.w	sub_1073A

loc_FB06:							; CODE XREF: ROM:0000FA0Cj
		add.w	#$80,a3					; ''
		dbf	d2,loc_F9FC
		bsr.w	sub_1702E
		move.l	(sp)+,a3
		bsr.w	sub_FD22
		move.w	($FFFFB026).w,d4
		move.l	#unk_55BFE,a2
		jsr	sub_11738
		move.w	d4,($FFFFB01C).w
		move.l	#unk_A78B6,a2
		bsr.w	sub_11738
		move.w	#$FFFF,($FFFFD6B4).w
		btst	#0,($FFFFC2FC).w
		beq.w	loc_FB54
		move.w	($FFFFD6BE).w,d0
		bmi.w	loc_FB54
		jsr	sub_FE510

loc_FB54:							; CODE XREF: ROM:0000FB42j
					; ROM:0000FB4Aj
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$36,d0					; '6'
		tst.w	($FFFFBF94).w
		bpl.w	loc_FB6C
		move.w	#$BE,d0					; ''

loc_FB6C:							; CODE XREF: ROM:0000FB64j
		move.w	d0,($FFFFB032).w
		sub.w	#$2E,d0					; '.'
		asr.w	#3,d0
		move.w	d0,($FFFFB028).w
		moveq	#$5C,d0					; '\'
		move.w	d0,($FFFFB034).w
		sub.w	#$44,d0					; 'D'
		asr.w	#3,d0
		move.w	d0,($FFFFB02A).w
		move.w	($FFFFB026).w,d4
		move.l	#unk_55BF6,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$C,d2
		moveq	#$A,d3
		moveq	#0,d5
		bset	#0,($FFFFC2F8).w
		bsr.w	sub_1169A
		bclr	#0,($FFFFC2F8).w
		tst.w	($FFFFD058).w
		bne.w	loc_FC08
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FA
		dc.b	$A
		dc.b	$FE
		dc.b	4
; ---------------------------------------------------------------------------
		moveq	#$C,d0
		moveq	#3,d1
		bsr.w	sub_119B8
		move.w	($FFFFC6E4).w,d0
		move.w	($FFFFCA48).w,d1
		btst	#1,($FFFFC2EA).w
		bne.w	loc_FBE4
		exg	d0,d1

loc_FBE4:							; CODE XREF: ROM:0000FBDEj
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FB
		dc.b	1
		dc.b	$FA
		dc.b	$FE
; ---------------------------------------------------------------------------
		move.l	#word_191A6,a1
		bsr.w	sub_13508
		addq.w	#4,($FFFFB028).w
		move.w	d1,d0
		move.l	#word_191A6,a1
		bsr.w	sub_13508

loc_FC08:							; CODE XREF: ROM:0000FBBAj
		move.w	#$78,d0					; 'x'
		bsr.w	sub_11086
		add.w	#$B4,d0					; ''
		cmp.w	#0,($FFFFD6B4).w
		beq.w	loc_FC46
		cmp.w	#1,($FFFFD6B4).w
		beq.w	loc_FC46
		cmp.w	#5,($FFFFD6B4).w
		beq.w	loc_FC46
		cmp.w	#4,($FFFFD6B4).w
		beq.w	loc_FC46
		cmp.w	#2,($FFFFD6B4).w
		bne.w	loc_FC52

loc_FC46:							; CODE XREF: ROM:0000FC1Aj
					; ROM:0000FC24j ...
		cmpnw	 $10E,d0
		bgt.w	loc_FC52
		move.w	#$10E,d0

loc_FC52:							; CODE XREF: ROM:0000FC42j
					; ROM:0000FC4Aj
		move.w	d0,$40(a3)
		move.w	#$18,($FFFFBD26).w
		move.l	#$FFFFBDA8,a0
		move.w	#1,(a0)
		move.w	#$8000,2(a0)
		move.w	#4,4(a0)
		move.w	#$A800,6(a0)
		move.w	#7,8(a0)
		move.w	#$8000,$A(a0)
		btst	#1,($FFFFC2EA).w
		bne.w	loc_FC9A
		eor.w	#$800,2(a0)
		eor.w	#$800,6(a0)

loc_FC9A:							; CODE XREF: ROM:0000FC8Aj
		move.w	#$FFFF,($FFFFBF98).w
		move.w	#$FFFF,($FFFFBF9A).w
		jsr	sub_FF7E2
		bclr	#6,($FFFFC2FE).w
		bne.w	loc_FCC0
		bclr	#4,($FFFFC2FE).w
		beq.w	loc_FCCE

loc_FCC0:							; CODE XREF: ROM:0000FCB2j
		bclr	#4,($FFFFC2FE).w
		move.w	($FFFFD6C8).w,-(sp)
		bsr.w	sub_11156

loc_FCCE:							; CODE XREF: ROM:0000FCBCj
		move.w	#$18,($FFFFBD26).w
		move	(sp)+,sr
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_FCDC:							; CODE XREF: ROM:0000F894j
		subq.w	#1,$40(a3)
		bpl.w	loc_FDBE
		jsr	sub_FE548
		bra.w	loc_FE18
; ---------------------------------------------------------------------------
		dc.b	0
		dc.b	1
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	5
		dc.b	6
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	5
		dc.b	3
		dc.b	4
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	3
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.w	$FF06
		dc.w	$FFDD
		dc.w	$FFCE
		dc.w	$23
		dc.w	$FFCE
		dc.w	$FFCE
		dc.w	$FFF6
		dc.w	0
		dc.w	$FFF1
		dc.w	$32
		dc.w	$FFF6
		dc.w	0
		dc.w	$FFC4

; =============== S U B	R O U T	I N E =======================================


sub_FD22:							; CODE XREF: sub_EF92+27Ap
					; ROM:0000FB14p
		move.w	#$FFFF,($FFFFC320).w
		move.w	#$FFFF,($FFFFC322).w
		tst.w	($FFFFC328).w
		beq.w	loc_FD3C
		clr.w	d4
		bsr.w	sub_BFBC

loc_FD3C:							; CODE XREF: sub_FD22+10j
		tst.w	($FFFFC32A).w
		beq.w	loc_FD4A
		moveq	#2,d4
		bsr.w	sub_BFBC

loc_FD4A:							; CODE XREF: sub_FD22+1Ej
		tst.w	($FFFFC32C).w
		beq.w	loc_FD82
		move.w	($FFFFC328).w,-(sp)
		move.w	($FFFFC32C).w,($FFFFC328).w
		move.w	($FFFFC320).w,-(sp)
		move.w	#$FFFF,($FFFFC320).w
		clr.w	d4
		jsr	sub_BFBC
		move.w	($FFFFC320).w,($FFFFC324).w
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	(sp)+,($FFFFC320).w
		move.w	(sp)+,($FFFFC328).w

loc_FD82:							; CODE XREF: sub_FD22+2Cj
		tst.w	($FFFFC32E).w
		beq.w	locret_FDBC
		move.w	($FFFFC32A).w,-(sp)
		move.w	($FFFFC32E).w,($FFFFC32A).w
		move.w	($FFFFC322).w,-(sp)
		move.w	#$FFFF,($FFFFC322).w
		move.w	#2,d4
		jsr	sub_BFBC
		move.w	($FFFFC322).w,($FFFFC326).w
		move.w	($FFFFC32A).w,($FFFFC32E).w
		move.w	(sp)+,($FFFFC322).w
		move.w	(sp)+,($FFFFC32A).w

locret_FDBC:							; CODE XREF: sub_FD22+64j
		rts
; End of function sub_FD22

; ---------------------------------------------------------------------------

loc_FDBE:							; CODE XREF: ROM:0000FCE0j
		move.w	$40(a3),d0
		beq.w	loc_FDDE
		addq.w	#6,d0
		lsr.w	#3,d0
		cmpnw	 2,d0
		bgt.w	locret_15464
		neg.w	d0
		add.w	#$A,d0
		move.w	d0,($FFFFBDB0).w
		rts
; ---------------------------------------------------------------------------

loc_FDDE:							; CODE XREF: ROM:0000FDC2j
		bclr	#4,($FFFFBF78).w
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFB032).w,d0
		sub.w	#$2E,d0					; '.'
		asr.w	#3,d0
		move.w	d0,($FFFFB028).w
		moveq	#$64,d0					; 'd'
		move.w	($FFFFB034).w,d0
		sub.w	#$44,d0					; 'D'
		asr.w	#3,d0
		move.w	d0,($FFFFB02A).w
		moveq	#$C,d0
		moveq	#$D,d1
		move.w	#$7FF,d2
		bra.w	sub_1197E
; ---------------------------------------------------------------------------

loc_FE18:							; CODE XREF: ROM:0000FCEAj
		move.w	#$2F,-(sp)				; '/'
		bsr.w	sub_11132
		bclr	#0,($FFFFC2EE).w
		move.w	#$3C,($FFFFD412).w			; '<'
		move.w	($FFFFB026).w,d4
		move.l	#unk_5C410,a2
		bsr.w	sub_11738
		bclr	#2,($FFFFC2EE).w
		bclr	#0,($FFFFC2EA).w
		bclr	#0,$63(a3)
		clr.w	($FFFFC304).w
		bset	#4,($FFFFC2F0).w
		move.w	($FFFFBF98).w,d3
		move.w	#$800,d4
		move.l	#$FF06,a0
		move.w	#$BDA8,a1
		moveq	#$10,d2
		move.w	(a1),d1
		sub.b	-1(a0,d1.w),d2
		move.w	4(a1),d1
		add.b	-1(a0,d1.w),d2
		moveq	#$21,d0					; '!'
		bsr.w	sub_11086
		cmp.b	d0,d2
		bls.w	loc_FE86
		addq.w	#4,a1

loc_FE86:							; CODE XREF: ROM:0000FE80j
		btst	#3,2(a1)
		beq.w	loc_FE96
		move.w	($FFFFBF9A).w,d3
		neg.w	d4

loc_FE96:							; CODE XREF: ROM:0000FE8Cj
		move.w	d3,d0
		btst	#3,d0
		bne.w	loc_FEB2
		and.w	#7,d0
		move.w	($C00008).l,d1
		and.w	#3,d1
		bne.w	loc_FEC8

loc_FEB2:							; CODE XREF: ROM:0000FE9Cj
		moveq	#5,d0
		bsr.w	sub_11086
		subq.w	#2,d0
		and.w	#7,d0
		tst.w	d4
		bmi.w	loc_FEC8
		eor.w	#4,d0

loc_FEC8:							; CODE XREF: ROM:0000FEAEj
					; ROM:0000FEC0j
		asl.w	#2,d0
		move.l	#unk_10E64,a0
		move.w	(a0,d0.w),d1
		asl.w	#5,d1
		move.w	d1,$28(a3)
		move.w	2(a0,d0.w),d1
		asl.w	#5,d1
		add.w	d4,d1

loc_FEE2:
		move.w	d1,$2A(a3)
		move.w	#$800,d0
		bsr.w	sub_11086
		move.w	d0,$2C(a3)
		clr.w	($FFFFB762).w
		bclr	#2,$62(a3)
		move.l	#$18,d0
		bra.w	sub_10662
; ---------------------------------------------------------------------------
		dc.b	0
		dc.b	8
		dc.b	$10
		dc.b	0
		dc.b	8
		dc.b	$10

; =============== S U B	R O U T	I N E =======================================


sub_FF0C:							; DATA XREF: ROM:00018DDCo

; FUNCTION CHUNK AT 00014D7A SIZE 00000096 BYTES

		btst	#2,($FFFFC2F2).w
		beq.w	loc_FF1A
		bsr.w	sub_F2F4

loc_FF1A:							; CODE XREF: sub_FF0C+6j
		bclr	#1,$62(a3)
		beq.w	loc_FF2E
		clr.w	$40(a3)
		move.w	#$78,$42(a3)				; 'x'

loc_FF2E:							; CODE XREF: sub_FF0C+14j
		move.w	#$BEE6,a1
		sub.w	d7,2(a1)
		sub.w	d7,6(a1)
		sub.w	d7,$40(a3)
		bpl.w	loc_FF4A
		addq.w	#5,$40(a3)
		bsr.w	sub_10346

loc_FF4A:							; CODE XREF: sub_FF0C+32j
		move.w	($FFFFB7AA).w,d0
		bmi.w	loc_FF86
		asl.w	#7,d0
		move.w	#$B04A,a2
		add.w	d0,a2
		bsr.w	sub_1013E
		move.l	a2,-(sp)
		bsr.w	sub_106E0
		add.w	(a2),d0
		sub.w	(a3),d0
		asr.w	#2,d0
		add.w	d0,(a3)
		add.w	$14(a2),d1
		sub.w	$14(a3),d1
		asr.w	#2,d1
		add.w	d1,$14(a3)
		move.w	$28(a2),$28(a3)
		move.w	$2A(a2),$2A(a3)

loc_FF86:							; CODE XREF: sub_FF0C+42j
		bsr.w	sub_10246
		bsr.w	sub_FFEA
		btst	#0,($FFFFC2EA).w
		bne.w	loc_FFD2
		tst.w	($FFFFB7AA).w
		bpl.w	loc_FFCC
		move.w	(a3),d0
		cmp.w	$1C(a3),d0
		bne.w	loc_FFCC
		move.w	$14(a3),d0
		cmp.w	$20(a3),d0
		bne.w	loc_FFCC
		move.l	#6,d0
		subq.w	#1,$42(a3)
		bpl.w	loc_FFC8
		bsr.w	sub_11F62

loc_FFC8:							; CODE XREF: sub_FF0C+B4j
		bra.w	loc_FFD2
; ---------------------------------------------------------------------------

loc_FFCC:							; CODE XREF: sub_FF0C+90j sub_FF0C+9Aj ...
		move.w	#$78,$42(a3)				; 'x'

loc_FFD2:							; CODE XREF: sub_FF0C+88j
					; sub_FF0C:loc_FFC8j
		tst.b	$2C(a3)
		bne.w	loc_14D8C
		tst.w	$18(a3)
		bne.w	loc_14D8C
		bsr.w	sub_102A8
		bra.w	loc_14D8C
; End of function sub_FF0C


; =============== S U B	R O U T	I N E =======================================


sub_FFEA:							; CODE XREF: sub_FF0C+7Ep
		btst	#5,($FFFFC2EA).w
		beq.w	locret_15464
		btst	#2,($FFFFC2F2).w
		bne.w	locret_15464

loc_FFFE:							; DATA XREF: sub_BC40+FAo
		move.w	#$C6CE,a1
		lea	$364(a1),a2
		bsr.w	sub_100A6
		exg	a1,a2
		bsr.w	sub_100A6
		move.w	#$54,d0					; 'T'
		cmp.w	$14(a3),d0
		bgt.w	loc_1005C
		cmp.w	$20(a3),d0
		ble.w	locret_15464
		add.w	#$A,d0
		btst	#1,($FFFFC2EA).w
		beq.w	loc_10034
		dc.w	$C549					; exg.l	a2,a1

loc_10034:							; CODE XREF: sub_FFEA+44j
		moveq	#5,d2
		move.w	$22(a2),a0

loc_1003A:							; CODE XREF: sub_FFEA+6Cj
		tst.w	$34(a0)
		bmi.w	loc_10052
		cmp.w	$14(a0),d0
		bge.w	loc_10052
		bset	#4,$30(a2)
		rts
; ---------------------------------------------------------------------------

loc_10052:							; CODE XREF: sub_FFEA+54j sub_FFEA+5Cj
		add.w	#$80,a0					; ''
		dbf	d2,loc_1003A
		rts
; ---------------------------------------------------------------------------

loc_1005C:							; CODE XREF: sub_FFEA+2Ej
		neg.w	d0
		cmp.w	$14(a3),d0
		blt.w	locret_15464
		cmp.w	$20(a3),d0
		bge.w	locret_15464
		sub.w	#$A,d0
		btst	#1,($FFFFC2EA).w
		bne.w	loc_1007E
		dc.w	$C549					; exg.l	a2,a1

loc_1007E:							; CODE XREF: sub_FFEA+8Ej
		moveq	#5,d2
		move.w	$22(a2),a0

loc_10084:							; CODE XREF: sub_FFEA+B6j
		tst.w	$34(a0)
		bmi.w	loc_1009C
		cmp.w	$14(a0),d0
		ble.w	loc_1009C
		bset	#4,$30(a2)
		rts
; ---------------------------------------------------------------------------

loc_1009C:							; CODE XREF: sub_FFEA+9Ej sub_FFEA+A6j
		add.w	#$80,a0					; ''
		dbf	d2,loc_10084
		rts
; End of function sub_FFEA


; =============== S U B	R O U T	I N E =======================================


sub_100A6:							; CODE XREF: sub_FFEA+1Cp sub_FFEA+22p
		btst	#4,$30(a2)
		beq.w	locret_15464
		move.w	$22(a2),a0
		moveq	#5,d1

loc_100B6:							; CODE XREF: sub_100A6+30j
		tst.w	$34(a0)
		bmi.w	loc_100CE
		move.w	$14(a0),d0
		btst	#7,$62(a0)
		bne.w	loc_100CE
		neg.w	d0

loc_100CE:							; CODE XREF: sub_100A6+14j
					; sub_100A6+22j
		add.w	#$80,a0					; ''
		cmpnw	 $58,d0					; 'X'
		dbgt	d1,loc_100B6
		bgt.w	locret_15464
		bclr	#4,$30(a2)
		rts
; End of function sub_100A6


; =============== S U B	R O U T	I N E =======================================


sub_100E6:							; CODE XREF: sub_1013E+6Cp
		btst	#5,($FFFFC2EA).w
		beq.w	locret_15464
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a2)
		bne.w	loc_10100
		neg.w	d0

loc_10100:							; CODE XREF: sub_100E6+14j
		cmpnw	 $68,d0					; 'h'
		blt.w	locret_15464
		move.w	#$C6CE,a0
		btst	#6,$62(a2)
		beq.w	loc_1011A
		add.w	#$364,a0

loc_1011A:							; CODE XREF: sub_100E6+2Cj
		btst	#4,$30(a0)
		beq.w	locret_15464
		btst	#4,($FFFFC2EA).w
		bne.w	locret_15464
		exg	a2,a3
		move.l	#$10,d0
		bsr.w	sub_11F2C
		exg	a2,a3
		rts
; End of function sub_100E6


; =============== S U B	R O U T	I N E =======================================


sub_1013E:							; CODE XREF: sub_FF0C+4Ep
					; sub_150E4+F2p ...
		move.w	(a2),($FFFFBF7C).w
		move.w	$14(a2),($FFFFBF7E).w
		move.w	$52(a2),($FFFFBF7A).w
		move.w	#$C6CE,a0
		btst	#6,$62(a2)
		beq.w	loc_10160
		lea	$364(a0),a0

loc_10160:							; CODE XREF: sub_1013E+1Aj
		clr.w	d0
		move.b	$66(a2),d0
		btst	#3,$64(a2)
		beq.w	loc_10176
		bset	#7,($FFFFC2FE).w

loc_10176:							; CODE XREF: sub_1013E+2Ej
		cmp.w	$18(a0),d0
		beq.w	loc_101A4
		bclr	#3,$30(a0)
		bne.w	loc_10194
		move.w	$1A(a0),$1C(a0)
		move.w	$18(a0),$1A(a0)

loc_10194:							; CODE XREF: sub_1013E+46j
		move.w	d0,$18(a0)
		cmp.w	$1C(a0),d0
		bne.w	loc_101A4
		st	$1C(a0)

loc_101A4:							; CODE XREF: sub_1013E+3Cj
					; sub_1013E+5Ej
		bclr	#4,($FFFFC2EE).w
		bsr.w	sub_100E6
		btst	#2,($FFFFBF80).w
		beq.w	loc_10202
		btst	#0,($FFFFBF80).w
		beq.w	loc_10202
		tst.w	$34(a2)
		beq.w	loc_10202
		btst	#1,($FFFFBF80).w
		bne.w	loc_101FA
		btst	#7,$62(a2)
		beq.w	loc_10202

loc_101DE:							; CODE XREF: sub_1013E:loc_10200j
		clr.w	d0
		move.b	($FFFFBF81).w,d0
		asl.w	#7,d0
		move.l	a3,-(sp)
		move.w	#$B04A,a3
		add.w	d0,a3
		move.w	#$C,d0
		bsr.w	sub_11F2C
		move.l	(sp)+,a3
		rts
; ---------------------------------------------------------------------------

loc_101FA:							; CODE XREF: sub_1013E+92j
		btst	#7,$62(a2)

loc_10200:							; DATA XREF: ROM:off_BA48o
		beq.s	loc_101DE

loc_10202:							; CODE XREF: sub_1013E+76j
					; sub_1013E+80j ...
		clr.b	($FFFFBF80).w
		move.b	$53(a2),($FFFFBF81).w
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a2)
		beq.w	loc_10222
		bset	#1,($FFFFBF80).w
		neg.w	d0

loc_10222:							; CODE XREF: sub_1013E+D8j
		bmi.w	locret_15464
		move.w	($FFFFC6F2).w,d0
		sub.w	($FFFFCA56).w,d0
		btst	#6,$62(a2)
		beq.w	loc_1023A
		neg.w	d0

loc_1023A:							; CODE XREF: sub_1013E+F6j
		bmi.w	locret_15464
		bset	#2,($FFFFBF80).w
		rts
; End of function sub_1013E


; =============== S U B	R O U T	I N E =======================================


sub_10246:							; CODE XREF: sub_FF0C:loc_FF86p
		btst	#2,($FFFFBF80).w
		beq.w	locret_15464
		btst	#0,($FFFFBF80).w
		bne.w	locret_15464
		tst.w	($FFFFB7AA).w
		bpl.w	locret_15464
		move.w	#$108,d0
		btst	#1,($FFFFBF80).w
		bne.w	loc_1027C
		neg.w	d0
		cmp.w	($FFFFB75E).w,d0
		bgt.w	loc_10284
		rts
; ---------------------------------------------------------------------------

loc_1027C:							; CODE XREF: sub_10246+26j
		cmp.w	($FFFFB75E).w,d0
		bgt.w	locret_15464

loc_10284:							; CODE XREF: sub_10246+30j
		cmp.w	#$2C,($FFFFB74A).w			; ','
		bgt.w	loc_102A0
		cmp.w	#$FFD4,($FFFFB74A).w
		blt.w	loc_102A0
		bclr	#2,($FFFFBF80).w
		rts
; ---------------------------------------------------------------------------

loc_102A0:							; CODE XREF: sub_10246+44j
					; sub_10246+4Ej
		bset	#0,($FFFFBF80).w
		rts
; End of function sub_10246


; =============== S U B	R O U T	I N E =======================================


sub_102A8:							; CODE XREF: sub_FF0C+D6p
					; DATA XREF: ROM:00018DE4o
		cmp.w	#8,$5A(a3)
		blt.w	loc_102C2
		cmp.w	#$18,$5A(a3)
		bge.w	loc_102C2
		eor.w	#2,$54(a3)

loc_102C2:							; CODE XREF: sub_102A8+6j
					; sub_102A8+10j
		or.w	#4,$54(a3)
		clr.w	$5A(a3)
		st	$5C(a3)
		rts
; End of function sub_102A8


; =============== S U B	R O U T	I N E =======================================


sub_102D2:							; CODE XREF: sub_14620+158j
					; sub_14C70+60p ...
		and.w	#1,d0
		eor.w	d0,$54(a3)
		and.w	#3,$54(a3)
		st	$5C(a3)
		move.w	#$46A,d1
		bra.w	sub_1073A
; End of function sub_102D2


; =============== S U B	R O U T	I N E =======================================


sub_102EC:							; DATA XREF: ROM:00018DE0o
		cmp.w	#$18A,6(a3)
		bne.w	loc_10318
		move.w	-$80(a3),(a3)
		move.w	-$6C(a3),$14(a3)
		clr.w	$18(a3)
		moveq	#$14,d0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_10312
		moveq	#0,d0

loc_10312:							; CODE XREF: sub_102EC+20j
		addq.w	#1,(a3,d0.w)
		rts
; ---------------------------------------------------------------------------

loc_10318:							; CODE XREF: sub_102EC+6j
		tst.w	6(a3)
		beq.w	locret_15464
		clr.w	(a3)
		move.w	#$12C,$14(a3)
		move.w	#$E,$18(a3)
		tst.w	-$6C(a3)
		bpl.w	locret_10344
		move.w	#$8000,4(a3)
		neg.w	$14(a3)
		subq.w	#1,$18(a3)

locret_10344:							; CODE XREF: sub_102EC+46j
		rts
; End of function sub_102EC


; =============== S U B	R O U T	I N E =======================================


sub_10346:							; CODE XREF: sub_FF0C+3Ap
		movem.l	d0-d4/a1-a2,-(sp)
		move.w	#$BEE6,a1
		move.w	#$88,d1					; ''
		move.w	#$108,d4
		bsr.w	sub_10366
		neg.w	d4
		bsr.w	sub_10366
		movem.l	(sp)+,d0-d4/a1-a2
		rts
; End of function sub_10346


; =============== S U B	R O U T	I N E =======================================


sub_10366:							; CODE XREF: sub_10346+10p
					; sub_10346+16p
		move.w	d4,d0
		sub.w	($FFFFB75E).w,d0
		tst.w	($FFFFB774).w
		beq.w	loc_103BE
		move.w	d0,d2
		swap	d2
		clr.w	d2
		asr.l	#4,d2
		divs.w	($FFFFB774).w,d2
		bmi.w	loc_103BE
		move.w	d2,2(a1)
		muls.w	($FFFFB772).w,d0
		divs.w	($FFFFB774).w,d0
		bvs.w	loc_103BE
		add.w	($FFFFB74A).w,d0
		cmp.w	d1,d0
		blt.w	loc_103A6
		neg.w	d0
		add.w	d1,d0
		add.w	d1,d0
		add.w	d1,d0

loc_103A6:							; CODE XREF: sub_10366+34j
		neg.w	d1
		cmp.w	d1,d0
		bgt.w	loc_103B6
		neg.w	d0
		add.w	d1,d0
		add.w	d1,d0
		add.w	d1,d0

loc_103B6:							; CODE XREF: sub_10366+44j
		neg.w	d1
		move.w	d0,(a1)
		bra.w	loc_103C4
; ---------------------------------------------------------------------------

loc_103BE:							; CODE XREF: sub_10366+Aj
					; sub_10366+1Aj ...
		move.w	#$FFFF,2(a1)

loc_103C4:							; CODE XREF: sub_10366+54j
		addq.w	#4,a1
		rts
; End of function sub_10366


; =============== S U B	R O U T	I N E =======================================


sub_103C8:							; CODE XREF: sub_C710+14Ej
					; sub_C8F0+118j ...
		sub.b	d7,$42(a3)
		bpl.w	loc_1047E
		add.b	#$C,$42(a3)
		bsr.w	sub_10488
		movem.w	d0-d1,-(sp)
		move.w	$28(a3),d0
		asr.w	#8,d0
		neg.w	d0
		add.w	(sp)+,d0
		sub.w	(a3),d0
		move.w	$2A(a3),d1
		asr.w	#8,d1
		neg.w	d1
		add.w	(sp)+,d1
		sub.w	$14(a3),d1
		cmpnw	 $C,d0
		bgt.w	loc_1041E
		cmpnw	 $FFF4,d0
		blt.w	loc_1041E
		cmpnw	 $C,d1
		bgt.w	loc_1041E
		cmpnw	 $FFF4,d1
		blt.w	loc_1041E
		moveq	#9,d0
		bra.w	loc_10422
; ---------------------------------------------------------------------------

loc_1041E:							; CODE XREF: sub_103C8+34j
					; sub_103C8+3Cj ...
		bsr.w	sub_10676

loc_10422:							; CODE XREF: sub_103C8+52j
		jsr	(a0)
		move.b	d0,$43(a3)
		cmpnw	 7,d0
		ble.w	loc_1047E
		move.w	$28(a3),d0
		or.w	$2A(a3),d0
		bne.w	loc_1047E
		move.w	($FFFFB74A).w,d0
		move.w	($FFFFB75E).w,d1
		btst	#0,($FFFFB7AD).w
		beq.w	loc_10456
		move.w	($FFFFBF8E).w,d0
		move.w	($FFFFBF90).w,d1

loc_10456:							; CODE XREF: sub_103C8+82j
		sub.w	(a3),d0
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		sub.w	$54(a3),d0
		beq.w	loc_1047E
		neg.w	d0
		and.w	#4,d0
		lsr.w	#1,d0
		subq.w	#1,d0
		add.w	$54(a3),d0
		and.w	#7,d0
		move.w	d0,$54(a3)

loc_1047E:							; CODE XREF: sub_103C8+4j
					; sub_103C8+64j ...
		clr.w	d0
		move.b	$43(a3),d0
		bra.w	loc_10750
; End of function sub_103C8


; =============== S U B	R O U T	I N E =======================================


sub_10488:							; CODE XREF: sub_103C8+Ep
					; sub_105A2+10p
		clr.w	($FFFFBF9C).w
		clr.w	($FFFFBF9E).w
		tst.w	$34(a3)
		beq.w	locret_15464
		move.w	(a3),d2
		eor.w	d0,d2
		bpl.w	loc_1051E
		move.w	d1,d2
		sub.w	$14(a3),d2
		move.w	(a3),d3
		muls.w	d3,d2
		sub.w	d0,d3
		divs.w	d3,d2
		add.w	$14(a3),d2
		cmpnw	 $121,d2
		bgt.w	loc_1051E
		cmpnw	 $DB,d2					; ''
		blt.w	loc_104EA
		move.w	#$144,d3
		cmpnw	 $FE,d2					; ''
		bgt.w	loc_104E0
		blt.w	loc_104DC
		cmp.w	#$FE,$14(a3)				; ''
		bgt.w	loc_104E0

loc_104DC:							; CODE XREF: sub_10488+46j
		move.w	#$B8,d3					; ''

loc_104E0:							; CODE XREF: sub_10488+42j
					; sub_10488+50j
		sub.w	d2,d3
		move.w	d3,($FFFFBF9E).w
		bra.w	loc_1051E
; ---------------------------------------------------------------------------

loc_104EA:							; CODE XREF: sub_10488+36j
		cmpnw	 $FF25,d2
		bgt.w	loc_1051E
		cmpnw	 $FEDF,d2
		blt.w	loc_1051E
		move.w	#$FF48,d3
		cmpnw	 $FF02,d2
		bgt.w	loc_10518
		blt.w	loc_10514
		cmp.w	#$FF02,$14(a3)
		bgt.w	loc_10518

loc_10514:							; CODE XREF: sub_10488+7Ej
		move.w	#$FEBC,d3

loc_10518:							; CODE XREF: sub_10488+7Aj
					; sub_10488+88j
		sub.w	d2,d3
		move.w	d3,($FFFFBF9E).w

loc_1051E:							; CODE XREF: sub_10488+14j
					; sub_10488+2Ej ...
		move.w	d1,d3
		sub.w	#$FE,d3					; ''
		move.w	$14(a3),d2
		sub.w	#$FE,d2					; ''
		bsr.w	sub_1054C
		move.w	d1,d3
		add.w	#$FE,d3					; ''
		move.w	$14(a3),d2
		add.w	#$FE,d2					; ''
		bsr.w	sub_1054C
		add.w	($FFFFBF9C).w,d0
		add.w	($FFFFBF9E).w,d1
		rts
; End of function sub_10488


; =============== S U B	R O U T	I N E =======================================


sub_1054C:							; CODE XREF: sub_10488+A4p
					; sub_10488+B6p
		move.w	d3,d4
		eor.w	d2,d4
		bpl.w	locret_15464
		move.w	d0,d4
		sub.w	(a3),d4
		muls.w	d2,d4
		sub.w	d3,d2
		divs.w	d2,d4
		add.w	(a3),d4
		cmpnw	 $50,d4					; 'P'
		bgt.w	locret_15464
		cmpnw	 $FFB0,d4
		blt.w	locret_15464
		moveq	#$50,d3					; 'P'
		tst.w	d4
		bne.w	loc_1057A
		tst.w	(a3)

loc_1057A:							; CODE XREF: sub_1054C+28j
		bpl.w	loc_10580
		neg.w	d3

loc_10580:							; CODE XREF: sub_1054C:loc_1057Aj
		sub.w	d4,d3
		move.w	d3,($FFFFBF9C).w
		rts
; End of function sub_1054C


; =============== S U B	R O U T	I N E =======================================


sub_10588:							; CODE XREF: sub_E6CE:loc_E9AEp
					; sub_105A2p
		move.b	($FFFFB772).w,d0
		asr.b	#1,d0
		ext.w	d0
		add.w	($FFFFB74A).w,d0
		move.b	($FFFFB774).w,d1
		asr.b	#1,d1
		ext.w	d1
		add.w	($FFFFB75E).w,d1
		rts
; End of function sub_10588


; =============== S U B	R O U T	I N E =======================================


sub_105A2:							; CODE XREF: sub_DD2E:loc_DDC0j
					; sub_E6CE:loc_EACAp
		bsr.s	sub_10588
		sub.b	d7,$42(a3)
		bpl.w	loc_1063E
		add.b	#$A,$42(a3)
		bsr.w	sub_10488
		movem.w	d0-d1,-(sp)
		move.l	a3,-(sp)
		bsr.w	sub_106E0
		neg.b	d0
		neg.b	d1
		sub.b	$28(a3),d0
		ext.w	d0
		add.w	(sp)+,d0
		sub.w	(a3),d0
		sub.b	$2A(a3),d1
		ext.w	d1
		add.w	(sp)+,d1
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		move.b	d0,$43(a3)
		tst.w	$34(a3)
		beq.w	loc_1063E
		move.w	($FFFFB772).w,d0
		move.w	($FFFFB774).w,d1
		bsr.w	sub_10676
		eor.w	#4,d0
		cmp.w	$54(a3),d0
		beq.w	loc_1063E
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		move.w	($FFFFB75E).w,d1
		sub.w	$14(a3),d1
		movem.w	d0-d1,-(sp)
		bsr.w	sub_10676
		cmp.w	$54(a3),d0
		movem.w	(sp)+,d0-d1
		bne.w	loc_1063E
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmpnl	 $384,d0
		bls.w	loc_1063E
		cmpnl	 $5A4,d0
		bls.w	loc_C0AE

loc_1063E:							; CODE XREF: sub_105A2+6j
					; sub_105A2+44j ...
		move.b	$43(a3),d0
		bra.w	loc_10750
; End of function sub_105A2


; =============== S U B	R O U T	I N E =======================================


sub_10646:							; CODE XREF: sub_C710+1Ej sub_C882+6Aj ...
		addq.w	#1,$36(a3)
		and.w	#7,$36(a3)
		bset	#1,$62(a3)
		rts
; End of function sub_10646


; =============== S U B	R O U T	I N E =======================================


sub_10658:							; CODE XREF: sub_7B30+36p sub_7B30+D2p ...
		subq.w	#1,$36(a3)
		and.w	#7,$36(a3)
; End of function sub_10658


; =============== S U B	R O U T	I N E =======================================


sub_10662:							; CODE XREF: sub_7CB0:loc_78CAp
					; sub_B0E8+14Cp ...
		move.l	d1,-(sp)
		move.w	$36(a3),d1
		move.b	d0,$38(a3,d1.w)
		bset	#1,$62(a3)
		move.l	(sp)+,d1
		rts
; End of function sub_10662


; =============== S U B	R O U T	I N E =======================================


sub_10676:							; CODE XREF: sub_9FD0+18Ap
					; sub_B0E8+404p ...
		movem.l	d2/a0,-(sp)
		move.w	d0,d2
		or.w	d1,d2
		beq.w	loc_106C8
		clr.w	d2
		tst.w	d0
		bpl.w	loc_10690
		neg.w	d0
		bset	#0,d2

loc_10690:							; CODE XREF: sub_10676+10j
		tst.w	d1
		bpl.w	loc_1069C
		neg.w	d1
		bset	#1,d2

loc_1069C:							; CODE XREF: sub_10676+1Cj
		asl.w	#1,d1
		cmp.w	d1,d0
		bhi.w	loc_106A8
		bset	#2,d2

loc_106A8:							; CODE XREF: sub_10676+2Aj
		lsr.w	#1,d1
		asl.w	#1,d0
		cmp.w	d0,d1
		bhi.w	loc_106B6
		bset	#3,d2

loc_106B6:							; CODE XREF: sub_10676+38j
		move.l	#byte_106D0,a0
		clr.w	d0
		move.b	(a0,d2.w),d0
		movem.l	(sp)+,d2/a0
		rts
; ---------------------------------------------------------------------------

loc_106C8:							; CODE XREF: sub_10676+8j
		moveq	#8,d0
		movem.l	(sp)+,d2/a0
		rts
; End of function sub_10676

; ---------------------------------------------------------------------------
byte_106D0:	dc.b	1					; DATA XREF: sub_10676:loc_106B6o
		dc.b	7
		dc.b	3
		dc.b	5
		dc.b	0
		dc.b	0
		dc.b	4
		dc.b	4
		dc.b	2
		dc.b	6
		dc.b	2
		dc.b	6
		dc.b	1
		dc.b	7
		dc.b	3
		dc.b	5

; =============== S U B	R O U T	I N E =======================================


sub_106E0:							; CODE XREF: sub_BE26+ACp
					; sub_E6CE+1BEp ...

arg_0		=	4

		movem.l	a0-a1,-(sp)
		move.l	8+arg_0(sp),a0
		clr.w	d0
		clr.w	d1
		tst.w	6(a0)
		ble.w	loc_10732
		move.l	#unk_A44C8,a1
		move.w	6(a0),d0
		add.w	d0,d0
		move.b	1(a1,d0.w),d1
		ext.w	d1
		move.b	(a1,d0.w),d0
		ext.w	d0
		btst	#3,4(a0)
		beq.w	loc_10718
		neg.w	d0

loc_10718:							; CODE XREF: sub_106E0+32j
		btst	#4,4(a0)
		bne.w	loc_10724
		neg.w	d1

loc_10724:							; CODE XREF: sub_106E0+3Ej
		btst	#7,($FFFFC2EC).w
		beq.w	loc_10732
		exg	d0,d1
		neg.w	d1

loc_10732:							; CODE XREF: sub_106E0+10j
					; sub_106E0+4Aj
		movem.l	(sp)+,a0-a1
		move.l	(sp)+,(sp)
		rts
; End of function sub_106E0


; =============== S U B	R O U T	I N E =======================================


sub_1073A:							; CODE XREF: sub_B0E8+370p
					; sub_B0E8+7FCj ...
		cmp.w	$58(a3),d1
		beq.w	locret_15464
		clr.w	$5A(a3)
		move.w	d1,$58(a3)
		st	$5C(a3)
		rts
; End of function sub_1073A

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_10750:							; CODE XREF: sub_B0E8+53Ej
					; sub_B0E8+5CEj ...
		tst.w	$34(a3)
		beq.w	loc_1090A
		move.w	#$50C,d1
		btst	#4,$62(a3)
		beq.w	loc_1076A
		move.w	#$A60,d1

loc_1076A:							; CODE XREF: sub_B0E8+567Aj
		and.w	#$F,d0
		cmpnw	 7,d0
		ble.w	loc_10794
		cmpnw	 9,d0
		bne.w	loc_1078A
		move.w	$28(a3),d0
		or.w	$2A(a3),d0
		bne.w	loc_10DB6

loc_1078A:							; CODE XREF: sub_B0E8+5692j
		btst	#1,$63(a3)
		beq.s	sub_1073A
		rts
; ---------------------------------------------------------------------------

loc_10794:							; CODE XREF: sub_B0E8+568Aj
		movem.w	d0-d1,-(sp)
		move.w	d0,d2
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		beq.w	loc_1085A
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		move.w	$14(a3),d3
		move.b	($FFFFB774).w,d1
		ext.w	d1
		add.w	($FFFFB75E).w,d1
		sub.w	d3,d1
		btst	#7,$62(a3)
		bne.w	loc_107D0
		neg.w	d0
		neg.w	d1
		neg.w	d3
		eor.w	#4,d2

loc_107D0:							; CODE XREF: sub_B0E8+56DAj
		btst	#4,$62(a3)
		bne.w	loc_107FE
		tst.w	d3
		bpl.w	loc_1085A
		cmpnw	 4,d2
		bne.w	loc_1085A
		bsr.w	sub_10676
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bhi.w	loc_1085A
		bra.w	loc_1081E
; ---------------------------------------------------------------------------

loc_107FE:							; CODE XREF: sub_B0E8+56EEj
		subq.w	#3,d2
		and.w	#7,d2
		cmpnw	 2,d2
		bhi.w	loc_1085A
		bsr.w	sub_10676
		addq.w	#2,d0
		and.w	#7,d0
		cmpnw	 4,d0
		bhi.w	loc_1085A

loc_1081E:							; CODE XREF: sub_B0E8+5712j
		btst	#4,$62(a3)
		bne.w	loc_10860
		move.w	$28(a3),d0
		or.w	$2A(a3),d0
		beq.w	loc_10850
		move.w	$28(a3),d0
		move.w	$2A(a3),d1
		bsr.w	sub_10676
		sub.w	(sp),d0
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bhi.w	loc_10860

loc_10850:							; CODE XREF: sub_B0E8+5748j
		bset	#4,$62(a3)
		bra.w	loc_10860
; ---------------------------------------------------------------------------

loc_1085A:							; CODE XREF: sub_B0E8+56BAj
					; sub_B0E8+56F4j ...
		bclr	#4,$62(a3)

loc_10860:							; CODE XREF: sub_B0E8+573Cj
					; sub_B0E8+5764j ...
		movem.w	(sp)+,d0-d1
		move.w	$54(a3),d2
		sub.w	d2,d0
		and.w	#7,d0
		move.l	#word_108FA,a0
		asl.w	#1,d0
		tst.w	(a0,d0.w)
		beq.w	loc_10B0A
		move.w	$28(a3),d4
		muls.w	d4,d4
		move.w	$2A(a3),d3
		muls.w	d3,d3
		add.l	d4,d3
		swap	d3
		move.w	#$300,d4
		sub.w	d3,d4
		cmpnw	 $180,d4
		bge.w	loc_108A0
		move.w	#$180,d4

loc_108A0:							; CODE XREF: sub_B0E8+57B0j
		muls.w	(a0,d0.w),d4
		btst	#4,$62(a3)
		beq.w	loc_108B0
		neg.l	d4

loc_108B0:							; CODE XREF: sub_B0E8+57C2j
		add.l	d4,$54(a3)
		and.w	#7,$54(a3)
		move.w	$54(a3),d2
		cmpnw	 $14,d3
		bls.w	loc_108EC
		clr.w	d1
		tst.w	(a0,d0.w)
		bpl.w	loc_108D4
		eor.w	#$FFCE,d1

loc_108D4:							; CODE XREF: sub_B0E8+57E4j
		btst	#3,4(a3)
		beq.w	loc_108E2
		eor.w	#$FFCE,d1

loc_108E2:							; CODE XREF: sub_B0E8+57F2j
		add.w	#$694,d1
		bset	#1,$63(a3)

loc_108EC:							; CODE XREF: sub_B0E8+57DAj
		bsr.w	sub_1073A
		cmpnw	 2,d3
		bhi.w	loc_10BB4
		rts
; END OF FUNCTION CHUNK	FOR sub_B0E8
; ---------------------------------------------------------------------------
word_108FA:	dc.w	0					; DATA XREF: sub_B0E8+5786o
		dc.w	$10
		dc.w	$10
		dc.w	$10
		dc.w	0
		dc.w	$FFF0
		dc.w	$FFF0
		dc.w	$FFF0
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_1090A:							; CODE XREF: sub_B0E8+566Cj
		btst	#3,$62(a3)
		beq.w	loc_10A42
		movem.w	d0,-(sp)
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		movem.w	(sp)+,d0
		beq.w	loc_10A42
		cmp.w	#$30,(a3)				; '0'
		bgt.w	loc_10A42
		cmp.w	#$FFD0,(a3)
		blt.w	loc_10A42
		cmp.w	#$FF40,$14(a3)
		bgt.w	loc_10950
		cmp.w	#$FEFA,$14(a3)
		blt.w	loc_10A42
		bra.w	loc_10964
; ---------------------------------------------------------------------------

loc_10950:							; CODE XREF: sub_B0E8+5856j
		cmp.w	#$C0,$14(a3)				; ''
		blt.w	loc_10A42
		cmp.w	#$106,$14(a3)
		bgt.w	loc_10A42

loc_10964:							; CODE XREF: sub_B0E8+5864j
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		move.w	($FFFFB75E).w,d1
		sub.w	$14(a3),d1
		cmpnw	 $10,d0
		bgt.w	loc_1099E
		cmpnw	 $FFF0,d0
		blt.w	loc_1099E
		cmpnw	 $10,d1
		bgt.w	loc_1099E
		cmpnw	 $FFF0,d1
		blt.w	loc_1099E
		move.w	$54(a3),d0
		bra.w	loc_109A2
; ---------------------------------------------------------------------------

loc_1099E:							; CODE XREF: sub_B0E8+5892j
					; sub_B0E8+589Aj ...
		bsr.w	sub_10676

loc_109A2:							; CODE XREF: sub_B0E8+58B2j
		btst	#0,($FFFFC2F4).w
		bne.w	loc_109B8
		bsr.w	loc_DB68
		movem.w	(sp)+,d0-d1
		bra.w	loc_109D0
; ---------------------------------------------------------------------------

loc_109B8:							; CODE XREF: sub_B0E8+58C0j
		movem.w	(sp)+,d0-d1
		cmpnw	 8,d0
		beq.w	loc_109D0
		movem.w	d0-d1,-(sp)
		bsr.w	loc_DB68
		movem.w	(sp)+,d0-d1

loc_109D0:							; CODE XREF: sub_B0E8+58CCj
					; sub_B0E8+58D8j
		move.w	#2,d1
		btst	#1,$63(a3)
		bne.w	locret_10ADA
		bsr.w	sub_1073A
		cmpnw	 8,d0
		bne.w	loc_10A3C
		tst.w	$34(a3)
		bne.w	loc_10A3C
		btst	#3,$62(a3)
		beq.w	loc_10A3C
		move.w	d0,-(sp)
		move.w	$14(a3),d0
		btst	#7,$62(a3)
		beq.w	loc_10A0E
		neg.w	d0

loc_10A0E:							; CODE XREF: sub_B0E8+5920j
		cmpnw	 $D8,d0					; ''
		blt.w	loc_10A34
		move.w	(a3),d0
		cmpnw	 $20,d0					; ' '
		bgt.w	loc_10A34
		cmpnw	 $FFE0,d0
		blt.w	loc_10A34
		move.w	(sp)+,d0
		jsr	sub_F6F8C
		bra.w	loc_10A3C
; ---------------------------------------------------------------------------

loc_10A34:							; CODE XREF: sub_B0E8+592Aj
					; sub_B0E8+5934j ...
		jsr	sub_F6F8C
		move.w	(sp)+,d0

loc_10A3C:							; CODE XREF: sub_B0E8+58FEj
					; sub_B0E8+5906j ...
		move.w	d0,d2
		bra.w	loc_10BC2
; ---------------------------------------------------------------------------

loc_10A42:							; CODE XREF: sub_B0E8+5828j
					; sub_B0E8+583Cj ...
		move.w	#2,d1
		btst	#3,$62(a3)
		beq.w	loc_10AAA
		btst	#2,$62(a3)
		bne.w	loc_10AD0
		cmpnw	 8,d0
		bne.w	loc_10AA6
		tst.w	$34(a3)
		bne.w	loc_10AD0
		move.w	d0,-(sp)
		move.w	$14(a3),d0
		btst	#7,$62(a3)
		beq.w	loc_10A7C
		neg.w	d0

loc_10A7C:							; CODE XREF: sub_B0E8+598Ej
		cmpnw	 $D8,d0					; ''
		blt.w	loc_10AA0
		move.w	(a3),d0
		cmpnw	 $20,d0					; ' '
		bgt.w	loc_10AA0
		cmpnw	 $FFE0,d0
		blt.w	loc_10AA0
		move.w	(sp)+,d0
		bsr.w	sub_10E1A
		bra.w	loc_10AD0
; ---------------------------------------------------------------------------

loc_10AA0:							; CODE XREF: sub_B0E8+5998j
					; sub_B0E8+59A2j ...
		move.w	(sp)+,d0
		bra.w	loc_10AD0
; ---------------------------------------------------------------------------

loc_10AA6:							; CODE XREF: sub_B0E8+5976j
		bra.w	loc_10ADC
; ---------------------------------------------------------------------------

loc_10AAA:							; CODE XREF: sub_B0E8+5964j
		and.w	#$F,d0
		cmpnw	 7,d0
		ble.w	loc_10ADC
		cmpnw	 9,d0
		bne.w	loc_10AD0
		move.w	$28(a3),d0
		or.w	$2A(a3),d0
		beq.w	loc_10AD0
		jmp	sub_F6F8C
; ---------------------------------------------------------------------------

loc_10AD0:							; CODE XREF: sub_B0E8+596Ej
					; sub_B0E8+597Ej ...
		btst	#1,$63(a3)
		beq.w	sub_1073A

locret_10ADA:							; CODE XREF: sub_B0E8+58F2j
		rts
; ---------------------------------------------------------------------------

loc_10ADC:							; CODE XREF: sub_B0E8:loc_10AA6j
					; sub_B0E8+59CAj
		sub.w	$54(a3),d0
		beq.w	loc_10AFA
		neg.w	d0
		and.w	#4,d0
		lsr.w	#1,d0
		subq.w	#1,d0
		add.w	$54(a3),d0
		and.w	#7,d0
		move.w	d0,$54(a3)

loc_10AFA:							; CODE XREF: sub_B0E8+59F8j
		move.w	#$3D8,d1
		bsr.w	sub_1073A
		move.w	$54(a3),d2
		bra.w	loc_10BC2
; ---------------------------------------------------------------------------

loc_10B0A:							; CODE XREF: sub_B0E8+5792j
		moveq	#2,d4
		btst	#4,$62(a3)
		beq.w	loc_10B1C
		addq.w	#4,d4
		eor.w	#8,d0

loc_10B1C:							; CODE XREF: sub_B0E8+5A2Aj
		tst.w	d0
		beq.w	loc_10B76
		move.w	$28(a3),d0
		move.w	$2A(a3),d1
		bsr.w	sub_10676
		btst	#3,d0
		bne.w	loc_10B48
		sub.w	$54(a3),d0
		add.w	d4,d0
		and.w	#7,d0
		cmpnw	 4,d0
		blt.w	loc_10DB6

loc_10B48:							; CODE XREF: sub_B0E8+5A4Aj
		addq.w	#1,$54(a3)
		btst	#3,4(a3)
		beq.w	loc_10B5A
		subq.w	#2,$54(a3)

loc_10B5A:							; CODE XREF: sub_B0E8+5A6Aj
		and.w	#7,$54(a3)
		move.w	#$50C,d1
		btst	#4,$62(a3)
		beq.w	sub_1073A
		move.w	#$A60,d1
		bra.w	sub_1073A
; ---------------------------------------------------------------------------

loc_10B76:							; CODE XREF: sub_B0E8+5A36j
		move.w	#$A92,d1
		btst	#4,$62(a3)
		bne.w	loc_10BA6
		move.w	#$5D0,d1
		btst	#6,$63(a3)
		beq.w	loc_10B96
		move.w	#$11E6,d1

loc_10B96:							; CODE XREF: sub_B0E8+5AA6j
		move.w	($FFFFB7AA).w,d4
		cmp.w	$52(a3),d4
		bne.w	loc_10BA6
		move.w	#$53E,d1

loc_10BA6:							; CODE XREF: sub_B0E8+5A98j
					; sub_B0E8+5AB6j
		btst	#1,$63(a3)
		bne.w	loc_10BB4
		bsr.w	sub_1073A

loc_10BB4:							; CODE XREF: sub_B0E8+580Cj
					; sub_B0E8+5AC4j
		btst	#4,$62(a3)
		beq.w	loc_10BC2
		eor.w	#4,d2

loc_10BC2:							; CODE XREF: ROM:0000CC4Aj
					; sub_D51C+592j ...
		asl.w	#2,d2
		lea	$10E64(pc),a0
		move.w	2(a0,d2.w),d1
		move.w	(a0,d2.w),d0
		move.w	$50(a3),d2
		beq.w	loc_10BE0
		eor.w	d0,d2
		bpl.w	loc_10BE0
		clr.w	d0

loc_10BE0:							; CODE XREF: sub_B0E8+5AECj
					; sub_B0E8+5AF2j
		move.w	$4E(a3),d2
		beq.w	loc_10BF0
		eor.w	d1,d2
		bmi.w	loc_10BF0
		clr.w	d1

loc_10BF0:							; CODE XREF: sub_B0E8+5AFCj
					; sub_B0E8+5B02j
		clr.w	d2
		move.b	$67(a3),d2
		lsr.w	#2,d2
		neg.w	d2
		add.w	#$40,d2					; '@'
		add.b	$68(a3),d2
		btst	#1,($FFFFC2FE).w
		bne.w	loc_10C16
		btst	#6,($FFFFC2FC).w
		beq.w	loc_10C18

loc_10C16:							; CODE XREF: sub_B0E8+5B20j
		addq.b	#2,d2

loc_10C18:							; CODE XREF: sub_B0E8+5B2Aj
		tst.w	$34(a3)
		bne.w	loc_10C28
		add.b	$68(a3),d2
		add.w	#$10,d2

loc_10C28:							; CODE XREF: sub_B0E8+5B34j
		asr.w	#1,d2
		muls.w	d2,d0
		muls.w	d2,d1
		asr.l	#5,d0
		asr.l	#5,d1
		muls.w	d7,d0
		muls.w	d7,d1
		tst.w	$34(a3)
		bne.w	loc_10C58
		btst	#3,$62(a3)
		beq.w	loc_10C58
		move.w	($FFFFB7AA).w,d2
		cmp.w	$52(a3),d2
		bne.w	loc_10C58
		asr.w	#1,d0
		asr.w	#1,d1

loc_10C58:							; CODE XREF: sub_B0E8+5B52j
					; sub_B0E8+5B5Cj ...
		add.w	$28(a3),d0
		add.w	$2A(a3),d1
		move.w	d0,d2
		move.w	d1,d3
		muls.w	d2,d2
		muls.w	d3,d3
		add.l	d2,d3
		movem.w	d0-d1,-(sp)
		bsr.w	sub_1575A
		btst	#4,($FFFFC2FC).w
		beq.w	loc_10C80
		move.w	#$1000,d0

loc_10C80:							; CODE XREF: sub_B0E8+5B90j
		clr.w	d2
		move.b	$69(a3),d2
		btst	#1,($FFFFC2FE).w
		bne.w	loc_10C9A
		btst	#6,($FFFFC2FC).w
		beq.w	loc_10CA8

loc_10C9A:							; CODE XREF: sub_B0E8+5BA4j
		addq.b	#2,d2
		cmpnb	 $1E,d2
		ble.w	loc_10CA8
		move.b	#$1E,d2

loc_10CA8:							; CODE XREF: sub_B0E8+5BAEj
					; sub_B0E8+5BB8j
		move.w	d2,($FFFFBF16).w
		lsr.w	#1,d2
		mulu.w	d0,d2
		asl.l	#4,d2
		swap	d2
		asl.w	#2,d2
		and.w	#$3F,d2					; '?'
		lea	dword_10D76(pc),a2
		move.w	d2,($FFFFBF1C).w
		move.l	(a2,d2.w),d2
		move.l	d2,($FFFFBF18).w
		cmp.w	#$3C,($FFFFBF1C).w			; '<'
		bge.w	loc_10CF2
		btst	#0,($FFFFBF17).w
		beq.w	loc_10CF2
		move.w	($FFFFBF1C).w,d2
		addq.w	#4,d2
		move.l	(a2,d2.w),d2
		sub.l	($FFFFBF18).w,d2
		asr.l	#1,d2
		add.l	($FFFFBF18).w,d2

loc_10CF2:							; CODE XREF: sub_B0E8+5BE8j
					; sub_B0E8+5BF2j
		btst	#6,$63(a3)
		beq.w	loc_10CFE
		lsr.l	#3,d2

loc_10CFE:							; CODE XREF: sub_B0E8+5C10j
		tst.w	$34(a3)
		bne.w	loc_10D14
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		bne.w	loc_10D14
		asr.l	#1,d2

loc_10D14:							; CODE XREF: sub_B0E8+5C1Aj
					; sub_B0E8+5C26j
		movem.w	(sp)+,d0-d1
		cmp.l	d2,d3
		bhi.w	loc_10D26
		move.w	d0,$28(a3)
		move.w	d1,$2A(a3)

loc_10D26:							; CODE XREF: sub_B0E8+5C32j
		tst.w	($FFFFD058).w
		bne.w	locret_15464
		move.w	($C00008).l,d0
		and.w	#$7F,d0					; ''
		bne.w	locret_15464
		bsr.w	sub_1575A
		sub.w	#$21,d0					; '!'
		cmpnw	 $C00,d0
		blt.w	sub_1577A
		move.b	$72(a3),d2
		ext.w	d2
		lsr.w	#1,d2
		add.w	d2,d0
		bsr.w	sub_1577A
		btst	#0,$72(a3)
		beq.w	locret_10D74
		btst	#0,($FFFFB03F).w
		beq.w	locret_10D74
		addq.w	#1,d0
		bra.w	sub_1577A
; ---------------------------------------------------------------------------

locret_10D74:							; CODE XREF: sub_B0E8+5C78j
					; sub_B0E8+5C82j
		rts
; END OF FUNCTION CHUNK	FOR sub_B0E8
; ---------------------------------------------------------------------------
dword_10D76:	dc.l	$1CD9410				; DATA XREF: sub_B0E8+5BD2o
		dc.l	$1FCE3E1
		dc.l	$22E8284
		dc.l	$2626FF9
		dc.l	$298AC40
		dc.l	$2D13759
		dc.l	$30C1144
		dc.l	$3493A01
		dc.l	$388B190
		dc.l	$3CA77F1
		dc.l	$40E8D24
		dc.l	$454F129
		dc.l	$49DA400
		dc.l	$4E8A5A9
		dc.l	$535F624
		dc.l	$5859571
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_10DB6:							; CODE XREF: sub_B0E8+569Ej
					; sub_B0E8+5A5Cj
		cmp.w	#$1000,$28(a3)
		bgt.w	loc_10DF0
		cmp.w	#$F000,$28(a3)
		blt.w	loc_10DF0
		cmp.w	#$1000,$2A(a3)
		bgt.w	loc_10DF0
		cmp.w	#$F000,$2A(a3)
		blt.w	loc_10DF0
		tst.w	$34(a3)
		bne.w	sub_10E1A
		jmp	sub_F6F8C
; END OF FUNCTION CHUNK	FOR sub_B0E8
; ---------------------------------------------------------------------------
		bra.w	sub_10E1A
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_B0E8

loc_10DF0:							; CODE XREF: sub_B0E8+5CD4j
					; sub_B0E8+5CDEj ...
		move.w	#$50C,d1
		btst	#4,$62(a3)
		bne.w	loc_10E08
		bset	#1,$63(a3)
		move.w	#$6C6,d1

loc_10E08:							; CODE XREF: sub_B0E8+5D12j
		bsr.w	sub_1073A
		tst.w	$34(a3)
		bne.w	sub_10E1A
		jmp	sub_F6F8C
; END OF FUNCTION CHUNK	FOR sub_B0E8

; =============== S U B	R O U T	I N E =======================================


sub_10E1A:							; CODE XREF: sub_D51C+596j
					; sub_B0E8+59B0p ...
		tst.w	$28(a3)
		bpl.w	loc_10E30
		add.w	#$96,$28(a3)				; ''
		bmi.w	loc_10E3E
		clr.w	$28(a3)

loc_10E30:							; CODE XREF: sub_10E1A+4j
		sub.w	#$96,$28(a3)				; ''
		bpl.w	loc_10E3E
		clr.w	$28(a3)

loc_10E3E:							; CODE XREF: sub_10E1A+Ej
					; sub_10E1A+1Cj
		tst.w	$2A(a3)
		bpl.w	loc_10E54
		add.w	#$96,$2A(a3)				; ''
		bmi.w	locret_15464
		clr.w	$2A(a3)

loc_10E54:							; CODE XREF: sub_10E1A+28j
		sub.w	#$96,$2A(a3)				; ''
		bpl.w	locret_15464
		clr.w	$2A(a3)
		rts
; End of function sub_10E1A

; ---------------------------------------------------------------------------
unk_10E64:	dc.b	0					; DATA XREF: sub_B0E8+AAAo
					; sub_BC40+132o ...
		dc.b	0
		dc.b	0
		dc.b	$C8					; 
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	$C8					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$8D					; 
		dc.b	$FF
		dc.b	$73					; s
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$73					; s
		dc.b	$FF
		dc.b	$73					; s
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$73					; s
		dc.b	0
		dc.b	$8D					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_10E88:							; CODE XREF: ROM:00017C30p
		movem.l	d0-d2/a0-a1,-(sp)
		move.w	#$D036,a1
		clr.w	d2
		bra.w	loc_10EAA
; ---------------------------------------------------------------------------

loc_10E96:							; CODE XREF: sub_10E88:loc_10EAAj
		move.b	(a0)+,d1
		bchg	#0,d2
		bne.w	loc_10EA4
		subq.w	#1,a0
		lsr.w	#4,d1

loc_10EA4:							; CODE XREF: sub_10E88+14j
		and.w	#$F,d1
		move.w	d1,(a1)+

loc_10EAA:							; CODE XREF: sub_10E88+Aj
		dbf	d0,loc_10E96
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_10E88


; =============== S U B	R O U T	I N E =======================================


sub_10EB4:							; CODE XREF: sub_132E2+52j
					; ROM:00017C34p
		movem.l	d1/a1,-(sp)
		move.w	#$D036,a1
		clr.w	d1
		bra.w	loc_10EC4
; ---------------------------------------------------------------------------

loc_10EC2:							; CODE XREF: sub_10EB4:loc_10EC4j
		add.w	(a1)+,d1

loc_10EC4:							; CODE XREF: sub_10EB4+Aj
		dbf	d0,loc_10EC2
		move.w	d1,d0
		bsr.w	sub_11086

loc_10ECE:							; CODE XREF: sub_10EB4+1Cj
		sub.w	-(a1),d0
		bpl.s	loc_10ECE
		suba.w	#$D036,a1
		move.w	a1,d0
		lsr.w	#1,d0
		movem.l	(sp)+,d1/a1
		rts
; End of function sub_10EB4

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_1173C

loc_10EE0:							; CODE XREF: sub_1173C+2Aj
					; DATA XREF: sub_1177A+Co
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.l	d0-d4/a0-a2,-(sp)
		dc.w	$C340					; exg.l	d1,d0
		move.l	a0,a2
		bsr.w	VDP_SetAddress
		subq.w	#1,d1

loc_10EF8:							; CODE XREF: sub_1173C-818j
		moveq	#3,d0
		move.w	(a2)+,d2
		clr.w	d3

loc_10EFE:							; CODE XREF: sub_1173C-81Ej
		move.w	d2,d4
		and.w	#$F,d4
		lsr.w	#1,d4
		move.b	(a1,d4.w),d4
		btst	#0,d2
		bne.w	loc_10F14
		lsr.w	#4,d4

loc_10F14:							; CODE XREF: sub_1173C-82Ej
		and.w	#$F,d4
		or.b	d4,d3
		ror.w	#4,d3
		ror.w	#4,d2
		dbf	d0,loc_10EFE
		move.w	d3,(a0)
		dbf	d1,loc_10EF8
		movem.l	(sp)+,d0-d4/a0-a2
		move.w	(sp)+,($FFFFBF78).w
		rts
; END OF FUNCTION CHUNK	FOR sub_1173C

; =============== S U B	R O U T	I N E =======================================


sub_10F32:							; CODE XREF: sub_7CF8p	sub_7CF8+6Ep ...
		movem.l	d0/a0,-(sp)
		move.w	#$BD28,a0
		moveq	#$1F,d0

loc_10F3C:							; CODE XREF: sub_10F32+Ej
		move.l	(a0),-(sp)
		clr.l	(a0)+
		dbf	d0,loc_10F3C
		move.w	#$18,($FFFFBD26).w			; set V-Blank routine to run
		bsr.w	WaitVBlank
		moveq	#$1F,d0

loc_10F50:							; CODE XREF: sub_10F32+20j
		move.l	(sp)+,-(a0)
		dbf	d0,loc_10F50
		movem.l	(sp)+,d0/a0
		rts
; End of function sub_10F32


; =============== S U B	R O U T	I N E =======================================


sub_10F5C:							; CODE XREF: sub_A528+1Cp
		movem.l	d0/a0,-(sp)
		move.w	#$BD28,a0
		moveq	#$1F,d0

loc_10F66:							; CODE XREF: sub_10F5C+Ej
		move.l	(a0),-(sp)
		clr.l	(a0)+
		dbf	d0,loc_10F66
		move.w	#$64,($FFFFBD26).w			; 'd'
		bsr.w	WaitVBlank
		moveq	#$1F,d0

loc_10F7A:							; CODE XREF: sub_10F5C+20j
		move.l	(sp)+,-(a0)
		dbf	d0,loc_10F7A
		movem.l	(sp)+,d0/a0
		rts
; End of function sub_10F5C


; =============== S U B	R O U T	I N E =======================================


WaitVBlank:							; CODE XREF: sub_10F32+18p
					; sub_10F5C+18p ...
		move	sr,-(sp)
		move.l	($FFFFB03A).w,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bclr	#2,($FFFFBF78).w
		move.l	#sub_15E4C,($FFFFB03A).l
		move	#$2500,sr

loc_10FA4:							; CODE XREF: WaitVBlank+22j
		tst.w	($FFFFBD26).w
		bpl.s	loc_10FA4
		move.w	(sp)+,($FFFFBF78).w
		move.l	(sp)+,($FFFFB03A).w
		move	(sp)+,sr
		rts
; End of function WaitVBlank


; =============== S U B	R O U T	I N E =======================================


sub_10FB6:							; CODE XREF: sub_15D9A:loc_15DB6p
					; sub_15E4C+Ep	...
		tst.w	($FFFFBD26).w
		bmi.w	locret_15464
		cmp.w	#$64,($FFFFBD26).w			; 'd'
		beq.w	VDP_FlushPalette
		subq.w	#1,($FFFFBD26).w
		bmi.w	locret_15464
		clr.l	d0
		move.w	($FFFFBD26).w,d0
		cmpnw	 $18,d0
		bgt.w	locret_15464
		divu.w	#3,d0
		swap	d0
		asl.w	#2,d0
		moveq	#2,d3
		asl.w	d0,d3
		moveq	#$E,d5
		asl.w	d0,d5
		move.w	d5,d4
		not.w	d4
		move.l	#$FFFFBD28,a1
		move.l	#$C00000,a0
		clr.w	d6

loc_11000:							; CODE XREF: sub_10FB6+8Aj
		move.w	d3,d2
		move.w	d6,d0
		swap	d0
		move.w	#$20,d0					; ' '
		move.l	d0,4(a0)
		move.w	(a0),d7
		move.w	d7,d0
		and.w	d5,d0
		move.w	(a1)+,d1
		and.w	d5,d1
		cmp.w	d1,d0
		beq.w	loc_1103A
		blt.w	loc_11024
		neg.w	d2

loc_11024:							; CODE XREF: sub_10FB6+68j
		add.w	d2,d0
		and.w	d4,d7
		or.w	d0,d7
		move.l	#$C000,d0
		move.b	d6,d0
		swap	d0
		move.l	d0,4(a0)
		move.w	d7,(a0)

loc_1103A:							; CODE XREF: sub_10FB6+64j
		addq.w	#2,d6
		cmpnw	 $80,d6					; ''
		bne.s	loc_11000
		rts
; End of function sub_10FB6


; =============== S U B	R O U T	I N E =======================================


VDP_FlushPalette:						; CODE XREF: sub_10FB6+Ej
					; sub_17A18+4Ap
		movem.l	d0/a0-a1,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	#$BD28,a1				; Palette buffer
		move.l	#$C00000,a0				; VDP data port
		move.l	#$C0000000,4(a0)			; Set VDP to CRAM write mode
		moveq	#$1F,d0

loc_11066:							; CODE XREF: VDP_FlushPalette+24j
		move.l	(a1)+,(a0)				; Copy palette of colours into CRAM
		dbf	d0,loc_11066
		st	($FFFFBD26).w
		move.w	(sp)+,($FFFFBF78).w
		movem.l	(sp)+,d0/a0-a1
		rts
; End of function VDP_FlushPalette


; =============== S U B	R O U T	I N E =======================================


sub_1107A:							; CODE XREF: sub_C2F2+182p
					; sub_C2F2:loc_C494p ...
		move.w	d0,-(sp)
		asl.w	#1,d0
		bsr.w	sub_11086
		sub.w	(sp)+,d0
		rts
; End of function sub_1107A


; =============== S U B	R O U T	I N E =======================================


sub_11086:							; CODE XREF: sub_7814+28p
					; sub_BC40+16Ap ...

var_A		=	-$A

		movem.l	d0-d2,-(sp)
		move.w	($FFFFD068).w,d0
		move.w	d0,d1
		move.w	($FFFFD066).w,d2
		mulu.w	#$E62D,d0
		mulu.w	#$BB40,d1
		mulu.w	#$E62D,d2
		add.w	d2,d1
		swap	d0
		add.w	d1,d0
		swap	d0
		addq.l	#1,d0
		move.l	d0,($FFFFD066).w
		asr.l	#8,d0
		mulu.w	$C+var_A(sp),d0
		swap	d0
		addq.w	#4,sp
		movem.l	(sp)+,d1-d2
		rts
; End of function sub_11086


; =============== S U B	R O U T	I N E =======================================


sub_110BE:							; CODE XREF: sub_9BB4+4p
					; sub_B0E8:loc_B55Ep ...
		tst.l	d0
		beq.w	locret_15464
		cmpnl	 $640,d0
		bhi.w	loc_110E0
		move.l	d1,-(sp)
		moveq	#$FFFFFFFF,d1

loc_110D2:							; CODE XREF: sub_110BE+18j
		addq.w	#2,d1
		sub.w	d1,d0
		bcc.s	loc_110D2
		lsr.w	#1,d1
		move.w	d1,d0
		move.l	(sp)+,d1
		rts
; ---------------------------------------------------------------------------

loc_110E0:							; CODE XREF: sub_110BE+Cj
		movem.l	d1-d4,-(sp)
		moveq	#9,d3
		move.w	#$8000,d1
		cmpnl	 $F00000,d0
		bhi.w	loc_11112
		move.l	d0,d1
		lsr.l	#8,d1
		addq.w	#2,d1

loc_110FA:							; CODE XREF: sub_110BE+48j
		move.w	d1,d2
		move.l	d0,d1
		divu.w	d2,d1
		add.w	d2,d1
		lsr.w	#1,d1
		cmp.w	d1,d2
		dbeq	d3,loc_110FA

loc_1110A:							; CODE XREF: sub_110BE+60j
		move.w	d1,d0
		movem.l	(sp)+,d1-d4
		rts
; ---------------------------------------------------------------------------

loc_11112:							; CODE XREF: sub_110BE+32j
		moveq	#0,d1
		moveq	#$FFFFFFFF,d2

loc_11116:							; CODE XREF: sub_110BE+6Ej
					; sub_110BE+72j
		move.w	d1,d3
		add.w	d2,d3
		roxr.w	#1,d3
		cmp.w	d3,d1
		beq.s	loc_1110A
		move.w	d3,d4
		mulu.w	d3,d3
		cmp.l	d3,d0
		bcc.w	loc_1112E
		move.w	d4,d2
		bra.s	loc_11116
; ---------------------------------------------------------------------------

loc_1112E:							; CODE XREF: sub_110BE+68j
		move.w	d4,d1
		bra.s	loc_11116
; End of function sub_110BE


; =============== S U B	R O U T	I N E =======================================


sub_11132:							; CODE XREF: sub_7B30+22p
					; sub_9FD0+34Cp ...

arg_0		=	4

		movem.l	d0-a6,-(sp)
		clr.l	d0
		move.w	$3C+arg_0(sp),d0
		bmi.w	loc_1114A
		move.w	d0,($FFFFC016).w
		jsr	sub_1A294

loc_1114A:							; CODE XREF: sub_11132+Aj
		movem.l	(sp)+,d0-a6
		move.l	(sp),2(sp)
		addq.w	#2,sp
		rts
; End of function sub_11132


; =============== S U B	R O U T	I N E =======================================


sub_11156:							; CODE XREF: sub_7CB0-38Ep
					; sub_794E+36p	...

arg_0		=	4

		movem.l	d0-a6,-(sp)
		clr.l	d0
		move.w	$3C+arg_0(sp),d0
		bmi.w	loc_1116A
		jsr	sub_1A294

loc_1116A:							; CODE XREF: sub_11156+Aj
		movem.l	(sp)+,d0-a6
		move.l	(sp),2(sp)
		addq.w	#2,sp
		rts
; End of function sub_11156


; =============== S U B	R O U T	I N E =======================================


sub_11176:							; CODE XREF: ROM:00009DE2j
					; sub_17A18+9Ap ...
		clr.w	($FFFFDED4).w
		neg.w	d0
		move.w	d0,($FFFFB03E).w

loc_11180:							; CODE XREF: sub_11176+56j
		bsr.w	sub_11340
		move.w	d3,($FFFFDED4).w
		tst.w	d1
		bne.w	locret_15464
		bsr.w	sub_11358
		or.w	d3,($FFFFDED4).w
		tst.w	d1
		bne.w	locret_15464
		tst.w	($FFFFD046).w
		beq.w	loc_111C0
		bsr.w	sub_11370
		or.w	d3,($FFFFDED4).w
		tst.w	d1
		bne.w	locret_15464
		bsr.w	sub_11388
		or.w	d3,($FFFFDED4).w
		tst.w	d1
		bne.w	locret_15464

loc_111C0:							; CODE XREF: sub_11176+2Aj
		move.w	($FFFFB03E).w,d0

loc_111C4:							; CODE XREF: sub_11176+52j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_111C4
		tst.w	d0
		bmi.s	loc_11180
		rts
; End of function sub_11176


; =============== S U B	R O U T	I N E =======================================


sub_111D0:							; CODE XREF: sub_130E6+92p
					; sub_130E6+A6p ...
		movem.l	d4-a3,-(sp)
		neg.w	d0
		move.w	d0,($FFFFB03E).w

loc_111DA:							; CODE XREF: sub_111D0+D0j
		cmp.w	#$708,($FFFFB8AC).w
		bhi.w	loc_111E8
		addq.w	#1,($FFFFB8AC).w

loc_111E8:							; CODE XREF: sub_111D0+10j
		moveq	#1,d7
		jsr	(sub_7A9E).w
		jsr	sub_FE2C8
		clr.w	($FFFFC316).w
		bclr	#1,($FFFFC2EC).w
		bsr.w	sub_11340
		bsr.w	sub_11318
		tst.w	d1
		bne.w	loc_11268
		clr.w	($FFFFC316).w

loc_11210:
		bset	#1,($FFFFC2EC).w
		bsr.w	sub_11358
		bsr.w	sub_11318
		tst.w	d1
		bne.w	loc_11268
		tst.w	($FFFFD046).w
		beq.w	loc_11290
		bclr	#1,($FFFFC2EC).w
		move.w	#3,($FFFFC316).w
		bsr.w	sub_11370
		bsr.w	sub_11318
		tst.w	d1
		bne.w	loc_11268
		tst.w	($FFFFD046).w
		beq.w	loc_11290
		bset	#1,($FFFFC2EC).w
		move.w	#4,($FFFFC316).w
		bsr.w	sub_11388
		bsr.w	sub_11318
		tst.w	d1
		beq.w	loc_11290

loc_11268:							; CODE XREF: sub_111D0+38j
					; sub_111D0+50j ...
		move.w	($FFFFB03E).w,d0
		cmpnw	 $FF88,d0
		blt.w	loc_11276
		moveq	#$FFFFFF88,d0

loc_11276:							; CODE XREF: sub_111D0+A0j
		movem.w	d0,-(sp)
		jsr	(loc_7E88).w
		bne.w	loc_1128C
		addq.w	#2,sp
		bset	#7,d1
		bra.w	loc_112A4
; ---------------------------------------------------------------------------

loc_1128C:							; CODE XREF: sub_111D0+AEj
		move.w	(sp)+,($FFFFB03E).w

loc_11290:							; CODE XREF: sub_111D0+58j
					; sub_111D0+7Aj ...
		bsr.w	sub_15EC0
		move.w	($FFFFB03E).w,d0

loc_11298:							; CODE XREF: sub_111D0+CCj
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_11298
		tst.w	d0
		bmi.w	loc_111DA

loc_112A4:							; CODE XREF: sub_111D0+B8j
		movem.l	(sp)+,d4-a3
		rts
; End of function sub_111D0

; ---------------------------------------------------------------------------

loc_112AA:							; CODE XREF: ROM:000112B8j
		move.w	($FFFFB03E).w,d0

loc_112AE:							; CODE XREF: ROM:000112B2j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_112AE
		bsr.w	sub_112BC
		beq.s	loc_112AA
		rts

; =============== S U B	R O U T	I N E =======================================


sub_112BC:							; CODE XREF: ROM:00007736p
					; ROM:000112B4p ...
		tst.w	($FFFFD046).w
		bne.w	loc_112D2
		bsr.w	sub_11340
		move.w	d1,-(sp)
		bsr.w	sub_11358
		or.w	(sp)+,d1
		rts
; ---------------------------------------------------------------------------

loc_112D2:							; CODE XREF: sub_112BC+4j
		bsr.w	sub_11340
		move.w	d1,-(sp)
		bsr.w	sub_11358
		move.w	d1,-(sp)
		bsr.w	sub_11370
		move.w	d1,-(sp)
		bsr.w	sub_11388
		or.w	(sp)+,d1
		or.w	(sp)+,d1
		or.w	(sp)+,d1
		rts
; End of function sub_112BC


; =============== S U B	R O U T	I N E =======================================


sub_112F0:							; CODE XREF: ROM:00008AA0p
					; ROM:0000951Cp ...
		movem.l	d0/d4-d5,-(sp)
		moveq	#3,d4
		move.w	d3,d0
		and.w	#$F,d0
		beq.w	loc_11312

loc_11300:							; CODE XREF: sub_112F0+16j
		clr.w	d5
		bset	d4,d5
		cmp.w	d5,d0
		dbeq	d4,loc_11300
		beq.w	loc_11312
		and.w	#$FFF0,d1

loc_11312:							; CODE XREF: sub_112F0+Cj
					; sub_112F0+1Aj
		movem.l	(sp)+,d0/d4-d5
		rts
; End of function sub_112F0


; =============== S U B	R O U T	I N E =======================================


sub_11318:							; CODE XREF: sub_7CF8+62p
					; ROM:0000832Ap ...
		bsr.s	sub_112F0
		tst.w	d3
		beq.w	locret_15464
		tst.w	d2
		bne.w	loc_11338
		subq.w	#1,($FFFFB042).w
		bpl.w	locret_15464
		move.w	#4,($FFFFB042).w
		move.w	d3,d1
		rts
; ---------------------------------------------------------------------------

loc_11338:							; CODE XREF: sub_11318+Aj
		move.w	#$F,($FFFFB042).w
		rts
; End of function sub_11318


; =============== S U B	R O U T	I N E =======================================


sub_11340:							; CODE XREF: sub_7CB0-55Ap
					; sub_7C58+10p	...
		move.b	($FFFFBEF6).w,d0
		bsr.w	sub_113A0
		move.w	($FFFFBEEE).w,d2
		move.w	d1,($FFFFBEEE).w
		move.w	d1,d3
		eor.w	d1,d2
		and.w	d2,d1
		rts
; End of function sub_11340


; =============== S U B	R O U T	I N E =======================================


sub_11358:							; CODE XREF: sub_7C58+22p sub_A41E+Ej	...
		move.b	($FFFFBEF7).w,d0
		bsr.w	sub_113A0
		move.w	($FFFFBEF0).w,d2
		move.w	d1,($FFFFBEF0).w
		move.w	d1,d3
		eor.w	d1,d2
		and.w	d2,d1
		rts
; End of function sub_11358


; =============== S U B	R O U T	I N E =======================================


sub_11370:							; CODE XREF: sub_7C58+38p sub_A41E+22j ...
		move.b	($FFFFBEF8).w,d0
		bsr.w	sub_113A0
		move.w	($FFFFBEF2).w,d2
		move.w	d1,($FFFFBEF2).w
		move.w	d1,d3
		eor.w	d1,d2
		and.w	d2,d1
		rts
; End of function sub_11370


; =============== S U B	R O U T	I N E =======================================


sub_11388:							; CODE XREF: sub_7C58+4Ap sub_A41E+26j ...
		move.b	($FFFFBEF9).w,d0
		bsr.w	sub_113A0
		move.w	($FFFFBEF4).w,d2
		move.w	d1,($FFFFBEF4).w
		move.w	d1,d3
		eor.w	d1,d2
		and.w	d2,d1
		rts
; End of function sub_11388


; =============== S U B	R O U T	I N E =======================================


sub_113A0:							; CODE XREF: sub_11340+4p sub_11358+4p ...
		not.b	d0
		clr.w	d1
		move.b	d0,d1
		move.w	d1,-(sp)
		and.w	#$F0,d0					; ''
		and.w	#$F,d1
		move.l	#unk_113C0,a0
		move.b	(a0,d1.w),d1
		or.w	d1,d0
		move.w	(sp)+,d1
		rts
; End of function sub_113A0

; ---------------------------------------------------------------------------
unk_113C0:	dc.b	8					; DATA XREF: sub_113A0+10o
					; sub_164D6+9Eo
		dc.b	0
		dc.b	4
		dc.b	8
		dc.b	6
		dc.b	7
		dc.b	5
		dc.b	8
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	8
		dc.b	8
		dc.b	8
		dc.b	8
		dc.b	8
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_1173C

loc_113D0:							; CODE XREF: sub_1173C+22j
					; DATA XREF: sub_1177A+12o
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		bsr.w	sub_113E4
		move.w	(sp)+,($FFFFBF78).w
		rts
; END OF FUNCTION CHUNK	FOR sub_1173C

; =============== S U B	R O U T	I N E =======================================


sub_113E4:							; CODE XREF: sub_1173C-362p
					; sub_15E6E+10p ...
		movem.l	d2-d3/a1,-(sp)
		move.w	d0,d2
		add.w	d2,d2
		add.w	a0,d2
		bcc.w	loc_11412
		beq.w	loc_11412
		lsr.w	#1,d2
		sub.w	d2,d0
		move.w	d0,-(sp)
		bsr.w	VDP_TransferDMA
		move.w	(sp)+,d0
		add.w	d0,d0
		add.w	d0,d1
		add.w	d0,a0
		move.w	d2,d0
		bra.w	loc_11412
; End of function sub_113E4


; =============== S U B	R O U T	I N E =======================================


VDP_TransferDMA:						; CODE XREF: sub_113E4+18p
		movem.l	d2-d3/a1,-(sp)				; d0 = size
					; a0 = source
					; d1 = destination (VRAM word)

loc_11412:							; CODE XREF: sub_113E4+Aj sub_113E4+Ej ...
		lea	($C00004).l,a1
		move.w	#$8154,(a1)
		move.w	#$8F02,(a1)
		move.w	#$9300,d2
		move.b	d0,d2
		move.w	d2,(a1)
		move.w	#$9400,d2
		lsr.w	#8,d0
		move.b	d0,d2
		move.w	d2,(a1)
		move.l	a0,d0
		lsr.l	#1,d0
		move.w	#$9500,d2
		move.b	d0,d2
		move.w	d2,(a1)
		lsr.l	#8,d0
		move.w	#$9600,d2
		move.b	d0,d2
		move.w	d2,(a1)
		lsr.l	#8,d0
		and.b	#$7F,d0					; ''
		move.w	#$9700,d2
		move.b	d0,d2
		move.w	d2,(a1)
		clr.l	d0
		move.w	d1,d0
		asl.l	#2,d0
		lsr.w	#2,d0
		move.l	d0,d3
		or.l	#$804000,d0
		move.l	d0,($FFFFD06A).w
		move	sr,-(sp)
		move	#$2700,sr
		move.w	#$100,($A11100).l

loc_11478:							; CODE XREF: VDP_TransferDMA+72j
		btst	#0,($A11100).l
		bne.s	loc_11478
		move.w	($FFFFD06C).w,(a1)
		move.w	($FFFFD06A).w,(a1)
		bsr.w	VDP_WaitDMA
		move.w	#$8164,(a1)
		clr.w	($A11100).l
		cmp.w	#0,a0
		bge.w	loc_114B0
		or.w	#$4000,d3
		swap	d3
		and.w	#3,d3
		move.l	d3,(a1)
		move.w	(a0),-4(a1)				; Transfer last	word (for the bug)

loc_114B0:							; CODE XREF: VDP_TransferDMA+8Ej
		move	(sp)+,sr
		movem.l	(sp)+,d2-d3/a1
		rts
; End of function VDP_TransferDMA


; =============== S U B	R O U T	I N E =======================================


sub_114B8:							; CODE XREF: sub_16D64+30p
					; sub_179D2+36p
		movem.l	d0-d3/a1,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		lea	($C00004).l,a1
		move.w	#$8154,(a1)
		move.w	#$8F01,(a1)
		move.w	#$9300,d3
		move.b	d0,d3
		move.w	d3,(a1)
		move.w	#$9400,d3
		lsr.w	#8,d0
		move.b	d0,d3
		move.w	d3,(a1)
		move.w	#$9500,d3
		move.b	d2,d3
		move.w	d3,(a1)
		move.w	#$9600,d3
		lsr.w	#8,d2
		move.b	d2,d3
		move.w	d3,(a1)
		move.w	#$97C0,(a1)
		clr.l	d0
		move.w	d1,d0
		asl.l	#2,d0
		lsr.w	#2,d0
		swap	d0
		or.w	#$C0,d0					; ''
		move.l	d0,($FFFFD06A).w
		move.w	#$100,($A11100).l

loc_11516:							; CODE XREF: sub_114B8+66j
		btst	#0,($A11100).l
		bne.s	loc_11516
		move.w	($FFFFD06A).w,(a1)
		move.w	($FFFFD06C).w,(a1)
		bsr.w	VDP_WaitDMA
		move.w	#$8164,(a1)
		move.w	#$8F02,(a1)
		clr.w	($A11100).l
		move.w	(sp)+,($FFFFBF78).w
		movem.l	(sp)+,d0-d3/a1
		rts
; End of function sub_114B8


; =============== S U B	R O U T	I N E =======================================


sub_11544:							; CODE XREF: sub_9FD0+2Cp
					; ROM:00012BFAp ...
		movem.l	d3/a1,-(sp)
		move.l	#$C00000,a1
		and.l	#$FFFF,d1
		asl.l	#2,d1
		lsr.w	#2,d1
		or.w	#$4000,d1
		swap	d1
		move.l	d1,4(a1)
		move.w	d2,d3
		swap	d3
		move.w	d2,d3
		lsr.w	#1,d0
		subq.w	#1,d0

loc_1156C:							; CODE XREF: sub_11544+34j
		move.l	d3,($FFFFD06A).w
		move.w	($FFFFD06A).w,(a1)
		move.w	($FFFFD06C).w,(a1)
		dbf	d0,loc_1156C
		movem.l	(sp)+,d3/a1
		rts
; End of function sub_11544


; =============== S U B	R O U T	I N E =======================================


VDP_WaitDMA:							; CODE XREF: VDP_TransferDMA+7Cp
					; sub_114B8+70p ...

var_2		=	-2

		move.w	($C00004).l,-(sp)
		btst	#1,1(sp)
		addq.w	#2,sp
		bne.s	VDP_WaitDMA
		rts
; End of function VDP_WaitDMA


; =============== S U B	R O U T	I N E =======================================


sub_11594:							; CODE XREF: ROM:00016A36p
					; ROM:00017382p ...
		move.w	#$BD28,a0
		moveq	#$3F,d1					; '?'

loc_1159A:							; CODE XREF: sub_11594+8j
		move.w	d0,(a0)+
		dbf	d1,loc_1159A
		move.w	#$18,($FFFFBD26).w
		bsr.w	WaitVBlank
; End of function sub_11594


; =============== S U B	R O U T	I N E =======================================


VDP_InitialSetup:						; CODE XREF: sub_17A18+4Ep
					; sub_FED70+50p
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	#$8F02,($C00004).l
		clr.w	d0
		bsr.w	VDP_SetAddress
		move.w	#$3FFF,d0
		clr.l	d1

loc_115C8:							; CODE XREF: VDP_InitialSetup+20j
		move.l	d1,(a0)					; Clear	VRAM 0000 - 3FFF
		dbf	d0,loc_115C8
		move.w	#$8C00,d0				; 20 resolution
		btst	#1,($FFFFBF78).w			; is the H-resolution set to 28?
		bne.w	loc_115E0				; if not, branch for 20
		or.w	#$81,d0					; ''   ; 28 resolution

loc_115E0:							; CODE XREF: VDP_InitialSetup+2Ej
		move.w	d0,4(a0)				; set resolution
		move.w	#$8004,4(a0)
		move.w	#$8164,4(a0)
		move.w	#$9001,d0				; $40 pixels horizontally
		cmp.w	#6,($FFFFB006).w			; is the plane	horizontal size	set to $80 pixels?
		beq.w	loc_11602				; if not, branch for $40 pixels
		move.w	#$9003,d0				; $80 pixels horizontally

loc_11602:							; CODE XREF: VDP_InitialSetup+50j
		move.w	d0,4(a0)				; set plane width size
		move.w	#$8200,d0
		move.b	($FFFFB004).w,d0			; Plane A address
		lsr.b	#2,d0
		and.b	#$38,d0					; '8'
		move.w	d0,4(a0)
		move.w	#$8400,d0
		move.b	($FFFFB008).w,d0			; Plane B address
		lsr.b	#5,d0
		move.w	d0,4(a0)
		move.w	#$8300,d0
		move.b	($FFFFB00C).w,d0			; Window Plane	address
		lsr.b	#2,d0
		and.b	#$3E,d0					; '>'
		move.w	d0,4(a0)
		move.w	#$8500,d0
		move.b	($FFFFB002).w,d0			; Sprite address
		lsr.b	#1,d0
		move.w	d0,4(a0)
		move.w	#$8D00,d0
		move.b	($FFFFB000).w,d0			; H-Scroll address
		lsr.b	#2,d0
		move.w	d0,4(a0)
		move.w	#$9100,4(a0)
		move.w	#$9200,4(a0)
		move.w	#$8700,4(a0)
		move.w	#$8B00,4(a0)
		move.l	#$40000010,4(a0)			; set VDP to VSRAM write mode
		move.l	#0,(a0)					; set FG and BG	Y position to 0
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function VDP_InitialSetup


; =============== S U B	R O U T	I N E =======================================


VDP_SetAddress:							; CODE XREF: ROM:00009BF4p
					; ROM:00009C12p ...
		move.l	#$C00000,a0				; load VDP data	port
		asl.l	#2,d0					; get upper bits of address into other word
		lsr.w	#2,d0					; send rest back
		or.w	#$4000,d0				; set "VRAM write mode"	bits
		swap	d0					; align	for VDP	port order
		and.w	#3,d0					; get only address (clear other	bit settings)
		move.l	d0,4(a0)				; set VDP's address
		rts
; End of function VDP_SetAddress


; =============== S U B	R O U T	I N E =======================================


sub_1169A:							; CODE XREF: ROM:00008120p
					; ROM:00008728p ...

var_2A		=	-$2A
var_22		=	-$22
var_1E		=	-$1E

		move.w	($FFFFB02A).w,-(sp)			; Possibly a decompression algorithm
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.l	d0-d3/d5-d6/a0-a3,-(sp)
		move.w	d1,d6
		move.w	#$BD28,a3
		bra.w	loc_116CC
; ---------------------------------------------------------------------------

loc_116B6:							; CODE XREF: sub_1169A+22j
					; sub_1169A+36j
		move.b	(a0,d0.w),(a3,d0.w)
		dbf	d0,loc_116B6

loc_116C0:							; CODE XREF: sub_1169A+38j
		add.l	#$20,a0					; ' '
		add.l	#$20,a3					; ' '

loc_116CC:							; CODE XREF: sub_1169A+18j
		moveq	#$1F,d0
		lsr.w	#1,d5
		bcs.s	loc_116B6
		bne.s	loc_116C0
		move.w	($FFFFB02C).w,d5
		and.w	#$F800,d5
		move.w	$2C+var_1E(sp),d2
		subq.w	#1,d2

loc_116E2:							; CODE XREF: sub_1169A+72j
		bsr.w	sub_11952
		move.w	d6,d0
		mulu.w	(a1),d0
		add.w	$2C+var_2A(sp),d0
		asl.w	#1,d0
		move.w	$2C+var_22(sp),d1
		subq.w	#1,d1

loc_116F6:							; CODE XREF: sub_1169A+68j
		move.w	4(a1,d0.w),d3
		add.w	d4,d3
		eor.w	d5,d3
		move.w	d3,(a0)
		addq.w	#2,d0
		dbf	d1,loc_116F6
		addq.w	#1,($FFFFB02A).w
		addq.w	#1,d6
		dbf	d2,loc_116E2
		btst	#0,($FFFFC2F8).w
		bne.w	loc_1171E
		bsr.w	sub_11738

loc_1171E:							; CODE XREF: sub_1169A+7Cj
		movem.l	(sp)+,d0-d3/d5-d6/a0-a3
		move.w	(sp)+,($FFFFBF78).w
		move.w	(sp)+,($FFFFB02A).w
		rts
; End of function sub_1169A


; =============== S U B	R O U T	I N E =======================================


sub_1172C:							; CODE XREF: ROM:00009C30p
					; ROM:00009CAAp ...
		move.l	(sp),($FFFFCF32).w
		bsr.w	sub_1173C
		addq.l	#8,(sp)
		rts
; End of function sub_1172C


; =============== S U B	R O U T	I N E =======================================


sub_11738:							; CODE XREF: ROM:loc_9D32p
					; sub_9FD0+3Ap	...
		clr.l	($FFFFCF32).w
; End of function sub_11738


; =============== S U B	R O U T	I N E =======================================


sub_1173C:							; CODE XREF: sub_1172C+4p

; FUNCTION CHUNK AT 00010EE0 SIZE 00000052 BYTES
; FUNCTION CHUNK AT 000113D0 SIZE 00000014 BYTES

		movem.l	d0-d1/a0-a6,-(sp)
		move.l	a2,a0
		move.w	d4,d1
		asl.w	#5,d1
		move.w	(a0)+,d0
		beq.w	loc_11774
		bmi.w	loc_1176A
		add.w	d0,d4
		asl.w	#4,d0
		pea	(loc_11774).l
		tst.l	($FFFFCF32).w
		beq.w	loc_113D0
		move.l	($FFFFCF32).w,a1
		bra.w	loc_10EE0
; ---------------------------------------------------------------------------

loc_1176A:							; CODE XREF: sub_1173C+10j
		and.w	#$7FFF,d0
		add.w	d0,d4
		bsr.w	sub_1177A

loc_11774:							; CODE XREF: sub_1173C+Cj
					; DATA XREF: sub_1173C+18o
		movem.l	(sp)+,d0-d1/a0-a6
		rts
; End of function sub_1173C


; =============== S U B	R O U T	I N E =======================================


sub_1177A:							; CODE XREF: sub_1173C+34p
		move.w	#$CF36,a1
		move.w	#$CF36,a3
		move.w	#$CF32,a4
		move.l	#loc_10EE0,a5
		move.l	#loc_113D0,a6
		movem.l	d0-d3/a0-a2,-(sp)
		move.w	d1,d3
		clr.w	d1
		clr.w	d2

loc_1179C:							; CODE XREF: sub_1177A+36j
		move.b	(a0)+,d0
		and.w	#$F0,d0					; ''
		lsr.w	#3,d0
		lea	word_117B2(pc),a2
		move.w	(a2,d0.w),d0
		jsr	(a2,d0.w)
		bra.s	loc_1179C
; End of function sub_1177A

; ---------------------------------------------------------------------------
word_117B2:	dc.w	$20					; DATA XREF: sub_1177A+2Ao
		dc.w	$20
		dc.w	$3C
		dc.w	$58
		dc.w	$78
		dc.w	$78
		dc.w	$78
		dc.w	$78
		dc.w	$AA
		dc.w	$B8
		dc.w	$D2
		dc.w	$EC
		dc.w	$106
		dc.w	$106
		dc.w	$138
		dc.w	$158
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#$1F,d0

loc_117DA:							; CODE XREF: ROM:loc_117E8j
		move.b	(a0)+,(a1,d1.w)
		addq.b	#1,d1
		bne.w	loc_117E8
		bsr.w	sub_11924

loc_117E8:							; CODE XREF: ROM:000117E0j
		dbf	d0,loc_117DA
		rts
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#$F,d0

loc_117F6:							; CODE XREF: ROM:loc_11804j
		clr.b	(a1,d1.w)
		addq.b	#1,d1
		bne.w	loc_11804
		bsr.w	sub_11924

loc_11804:							; CODE XREF: ROM:000117FCj
		dbf	d0,loc_117F6
		rts
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#$F,d0
		addq.w	#2,d0
		move.b	(a0)+,d2

loc_11816:							; CODE XREF: ROM:loc_11824j
		move.b	d2,(a1,d1.w)
		addq.b	#1,d1
		bne.w	loc_11824
		bsr.w	sub_11924

loc_11824:							; CODE XREF: ROM:0001181Cj
		dbf	d0,loc_11816
		rts
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#7,d0
		addq.w	#1,d0
		move.b	-1(a0),d2
		lsr.w	#3,d2
		and.w	#7,d2
		addq.w	#1,d2

loc_11840:							; CODE XREF: ROM:00011868j
					; ROM:00011882j ...
		neg.b	d2
		add.b	d1,d2

loc_11844:							; CODE XREF: ROM:loc_11856j
		move.b	(a1,d2.w),(a1,d1.w)
		addq.b	#1,d2
		addq.b	#1,d1
		bne.w	loc_11856
		bsr.w	sub_11924

loc_11856:							; CODE XREF: ROM:0001184Ej
		dbf	d0,loc_11844
		rts
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#$F,d0
		addq.w	#2,d0
		move.b	(a0)+,d2
		bra.s	loc_11840
; ---------------------------------------------------------------------------
		move.b	(a0),d0
		asl.b	#1,d0
		move.b	-1(a0),d0
		roxl.b	#1,d0
		and.w	#$1F,d0
		addq.w	#2,d0
		move.b	(a0)+,d2
		and.w	#$7F,d2					; ''
		addq.w	#1,d2
		bra.s	loc_11840
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		asl.w	#8,d0
		move.b	(a0),d0
		lsr.w	#6,d0
		and.w	#$3F,d0					; '?'
		addq.w	#2,d0
		move.b	(a0)+,d2
		and.w	#$3F,d2					; '?'
		addq.w	#1,d2
		bra.s	loc_11840
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		asl.w	#8,d0
		move.b	(a0),d0
		lsr.w	#5,d0
		and.w	#$7F,d0					; ''
		addq.w	#2,d0
		move.b	(a0)+,d2
		and.w	#$1F,d2
		addq.w	#1,d2
		bra.s	loc_11840
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#3,d0
		addq.w	#1,d0
		move.b	-1(a0),d2
		lsr.w	#2,d2
		and.w	#7,d2
		addq.w	#1,d2

loc_118CE:							; CODE XREF: ROM:000118F6j
					; ROM:00011922j
		neg.b	d2
		add.b	d1,d2

loc_118D2:							; CODE XREF: ROM:loc_118E4j
		move.b	(a1,d2.w),(a1,d1.w)
		subq.b	#1,d2
		addq.b	#1,d1
		bne.w	loc_118E4
		bsr.w	sub_11924

loc_118E4:							; CODE XREF: ROM:000118DCj
		dbf	d0,loc_118D2
		rts
; ---------------------------------------------------------------------------
		move.b	-1(a0),d0
		and.w	#$F,d0
		addq.w	#2,d0
		move.b	(a0)+,d2
		bne.s	loc_118CE
		tst.w	d1
		beq.w	loc_11902
		bsr.w	sub_11924

loc_11902:							; CODE XREF: ROM:000118FAj
		addq.w	#4,sp
		movem.l	(sp)+,d0-d3/a0-a2
		rts
; ---------------------------------------------------------------------------
		move.b	(a0),d0
		asl.b	#1,d0
		move.b	-1(a0),d0
		roxl.b	#1,d0
		and.w	#$1F,d0
		addq.w	#2,d0
		move.b	(a0)+,d2
		and.w	#$7F,d2					; ''
		addq.w	#1,d2
		bra.s	loc_118CE

; =============== S U B	R O U T	I N E =======================================


sub_11924:							; CODE XREF: ROM:000117E4p
					; ROM:00011800p ...
		movem.l	d0-d1/a0-a1,-(sp)
		move.w	d1,d0
		bne.w	loc_11932
		move.w	#$100,d0

loc_11932:							; CODE XREF: sub_11924+6j
		lsr.w	#1,d0
		move.w	d3,d1
		add.w	d0,d3
		add.w	d0,d3
		move.l	a1,a0
		tst.l	(a4)
		beq.w	loc_1194A
		move.l	(a4),a1
		jsr	(a5)
		bra.w	loc_1194C
; ---------------------------------------------------------------------------

loc_1194A:							; CODE XREF: sub_11924+1Aj
		jsr	(a6)

loc_1194C:							; CODE XREF: sub_11924+22j
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_11924


; =============== S U B	R O U T	I N E =======================================


sub_11952:							; CODE XREF: sub_8078:loc_808Ep
					; sub_1169A:loc_116E2p	...
		movem.l	d0-d2,-(sp)
		move.w	($FFFFB028).w,d0
		move.w	($FFFFB02A).w,d1
		move.l	#$FFFFB004,a0
		add.w	($FFFFB02E).w,a0
		move.w	2(a0),d2
		asl.w	d2,d1
		add.w	d1,d0
		asl.w	#1,d0
		add.w	(a0),d0
		bsr.w	VDP_SetAddress
		movem.l	(sp)+,d0-d2
		rts
; End of function sub_11952


; =============== S U B	R O U T	I N E =======================================


sub_1197E:							; CODE XREF: ROM:00007E08j
					; ROM:0000830Cp ...

var_18		=	-$18

		movem.l	d0-d2/a0,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.w	d0-d1,-(sp)

loc_11990:							; CODE XREF: sub_1197E+2Cj
		bsr.s	sub_11952
		move.w	(sp),d0
		subq.w	#1,d0

loc_11996:							; CODE XREF: sub_1197E+1Aj
		move.w	d2,(a0)
		dbf	d0,loc_11996
		addq.w	#1,($FFFFB02A).w
		and.w	#$1F,($FFFFB02A).w
		subq.w	#1,$1A+var_18(sp)
		bne.s	loc_11990
		addq.w	#4,sp
		move.w	(sp)+,($FFFFBF78).w
		movem.l	(sp)+,d0-d2/a0
		rts
; End of function sub_1197E


; =============== S U B	R O U T	I N E =======================================


sub_119B8:							; CODE XREF: ROM:00007E60p
					; ROM:000080ECp ...

var_24		=	-$24

		movem.l	d0-d4/a0-a1,-(sp)
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB02C).w,d2
		add.w	($FFFFB01E).w,d2
		move.l	#unk_55B7E,a1
		add.l	4(a1),a1
		addq.w	#4,a1
		clr.w	d4
		bsr.w	sub_11A06
		subq.w	#3,$26+var_24(sp)

loc_119E8:							; CODE XREF: sub_119B8+3Aj
		bsr.w	sub_11A06
		subq.w	#6,d4
		subq.w	#1,$26+var_24(sp)
		bpl.s	loc_119E8
		addq.w	#6,d4
		bsr.w	sub_11A06
		addq.w	#4,sp
		move.w	(sp)+,($FFFFBF78).w
		movem.l	(sp)+,d0-d4/a0-a1
		rts
; End of function sub_119B8


; =============== S U B	R O U T	I N E =======================================


sub_11A06:							; CODE XREF: sub_119B8+28p
					; sub_119B8:loc_119E8p	...

arg_0		=	4

		bsr.w	sub_11952
		addq.w	#1,($FFFFB02A).w
		bsr.w	sub_11A2C
		addq.w	#2,d4
		move.w	arg_0(sp),d0
		subq.w	#3,d0

loc_11A1A:							; CODE XREF: sub_11A06+18j
		bsr.w	sub_11A2C
		dbf	d0,loc_11A1A
		addq.w	#2,d4
		bsr.w	sub_11A2C
		addq.w	#2,d4
		rts
; End of function sub_11A06


; =============== S U B	R O U T	I N E =======================================


sub_11A2C:							; CODE XREF: sub_11A06+8p
					; sub_11A06:loc_11A1Ap	...
		move.w	(a1,d4.w),d3
		add.w	d2,d3
		move.w	d3,(a0)
		rts
; End of function sub_11A2C


; =============== S U B	R O U T	I N E =======================================


sub_11A36:							; CODE XREF: ROM:00007DF2p
					; ROM:00007E4Cp ...
		move.l	a1,-(sp)
		move.l	4(sp),a1
		bsr.w	sub_11A48
		move.l	a1,4(sp)
		move.l	(sp)+,a1
		rts
; End of function sub_11A36


; =============== S U B	R O U T	I N E =======================================


sub_11A48:							; CODE XREF: ROM:00007F4Ap
					; ROM:00007FCAp ...
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.l	d0-d3/a0/a2-a3,-(sp)			; store register data
		move.w	#$B012,a3
		btst	#3,($FFFFC2F8).w
		beq.w	loc_11A68
		move.w	#$BF52,a3

loc_11A68:							; CODE XREF: sub_11A48+18j
		bsr.w	sub_11952
		move.w	($FFFFB02C).w,d2
		move.w	(a1)+,d3				; load size of string
		subq.w	#$02,d3					; minus 2 (for the first word)
		bra.w	loc_11ADC				; continue
; ---------------------------------------------------------------------------

loc_11A78:							; CODE XREF: sub_11A48:loc_11ADCj
		move.b	(a1)+,d0				; load character
		ext.w	d0					; extend to word
		bgt.w	loc_11A94				; if it's from 01 to 7F, branch
		neg.w	d0
		asl.w	#2,d0
		move.l	#dword_11AF4,a2
		move.l	(a2,d0.w),a2
		jsr	(a2)
		bra.w	loc_11ADC
; ---------------------------------------------------------------------------

loc_11A94:							; CODE XREF: sub_11A48+34j
		cmpnb	 $40,d0					; is it an '@' symbol?
		bne.w	loc_11AA4				; if not, branch
		move.w	#$7FF,d0
		bra.w	loc_11AD4
; ---------------------------------------------------------------------------

loc_11AA4:							; CODE XREF: sub_11A48+50j
		cmpnb	 $5E,d0					; is it a '^' symbol?
		beq.w	loc_11AEA				; if so, branch
		asl.w	#$01,d0					; multiply character by 2
		move.l	#Art_BoardText,a2			; load uncompressed text art for the board
		btst	#3,($FFFFC2F8).w
		beq.w	loc_11AC4
		move.l	#off_BE26A,a2

loc_11AC4:							; CODE XREF: sub_11A48+72j
		add.l	4(a2),a2
		move.w	4(a2,d0.w),d0
		move.w	($FFFFB030).w,d1
		add.w	(a3,d1.w),d0

loc_11AD4:							; CODE XREF: sub_11A48+58j
		add.w	d2,d0
		move.w	d0,(a0)
		addq.w	#1,($FFFFB028).w

loc_11ADC:							; CODE XREF: sub_11A48+2Cj
		dbf	d3,loc_11A78				; repeat for all characters available
		movem.l	(sp)+,d0-d3/a0/a2-a3			; restore register data
		move.w	(sp)+,($FFFFBF78).w
		rts						; return

; ---------------------------------------------------------------------------

loc_11AEA:							; CODE XREF: sub_11A48+60j
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_11952
		bra.s	loc_11ADC
; End of function sub_11A48

; ---------------------------------------------------------------------------
dword_11AF4:	dc.l	$15464					; DATA XREF: sub_11A48+3Co
		dc.l	sub_11B28
		dc.l	sub_11B3C
		dc.l	sub_11B4C
		dc.l	loc_11B5A
		dc.l	loc_11B68
		dc.l	loc_11B76
		dc.l	loc_11B84
		dc.l	loc_11B18
; ---------------------------------------------------------------------------

loc_11B18:							; DATA XREF: ROM:00011B14o
		bsr.w	sub_11B3C
		bsr.w	sub_11B28
		bsr.w	sub_11B4C
		bra.w	loc_11B5A

; =============== S U B	R O U T	I N E =======================================


sub_11B28:							; CODE XREF: ROM:00011B1Cp
					; DATA XREF: ROM:00011AF8o
		move.b	(a1)+,d0
		subq.w	#1,d3
		and.w	#3,d0
		asl.w	#2,d0
		subq.w	#4,d0
		move.w	d0,($FFFFB02E).w
		bra.w	sub_11952
; End of function sub_11B28


; =============== S U B	R O U T	I N E =======================================


sub_11B3C:							; CODE XREF: ROM:loc_11B18p
					; DATA XREF: ROM:00011AFCo
		move.b	(a1)+,d2
		and.w	#7,d2
		subq.w	#1,d3
		ror.w	#3,d2
		move.w	d2,($FFFFB02C).w
		rts
; End of function sub_11B3C


; =============== S U B	R O U T	I N E =======================================


sub_11B4C:							; CODE XREF: ROM:00011B20p
					; DATA XREF: ROM:00011B00o
		clr.w	d0
		move.b	(a1)+,d0
		subq.w	#1,d3
		move.w	d0,($FFFFB028).w
		bra.w	sub_11952
; End of function sub_11B4C

; ---------------------------------------------------------------------------

loc_11B5A:							; CODE XREF: ROM:00011B24j
					; DATA XREF: ROM:00011B04o
		clr.w	d0
		move.b	(a1)+,d0
		subq.w	#1,d3
		move.w	d0,($FFFFB02A).w
		bra.w	sub_11952
; ---------------------------------------------------------------------------

loc_11B68:							; DATA XREF: ROM:00011B08o
		move.b	(a1)+,d0
		ext.w	d0
		subq.w	#1,d3
		add.w	d0,($FFFFB028).w
		bra.w	sub_11952
; ---------------------------------------------------------------------------

loc_11B76:							; DATA XREF: ROM:00011B0Co
		move.b	(a1)+,d0
		ext.w	d0
		subq.w	#1,d3
		add.w	d0,($FFFFB02A).w
		bra.w	sub_11952
; ---------------------------------------------------------------------------

loc_11B84:							; DATA XREF: ROM:00011B10o
		clr.w	d0
		move.b	(a1)+,d0
		subq.w	#1,d3
		asl.w	#1,d0
		move.w	d0,($FFFFB030).w
		rts

; =============== S U B	R O U T	I N E =======================================


sub_11B92:							; CODE XREF: ROM:000080DCp
					; ROM:loc_84CEp ...
		move.l	a1,-(sp)
		move.l	4(sp),a1
		bsr.w	sub_11BA4
		move.l	a1,4(sp)
		move.l	(sp)+,a1
		rts
; End of function sub_11B92


; =============== S U B	R O U T	I N E =======================================


sub_11BA4:							; CODE XREF: sub_8250+14p sub_8250+2Aj ...
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.l	d0-d3/a0/a2,-(sp)
		bsr.w	sub_11952
		move.w	($FFFFB02C).w,d2
		move.w	(a1)+,d3
		subq.w	#2,d3
		bra.w	loc_11C5A
; ---------------------------------------------------------------------------

loc_11BC2:							; CODE XREF: sub_11BA4:loc_11C5Aj
		move.b	(a1)+,d0
		beq.w	loc_11C5A
		ext.w	d0
		bpl.w	loc_11C04
		neg.w	d0
		move.w	d0,d2
		asl.w	#8,d2
		asl.w	#1,d2
		and.w	#$F800,d2
		move.w	d2,($FFFFB02C).w
		and.w	#3,d0
		asl.w	#2,d0
		subq.w	#4,d0
		move.w	d0,($FFFFB02E).w
		move.b	(a1)+,d0
		ext.w	d0
		move.w	d0,($FFFFB028).w
		move.b	(a1)+,d1
		ext.w	d1
		move.w	d1,($FFFFB02A).w
		bsr.w	sub_11952
		subq.w	#2,d3
		bra.w	loc_11C5A
; ---------------------------------------------------------------------------

loc_11C04:							; CODE XREF: sub_11BA4+26j
		cmpnb	 $40,d0					; '@'
		bne.w	loc_11C14
		move.w	#$7FF,d0
		bra.w	loc_11C52
; ---------------------------------------------------------------------------

loc_11C14:							; CODE XREF: sub_11BA4+64j
		cmpnb	 $5E,d0					; '^'
		beq.w	loc_11C68
		asl.w	#1,d0
		move.l	#Art_BoardText,a2
		btst	#3,($FFFFC2F8).w
		beq.w	loc_11C34
		move.l	#off_BE26A,a2

loc_11C34:							; CODE XREF: sub_11BA4+86j
		add.l	4(a2),a2
		move.w	4(a2,d0.w),d0
		btst	#3,($FFFFC2F8).w
		beq.w	loc_11C4E
		add.w	($FFFFBF52).w,d0
		bra.w	loc_11C52
; ---------------------------------------------------------------------------

loc_11C4E:							; CODE XREF: sub_11BA4+9Ej
		add.w	($FFFFB012).w,d0

loc_11C52:							; CODE XREF: sub_11BA4+6Cj
					; sub_11BA4+A6j
		add.w	d2,d0
		move.w	d0,(a0)
		addq.w	#1,($FFFFB028).w

loc_11C5A:							; CODE XREF: sub_11BA4+1Aj
					; sub_11BA4+20j ...
		dbf	d3,loc_11BC2
		movem.l	(sp)+,d0-d3/a0/a2
		move.w	(sp)+,($FFFFBF78).w
		rts
; ---------------------------------------------------------------------------

loc_11C68:							; CODE XREF: sub_11BA4+74j
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_11952
		bra.s	loc_11C5A
; End of function sub_11BA4


; =============== S U B	R O U T	I N E =======================================


sub_11C72:							; CODE XREF: sub_907C+24p sub_9316+24p
		swap	d0
		clr.w	d0
		rol.l	#2,d0
		move.l	#off_11C92,a1
		bsr.w	sub_13508
		addq.w	#2,($FFFFB028).w
		swap	d0
		lsr.w	#2,d0
		bsr.w	sub_11CA2
		bra.w	sub_11BA4
; End of function sub_11C72

; ---------------------------------------------------------------------------
off_11C92:	dc.l	unk_42031				; DATA XREF: sub_11C72+6o
		dc.l	unk_42032
		dc.l	unk_42033
		dc.l	unk_44F54

; =============== S U B	R O U T	I N E =======================================


sub_11CA2:							; CODE XREF: ROM:00008DCAp
					; sub_11C72+18p ...
		move.w	#$BFC2,a1
		move.l	d0,-(sp)
		move.l	a1,-(sp)
		ext.l	d0
		divu.w	#$A,d0
		swap	d0
		add.w	#$30,d0					; '0'
		move.b	d0,-(a1)
		swap	d0
		ext.l	d0
		divu.w	#6,d0
		swap	d0
		add.w	#$30,d0					; '0'
		move.b	d0,-(a1)
		swap	d0
		move.b	#$3A,-(a1)				; ':'
		ext.l	d0
		divu.w	#$A,d0
		swap	d0
		add.w	#$30,d0					; '0'
		move.b	d0,-(a1)
		swap	d0
		move.b	#$20,-(a1)				; ' '
		tst.w	d0
		beq.w	loc_11CEE
		add.w	#$30,d0					; '0'
		move.b	d0,(a1)

loc_11CEE:							; CODE XREF: sub_11CA2+42j
		move.l	(sp)+,d0
		sub.l	a1,d0
		addq.w	#2,d0
		btst	#0,d0
		beq.w	loc_11D00
		clr.b	-(a1)
		addq.w	#1,d0

loc_11D00:							; CODE XREF: sub_11CA2+56j
		move.w	d0,-(a1)
		move.l	(sp)+,d0
		rts
; End of function sub_11CA2


; =============== S U B	R O U T	I N E =======================================


sub_11D06:							; CODE XREF: ROM:00008E94p
					; sub_9316+6Ap	...
		move.w	#$C010,a1
		move.l	d0,-(sp)
		move.l	a1,-(sp)

loc_11D0E:							; CODE XREF: sub_11D06+1Aj
		ext.l	d0
		divu.w	#$A,d0
		swap	d0
		add.w	#$30,d0					; '0'
		move.b	d0,-(a1)
		swap	d0
		tst.w	d0
		bne.s	loc_11D0E
		move.l	(sp)+,d0
		sub.l	a1,d0
		addq.w	#2,d0
		btst	#0,d0
		beq.w	loc_11D34
		clr.b	-(a1)
		addq.w	#1,d0

loc_11D34:							; CODE XREF: sub_11D06+26j
		move.w	d0,-(a1)
		move.l	(sp)+,d0
		rts
; End of function sub_11D06


; =============== S U B	R O U T	I N E =======================================


sub_11D3A:							; CODE XREF: sub_8250+24p
					; ROM:00008E4Ep ...
		movem.l	d0-d3,-(sp)
		move.w	#$C00A,a1
		moveq	#1,d2
		sub.w	d2,d1
		bra.w	loc_11D4E
; ---------------------------------------------------------------------------

loc_11D4A:							; CODE XREF: sub_11D3A:loc_11D4Ej
		mulu.w	#$A,d2

loc_11D4E:							; CODE XREF: sub_11D3A+Cj
		dbf	d1,loc_11D4A
		moveq	#$20,d3					; ' '

loc_11D54:							; CODE XREF: sub_11D3A+3Cj
		ext.l	d0
		divu.w	d2,d0
		bne.w	loc_11D6A
		cmpnw	 1,d2
		beq.w	loc_11D6A
		move.w	d3,d0
		bra.w	loc_11D6E
; ---------------------------------------------------------------------------

loc_11D6A:							; CODE XREF: sub_11D3A+1Ej
					; sub_11D3A+26j
		moveq	#$30,d3					; '0'
		add.w	d3,d0

loc_11D6E:							; CODE XREF: sub_11D3A+2Cj
		move.b	d0,(a1)+
		swap	d0
		divu.w	#$A,d2
		bne.s	loc_11D54
		move.l	a1,d0
		sub.w	#$C008,d0
		btst	#0,d0
		beq.w	loc_11D8A
		clr.b	(a1)+
		addq.w	#1,d0

loc_11D8A:							; CODE XREF: sub_11D3A+48j
		move.w	#$C008,a1
		move.w	d0,(a1)
		movem.l	(sp)+,d0-d3
		rts
; End of function sub_11D3A


; =============== S U B	R O U T	I N E =======================================


sub_11D96:							; CODE XREF: sub_799E+B178p
					; sub_18A90+1Ep ...
		move.l	(sp)+,a1
		bsr.w	sub_11D9E
		jmp	(a1)
; End of function sub_11D96


; =============== S U B	R O U T	I N E =======================================


sub_11D9E:							; CODE XREF: sub_11D96+2p
					; sub_799E+B18Ap ...
		movem.l	d0/a0,-(sp)
		lea	2(a3),a0
		move.w	(a3),d0
		subq.w	#3,d0
		bmi.w	loc_11DB6

loc_11DAE:							; CODE XREF: sub_11D9E+14j
		addq.w	#1,a0
		tst.b	(a0)
		dbeq	d0,loc_11DAE

loc_11DB6:							; CODE XREF: sub_11D9E+Cj
		move.w	(a1)+,d0
		subq.w	#3,d0
		bmi.w	loc_11DDC

loc_11DBE:							; CODE XREF: sub_11D9E:loc_11DC6j
		move.b	(a1)+,(a0)+
		bne.w	loc_11DC6
		subq.w	#1,a0

loc_11DC6:							; CODE XREF: sub_11D9E+22j
		dbf	d0,loc_11DBE
		move.l	a0,d0
		btst	#0,d0
		beq.w	loc_11DD8
		clr.b	(a0)+
		addq.l	#1,d0

loc_11DD8:							; CODE XREF: sub_11D9E+32j
		sub.l	a3,d0
		move.w	d0,(a3)

loc_11DDC:							; CODE XREF: sub_11D9E+1Cj
		movem.l	(sp)+,d0/a0
		rts
; End of function sub_11D9E


; =============== S U B	R O U T	I N E =======================================


sub_11DE2:							; CODE XREF: ROM:000080F2p
					; ROM:00008748p ...
		move.l	a1,-(sp)
		move.l	4(sp),a1
		bsr.w	sub_11DF4
		move.l	a1,4(sp)
		move.l	(sp)+,a1
		rts
; End of function sub_11DE2


; =============== S U B	R O U T	I N E =======================================


sub_11DF4:							; CODE XREF: sub_11DE2+6p
					; sub_12D30+3Cj ...
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		movem.l	d0-a0/a2,-(sp)
		move.w	($FFFFB028).w,d4
		move.w	($FFFFB02A).w,d5
		move.w	($FFFFB02C).w,d6
		add.w	($FFFFB010).w,d6
		move.w	(a1)+,d3
		subq.w	#2,d3
		bra.w	loc_11E78
; ---------------------------------------------------------------------------

loc_11E1A:							; CODE XREF: sub_11DF4:loc_11E78j
		move.b	(a1)+,d0				; load text character
		beq.w	loc_11E78				; if it's the end (null terminal 0), branch
		ext.w	d0					; extend to word
		bpl.w	loc_11E5C				; if it's positive, branch for ASCII character
		neg.w	d0
		move.w	d0,d6
		asl.w	#8,d6
		asl.w	#1,d6
		and.w	#$F800,d6
		move.w	d6,($FFFFB02C).w
		add.w	($FFFFB010).w,d6
		and.w	#3,d0
		asl.w	#2,d0
		subq.w	#4,d0
		move.w	d0,($FFFFB02E).w
		move.b	(a1)+,d4
		ext.w	d4
		move.w	d4,($FFFFB028).w
		move.b	(a1)+,d5
		ext.w	d5
		move.w	d5,($FFFFB02A).w
		subq.w	#2,d3
		bra.w	loc_11E78
; ---------------------------------------------------------------------------

loc_11E5C:							; CODE XREF: sub_11DF4+2Ej
		cmpnb	 $61,d0					; 'a'   ; is it lower than lower case 'a'?
		blt.w	loc_11E70				; if so, branch
		cmpnb	 $7A,d0					; 'z'   ; is it higher than lower case 'z'?
		bgt.w	loc_11E70				; if so, branch
		add.b	#-$20,d0				; convert the character	to upper case 'A' to 'Z'

loc_11E70:							; CODE XREF: sub_11DF4+6Cj
					; sub_11DF4+74j
		move.w	d3,-(sp)				; store	counter
		bsr.w	sub_11E8E
		move.w	(sp)+,d3				; restore counter

loc_11E78:							; CODE XREF: sub_11DF4+22j
					; sub_11DF4+28j ...
		dbf	d3,loc_11E1A
		move.w	d4,($FFFFB028).w
		move.w	d5,($FFFFB02A).w
		movem.l	(sp)+,d0-a0/a2
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_11DF4


; =============== S U B	R O U T	I N E =======================================


sub_11E8E:							; CODE XREF: sub_11DF4+7Ep
		sub.w	#$20,d0					; ' '   ; subtract starting ASCII character (space)
		move.l	#unk_1916A,a0
		moveq	#1,d2
		move.b	(a0,d0.w),d1
		ext.w	d1
		bpl.w	loc_11EA8
		neg.w	d1
		clr.w	d2

loc_11EA8:							; CODE XREF: sub_11E8E+12j
		asl.w	#1,d1
		move.l	#off_A9A10,a0
		add.l	4(a0),a0

loc_11EB4:							; CODE XREF: sub_11E8E+46j
		move.w	4(a0,d1.w),d3
		bsr.w	sub_11EDA
		move.w	(a0),d7
		asl.w	#1,d7
		add.w	d7,d1
		move.w	4(a0,d1.w),d3
		sub.w	d7,d1
		addq.w	#1,d5
		bsr.w	sub_11EDA
		subq.w	#1,d5
		addq.w	#1,d4
		addq.w	#2,d1
		dbf	d2,loc_11EB4
		rts
; End of function sub_11E8E


; =============== S U B	R O U T	I N E =======================================


sub_11EDA:							; CODE XREF: sub_11E8E+2Ap
					; sub_11E8E+3Cp
		add.w	d6,d3
		movem.l	d1/a0,-(sp)
		move.w	d5,d0
		move.l	#$FFFFB004,a0
		add.w	($FFFFB02E).w,a0
		move.w	2(a0),d1
		asl.w	d1,d0
		add.w	d4,d0
		asl.w	#1,d0
		add.w	(a0),d0
		bsr.w	VDP_SetAddress
		move.w	d3,(a0)
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_11EDA


; =============== S U B	R O U T	I N E =======================================


sub_11F04:							; CODE XREF: ROM:00009C40p
					; ROM:000173BAp ...
		move.w	d4,($FFFFB012).w
		move.l	#unk_AAC5A,a2
		bra.w	sub_11738
; End of function sub_11F04


; =============== S U B	R O U T	I N E =======================================


sub_11F12:							; CODE XREF: ROM:00009D56p
					; ROM:00016ACAp
		move.l	#unk_55B86,a2
		move.w	d4,($FFFFB01E).w
		bra.w	sub_11738
; End of function sub_11F12


; =============== S U B	R O U T	I N E =======================================


sub_11F20:							; CODE XREF: ROM:000173B6p
		moveq	#2,d4
		move.l	#unk_ABA1C,a2
		bra.w	sub_11738
; End of function sub_11F20


; =============== S U B	R O U T	I N E =======================================


sub_11F2C:							; CODE XREF: sub_100E6+50p
					; sub_1013E+B4p ...
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		cmpnw	 $C,d0
		beq.w	sub_11F62
		tst.w	($FFFFD056).w
		beq.w	locret_15464
		btst	#2,$63(a3)
		bne.w	locret_15464
		btst	#5,($FFFFC2EA).w
		bne.w	sub_11F62
		cmpnw	 $10,d0
		beq.w	locret_15464
; End of function sub_11F2C


; =============== S U B	R O U T	I N E =======================================


sub_11F62:							; CODE XREF: sub_7B30+108j
					; sub_7B30+122j ...
		btst	#7,($FFFFC2EC).w
		bne.w	locret_15464
		movem.l	d1/a0-a1,-(sp)
		cmpnw	 $E,d0
		blt.w	loc_11F9A
		add.w	#$C8,($FFFFB89C).w			; ''
		move.w	#$C,-(sp)
		btst	#6,$62(a3)
		beq.w	loc_11F96
		add.w	#$14,($FFFFB8A2).w
		move.w	#$B,(sp)

loc_11F96:							; CODE XREF: sub_11F62+26j
		bsr.w	sub_11156

loc_11F9A:							; CODE XREF: sub_11F62+12j
		move.w	#$C3A4,a1
		moveq	#$1F,d1

loc_11FA0:							; CODE XREF: sub_11F62+40j
		tst.w	(a1)+
		dbeq	d1,loc_11FA0
		bne.w	loc_11FEC
		move.b	$53(a3),-(a1)
		move.b	d0,-(a1)
		move.l	#word_18E0C,a0
		add.w	(a0,d0.w),a0
		tst.b	1(a0)
		beq.w	loc_11FEC
		bmi.w	loc_11FEC
		btst	#6,$62(a3)
		bne.w	loc_11FDA
		bset	#1,($FFFFC300).w
		bra.w	loc_11FE0
; ---------------------------------------------------------------------------

loc_11FDA:							; CODE XREF: sub_11F62+6Aj
		bset	#2,($FFFFC300).w

loc_11FE0:							; CODE XREF: sub_11F62+74j
		bset	#4,$63(a3)
		beq.w	loc_11FEC
		clr.w	(a1)

loc_11FEC:							; CODE XREF: sub_11F62+44j
					; sub_11F62+5Cj ...
		movem.l	(sp)+,d1/a0-a1
		rts
; End of function sub_11F62


; =============== S U B	R O U T	I N E =======================================


sub_11FF2:							; CODE XREF: sub_799Ep
		bsr.w	sub_128B2
		bsr.w	sub_12418
		bsr.w	sub_12002
		bra.w	sub_125E8
; End of function sub_11FF2


; =============== S U B	R O U T	I N E =======================================


sub_12002:							; CODE XREF: sub_11FF2+8p

; FUNCTION CHUNK AT 0001221C SIZE 000001C2 BYTES

		btst	#2,($FFFFC2F2).w
		beq.w	loc_12074
		btst	#7,($FFFFC2F2).w
		bne.w	loc_12066
		tst.w	($FFFFC31A).w
		bmi.w	loc_12074
		subq.w	#1,($FFFFC31A).w
		bpl.w	locret_12072
		movem.l	d0-a6,-(sp)
		bclr	#2,($FFFFC2EE).w
		bclr	#7,($FFFFC2FA).w
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	3
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1B,d0
		moveq	#8,d1
		btst	#0,($FFFFC2FA).w
		beq.w	loc_12052
		move.w	#$C,d1

loc_12052:							; CODE XREF: sub_12002+48j
		move.l	#$7FF,d2
		jsr	sub_1197E
		movem.l	(sp)+,d0-a6
		bra.w	locret_12072
; ---------------------------------------------------------------------------

loc_12066:							; CODE XREF: sub_12002+10j
		jsr	loc_187B8
		bclr	#7,($FFFFC2F2).w

locret_12072:							; CODE XREF: sub_12002+20j
					; sub_12002+60j
		rts
; ---------------------------------------------------------------------------

loc_12074:							; CODE XREF: sub_12002+6j
					; sub_12002+18j
		btst	#2,($FFFFC2EA).w
		beq.w	locret_15464
		tst.w	($FFFFC3E6).w
		bmi.w	loc_1221C
		sub.w	d7,($FFFFC3E6).w
		bpl.w	locret_15464
		bclr	#3,($FFFFC2EA).w
		move.w	#$C3A4,a0

loc_12098:							; CODE XREF: sub_12002+B0j
					; sub_12002+B2j
		tst.w	(a0)+
		beq.w	locret_15464
		clr.w	d0
		move.b	-2(a0),d0
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		tst.b	1(a1)
		beq.s	loc_12098
		bmi.s	loc_12098
		bset	#2,($FFFFC2EE).w
		move.w	($FFFFB03E).w,-(sp)
		bsr.w	sub_10F32
		move.w	(sp)+,($FFFFB03E).w
		bsr.w	sub_12B94
		move.w	($FFFFB026).w,d4
		move.l	#unk_5CF6C,a2
		bsr.w	sub_11738
		bsr.w	sub_13378
		bsr.w	sub_133B2
		st	($FFFFB7AA).w
		move.w	#$B74A,a3
		move.l	#$1A,d0
		bsr.w	sub_10658
		move.w	#$1C20,$40(a3)
		st	($FFFFC3F2).w
		clr.w	($FFFFC3F0).w
		bsr.w	sub_125E8
		move.w	#$32,($FFFFC3F0).w			; '2'
		clr.w	d0
		bsr.w	sub_126F8
		move.w	#$18,($FFFFBD26).w
		rts
; End of function sub_12002


; =============== S U B	R O U T	I N E =======================================


sub_1211A:							; CODE XREF: sub_13CD8+102p
					; sub_13CD8+130p ...
		movem.l	d1-d3/a0,-(sp)
		btst	#1,$64(a2)
		beq.w	loc_12132
		btst	#3,($FFFFC2F2).w
		beq.w	loc_1213A

loc_12132:							; CODE XREF: sub_1211A+Aj
					; sub_1211A+4Cj
		move.w	#$FFFF,d1
		bra.w	loc_1216E
; ---------------------------------------------------------------------------

loc_1213A:							; CODE XREF: sub_1211A+14j
		move.l	#unk_1913A,a0
		move.w	(a0,d0.w),d1
		bmi.w	loc_1216E
		move.w	$52(a2),d2
		movem.w	d0-d1,-(sp)
		move.w	d2,d0
		jsr	sub_12174
		movem.w	(sp)+,d0-d1
		bpl.w	loc_12168
		bclr	#3,($FFFFC2F2).w
		bra.s	loc_12132
; ---------------------------------------------------------------------------

loc_12168:							; CODE XREF: sub_1211A+42j
		move.w	d1,($FFFFD410).w
		move.w	d1,d0

loc_1216E:							; CODE XREF: sub_1211A+1Cj
					; sub_1211A+2Aj
		movem.l	(sp)+,d1-d3/a0
		rts
; End of function sub_1211A


; =============== S U B	R O U T	I N E =======================================


sub_12174:							; CODE XREF: sub_1211A+38p
		movem.l	d1-d3/a1,-(sp)
		tst.w	($FFFFD056).w
		beq.w	loc_12216
		bset	#3,($FFFFC2F2).w
		movem.w	d0,-(sp)
		move.w	#5,d1
		move.w	#0,d2
		cmpnw	 5,d0
		bgt.w	loc_121A2
		move.w	#$B,d1
		move.w	#1,d2

loc_121A2:							; CODE XREF: sub_12174+22j
		move.w	d1,d0
		jsr	sub_B86A
		tst.w	d0
		bpl.w	loc_121BE
		movem.w	(sp)+,d0
		bclr	#3,($FFFFC2F2).w
		bra.w	loc_12210
; ---------------------------------------------------------------------------

loc_121BE:							; CODE XREF: sub_12174+38j
		move.w	d0,d1
		move.w	(sp)+,d0
		move.w	d0,($FFFFD406).w
		move.l	#$FFFFB04A,a1
		move.w	d0,d3
		asl.w	#7,d3
		move.w	#0,($FFFFD404).w
		btst	#6,$62(a1,d3.w)
		beq.w	loc_121E6
		move.w	#1,($FFFFD404).w

loc_121E6:							; CODE XREF: sub_12174+68j
		move.w	#0,($FFFFD40A).w
		move.b	$66(a1,d3.w),($FFFFD40B).w
		move.w	d1,($FFFFD408).w
		asl.w	#7,d1
		move.w	#0,($FFFFD40C).w
		move.b	$66(a1,d1.w),($FFFFD40D).w
		move.w	#0,($FFFFD40E).w
		move.b	$66(a3),($FFFFD40F).w

loc_12210:							; CODE XREF: sub_12174+46j
					; sub_12174+A6j
		movem.l	(sp)+,d1-d3/a1
		rts
; ---------------------------------------------------------------------------

loc_12216:							; CODE XREF: sub_12174+8j
		move.w	#$FFFF,d1
		bra.s	loc_12210
; End of function sub_12174

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_12002

loc_1221C:							; CODE XREF: sub_12002+80j
		tst.w	($FFFFC3F0).w
		bpl.w	locret_15464
		tst.w	($FFFFC3F2).w
		bpl.w	locret_15464
		movem.l	d0/a0-a3,-(sp)

loc_12230:							; CODE XREF: sub_12002+394j
		move.w	#$C3A4,a0
		tst.w	(a0)+
		beq.w	loc_123A2

loc_1223A:							; CODE XREF: sub_12002+23Aj
		tst.w	(a0)+
		bne.s	loc_1223A
		subq.w	#4,a0
		bclr	#7,1(a0)
		bne.w	loc_1226A
		move.l	a3,-(sp)
		clr.w	d1
		move.b	1(a0),d1
		asl.w	#7,d1
		move.w	#$B04A,a3
		add.w	d1,a3
		tst.w	$34(a3)
		move.l	(sp)+,a3
		bpl.w	loc_123D8
		clr.w	(a0)
		bra.w	loc_123D8
; ---------------------------------------------------------------------------

loc_1226A:							; CODE XREF: sub_12002+244j
		clr.w	d0
		move.b	(a0),d0
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		bclr	#5,($FFFFC2F6).w
		clr.w	d2
		move.b	1(a1),d2
		beq.w	loc_1238E
		bmi.w	loc_1238E
		cmpnb	 5,d2
		bne.w	loc_1229A
		bset	#5,($FFFFC2F6).w

loc_1229A:							; CODE XREF: sub_12002+28Ej
		movem.l	d0-d1/a1-a4,-(sp)
		bsr.w	sub_14A94
		move.w	#$C5DE,a4
		add.w	($FFFFC5DC).w,a4
		cmp.w	#$EC,($FFFFC5DC).w			; ''
		beq.w	loc_122B8
		addq.w	#4,($FFFFC5DC).w

loc_122B8:							; CODE XREF: sub_12002+2AEj
		move.w	d0,(a4)+
		move.b	(a0),(a4)+
		clr.w	d1
		move.b	1(a0),d1
		asl.w	#7,d1
		move.w	#$B04A,a3
		add.w	d1,a3
		clr.w	d0
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		btst	#6,$62(a3)
		beq.w	loc_122EA
		bset	#7,-1(a4)
		move.w	#$8000,d0
		exg	a1,a2

loc_122EA:							; CODE XREF: sub_12002+2D8j
		addq.w	#1,6(a2)
		add.w	d2,8(a2)
		move.b	$66(a3),d0
		move.b	d0,(a4)
		move.w	d0,($FFFFC470).w
		ext.w	d0
		add.w	#$102,d0
		add.b	d2,(a2,d0.w)
		sub.w	#$102,d0
		asl.w	#1,d0
		ext.w	d2
		mulu.w	#$3C,d2					; '<'
		bset	#$D,d2
		tst.w	$66(a2,d0.w)
		bmi.w	loc_1232C
		btst	#4,$66(a2,d0.w)
		beq.w	loc_1232C
		bset	#$C,d2

loc_1232C:							; CODE XREF: sub_12002+318j
					; sub_12002+322j
		move.w	d2,$66(a2,d0.w)
		and.w	#$EFFF,d2
		moveq	#$34,d1					; '4'

loc_12336:							; CODE XREF: sub_12002+344j
					; sub_12002+34Cj
		subq.w	#2,d1
		bmi.w	loc_1235C
		move.w	$66(a1,d1.w),d3
		and.w	#$EFFF,d3
		cmp.w	d3,d2
		bne.s	loc_12336
		btst	#5,($FFFFC2F6).w
		beq.s	loc_12336
		bset	#6,$66(a1,d1.w)
		bset	#6,$66(a2,d0.w)

loc_1235C:							; CODE XREF: sub_12002+336j
		movem.l	a0,-(sp)
		lea	$9A(a2),a0
		moveq	#$18,d1

loc_12366:							; CODE XREF: sub_12002+366j
		tst.b	(a0)+
		dbmi	d1,loc_12366
		move.b	d0,-1(a0)
		st	(a0)
		movem.l	(sp)+,a0
		move.w	#$C,d0
		bsr.w	sub_10662
		movem.l	(sp)+,d0-d1/a1-a4
		bsr.w	sub_12638
		bsr.w	loc_12DA6
		bra.w	loc_123D8
; ---------------------------------------------------------------------------

loc_1238E:							; CODE XREF: sub_12002+282j
					; sub_12002+286j
		clr.w	(a0)
		btst	#7,($FFFFC2EC).w
		bne.w	loc_12230
		bsr.w	sub_12638
		bra.w	loc_123D8
; ---------------------------------------------------------------------------

loc_123A2:							; CODE XREF: sub_12002+234j
		move.w	#$C6CE,a2
		bsr.w	sub_123DE
		add.w	#$364,a2
		bsr.w	sub_123DE
		bclr	#2,($FFFFC2EA).w
		move.w	#$B74A,a3
		bclr	#6,($FFFFC2F2).w
		move.w	#$1B,d0
		btst	#3,($FFFFC2F2).w
		beq.w	loc_123D4
		move.w	#$1E,d0

loc_123D4:							; CODE XREF: sub_12002+3CAj
		bsr.w	sub_10662

loc_123D8:							; CODE XREF: sub_12002+25Ej
					; sub_12002+264j ...
		movem.l	(sp)+,d0/a0-a3
		rts
; END OF FUNCTION CHUNK	FOR sub_12002

; =============== S U B	R O U T	I N E =======================================


sub_123DE:							; CODE XREF: sub_12002+3A4p
					; sub_12002+3ACp
		moveq	#6,d1
		moveq	#$32,d0					; '2'

loc_123E2:							; CODE XREF: sub_123DE+32j
		tst.w	$66(a2,d0.w)
		ble.w	loc_1240E
		bclr	#5,$66(a2,d0.w)
		btst	#6,$66(a2,d0.w)
		bne.w	loc_1240E
		btst	#4,$66(a2,d0.w)
		bne.w	loc_1240E
		cmpnw	 4,d1
		beq.w	loc_1240E
		subq.w	#1,d1

loc_1240E:							; CODE XREF: sub_123DE+8j
					; sub_123DE+18j ...
		subq.w	#2,d0
		bpl.s	loc_123E2
		move.w	d1,$24(a2)
		rts
; End of function sub_123DE


; =============== S U B	R O U T	I N E =======================================


sub_12418:							; CODE XREF: sub_11FF2+4p
		btst	#7,($FFFFC2EC).w
		bne.w	locret_15464
		move.w	#$C3A4,a0

loc_12426:							; CODE XREF: sub_12418+58j
					; sub_12418+74j ...
		tst.w	(a0)+
		beq.w	locret_15464
		btst	#2,($FFFFC2EA).w
		bne.w	loc_1246A
		clr.w	d0
		move.b	-2(a0),d0
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		tst.b	1(a1)
		beq.w	loc_12466
		move.w	($FFFFB7AA).w,d0
		bmi.w	loc_12494
		subq.w	#6,d0
		move.b	-1(a0),d1
		ext.w	d1
		subq.w	#6,d1
		eor.w	d1,d0
		bmi.w	loc_12494

loc_12466:							; CODE XREF: sub_12418+32j
		bsr.w	sub_124A6

loc_1246A:							; CODE XREF: sub_12418+1Aj
		bset	#7,-1(a0)
		bne.s	loc_12426
		clr.w	d0
		move.b	-2(a0),d0
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		clr.w	d1
		move.b	(a1),d1
		asl.w	#5,d1
		cmp.w	($FFFFC3E6).w,d1
		ble.s	loc_12426
		move.w	d1,($FFFFC3E6).w
		bra.s	loc_12426
; ---------------------------------------------------------------------------

loc_12494:							; CODE XREF: sub_12418+3Aj
					; sub_12418+4Aj
		bset	#3,($FFFFC2EA).w
		bne.s	loc_12426
		move.w	#$2C,d0					; ','
		bsr.w	sub_12638
		bra.s	loc_12426
; End of function sub_12418


; =============== S U B	R O U T	I N E =======================================


sub_124A6:							; CODE XREF: ROM:0000ED32j
					; sub_F3E4+BCj	...
		bset	#0,($FFFFC2EA).w
		bne.w	loc_1255E
		clr.w	d0
		clr.w	d1
		cmp.b	#$E,-2(a0)
		beq.w	loc_12504
		move.w	($FFFFBF7C).w,d0
		move.w	($FFFFBF7E).w,d1
		cmp.b	#6,-2(a0)
		beq.w	loc_12504
		move.w	($FFFFB74A).w,d0
		move.w	($FFFFB75E).w,d1
		cmp.b	#$C,-2(a0)
		bne.w	loc_12504
		move.l	a3,-(sp)
		move.w	#$B04A,a3
		move.b	-1(a0),d1
		ext.w	d1
		asl.w	#7,d1
		add.w	d1,a3
		move.w	#$258,d1
		btst	#7,$62(a3)
		move.l	(sp)+,a3
		beq.w	loc_12504
		neg.w	d1

loc_12504:							; CODE XREF: sub_124A6+14j
					; sub_124A6+26j ...
		movem.w	d0-d1,-(sp)
		cmpnw	 $46,d0					; 'F'
		blt.w	loc_12514
		move.w	#$46,d0					; 'F'

loc_12514:							; CODE XREF: sub_124A6+66j
		cmpnw	 $FFBA,d0
		bgt.w	loc_12520
		move.w	#$FFBA,d0

loc_12520:							; CODE XREF: sub_124A6+72j
		cmpnw	 $A6,d1					; ''
		blt.w	loc_12538
		move.w	#$CE,d1					; ''
		move.w	#$46,d0					; 'F'
		tst.w	(sp)
		bpl.w	loc_12538
		neg.w	d0

loc_12538:							; CODE XREF: sub_124A6+7Ej
					; sub_124A6+8Cj
		cmpnw	 $FF5A,d1
		bgt.w	loc_12550
		move.w	#$FF32,d1
		move.w	#$46,d0					; 'F'
		tst.w	(sp)
		bpl.w	loc_12550
		neg.w	d0

loc_12550:							; CODE XREF: sub_124A6+96j
					; sub_124A6+A4j
		move.w	d0,($FFFFBF94).w
		move.w	d1,($FFFFBF96).w
		addq.w	#4,sp
		bsr.w	sub_1257E

loc_1255E:							; CODE XREF: sub_124A6+6j
		clr.w	($FFFFC3E6).w
		bset	#2,($FFFFC2EA).w
		jsr	sub_1A304
		move.w	#3,-(sp)
		bsr.w	sub_11132
		move.w	#$A,d0
		bra.w	sub_12638
; End of function sub_124A6


; =============== S U B	R O U T	I N E =======================================


sub_1257E:							; CODE XREF: sub_124A6+B4p
		movem.l	d0-d1/a0-a1,-(sp)
		move.w	#$C3A4,a0

loc_12586:							; CODE XREF: sub_1257E+10j
		bsr.w	sub_12596
		addq.w	#2,a0
		tst.w	(a0)
		bne.s	loc_12586
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_1257E


; =============== S U B	R O U T	I N E =======================================


sub_12596:							; CODE XREF: sub_1257E:loc_12586p
		move.b	1(a0),d0
		and.w	#$7F,d0					; ''
		asl.w	#7,d0
		move.w	#$B04A,a1
		move.w	#$58,d1					; 'X'
		btst	#7,$62(a1,d0.w)
		bne.w	loc_125C6
		neg.w	d1
		cmp.w	($FFFFBF96).w,d1
		blt.w	locret_15464
		move.w	#$FFBF,($FFFFBF96).w
		bra.w	loc_125D4
; ---------------------------------------------------------------------------

loc_125C6:							; CODE XREF: sub_12596+18j
		cmp.w	($FFFFBF96).w,d1
		bgt.w	locret_15464
		move.w	#$41,($FFFFBF96).w			; 'A'

loc_125D4:							; CODE XREF: sub_12596+2Cj
		move.w	#$46,d0					; 'F'
		tst.w	($FFFFBF94).w
		bpl.w	loc_125E2
		neg.w	d0

loc_125E2:							; CODE XREF: sub_12596+46j
		move.w	d0,($FFFFBF94).w
		rts
; End of function sub_12596


; =============== S U B	R O U T	I N E =======================================


sub_125E8:							; CODE XREF: sub_11FF2+Cj
					; sub_12002+100p
		tst.w	($FFFFC3F0).w
		bmi.w	locret_15464
		sub.w	d7,($FFFFC3F0).w
		bpl.w	loc_125FC
		bsr.w	sub_1268A

loc_125FC:							; CODE XREF: sub_125E8+Cj
		cmp.w	#4,($FFFFC3F4).l
		bne.w	loc_1260C
		bsr.w	loc_1850A

loc_1260C:							; CODE XREF: sub_125E8+1Cj
		sub.w	d7,($FFFFC3EE).w
		bpl.w	locret_15464
		move.w	#$7FFF,($FFFFC3EE).w
		bsr.w	sub_12828
		bsr.w	sub_11B92
; End of function sub_125E8

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$11
		dc.b	$B
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_18A6E
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bra.w	sub_11BA4

; =============== S U B	R O U T	I N E =======================================


sub_12638:							; CODE XREF: sub_12002+380p
					; sub_12002+398p ...
		move.w	d0,($FFFFC3F4).w
		cmpnw	 $2E,d0					; '.'
		blt.w	loc_12646
		rts
; ---------------------------------------------------------------------------

loc_12646:							; CODE XREF: sub_12638+8j
		clr.w	($FFFFC3F2).w
		bsr.w	sub_1277A
		cmpnw	 $E,d0
		bne.w	loc_1265A
		bsr.w	loc_1889A

loc_1265A:							; CODE XREF: sub_12638+1Aj
		btst	#0,($FFFFC2FA).w
		beq.w	loc_12674
		btst	#3,($FFFFC2FA).w
		beq.w	loc_12674
		jsr	loc_FC5AE

loc_12674:							; CODE XREF: sub_12638+28j
					; sub_12638+32j
		move.w	#$7FFF,($FFFFC3EE).w
		btst	#7,($FFFFC2EC).w
		beq.w	sub_1268A
		move.w	#$3C,($FFFFC3EE).w			; '<'
; End of function sub_12638


; =============== S U B	R O U T	I N E =======================================


sub_1268A:							; CODE XREF: sub_125E8+10p
					; sub_12638+48j
		movem.l	d0-d2/a0-a1,-(sp)
		moveq	#$40,d0					; '@'
		tst.w	($FFFFC3F2).w
		bmi.w	loc_126EE
		move.w	($FFFFC3F2).w,d0
		addq.w	#2,($FFFFC3F2).w
		move.w	($FFFFC3F4).w,d1
		move.l	#word_18E0C,a0
		add.w	(a0,d1.w),a0
		addq.w	#2,a0
		add.w	(a0),a0
		move.w	(a0,d0.w),d0
		bpl.w	loc_126C0
		neg.w	d0
		st	($FFFFC3F2).w

loc_126C0:							; CODE XREF: sub_1268A+2Cj
		clr.w	d1
		move.b	d0,d1
		asl.w	#3,d1
		move.w	d1,($FFFFC3F0).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_126EC
		btst	#3,($FFFFC2FA).w
		beq.w	loc_126EC
		move.w	d0,-(sp)
		move.w	($FFFFC3F0).w,d0
		add.w	d0,d0
		move.w	d0,($FFFFC3F0).w
		move.w	(sp)+,d0

loc_126EC:							; CODE XREF: sub_1268A+46j
					; sub_1268A+50j
		lsr.w	#8,d0

loc_126EE:							; CODE XREF: sub_1268A+Aj
		bsr.w	sub_126F8
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_1268A


; =============== S U B	R O U T	I N E =======================================


sub_126F8:							; CODE XREF: sub_E62E+7Cp
					; sub_12002+10Cp ...
		movem.l	d0-d2/a0-a1,-(sp)
		cmpnw	 $40,d0					; '@'
		beq.w	loc_12750
		mulu.w	#$70,d0					; 'p'
		move.l	#unk_5C408,a0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_1271E
		move.l	#unk_5CF64,a0

loc_1271E:							; CODE XREF: sub_126F8+1Cj
		add.l	4(a0),a0
		addq.w	#4,a0
		add.w	d0,a0
		move.w	#$C3F6,a1
		move.w	($FFFFB026).w,d2
		btst	#7,($FFFFC2EC).w
		bne.w	loc_1273C
		or.w	#$8000,d2

loc_1273C:							; CODE XREF: sub_126F8+3Cj
		moveq	#$37,d0					; '7'

loc_1273E:							; CODE XREF: sub_126F8+4Aj
		move.w	(a0)+,(a1)
		add.w	d2,(a1)+
		dbf	d0,loc_1273E
		bset	#1,($FFFFC2EE).w
		bra.w	loc_12774
; ---------------------------------------------------------------------------

loc_12750:							; CODE XREF: sub_126F8+8j
		btst	#7,($FFFFC2EC).w
		bne.w	loc_1276E
		move.w	#$C3F6,a1
		moveq	#$37,d0					; '7'

loc_12760:							; CODE XREF: sub_126F8+6Cj
		move.w	#$7FF,(a1)+
		dbf	d0,loc_12760
		bset	#1,($FFFFC2EE).w

loc_1276E:							; CODE XREF: sub_126F8+5Ej
		moveq	#$FFFFFFFF,d0
		bsr.w	sub_1277A

loc_12774:							; CODE XREF: sub_126F8+54j
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_126F8


; =============== S U B	R O U T	I N E =======================================


sub_1277A:							; CODE XREF: sub_12638+12p
					; sub_126F8+78p ...
		movem.l	d0-d2/a1,-(sp)
		btst	#7,($FFFFC2EC).w
		bne.w	loc_127F6
		tst.w	d0
		bpl.w	loc_127A8
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	0
		dc.b	$A
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$D,d0
		moveq	#3,d1
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		bra.w	loc_12822
; ---------------------------------------------------------------------------

loc_127A8:							; CODE XREF: sub_1277A+10j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	5
		dc.b	$A
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		addq.w	#2,a1
		move.w	(a1),d0
		subq.w	#2,d0
		beq.w	loc_12822
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bpl.w	loc_127D4
		clr.w	($FFFFB028).w

loc_127D4:							; CODE XREF: sub_1277A+52j
		move.w	(a1),d0
		tst.b	-1(a1,d0.w)
		bne.w	loc_127E0
		subq.w	#1,d0

loc_127E0:							; CODE XREF: sub_1277A+60j
		moveq	#3,d1
		bsr.w	sub_119B8
		subq.w	#2,($FFFFB02A).w
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_11BA4
		bra.w	loc_12822
; ---------------------------------------------------------------------------

loc_127F6:							; CODE XREF: sub_1277A+Aj
		bsr.w	sub_12828
		tst.w	d0
		bmi.w	loc_12822
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$11
		dc.b	$B
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		addq.w	#2,a1
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bsr.w	sub_11BA4

loc_12822:							; CODE XREF: sub_1277A+2Aj
					; sub_1277A+48j ...
		movem.l	(sp)+,d0-d2/a1
		rts
; End of function sub_1277A


; =============== S U B	R O U T	I N E =======================================


sub_12828:							; CODE XREF: sub_125E8+32p
					; sub_1277A:loc_127F6p
		bsr.w	sub_11B92
; End of function sub_12828

; ---------------------------------------------------------------------------
		dc.w	$1C
		dc.b	$BF
		dc.b	5
		dc.b	$B
		dc.b	'                      ',0
; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_1284A:							; CODE XREF: sub_14620+39Ep
		movem.l	d0-d2/a0,-(sp)
		cmp.w	#6,$24(a1)
		bne.w	loc_1285C
		bsr.w	sub_128A4

loc_1285C:							; CODE XREF: sub_1284A+Aj
		move.w	$24(a2),d2
		cmp.w	$24(a1),d2
		ble.w	loc_1289E
		lea	$9A(a1),a0

loc_1286C:							; CODE XREF: sub_1284A+30j
		clr.w	d2
		move.b	(a0)+,d2
		bmi.w	loc_1289E
		btst	#6,$66(a1,d2.w)
		bne.s	loc_1286C
		clr.w	$66(a1,d2.w)
		bsr.w	sub_1298C
		bset	#0,($FFFFC300).w
		addq.w	#1,$24(a1)
		addq.w	#1,2(a2)
		bset	#0,($FFFFC6FE).w
		bset	#0,($FFFFCA62).w

loc_1289E:							; CODE XREF: sub_1284A+1Aj
					; sub_1284A+26j
		movem.l	(sp)+,d0-d2/a0
		rts
; End of function sub_1284A


; =============== S U B	R O U T	I N E =======================================


sub_128A4:							; CODE XREF: sub_7B30:loc_7C16p
					; sub_1284A+Ep
		moveq	#$1F,d0
		move.w	#$C3A4,a0

loc_128AA:							; CODE XREF: sub_128A4+8j
		clr.w	(a0)+
		dbf	d0,loc_128AA
		rts
; End of function sub_128A4


; =============== S U B	R O U T	I N E =======================================


sub_128B2:							; CODE XREF: sub_11FF2p
		bclr	#6,($FFFFC2F0).w
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		sub.w	d7,($FFFFC3E8).w
		bpl.w	locret_15464
		add.w	#$18,($FFFFC3E8).w
		bset	#6,($FFFFC2F0).w
		bsr.w	sub_1299C
		jsr	sub_FE14C
		move.w	#$C6CE,a2
		bsr.w	sub_128EC
		add.w	#$364,a2
; End of function sub_128B2


; =============== S U B	R O U T	I N E =======================================


sub_128EC:							; CODE XREF: sub_128B2+32p

; FUNCTION CHUNK AT 00012968 SIZE 00000024 BYTES

		lea	$9A(a2),a0
		move.w	#$BFA4,a1
		moveq	#2,d1
		moveq	#2,d3

loc_128F8:							; CODE XREF: sub_128EC+20j
					; sub_128EC:loc_1291Cj	...
		clr.w	d0
		move.b	(a0)+,d0
		bmi.w	loc_1291E
		btst	#6,$66(a2,d0.w)
		bne.w	loc_12968
		subq.w	#1,d1
		bmi.s	loc_128F8
		move.w	d0,(a1)+
		subq.w	#1,$66(a2,d0.w)
		bne.w	loc_1291C
		bsr.w	sub_1298C

loc_1291C:							; CODE XREF: sub_128EC+28j
		bra.s	loc_128F8
; ---------------------------------------------------------------------------

loc_1291E:							; CODE XREF: sub_128EC+10j
		move.w	($FFFFBFA4).w,d0
		cmpnw	 1,d1
		beq.w	sub_12944
		tst.w	d1
		bne.w	loc_12938
		bsr.w	sub_12944
		bra.w	loc_12940
; ---------------------------------------------------------------------------

loc_12938:							; CODE XREF: sub_128EC+40j
		cmpnw	 $FFFF,d1
		bne.w	locret_15464

loc_12940:							; CODE XREF: sub_128EC+48j
		move.w	($FFFFBFA6).w,d0
; End of function sub_128EC


; =============== S U B	R O U T	I N E =======================================


sub_12944:							; CODE XREF: sub_128EC+3Aj
					; sub_128EC+44p
		cmp.w	#5,$66(a2,d0.w)
		bgt.w	locret_15464
		move.w	#1,-(sp)
		tst.w	$66(a2,d0.w)
		bne.w	loc_12962
		bsr.w	sub_129D2
		move.w	#2,(sp)

loc_12962:							; CODE XREF: sub_12944+12j
		bsr.w	sub_11132
		rts
; End of function sub_12944

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_128EC

loc_12968:							; CODE XREF: sub_128EC+1Aj
		subq.w	#1,$66(a2,d0.w)
		btst	#3,$66(a2,d0.w)
		beq.s	loc_128F8
		btst	#4,$66(a2,d0.w)
		bne.w	loc_12988
		move.w	#$1000,$66(a2,d0.w)
		bra.w	sub_1298C
; ---------------------------------------------------------------------------

loc_12988:							; CODE XREF: sub_128EC+8Ej
		clr.w	$66(a2,d0.w)
; END OF FUNCTION CHUNK	FOR sub_128EC

; =============== S U B	R O U T	I N E =======================================


sub_1298C:							; CODE XREF: sub_1284A+36p
					; sub_128EC+2Cp ...
		moveq	#$FFFFFFFF,d2

loc_1298E:							; CODE XREF: sub_1298C+Aj
		addq.w	#1,d2
		move.b	(a0,d2.w),-1(a0,d2.w)
		bpl.s	loc_1298E
		subq.w	#1,a0
		rts
; End of function sub_1298C


; =============== S U B	R O U T	I N E =======================================


sub_1299C:							; CODE XREF: sub_128B2+24p
		moveq	#0,d1
		move.w	($FFFFB75E).w,d0
		cmpnw	 $58,d0					; 'X'
		bgt.w	loc_129BA
		move.l	#$364,d1
		neg.w	d0
		cmpnw	 $58,d0					; 'X'
		blt.w	locret_15464

loc_129BA:							; CODE XREF: sub_1299C+Aj
		btst	#1,($FFFFC2EA).w
		beq.w	loc_129C8
		eor.w	#$364,d1

loc_129C8:							; CODE XREF: sub_1299C+24j
		move.w	#$C6CE,a2
		addq.w	#1,$A(a2,d1.w)
		rts
; End of function sub_1299C


; =============== S U B	R O U T	I N E =======================================


sub_129D2:							; CODE XREF: sub_12944+16p
		movem.l	d0-d3/a0-a3,-(sp)
		move.w	$22(a2),a3
		suba.w	#$80,a3					; ''

loc_129DE:							; CODE XREF: sub_129D2+14j
		add.w	#$80,a3					; ''
		tst.w	$34(a3)
		bpl.s	loc_129DE
		move.w	d0,d3
		lsr.w	#1,d3
		move.w	$24(a2),d1
		addq.w	#1,$24(a2)
		bset	#0,($FFFFC6FE).w
		bset	#0,($FFFFCA62).w
		move.l	#off_19286,a0
		tst.w	$26(a2)
		bpl.w	loc_12A10
		addq.w	#1,a0

loc_12A10:							; CODE XREF: sub_129D2+38j
		clr.w	$34(a3)
		move.b	(a0,d1.w),$35(a3)
		bsr.w	sub_15A88
		bsr.w	sub_15AA4
		bset	#2,$63(a3)
		movem.l	(sp)+,d0-d3/a0-a3
		rts
; End of function sub_129D2


; =============== S U B	R O U T	I N E =======================================


sub_12A2E:							; CODE XREF: sub_799E+B160p
		clr.w	d0
		clr.w	d3
		lea	$9A(a2),a0

loc_12A36:							; CODE XREF: sub_12A2E+18j
					; sub_12A2E+20j
		clr.w	d2
		move.b	(a0)+,d2
		bmi.w	loc_12A50
		move.w	$66(a2,d2.w),d2
		btst	#$E,d2
		bne.s	loc_12A36
		sub.w	d3,d2
		add.w	d2,d0
		move.w	d2,d3
		bra.s	loc_12A36
; ---------------------------------------------------------------------------

loc_12A50:							; CODE XREF: sub_12A2E+Cj
		cmp.w	#6,$24(a3)
		beq.w	locret_15464
		sub.w	d3,d0
		lea	$9A(a3),a0
		clr.w	d2

loc_12A62:							; CODE XREF: sub_12A2E+40j
		move.b	(a0)+,d2
		bmi.w	locret_15464
		btst	#6,$66(a3,d2.w)
		bne.s	loc_12A62
		cmp.w	$66(a3,d2.w),d0
		blt.w	locret_15464
		add.w	d3,d0
		rts
; End of function sub_12A2E

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_799E

loc_12A7C:							; CODE XREF: sub_799E+46j
		btst	#1,($FFFFC2FA).w
		bne.w	locret_15464
		move.w	#$C6CE,a2
		lea	$364(a2),a3
		move.w	$24(a2),d0
		sub.w	$24(a3),d0
		beq.w	sub_12B32
		bpl.w	loc_12AB6
		btst	#6,($FFFFC2EE).w
		bne.w	loc_12ACC
		bsr.w	sub_12B32
		bset	#6,($FFFFC2EE).w
		bra.w	loc_12ACC
; ---------------------------------------------------------------------------

loc_12AB6:							; CODE XREF: sub_799E+B0FCj
		exg	a2,a3
		btst	#6,($FFFFC2EE).w
		beq.w	loc_12ACC
		bsr.w	sub_12B32
		bclr	#6,($FFFFC2EE).w

loc_12ACC:							; CODE XREF: sub_799E+B106j
					; sub_799E+B114j ...
		bset	#5,($FFFFC2EE).w
		bne.w	loc_12AE8
		addq.w	#1,4(a3)
		cmp.w	#$C6CE,a3
		bne.w	loc_12AE6

loc_12AE2:							; CODE XREF: sub_799E:loc_12AE6j
		bra.w	loc_12AE8
; ---------------------------------------------------------------------------

loc_12AE6:							; CODE XREF: sub_799E+B140j
		bra.s	loc_12AE2
; ---------------------------------------------------------------------------

loc_12AE8:							; CODE XREF: sub_799E+B134j
					; sub_799E:loc_12AE2j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$12
		dc.b	$BF
		dc.b	1
		dc.b	$19
		dc.b	$20
		dc.b	$20
		dc.b	$20
		dc.b	$16
		dc.b	$17
		dc.b	$18
		dc.b	$19
		dc.b	$BF
		dc.b	1
		dc.b	$1A
		dc.b	$20
		dc.b	$20
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_12A2E
		bsr.w	sub_11CA2
		bsr.w	sub_11BA4
		move.l	$1E(a3),a0
		move.w	#$BFA4,a3
		move.w	#2,(a3)
		bsr.w	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	1
		dc.b	$19
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	4(a0),a0
		add.w	(a0),a0
		move.l	a0,a1
		bsr.w	sub_11D9E
		move.w	a3,a1
		bra.w	sub_11BA4
; END OF FUNCTION CHUNK	FOR sub_799E

; =============== S U B	R O U T	I N E =======================================


sub_12B32:							; CODE XREF: sub_799E+B0F8j
					; sub_799E+B10Ap ...
		bclr	#5,($FFFFC2EE).w
		beq.w	locret_15464
		bra.w	sub_12D70
; End of function sub_12B32


; =============== S U B	R O U T	I N E =======================================


sub_12B40:							; CODE XREF: sub_7CF8:loc_7D8Ap
					; sub_9FD0+1Ap	...
		movem.l	d0-a6,-(sp)
		bclr	#7,($FFFFC2EC).w
		move.w	#$3E8,($FFFFBD1A).w
		move.w	($FFFFB018).w,d4
		move.l	#unk_56062,a2
		jsr	sub_11738
		jsr	sub_FEA52
		bsr.w	sub_1702E
		btst	#0,($FFFFC2EC).w
		bne.w	loc_12B8E
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#$1C,d1
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		bsr.w	loc_12C04

loc_12B8E:							; CODE XREF: sub_12B40+30j
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_12B40


; =============== S U B	R O U T	I N E =======================================


sub_12B94:							; CODE XREF: sub_7CF8+88p
					; ROM:00007DE0p ...
		movem.l	d0-a6,-(sp)
		bset	#7,($FFFFC2EC).w

loc_12B9E:							; CODE XREF: sub_12B94+10j
		btst	#0,($FFFFBF78).w
		bne.s	loc_12B9E
		clr.w	($FFFFBD1E).w
		clr.w	($FFFFBD20).w
		bsr.w	sub_1702E
		bsr.w	sub_11B92
; End of function sub_12B94

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_BC05C,a1
		add.l	4(a1),a1
		move.l	#unk_BC064,a2
		clr.w	d0
		move.w	#0,d1
		moveq	#$20,d2					; ' '
		moveq	#$1C,d3
		move.w	($FFFFB018).w,d4
		moveq	#0,d5
		bsr.w	sub_1169A
		bsr.w	loc_12DA6
		btst	#0,($FFFFC2EC).w
		bne.w	loc_12BFE
		move.w	#$800,d0
		move.w	($FFFFB004).w,d1
		move.w	#$7FF,d2
		bsr.w	sub_11544

loc_12BFE:							; CODE XREF: ROM:00012BEAj
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_12C04:							; CODE XREF: sub_7CF8+B6p sub_BB36+28j ...
		movem.l	d0-d2/a0-a3,-(sp)
		btst	#7,($FFFFC2EC).w
		bne.w	loc_12CB0
		bclr	#7,($FFFFC2F6).w
		beq.w	loc_12C22
		move.w	#$78,($FFFFC304).w			; 'x'

loc_12C22:							; CODE XREF: ROM:00012C18j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	0
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#9,d0
		moveq	#5,d1
		bsr.w	sub_119B8
		bset	#3,($FFFFBF78).w
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	1
		dc.b	$18
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC466).w,d0
		move.l	#word_191E4,a1
		btst	#1,($FFFFC2FA).w
		beq.w	loc_12C62
		move.w	#4,d0
		move.l	#word_191F8,a1

loc_12C62:							; CODE XREF: ROM:00012C54j
		bsr.w	sub_13508
		bsr.w	sub_12D70
		btst	#0,($FFFFC2F0).w
		bne.w	loc_12CAA
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$17
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#8,d0
		moveq	#5,d1
		bsr.w	sub_119B8
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$18
		dc.b	$1A
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$C6CE,a2
		bsr.w	sub_12D0E
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$18
		dc.b	$18
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	#$364,a2
		bsr.w	sub_12D0E

loc_12CAA:							; CODE XREF: ROM:00012C70j
					; ROM:00012D0Cj
		movem.l	(sp)+,d0-d2/a0-a3
		rts
; ---------------------------------------------------------------------------

loc_12CB0:							; CODE XREF: ROM:00012C0Ej
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$11
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC466).w,d0
		btst	#0,($FFFFC2FA).w
		beq.w	loc_12CCC
		move.w	#4,d0

loc_12CCC:							; CODE XREF: ROM:00012CC4j
		move.l	#word_191F8,a1
		bsr.w	sub_13510
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bsr.w	sub_11BA4
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$17
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$C6CE,a2
		clr.w	d0
		bsr.w	sub_12D30
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	8
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	#$364,a2
		move.w	#$2C,d0					; ','
		bsr.w	sub_12D30
		bra.s	loc_12CAA

; =============== S U B	R O U T	I N E =======================================


sub_12D0E:							; CODE XREF: ROM:00012C94p
					; ROM:00012CA6p
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		bsr.w	sub_11BA4
		move.w	#$1C,($FFFFB028).w
		move.w	$C(a2),d0
		moveq	#2,d1
		bsr.w	sub_11D3A
		bra.w	sub_11BA4
; End of function sub_12D0E


; =============== S U B	R O U T	I N E =======================================


sub_12D30:							; CODE XREF: ROM:00012CF2p
					; ROM:00012D08p
		move.w	($FFFFB028).w,-(sp)
		subq.w	#5,($FFFFB028).w
		jsr	sub_8078
		move.w	(sp)+,($FFFFB028).w
		addq.w	#2,($FFFFB02A).w
		move.w	$C(a2),d0
		cmpnw	 $A,d0
		bge.w	loc_12D56
		bra.w	loc_12D5A
; ---------------------------------------------------------------------------

loc_12D56:							; CODE XREF: sub_12D30+1Ej
		subq.w	#1,($FFFFB028).w

loc_12D5A:							; CODE XREF: sub_12D30+22j
		bsr.w	sub_11D06
		move.w	(a1),d0
		subq.w	#2,d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		addq.w	#1,($FFFFB028).w
		bra.w	sub_11DF4
; End of function sub_12D30


; =============== S U B	R O U T	I N E =======================================


sub_12D70:							; CODE XREF: sub_12B32+Aj
					; ROM:00012C66p
		btst	#5,($FFFFC2EE).w
		bne.w	locret_15464
		bsr.w	sub_11B92
; End of function sub_12D70

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	1
		dc.b	$19
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#unk_B3530,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		move.w	($FFFFB020).w,d4
		clr.w	d5
		bra.w	sub_1169A
; ---------------------------------------------------------------------------

loc_12DA6:							; CODE XREF: sub_12002+384p
					; ROM:00012BE0p
		movem.l	d0-a3,-(sp)
		bset	#3,($FFFFBF78).w
		bsr.w	loc_12C04
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$14
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$C6CE,a2
		bsr.w	sub_12DDE
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	5
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$CA32,a2
		bsr.w	sub_12DDE
		movem.l	(sp)+,d0-a3
		rts

; =============== S U B	R O U T	I N E =======================================


sub_12DDE:							; CODE XREF: ROM:00012DC2p
					; ROM:00012DD4p
		lea	$9A(a2),a0

loc_12DE2:							; CODE XREF: sub_12DDE+12j
					; sub_12DDE+18j
		clr.w	d0
		move.b	(a0)+,d0
		bmi.w	loc_12DF8
		btst	#6,$66(a2,d0.w)
		bne.s	loc_12DE2
		bsr.w	sub_12E12
		bra.s	loc_12DE2
; ---------------------------------------------------------------------------

loc_12DF8:							; CODE XREF: sub_12DDE+8j
		lea	$9A(a2),a0

loc_12DFC:							; CODE XREF: sub_12DDE+2Cj
					; sub_12DDE+32j
		clr.w	d0
		move.b	(a0)+,d0
		bmi.w	locret_15464
		btst	#6,$66(a2,d0.w)
		beq.s	loc_12DFC
		bsr.w	sub_12E12
		bra.s	loc_12DFC
; End of function sub_12DDE


; =============== S U B	R O U T	I N E =======================================


sub_12E12:							; CODE XREF: sub_12DDE+14p
					; sub_12DDE+2Ep
		cmp.w	#$A,($FFFFB02A).w
		bhi.w	locret_15464
		move.w	$66(a2,d0.w),d2
		and.w	#$7FF,d2
		move.l	$1E(a2),a1
		add.w	(a1),a1
		lsr.w	#1,d0

loc_12E2C:							; CODE XREF: sub_12E12+1Ej
		add.w	(a1),a1
		addq.w	#8,a1
		dbf	d0,loc_12E2C
		clr.w	d0
		move.b	-8(a1),d0
		move.w	($FFFFB028).w,-(sp)
		move.w	#$BFA6,a1
		bsr.w	sub_18BDC
		move.w	#$BFA4,a1
		move.w	#4,(a1)
		bsr.w	sub_11BA4
		move.w	d2,d0
		bsr.w	sub_11CA2
		bsr.w	sub_11BA4
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		rts
; End of function sub_12E12


; =============== S U B	R O U T	I N E =======================================


sub_12E66:							; CODE XREF: sub_B92E+72p
		movem.l	d0-d5/a0-a2,-(sp)
		bsr.w	sub_12EB4
		move.w	($FFFFB02C).w,-(sp)
		move.w	#$8000,($FFFFB02C).w
		ext.l	d0
		divu.w	#$100,d0
		cmpnw	 $F,d0
		bls.w	loc_12E88
		moveq	#$F,d0

loc_12E88:							; CODE XREF: sub_12E66+1Cj
		moveq	#$F,d1
		sub.w	d0,d1
		clr.w	d0
		move.l	#unk_AB920,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		move.w	(a1),d2
		moveq	#1,d3
		move.w	($FFFFB016).w,d4
		moveq	#0,d5
		bsr.w	sub_1169A
		move.w	(sp)+,($FFFFB02C).w
		movem.l	(sp)+,d0-d5/a0-a2
		rts
; End of function sub_12E66


; =============== S U B	R O U T	I N E =======================================


sub_12EB4:							; CODE XREF: sub_F790+1Ap sub_F790+24p ...
		movem.l	d1-d5/a0-a3,-(sp)
		lea	$16A(a2),a1
		asl.w	#3,d0
		add.w	d0,a1
		clr.l	d0
		clr.w	d1
		move.l	#off_19286,a0
		move.w	$24(a2),d4
		bra.w	loc_12EEA
; ---------------------------------------------------------------------------

loc_12ED2:							; CODE XREF: sub_12EB4:loc_12EEAj
		clr.w	d5
		move.b	(a0,d4.w),d5
		beq.w	loc_12EEA
		clr.w	d3
		move.b	(a1,d5.w),d3
		asl.w	#1,d3
		addq.w	#1,d1
		add.w	$30(a2,d3.w),d0

loc_12EEA:							; CODE XREF: sub_12EB4+1Aj
					; sub_12EB4+24j
		dbf	d4,loc_12ED2
		divu.w	d1,d0
		movem.l	(sp)+,d1-d5/a0-a3
		rts
; End of function sub_12EB4


; =============== S U B	R O U T	I N E =======================================


sub_12EF6:							; CODE XREF: sub_E1F4:loc_E248p
		movem.l	d1-d3/a0,-(sp)
		clr.l	d0
		clr.w	d1
		moveq	#5,d2
		move.w	$22(a2),a0

loc_12F04:							; CODE XREF: sub_12EF6+28j
		tst.w	$34(a0)
		ble.w	loc_12F1A
		clr.w	d3
		move.b	$66(a0),d3
		add.w	d3,d3
		add.w	$32(a2,d3.w),d0
		addq.w	#1,d1

loc_12F1A:							; CODE XREF: sub_12EF6+12j
		add.w	#$80,a0					; ''
		dbf	d2,loc_12F04
		tst.w	d1
		beq.w	loc_12F2A
		divu.w	d1,d0

loc_12F2A:							; CODE XREF: sub_12EF6+2Ej
		movem.l	(sp)+,d1-d3/a0
		rts
; End of function sub_12EF6


; =============== S U B	R O U T	I N E =======================================


sub_12F30:							; CODE XREF: sub_14620:loc_14724p
					; sub_14620+214p ...
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		btst	#4,($FFFFC2EA).w
		bne.w	locret_15464
		bclr	#4,($FFFFC2EE).w
		bne.w	loc_12FBC
		btst	#0,($FFFFC2F4).w
		beq.w	locret_15464
		bclr	#3,($FFFFC2FE).w
		bne.w	loc_12F6C
		bset	#3,($FFFFC2FE).w
		bra.w	locret_15464
; ---------------------------------------------------------------------------

loc_12F6C:							; CODE XREF: sub_12F30+2Ej
		movem.l	d0-d1/a1-a3,-(sp)
		move.l	a4,-(sp)
		move.l	#$FFFFC46E,a4
		add.w	($FFFFC472).w,a4
		move.l	#$FFFFC6CE,a2
		btst	#7,2(a4)
		beq.w	loc_12F90
		add.w	#$364,a2

loc_12F90:							; CODE XREF: sub_12F30+58j
		clr.w	d0
		move.b	3(a4),d0
		move.l	(sp)+,a4
		move.w	$22(a2),a2
		move.w	#5,d1

loc_12FA0:							; CODE XREF: sub_12F30+7Cj
		cmp.b	$66(a2),d0
		beq.w	loc_12FB4
		add.w	#$80,a2					; ''
		dbf	d1,loc_12FA0
		bra.w	loc_1303A
; ---------------------------------------------------------------------------

loc_12FB4:							; CODE XREF: sub_12F30+74j
		move.w	$52(a2),d0
		bra.w	loc_12FD0
; ---------------------------------------------------------------------------

loc_12FBC:							; CODE XREF: sub_12F30+1Aj
		movem.l	d0-d1/a1-a3,-(sp)
		add.w	#$64,($FFFFB89C).w			; 'd'
		add.w	#$A,($FFFFB8A2).w
		move.w	($FFFFBED8).w,d0

loc_12FD0:							; CODE XREF: sub_12F30+88j
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		bsr.w	sub_13040
		addq.w	#1,(a2)
		btst	#5,($FFFFC2EE).w
		beq.w	loc_1300C
		btst	#6,($FFFFC2EE).w
		bne.w	loc_13004
		btst	#6,$62(a3)
		bne.w	loc_1300C

loc_12FFC:							; CODE XREF: sub_12F30+DAj
		addq.w	#1,$354(a2)
		bra.w	loc_1300C
; ---------------------------------------------------------------------------

loc_13004:							; CODE XREF: sub_12F30+BEj
		btst	#6,$62(a3)
		bne.s	loc_12FFC

loc_1300C:							; CODE XREF: sub_12F30+B4j
					; sub_12F30+C8j ...
		move.l	a2,-(sp)
		move.w	($FFFFC466).w,d0
		add.w	d0,d0
		add.w	d0,a2
		addq.w	#1,$34A(a2)
		move.l	(sp)+,a2
		clr.w	d0
		move.b	$66(a3),d0
		add.w	#$E8,d0					; ''
		addq.b	#1,(a2,d0.w)
		move.w	$26(a1),d0
		bmi.w	loc_1303A
		add.w	#$E8,d0					; ''
		addq.b	#1,(a1,d0.w)

loc_1303A:							; CODE XREF: sub_12F30+80j
					; sub_12F30+FEj
		movem.l	(sp)+,d0-d1/a1-a3
		rts
; End of function sub_12F30


; =============== S U B	R O U T	I N E =======================================


sub_13040:							; CODE XREF: sub_A9D6+206p
					; sub_A9D6+4BCp ...
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		btst	#6,$62(a3)
		beq.w	locret_15464
		exg	a1,a2
		rts
; End of function sub_13040


; =============== S U B	R O U T	I N E =======================================


sub_13056:							; CODE XREF: sub_130E6:loc_130FCp
		bsr.w	sub_15A24
		move.w	#$C6CE,a2
		lea	$364(a2),a3
		bsr.w	sub_13068
		exg	a2,a3
; End of function sub_13056


; =============== S U B	R O U T	I N E =======================================


sub_13068:							; CODE XREF: sub_13056+Cp
		bsr.w	sub_130BE
		clr.w	$16(a3)
		tst.w	($FFFFD058).w
		bne.w	locret_15464
		move.w	$24(a3),d0
		sub.w	$24(a2),d0
		beq.w	locret_15464
		move.w	#3,$16(a3)
		tst.w	d0
		bpl.w	locret_15464
		move.w	#5,$16(a3)
		rts
; End of function sub_13068


; =============== S U B	R O U T	I N E =======================================


sub_13098:							; CODE XREF: ROM:00009DD0p
					; ROM:00009DDAp
		moveq	#$32,d0					; '2'

loc_1309A:							; CODE XREF: sub_13098+22j
		cmp.w	#$FFFD,$66(a2,d0.w)
		bne.w	loc_130B2
		cmp.w	#$FFFC,$66(a2,d0.w)
		bne.w	loc_130B2
		bra.w	loc_130B8
; ---------------------------------------------------------------------------

loc_130B2:							; CODE XREF: sub_13098+8j
					; sub_13098+12j
		move.w	#$1000,$32(a2,d0.w)

loc_130B8:							; CODE XREF: sub_13098+16j
		subq.w	#2,d0
		bpl.s	loc_1309A
		rts
; End of function sub_13098


; =============== S U B	R O U T	I N E =======================================


sub_130BE:							; CODE XREF: sub_13068p sub_137FCp
		moveq	#$32,d0					; '2'

loc_130C0:							; CODE XREF: sub_130BE+24j
		cmp.w	#$FFFC,$66(a2,d0.w)
		beq.w	loc_130E0
		move.w	#$1000,$32(a2,d0.w)
		cmp.w	#$FFFD,$66(a2,d0.w)
		bne.w	loc_130E0
		move.w	#$FFFE,$66(a2,d0.w)

loc_130E0:							; CODE XREF: sub_130BE+8j
					; sub_130BE+18j
		subq.w	#2,d0
		bpl.s	loc_130C0
		rts
; End of function sub_130BE


; =============== S U B	R O U T	I N E =======================================


sub_130E6:							; CODE XREF: sub_7CB0:loc_172B6p
		cmp.w	#4,($FFFFC466).w
		bne.w	loc_130FC
		move.w	#0,($FFFFC468).w
		jsr	sub_F9CDE

loc_130FC:							; CODE XREF: sub_130E6+6j
		bsr.w	sub_13056
		moveq	#$F,d0
		move.w	#$B04A,a0

loc_13106:							; CODE XREF: sub_130E6+26j
		clr.w	(a0)
		add.w	#$80,a0					; ''
		dbf	d0,loc_13106
		move.w	($FFFFB026).w,d4
		move.l	#unk_A892A,a2
		bsr.w	sub_11738
		move.w	#$140,($FFFFB8AC).w
		move.w	($FFFFC466).w,d0
		tst.w	($FFFFD048).w
		beq.w	loc_13132
		addq.w	#5,d0

loc_13132:							; CODE XREF: sub_130E6+46j
		asl.w	#2,d0
		lea	off_131CC(pc),a0
		move.l	(a0,d0.w),a0
		btst	#0,($FFFFC2FA).w
		beq.w	loc_1314C
		move.l	#off_19A00,a0

loc_1314C:							; CODE XREF: sub_130E6+5Cj
		bset	#0,($FFFFC2EC).w
		move.l	#$7DCE,a1
		jsr	(sub_7E36).w
		bsr.w	sub_1828A
		move.w	d1,($FFFFCF2C).w

loc_13164:							; CODE XREF: sub_130E6+AEj
		tst.w	($FFFFCF2C).w
		bmi.w	loc_1319C
		bsr.w	sub_13384
		bsr.w	sub_133B2
		move.w	#$96,d0					; ''
		bsr.w	sub_111D0
		btst	#7,d1
		bne.w	loc_13196
		bsr.w	sub_134D8
		move.w	#$3C,d0					; '<'
		bsr.w	sub_111D0
		btst	#7,d1
		beq.s	loc_13164

loc_13196:							; CODE XREF: sub_130E6+9Aj
		bset	#3,($FFFFC2F0).w

loc_1319C:							; CODE XREF: sub_130E6+82j
		bsr.w	sub_1351C
		bclr	#3,($FFFFC2F0).w
		bne.w	loc_131C6

loc_131AA:							; CODE XREF: sub_130E6+D8j
					; sub_130E6+DEj
		move.w	#$1E0,d0
		bsr.w	sub_111D0
		btst	#7,d1
		bne.w	loc_131C6
		tst.w	($FFFFC328).w
		bne.s	loc_131AA
		tst.w	($FFFFC32A).w
		bne.s	loc_131AA

loc_131C6:							; CODE XREF: sub_130E6+C0j
					; sub_130E6+D0j
		st	($FFFFB8AC).w
		rts
; End of function sub_130E6

; ---------------------------------------------------------------------------
off_131CC:	dc.l	off_19A84				; DATA XREF: sub_130E6+4Eo
		dc.l	off_19C04
		dc.l	off_19C04				; Menu related string pointers
		dc.l	off_19C04
		dc.l	off_19D60
		dc.l	off_19B38
		dc.l	off_19C04
		dc.l	off_19C04
		dc.l	off_19C04
		dc.l	off_19E74

; =============== S U B	R O U T	I N E =======================================


sub_131F4:							; CODE XREF: sub_7CB0-4E4p
		cmp.w	#1,($FFFFCEEC).w
		bgt.w	locret_15464
		bsr.w	sub_1828A
		move.w	#$CE66,a0
		moveq	#$10,d0
		mulu.w	d1,d0
		add.w	d0,a0

loc_1320C:							; CODE XREF: sub_131F4+44j
		cmp.w	($FFFFCEE6).w,d1
		beq.w	loc_13234
		btst	#2,$E(a0)
		bne.w	loc_13234
		clr.w	8(a0)
		moveq	#3,d0
		bsr.w	sub_11086
		bra.w	loc_13230
; ---------------------------------------------------------------------------

loc_1322C:							; CODE XREF: sub_131F4:loc_13230j
		bsr.w	sub_13276

loc_13230:							; CODE XREF: sub_131F4+34j
		dbf	d0,loc_1322C

loc_13234:							; CODE XREF: sub_131F4+1Cj
					; sub_131F4+26j
		suba.w	#$10,a0
		dbf	d1,loc_1320C
		rts
; End of function sub_131F4


; =============== S U B	R O U T	I N E =======================================


sub_1323E:							; CODE XREF: sub_7CB0+F5F4p
		bsr.w	sub_1828A
		move.w	d1,($FFFFCF2C).w
		move.w	#$CE66,a0
		moveq	#$10,d0
		mulu.w	d1,d0
		add.w	d0,a0

loc_13250:							; CODE XREF: sub_1323E+32j
		cmp.w	($FFFFCEE6).w,d1
		beq.w	loc_1326C
		bclr	#1,$E(a0)
		btst	#2,$E(a0)
		bne.w	loc_1326C
		bsr.w	sub_13276

loc_1326C:							; CODE XREF: sub_1323E+16j
					; sub_1323E+26j
		suba.w	#$10,a0
		dbf	d1,loc_13250
		rts
; End of function sub_1323E


; =============== S U B	R O U T	I N E =======================================


sub_13276:							; CODE XREF: sub_131F4:loc_1322Cp
					; sub_1323E+2Ap
		cmp.w	#4,8(a0)
		bge.w	locret_15464
		movem.l	d0-d1/a0-a1,-(sp)
		cmp.w	#3,8(a0)
		bne.w	loc_132BC
		move.w	#5,8(a0)
		move.w	$A(a0),d0
		sub.w	$C(a0),d0
		cmpnw	 1,d0
		bgt.w	loc_132DC
		cmpnw	 $FFFF,d0
		blt.w	loc_132DC
		move.w	#3,8(a0)
		bset	#1,$E(a0)
		bra.w	loc_132DC
; ---------------------------------------------------------------------------

loc_132BC:							; CODE XREF: sub_13276+14j
		addq.w	#1,8(a0)
		move.w	(a0),d0
		move.w	2(a0),d1
		bsr.w	sub_132E2
		add.w	d0,$A(a0)
		move.w	2(a0),d0
		move.w	(a0),d1
		bsr.w	sub_132E2
		add.w	d0,$C(a0)

loc_132DC:							; CODE XREF: sub_13276+2Aj
					; sub_13276+32j ...
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_13276


; =============== S U B	R O U T	I N E =======================================


sub_132E2:							; CODE XREF: sub_13276+50p
					; sub_13276+5Ep
		asl.w	#2,d0
		move.w	#$30E,a1
		move.l	(a1,d0.w),a1
		add.w	8(a1),a1
		move.b	(a1),d0
		and.w	#$70,d0					; 'p'
		lsr.w	#1,d0
		lea	word_13338(pc),a1
		move.l	(a1,d0.w),($FFFFD036).w
		move.l	4(a1,d0.w),($FFFFD03A).w
		asl.w	#2,d1
		move.w	#$30E,a1
		move.l	(a1,d1.w),a1
		add.w	8(a1),a1
		move.b	(a1),d0
		and.w	#7,d0
		asl.w	#3,d0
		lea	word_13338(pc),a1
		move.l	(a1,d0.w),d1
		add.l	d1,($FFFFD036).w
		move.l	4(a1,d0.w),d1
		add.l	d1,($FFFFD03A).w
		moveq	#4,d0
		bra.w	sub_10EB4
; End of function sub_132E2

; ---------------------------------------------------------------------------
word_13338:	dc.w	$2B					; DATA XREF: sub_132E2+16o
					; sub_132E2+3Co
		dc.w	$1E
		dc.w	$17
		dc.w	4
		dc.w	$27
		dc.w	$20
		dc.w	$18
		dc.w	5
		dc.w	$23
		dc.w	$21
		dc.w	$1A
		dc.w	6
		dc.w	$1E
		dc.w	$22
		dc.w	$1E
		dc.w	6
		dc.w	$1A
		dc.w	$24
		dc.w	$1E
		dc.w	8
		dc.w	$16
		dc.w	$25
		dc.w	$1F
		dc.w	$A
		dc.w	$13
		dc.w	$26
		dc.w	$1F
		dc.w	$C
		dc.w	$F
		dc.w	$26
		dc.w	$20
		dc.w	$F

; =============== S U B	R O U T	I N E =======================================


sub_13378:							; CODE XREF: sub_12002+D8p
		moveq	#$FFFFFFFF,d3
		cmp.w	#$1E0,($FFFFC468).w
		bgt.w	locret_15464
; End of function sub_13378


; =============== S U B	R O U T	I N E =======================================


sub_13384:							; CODE XREF: sub_130E6+86p
					; sub_13384+10j ...
		move.w	($FFFFCF2C).w,d3
		bmi.w	locret_15464
		subq.w	#1,($FFFFCF2C).w
		cmp.w	($FFFFCEE6).w,d3
		beq.s	sub_13384
; End of function sub_13384


; =============== S U B	R O U T	I N E =======================================


sub_13396:							; CODE XREF: sub_1354C+14p
					; ROM:000137B4p
		moveq	#$10,d0
		mulu.w	d3,d0
		move.w	#$CE66,a0
		add.w	d0,a0
		btst	#1,$E(a0)
		bne.s	sub_13384
		btst	#2,$E(a0)
		bne.s	sub_13384
		rts
; End of function sub_13396


; =============== S U B	R O U T	I N E =======================================


sub_133B2:							; CODE XREF: sub_12002+DCp
					; sub_130E6+8Ap ...
		tst.w	d3
		bmi.w	locret_15464
		bsr.w	sub_134E4
		bsr.w	sub_119B8
		addq.w	#1,($FFFFB028).w
		subq.w	#4,($FFFFB02A).w
		move.w	($FFFFB028).w,-(sp)
		bsr.w	sub_1828A
		move.w	#$A000,($FFFFB02C).w
		bsr.w	sub_11B92
; End of function sub_133B2

; ---------------------------------------------------------------------------
		dc.w	$1A
		dc.b	'                        '
; ---------------------------------------------------------------------------
		move.w	(sp),($FFFFB028).w
		moveq	#1,d0
		add.w	($FFFFCEEC).w,d0
		tst.w	($FFFFD048).w
		bne.w	loc_13408
		clr.w	d0

loc_13408:							; CODE XREF: ROM:00013402j
		lea	word_13488(pc),a1
		bsr.w	sub_13508
		move.w	8(a0),d0
		subq.w	#1,d0
		move.l	#word_191E4,a1
		bsr.w	sub_13510
		move.w	(a1),d0
		lsr.w	#1,d0
		neg.w	d0
		add.w	(sp),d0
		add.w	#$17,d0
		move.w	d0,($FFFFB028).w
		bsr.w	sub_11BA4
		move.w	#$8000,($FFFFB02C).w
		move.w	2(a0),d0
		move.w	$C(a0),d1
		bsr.w	sub_13454
		move.w	(a0),d0
		move.w	$A(a0),d1
		bsr.w	sub_13454
		addq.w	#2,sp
		rts

; =============== S U B	R O U T	I N E =======================================


sub_13454:							; CODE XREF: ROM:00013442p
					; ROM:0001344Cp

arg_0		=	4

		move.w	arg_0(sp),($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		move.w	#$30E,a1
		asl.w	#2,d0
		move.l	(a1,d0.w),a1
		add.w	4(a1),a1
		bsr.w	sub_11BA4
		move.w	arg_0(sp),($FFFFB028).w
		add.w	#$15,($FFFFB028).w
		move.w	d1,d0
		moveq	#2,d1
		bsr.w	sub_11D3A
		bra.w	sub_11BA4
; End of function sub_13454

; ---------------------------------------------------------------------------
word_13488:	dc.w	$12					; DATA XREF: ROM:loc_13408o
aEaHockeyNight:	dc.b	'EA Hockey Night',0
		dc.w	$18
aStanleyCupQualifier: dc.b 'Stanley Cup Qualifier',0
		dc.w	$E
aQuarterfinal:	dc.b	'Quarterfinal'
		dc.w	$18
aStanleyCupSemifinal: dc.b 'Stanley Cup Semifinal',0

; =============== S U B	R O U T	I N E =======================================


sub_134D8:							; CODE XREF: sub_130E6+9Ep
					; ROM:000137CCp
		bsr.w	sub_134E4
		move.w	#$7FF,d2
		bra.w	sub_1197E
; End of function sub_134D8


; =============== S U B	R O U T	I N E =======================================


sub_134E4:							; CODE XREF: sub_133B2+6p sub_134D8p
		bsr.w	sub_11B92
; End of function sub_134E4

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	3
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		btst	#0,($FFFFC2EC).w
		bne.w	loc_13502
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	3
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------

loc_13502:							; CODE XREF: ROM:000134F4j
		moveq	#$1A,d0
		moveq	#5,d1
		rts

; =============== S U B	R O U T	I N E =======================================


sub_13508:							; CODE XREF: sub_81F2+52p
					; ROM:000082B0p ...
		bsr.w	sub_13510
		bra.w	sub_11A48
; End of function sub_13508


; =============== S U B	R O U T	I N E =======================================


sub_13510:							; CODE XREF: ROM:00008BBEp
					; ROM:00012CD2p ...
		bra.w	loc_13516
; ---------------------------------------------------------------------------

loc_13514:							; CODE XREF: sub_13510:loc_13516j
		add.w	(a1),a1

loc_13516:							; CODE XREF: sub_13510j
		dbf	d0,loc_13514
		rts
; End of function sub_13510


; =============== S U B	R O U T	I N E =======================================


sub_1351C:							; CODE XREF: sub_130E6:loc_1319Cp
		bsr.w	sub_1828A
		moveq	#$10,d3
		mulu.w	d1,d3
		move.w	#$CE66,a0
		add.w	d3,a0

loc_1352A:							; CODE XREF: sub_1351C+2Aj
		btst	#2,$E(a0)
		bne.w	loc_13542
		bclr	#1,$E(a0)
		beq.w	loc_13542
		bsr.w	sub_1354C

loc_13542:							; CODE XREF: sub_1351C+14j
					; sub_1351C+1Ej
		suba.w	#$10,a0
		dbf	d1,loc_1352A
		rts
; End of function sub_1351C


; =============== S U B	R O U T	I N E =======================================


sub_1354C:							; CODE XREF: sub_1351C+22p

; FUNCTION CHUNK AT 000137E0 SIZE 0000001C BYTES

		movem.l	d0-a6,-(sp)

loc_13550:							; CODE XREF: ROM:000137A8j
		btst	#3,($FFFFC2F0).w
		bne.w	loc_137F6
		bsr.w	sub_13848
		move.w	d1,d3
		bsr.w	sub_13396
		bsr.w	sub_133B2
		btst	#3,($FFFFC2F0).w
		bne.w	loc_137EC
		move.w	#4,($FFFFB028).w
		subq.w	#6,($FFFFB02A).w
		moveq	#$18,d0
		moveq	#3,d1
		bsr.w	sub_119B8
		addq.w	#1,($FFFFB028).w
		subq.w	#2,($FFFFB02A).w
		bsr.w	sub_11B92
; End of function sub_1354C

; ---------------------------------------------------------------------------
		dc.w	$16
aHighlightFromGame: dc.b 'Highlight from game:'
; ---------------------------------------------------------------------------
		move.w	#$B4,d0					; ''
		bsr.w	sub_111D0
		btst	#7,d1
		bne.w	loc_137EC
		st	($FFFFB8AC).w
		clr.w	($FFFFB8A2).w
		move.w	(a0),($FFFFC330).w
		move.w	2(a0),($FFFFC332).w
		move.l	a0,-(sp)
		clr.w	($FFFFC6F4).w
		clr.w	($FFFFCA58).w
		jsr	sub_17102
		jsr	(sub_77F4).w
		st	($FFFFC320).w
		st	($FFFFC322).w
		clr.w	($FFFFC328).w
		clr.w	($FFFFC32A).w
		moveq	#$78,d0					; 'x'
		bsr.w	sub_11086
		add.w	#$3C,d0					; '<'
		move.w	d0,($FFFFC468).w
		move.l	(sp),a0
		move.w	8(a0),($FFFFC466).w
		subq.w	#1,($FFFFC466).w
		move.w	$A(a0),($FFFFC6DA).w
		move.w	$C(a0),($FFFFCA3E).w
		move.b	#$10,($FFFFC2EA).w
		btst	#0,($FFFFC467).w
		beq.w	loc_13628
		bset	#1,($FFFFC2EA).w

loc_13628:							; CODE XREF: ROM:0001361Ej
		clr.b	($FFFFC2EC).w
		move.b	#4,($FFFFC2EE).w
		clr.b	($FFFFC2F0).w
		bclr	#4,($FFFFBF78).w
		bset	#3,($FFFFBF78).w
		clr.w	($FFFFBEE2).w
		clr.b	($FFFFBF80).w
		st	($FFFFC3F0).w
		st	($FFFFBEE8).w
		st	($FFFFBEEC).w
		jsr	Z80_MuteAllFM
		bsr.w	sub_16BAC
		bsr.w	sub_12B40
		move.l	#$C00000,a0
		move.w	#$9100,4(a0)
		move.w	#$9200,4(a0)
		move.w	($FFFFB026).w,d4
		move.l	#unk_5C410,a2
		bsr.w	sub_11738
		clr.w	($FFFFBD18).w
		clr.w	($FFFFBD1C).w
		bsr.w	sub_15A24
		move.w	#$C6CE,a2
		bsr.w	sub_137FC
		add.w	#$364,a2
		bsr.w	sub_137FC
		bsr.w	sub_170B0
		moveq	#$B,d0
		move.l	#word_1387C,a0
		move.w	#$B04A,a1

loc_136B0:							; CODE XREF: ROM:000136D8j
		move.w	$34(a1),d1
		btst	#7,$62(a1)
		bne.w	loc_136C0
		addq.w	#6,d1

loc_136C0:							; CODE XREF: ROM:000136BAj
		asl.w	#2,d1
		move.w	(a0,d1.w),(a1)
		move.w	2(a0,d1.w),$14(a1)
		clr.w	$28(a1)
		clr.w	$2A(a1)
		add.w	#$80,a1					; ''
		dbf	d0,loc_136B0
		bsr.w	sub_1702E
		move.w	#$18,($FFFFBD26).w
		move.w	($FFFFB03E).w,($FFFFB040).w
		move.w	#$B4,-(sp)				; ''

loc_136F0:							; CODE XREF: ROM:00013714j
		jsr	(sub_794E).w
		btst	#0,($FFFFC2EA).w
		beq.w	loc_13704
		subq.w	#1,(sp)
		bmi.w	loc_13738

loc_13704:							; CODE XREF: ROM:000136FAj
		bsr.w	sub_112BC
		btst	#5,d1
		bne.w	loc_13716
		btst	#7,d1
		beq.s	loc_136F0

loc_13716:							; CODE XREF: ROM:0001370Cj
		move.w	($FFFFC6DA).w,d0
		cmp.w	($FFFFCA3E).w,d0
		bne.w	loc_13738
		move.w	($C00008).l,d0
		and.w	#1,d0
		add.w	d0,($FFFFC6DA).w
		eor.w	#1,d0
		add.w	d0,($FFFFCA3E).w

loc_13738:							; CODE XREF: ROM:00013700j
					; ROM:0001371Ej
		addq.w	#2,sp
		move.l	(sp)+,a0
		movem.l	d1/a0,-(sp)
		jsr	PSG_SetupNoise
		jsr	Z80_MuteAllFM
		movem.l	(sp)+,d1/a0
		move.w	($FFFFC6DA).w,$A(a0)
		move.w	($FFFFCA3E).w,$C(a0)
		bsr.w	sub_10F32
		moveq	#$F,d0
		move.w	#$B04A,a0

loc_13766:							; CODE XREF: ROM:0001376Cj
		clr.w	(a0)
		add.w	#$80,a0					; ''
		dbf	d0,loc_13766
		bsr.w	sub_13862
		bsr.w	sub_1720C
		bsr.w	sub_12B94
		bsr.w	sub_15EC0
		jsr	(loc_7DCE).w
		jsr	(loc_7E46).w
		move.w	#$18,($FFFFBD26).w
		move.w	#$79,-(sp)				; 'y'
		bsr.w	sub_11156
		movem.l	(sp),d0-a6
		move.w	#4,8(a0)
		move.w	$A(a0),d0
		cmp.w	$C(a0),d0
		beq.w	loc_13550
		move.w	#5,8(a0)
		move.w	d1,d3
		bsr.w	sub_13396
		bsr.w	sub_133B2
		move.w	#$B4,d0					; ''
		bsr.w	sub_111D0
		btst	#7,d1
		bne.w	loc_137E0
		bsr.w	sub_134D8
		move.w	#$3C,d0					; '<'
		bsr.w	sub_111D0
		btst	#7,d1
		beq.w	loc_137E6
; START	OF FUNCTION CHUNK FOR sub_1354C

loc_137E0:							; CODE XREF: ROM:000137C8j
					; sub_1354C+2A8j
		bset	#3,($FFFFC2F0).w

loc_137E6:							; CODE XREF: ROM:000137DCj
					; sub_1354C+2AEj
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_137EC:							; CODE XREF: sub_1354C+22j
					; ROM:000135B2j
		bsr.w	sub_13862
		bsr.w	VDP_ReadHVCounter
		bra.s	loc_137E0
; ---------------------------------------------------------------------------

loc_137F6:							; CODE XREF: sub_1354C+Aj
		bsr.w	VDP_ReadHVCounter
		bra.s	loc_137E6
; END OF FUNCTION CHUNK	FOR sub_1354C

; =============== S U B	R O U T	I N E =======================================


sub_137FC:							; CODE XREF: ROM:00013694p
					; ROM:0001369Cp
		jsr	sub_130BE
		bsr.w	sub_15788
		bsr.w	sub_159A8
		st	$18(a2)
		bclr	#7,($FFFFC2FE).w
		st	$1A(a2)
		st	$1C(a2)
		rts
; End of function sub_137FC


; =============== S U B	R O U T	I N E =======================================


VDP_ReadHVCounter:						; CODE XREF: sub_1354C+2A4p
					; sub_1354C:loc_137F6p
		move.w	$A(a0),d0
		cmp.w	$C(a0),d0
		bne.w	loc_13840
		move.w	($C00008).l,d0
		and.w	#1,d0
		add.w	d0,$A(a0)
		eor.w	#1,d0
		add.w	d0,$C(a0)

loc_13840:							; CODE XREF: VDP_ReadHVCounter+8j
		move.w	#5,8(a0)
		rts
; End of function VDP_ReadHVCounter


; =============== S U B	R O U T	I N E =======================================


sub_13848:							; CODE XREF: sub_1354C+Ep
		move.w	#$5BD,d0
		move.w	#$C2EA,a1
		move.l	#$FFFF0000,a2

loc_13856:							; CODE XREF: sub_13848+10j
		move.w	(a1)+,(a2)+
		dbf	d0,loc_13856
		move.w	($FFFFC472).w,(a2)+
		rts
; End of function sub_13848


; =============== S U B	R O U T	I N E =======================================


sub_13862:							; CODE XREF: ROM:00013770p
					; sub_1354C:loc_137ECp
		move.w	#$5BD,d0
		move.w	#$C2EA,a2
		move.l	#$FFFF0000,a1

loc_13870:							; CODE XREF: sub_13862+10j
		move.w	(a1)+,(a2)+
		dbf	d0,loc_13870
		move.w	(a1)+,($FFFFC472).w
		rts
; End of function sub_13862

; ---------------------------------------------------------------------------
word_1387C:	dc.w	0					; DATA XREF: ROM:000136A6o
		dc.w	$FF10
		dc.w	$FFBA
		dc.w	$FF74
		dc.w	$14
		dc.w	$FF92
		dc.w	$FF9C
		dc.w	$FFC4
		dc.w	$19
		dc.w	$FFD8
		dc.w	$46
		dc.w	$FFB0
		dc.w	0
		dc.w	$F0
		dc.w	$22
		dc.w	$5A
		dc.w	$FFC9
		dc.w	$50
		dc.w	$64
		dc.w	$28
		dc.w	$FFF6
		dc.w	$A
		dc.w	$FFCE
		dc.w	0

; =============== S U B	R O U T	I N E =======================================


sub_138AC:							; CODE XREF: sub_A9D6:loc_AEB8p
		clr.w	($FFFFBFA0).w
		btst	#7,($FFFFC2EC).w
		beq.w	loc_138BC
		exg	d2,d3

loc_138BC:							; CODE XREF: sub_138AC+Aj
		btst	#2,$62(a3)
		bne.w	loc_13922
		movem.l	d0-d7,-(sp)
		move.w	(a3),d2
		move.w	$14(a3),d3
		move.w	$4A(a3),($FFFFBD22).w
		move.w	$4C(a3),($FFFFBD24).w
		bsr.w	sub_1454C
		move.w	$4E(a3),d0
		or.w	$50(a3),d0
		bne.w	loc_1391A
		cmp.w	#$B,$52(a3)
		bgt.w	loc_1391A
		movem.l	(sp),d0-d7
		move.l	a3,-(sp)
		bsr.w	sub_106E0
		move.w	(a3),d2
		move.w	$14(a3),d3
		add.w	d0,d2
		add.w	d1,d3
		move.w	#1,($FFFFBD22).w
		move.w	#1,($FFFFBD24).w
		bsr.w	sub_1454C

loc_1391A:							; CODE XREF: sub_138AC+3Cj
					; sub_138AC+46j
		movem.l	(sp)+,d0-d7
		bsr.w	sub_139A6

loc_13922:							; CODE XREF: sub_138AC+16j
		tst.w	($FFFFBFA0).w
		bne.w	loc_1399A
		move.w	$52(a3),d0
		asl.w	#1,d0
		move.w	#$B86A,a0
		move.w	#$B88A,a1
		move.w	#$B84A,a2
		move.w	(a0,d0.w),d1

loc_13940:							; CODE XREF: sub_138AC+BCj
		cmpnw	 $F,d1
		beq.w	loc_1396A
		clr.w	d4
		move.b	1(a1,d1.w),d4
		cmp.w	(a2,d4.w),d3
		ble.w	loc_1396A
		addq.w	#1,(a0,d0.w)
		subq.w	#1,(a0,d4.w)
		move.b	d4,(a1,d1.w)
		move.b	d0,1(a1,d1.w)
		addq.w	#1,d1
		bra.s	loc_13940
; ---------------------------------------------------------------------------

loc_1396A:							; CODE XREF: sub_138AC+98j
					; sub_138AC+A6j
		move.w	(a0,d0.w),d1
		beq.w	loc_13994

loc_13972:							; CODE XREF: sub_138AC+E6j
		clr.w	d4
		move.b	-1(a1,d1.w),d4
		cmp.w	(a2,d4.w),d3
		bge.w	loc_13994
		subq.w	#1,(a0,d0.w)
		addq.w	#1,(a0,d4.w)
		move.b	d4,(a1,d1.w)
		move.b	d0,-1(a1,d1.w)
		subq.w	#1,d1
		bne.s	loc_13972

loc_13994:							; CODE XREF: sub_138AC+C2j
					; sub_138AC+D0j
		move.w	d3,(a2,d0.w)
		rts
; ---------------------------------------------------------------------------

loc_1399A:							; CODE XREF: sub_138AC+7Aj
		move.w	$1C(a3),(a3)
		move.w	$20(a3),$14(a3)
		rts
; End of function sub_138AC


; =============== S U B	R O U T	I N E =======================================


sub_139A6:							; CODE XREF: sub_138AC+72p
		btst	#5,$63(a3)
		bne.w	locret_13A18
		cmp.w	#$B,$52(a3)
		bgt.w	locret_13A18
		move.w	$52(a3),d0
		asl.w	#1,d0
		move.w	#$B86A,a0
		move.w	#$B88A,a1
		move.w	#$B84A,a2
		move.w	(a0,d0.w),d1

loc_139D0:							; CODE XREF: sub_139A6+4Cj
		cmpnw	 $F,d1
		beq.w	loc_139F4
		clr.w	d4
		move.b	1(a1,d1.w),d4
		move.w	(a2,d4.w),d5
		sub.w	d3,d5
		cmpnw	 $10,d5
		bgt.w	loc_139F4
		bsr.w	sub_13A1A
		addq.w	#1,d1
		bra.s	loc_139D0
; ---------------------------------------------------------------------------

loc_139F4:							; CODE XREF: sub_139A6+2Ej
					; sub_139A6+42j
		move.w	(a0,d0.w),d1
		beq.w	locret_13A18

loc_139FC:							; CODE XREF: sub_139A6+70j
		clr.w	d4
		move.b	-1(a1,d1.w),d4
		move.w	d3,d5
		sub.w	(a2,d4.w),d5
		cmpnw	 $10,d5
		bgt.w	locret_13A18
		bsr.w	sub_13A1A
		subq.w	#1,d1
		bne.s	loc_139FC

locret_13A18:							; CODE XREF: sub_139A6+6j
					; sub_139A6+10j ...
		rts
; End of function sub_139A6


; =============== S U B	R O U T	I N E =======================================


sub_13A1A:							; CODE XREF: sub_139A6+46p
					; sub_139A6+6Ap
		movem.l	d0-a3,-(sp)
		asl.w	#6,d4
		move.l	#$FFFFB04A,a2
		add.w	d4,a2
		btst	#2,$62(a2)
		bne.w	loc_13BA8
		btst	#5,$63(a2)
		bne.w	loc_13BA8
		cmp.w	#$B,$52(a2)
		bgt.w	loc_13BA8
		move.w	(a2),d0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_13A56
		move.w	$14(a2),d0

loc_13A56:							; CODE XREF: sub_13A1A+34j
		sub.w	d2,d0
		cmpnw	 $FFF0,d0
		blt.w	loc_13BA8
		cmpnw	 $10,d0
		bgt.w	loc_13BA8
		muls.w	d5,d5
		muls.w	d0,d0
		add.l	d5,d0
		cmpnl	 $100,d0
		bgt.w	loc_13BA8
		move.w	#0,($FFFFBFA2).w
		move.b	$62(a3),d6
		move.b	$62(a2),d0
		eor.b	d0,d6
		move.w	$28(a3),d0
		sub.w	$28(a2),d0
		move.w	$2A(a3),d1
		sub.w	$2A(a2),d1
		move.w	(a3),d2
		sub.w	(a2),d2
		neg.w	d2
		move.w	$14(a3),d3
		sub.w	$14(a2),d3
		neg.w	d3
		movem.w	d0-d1,-(sp)
		muls.w	d3,d1
		muls.w	d2,d0
		add.l	d0,d1
		bmi.w	loc_13BAE
		asr.l	#4,d1
		btst	#6,d6
		beq.w	loc_13B22
		move.w	d1,d4
		lsr.w	#8,d4
		cmpnw	 5,d4
		bgt.w	loc_13ACE
		moveq	#5,d4

loc_13ACE:							; CODE XREF: sub_13A1A+AEj
		add.w	d4,$32(a3)
		add.w	d4,$32(a2)
		btst	#0,$63(a2)
		bne.w	loc_13AE6
		move.w	$52(a3),$2E(a2)

loc_13AE6:							; CODE XREF: sub_13A1A+C2j
		btst	#0,$63(a3)
		bne.w	loc_13AF6
		move.w	$52(a2),$2E(a3)

loc_13AF6:							; CODE XREF: sub_13A1A+D2j
		cmpnw	 $14,d4
		blt.w	loc_13B16
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		beq.w	loc_13B12
		cmp.w	$52(a2),d0
		bne.w	loc_13B16

loc_13B12:							; CODE XREF: sub_13A1A+ECj
		bsr.w	sub_13BB2

loc_13B16:							; CODE XREF: sub_13A1A+E0j
					; sub_13A1A+F4j
		bsr.w	sub_13BD6
		bsr.w	sub_13CC4
		bsr.w	nullsub_1

loc_13B22:							; CODE XREF: sub_13A1A+A2j
		move.w	d1,d4
		movem.w	(sp)+,d0-d1
		tst.w	($FFFFBFA0).w
		bmi.w	loc_13BA8
		muls.w	d2,d1
		muls.w	d3,d0
		sub.l	d1,d0
		asr.l	#4,d0
		move.w	d0,d5
		clr.w	d0
		move.b	$67(a3),d0
		add.w	#$8C,d0					; ''
		clr.w	d1
		move.b	$67(a2),d1
		add.w	#$8C,d1					; ''
		move.w	($FFFFBFA2).w,d7
		mulu.w	d1,d7
		lsr.w	#4,d7
		add.w	d0,d1
		sub.w	d7,d0
		muls.w	d4,d0
		divs.w	d1,d0
		move.w	($FFFFBFA2).w,d1
		muls.w	d4,d1
		asr.w	#4,d1
		add.w	d0,d1
		movem.w	d2-d3,-(sp)
		muls.w	d5,d3
		muls.w	d0,d2
		add.l	d2,d3
		asr.l	#4,d3
		add.w	$28(a2),d3
		move.w	d3,$28(a3)
		movem.w	(sp),d2-d3
		muls.w	d5,d2
		muls.w	d0,d3
		sub.l	d2,d3
		asr.l	#4,d3
		add.w	$2A(a2),d3
		move.w	d3,$2A(a3)
		movem.w	(sp)+,d2-d3
		muls.w	d1,d2
		asr.l	#4,d2
		add.w	d2,$28(a2)
		muls.w	d1,d3
		asr.l	#4,d3
		add.w	d3,$2A(a2)
		st	($FFFFBFA0).w

loc_13BA8:							; CODE XREF: sub_13A1A+14j
					; sub_13A1A+1Ej ...
		movem.l	(sp)+,d0-a3
		rts
; ---------------------------------------------------------------------------

loc_13BAE:							; CODE XREF: sub_13A1A+98j
		addq.w	#4,sp
		bra.s	loc_13BA8
; End of function sub_13A1A


; =============== S U B	R O U T	I N E =======================================


sub_13BB2:							; CODE XREF: sub_13A1A:loc_13B12p
					; sub_14062+38Cj ...
		move.l	d0,-(sp)
		moveq	#3,d0
		bsr.w	sub_11086
		addq.w	#1,d0
		add.w	($FFFFC014).w,d0
		and.w	#3,d0
		move.w	d0,($FFFFC014).w
		add.w	#$1C,d0
		move.w	d0,-(sp)
		bsr.w	sub_11132
		move.l	(sp)+,d0
		rts
; End of function sub_13BB2


; =============== S U B	R O U T	I N E =======================================


sub_13BD6:							; CODE XREF: sub_13A1A:loc_13B16p
		move.l	d0,-(sp)
		bsr.w	sub_13BE8
		exg	a2,a3
		bsr.w	sub_13BE8
		exg	a2,a3
		move.l	(sp)+,d0
		rts
; End of function sub_13BD6


; =============== S U B	R O U T	I N E =======================================


sub_13BE8:							; CODE XREF: sub_13BD6+2p sub_13BD6+8p

; FUNCTION CHUNK AT 00015464 SIZE 00000002 BYTES

		tst.w	$34(a2)
		bne.w	locret_15464
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		beq.w	loc_13C10
		cmp.w	#$19,$32(a3)
		ble.w	locret_15464

loc_13C10:							; CODE XREF: sub_13BE8+1Aj
		cmp.w	#2,$32(a3)
		ble.w	locret_15464
		exg	a2,a3
		bsr.w	sub_14062
		exg	a2,a3
		cmp.w	#$1E,$32(a2)
		ble.w	locret_15464
		btst	#4,$63(a3)
		bne.w	locret_15464
		tst.w	$34(a2)
		bne.w	locret_15464
		move.w	#1,d0
		btst	#6,$62(a3)
		beq.w	loc_13C50
		move.w	#2,d0

loc_13C50:							; CODE XREF: sub_13BE8+60j
		cmp.w	($FFFFC328).w,d0
		beq.w	loc_13C60
		cmp.w	($FFFFC32A).w,d0
		bne.w	loc_13C6A

loc_13C60:							; CODE XREF: sub_13BE8+6Cj
		btst	#3,$62(a3)
		beq.w	locret_15464

loc_13C6A:							; CODE XREF: sub_13BE8+74j
		move.w	$14(a2),d0
		btst	#7,$62(a2)
		beq.w	loc_13C7A
		neg.w	d0

loc_13C7A:							; CODE XREF: sub_13BE8+8Cj
		cmpnw	 $EF,d0					; ''
		blt.w	locret_15464
		cmpnw	 $10A,d0
		bgt.w	locret_15464
		move.w	(a2),d0
		cmpnw	 $14,d0
		bgt.w	locret_15464
		cmpnw	 $FFEC,d0
		blt.w	locret_15464
		move.w	#$28,d0					; '('
		sub.b	$73(a3),d0
		bsr.w	sub_11086
		cmpnw	 4,d0
		bhi.w	locret_15464
		btst	#4,($FFFFC2EA).w
		bne.w	locret_15464
		move.l	#$22,d0					; '"'
		bra.w	sub_11F2C
; End of function sub_13BE8


; =============== S U B	R O U T	I N E =======================================


sub_13CC4:							; CODE XREF: sub_13A1A+100p
		movem.l	d0-d4/a0-a3,-(sp)
		bsr.w	sub_13CD8
		exg	a2,a3
		bsr.w	sub_13CD8
		movem.l	(sp)+,d0-d4/a0-a3
		rts
; End of function sub_13CC4


; =============== S U B	R O U T	I N E =======================================


sub_13CD8:							; CODE XREF: sub_13CC4+4p sub_13CC4+Ap

; FUNCTION CHUNK AT 00013EE2 SIZE 00000180 BYTES

		cmp.w	#$C90,$58(a2)
		beq.w	loc_13EE2
		cmp.w	#$1122,$58(a2)
		beq.w	loc_13EE2
		cmp.w	#$B24,$58(a2)
		beq.w	loc_13FBC
		cmp.w	#$C5E,$58(a3)
		bne.w	locret_15464
		btst	#0,($FFFFC2EA).w
		beq.w	loc_13D16
		move.w	#$100,$32(a3)
		move.w	#$100,$32(a2)

loc_13D16:							; CODE XREF: sub_13CD8+2Ej
		move.w	(a2),d0
		sub.w	(a3),d0
		move.w	$14(a2),d1
		sub.w	$14(a3),d1
		bsr.w	sub_10676
		sub.w	$54(a3),d0
		and.w	#7,d0
		btst	#3,4(a3)
		beq.w	loc_13D40
		neg.w	d0
		addq.w	#8,d0
		and.w	#7,d0

loc_13D40:							; CODE XREF: sub_13CD8+5Cj
		asl.w	#1,d0
		lea	word_13E16(pc),a0
		move.w	(a0,d0.w),d1
		bset	#5,$62(a3)
		bsr.w	sub_1073A
		tst.w	$34(a2)
		beq.w	locret_15464
		cmpnw	 $14,d4
		blt.w	locret_15464
		moveq	#$78,d0					; 'x'
		btst	#3,$62(a3)
		beq.w	loc_13D72
		asl.w	#1,d0

loc_13D72:							; CODE XREF: sub_13CD8+94j
		sub.b	$67(a3),d0
		add.b	$67(a2),d0
		lsr.w	#1,d0
		exg	a2,a3
		jsr	sub_F8B5A
		exg	a2,a3
		sub.w	$32(a2),d0
		beq.w	loc_13DE4
		bmi.w	loc_13DE4
		btst	#4,$64(a3)
		bne.w	loc_13DE4
		bsr.w	sub_11086
		clr.w	($FFFFBF12).w
		move.b	$75(a3),($FFFFBF13).w
		lsr	($FFFFBF12).w
		cmp.b	($FFFFBF13).w,d0
		ble.w	loc_13DE4
		bsr.w	sub_13E26
		cmpnw	 4,d0
		bhi.w	locret_15464
		btst	#4,($FFFFC2EA).w
		bne.w	locret_15464
		move.b	($C00008).l,d0
		and.w	#2,d0
		add.w	#$16,d0
		jsr	sub_1211A
		bra.w	sub_11F2C
; ---------------------------------------------------------------------------

loc_13DE4:							; CODE XREF: sub_13CD8+B2j
					; sub_13CD8+B6j ...
		bsr.w	sub_13E26
		cmpnw	 3,d0
		bhi.w	loc_13E12
		btst	#4,($FFFFC2EA).w
		bne.w	loc_13E12
		move.b	($C00008).l,d0
		and.w	#2,d0
		add.w	#$1A,d0
		jsr	sub_1211A
		bsr.w	sub_11F2C

loc_13E12:							; CODE XREF: sub_13CD8+114j
					; sub_13CD8+11Ej
		bra.w	sub_14062
; End of function sub_13CD8

; ---------------------------------------------------------------------------
word_13E16:	dc.w	$B96					; DATA XREF: sub_13CD8+6Ao
		dc.w	$BC8
		dc.w	$C2C
		dc.w	$C2C
		dc.w	$C2C
		dc.w	$BFA
		dc.w	$BFA
		dc.w	$B96

; =============== S U B	R O U T	I N E =======================================


sub_13E26:							; CODE XREF: sub_13CD8+DEp
					; sub_13CD8:loc_13DE4p	...
		move.w	#$28,d0					; '('
		sub.b	$73(a3),d0
		lsr.b	#1,d0
		mulu.w	#$D,d0
		btst	#3,$62(a3)
		beq.w	loc_13E40
		asl.w	#1,d0

loc_13E40:							; CODE XREF: sub_13E26+14j
		move.w	(a3),d1
		sub.w	($FFFFB74A).w,d1
		cmpnw	 $28,d1					; '('
		bgt.w	loc_13E70
		cmpnw	 $FFD8,d1
		blt.w	loc_13E70
		move.w	$14(a3),d1
		sub.w	($FFFFB75E).w,d1
		cmpnw	 $28,d1					; '('
		bgt.w	loc_13E70
		cmpnw	 $FFD8,d1
		blt.w	loc_13E70
		asr.w	#1,d0

loc_13E70:							; CODE XREF: sub_13E26+24j
					; sub_13E26+2Cj ...
		bsr.w	sub_11086
		btst	#1,$64(a2)
		beq.w	loc_13EBA
		move.w	$14(a2),d1
		bpl.w	loc_13E88
		neg.w	d1

loc_13E88:							; CODE XREF: sub_13E26+5Cj
		sub.w	#$108,d1
		neg.w	d1
		move.w	#8,d0
		cmpnw	 $75,d1					; 'u'
		bgt.w	loc_13EB6
		move.w	#7,d0
		cmpnw	 $3A,d1					; ':'
		bgt.w	loc_13EB6
		move.w	#4,d0
		cmpnw	 $2C,d1					; ','
		bgt.w	loc_13EB6
		move.w	#3,d0

loc_13EB6:							; CODE XREF: sub_13E26+70j
					; sub_13E26+7Cj ...
		bsr.w	sub_11086

loc_13EBA:							; CODE XREF: sub_13E26+54j
		btst	#6,$62(a3)
		bne.w	loc_13ED2
		btst	#2,($FFFFC300).w
		bne.w	loc_13EDC
		bra.w	locret_13EE0
; ---------------------------------------------------------------------------

loc_13ED2:							; CODE XREF: sub_13E26+9Aj
		btst	#1,($FFFFC300).w
		beq.w	locret_13EE0

loc_13EDC:							; CODE XREF: sub_13E26+A4j
		move.w	#$7F,d0					; ''

locret_13EE0:							; CODE XREF: sub_13E26+A8j
					; sub_13E26+B2j
		rts
; End of function sub_13E26

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_13CD8

loc_13EE2:							; CODE XREF: sub_13CD8+6j
					; sub_13CD8+10j
		btst	#5,$62(a3)
		bne.w	locret_15464
		tst.w	$34(a3)
		beq.w	locret_15464
		btst	#0,$63(a3)
		bne.w	locret_15464
		move.w	(a3),d0
		sub.w	(a2),d0
		move.w	$14(a3),d1
		sub.w	$14(a2),d1
		bsr.w	sub_10676
		sub.w	$54(a2),d0
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bhi.w	locret_15464
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a3),d0
		bne.w	loc_13F32
		bclr	#3,($FFFFC2EC).w

loc_13F32:							; CODE XREF: sub_13CD8+250j
		move.w	$28(a3),d0
		add.w	$28(a2),d0
		asr.w	#1,d0
		move.w	d0,$28(a3)
		move.w	d0,$28(a2)
		move.w	$2A(a3),d0
		add.w	$2A(a2),d0
		asr.w	#1,d0
		move.w	d0,$2A(a3)
		move.w	d0,$2A(a2)
		bset	#5,$62(a3)
		move.w	#$CF4,d1
		bsr.w	sub_1073A
		exg	a2,a3
		bset	#5,$62(a3)
		move.w	#$CC2,d1
		cmp.w	#$C90,$58(a3)
		beq.w	loc_13F7E
		move.w	#$1154,d1

loc_13F7E:							; CODE XREF: sub_13CD8+29Ej
		bsr.w	sub_1073A
		bsr.w	sub_13E26
		cmpnw	 6,d0
		bhi.w	loc_13FB4
		btst	#4,($FFFFC2EA).w
		bne.w	loc_13FB4
		move.w	#$24,d0					; '$'
		cmp.w	#$CC2,$58(a3)
		beq.w	loc_13FAA
		move.w	#$1E,d0

loc_13FAA:							; CODE XREF: sub_13CD8+2CAj
		jsr	sub_1211A
		bsr.w	sub_11F2C

loc_13FB4:							; CODE XREF: sub_13CD8+2B2j
					; sub_13CD8+2BCj
		exg	a2,a3
		st	($FFFFBFA0).w
		rts
; ---------------------------------------------------------------------------

loc_13FBC:							; CODE XREF: sub_13CD8+1Aj
		btst	#5,$62(a3)
		bne.w	locret_15464
		tst.w	$34(a3)
		beq.w	locret_15464
		btst	#0,$63(a3)
		bne.w	locret_15464
		tst.w	($FFFFD056).w
		bne.w	loc_14018
		btst	#3,$62(a2)
		beq.w	loc_14018
		move.w	#$20,d0					; ' '
		add.b	$75(a2),d0
		sub.b	$68(a3),d0
		btst	#1,($FFFFC2FE).w
		bne.w	loc_1400A
		btst	#6,($FFFFC2FC).w
		beq.w	loc_1400C

loc_1400A:							; CODE XREF: sub_13CD8+324j
		subq.b	#2,d0

loc_1400C:							; CODE XREF: sub_13CD8+32Ej
		bsr.w	sub_11086
		cmpnw	 $18,d0
		blt.w	locret_15464

loc_14018:							; CODE XREF: sub_13CD8+304j
					; sub_13CD8+30Ej
		move.w	(a3),d0
		sub.w	(a2),d0
		move.w	$14(a3),d1
		sub.w	$14(a2),d1
		bsr.w	sub_10676
		sub.w	$54(a2),d0
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bhi.w	locret_15464
		exg	a2,a3
		bsr.w	sub_14062
		bsr.w	sub_13E26
		cmpnw	 4,d0
		bhi.w	loc_1405A
		move.w	#$20,d0					; ' '
		jsr	sub_1211A
		bsr.w	sub_11F2C

loc_1405A:							; CODE XREF: sub_13CD8+370j
		exg	a2,a3
		st	($FFFFBFA0).w
		rts
; END OF FUNCTION CHUNK	FOR sub_13CD8

; =============== S U B	R O U T	I N E =======================================


sub_14062:							; CODE XREF: sub_13BE8+34p
					; sub_13CD8:loc_13E12j	...
		cmp.w	#$B,$52(a2)
		bgt.w	locret_15464
		cmp.w	#$145C,$58(a2)
		beq.w	locret_15464
		cmp.w	#$1AF4,$58(a2)
		beq.w	locret_15464
		cmp.w	#$11E6,$58(a2)
		beq.w	locret_15464
		cmp.w	#$D26,$58(a2)
		beq.w	locret_15464
		cmp.w	#$DD8,$58(a2)
		beq.w	locret_15464
		cmp.w	#$136A,$58(a2)
		beq.w	locret_15464
		cmp.l	($FFFFDEDC).w,a2
		bne.w	loc_140BA
		cmp.l	($FFFFDEE0).w,a3
		bne.w	loc_140BA
		rts
; ---------------------------------------------------------------------------

loc_140BA:							; CODE XREF: sub_14062+4Aj
					; sub_14062+52j
		move.l	a2,($FFFFDEDC).w
		move.l	a3,($FFFFDEE0).w
		btst	#3,$64(a2)
		bne.w	locret_15464
		btst	#3,$64(a3)
		bne.w	locret_15464
		move.w	#$11E6,d1
		btst	#4,$64(a3)
		bne.w	loc_14116
		cmp.b	#$18,$71(a2)
		blt.w	loc_14116
		tst.b	$5F(a2)
		bne.w	loc_14116
		move.w	($C00008).l,d0
		and.w	#$F,d0
		add.w	#$10,d0
		cmp.w	$32(a2),d0
		ble.w	loc_14116
		move.b	#$3C,$5F(a2)				; '<'
		bra.w	loc_143C0
; ---------------------------------------------------------------------------

loc_14116:							; CODE XREF: sub_14062+7Ej
					; sub_14062+88j ...
		tst.w	$34(a3)
		beq.w	loc_14150
		move.w	#$C6CE,a0
		btst	#6,$62(a3)
		beq.w	loc_14130
		add.w	#$364,a0

loc_14130:							; CODE XREF: sub_14062+C6j
		addq.w	#1,$10(a0)
		clr.w	d0
		move.b	$66(a3),d0
		add.w	d0,a0
		addq.b	#1,$11C(a0)
		addq.w	#1,($FFFFC46E).w
		tst.b	$74(a3)
		bne.w	loc_14150
		addq.w	#2,($FFFFC46E).w

loc_14150:							; CODE XREF: sub_14062+B8j
					; sub_14062+E6j
		move.b	#$78,$5E(a2)				; 'x'
		move.w	(a3),d0
		sub.w	(a2),d0
		move.w	$14(a3),d1
		sub.w	$14(a2),d1
		bsr.w	sub_10676
		tst.w	$34(a2)
		beq.w	loc_14378
		jsr	sub_F8B5A
		btst	#4,$64(a2)
		beq.w	loc_14378
		movem.l	d1-d4,-(sp)
		move.w	$28(a2),d1
		bpl.w	loc_1418C
		neg.w	d1

loc_1418C:							; CODE XREF: sub_14062+124j
		move.w	$2A(a2),d2
		bpl.w	loc_14196
		neg.w	d2

loc_14196:							; CODE XREF: sub_14062+12Ej
		move.w	(a2),d3
		sub.w	(a3),d3
		move.w	$14(a2),d4
		sub.w	$14(a3),d4
		cmp.w	#$108,$14(a3)
		bgt.w	loc_141EA
		cmp.w	#$FEF8,$14(a3)
		blt.w	loc_141E0
		tst.w	(a3)
		bpl.w	loc_141CE
		tst.w	d3
		bpl.w	loc_141F4
		cmpnw	 $FFFD,d3
		bgt.w	loc_141F4
		bra.w	loc_141FC
; ---------------------------------------------------------------------------

loc_141CE:							; CODE XREF: sub_14062+156j
		tst.w	d3
		bmi.w	loc_141F4
		cmpnw	 3,d3
		blt.w	loc_141F4
		bra.w	loc_141FC
; ---------------------------------------------------------------------------

loc_141E0:							; CODE XREF: sub_14062+150j
		tst.w	d4
		bpl.w	loc_141F4
		bra.w	loc_141FC
; ---------------------------------------------------------------------------

loc_141EA:							; CODE XREF: sub_14062+146j
		tst.w	d4
		bmi.w	loc_141F4
		bra.w	loc_141FC
; ---------------------------------------------------------------------------

loc_141F4:							; CODE XREF: sub_14062+15Cj
					; sub_14062+164j ...
		movem.l	(sp)+,d1-d4
		bra.w	loc_14378
; ---------------------------------------------------------------------------

loc_141FC:							; CODE XREF: sub_14062+168j
					; sub_14062+17Aj ...
		movem.l	(sp)+,d1-d4
		bra.w	*+4
		cmp.w	#$10C,$14(a2)
		bgt.w	loc_1422C
		cmp.w	#$FEF4,$14(a2)
		blt.w	loc_1422C
		cmp.w	#$68,(a2)				; 'h'
		bgt.w	loc_1422C
		cmp.w	#$FF98,(a2)
		blt.w	loc_1422C
		bra.w	loc_14378
; ---------------------------------------------------------------------------

loc_1422C:							; CODE XREF: sub_14062+1A8j
					; sub_14062+1B2j ...
		cmp.w	#$B,$52(a3)
		bgt.w	loc_14264
		add.w	#$A,($FFFFB8A2).w
		add.w	#$96,($FFFFB89C).w			; ''
		move.w	d1,-(sp)
		move.b	#0,$40(a3)
		move.b	#0,$65(a3)
		move.w	#$1AC2,d1
		jsr	sub_1073A
		clr.w	$28(a3)
		clr.w	$2A(a3)
		move.w	(sp)+,d1

loc_14264:							; CODE XREF: sub_14062+1D0j
		movem.l	d0/a0,-(sp)
		cmp.w	#$10C,$14(a2)
		bgt.w	loc_14292
		cmp.w	#$FEF4,$14(a2)
		blt.w	loc_1429A
		tst.w	(a2)
		bmi.w	loc_1428A
		move.w	#2,d0
		bra.w	loc_1429E
; ---------------------------------------------------------------------------

loc_1428A:							; CODE XREF: sub_14062+21Cj
		move.w	#6,d0
		bra.w	loc_1429E
; ---------------------------------------------------------------------------

loc_14292:							; CODE XREF: sub_14062+20Cj
		move.w	#0,d0
		bra.w	loc_1429E
; ---------------------------------------------------------------------------

loc_1429A:							; CODE XREF: sub_14062+216j
		move.w	#4,d0

loc_1429E:							; CODE XREF: sub_14062+224j
					; sub_14062+22Cj ...
		add.w	d0,d0
		move.l	#word_14368,a0
		move.w	(a0,d0.w),d1
		move.w	$14(a2),($FFFFD41A).w
		move.w	(a2),($FFFFD41C).w
		cmpnw	 $18CC,d1
		bne.w	loc_142F4
		cmp.w	#$59,$14(a2)				; 'Y'
		bgt.w	loc_142F4
		cmp.w	#$FFA7,$14(a2)
		blt.w	loc_142F4
		move.w	d0,-(sp)
		move.w	#7,d0
		jsr	sub_FE510
		move.w	(sp)+,d0
		move.w	#$193E,d1
		btst	#3,4(a2)
		beq.w	loc_1434C
		move.w	#$1A00,d1
		bra.w	loc_1434C
; ---------------------------------------------------------------------------

loc_142F4:							; CODE XREF: sub_14062+256j
					; sub_14062+260j ...
		cmpnw	 $17E8,d1
		bne.w	loc_14326
		cmp.w	#$38,$14(a2)				; '8'
		bgt.w	loc_14326
		cmp.w	#$FFC8,$14(a2)
		blt.w	loc_14326
		move.w	#$1A00,d1
		btst	#3,4(a2)
		beq.w	loc_1434C
		move.w	#$193E,d1
		bra.w	loc_1434C
; ---------------------------------------------------------------------------

loc_14326:							; CODE XREF: sub_14062+296j
					; sub_14062+2A0j ...
		btst	#3,4(a2)
		beq.w	loc_1434C
		cmpnw	 $17E8,d1
		beq.w	loc_14348
		cmpnw	 $18CC,d1
		bne.w	loc_1434C
		move.w	#$17E8,d1
		bra.w	loc_1434C
; ---------------------------------------------------------------------------

loc_14348:							; CODE XREF: sub_14062+2D2j
		move.w	#$18CC,d1

loc_1434C:							; CODE XREF: sub_14062+286j
					; sub_14062+28Ej ...
		clr.w	$28(a2)
		clr.w	$2A(a2)
		bset	#5,$64(a2)
		bset	#0,$62(a2)
		movem.l	(sp)+,d0/a0
		bra.w	loc_14392
; ---------------------------------------------------------------------------
word_14368:	dc.w	$1776					; DATA XREF: sub_14062+23Eo
		dc.w	$1776
		dc.w	$17E8
		dc.w	$17E8
		dc.w	$185A
		dc.w	$185A
		dc.w	$18CC
		dc.w	$18CC
; ---------------------------------------------------------------------------

loc_14378:							; CODE XREF: sub_14062+108j
					; sub_14062+118j ...
		move.w	#$DD8,d1
		sub.w	$54(a2),d0
		addq.w	#1,d0
		and.w	#7,d0
		cmpnw	 2,d0
		bls.w	loc_143C0
		move.w	#$D26,d1

loc_14392:							; CODE XREF: sub_14062+302j
		move.w	$54(a2),d0
		and.w	#3,d0
		bne.w	loc_143C0
		cmp.b	#$14,$75(a3)
		blt.w	loc_143C0
		cmp.w	#$B,$52(a3)
		bgt.w	loc_143C0
		btst	#5,$64(a2)
		bne.w	loc_143C0
		move.w	#$136A,d1

loc_143C0:							; CODE XREF: sub_14062+B0j
					; sub_14062+328j ...
		exg	a2,a3
		bset	#5,$62(a3)
		bsr.w	sub_1073A
		exg	a2,a3
		tst.w	$34(a3)
		beq.w	loc_143EA
		tst.w	$34(a2)
		beq.w	loc_143EA
		add.w	#$12C,($FFFFB89C).w
		add.w	#$F,($FFFFB8A2).w

loc_143EA:							; CODE XREF: sub_14062+370j
					; sub_14062+378j
		move.w	($FFFFB7AA).w,d0
		bmi.w	sub_13BB2
		cmp.w	$52(a2),d0
		bne.w	sub_13BB2
		st	($FFFFB7AA).w
		btst	#4,($FFFFC2EA).w
		bne.w	loc_14494
		cmp.w	#$DD8,$58(a2)
		bne.w	loc_14494
		move.w	$54(a2),d0
		and.w	#3,d0
		bne.w	loc_14494
		btst	#4,$63(a2)
		bne.w	loc_14494
		move.l	#$A0,d0					; ''
		bsr.w	sub_11086
		cmp.w	$32(a2),d0
		bgt.w	loc_14494
		btst	#0,($FFFFC2EA).w
		bne.w	loc_14494
		exg	a2,a3
		move.w	#$145C,d1
		bsr.w	sub_1073A
		exg	a2,a3
		bsr.w	sub_144AC
		btst	#5,($FFFFC2FC).w
		bne.w	loc_1446A
		exg	a2,a3
		move.w	#$1AF4,d1
		bsr.w	sub_1073A
		exg	a2,a3

loc_1446A:							; CODE XREF: sub_14062+3F8j
		move.w	#4,($FFFFC3EC).w
		move.l	#$14,d0
		tst.w	$34(a3)
		beq.w	sub_11F62
		tst.w	($FFFFD056).w
		beq.w	sub_11F62
		move.l	#$12,d0
		bsr.w	sub_11F62
		bra.w	sub_124A6
; ---------------------------------------------------------------------------

loc_14494:							; CODE XREF: sub_14062+3A2j
					; sub_14062+3ACj ...
		move.w	#$B,-(sp)
		btst	#6,$62(a2)
		bne.w	loc_144A6
		move.w	#$C,(sp)

loc_144A6:							; CODE XREF: sub_14062+43Cj
		bsr.w	sub_11156
		rts
; End of function sub_14062


; =============== S U B	R O U T	I N E =======================================


sub_144AC:							; CODE XREF: sub_14062+3EEp
		move.w	d0,-(sp)
		bset	#2,$63(a2)
		add.w	#$12C,($FFFFB89C).w
		add.w	#$1E,($FFFFB8A2).w
		move.w	#$D,-(sp)
		bsr.w	sub_11132
		move.w	(a2),($FFFFBF8E).w
		move.w	$14(a2),($FFFFBF90).w
		bset	#6,($FFFFC2EC).w
		clr.w	d1
		move.w	#$C6CE,a0
		btst	#6,$62(a2)
		beq.w	loc_144F0
		move.w	#$8000,d1
		add.w	#$364,a0

loc_144F0:							; CODE XREF: sub_144AC+38j
		move.b	$66(a2),d1
		move.w	d1,($FFFFC470).w
		ext.w	d1
		add.w	d1,d1
		exg	a2,a3
		jsr	sub_FEC98
		exg	a2,a3
		tst.w	d0
		bne.w	loc_1451C
		bclr	#5,($FFFFC2FC).w
		move.w	#$FFFD,$66(a0,d1.w)
		bra.w	loc_14546
; ---------------------------------------------------------------------------

loc_1451C:							; CODE XREF: sub_144AC+5Cj
		cmpnw	 3,d0
		beq.w	loc_1453A
		bclr	#5,($FFFFC2FC).w
		move.w	#$FFFD,$66(a0,d1.w)
		jsr	sub_FECA2
		beq.w	loc_14546

loc_1453A:							; CODE XREF: sub_144AC+74j
		bset	#5,($FFFFC2FC).w
		move.w	#$FFFC,$66(a0,d1.w)

loc_14546:							; CODE XREF: sub_144AC+6Cj
					; sub_144AC+8Aj
		move.w	(sp)+,d0
		rts
; End of function sub_144AC


; =============== S U B	R O U T	I N E =======================================


nullsub_1:							; CODE XREF: sub_13A1A+104p
		rts
; End of function nullsub_1


; =============== S U B	R O U T	I N E =======================================


sub_1454C:							; CODE XREF: sub_138AC+30p
					; sub_138AC+6Ap
		bclr	#4,$64(a3)
		move.w	#$88,d4					; ''
		sub.w	($FFFFBD22).w,d4
		move.w	#$12A,d5
		sub.w	($FFFFBD24).w,d5
		movem.w	d2-d5,-(sp)
		neg.w	d4
		neg.w	d5
		add.w	#$40,d4					; '@'
		add.w	#$40,d5					; '@'
		cmp.w	d5,d3
		bgt.w	loc_14592
		cmp.w	d4,d2
		blt.w	loc_145B4
		neg.w	d4
		cmp.w	d4,d2
		bgt.w	loc_145B4
		move.w	#$B6CA,a2
		bsr.w	sub_14620
		bra.w	loc_145E6
; ---------------------------------------------------------------------------

loc_14592:							; CODE XREF: sub_1454C+28j
		neg.w	d5
		cmp.w	d5,d3
		blt.w	loc_145E6
		cmp.w	d4,d2
		blt.w	loc_145B4
		neg.w	d4
		cmp.w	d4,d2
		bgt.w	loc_145B4
		move.w	#$B64A,a2
		bsr.w	sub_14620
		bra.w	loc_145E6
; ---------------------------------------------------------------------------

loc_145B4:							; CODE XREF: sub_1454C+2Ej
					; sub_1454C+36j ...
		sub.w	d4,d2
		sub.w	d5,d3
		move.w	d3,d0
		move.w	d2,d1
		neg.w	d1
		muls.w	d3,d3
		muls.w	d2,d2
		add.l	d2,d3
		cmpnl	 $1000,d3
		bls.w	loc_145E6
		exg	d0,d3
		bsr.w	sub_110BE
		exg	d0,d3
		ext.l	d0
		asl.l	#8,d0
		divs.w	d3,d0
		ext.l	d1
		asl.l	#8,d1
		divs.w	d3,d1
		bsr.w	sub_14BC2

loc_145E6:							; CODE XREF: sub_1454C+42j
					; sub_1454C+4Aj ...
		movem.w	(sp)+,d2-d5
		move.w	$4E(a3),d0
		or.w	$50(a3),d0
		bne.w	locret_15464
		move.w	#$100,d0
		clr.w	d1
		cmp.w	d5,d3
		bge.w	sub_14BC2
		neg.w	d5
		neg.w	d0
		cmp.w	d5,d3
		ble.w	sub_14BC2
		exg	d0,d1
		cmp.w	d4,d2
		bge.w	sub_14BC2
		neg.w	d4
		neg.w	d1
		cmp.w	d4,d2
		ble.w	sub_14BC2
		rts
; End of function sub_1454C


; =============== S U B	R O U T	I N E =======================================


sub_14620:							; CODE XREF: sub_1454C+3Ep
					; sub_1454C+60p

; FUNCTION CHUNK AT 00014AA8 SIZE 00000090 BYTES

		cmp.w	#$D,$18(a3)
		bgt.w	locret_15464
		cmp.w	#$E,$52(a3)
		bne.w	loc_14AA8
		sub.w	(a2),d2
		moveq	#$10,d4
		add.w	($FFFFBD22).w,d4
		cmp.w	d4,d2
		bgt.w	locret_15464
		neg.w	d4
		cmp.w	d4,d2
		blt.w	locret_15464
		sub.w	$14(a2),d3
		move.w	#2,d5
		add.w	($FFFFBD24).w,d5
		cmp.w	d5,d3
		bgt.w	locret_15464
		neg.w	d5
		cmp.w	d5,d3
		blt.w	locret_15464
		st	($FFFFBFA0).w
		cmp.w	#$D,$24(a3)
		blt.w	loc_1467C
		move.w	$24(a3),$18(a3)
		bra.w	loc_1477C
; ---------------------------------------------------------------------------

loc_1467C:							; CODE XREF: sub_14620+4Ej
		bclr	#7,$62(a3)
		move.w	($FFFFB7AA).w,d0
		bmi.w	loc_146B6
		st	($FFFFB7AA).w
		asl.w	#7,d0
		move.w	#$B04A,a0
		move.b	#8,$5E(a0,d0.w)
		move.w	($FFFFB75E).w,d1
		btst	#7,$62(a0,d0.w)
		bne.w	loc_146AA
		neg.w	d1

loc_146AA:							; CODE XREF: sub_14620+84j
		tst.w	d1
		bpl.w	loc_146B6
		bset	#7,$62(a3)

loc_146B6:							; CODE XREF: sub_14620+66j
					; sub_14620+8Cj
		move.w	#$FF00,d0
		move.l	$14(a3),d1
		sub.l	$20(a3),d1
		asr.l	#8,d1
		beq.w	loc_1478A
		bmi.w	loc_146D0
		neg.w	d0
		neg.w	d5

loc_146D0:							; CODE XREF: sub_14620+A8j
		add.w	d3,d5
		move.l	(a3),d3
		sub.l	$1C(a3),d3
		asr.l	#8,d3
		muls.w	d3,d5
		divs.w	d1,d5
		bvs.w	loc_1478A
		sub.w	d5,d2
		cmp.w	d4,d2
		blt.w	loc_1478A
		neg.w	d4
		cmp.w	d4,d2
		bgt.w	loc_1478A
		clr.w	d1
		move.w	$14(a3),d3
		eor.w	d0,d3
		bmi.w	sub_14C70
		neg.w	d0
		btst	#7,$62(a3)
		bne.w	sub_14C70
		cmp.w	#$D,$18(a3)
		beq.w	loc_14724
		subq.w	#1,d4
		cmp.w	d4,d2
		bgt.w	loc_14724
		neg.w	d4
		cmp.w	d4,d2
		bge.w	loc_147A0

loc_14724:							; CODE XREF: sub_14620+F0j
					; sub_14620+F8j
		bsr.w	sub_12F30
		move.w	#$25,-(sp)				; '%'
		btst	#0,($FFFFC2EA).w
		bne.w	loc_14746
		move.w	#8,(sp)
		add.w	#$12C,($FFFFB89C).w
		add.w	#$28,($FFFFB8A2).w			; '('

loc_14746:							; CODE XREF: sub_14620+112j
		bsr.w	sub_11132
		move.w	#$1000,d0
		bsr.w	sub_11086
		tst.w	$14(a3)
		bmi.w	loc_1475C
		neg.w	d0

loc_1475C:							; CODE XREF: sub_14620+136j
		move.w	d0,$2A(a3)
		move.w	#$1000,d0
		bsr.w	sub_1107A
		move.w	d0,$28(a3)
		move.w	#$1000,d0
		bsr.w	sub_1107A
		move.w	d0,$2C(a3)
		bra.w	sub_102D2
; ---------------------------------------------------------------------------

loc_1477C:							; CODE XREF: sub_14620+58j
		neg.w	$2C(a3)
		bpl.w	locret_15464
		neg.w	$2C(a3)
		rts
; ---------------------------------------------------------------------------

loc_1478A:							; CODE XREF: sub_14620+A4j
					; sub_14620+BEj ...
		clr.w	d0
		move.w	#$100,d1
		move.w	(a3),d2
		sub.w	$1C(a3),d2
		bmi.w	sub_14C70
		neg.w	d1
		bra.w	sub_14C70
; ---------------------------------------------------------------------------

loc_147A0:							; CODE XREF: sub_14620+100j
		btst	#2,($FFFFC2F2).w
		beq.w	loc_1481E
		btst	#0,($FFFFC2FA).w
		bne.w	loc_147D8
		movem.l	d0/a0,-(sp)
		move.w	($FFFFD406).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a0
		add.w	d0,a0
		btst	#7,$62(a0)
		movem.l	(sp)+,d0/a0
		bne.w	loc_147E0
		bra.w	loc_147EC
; ---------------------------------------------------------------------------

loc_147D8:							; CODE XREF: sub_14620+190j
		tst.w	($FFFFD594).w
		bne.w	loc_147EC

loc_147E0:							; CODE XREF: sub_14620+1B0j
		tst.w	($FFFFB75E).w
		bmi.w	locret_15464
		bra.w	loc_147F4
; ---------------------------------------------------------------------------

loc_147EC:							; CODE XREF: sub_14620+1B4j
					; sub_14620+1BCj
		tst.w	($FFFFB75E).w
		bpl.w	locret_15464

loc_147F4:							; CODE XREF: sub_14620+1C8j
		bset	#5,($FFFFC2F4).w
		tst.w	($FFFFD594).w
		beq.w	loc_1480A
		addq.w	#1,($FFFFD576).w
		bra.w	loc_1480E
; ---------------------------------------------------------------------------

loc_1480A:							; CODE XREF: sub_14620+1DEj
		addq.w	#1,($FFFFD574).w

loc_1480E:							; CODE XREF: sub_14620+1E6j
		bset	#0,($FFFFC2FE).w
		jsr	sub_F37C
		bra.w	loc_14828
; ---------------------------------------------------------------------------

loc_1481E:							; CODE XREF: sub_14620+186j
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464

loc_14828:							; CODE XREF: sub_14620+1FAj
		bset	#0,($FFFFC2F4).w
		bclr	#3,($FFFFC2FE).w
		bsr.w	sub_12F30
		bsr.w	sub_1A304
		move.w	d0,-(sp)
		move.w	($FFFFB03E).w,d0

loc_14842:							; CODE XREF: sub_14620+226j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_14842
		move.w	(sp)+,d0
		move.w	#0,-(sp)
		bsr.w	sub_11132
		jsr	sub_AFB6
		add.w	#$1F4,($FFFFB89C).w
		move.w	#$C6CE,a2
		lea	$364(a2),a1
		tst.w	$14(a3)
		bpl.w	loc_14870
		dc.w	$C549					; exg.l	a2,a1

loc_14870:							; CODE XREF: sub_14620+24Aj
		btst	#1,($FFFFC2EA).w
		beq.w	loc_1487C
		dc.w	$C549					; exg.l	a2,a1

loc_1487C:							; CODE XREF: sub_14620+256j
		addq.w	#1,$C(a2)
		btst	#5,($FFFFC2F4).w
		beq.w	loc_14898
		btst	#0,($FFFFC2FA).w
		bne.w	loc_14898
		addq.w	#1,$362(a2)

loc_14898:							; CODE XREF: sub_14620+266j
					; sub_14620+270j
		bclr	#4,($FFFFC2FA).w
		beq.w	loc_148A6
		addq.w	#1,$35A(a2)

loc_148A6:							; CODE XREF: sub_14620+27Ej
		btst	#5,($FFFFC2EE).w
		beq.w	loc_148D4
		btst	#6,($FFFFC2EE).w
		bne.w	loc_148CC
		cmp.l	#$FFFFCA32,a2
		bne.w	loc_148D4

loc_148C4:							; CODE XREF: sub_14620+2B2j
		addq.w	#1,$356(a2)
		bra.w	loc_148D4
; ---------------------------------------------------------------------------

loc_148CC:							; CODE XREF: sub_14620+296j
		cmp.l	#$FFFFC6CE,a2
		beq.s	loc_148C4

loc_148D4:							; CODE XREF: sub_14620+28Cj
					; sub_14620+2A0j ...
		movem.l	d0/a2,-(sp)
		move.w	($FFFFC466).w,d0
		add.w	d0,d0
		add.w	d0,a2
		addq.w	#1,$342(a2)
		movem.l	(sp)+,d0/a2
		bclr	#7,($FFFFC2FE).w
		beq.w	loc_148F6
		addq.w	#1,$35E(a2)

loc_148F6:							; CODE XREF: sub_14620+2CEj
		cmp.w	#$C6CE,a2
		bne.w	loc_1491E
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#3,($FFFFD6CC).w
		jsr	sub_FE556
		move.w	#$78,($FFFFDECC).w			; 'x'
		move.w	($FFFFD6C8).w,-(sp)
		move.w	(sp)+,($FFFFDECE).w

loc_1491E:							; CODE XREF: sub_14620+2DAj
		cmp.w	#$168,($FFFFC472).w
		bne.w	loc_1492C
		subq.w	#6,($FFFFC472).w

loc_1492C:							; CODE XREF: sub_14620+304j
		move.w	#$C474,a0
		add.w	($FFFFC472).w,a0
		addq.w	#6,($FFFFC472).w
		bsr.w	sub_14A94
		move.w	d0,(a0)+
		moveq	#2,d0
		add.w	$24(a2),d0
		sub.w	$24(a1),d0
		move.b	d0,(a0)+
		add.w	#$1E,($FFFFB8A2).w
		cmp.w	#$C6CE,a2
		beq.w	loc_14964
		sub.w	#$14,($FFFFB8A2).w
		bset	#7,-1(a0)

loc_14964:							; CODE XREF: sub_14620+334j
		move.w	$18(a2),d0
		move.b	d0,(a0)+
		move.w	#$FFFF,(a0)
		add.w	#$B4,d0					; ''
		addq.b	#1,(a2,d0.w)
		move.w	$1A(a2),d0
		bmi.w	loc_149AA
		cmp.w	$18(a2),d0
		beq.w	loc_149AA
		move.b	d0,(a0)+
		add.w	#$CE,d0					; ''
		addq.b	#1,(a2,d0.w)
		move.w	$1C(a2),d0
		bmi.w	loc_149AA
		cmp.w	$18(a2),d0
		beq.w	loc_149AA
		move.b	d0,(a0)
		add.w	#$CE,d0					; ''
		addq.b	#1,(a2,d0.w)

loc_149AA:							; CODE XREF: sub_14620+35Aj
					; sub_14620+362j ...
		move.w	$26(a1),d0
		bmi.w	loc_149BA
		add.w	#$B4,d0					; ''
		addq.b	#1,(a1,d0.w)

loc_149BA:							; CODE XREF: sub_14620+38Ej
		bsr.w	sub_12F30
		bsr.w	sub_1284A
		bclr	#3,($FFFFC2F2).w
		bsr.w	loc_12C04
		move.w	#$2710,($FFFFC304).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_149E0
		bra.w	loc_149EA
; ---------------------------------------------------------------------------

loc_149E0:							; CODE XREF: sub_14620+3B8j
		move.l	#7,d0
		bsr.w	sub_14A54

loc_149EA:							; CODE XREF: sub_14620+3BCj
		clr.w	($FFFFBFA0).w
		clr.w	$28(a3)
		clr.w	$2A(a3)
		moveq	#6,d0
		tst.w	(a3)
		bpl.w	loc_14A00
		neg.w	d0

loc_14A00:							; CODE XREF: sub_14620+3DAj
		move.w	d0,(a3)
		move.w	#$110,d0
		tst.w	$14(a3)
		bpl.w	loc_14A10
		neg.w	d0

loc_14A10:							; CODE XREF: sub_14620+3EAj
		move.w	d0,$14(a3)
		move.w	#$600,$2C(a3)
		clr.w	$18(a3)
		st	($FFFFBEE8).w
		st	($FFFFBEEC).w
		bset	#2,$62(a3)
		move.w	#$1A,d0
		bsr.w	sub_10662
		move.l	a3,-(sp)
		add.w	#$80,a3					; ''
		move.w	#$102E,d1
		bsr.w	sub_1073A
		move.w	$22(a1),a3
		move.l	#$E,d0
		bsr.w	sub_11F62
		move.l	(sp)+,a3
		rts
; End of function sub_14620


; =============== S U B	R O U T	I N E =======================================


sub_14A54:							; CODE XREF: sub_14620+3C6p
		move.l	a3,-(sp)
		move.w	$22(a2),a3
		moveq	#5,d3

loc_14A5C:							; CODE XREF: sub_14A54+38j
		tst.w	$34(a3)
		ble.w	loc_14A88
		btst	#0,$63(a3)
		bne.w	loc_14A88
		bclr	#2,$62(a3)
		bclr	#3,$64(a3)
		beq.w	loc_14A84
		jsr	sub_FEFF0

loc_14A84:							; CODE XREF: sub_14A54+26j
		bsr.w	sub_10658

loc_14A88:							; CODE XREF: sub_14A54+Cj
					; sub_14A54+16j
		add.w	#$80,a3					; ''
		dbf	d3,loc_14A5C
		move.l	(sp)+,a3
		rts
; End of function sub_14A54


; =============== S U B	R O U T	I N E =======================================


sub_14A94:							; CODE XREF: sub_12002+29Cp
					; sub_14620+318p
		move.w	($FFFFC466).w,d0
		swap	d0
		clr.w	d0
		lsr.l	#2,d0
		or.w	($FFFFC46C).w,d0
		sub.w	($FFFFC468).w,d0
		rts
; End of function sub_14A94

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_14620

loc_14AA8:							; CODE XREF: sub_14620+10j
		btst	#5,$63(a3)
		bne.w	locret_15464
		cmp.w	#$A,$18(a3)
		bgt.w	locret_15464
		cmp.w	#$B,$52(a3)
		bgt.w	locret_15464
		movem.w	d2-d3,-(sp)
		sub.w	$14(a2),d3
		move.w	d3,d0
		sub.w	(a2),d2
		move.w	d2,d1
		neg.w	d0
		asl.w	#4,d2
		muls.w	d2,d2
		divu.w	#$400,d2
		cmpnw	 $100,d2
		bhi.w	loc_14B32
		asl.w	#4,d3
		muls.w	d3,d3
		divu.w	#$79,d3					; 'y'
		add.w	d2,d3
		cmpnw	 $100,d3
		bhi.w	loc_14B32
		movem.w	(sp)+,d2-d3
		bsr.w	sub_14B38
		movem.w	d2-d3,-(sp)
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		bsr.w	sub_110BE
		move.w	d0,d2
		movem.w	(sp)+,d0-d1
		addq.w	#1,d2
		asl.l	#8,d0
		divs.w	d2,d0
		asl.l	#8,d1
		divs.w	d2,d1
		bset	#4,($FFFFC2F8).w
		bsr.w	sub_14C70
		bclr	#4,($FFFFC2F8).w

loc_14B32:							; CODE XREF: sub_14620+4C2j
					; sub_14620+4D4j
		movem.w	(sp)+,d2-d3
		rts
; END OF FUNCTION CHUNK	FOR sub_14620

; =============== S U B	R O U T	I N E =======================================


sub_14B38:							; CODE XREF: sub_14620+4DCp
		movem.l	d0-d1,-(sp)
		move.w	($C00008).l,d0
		and.w	#$1F,d0
		bne.w	loc_14B8A
		move.w	($FFFFB75E).w,d0
		sub.w	$14(a2),d0
		cmpnw	 $28,d0					; '('
		bgt.w	loc_14B8A
		cmpnw	 $FFD8,d0
		blt.w	loc_14B8A
		move.w	$28(a3),d0
		move.w	$2A(a3),d1
		cmpnw	 $2000,d0
		bgt.w	loc_14B90
		cmpnw	 $E000,d0
		blt.w	loc_14B90
		cmpnw	 $2000,d1
		bgt.w	loc_14B90
		cmpnw	 $E000,d1
		blt.w	loc_14B90

loc_14B8A:							; CODE XREF: sub_14B38+Ej
					; sub_14B38+1Ej ...
		movem.l	(sp)+,d0-d1
		rts
; ---------------------------------------------------------------------------

loc_14B90:							; CODE XREF: sub_14B38+36j
					; sub_14B38+3Ej ...
		add.w	#$C,sp
		asr.w	#2,d0
		asr.w	#2,d1
		move.w	d0,$28(a2)
		move.w	d1,$2A(a2)
		clr.w	$28(a3)
		clr.w	$2A(a3)
		bset	#6,($FFFFC2EC).w
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		move.l	#8,d0
		bra.w	sub_11F62
; End of function sub_14B38


; =============== S U B	R O U T	I N E =======================================


sub_14BC2:							; CODE XREF: sub_1454C+96p
					; sub_1454C+B2j ...
		cmp.w	#$E,$52(a3)
		bne.w	sub_14C70
		cmp.w	#$1D,$18(a3)
		bgt.w	loc_14C2C
		cmp.w	#$12,$18(a3)
		bls.w	sub_14C70
		cmp.w	#$118,$14(a3)
		blt.w	loc_14C2C
		cmp.w	#$D,(a3)
		blt.w	sub_14C70
		cmp.w	#$F,(a3)
		bgt.w	sub_14C70
		cmp.w	#$FA0,$2A(a3)
		blt.w	sub_14C70
		move.l	a3,-(sp)
		asr	$2A(a3)
		add.w	#$80,a3					; ''
		move.w	#$1078,d1
		bsr.w	sub_1073A
		move.l	(sp)+,a3
		move.w	#$E,-(sp)
		bsr.w	sub_11132
		add.w	#$258,($FFFFB89C).w
		add.w	#$F,($FFFFB8A2).w

loc_14C2C:							; CODE XREF: sub_14BC2+10j
					; sub_14BC2+24j
		bset	#6,($FFFFC2EC).w
		bset	#2,$62(a3)
		tst.w	$14(a3)
		bpl.w	loc_14C46
		or.w	#$8000,4(a3)

loc_14C46:							; CODE XREF: sub_14BC2+7Aj
		clr.w	$86(a3)
		btst	#0,($FFFFC2EA).w
		bne.w	locret_15464
		move.l	a3,-(sp)
		move.w	($FFFFBF7A).w,d0
		asl.w	#7,d0
		move.w	#$B04A,a3
		add.w	d0,a3
		move.l	#6,d0
		bsr.w	sub_11F62
		move.l	(sp)+,a3
		rts
; End of function sub_14BC2


; =============== S U B	R O U T	I N E =======================================


sub_14C70:							; CODE XREF: sub_14620+DAj
					; sub_14620+E6j ...
		move.w	d0,$4E(a3)
		move.w	d1,$50(a3)
		movem.l	d2-d3,-(sp)
		movem.w	d0-d1,-(sp)
		muls.w	$2A(a3),d0
		muls.w	$28(a3),d1
		sub.l	d1,d0
		asr.l	#8,d0
		move.w	d0,d2
		movem.w	(sp),d0-d1
		muls.w	$28(a3),d0
		muls.w	$2A(a3),d1
		add.l	d1,d0
		asr.l	#8,d0
		move.w	d0,d3
		neg.w	d2
		cmp.w	#$E,$52(a3)
		bne.w	loc_14CFE
		bclr	#4,($FFFFC2EE).w
		tst.w	d2
		bpl.w	loc_14D72
		asr.w	#2,d2
		cmpnw	 $FC00,d2
		bgt.w	loc_14CF0
		move.w	#$800,d0
		bsr.w	sub_11086
		neg.w	d0
		move.w	d0,$2C(a3)
		bsr.w	sub_102D2
		move.w	d2,d0
		asr.w	#8,d0
		asr.w	#2,d0
		addq.w	#4,d0
		bpl.w	loc_14CE2
		clr.w	d0

loc_14CE2:							; CODE XREF: sub_14C70+6Cj
		and.w	#3,d0
		add.w	#$28,d0					; '('
		move.w	d0,-(sp)
		bsr.w	sub_11132

loc_14CF0:							; CODE XREF: sub_14C70+4Ej
		move.w	d3,d0
		asr.w	#6,d0
		sub.w	d0,d3
		asr.w	#1,d0
		sub.w	d0,d3
		bra.w	loc_14D3E
; ---------------------------------------------------------------------------

loc_14CFE:							; CODE XREF: sub_14C70+38j
		cmpnw	 $3E8,d2
		bgt.w	loc_14D72
		bclr	#4,($FFFFC2F8).w
		bne.w	loc_14D16
		bset	#4,$64(a3)

loc_14D16:							; CODE XREF: sub_14C70+9Cj
		cmpnw	 $F000,d2
		bgt.w	loc_14D30
		cmp.w	#$A,$32(a3)
		blt.w	loc_14D30
		move.w	#$20,-(sp)				; ' '
		bsr.w	sub_11132

loc_14D30:							; CODE XREF: sub_14C70+AAj
					; sub_14C70+B4j
		asr.w	#2,d2
		cmpnw	 $FC7C,d2
		blt.w	loc_14D3E
		move.w	#$FC18,d2

loc_14D3E:							; CODE XREF: sub_14C70+8Aj
					; sub_14C70+C6j
		movem.w	(sp),d0-d1
		movem.w	d2-d3,-(sp)
		muls.w	d0,d3
		muls.w	d1,d2
		sub.l	d2,d3
		asr.l	#8,d3
		move.w	d3,$28(a3)
		movem.w	(sp)+,d2-d3
		movem.w	(sp),d0-d1
		muls.w	d1,d3
		muls.w	d0,d2
		add.l	d2,d3
		asr.l	#8,d3
		move.w	d3,$2A(a3)
		tst.w	$2C(a3)
		bmi.w	loc_14D72
		clr.w	$2C(a3)

loc_14D72:							; CODE XREF: sub_14C70+44j
					; sub_14C70+92j ...
		addq.w	#4,sp
		movem.l	(sp)+,d2-d3
		rts
; End of function sub_14C70

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_FF0C

loc_14D7A:							; CODE XREF: sub_FF0C+4E9Ej
		bclr	#2,($FFFFC2F4).w
		beq.w	locret_15464
		move.w	#5,-(sp)
		bra.w	sub_11132
; ---------------------------------------------------------------------------

loc_14D8C:							; CODE XREF: sub_FF0C+CAj sub_FF0C+D2j ...
		cmp.w	#$190,$14(a3)
		bgt.w	loc_14DA0
		cmp.w	#$FE70,$14(a3)
		bgt.w	loc_14DA4

loc_14DA0:							; CODE XREF: sub_FF0C+4E86j
		clr.w	$2A(a3)

loc_14DA4:							; CODE XREF: sub_FF0C+4E90j
		cmp.w	#$10,$18(a3)
		bgt.s	loc_14D7A
		move.w	$52(a3),d0
		asl.w	#1,d0
		move.w	#$B86A,a0
		move.w	#$B88A,a1
		move.w	#$B84A,a2
		move.w	(a0,d0.w),d1

loc_14DC2:							; CODE XREF: sub_FF0C+4EDAj
		cmpnw	 $F,d1
		beq.w	loc_14DE8
		clr.w	d4
		move.b	1(a1,d1.w),d4
		move.w	(a2,d4.w),d5
		sub.w	$14(a3),d5
		cmpnw	 $16,d5
		bgt.w	loc_14DE8
		bsr.w	sub_14E10
		addq.w	#1,d1
		bra.s	loc_14DC2
; ---------------------------------------------------------------------------

loc_14DE8:							; CODE XREF: sub_FF0C+4EBAj
					; sub_FF0C+4ED0j
		move.w	(a0,d0.w),d1
		beq.w	locret_14E0E

loc_14DF0:							; CODE XREF: sub_FF0C+4F00j
		clr.w	d4
		move.b	-1(a1,d1.w),d4
		move.w	$14(a3),d5
		sub.w	(a2,d4.w),d5
		cmpnw	 $16,d5
		bgt.w	locret_14E0E
		bsr.w	sub_14E10
		subq.w	#1,d1
		bne.s	loc_14DF0

locret_14E0E:							; CODE XREF: sub_FF0C+4EE0j
					; sub_FF0C+4EF6j
		rts
; END OF FUNCTION CHUNK	FOR sub_FF0C

; =============== S U B	R O U T	I N E =======================================


sub_14E10:							; CODE XREF: sub_FF0C+4ED4p
					; sub_FF0C+4EFAp
		movem.l	d0-a3,-(sp)
		lsr.w	#1,d4
		cmp.w	($FFFFB7AA).w,d4
		beq.w	loc_14F22
		cmpnw	 $B,d4
		bgt.w	loc_14F22
		asl.w	#7,d4
		move.w	#$B04A,a2
		add.w	d4,a2
		btst	#2,($FFFFC2F2).w
		beq.w	loc_14E4A
		asr.w	#7,d4
		cmp.w	($FFFFD406).w,d4
		beq.w	loc_14E4A
		cmp.w	($FFFFD408).w,d4
		bne.w	loc_14F22

loc_14E4A:							; CODE XREF: sub_14E10+24j
					; sub_14E10+2Ej
		btst	#2,$62(a2)
		bne.w	loc_14F22
		tst.b	$5E(a2)
		bne.w	loc_14F22
		btst	#2,$63(a2)
		bne.w	loc_14ED0
		cmp.w	#5,$18(a3)
		bgt.w	loc_14ED0
		cmp.w	#$200,$2C(a3)
		bgt.w	loc_14ED0
		move.l	a2,-(sp)
		bsr.w	sub_106E0
		add.w	(a2),d0
		sub.w	(a3),d0
		cmpnw	 $E,d0
		bgt.w	loc_14ED0
		cmpnw	 $FFF2,d0
		blt.w	loc_14ED0
		add.w	$14(a2),d1
		sub.w	$14(a3),d1
		cmpnw	 $E,d1
		bgt.w	loc_14ED0
		cmpnw	 $FFF2,d1
		blt.w	loc_14ED0
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		move.l	#$C4,d1					; ''
		tst.b	$5E(a3)
		ble.w	loc_14EC2
		lsr.w	#2,d1

loc_14EC2:							; CODE XREF: sub_14E10+ACj
		cmp.l	d1,d0
		bhi.w	loc_14ED0
		bsr.w	sub_150E4
		bra.w	loc_14F22
; ---------------------------------------------------------------------------

loc_14ED0:							; CODE XREF: sub_14E10+52j
					; sub_14E10+5Cj ...
		tst.w	$34(a2)
		beq.w	loc_14F78
		btst	#3,$64(a2)
		bne.w	loc_14F22
		move.w	(a2),d0
		sub.w	(a3),d0
		cmpnw	 8,d0
		bgt.w	loc_14F22
		cmpnw	 $FFF8,d0
		blt.w	loc_14F22
		move.w	$14(a2),d1
		sub.w	$14(a3),d1
		cmpnw	 8,d1
		bgt.w	loc_14F22
		cmpnw	 $FFF8,d1
		blt.w	loc_14F22
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmpnl	 $40,d0					; '@'
		bhi.w	loc_14F22
		bsr.w	sub_153CA

loc_14F22:							; CODE XREF: sub_14E10+Aj
					; sub_14E10+12j ...
		movem.l	(sp)+,d0-a3
		rts
; ---------------------------------------------------------------------------
word_14F28:	dc.w	$C					; DATA XREF: sub_14E10+16Ao
		dc.w	$C
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$D
		dc.w	$E
		dc.w	$E
		dc.w	$E
		dc.w	$E
		dc.w	$F
		dc.w	$F
		dc.w	$F
		dc.w	$F
		dc.w	$F
		dc.w	$F
word_14F50:	dc.w	$90					; DATA XREF: sub_14E10+1C2o
		dc.w	$90
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$A9
		dc.w	$C4
		dc.w	$C4
		dc.w	$C4
		dc.w	$C4
		dc.w	$E1
		dc.w	$E1
		dc.w	$E1
		dc.w	$E1
		dc.w	$E1
		dc.w	$E1
; ---------------------------------------------------------------------------

loc_14F78:							; CODE XREF: sub_14E10+C4j
		move.l	a0,-(sp)
		move.l	#word_14F28,a0
		clr.w	d0
		move.b	$68(a2),d0
		btst	#1,($FFFFC2FE).w
		bne.w	loc_14F9A
		btst	#6,($FFFFC2FC).w
		beq.w	loc_14F9C

loc_14F9A:							; CODE XREF: sub_14E10+17Cj
		addq.b	#2,d0

loc_14F9C:							; CODE XREF: sub_14E10+186j
		lsr.w	#1,d0
		btst	#3,$62(a2)
		beq.w	loc_14FAC
		move.w	#$F,d0

loc_14FAC:							; CODE XREF: sub_14E10+194j
		add.w	d0,d0
		btst	#0,($FFFFC2FA).w
		beq.w	loc_14FC0
		tst.w	d0
		beq.w	loc_14FC0
		subq.w	#2,d0

loc_14FC0:							; CODE XREF: sub_14E10+1A4j
					; sub_14E10+1AAj
		cmpnw	 $1E,d0
		blt.w	loc_14FCC
		move.w	#$1E,d0

loc_14FCC:							; CODE XREF: sub_14E10+1B4j
		move.w	(a0,d0.w),($FFFFBF64).w
		move.l	#word_14F50,a0
		clr.l	($FFFFBF66).w
		move.w	(a0,d0.w),($FFFFBF68).w
		move.w	($FFFFBF64).w,($FFFFBF6A).w
		neg.w	($FFFFBF6A).w
		btst	#1,($FFFFC2F8).w
		beq.w	loc_1500A
		move.w	#$C,($FFFFBF64).w
		move.l	#$90,($FFFFBF66).w			; ''
		move.w	#$FFF4,($FFFFBF6A).w

loc_1500A:							; CODE XREF: sub_14E10+1E2j
		cmp.w	#$250,$58(a2)
		beq.w	loc_1501E
		cmp.w	#$2A2,$58(a2)
		bne.w	loc_15032

loc_1501E:							; CODE XREF: sub_14E10+200j
		move.w	#$12,($FFFFBF64).w
		move.l	#$144,($FFFFBF66).w
		move.w	#$FFEE,($FFFFBF6A).w

loc_15032:							; CODE XREF: sub_14E10+20Aj
		move.l	(sp)+,a0
		cmp.w	#$F,($FFFFB762).w
		bgt.w	loc_150C2
		move.w	(a2),d0
		cmp.w	#$250,$58(a2)
		beq.w	loc_1505C
		cmp.w	#$2A2,$58(a2)
		bne.w	loc_1507A
		move.w	#6,d1
		bra.w	loc_15060
; ---------------------------------------------------------------------------

loc_1505C:							; CODE XREF: sub_14E10+236j
		move.w	#$FFFA,d1

loc_15060:							; CODE XREF: sub_14E10+248j
		btst	#7,$62(a2)
		bne.w	loc_1506C
		neg.w	d1

loc_1506C:							; CODE XREF: sub_14E10+256j
		btst	#0,$76(a2)
		beq.w	loc_15078
		neg.w	d1

loc_15078:							; CODE XREF: sub_14E10+262j
		add.w	d1,d0

loc_1507A:							; CODE XREF: sub_14E10+240j
		sub.w	(a3),d0
		cmp.w	($FFFFBF64).w,d0
		bgt.w	loc_150C2
		cmp.w	($FFFFBF6A).w,d0
		blt.w	loc_150C2
		move.w	$14(a2),d1
		sub.w	$14(a3),d1
		cmp.w	($FFFFBF64).w,d1
		bgt.w	loc_150C2
		cmp.w	($FFFFBF6A).w,d1
		blt.w	loc_150C2
		movem.w	d0-d1,-(sp)
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		cmp.l	($FFFFBF66).w,d0
		movem.w	(sp)+,d0-d1
		bhi.w	loc_150C2
		bsr.w	sub_15466
		bra.w	loc_14F22
; ---------------------------------------------------------------------------

loc_150C2:							; CODE XREF: sub_14E10+22Aj
					; sub_14E10+270j ...
		bclr	#2,($FFFFC2F4).w
		beq.w	loc_14F22
		move.w	($FFFFB7AA).w,d0
		cmp.w	$52(a2),d0
		beq.w	loc_14F22
		move.w	#5,-(sp)
		bsr.w	sub_11132
		bra.w	loc_14F22
; End of function sub_14E10


; =============== S U B	R O U T	I N E =======================================


sub_150E4:							; CODE XREF: sub_14E10+B8p

; FUNCTION CHUNK AT 00015704 SIZE 0000002C BYTES

		tst.w	$34(a2)
		bne.w	loc_150F6
		btst	#4,($FFFFC2EA).w
		bne.w	locret_15464

loc_150F6:							; CODE XREF: sub_150E4+4j
		bclr	#0,$62(a2)
		move.w	($FFFFB7AA).w,d1
		bmi.w	loc_151DE
		asl.w	#7,d1
		move.w	#$B04A,a0
		add.w	d1,a0
		move.b	$62(a0),d1
		move.b	$62(a2),d2
		eor.w	d1,d2
		btst	#6,d2
		beq.w	locret_15464
		btst	#5,($FFFFC2F8).w
		beq.w	loc_15136
		cmpnl	 $40,d0					; '@'
		bhi.w	locret_15464
		bra.w	loc_15140
; ---------------------------------------------------------------------------

loc_15136:							; CODE XREF: sub_150E4+40j
		cmpnl	 $24,d0					; '$'
		bhi.w	locret_15464

loc_15140:							; CODE XREF: sub_150E4+4Ej
		tst.w	$34(a2)
		beq.w	loc_15194
		tst.w	$34(a0)
		beq.w	locret_15464
		move.b	$71(a0),d0
		lsr.b	#1,d0
		bsr.w	sub_15730
		move.w	d0,-(sp)
		exg	a0,a2
		move.b	$71(a0),d0
		lsr.b	#1,d0
		bsr.w	sub_15730
		exg	a0,a2
		neg.w	d0
		add.w	(sp)+,d0
		add.w	#$24,d0					; '$'
		bsr.w	sub_11086
		btst	#5,($FFFFC2F8).w
		beq.w	loc_1518C
		cmpnw	 4,d0
		bhi.w	locret_15464
		bra.w	loc_15194
; ---------------------------------------------------------------------------

loc_1518C:							; CODE XREF: sub_150E4+98j
		cmpnw	 2,d0
		bhi.w	locret_15464

loc_15194:							; CODE XREF: sub_150E4+60j
					; sub_150E4+A4j
		move.b	$71(a0),d0
		lsr.b	#1,d0
		bsr.w	sub_15730
		add.w	#$14,d0
		move.b	d0,$5E(a0)
		exg	a0,a2
		move.b	$71(a0),d0
		lsr.b	#1,d0
		bsr.w	sub_15730
		add.w	#$14,d0
		move.b	d0,$5E(a0)
		exg	a0,a2
		move.w	$52(a0),($FFFFBEDA).w
		bclr	#2,($FFFFC2EC).w
		bclr	#3,($FFFFC2EC).w

loc_151CE:							; CODE XREF: sub_150E4+174j
		move.w	#6,-(sp)
		bsr.w	sub_11132
		bsr.w	sub_1013E
		bra.w	loc_15704
; ---------------------------------------------------------------------------

loc_151DE:							; CODE XREF: sub_150E4+1Cj
		tst.w	$34(a2)
		bne.w	loc_151F0
		cmpnl	 $40,d0					; '@'
		bhi.w	locret_15464

loc_151F0:							; CODE XREF: sub_150E4+FEj
		btst	#3,$64(a2)
		beq.w	loc_1521A
		cmpnl	 $64,d0					; 'd'
		bhi.w	loc_15268
		cmp.w	#$18,$5A(a2)
		bge.w	loc_1521A
		move.w	#$18,$5A(a2)
		bset	#7,($FFFFC2FE).w

loc_1521A:							; CODE XREF: sub_150E4+112j
					; sub_150E4+126j
		move.w	($FFFFB772).w,d0
		move.w	($FFFFB774).w,d1
		muls.w	d0,d0
		muls.w	d1,d1
		add.l	d1,d0
		clr.w	d1
		btst	#4,($FFFFC2EE).w
		bne.w	loc_15246
		btst	#3,$64(a2)
		bne.w	loc_1525C
		move.b	$71(a2),d1
		mulu.w	#$15E,d1

loc_15246:							; CODE XREF: sub_150E4+14Cj
		add.w	#$32C8,d1
		mulu.w	d1,d1
		cmp.l	d1,d0
		bls.w	loc_15284
		move.b	#8,$5E(a2)
		bra.w	loc_151CE
; ---------------------------------------------------------------------------

loc_1525C:							; CODE XREF: sub_150E4+156j
		exg	a2,a3
		jsr	sub_F6CBA
		exg	a2,a3
		rts
; ---------------------------------------------------------------------------

loc_15268:							; CODE XREF: sub_150E4+11Cj
		cmpnl	 $144,d0
		bhi.w	locret_15282
		cmp.w	#$18,$5A(a2)
		bge.w	locret_15282
		move.w	#$18,$5A(a2)

locret_15282:							; CODE XREF: sub_150E4+18Aj
					; sub_150E4+194j
		rts
; ---------------------------------------------------------------------------

loc_15284:							; CODE XREF: sub_150E4+16Aj
					; sub_15466+1C6j
		move.w	#7,-(sp)
		move.w	$52(a2),d0
		move.w	d0,($FFFFB7AA).w
		move.w	#$C6CE,a0
		lea	$364(a0),a1
		btst	#6,$62(a2)
		bne.w	loc_152A4
		exg	a0,a1

loc_152A4:							; CODE XREF: sub_150E4+1BAj
		st	$1A(a0)
		st	$1C(a0)
		bset	#3,$30(a0)
		bclr	#4,($FFFFC2F0).w
		beq.w	loc_152DC
		addq.w	#1,$E(a1)
		add.w	#$C8,($FFFFB89C).w			; ''
		move.w	#$C,(sp)
		cmp.w	#$C6CE,a1
		bne.w	loc_152DC
		add.w	#$A,($FFFFB8A2).w
		move.w	#$B,(sp)

loc_152DC:							; CODE XREF: sub_150E4+1D4j
					; sub_150E4+1EAj
		bsr.w	sub_11156
		move.w	$52(a2),d0
		cmp.w	($FFFFBEE0).w,d0
		bne.w	loc_152F0
		addq.w	#1,$14(a1)

loc_152F0:							; CODE XREF: sub_150E4+204j
		st	($FFFFBEE0).w
		bclr	#2,($FFFFC2EC).w
		bclr	#3,($FFFFC2EC).w
		tst.w	$34(a2)
		bne.w	loc_15322
		bsr.w	sub_12F30
		move.w	#$8C,$48(a2)				; ''
		cmp.w	#$2F4,$58(a2)
		bne.w	loc_15322
		move.w	#5,$48(a2)

loc_15322:							; CODE XREF: sub_150E4+220j
					; sub_150E4+234j
		cmp.w	($FFFFC320).w,d0
		beq.w	locret_15464
		cmp.w	($FFFFC322).w,d0
		beq.w	locret_15464
		tst.w	($FFFFC32C).w
		beq.w	loc_15342
		cmp.w	($FFFFC324).w,d0
		beq.w	locret_15464

loc_15342:							; CODE XREF: sub_150E4+252j
		tst.w	($FFFFC32E).w
		beq.w	loc_15352
		cmp.w	($FFFFC326).w,d0
		beq.w	locret_15464

loc_15352:							; CODE XREF: sub_150E4+262j
		tst.w	($FFFFC32C).w
		bne.w	loc_15362
		tst.w	($FFFFC32E).w
		beq.w	loc_1538A

loc_15362:							; CODE XREF: sub_150E4+272j
		movem.l	d0/a3,-(sp)
		asl.w	#7,d0
		move.l	#$FFFFB04A,a3
		tst.w	$34(a3,d0.w)
		bne.w	loc_15386
		btst	#3,$62(a3,d0.w)
		beq.w	loc_15386
		movem.l	(sp)+,d0/a3
		rts
; ---------------------------------------------------------------------------

loc_15386:							; CODE XREF: sub_150E4+28Ej
					; sub_150E4+298j
		movem.l	(sp)+,d0/a3

loc_1538A:							; CODE XREF: sub_150E4+27Aj
		cmpnw	 6,d0
		slt	d1
		ext.w	d1
		addq.w	#2,d1
		move.w	($FFFFBEDA).w,d2
		cmp.w	($FFFFC322).w,d2
		beq.w	loc_153AE
		cmp.w	($FFFFC328).w,d1
		bne.w	loc_153AE
		jmp	sub_C0BC
; ---------------------------------------------------------------------------

loc_153AE:							; CODE XREF: sub_150E4+2B8j
					; sub_150E4+2C0j
		cmp.w	($FFFFC32A).w,d1
		bne.w	loc_153BC
		jmp	sub_C0DA
; ---------------------------------------------------------------------------

loc_153BC:							; CODE XREF: sub_150E4+2CEj
		cmp.w	($FFFFC328).w,d1
		bne.w	locret_15464
		jmp	sub_C0BC
; End of function sub_150E4


; =============== S U B	R O U T	I N E =======================================


sub_153CA:							; CODE XREF: sub_14E10+10Ep
		btst	#3,$64(a2)
		bne.w	locret_15464
		cmp.w	#8,$18(a3)
		bgt.w	loc_153E8
		move.w	d0,d1
		and.w	#$F,d1
		bne.w	locret_15464

loc_153E8:							; CODE XREF: sub_153CA+10j
		bsr.w	sub_1013E
		move.w	#$24,-(sp)				; '$'
		bsr.w	sub_11132
		clr.w	$2C(a3)
		move.b	#8,$5E(a2)
		move.w	(a3),d0
		sub.w	(a2),d0
		move.w	$14(a3),d1
		sub.w	$14(a2),d1
		bne.w	loc_15414
		move.l	a2,-(sp)
		bsr.w	sub_106E0

loc_15414:							; CODE XREF: sub_153CA+40j
		move.w	$28(a3),d2
		move.w	$2A(a3),d3
		move.b	d0,$28(a3)
		move.b	d1,$2A(a3)
		bsr.w	sub_102D2
		cmp.w	#8,$18(a3)
		ble.w	locret_15464
		muls.w	d2,d2
		muls.w	d3,d3
		add.l	d3,d2
		cmpnl	 $9000000,d2
		bls.w	loc_1544E
		cmp.w	#$C,$18(a3)
		bgt.w	sub_14062
		rts
; ---------------------------------------------------------------------------

loc_1544E:							; CODE XREF: sub_153CA+74j
		bset	#5,$62(a2)
		bne.w	locret_15464
		move.w	#$10D0,d1
		exg	a2,a3
		bsr.w	sub_1073A
		exg	a2,a3
; End of function sub_153CA

; START	OF FUNCTION CHUNK FOR sub_13BE8

locret_15464:							; CODE XREF: sub_D51C-A6j sub_D51C-80j ...
		rts
; END OF FUNCTION CHUNK	FOR sub_13BE8

; =============== S U B	R O U T	I N E =======================================


sub_15466:							; CODE XREF: sub_14E10+2AAp
		btst	#4,($FFFFC2EA).w
		bne.s	locret_15464
		clr.w	d2
		cmp.w	#8,$18(a3)
		bgt.w	loc_1547C
		addq.w	#2,d2

loc_1547C:							; CODE XREF: sub_15466+10j
		neg.w	d0
		neg.w	d1
		bsr.w	sub_10676
		sub.w	$54(a2),d0
		and.w	#7,d0
		move.w	d0,d1
		and.w	#3,d1
		bne.w	loc_1549C
		move.w	($C00008).l,d0

loc_1549C:							; CODE XREF: sub_15466+2Cj
		and.w	#4,d0
		lsr.w	#2,d0
		eor.w	#1,d0
		add.w	d0,d2
		add.w	d2,d2
		lea	word_15630(pc),a0
		move.w	(a0,d2.w),d0
		moveq	#$F,d1
		add.b	(a2,d0.w),d1
		lea	unk_15638(pc),a0
		btst	#3,4(a2)
		beq.w	loc_154CA
		eor.w	#2,d2

loc_154CA:							; CODE XREF: sub_15466+5Cj
		move.w	6(a2),d0
		move.w	#$197,-(sp)
		cmpnw	 $233,d0
		blt.w	loc_154E2
		move.w	#$292,(sp)
		lea	unk_156D4(pc),a0

loc_154E2:							; CODE XREF: sub_15466+70j
		sub.w	(sp)+,d0
		move.b	(a0,d0.w),d0
		lsr.w	d2,d0
		and.w	#3,d0
		bclr	#2,($FFFFC2F4).w
		bsr.w	sub_12F30
		bsr.w	sub_1013E
		cmp.w	#$3000,($FFFFB774).w
		bgt.w	loc_15510
		cmp.w	#$D000,($FFFFB774).w
		bgt.w	loc_15526

loc_15510:							; CODE XREF: sub_15466+9Cj
		move.w	#$B,-(sp)
		btst	#6,$62(a2)
		beq.w	loc_15522
		move.w	#$D,(sp)

loc_15522:							; CODE XREF: sub_15466+B4j
		bsr.w	sub_11156

loc_15526:							; CODE XREF: sub_15466+A6j
		move.w	#$24,-(sp)				; '$'
		bsr.w	sub_11132
		move.w	($FFFFB7AA).w,d2
		bmi.w	loc_15594
		st	($FFFFB7AA).w
		asl.w	#7,d2
		move.w	#$B04A,a0
		add.w	d2,a0
		move.b	#$14,$5E(a0)
		move.w	$52(a0),($FFFFBEDA).w
		bclr	#2,($FFFFC2EC).w
		bclr	#3,($FFFFC2EC).w

loc_1555A:							; CODE XREF: sub_15466+134j
					; sub_15466+19Aj ...
		clr.w	$2C(a3)
		move.b	#$A,$5E(a2)
		move.w	#4,$46(a2)
		move.w	(a3),d0
		sub.w	(a2),d0
		move.w	$14(a3),d1
		sub.w	$14(a2),d1
		bne.w	loc_15580
		move.l	a2,-(sp)
		bsr.w	sub_106E0

loc_15580:							; CODE XREF: sub_15466+110j
		move.b	d0,$28(a3)
		move.b	d1,$2A(a3)
		bra.w	sub_102D2
; ---------------------------------------------------------------------------
		bset	#2,($FFFFC2F4).w
		rts
; ---------------------------------------------------------------------------

loc_15594:							; CODE XREF: sub_15466+CCj
		cmp.w	#$2F4,$58(a2)
		beq.s	loc_1555A
		moveq	#2,d0
		add.b	$6C(a2),d0
		asl.w	#8,d0
		asl.w	#1,d0
		bsr.w	sub_11086
		add.w	#$C00,d0
		add.w	d0,d0
		bpl.w	loc_155B8
		move.w	#$7FFF,d0

loc_155B8:							; CODE XREF: sub_15466+14Aj
		btst	#1,$63(a2)
		bne.w	loc_155FC
		cmp.w	#$2C,(a3)				; ','
		bgt.w	loc_155FC
		cmp.w	#$FFD4,(a3)
		blt.w	loc_155FC
		cmp.w	#$10E,$14(a3)
		bgt.w	loc_155FC
		cmp.w	#$FEF2,$14(a3)
		blt.w	loc_155FC
		cmp.w	#$D2,$14(a3)				; ''
		bgt.w	loc_155FA
		cmp.w	#$FF2E,$14(a3)
		bgt.w	loc_155FC

loc_155FA:							; CODE XREF: sub_15466+186j
		asr.w	#2,d0

loc_155FC:							; CODE XREF: sub_15466+158j
					; sub_15466+160j ...
		cmp.w	($FFFFB772).w,d0
		blt.w	loc_1555A
		cmp.w	($FFFFB774).w,d0
		blt.w	loc_1555A
		neg.w	d0
		cmp.w	($FFFFB772).w,d0
		bgt.w	loc_1555A
		cmp.w	($FFFFB774).w,d0
		bgt.w	loc_1555A
		btst	#2,$63(a2)
		bne.w	loc_1555A
		clr.w	$2C(a3)
		bra.w	loc_15284
; End of function sub_15466

; ---------------------------------------------------------------------------
word_15630:	dc.w	$73					; DATA XREF: sub_15466+44o
		dc.w	$6E
		dc.w	$70
		dc.w	$72
unk_15638:	dc.b	$55					; U		; DATA XREF: sub_15466+52o
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$F0					; 
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
unk_156D4:	dc.b	$9D					; 		; DATA XREF: sub_15466+78o
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$75					; u
		dc.b	$75					; u
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_150E4

loc_15704:							; CODE XREF: sub_150E4+F6j
		st	($FFFFB7AA).w
		move.w	#$1000,d0
		bsr.w	sub_1107A
		move.w	d0,$2A(a3)
		move.w	#$1000,d0
		bsr.w	sub_1107A
		move.w	d0,$28(a3)
		move.w	#$1000,d0
		bsr.w	sub_11086
		move.w	d0,$2C(a3)
		bra.w	sub_102D2
; END OF FUNCTION CHUNK	FOR sub_150E4

; =============== S U B	R O U T	I N E =======================================


sub_15730:							; CODE XREF: sub_C2F2+90p
					; sub_150E4+72p ...
		ext.w	d0
		move.w	d0,-(sp)
		movem.l	d1/a2-a3,-(sp)
		move.l	a0,a3
		bsr.w	sub_1575A
		btst	#4,($FFFFC2FC).w
		beq.w	loc_1574C
		move.w	#$1000,d0

loc_1574C:							; CODE XREF: sub_15730+14j
		movem.l	(sp)+,d1/a2-a3
		muls.w	(sp)+,d0
		asl.l	#4,d0
		swap	d0
		ext.l	d0
		rts
; End of function sub_15730


; =============== S U B	R O U T	I N E =======================================


sub_1575A:							; CODE XREF: sub_B0E8:loc_BB62p
					; sub_B0E8+5B86p ...
		move.w	#$C6CE,a2
		btst	#6,$62(a3)
		beq.w	loc_1576C
		add.w	#$364,a2

loc_1576C:							; CODE XREF: sub_1575A+Aj
		move.b	$66(a3),d1
		ext.w	d1
		add.w	d1,d1
		move.w	$32(a2,d1.w),d0
		rts
; End of function sub_1575A


; =============== S U B	R O U T	I N E =======================================


sub_1577A:							; CODE XREF: sub_B0E8+A8Cp
					; sub_B0E8+5C60j ...
		tst.w	d0
		bpl.w	loc_15782
		clr.w	d0

loc_15782:							; CODE XREF: sub_1577A+2j
		move.w	d0,$32(a2,d1.w)
		rts
; End of function sub_1577A


; =============== S U B	R O U T	I N E =======================================


sub_15788:							; CODE XREF: ROM:00009E76p
					; sub_BB06+2Ap	...
		movem.l	d0-d5/a0-a4,-(sp)
		move.w	$22(a2),a3
		moveq	#5,d4

loc_15792:							; CODE XREF: sub_15788+16j
		st	$60(a3)
		st	$61(a3)
		add.w	#$80,a3					; ''
		dbf	d4,loc_15792
		bsr.w	sub_15832
		moveq	#5,d4
		move.w	#$CF14,a4

loc_157AC:							; CODE XREF: sub_15788:loc_157E0j
		clr.w	d5
		move.b	(a4,d4.w),d5
		beq.w	loc_157E0
		subq.w	#1,d5
		moveq	#5,d3
		move.w	$22(a2),a3
		suba.w	#$80,a3					; ''

loc_157C2:							; CODE XREF: sub_15788+42j
		add.w	#$80,a3					; ''
		cmp.b	$66(a3),d5
		dbeq	d3,loc_157C2
		bne.w	loc_157E0
		move.b	6(a4,d4.w),$60(a3)
		move.b	d5,$61(a3)
		clr.b	(a4,d4.w)

loc_157E0:							; CODE XREF: sub_15788+2Aj
					; sub_15788+46j
		dbf	d4,loc_157AC
		moveq	#5,d4
		move.w	#$CF14,a4

loc_157EA:							; CODE XREF: sub_15788:loc_15828j
		clr.w	d5
		move.b	(a4,d4.w),d5
		beq.w	loc_15828
		subq.w	#1,d5
		moveq	#5,d3
		move.w	$22(a2),a3
		suba.w	#$80,a3					; ''

loc_15800:							; CODE XREF: sub_15788+80j
					; sub_15788+8Ej
		add.w	#$80,a3					; ''
		tst.b	$61(a3)
		dbmi	d3,loc_15800
		bpl.w	loc_15812
		move.w	a3,a0

loc_15812:							; CODE XREF: sub_15788+84j
		tst.w	$34(a3)
		dbpl	d3,loc_15800
		move.b	6(a4,d4.w),$60(a0)
		move.b	d5,$61(a0)
		clr.b	(a4,d4.w)

loc_15828:							; CODE XREF: sub_15788+68j
		dbf	d4,loc_157EA
		movem.l	(sp)+,d0-d5/a0-a4
		rts
; End of function sub_15788


; =============== S U B	R O U T	I N E =======================================


sub_15832:							; CODE XREF: sub_15788+1Ap
		move.w	#$CF14,a4
		clr.l	(a4)
		clr.w	4(a4)
		move.l	#off_19286,a0
		tst.w	$26(a2)
		bpl.w	loc_1584C
		addq.w	#1,a0

loc_1584C:							; CODE XREF: sub_15832+14j
		lea	$16A(a2),a1
		move.w	$16(a2),d0
		asl.w	#3,d0
		add.w	d0,a1
		move.w	$24(a2),d4
		bra.w	loc_1587E
; ---------------------------------------------------------------------------

loc_15860:							; CODE XREF: sub_15832:loc_1587Ej
		clr.w	d5
		move.b	(a0,d4.w),d5
		move.b	(a1,d5.w),(a4,d4.w)
		move.b	d5,6(a4,d4.w)
		bne.w	loc_1587E
		moveq	#1,d3
		add.w	$26(a2),d3
		move.b	d3,(a4,d4.w)

loc_1587E:							; CODE XREF: sub_15832+2Aj
					; sub_15832+3Ej
		dbf	d4,loc_15860
		moveq	#5,d4

loc_15884:							; CODE XREF: sub_15832:loc_158D6j
		move.b	(a4,d4.w),d3
		beq.w	loc_158D6
		ext.w	d3
		subq.w	#1,d3
		add.w	d3,d3
		cmp.w	#$FFFD,$66(a2,d3.w)
		beq.w	loc_158AE
		cmp.w	#$FFFC,$66(a2,d3.w)
		beq.w	loc_158AE
		tst.w	$66(a2,d3.w)
		ble.w	loc_158D6

loc_158AE:							; CODE XREF: sub_15832+66j
					; sub_15832+70j
		lea	$16A(a2),a1
		move.b	6(a4,d4.w),d0
		ext.w	d0
		asl.w	#1,d0
		move.l	#word_1921C,a0
		add.w	(a0,d0.w),a0

loc_158C4:							; CODE XREF: sub_15832+A2j
		clr.w	d0
		move.b	(a0)+,d0
		bmi.w	loc_158DC
		move.b	(a1,d0.w),d0
		bsr.w	sub_158F2
		beq.s	loc_158C4

loc_158D6:							; CODE XREF: sub_15832+56j
					; sub_15832+78j ...
		dbf	d4,loc_15884
		rts
; ---------------------------------------------------------------------------

loc_158DC:							; CODE XREF: sub_15832+96j
		jsr	sub_9F9A
		move.w	d0,d3

loc_158E4:							; CODE XREF: sub_15832+BCj
		move.w	d3,d0
		subq.w	#1,d3
		bmi.s	loc_158D6
		bsr.w	sub_158F2
		beq.s	loc_158E4
		bra.s	loc_158D6
; End of function sub_15832


; =============== S U B	R O U T	I N E =======================================


sub_158F2:							; CODE XREF: sub_15832+9Ep
					; sub_15832+B8p
		move.w	d0,d1
		subq.w	#1,d0
		add.w	d0,d0
		tst.w	$66(a2,d0.w)
		bgt.w	loc_15928
		cmp.w	#$FFFD,$66(a2,d0.w)
		beq.w	loc_15928
		cmp.w	#$FFFC,$66(a2,d0.w)
		beq.w	loc_15928
		moveq	#5,d0

loc_15916:							; CODE XREF: sub_158F2+28j
		cmp.b	(a4,d0.w),d1
		dbeq	d0,loc_15916
		beq.w	loc_15928
		move.b	d1,(a4,d4.w)
		rts
; ---------------------------------------------------------------------------

loc_15928:							; CODE XREF: sub_158F2+Aj
					; sub_158F2+14j ...
		clr.w	d1
		rts
; End of function sub_158F2


; =============== S U B	R O U T	I N E =======================================


sub_1592C:							; CODE XREF: sub_EF92+F4p
					; sub_EF92+100p
		movem.l	d0-d5/a0-a3,-(sp)
		move.w	$22(a2),a3
		moveq	#5,d4

loc_15936:							; CODE XREF: sub_1592C+72j
		move.w	$52(a3),d0
		cmp.w	($FFFFD406).w,d0
		beq.w	loc_1595C
		cmp.w	($FFFFD408).w,d0
		beq.w	loc_15976
		bset	#2,$63(a3)
		move.w	#$20,d0					; ' '
		bsr.w	sub_10662
		bra.w	loc_15992
; ---------------------------------------------------------------------------

loc_1595C:							; CODE XREF: sub_1592C+12j
		bsr.w	sub_15A88
		move.b	$61(a3),($FFFFC31E).w
		move.b	($FFFFD40B).w,$61(a3)
		move.w	#4,$34(a3)
		bra.w	loc_1597E
; ---------------------------------------------------------------------------

loc_15976:							; CODE XREF: sub_1592C+1Aj
		bsr.w	sub_15A88
		clr.w	$34(a3)

loc_1597E:							; CODE XREF: sub_1592C+46j
		clr.w	d3
		move.b	$61(a3),d3
		add.w	d3,d3
		move.w	#$FFFF,$66(a2,d3.w)
		lsr.w	#1,d3
		bsr.w	sub_15AA4

loc_15992:							; CODE XREF: sub_1592C+2Cj
		st	$61(a3)
		st	$60(a3)
		add.w	#$80,a3					; ''
		dbf	d4,loc_15936
		movem.l	(sp)+,d0-d5/a0-a3
		rts
; End of function sub_1592C


; =============== S U B	R O U T	I N E =======================================


sub_159A8:							; CODE XREF: ROM:0000F9E0p
					; ROM:0000F9ECp ...
		movem.l	d0-d4/a0-a3,-(sp)
		move.w	$22(a2),a3
		moveq	#5,d4

loc_159B2:							; CODE XREF: sub_159A8+72j
		move.b	$60(a3),d0
		ext.w	d0
		tst.w	$34(a3)
		bne.w	loc_159DA
		tst.w	d0
		beq.w	loc_159DA
		move.w	$52(a3),-(sp)
		move.w	#$F,$52(a3)
		jsr	nullsub_2
		move.w	(sp)+,$52(a3)

loc_159DA:							; CODE XREF: sub_159A8+14j
					; sub_159A8+1Aj
		move.w	d0,$34(a3)
		bmi.w	loc_15A0E
		bsr.w	sub_15A88
		cmp.w	#4,$34(a3)
		bne.w	loc_159FA
		move.l	#$11,d0
		bsr.w	sub_10658

loc_159FA:							; CODE XREF: sub_159A8+44j
		clr.w	d3
		move.b	$61(a3),d3
		add.w	d3,d3
		move.w	#$FFFF,$66(a2,d3.w)
		lsr.w	#1,d3
		bsr.w	sub_15AA4

loc_15A0E:							; CODE XREF: sub_159A8+36j
		st	$61(a3)
		st	$60(a3)
		add.w	#$80,a3					; ''
		dbf	d4,loc_159B2
		movem.l	(sp)+,d0-d4/a0-a3
		rts
; End of function sub_159A8


; =============== S U B	R O U T	I N E =======================================


sub_15A24:							; CODE XREF: sub_EF92+E8p
					; ROM:0000F9D4p ...
		clr.b	($FFFFC3EA).w
		moveq	#$10,d1
		move.w	#$C6CE,a0
		bsr.w	sub_15A38
		moveq	#1,d1
		add.w	#$364,a0
; End of function sub_15A24


; =============== S U B	R O U T	I N E =======================================


sub_15A38:							; CODE XREF: sub_15A24+Ap
		moveq	#$32,d0					; '2'

loc_15A3A:							; CODE XREF: sub_15A38+4Cj
		add.b	d1,($FFFFC3EA).w
		tst.w	$66(a0,d0.w)
		ble.w	loc_15A64
		btst	#4,$66(a0,d0.w)
		beq.w	loc_15A82
		move.w	$66(a0,d0.w),d2
		and.w	#$7FF,d2
		bne.w	loc_15A82
		sub.b	d1,($FFFFC3EA).w
		bra.w	loc_15A82
; ---------------------------------------------------------------------------

loc_15A64:							; CODE XREF: sub_15A38+Aj
		sub.b	d1,($FFFFC3EA).w
		cmp.w	#$FFFD,$66(a0,d0.w)
		beq.w	loc_15A82
		cmp.w	#$FFFC,$66(a0,d0.w)
		beq.w	loc_15A82
		move.w	#$FFFE,$66(a0,d0.w)

loc_15A82:							; CODE XREF: sub_15A38+14j
					; sub_15A38+20j ...
		subq.w	#2,d0
		bpl.s	loc_15A3A
		rts
; End of function sub_15A38


; =============== S U B	R O U T	I N E =======================================


sub_15A88:							; CODE XREF: sub_C67A+80j sub_C67A+90j ...
		move.w	$34(a3),d0
		bmi.w	locret_15464
		lea	byte_15A9C(pc),a0
		move.b	(a0,d0.w),d0
		bra.w	sub_10662
; End of function sub_15A88

; ---------------------------------------------------------------------------
byte_15A9C:	dc.b	$E					; DATA XREF: sub_15A88+8o
		dc.b	2
		dc.b	2
		dc.b	3
		dc.b	5
		dc.b	3
		dc.b	5
		dc.b	$FF

; =============== S U B	R O U T	I N E =======================================


sub_15AA4:							; CODE XREF: sub_C710+132j
					; sub_EF92+15Ap ...
		bclr	#6,$63(a3)
		move.w	#$C6CE,a0
		btst	#6,$62(a3)
		beq.w	loc_15ABC
		add.w	#$364,a0

loc_15ABC:							; CODE XREF: sub_15AA4+10j
		move.b	d3,$66(a3)
		move.l	#$A,d0
		ext.w	d3
		add.w	d3,d3
		move.w	$66(a0,d3.w),d1
		bpl.w	loc_15AE0
		move.l	#9,d0
		cmpnw	 $FFFE,d1
		bne.w	loc_15AEE

loc_15AE0:							; CODE XREF: sub_15AA4+2Aj
		bsr.w	sub_10658
		bclr	#5,$62(a3)
		clr.w	$58(a3)

loc_15AEE:							; CODE XREF: sub_15AA4+38j
		move.w	#$FFFF,$66(a0,d3.w)
		lsr.w	#1,d3
		move.l	$1E(a0),a0
		move.l	a0,-(sp)
		add.w	8(a0),a0
		clr.l	($FFFFCF2E).w
		tst.w	$34(a3)
		beq.w	loc_15B98
		btst	#5,($FFFFC2EE).w
		beq.w	loc_15B3A
		bsr.w	sub_E274
		beq.w	loc_15B2E
		move.b	1(a0),($FFFFCF2E).w
		and.b	#$F,($FFFFCF2E).w
		bra.w	loc_15B3A
; ---------------------------------------------------------------------------

loc_15B2E:							; CODE XREF: sub_15AA4+76j
		move.b	1(a0),d0
		lsr.b	#4,d0
		neg.b	d0
		move.b	d0,($FFFFCF2F).w

loc_15B3A:							; CODE XREF: sub_15AA4+6Ej
					; sub_15AA4+86j
		move.b	2(a0),d0
		and.b	#$F,d0
		neg.b	d0
		btst	#6,$62(a3)
		bne.w	loc_15B54
		move.b	2(a0),d0
		lsr.b	#4,d0

loc_15B54:							; CODE XREF: sub_15AA4+A6j
		move.b	d0,($FFFFCF31).w
		cmp.w	#2,($FFFFC466).w
		blt.w	loc_15B98
		bgt.w	loc_15B74
		jsr	(sub_784E).w
		lsr.w	#1,d0
		cmp.w	($FFFFC468).w,d0
		bgt.w	loc_15B98

loc_15B74:							; CODE XREF: sub_15AA4+BEj
		move.w	($FFFFC6DA).w,d0
		sub.w	($FFFFCA3E).w,d0
		beq.w	loc_15B92
		btst	#6,$62(a3)
		beq.w	loc_15B8E
		eor	#8,ccr

loc_15B8E:							; CODE XREF: sub_15AA4+E2j
		bpl.w	loc_15B98

loc_15B92:							; CODE XREF: sub_15AA4+D8j
		move.b	#2,($FFFFCF30).w

loc_15B98:							; CODE XREF: sub_15AA4+64j
					; sub_15AA4+BAj ...
		move.l	(sp)+,a0
		add.w	(a0),a0

loc_15B9C:							; CODE XREF: sub_15AA4+FCj
		add.w	(a0),a0
		addq.w	#8,a0
		dbf	d3,loc_15B9C
		subq.w	#8,a0
		move.b	(a0),$6F(a3)
		move.b	1(a0),d3
		and.w	#$F0,d3					; ''
		lsr.w	#1,d3
		move.b	d3,$67(a3)
		move.b	1(a0),d3
		and.b	#$F,d3
		move.w	#3,($FFFFBF14).w
		jsr	sub_F70E4
		move.b	d3,$68(a3)
		move.b	2(a0),d3
		lsr.b	#4,d3
		move.w	#4,($FFFFBF14).w
		jsr	sub_F70E4
		move.b	d3,$69(a3)
		move.b	2(a0),d3
		and.b	#$F,d3
		move.w	#5,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF2E).w,d3
		add.b	($FFFFCF2F).w,d3
		add.b	($FFFFCF31).w,d3
		add.b	($FFFFCF30).w,d3
		bsr.w	sub_15D84
		lsr.b	#1,d3
		eor.b	#$F,d3
		add.b	#$F,d3
		lsr.b	#1,d3
		move.b	d3,$6A(a3)
		move.b	3(a0),d3
		lsr.b	#4,d3
		move.w	#6,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF31).w,d3
		bsr.w	sub_15D84
		lsr.b	#1,d3
		eor.b	#$F,d3
		add.b	#$F,d3
		lsr.b	#1,d3
		move.b	d3,$6B(a3)
		move.b	3(a0),$6C(a3)
		and.b	#$F,$6C(a3)
		move.b	$6C(a3),d3
		move.w	#7,($FFFFBF14).w
		jsr	sub_F70E4
		move.b	d3,$6C(a3)
		move.b	4(a0),d3
		lsr.b	#4,d3
		move.w	#8,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF30).w,d3
		bsr.w	sub_15D84
		move.b	d3,$75(a3)
		bclr	#3,4(a3)
		move.b	4(a0),$76(a3)
		and.b	#1,$76(a3)
		eor.b	#1,$76(a3)
		bne.w	loc_15CA8
		bset	#3,4(a3)

loc_15CA8:							; CODE XREF: sub_15AA4+1FAj
		move.b	4(a0),$74(a3)
		and.b	#$E,$74(a3)
		move.b	5(a0),d3
		lsr.b	#4,d3
		move.w	#$A,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF2E).w,d3
		add.b	($FFFFCF2F).w,d3
		add.b	($FFFFCF31).w,d3
		bsr.w	sub_15D84
		move.b	d3,$71(a3)
		move.b	5(a0),d3
		and.b	#$F,d3
		move.w	#$B,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF2E).w,d3
		add.b	($FFFFCF2F).w,d3
		add.b	($FFFFCF31).w,d3
		bsr.w	sub_15D84
		move.b	d3,$6D(a3)
		move.b	6(a0),d3
		lsr.b	#4,d3
		move.w	#$C,($FFFFBF14).w
		jsr	sub_F70E4
		move.b	d3,$72(a3)
		move.b	6(a0),d3
		and.b	#$F,d3
		move.w	#$D,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF30).w,d3
		add.b	($FFFFCF30).w,d3
		bsr.w	sub_15D84
		move.b	d3,$70(a3)
		move.b	7(a0),d3
		lsr.b	#4,d3
		move.w	#$E,($FFFFBF14).w
		jsr	sub_F70E4
		add.b	($FFFFCF2E).w,d3
		add.b	($FFFFCF31).w,d3
		bsr.w	sub_15D84
		move.b	d3,$6E(a3)
		move.b	7(a0),$73(a3)
		move.b	$73(a3),d3
		and.b	#$F,d3
		move.w	#$F,($FFFFBF14).w
		jsr	sub_F70E4
		move.b	d3,$73(a3)
		and.b	#$F,$73(a3)
		rts
; End of function sub_15AA4


; =============== S U B	R O U T	I N E =======================================


sub_15D84:							; CODE XREF: sub_15AA4+166p
					; sub_15AA4+190p ...
		tst.b	d3
		bpl.w	loc_15D8C
		clr.w	d3

loc_15D8C:							; CODE XREF: sub_15D84+2j
		cmpnb	 $1E,d3
		ble.w	locret_15D98
		move.w	#$1E,d3

locret_15D98:							; CODE XREF: sub_15D84+Cj
		rts
; End of function sub_15D84


; =============== S U B	R O U T	I N E =======================================


sub_15D9A:							; DATA XREF: ROM:00016B8Eo
					; sub_16BAC+CCo
		movem.l	d0-a6,-(sp)
		btst	#2,($FFFFBF78).w
		bne.w	loc_15DBA
		bclr	#0,($FFFFBF78).w
		beq.w	loc_15DB6
		bsr.w	sub_15E6E

loc_15DB6:							; CODE XREF: sub_15D9A+14j
		bsr.w	sub_10FB6

loc_15DBA:							; CODE XREF: sub_15D9A+Aj
		btst	#0,($FFFFC2EC).w
		bne.w	loc_15E1A
		btst	#2,($FFFFC2FA).w
		bne.w	loc_15DD8
		btst	#0,($FFFFC2EA).w
		bne.w	loc_15E1A

loc_15DD8:							; CODE XREF: sub_15D9A+30j
		btst	#1,($FFFFC2FA).w
		bne.w	loc_15E2A
		tst.w	($FFFFC468).w
		beq.w	loc_15E1A
		sub.w	#$AAA,($FFFFC46A).w
		bcc.w	loc_15E1A
		bset	#3,($FFFFBF78).w
		subq.w	#1,($FFFFC468).w
		cmp.w	#$3D,($FFFFC468).w			; '='
		bgt.w	loc_15E1A
		cmp.w	#$3C,($FFFFC468).w			; '<'
		blt.w	loc_15E1A
		move.w	#2,-(sp)
		bsr.w	sub_11132

loc_15E1A:							; CODE XREF: sub_15D9A+26j
					; sub_15D9A+3Aj ...
		addq.w	#1,($FFFFB03E).w
		jsr	sub_1A50A
		movem.l	(sp)+,d0-a6
		rte
; ---------------------------------------------------------------------------

loc_15E2A:							; CODE XREF: sub_15D9A+44j
		tst.w	($FFFFD454).w
		beq.s	loc_15E1A
		sub.w	#$AAA,($FFFFD456).w
		bcc.s	loc_15E1A
		bset	#3,($FFFFBF78).w
		btst	#7,($FFFFC2FA).w
		bne.s	loc_15E1A
		subq.w	#1,($FFFFD454).w
		bra.s	loc_15E1A
; End of function sub_15D9A


; =============== S U B	R O U T	I N E =======================================


sub_15E4C:							; DATA XREF: WaitVBlank+10o sub_17A18o ...
		movem.l	d0-a6,-(sp)
		btst	#2,($FFFFBF78).w
		bne.w	loc_15E5E
		bsr.w	sub_10FB6

loc_15E5E:							; CODE XREF: sub_15E4C+Aj
		addq.w	#1,($FFFFB03E).w
		jsr	sub_1A50A
		movem.l	(sp)+,d0-a6

ExpReturn:							; DATA XREF: ROM:00000060o
					; ROM:00000064o ...
		rte
; End of function sub_15E4C


; =============== S U B	R O U T	I N E =======================================


sub_15E6E:							; CODE XREF: sub_15D9A+18p
		bsr.w	sub_15EA4

loc_15E72:							; CODE XREF: ROM:00017C5Ap
		move.w	#$C018,a0
		move.w	($FFFFC2E8).w,d0
		move.w	($FFFFB002).w,d1
		bsr.w	sub_113E4
; End of function sub_15E6E


; =============== S U B	R O U T	I N E =======================================


sub_15E82:							; CODE XREF: ROM:00016B86p
					; sub_16BAC+BEp
		move.l	($FFFFBD14).w,a6
		cmp.l	#$FFFFB8B4,a6
		beq.w	locret_15464

loc_15E90:							; CODE XREF: sub_15E82+1Ej
		move.w	-(a6),d1
		move.w	-(a6),d0
		move.l	-(a6),a0
		bsr.w	sub_113E4
		cmp.l	#$FFFFB8B4,a6
		bne.s	loc_15E90
		rts
; End of function sub_15E82


; =============== S U B	R O U T	I N E =======================================


sub_15EA4:							; CODE XREF: sub_15E6Ep
		move.w	($FFFFB000).w,d0
		addq.w	#2,d0
		bsr.w	VDP_SetAddress
		move.w	($FFFFBD1E).w,(a0)
		move.l	#$40020010,4(a0)
		move.w	($FFFFBD20).w,(a0)
		rts
; End of function sub_15EA4


; =============== S U B	R O U T	I N E =======================================


sub_15EC0:							; CODE XREF: sub_794E+4Aj sub_7CF8+BCp ...
		movem.l	d0-a6,-(sp)

loc_15EC4:							; CODE XREF: sub_15EC0+Aj
		btst	#0,($FFFFBF78).w
		bne.s	loc_15EC4
		move.w	#$B8B4,a5
		bsr.w	sub_15F34
		move.w	#$C018,a6
		moveq	#1,d6
		bsr.w	sub_16044
		bsr.w	sub_16480
		bsr.w	sub_165FC
		bsr.w	sub_1661A
		bsr.w	sub_162FE
		bsr.w	sub_160F4
		jsr	sub_FD78A
		bsr.w	sub_161D0
		bsr.w	sub_15FF0
		jsr	sub_F6EBE
		cmp.w	#$C018,a6
		bne.w	loc_15F12
		clr.l	(a6)+
		clr.l	(a6)+

loc_15F12:							; CODE XREF: sub_15EC0+4Aj
		clr.b	-5(a6)
		move.l	a6,d0
		sub.l	#$FFFFC018,d0
		lsr.w	#1,d0
		move.w	d0,($FFFFC2E8).w
		move.l	a5,($FFFFBD14).w
		bset	#0,($FFFFBF78).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_15EC0


; =============== S U B	R O U T	I N E =======================================


sub_15F34:							; CODE XREF: sub_15EC0+10p
		btst	#7,($FFFFC2EC).w
		bne.w	locret_15464
		moveq	#$FFFFFFC0,d0
		sub.w	($FFFFBD1C).w,d0
		move.w	($FFFFBD18).w,d1
		move.w	d0,($FFFFBD1E).w
		move.w	d1,($FFFFBD20).w
		neg.w	($FFFFBD20).w
		move.w	($FFFFBD1A).w,d4
		asr.w	#3,d1
		move.w	d1,($FFFFBD1A).w
		moveq	#$1F,d3
		cmp.w	d4,d1
		beq.w	locret_15464
		move.l	#off_5605A,a0
		btst	#4,($FFFFC2F4).w
		beq.w	loc_15F7C
		move.l	#off_B5180,a0

loc_15F7C:							; CODE XREF: sub_15F34+3Ej
		add.l	4(a0),a0
		blt.w	loc_15FBA

loc_15F84:							; CODE XREF: sub_15F34+80j
		move.w	d1,d0
		neg.w	d0
		add.w	#$1E,d0
		and.w	#$1F,d0
		asl.w	#7,d0
		add.w	($FFFFB008).w,d0
		move.w	d0,6(a5)
		move.w	#$1E,d0
		sub.w	d1,d0
		move.w	(a0),4(a5)
		mulu.w	(a0),d0
		add.w	d0,d0
		lea	4(a0,d0.w),a1
		move.l	a1,(a5)
		addq.w	#8,a5
		subq.w	#1,d1
		cmp.w	d4,d1
		dbeq	d3,loc_15F84
		rts
; ---------------------------------------------------------------------------

loc_15FBA:							; CODE XREF: sub_15F34+4Cj
					; sub_15F34+B6j
		move.w	d1,d0
		neg.w	d0
		add.w	#$1D,d0
		and.w	#$1F,d0
		asl.w	#7,d0
		add.w	($FFFFB008).w,d0
		move.w	d0,6(a5)
		move.w	#$3D,d0					; '='
		sub.w	d1,d0
		move.w	(a0),4(a5)
		mulu.w	(a0),d0
		add.w	d0,d0
		lea	4(a0,d0.w),a1
		move.l	a1,(a5)
		addq.w	#8,a5
		addq.w	#1,d1
		cmp.w	d4,d1
		dbeq	d3,loc_15FBA
		rts
; End of function sub_15F34


; =============== S U B	R O U T	I N E =======================================


sub_15FF0:							; CODE XREF: sub_15EC0+3Cp
		bclr	#1,($FFFFC2EE).w
		beq.w	locret_15464
		move.w	#$C3F6,a0
		move.w	#$B004,a1
		move.w	2(a1),d2
		moveq	#2,d0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_16014
		moveq	#$F,d0

loc_16014:							; CODE XREF: sub_15FF0+1Ej
		asl.w	d2,d0
		moveq	#2,d1
		asl.w	d2,d1
		addq.w	#2,d0
		btst	#7,($FFFFC2EC).w
		beq.w	loc_1602A
		add.w	#$B,d0

loc_1602A:							; CODE XREF: sub_15FF0+32j
		asl.w	#1,d0
		add.w	(a1),d0
		moveq	#7,d2

loc_16030:							; CODE XREF: sub_15FF0+4Ej
		move.l	a0,(a5)+
		move.w	#7,(a5)+
		move.w	d0,(a5)+
		add.w	#$E,a0
		add.w	d1,d0
		dbf	d2,loc_16030
		rts
; End of function sub_15FF0


; =============== S U B	R O U T	I N E =======================================


sub_16044:							; CODE XREF: sub_15EC0+1Ap
		btst	#0,($FFFFC2EC).w
		bne.w	locret_15464
		btst	#4,($FFFFBF78).w
		beq.w	locret_15464
		move.w	#$BDA8,a3
		moveq	#2,d0

loc_1605E:							; CODE XREF: sub_16044:loc_160EEj
		move.w	(a3),d4
		bmi.w	loc_160EE
		beq.w	loc_160EE
		move.l	#unk_A78AE,a2
		add.l	4(a2),a2
		add.w	d4,d4
		move.w	2(a2,d4.w),d5
		sub.w	(a2,d4.w),d5
		lsr.w	#3,d5
		subq.w	#1,d5
		add.w	(a2,d4.w),a2

loc_16084:							; CODE XREF: sub_16044+A4j
		move.w	2(a2),d2
		add.w	#$80,d2					; ''
		add.w	($FFFFB034).w,d2
		move.w	d2,(a6)
		move.w	(a2),d2
		btst	#3,2(a3)
		beq.w	loc_160AE
		move.b	7(a2),d2
		and.w	#$C,d2
		addq.w	#4,d2
		asl.w	#1,d2
		neg.w	d2
		sub.w	(a2),d2

loc_160AE:							; CODE XREF: sub_16044+56j
		add.w	#$80,d2					; ''
		add.w	($FFFFB032).w,d2
		move.w	d2,6(a6)
		move.b	7(a2),2(a6)
		move.b	d6,3(a6)
		move.w	6(a2),d2
		and.w	#$F800,d2
		or.w	4(a2),d2
		move.w	2(a3),d1
		and.w	#$F800,d1
		eor.w	d1,d2
		add.w	($FFFFB01C).w,d2
		move.w	d2,4(a6)
		addq.w	#1,d6
		addq.w	#8,a6
		addq.w	#8,a2
		dbf	d5,loc_16084
		addq.w	#4,a3

loc_160EE:							; CODE XREF: sub_16044+1Cj
					; sub_16044+20j
		dbf	d0,loc_1605E
		rts
; End of function sub_16044


; =============== S U B	R O U T	I N E =======================================


sub_160F4:							; CODE XREF: sub_15EC0+2Ep
		move.w	($FFFFB8AC).w,d0
		bmi.w	locret_15464
		move.l	#off_A8922,a0
		lsr.w	#2,d0
		move.w	#$12A,d1
		moveq	#1,d2
		move.w	($FFFFB026).w,d3
		bsr.w	sub_16178
		move.w	d0,d2
		lsr.w	#1,d2
		ext.l	d2
		divu.w	#3,d2
		swap	d2
		addq.w	#2,d2
		bsr.w	sub_16178
		moveq	#5,d2
		move.w	($FFFFC6DA).w,d4
		cmp.w	($FFFFCA3E).w,d4
		bls.w	sub_16178
		move.w	d0,d2
		sub.w	#$DA,d2					; ''
		bpl.w	loc_1613E
		clr.w	d2

loc_1613E:							; CODE XREF: sub_160F4+44j
		lsr.w	#2,d2
		add.w	d2,d2
		cmpnw	 $1A,d2
		blt.w	loc_16150
		move.l	#$18,d2

loc_16150:							; CODE XREF: sub_160F4+52j
		move.l	#word_1615E,a1
		move.w	(a1,d2.w),d2
		bra.w	sub_16178
; End of function sub_160F4

; ---------------------------------------------------------------------------
word_1615E:	dc.w	5					; DATA XREF: sub_160F4:loc_16150o
		dc.w	6
		dc.w	7
		dc.w	8
		dc.w	7
		dc.w	8
		dc.w	7
		dc.w	8
		dc.w	7
		dc.w	8
		dc.w	7
		dc.w	6
		dc.w	5

; =============== S U B	R O U T	I N E =======================================


sub_16178:							; CODE XREF: sub_160F4+1Ap
					; sub_160F4+2Cp ...
		cmpnw	 $40,d6					; '@'
		bge.w	locret_15464
		movem.l	d0-d5/a0,-(sp)
		add.l	4(a0),a0
		add.w	d2,d2
		move.w	2(a0,d2.w),d4
		sub.w	(a0,d2.w),d4
		lsr.w	#3,d4
		subq.w	#1,d4
		add.w	(a0,d2.w),a0

loc_1619A:							; CODE XREF: sub_16178+4Ej
		move.w	2(a0),(a6)
		add.w	d1,(a6)+
		move.b	7(a0),(a6)+
		move.b	d6,(a6)+
		move.w	6(a0),d2
		and.w	#$F800,d2
		add.w	4(a0),d2
		add.w	d3,d2
		move.w	d2,(a6)+
		move.w	(a0),(a6)
		add.w	d0,(a6)+
		addq.w	#1,d6
		cmpnw	 $40,d6					; '@'
		beq.w	loc_161CA
		addq.w	#8,a0
		dbf	d4,loc_1619A

loc_161CA:							; CODE XREF: sub_16178+48j
		movem.l	(sp)+,d0-d5/a0
		rts
; End of function sub_16178


; =============== S U B	R O U T	I N E =======================================


sub_161D0:							; CODE XREF: sub_15EC0+38p
		btst	#4,($FFFFC2F4).w
		bne.w	locret_162FC
		move.l	#off_A4B54,a1
		add.l	4(a1),a1
		move.w	($FFFFBD1C).w,d4
		move.w	($FFFFBD18).w,d5
		btst	#7,($FFFFC2EC).w
		beq.w	loc_161FE
		moveq	#$FFFFFFC0,d4
		move.l	#$100,d5

loc_161FE:							; CODE XREF: sub_161D0+22j
		clr.w	d0
		move.b	($FFFFC3EA).w,d2
		moveq	#$1A,d3
		bsr.w	sub_16226
		move.b	($FFFFC3EA).w,d2
		lsr.w	#4,d2
		moveq	#$1D,d3
		bsr.w	sub_16226
		move.b	($FFFFB89A).w,d0
		bsr.w	sub_16246
		move.b	($FFFFB89B).w,d0
		bra.w	sub_16246
; End of function sub_161D0


; =============== S U B	R O U T	I N E =======================================


sub_16226:							; CODE XREF: sub_161D0+36p
					; sub_161D0+42p
		and.w	#$F,d2
		cmpnw	 3,d2
		bls.w	loc_16234
		moveq	#3,d2

loc_16234:							; CODE XREF: sub_16226+8j
		bra.w	loc_16240
; ---------------------------------------------------------------------------

loc_16238:							; CODE XREF: sub_16226:loc_16240j
		move.w	d3,d0
		add.w	d2,d0
		bsr.w	sub_16246

loc_16240:							; CODE XREF: sub_16226:loc_16234j
		dbf	d2,loc_16238
		rts
; End of function sub_16226


; =============== S U B	R O U T	I N E =======================================


sub_16246:							; CODE XREF: sub_161D0+4Ap
					; sub_161D0+52j ...
		ext.w	d0
		beq.w	locret_15464
		btst	#7,($FFFFC2EC).w
		beq.w	loc_1625A
		add.w	#$1F,d0

loc_1625A:							; CODE XREF: sub_16246+Cj
		cmpnw	 $40,d6					; '@'
		bge.w	locret_15464
		movem.l	d0-d5,-(sp)
		add.w	d0,d0
		move.l	a1,a0
		move.w	2(a0,d0.w),d1
		sub.w	(a0,d0.w),d1
		lsr.w	#3,d1
		subq.w	#1,d1
		move.w	d1,-(sp)
		add.w	(a0,d0.w),a0
		move.w	d4,d0
		add.w	#$C0,d0					; ''
		move.w	d0,d1
		sub.w	#$90,d0					; ''
		add.w	#$80,d1					; ''
		move.w	#$170,d2
		sub.w	d5,d2
		move.w	d2,d3
		sub.w	#$80,d2					; ''
		add.w	#$70,d3					; 'p'
		move.w	(sp)+,d4

loc_1629E:							; CODE XREF: sub_16246+AEj
		cmp.w	2(a0),d2
		bgt.w	loc_162F2
		cmp.w	2(a0),d3
		blt.w	loc_162F2
		cmp.w	(a0),d0
		bgt.w	loc_162F2
		cmp.w	(a0),d1
		blt.w	loc_162F2
		move.w	2(a0),d5
		add.w	#$70,d5					; 'p'
		sub.w	d2,d5
		move.w	d5,(a6)+
		move.b	7(a0),(a6)+
		move.b	d6,(a6)+
		move.w	6(a0),d5
		and.w	#$F800,d5
		add.w	4(a0),d5
		add.w	($FFFFB01A).w,d5
		move.w	d5,(a6)+
		move.w	(a0),d5
		add.w	#$70,d5					; 'p'
		sub.w	d0,d5
		move.w	d5,(a6)+
		addq.w	#1,d6
		cmpnw	 $40,d6					; '@'
		beq.w	loc_162F8

loc_162F2:							; CODE XREF: sub_16246+5Cj
					; sub_16246+64j ...
		addq.w	#8,a0
		dbf	d4,loc_1629E

loc_162F8:							; CODE XREF: sub_16246+A8j
		movem.l	(sp)+,d0-d5

locret_162FC:							; CODE XREF: sub_161D0+6j
		rts
; End of function sub_16246


; =============== S U B	R O U T	I N E =======================================


sub_162FE:							; CODE XREF: sub_7CF8+5Cp
					; sub_15EC0+2Ap

; FUNCTION CHUNK AT 000163BE SIZE 000000AA BYTES

		btst	#7,($FFFFC2EC).w
		beq.w	loc_163BE
		btst	#0,($FFFFC2FA).w
		bne.w	locret_16382
		move.l	a1,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$D
		dc.b	5
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC468).w,d0
		ext.l	d0
		divu.w	#$258,d0
		move.l	d0,-(sp)
		tst.w	d0
		bne.w	loc_1633A
		addq.w	#1,($FFFFB028).w
		bra.w	loc_16340
; ---------------------------------------------------------------------------

loc_1633A:							; CODE XREF: sub_162FE+30j
		jsr	sub_16384

loc_16340:							; CODE XREF: sub_162FE+38j
		move.l	(sp)+,d0
		swap	d0
		ext.l	d0
		divu.w	#$3C,d0					; '<'
		move.l	d0,-(sp)
		jsr	sub_16384
		move.l	(sp)+,d0
		swap	d0
		ext.l	d0
		addq.w	#2,($FFFFB028).w
		divu.w	#$A,d0
		move.l	d0,-(sp)
		jsr	sub_16384
		move.l	(sp)+,d0
		swap	d0
		ext.l	d0
		jsr	sub_16384
		move.l	(sp)+,a1
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	$D
		dc.b	5
		dc.b	0
; ---------------------------------------------------------------------------

locret_16382:							; CODE XREF: sub_162FE+10j
		rts
; End of function sub_162FE


; =============== S U B	R O U T	I N E =======================================


sub_16384:							; CODE XREF: sub_162FE:loc_1633Ap
					; sub_162FE+4Ep ...
		move.l	#off_16396,a1
		asl.w	#2,d0
		add.w	d0,a1
		jsr	sub_11BA4
		rts
; End of function sub_16384

; ---------------------------------------------------------------------------
off_16396:	dc.l	unk_43000				; DATA XREF: sub_16384o
		dc.l	unk_43100
		dc.l	unk_43200
		dc.l	unk_43300
		dc.l	unk_43400
		dc.l	unk_43500
		dc.l	unk_43600
		dc.l	unk_43700
		dc.l	unk_43800
		dc.l	unk_43900
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_162FE

loc_163BE:							; CODE XREF: sub_162FE+6j
		bclr	#3,($FFFFBF78).w
		beq.w	locret_15464
		btst	#3,($FFFFC2EE).w
		bne.w	locret_15464
		cmp.w	#4,($FFFFC466).w
		beq.w	locret_15464
		move.w	#$BF8C,a0
		move.l	#Art_BoardText,a1
		add.l	4(a1),a1
		move.w	$78(a1),d0
		add.w	($FFFFB012).w,d0
		or.w	#$8000,d0
		move.w	d0,($FFFFBF86).w
		move.w	($FFFFC468).w,d0
		btst	#1,($FFFFC2FA).w
		beq.w	loc_1640C
		move.w	($FFFFD454).w,d0

loc_1640C:							; CODE XREF: sub_162FE+106j
		ext.l	d0
		divu.w	#$A,d0
		bsr.w	sub_16468
		divu.w	#6,d0
		bsr.w	sub_16468
		subq.w	#2,a0
		divu.w	#$A,d0
		bsr.w	sub_16468
		swap	d0
		tst.l	d0
		bne.w	loc_16434
		moveq	#$FFFFFFF0,d0
		swap	d0

loc_16434:							; CODE XREF: sub_162FE+12Ej
		bsr.w	sub_16468
		move.l	a0,(a5)+
		move.w	#5,(a5)+
		move.w	#$B004,a1
		moveq	#$18,d0
		moveq	#3,d2
		btst	#7,($FFFFC2EC).w
		beq.w	loc_16458
		move.w	#$B008,a1
		moveq	#5,d0
		moveq	#$D,d2

loc_16458:							; CODE XREF: sub_162FE+14Ej
		move.w	2(a1),d1
		asl.w	d1,d0
		add.w	d2,d0
		asl.w	#1,d0
		add.w	(a1),d0
		move.w	d0,(a5)+
		rts
; END OF FUNCTION CHUNK	FOR sub_162FE

; =============== S U B	R O U T	I N E =======================================


sub_16468:							; CODE XREF: sub_162FE+114p
					; sub_162FE+11Cp ...
		swap	d0
		asl.w	#1,d0
		move.w	$64(a1,d0.w),d0
		add.w	($FFFFB012).w,d0
		or.w	#$8000,d0
		move.w	d0,-(a0)
		swap	d0
		ext.l	d0
		rts
; End of function sub_16468


; =============== S U B	R O U T	I N E =======================================


sub_16480:							; CODE XREF: sub_15EC0+1Ep
		btst	#7,($FFFFC2EC).w
		bne.w	locret_15464
		move.w	#$BDD0,a0
		move.w	#$BE88,a3
		move.w	#$180,d3
		bsr.w	sub_164D6
		add.w	#$1C,a0
		add.w	#$14,a3
		move.w	#$183,d3
		bsr.w	sub_164D6
		tst.w	($FFFFC32C).w
		beq.w	locret_15464
		move.w	#$BE24,a0
		add.w	#$14,a3
		move.w	#$346,d3
		bsr.w	sub_164D6
		tst.w	($FFFFC32E).w
		beq.w	locret_15464
		move.w	#$BE40,a0
		add.w	#$14,a3
		move.w	#$349,d3
; End of function sub_16480


; =============== S U B	R O U T	I N E =======================================


sub_164D6:							; CODE XREF: sub_16480+16p
					; sub_16480+26p ...
		tst.w	$18(a0)
		bmi.w	locret_15464
		st	6(a3)
		move.w	(a0),d0
		btst	#0,($FFFFC2FA).w
		bne.w	loc_16502
		btst	#3,($FFFFC2EE).w
		bne.w	loc_16502
		btst	#2,($FFFFC2F2).w
		beq.w	loc_1651A

loc_16502:							; CODE XREF: sub_164D6+14j
					; sub_164D6+1Ej
		move.w	d0,-(sp)
		tst.w	d0
		bpl.w	loc_1650C
		neg.w	d0

loc_1650C:							; CODE XREF: sub_164D6+30j
		cmpnw	 $A6,d0					; ''
		blt.w	loc_16518
		move.w	(sp)+,d0
		rts
; ---------------------------------------------------------------------------

loc_16518:							; CODE XREF: sub_164D6+3Aj
		move.w	(sp)+,d0

loc_1651A:							; CODE XREF: sub_164D6+28j
		move.w	$14(a0),d1
		btst	#7,($FFFFC2EC).w
		beq.w	loc_16534
		exg	d0,d1
		neg.w	d0
		sub.w	#$C5,d1					; ''
		bra.w	loc_1653C
; ---------------------------------------------------------------------------

loc_16534:							; CODE XREF: sub_164D6+4Ej
		sub.w	($FFFFBD1C).w,d0
		sub.w	($FFFFBD18).w,d1

loc_1653C:							; CODE XREF: sub_164D6+5Aj
		clr.w	d2
		cmpnw	 $74,d0					; 't'
		blt.w	loc_1654A
		bset	#3,d2

loc_1654A:							; CODE XREF: sub_164D6+6Cj
		cmpnw	 $FF8C,d0
		bgt.w	loc_16556
		bset	#2,d2

loc_16556:							; CODE XREF: sub_164D6+78j
		cmpnw	 $64,d1					; 'd'
		blt.w	loc_16562
		bset	#0,d2

loc_16562:							; CODE XREF: sub_164D6+84j
		cmpnw	 $FF9C,d1
		bgt.w	loc_1656E
		bset	#1,d2

loc_1656E:							; CODE XREF: sub_164D6+90j
		tst.w	d2
		beq.w	locret_15464
		move.l	#unk_113C0,a1
		move.b	(a1,d2.w),d2
		asl.w	#3,d2
		move.l	#word_165BC,a1
		move.w	(a1,d2.w),d4
		beq.w	loc_16590
		move.w	d4,d0

loc_16590:							; CODE XREF: sub_164D6+B4j
		add.w	#$100,d0
		move.w	d0,(a3)
		move.w	2(a1,d2.w),d4
		beq.w	loc_165A0
		move.w	d4,d1

loc_165A0:							; CODE XREF: sub_164D6+C4j
		neg.w	d1
		add.w	#$F0,d1					; ''
		move.w	d1,2(a3)
		add.w	4(a1,d2.w),d3
		move.w	d3,6(a3)
		move.w	6(a1,d2.w),4(a3)
		bra.w	sub_167AA
; End of function sub_164D6

; ---------------------------------------------------------------------------
word_165BC:	dc.w	0					; DATA XREF: sub_164D6+AAo
		dc.w	$64
		dc.w	0
		dc.w	0
		dc.w	$74
		dc.w	$64
		dc.w	1
		dc.w	0
		dc.b	0
		dc.b	$74					; t
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$74					; t
		dc.b	$FF
		dc.b	$9C					; 
		dc.b	0
		dc.b	1
		dc.b	$10
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$9C					; 
		dc.b	0
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$FF
		dc.b	$8C					; 
		dc.b	$FF
		dc.b	$9C					; 
		dc.b	0
		dc.b	1
		dc.b	$18
		dc.b	0
		dc.b	$FF
		dc.b	$8C					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	8
		dc.b	0
		dc.b	$FF
		dc.b	$8C					; 
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	1
		dc.b	8
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_165FC:							; CODE XREF: sub_15EC0+22p
		move.w	#$B88A,a4
		move.w	#$F,d0

loc_16604:							; CODE XREF: sub_165FC+18j
		clr.w	d1
		move.b	(a4)+,d1
		asl.w	#6,d1
		move.w	#$B04A,a3
		add.w	d1,a3
		bsr.w	sub_16782
		dbf	d0,loc_16604
		rts
; End of function sub_165FC


; =============== S U B	R O U T	I N E =======================================


sub_1661A:							; CODE XREF: sub_15EC0+26p
		bsr.w	sub_16648
		move.w	#6,d0
		move.w	#$BDB4,a3

loc_16626:							; CODE XREF: sub_1661A+28j
		move.w	a6,a0
		bsr.w	sub_16782
		cmp.w	a6,a0
		beq.w	loc_1663E
		move.w	2(a3),d1
		add.w	d1,6(a0)
		add.w	d1,$E(a0)

loc_1663E:							; CODE XREF: sub_1661A+14j
		add.w	#$1C,a3
		dbf	d0,loc_16626
		rts
; End of function sub_1661A


; =============== S U B	R O U T	I N E =======================================


sub_16648:							; CODE XREF: sub_1661Ap
		move.w	#$BE5C,a0
		st	$18(a0)
		move.b	($FFFFBEE2).w,d0
		beq.w	loc_1666E
		move.b	($FFFFBEE3).w,d1
		ext.w	d0
		asl.w	#2,d0
		move.w	d0,(a0)
		ext.w	d1
		asl.w	#2,d1
		move.w	d1,$14(a0)
		clr.w	$18(a0)

loc_1666E:							; CODE XREF: sub_16648+Cj
		move.w	#5,d4
		move.w	#$BDB4,a0
		move.w	#$BE7A,a1
		move.w	#$B04A,a2
		move.w	($FFFFBE78).w,d3

loc_16682:							; CODE XREF: sub_16648+98j
		cmpnw	 1,d4
		bne.w	loc_16694
		move.w	($FFFFBE86).w,d3
		lsr.w	#4,d3
		bra.w	*+4

loc_16694:							; CODE XREF: sub_16648+3Ej
		move.w	d3,d0
		and.w	#$F,d0
		move.w	#$F,d1
		cmpnw	 $E,d0
		beq.w	loc_166CC
		st	$18(a0)
		cmpnw	 $F,d0
		beq.w	loc_166D8
		asl.w	#7,d0
		move.w	(a2,d0.w),(a0)
		move.w	$14(a2,d0.w),$14(a0)
		clr.w	$18(a0)
		move.b	$35(a2,d0.w),d1
		asl.w	#8,d1
		move.b	$6F(a2,d0.w),d1

loc_166CC:							; CODE XREF: sub_16648+5Aj
		cmp.w	(a1),d1
		beq.w	loc_166D8
		move.w	d1,(a1)
		bsr.w	sub_166E6

loc_166D8:							; CODE XREF: sub_16648+66j
					; sub_16648+86j
		lsr.w	#4,d3
		add.w	#$1C,a0
		addq.w	#2,a1
		dbf	d4,loc_16682
		rts
; End of function sub_16648


; =============== S U B	R O U T	I N E =======================================


sub_166E6:							; CODE XREF: sub_16648+8Cp
		lea	unk_1677A(pc),a4
		clr.w	2(a0)
		move.w	d1,d2
		lsr.w	#4,d2
		and.w	#$F,d2
		bne.w	loc_16702
		move.w	#$FFF0,d2
		subq.w	#4,2(a0)

loc_16702:							; CODE XREF: sub_166E6+10j
		add.w	#$30,d2					; '0'
		clr.w	d0
		bsr.w	sub_1674A
		move.w	d1,d2
		and.w	#$F,d2
		cmpnw	 $F,d2
		bne.w	loc_1671E
		move.w	#$FFF0,d2

loc_1671E:							; CODE XREF: sub_166E6+30j
		add.w	#$30,d2					; '0'
		moveq	#1,d0
		bsr.w	sub_1674A
		move.w	d1,d2
		btst	#7,($FFFFC2FC).w
		beq.w	loc_16736
		clr.w	d2

loc_16736:							; CODE XREF: sub_166E6+4Aj
		lsr.w	#8,d2
		and.w	#7,d2
		bne.w	loc_16744
		addq.w	#4,2(a0)

loc_16744:							; CODE XREF: sub_166E6+56j
		move.b	(a4,d2.w),d2
		moveq	#2,d0
; End of function sub_166E6


; =============== S U B	R O U T	I N E =======================================


sub_1674A:							; CODE XREF: sub_166E6+22p
					; sub_166E6+3Ep
		move.l	#Art_BoardText,a3
		add.l	4(a3),a3
		add.w	d2,d2
		move.w	4(a3,d2.w),d2
		and.w	#$7FF,d2
		asl.w	#5,d2
		move.l	#Art_BoardText,a3
		lea	$A(a3,d2.w),a3
		move.l	a3,(a5)+
		move.w	#$10,(a5)+
		add.w	$12(a0),d0
		asl.w	#5,d0
		move.w	d0,(a5)+
		rts
; End of function sub_1674A

; ---------------------------------------------------------------------------
unk_1677A:	dc.b	$20					; DATA XREF: sub_166E6o
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$4C					; L
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$58					; X
		dc.b	$FF

; =============== S U B	R O U T	I N E =======================================


sub_16782:							; CODE XREF: sub_165FC+14p
					; sub_1661A+Ep
		movem.l	d0-d2,-(sp)
		move.w	(a3),d0
		move.w	$14(a3),d1
		move.w	$18(a3),d2
		bmi.w	loc_167A4
		bsr.w	sub_16920
		cmpnw	 $4E20,d1
		beq.w	loc_167A4
		bsr.w	sub_167AA

loc_167A4:							; CODE XREF: sub_16782+Ej
					; sub_16782+1Aj
		movem.l	(sp)+,d0-d2
		rts
; End of function sub_16782


; =============== S U B	R O U T	I N E =======================================


sub_167AA:							; CODE XREF: sub_164D6+E2j
					; sub_16782+1Ep ...

var_38		=	-$38
var_34		=	-$34

		move.w	4(a3),-(sp)
		movem.l	d0-d5/a0-a2,-(sp)
		move.w	6(a3),d4
		bmi.w	loc_16916
		beq.w	loc_16916
		and.w	#$F800,d4
		eor.w	d4,4(a3)
		move.w	6(a3),d4
		and.w	#$7FF,d4
		move.l	#off_5DE7A,a2
		add.l	4(a2),a2
		add.w	d4,d4
		cmp.w	2(a2),d4
		bge.w	loc_16916
		move.w	2(a2,d4.w),d5
		sub.w	(a2,d4.w),d5
		lsr.w	#3,d5
		subq.w	#1,d5
		add.w	(a2,d4.w),a2
		clr.w	d3
		clr.w	d4

loc_167F6:							; CODE XREF: sub_167AA+168j
		move.w	d0,-(sp)
		move.w	6(a3),d0
		and.w	#$7FF,d0
		cmp.w	8(a3),d0
		beq.w	loc_16882
		tst.w	d5
		bne.w	loc_16812
		move.w	d0,8(a3)

loc_16812:							; CODE XREF: sub_167AA+60j
		movem.w	d0-d4,-(sp)
		move.w	4(a2),d2
		clr.w	d4
		move.b	7(a2),d4
		move.l	#byte_1920C,a0
		move.b	(a0,d4.w),d4
		cmp.w	$3C+var_34(sp),d4
		bgt.w	loc_16854
		cmp.w	$3C+var_38(sp),d2
		blt.w	loc_16854
		move.w	$3C+var_38(sp),d0
		add.w	$3C+var_34(sp),d0
		sub.w	d2,d0
		sub.w	d4,d0
		bmi.w	loc_16854
		movem.w	(sp)+,d0-d1
		addq.w	#6,sp
		bra.w	loc_1687E
; ---------------------------------------------------------------------------

loc_16854:							; CODE XREF: sub_167AA+84j
					; sub_167AA+8Cj ...
		add.w	$3C+var_34(sp),d3
		movem.w	(sp)+,d0-d1
		addq.w	#6,sp
		movem.w	d0-d4,-(sp)
		add.w	$12(a3),d3
		ext.l	d2
		asl.l	#5,d2
		add.l	#off_5DE84,d2
		asl.w	#4,d4
		asl.w	#5,d3
		move.l	d2,(a5)+
		move.w	d4,(a5)+
		move.w	d3,(a5)+
		movem.w	(sp)+,d0-d4

loc_1687E:							; CODE XREF: sub_167AA+A6j
		move.b	d3,$A(a3,d5.w)

loc_16882:							; CODE XREF: sub_167AA+5Aj
		move.w	(sp)+,d0
		movem.w	d0-d2,-(sp)
		move.w	2(a2),d2
		btst	#4,4(a3)
		beq.w	loc_168A8
		move.b	7(a2),d2
		and.w	#3,d2
		addq.w	#1,d2
		asl.w	#3,d2
		neg.w	d2
		sub.w	2(a2),d2

loc_168A8:							; CODE XREF: sub_167AA+E8j
		add.w	d2,d1
		move.w	d1,(a6)
		move.w	(a2),d2
		btst	#3,4(a3)
		beq.w	loc_168C8
		move.b	7(a2),d2
		and.w	#$C,d2
		addq.w	#4,d2
		asl.w	#1,d2
		neg.w	d2
		sub.w	(a2),d2

loc_168C8:							; CODE XREF: sub_167AA+10Aj
		add.w	d2,d0
		move.w	d0,6(a6)
		move.b	7(a2),2(a6)
		move.b	d6,3(a6)
		move.w	6(a2),d2
		move.w	4(a3),d0
		eor.w	d0,d2
		and.w	#$F800,d2
		btst	#0,5(a3)
		beq.w	loc_168FC
		btst	#$E,d2
		beq.w	loc_168FC
		bset	#$D,d2

loc_168FC:							; CODE XREF: sub_167AA+142j
					; sub_167AA+14Aj
		or.b	$A(a3,d5.w),d2
		add.w	$12(a3),d2
		move.w	d2,4(a6)
		movem.w	(sp)+,d0-d2
		addq.w	#1,d6
		addq.w	#8,a6
		addq.w	#8,a2
		dbf	d5,loc_167F6

loc_16916:							; CODE XREF: sub_167AA+Cj
					; sub_167AA+10j ...
		movem.l	(sp)+,d0-d5/a0-a2
		move.w	(sp)+,4(a3)
		rts
; End of function sub_167AA


; =============== S U B	R O U T	I N E =======================================


sub_16920:							; CODE XREF: sub_16782+12p
		btst	#7,($FFFFC2EC).w
		beq.w	loc_16936
		exg	d0,d1
		neg.w	d0
		sub.w	#$C5,d1					; ''
		bra.w	loc_1693E
; ---------------------------------------------------------------------------

loc_16936:							; CODE XREF: sub_16920+6j
		sub.w	($FFFFBD1C).w,d0
		sub.w	($FFFFBD18).w,d1

loc_1693E:							; CODE XREF: sub_16920+12j
		cmpnw	 $90,d0					; ''
		bgt.w	loc_16970
		cmpnw	 $FF70,d0
		blt.w	loc_16970
		add.w	#$100,d0
		add.w	d2,d1
		asr.w	#1,d2
		add.w	d2,d1
		cmpnw	 $90,d1					; ''
		bgt.w	loc_16970
		cmpnw	 $FF70,d1
		blt.w	loc_16970
		neg.w	d1
		add.w	#$F0,d1					; ''
		rts
; ---------------------------------------------------------------------------

loc_16970:							; CODE XREF: sub_16920+22j
					; sub_16920+2Aj ...
		move.w	#$4E20,d1
		rts
; End of function sub_16920


; =============== S U B	R O U T	I N E =======================================


sub_16976:							; CODE XREF: sub_799E+10p
		move.w	($FFFFB89C).w,d0
		asl.w	#3,d0
		add.w	#$400,d0
		cmpnw	 $EFF,d0
		bls.w	loc_1698C
		move.w	#$EFF,d0

loc_1698C:							; CODE XREF: sub_16976+Ej
		moveq	#$28,d2					; '('
		sub.w	($FFFFB044).w,d0
		cmp.w	d2,d0
		bgt.w	loc_169A0
		neg.w	d2
		cmp.w	d2,d0
		bge.w	loc_169AC

loc_169A0:							; CODE XREF: sub_16976+1Ej
		add.w	d2,($FFFFB044).w
		bpl.w	loc_169AC
		clr.w	($FFFFB044).w

loc_169AC:							; CODE XREF: sub_16976+26j
					; sub_16976+2Ej
		move.b	#$C8,($C00011).l
		move.b	#1,($C00011).l
		move.b	($FFFFB044).w,d0
		eor.b	#$F,d0
		or.b	#$F0,d0
		move.b	d0,($C00011).l
		rts
; End of function sub_16976


; =============== S U B	R O U T	I N E =======================================


PSG_SetupNoise:							; CODE XREF: ROM:000076FAp
					; ROM:00007DECp ...
		move.b	#$E7,($C00011).l			; Set PSG 4 to	white noise, and to use	PSG 3's frequency
		move.b	#$DF,($C00011).l			; Mute	PSG 3
		move.b	#$C8,($C00011).l			; Set PSG 3 frequency to 018
		move.b	#1,($C00011).l				; ''
		move.b	#$FF,($C00011).l			; Mute	PSG 4
		rts
; End of function PSG_SetupNoise

; ---------------------------------------------------------------------------

loc_169FA:							; CODE XREF: sub_7CB0-4DEp
					; sub_7CB0-42Cp
		movem.l	d0-a6,-(sp)
		bset	#1,($FFFFBF78).w
		move.w	#$C000,($FFFFB008).w
		move.w	#6,($FFFFB00A).w
		move.w	#$DC00,($FFFFB002).w
		move.w	#$E000,($FFFFB004).w
		move.w	#6,($FFFFB006).w
		move.w	#$F000,($FFFFB00C).w
		move.w	#5,($FFFFB00E).w
		move.w	#$FC00,($FFFFB000).w
		moveq	#0,d0
		bsr.w	sub_11594
		bclr	#0,($FFFFC2EC).w
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		clr.w	($FFFFBD1C).w
		clr.w	($FFFFBD18).w
		move.w	#$7D0,($FFFFBD1A).w
		st	($FFFFB8AC).w
		move.w	#$800,d0
		move.w	($FFFFB004).w,d1
		move.w	#$7FF,d2
		bsr.w	sub_11544
		clr.w	d4
		move.w	d4,($FFFFB018).w
		move.l	#unk_56062,a2
		btst	#4,($FFFFC2F4).w
		beq.w	loc_16A88
		move.l	#unk_B5188,a2

loc_16A88:							; CODE XREF: ROM:00016A7Ej
		bsr.w	sub_11738
		jsr	sub_FEA52
		move.w	d4,($FFFFB020).w
		bsr.w	loc_16CAC
		move.w	d4,($FFFFB016).w
		move.l	#unk_AB928,a2
		bsr.w	sub_11738
		move.w	d4,($FFFFB01A).w
		move.l	#unk_A4B5C,a2
		bsr.w	sub_11738
		move.w	d4,($FFFFD6AC).w
		add.w	#$32,d4					; '2'
		move.w	d4,($FFFFB024).w
		bsr.w	sub_16DA2
		bsr.w	sub_16C96
		bsr.w	sub_11F12
		move.w	d4,($FFFFB012).w
		move.l	#unk_AAC5A,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$4343
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFB010).w
		move.l	#unk_A9A18,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$7123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		bsr.w	sub_16D18
		move.w	d4,($FFFFB026).w
		btst	#1,($FFFFC2EA).w
		beq.w	loc_16B20
		move.w	#$B04A,a3
		moveq	#$B,d0

loc_16B12:							; CODE XREF: ROM:00016B1Cj
		bchg	#7,$62(a3)
		add.w	#$80,a3					; ''
		dbf	d0,loc_16B12

loc_16B20:							; CODE XREF: ROM:00016B08j
		bsr.w	sub_1720C
		move.w	#$FFFF,($FFFFBE78).w
		move.w	#$FFFF,($FFFFBE86).w
		clr.l	($FFFFBE7A).w
		clr.l	($FFFFBE7E).w
		clr.l	($FFFFBE82).w
		st	($FFFFC320).w
		st	($FFFFC322).w
		move.w	#$B8B4,a5
		move.w	#$C018,a6
		moveq	#1,d6
		move.w	#$BDB4,a3
		clr.w	d0
		clr.w	d1
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		move.l	a5,($FFFFBD14).w
		bsr.w	sub_15E82
		move.w	(sp)+,($FFFFBF78).w
		move.l	#sub_15D9A,($FFFFB03A).w
		bclr	#0,($FFFFBF78).w
		bclr	#2,($FFFFBF78).w
		move	#$2300,sr
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_16BAC:							; CODE XREF: ROM:0001365Cp
		movem.l	d0-a6,-(sp)
		bsr.w	sub_10F32
		bclr	#0,($FFFFC2EC).w
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		clr.w	($FFFFBD1C).w
		clr.w	($FFFFBD18).w
		move.w	#$7D0,($FFFFBD1A).w
		st	($FFFFB8AC).w
		move.w	($FFFFB024).w,d4
		bsr.w	sub_16DA2
		bsr.w	sub_16C96
		btst	#1,($FFFFC2EA).w
		beq.w	loc_16C00
		move.w	#$B04A,a3
		moveq	#$B,d0

loc_16BF2:							; CODE XREF: sub_16BAC+50j
		bchg	#7,$62(a3)
		add.w	#$80,a3					; ''
		dbf	d0,loc_16BF2

loc_16C00:							; CODE XREF: sub_16BAC+3Cj
		bsr.w	sub_1720C
		move.w	#$FFFF,($FFFFBE78).w
		move.w	#$FFFF,($FFFFBE86).w
		clr.l	($FFFFBE7A).w
		clr.l	($FFFFBE7E).w
		clr.l	($FFFFBE82).w
		st	($FFFFC320).w
		st	($FFFFC322).w
		bsr.w	loc_16CAC
		move.w	#$B8B4,a5
		move.w	#$C018,a6
		moveq	#1,d6
		move.w	#$BDB4,a3
		clr.w	d0
		clr.w	d1
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		add.w	#$1C,a3
		bsr.w	sub_167AA
		move.l	a5,($FFFFBD14).w
		bsr.w	sub_15E82
		move.w	#$1C,($FFFFBD26).w
		move.w	(sp)+,($FFFFBF78).w
		move.l	#sub_15D9A,($FFFFB03A).w
		bclr	#0,($FFFFBF78).w
		bclr	#2,($FFFFBF78).w
		move	#$2300,sr
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_16BAC


; =============== S U B	R O U T	I N E =======================================


sub_16C96:							; CODE XREF: ROM:00009D6Ep
					; ROM:00016AC6p ...
		move.l	#off_5605A,a0
		add.l	(a0),a0
		moveq	#$F,d0
		move.w	#$BD28,a1

loc_16CA4:							; CODE XREF: sub_16C96+10j
		move.l	(a0)+,(a1)+
		dbf	d0,loc_16CA4
		rts
; End of function sub_16C96

; ---------------------------------------------------------------------------

loc_16CAC:							; CODE XREF: ROM:00009D5Cp
					; sub_9FD0+40Cp ...
		move.w	($FFFFB020).w,d4
		move.l	#unk_B3538,a2
		bsr.w	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$7123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_16CC4:							; CODE XREF: ROM:00009D62p
					; sub_9FD0+412p
		move.w	($FFFFB016).w,d4
		move.l	#unk_AB928,a2
		bra.w	sub_11738
; End of function sub_16CC4


; =============== S U B	R O U T	I N E =======================================


sub_16CD2:							; CODE XREF: ROM:00009D68p
					; sub_9FD0+418p
		move.w	($FFFFB01A).w,d4
		move.l	#unk_A4B5C,a2
		bra.w	sub_11738
; End of function sub_16CD2


; =============== S U B	R O U T	I N E =======================================


sub_16CE0:							; CODE XREF: sub_7CF8+82p
		move.w	($FFFFB026).w,d4
		move.l	#unk_5CF6C,a2
		bra.w	sub_11738
; End of function sub_16CE0


; =============== S U B	R O U T	I N E =======================================


sub_16CEE:							; CODE XREF: sub_9FD0+41Ep
		move.w	($FFFFB026).w,d4
		move.l	#unk_5C410,a2
		bra.w	sub_11738
; End of function sub_16CEE


; =============== S U B	R O U T	I N E =======================================


sub_16CFC:							; CODE XREF: sub_9FD0+434p
		move.w	($FFFFB01C).w,d4
		move.l	#unk_A78B6,a2
		bra.w	sub_11738
; End of function sub_16CFC


; =============== S U B	R O U T	I N E =======================================


sub_16D0A:							; CODE XREF: sub_9FD0+42Ep
		move.w	($FFFFB026).w,d4
		move.l	#unk_55BFE,a2
		bra.w	sub_11738
; End of function sub_16D0A


; =============== S U B	R O U T	I N E =======================================


sub_16D18:							; CODE XREF: ROM:00016AFAp
		move.w	d4,($FFFFB022).w
		add.w	#$2C,d4					; ','
		move.w	#$C334,a1
		move.l	#off_ABA14,a0
		lea	8(a0),a2
		bsr.w	sub_11738
		add.l	4(a0),a0
		move.l	a0,-(sp)
		move.w	($FFFFC330).w,d0
		move.w	($FFFFB022).w,d1
		asl.w	#5,d1
		bsr.w	sub_16D64
		move.l	(sp)+,a0
		move.w	($FFFFC332).w,d0
		move.w	($FFFFB022).w,d1
		add.w	#$16,d1
		asl.w	#5,d1
		bsr.w	sub_16D64
		move.w	($FFFFB022).w,d4
		add.w	#$2C,d4					; ','
		rts
; End of function sub_16D18


; =============== S U B	R O U T	I N E =======================================


sub_16D64:							; CODE XREF: sub_16D18+2Ap
					; sub_16D18+3Ep
		mulu.w	#$2C,d0					; ','
		lea	4(a0,d0.w),a0
		move.w	#$15,d3

loc_16D70:							; CODE XREF: sub_16D64+38j
		moveq	#$20,d0					; ' '
		move.w	d1,-(sp)
		move.b	(a0),(a1)
		and.w	#$F800,(a1)
		lsr.w	#5,d1
		or.w	#$8000,d1
		or.w	d1,(a1)+
		move.w	(sp)+,d1
		move.w	(a0)+,d2
		and.w	#$7FF,d2
		add.w	($FFFFB022).w,d2
		add.w	#$2C,d2					; ','
		asl.w	#5,d2
		bsr.w	sub_114B8
		add.w	#$20,d1					; ' '
		dbf	d3,loc_16D70
		rts
; End of function sub_16D64


; =============== S U B	R O U T	I N E =======================================


sub_16DA2:							; CODE XREF: ROM:00016AC2p
					; sub_16BAC+2Ep
		move.l	#unk_16E06,a2
		move.w	#$BE88,a3
		moveq	#3,d0

loc_16DAE:							; CODE XREF: sub_16DA2+2Aj
		move.w	#$FFFF,8(a3)
		move.w	(a2)+,(a3)
		move.w	(a2)+,2(a3)
		move.w	(a2)+,6(a3)
		move.w	(a2)+,4(a3)
		move.w	d4,$12(a3)
		add.w	(a2)+,d4
		add.w	#$14,a3
		dbf	d0,loc_16DAE
		move.l	#unk_16E2E,a2
		move.l	#$FFFFBDB4,a3
		moveq	#6,d0

loc_16DDE:							; CODE XREF: sub_16DA2+5Cj
		st	8(a3)
		move.w	(a2)+,(a3)
		move.w	(a2)+,$14(a3)
		move.w	(a2)+,$18(a3)
		move.w	(a2)+,6(a3)
		move.w	(a2)+,4(a3)
		move.w	d4,$12(a3)
		add.w	(a2)+,d4
		add.w	#$1C,a3
		dbf	d0,loc_16DDE
		bra.w	loc_16E82
; ---------------------------------------------------------------------------
unk_16E06:	dc.b	0					; DATA XREF: sub_16DA2o
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	9
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	9
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	9
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	9
unk_16E2E:	dc.b	0					; DATA XREF: sub_16DA2+2Eo
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	$88					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	$86					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	$87					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	3
		dc.b	$4C					; L
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	3
		dc.b	$4D					; M
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	1
		dc.b	$61					; a
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	5
; ---------------------------------------------------------------------------

loc_16E82:							; CODE XREF: sub_16DA2+60j
					; ROM:000F82E8p
		clr.w	d6
		move.w	#$B88A,a1
		lea	unk_16EEE(pc),a2
		move.w	#$B04A,a3
		move.w	#$B86A,a4

loc_16E94:							; CODE XREF: sub_16DA2+146j
		moveq	#$1F,d0
		move.w	a3,a0

loc_16E98:							; CODE XREF: sub_16DA2+F8j
		clr.l	(a0)+
		dbf	d0,loc_16E98
		move.w	d6,$52(a3)
		st	8(a3)
		st	$66(a3)
		move.w	(a2)+,(a3)
		move.w	(a2)+,$14(a3)
		move.w	(a2)+,$18(a3)
		move.w	(a2)+,6(a3)
		move.w	(a2)+,4(a3)
		move.w	d4,$12(a3)
		add.w	(a2)+,d4
		move.w	(a2)+,$4A(a3)
		move.w	(a2)+,$4C(a3)
		addq.w	#1,a2
		move.b	(a2)+,$38(a3)
		addq.w	#1,a2
		move.b	(a2)+,$62(a3)
		add.w	#$80,a3					; ''
		asl.w	#1,d6
		move.b	d6,(a1)+
		lsr.w	#1,d6
		move.w	d6,(a4)+
		addq.w	#1,d6
		cmpnw	 $10,d6
		bne.s	loc_16E94
		bra.w	sub_1702E
; End of function sub_16DA2

; ---------------------------------------------------------------------------
unk_16EEE:	dc.b	1					; DATA XREF: sub_16DA2+E6o
		dc.b	$80					; 
		dc.b	0
		dc.b	$C0					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$9C					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$B0					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$C4					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$D8					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	$FF
		dc.b	$EC					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$80					; 
		dc.b	0
		dc.b	$C0					; 
		dc.b	0
		dc.b	$C0					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	$28					; (
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	$3C					; <
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	$50					; P
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	$64					; d
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	$FF
		dc.b	$38					; 8
		dc.b	0
		dc.b	$78					; x
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$40					; @
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$C
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$95					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$B
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$96					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$8B					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	5
		dc.b	0
		dc.b	5
		dc.b	0
		dc.b	$18
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	$8A					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$11
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	$19
		dc.b	0
		dc.b	4

; =============== S U B	R O U T	I N E =======================================


sub_1702E:							; CODE XREF: sub_9FD0+62p
					; sub_9FD0+2CCp ...
		movem.l	d0-d4/a0-a2,-(sp)
		move.l	#$FFFFB86A,a2
		move.l	#$FFFFB84A,a1
		move.l	#$FFFFB04A,a0
		move.w	#$F,d3

loc_17048:							; CODE XREF: sub_1702E+30j
		move.w	$14(a0),d4
		btst	#7,($FFFFC2EC).w
		beq.w	loc_17058
		move.w	(a0),d4

loc_17058:							; CODE XREF: sub_1702E+24j
		move.w	d4,(a1)+
		add.w	#$80,a0					; ''
		dbf	d3,loc_17048
		move.l	#$FFFFB84A,a1

loc_17068:							; CODE XREF: sub_1702E+7Aj
		clr.w	d4
		move.l	#$FFFFB88A,a0
		move.w	#$E,d3
		clr.w	d0
		clr.w	d1

loc_17078:							; CODE XREF: sub_1702E:loc_170A2j
		move.b	(a0)+,d0
		move.b	(a0),d1
		move.w	(a1,d0.w),d2
		cmp.w	(a1,d1.w),d2
		ble.w	loc_170A2
		move.b	d0,(a0)
		move.b	d1,-1(a0)
		move.l	a0,d2
		sub.l	#$FFFFB88A,d2
		move.w	d2,(a2,d0.w)
		subq.w	#1,d2
		move.w	d2,(a2,d1.w)
		st	d4

loc_170A2:							; CODE XREF: sub_1702E+56j
		dbf	d3,loc_17078
		tst.w	d4
		bne.s	loc_17068
		movem.l	(sp)+,d0-d4/a0-a2
		rts
; End of function sub_1702E


; =============== S U B	R O U T	I N E =======================================


sub_170B0:							; CODE XREF: sub_EF92+104p
					; ROM:0000F9F0p ...
		movem.l	d0-d2/a0-a3,-(sp)
		move.w	#$C6CE,a2
		bsr.w	sub_170CA
		move.w	#$CA32,a2
		bsr.w	sub_170CA
		movem.l	(sp)+,d0-d2/a0-a3
		rts
; End of function sub_170B0


; =============== S U B	R O U T	I N E =======================================


sub_170CA:							; CODE XREF: sub_170B0+8p
					; sub_170B0+10p
		bclr	#4,$30(a2)
		moveq	#5,d2
		move.w	$22(a2),a3

loc_170D6:							; CODE XREF: sub_170CA+32j
		clr.b	$63(a3)
		tst.w	$34(a3)
		bmi.w	loc_170F8
		move.w	#$50C,d1
		bsr.w	sub_1073A
		clr.w	$32(a3)
		clr.b	$5E(a3)
		and.b	#$C2,$62(a3)

loc_170F8:							; CODE XREF: sub_170CA+14j
		add.w	#$80,a3					; ''
		dbf	d2,loc_170D6
		rts
; End of function sub_170CA


; =============== S U B	R O U T	I N E =======================================


sub_17102:							; CODE XREF: sub_7CB0:loc_778Cp
					; ROM:000135D2p ...
		move.w	($FFFFC6F4).w,-(sp)
		move.w	($FFFFCA58).w,-(sp)
		movem.l	a1-a3,-(sp)
		move.w	#$19F,d0
		move.l	#$FFFFCBD4,a1
		move.l	#$FFFFC870,a0
		move.l	#$FFFFD6D2,a2
		move.l	#$FFFFD872,a3

loc_1712A:							; CODE XREF: sub_17102+2Cj
		move.b	(a0)+,(a2)+
		move.b	(a1)+,(a3)+
		dbf	d0,loc_1712A
		movem.l	(sp)+,a1-a3
		move.l	#$363,d0
		move.w	#$C6CE,a0

loc_17140:							; CODE XREF: sub_17102+40j
		clr.b	(a0)+
		dbf	d0,loc_17140
		move.w	#$CA32,a0
		move.w	#$363,d0

loc_1714E:							; CODE XREF: sub_17102+4Ej
		clr.b	(a0)+
		dbf	d0,loc_1714E
		movem.l	a1-a3,-(sp)
		move.w	#$19F,d0
		move.l	#$FFFFCBD4,a1
		move.l	#$FFFFC870,a0
		move.l	#$FFFFD6D2,a2
		move.l	#$FFFFD872,a3

loc_17174:							; CODE XREF: sub_17102+76j
		move.b	(a2)+,(a0)+
		move.b	(a3)+,(a1)+
		dbf	d0,loc_17174
		movem.l	(sp)+,a1-a3
		move.w	(sp)+,($FFFFCA58).w
		move.w	(sp)+,($FFFFC6F4).w
		st	($FFFFC768).w
		st	($FFFFCACC).w
; End of function sub_17102


; =============== S U B	R O U T	I N E =======================================


sub_17190:							; CODE XREF: ROM:00017B68p
		movem.l	d0/a0-a2,-(sp)
		move.w	#$C6CE,a2
		move.w	($FFFFC330).w,d0
		move.w	#$B04A,$22(a2)
		bsr.w	sub_171BE
		move.w	#$CA32,a2
		move.w	($FFFFC332).w,d0
		move.w	#$B34A,$22(a2)
		bsr.w	sub_171BE
		movem.l	(sp)+,d0/a0-a2
		rts
; End of function sub_17190


; =============== S U B	R O U T	I N E =======================================


sub_171BE:							; CODE XREF: sub_17190+12p
					; sub_17190+24p
		move.w	d0,$28(a2)
		move.w	#$30E,a0
		asl.w	#2,d0
		move.l	(a0,d0.w),$1E(a2)
		tst.w	($FFFFD058).w
		beq.w	loc_171F4
		move.l	$1E(a2),a0
		add.w	6(a0),a0
		lea	$16A(a2),a1
		move.l	(a0)+,(a1)+
		move.l	(a0)+,(a1)+
		addq.w	#8,a0
		move.w	#$B,d0

loc_171EC:							; CODE XREF: sub_171BE+30j
		move.l	(a0)+,(a1)+
		dbf	d0,loc_171EC
		rts
; ---------------------------------------------------------------------------

loc_171F4:							; CODE XREF: sub_171BE+14j
		moveq	#$D,d0
		move.l	$1E(a2),a0
		add.w	6(a0),a0
		addq.w	#8,a0
		lea	$16A(a2),a1

loc_17204:							; CODE XREF: sub_171BE+48j
		move.l	(a0)+,(a1)+
		dbf	d0,loc_17204
		rts
; End of function sub_171BE


; =============== S U B	R O U T	I N E =======================================


sub_1720C:							; CODE XREF: ROM:00009D74j
					; ROM:00013774p ...
		clr.w	d1
		move.w	#$C6CE,a0
		bsr.w	sub_1721C
		moveq	#$20,d1					; ' '
		add.w	#$364,a0
; End of function sub_1720C


; =============== S U B	R O U T	I N E =======================================


sub_1721C:							; CODE XREF: sub_1720C+6p
		move.l	$1E(a0),a2
		add.w	2(a2),a2
		add.w	d1,a2
		move.w	#$BD68,a1
		add.w	d1,a1
		moveq	#7,d0

loc_1722E:							; CODE XREF: sub_1721C+14j
		move.l	(a2)+,(a1)+
		dbf	d0,loc_1722E
		rts
; End of function sub_1721C

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_F3E4

loc_17236:							; CODE XREF: sub_F3E4+92j sub_F3E4+9Cj
		addq.w	#1,($FFFFC466).w
		bchg	#1,($FFFFC2EA).w
		cmp.w	#3,($FFFFC466).w
		blt.w	loc_17274
		beq.w	loc_17262
		move.w	#3,($FFFFC466).w
		tst.w	($FFFFD048).w
		bne.w	loc_17262
		move.w	#4,($FFFFC466).w

loc_17262:							; CODE XREF: sub_F3E4+7E66j
					; sub_F3E4+7E74j
		move.w	($FFFFC6DA).w,d0
		sub.w	($FFFFCA3E).w,d0
		beq.w	loc_17274
		move.w	#4,($FFFFC466).w

loc_17274:							; CODE XREF: sub_F3E4+7E62j
					; sub_F3E4+7E86j
		bsr.w	sub_10F32
; END OF FUNCTION CHUNK	FOR sub_F3E4
; START	OF FUNCTION CHUNK FOR sub_7CB0

loc_17278:							; CODE XREF: sub_7CB0-4D2j
		jsr	(sub_7814).w
		move.w	d0,-(sp)
		move.w	($FFFFB03E).w,d0

loc_17282:							; CODE XREF: sub_7CB0+F5D6j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_17282
		jsr	Z80_MuteAllFM
		move.w	d0,-(sp)
		move.w	($FFFFB03E).w,d0

loc_17294:							; CODE XREF: sub_7CB0+F5E8j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_17294
		move.w	(sp)+,d0
		move.w	#$79,-(sp)				; 'y'
		bsr.w	sub_11156
		bsr.w	sub_1323E
		cmp.w	#4,($FFFFC466).w
		bne.w	loc_172B6
		bsr.w	sub_182A2

loc_172B6:							; CODE XREF: sub_7CB0+F5FEj
		bsr.w	sub_130E6
		cmp.w	#4,($FFFFC466).w
		beq.w	loc_172C8
		jmp	(loc_7866).w
; ---------------------------------------------------------------------------

loc_172C8:							; CODE XREF: sub_7CB0+F610j
		bsr.w	sub_180FC
		tst.w	($FFFFD048).w
		beq.w	loc_172E0
		bclr	#1,($FFFFC2EC).w
		jsr	sub_9428

loc_172E0:							; CODE XREF: sub_7CB0+F620j
		bsr.w	loc_17336

loc_172E4:							; CODE XREF: sub_7CB0+6j sub_FC4C0+Aj
		move.w	#$78,-(sp)				; 'x'
		bsr.w	sub_11156
		bra.w	loc_172FA
; END OF FUNCTION CHUNK	FOR sub_7CB0
; ---------------------------------------------------------------------------

loc_172F0:							; CODE XREF: ROM:0000773Cj
		bsr.w	PSG_SetupNoise
		jsr	loc_FF042
; START	OF FUNCTION CHUNK FOR sub_7CB0

loc_172FA:							; CODE XREF: sub_7CB0+F63Cj
		bsr.w	PSG_SetupNoise
		move	#$2700,sr
		move.w	#$FFFE,sp
		move.w	#$B000,a0

loc_1730A:							; CODE XREF: sub_7CB0+F660j
		clr.l	(a0)+
		cmp.w	#$D03E,a0
		blt.s	loc_1730A
		jsr	sub_F739E
		jsr	sub_FBB88
		bsr.w	loc_17336
		btst	#0,($FFFFC2FA).w
		bne.w	loc_17332
		jsr	loc_FCC76

loc_17332:							; CODE XREF: sub_7CB0+F678j
		jmp	(loc_7742).w
; END OF FUNCTION CHUNK	FOR sub_7CB0
; ---------------------------------------------------------------------------

loc_17336:							; CODE XREF: sub_7CB0:loc_172E0p
					; sub_7CB0+F66Ep
		tst.w	($FFFFD048).w
		beq.w	locret_15464
		move.l	#sub_17692,($FFFFB03A).l
		bclr	#1,($FFFFBF78).w
		move.w	#0,($FFFFB000).w
		move.w	#$BC00,($FFFFB002).w
		move.w	#$B000,($FFFFB00C).w
		move.w	#6,($FFFFB00E).w
		move.w	#$C000,($FFFFB008).w
		move.w	#7,($FFFFB00A).w
		move.w	#$E000,($FFFFB004).w
		move.w	#7,($FFFFB006).w
		move.w	#0,d0
		bsr.w	sub_11594
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$80,d0					; ''
		moveq	#$1C,d1
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#2,d1
		move.w	#$7FF,d2
		bsr.w	sub_1197E
		bsr.w	sub_11F20
		bsr.w	sub_11F04
		move.l	#unk_B3648,a2
		move.w	d4,($FFFFB026).w
		bsr.w	sub_11738
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$CE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#3,d5
		bsr.w	sub_1169A
		move.l	#unk_F3098,a1
		lea	8(a1),a2
		add.l	(a1),a1
		moveq	#7,d0
		move.w	#$BD68,a0

loc_17406:							; CODE XREF: ROM:00017414j
		move.l	-$40(a0),$20(a0)
		move.l	(a1)+,-$40(a0)
		move.l	-$20(a0),(a0)+
		dbf	d0,loc_17406
		move.w	d4,($FFFFB016).w
		bsr.w	sub_11738
		move.w	#$104,($FFFFB8B2).w
		move.w	#$A0,($FFFFB8B0).w			; ''
		move.l	#$C00004,a0
		move.w	#$9202,(a0)
		move.w	#$CEF4,a1
		move.l	#off_1928E,a0
		move.w	($FFFFCEEC).w,d0
		asl.w	#1,d0
		add.w	(a0,d0.w),a0
		clr.w	d4
		move.b	(a0)+,d4

loc_1744E:							; CODE XREF: ROM:0001746Aj
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.b	(a0)+,($FFFFB029).w
		move.b	(a0)+,($FFFFB02B).w
		clr.w	d1
		move.b	(a1)+,d1
		add.w	d1,d1
		bsr.w	sub_17652
		dbf	d4,loc_1744E
		clr.w	d4
		move.b	(a0)+,d4

loc_17472:							; CODE XREF: ROM:00017488j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		move.b	(a0)+,($FFFFB029).w
		clr.w	d0
		move.b	(a0)+,d0
		bsr.w	sub_17626
		dbf	d4,loc_17472
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		cmp.w	#7,($FFFFCEEA).w
		beq.w	loc_174C0
		clr.w	d4
		move.b	(a0)+,d4
		bmi.w	loc_174C0
		move.w	#$CE66,a2

loc_174AC:							; CODE XREF: ROM:000174BCj
		move.b	(a0)+,($FFFFB029).w
		move.b	(a0)+,($FFFFB02B).w
		bsr.w	sub_175FC
		add.w	#$10,a2
		dbf	d4,loc_174AC

loc_174C0:							; CODE XREF: ROM:0001749Cj
					; ROM:000174A4j
		bsr.w	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	1
		dc.b	1
		dc.b	$41
		dc.b	$1A
		dc.b	0
; ---------------------------------------------------------------------------
		lea	word_176DE(pc),a1
		move.w	($FFFFCEEC).w,d0
		bsr.w	sub_13510
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		bsr.w	sub_11A48
		tst.w	($FFFFCEEC).w
		beq.w	loc_1750A
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1A
		dc.b	$CD
		dc.b	$A
		dc.b	1
aPressOrToPage:	dc.b	'Press [ or ] to page',0
; ---------------------------------------------------------------------------

loc_1750A:							; CODE XREF: ROM:000174E8j
		move.w	#$18,($FFFFBD26).w
		move.w	#$79,-(sp)				; 'y'
		bsr.w	sub_11156
		clr.w	d0
		clr.w	d4
		move.w	#$FFFF,($FFFFD5B0).w
		move.w	#1,($FFFFD5AC).w
		bsr.w	sub_17572

loc_1752C:							; CODE XREF: ROM:00017540j
		bsr.w	sub_175D8
		move.l	#locret_15464,a0
		jsr	sub_FED2A
		bsr.w	sub_17542
		bra.s	loc_1752C

; =============== S U B	R O U T	I N E =======================================


sub_17542:							; CODE XREF: ROM:0001753Cp

; FUNCTION CHUNK AT 000175D4 SIZE 00000004 BYTES

		bsr.w	sub_11340
		move.w	d3,-(sp)
		bsr.w	sub_11358
		or.w	(sp)+,d3
		btst	#7,d3
		bne.w	loc_175D4
		btst	#3,d3
		beq.w	loc_17564
		move.w	#$FFFE,($FFFFD5B0).w

loc_17564:							; CODE XREF: sub_17542+18j
		btst	#2,d3
		beq.w	sub_17572
		move.w	#2,($FFFFD5B0).w
; End of function sub_17542


; =============== S U B	R O U T	I N E =======================================


sub_17572:							; CODE XREF: ROM:00017528p
					; sub_17542+26j
		move.w	($FFFFD5B0).w,d0
		beq.w	locret_15464
		add.w	($FFFFD5AC).w,d0
		move.w	($FFFFCEEC).w,d1
		cmpnw	 3,d1
		bls.w	loc_1758C
		moveq	#3,d1

loc_1758C:							; CODE XREF: sub_17572+14j
		mulu.w	#$70,d1					; 'p'
		cmp.w	d1,d0
		bgt.w	locret_15464
		neg.w	d1
		cmp.w	d1,d0
		blt.w	locret_15464
		move.w	d0,($FFFFD5AC).w
		clr.w	($FFFFB8AE).w
		move.w	d0,d1
		add.w	#$100,d1
		cmpnw	 $40,d1					; '@'
		blt.w	loc_175C0
		cmpnw	 $200,d1
		bgt.w	loc_175C0
		move.w	d1,($FFFFB8AE).w

loc_175C0:							; CODE XREF: sub_17572+3Ej
					; sub_17572+46j
		ext.l	d0
		divs.w	#$70,d0					; 'p'
		swap	d0
		tst.w	d0
		bne.w	locret_15464
		clr.w	($FFFFD5B0).w
		rts
; End of function sub_17572

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_17542

loc_175D4:							; CODE XREF: sub_17542+10j
		addq.w	#4,sp
		rts
; END OF FUNCTION CHUNK	FOR sub_17542

; =============== S U B	R O U T	I N E =======================================


sub_175D8:							; CODE XREF: ROM:loc_1752Cp
		tst.w	($FFFFBD26).w
		bpl.w	loc_175EC
		eor.w	#$EE,($FFFFBD6A).w			; ''
		move.w	#$18,($FFFFBD26).w

loc_175EC:							; CODE XREF: sub_175D8+4j
					; sub_175D8+1Cj
		move.w	($FFFFB03E).w,d0
		cmp.w	($FFFFB040).w,d0
		beq.s	loc_175EC
		move.w	d0,($FFFFB040).w
		rts
; End of function sub_175D8


; =============== S U B	R O U T	I N E =======================================


sub_175FC:							; CODE XREF: ROM:000174B4p
		move.w	#$BFA4,a1
		move.w	#6,(a1)+
		move.w	4(a2),d0
		add.w	#$30,d0					; '0'
		move.b	d0,(a1)+
		move.b	#$2D,(a1)+				; '-'
		move.w	6(a2),d0
		add.w	#$30,d0					; '0'
		move.b	d0,(a1)+
		clr.b	(a1)+
		move.w	#$BFA4,a1
		bra.w	sub_11BA4
; End of function sub_175FC


; =============== S U B	R O U T	I N E =======================================


sub_17626:							; CODE XREF: ROM:00017484p
		movem.l	d0-a3,-(sp)
		move.l	#unk_B3640,a0
		move.l	a0,a1
		add.l	(a0),a0
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d1
		moveq	#2,d2
		moveq	#$17,d3
		move.w	($FFFFB026).w,d4
		moveq	#0,d5
		bsr.w	sub_1169A
		movem.l	(sp)+,d0-a3
		rts
; End of function sub_17626


; =============== S U B	R O U T	I N E =======================================


sub_17652:							; CODE XREF: ROM:00017466p
		movem.l	d0-a3,-(sp)
		move.w	#$CEF4,a0
		move.w	($FFFFCEEE).w,d0
		move.b	(a0,d0.w),d0
		add.b	d0,d0
		cmp.b	d0,d1
		bne.w	loc_17670
		move.w	#$6000,($FFFFB02C).w

loc_17670:							; CODE XREF: sub_17652+14j
		move.l	#off_ABA14,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d0
		move.w	(a1),d2
		moveq	#2,d3
		moveq	#2,d4
		moveq	#0,d5
		bsr.w	sub_1169A
		movem.l	(sp)+,d0-a3
		rts
; End of function sub_17652


; =============== S U B	R O U T	I N E =======================================


sub_17692:							; DATA XREF: ROM:0001733Eo
		movem.l	d0-a6,-(sp)
		btst	#2,($FFFFBF78).w
		bne.w	loc_176CE
		move.w	#$C018,a0
		move.w	($FFFFC2E8).w,d0
		beq.w	loc_176CA
		clr.w	($FFFFC2E8).w
		move.w	($FFFFB002).w,d1
		bsr.w	sub_113E4
		move.w	($FFFFB000).w,d0
		bsr.w	VDP_SetAddress
		move.w	#$FEA0,d0
		add.w	($FFFFD5AC).w,d0
		move.w	d0,(a0)

loc_176CA:							; CODE XREF: sub_17692+16j
		bsr.w	sub_10FB6

loc_176CE:							; CODE XREF: sub_17692+Aj
		addq.w	#1,($FFFFB03E).w
		jsr	sub_1A50A
		movem.l	(sp)+,d0-a6
		rte
; End of function sub_17692

; ---------------------------------------------------------------------------
word_176DE:	dc.w	$A					; DATA XREF: ROM:000174CCo
aPlayoffs:	dc.b	'Playoffs'
		dc.w	$10
aQuarterfinals:	dc.b	'Quarterfinals',0
		dc.w	$C
aSemifinals:	dc.b	'Semifinals'
		dc.w	8
aFinals:	dc.b	'Finals'
		dc.w	$C
aChampions:	dc.b	'Champions',0

; =============== S U B	R O U T	I N E =======================================


sub_17718:							; CODE XREF: ROM:000FCE5Cp
		move.w	#$FFFF,($FFFFD5AE).w
		st	($FFFFD5B0).w
		clr.w	($FFFFD5B2).w
		clr.w	($FFFFD5B4).w
		move.w	#$D5B8,a0
		rts
; End of function sub_17718

; ---------------------------------------------------------------------------

loc_17730:							; CODE XREF: ROM:loc_FCF5Ap
		cmp.w	#$1000,($FFFFD5B4).w
		bgt.w	locret_15464
		tst.w	($FFFFB044).w
		bmi.w	loc_1774A
		subq.w	#1,($FFFFD5B4).w
		bpl.w	locret_15464

loc_1774A:							; CODE XREF: ROM:0001773Ej
		move.l	#aWelcomeToASoldOut_Ton,a1		; "\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\rWelcome to a "...
		move.w	($FFFFD5B0).w,d0
		bpl.w	loc_17796
		clr.w	($FFFFD5AC).w
		move.w	($FFFFD5B2).w,d0
		addq.w	#2,($FFFFD5B2).w
		move.w	#$D5B8,a0
		move.w	(a0,d0.w),d0
		bpl.w	loc_1777E
		move.w	#$7FFF,($FFFFD5B4).w
		move.w	#$1E0,($FFFFD5B6).w
		rts
; ---------------------------------------------------------------------------

loc_1777E:							; CODE XREF: ROM:0001776Cj
		bsr.w	sub_179D2
		move.l	a1,a2
		bra.w	loc_1778E
; ---------------------------------------------------------------------------

loc_17788:							; CODE XREF: ROM:0001778Cj
					; ROM:loc_1778Ej
		cmp.b	#$D,(a2)+
		bne.s	loc_17788

loc_1778E:							; CODE XREF: ROM:00017784j
		dbf	d0,loc_17788
		suba.l	a1,a2
		move.w	a2,d0

loc_17796:							; CODE XREF: ROM:00017754j
		lea	(a1,d0.w),a2
		move.w	#$A,($FFFFD5B4).w
		move.w	($FFFFD5AC).w,d0
		cmp.b	#$5F,(a2)				; '_'
		beq.w	loc_179B0
		cmp.b	#$2B,(a2)				; '+'
		beq.w	loc_1799C
		cmp.b	#$24,(a2)				; '$'
		beq.w	loc_17956
		cmp.b	#$7B,(a2)				; '{'
		beq.w	loc_179B0
		cmp.b	#$7D,(a2)				; '}'
		beq.w	loc_179A8
		cmp.b	#$5B,(a2)				; '['
		beq.w	loc_1797E
		cmp.b	#$5D,(a2)				; ']'
		beq.w	loc_17976
		cmp.b	#$3C,(a2)				; '<'
		beq.w	loc_1792E
		cmp.b	#$3E,(a2)				; '>'
		beq.w	loc_17938
		cmp.b	#$7C,(a2)				; '|'
		beq.w	loc_17942
		cmp.b	#$5C,(a2)				; '\'
		beq.w	loc_1794C
		cmp.b	#$40,(a2)				; '@'
		beq.w	loc_178C0
		cmp.b	#$23,(a2)				; '#'
		beq.w	loc_178CA
		cmp.b	#$25,(a2)				; '%'
		beq.w	loc_178D4
		cmp.b	#$3D,(a2)				; '='
		beq.w	loc_178DE
		cmp.b	#$2A,(a2)				; '*'
		beq.w	loc_178F2
		cmp.b	#$5E,(a2)				; '^'
		beq.w	loc_17906
		cmp.b	#$3B,(a2)				; ';'
		beq.w	loc_1791A
		st	($FFFFD5B0).w
		move.w	#$BFA6,a0
		clr.w	d1

loc_1783E:							; CODE XREF: ROM:00017866j
		cmp.b	#$D,(a2)
		beq.w	loc_17876
		addq.w	#1,d0

loc_17848:							; CODE XREF: ROM:0001786Cj
		move.b	(a2),(a0)+
		addq.w	#1,d1
		cmp.b	#$2C,(a2)				; ','
		beq.w	loc_1785C
		cmp.b	#$2E,(a2)				; '.'
		bne.w	loc_17862

loc_1785C:							; CODE XREF: ROM:00017850j
		add.w	#$28,($FFFFD5B4).w			; '('

loc_17862:							; CODE XREF: ROM:00017858j
		cmp.b	#$20,(a2)+				; ' '
		bne.s	loc_1783E
		cmp.b	#$20,(a2)				; ' '
		beq.s	loc_17848
		subq.w	#1,d0
		suba.w	a1,a2
		move.w	a2,($FFFFD5B0).w

loc_17876:							; CODE XREF: ROM:00017842j
		move.w	d1,($FFFFBFA4).w
		addq.w	#2,($FFFFBFA4).w
		btst	#0,d1
		beq.w	loc_1788C
		clr.b	(a0)
		addq.w	#1,($FFFFBFA4).w

loc_1788C:							; CODE XREF: ROM:00017882j
		move.w	#$BFA4,a1

loc_17890:							; CODE XREF: ROM:000179C8j
					; ROM:000179CEj
		cmpnw	 $1D,d0
		ble.w	loc_1789C
		bsr.w	sub_179D2

loc_1789C:							; CODE XREF: ROM:00017894j
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	9
		dc.b	4
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD5AC).w,d0
		add.w	d0,($FFFFB028).w
		move.w	($FFFFD5AE).w,d0
		add.w	d0,($FFFFB02A).w
		bsr.w	sub_11BA4
		add.w	d1,($FFFFD5AC).w
		rts
; ---------------------------------------------------------------------------

loc_178C0:							; CODE XREF: ROM:00017800j
		jsr	sub_FCB9A
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_178CA:							; CODE XREF: ROM:00017808j
		jsr	sub_FEEC8
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_178D4:							; CODE XREF: ROM:00017810j
		jsr	sub_FEF5A
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_178DE:							; CODE XREF: ROM:00017818j
		move.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		jsr	sub_FD5AE
		move.l	(sp)+,a2
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_178F2:							; CODE XREF: ROM:00017820j
		move.l	a2,-(sp)
		move.l	#$FFFFCA32,a2
		jsr	sub_FD5AE
		move.l	(sp)+,a2
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_17906:							; CODE XREF: ROM:00017828j
		move.l	a2,-(sp)
		move.l	#$FFFFC6CE,a2
		jsr	sub_FD5F4
		move.l	(sp)+,a2
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_1791A:							; CODE XREF: ROM:00017830j
		move.l	a2,-(sp)
		move.l	#$FFFFCA32,a2
		jsr	sub_FD5F4
		move.l	(sp)+,a2
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_1792E:							; CODE XREF: ROM:000177E0j
		jsr	sub_F7144
		bra.w	loc_17986
; ---------------------------------------------------------------------------

loc_17938:							; CODE XREF: ROM:000177E8j
		jsr	sub_F7172
		bra.w	loc_17986
; ---------------------------------------------------------------------------

loc_17942:							; CODE XREF: ROM:000177F0j
		jsr	sub_F727C
		bra.w	loc_17986
; ---------------------------------------------------------------------------

loc_1794C:							; CODE XREF: ROM:000177F8j
		jsr	sub_F72AA
		bra.w	loc_17986
; ---------------------------------------------------------------------------

loc_17956:							; CODE XREF: ROM:000177B8j
		move.l	#$FFFFC6CE,a1
		tst.w	($FFFFBF50).w
		beq.w	loc_1796A
		move.l	#$FFFFCA32,a1

loc_1796A:							; CODE XREF: ROM:00017960j
		move.l	$1E(a1),a1
		add.w	4(a1),a1
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_17976:							; CODE XREF: ROM:000177D8j
		move.w	#$CA32,a1
		bra.w	loc_17982
; ---------------------------------------------------------------------------

loc_1797E:							; CODE XREF: ROM:000177D0j
		move.w	#$C6CE,a1

loc_17982:							; CODE XREF: ROM:0001797Aj
		move.w	$26(a1),d1

loc_17986:							; CODE XREF: ROM:00017934j
					; ROM:0001793Ej ...
		move.l	$1E(a1),a1
		add.w	(a1),a1
		bra.w	loc_17994
; ---------------------------------------------------------------------------

loc_17990:							; CODE XREF: ROM:loc_17994j
		add.w	(a1),a1
		addq.w	#8,a1

loc_17994:							; CODE XREF: ROM:0001798Cj
		dbf	d1,loc_17990
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_1799C:							; CODE XREF: ROM:000177B0j
		move.l	($FFFFCA50).w,a1
		add.w	4(a1),a1
		bra.w	loc_179B8
; ---------------------------------------------------------------------------

loc_179A8:							; CODE XREF: ROM:000177C8j
		move.l	($FFFFCA50).w,a1
		bra.w	loc_179B4
; ---------------------------------------------------------------------------

loc_179B0:							; CODE XREF: ROM:000177A8j
					; ROM:000177C0j
		move.l	($FFFFC6EC).w,a1

loc_179B4:							; CODE XREF: ROM:000179ACj
		add.w	4(a1),a1

loc_179B8:							; CODE XREF: ROM:000178C6j
					; ROM:000178D0j ...
		addq.w	#1,($FFFFD5B0).w
		add.w	(a1),d0
		subq.w	#1,d0
		move.w	(a1),d1
		subq.w	#2,d1
		tst.b	1(a1,d1.w)
		bne.w	loc_17890
		subq.w	#1,d1
		bra.w	loc_17890

; =============== S U B	R O U T	I N E =======================================


sub_179D2:							; CODE XREF: ROM:loc_1777Ep
					; ROM:00017898p
		clr.w	($FFFFD5AC).w
		addq.w	#1,($FFFFD5AE).w
		cmp.w	#7,($FFFFD5AE).l
		blt.w	locret_15464
		subq.w	#1,($FFFFD5AE).w
		movem.l	d0-d3,-(sp)
		move.l	#7,d3
		move.w	($FFFFB004).w,d1
		add.w	#$212,d1

loc_179FC:							; CODE XREF: sub_179D2+3Cj
		move.l	#$3A,d0					; ':'
		move.w	d1,d2
		add.w	#$80,d2					; ''
		bsr.w	sub_114B8
		move.w	d2,d1
		dbf	d3,loc_179FC
		movem.l	(sp)+,d0-d3
		rts
; End of function sub_179D2


; =============== S U B	R O U T	I N E =======================================


sub_17A18:							; CODE XREF: ROM:0000770Cp
		move.l	#sub_15E4C,($FFFFB03A).w
		bclr	#1,($FFFFBF78).w
		move.w	#5,($FFFFB00E).w
		move.w	#$A000,($FFFFB008).w
		move.w	#7,($FFFFB00A).w
		move.w	#$C000,($FFFFB004).w
		move.w	#7,($FFFFB006).w
		move.w	#$F000,($FFFFB00C).w
		move.w	#$F800,($FFFFB002).w
		move.w	#$FC00,($FFFFB000).w
		move.w	#$BD28,a0
		moveq	#$1F,d1

loc_17A5C:							; CODE XREF: sub_17A18+46j
		clr.l	(a0)+
		dbf	d1,loc_17A5C
		bsr.w	VDP_FlushPalette
		bsr.w	VDP_InitialSetup
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_B425A,a2
		move.l	a2,a0
		move.l	a2,a1
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		clr.w	d4
		moveq	#$F,d5
		bsr.w	sub_1169A
		move.w	#$EEE,($FFFFBD28).w
		move.w	#0,($FFFFBD36).w
		move.w	#$18,($FFFFBD26).w
		move	#$2500,sr
		move.w	#$50,($FFFFD066).w			; 'P'

loc_17AB0:							; CODE XREF: sub_17A18+A8j
		moveq	#4,d0
		bsr.w	sub_11176
		tst.w	d1
		bne.w	loc_17AC2
		subq.w	#1,($FFFFD066).w
		bpl.s	loc_17AB0

loc_17AC2:							; CODE XREF: sub_17A18+A0j
		move	#$2700,sr
		rts
; End of function sub_17A18


; =============== S U B	R O U T	I N E =======================================


sub_17AC8:							; CODE XREF: sub_F739E+10ACp
		cmp.w	#2,($FFFFD048).w
		blt.w	locret_15464
		cmp.w	#4,($FFFFD048).w
		beq.w	locret_15464
		moveq	#7,d0
		tst.w	($FFFFD046).w
		beq.w	loc_17AEA
		move.w	#$B,d0

loc_17AEA:							; CODE XREF: sub_17AC8+1Aj
		sub.w	($FFFFD04A).w,d0
		move.w	d0,($FFFFCEF0).w
		rts
; End of function sub_17AC8


; =============== S U B	R O U T	I N E =======================================


sub_17AF4:							; CODE XREF: sub_F7B20+2Ep
		movem.l	d0-a2,-(sp)
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_17B0A
		tst.w	($FFFFD048).w
		bne.w	loc_17B16

loc_17B0A:							; CODE XREF: sub_17AF4+Aj
		move.w	($FFFFD04C).w,($FFFFC330).w
		move.w	($FFFFD04E).w,($FFFFC332).w

loc_17B16:							; CODE XREF: sub_17AF4+12j
		bsr.w	sub_11B92
; End of function sub_17AF4

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	7
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC332).w,d1
		move.w	d1,d0
		asl.w	#6,d0
		move.l	#word_F8BF4,a0
		move.l	$26(a0,d0.w),($FFFFBD4E).w
		move.w	($FFFFD43A).w,d4
		bsr.w	sub_17B78
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$16
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d1
		move.w	d1,d0
		asl.w	#6,d0
		move.l	#word_F8BF4,a0
		move.l	2(a0,d0.w),($FFFFBD4A).w
		move.w	#$EEE,($FFFFBD52).w
		move.w	#2,d4
		bsr.w	sub_17B78
		bsr.w	sub_17190
		move.w	#$64,($FFFFBD26).w			; 'd'
		movem.l	(sp)+,d0-a2
		rts

; =============== S U B	R O U T	I N E =======================================


sub_17B78:							; CODE XREF: ROM:00017B38p
					; ROM:00017B64p
		clr.w	d0
		asl.w	#1,d1
		move.l	#off_AFE12,a0
		move.l	a0,a1
		add.l	(a0),a0
		add.l	4(a1),a1
		move.w	#$30A,a2
		move.w	(a1),d2
		moveq	#2,d3
		moveq	#0,d5
		bra.w	sub_1169A
; End of function sub_17B78

; ---------------------------------------------------------------------------
		clr.w	d0
		asl.w	#1,d1
		move.l	#off_ABA14,a0
		move.l	a0,a1
		add.l	(a0),a0
		add.l	4(a1),a1
		move.w	#$30A,a2
		move.w	(a1),d2
		moveq	#2,d3
		moveq	#2,d4
		moveq	#2,d5
		bra.w	sub_1169A
; ---------------------------------------------------------------------------
		bsr.w	loc_17BCA
		moveq	#$C,d0
		moveq	#2,d1
		move.w	#$7FF,d2
		bra.w	sub_1197E
; ---------------------------------------------------------------------------

loc_17BCA:							; CODE XREF: ROM:00017BBAp
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	2
		dc.b	$A
		dc.b	0
; ---------------------------------------------------------------------------
		cmp.w	#$C6CE,a2
		bne.w	locret_15464
		move.w	#$1A,($FFFFB028).w
		rts
; ---------------------------------------------------------------------------
		movem.l	d0/a0,-(sp)
		clr.w	$26(a2)
		bsr.w	sub_17E42
		cmp.w	#$C6CE,a2
		seq	d0
		ext.w	d0
		addq.w	#2,d0
		cmp.w	($FFFFC328).w,d0
		beq.w	loc_17C3C
		cmp.w	($FFFFC32A).w,d0
		beq.w	loc_17C3C
		moveq	#0,d0
		cmp.w	#2,($FFFFCEEC).w
		bgt.w	loc_17C38
		tst.w	($FFFFD058).w
		bne.w	loc_17C38
		move.w	#$30E,a0
		move.w	(a2),d0
		asl.w	#2,d0
		move.l	(a0,d0.w),a0
		add.w	$A(a0),a0
		moveq	#4,d0
		bsr.w	sub_10E88
		bsr.w	sub_10EB4

loc_17C38:							; CODE XREF: ROM:00017C12j
					; ROM:00017C1Aj
		move.w	d0,$26(a2)

loc_17C3C:							; CODE XREF: ROM:00017BFEj
					; ROM:00017C06j
		movem.l	(sp)+,d0/a0
		rts
; ---------------------------------------------------------------------------

loc_17C42:							; DATA XREF: ROM:loc_F8168o
					; ROM:000FB024o
		movem.l	d0-a6,-(sp)
		btst	#2,($FFFFBF78).w
		bne.w	loc_17C62
		bclr	#0,($FFFFBF78).w
		beq.w	loc_17C5E
		bsr.w	loc_15E72

loc_17C5E:							; CODE XREF: ROM:00017C56j
		bsr.w	sub_10FB6

loc_17C62:							; CODE XREF: ROM:00017C4Cj
		addq.w	#1,($FFFFB03E).w
		jsr	sub_1A50A
		movem.l	(sp)+,d0-a6
		rte

; =============== S U B	R O U T	I N E =======================================


sub_17C72:							; CODE XREF: ROM:00007724p
		st	($FFFFD064).w
		move.l	#$FFFFD048,a0
		move.l	#word_17C8E,a1
		move.w	#8,d0

loc_17C86:							; CODE XREF: sub_17C72+16j
		move.w	(a1)+,(a0)+
		dbf	d0,loc_17C86
		rts
; End of function sub_17C72

; ---------------------------------------------------------------------------
word_17C8E:	dc.w	0					; DATA XREF: sub_17C72+Ao
		dc.w	1
		dc.w	$B
		dc.w	$A
		dc.w	1
		dc.w	0
		dc.w	1
		dc.w	0
		dc.w	1

; =============== S U B	R O U T	I N E =======================================


sub_17CA0:							; CODE XREF: sub_F76CEj sub_F7942+192p
		movem.l	d0-a3,-(sp)
		move.w	#$D088,a3
		bsr.w	sub_1803E
		move.w	($FFFFCEEC).w,d0
		or.w	($FFFFCEEA).w,d0
		beq.w	loc_17D10
		bsr.w	sub_17D80
		moveq	#7,d0
		tst.w	($FFFFD046).w
		beq.w	loc_17CCA
		move.w	#$B,d0

loc_17CCA:							; CODE XREF: sub_17CA0+22j
		tst.w	($FFFFD046).w
		bne.w	loc_17CE4
		cmp.w	#2,($FFFFCEF0).w
		blt.w	loc_17D08
		subq.w	#2,($FFFFCEF0).w
		bra.w	loc_17D08
; ---------------------------------------------------------------------------

loc_17CE4:							; CODE XREF: sub_17CA0+2Ej
		cmp.w	#1,($FFFFCEF0).w
		bne.w	loc_17CF8
		move.w	#3,($FFFFCEF0).w
		bra.w	loc_17D08
; ---------------------------------------------------------------------------

loc_17CF8:							; CODE XREF: sub_17CA0+4Aj
		cmp.w	#2,($FFFFCEF0).w
		bne.w	loc_17D08
		move.w	#4,($FFFFCEF0).w

loc_17D08:							; CODE XREF: sub_17CA0+38j
					; sub_17CA0+40j ...
		sub.w	($FFFFCEF0).w,d0
		move.w	d0,($FFFFD04A).w

loc_17D10:							; CODE XREF: sub_17CA0+14j
		movem.l	(sp)+,d0-a3
		rts
; End of function sub_17CA0


; =============== S U B	R O U T	I N E =======================================


sub_17D16:							; CODE XREF: sub_F76C8j
		moveq	#$20,d0					; ' '
		bsr.w	sub_11086

loc_17D1C:							; CODE XREF: sub_17D16+2Ej
		addq.w	#1,d0
		and.w	#$1F,d0
		asl.w	#4,d0
		move.l	#$5576,a0
		add.w	d0,a0
		lsr.w	#4,d0
		moveq	#$F,d1
		move.w	($FFFFD04C).w,d2
		cmpnw	 $19,d2
		bls.w	loc_17D3E
		moveq	#$19,d2

loc_17D3E:							; CODE XREF: sub_17D16+22j
					; sub_17D16+2Aj
		cmp.b	(a0)+,d2
		dbeq	d1,loc_17D3E
		bne.s	loc_17D1C
		eor.w	#$F,d1
		move.w	d1,($FFFFCEEE).w
		move.w	d0,($FFFFCEE8).w
		clr.w	($FFFFCEEC).w
		move.w	#7,($FFFFCEEA).w
		cmp.w	#2,($FFFFD048).w
		beq.w	sub_17D80
		clr.w	($FFFFCEEA).w
		moveq	#7,d0
		move.w	#$CE66,a0

loc_17D70:							; CODE XREF: sub_17D16+66j
		clr.w	4(a0)
		clr.w	6(a0)
		add.w	#$10,a0
		dbf	d0,loc_17D70
; End of function sub_17D16


; =============== S U B	R O U T	I N E =======================================


sub_17D80:							; CODE XREF: sub_17CA0+18p
					; sub_17D16+4Cj ...
		movem.l	d0-d4/a0-a3,-(sp)
		move.w	($FFFFCEE8).w,d1
		asl.w	#4,d1
		move.w	#$CEF4,a0
		move.l	#$5576,a1
		add.w	d1,a1
		move.l	(a1),(a0)
		move.l	4(a1),4(a0)
		move.l	8(a1),8(a0)
		move.l	$C(a1),$C(a0)
		lea	$10(a0),a1
		moveq	#$E,d2
		move.w	($FFFFCEF2).w,d0

loc_17DB4:							; CODE XREF: sub_17D80+42j
		move.w	d0,d1
		and.w	#1,d1
		move.b	(a0,d1.w),(a1)+
		addq.w	#2,a0
		lsr.w	#1,d0
		dbf	d2,loc_17DB4
		bsr.w	sub_1828A
		tst.w	d1
		bmi.w	loc_17DFE
		move.w	#$CEF4,a0
		add.w	d2,a0
		add.w	d2,a0
		move.w	#$CE66,a1

loc_17DDC:							; CODE XREF: sub_17D80+76j
		clr.w	$A(a1)
		clr.w	$C(a1)
		clr.w	8(a1)
		bclr	#1,$E(a1)
		bsr.w	sub_17E04
		add.w	#$10,a1
		dbf	d1,loc_17DDC
		bsr.w	sub_17E42

loc_17DFE:							; CODE XREF: sub_17D80+4Cj
		movem.l	(sp)+,d0-d4/a0-a3
		rts
; End of function sub_17D80


; =============== S U B	R O U T	I N E =======================================


sub_17E04:							; CODE XREF: sub_17D80+6Ep
		cmp.w	#2,($FFFFCEEA).w
		beq.w	loc_17E32
		cmp.w	#3,($FFFFCEEA).w
		beq.w	loc_17E32
		cmp.w	#5,($FFFFCEEA).w
		beq.w	loc_17E32
		bset	#0,$E(a1)
		move.b	(a0)+,3(a1)
		move.b	(a0)+,1(a1)
		rts
; ---------------------------------------------------------------------------

loc_17E32:							; CODE XREF: sub_17E04+6j
					; sub_17E04+10j ...
		bclr	#0,$E(a1)
		move.b	(a0)+,1(a1)
		move.b	(a0)+,3(a1)
		rts
; End of function sub_17E04


; =============== S U B	R O U T	I N E =======================================


sub_17E42:							; CODE XREF: ROM:00017BECp
					; sub_17D80+7Ap ...
		movem.l	d0-d3/a0-a1,-(sp)
		clr.w	($FFFFC328).w
		clr.w	($FFFFC32A).w
		clr.w	($FFFFC32C).w
		clr.w	($FFFFC32E).w
		tst.w	($FFFFD064).w
		beq.w	loc_17FDC
		tst.w	($FFFFD048).w
		beq.w	loc_17F9A
		btst	#0,($FFFFC2FA).w
		bne.w	loc_17F9A
		bsr.w	sub_1828A
		move.w	#$CEF4,a0
		move.w	($FFFFCEEE).w,d2
		move.b	(a0,d2.w),d2
		move.w	#$CE66,a1
		moveq	#$10,d4
		mulu.w	d1,d4
		add.w	d4,a1
		st	($FFFFCEE6).w
		clr.w	d0

loc_17E90:							; CODE XREF: sub_17E42+60j
		cmp.w	(a1),d2
		beq.w	loc_17EC6
		cmp.w	2(a1),d2
		beq.w	loc_17EAA
		suba.w	#$10,a1
		dbf	d1,loc_17E90
		bra.w	loc_17FE0
; ---------------------------------------------------------------------------

loc_17EAA:							; CODE XREF: sub_17E42+58j
		moveq	#4,d0
		tst.w	($FFFFD046).w
		beq.w	loc_17EB8
		move.w	#5,d0

loc_17EB8:							; CODE XREF: sub_17E42+6Ej
		move.w	(a1),($FFFFD04E).w
		move.w	2(a1),($FFFFD04C).w
		bra.w	loc_17ED2
; ---------------------------------------------------------------------------

loc_17EC6:							; CODE XREF: sub_17E42+50j
		clr.w	d0
		move.w	2(a1),($FFFFD04E).w
		move.w	(a1),($FFFFD04C).w

loc_17ED2:							; CODE XREF: sub_17E42+80j
		add.w	($FFFFCEF0).w,d0
		move.w	d1,($FFFFCEE6).w
		move.w	(a1),($FFFFC330).w
		move.w	2(a1),($FFFFC332).w
		asl.w	#2,d0
		lea	word_17F52(pc),a0
		tst.w	($FFFFD046).w
		beq.w	loc_17EF6
		lea	word_17F72(pc),a0

loc_17EF6:							; CODE XREF: sub_17E42+ACj
		move.w	(a0,d0.w),($FFFFC328).w
		move.w	2(a0,d0.w),($FFFFC32A).w
		tst.w	($FFFFD046).w
		beq.w	loc_17FE0
		cmpnw	 4,d0
		beq.w	loc_17F1E
		cmpnw	 $18,d0
		beq.w	loc_17F1E
		bra.w	loc_17F2E
; ---------------------------------------------------------------------------

loc_17F1E:							; CODE XREF: sub_17E42+CCj
					; sub_17E42+D4j
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	($FFFFC32A).w,($FFFFC32E).w
		bra.w	loc_17FE0
; ---------------------------------------------------------------------------

loc_17F2E:							; CODE XREF: sub_17E42+D8j
		cmpnw	 8,d0
		beq.w	loc_17F42
		cmpnw	 $1C,d0
		beq.w	loc_17F42
		bra.w	loc_17FE0
; ---------------------------------------------------------------------------

loc_17F42:							; CODE XREF: sub_17E42+F0j
					; sub_17E42+F8j
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	#0,($FFFFC32E).w
		bra.w	loc_17FE0
; ---------------------------------------------------------------------------
word_17F52:	dc.w	1					; DATA XREF: sub_17E42+A4o
		dc.w	0
		dc.w	1
		dc.w	1
		dc.w	1
		dc.w	2
		dc.w	2
		dc.w	1
		dc.w	2
		dc.w	0
		dc.w	2
		dc.w	2
		dc.w	2
		dc.w	1
		dc.w	1
		dc.w	2
word_17F72:	dc.w	1					; DATA XREF: sub_17E42+B0o
		dc.w	0
		dc.w	1
		dc.w	2
		dc.w	1
		dc.w	2
		dc.w	1
		dc.w	1
		dc.w	1
		dc.w	2
		dc.w	2
		dc.w	0
		dc.w	2
		dc.w	1
		dc.w	2
		dc.w	1
		dc.w	2
		dc.w	2
		dc.w	2
		dc.w	1
; ---------------------------------------------------------------------------

loc_17F9A:							; CODE XREF: sub_17E42+20j
					; sub_17E42+2Aj
		move.w	($FFFFD04A).w,d0
		asl.w	#2,d0
		lea	unk_17FE6(pc),a0
		move.w	(a0,d0.w),($FFFFC328).w
		move.w	2(a0,d0.w),($FFFFC32A).w
		cmpnw	 $14,d0
		bne.w	loc_17FC8
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	#0,($FFFFC32E).w
		bra.w	loc_17FDC
; ---------------------------------------------------------------------------

loc_17FC8:							; CODE XREF: sub_17E42+172j
		cmpnw	 $18,d0
		bne.w	loc_17FDC
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	($FFFFC32A).w,($FFFFC32E).w

loc_17FDC:							; CODE XREF: sub_17E42+18j
					; sub_17E42+182j ...
		bsr.w	sub_18002

loc_17FE0:							; CODE XREF: sub_17E42+64j
					; sub_17E42+C4j ...
		movem.l	(sp)+,d0-d3/a0-a1
		rts
; End of function sub_17E42

; ---------------------------------------------------------------------------
unk_17FE6:	dc.b	0					; DATA XREF: sub_17E42+15Eo
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	2

; =============== S U B	R O U T	I N E =======================================


sub_18002:							; CODE XREF: sub_17E42:loc_17FDCp
		st	($FFFFCEE6).w
		clr.l	d3
		move.w	($FFFFC330).w,d1
		bset	d1,d3
		move.w	($FFFFC332).w,d1
		bset	d1,d3
		move.w	#$CE66,a1
		moveq	#7,d2

loc_1801A:							; CODE XREF: sub_18002+2Aj
		bsr.w	sub_18032
		move.w	d0,(a1)
		bsr.w	sub_18032
		move.w	d0,2(a1)
		add.w	#$10,a1
		dbf	d2,loc_1801A
		rts
; End of function sub_18002


; =============== S U B	R O U T	I N E =======================================


sub_18032:							; CODE XREF: sub_18002:loc_1801Ap
					; sub_18002+1Ep ...
		moveq	#$1A,d0
		bsr.w	sub_11086
		bset	d0,d3
		bne.s	sub_18032
		rts
; End of function sub_18032


; =============== S U B	R O U T	I N E =======================================


sub_1803E:							; CODE XREF: sub_17CA0+8p
					; sub_180FC+5Cp ...
		moveq	#4,d0
		lea	$A(a3),a0

loc_18044:							; CODE XREF: sub_1803E+8j
		move.w	-(a0),-(sp)
		dbf	d0,loc_18044
		moveq	#7,d2
		move.w	#$CED6,a1

loc_18050:							; CODE XREF: sub_1803E+4Cj
		bclr	#2,$E(a1)
		moveq	#5,d0
		bsr.w	sub_1826C
		move.w	d0,6(a1)
		cmpnw	 4,d0
		bne.w	loc_1806E
		bset	#2,$E(a1)

loc_1806E:							; CODE XREF: sub_1803E+26j
		moveq	#5,d0
		bsr.w	sub_1826C
		move.w	d0,4(a1)
		cmpnw	 4,d0
		bne.w	loc_18086
		bset	#2,$E(a1)

loc_18086:							; CODE XREF: sub_1803E+3Ej
		suba.w	#$10,a1
		dbf	d2,loc_18050
		move.w	#$4000,d0
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEF2).w
		move.w	#8,d0
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEF0).w
		tst.w	($FFFFD046).w
		bne.w	loc_180C8
		cmp.w	#1,($FFFFCEF0).w
		beq.w	loc_180C2
		cmp.w	#2,($FFFFCEF0).w
		bne.w	loc_180C8

loc_180C2:							; CODE XREF: sub_1803E+76j
		move.w	#4,($FFFFCEF0).w

loc_180C8:							; CODE XREF: sub_1803E+6Cj
					; sub_1803E+80j
		moveq	#$10,d0
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEEE).w
		moveq	#4,d0
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEEC).w
		moveq	#8,d0
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEEA).w
		moveq	#$20,d0					; ' '
		bsr.w	sub_1826C
		move.w	d0,($FFFFCEE8).w
		moveq	#4,d0
		lea	(a3),a0

loc_180F4:							; CODE XREF: sub_1803E+B8j
		move.w	(sp)+,(a0)+
		dbf	d0,loc_180F4
		rts
; End of function sub_1803E


; =============== S U B	R O U T	I N E =======================================


sub_180FC:							; CODE XREF: sub_7CB0:loc_172C8p
		tst.w	($FFFFD04A).w
		beq.w	locret_15464
		tst.w	($FFFFD048).w
		beq.w	locret_15464
		cmp.w	#4,($FFFFD048).w
		beq.w	locret_15464
		move.w	#1,($FFFFD048).w
		move.w	#1,($FFFFD074).w
		bsr.w	sub_18380
		bsr.w	sub_17D80
		cmp.w	#4,($FFFFCEEC).w
		beq.w	loc_18160
		tst.w	($FFFFCEE6).w
		bmi.w	loc_18160
		move.w	#$D088,a3
		bsr.w	sub_18192
		jsr	sub_FE696
		btst	#2,($FFFFC2F0).w
		beq.w	locret_15464
		move.w	#$D176,a3
		bsr.w	sub_1803E
		bra.w	sub_17D80
; ---------------------------------------------------------------------------

loc_18160:							; CODE XREF: sub_180FC+34j
					; sub_180FC+3Cj
		move.w	#$D088,a3
		bsr.w	sub_1820E
		jsr	sub_FE696
		move.w	#2,($FFFFD048).w
		move.w	#2,($FFFFD074).w
		cmp.w	#7,($FFFFCEEA).w
		beq.w	locret_15464
		move.w	#3,($FFFFD048).w
		move.w	#3,($FFFFD074).w
		rts
; End of function sub_180FC


; =============== S U B	R O U T	I N E =======================================


sub_18192:							; CODE XREF: sub_180FC+44p
					; sub_18380+100p
		bsr.w	sub_1820E
		move.w	($FFFFCEE8).w,d0
		moveq	#$20,d1					; ' '
		bsr.w	sub_181F6
		move.w	($FFFFCEEA).w,d0
		moveq	#8,d1
		bsr.w	sub_181F6
		move.w	($FFFFCEEC).w,d0
		moveq	#4,d1
		bsr.w	sub_181F6
		move.w	($FFFFCEEE).w,d0
		moveq	#$10,d1
		bsr.w	sub_181F6
		move.w	($FFFFCEF0).w,d0
		moveq	#8,d1
		bsr.w	sub_181F6
		move.w	($FFFFCEF2).w,d0
		move.w	#$4000,d1
		bsr.w	sub_181F6
		moveq	#5,d1
		moveq	#7,d2
		move.w	#$CE66,a1

loc_181DC:							; CODE XREF: sub_18192+5Ej
		move.w	4(a1),d0
		bsr.w	sub_181F6
		move.w	6(a1),d0
		bsr.w	sub_181F6
		add.w	#$10,a1
		dbf	d2,loc_181DC
		rts
; End of function sub_18192


; =============== S U B	R O U T	I N E =======================================


sub_181F6:							; CODE XREF: sub_18192+Ap
					; sub_18192+14p ...
		movem.l	d0-d1,-(sp)
		exg	d0,d1
		bsr.w	sub_18236
		clr.l	d0
		move.w	d1,d0
		bsr.w	sub_1821A
		movem.l	(sp)+,d0-d1
		rts
; End of function sub_181F6


; =============== S U B	R O U T	I N E =======================================


sub_1820E:							; CODE XREF: sub_180FC+68p sub_18192p
		move.w	a3,a0
		moveq	#4,d0

loc_18212:							; CODE XREF: sub_1820E+6j
		clr.w	(a0)+
		dbf	d0,loc_18212
		rts
; End of function sub_1820E


; =============== S U B	R O U T	I N E =======================================


sub_1821A:							; CODE XREF: sub_181F6+Ep
		movem.l	d1/a0,-(sp)
		lea	$A(a3),a0
		moveq	#3,d1
		add.l	d0,-(a0)
		bra.w	loc_1822C
; ---------------------------------------------------------------------------

loc_1822A:							; CODE XREF: sub_1821A:loc_1822Cj
		addq.w	#1,-(a0)

loc_1822C:							; CODE XREF: sub_1821A+Cj
		dbcc	d1,loc_1822A
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_1821A


; =============== S U B	R O U T	I N E =======================================


sub_18236:							; CODE XREF: sub_181F6+6p
		movem.l	d1-d4/a0,-(sp)
		move.w	a3,a0
		moveq	#4,d4

loc_1823E:							; CODE XREF: sub_18236+Cj
		move.w	(a0),-(sp)
		clr.w	(a0)+
		dbf	d4,loc_1823E
		moveq	#4,d4

loc_18248:							; CODE XREF: sub_18236+2Cj
		move.w	d0,d1
		mulu.w	(sp)+,d1
		lea	2(a3),a0
		add.w	d4,a0
		add.w	d4,a0
		move.w	d4,d2
		add.l	d1,-(a0)
		bra.w	loc_1825E
; ---------------------------------------------------------------------------

loc_1825C:							; CODE XREF: sub_18236:loc_1825Ej
		addq.w	#1,-(a0)

loc_1825E:							; CODE XREF: sub_18236+22j
		dbcc	d2,loc_1825C
		dbf	d4,loc_18248
		movem.l	(sp)+,d1-d4/a0
		rts
; End of function sub_18236


; =============== S U B	R O U T	I N E =======================================


sub_1826C:							; CODE XREF: sub_1803E+1Ap
					; sub_1803E+32p ...
		movem.l	d1-d2/a0,-(sp)
		move.w	a3,a0
		moveq	#4,d1
		clr.l	d2

loc_18276:							; CODE XREF: sub_1826C+10j
		move.w	(a0),d2
		divu.w	d0,d2
		move.w	d2,(a0)+
		dbf	d1,loc_18276
		swap	d2
		move.w	d2,d0
		movem.l	(sp)+,d1-d2/a0
		rts
; End of function sub_1826C


; =============== S U B	R O U T	I N E =======================================


sub_1828A:							; CODE XREF: ROM:0000813Cp
					; sub_130E6+76p ...
		move.l	d0,-(sp)
		moveq	#$FFFFFFF0,d2
		moveq	#$10,d1
		move.w	($FFFFCEEC).w,d0

loc_18294:							; CODE XREF: sub_1828A+Ej
		add.w	d1,d2
		lsr.w	#1,d1
		dbf	d0,loc_18294
		subq.w	#1,d1
		move.l	(sp)+,d0
		rts
; End of function sub_1828A


; =============== S U B	R O U T	I N E =======================================


sub_182A2:							; CODE XREF: sub_7CB0+F602p
		cmp.w	#1,($FFFFD048).w
		blt.w	locret_15464
		bsr.w	sub_1833A
		move.w	#$CEF4,a0
		move.w	($FFFFCEEE).w,d2
		move.b	(a0,d2.w),d2
		move.w	#$C6CE,a2
		cmp.w	$28(a2),d2
		beq.w	loc_182CC
		add.w	#$364,a2

loc_182CC:							; CODE XREF: sub_182A2+22j
		add.w	#$B4,a2					; ''
		moveq	#$67,d0					; 'g'
		move.w	#$CD96,a1

loc_182D6:							; CODE XREF: sub_182A2+3Aj
		clr.w	d1
		move.b	(a2)+,d1
		add.w	d1,(a1)+
		dbf	d0,loc_182D6
		move.w	#$CD96,a0
		move.l	#byte_18336,a1
		move.w	#$D092,a2
		moveq	#$67,d0					; 'g'
		clr.w	d4

loc_182F2:							; CODE XREF: sub_182A2+8Ej
		clr.l	d1
		move.w	(a0)+,d1
		move.w	d0,d2
		and.w	#3,d2
		move.b	(a1,d2.w),d2
		clr.l	d3
		bset	d2,d3
		subq.w	#1,d3
		cmp.w	d3,d1
		ble.w	loc_1830E
		move.w	d3,d1

loc_1830E:							; CODE XREF: sub_182A2+66j
		not.l	d3
		move.w	d4,d5
		and.w	#$F,d5
		rol.l	d5,d3
		rol.l	d5,d1
		move.w	d4,d5
		lsr.w	#4,d5
		add.w	d5,d5
		neg.w	d5
		add.w	#$100,d5
		and.l	d3,-$22(a2,d5.w)
		or.l	d1,-$22(a2,d5.w)
		add.w	d2,d4
		dbf	d0,loc_182F2
		rts
; End of function sub_182A2

; ---------------------------------------------------------------------------
byte_18336:	dc.b	$C					; DATA XREF: sub_182A2+42o
					; sub_1833A+4o
		dc.b	$E
		dc.b	$A
		dc.b	$A

; =============== S U B	R O U T	I N E =======================================


sub_1833A:							; CODE XREF: sub_9428+4p sub_182A2+Ap
		move.w	#$CD96,a0
		move.l	#byte_18336,a1
		move.w	#$D092,a2
		moveq	#$67,d0					; 'g'
		clr.w	d4

loc_1834C:							; CODE XREF: sub_1833A+40j
		move.w	d4,d5
		lsr.w	#4,d5
		add.w	d5,d5
		neg.w	d5
		add.w	#$100,d5
		move.l	-$22(a2,d5.w),d1
		move.w	d4,d5
		and.w	#$F,d5
		lsr.l	d5,d1
		move.w	d0,d2
		and.w	#3,d2
		move.b	(a1,d2.w),d2
		add.w	d2,d4
		clr.w	d3
		bset	d2,d3
		subq.w	#1,d3
		and.w	d3,d1
		move.w	d1,(a0)+
		dbf	d0,loc_1834C
		rts
; End of function sub_1833A


; =============== S U B	R O U T	I N E =======================================


sub_18380:							; CODE XREF: sub_180FC+26p
		move.w	($FFFFCEE6).w,d0
		mulu.w	#$10,d0
		move.w	#$CE66,a0
		move.w	($FFFFC6DA).w,$A(a0,d0.w)
		move.w	($FFFFCA3E).w,$C(a0,d0.w)
		bsr.w	sub_1828A
		move.w	#$CE66,a0
		moveq	#$10,d3
		mulu.w	d1,d3
		add.w	d3,a0
		cmp.w	#7,($FFFFCEEA).w
		beq.w	loc_184D0

loc_183B0:							; CODE XREF: sub_18380+6Cj
		cmp.w	#4,4(a0)
		beq.w	loc_183E8
		cmp.w	#4,6(a0)
		beq.w	loc_183E8
		clr.w	d3
		btst	#0,$E(a0)
		beq.w	loc_183D4
		eor.w	#2,d3

loc_183D4:							; CODE XREF: sub_18380+4Cj
		move.w	$A(a0),d0
		sub.w	$C(a0),d0
		bpl.w	loc_183E4
		eor.w	#2,d3

loc_183E4:							; CODE XREF: sub_18380+5Cj
		addq.w	#1,4(a0,d3.w)

loc_183E8:							; CODE XREF: sub_18380+36j
					; sub_18380+40j
		suba.w	#$10,a0
		dbf	d1,loc_183B0
		addq.w	#1,($FFFFCEEA).w
		cmp.w	#7,($FFFFCEEA).w
		beq.w	loc_1848A
		move.w	($FFFFCEE6).w,d0
		mulu.w	#$10,d0
		move.w	#$CE66,a0
		add.w	d0,a0
		cmp.w	#4,4(a0)
		beq.w	loc_18420
		cmp.w	#4,6(a0)
		bne.w	locret_15464

loc_18420:							; CODE XREF: sub_18380+92j
		cmp.w	#3,($FFFFCEEC).w
		bge.w	loc_1848A
		bsr.w	sub_1828A
		move.w	#$CE66,a0
		moveq	#$10,d3
		mulu.w	d1,d3
		add.w	d3,a0

loc_18438:							; CODE XREF: sub_18380+E8j
					; sub_18380+F2j ...
		cmp.w	($FFFFCEE6).w,d1
		beq.w	loc_18474
		cmp.w	#4,4(a0)
		beq.w	loc_18474
		cmp.w	#4,6(a0)
		beq.w	loc_18474
		addq.w	#1,4(a0)
		move.l	#$C8,d0					; ''
		jsr	sub_11086
		and.w	#1,d0
		beq.s	loc_18438
		subq.w	#1,4(a0)
		addq.w	#1,6(a0)
		bra.s	loc_18438
; ---------------------------------------------------------------------------

loc_18474:							; CODE XREF: sub_18380+BCj
					; sub_18380+C6j ...
		suba.w	#$10,a0
		dbf	d1,loc_18438
		move.w	#$D176,a3
		bsr.w	sub_18192
		bset	#2,($FFFFC2F0).w

loc_1848A:							; CODE XREF: sub_18380+7Aj
					; sub_18380+A6j
		clr.w	($FFFFCEEA).w
		bsr.w	sub_1828A
		move.w	#$CE66,a0
		moveq	#$10,d3
		mulu.w	d1,d3
		add.w	d3,a0
		clr.w	d3

loc_1849E:							; CODE XREF: sub_18380+136j
		cmp.w	#4,4(a0)
		beq.w	loc_184AA
		bset	d1,d3

loc_184AA:							; CODE XREF: sub_18380+124j
		clr.w	4(a0)
		clr.w	6(a0)
		suba.w	#$10,a0
		dbf	d1,loc_1849E
		moveq	#1,d1
		asl.w	d2,d1
		subq.w	#1,d1
		and.w	d1,($FFFFCEF2).w
		asl.w	d2,d3
		or.w	d3,($FFFFCEF2).w
		addq.w	#1,($FFFFCEEC).w
		rts
; ---------------------------------------------------------------------------

loc_184D0:							; CODE XREF: sub_18380+2Cj
		clr.w	d3

loc_184D2:							; CODE XREF: sub_18380+170j
		move.w	$A(a0),d0
		cmp.w	$C(a0),d0
		bhi.w	loc_184E0
		bset	d1,d3

loc_184E0:							; CODE XREF: sub_18380+15Aj
		btst	#0,$E(a0)
		beq.w	loc_184EC
		bchg	d1,d3

loc_184EC:							; CODE XREF: sub_18380+166j
		suba.w	#$10,a0
		dbf	d1,loc_184D2
		moveq	#1,d1
		asl.w	d2,d1
		subq.w	#1,d1
		and.w	d1,($FFFFCEF2).w
		asl.w	d2,d3
		or.w	d3,($FFFFCEF2).w
		addq.w	#1,($FFFFCEEC).w
		rts
; End of function sub_18380

; ---------------------------------------------------------------------------

loc_1850A:							; CODE XREF: sub_125E8+20p
		cmp.w	#$40,($FFFFC3F0).w			; '@'
		bgt.w	locret_15464
		bset	#7,($FFFFC2EA).w
		bne.w	locret_15464
		movem.l	d0/a1,-(sp)
		bset	#3,($FFFFBF78).w
		move.w	($FFFFB03E).w,d0

loc_1852C:							; CODE XREF: ROM:00018530j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_1852C
		movem.l	(sp)+,d0/a1
		movem.l	d0-d5/a0-a4,-(sp)
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	2
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1C,d0
		moveq	#9,d1
		bsr.w	sub_119B8
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1A
		dc.b	$BF
		dc.b	9
		dc.b	$10
aStarsOfTheGame: dc.b	'Stars of the Game'
		dc.b	$BF
		dc.b	3
		dc.b	$F
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_5605A,a1
		move.w	#$30A,a2
		add.l	4(a1),a1
		moveq	#$D,d0
		moveq	#$5B,d1					; '['
		moveq	#6,d2
		moveq	#4,d3
		move.w	($FFFFB018).w,d4
		clr.w	d5
		bsr.w	sub_1169A
		bsr.w	sub_18620
		move.w	#$12,($FFFFB02A).w
		moveq	#2,d2

loc_18596:							; CODE XREF: ROM:000185C0j
		bsr.w	sub_185E0
		move.w	#$1A,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		move.l	$1E(a2),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		bsr.w	sub_11BA4
		bsr.w	sub_18A90
		move.w	#3,($FFFFB028).w
		bsr.w	sub_11BA4
		dbf	d2,loc_18596
		move.w	($FFFFC6DA).w,d0
		sub.w	($FFFFCA3E).w,d0
		ble.w	loc_185DA
		move.w	#$F,-(sp)
		jsr	sub_11156

loc_185DA:							; CODE XREF: ROM:000185CCj
		movem.l	(sp)+,d0-d5/a0-a4
		rts

; =============== S U B	R O U T	I N E =======================================


sub_185E0:							; CODE XREF: ROM:loc_18596p
		movem.l	d1-d2/a1/a4,-(sp)
		move.w	#$CF36,a4
		clr.l	d0
		moveq	#$33,d2					; '3'

loc_185EC:							; CODE XREF: sub_185E0:loc_185F8j
		cmp.l	(a4)+,d0
		bge.w	loc_185F8
		lea	-4(a4),a1
		move.l	(a1),d0

loc_185F8:							; CODE XREF: sub_185E0+Ej
		dbf	d2,loc_185EC
		clr.l	(a1)
		move.w	a1,d0
		sub.w	#$CF36,d0
		lsr.w	#2,d0
		move.w	#$C6CE,a2
		cmpnw	 $1A,d0
		blt.w	loc_1861A
		sub.w	#$1A,d0
		add.w	#$364,a2

loc_1861A:							; CODE XREF: sub_185E0+2Ej
		movem.l	(sp)+,d1-d2/a1/a4
		rts
; End of function sub_185E0


; =============== S U B	R O U T	I N E =======================================


sub_18620:							; CODE XREF: ROM:0001858Ap
		move.w	#$CF36,a4
		jsr	(sub_784E).w
		move.w	d0,d5
		add.w	d2,d5
		move.w	#$C6CE,a2
		lea	$364(a2),a3
		bsr.w	sub_1867E
		move.w	a3,a2
		lea	-$364(a2),a3
		bsr.w	sub_1867E
		cmp.w	#3,($FFFFC466).w
		bne.w	locret_1867C
		tst.w	d3
		beq.w	locret_1867C
		move.w	#$C46E,a0
		add.w	($FFFFC472).w,a0
		clr.w	d0
		btst	#7,2(a0)
		beq.w	loc_1866A
		add.w	#$1A,d0

loc_1866A:							; CODE XREF: sub_18620+42j
		add.b	3(a0),d0
		asl.w	#2,d0
		move.w	#$CF36,a0
		move.l	#$7FFFFFFF,(a0,d0.w)

locret_1867C:							; CODE XREF: sub_18620+28j
					; sub_18620+2Ej
		rts
; End of function sub_18620


; =============== S U B	R O U T	I N E =======================================


sub_1867E:							; CODE XREF: sub_18620+14p
					; sub_18620+1Ep
		move.w	a2,a1
		move.w	$C(a2),d3
		sub.w	$C(a3),d3
		ext.l	d3
		moveq	#$19,d4
		jsr	sub_9F40
		neg.w	d0
		add.w	d4,d0

loc_18696:							; CODE XREF: sub_1867E+98j
		move.l	d3,(a4)
		cmp.w	d0,d4
		bhi.w	loc_186D8
		clr.w	d1
		move.b	$B4(a2),d1
		mulu.w	#$2AF8,d1
		add.l	d1,(a4)
		clr.w	d1
		move.b	$CE(a2),d1
		mulu.w	#$2774,d1
		add.l	d1,(a4)
		clr.w	d1
		move.b	$E8(a2),d1
		mulu.w	#$A,d1
		tst.w	d3
		bne.w	loc_186D2
		mulu.w	#$64,d1					; 'd'
		add.l	d1,(a4)
		clr.l	d1
		add.w	$136(a1),d1

loc_186D2:							; CODE XREF: sub_1867E+44j
		add.l	d1,(a4)
		bra.w	loc_18710
; ---------------------------------------------------------------------------

loc_186D8:							; CODE XREF: sub_1867E+1Cj
		cmp.w	$136(a1),d5
		bhi.w	loc_18710
		clr.w	d1
		move.b	$B4(a2),d1
		mulu.w	#$64,d1					; 'd'
		clr.w	d2
		move.b	$E8(a2),d2
		beq.w	loc_18710
		divu.w	d2,d1
		cmpnw	 4,d1
		bhi.w	loc_18710
		add.l	#$7D00,(a4)
		tst.w	d1
		bne.w	loc_18710
		add.l	#$124F8,(a4)

loc_18710:							; CODE XREF: sub_1867E+56j
					; sub_1867E+5Ej ...
		addq.w	#2,a1
		addq.w	#1,a2
		addq.w	#4,a4
		dbf	d4,loc_18696
		rts
; End of function sub_1867E

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_79EA

loc_1871C:							; CODE XREF: sub_79EA+8j
		movem.l	d0-d2/a0-a4,-(sp)
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$B
		dc.b	3
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$12,d0
		moveq	#5,d1
		bsr.w	sub_119B8
		btst	#5,($FFFFC2FC).w
		bne.w	loc_18768
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$24
		dc.b	$BF
		dc.b	$C
		dc.b	4
aInjuryTo:	dc.b	'Injury to:'
; END OF FUNCTION CHUNK	FOR sub_79EA
		dc.b	$BF
		dc.b	$C
		dc.b	6
aOutForPeriod:	dc.b	'Out for period'
		dc.b	$BF
		dc.b	$C
		dc.b	5
		dc.b	0
; ---------------------------------------------------------------------------
		bra.w	loc_1878E
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_79EA

loc_18768:							; CODE XREF: sub_79EA+10D4Ej
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$22
		dc.b	$BF
		dc.b	$C
		dc.b	4
aInjuryTo_0:	dc.b	'Injury to:'
; END OF FUNCTION CHUNK	FOR sub_79EA
		dc.b	$BF
		dc.b	$C
		dc.b	6
aOutForGame:	dc.b	'Out for game'
		dc.b	$BF
		dc.b	$C
		dc.b	5
		dc.b	0
; ---------------------------------------------------------------------------

loc_1878E:							; CODE XREF: ROM:00018764j
		bsr.w	sub_18AC6
		bsr.w	sub_11BA4
		btst	#0,($FFFFB7AD).w
		beq.w	loc_187B2
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	$BF
		dc.b	$14
		dc.b	6
aTheGame:	dc.b	'the game',0
; ---------------------------------------------------------------------------

loc_187B2:							; CODE XREF: ROM:0001879Cj
		movem.l	(sp)+,d0-d2/a0-a4
		rts
; ---------------------------------------------------------------------------

loc_187B8:							; CODE XREF: sub_12002:loc_12066p
		bset	#7,($FFFFC2FA).w
		btst	#0,($FFFFC2FA).w
		beq.w	loc_187CE
		jmp	loc_FC320
; ---------------------------------------------------------------------------

loc_187CE:							; CODE XREF: ROM:000187C4j
		movem.l	d0-d2/a0-a4,-(sp)
		move.w	#$FFFF,d0
		jsr	sub_1277A
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	3
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1B,d0
		moveq	#8,d1
		bsr.w	sub_119B8
		lea	word_18884(pc),a1
		bsr.w	sub_11DF4
		move.w	($FFFFD406).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a2
		add.w	d0,a2
		clr.w	d0
		move.b	$66(a2),d0
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD404).w
		beq.w	loc_1881E
		move.l	#$FFFFCA32,a2

loc_1881E:							; CODE XREF: ROM:00018814j
		bsr.w	sub_18AE8
		bsr.w	sub_11BA4
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	4
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		movem.l	a1/a3,-(sp)
		move.w	($FFFFD410).w,d0
		move.l	#word_18E0C,a1
		add.w	(a1,d0.w),a1
		lea	2(a1),a1
		bsr.w	sub_11BA4
		movem.l	(sp)+,a1/a3
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$20
		dc.b	$62
		dc.b	$79
		dc.b	$BF
		dc.b	4
		dc.b	8
; ---------------------------------------------------------------------------
		move.w	($FFFFD40E).w,d0
		move.l	#$FFFFCA32,a2
		tst.w	($FFFFD404).w
		beq.w	loc_18872
		move.l	#$FFFFC6CE,a2

loc_18872:							; CODE XREF: ROM:00018868j
		jsr	sub_18AE8
		jsr	sub_11BA4
		movem.l	(sp)+,d0-d2/a0-a4
		rts
; ---------------------------------------------------------------------------
word_18884:	dc.w	$16					; DATA XREF: ROM:000187EEo
		dc.b	$BF
		dc.b	4
		dc.b	3
aPenaltyShot:	dc.b	'PENALTY SHOT!'
		dc.b	$BF
		dc.b	4
		dc.b	5
		dc.b	0
; ---------------------------------------------------------------------------

loc_1889A:							; CODE XREF: sub_12638+1Ep
		movem.l	d0-d2/a0-a4,-(sp)
		btst	#2,($FFFFC2F2).w
		beq.w	loc_188AE
		bset	#2,($FFFFC2EE).w

loc_188AE:							; CODE XREF: ROM:000188A4j
		move.w	#$C6CE,a2
		jsr	sub_BB36
		add.w	#$364,a2
		jsr	sub_BB36
		move.w	#$C46E,a4
		add.w	($FFFFC472).w,a4
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$B
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$13,d0
		move.w	#5,d1
		tst.b	4(a4)
		bmi.w	loc_188EE
		addq.w	#2,d1
		tst.b	5(a4)
		bmi.w	loc_188EE
		addq.w	#1,d1

loc_188EE:							; CODE XREF: ROM:000188DEj
					; ROM:000188E8j
		bsr.w	sub_119B8
		jsr	sub_F9FC0
		move.w	#$C6CE,a2
		move.w	($FFFFD448).w,($FFFFBF48).w
		btst	#7,2(a4)
		beq.w	loc_18916
		add.w	#$364,a2
		move.w	($FFFFD44A).w,($FFFFBF48).w

loc_18916:							; CODE XREF: ROM:00018908j
		lea	word_18A14(pc),a1
		bclr	#0,($FFFFC300).w
		beq.w	loc_18928
		lea	word_18A34(pc),a1

loc_18928:							; CODE XREF: ROM:00018920j
		clr.w	d0
		move.b	3(a4),d0
		add.w	#$B4,d0					; ''
		cmp.b	#3,(a2,d0.w)
		bne.w	loc_1896E
		movem.w	d1,-(sp)
		clr.w	d1
		move.b	3(a4),d1
		cmp.w	($FFFFBF48).w,d1
		movem.w	(sp)+,d1
		blt.w	loc_1896E
		add.w	(a1),a1
		move.w	d0,-(sp)
		move.w	$28(a2),d0
		cmp.w	($FFFFC330).w,d0
		bne.w	loc_1896C
		move.w	#0,d0
		jsr	sub_FE510

loc_1896C:							; CODE XREF: ROM:0001895Ej
		move.w	(sp)+,d0

loc_1896E:							; CODE XREF: ROM:00018938j
					; ROM:0001894Ej
		bsr.w	sub_11DF4
		clr.w	d0
		move.b	3(a4),d0
		move.w	d0,-(sp)
		move.w	#$C,($FFFFB028).w
		bsr.w	sub_18B26
		bsr.w	sub_11BA4
		move.w	(sp)+,d0
		movem.w	d1,-(sp)
		clr.w	d1
		move.b	3(a4),d1
		cmp.w	($FFFFBF48).w,d1
		movem.w	(sp)+,d1
		blt.w	loc_189A6
		jsr	sub_FEAFA

loc_189A6:							; CODE XREF: ROM:0001899Cj
		clr.w	d0
		move.b	4(a4),d0
		bmi.w	loc_18A08
		bclr	#5,($FFFFC2F4).w
		bne.w	loc_18A08
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$12
		dc.b	$BF
		dc.b	$E
		dc.b	6
aAssistBy:	dc.b	'Assist by:'
		dc.b	$BF
		dc.b	$C
		dc.b	7
; ---------------------------------------------------------------------------
		move.w	d0,-(sp)
		bsr.w	sub_18B26
		bsr.w	sub_11BA4
		move.w	(sp)+,d0
		jsr	sub_FEAE4
		clr.w	d0
		move.b	5(a4),d0
		bmi.w	loc_18A08
		bsr.w	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$C
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	d0,-(sp)
		bsr.w	sub_18B26
		bsr.w	sub_11BA4
		move.w	(sp)+,d0
		jsr	sub_FEAE4

loc_18A08:							; CODE XREF: ROM:000189ACj
					; ROM:000189B6j ...
		bclr	#5,($FFFFC2F4).w
		movem.l	(sp)+,d0-d2/a0-a4
		rts
; ---------------------------------------------------------------------------
word_18A14:	dc.w	$E					; DATA XREF: ROM:loc_18916o
		dc.b	$BF
		dc.b	$F
		dc.b	3
aGoal:		dc.b	'GOAL!'
		dc.b	$BF
		dc.b	$C
		dc.b	5
		dc.b	0
		dc.w	$12
		dc.b	$BF
		dc.b	$C
		dc.b	3
aHatTrick:	dc.b	'HAT TRICK!'
		dc.b	$BF
		dc.b	$C
		dc.b	5
word_18A34:	dc.w	$10					; DATA XREF: ROM:00018924o
		dc.b	$BF
		dc.b	$D
		dc.b	3
aPpGoal:	dc.b	'PP GOAL!'
		dc.b	$BF
		dc.b	$C
		dc.b	5
		dc.w	$12
		dc.b	$BF
		dc.b	$C
		dc.b	3
aHatTrick_0:	dc.b	'HAT TRICK!'
		dc.b	$BF
		dc.b	$C
		dc.b	5

; =============== S U B	R O U T	I N E =======================================


sub_18A56:							; CODE XREF: sub_B92E:loc_B946p
		bsr.w	sub_11B92
; End of function sub_18A56

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$B
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$13,d0
		moveq	#8,d1
		move.l	#$7FF,d2
		bra.w	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_18A6E:							; CODE XREF: ROM:00012628p
		movem.l	d0/a2,-(sp)
		move.w	#$C6CE,a2
		move.w	($FFFFC470).w,d0
		bpl.w	loc_18A86
		and.w	#$FF,d0
		add.w	#$364,a2

loc_18A86:							; CODE XREF: sub_18A6E+Cj
		bsr.w	sub_18A90
		movem.l	(sp)+,d0/a2
		rts
; End of function sub_18A6E


; =============== S U B	R O U T	I N E =======================================


sub_18A90:							; CODE XREF: ROM:000085C8p sub_8D4E+8p ...
		movem.l	d0-d3/a0/a2-a3,-(sp)
		bsr.w	sub_18BC8
		move.l	a0,-(sp)
		move.w	#$BFA4,a3
		move.w	#4,(a3)
		lea	2(a3),a1
		add.w	(a0),a0
		move.b	(a0),d0
		bsr.w	sub_18BDC
		bsr.w	sub_11D96
; End of function sub_18A90

; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	$20
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	(sp)+,a1
		bsr.w	sub_11D9E
		move.w	#$BFA4,a1
		movem.l	(sp)+,d0-d3/a0/a2-a3
		rts

; =============== S U B	R O U T	I N E =======================================


sub_18AC6:							; CODE XREF: ROM:loc_1878Ep
		movem.l	d0/a2,-(sp)
		move.w	#$C6CE,a2
		move.w	($FFFFC470).w,d0
		bpl.w	loc_18ADE
		and.w	#$FF,d0
		add.w	#$364,a2

loc_18ADE:							; CODE XREF: sub_18AC6+Cj
		bsr.w	sub_18AE8
		movem.l	(sp)+,d0/a2
		rts
; End of function sub_18AC6


; =============== S U B	R O U T	I N E =======================================


sub_18AE8:							; CODE XREF: sub_9112+6p sub_9316+88p	...
		movem.l	d0-d3/a0/a2,-(sp)
		bsr.w	sub_18BC8
		move.l	a0,-(sp)
		move.w	#$BFA6,a1
		add.w	(a0),a0
		move.b	(a0),d0
		bsr.w	sub_18BDC
		move.b	#$20,(a1)+				; ' '
		move.l	(sp)+,a0
		move.w	(a0)+,d0
		lea	-2(a0,d0.w),a2
		move.b	(a0)+,(a1)+
		move.w	#$2E20,(a1)+

loc_18B10:							; CODE XREF: sub_18AE8+2Cj
		cmp.b	#$20,(a0)+				; ' '
		bne.s	loc_18B10

loc_18B16:							; CODE XREF: sub_18AE8+32j
		move.b	(a0)+,(a1)+
		cmp.l	a0,a2
		bne.s	loc_18B16
		bsr.w	sub_18BAE
		movem.l	(sp)+,d0-d3/a0/a2
		rts
; End of function sub_18AE8


; =============== S U B	R O U T	I N E =======================================


sub_18B26:							; CODE XREF: sub_9744+58p
					; ROM:00018980p ...
		movem.l	d0-d3/a0/a2,-(sp)
		bsr.w	sub_18BC8
		move.l	a0,-(sp)
		move.w	#$BFA6,a1
		add.w	(a0),a0
		move.b	(a0),d0
		bsr.w	sub_18BDC
		move.b	#$20,(a1)+				; ' '
		move.l	(sp)+,a0
		move.w	(a0)+,d0
		lea	-2(a0,d0.w),a2

loc_18B48:							; CODE XREF: sub_18B26+26j
		cmp.b	#$20,(a0)+				; ' '
		bne.s	loc_18B48

loc_18B4E:							; CODE XREF: sub_18B26+2Cj
		move.b	(a0)+,(a1)+
		cmp.l	a0,a2
		bne.s	loc_18B4E
		bsr.w	sub_18BAE
		movem.l	(sp)+,d0-d3/a0/a2
		rts
; End of function sub_18B26


; =============== S U B	R O U T	I N E =======================================


sub_18B5E:							; CODE XREF: ROM:loc_FA3F0p
		movem.l	d0-d3/a0/a2,-(sp)
		bsr.w	sub_18BC8
		move.w	#$BFA6,a1
		bra.w	loc_18B7E
; End of function sub_18B5E


; =============== S U B	R O U T	I N E =======================================


sub_18B6E:							; CODE XREF: ROM:00008660p
					; ROM:000FA9B0p ...
		movem.l	d0-d3/a0/a2,-(sp)
		bsr.w	sub_18BC8
		move.w	#$BFA6,a1
		move.b	#$20,(a1)+				; ' '

loc_18B7E:							; CODE XREF: sub_18B5E+Cj
		move.w	(a0)+,d0
		lea	-2(a0,d0.w),a2

loc_18B84:							; CODE XREF: sub_18B6E+1Aj
		cmp.b	#$20,(a0)+				; ' '
		bne.s	loc_18B84

loc_18B8A:							; CODE XREF: sub_18B6E+26j
		move.b	(a0)+,(a1)+
		cmp.l	a0,a2
		beq.w	loc_18B9E
		tst.b	(a0)
		bne.s	loc_18B8A
		bra.w	loc_18B9E
; ---------------------------------------------------------------------------

loc_18B9A:							; CODE XREF: sub_18B6E+34j
		move.b	#$20,(a1)+				; ' '

loc_18B9E:							; CODE XREF: sub_18B6E+20j
					; sub_18B6E+28j
		cmp.w	#$BFB2,a1
		blt.s	loc_18B9A
		bsr.w	sub_18BAE
		movem.l	(sp)+,d0-d3/a0/a2
		rts
; End of function sub_18B6E


; =============== S U B	R O U T	I N E =======================================


sub_18BAE:							; CODE XREF: sub_18AE8+34p
					; sub_18B26+2Ep ...
		move.w	a1,d0
		sub.w	#$BFA4,d0
		btst	#0,d0
		beq.w	loc_18BC0
		clr.b	(a1)+
		addq.w	#1,d0

loc_18BC0:							; CODE XREF: sub_18BAE+Aj
		move.w	#$BFA4,a1
		move.w	d0,(a1)
		rts
; End of function sub_18BAE


; =============== S U B	R O U T	I N E =======================================


sub_18BC8:							; CODE XREF: sub_18A90+4p sub_18AE8+4p ...
		move.l	$1E(a2),a0
		add.w	(a0),a0
		bra.w	loc_18BD6
; ---------------------------------------------------------------------------

loc_18BD2:							; CODE XREF: sub_18BC8:loc_18BD6j
		add.w	(a0),a0
		addq.w	#8,a0

loc_18BD6:							; CODE XREF: sub_18BC8+6j
		dbf	d0,loc_18BD2
		rts
; End of function sub_18BC8


; =============== S U B	R O U T	I N E =======================================


sub_18BDC:							; CODE XREF: sub_12E12+30p
					; sub_18A90+1Ap ...
		move.w	d0,-(sp)
		lsr.b	#4,d0
		bne.w	loc_18BE8
		move.b	#$F0,d0

loc_18BE8:							; CODE XREF: sub_18BDC+4j
		add.b	#$30,d0					; '0'
		move.b	d0,(a1)+
		move.w	(sp)+,d0
		and.w	#$F,d0
		add.b	#$30,d0					; '0'
		move.b	d0,(a1)+
		rts
; End of function sub_18BDC

; ===========================================================================
; ---------------------------------------------------------------------------
; Error Routine - Address Error (Also used for BUS Error)
; ---------------------------------------------------------------------------

ErAddress:
		move	#$2700,sr
		bsr.w	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$12
		dc.b	$BD
		dc.b	0
		dc.b	0
aAddressError:	dc.b	'Address Error'
; ---------------------------------------------------------------------------
		move.l	$A(sp),d0
		move.l	2(sp),d1
		bra.w	loc_18C70
; ---------------------------------------------------------------------------

ErIllegal:							; DATA XREF: ROM:00000010o
		move	#$2700,sr
		bsr.w	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	0
		dc.b	0
aIllegalInstruction: dc.b 'Illegal Instruction'
; ---------------------------------------------------------------------------
		move.l	2(sp),d0
		move.l	d0,d1
		bra.w	loc_18C70
; ---------------------------------------------------------------------------

ErrDivide:							; DATA XREF: ROM:00000014o
		move	#$2700,sr
		bsr.w	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$BD
		dc.b	0
		dc.b	0
aDivisionByZero: dc.b	'Division by zero',0
; ---------------------------------------------------------------------------
		move.l	2(sp),d0
		move.l	d0,d1

loc_18C70:							; CODE XREF: ROM:00018C1Ej
					; ROM:00018C48j
		move.w	#$BFA4,a0
		move.w	#$18,(a0)+
		move.b	#$BD,(a0)+
		move.b	#0,(a0)+
		move.b	#2,(a0)+
		move.l	d1,-(sp)
		bsr.w	sub_18CDC
		move.b	#$BD,(a0)+
		move.b	#0,(a0)+
		move.b	#4,(a0)+
		move.l	(sp)+,d0
		bsr.w	sub_18CDC
		move.w	#$BFA4,a1
		bsr.w	sub_11DF4
		move.l	#$C00000,a0

loc_18CAA:							; set window X position
		move.w	#$9100,4(a0)
		move.w	#$9206,4(a0)				; set window Y position
		move.w	#$8F02,4(a0)				; set increment	mode
		move.l	#$C0200000,4(a0)			; set CRAM write address
		move.l	#off_5605A,a1
		add.l	(a1),a1

loc_18CCC:
		add.w	#$20,a1					; ' '

loc_18CD0:
		moveq	#7,d0

loc_18CD2:							; CODE XREF: ROM:00018CD4j
		move.l	(a1)+,(a0)
		dbf	d0,loc_18CD2

loc_18CD8:							; CODE XREF: ROM:loc_18CD8j
		bra.w	loc_18CD8

; =============== S U B	R O U T	I N E =======================================


sub_18CDC:							; CODE XREF: ROM:00018C86p
					; ROM:00018C98p
		moveq	#7,d2					; Subroutine to	convert	a hex number to	ASCII

loc_18CDE:							; CODE XREF: sub_18CDC+1Aj
		rol.l	#4,d0					; get digit (nybble)
		move.w	d0,d1
		and.w	#$F,d1
		add.w	#$30,d1					; '0'   ; advance to ASCII character '0'
		cmpnw	 $39,d1					; '9'   ; is it higher than '9'?
		ble.w	loc_18CF4				; if not, branch
		addq.w	#7,d1					; advance to letters 'A' to 'F'

loc_18CF4:							; CODE XREF: sub_18CDC+12j
		move.b	d1,(a0)+				; save ASCII character
		dbf	d2,loc_18CDE				; repeat for all 8 digits
		rts
; End of function sub_18CDC

; ---------------------------------------------------------------------------
unk_18CFC:	dc.b	1					; DATA XREF: sub_7A9E+44o
		dc.b	$C
		dc.b	2
		dc.b	$C
		dc.b	3
		dc.b	$C
		dc.b	4
		dc.b	$C
		dc.b	5
		dc.b	$C
		dc.b	$A
		dc.b	$C
		dc.b	9
		dc.b	$C
		dc.b	8
		dc.b	$C
		dc.b	6
		dc.b	$C
		dc.b	7
		dc.b	$C
		dc.b	8
		dc.b	$C
		dc.b	9
		dc.b	$C
		dc.b	$A
		dc.b	$C
		dc.b	5
		dc.b	$C
		dc.b	4
		dc.b	$C
		dc.b	3
		dc.b	$C
		dc.b	1
		dc.b	7
		dc.b	2
		dc.b	7
		dc.b	3
		dc.b	7
		dc.b	4
		dc.b	7
		dc.b	5
		dc.b	7
		dc.b	$B
		dc.b	7
		dc.b	$C
		dc.b	7
		dc.b	$D
		dc.b	7
		dc.b	6
		dc.b	7
		dc.b	7
		dc.b	7
		dc.b	8
		dc.b	7
		dc.b	9
		dc.b	7
		dc.b	$A
		dc.b	7
		dc.b	$B
		dc.b	7
		dc.b	$C
		dc.b	7
		dc.b	$D
		dc.b	7
		dc.b	$E
		dc.b	7
		dc.b	$F
		dc.b	7
		dc.b	$10
		dc.b	7
		dc.b	$11
		dc.b	7
		dc.b	$12
		dc.b	7
		dc.b	$B
		dc.b	7
		dc.b	$C
		dc.b	7
		dc.b	$D
		dc.b	7
		dc.b	$12
		dc.b	7
		dc.b	$11
		dc.b	7
		dc.b	$10
		dc.b	7
		dc.b	$F
		dc.b	7
		dc.b	$E
		dc.b	7
		dc.b	$D
		dc.b	7
		dc.b	$C
		dc.b	7
		dc.b	$B
		dc.b	7
		dc.b	$E
		dc.b	3
		dc.b	$F
		dc.b	3
		dc.b	$10
		dc.b	3
		dc.b	$11
		dc.b	3
		dc.b	$12
		dc.b	3
		dc.b	$B
		dc.b	3
		dc.b	$C
		dc.b	3
		dc.b	$D
		dc.b	3
		dc.b	$12
		dc.b	3
		dc.b	$11
		dc.b	3
		dc.b	$10
		dc.b	3
		dc.b	$F
		dc.b	3
		dc.b	$E
		dc.b	3
		dc.b	$D
		dc.b	3
		dc.b	$C
		dc.b	3
		dc.b	$B
		dc.b	3
ObjListMaybe:	dc.l	locret_15464				; DATA XREF: sub_A9D6+4B2o
		dc.l	loc_CD00
		dc.l	loc_CDE0
		dc.l	loc_CF9A
		dc.l	loc_D09C
		dc.l	loc_D1E8
		dc.l	loc_D294
		dc.l	loc_CC4E
		dc.l	loc_CBFA
		dc.l	sub_C882
		dc.l	sub_CA50
		dc.l	sub_C710
		dc.l	sub_C8F0
		dc.l	sub_CA2A
		dc.l	sub_D51C
		dc.l	sub_DD2E
		dc.l	$DF18
		dc.l	$E6F4
		dc.l	loc_EC82
		dc.l	sub_EB84
		dc.l	nullsub_4
		dc.l	nullsub_5
		dc.l	sub_CB3A
		dc.l	sub_CB50
		dc.l	sub_FF0C
		dc.l	sub_102EC
		dc.l	sub_102A8
		dc.l	sub_F3E4
		dc.l	loc_F88E
		dc.l	loc_D3C6
		dc.l	loc_ECB6
		dc.l	sub_EF92
		dc.l	sub_CAE0
		dc.l	sub_DEEE
		dc.l	sub_E6CE
		dc.l	sub_F693E
word_18E0C:	dc.w	0					; DATA XREF: sub_9316+5Ao
					; sub_11F62+4Eo ...
		dc.w	$3E
		dc.w	$50
		dc.w	$94
		dc.w	$94
		dc.w	$C2
		dc.w	$72
		dc.w	$62
		dc.w	$82
		dc.w	$190
		dc.w	$6A
		dc.w	$C8
		dc.w	$10C
		dc.w	$1B0
		dc.w	$234
		dc.w	$1F0
		dc.w	$150
		dc.w	$280
		dc.w	$2B0
		dc.w	$2C4
		dc.w	$2E4
		dc.w	$306
		dc.w	$B0
		dc.w	$212
		dc.w	$170
		dc.w	$A2
		dc.w	$298
		dc.w	$1D0
		dc.w	$EA
		dc.w	$25A
		dc.w	$12E
		dc.w	$100
		dc.w	$E
aPeriodOver:	dc.b	'Period Over',0
		dc.b	$FA					; 
		dc.b	$F0					; 
		dc.b	1
		dc.b	0
		dc.w	$C
aGameOver:	dc.b	'Game Over',0
		dc.b	5
		dc.b	$10
		dc.b	$BF					; 
		dc.b	$A0					; 
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	5
		dc.b	$A
		dc.b	$BF					; 
		dc.b	$E8					; 
		dc.b	1
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	5
		dc.b	$10
		dc.b	$BF					; 
		dc.b	$E8					; 
		dc.b	1
		dc.b	0
		dc.w	8
aIcing:		dc.b	'Icing',0
		dc.b	0
		dc.b	2
		dc.b	1
		dc.b	1
		dc.b	$FD					; 
		dc.b	$FA					; 
		dc.b	1
		dc.b	0
		dc.w	$A
aOffSide:	dc.b	'Off-side'
		dc.b	0
		dc.b	2
		dc.b	3
		dc.b	1
		dc.b	$FB					; 
		dc.b	$F1					; 
		dc.b	1
		dc.b	0
		dc.w	$A
aFaceOff:	dc.b	'Face Off'
		dc.b	$FA					; 
		dc.b	$F6					; 
		dc.b	0
		dc.b	0
		dc.w	$A
aFaceOff_0:	dc.b	'Face Off'
		dc.b	$FA					; 
		dc.b	$F6					; 
		dc.b	0
		dc.b	0
		dc.w	$A
aPenalty:	dc.b	'Penalty',0
		dc.b	0
		dc.b	3
		dc.b	3
		dc.b	1
		dc.b	$FB					; 
		dc.b	$F1					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	2
		dc.b	$FA					; 
		dc.b	$F6					; 
		dc.b	4
		dc.b	2
		dc.w	$A
aCharging:	dc.b	'Charging'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$A
aCharging_0:	dc.b	'Charging'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$A
aSlashing:	dc.b	'Slashing'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$A
aSlashing_0:	dc.b	'Slashing'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$A
aTripping:	dc.b	'Tripping'
		dc.b	0
		dc.b	4
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$A
aTripping_0:	dc.b	'Tripping'
		dc.b	0
		dc.b	4
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	5
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$A
		dc.b	2
		dc.w	$A
aRoughing:	dc.b	'Roughing'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$A
aRoughing_0:	dc.b	'Roughing'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$A
aRoughing_1:	dc.b	'Roughing'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$A
aHooking:	dc.b	'Hooking',0
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$A
aHooking_0:	dc.b	'Hooking',0
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	9
		dc.b	1
		dc.b	8
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$E
aCrossCheck:	dc.b	'Cross Check',0
		dc.b	0
		dc.b	4
		dc.b	6
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$E
aCrossCheck_0:	dc.b	'Cross Check',0
		dc.b	0
		dc.b	4
		dc.b	6
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$F
		dc.b	1
		dc.b	$10
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$E
aInterference:	dc.b	'Interference'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	$C
		dc.b	6
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	$FF
		dc.w	$E
aInterference_0: dc.b	'Interference'
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	$C
		dc.b	6
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	4
		dc.b	2
		dc.w	$A
aHolding:	dc.b	'Holding',0
		dc.b	0
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	7
		dc.b	6
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$28					; (
		dc.b	5
		dc.w	$A
aFighting:	dc.b	'Fighting'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$28					; (
		dc.b	5
		dc.w	$C
aFighting_0:	dc.b	'Fighting *'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
		dc.b	$28					; (
		dc.b	2
		dc.w	$12
aFightInstigator: dc.b	'Fight Instigator'
		dc.b	0
		dc.b	4
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$D
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FF
unk_1913A:	dc.b	$FF					; DATA XREF: sub_1211A:loc_1213Ao
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	$38					; 8
		dc.b	0
		dc.b	$3C					; <
		dc.b	0
		dc.b	$36					; 6
		dc.b	0
		dc.b	$3A					; :
		dc.b	0
		dc.b	$2E					; .
		dc.b	0
		dc.b	$30					; 0
		dc.b	0
		dc.b	$34					; 4
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
unk_1916A:	dc.b	$B4					; 		; DATA XREF: sub_11E8E+4o
		dc.b	$B7					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$B9					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$B8					; 
		dc.b	0
		dc.b	$33					; 3
		dc.b	$CB					; 
		dc.b	$36					; 6
		dc.b	$38					; 8
		dc.b	$3A					; :
		dc.b	$3C					; <
		dc.b	$3E					; >
		dc.b	$40					; @
		dc.b	$42					; B
		dc.b	$44					; D
		dc.b	$B9					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$4A					; J
		dc.b	$4A					; J
		dc.b	0
		dc.b	2
		dc.b	4
		dc.b	6
		dc.b	8
		dc.b	$A
		dc.b	$C
		dc.b	$E
		dc.b	$F0					; 
		dc.b	$11
		dc.b	$13
		dc.b	$15
		dc.b	$17
		dc.b	$19
		dc.b	$1B
		dc.b	$1D
		dc.b	$1F
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$27					; '
		dc.b	$29					; )
		dc.b	$2B					; +
		dc.b	$2D					; -
		dc.b	$2F					; /
		dc.b	$31					; 1
		dc.b	$FF
word_191A6:	dc.w	6					; DATA XREF: ROM:0000860Eo
					; sub_B92E+66o	...
aSc1:		dc.b	'Sc1',0
		dc.w	6
aSc2:		dc.b	'Sc2',0
		dc.w	6
aChk:		dc.b	'Chk',0
		dc.w	6
aPp1:		dc.b	'PP1',0
		dc.w	6
aPp2_0:		dc.b	'PP2',0
		dc.w	6
aPk1:		dc.b	'PK1',0
		dc.w	6
aPk2:		dc.b	'PK2',0
word_191D0:	dc.w	4					; DATA XREF: ROM:00008C70o
aLd:		dc.b	'LD'
		dc.w	4
aRd:		dc.b	'RD'
		dc.w	4
aLw:		dc.b	'LW'
		dc.w	4
		dc.b	'C',0
		dc.w	4
aRw:		dc.b	'RW'
word_191E4:	dc.w	4					; DATA XREF: sub_81F2+46o
					; ROM:00012C48o ...
		dc.b	'1'
		dc.b	$12
		dc.w	4
		dc.b	'2'
		dc.b	$13
		dc.w	4
		dc.b	'3'
		dc.b	$14
		dc.w	4
aOt:		dc.b	'OT'
		dc.w	4
aF:		dc.b	' F'
word_191F8:	dc.w	4					; DATA XREF: ROM:00012C5Co
					; ROM:loc_12CCCo
		dc.b	'1'
		dc.b	$12
		dc.w	4
		dc.b	'2'
		dc.b	$13
		dc.w	4
		dc.b	'3'
		dc.b	$14
		dc.w	4
aOt_0:		dc.b	'OT'
		dc.w	4
		dc.b	'  '
byte_1920C:	dc.b	1					; DATA XREF: sub_167AA+76o
		dc.b	2
		dc.b	3
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	6
		dc.b	8
		dc.b	3
		dc.b	6
		dc.b	9
		dc.b	$C
		dc.b	4
		dc.b	8
		dc.b	$C
		dc.b	$10
word_1921C:	dc.w	$E					; DATA XREF: sub_15832+88o
		dc.w	$E
		dc.w	$1C
		dc.w	$2A
		dc.w	$54
		dc.w	$3F
		dc.w	$54
		dc.w	$109
		dc.b	$11
		dc.b	$19
		dc.b	$21					; !
		dc.b	$29					; )
		dc.b	$31					; 1
		dc.b	2
		dc.b	$A
		dc.b	$12
		dc.b	$1A
		dc.b	$22					; "
		dc.b	$2A					; *
		dc.b	$32					; 2
		dc.b	2
		dc.b	$A
		dc.b	$12
		dc.b	$1A
		dc.b	$22					; "
		dc.b	$2A					; *
		dc.b	$32					; 2
		dc.b	1
		dc.b	9
		dc.b	$11
		dc.b	$19
		dc.b	$21					; !
		dc.b	$29					; )
		dc.b	$31					; 1
		dc.b	3
		dc.b	$B
		dc.b	$13
		dc.b	$1B
		dc.b	$23					; #
		dc.b	$2B					; +
		dc.b	$33					; 3
		dc.b	5
		dc.b	$D
		dc.b	$15
		dc.b	$1D
		dc.b	$25					; %
		dc.b	$2D					; -
		dc.b	$35					; 5
		dc.b	4
		dc.b	$C
		dc.b	$14
		dc.b	$1C
		dc.b	$24					; $
		dc.b	$2C					; ,
		dc.b	$34					; 4
		dc.b	5
		dc.b	$D
		dc.b	$15
		dc.b	$1D
		dc.b	$25					; %
		dc.b	$2D					; -
		dc.b	$35					; 5
		dc.b	3
		dc.b	$B
		dc.b	$13
		dc.b	$1B
		dc.b	$23					; #
		dc.b	$2B					; +
		dc.b	$33					; 3
		dc.b	4
		dc.b	$C
		dc.b	$14
		dc.b	$1C
		dc.b	$24					; $
		dc.b	$2C					; ,
		dc.b	$34					; 4
		dc.b	4
		dc.b	$C
		dc.b	$14
		dc.b	$1C
		dc.b	$24					; $
		dc.b	$2C					; ,
		dc.b	$34					; 4
		dc.b	3
		dc.b	$B
		dc.b	$13
		dc.b	$1B
		dc.b	$23					; #
		dc.b	$2B					; +
		dc.b	$33					; 3
		dc.b	5
		dc.b	$D
		dc.b	$15
		dc.b	$1D
		dc.b	$25					; %
		dc.b	$2D					; -
		dc.b	$35					; 5
		dc.b	$FF
off_19286:	dc.l	loc_10202+2				; DATA XREF: sub_129D2+2Eo
					; sub_12EB4+10o ...
		dc.b	3
		dc.b	5
		dc.b	6
		dc.b	$FF
off_1928E:	dc.l	unk_A0041				; DATA XREF: ROM:0001743Ao
		dc.b	0
		dc.b	$84					; 
		dc.b	0
		dc.b	$CF					; 
		dc.b	1
		dc.b	$1C
		dc.b	$F
		dc.b	$2E					; .
		dc.b	2
		dc.b	$2E					; .
		dc.b	5
		dc.b	$2E					; .
		dc.b	8
		dc.b	$2E					; .
		dc.b	$B
		dc.b	$2E					; .
		dc.b	$E
		dc.b	$2E					; .
		dc.b	$11
		dc.b	$2E					; .
		dc.b	$14
		dc.b	$2E					; .
		dc.b	$17
		dc.b	$47					; G
		dc.b	2
		dc.b	$47					; G
		dc.b	5
		dc.b	$47					; G
		dc.b	8
		dc.b	$47					; G
		dc.b	$B
		dc.b	$47					; G
		dc.b	$E
		dc.b	$47					; G
		dc.b	$11
		dc.b	$47					; G
		dc.b	$14
		dc.b	$47					; G
		dc.b	$17
		dc.b	1
		dc.b	$39					; 9
		dc.b	0
		dc.b	$45					; E
		dc.b	$A
		dc.b	7
		dc.b	$32					; 2
		dc.b	4
		dc.b	$32					; 2
		dc.b	$A
		dc.b	$32					; 2
		dc.b	$10
		dc.b	$32					; 2
		dc.b	$16
		dc.b	$4B					; K
		dc.b	4
		dc.b	$4B					; K
		dc.b	$A
		dc.b	$4B					; K
		dc.b	$10
		dc.b	$4B					; K
		dc.b	$16
		dc.b	$17
		dc.b	$20
		dc.b	2
		dc.b	$20
		dc.b	5
		dc.b	$20
		dc.b	8
		dc.b	$20
		dc.b	$B
		dc.b	$20
		dc.b	$E
		dc.b	$20
		dc.b	$11
		dc.b	$20
		dc.b	$14
		dc.b	$20
		dc.b	$17
		dc.b	$55					; U
		dc.b	2
		dc.b	$55					; U
		dc.b	5
		dc.b	$55					; U
		dc.b	8
		dc.b	$55					; U
		dc.b	$B
		dc.b	$55					; U
		dc.b	$E
		dc.b	$55					; U
		dc.b	$11
		dc.b	$55					; U
		dc.b	$14
		dc.b	$55					; U
		dc.b	$17
		dc.b	$2E					; .
		dc.b	4
		dc.b	$2E					; .
		dc.b	$A
		dc.b	$2E					; .
		dc.b	$F
		dc.b	$2E					; .
		dc.b	$15
		dc.b	$47					; G
		dc.b	4
		dc.b	$47					; G
		dc.b	$A
		dc.b	$47					; G
		dc.b	$F
		dc.b	$47					; G
		dc.b	$15
		dc.b	3
		dc.b	$2B					; +
		dc.b	0
		dc.b	$39					; 9
		dc.b	2
		dc.b	$45					; E
		dc.b	8
		dc.b	$53					; S
		dc.b	$A
		dc.b	3
		dc.b	$32					; 2
		dc.b	8
		dc.b	$32					; 2
		dc.b	$12
		dc.b	$4B					; K
		dc.b	8
		dc.b	$4B					; K
		dc.b	$12
		dc.b	$1B
		dc.b	$12
		dc.b	2
		dc.b	$12
		dc.b	5
		dc.b	$12
		dc.b	8
		dc.b	$12
		dc.b	$B
		dc.b	$12
		dc.b	$E
		dc.b	$12
		dc.b	$11
		dc.b	$12
		dc.b	$14
		dc.b	$12
		dc.b	$17
		dc.b	$63					; c
		dc.b	2
		dc.b	$63					; c
		dc.b	5
		dc.b	$63					; c
		dc.b	8
		dc.b	$63					; c
		dc.b	$B
		dc.b	$63					; c
		dc.b	$E
		dc.b	$63					; c
		dc.b	$11
		dc.b	$63					; c
		dc.b	$14
		dc.b	$63					; c
		dc.b	$17
		dc.b	$20
		dc.b	4
		dc.b	$20
		dc.b	$A
		dc.b	$20
		dc.b	$F
		dc.b	$20
		dc.b	$15
		dc.b	$55					; U
		dc.b	4
		dc.b	$55					; U
		dc.b	$A
		dc.b	$55					; U
		dc.b	$F
		dc.b	$55					; U
		dc.b	$15
		dc.b	$2E					; .
		dc.b	7
		dc.b	$2E					; .
		dc.b	$12
		dc.b	$47					; G
		dc.b	7
		dc.b	$47					; G
		dc.b	$12
		dc.b	5
		dc.b	$1D
		dc.b	0
		dc.b	$2B					; +
		dc.b	2
		dc.b	$39					; 9
		dc.b	4
		dc.b	$45					; E
		dc.b	6
		dc.b	$53					; S
		dc.b	8
		dc.b	$61					; a
		dc.b	$A
		dc.b	1
		dc.b	$32					; 2
		dc.b	$D
		dc.b	$4B					; K
		dc.b	$D
		dc.b	$1D
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	5
		dc.b	4
		dc.b	8
		dc.b	4
		dc.b	$B
		dc.b	4
		dc.b	$E
		dc.b	4
		dc.b	$11
		dc.b	4
		dc.b	$14
		dc.b	4
		dc.b	$17
		dc.b	$71					; q
		dc.b	2
		dc.b	$71					; q
		dc.b	5
		dc.b	$71					; q
		dc.b	8
		dc.b	$71					; q
		dc.b	$B
		dc.b	$71					; q
		dc.b	$E
		dc.b	$71					; q
		dc.b	$11
		dc.b	$71					; q
		dc.b	$14
		dc.b	$71					; q
		dc.b	$17
		dc.b	$12
		dc.b	4
		dc.b	$12
		dc.b	$A
		dc.b	$12
		dc.b	$F
		dc.b	$12
		dc.b	$15
		dc.b	$63					; c
		dc.b	4
		dc.b	$63					; c
		dc.b	$A
		dc.b	$63					; c
		dc.b	$F
		dc.b	$63					; c
		dc.b	$15
		dc.b	$20
		dc.b	7
		dc.b	$20
		dc.b	$12
		dc.b	$55					; U
		dc.b	7
		dc.b	$55					; U
		dc.b	$12
		dc.b	$2E					; .
		dc.b	$C
		dc.b	$47					; G
		dc.b	$C
		dc.b	5
		dc.b	$F
		dc.b	0
		dc.b	$1D
		dc.b	2
		dc.b	$2B					; +
		dc.b	4
		dc.b	$53					; S
		dc.b	6
		dc.b	$61					; a
		dc.b	8
		dc.b	$6F					; o
		dc.b	$A
		dc.b	0
		dc.b	$3F					; ?
		dc.b	$17
		dc.b	$1E
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	5
		dc.b	4
		dc.b	8
		dc.b	4
		dc.b	$B
		dc.b	4
		dc.b	$E
		dc.b	4
		dc.b	$11
		dc.b	4
		dc.b	$14
		dc.b	4
		dc.b	$17
		dc.b	$71					; q
		dc.b	2
		dc.b	$71					; q
		dc.b	5
		dc.b	$71					; q
		dc.b	8
		dc.b	$71					; q
		dc.b	$B
		dc.b	$71					; q
		dc.b	$E
		dc.b	$71					; q
		dc.b	$11
		dc.b	$71					; q
		dc.b	$14
		dc.b	$71					; q
		dc.b	$17
		dc.b	$12
		dc.b	4
		dc.b	$12
		dc.b	$A
		dc.b	$12
		dc.b	$F
		dc.b	$12
		dc.b	$15
		dc.b	$63					; c
		dc.b	4
		dc.b	$63					; c
		dc.b	$A
		dc.b	$63					; c
		dc.b	$F
		dc.b	$63					; c
		dc.b	$15
		dc.b	$20
		dc.b	7
		dc.b	$20
		dc.b	$12
		dc.b	$55					; U
		dc.b	7
		dc.b	$55					; U
		dc.b	$12
		dc.b	$2E					; .
		dc.b	$C
		dc.b	$47					; G
		dc.b	$C
		dc.b	$3A					; :
		dc.b	$17
		dc.b	5
		dc.b	$F
		dc.b	0
		dc.b	$1D
		dc.b	2
		dc.b	$2B					; +
		dc.b	4
		dc.b	$53					; S
		dc.b	6
		dc.b	$61					; a
		dc.b	8
		dc.b	$6F					; o
		dc.b	$A
		dc.b	$FF
		dc.b	$FF
word_193F8:	dc.w	$12					; DATA XREF: ROM:loc_84DAo
					; ROM:loc_8BD6o
aStatus:	dc.b	'     Status    ]'
		dc.w	0
		dc.w	0
word_1940E:	dc.w	$12					; DATA XREF: ROM:loc_FC85Co
					; ROM:loc_FC88Eo
aOverall:	dc.b	'[   Overall    ]'
dword_19420:	dc.l	$1FBA000A				; DATA XREF: ROM:loc_FA8F8r
					; sub_FA9F8+10r ...
		dc.w	$12
aEnergy:	dc.b	'[   Energy     ]'
		dc.w	0
		dc.w	2
		dc.w	$12
aAgility:	dc.b	'[   Agility    ]'
		dc.w	$1000
		dc.w	$A
		dc.w	$12
aSpeed:		dc.b	'[    Speed     ]'
		dc.w	$800
		dc.w	$A
		dc.w	$12
aHanded:	dc.b	'[   Handed     ]'
		dc.w	$40
		dc.w	4
		dc.w	$12
aOff_Awareness:	dc.b	'[Off. Awareness]'
		dc.w	$400
		dc.w	$A
		dc.w	$12
aDef_Awareness:	dc.b	'[Def. Awareness]'
		dc.w	$200
		dc.w	$A
		dc.w	$12
aShotPower:	dc.b	'[  Shot Power  ]'
		dc.w	$100
		dc.w	$A
		dc.w	$12
aShotAccuracy:	dc.b	'[Shot  Accuracy]'
		dc.w	$10
		dc.w	$A
		dc.w	$12
aPassAccuracy:	dc.b	'[Pass  Accuracy]'
		dc.w	2
		dc.w	$A
		dc.w	$12
aStickHandling:	dc.b	'[Stick Handling]'
		dc.w	$20
		dc.w	$A
		dc.w	$12
aWeight:	dc.b	'[    Weight    ]'
		dc.w	$2000
		dc.w	6
		dc.w	$12
aEndurance:	dc.b	'[  Endurance   ]'
		dc.w	8
		dc.w	$A
		dc.w	$12
aAggressiveness: dc.b	'[Aggressiveness]'
		dc.w	1
		dc.w	$A
		dc.w	$12
aChecking_0:	dc.b	'[   Checking    '
		dc.w	$80
		dc.w	$A
		dc.w	$FFFF
word_1955A:	dc.w	$12					; DATA XREF: ROM:00008BE8o
aStatus_0:	dc.b	'     Status    ]'
		dc.w	0
		dc.w	0
word_19570:	dc.w	$12					; DATA XREF: ROM:000FC86Co
					; ROM:loc_FC89Eo
aOverall_0:	dc.b	'[   Overall    ]'
dword_19582:	dc.l	$130F000A				; DATA XREF: ROM:000FA906r
					; sub_FA9F8:loc_FAA18r	...
		dc.w	$12
aAgility_0:	dc.b	'[   Agility    ]'
		dc.w	$1000
		dc.w	$A
		dc.w	$12
aSpeed_0:	dc.b	'[    Speed     ]'
		dc.w	$800
		dc.w	$A
		dc.w	$12
aGloveHand:	dc.b	'[  Glove Hand  ]'
		dc.w	$40
		dc.w	4
		dc.w	$12
aDef_Awareness_0: dc.b	'[Def. Awareness]'
		dc.w	$200
		dc.w	$A
		dc.w	$12
aPuckControl:	dc.b	'[ Puck Control ]'
		dc.w	$100
		dc.w	$A
		dc.w	$12
aStickRight:	dc.b	'[ Stick  Right ]'
		dc.l	off_8000A
		dc.w	$12
aStickLeft:	dc.b	'[  Stick Left  ]'
		dc.w	4
		dc.w	$A
		dc.w	$12
aGloveRight:	dc.b	'[ Glove  Right ]'
		dc.w	2
		dc.w	$A
		dc.w	$12
aGloveLeft:	dc.b	'[  Glove Left  ]'
		dc.w	1
		dc.w	$A
		dc.w	$12
aWeight_0:	dc.b	'[    Weight     '
		dc.w	$2000
		dc.w	6
		dc.w	$FFFF
word_19664:	dc.w	4					; DATA XREF: sub_7CF8+36o
		dc.b	$FE
		dc.b	5
		dc.w	4
		dc.b	$FE
		dc.b	4
		dc.w	$14
aResumeGame:	dc.b	'   Resume Game    '
		dc.l	$15464
		dc.w	$14
aInstantReplay:	dc.b	'  Instant Replay  '
		dc.l	sub_9FD0
		dc.w	$14
aTeamRoster:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aPlayerCards:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders:	dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aXManualGoalie:	dc.b	'x Manual Goalie   '
		dc.l	sub_FE1D8
		dc.l	unk_4FF00
off_19700:	dc.l	unk_4FE05				; DATA XREF: sub_7CF8+22o
		dc.l	unk_4FE04
		dc.w	$14
aResumeGame_0:	dc.b	'   Resume Game    '
		dc.l	$15464
		dc.w	$14
aInstantReplay_0: dc.b	'  Instant Replay  '
		dc.l	sub_9FD0
		dc.w	$14
aChangeGoalie:	dc.b	'  Change Goalie   '
		dc.l	loc_9DE6
		dc.w	$14
aEditLines:	dc.b	'    Edit Lines    '
		dc.l	loc_82DA
		dc.w	$14
aGameStats:	dc.b	'    Game Stats    '
		dc.l	loc_FDC5A
		dc.w	$14
aPlayerStats_0:	dc.b	'   Player Stats   '
		dc.l	loc_945A
		dc.w	$14
aScoringSummary_0: dc.b	' Scoring Summary  '
		dc.l	loc_8EB0
		dc.w	$14
aPenaltySummary_0: dc.b	' Penalty Summary  '
		dc.l	loc_9142
		dc.w	$14
aTeamRoster_0:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aCrowdMeter_0:	dc.b	'   Crowd Meter    '
		dc.l	loc_9A2A
		dc.w	$14
aTimeout_0:	dc.b	'     Timeout      '
		dc.l	loc_9D7A
		dc.w	$14
aPlayerCards_0:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_0: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aPeriodStats:	dc.b	'   Period Stats   '
		dc.l	loc_FD90C
		dc.w	$14
aXManualGoalie_0: dc.b	'x Manual Goalie   '
		dc.l	sub_FE1D8
		dc.l	unk_4FF00
off_1988C:	dc.l	unk_4FE05				; DATA XREF: sub_7CF8+4Ao
					; ROM:00009D82o
		dc.l	unk_4FE04
		dc.w	$14
aResumeGame_1:	dc.b	'   Resume Game    '
		dc.l	$15464
		dc.w	$14
aInstantReplay_1: dc.b	'  Instant Replay  '
		dc.l	sub_9FD0
		dc.w	$14
aChangeGoalie_0: dc.b	'  Change Goalie   '
		dc.l	loc_9DE6
		dc.w	$14
aEditLines_0:	dc.b	'    Edit Lines    '
		dc.l	loc_82DA
		dc.w	$14
aGameStats_0:	dc.b	'    Game Stats    '
		dc.l	loc_FDC5A
		dc.w	$14
aPlayerStats_1:	dc.b	'   Player Stats   '
		dc.l	loc_945A
		dc.w	$14
aScoringSummary_1: dc.b	' Scoring Summary  '
		dc.l	loc_8EB0
		dc.w	$14
aPenaltySummary_1: dc.b	' Penalty Summary  '
		dc.l	loc_9142
		dc.w	$14
aTeamRoster_1:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores_0:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aCrowdMeter_1:	dc.b	'   Crowd Meter    '
		dc.l	loc_9A2A
		dc.w	$14
aPlayerCards_1:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_1: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aPeriodStats_0:	dc.b	'   Period Stats   '
		dc.l	loc_FD90C
		dc.w	$14
aXManualGoalie_1: dc.b	'x Manual Goalie   '
		dc.l	sub_FE1D8
		dc.l	unk_4FF00
off_19A00:	dc.l	unk_4FE05				; DATA XREF: sub_130E6+60o
		dc.l	unk_4FE04
		dc.w	$14
aStartShootout:	dc.b	'  Start Shootout  '
		dc.l	$15464
		dc.w	$14
aShootoutSetup:	dc.b	'  Shootout SetUp  '
		dc.l	loc_FC620
		dc.w	$14
aTeamRoster_2:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aPlayerCards_2:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_2: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.l	unk_4FF00
off_19A84:	dc.l	unk_4FE05				; DATA XREF: ROM:off_131CCo
		dc.l	unk_4FE04
		dc.w	$14
aStartGame:	dc.b	'    Start Game    '
		dc.l	$15464
		dc.w	$14
aChangeGoalie_1: dc.b	'  Change Goalie   '
		dc.l	loc_9DE6
		dc.w	$14
aEditLines_1:	dc.b	'    Edit Lines    '
		dc.l	loc_82DA
		dc.w	$14
aTeamRoster_3:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores_1:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aPlayerCards_3:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_3: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.l	unk_4FF00
off_19B38:	dc.l	unk_4FE05				; DATA XREF: ROM:000131E0o
		dc.l	unk_4FE04
		dc.w	$14
aStartGame_0:	dc.b	'    Start Game    '
		dc.l	$15464
		dc.w	$14
aChangeGoalie_2: dc.b	'  Change Goalie   '
		dc.l	loc_9DE6
		dc.w	$14
aEditLines_2:	dc.b	'    Edit Lines    '
		dc.l	loc_82DA
		dc.w	$14
aTeamRoster_4:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aPlayoffStats_0: dc.b	'  Playoff Stats   '
		dc.l	sub_9428
		dc.w	$14
aOtherScores_2:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aPlayerCards_4:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_4: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.l	unk_4FF00
off_19C04:	dc.l	unk_4FE05				; DATA XREF: ROM:000131D0o
					; ROM:000131D4o ...
		dc.l	unk_4FE04
		dc.w	$14
aResumeGame_2:	dc.b	'   Resume Game    '
		dc.l	$15464
		dc.w	$14
aGameStats_1:	dc.b	'    Game Stats    '
		dc.l	loc_FDC5A
		dc.w	$14
aPlayerStats_2:	dc.b	'   Player Stats   '
		dc.l	loc_945A
		dc.w	$14
aScoringSummary_2: dc.b	' Scoring Summary  '
		dc.l	loc_8EB0
		dc.w	$14
aPenaltySummary_2: dc.b	' Penalty Summary  '
		dc.l	loc_9142
		dc.w	$14
aTeamRoster_5:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores_3:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aCrowdMeter_2:	dc.b	'   Crowd Meter    '
		dc.l	loc_9A2A
		dc.w	$14
aChangeGoalie_3: dc.b	'  Change Goalie   '
		dc.l	loc_9DE6
		dc.w	$14
aEditLines_3:	dc.b	'    Edit Lines    '
		dc.l	loc_82DA
		dc.w	$14
aPlayerCards_5:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_5: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aPeriodStats_1:	dc.b	'   Period Stats   '
		dc.l	loc_FD90C
		dc.w	$14
aXManualGoalie_2: dc.b	'x Manual Goalie   '
		dc.l	sub_FE1D8
		dc.l	unk_4FF00
off_19D60:	dc.l	unk_4FE05				; DATA XREF: ROM:000131DCo
		dc.l	unk_4FE04
		dc.w	$14
aExitGame:	dc.b	'    Exit Game     '
		dc.l	$15464
		dc.w	$14
aGameStats_2:	dc.b	'    Game Stats    '
		dc.l	loc_FDC5A
		dc.w	$14
aPlayerStats_3:	dc.b	'   Player Stats   '
		dc.l	loc_945A
		dc.w	$14
aScoringSummary_3: dc.b	' Scoring Summary  '
		dc.l	loc_8EB0
		dc.w	$14
aPenaltySummary_3: dc.b	' Penalty Summary  '
		dc.l	loc_9142
		dc.w	$14
aTeamRoster_6:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores_4:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aCrowdMeter_3:	dc.b	'   Crowd Meter    '
		dc.l	loc_9A2A
		dc.w	$14
aPlayerCards_6:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_6: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aPeriodStats_2:	dc.b	'   Period Stats   '
		dc.l	loc_FD90C
		dc.l	unk_4FF00
off_19E74:	dc.l	unk_4FE05				; DATA XREF: ROM:000131F0o
		dc.l	unk_4FE04
		dc.w	$14
aExitGame_0:	dc.b	'    Exit Game     '
		dc.l	$15464
		dc.w	$14
aGameStats_3:	dc.b	'    Game Stats    '
		dc.l	loc_FDC5A
		dc.w	$14
aPlayerStats_4:	dc.b	'   Player Stats   '
		dc.l	loc_945A
		dc.w	$14
aScoringSummary_4: dc.b	' Scoring Summary  '
		dc.l	loc_8EB0
		dc.w	$14
aPenaltySummary_4: dc.b	' Penalty Summary  '
		dc.l	loc_9142
		dc.w	$14
aTeamRoster_7:	dc.b	'   Team Roster    '
		dc.l	loc_89AC
		dc.w	$14
aOtherScores_5:	dc.b	'   Other Scores   '
		dc.l	loc_80D4
		dc.w	$14
aCrowdMeter_4:	dc.b	'   Crowd Meter    '
		dc.l	loc_9A2A
		dc.w	$14
aPlayerCards_7:	dc.b	'   Player Cards   '
		dc.l	loc_FA07E
		dc.w	$14
aRecordHolders_7: dc.b	'  Record Holders  '
		dc.l	loc_FBC14
		dc.w	$14
aPeriodStats_3:	dc.b	'   Period Stats   '
		dc.l	loc_FD90C
		dc.l	unk_4FF00
word_19F88:	dc.w	6					; DATA XREF: ROM:0000882Eo
		dc.b	$FE
		dc.b	6
		dc.b	$F9
		dc.b	1
		dc.w	6
		dc.b	$FE
		dc.b	4
		dc.b	$F9
		dc.b	1
		dc.w	$14
aExit:		dc.b	'       Exit       '
		dc.l	$15464
		dc.w	$14
aSetOriginalLines: dc.b	'Set Original lines'
		dc.l	$171CE
		dc.w	$14
aSaveTeamLine:	dc.b	'  Save Team Line  '
		dc.l	sub_8928
		dc.w	$14
aLoadTeamLine:	dc.b	'  Load Team Line  '
		dc.l	sub_88C8
		dc.l	unk_4FF00
word_19FF8:	dc.w	6					; DATA XREF: ROM:00008844o
		dc.b	$FE
		dc.b	6
		dc.b	$F9
		dc.b	1
		dc.w	6
		dc.b	$FE
		dc.b	4
		dc.b	$F9
		dc.b	1
		dc.w	$14
aExit_0:	dc.b	'       Exit       '
		dc.l	$15464
		dc.w	$14
aSetOriginalLines_0: dc.b 'Set Original lines'
		dc.l	$171CE
		dc.w	$14
aSaveTeamLine_0: dc.b	'  Save Team Line  '
		dc.l	sub_8928
		dc.l	unk_4FF00

; =============== S U B	R O U T	I N E =======================================


sub_1A050:							; CODE XREF: ROM:00007712p
		move.l	#sub_1A148,($FFFFB03A).l		; set interrupt address?
		move	#$2500,sr
		clr.w	($FFFFD458).w
		jsr	sub_11340
		move.w	d3,d0
		move.l	#$200000,a0
		move.w	#$E,d3
		move.w	#2,d4
		cmpnb	 $E0,d0
		beq.w	loc_1A0C2
		cmpnb	 $B0,d0
		beq.w	loc_1A11E
		moveq	#0,d0
		move.l	#$2000,d1
		move.l	#$FFFF0000,a0
		bsr.w	sub_1A244
		bsr.w	sub_1A14E
		tst.w	($FFFFD458).w
		bpl.w	locret_1A0C0
		bsr.w	sub_1A19C
		moveq	#0,d0
		move.l	#$2000,d1
		move.l	#$FFFF0000,a0
		bsr.w	sub_1A244
		bsr.w	sub_1A14E

locret_1A0C0:							; CODE XREF: sub_1A050+52j
		rts
; ---------------------------------------------------------------------------

loc_1A0C2:							; CODE XREF: sub_1A050+2Cj
		move.w	#$1FFF,d2

loc_1A0C6:							; CODE XREF: sub_1A050+92j
		move.w	#1,d1
		move.w	#7,d0

loc_1A0CE:							; CODE XREF: sub_1A050+8Aj
		move.w	d1,(a0)
		cmp.b	1(a0),d1
		bne.w	loc_1A100
		lsl.w	#1,d1
		dbf	d0,loc_1A0CE
		add.w	#2,a0
		dbf	d2,loc_1A0C6
		move.l	#$200000,a0
		move.l	#$120034,(a0)+
		move.l	#$560078,(a0)+
		move.w	#$E0,d3					; ''
		move.w	#$20,d4					; ' '

loc_1A100:							; CODE XREF: sub_1A050+84j
					; sub_1A050+E4j ...
		move.w	d3,d0

loc_1A102:							; CODE XREF: sub_1A050+CCj
		move.l	#$C0000000,($C00004).l
		move.w	d0,d1
		and.w	d3,d1
		move.w	d1,($C00000).l
		bsr.w	sub_1A140
		sub.w	d4,d0
		bra.s	loc_1A102
; ---------------------------------------------------------------------------

loc_1A11E:							; CODE XREF: sub_1A050+34j
		move.w	#3,d1

loc_1A122:							; CODE XREF: sub_1A050+DAj
		add.w	#1,a0
		lsl.l	#8,d0
		move.b	(a0)+,d0
		dbf	d1,loc_1A122
		cmpnl	 $12345678,d0
		bne.s	loc_1A100
		move.w	#$E0,d3					; ''
		move.w	#$20,d4					; ' '
		bra.s	loc_1A100
; End of function sub_1A050


; =============== S U B	R O U T	I N E =======================================


sub_1A140:							; CODE XREF: sub_1A050+C6p
		jsr	sub_80BA
		rts
; End of function sub_1A140


; =============== S U B	R O U T	I N E =======================================


sub_1A148:							; DATA XREF: sub_1A050o
		addq.w	#1,($FFFFB03E).w
		rte
; End of function sub_1A148


; =============== S U B	R O U T	I N E =======================================


sub_1A14E:							; CODE XREF: sub_1A050+4Ap
					; sub_1A050+6Cp
		move.w	#$1FFD,d1
		clr.w	d0
		lea	($FFFF0000).l,a0

loc_1A15A:							; CODE XREF: sub_1A14E+Ej
		add.b	(a0)+,d0
		dbf	d1,loc_1A15A
		clr.w	d1
		cmp.b	1(a0),d0
		beq.w	loc_1A16C
		addq.w	#1,d1

loc_1A16C:							; CODE XREF: sub_1A14E+18j
		not.w	d0
		cmp.b	(a0),d0
		beq.w	loc_1A176
		addq.w	#1,d1

loc_1A176:							; CODE XREF: sub_1A14E+22j
		swap	d0
		move.b	(a0),d0
		not.b	d0
		cmp.b	1(a0),d0
		beq.w	loc_1A186
		addq.w	#1,d1

loc_1A186:							; CODE XREF: sub_1A14E+32j
		swap	d0
		tst.w	d1
		bne.w	loc_1A196
		clr.w	($FFFFD458).w
		bra.w	locret_1A19A
; ---------------------------------------------------------------------------

loc_1A196:							; CODE XREF: sub_1A14E+3Cj
		st	($FFFFD458).w

locret_1A19A:							; CODE XREF: sub_1A14E+44j
		rts
; End of function sub_1A14E


; =============== S U B	R O U T	I N E =======================================


sub_1A19C:							; CODE XREF: sub_1A050+56p
		lea	($FFFF0000).l,a0
		move.w	#$1FFF,d0
		clr.l	d1

loc_1A1A8:							; CODE XREF: sub_1A19C+Ej
		move.b	d1,(a0)+
		dbf	d0,loc_1A1A8
		move.b	#$FF,($FFFF1FFE).l
		moveq	#0,d0
		move.l	#$2000,d1
		move.l	#$FFFF0000,a0
		bsr.w	sub_1A1E4
		moveq	#1,d0
		moveq	#1,d1
		move.b	#1,($FFFF0000).l
		move.l	#$FFFF0000,a0
		bsr.w	sub_1A1E4
		bsr.w	sub_1A206
		rts
; End of function sub_1A19C


; =============== S U B	R O U T	I N E =======================================


sub_1A1E4:							; CODE XREF: sub_1A19C+28p
					; sub_1A19C+3Ep ...
		movem.l	d0-d2/a0-a1,-(sp)
		move.l	#$200000,a1
		add.l	d0,d0
		subq.l	#1,d1
		clr.w	d2

loc_1A1F4:							; CODE XREF: sub_1A1E4+18j
		move.b	(a0)+,d2
		move.w	d2,(a1,d0.w)
		addq.w	#2,d0
		dbf	d1,loc_1A1F4
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_1A1E4


; =============== S U B	R O U T	I N E =======================================


sub_1A206:							; CODE XREF: sub_1A19C+42p
					; sub_F9CDE+F0p ...
		lea	($FFFF0000).l,a0
		move.l	#$2000,d1
		clr.l	d0
		bsr.w	sub_1A244
		lea	($FFFF0000).l,a0
		clr.w	d0
		move.w	#$1FFD,d1

loc_1A224:							; CODE XREF: sub_1A206+20j
		add.b	(a0)+,d0
		dbf	d1,loc_1A224
		move.b	d0,1(a0)
		not.w	d0
		move.b	d0,(a0)
		move.l	#$FFFF1FFE,a0
		moveq	#2,d1
		move.l	#$1FFE,d0
		bsr.s	sub_1A1E4
		rts
; End of function sub_1A206


; =============== S U B	R O U T	I N E =======================================


sub_1A244:							; CODE XREF: sub_1A050+46p
					; sub_1A050+68p ...
		movem.l	d0-d2/a0-a1,-(sp)
		move.l	#$200000,a1
		add.l	d0,d0
		subq.l	#1,d1

loc_1A252:							; CODE XREF: sub_1A244+16j
		move.b	1(a1,d0.w),d2
		move.b	d2,(a0)+
		addq.w	#2,d0
		dbf	d1,loc_1A252
		movem.l	(sp)+,d0-d2/a0-a1
		rts
; End of function sub_1A244


; =============== S U B	R O U T	I N E =======================================


Z80_MuteAllFM:							; CODE XREF: ROM:000076EEp
					; sub_7CB0-432p ...
		movem.l	d0-a2,-(sp)
		bsr.w	sub_1AD54
		move.b	#$77,($FFFFD182).w			; 'w'
		move.b	#$77,($FFFFD184).w			; 'w'
		moveq	#6,d0					; FM channels
		move.w	#$D187,a0

loc_1A27E:							; CODE XREF: Z80_MuteAllFM+1Ej
		move.b	#$7F,(a0)+				; '' ; Volume of FM channel
		dbf	d0,loc_1A27E
		bsr.w	Z80_WriteVolume
		bsr.w	Z80_ClearFlagUnknown
		movem.l	(sp)+,d0-a2

locret_1A292:							; CODE XREF: sub_1A294+4j sub_1A5C0+4j ...
		rts
; End of function Z80_MuteAllFM


; =============== S U B	R O U T	I N E =======================================


sub_1A294:							; CODE XREF: sub_11132+12p
					; sub_11156+Ep
		cmpnw	 $7A,d0					; 'z'
		bhi.s	locret_1A292
		movem.l	d0-d3/a0-a2,-(sp)
		cmpnw	 $30,d0					; '0'
		blt.w	loc_1A2AA
		bsr.w	sub_1A304

loc_1A2AA:							; CODE XREF: sub_1A294+Ej
		lea	($FFFFD3D4).w,a1
		moveq	#7,d3

loc_1A2B0:							; CODE XREF: sub_1A294+24j
		move.l	(a1),d1
		move.w	d3,d2
		move.w	a1,a2

loc_1A2B6:							; CODE XREF: sub_1A294+28j
		cmp.l	(a1),d1
		bgt.s	loc_1A2B0
		addq.w	#6,a1
		dbf	d3,loc_1A2B6
		tst.w	(a2)
		bpl.w	loc_1A2FE
		asl.w	#2,d0
		lea	(off_2C648).l,a0
		move.l	(a0,d0.w),a0
		move.l	a0,(a2)
		clr.w	4(a2)
		move.b	(a0),5(a2)
		lea	($FFFFD1A4).w,a0
		asl.w	#3,d2
		add.w	d2,a0
		moveq	#7,d0

loc_1A2E6:							; CODE XREF: sub_1A294+66j
		clr.w	(a0)
		clr.b	2(a0)
		clr.b	3(a0)
		move.w	#$7F,4(a0)				; ''
		add.w	#$40,a0					; '@'
		dbf	d0,loc_1A2E6

loc_1A2FE:							; CODE XREF: sub_1A294+2Ej
		movem.l	(sp)+,d0-d3/a0-a2
		rts
; End of function sub_1A294


; =============== S U B	R O U T	I N E =======================================


sub_1A304:							; CODE XREF: sub_794E+2Cp sub_7B30+18p ...
		movem.l	d0-d3/a0-a3,-(sp)
		move	sr,-(sp)
		move	#$2700,sr
		lea	($FFFFD3D4).w,a1
		moveq	#7,d3
		move.l	(off_2C708).l,a0

loc_1A31A:							; CODE XREF: sub_1A304+1Aj
		cmp.l	(a1),a0
		addq.w	#6,a1
		dble	d3,loc_1A31A
		bgt.w	loc_1A368
		clr.l	($FFFFD182).w
		clr.b	($FFFFD186).w
		move.l	#$FFFFFFFF,-6(a1)
		moveq	#5,d1
		lea	($FFFFD3A4).w,a2

loc_1A33C:							; CODE XREF: sub_1A304+5Cj
		move.b	(a2),d2
		and.w	#7,d2
		cmp.w	d2,d3
		bne.w	loc_1A35E
		bsr.w	sub_1A64C
		move.b	4(a2),d0
		lea	($FFFFD187).w,a1
		move.b	#$7F,(a1,d0.w)				; ''
		bset	d0,($FFFFD184).w

loc_1A35E:							; CODE XREF: sub_1A304+40j
		addq.w	#8,a2
		dbf	d1,loc_1A33C
		bsr.w	Z80_WriteVolume

loc_1A368:							; CODE XREF: sub_1A304+1Ej
		move	(sp)+,sr
		movem.l	(sp)+,d0-d3/a0-a3
		rts
; End of function sub_1A304


; =============== S U B	R O U T	I N E =======================================


ReadControls:							; CODE XREF: sub_1A50Ap
		movem.l	d1/a0,-(sp)
		tst.w	($FFFFD046).w
		beq.w	loc_1A39C
		bsr.s	sub_1A3B2
		move.b	d0,($FFFFBEF6).w
		bsr.s	sub_1A3DC
		move.b	d0,($FFFFBEF7).w
		bsr.w	sub_1A406
		move.b	d0,($FFFFBEF8).w
		bsr.w	sub_1A430
		move.b	d0,($FFFFBEF9).w
		bra.w	loc_1A3AC
; ---------------------------------------------------------------------------

loc_1A39C:							; CODE XREF: ReadControls+8j
		bsr.w	sub_1A470
		move.b	d0,($FFFFBEF6).w
		bsr.w	sub_1A464
		move.b	d0,($FFFFBEF7).w

loc_1A3AC:							; CODE XREF: ReadControls+28j
		movem.l	(sp)+,d1/a0
		rts
; End of function ReadControls


; =============== S U B	R O U T	I N E =======================================


sub_1A3B2:							; CODE XREF: ReadControls+Cp
		move.w	#$100,($A11100).l
		move.w	#$64,d0					; 'd'

loc_1A3BE:							; CODE XREF: sub_1A3B2+16j
		btst	#0,($A11100).l
		beq.s	loc_1A3D0
		dbf	d0,loc_1A3BE
		bsr.w	sub_1A4EE

loc_1A3D0:							; CODE XREF: sub_1A3B2+14j
		move.b	#$C,($A10005).l
		bra.w	loc_1A456
; End of function sub_1A3B2


; =============== S U B	R O U T	I N E =======================================


sub_1A3DC:							; CODE XREF: ReadControls+12p
		move.w	#$100,($A11100).l
		move.w	#$64,d0					; 'd'

loc_1A3E8:							; CODE XREF: sub_1A3DC+16j
		btst	#0,($A11100).l
		beq.s	loc_1A3FA
		dbf	d0,loc_1A3E8
		bsr.w	sub_1A4EE

loc_1A3FA:							; CODE XREF: sub_1A3DC+14j
		move.b	#$1C,($A10005).l
		bra.w	loc_1A456
; End of function sub_1A3DC


; =============== S U B	R O U T	I N E =======================================


sub_1A406:							; CODE XREF: ReadControls+18p
		move.w	#$100,($A11100).l
		move.w	#$64,d0					; 'd'

loc_1A412:							; CODE XREF: sub_1A406+16j
		btst	#0,($A11100).l
		beq.s	loc_1A424
		dbf	d0,loc_1A412
		bsr.w	sub_1A4EE

loc_1A424:							; CODE XREF: sub_1A406+14j
		move.b	#$2C,($A10005).l			; ','
		bra.w	loc_1A456
; End of function sub_1A406


; =============== S U B	R O U T	I N E =======================================


sub_1A430:							; CODE XREF: ReadControls+20p
		move.w	#$100,($A11100).l
		move.w	#$64,d0					; 'd'

loc_1A43C:							; CODE XREF: sub_1A430+16j
		btst	#0,($A11100).l
		beq.s	loc_1A44E
		dbf	d0,loc_1A43C
		bsr.w	sub_1A4EE

loc_1A44E:							; CODE XREF: sub_1A430+14j
		move.b	#$3C,($A10005).l			; '<'

loc_1A456:							; CODE XREF: sub_1A3B2+26j
					; sub_1A3DC+26j ...
		movem.l	d1/a0,-(sp)
		lea	($A10003).l,a0
		bra.w	loc_1A498
; End of function sub_1A430


; =============== S U B	R O U T	I N E =======================================


sub_1A464:							; CODE XREF: ReadControls+34p
		movem.l	d1/a0,-(sp)
		lea	($A10005).l,a0
		bra.s	loc_1A47A
; End of function sub_1A464


; =============== S U B	R O U T	I N E =======================================


sub_1A470:							; CODE XREF: ReadControls:loc_1A39Cp
		movem.l	d1/a0,-(sp)
		lea	($A10003).l,a0

loc_1A47A:							; CODE XREF: sub_1A464+Aj
		move.w	#$100,($A11100).l
		move.w	#$64,d0					; 'd'

loc_1A486:							; CODE XREF: sub_1A470+20j
		btst	#0,($A11100).l
		beq.s	loc_1A498
		dbf	d0,loc_1A486
		bsr.w	sub_1A4EE

loc_1A498:							; CODE XREF: sub_1A430+30j
					; sub_1A470+1Ej
		moveq	#0,d0
		move.b	#0,(a0)
		nop
		nop
		move.b	(a0),d0
		move.b	#$40,(a0)				; '@'
		nop
		nop
		move.b	(a0),d1
		move.w	#0,($A11100).l
		asl.b	#2,d0
		and.b	#$C0,d0
		and.b	#$3F,d1					; '?'
		or.b	d1,d0
		not.b	d0
		not.b	d1
		and.w	#$F,d1
		lea	byte_1A4DE(pc),a0
		and.b	#$F0,d0
		or.b	(a0,d1.w),d0
		not.b	d0
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_1A470

; ---------------------------------------------------------------------------
byte_1A4DE:	dc.b	0					; DATA XREF: sub_1A470+5Ao
		dc.b	1
		dc.b	2
		dc.b	1
		dc.b	4
		dc.b	5
		dc.b	6
		dc.b	6
		dc.b	8
		dc.b	9
		dc.b	$A
		dc.b	$A
		dc.b	8
		dc.b	9
		dc.b	$A
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_1A4EE:							; CODE XREF: sub_1A3B2+1Ap
					; sub_1A3DC+1Ap ...
		move.w	#$100,($A11200).l
		move.w	#$100,($A11100).l

loc_1A4FE:							; CODE XREF: sub_1A4EE+18j
		btst	#0,($A11100).l
		bne.s	loc_1A4FE
		rts
; End of function sub_1A4EE


; =============== S U B	R O U T	I N E =======================================


sub_1A50A:							; CODE XREF: ROM:000076F4p
					; sub_15D9A+84p ...
		bsr.w	ReadControls
		clr.b	($FFFFD180).w
		clr.l	($FFFFD182).w
		clr.b	($FFFFD186).w

loc_1A51A:							; CODE XREF: sub_1A50A+36j
		lea	($FFFFD3D4).w,a5
		moveq	#7,d7

loc_1A520:							; CODE XREF: sub_1A50A+1Cj
		bsr.w	sub_1A5C0
		addq.w	#6,a5
		dbf	d7,loc_1A520
		btst	#0,($FFFFDEF2).w
		beq.w	loc_1A542
		subq.w	#1,($FFFFD070).w
		bpl.w	loc_1A542
		addq.w	#6,($FFFFD070).w
		bra.s	loc_1A51A
; ---------------------------------------------------------------------------

loc_1A542:							; CODE XREF: sub_1A50A+26j
					; sub_1A50A+2Ej
		tst.b	($FFFFD180).w
		beq.w	loc_1A54E
		bsr.w	Z80_WriteVolume

loc_1A54E:							; CODE XREF: sub_1A50A+3Cj
		move.w	#$D3A4,a0
		moveq	#5,d0

loc_1A554:							; CODE XREF: sub_1A50A+58j
		addq.b	#1,5(a0)
		bne.w	loc_1A560
		subq.b	#1,5(a0)

loc_1A560:							; CODE XREF: sub_1A50A+4Ej
		addq.w	#8,a0
		dbf	d0,loc_1A554
		rts
; End of function sub_1A50A

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR Z80_WriteVolume

loc_1A568:							; CODE XREF: Z80_WriteVolume+1Ej
					; Z80_WriteVolume+26j
		clr.w	($A11100).l
		moveq	#$64,d0					; 'd'

loc_1A570:							; CODE XREF: Z80_WriteVolume:loc_1A570j
		dbf	d0,loc_1A570
; END OF FUNCTION CHUNK	FOR Z80_WriteVolume

; =============== S U B	R O U T	I N E =======================================


Z80_WriteVolume:						; CODE XREF: Z80_MuteAllFM+22p
					; sub_1A304+60p ...

; FUNCTION CHUNK AT 0001A568 SIZE 0000000C BYTES

		move.w	#$100,($A11100).l

loc_1A57C:							; CODE XREF: Z80_WriteVolume+10j
		btst	#0,($A11100).l
		bne.s	loc_1A57C
		move.l	#$A00000,a0
		cmp.b	#0,$97(a0)				; These	are likely busy	checks (to see if the z80 is accessing YM2612, or even accessing the data we're about the write, if it is, the z80 will resume, and the 68k will delay for a while until the z80 is finished.
		bne.s	loc_1A568
		cmp.b	#$7D,$96(a0)				; '}'
		bne.s	loc_1A568
		move.b	#$D1,$96(a0)
		move.b	#0,$97(a0)
		add.w	#2,a0
		move.w	#$D182,a1
		moveq	#$20,d0					; ' '

loc_1A5B2:							; CODE XREF: Z80_WriteVolume+40j
		move.b	(a1)+,(a0)+
		dbf	d0,loc_1A5B2
		clr.w	($A11100).l

locret_1A5BE:							; CODE XREF: ROM:0001A75Cj
		rts
; End of function Z80_WriteVolume


; =============== S U B	R O U T	I N E =======================================


sub_1A5C0:							; CODE XREF: sub_1A50A:loc_1A520p
					; sub_1A5C0+48j
		subq.w	#1,4(a5)
		bpl.w	locret_1A292

loc_1A5C8:							; CODE XREF: sub_1A5C0+32j
		tst.w	(a5)
		bmi.w	locret_1A292
		move.l	(a5),a0
		addq.l	#4,(a5)
		clr.w	4(a5)
		move.b	4(a0),5(a5)
		move.b	1(a0),d0
		bne.w	loc_1A5F4
		st	(a5)
		tst.w	2(a0)
		bmi.w	locret_1A292
		move.l	2(a0),(a5)
		bra.s	loc_1A5C8
; ---------------------------------------------------------------------------

loc_1A5F4:							; CODE XREF: sub_1A5C0+20j
		move.b	1(a0),d0
		and.w	#$70,d0					; 'p'
		lsr.w	#3,d0
		lea	word_1A60A(pc),a2
		add.w	(a2,d0.w),a2
		jsr	(a2)
		bra.s	sub_1A5C0
; End of function sub_1A5C0

; ---------------------------------------------------------------------------
word_1A60A:	dc.w	$10					; DATA XREF: sub_1A5C0+3Eo
		dc.w	$84
		dc.w	$6B6
		dc.w	$678
		dc.w	$626
		dc.w	$6B6
		dc.w	$640
		dc.w	$6B6
; ---------------------------------------------------------------------------

loc_1A61A:							; CODE XREF: ROM:0001A692j
		move.b	1(a0),d0
		and.w	#$F,d0
		asl.w	#3,d0
		or.w	d7,d0
		asl.w	#8,d0
		move.b	2(a0),d0
		move.w	#$D3D4,a2
		moveq	#5,d1

loc_1A632:							; CODE XREF: ROM:0001A636j
					; ROM:0001A644j
		subq.w	#8,a2
		cmp.w	(a2),d0
		dbeq	d1,loc_1A632
		bne.w	locret_1A292
		btst	#0,6(a2)
		dbne	d1,loc_1A632
		beq.w	locret_1A292

; =============== S U B	R O U T	I N E =======================================


sub_1A64C:							; CODE XREF: sub_1A304+44p
		bclr	#0,6(a2)
		beq.w	locret_1A292
		cmp.b	#$60,3(a2)				; '`'
		bge.w	Z80_ClearFlagUnknown
		move.b	4(a2),d0
		bset	d0,($FFFFD182).w
		st	($FFFFD180).w
		rts
; End of function sub_1A64C


; =============== S U B	R O U T	I N E =======================================


Z80_ClearFlagUnknown:						; CODE XREF: Z80_MuteAllFM+26p
					; sub_1A64C+10j
		move.w	#$100,($A11100).l

loc_1A676:							; CODE XREF: Z80_ClearFlagUnknown+10j
		btst	#0,($A11100).l
		bne.s	loc_1A676
		clr.b	($A0008E).l
		clr.w	($A11100).l
		rts
; End of function Z80_ClearFlagUnknown

; ---------------------------------------------------------------------------
		tst.b	3(a0)
		beq.s	loc_1A61A
		move.w	#$D3CC,a2
		lea	($FFFFD1A4).w,a3
		move.b	1(a0),d0
		and.w	#$F,d0
		asl.w	#3,d0
		or.w	d7,d0
		move.w	d0,d6
		asl.w	#3,d0
		move.b	3(a3,d0.w),d0
		cmpnb	 $60,d0					; '`'
		bge.w	Unk_PCMSampleWrite
		subq.w	#8,a2
		moveq	#4,d1
		bra.w	loc_1A6C8
; ---------------------------------------------------------------------------

loc_1A6C0:							; CODE XREF: ROM:0001A6D0j
		cmp.b	5(a2),d2
		bhi.w	loc_1A6CE

loc_1A6C8:							; CODE XREF: ROM:0001A6BCj
		move.w	a2,a4
		move.b	5(a4),d2

loc_1A6CE:							; CODE XREF: ROM:0001A6C4j
		subq.w	#8,a2
		dbf	d1,loc_1A6C0
		moveq	#4,d1
		move.w	#$D3CC,a2

loc_1A6DA:							; CODE XREF: ROM:0001A6E2j
					; ROM:0001A6F0j
		subq.w	#8,a2
		btst	#0,6(a2)
		dbeq	d1,loc_1A6DA
		bne.w	loc_1A6FC
		move.w	a2,a4
		cmp.b	3(a4),d0
		dbeq	d1,loc_1A6DA
		beq.w	loc_1A720
		bra.w	loc_1A704
; ---------------------------------------------------------------------------

loc_1A6FC:							; CODE XREF: ROM:0001A6E6j
		cmp.b	3(a4),d0
		beq.w	loc_1A720

loc_1A704:							; CODE XREF: ROM:0001A6F8j
		move.w	a4,a2
		move.b	d0,3(a2)
		clr.w	d1
		move.b	4(a2),d1
		bset	d1,($FFFFD186).w
		move.w	#$D19C,a4
		move.b	d0,(a4,d1.w)
		st	($FFFFD180).w

loc_1A720:							; CODE XREF: ROM:0001A6F4j
					; ROM:0001A700j
		clr.b	5(a2)
		bset	#0,6(a2)
		move.b	d6,(a2)
		move.b	2(a0),1(a2)
		move.b	3(a0),2(a2)
		clr.w	d1
		move.b	4(a2),d1
		bset	d1,($FFFFD182).w
		bset	d1,($FFFFD183).w
		bsr.w	sub_1ABAA
		bra.w	sub_1A7D8
; ---------------------------------------------------------------------------

Unk_PCMSampleWrite:						; CODE XREF: ROM:0001A6B4j
		bset	#0,6(a2)
		beq.w	loc_1A760
		cmp.b	3(a2),d0
		ble.w	locret_1A5BE

loc_1A760:							; CODE XREF: ROM:0001A754j
		move.b	d0,3(a2)
		clr.b	5(a2)
		move.b	d6,(a2)
		move.b	2(a0),1(a2)
		move.b	3(a0),2(a2)
		ext.w	d0
		sub.w	#$60,d0					; '`'
		asl.w	#3,d0
		lea	$1B01C(pc),a1				; Probably samples address
		move.l	4(a1,d0.w),d1
		move.l	(a1,d0.w),d0
		move.w	#$100,($A11100).l

loc_1A792:							; CODE XREF: ROM:0001A79Aj
		btst	#0,($A11100).l
		bne.s	loc_1A792
		move.l	#$A00000,a1
		move.b	d0,$25(a1)
		lsr.w	#8,d0
		move.b	d0,$24(a1)
		swap	d0
		move.b	d0,$23(a1)
		move.b	d1,$28(a1)
		lsr.w	#8,d1
		move.b	d1,$27(a1)
		swap	d1
		move.b	d1,$26(a1)
		move.b	#$29,$96(a1)				; ')'
		move.b	#0,$97(a1)
		clr.w	($A11100).l
		bsr.w	sub_1ABAA

; =============== S U B	R O U T	I N E =======================================


sub_1A7D8:							; CODE XREF: ROM:0001A74Aj
					; ROM:0001AC76p
		clr.l	d3
		move.b	1(a2),d3
		divu.w	#$C,d3
		move.w	d3,-(sp)
		swap	d3
		add.w	d3,d3
		lea	word_1A890(pc),a4
		move.w	(a4,d3.w),d2
		clr.w	d1
		move.b	(a2),d1
		asl.w	#3,d1
		move.w	(a3,d1.w),d3
		beq.w	loc_1A834
		moveq	#$C,d1
		cmp.b	#$60,3(a2)				; '`'
		bge.w	loc_1A81C
		move.b	3(a2),d1
		asl.w	#5,d1
		lea	(unk_2C248).l,a4
		move.b	$1E(a4,d1.w),d1
		ext.w	d1

loc_1A81C:							; CODE XREF: sub_1A7D8+2Ej
		muls.w	d1,d3
		asr.l	#2,d3
		asr.w	#7,d3
		add.w	#$C0,d3					; ''
		add.w	d3,d3
		lea	word_1A8A8(pc),a4
		mulu.w	(a4,d3.w),d2
		asl.l	#1,d2
		swap	d2

loc_1A834:							; CODE XREF: sub_1A7D8+22j
		move.w	(sp)+,d3
		cmp.b	#$60,3(a2)				; '`'
		bge.w	loc_1A860
		asl.w	#3,d3
		asl.w	#8,d3
		or.w	d3,d2
		clr.w	d3
		move.b	4(a2),d3
		bset	d3,($FFFFD185).w
		add.w	d3,d3
		move.w	#$D18E,a4
		move.w	d2,(a4,d3.w)
		st	($FFFFD180).w
		rts
; ---------------------------------------------------------------------------

loc_1A860:							; CODE XREF: sub_1A7D8+64j
		btst	#0,6(a2)
		beq.w	locret_1A88E
		move.w	#$100,($A11100).l

loc_1A872:							; CODE XREF: sub_1A7D8+A2j
		btst	#0,($A11100).l
		bne.s	loc_1A872
		neg.w	d3
		addq.w	#8,d3
		lsr.w	d3,d2
		move.b	d2,($A0008E).l
		clr.w	($A11100).l

locret_1A88E:							; CODE XREF: sub_1A7D8+8Ej
		rts
; End of function sub_1A7D8

; ---------------------------------------------------------------------------
word_1A890:	dc.w	$146					; DATA XREF: sub_1A7D8+10o
		dc.w	$159
		dc.w	$16E
		dc.w	$184
		dc.w	$19B
		dc.w	$1B3
		dc.w	$1CD
		dc.w	$1E8
		dc.w	$205
		dc.w	$224
		dc.w	$245
		dc.w	$268
word_1A8A8:	dc.w	$4000					; DATA XREF: sub_1A7D8+50o
		dc.w	$403B
		dc.w	$4076
		dc.w	$40B2
		dc.w	$40EE
		dc.w	$412A
		dc.w	$4166
		dc.w	$41A3
		dc.w	$41E0
		dc.w	$421D
		dc.w	$425A
		dc.w	$4297
		dc.w	$42D5
		dc.w	$4313
		dc.w	$4351
		dc.w	$438F
		dc.w	$43CE
		dc.w	$440D
		dc.w	$444C
		dc.w	$448B
		dc.w	$44CA
		dc.w	$450A
		dc.w	$454A
		dc.w	$458A
		dc.w	$45CA
		dc.w	$460B
		dc.w	$464C
		dc.w	$468D
		dc.w	$46CE
		dc.w	$4710
		dc.w	$4752
		dc.w	$4794
		dc.w	$47D6
		dc.w	$4818
		dc.w	$485B
		dc.w	$489E
		dc.w	$48E1
		dc.w	$4925
		dc.w	$4969
		dc.w	$49AD
		dc.w	$49F1
		dc.w	$4A35
		dc.w	$4A7A
		dc.w	$4ABF
		dc.w	$4B04
		dc.w	$4B4A
		dc.w	$4B8F
		dc.w	$4BD5
		dc.w	$4C1B
		dc.w	$4C62
		dc.w	$4CA9
		dc.w	$4CF0
		dc.w	$4D37
		dc.w	$4D7E
		dc.w	$4DC6
		dc.w	$4E0E
		dc.w	$4E56
		dc.w	$4E9F
		dc.w	$4EE8
		dc.w	$4F31
		dc.w	$4F7A
		dc.w	$4FC4
		dc.w	$500E
		dc.w	$5058
		dc.w	$50A2
		dc.w	$50ED
		dc.w	$5138
		dc.w	$5183
		dc.w	$51CE
		dc.w	$521A
		dc.w	$5266
		dc.w	$52B2
		dc.w	$52FF
		dc.w	$534C
		dc.w	$5399
		dc.w	$53E6
		dc.w	$5434
		dc.w	$5482
		dc.w	$54D0
		dc.w	$551F
		dc.w	$556E
		dc.w	$55BD
		dc.w	$560C
		dc.w	$565C
		dc.w	$56AC
		dc.w	$56FC
		dc.w	$574C
		dc.w	$579D
		dc.w	$57EE
		dc.w	$5840
		dc.w	$5891
		dc.w	$58E3
		dc.w	$5936
		dc.w	$5988
		dc.w	$59DB
		dc.w	$5A2E
		dc.w	$5A82
		dc.w	$5AD6
		dc.w	$5B2A
		dc.w	$5B7E
		dc.w	$5BD3
		dc.w	$5C28
		dc.w	$5C7D
		dc.w	$5CD3
		dc.w	$5D29
		dc.w	$5D7F
		dc.w	$5DD6
		dc.w	$5E2D
		dc.w	$5E84
		dc.w	$5EDB
		dc.w	$5F33
		dc.w	$5F8B
		dc.w	$5FE4
		dc.w	$603D
		dc.w	$6096
		dc.w	$60EF
		dc.w	$6149
		dc.w	$61A3
		dc.w	$61FD
		dc.w	$6258
		dc.w	$62B3
		dc.w	$630E
		dc.w	$636A
		dc.w	$63C6
		dc.w	$6423
		dc.w	$647F
		dc.w	$64DC
		dc.w	$653A
		dc.w	$6597
		dc.w	$65F6
		dc.w	$6654
		dc.w	$66B3
		dc.w	$6712
		dc.w	$6771
		dc.w	$67D1
		dc.w	$6831
		dc.w	$6892
		dc.w	$68F2
		dc.w	$6954
		dc.w	$69B5
		dc.w	$6A17
		dc.w	$6A79
		dc.w	$6ADC
		dc.w	$6B3F
		dc.w	$6BA2
		dc.w	$6C06
		dc.w	$6C6A
		dc.w	$6CCE
		dc.w	$6D33
		dc.w	$6D98
		dc.w	$6DFD
		dc.w	$6E63
		dc.w	$6EC9
		dc.w	$6F30
		dc.w	$6F97
		dc.w	$6FFE
		dc.w	$7066
		dc.w	$70CE
		dc.w	$7136
		dc.w	$719F
		dc.w	$7208
		dc.w	$7272
		dc.w	$72DC
		dc.w	$7346
		dc.w	$73B1
		dc.w	$741C
		dc.w	$7488
		dc.w	$74F4
		dc.w	$7560
		dc.w	$75CD
		dc.w	$763A
		dc.w	$76A7
		dc.w	$7715
		dc.w	$7783
		dc.w	$77F2
		dc.w	$7861
		dc.w	$78D0
		dc.w	$7940
		dc.w	$79B0
		dc.w	$7A21
		dc.w	$7A92
		dc.w	$7B04
		dc.w	$7B76
		dc.w	$7BE8
		dc.w	$7C5B
		dc.w	$7CCE
		dc.w	$7D41
		dc.w	$7DB5
		dc.w	$7E2A
		dc.w	$7E9F
		dc.w	$7F14
		dc.w	$7F89
		dc.w	$8000
		dc.w	$8076
		dc.w	$80ED
		dc.w	$8164
		dc.w	$81DC
		dc.w	$8254
		dc.w	$82CD
		dc.w	$8346
		dc.w	$83C0
		dc.w	$843A
		dc.w	$84B4
		dc.w	$852F
		dc.w	$85AA
		dc.w	$8626
		dc.w	$86A2
		dc.w	$871F
		dc.w	$879C
		dc.w	$881A
		dc.w	$8898
		dc.w	$8916
		dc.w	$8995
		dc.w	$8A14
		dc.w	$8A94
		dc.w	$8B14
		dc.w	$8B95
		dc.w	$8C16
		dc.w	$8C98
		dc.w	$8D1A
		dc.w	$8D9D
		dc.w	$8E20
		dc.w	$8EA4
		dc.w	$8F28
		dc.w	$8FAC
		dc.w	$9031
		dc.w	$90B7
		dc.w	$913D
		dc.w	$91C3
		dc.w	$924A
		dc.w	$92D2
		dc.w	$935A
		dc.w	$93E2
		dc.w	$946B
		dc.w	$94F4
		dc.w	$957E
		dc.w	$9609
		dc.w	$9694
		dc.w	$971F
		dc.w	$97AB
		dc.w	$9837
		dc.w	$98C4
		dc.w	$9952
		dc.w	$99E0
		dc.w	$9A6E
		dc.w	$9AFD
		dc.w	$9B8D
		dc.w	$9C1D
		dc.w	$9CAD
		dc.w	$9D3E
		dc.w	$9DD0
		dc.w	$9E62
		dc.w	$9EF5
		dc.w	$9F88
		dc.w	$A01C
		dc.w	$A0B0
		dc.w	$A145
		dc.w	$A1DA
		dc.w	$A270
		dc.w	$A306
		dc.w	$A39D
		dc.w	$A435
		dc.w	$A4CD
		dc.w	$A565
		dc.w	$A5FE
		dc.w	$A698
		dc.w	$A732
		dc.w	$A7CD
		dc.w	$A868
		dc.w	$A904
		dc.w	$A9A1
		dc.w	$AA3E
		dc.w	$AADC
		dc.w	$AB7A
		dc.w	$AC18
		dc.w	$ACB8
		dc.w	$AD58
		dc.w	$ADF8
		dc.w	$AE99
		dc.w	$AF3B
		dc.w	$AFDD
		dc.w	$B080
		dc.w	$B123
		dc.w	$B1C7
		dc.w	$B26C
		dc.w	$B311
		dc.w	$B3B7
		dc.w	$B45D
		dc.w	$B504
		dc.w	$B5AC
		dc.w	$B654
		dc.w	$B6FD
		dc.w	$B7A7
		dc.w	$B851
		dc.w	$B8FB
		dc.w	$B9A6
		dc.w	$BA52
		dc.w	$BAFF
		dc.w	$BBAC
		dc.w	$BC5A
		dc.w	$BD08
		dc.w	$BDB7
		dc.w	$BE67
		dc.w	$BF17
		dc.w	$BFC8
		dc.w	$C07A
		dc.w	$C12C
		dc.w	$C1DF
		dc.w	$C292
		dc.w	$C346
		dc.w	$C3FB
		dc.w	$C4B1
		dc.w	$C567
		dc.w	$C61D
		dc.w	$C6D5
		dc.w	$C78D
		dc.w	$C846
		dc.w	$C8FF
		dc.w	$C9B9
		dc.w	$CA74
		dc.w	$CB2F
		dc.w	$CBEC
		dc.w	$CCA8
		dc.w	$CD66
		dc.w	$CE24
		dc.w	$CEE3
		dc.w	$CFA2
		dc.w	$D063
		dc.w	$D124
		dc.w	$D1E5
		dc.w	$D2A8
		dc.w	$D36B
		dc.w	$D42E
		dc.w	$D4F3
		dc.w	$D5B8
		dc.w	$D67E
		dc.w	$D744
		dc.w	$D80C
		dc.w	$D8D4
		dc.w	$D99D
		dc.w	$DA66
		dc.w	$DB30
		dc.w	$DBFB
		dc.w	$DCC7
		dc.w	$DD93
		dc.w	$DE60
		dc.w	$DF2E
		dc.w	$DFFD
		dc.w	$E0CC
		dc.w	$E19D
		dc.w	$E26D
		dc.w	$E33F
		dc.w	$E411
		dc.w	$E4E5
		dc.w	$E5B9
		dc.w	$E68D
		dc.w	$E763
		dc.w	$E839
		dc.w	$E910
		dc.w	$E9E8
		dc.w	$EAC0
		dc.w	$EB9A
		dc.w	$EC74
		dc.w	$ED4F
		dc.w	$EE2A
		dc.w	$EF07
		dc.w	$EFE4
		dc.w	$F0C2
		dc.w	$F1A1
		dc.w	$F281
		dc.w	$F361
		dc.w	$F443
		dc.w	$F525
		dc.w	$F608
		dc.w	$F6EC
		dc.w	$F7D0
		dc.w	$F8B6
		dc.w	$F99C
		dc.w	$FA83
		dc.w	$FB6B
		dc.w	$FC54
		dc.w	$FD3E
		dc.w	$FE28
		dc.w	$FF13
		dc.w	$FF13

; =============== S U B	R O U T	I N E =======================================


sub_1ABAA:							; CODE XREF: ROM:0001A746p
					; ROM:0001A7D4p ...
		movem.l	d0,-(sp)
		clr.w	d0
		move.b	2(a2),d0
		clr.w	d1
		move.b	(a2),d1
		asl.w	#3,d1
		mulu.w	4(a3,d1.w),d0
		lsr.w	#7,d0
		cmp.b	#$60,3(a2)				; '`'
		bge.w	loc_1AC00
		lea	byte_1ABF0(pc),a4
		lsr.w	#3,d0
		move.b	(a4,d0.w),d0
		clr.w	d3
		move.b	4(a2),d3
		move.w	#$D187,a4
		move.b	d0,(a4,d3.w)
		bset	d3,($FFFFD184).w
		st	($FFFFD180).w
		movem.l	(sp)+,d0
		rts
; ---------------------------------------------------------------------------
byte_1ABF0:	dc.b	$1A					; DATA XREF: sub_1ABAA+20o
		dc.b	$18
		dc.b	$16
		dc.b	$14
		dc.b	$12
		dc.b	$10
		dc.b	$E
		dc.b	$C
		dc.b	$A
		dc.b	8
		dc.b	6
		dc.b	4
		dc.b	3
		dc.b	2
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------

loc_1AC00:							; CODE XREF: sub_1ABAA+1Cj
		move.w	#$100,($A11100).l

loc_1AC08:							; CODE XREF: sub_1ABAA+66j
		btst	#0,($A11100).l
		bne.s	loc_1AC08
		lsr.b	#2,d0
		cmpnw	 3,d0
		bgt.w	loc_1AC1E
		moveq	#3,d0

loc_1AC1E:							; CODE XREF: sub_1ABAA+6Ej
		move.b	d0,($A00083).l
		clr.w	($A11100).l
		movem.l	(sp)+,d0
		rts
; End of function sub_1ABAA

; ---------------------------------------------------------------------------
		lea	($FFFFD1A4).w,a3
		move.b	1(a0),d0
		and.w	#$F,d0
		asl.w	#3,d0
		or.w	d7,d0
		asl.w	#3,d0
		move.b	2(a0),3(a3,d0.w)
		rts
; ---------------------------------------------------------------------------
		lea	($FFFFD1A4).w,a3
		move.b	1(a0),d0
		and.w	#$F,d0
		asl.w	#3,d0
		or.w	d7,d0
		move.w	d0,d4
		asl.w	#3,d0
		move.w	2(a0),(a3,d0.w)
		sub.w	#$2000,(a3,d0.w)
		move.w	#$D3A4,a2
		moveq	#5,d0

loc_1AC70:							; CODE XREF: ROM:0001AC7Cj
		cmp.b	(a2),d4
		bne.w	loc_1AC7A
		bsr.w	sub_1A7D8

loc_1AC7A:							; CODE XREF: ROM:0001AC72j
		addq.w	#8,a2
		dbf	d0,loc_1AC70
		rts
; ---------------------------------------------------------------------------
		cmp.b	#7,2(a0)
		beq.w	loc_1AC8E
		rts
; ---------------------------------------------------------------------------

loc_1AC8E:							; CODE XREF: ROM:0001AC88j
		lea	($FFFFD1A4).w,a3
		move.b	1(a0),d0
		and.w	#$F,d0
		asl.w	#3,d0
		or.w	d7,d0
		move.w	d0,d4
		asl.w	#3,d0
		move.b	3(a0),5(a3,d0.w)
		move.w	#$D3A4,a2
		moveq	#5,d0

loc_1ACAE:							; CODE XREF: ROM:0001ACBAj
		cmp.b	(a2),d4
		bne.w	loc_1ACB8
		bsr.w	sub_1ABAA

loc_1ACB8:							; CODE XREF: ROM:0001ACB0j
		addq.w	#8,a2
		dbf	d0,loc_1ACAE
		rts
; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


Z80_LoadROM:							; CODE XREF: ROM:loc_76E8p
		movem.l	d0-d2/a0-a2,-(sp)
		bsr.w	sub_1AD54
		move.w	#$100,($A11200).l
		move.w	#$100,($A11100).l

loc_1ACDA:							; CODE XREF: Z80_LoadROM+20j
		btst	#0,($A11100).l
		bne.s	loc_1ACDA
		move.l	#Z80ROM,a1
		move.l	#$A00000,a2
		move.w	#$294,d0

loc_1ACF4:							; CODE XREF: Z80_LoadROM+34j
		move.b	(a1)+,(a2)+
		dbf	d0,loc_1ACF4
		move.l	#$A02000,a2
		moveq	#8,d2
		move.l	#$1C,d3

loc_1AD08:							; CODE XREF: Z80_LoadROM+64j
		move.w	#$FF,d0

loc_1AD0C:							; CODE XREF: Z80_LoadROM+5Cj
		move.w	d0,d1
		sub.w	#$80,d1					; ''
		asl.w	#3,d1
		ext.l	d1
		divs.w	d2,d1
		add.w	#$80,d1					; ''
		move.b	d1,-(a2)
		dbf	d0,loc_1AD0C
		clr.b	(a2)
		addq.w	#1,d2
		dbf	d3,loc_1AD08
		move.w	#0,($A11200).l
		move.w	#0,($A11100).l
		move.w	#$1F4,d0

loc_1AD3E:							; CODE XREF: Z80_LoadROM:loc_1AD3Ej
		dbf	d0,loc_1AD3E
		move.w	#$100,($A11200).l
		clr.b	($FFFFD180).w
		movem.l	(sp)+,d0-d2/a0-a2
		rts
; End of function Z80_LoadROM


; =============== S U B	R O U T	I N E =======================================


sub_1AD54:							; CODE XREF: Z80_MuteAllFM+4p
					; Z80_LoadROM+4p
		lea	($FFFFD3D4).w,a0
		moveq	#7,d0
		moveq	#$FFFFFFFF,d1

loc_1AD5C:							; CODE XREF: sub_1AD54+Cj
		move.l	d1,(a0)
		addq.w	#6,a0
		dbf	d0,loc_1AD5C
		moveq	#5,d0
		move.w	#$D3D4,a0

loc_1AD6A:							; CODE XREF: sub_1AD54+36j
		subq.w	#8,a0
		move.b	d0,4(a0)
		cmpnw	 3,d0
		blt.w	loc_1AD7C
		addq.b	#1,4(a0)

loc_1AD7C:							; CODE XREF: sub_1AD54+20j
		st	3(a0)
		st	(a0)
		clr.b	1(a0)
		clr.b	6(a0)
		dbf	d0,loc_1AD6A
		rts
; End of function sub_1AD54

; ---------------------------------------------------------------------------
Z80ROM:		dc.b	$18					; DATA XREF: Z80_LoadROM+22o
		dc.b	$27					; '
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$F3					; 
		dc.b	$31					; 1
		dc.b	$95					; 
		dc.b	2
		dc.b	$ED					; 
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$7D					; }
		dc.b	0
		dc.b	$22					; "
		dc.b	$96					; 
		dc.b	0
		dc.b	$DD					; 
		dc.b	$21					; !
		dc.b	0
		dc.b	$40					; @
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$2B					; +
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	1
		dc.b	$80					; 
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$22					; "
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	1
		dc.b	$C
		dc.b	$FD					; 
		dc.b	$21					; !
		dc.b	$23					; #
		dc.b	0
		dc.b	$18
		dc.b	4
		dc.b	$FD					; 
		dc.b	$21					; !
		dc.b	$26					; &
		dc.b	0
		dc.b	$FD					; 
		dc.b	$66					; f
		dc.b	1
		dc.b	$FD					; 
		dc.b	$6E					; n
		dc.b	2
		dc.b	$7C					; |
		dc.b	$17
		dc.b	$FD					; 
		dc.b	$7E					; ~
		dc.b	0
		dc.b	$17
		dc.b	$32					; 2
		dc.b	$9D					; 
		dc.b	0
		dc.b	$C3					; 
		dc.b	$98					; 
		dc.b	0
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$2A					; *
		dc.b	$5E					; ^
		dc.b	$16
		dc.b	$1F
		dc.b	$1A
		dc.b	$B7					; 
		dc.b	$CA					; 
		dc.b	$67					; g
		dc.b	0
		dc.b	$32					; 2
		dc.b	1
		dc.b	$40					; @
		dc.b	$79					; y
		dc.b	$C6					; 
		dc.b	0
		dc.b	$4F					; O
		dc.b	$11
		dc.b	0
		dc.b	0
		dc.b	$ED					; 
		dc.b	$5A					; Z
		dc.b	$D2					; 
		dc.b	$7D					; }
		dc.b	0
		dc.b	$7C					; |
		dc.b	$F6					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$3E					; >
		dc.b	0
		dc.b	$CD					; 
		dc.b	$AA					; 
		dc.b	0
		dc.b	$3C					; <
		dc.b	$32					; 2
		dc.b	$9D					; 
		dc.b	0
		dc.b	$37					; 7
		dc.b	$3F					; ?
		dc.b	$C3					; 
		dc.b	$95					; 
		dc.b	0
		dc.b	$37					; 7
		dc.b	$3F					; ?
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$32					; 2
		dc.b	0
		dc.b	$60					; `
		dc.b	$1F
		dc.b	$C9					; 
		dc.b	$E5					; 
		dc.b	$C5					; 
		dc.b	$1E
		dc.b	$FF
		dc.b	$3A					; :
		dc.b	2
		dc.b	0
		dc.b	$1C
		dc.b	$CB					; 
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$14
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$28					; (
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$73					; s
		dc.b	1
		dc.b	$B7					; 
		dc.b	$20
		dc.b	$E5					; 
		dc.b	$1E
		dc.b	$FF
		dc.b	$3A					; :
		dc.b	6
		dc.b	0
		dc.b	$1C
		dc.b	$CB					; 
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$F5					; 
		dc.b	$CD					; 
		dc.b	$5B					; [
		dc.b	2
		dc.b	$CD					; 
		dc.b	$80					; 
		dc.b	2
		dc.b	$C6					; 
		dc.b	$30					; 0
		dc.b	$57					; W
		dc.b	6
		dc.b	$1E
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	0
		dc.b	$4E					; N
		dc.b	$23					; #
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$71					; q
		dc.b	1
		dc.b	$14
		dc.b	$14
		dc.b	$14
		dc.b	$14
		dc.b	$78					; x
		dc.b	$FE					; 
		dc.b	3
		dc.b	$20
		dc.b	4
		dc.b	$7A					; z
		dc.b	$C6					; 
		dc.b	$10
		dc.b	$57					; W
		dc.b	$10
		dc.b	$DD					; 
		dc.b	$DD					; 
		dc.b	$21					; !
		dc.b	0
		dc.b	$40					; @
		dc.b	$F1					; 
		dc.b	$20
		dc.b	$C5					; 
		dc.b	$1E
		dc.b	$FF
		dc.b	$3A					; :
		dc.b	5
		dc.b	0
		dc.b	$1C
		dc.b	$CB					; 
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$3C					; <
		dc.b	$F5					; 
		dc.b	$21					; !
		dc.b	$E
		dc.b	0
		dc.b	$7B					; {
		dc.b	$CB					; 
		dc.b	$27					; '
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$CD					; 
		dc.b	$80					; 
		dc.b	2
		dc.b	$C6					; 
		dc.b	$A4					; 
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$56					; V
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$D6					; 
		dc.b	4
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$23					; #
		dc.b	$56					; V
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$DD					; 
		dc.b	$21					; !
		dc.b	0
		dc.b	$40					; @
		dc.b	$F1					; 
		dc.b	$20
		dc.b	$BD					; 
		dc.b	$1E
		dc.b	$FF
		dc.b	$3A					; :
		dc.b	4
		dc.b	0
		dc.b	$1C
		dc.b	$CB					; 
		dc.b	$3F					; ?
		dc.b	$D2					; 
		dc.b	$E
		dc.b	2
		dc.b	$F5					; 
		dc.b	$CD					; 
		dc.b	$5B					; [
		dc.b	2
		dc.b	$7D					; }
		dc.b	$C6					; 
		dc.b	$1C
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$CE					; 
		dc.b	0
		dc.b	$67					; g
		dc.b	$7E					; ~
		dc.b	$E6					; 
		dc.b	7
		dc.b	$E
		dc.b	1
		dc.b	$FE					; 
		dc.b	4
		dc.b	$38					; 8
		dc.b	$B
		dc.b	$C
		dc.b	$FE					; 
		dc.b	5
		dc.b	$38					; 8
		dc.b	6
		dc.b	$C
		dc.b	$FE					; 
		dc.b	7
		dc.b	$38					; 8
		dc.b	1
		dc.b	$C
		dc.b	$21					; !
		dc.b	7
		dc.b	0
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$CD					; 
		dc.b	$80					; 
		dc.b	2
		dc.b	$C6					; 
		dc.b	$4C					; L
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$56					; V
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$D
		dc.b	$28					; (
		dc.b	$42					; B
		dc.b	$D6					; 
		dc.b	4
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$D
		dc.b	$28					; (
		dc.b	$2B					; +
		dc.b	$D6					; 
		dc.b	4
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$D
		dc.b	$28					; (
		dc.b	$14
		dc.b	$D6					; 
		dc.b	4
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	0
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$DD					; 
		dc.b	$21					; !
		dc.b	0
		dc.b	$40					; @
		dc.b	$F1					; 
		dc.b	$C2					; 
		dc.b	$80					; 
		dc.b	1
		dc.b	$1E
		dc.b	$EF					; 
		dc.b	$3A					; :
		dc.b	3
		dc.b	0
		dc.b	$CB					; 
		dc.b	$77					; w
		dc.b	$28					; (
		dc.b	$14
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$2B					; +
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	1
		dc.b	0
		dc.b	$1C
		dc.b	$CB					; 
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$14
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$36					; 6
		dc.b	0
		dc.b	$28					; (
		dc.b	$DD					; 
		dc.b	$CB					; 
		dc.b	0
		dc.b	$7E					; ~
		dc.b	$20
		dc.b	$FA					; 
		dc.b	$DD					; 
		dc.b	$73					; s
		dc.b	1
		dc.b	$B7					; 
		dc.b	$20
		dc.b	$E5					; 
		dc.b	$3A					; :
		dc.b	$9D					; 
		dc.b	0
		dc.b	$3D					; =
		dc.b	$32					; 2
		dc.b	$9D					; 
		dc.b	0
		dc.b	$21					; !
		dc.b	$7D					; }
		dc.b	0
		dc.b	$22					; "
		dc.b	$96					; 
		dc.b	0
		dc.b	$C1					; 
		dc.b	$E1					; 
		dc.b	$C3					; 
		dc.b	$98					; 
		dc.b	0
		dc.b	$21					; !
		dc.b	$1C
		dc.b	0
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$7E					; ~
		dc.b	$6F					; o
		dc.b	$CB					; 
		dc.b	$25					; %
		dc.b	$CB					; 
		dc.b	$14
		dc.b	$CB					; 
		dc.b	$25					; %
		dc.b	$CB					; 
		dc.b	$14
		dc.b	$CB					; 
		dc.b	$25					; %
		dc.b	$CB					; 
		dc.b	$14
		dc.b	$CB					; 
		dc.b	$25					; %
		dc.b	$CB					; 
		dc.b	$14
		dc.b	$CB					; 
		dc.b	$25					; %
		dc.b	$CB					; 
		dc.b	$14
		dc.b	1
		dc.b	$48					; H
		dc.b	$C2					; 
		dc.b	9
		dc.b	$3E					; >
		dc.b	5
		dc.b	$C3					; 
		dc.b	$AA					; 
		dc.b	0
		dc.b	$7B					; {
		dc.b	$FE					; 
		dc.b	3
		dc.b	$D8					; 
		dc.b	$DD					; 
		dc.b	$23					; #
		dc.b	$DD					; 
		dc.b	$23					; #
		dc.b	$D6					; 
		dc.b	4
		dc.b	$C9					; 
		dc.b	$FF
off_1B01C:	dc.l	unk_2BFE8				; 11
		dc.l	0
		dc.l	PCM_CrowdGeneral			; 2
		dc.l	0
		dc.l	PCM_Clack				; 1
		dc.l	0
		dc.l	unk_2B430
		dc.l	0
		dc.l	unk_2A4AA				; 9
		dc.l	0
		dc.l	PCM_CrowdCheer				; 5
		dc.l	0
		dc.l	unk_29FFA				; 8
		dc.l	0
		dc.l	unk_2AEDA				; 10
		dc.l	0
		dc.l	0
		dc.l	0
		dc.l	0
		dc.l	0
		dc.l	unk_2AEDA				; 10 (again)
		dc.l	0
		dc.l	PCM_Whooaahh				; 4
		dc.l	0
		dc.l	PCM_CrowdBoo				; 3
		dc.l	0
		dc.l	PCM_Oooff				; 6
		dc.l	0
		dc.l	PCM_Punch				; 7
		dc.l	0
PCM_Clack:	dc.b	$80					; 		; DATA XREF: ROM:0001B02Co
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$D3					; 
		dc.b	1
		dc.b	1
		dc.b	$E8					; 
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$83					; 
		dc.b	$37					; 7
		dc.b	$84					; 
		dc.b	$A9					; 
		dc.b	$65					; e
		dc.b	$94					; 
		dc.b	$FC					; 
		dc.b	$9B					; 
		dc.b	$35					; 5
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$30					; 0
		dc.b	$75					; u
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$59					; Y
		dc.b	$95					; 
		dc.b	$E1					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$2E					; .
		dc.b	$4F					; O
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$60					; `
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$B0					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$D7					; 
		dc.b	$B
		dc.b	$41					; A
		dc.b	$FE					; 
		dc.b	$6C					; l
		dc.b	2
		dc.b	$FE					; 
		dc.b	$2E					; .
		dc.b	$BD					; 
		dc.b	$83					; 
		dc.b	$18
		dc.b	7
		dc.b	$FE					; 
		dc.b	$4C					; L
		dc.b	$46					; F
		dc.b	$9B					; 
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$A0					; 
		dc.b	$3A					; :
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$67					; g
		dc.b	$B4					; 
		dc.b	$B1					; 
		dc.b	$3A					; :
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$57					; W
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$29					; )
		dc.b	$2D					; -
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$DD					; 
		dc.b	$D4					; 
		dc.b	1
		dc.b	$9D					; 
		dc.b	$FE					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$BD					; 
		dc.b	$AB					; 
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$B0					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$53					; S
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$86					; 
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$B4					; 
		dc.b	$73					; s
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$A2					; 
		dc.b	$7F					; 
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$48					; H
		dc.b	$31					; 1
		dc.b	$9D					; 
		dc.b	$C1					; 
		dc.b	$7A					; z
		dc.b	1
		dc.b	$A6					; 
		dc.b	$C4					; 
		dc.b	$D6					; 
		dc.b	9
		dc.b	$35					; 5
		dc.b	$8D					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$52					; R
		dc.b	$57					; W
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$56					; V
		dc.b	$56					; V
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$7A					; z
		dc.b	$41					; A
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$B3					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$6A					; j
		dc.b	$49					; I
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$B4					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$8D					; 
		dc.b	$6F					; o
		dc.b	$E
		dc.b	1
		dc.b	$C8					; 
		dc.b	$DA					; 
PCM_CrowdGeneral: dc.b	$F2					; 		; DATA XREF: ROM:0001B024o
		dc.b	$D4					; 
		dc.b	$B0					; 
		dc.b	$55					; U
		dc.b	$3F					; ?
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$41					; A
		dc.b	$57					; W
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$2C					; ,
		dc.b	$5A					; Z
		dc.b	$C1					; 
		dc.b	$E7					; 
		dc.b	$C2					; 
		dc.b	$A2					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$53					; S
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$AB					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$4F					; O
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$86					; 
		dc.b	$AE					; 
		dc.b	$4F					; O
		dc.b	$15
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$B1					; 
		dc.b	$BE					; 
		dc.b	$96					; 
		dc.b	$37					; 7
		dc.b	1
		dc.b	$98					; 
		dc.b	$FE					; 
		dc.b	$BB					; 
		dc.b	$60					; `
		dc.b	$12
		dc.b	1
		dc.b	$48					; H
		dc.b	$B0					; 
		dc.b	$C9					; 
		dc.b	$D9					; 
		dc.b	$A9					; 
		dc.b	$5A					; Z
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$37					; 7
		dc.b	$6B					; k
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$8A					; 
		dc.b	$42					; B
		dc.b	$75					; u
		dc.b	$9C					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$83					; 
		dc.b	$94					; 
		dc.b	$A7					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$48					; H
		dc.b	$3B					; ;
		dc.b	$9F					; 
		dc.b	$C9					; 
		dc.b	$78					; x
		dc.b	$49					; I
		dc.b	$57					; W
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$B4					; 
		dc.b	$BE					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$67					; g
		dc.b	$34					; 4
		dc.b	$80					; 
		dc.b	$B7					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$94					; 
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$B0					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$72					; r
		dc.b	$22					; "
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$50					; P
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$B7					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$57					; W
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$AA					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$4A					; J
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$C1					; 
		dc.b	$C6					; 
		dc.b	$72					; r
		dc.b	$D2					; 
		dc.b	$FD					; 
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$D2					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$C1					; 
		dc.b	$A7					; 
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$56					; V
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$42					; B
		dc.b	$71					; q
		dc.b	$B1					; 
		dc.b	$EF					; 
		dc.b	$5A					; Z
		dc.b	$94					; 
		dc.b	$DA					; 
		dc.b	$46					; F
		dc.b	$56					; V
		dc.b	$17
		dc.b	1
		dc.b	$65					; e
		dc.b	$F6					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$64					; d
		dc.b	$2D					; -
		dc.b	$41					; A
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$BD					; 
		dc.b	$AD					; 
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$B7					; 
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$59					; Y
		dc.b	$B
		dc.b	$41					; A
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$B4					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$49					; I
		dc.b	$73					; s
		dc.b	$AE					; 
		dc.b	$6E					; n
		dc.b	$15
		dc.b	$1C
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$A7					; 
		dc.b	$96					; 
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$3B					; ;
		dc.b	$59					; Y
		dc.b	$A7					; 
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$BA					; 
		dc.b	$76					; v
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$98					; 
		dc.b	$5E					; ^
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$59					; Y
		dc.b	$94					; 
		dc.b	$BB					; 
		dc.b	$29					; )
		dc.b	$1F
		dc.b	$61					; a
		dc.b	$88					; 
		dc.b	$BD					; 
		dc.b	$8F					; 
		dc.b	$63					; c
		dc.b	$C2					; 
		dc.b	$A7					; 
		dc.b	$3C					; <
		dc.b	$33					; 3
		dc.b	$5A					; Z
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$A4					; 
		dc.b	$E0					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$61					; a
		dc.b	$3C					; <
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$A7					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$AD					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$6A					; j
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$BD					; 
		dc.b	$8D					; 
		dc.b	$4C					; L
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$8F					; 
		dc.b	$A6					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$56					; V
		dc.b	$4C					; L
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$8D					; 
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$64					; d
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$83					; 
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$87					; 
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$A0					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$B3					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$73					; s
		dc.b	$A6					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$60					; `
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$AB					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$5C					; \
		dc.b	$42					; B
		dc.b	$6A					; j
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$5D					; ]
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$9F					; 
		dc.b	$C1					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$6E					; n
		dc.b	$46					; F
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$AA					; 
		dc.b	$6C					; l
		dc.b	$45					; E
		dc.b	$46					; F
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$BD					; 
		dc.b	$A9					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$50					; P
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$5D					; ]
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$56					; V
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$48					; H
		dc.b	$48					; H
		dc.b	$87					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$B1					; 
		dc.b	$C5					; 
		dc.b	$8A					; 
		dc.b	$49					; I
		dc.b	$2E					; .
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$A9					; 
		dc.b	$B3					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$73					; s
		dc.b	$B8					; 
		dc.b	$A6					; 
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$4C					; L
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$9F					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$91					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$8F					; 
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$A6					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$4A					; J
		dc.b	$7A					; z
		dc.b	$B3					; 
		dc.b	$95					; 
		dc.b	$6A					; j
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$9F					; 
		dc.b	$79					; y
		dc.b	$10
		dc.b	$99					; 
		dc.b	$FE					; 
		dc.b	$BA					; 
		dc.b	$E
		dc.b	$27					; '
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$4A					; J
		dc.b	$86					; 
		dc.b	$BF					; 
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$9F					; 
		dc.b	$8D					; 
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$61					; a
		dc.b	$8E					; 
		dc.b	$B3					; 
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$BF					; 
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$56					; V
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$7D					; }
		dc.b	$50					; P
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$A9					; 
		dc.b	$AA					; 
		dc.b	$7D					; }
		dc.b	$63					; c
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$B4					; 
		dc.b	$A4					; 
		dc.b	$60					; `
		dc.b	$48					; H
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$AB					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$3F					; ?
		dc.b	$4A					; J
		dc.b	$6B					; k
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$7F					; 
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$D4					; 
		dc.b	$3E					; >
		dc.b	1
		dc.b	$C2					; 
		dc.b	$FE					; 
		dc.b	$E2					; 
		dc.b	$4F					; O
		dc.b	$1B
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$48					; H
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$6F					; o
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$48					; H
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$92					; 
		dc.b	$9F					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$A4					; 
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$AD					; 
		dc.b	$84					; 
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$A7					; 
		dc.b	$C8					; 
		dc.b	$A4					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$63					; c
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$5C					; \
		dc.b	$42					; B
		dc.b	$5E					; ^
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$87					; 
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$9D					; 
		dc.b	$AD					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$46					; F
		dc.b	$5E					; ^
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$63					; c
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$81					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$64					; d
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
PCM_CrowdBoo:	dc.b	$80					; 		; DATA XREF: ROM:0001B07Co
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7C					; |
unk_1C000:	dc.b	$80					; 		; DATA XREF: sub_FE98A+88o
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$89					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$70					; p
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$91					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$96					; 
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$97					; 
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$69					; i
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$5C					; \
		dc.b	$5B					; [
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$A1					; 
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$97					; 
		dc.b	$AD					; 
		dc.b	$A1					; 
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$5C					; \
		dc.b	$4A					; J
		dc.b	$58					; X
		dc.b	$5A					; Z
		dc.b	$70					; p
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$9F					; 
		dc.b	$B8					; 
		dc.b	$8D					; 
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$9C					; 
		dc.b	$BA					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$7B					; {
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$A4					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$66					; f
		dc.b	$80					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$47					; G
		dc.b	$54					; T
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$85					; 
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$6C					; l
		dc.b	$44					; D
		dc.b	$69					; i
		dc.b	$8E					; 
		dc.b	$AC					; 
		dc.b	$9C					; 
		dc.b	$69					; i
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$5A					; Z
		dc.b	$3E					; >
		dc.b	$63					; c
		dc.b	$90					; 
		dc.b	$A9					; 
		dc.b	$BB					; 
		dc.b	$A6					; 
		dc.b	$93					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$4B					; K
		dc.b	$7B					; {
		dc.b	$BC					; 
		dc.b	$AD					; 
		dc.b	$71					; q
		dc.b	$54					; T
		dc.b	$6C					; l
		dc.b	$98					; 
		dc.b	$90					; 
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$61					; a
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$8C					; 
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$70					; p
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$69					; i
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$6E					; n
		dc.b	$66					; f
		dc.b	$8A					; 
		dc.b	$AF					; 
		dc.b	$BD					; 
		dc.b	$AD					; 
		dc.b	$70					; p
		dc.b	$3F					; ?
		dc.b	$59					; Y
		dc.b	$8E					; 
		dc.b	$A7					; 
		dc.b	$94					; 
		dc.b	$6E					; n
		dc.b	$70					; p
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$8C					; 
		dc.b	$61					; a
		dc.b	$3A					; :
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$85					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$8A					; 
		dc.b	$B2					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$BA					; 
		dc.b	$CA					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$77					; w
		dc.b	$C5					; 
		dc.b	$E8					; 
		dc.b	$B7					; 
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$41					; A
		dc.b	$57					; W
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$90					; 
		dc.b	$B8					; 
		dc.b	$93					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$70					; p
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$5B					; [
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$90					; 
		dc.b	$91					; 
		dc.b	$82					; 
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$90					; 
		dc.b	$75					; u
		dc.b	$52					; R
		dc.b	$61					; a
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$64					; d
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$70					; p
		dc.b	$82					; 
		dc.b	$A0					; 
		dc.b	$A5					; 
		dc.b	$92					; 
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$97					; 
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$74					; t
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$6D					; m
		dc.b	$7F					; 
		dc.b	$9A					; 
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$5B					; [
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$70					; p
		dc.b	$88					; 
		dc.b	$9E					; 
		dc.b	$A8					; 
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$64					; d
		dc.b	$5B					; [
		dc.b	$47					; G
		dc.b	$71					; q
		dc.b	$BB					; 
		dc.b	$BC					; 
		dc.b	$99					; 
		dc.b	$72					; r
		dc.b	$54					; T
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$AE					; 
		dc.b	$B0					; 
		dc.b	$7E					; ~
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$90					; 
		dc.b	$94					; 
		dc.b	$7E					; ~
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$85					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$5F					; _
		dc.b	$7B					; {
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$6E					; n
		dc.b	$8F					; 
		dc.b	$B2					; 
		dc.b	$BF					; 
		dc.b	$B0					; 
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$4D					; M
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$53					; S
		dc.b	$3C					; <
		dc.b	$5C					; \
		dc.b	$90					; 
		dc.b	$A5					; 
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$8C					; 
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$90					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$89					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$97					; 
		dc.b	$6E					; n
		dc.b	$7E					; ~
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$9E					; 
		dc.b	$A6					; 
		dc.b	$93					; 
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$5B					; [
		dc.b	$3D					; =
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$CA					; 
		dc.b	$DB					; 
		dc.b	$A9					; 
		dc.b	$90					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$97					; 
		dc.b	$C3					; 
		dc.b	$B4					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$61					; a
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$7B					; {
		dc.b	$5F					; _
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$92					; 
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$5D					; ]
		dc.b	$45					; E
		dc.b	$54					; T
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$BF					; 
		dc.b	$96					; 
		dc.b	$59					; Y
		dc.b	$39					; 9
		dc.b	$48					; H
		dc.b	$62					; b
		dc.b	$8D					; 
		dc.b	$AC					; 
		dc.b	$94					; 
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$B5					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$77					; w
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$75					; u
		dc.b	$56					; V
		dc.b	$58					; X
		dc.b	$83					; 
		dc.b	$B0					; 
		dc.b	$AC					; 
		dc.b	$76					; v
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$72					; r
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$A1					; 
		dc.b	$93					; 
		dc.b	$5B					; [
		dc.b	$40					; @
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$A0					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$7F					; 
		dc.b	$AD					; 
		dc.b	$C8					; 
		dc.b	$B7					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$6D					; m
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$40					; @
		dc.b	$50					; P
		dc.b	$7D					; }
		dc.b	$89					; 
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$5B					; [
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$9A					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$C2					; 
		dc.b	$D1					; 
		dc.b	$B5					; 
		dc.b	$B6					; 
		dc.b	$B3					; 
		dc.b	$9A					; 
		dc.b	$79					; y
		dc.b	$4B					; K
		dc.b	$3C					; <
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$85					; 
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$9E					; 
		dc.b	$68					; h
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$85					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$7E					; ~
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$62					; b
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$65					; e
		dc.b	$37					; 7
		dc.b	$3C					; <
		dc.b	$5E					; ^
		dc.b	$91					; 
		dc.b	$C4					; 
		dc.b	$C9					; 
		dc.b	$AA					; 
		dc.b	$9A					; 
		dc.b	$9D					; 
		dc.b	$92					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$54					; T
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$AD					; 
		dc.b	$A9					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$71					; q
		dc.b	$66					; f
		dc.b	$95					; 
		dc.b	$B5					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$85					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$A8					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$93					; 
		dc.b	$7E					; ~
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$5F					; _
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$5F					; _
		dc.b	$66					; f
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$6C					; l
		dc.b	$58					; X
		dc.b	$6D					; m
		dc.b	$75					; u
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$A1					; 
		dc.b	$80					; 
		dc.b	$6D					; m
		dc.b	$66					; f
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$9C					; 
		dc.b	$AB					; 
		dc.b	$A4					; 
		dc.b	$9A					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$4E					; N
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$93					; 
		dc.b	$C1					; 
		dc.b	$D2					; 
		dc.b	$BF					; 
		dc.b	$BB					; 
		dc.b	$8D					; 
		dc.b	$3A					; :
		dc.b	$2C					; ,
		dc.b	$43					; C
		dc.b	$5D					; ]
		dc.b	$AA					; 
		dc.b	$DE					; 
		dc.b	$CB					; 
		dc.b	$AF					; 
		dc.b	$74					; t
		dc.b	$30					; 0
		dc.b	$2B					; +
		dc.b	$51					; Q
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$70					; p
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$95					; 
		dc.b	$BF					; 
		dc.b	$CE					; 
		dc.b	$B0					; 
		dc.b	$6B					; k
		dc.b	$45					; E
		dc.b	$49					; I
		dc.b	$3E					; >
		dc.b	$57					; W
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$AB					; 
		dc.b	$A3					; 
		dc.b	$A0					; 
		dc.b	$64					; d
		dc.b	$41					; A
		dc.b	$65					; e
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$C2					; 
		dc.b	$C9					; 
		dc.b	$A2					; 
		dc.b	$77					; w
		dc.b	$54					; T
		dc.b	$3E					; >
		dc.b	$48					; H
		dc.b	$7C					; |
		dc.b	$A9					; 
		dc.b	$AC					; 
		dc.b	$A6					; 
		dc.b	$72					; r
		dc.b	$4B					; K
		dc.b	$8C					; 
		dc.b	$BC					; 
		dc.b	$A0					; 
		dc.b	$77					; w
		dc.b	$53					; S
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$6D					; m
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$66					; f
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$9E					; 
		dc.b	$93					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$50					; P
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$AA					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$6A					; j
		dc.b	$48					; H
		dc.b	$69					; i
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$9E					; 
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$70					; p
		dc.b	$55					; U
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$A8					; 
		dc.b	$C1					; 
		dc.b	$B6					; 
		dc.b	$8D					; 
		dc.b	$60					; `
		dc.b	$32					; 2
		dc.b	$2C					; ,
		dc.b	$63					; c
		dc.b	$99					; 
		dc.b	$AF					; 
		dc.b	$C1					; 
		dc.b	$AE					; 
		dc.b	$7C					; |
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$4D					; M
		dc.b	$54					; T
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$B9					; 
		dc.b	$98					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$62					; b
		dc.b	$76					; v
		dc.b	$8C					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$A8					; 
		dc.b	$84					; 
		dc.b	$5A					; Z
		dc.b	$5F					; _
		dc.b	$70					; p
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$92					; 
		dc.b	$97					; 
		dc.b	$85					; 
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$70					; p
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$89					; 
		dc.b	$AA					; 
		dc.b	$C1					; 
		dc.b	$BF					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$6A					; j
		dc.b	$4E					; N
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$90					; 
		dc.b	$9D					; 
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$A1					; 
		dc.b	$AF					; 
		dc.b	$8C					; 
		dc.b	$5D					; ]
		dc.b	$45					; E
		dc.b	$3D					; =
		dc.b	$55					; U
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$B8					; 
		dc.b	$B3					; 
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$66					; f
		dc.b	$51					; Q
		dc.b	$65					; e
		dc.b	$8A					; 
		dc.b	$A1					; 
		dc.b	$B4					; 
		dc.b	$B5					; 
		dc.b	$A8					; 
		dc.b	$89					; 
		dc.b	$66					; f
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$B1					; 
		dc.b	$C3					; 
		dc.b	$A3					; 
		dc.b	$6B					; k
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$6D					; m
		dc.b	$7B					; {
		dc.b	$90					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$7E					; ~
		dc.b	$A1					; 
		dc.b	$A1					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$58					; X
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$9E					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$AB					; 
		dc.b	$84					; 
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$85					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$A7					; 
		dc.b	$A5					; 
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$82					; 
		dc.b	$4E					; N
		dc.b	$47					; G
		dc.b	$49					; I
		dc.b	$7B					; {
		dc.b	$B6					; 
		dc.b	$95					; 
		dc.b	$85					; 
		dc.b	$A5					; 
		dc.b	$A3					; 
		dc.b	$84					; 
		dc.b	$67					; g
		dc.b	$50					; P
		dc.b	$42					; B
		dc.b	$5E					; ^
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$97					; 
		dc.b	$B5					; 
		dc.b	$AC					; 
		dc.b	$85					; 
		dc.b	$6C					; l
		dc.b	$55					; U
		dc.b	$4B					; K
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$9E					; 
		dc.b	$BC					; 
		dc.b	$D3					; 
		dc.b	$B1					; 
		dc.b	$67					; g
		dc.b	$31					; 1
		dc.b	$28					; (
		dc.b	$47					; G
		dc.b	$87					; 
		dc.b	$B0					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$AB					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$58					; X
		dc.b	$5D					; ]
		dc.b	$5A					; Z
		dc.b	$3A					; :
		dc.b	$4A					; J
		dc.b	$84					; 
		dc.b	$B7					; 
		dc.b	$BE					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$9A					; 
		dc.b	$9B					; 
		dc.b	$64					; d
		dc.b	$51					; Q
		dc.b	$69					; i
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$A6					; 
		dc.b	$B2					; 
		dc.b	$77					; w
		dc.b	$4D					; M
		dc.b	$66					; f
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$62					; b
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$9A					; 
		dc.b	$A7					; 
		dc.b	$A5					; 
		dc.b	$79					; y
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$53					; S
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$B0					; 
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$A1					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$64					; d
		dc.b	$3B					; ;
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$62					; b
		dc.b	$82					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$B5					; 
		dc.b	$96					; 
		dc.b	$4E					; N
		dc.b	$61					; a
		dc.b	$A6					; 
		dc.b	$9C					; 
		dc.b	$85					; 
		dc.b	$94					; 
		dc.b	$7C					; |
		dc.b	$69					; i
		dc.b	$8C					; 
		dc.b	$A1					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$B2					; 
		dc.b	$B7					; 
		dc.b	$80					; 
		dc.b	$58					; X
		dc.b	$3D					; =
		dc.b	$34					; 4
		dc.b	$5E					; ^
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$85					; 
		dc.b	$A4					; 
		dc.b	$B5					; 
		dc.b	$90					; 
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$69					; i
		dc.b	$4F					; O
		dc.b	$5F					; _
		dc.b	$88					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$9A					; 
		dc.b	$8C					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$69					; i
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$81					; 
		dc.b	$B9					; 
		dc.b	$C4					; 
		dc.b	$B0					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$44					; D
		dc.b	$46					; F
		dc.b	$89					; 
		dc.b	$B1					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$4E					; N
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$54					; T
		dc.b	$86					; 
		dc.b	$C5					; 
		dc.b	$BA					; 
		dc.b	$B8					; 
		dc.b	$8D					; 
		dc.b	$41					; A
		dc.b	$50					; P
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$6E					; n
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$A6					; 
		dc.b	$AF					; 
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$87					; 
		dc.b	$4C					; L
		dc.b	$3D					; =
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$9A					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$5B					; [
		dc.b	$49					; I
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$A4					; 
		dc.b	$C4					; 
		dc.b	$BE					; 
		dc.b	$CA					; 
		dc.b	$A4					; 
		dc.b	$58					; X
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	$AD					; 
		dc.b	$C9					; 
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$52					; R
		dc.b	$44					; D
		dc.b	$3C					; <
		dc.b	$65					; e
		dc.b	$92					; 
		dc.b	$9E					; 
		dc.b	$9F					; 
		dc.b	$8A					; 
		dc.b	$6D					; m
		dc.b	$7D					; }
		dc.b	$A5					; 
		dc.b	$AB					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$89					; 
		dc.b	$A7					; 
		dc.b	$94					; 
		dc.b	$65					; e
		dc.b	$30					; 0
		dc.b	$37					; 7
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$A9					; 
		dc.b	$BC					; 
		dc.b	$AE					; 
		dc.b	$81					; 
		dc.b	$57					; W
		dc.b	$5D					; ]
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$BB					; 
		dc.b	$D0					; 
		dc.b	$AC					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$4B					; K
		dc.b	$5F					; _
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$9C					; 
		dc.b	$BA					; 
		dc.b	$95					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$6D					; m
		dc.b	$77					; w
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$85					; 
		dc.b	$9E					; 
		dc.b	$AA					; 
		dc.b	$BE					; 
		dc.b	$B2					; 
		dc.b	$8B					; 
		dc.b	$69					; i
		dc.b	$46					; F
		dc.b	$2C					; ,
		dc.b	$3B					; ;
		dc.b	$74					; t
		dc.b	$B4					; 
		dc.b	$C0					; 
		dc.b	$97					; 
		dc.b	$8D					; 
		dc.b	$90					; 
		dc.b	$70					; p
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$77					; w
		dc.b	$5B					; [
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$5F					; _
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$B7					; 
		dc.b	$B8					; 
		dc.b	$B8					; 
		dc.b	$AF					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$75					; u
		dc.b	$49					; I
		dc.b	$3D					; =
		dc.b	$56					; V
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$A1					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$A6					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$3D					; =
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$D0					; 
		dc.b	$DD					; 
		dc.b	$BB					; 
		dc.b	$86					; 
		dc.b	$3F					; ?
		dc.b	$20
		dc.b	$3E					; >
		dc.b	$6D					; m
		dc.b	$9B					; 
		dc.b	$D1					; 
		dc.b	$E0					; 
		dc.b	$A2					; 
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$AB					; 
		dc.b	$D4					; 
		dc.b	$C8					; 
		dc.b	$B1					; 
		dc.b	$97					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$54					; T
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$97					; 
		dc.b	$A3					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$89					; 
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$62					; b
		dc.b	$5E					; ^
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$57					; W
		dc.b	$4B					; K
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$93					; 
		dc.b	$9A					; 
		dc.b	$84					; 
		dc.b	$4E					; N
		dc.b	$39					; 9
		dc.b	$3A					; :
		dc.b	$47					; G
		dc.b	$7A					; z
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$AF					; 
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$69					; i
		dc.b	$56					; V
		dc.b	$58					; X
		dc.b	$7D					; }
		dc.b	$A9					; 
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$70					; p
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$C7					; 
		dc.b	$A6					; 
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$49					; I
		dc.b	$56					; V
		dc.b	$80					; 
		dc.b	$B3					; 
		dc.b	$D3					; 
		dc.b	$C9					; 
		dc.b	$AA					; 
		dc.b	$72					; r
		dc.b	$39					; 9
		dc.b	$4D					; M
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$95					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$93					; 
		dc.b	$54					; T
		dc.b	$35					; 5
		dc.b	$5B					; [
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$B3					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$93					; 
		dc.b	$B3					; 
		dc.b	$92					; 
		dc.b	$68					; h
		dc.b	$5B					; [
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$74					; t
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$7B					; {
		dc.b	$92					; 
		dc.b	$AF					; 
		dc.b	$A2					; 
		dc.b	$76					; v
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$97					; 
		dc.b	$7E					; ~
		dc.b	$90					; 
		dc.b	$A7					; 
		dc.b	$7C					; |
		dc.b	$39					; 9
		dc.b	$33					; 3
		dc.b	$5F					; _
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$9B					; 
		dc.b	$95					; 
		dc.b	$5C					; \
		dc.b	$52					; R
		dc.b	$51					; Q
		dc.b	$4A					; J
		dc.b	$8D					; 
		dc.b	$C4					; 
		dc.b	$B5					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$37					; 7
		dc.b	$69					; i
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$DF					; 
		dc.b	$BB					; 
		dc.b	$70					; p
		dc.b	$45					; E
		dc.b	$3F					; ?
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$9B					; 
		dc.b	$CD					; 
		dc.b	$C2					; 
		dc.b	$B6					; 
		dc.b	$87					; 
		dc.b	$44					; D
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$B5					; 
		dc.b	$B5					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$69					; i
		dc.b	$55					; U
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$98					; 
		dc.b	$90					; 
		dc.b	$7F					; 
		dc.b	$93					; 
		dc.b	$7E					; ~
		dc.b	$69					; i
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$83					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$AA					; 
		dc.b	$98					; 
		dc.b	$6C					; l
		dc.b	$39					; 9
		dc.b	$22					; "
		dc.b	$2A					; *
		dc.b	$4D					; M
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$B5					; 
		dc.b	$B6					; 
		dc.b	$99					; 
		dc.b	$7B					; {
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$74					; t
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$5F					; _
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$A0					; 
		dc.b	$A8					; 
		dc.b	$9B					; 
		dc.b	$82					; 
		dc.b	$61					; a
		dc.b	$45					; E
		dc.b	$4A					; J
		dc.b	$74					; t
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$A3					; 
		dc.b	$BD					; 
		dc.b	$9E					; 
		dc.b	$6B					; k
		dc.b	$58					; X
		dc.b	$68					; h
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$6D					; m
		dc.b	$57					; W
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$A3					; 
		dc.b	$A7					; 
		dc.b	$97					; 
		dc.b	$9A					; 
		dc.b	$75					; u
		dc.b	$51					; Q
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$B5					; 
		dc.b	$78					; x
		dc.b	$3D					; =
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$5F					; _
		dc.b	$93					; 
		dc.b	$C4					; 
		dc.b	$C8					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$54					; T
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$8D					; 
		dc.b	$4A					; J
		dc.b	$48					; H
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$AB					; 
		dc.b	$AF					; 
		dc.b	$9D					; 
		dc.b	$61					; a
		dc.b	$49					; I
		dc.b	$75					; u
		dc.b	$9C					; 
		dc.b	$B5					; 
		dc.b	$A8					; 
		dc.b	$74					; t
		dc.b	$64					; d
		dc.b	$70					; p
		dc.b	$78					; x
		dc.b	$89					; 
		dc.b	$70					; p
		dc.b	$64					; d
		dc.b	$85					; 
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$6D					; m
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$5F					; _
		dc.b	$33					; 3
		dc.b	$50					; P
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$A1					; 
		dc.b	$6E					; n
		dc.b	$58					; X
		dc.b	$82					; 
		dc.b	$AD					; 
		dc.b	$9C					; 
		dc.b	$74					; t
		dc.b	$59					; Y
		dc.b	$51					; Q
		dc.b	$72					; r
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$85					; 
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$75					; u
		dc.b	$62					; b
		dc.b	$9A					; 
		dc.b	$B5					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$91					; 
		dc.b	$BB					; 
		dc.b	$B5					; 
		dc.b	$88					; 
		dc.b	$65					; e
		dc.b	$51					; Q
		dc.b	$4D					; M
		dc.b	$74					; t
		dc.b	$A5					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$AF					; 
		dc.b	$D3					; 
		dc.b	$C5					; 
		dc.b	$77					; w
		dc.b	$2E					; .
		dc.b	$20
		dc.b	$47					; G
		dc.b	$76					; v
		dc.b	$B8					; 
		dc.b	$D6					; 
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$93					; 
		dc.b	$97					; 
		dc.b	$89					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$5B					; [
		dc.b	$4F					; O
		dc.b	$4E					; N
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$88					; 
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$8C					; 
		dc.b	$6F					; o
		dc.b	$4A					; J
		dc.b	$5C					; \
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$7D					; }
		dc.b	$59					; Y
		dc.b	$30					; 0
		dc.b	$3B					; ;
		dc.b	$6E					; n
		dc.b	$93					; 
		dc.b	$C9					; 
		dc.b	$F1					; 
		dc.b	$CC					; 
		dc.b	$8C					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$70					; p
		dc.b	$84					; 
		dc.b	$5F					; _
		dc.b	$56					; V
		dc.b	$76					; v
		dc.b	$9D					; 
		dc.b	$8C					; 
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$78					; x
		dc.b	$89					; 
		dc.b	$A3					; 
		dc.b	$7E					; ~
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$79					; y
		dc.b	$A9					; 
		dc.b	$B4					; 
		dc.b	$90					; 
		dc.b	$4B					; K
		dc.b	$3F					; ?
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$BD					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$90					; 
		dc.b	$93					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$69					; i
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$A5					; 
		dc.b	$A4					; 
		dc.b	$93					; 
		dc.b	$53					; S
		dc.b	$2F					; /
		dc.b	$5C					; \
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$93					; 
		dc.b	$77					; w
		dc.b	$85					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$6D					; m
		dc.b	$9E					; 
		dc.b	$94					; 
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$89					; 
		dc.b	$96					; 
		dc.b	$97					; 
		dc.b	$74					; t
		dc.b	$5B					; [
		dc.b	$63					; c
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$A8					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$74					; t
		dc.b	$A0					; 
		dc.b	$C3					; 
		dc.b	$BC					; 
		dc.b	$6E					; n
		dc.b	$3F					; ?
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$B3					; 
		dc.b	$93					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$66					; f
		dc.b	$5A					; Z
		dc.b	$56					; V
		dc.b	$80					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$CB					; 
		dc.b	$DA					; 
		dc.b	$93					; 
		dc.b	$4F					; O
		dc.b	$3D					; =
		dc.b	$4C					; L
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$69					; i
		dc.b	$7D					; }
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$82					; 
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$49					; I
		dc.b	$43					; C
		dc.b	$65					; e
		dc.b	$B4					; 
		dc.b	$C2					; 
		dc.b	$A6					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$89					; 
		dc.b	$7A					; z
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$7B					; {
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$5F					; _
		dc.b	$6D					; m
		dc.b	$9F					; 
		dc.b	$BC					; 
		dc.b	$D5					; 
		dc.b	$BA					; 
		dc.b	$67					; g
		dc.b	$54					; T
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$A6					; 
		dc.b	$D1					; 
		dc.b	$D3					; 
		dc.b	$93					; 
		dc.b	$3D					; =
		dc.b	$1B
		dc.b	$33					; 3
		dc.b	$66					; f
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$D2					; 
		dc.b	$C7					; 
		dc.b	$9F					; 
		dc.b	$53					; S
		dc.b	$35					; 5
		dc.b	$59					; Y
		dc.b	$7B					; {
		dc.b	$8F					; 
		dc.b	$9E					; 
		dc.b	$91					; 
		dc.b	$6F					; o
		dc.b	$58					; X
		dc.b	$3F					; ?
		dc.b	$24					; $
		dc.b	$2F					; /
		dc.b	$5D					; ]
		dc.b	$9D					; 
		dc.b	$D1					; 
		dc.b	$D6					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$74					; t
		dc.b	$54					; T
		dc.b	$4C					; L
		dc.b	$54					; T
		dc.b	$3E					; >
		dc.b	$42					; B
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$CA					; 
		dc.b	$D7					; 
		dc.b	$AE					; 
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$78					; x
		dc.b	$C3					; 
		dc.b	$F4					; 
		dc.b	$E3					; 
		dc.b	$A6					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$4E					; N
		dc.b	$69					; i
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$BB					; 
		dc.b	$A5					; 
		dc.b	$98					; 
		dc.b	$9E					; 
		dc.b	$87					; 
		dc.b	$51					; Q
		dc.b	$3B					; ;
		dc.b	$5B					; [
		dc.b	$9B					; 
		dc.b	$D4					; 
		dc.b	$D2					; 
		dc.b	$90					; 
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$58					; X
		dc.b	$63					; c
		dc.b	$49					; I
		dc.b	$64					; d
		dc.b	$AF					; 
		dc.b	$C0					; 
		dc.b	$B8					; 
		dc.b	$AA					; 
		dc.b	$73					; s
		dc.b	$4D					; M
		dc.b	$56					; V
		dc.b	$66					; f
		dc.b	$8A					; 
		dc.b	$B2					; 
		dc.b	$A5					; 
		dc.b	$80					; 
		dc.b	$69					; i
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$AB					; 
		dc.b	$89					; 
		dc.b	$72					; r
		dc.b	$59					; Y
		dc.b	$31					; 1
		dc.b	$35					; 5
		dc.b	$4D					; M
		dc.b	$7B					; {
		dc.b	$B5					; 
		dc.b	$C3					; 
		dc.b	$C0					; 
		dc.b	$92					; 
		dc.b	$4A					; J
		dc.b	$3D					; =
		dc.b	$5E					; ^
		dc.b	$85					; 
		dc.b	$97					; 
		dc.b	$9F					; 
		dc.b	$9A					; 
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$5F					; _
		dc.b	$58					; X
		dc.b	$72					; r
		dc.b	$C2					; 
		dc.b	$D1					; 
		dc.b	$A5					; 
		dc.b	$76					; v
		dc.b	$62					; b
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$4B					; K
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$A5					; 
		dc.b	$B2					; 
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$6F					; o
		dc.b	$66					; f
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$66					; f
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$A1					; 
		dc.b	$AA					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$89					; 
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$8C					; 
		dc.b	$B9					; 
		dc.b	$CB					; 
		dc.b	$BD					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$42					; B
		dc.b	$5B					; [
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$47					; G
		dc.b	$4F					; O
		dc.b	$6F					; o
		dc.b	$9B					; 
		dc.b	$B6					; 
		dc.b	$AA					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$A6					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$4C					; L
		dc.b	$44					; D
		dc.b	$4C					; L
		dc.b	$85					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$A7					; 
		dc.b	$78					; x
		dc.b	$3D					; =
		dc.b	$37					; 7
		dc.b	$74					; t
		dc.b	$B5					; 
		dc.b	$BD					; 
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$68					; h
		dc.b	$26					; &
		dc.b	$1D
		dc.b	$4D					; M
		dc.b	$91					; 
		dc.b	$CE					; 
		dc.b	$CE					; 
		dc.b	$91					; 
		dc.b	$59					; Y
		dc.b	$3C					; <
		dc.b	$37					; 7
		dc.b	$6B					; k
		dc.b	$94					; 
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$A7					; 
		dc.b	$A1					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$50					; P
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$C4					; 
		dc.b	$B3					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$54					; T
		dc.b	$4C					; L
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$BE					; 
		dc.b	$B8					; 
		dc.b	$8C					; 
		dc.b	$6D					; m
		dc.b	$59					; Y
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$54					; T
		dc.b	$7F					; 
		dc.b	$AE					; 
		dc.b	$C0					; 
		dc.b	$B5					; 
		dc.b	$94					; 
		dc.b	$56					; V
		dc.b	$35					; 5
		dc.b	$4D					; M
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$6D					; m
		dc.b	$77					; w
		dc.b	$B4					; 
		dc.b	$DA					; 
		dc.b	$B6					; 
		dc.b	$81					; 
		dc.b	$56					; V
		dc.b	$31					; 1
		dc.b	$3D					; =
		dc.b	$70					; p
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$A5					; 
		dc.b	$9A					; 
		dc.b	$97					; 
		dc.b	$97					; 
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$58					; X
		dc.b	$46					; F
		dc.b	$58					; X
		dc.b	$74					; t
		dc.b	$AF					; 
		dc.b	$CD					; 
		dc.b	$A7					; 
		dc.b	$89					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$53					; S
		dc.b	$7D					; }
		dc.b	$B7					; 
		dc.b	$CA					; 
		dc.b	$B3					; 
		dc.b	$82					; 
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$42					; B
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$B2					; 
		dc.b	$BB					; 
		dc.b	$AA					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$70					; p
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$7E					; ~
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$58					; X
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$A4					; 
		dc.b	$C8					; 
		dc.b	$B9					; 
		dc.b	$84					; 
		dc.b	$5F					; _
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$9F					; 
		dc.b	$D9					; 
		dc.b	$C1					; 
		dc.b	$A3					; 
		dc.b	$79					; y
		dc.b	$40					; @
		dc.b	$3F					; ?
		dc.b	$65					; e
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$9B					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$98					; 
		dc.b	$6F					; o
		dc.b	$42					; B
		dc.b	$3F					; ?
		dc.b	$57					; W
		dc.b	$9D					; 
		dc.b	$DB					; 
		dc.b	$E4					; 
		dc.b	$B5					; 
		dc.b	$7D					; }
		dc.b	$60					; `
		dc.b	$35					; 5
		dc.b	$1C
		dc.b	$39					; 9
		dc.b	$6F					; o
		dc.b	$AB					; 
		dc.b	$DC					; 
		dc.b	$EA					; 
		dc.b	$BF					; 
		dc.b	$73					; s
		dc.b	$3B					; ;
		dc.b	$38					; 8
		dc.b	$61					; a
		dc.b	$89					; 
		dc.b	$9A					; 
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$83					; 
		dc.b	$B8					; 
		dc.b	$C7					; 
		dc.b	$A0					; 
		dc.b	$7C					; |
		dc.b	$51					; Q
		dc.b	$46					; F
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$5F					; _
		dc.b	$96					; 
		dc.b	$BA					; 
		dc.b	$B8					; 
		dc.b	$BB					; 
		dc.b	$A0					; 
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$7C					; |
		dc.b	$41					; A
		dc.b	$30					; 0
		dc.b	$2D					; -
		dc.b	$57					; W
		dc.b	$8C					; 
		dc.b	$9E					; 
		dc.b	$B4					; 
		dc.b	$B6					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$5F					; _
		dc.b	$4A					; J
		dc.b	$63					; c
		dc.b	$90					; 
		dc.b	$A0					; 
		dc.b	$B3					; 
		dc.b	$AC					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$46					; F
		dc.b	$5D					; ]
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$9E					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$69					; i
		dc.b	$82					; 
		dc.b	$9F					; 
		dc.b	$C6					; 
		dc.b	$CA					; 
		dc.b	$94					; 
		dc.b	$59					; Y
		dc.b	$39					; 9
		dc.b	$46					; F
		dc.b	$6D					; m
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$A4					; 
		dc.b	$AE					; 
		dc.b	$B0					; 
		dc.b	$8C					; 
		dc.b	$59					; Y
		dc.b	$35					; 5
		dc.b	$36					; 6
		dc.b	$65					; e
		dc.b	$B6					; 
		dc.b	$C5					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$9A					; 
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$69					; i
		dc.b	$55					; U
		dc.b	$49					; I
		dc.b	$7B					; {
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$79					; y
		dc.b	$47					; G
		dc.b	$54					; T
		dc.b	$64					; d
		dc.b	$62					; b
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$C3					; 
		dc.b	$D6					; 
		dc.b	$9E					; 
		dc.b	$55					; U
		dc.b	$32					; 2
		dc.b	$37					; 7
		dc.b	$74					; t
		dc.b	$C9					; 
		dc.b	$E7					; 
		dc.b	$D9					; 
		dc.b	$9D					; 
		dc.b	$51					; Q
		dc.b	$48					; H
		dc.b	$62					; b
		dc.b	$8B					; 
		dc.b	$AF					; 
		dc.b	$A1					; 
		dc.b	$7E					; ~
		dc.b	$67					; g
		dc.b	$77					; w
		dc.b	$A8					; 
		dc.b	$AA					; 
		dc.b	$7A					; z
		dc.b	$4E					; N
		dc.b	$28					; (
		dc.b	$3D					; =
		dc.b	$82					; 
		dc.b	$A6					; 
		dc.b	$C8					; 
		dc.b	$DE					; 
		dc.b	$CA					; 
		dc.b	$A6					; 
		dc.b	$56					; V
		dc.b	$17
		dc.b	$3A					; :
		dc.b	$80					; 
		dc.b	$97					; 
		dc.b	$99					; 
		dc.b	$AB					; 
		dc.b	$86					; 
		dc.b	$52					; R
		dc.b	$6D					; m
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$89					; 
		dc.b	$68					; h
		dc.b	$4A					; J
		dc.b	$7C					; |
		dc.b	$B9					; 
		dc.b	$C9					; 
		dc.b	$B1					; 
		dc.b	$6A					; j
		dc.b	$28					; (
		dc.b	$3A					; :
		dc.b	$89					; 
		dc.b	$C3					; 
		dc.b	$CB					; 
		dc.b	$9C					; 
		dc.b	$4A					; J
		dc.b	$35					; 5
		dc.b	$55					; U
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$6A					; j
		dc.b	$82					; 
		dc.b	$97					; 
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$A1					; 
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$84					; 
		dc.b	$A1					; 
		dc.b	$AC					; 
		dc.b	$98					; 
		dc.b	$6D					; m
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$8F					; 
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$BA					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$70					; p
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$74					; t
		dc.b	$90					; 
		dc.b	$C2					; 
		dc.b	$C3					; 
		dc.b	$92					; 
		dc.b	$74					; t
		dc.b	$5D					; ]
		dc.b	$3C					; <
		dc.b	$41					; A
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$8C					; 
		dc.b	$B5					; 
		dc.b	$C7					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$66					; f
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$4E					; N
		dc.b	$5E					; ^
		dc.b	$98					; 
		dc.b	$C2					; 
		dc.b	$AF					; 
		dc.b	$85					; 
		dc.b	$70					; p
		dc.b	$8F					; 
		dc.b	$9E					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$9D					; 
		dc.b	$C1					; 
		dc.b	$C6					; 
		dc.b	$A4					; 
		dc.b	$54					; T
		dc.b	$2C					; ,
		dc.b	$40					; @
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$73					; s
		dc.b	$B5					; 
		dc.b	$E9					; 
		dc.b	$F2					; 
		dc.b	$D4					; 
		dc.b	$92					; 
		dc.b	$3D					; =
		dc.b	$E
		dc.b	$2E					; .
		dc.b	$5F					; _
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$C6					; 
		dc.b	$BD					; 
		dc.b	$B0					; 
		dc.b	$97					; 
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$61					; a
		dc.b	$58					; X
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$3D					; =
		dc.b	$75					; u
		dc.b	$AB					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$60					; `
		dc.b	$51					; Q
		dc.b	$88					; 
		dc.b	$9A					; 
		dc.b	$99					; 
		dc.b	$A8					; 
		dc.b	$90					; 
		dc.b	$6D					; m
		dc.b	$59					; Y
		dc.b	$4B					; K
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$6D					; m
		dc.b	$82					; 
		dc.b	$9A					; 
		dc.b	$9F					; 
		dc.b	$AF					; 
		dc.b	$BB					; 
		dc.b	$98					; 
		dc.b	$5A					; Z
		dc.b	$48					; H
		dc.b	$55					; U
		dc.b	$77					; w
		dc.b	$B6					; 
		dc.b	$C4					; 
		dc.b	$BC					; 
		dc.b	$B8					; 
		dc.b	$91					; 
		dc.b	$82					; 
		dc.b	$6D					; m
		dc.b	$33					; 3
		dc.b	$1F
		dc.b	$40					; @
		dc.b	$75					; u
		dc.b	$A3					; 
		dc.b	$C2					; 
		dc.b	$A7					; 
		dc.b	$69					; i
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$6D					; m
		dc.b	$63					; c
		dc.b	$58					; X
		dc.b	$4F					; O
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$58					; X
		dc.b	$77					; w
		dc.b	$93					; 
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$6C					; l
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$75					; u
		dc.b	$93					; 
		dc.b	$B9					; 
		dc.b	$BA					; 
		dc.b	$A5					; 
		dc.b	$B4					; 
		dc.b	$96					; 
		dc.b	$4E					; N
		dc.b	$29					; )
		dc.b	$1B
		dc.b	$2A					; *
		dc.b	$73					; s
		dc.b	$D0					; 
		dc.b	$F4					; 
		dc.b	$D8					; 
		dc.b	$9D					; 
		dc.b	$52					; R
		dc.b	$2E					; .
		dc.b	$3F					; ?
		dc.b	$59					; Y
		dc.b	$82					; 
		dc.b	$AB					; 
		dc.b	$9E					; 
		dc.b	$72					; r
		dc.b	$54					; T
		dc.b	$63					; c
		dc.b	$A0					; 
		dc.b	$B7					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$6D					; m
		dc.b	$44					; D
		dc.b	$51					; Q
		dc.b	$6A					; j
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$82					; 
		dc.b	$60					; `
		dc.b	$5B					; [
		dc.b	$66					; f
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$9A					; 
		dc.b	$B4					; 
		dc.b	$B2					; 
		dc.b	$80					; 
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$C8					; 
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$58					; X
		dc.b	$58					; X
		dc.b	$6B					; k
		dc.b	$6D					; m
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$A0					; 
		dc.b	$9D					; 
		dc.b	$82					; 
		dc.b	$6D					; m
		dc.b	$5B					; [
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$97					; 
		dc.b	$8B					; 
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$8F					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$74					; t
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$51					; Q
		dc.b	$74					; t
		dc.b	$9E					; 
		dc.b	$AA					; 
		dc.b	$A1					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$87					; 
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$56					; V
		dc.b	$4E					; N
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$B4					; 
		dc.b	$8E					; 
		dc.b	$65					; e
		dc.b	$5F					; _
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$A8					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$4E					; N
		dc.b	$7E					; ~
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$6C					; l
		dc.b	$90					; 
		dc.b	$CE					; 
		dc.b	$D1					; 
		dc.b	$7A					; z
		dc.b	$43					; C
		dc.b	$35					; 5
		dc.b	$55					; U
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$5B					; [
		dc.b	$39					; 9
		dc.b	$47					; G
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$90					; 
		dc.b	$B4					; 
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$56					; V
		dc.b	$98					; 
		dc.b	$D9					; 
		dc.b	$D4					; 
		dc.b	$B1					; 
		dc.b	$6E					; n
		dc.b	$2A					; *
		dc.b	$1A
		dc.b	$3B					; ;
		dc.b	$88					; 
		dc.b	$C7					; 
		dc.b	$B5					; 
		dc.b	$78					; x
		dc.b	$62					; b
		dc.b	$7E					; ~
		dc.b	$AB					; 
		dc.b	$BA					; 
		dc.b	$8E					; 
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$51					; Q
		dc.b	$74					; t
		dc.b	$91					; 
		dc.b	$89					; 
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$AE					; 
		dc.b	$A5					; 
		dc.b	$47					; G
		dc.b	$27					; '
		dc.b	$48					; H
		dc.b	$60					; `
		dc.b	$8D					; 
		dc.b	$B2					; 
		dc.b	$A8					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$A7					; 
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$2A					; *
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$97					; 
		dc.b	$98					; 
		dc.b	$9E					; 
		dc.b	$C5					; 
		dc.b	$D0					; 
		dc.b	$AA					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$56					; V
		dc.b	$5F					; _
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$C4					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$93					; 
		dc.b	$9A					; 
		dc.b	$8C					; 
		dc.b	$61					; a
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$9B					; 
		dc.b	$AA					; 
		dc.b	$B8					; 
		dc.b	$B7					; 
		dc.b	$6E					; n
		dc.b	$50					; P
		dc.b	$63					; c
		dc.b	$68					; h
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$70					; p
		dc.b	$A1					; 
		dc.b	$98					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$A7					; 
		dc.b	$B4					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$54					; T
		dc.b	$72					; r
		dc.b	$94					; 
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$70					; p
		dc.b	$62					; b
		dc.b	$5F					; _
		dc.b	$82					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$89					; 
		dc.b	$A0					; 
		dc.b	$7B					; {
		dc.b	$44					; D
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$A8					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$8F					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$77					; w
		dc.b	$3C					; <
		dc.b	$30					; 0
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$B4					; 
		dc.b	$D6					; 
		dc.b	$E3					; 
		dc.b	$C3					; 
		dc.b	$7B					; {
		dc.b	$45					; E
		dc.b	$38					; 8
		dc.b	$40					; @
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$6F					; o
		dc.b	$60					; `
		dc.b	$92					; 
		dc.b	$CC					; 
		dc.b	$DF					; 
		dc.b	$E3					; 
		dc.b	$CD					; 
		dc.b	$89					; 
		dc.b	$39					; 9
		dc.b	$F
		dc.b	$10
		dc.b	$3D					; =
		dc.b	$82					; 
		dc.b	$B8					; 
		dc.b	$D2					; 
		dc.b	$C3					; 
		dc.b	$97					; 
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$69					; i
		dc.b	$47					; G
		dc.b	$4D					; M
		dc.b	$54					; T
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$9D					; 
		dc.b	$B3					; 
		dc.b	$C4					; 
		dc.b	$9E					; 
		dc.b	$66					; f
		dc.b	$5C					; \
		dc.b	$70					; p
		dc.b	$77					; w
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$89					; 
		dc.b	$93					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$8C					; 
		dc.b	$71					; q
		dc.b	$5F					; _
		dc.b	$86					; 
		dc.b	$B2					; 
		dc.b	$B6					; 
		dc.b	$B2					; 
		dc.b	$9F					; 
		dc.b	$5F					; _
		dc.b	$23					; #
		dc.b	$1B
		dc.b	$3E					; >
		dc.b	$60					; `
		dc.b	$7B					; {
		dc.b	$B0					; 
		dc.b	$E8					; 
		dc.b	$EB					; 
		dc.b	$C6					; 
		dc.b	$97					; 
		dc.b	$67					; g
		dc.b	$40					; @
		dc.b	$2C					; ,
		dc.b	$28					; (
		dc.b	$48					; H
		dc.b	$8F					; 
		dc.b	$BC					; 
		dc.b	$CB					; 
		dc.b	$D0					; 
		dc.b	$C6					; 
		dc.b	$AD					; 
		dc.b	$7E					; ~
		dc.b	$4C					; L
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$60					; `
		dc.b	$93					; 
		dc.b	$A7					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$5F					; _
		dc.b	$72					; r
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$62					; b
		dc.b	$67					; g
		dc.b	$51					; Q
		dc.b	$59					; Y
		dc.b	$9E					; 
		dc.b	$B7					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$86					; 
		dc.b	$39					; 9
		dc.b	$19
		dc.b	$3E					; >
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$C2					; 
		dc.b	$C4					; 
		dc.b	$CF					; 
		dc.b	$83					; 
		dc.b	$2A					; *
		dc.b	$23					; #
		dc.b	$44					; D
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$D2					; 
		dc.b	$EB					; 
		dc.b	$CC					; 
		dc.b	$A8					; 
		dc.b	$85					; 
		dc.b	$57					; W
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$4B					; K
		dc.b	$54					; T
		dc.b	$8B					; 
		dc.b	$BC					; 
		dc.b	$A1					; 
		dc.b	$74					; t
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$79					; y
		dc.b	$46					; F
		dc.b	$3B					; ;
		dc.b	$63					; c
		dc.b	$98					; 
		dc.b	$A8					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$A5					; 
		dc.b	$A5					; 
		dc.b	$6E					; n
		dc.b	$3E					; >
		dc.b	$37					; 7
		dc.b	$55					; U
		dc.b	$8B					; 
		dc.b	$BE					; 
		dc.b	$C7					; 
		dc.b	$90					; 
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$A8					; 
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$66					; f
		dc.b	$6F					; o
		dc.b	$82					; 
		dc.b	$62					; b
		dc.b	$4A					; J
		dc.b	$56					; V
		dc.b	$62					; b
		dc.b	$81					; 
		dc.b	$AD					; 
		dc.b	$B3					; 
		dc.b	$A6					; 
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$B0					; 
		dc.b	$A4					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$6D					; m
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$61					; a
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$9B					; 
		dc.b	$A3					; 
		dc.b	$7E					; ~
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$8C					; 
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$58					; X
		dc.b	$72					; r
		dc.b	$97					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$4B					; K
		dc.b	$67					; g
		dc.b	$93					; 
		dc.b	$CF					; 
		dc.b	$D9					; 
		dc.b	$9E					; 
		dc.b	$61					; a
		dc.b	$4E					; N
		dc.b	$63					; c
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$CA					; 
		dc.b	$AA					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$58					; X
		dc.b	$74					; t
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$90					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$5C					; \
		dc.b	$1D
		dc.b	$2D					; -
		dc.b	$69					; i
		dc.b	$B5					; 
		dc.b	$DF					; 
		dc.b	$C5					; 
		dc.b	$85					; 
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$6D					; m
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$A3					; 
		dc.b	$85					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$82					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$8C					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$8E					; 
		dc.b	$93					; 
		dc.b	$6C					; l
		dc.b	$74					; t
		dc.b	$A2					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$AE					; 
		dc.b	$C5					; 
		dc.b	$A1					; 
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$6D					; m
		dc.b	$8C					; 
		dc.b	$6E					; n
		dc.b	$26					; &
		dc.b	$30					; 0
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$C8					; 
		dc.b	$D1					; 
		dc.b	$CC					; 
		dc.b	$BA					; 
		dc.b	$8B					; 
		dc.b	$48					; H
		dc.b	$22					; "
		dc.b	$2C					; ,
		dc.b	$68					; h
		dc.b	$A8					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$57					; W
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$A6					; 
		dc.b	$A8					; 
		dc.b	$90					; 
		dc.b	$77					; w
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$5F					; _
		dc.b	$4E					; N
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$C5					; 
		dc.b	$E5					; 
		dc.b	$B3					; 
		dc.b	$82					; 
		dc.b	$63					; c
		dc.b	$51					; Q
		dc.b	$6A					; j
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$4C					; L
		dc.b	$30					; 0
		dc.b	$50					; P
		dc.b	$8A					; 
		dc.b	$AF					; 
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$93					; 
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$77					; w
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$58					; X
		dc.b	$61					; a
		dc.b	$8B					; 
		dc.b	$C4					; 
		dc.b	$DE					; 
		dc.b	$AF					; 
		dc.b	$51					; Q
		dc.b	$18
		dc.b	$22					; "
		dc.b	$57					; W
		dc.b	$A5					; 
		dc.b	$E0					; 
		dc.b	$DD					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$5D					; ]
		dc.b	$2F					; /
		dc.b	$20
		dc.b	$25					; %
		dc.b	$66					; f
		dc.b	$D1					; 
		dc.b	$FC					; 
		dc.b	$E8					; 
		dc.b	$B6					; 
		dc.b	$5A					; Z
		dc.b	$1A
		dc.b	$22					; "
		dc.b	$49					; I
		dc.b	$6C					; l
		dc.b	$A4					; 
		dc.b	$D6					; 
		dc.b	$C7					; 
		dc.b	$98					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$51					; Q
		dc.b	$3D					; =
		dc.b	$3E					; >
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$AF					; 
		dc.b	$B8					; 
		dc.b	$B8					; 
		dc.b	$AA					; 
		dc.b	$8B					; 
		dc.b	$45					; E
		dc.b	$28					; (
		dc.b	$4F					; O
		dc.b	$7F					; 
		dc.b	$AD					; 
		dc.b	$BD					; 
		dc.b	$A2					; 
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$54					; T
		dc.b	$35					; 5
		dc.b	$2B					; +
		dc.b	$42					; B
		dc.b	$8B					; 
		dc.b	$BF					; 
		dc.b	$BE					; 
		dc.b	$AA					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$85					; 
		dc.b	$59					; Y
		dc.b	$39					; 9
		dc.b	$3A					; :
		dc.b	$72					; r
		dc.b	$AE					; 
		dc.b	$BB					; 
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$9E					; 
		dc.b	$8C					; 
		dc.b	$55					; U
		dc.b	$36					; 6
		dc.b	$4D					; M
		dc.b	$66					; f
		dc.b	$86					; 
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$B8					; 
		dc.b	$96					; 
		dc.b	$4D					; M
		dc.b	$2C					; ,
		dc.b	$42					; B
		dc.b	$78					; x
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$99					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$A9					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$3B					; ;
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$90					; 
		dc.b	$8C					; 
		dc.b	$6D					; m
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$50					; P
		dc.b	$6A					; j
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$77					; w
		dc.b	$5F					; _
		dc.b	$64					; d
		dc.b	$7E					; ~
		dc.b	$96					; 
		dc.b	$AB					; 
		dc.b	$9A					; 
		dc.b	$76					; v
		dc.b	$55					; U
		dc.b	$51					; Q
		dc.b	$88					; 
		dc.b	$CE					; 
		dc.b	$D7					; 
		dc.b	$8C					; 
		dc.b	$43					; C
		dc.b	$37					; 7
		dc.b	$61					; a
		dc.b	$A6					; 
		dc.b	$D0					; 
		dc.b	$CC					; 
		dc.b	$97					; 
		dc.b	$5D					; ]
		dc.b	$42					; B
		dc.b	$2D					; -
		dc.b	$47					; G
		dc.b	$8E					; 
		dc.b	$B7					; 
		dc.b	$C7					; 
		dc.b	$BC					; 
		dc.b	$7B					; {
		dc.b	$36					; 6
		dc.b	$28					; (
		dc.b	$49					; I
		dc.b	$7E					; ~
		dc.b	$BB					; 
		dc.b	$C8					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$82					; 
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$69					; i
		dc.b	$52					; R
		dc.b	$3F					; ?
		dc.b	$62					; b
		dc.b	$94					; 
		dc.b	$AB					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$66					; f
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$66					; f
		dc.b	$86					; 
		dc.b	$D3					; 
		dc.b	$C5					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$69					; i
		dc.b	$59					; Y
		dc.b	$50					; P
		dc.b	$58					; X
		dc.b	$83					; 
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$70					; p
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$B9					; 
		dc.b	$CF					; 
		dc.b	$91					; 
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$66					; f
		dc.b	$5B					; [
		dc.b	$78					; x
		dc.b	$9A					; 
		dc.b	$BA					; 
		dc.b	$D6					; 
		dc.b	$B2					; 
		dc.b	$56					; V
		dc.b	$1C
		dc.b	$1E
		dc.b	$2C					; ,
		dc.b	$67					; g
		dc.b	$D2					; 
		dc.b	$F9					; 
		dc.b	$D1					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$48					; H
		dc.b	$1E
		dc.b	$2A					; *
		dc.b	$4E					; N
		dc.b	$83					; 
		dc.b	$B7					; 
		dc.b	$BA					; 
		dc.b	$AC					; 
		dc.b	$AF					; 
		dc.b	$95					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$3D					; =
		dc.b	$31					; 1
		dc.b	$73					; s
		dc.b	$D2					; 
		dc.b	$E9					; 
		dc.b	$A9					; 
		dc.b	$56					; V
		dc.b	$3C					; <
		dc.b	$5A					; Z
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$C9					; 
		dc.b	$B1					; 
		dc.b	$5F					; _
		dc.b	$2B					; +
		dc.b	$30					; 0
		dc.b	$52					; R
		dc.b	$8C					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$A8					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$52					; R
		dc.b	$38					; 8
		dc.b	$65					; e
		dc.b	$98					; 
		dc.b	$BF					; 
		dc.b	$B6					; 
		dc.b	$97					; 
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$4B					; K
		dc.b	$42					; B
		dc.b	$4E					; N
		dc.b	$7E					; ~
		dc.b	$BC					; 
		dc.b	$E2					; 
		dc.b	$C4					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$8C					; 
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$69					; i
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$62					; b
		dc.b	$69					; i
		dc.b	$AC					; 
		dc.b	$DF					; 
		dc.b	$CE					; 
		dc.b	$B7					; 
		dc.b	$98					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$66					; f
		dc.b	$4D					; M
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$96					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$A8					; 
		dc.b	$85					; 
		dc.b	$3B					; ;
		dc.b	$28					; (
		dc.b	$54					; T
		dc.b	$A0					; 
		dc.b	$D7					; 
		dc.b	$C3					; 
		dc.b	$87					; 
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$69					; i
		dc.b	$4B					; K
		dc.b	$47					; G
		dc.b	$58					; X
		dc.b	$8A					; 
		dc.b	$D0					; 
		dc.b	$DD					; 
		dc.b	$A5					; 
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$6D					; m
		dc.b	$AA					; 
		dc.b	$CF					; 
		dc.b	$92					; 
		dc.b	$5B					; [
		dc.b	$51					; Q
		dc.b	$4A					; J
		dc.b	$73					; s
		dc.b	$B9					; 
		dc.b	$BC					; 
		dc.b	$88					; 
		dc.b	$5D					; ]
		dc.b	$3E					; >
		dc.b	$4F					; O
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$97					; 
		dc.b	$8D					; 
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$3F					; ?
		dc.b	$3E					; >
		dc.b	$79					; y
		dc.b	$A4					; 
		dc.b	$B8					; 
		dc.b	$8D					; 
		dc.b	$60					; `
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$90					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$67					; g
		dc.b	$3A					; :
		dc.b	$45					; E
		dc.b	$74					; t
		dc.b	$82					; 
		dc.b	$5C					; \
		dc.b	$51					; Q
		dc.b	$6E					; n
		dc.b	$A8					; 
		dc.b	$F1					; 
		dc.b	$EB					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$58					; X
		dc.b	$5F					; _
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$5F					; _
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$7F					; 
		dc.b	$C0					; 
		dc.b	$C7					; 
		dc.b	$9E					; 
		dc.b	$67					; g
		dc.b	$39					; 9
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$82					; 
		dc.b	$9F					; 
		dc.b	$C5					; 
		dc.b	$CE					; 
		dc.b	$AF					; 
		dc.b	$83					; 
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$62					; b
		dc.b	$75					; u
		dc.b	$97					; 
		dc.b	$A4					; 
		dc.b	$AD					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$4C					; L
		dc.b	$5B					; [
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$7D					; }
		dc.b	$AE					; 
		dc.b	$D0					; 
		dc.b	$BD					; 
		dc.b	$93					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$4B					; K
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$90					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$9E					; 
		dc.b	$91					; 
		dc.b	$5A					; Z
		dc.b	$2D					; -
		dc.b	$4D					; M
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$A5					; 
		dc.b	$93					; 
		dc.b	$66					; f
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$AA					; 
		dc.b	$B9					; 
		dc.b	$A2					; 
		dc.b	$62					; b
		dc.b	$31					; 1
		dc.b	$40					; @
		dc.b	$73					; s
		dc.b	$B4					; 
		dc.b	$D9					; 
		dc.b	$B4					; 
		dc.b	$82					; 
		dc.b	$70					; p
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$44					; D
		dc.b	$5A					; Z
		dc.b	$6D					; m
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$D0					; 
		dc.b	$CC					; 
		dc.b	$92					; 
		dc.b	$34					; 4
		dc.b	$15
		dc.b	$43					; C
		dc.b	$91					; 
		dc.b	$C7					; 
		dc.b	$BB					; 
		dc.b	$86					; 
		dc.b	$58					; X
		dc.b	$55					; U
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$9F					; 
		dc.b	$9E					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$98					; 
		dc.b	$D9					; 
		dc.b	$D1					; 
		dc.b	$93					; 
		dc.b	$41					; A
		dc.b	$2D					; -
		dc.b	$76					; v
		dc.b	$93					; 
		dc.b	$69					; i
		dc.b	$54					; T
		dc.b	$59					; Y
		dc.b	$94					; 
		dc.b	$EB					; 
		dc.b	$F5					; 
		dc.b	$AC					; 
		dc.b	$4B					; K
		dc.b	$1C
		dc.b	$23					; #
		dc.b	$4C					; L
		dc.b	$98					; 
		dc.b	$D8					; 
		dc.b	$E0					; 
		dc.b	$BA					; 
		dc.b	$86					; 
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$89					; 
		dc.b	$82					; 
		dc.b	$66					; f
		dc.b	$53					; S
		dc.b	$58					; X
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$8A					; 
		dc.b	$AC					; 
		dc.b	$90					; 
		dc.b	$7A					; z
		dc.b	$4E					; N
		dc.b	$2F					; /
		dc.b	$5C					; \
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$C1					; 
		dc.b	$B4					; 
		dc.b	$89					; 
		dc.b	$95					; 
		dc.b	$A3					; 
		dc.b	$77					; w
		dc.b	$5F					; _
		dc.b	$40					; @
		dc.b	$2C					; ,
		dc.b	$7D					; }
		dc.b	$D3					; 
		dc.b	$D4					; 
		dc.b	$90					; 
		dc.b	$42					; B
		dc.b	$40					; @
		dc.b	$67					; g
		dc.b	$6D					; m
		dc.b	$85					; 
		dc.b	$A6					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$93					; 
		dc.b	$62					; b
		dc.b	$37					; 7
		dc.b	$3F					; ?
		dc.b	$59					; Y
		dc.b	$9F					; 
		dc.b	$DB					; 
		dc.b	$E4					; 
		dc.b	$D9					; 
		dc.b	$AF					; 
		dc.b	$67					; g
		dc.b	$2F					; /
		dc.b	$33					; 3
		dc.b	$5A					; Z
		dc.b	$85					; 
		dc.b	$9B					; 
		dc.b	$6E					; n
		dc.b	$4B					; K
		dc.b	$79					; y
		dc.b	$AE					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$5F					; _
		dc.b	$32					; 2
		dc.b	$33					; 3
		dc.b	$59					; Y
		dc.b	$9B					; 
		dc.b	$CD					; 
		dc.b	$D4					; 
		dc.b	$C9					; 
		dc.b	$AA					; 
		dc.b	$7F					; 
		dc.b	$4A					; J
		dc.b	$1B
		dc.b	$26					; &
		dc.b	$60					; `
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$B9					; 
		dc.b	$B3					; 
		dc.b	$A8					; 
		dc.b	$6B					; k
		dc.b	$3E					; >
		dc.b	$47					; G
		dc.b	$56					; V
		dc.b	$7A					; z
		dc.b	$A1					; 
		dc.b	$A9					; 
		dc.b	$9A					; 
		dc.b	$79					; y
		dc.b	$5F					; _
		dc.b	$55					; U
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$B7					; 
		dc.b	$C9					; 
		dc.b	$CD					; 
		dc.b	$BC					; 
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$1D
		dc.b	$21					; !
		dc.b	$61					; a
		dc.b	$A9					; 
		dc.b	$CE					; 
		dc.b	$C2					; 
		dc.b	$80					; 
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$80					; 
		dc.b	$A9					; 
		dc.b	$8F					; 
		dc.b	$63					; c
		dc.b	$55					; U
		dc.b	$57					; W
		dc.b	$7E					; ~
		dc.b	$BA					; 
		dc.b	$C7					; 
		dc.b	$9D					; 
		dc.b	$70					; p
		dc.b	$4E					; N
		dc.b	$39					; 9
		dc.b	$54					; T
		dc.b	$8D					; 
		dc.b	$AF					; 
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$76					; v
		dc.b	$36					; 6
		dc.b	$2C					; ,
		dc.b	$53					; S
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$A5					; 
		dc.b	$C0					; 
		dc.b	$C8					; 
		dc.b	$C0					; 
		dc.b	$9B					; 
		dc.b	$62					; b
		dc.b	$3A					; :
		dc.b	$41					; A
		dc.b	$4B					; K
		dc.b	$5E					; ^
		dc.b	$9D					; 
		dc.b	$C2					; 
		dc.b	$B6					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$85					; 
		dc.b	$91					; 
		dc.b	$AC					; 
		dc.b	$BF					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$5F					; _
		dc.b	$48					; H
		dc.b	$39					; 9
		dc.b	$46					; F
		dc.b	$7E					; ~
		dc.b	$B6					; 
		dc.b	$B7					; 
		dc.b	$8B					; 
		dc.b	$6D					; m
		dc.b	$7E					; ~
		dc.b	$AF					; 
		dc.b	$BB					; 
		dc.b	$87					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$51					; Q
		dc.b	$63					; c
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$9A					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$A8					; 
		dc.b	$93					; 
		dc.b	$68					; h
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$4B					; K
		dc.b	$56					; V
		dc.b	$89					; 
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$9A					; 
		dc.b	$71					; q
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$93					; 
		dc.b	$74					; t
		dc.b	$63					; c
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$6E					; n
		dc.b	$74					; t
		dc.b	$5F					; _
		dc.b	$45					; E
		dc.b	$74					; t
		dc.b	$B6					; 
		dc.b	$E1					; 
		dc.b	$E7					; 
		dc.b	$B2					; 
		dc.b	$84					; 
		dc.b	$69					; i
		dc.b	$35					; 5
		dc.b	$1C
		dc.b	$29					; )
		dc.b	$4C					; L
		dc.b	$9F					; 
		dc.b	$E3					; 
		dc.b	$D1					; 
		dc.b	$A1					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	$3E					; >
		dc.b	$7B					; {
		dc.b	$C4					; 
		dc.b	$CA					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$54					; T
		dc.b	$4E					; N
		dc.b	$64					; d
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$C3					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$B6					; 
		dc.b	$98					; 
		dc.b	$69					; i
		dc.b	$6D					; m
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$9A					; 
		dc.b	$AB					; 
		dc.b	$B6					; 
		dc.b	$93					; 
		dc.b	$59					; Y
		dc.b	$2E					; .
		dc.b	$27					; '
		dc.b	$42					; B
		dc.b	$7B					; {
		dc.b	$C7					; 
		dc.b	$E9					; 
		dc.b	$D7					; 
		dc.b	$9F					; 
		dc.b	$48					; H
		dc.b	$25					; %
		dc.b	$3D					; =
		dc.b	$4B					; K
		dc.b	$61					; a
		dc.b	$8C					; 
		dc.b	$9E					; 
		dc.b	$9D					; 
		dc.b	$A7					; 
		dc.b	$A0					; 
		dc.b	$9E					; 
		dc.b	$97					; 
		dc.b	$82					; 
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$5D					; ]
		dc.b	$70					; p
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$BF					; 
		dc.b	$B5					; 
		dc.b	$7F					; 
		dc.b	$44					; D
		dc.b	$2C					; ,
		dc.b	$43					; C
		dc.b	$7A					; z
		dc.b	$A4					; 
		dc.b	$C1					; 
		dc.b	$C7					; 
		dc.b	$A6					; 
		dc.b	$6C					; l
		dc.b	$33					; 3
		dc.b	$2B					; +
		dc.b	$4D					; M
		dc.b	$76					; v
		dc.b	$90					; 
		dc.b	$8E					; 
		dc.b	$9A					; 
		dc.b	$B4					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$93					; 
		dc.b	$99					; 
		dc.b	$90					; 
		dc.b	$76					; v
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$7B					; {
		dc.b	$BF					; 
		dc.b	$E9					; 
		dc.b	$CF					; 
		dc.b	$8B					; 
		dc.b	$5E					; ^
		dc.b	$35					; 5
		dc.b	$26					; &
		dc.b	$4D					; M
		dc.b	$79					; y
		dc.b	$A3					; 
		dc.b	$9E					; 
		dc.b	$8C					; 
		dc.b	$AD					; 
		dc.b	$B5					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$49					; I
		dc.b	$51					; Q
		dc.b	$64					; d
		dc.b	$69					; i
		dc.b	$98					; 
		dc.b	$B9					; 
		dc.b	$AE					; 
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$4C					; L
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$6D					; m
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$8C					; 
		dc.b	$A8					; 
		dc.b	$95					; 
		dc.b	$90					; 
		dc.b	$A4					; 
		dc.b	$AF					; 
		dc.b	$9D					; 
		dc.b	$72					; r
		dc.b	$43					; C
		dc.b	$3A					; :
		dc.b	$5A					; Z
		dc.b	$89					; 
		dc.b	$C4					; 
		dc.b	$BE					; 
		dc.b	$93					; 
		dc.b	$92					; 
		dc.b	$64					; d
		dc.b	$39					; 9
		dc.b	$59					; Y
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$A6					; 
		dc.b	$B7					; 
		dc.b	$AE					; 
		dc.b	$9E					; 
		dc.b	$73					; s
		dc.b	$66					; f
		dc.b	$55					; U
		dc.b	$24					; $
		dc.b	$21					; !
		dc.b	$4D					; M
		dc.b	$7C					; |
		dc.b	$A3					; 
		dc.b	$AC					; 
		dc.b	$9C					; 
		dc.b	$AC					; 
		dc.b	$C2					; 
		dc.b	$B7					; 
		dc.b	$B0					; 
		dc.b	$74					; t
		dc.b	$2A					; *
		dc.b	$42					; B
		dc.b	$5F					; _
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$78					; x
		dc.b	$55					; U
		dc.b	$51					; Q
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$8E					; 
		dc.b	$CE					; 
		dc.b	$E9					; 
		dc.b	$C0					; 
		dc.b	$74					; t
		dc.b	$5B					; [
		dc.b	$52					; R
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$BA					; 
		dc.b	$CB					; 
		dc.b	$AA					; 
		dc.b	$97					; 
		dc.b	$93					; 
		dc.b	$6B					; k
		dc.b	$41					; A
		dc.b	$25					; %
		dc.b	$34					; 4
		dc.b	$7E					; ~
		dc.b	$B7					; 
		dc.b	$C9					; 
		dc.b	$B1					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.b	$5E					; ^
		dc.b	$91					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$53					; S
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$A1					; 
		dc.b	$D5					; 
		dc.b	$CD					; 
		dc.b	$86					; 
		dc.b	$42					; B
		dc.b	$24					; $
		dc.b	$39					; 9
		dc.b	$82					; 
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$94					; 
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$4A					; J
		dc.b	$7B					; {
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$A3					; 
		dc.b	$C2					; 
		dc.b	$B4					; 
		dc.b	$84					; 
		dc.b	$5E					; ^
		dc.b	$7E					; ~
		dc.b	$97					; 
		dc.b	$7F					; 
		dc.b	$62					; b
		dc.b	$54					; T
		dc.b	$7F					; 
		dc.b	$A9					; 
		dc.b	$9C					; 
		dc.b	$8C					; 
		dc.b	$75					; u
		dc.b	$58					; X
		dc.b	$61					; a
		dc.b	$54					; T
		dc.b	$3B					; ;
		dc.b	$54					; T
		dc.b	$7B					; {
		dc.b	$BB					; 
		dc.b	$F1					; 
		dc.b	$E1					; 
		dc.b	$BA					; 
		dc.b	$87					; 
		dc.b	$54					; T
		dc.b	$50					; P
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$6D					; m
		dc.b	$8A					; 
		dc.b	$B2					; 
		dc.b	$C0					; 
		dc.b	$94					; 
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$85					; 
		dc.b	$97					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$70					; p
		dc.b	$97					; 
		dc.b	$B0					; 
		dc.b	$8E					; 
		dc.b	$70					; p
		dc.b	$57					; W
		dc.b	$3C					; <
		dc.b	$60					; `
		dc.b	$B0					; 
		dc.b	$AD					; 
		dc.b	$61					; a
		dc.b	$38					; 8
		dc.b	$3F					; ?
		dc.b	$76					; v
		dc.b	$B4					; 
		dc.b	$B4					; 
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$9E					; 
		dc.b	$7F					; 
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$44					; D
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$93					; 
		dc.b	$85					; 
		dc.b	$4A					; J
		dc.b	$2B					; +
		dc.b	$43					; C
		dc.b	$6E					; n
		dc.b	$AE					; 
		dc.b	$E7					; 
		dc.b	$EB					; 
		dc.b	$BF					; 
		dc.b	$83					; 
		dc.b	$5B					; [
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$30					; 0
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$BB					; 
		dc.b	$C0					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$49					; I
		dc.b	$24					; $
		dc.b	$45					; E
		dc.b	$85					; 
		dc.b	$C1					; 
		dc.b	$EC					; 
		dc.b	$F0					; 
		dc.b	$CB					; 
		dc.b	$8A					; 
		dc.b	$3D					; =
		dc.b	$21					; !
		dc.b	$2F					; /
		dc.b	$59					; Y
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$A3					; 
		dc.b	$CD					; 
		dc.b	$D4					; 
		dc.b	$AC					; 
		dc.b	$7E					; ~
		dc.b	$5B					; [
		dc.b	$44					; D
		dc.b	$3F					; ?
		dc.b	$69					; i
		dc.b	$AF					; 
		dc.b	$BC					; 
		dc.b	$8E					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$66					; f
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$48					; H
		dc.b	$74					; t
		dc.b	$A9					; 
		dc.b	$C2					; 
		dc.b	$CC					; 
		dc.b	$AC					; 
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$70					; p
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$5F					; _
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$B3					; 
		dc.b	$AF					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$49					; I
		dc.b	$45					; E
		dc.b	$39					; 9
		dc.b	$5A					; Z
		dc.b	$98					; 
		dc.b	$B2					; 
		dc.b	$C9					; 
		dc.b	$DE					; 
		dc.b	$B5					; 
		dc.b	$6E					; n
		dc.b	$35					; 5
		dc.b	$1B
		dc.b	$3A					; :
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$97					; 
		dc.b	$60					; `
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$67					; g
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$C5					; 
		dc.b	$B3					; 
		dc.b	$A8					; 
		dc.b	$7A					; z
		dc.b	$40					; @
		dc.b	$5B					; [
		dc.b	$96					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$56					; V
		dc.b	$51					; Q
		dc.b	$78					; x
		dc.b	$99					; 
		dc.b	$A1					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$59					; Y
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$80					; 
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$89					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$5D					; ]
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$A5					; 
		dc.b	$9E					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$B2					; 
		dc.b	$E1					; 
		dc.b	$C7					; 
		dc.b	$A1					; 
		dc.b	$66					; f
		dc.b	$32					; 2
		dc.b	$4A					; J
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$A1					; 
		dc.b	$82					; 
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$81					; 
		dc.b	$A0					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$49					; I
		dc.b	$62					; b
		dc.b	$93					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$5F					; _
		dc.b	$5E					; ^
		dc.b	$8C					; 
		dc.b	$B7					; 
		dc.b	$B9					; 
		dc.b	$A9					; 
		dc.b	$82					; 
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$4D					; M
		dc.b	$47					; G
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$B9					; 
		dc.b	$CF					; 
		dc.b	$CF					; 
		dc.b	$BB					; 
		dc.b	$88					; 
		dc.b	$45					; E
		dc.b	$28					; (
		dc.b	$61					; a
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$62					; b
		dc.b	$5F					; _
		dc.b	$7E					; ~
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$B5					; 
		dc.b	$A8					; 
		dc.b	$82					; 
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$97					; 
		dc.b	$C0					; 
		dc.b	$A0					; 
		dc.b	$5A					; Z
		dc.b	$2C					; ,
		dc.b	$45					; E
		dc.b	$92					; 
		dc.b	$B9					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$65					; e
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$90					; 
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$5F					; _
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$5F					; _
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$8D					; 
		dc.b	$B8					; 
		dc.b	$9A					; 
		dc.b	$5F					; _
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$90					; 
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$9E					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$48					; H
		dc.b	$20
		dc.b	$29					; )
		dc.b	$70					; p
		dc.b	$CC					; 
		dc.b	$FF
		dc.b	$EF					; 
		dc.b	$B9					; 
		dc.b	$77					; w
		dc.b	$36					; 6
		dc.b	$16
		dc.b	$18
		dc.b	$2C					; ,
		dc.b	$66					; f
		dc.b	$B1					; 
		dc.b	$E0					; 
		dc.b	$DF					; 
		dc.b	$A8					; 
		dc.b	$71					; q
		dc.b	$50					; P
		dc.b	$41					; A
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$AC					; 
		dc.b	$C5					; 
		dc.b	$CC					; 
		dc.b	$C7					; 
		dc.b	$9C					; 
		dc.b	$40					; @
		dc.b	$11
		dc.b	$24					; $
		dc.b	$48					; H
		dc.b	$77					; w
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$D3					; 
		dc.b	$C7					; 
		dc.b	$8B					; 
		dc.b	$5A					; Z
		dc.b	$49					; I
		dc.b	$46					; F
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$AF					; 
		dc.b	$9E					; 
		dc.b	$5A					; Z
		dc.b	$32					; 2
		dc.b	$39					; 9
		dc.b	$4E					; N
		dc.b	$79					; y
		dc.b	$B0					; 
		dc.b	$D8					; 
		dc.b	$D5					; 
		dc.b	$AE					; 
		dc.b	$84					; 
		dc.b	$5F					; _
		dc.b	$4C					; L
		dc.b	$54					; T
		dc.b	$58					; X
		dc.b	$5F					; _
		dc.b	$99					; 
		dc.b	$C7					; 
		dc.b	$BD					; 
		dc.b	$C0					; 
		dc.b	$B9					; 
		dc.b	$71					; q
		dc.b	$29					; )
		dc.b	$18
		dc.b	$21					; !
		dc.b	$5A					; Z
		dc.b	$AC					; 
		dc.b	$CA					; 
		dc.b	$B8					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$7E					; ~
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$8D					; 
		dc.b	$9A					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$70					; p
		dc.b	$52					; R
		dc.b	$49					; I
		dc.b	$4C					; L
		dc.b	$6D					; m
		dc.b	$9A					; 
		dc.b	$BE					; 
		dc.b	$D5					; 
		dc.b	$B4					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$A5					; 
		dc.b	$C9					; 
		dc.b	$B6					; 
		dc.b	$BB					; 
		dc.b	$A3					; 
		dc.b	$4E					; N
		dc.b	$26					; &
		dc.b	$28					; (
		dc.b	$4F					; O
		dc.b	$8C					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$A8					; 
		dc.b	$A0					; 
		dc.b	$90					; 
		dc.b	$66					; f
		dc.b	$30					; 0
		dc.b	$31					; 1
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$B8					; 
		dc.b	$D7					; 
		dc.b	$B9					; 
		dc.b	$83					; 
		dc.b	$57					; W
		dc.b	$47					; G
		dc.b	$53					; S
		dc.b	$7B					; {
		dc.b	$B8					; 
		dc.b	$C9					; 
		dc.b	$A8					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$27					; '
		dc.b	$2A					; *
		dc.b	$5A					; Z
		dc.b	$87					; 
		dc.b	$B9					; 
		dc.b	$AE					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$8B					; 
		dc.b	$AB					; 
		dc.b	$C1					; 
		dc.b	$95					; 
		dc.b	$4C					; L
		dc.b	$2B					; +
		dc.b	$29					; )
		dc.b	$4A					; J
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$BD					; 
		dc.b	$A5					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$66					; f
		dc.b	$51					; Q
		dc.b	$4B					; K
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$8A					; 
		dc.b	$D9					; 
		dc.b	$E8					; 
		dc.b	$D1					; 
		dc.b	$8F					; 
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$66					; f
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$B2					; 
		dc.b	$B3					; 
		dc.b	$8D					; 
		dc.b	$69					; i
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$87					; 
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$87					; 
		dc.b	$66					; f
		dc.b	$65					; e
		dc.b	$90					; 
		dc.b	$99					; 
		dc.b	$85					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$66					; f
		dc.b	$98					; 
		dc.b	$C0					; 
		dc.b	$BF					; 
		dc.b	$7F					; 
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$51					; Q
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$B5					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$68					; h
		dc.b	$43					; C
		dc.b	$47					; G
		dc.b	$83					; 
		dc.b	$BB					; 
		dc.b	$CC					; 
		dc.b	$C4					; 
		dc.b	$B0					; 
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$48					; H
		dc.b	$20
		dc.b	$27					; '
		dc.b	$3D					; =
		dc.b	$68					; h
		dc.b	$B4					; 
		dc.b	$D8					; 
		dc.b	$BF					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$52					; R
		dc.b	$40					; @
		dc.b	$66					; f
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$92					; 
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$8F					; 
		dc.b	$AF					; 
		dc.b	$B9					; 
		dc.b	$98					; 
		dc.b	$63					; c
		dc.b	$54					; T
		dc.b	$47					; G
		dc.b	$49					; I
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$9E					; 
		dc.b	$A3					; 
		dc.b	$93					; 
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$37					; 7
		dc.b	$3A					; :
		dc.b	$76					; v
		dc.b	$C6					; 
		dc.b	$E4					; 
		dc.b	$A0					; 
		dc.b	$6B					; k
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$65					; e
		dc.b	$55					; U
		dc.b	$70					; p
		dc.b	$86					; 
		dc.b	$9F					; 
		dc.b	$C4					; 
		dc.b	$BA					; 
		dc.b	$89					; 
		dc.b	$68					; h
		dc.b	$58					; X
		dc.b	$5D					; ]
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$89					; 
		dc.b	$A2					; 
		dc.b	$AC					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$77					; w
		dc.b	$50					; P
		dc.b	$4C					; L
		dc.b	$89					; 
		dc.b	$B8					; 
		dc.b	$C3					; 
		dc.b	$BA					; 
		dc.b	$7E					; ~
		dc.b	$3B					; ;
		dc.b	$37					; 7
		dc.b	$4F					; O
		dc.b	$93					; 
		dc.b	$E5					; 
		dc.b	$E2					; 
		dc.b	$B2					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$41					; A
		dc.b	$44					; D
		dc.b	$5C					; \
		dc.b	$58					; X
		dc.b	$4D					; M
		dc.b	$73					; s
		dc.b	$B8					; 
		dc.b	$DC					; 
		dc.b	$CE					; 
		dc.b	$9C					; 
		dc.b	$55					; U
		dc.b	$33					; 3
		dc.b	$55					; U
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$9E					; 
		dc.b	$94					; 
		dc.b	$AD					; 
		dc.b	$BB					; 
		dc.b	$AC					; 
		dc.b	$A5					; 
		dc.b	$6E					; n
		dc.b	$27					; '
		dc.b	$1C
		dc.b	$3A					; :
		dc.b	$76					; v
		dc.b	$AA					; 
		dc.b	$BB					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$97					; 
		dc.b	$56					; V
		dc.b	$2F					; /
		dc.b	$42					; B
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$AE					; 
		dc.b	$C7					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$6D					; m
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$3C					; <
		dc.b	$41					; A
		dc.b	$82					; 
		dc.b	$B1					; 
		dc.b	$C7					; 
		dc.b	$DF					; 
		dc.b	$B0					; 
		dc.b	$64					; d
		dc.b	$3B					; ;
		dc.b	$25					; %
		dc.b	$37					; 7
		dc.b	$5F					; _
		dc.b	$75					; u
		dc.b	$A5					; 
		dc.b	$DE					; 
		dc.b	$E7					; 
		dc.b	$C3					; 
		dc.b	$93					; 
		dc.b	$70					; p
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$57					; W
		dc.b	$5D					; ]
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$A0					; 
		dc.b	$B9					; 
		dc.b	$D3					; 
		dc.b	$BE					; 
		dc.b	$8D					; 
		dc.b	$5E					; ^
		dc.b	$34					; 4
		dc.b	$1D
		dc.b	$26					; &
		dc.b	$4E					; N
		dc.b	$9D					; 
		dc.b	$E4					; 
		dc.b	$E9					; 
		dc.b	$AD					; 
		dc.b	$57					; W
		dc.b	$38					; 8
		dc.b	$61					; a
		dc.b	$7B					; {
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$6A					; j
		dc.b	$99					; 
		dc.b	$D3					; 
		dc.b	$EB					; 
		dc.b	$BF					; 
		dc.b	$85					; 
		dc.b	$51					; Q
		dc.b	$25					; %
		dc.b	$2E					; .
		dc.b	$5A					; Z
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$A3					; 
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$74					; t
		dc.b	$99					; 
		dc.b	$93					; 
		dc.b	$6A					; j
		dc.b	$44					; D
		dc.b	$6D					; m
		dc.b	$BD					; 
		dc.b	$D2					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$3E					; >
		dc.b	$2E					; .
		dc.b	$4B					; K
		dc.b	$67					; g
		dc.b	$69					; i
		dc.b	$94					; 
		dc.b	$D7					; 
		dc.b	$BA					; 
		dc.b	$6D					; m
		dc.b	$4B					; K
		dc.b	$4E					; N
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$AF					; 
		dc.b	$95					; 
		dc.b	$90					; 
		dc.b	$9F					; 
		dc.b	$AD					; 
		dc.b	$9B					; 
		dc.b	$56					; V
		dc.b	$22					; "
		dc.b	$2D					; -
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$B0					; 
		dc.b	$C3					; 
		dc.b	$BA					; 
		dc.b	$AF					; 
		dc.b	$8B					; 
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$95					; 
		dc.b	$B6					; 
		dc.b	$C1					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$93					; 
		dc.b	$BF					; 
		dc.b	$BB					; 
		dc.b	$8F					; 
		dc.b	$5D					; ]
		dc.b	$2A					; *
		dc.b	$24					; $
		dc.b	$4B					; K
		dc.b	$7B					; {
		dc.b	$A1					; 
		dc.b	$B5					; 
		dc.b	$BC					; 
		dc.b	$B3					; 
		dc.b	$77					; w
		dc.b	$42					; B
		dc.b	$47					; G
		dc.b	$49					; I
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$77					; w
		dc.b	$8B					; 
		dc.b	$AB					; 
		dc.b	$A8					; 
		dc.b	$B3					; 
		dc.b	$A0					; 
		dc.b	$77					; w
		dc.b	$6B					; k
		dc.b	$5B					; [
		dc.b	$4B					; K
		dc.b	$65					; e
		dc.b	$94					; 
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$8D					; 
		dc.b	$A0					; 
		dc.b	$AF					; 
		dc.b	$8E					; 
		dc.b	$41					; A
		dc.b	$19
		dc.b	$22					; "
		dc.b	$4A					; J
		dc.b	$A1					; 
		dc.b	$EB					; 
		dc.b	$EC					; 
		dc.b	$CB					; 
		dc.b	$9B					; 
		dc.b	$62					; b
		dc.b	$3F					; ?
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$54					; T
		dc.b	$88					; 
		dc.b	$D1					; 
		dc.b	$EF					; 
		dc.b	$D2					; 
		dc.b	$9A					; 
		dc.b	$75					; u
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$5F					; _
		dc.b	$3F					; ?
		dc.b	$2D					; -
		dc.b	$50					; P
		dc.b	$96					; 
		dc.b	$B4					; 
		dc.b	$BC					; 
		dc.b	$C5					; 
		dc.b	$AA					; 
		dc.b	$7A					; z
		dc.b	$70					; p
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$4E					; N
		dc.b	$66					; f
		dc.b	$85					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$97					; 
		dc.b	$B3					; 
		dc.b	$B8					; 
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$3C					; <
		dc.b	$6A					; j
		dc.b	$95					; 
		dc.b	$AC					; 
		dc.b	$80					; 
		dc.b	$3D					; =
		dc.b	$53					; S
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$A2					; 
		dc.b	$D1					; 
		dc.b	$D2					; 
		dc.b	$B2					; 
		dc.b	$73					; s
		dc.b	$54					; T
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$70					; p
		dc.b	$43					; C
		dc.b	$4B					; K
		dc.b	$5C					; \
		dc.b	$6D					; m
		dc.b	$82					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$CA					; 
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$54					; T
		dc.b	$40					; @
		dc.b	$5B					; [
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$69					; i
		dc.b	$7E					; ~
		dc.b	$A7					; 
		dc.b	$CB					; 
		dc.b	$CA					; 
		dc.b	$A4					; 
		dc.b	$74					; t
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$54					; T
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$85					; 
		dc.b	$AC					; 
		dc.b	$A6					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$9E					; 
		dc.b	$8C					; 
		dc.b	$4D					; M
		dc.b	$21					; !
		dc.b	$38					; 8
		dc.b	$65					; e
		dc.b	$9B					; 
		dc.b	$C3					; 
		dc.b	$B9					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$65					; e
		dc.b	$29					; )
		dc.b	$1E
		dc.b	$4D					; M
		dc.b	$A0					; 
		dc.b	$E4					; 
		dc.b	$F6					; 
		dc.b	$DA					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$37					; 7
		dc.b	$49					; I
		dc.b	$7E					; ~
		dc.b	$9D					; 
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$97					; 
		dc.b	$9E					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$97					; 
		dc.b	$5F					; _
		dc.b	$26					; &
		dc.b	$24					; $
		dc.b	$45					; E
		dc.b	$80					; 
		dc.b	$CB					; 
		dc.b	$DA					; 
		dc.b	$C1					; 
		dc.b	$B8					; 
		dc.b	$97					; 
		dc.b	$65					; e
		dc.b	$6D					; m
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$51					; Q
		dc.b	$2D					; -
		dc.b	$47					; G
		dc.b	$58					; X
		dc.b	$8F					; 
		dc.b	$BC					; 
		dc.b	$B1					; 
		dc.b	$B6					; 
		dc.b	$A4					; 
		dc.b	$5B					; [
		dc.b	$3C					; <
		dc.b	$4A					; J
		dc.b	$58					; X
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$A1					; 
		dc.b	$97					; 
		dc.b	$84					; 
		dc.b	$69					; i
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$72					; r
		dc.b	$A4					; 
		dc.b	$D0					; 
		dc.b	$D6					; 
		dc.b	$B6					; 
		dc.b	$81					; 
		dc.b	$47					; G
		dc.b	$41					; A
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$92					; 
		dc.b	$BD					; 
		dc.b	$C5					; 
		dc.b	$A4					; 
		dc.b	$82					; 
		dc.b	$63					; c
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$7A					; z
		dc.b	$A1					; 
		dc.b	$B3					; 
		dc.b	$A5					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$4C					; L
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$AF					; 
		dc.b	$B5					; 
		dc.b	$B8					; 
		dc.b	$9D					; 
		dc.b	$6B					; k
		dc.b	$58					; X
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$62					; b
		dc.b	$57					; W
		dc.b	$81					; 
		dc.b	$A9					; 
		dc.b	$A4					; 
		dc.b	$83					; 
		dc.b	$50					; P
		dc.b	$5D					; ]
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$97					; 
		dc.b	$7A					; z
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$89					; 
		dc.b	$BB					; 
		dc.b	$D5					; 
		dc.b	$AA					; 
		dc.b	$70					; p
		dc.b	$43					; C
		dc.b	$24					; $
		dc.b	$29					; )
		dc.b	$56					; V
		dc.b	$8B					; 
		dc.b	$BD					; 
		dc.b	$CD					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$C1					; 
		dc.b	$A2					; 
		dc.b	$6F					; o
		dc.b	$30					; 0
		dc.b	$29					; )
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$9C					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$70					; p
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$93					; 
		dc.b	$BD					; 
		dc.b	$B1					; 
		dc.b	$89					; 
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$61					; a
		dc.b	$7B					; {
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$7B					; {
		dc.b	$A7					; 
		dc.b	$B9					; 
		dc.b	$B5					; 
		dc.b	$95					; 
		dc.b	$64					; d
		dc.b	$4E					; N
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$90					; 
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$4C					; L
		dc.b	$7C					; |
		dc.b	$BD					; 
		dc.b	$D8					; 
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$5F					; _
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$9A					; 
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$4B					; K
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$9E					; 
		dc.b	$CE					; 
		dc.b	$BD					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$9A					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$6D					; m
		dc.b	$2E					; .
		dc.b	$37					; 7
		dc.b	$6D					; m
		dc.b	$A8					; 
		dc.b	$D5					; 
		dc.b	$CF					; 
		dc.b	$97					; 
		dc.b	$62					; b
		dc.b	$61					; a
		dc.b	$5B					; [
		dc.b	$3D					; =
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$81					; 
		dc.b	$A4					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$B3					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$53					; S
		dc.b	$47					; G
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$77					; w
		dc.b	$C0					; 
		dc.b	$D0					; 
		dc.b	$A7					; 
		dc.b	$69					; i
		dc.b	$44					; D
		dc.b	$55					; U
		dc.b	$86					; 
		dc.b	$C2					; 
		dc.b	$CB					; 
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$3E					; >
		dc.b	$3F					; ?
		dc.b	$66					; f
		dc.b	$A1					; 
		dc.b	$BA					; 
		dc.b	$88					; 
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$A8					; 
		dc.b	$CC					; 
		dc.b	$B0					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$4C					; L
		dc.b	$4C					; L
		dc.b	$54					; T
		dc.b	$66					; f
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$B4					; 
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$1D
		dc.b	$2B					; +
		dc.b	$70					; p
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$E0					; 
		dc.b	$E2					; 
		dc.b	$B9					; 
		dc.b	$8B					; 
		dc.b	$4D					; M
		dc.b	$32					; 2
		dc.b	$57					; W
		dc.b	$66					; f
		dc.b	$48					; H
		dc.b	$42					; B
		dc.b	$62					; b
		dc.b	$99					; 
		dc.b	$B7					; 
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$7E					; ~
		dc.b	$9A					; 
		dc.b	$AC					; 
		dc.b	$CD					; 
		dc.b	$CF					; 
		dc.b	$9D					; 
		dc.b	$6A					; j
		dc.b	$44					; D
		dc.b	$4B					; K
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$9A					; 
		dc.b	$6D					; m
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$8B					; 
		dc.b	$B5					; 
		dc.b	$CE					; 
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	$3F					; ?
		dc.b	$31					; 1
		dc.b	$40					; @
		dc.b	$5A					; Z
		dc.b	$96					; 
		dc.b	$A1					; 
		dc.b	$89					; 
		dc.b	$9A					; 
		dc.b	$89					; 
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$8E					; 
		dc.b	$93					; 
		dc.b	$7E					; ~
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$5A					; Z
		dc.b	$8E					; 
		dc.b	$A1					; 
		dc.b	$9A					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$97					; 
		dc.b	$7D					; }
		dc.b	$69					; i
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$57					; W
		dc.b	$74					; t
		dc.b	$A9					; 
		dc.b	$BE					; 
		dc.b	$C9					; 
		dc.b	$BF					; 
		dc.b	$8B					; 
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$69					; i
		dc.b	$6F					; o
		dc.b	$7B					; {
		dc.b	$AA					; 
		dc.b	$B5					; 
		dc.b	$7F					; 
		dc.b	$51					; Q
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$5F					; _
		dc.b	$82					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$A1					; 
		dc.b	$B5					; 
		dc.b	$C3					; 
		dc.b	$A7					; 
		dc.b	$81					; 
		dc.b	$6D					; m
		dc.b	$5A					; Z
		dc.b	$5B					; [
		dc.b	$66					; f
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$B8					; 
		dc.b	$B9					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$5F					; _
		dc.b	$37					; 7
		dc.b	$31					; 1
		dc.b	$3B					; ;
		dc.b	$76					; v
		dc.b	$C0					; 
		dc.b	$E2					; 
		dc.b	$BA					; 
		dc.b	$72					; r
		dc.b	$3D					; =
		dc.b	$28					; (
		dc.b	$4B					; K
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$A2					; 
		dc.b	$BC					; 
		dc.b	$9A					; 
		dc.b	$6B					; k
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$BB					; 
		dc.b	$C4					; 
		dc.b	$C2					; 
		dc.b	$B1					; 
		dc.b	$75					; u
		dc.b	$50					; P
		dc.b	$5B					; [
		dc.b	$80					; 
		dc.b	$A5					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$47					; G
		dc.b	$41					; A
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$AF					; 
		dc.b	$D2					; 
		dc.b	$BB					; 
		dc.b	$90					; 
		dc.b	$7D					; }
		dc.b	$5F					; _
		dc.b	$6F					; o
		dc.b	$8C					; 
		dc.b	$63					; c
		dc.b	$33					; 3
		dc.b	$2D					; -
		dc.b	$61					; a
		dc.b	$C2					; 
		dc.b	$ED					; 
		dc.b	$BA					; 
		dc.b	$6D					; m
		dc.b	$4E					; N
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$BF					; 
		dc.b	$C6					; 
		dc.b	$9A					; 
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$79					; y
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$A5					; 
		dc.b	$C3					; 
		dc.b	$B8					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$58					; X
		dc.b	$49					; I
		dc.b	$40					; @
		dc.b	$3B					; ;
		dc.b	$47					; G
		dc.b	$72					; r
		dc.b	$A6					; 
		dc.b	$C0					; 
		dc.b	$BD					; 
		dc.b	$B2					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$76					; v
		dc.b	$49					; I
		dc.b	$2A					; *
		dc.b	$58					; X
		dc.b	$8C					; 
		dc.b	$A1					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$70					; p
		dc.b	$58					; X
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$AB					; 
		dc.b	$C8					; 
		dc.b	$A5					; 
		dc.b	$62					; b
		dc.b	$2E					; .
		dc.b	$2A					; *
		dc.b	$47					; G
		dc.b	$84					; 
		dc.b	$D2					; 
		dc.b	$E0					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$62					; b
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$9A					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$6A					; j
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$76					; v
		dc.b	$97					; 
		dc.b	$A2					; 
		dc.b	$B2					; 
		dc.b	$AD					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$5B					; [
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$76					; v
		dc.b	$B3					; 
		dc.b	$CA					; 
		dc.b	$97					; 
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$60					; `
		dc.b	$58					; X
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$5D					; ]
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$62					; b
		dc.b	$42					; B
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$93					; 
		dc.b	$D0					; 
		dc.b	$DE					; 
		dc.b	$D6					; 
		dc.b	$C6					; 
		dc.b	$9D					; 
		dc.b	$73					; s
		dc.b	$56					; V
		dc.b	$3C					; <
		dc.b	$2A					; *
		dc.b	$33					; 3
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$85					; 
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$89					; 
		dc.b	$AA					; 
		dc.b	$C5					; 
		dc.b	$BF					; 
		dc.b	$7E					; ~
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$58					; X
		dc.b	$7A					; z
		dc.b	$8C					; 
		dc.b	$93					; 
		dc.b	$A9					; 
		dc.b	$B5					; 
		dc.b	$B2					; 
		dc.b	$A2					; 
		dc.b	$75					; u
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$49					; I
		dc.b	$68					; h
		dc.b	$A1					; 
		dc.b	$D8					; 
		dc.b	$D1					; 
		dc.b	$BE					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$79					; y
		dc.b	$3A					; :
		dc.b	$1C
		dc.b	$25					; %
		dc.b	$4E					; N
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$8C					; 
		dc.b	$AF					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$69					; i
		dc.b	$5C					; \
		dc.b	$51					; Q
		dc.b	$57					; W
		dc.b	$8C					; 
		dc.b	$CE					; 
		dc.b	$E2					; 
		dc.b	$BE					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$3F					; ?
		dc.b	$49					; I
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$BB					; 
		dc.b	$C5					; 
		dc.b	$97					; 
		dc.b	$58					; X
		dc.b	$40					; @
		dc.b	$43					; C
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$8C					; 
		dc.b	$AC					; 
		dc.b	$C2					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$6D					; m
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$9E					; 
		dc.b	$7E					; ~
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$AB					; 
		dc.b	$AC					; 
		dc.b	$93					; 
		dc.b	$76					; v
		dc.b	$5B					; [
		dc.b	$57					; W
		dc.b	$81					; 
		dc.b	$A1					; 
		dc.b	$A5					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$5B					; [
		dc.b	$64					; d
		dc.b	$76					; v
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$A5					; 
		dc.b	$BC					; 
		dc.b	$97					; 
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$9C					; 
		dc.b	$A5					; 
		dc.b	$8E					; 
		dc.b	$53					; S
		dc.b	$2B					; +
		dc.b	$4E					; N
		dc.b	$84					; 
		dc.b	$AF					; 
		dc.b	$CE					; 
		dc.b	$BB					; 
		dc.b	$89					; 
		dc.b	$49					; I
		dc.b	$1C
		dc.b	$37					; 7
		dc.b	$66					; f
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$A1					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$9A					; 
		dc.b	$7E					; ~
		dc.b	$45					; E
		dc.b	$40					; @
		dc.b	$66					; f
		dc.b	$9B					; 
		dc.b	$D1					; 
		dc.b	$C2					; 
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$48					; H
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$CF					; 
		dc.b	$DE					; 
		dc.b	$AF					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$54					; T
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$B5					; 
		dc.b	$A8					; 
		dc.b	$A3					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$9A					; 
		dc.b	$86					; 
		dc.b	$5F					; _
		dc.b	$3D					; =
		dc.b	$41					; A
		dc.b	$4F					; O
		dc.b	$5F					; _
		dc.b	$82					; 
		dc.b	$8C					; 
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$BF					; 
		dc.b	$BC					; 
		dc.b	$84					; 
		dc.b	$51					; Q
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$69					; i
		dc.b	$A2					; 
		dc.b	$CA					; 
		dc.b	$D1					; 
		dc.b	$AE					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$6D					; m
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$58					; X
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$C7					; 
		dc.b	$D8					; 
		dc.b	$CF					; 
		dc.b	$B9					; 
		dc.b	$76					; v
		dc.b	$28					; (
		dc.b	$1C
		dc.b	$3D					; =
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$9D					; 
		dc.b	$A8					; 
		dc.b	$8D					; 
		dc.b	$55					; U
		dc.b	$3E					; >
		dc.b	$48					; H
		dc.b	$79					; y
		dc.b	$AB					; 
		dc.b	$BB					; 
		dc.b	$BA					; 
		dc.b	$99					; 
		dc.b	$58					; X
		dc.b	$3E					; >
		dc.b	$47					; G
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$AF					; 
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$9A					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$7B					; {
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$62					; b
		dc.b	$8C					; 
		dc.b	$98					; 
		dc.b	$A7					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$A5					; 
		dc.b	$71					; q
		dc.b	$34					; 4
		dc.b	$2B					; +
		dc.b	$42					; B
		dc.b	$82					; 
		dc.b	$C6					; 
		dc.b	$C1					; 
		dc.b	$AF					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$41					; A
		dc.b	$40					; @
		dc.b	$54					; T
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$B5					; 
		dc.b	$C8					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$58					; X
		dc.b	$42					; B
		dc.b	$59					; Y
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$97					; 
		dc.b	$A8					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$74					; t
		dc.b	$61					; a
		dc.b	$3F					; ?
		dc.b	$4D					; M
		dc.b	$95					; 
		dc.b	$C3					; 
		dc.b	$B6					; 
		dc.b	$8D					; 
		dc.b	$57					; W
		dc.b	$2C					; ,
		dc.b	$34					; 4
		dc.b	$70					; p
		dc.b	$9E					; 
		dc.b	$B2					; 
		dc.b	$BE					; 
		dc.b	$B2					; 
		dc.b	$8D					; 
		dc.b	$74					; t
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$9F					; 
		dc.b	$D2					; 
		dc.b	$DA					; 
		dc.b	$A8					; 
		dc.b	$6C					; l
		dc.b	$4F					; O
		dc.b	$47					; G
		dc.b	$51					; Q
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$70					; p
		dc.b	$61					; a
		dc.b	$69					; i
		dc.b	$9C					; 
		dc.b	$C4					; 
		dc.b	$B7					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$8F					; 
		dc.b	$C6					; 
		dc.b	$D2					; 
		dc.b	$B0					; 
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$5F					; _
		dc.b	$6D					; m
		dc.b	$71					; q
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$5E					; ^
		dc.b	$53					; S
		dc.b	$74					; t
		dc.b	$A3					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$63					; c
		dc.b	$31					; 1
		dc.b	$20
		dc.b	$2D					; -
		dc.b	$63					; c
		dc.b	$B6					; 
		dc.b	$D8					; 
		dc.b	$CE					; 
		dc.b	$BD					; 
		dc.b	$9E					; 
		dc.b	$7E					; ~
		dc.b	$6F					; o
		dc.b	$54					; T
		dc.b	$39					; 9
		dc.b	$47					; G
		dc.b	$65					; e
		dc.b	$6D					; m
		dc.b	$7E					; ~
		dc.b	$AF					; 
		dc.b	$BD					; 
		dc.b	$9A					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$66					; f
		dc.b	$5B					; [
		dc.b	$6A					; j
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$9E					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$54					; T
		dc.b	$6F					; o
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$A1					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$4A					; J
		dc.b	$79					; y
		dc.b	$C9					; 
		dc.b	$E6					; 
		dc.b	$D1					; 
		dc.b	$AD					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$5B					; [
		dc.b	$33					; 3
		dc.b	$2E					; .
		dc.b	$3B					; ;
		dc.b	$66					; f
		dc.b	$A1					; 
		dc.b	$BD					; 
		dc.b	$B7					; 
		dc.b	$B2					; 
		dc.b	$AA					; 
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$69					; i
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$5C					; \
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$C7					; 
		dc.b	$C1					; 
		dc.b	$9E					; 
		dc.b	$7E					; ~
		dc.b	$65					; e
		dc.b	$62					; b
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$A6					; 
		dc.b	$BF					; 
		dc.b	$AC					; 
		dc.b	$A3					; 
		dc.b	$7E					; ~
		dc.b	$45					; E
		dc.b	$29					; )
		dc.b	$33					; 3
		dc.b	$69					; i
		dc.b	$A3					; 
		dc.b	$B8					; 
		dc.b	$A2					; 
		dc.b	$74					; t
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$68					; h
		dc.b	$5B					; [
		dc.b	$76					; v
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$9F					; 
		dc.b	$93					; 
		dc.b	$77					; w
		dc.b	$62					; b
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$B5					; 
		dc.b	$AA					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$5F					; _
		dc.b	$A1					; 
		dc.b	$BB					; 
		dc.b	$99					; 
		dc.b	$6E					; n
		dc.b	$4A					; J
		dc.b	$51					; Q
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$92					; 
		dc.b	$77					; w
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$74					; t
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$66					; f
		dc.b	$5E					; ^
		dc.b	$6D					; m
		dc.b	$6E					; n
		dc.b	$74					; t
		dc.b	$93					; 
		dc.b	$A7					; 
		dc.b	$9E					; 
		dc.b	$91					; 
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$69					; i
		dc.b	$4B					; K
		dc.b	$58					; X
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$9A					; 
		dc.b	$88					; 
		dc.b	$66					; f
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$AD					; 
		dc.b	$BF					; 
		dc.b	$B9					; 
		dc.b	$8D					; 
		dc.b	$56					; V
		dc.b	$31					; 1
		dc.b	$37					; 7
		dc.b	$68					; h
		dc.b	$9B					; 
		dc.b	$C0					; 
		dc.b	$CA					; 
		dc.b	$97					; 
		dc.b	$49					; I
		dc.b	$36					; 6
		dc.b	$63					; c
		dc.b	$89					; 
		dc.b	$92					; 
		dc.b	$7B					; {
		dc.b	$55					; U
		dc.b	$47					; G
		dc.b	$49					; I
		dc.b	$6D					; m
		dc.b	$BD					; 
		dc.b	$EB					; 
		dc.b	$E9					; 
		dc.b	$CB					; 
		dc.b	$88					; 
		dc.b	$4E					; N
		dc.b	$3E					; >
		dc.b	$48					; H
		dc.b	$56					; V
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$B4					; 
		dc.b	$C3					; 
		dc.b	$CF					; 
		dc.b	$A5					; 
		dc.b	$6A					; j
		dc.b	$44					; D
		dc.b	$35					; 5
		dc.b	$51					; Q
		dc.b	$7B					; {
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$7D					; }
		dc.b	$A0					; 
		dc.b	$CF					; 
		dc.b	$E2					; 
		dc.b	$B9					; 
		dc.b	$5F					; _
		dc.b	$1C
		dc.b	$26					; &
		dc.b	$59					; Y
		dc.b	$95					; 
		dc.b	$B2					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$A5					; 
		dc.b	$9E					; 
		dc.b	$79					; y
		dc.b	$45					; E
		dc.b	$28					; (
		dc.b	$42					; B
		dc.b	$6D					; m
		dc.b	$99					; 
		dc.b	$C6					; 
		dc.b	$C6					; 
		dc.b	$A5					; 
		dc.b	$70					; p
		dc.b	$2F					; /
		dc.b	$27					; '
		dc.b	$56					; V
		dc.b	$82					; 
		dc.b	$AF					; 
		dc.b	$BA					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$B2					; 
		dc.b	$92					; 
		dc.b	$5B					; [
		dc.b	$36					; 6
		dc.b	$37					; 7
		dc.b	$55					; U
		dc.b	$83					; 
		dc.b	$AC					; 
		dc.b	$C3					; 
		dc.b	$B9					; 
		dc.b	$93					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$54					; T
		dc.b	$6B					; k
		dc.b	$97					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$62					; b
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$90					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$90					; 
		dc.b	$77					; w
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$58					; X
		dc.b	$83					; 
		dc.b	$C6					; 
		dc.b	$D0					; 
		dc.b	$C4					; 
		dc.b	$93					; 
		dc.b	$58					; X
		dc.b	$62					; b
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$43					; C
		dc.b	$51					; Q
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$A5					; 
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$B5					; 
		dc.b	$B8					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$44					; D
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$7A					; z
		dc.b	$93					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$9E					; 
		dc.b	$A4					; 
		dc.b	$6E					; n
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$44					; D
		dc.b	$5E					; ^
		dc.b	$9D					; 
		dc.b	$C1					; 
		dc.b	$CA					; 
		dc.b	$BF					; 
		dc.b	$BD					; 
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$32					; 2
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$5B					; [
		dc.b	$A1					; 
		dc.b	$BA					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$5F					; _
		dc.b	$45					; E
		dc.b	$56					; V
		dc.b	$85					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$C6					; 
		dc.b	$93					; 
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$49					; I
		dc.b	$5A					; Z
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$69					; i
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$7B					; {
		dc.b	$A2					; 
		dc.b	$93					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$56					; V
		dc.b	$77					; w
		dc.b	$97					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$85					; 
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$6F					; o
		dc.b	$62					; b
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$54					; T
		dc.b	$56					; V
		dc.b	$79					; y
		dc.b	$A4					; 
		dc.b	$CF					; 
		dc.b	$D8					; 
		dc.b	$BE					; 
		dc.b	$8C					; 
		dc.b	$55					; U
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$5F					; _
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$97					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$5F					; _
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$83					; 
		dc.b	$9E					; 
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$97					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$4D					; M
		dc.b	$42					; B
		dc.b	$70					; p
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$5F					; _
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$82					; 
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$A8					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$70					; p
		dc.b	$6D					; m
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$86					; 
		dc.b	$A7					; 
		dc.b	$B9					; 
		dc.b	$87					; 
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$B3					; 
		dc.b	$BC					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$6C					; l
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$70					; p
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$B2					; 
		dc.b	$AB					; 
		dc.b	$8B					; 
		dc.b	$5B					; [
		dc.b	$59					; Y
		dc.b	$91					; 
		dc.b	$A9					; 
		dc.b	$86					; 
		dc.b	$59					; Y
		dc.b	$42					; B
		dc.b	$5A					; Z
		dc.b	$93					; 
		dc.b	$A2					; 
		dc.b	$89					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$92					; 
		dc.b	$A1					; 
		dc.b	$8E					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$AC					; 
		dc.b	$B7					; 
		dc.b	$A3					; 
		dc.b	$8A					; 
		dc.b	$5E					; ^
		dc.b	$3D					; =
		dc.b	$5E					; ^
		dc.b	$95					; 
		dc.b	$AB					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$62					; b
		dc.b	$9B					; 
		dc.b	$C2					; 
		dc.b	$D5					; 
		dc.b	$B9					; 
		dc.b	$72					; r
		dc.b	$59					; Y
		dc.b	$7A					; z
		dc.b	$89					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$BD					; 
		dc.b	$C5					; 
		dc.b	$A5					; 
		dc.b	$77					; w
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$66					; f
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$B7					; 
		dc.b	$D8					; 
		dc.b	$D7					; 
		dc.b	$B2					; 
		dc.b	$6D					; m
		dc.b	$43					; C
		dc.b	$3B					; ;
		dc.b	$56					; V
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$93					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$78					; x
		dc.b	$B6					; 
		dc.b	$D7					; 
		dc.b	$D1					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$2F					; /
		dc.b	$2D					; -
		dc.b	$4A					; J
		dc.b	$74					; t
		dc.b	$9F					; 
		dc.b	$B2					; 
		dc.b	$AB					; 
		dc.b	$A2					; 
		dc.b	$AF					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$49					; I
		dc.b	$3E					; >
		dc.b	$5E					; ^
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$69					; i
		dc.b	$8F					; 
		dc.b	$CA					; 
		dc.b	$CD					; 
		dc.b	$AA					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$6F					; o
		dc.b	$58					; X
		dc.b	$4B					; K
		dc.b	$52					; R
		dc.b	$84					; 
		dc.b	$B7					; 
		dc.b	$A0					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$6D					; m
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$54					; T
		dc.b	$57					; W
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$B2					; 
		dc.b	$AF					; 
		dc.b	$93					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$69					; i
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$60					; `
		dc.b	$51					; Q
		dc.b	$6D					; m
		dc.b	$8C					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$9A					; 
		dc.b	$A4					; 
		dc.b	$AD					; 
		dc.b	$86					; 
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$93					; 
		dc.b	$6E					; n
		dc.b	$5B					; [
		dc.b	$4A					; J
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$53					; S
		dc.b	$46					; F
		dc.b	$66					; f
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$89					; 
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$5B					; [
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$6C					; l
		dc.b	$94					; 
		dc.b	$CC					; 
		dc.b	$C9					; 
		dc.b	$88					; 
		dc.b	$3F					; ?
		dc.b	$27					; '
		dc.b	$31					; 1
		dc.b	$5B					; [
		dc.b	$9D					; 
		dc.b	$D3					; 
		dc.b	$E3					; 
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$99					; 
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$49					; I
		dc.b	$56					; V
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$97					; 
		dc.b	$9B					; 
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$4B					; K
		dc.b	$65					; e
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$A0					; 
		dc.b	$A3					; 
		dc.b	$7E					; ~
		dc.b	$5A					; Z
		dc.b	$58					; X
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$BB					; 
		dc.b	$C7					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$85					; 
		dc.b	$99					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$A1					; 
		dc.b	$B2					; 
		dc.b	$A3					; 
		dc.b	$6D					; m
		dc.b	$48					; H
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$A9					; 
		dc.b	$B0					; 
		dc.b	$9D					; 
		dc.b	$69					; i
		dc.b	$40					; @
		dc.b	$54					; T
		dc.b	$7E					; ~
		dc.b	$AE					; 
		dc.b	$D3					; 
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$58					; X
		dc.b	$66					; f
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$A8					; 
		dc.b	$BD					; 
		dc.b	$98					; 
		dc.b	$70					; p
		dc.b	$4E					; N
		dc.b	$47					; G
		dc.b	$6C					; l
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$7E					; ~
		dc.b	$93					; 
		dc.b	$A5					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$8C					; 
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$66					; f
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$51					; Q
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$82					; 
		dc.b	$B0					; 
		dc.b	$D5					; 
		dc.b	$BD					; 
		dc.b	$93					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$77					; w
		dc.b	$9A					; 
		dc.b	$9E					; 
		dc.b	$93					; 
		dc.b	$78					; x
		dc.b	$5B					; [
		dc.b	$55					; U
		dc.b	$5B					; [
		dc.b	$73					; s
		dc.b	$A1					; 
		dc.b	$C5					; 
		dc.b	$B2					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$62					; b
		dc.b	$49					; I
		dc.b	$45					; E
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$BC					; 
		dc.b	$C7					; 
		dc.b	$A0					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$5B					; [
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$AA					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$5B					; [
		dc.b	$41					; A
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$4B					; K
		dc.b	$83					; 
		dc.b	$AB					; 
		dc.b	$CB					; 
		dc.b	$E3					; 
		dc.b	$D2					; 
		dc.b	$A8					; 
		dc.b	$71					; q
		dc.b	$40					; @
		dc.b	$31					; 1
		dc.b	$40					; @
		dc.b	$5F					; _
		dc.b	$6F					; o
		dc.b	$74					; t
		dc.b	$96					; 
		dc.b	$B5					; 
		dc.b	$A9					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$5F					; _
		dc.b	$69					; i
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$69					; i
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$B9					; 
		dc.b	$CE					; 
		dc.b	$CC					; 
		dc.b	$B6					; 
		dc.b	$7C					; |
		dc.b	$40					; @
		dc.b	$32					; 2
		dc.b	$35					; 5
		dc.b	$3F					; ?
		dc.b	$5E					; ^
		dc.b	$74					; t
		dc.b	$8E					; 
		dc.b	$BC					; 
		dc.b	$D1					; 
		dc.b	$CC					; 
		dc.b	$AF					; 
		dc.b	$77					; w
		dc.b	$48					; H
		dc.b	$42					; B
		dc.b	$52					; R
		dc.b	$5F					; _
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$70					; p
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$5F					; _
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$B7					; 
		dc.b	$CD					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$4C					; L
		dc.b	$2B					; +
		dc.b	$35					; 5
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$B9					; 
		dc.b	$AB					; 
		dc.b	$9A					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$6E					; n
		dc.b	$42					; B
		dc.b	$5F					; _
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$AB					; 
		dc.b	$82					; 
		dc.b	$58					; X
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$97					; 
		dc.b	$B3					; 
		dc.b	$B6					; 
		dc.b	$A1					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$66					; f
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$77					; w
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$43					; C
		dc.b	$3D					; =
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$AD					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$74					; t
		dc.b	$5C					; \
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$74					; t
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$69					; i
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$B0					; 
		dc.b	$C7					; 
		dc.b	$A8					; 
		dc.b	$7C					; |
		dc.b	$4B					; K
		dc.b	$47					; G
		dc.b	$82					; 
		dc.b	$A8					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$B8					; 
		dc.b	$C5					; 
		dc.b	$B0					; 
		dc.b	$7B					; {
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$5B					; [
		dc.b	$5A					; Z
		dc.b	$82					; 
		dc.b	$9C					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$77					; w
		dc.b	$51					; Q
		dc.b	$4E					; N
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$79					; y
		dc.b	$50					; P
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$AA					; 
		dc.b	$BD					; 
		dc.b	$B9					; 
		dc.b	$AC					; 
		dc.b	$87					; 
		dc.b	$66					; f
		dc.b	$52					; R
		dc.b	$3C					; <
		dc.b	$4F					; O
		dc.b	$88					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$6D					; m
		dc.b	$58					; X
		dc.b	$76					; v
		dc.b	$A3					; 
		dc.b	$B5					; 
		dc.b	$A6					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$A5					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$5F					; _
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$87					; 
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$8B					; 
		dc.b	$9E					; 
		dc.b	$9F					; 
		dc.b	$9A					; 
		dc.b	$9A					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$5F					; _
		dc.b	$5B					; [
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$C7					; 
		dc.b	$AC					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$6D					; m
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$82					; 
		dc.b	$97					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$C4					; 
		dc.b	$C5					; 
		dc.b	$A8					; 
		dc.b	$70					; p
		dc.b	$39					; 9
		dc.b	$3C					; <
		dc.b	$62					; b
		dc.b	$8E					; 
		dc.b	$A3					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$A1					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$95					; 
		dc.b	$6D					; m
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$3A					; :
		dc.b	$59					; Y
		dc.b	$85					; 
		dc.b	$B5					; 
		dc.b	$D6					; 
		dc.b	$CC					; 
		dc.b	$94					; 
		dc.b	$58					; X
		dc.b	$3F					; ?
		dc.b	$4C					; L
		dc.b	$7A					; z
		dc.b	$9E					; 
		dc.b	$A1					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$97					; 
		dc.b	$7B					; {
		dc.b	$43					; C
		dc.b	$24					; $
		dc.b	$37					; 7
		dc.b	$54					; T
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$AA					; 
		dc.b	$BF					; 
		dc.b	$CB					; 
		dc.b	$D6					; 
		dc.b	$C8					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$4A					; J
		dc.b	$38					; 8
		dc.b	$30					; 0
		dc.b	$35					; 5
		dc.b	$51					; Q
		dc.b	$7F					; 
		dc.b	$B0					; 
		dc.b	$C3					; 
		dc.b	$AE					; 
		dc.b	$9D					; 
		dc.b	$96					; 
		dc.b	$93					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$47					; G
		dc.b	$4B					; K
		dc.b	$42					; B
		dc.b	$56					; V
		dc.b	$9B					; 
		dc.b	$C5					; 
		dc.b	$CB					; 
		dc.b	$B9					; 
		dc.b	$7F					; 
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$4B					; K
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$A1					; 
		dc.b	$A6					; 
		dc.b	$9A					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$B4					; 
		dc.b	$96					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$3F					; ?
		dc.b	$5B					; [
		dc.b	$98					; 
		dc.b	$C0					; 
		dc.b	$BD					; 
		dc.b	$AE					; 
		dc.b	$89					; 
		dc.b	$47					; G
		dc.b	$2D					; -
		dc.b	$40					; @
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$82					; 
		dc.b	$A4					; 
		dc.b	$C8					; 
		dc.b	$CB					; 
		dc.b	$BB					; 
		dc.b	$9F					; 
		dc.b	$6A					; j
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$9C					; 
		dc.b	$97					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$89					; 
		dc.b	$9C					; 
		dc.b	$8D					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$8F					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$85					; 
		dc.b	$93					; 
		dc.b	$AC					; 
		dc.b	$BF					; 
		dc.b	$9B					; 
		dc.b	$65					; e
		dc.b	$4C					; L
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$69					; i
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$AA					; 
		dc.b	$A1					; 
		dc.b	$6B					; k
		dc.b	$4D					; M
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$A4					; 
		dc.b	$A8					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$70					; p
		dc.b	$82					; 
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$7B					; {
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$7B					; {
		dc.b	$55					; U
		dc.b	$38					; 8
		dc.b	$38					; 8
		dc.b	$61					; a
		dc.b	$AC					; 
		dc.b	$D6					; 
		dc.b	$C0					; 
		dc.b	$99					; 
		dc.b	$74					; t
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$93					; 
		dc.b	$96					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$7B					; {
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$9D					; 
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$85					; 
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$85					; 
		dc.b	$A4					; 
		dc.b	$B3					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$65					; e
		dc.b	$3B					; ;
		dc.b	$2C					; ,
		dc.b	$3E					; >
		dc.b	$70					; p
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$A5					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$69					; i
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$89					; 
		dc.b	$9E					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$6D					; m
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$83					; 
		dc.b	$B2					; 
		dc.b	$BD					; 
		dc.b	$9D					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$AF					; 
		dc.b	$B6					; 
		dc.b	$9C					; 
		dc.b	$7E					; ~
		dc.b	$66					; f
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$A4					; 
		dc.b	$CB					; 
		dc.b	$C6					; 
		dc.b	$91					; 
		dc.b	$46					; F
		dc.b	$24					; $
		dc.b	$31					; 1
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$BC					; 
		dc.b	$C0					; 
		dc.b	$9B					; 
		dc.b	$66					; f
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$A1					; 
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$74					; t
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$74					; t
		dc.b	$50					; P
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$9C					; 
		dc.b	$AC					; 
		dc.b	$B5					; 
		dc.b	$B0					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$5C					; \
		dc.b	$40					; @
		dc.b	$46					; F
		dc.b	$60					; `
		dc.b	$93					; 
		dc.b	$B7					; 
		dc.b	$A8					; 
		dc.b	$86					; 
		dc.b	$77					; w
		dc.b	$8D					; 
		dc.b	$AA					; 
		dc.b	$B0					; 
		dc.b	$A9					; 
		dc.b	$7E					; ~
		dc.b	$3D					; =
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$AB					; 
		dc.b	$B6					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$53					; S
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$B3					; 
		dc.b	$9E					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$74					; t
		dc.b	$60					; `
		dc.b	$57					; W
		dc.b	$4A					; J
		dc.b	$5B					; [
		dc.b	$95					; 
		dc.b	$BC					; 
		dc.b	$C1					; 
		dc.b	$BA					; 
		dc.b	$96					; 
		dc.b	$6A					; j
		dc.b	$54					; T
		dc.b	$48					; H
		dc.b	$59					; Y
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$85					; 
		dc.b	$96					; 
		dc.b	$97					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$A5					; 
		dc.b	$C2					; 
		dc.b	$BF					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$62					; b
		dc.b	$4A					; J
		dc.b	$46					; F
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$97					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$89					; 
		dc.b	$5B					; [
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$93					; 
		dc.b	$C5					; 
		dc.b	$B4					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$69					; i
		dc.b	$5F					; _
		dc.b	$5B					; [
		dc.b	$6C					; l
		dc.b	$85					; 
		dc.b	$9B					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$9A					; 
		dc.b	$A3					; 
		dc.b	$AA					; 
		dc.b	$99					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$85					; 
		dc.b	$B2					; 
		dc.b	$B1					; 
		dc.b	$9E					; 
		dc.b	$80					; 
		dc.b	$54					; T
		dc.b	$56					; V
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$4A					; J
		dc.b	$6C					; l
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$9A					; 
		dc.b	$AB					; 
		dc.b	$A8					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$5F					; _
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$92					; 
		dc.b	$A1					; 
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$B8					; 
		dc.b	$9E					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$4B					; K
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$77					; w
		dc.b	$6D					; m
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$8B					; 
		dc.b	$A6					; 
		dc.b	$AC					; 
		dc.b	$9A					; 
		dc.b	$7E					; ~
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$69					; i
		dc.b	$89					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$85					; 
		dc.b	$AB					; 
		dc.b	$BF					; 
		dc.b	$A0					; 
		dc.b	$6D					; m
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$9E					; 
		dc.b	$A1					; 
		dc.b	$84					; 
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$B8					; 
		dc.b	$98					; 
		dc.b	$62					; b
		dc.b	$52					; R
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$74					; t
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$9E					; 
		dc.b	$B7					; 
		dc.b	$AB					; 
		dc.b	$8B					; 
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$4A					; J
		dc.b	$4B					; K
		dc.b	$7B					; {
		dc.b	$B6					; 
		dc.b	$B1					; 
		dc.b	$74					; t
		dc.b	$4E					; N
		dc.b	$4A					; J
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$9C					; 
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$69					; i
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$7B					; {
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$A1					; 
		dc.b	$82					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$39					; 9
		dc.b	$46					; F
		dc.b	$6E					; n
		dc.b	$A3					; 
		dc.b	$C6					; 
		dc.b	$BB					; 
		dc.b	$9E					; 
		dc.b	$97					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$5F					; _
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$74					; t
		dc.b	$53					; S
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$A9					; 
		dc.b	$93					; 
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$97					; 
		dc.b	$A6					; 
		dc.b	$AD					; 
		dc.b	$95					; 
		dc.b	$70					; p
		dc.b	$46					; F
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$CC					; 
		dc.b	$BD					; 
		dc.b	$8C					; 
		dc.b	$5E					; ^
		dc.b	$47					; G
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$4C					; L
		dc.b	$45					; E
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$BA					; 
		dc.b	$BE					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$5B					; [
		dc.b	$68					; h
		dc.b	$85					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$75					; u
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$99					; 
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$4C					; L
		dc.b	$4C					; L
		dc.b	$70					; p
		dc.b	$A2					; 
		dc.b	$C0					; 
		dc.b	$B4					; 
		dc.b	$91					; 
		dc.b	$6B					; k
		dc.b	$44					; D
		dc.b	$37					; 7
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$AC					; 
		dc.b	$BC					; 
		dc.b	$B7					; 
		dc.b	$AB					; 
		dc.b	$9E					; 
		dc.b	$9C					; 
		dc.b	$82					; 
		dc.b	$51					; Q
		dc.b	$3A					; :
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$97					; 
		dc.b	$AC					; 
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$47					; G
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$9A					; 
		dc.b	$9A					; 
		dc.b	$8C					; 
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$7C					; |
		dc.b	$50					; P
		dc.b	$57					; W
		dc.b	$7B					; {
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$5F					; _
		dc.b	$58					; X
		dc.b	$61					; a
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$A7					; 
		dc.b	$BC					; 
		dc.b	$A4					; 
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$69					; i
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$77					; w
		dc.b	$4C					; L
		dc.b	$37					; 7
		dc.b	$4B					; K
		dc.b	$7E					; ~
		dc.b	$AD					; 
		dc.b	$C1					; 
		dc.b	$AD					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$4E					; N
		dc.b	$5B					; [
		dc.b	$81					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$69					; i
		dc.b	$50					; P
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$B0					; 
		dc.b	$AF					; 
		dc.b	$97					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$98					; 
		dc.b	$89					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$97					; 
		dc.b	$B4					; 
		dc.b	$BE					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$4E					; N
		dc.b	$57					; W
		dc.b	$74					; t
		dc.b	$93					; 
		dc.b	$AD					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$53					; S
		dc.b	$61					; a
		dc.b	$7B					; {
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$A2					; 
		dc.b	$A6					; 
		dc.b	$9C					; 
		dc.b	$7E					; ~
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$43					; C
		dc.b	$63					; c
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$94					; 
		dc.b	$BE					; 
		dc.b	$CC					; 
		dc.b	$AB					; 
		dc.b	$7F					; 
		dc.b	$51					; Q
		dc.b	$41					; A
		dc.b	$5F					; _
		dc.b	$89					; 
		dc.b	$9D					; 
		dc.b	$9A					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$94					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$82					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$9A					; 
		dc.b	$BF					; 
		dc.b	$BA					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$58					; X
		dc.b	$63					; c
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$5F					; _
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$AB					; 
		dc.b	$B3					; 
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$63					; c
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$8D					; 
		dc.b	$93					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$93					; 
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$41					; A
		dc.b	$36					; 6
		dc.b	$4D					; M
		dc.b	$80					; 
		dc.b	$B0					; 
		dc.b	$BD					; 
		dc.b	$B6					; 
		dc.b	$A3					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$6D					; m
		dc.b	$57					; W
		dc.b	$59					; Y
		dc.b	$5B					; [
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$9D					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$5B					; [
		dc.b	$4D					; M
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$9B					; 
		dc.b	$97					; 
		dc.b	$8B					; 
		dc.b	$9E					; 
		dc.b	$9D					; 
		dc.b	$86					; 
		dc.b	$69					; i
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$85					; 
		dc.b	$92					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$41					; A
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$98					; 
		dc.b	$D1					; 
		dc.b	$DE					; 
		dc.b	$BA					; 
		dc.b	$94					; 
		dc.b	$6D					; m
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$58					; X
		dc.b	$60					; `
		dc.b	$69					; i
		dc.b	$88					; 
		dc.b	$A5					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$AF					; 
		dc.b	$B9					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$54					; T
		dc.b	$43					; C
		dc.b	$4E					; N
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$7B					; {
		dc.b	$90					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$93					; 
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$86					; 
		dc.b	$A0					; 
		dc.b	$A1					; 
		dc.b	$89					; 
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$6E					; n
		dc.b	$70					; p
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$9E					; 
		dc.b	$77					; w
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$7E					; ~
		dc.b	$9B					; 
		dc.b	$9A					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$5B					; [
		dc.b	$54					; T
		dc.b	$63					; c
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$C0					; 
		dc.b	$AD					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$97					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$B8					; 
		dc.b	$97					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$5B					; [
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$50					; P
		dc.b	$6E					; n
		dc.b	$90					; 
		dc.b	$A8					; 
		dc.b	$B7					; 
		dc.b	$AF					; 
		dc.b	$9D					; 
		dc.b	$A0					; 
		dc.b	$93					; 
		dc.b	$69					; i
		dc.b	$48					; H
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$89					; 
		dc.b	$A3					; 
		dc.b	$A8					; 
		dc.b	$8E					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$9E					; 
		dc.b	$A1					; 
		dc.b	$93					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$58					; X
		dc.b	$55					; U
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$C0					; 
		dc.b	$C1					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$76					; v
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$9D					; 
		dc.b	$D2					; 
		dc.b	$D6					; 
		dc.b	$AD					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$62					; b
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$90					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$9A					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$97					; 
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$5F					; _
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$70					; p
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$A3					; 
		dc.b	$9A					; 
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$45					; E
		dc.b	$5D					; ]
		dc.b	$8E					; 
		dc.b	$AA					; 
		dc.b	$B2					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$54					; T
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$77					; w
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$70					; p
		dc.b	$6C					; l
		dc.b	$69					; i
		dc.b	$65					; e
		dc.b	$79					; y
		dc.b	$9B					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$AC					; 
		dc.b	$AC					; 
		dc.b	$9D					; 
		dc.b	$75					; u
		dc.b	$45					; E
		dc.b	$30					; 0
		dc.b	$39					; 9
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$97					; 
		dc.b	$C6					; 
		dc.b	$D1					; 
		dc.b	$B2					; 
		dc.b	$76					; v
		dc.b	$54					; T
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$90					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$5F					; _
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$77					; w
		dc.b	$8D					; 
		dc.b	$9A					; 
		dc.b	$AB					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$81					; 
		dc.b	$5A					; Z
		dc.b	$46					; F
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$74					; t
		dc.b	$90					; 
		dc.b	$97					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$9A					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$7E					; ~
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$7E					; ~
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$9E					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$8E					; 
		dc.b	$9A					; 
		dc.b	$90					; 
		dc.b	$93					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$7B					; {
		dc.b	$8F					; 
		dc.b	$9E					; 
		dc.b	$97					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$69					; i
		dc.b	$70					; p
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$97					; 
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$51					; Q
		dc.b	$55					; U
		dc.b	$7C					; |
		dc.b	$A9					; 
		dc.b	$BD					; 
		dc.b	$D3					; 
		dc.b	$CD					; 
		dc.b	$A4					; 
		dc.b	$8C					; 
		dc.b	$7F					; 
		dc.b	$70					; p
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$54					; T
		dc.b	$50					; P
		dc.b	$69					; i
		dc.b	$93					; 
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$A4					; 
		dc.b	$82					; 
		dc.b	$61					; a
		dc.b	$62					; b
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$90					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$8C					; 
		dc.b	$77					; w
		dc.b	$6C					; l
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$A4					; 
		dc.b	$B9					; 
		dc.b	$B7					; 
		dc.b	$A7					; 
		dc.b	$7E					; ~
		dc.b	$46					; F
		dc.b	$39					; 9
		dc.b	$45					; E
		dc.b	$5B					; [
		dc.b	$85					; 
		dc.b	$9E					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$74					; t
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$8C					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$89					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$A1					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$87					; 
		dc.b	$6C					; l
		dc.b	$5F					; _
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$69					; i
		dc.b	$7B					; {
		dc.b	$A0					; 
		dc.b	$B7					; 
		dc.b	$B7					; 
		dc.b	$AF					; 
		dc.b	$9A					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$4E					; N
		dc.b	$40					; @
		dc.b	$40					; @
		dc.b	$51					; Q
		dc.b	$78					; x
		dc.b	$9D					; 
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$A8					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$5B					; [
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$90					; 
		dc.b	$6E					; n
		dc.b	$4E					; N
		dc.b	$40					; @
		dc.b	$4F					; O
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$67					; g
		dc.b	$4B					; K
		dc.b	$44					; D
		dc.b	$4A					; J
		dc.b	$70					; p
		dc.b	$95					; 
		dc.b	$AF					; 
		dc.b	$CF					; 
		dc.b	$CA					; 
		dc.b	$AE					; 
		dc.b	$9D					; 
		dc.b	$85					; 
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$40					; @
		dc.b	$41					; A
		dc.b	$5F					; _
		dc.b	$80					; 
		dc.b	$93					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$91					; 
		dc.b	$B1					; 
		dc.b	$BC					; 
		dc.b	$B5					; 
		dc.b	$9B					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$93					; 
		dc.b	$9F					; 
		dc.b	$94					; 
		dc.b	$97					; 
		dc.b	$81					; 
		dc.b	$5F					; _
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$77					; w
		dc.b	$6D					; m
		dc.b	$78					; x
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$9F					; 
		dc.b	$AE					; 
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$6D					; m
		dc.b	$5B					; [
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$AC					; 
		dc.b	$A5					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$82					; 
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$90					; 
		dc.b	$97					; 
		dc.b	$9E					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$62					; b
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$A1					; 
		dc.b	$A6					; 
		dc.b	$93					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$7E					; ~
		dc.b	$76					; v
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$97					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$5B					; [
		dc.b	$52					; R
		dc.b	$5F					; _
		dc.b	$7B					; {
		dc.b	$95					; 
		dc.b	$9A					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$85					; 
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$BB					; 
		dc.b	$BA					; 
		dc.b	$A5					; 
		dc.b	$82					; 
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$A3					; 
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$67					; g
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$90					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$62					; b
		dc.b	$6A					; j
		dc.b	$8C					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$90					; 
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$90					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$97					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$70					; p
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$62					; b
		dc.b	$6D					; m
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$90					; 
		dc.b	$7E					; ~
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$A0					; 
		dc.b	$AB					; 
		dc.b	$AA					; 
		dc.b	$A1					; 
		dc.b	$89					; 
		dc.b	$60					; `
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$4E					; N
		dc.b	$6A					; j
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$90					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$AE					; 
		dc.b	$9E					; 
		dc.b	$8A					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$94					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A5					; 
		dc.b	$9B					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$74					; t
		dc.b	$61					; a
		dc.b	$57					; W
		dc.b	$5B					; [
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$64					; d
		dc.b	$54					; T
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$B0					; 
		dc.b	$A7					; 
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$74					; t
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$A8					; 
		dc.b	$9A					; 
		dc.b	$77					; w
		dc.b	$5D					; ]
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$9E					; 
		dc.b	$90					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$6B					; k
		dc.b	$5F					; _
		dc.b	$74					; t
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$9A					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$7D					; }
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$9A					; 
		dc.b	$86					; 
		dc.b	$70					; p
		dc.b	$61					; a
		dc.b	$69					; i
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$74					; t
		dc.b	$88					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$9D					; 
		dc.b	$A6					; 
		dc.b	$97					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$82					; 
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$65					; e
		dc.b	$6D					; m
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$77					; w
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$69					; i
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$97					; 
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$85					; 
		dc.b	$6D					; m
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$5B					; [
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A1					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$9E					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$89					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$93					; 
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$89					; 
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$8C					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$7E					; ~
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$97					; 
		dc.b	$97					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$6D					; m
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$58					; X
		dc.b	$59					; Y
		dc.b	$82					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$89					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$97					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$90					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$56					; V
		dc.b	$62					; b
		dc.b	$82					; 
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$AD					; 
		dc.b	$A9					; 
		dc.b	$A1					; 
		dc.b	$88					; 
		dc.b	$66					; f
		dc.b	$5B					; [
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$70					; p
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$9E					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$A1					; 
		dc.b	$A1					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$5E					; ^
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$69					; i
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$5F					; _
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$93					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$69					; i
		dc.b	$66					; f
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$6D					; m
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$85					; 
		dc.b	$94					; 
		dc.b	$9E					; 
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$68					; h
		dc.b	$7E					; ~
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$8B					; 
		dc.b	$77					; w
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$77					; w
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$8C					; 
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$70					; p
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$94					; 
		dc.b	$7B					; {
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$89					; 
		dc.b	$95					; 
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$92					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$97					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$74					; t
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$97					; 
		dc.b	$7E					; ~
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$59					; Y
		dc.b	$68					; h
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$8E					; 
		dc.b	$93					; 
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$6D					; m
		dc.b	$66					; f
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$6D					; m
		dc.b	$81					; 
		dc.b	$90					; 
		dc.b	$93					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$7E					; ~
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$97					; 
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$7B					; {
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$90					; 
		dc.b	$7A					; z
		dc.b	$66					; f
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$69					; i
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$66					; f
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$8C					; 
		dc.b	$97					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$5F					; _
		dc.b	$57					; W
		dc.b	$62					; b
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$9A					; 
		dc.b	$91					; 
		dc.b	$93					; 
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$69					; i
		dc.b	$6E					; n
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$8C					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$82					; 
		dc.b	$90					; 
		dc.b	$9A					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$92					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$62					; b
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$A8					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$7B					; {
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$5F					; _
		dc.b	$63					; c
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$70					; p
		dc.b	$76					; v
		dc.b	$82					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$93					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$70					; p
		dc.b	$85					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$93					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$76					; v
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$85					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$7D					; }
		dc.b	$70					; p
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$9E					; 
		dc.b	$9C					; 
		dc.b	$93					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$93					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$82					; 
		dc.b	$8C					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$6D					; m
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$69					; i
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$92					; 
		dc.b	$97					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$93					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7B					; {
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$81					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$93					; 
		dc.b	$95					; 
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$7B					; {
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$8C					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8C					; 
		dc.b	$87					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$74					; t
		dc.b	$74					; t
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$8C					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$74					; t
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$89					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.l	$7F7F7F80
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
PCM_Whooaahh:	dc.b	$80					; 		; DATA XREF: ROM:0001B074o
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$A0					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$64					; d
		dc.b	$4C					; L
		dc.b	$63					; c
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$76					; v
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$53					; S
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$A4					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$91					; 
		dc.b	$B1					; 
		dc.b	$B7					; 
		dc.b	$A2					; 
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$2E					; .
		dc.b	$2D					; -
		dc.b	$49					; I
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$A0					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$9B					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$96					; 
		dc.b	$AE					; 
		dc.b	$AA					; 
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$81					; 
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$76					; v
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$52					; R
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$43					; C
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$9C					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$61					; a
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$B6					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$49					; I
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$73					; s
		dc.b	$A9					; 
		dc.b	$C6					; 
		dc.b	$B8					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$42					; B
		dc.b	$2C					; ,
		dc.b	$42					; B
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$A0					; 
		dc.b	$BD					; 
		dc.b	$B6					; 
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$3F					; ?
		dc.b	$49					; I
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$B6					; 
		dc.b	$C2					; 
		dc.b	$B3					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$A4					; 
		dc.b	$95					; 
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$3E					; >
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$8F					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$BA					; 
		dc.b	$9B					; 
		dc.b	$65					; e
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$94					; 
		dc.b	$B6					; 
		dc.b	$A6					; 
		dc.b	$9B					; 
		dc.b	$AE					; 
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$7A					; z
		dc.b	$59					; Y
		dc.b	$50					; P
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$AA					; 
		dc.b	$BF					; 
		dc.b	$BB					; 
		dc.b	$87					; 
		dc.b	$55					; U
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$B8					; 
		dc.b	$A9					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$B0					; 
		dc.b	$B4					; 
		dc.b	$B7					; 
		dc.b	$A6					; 
		dc.b	$6B					; k
		dc.b	$41					; A
		dc.b	$3F					; ?
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$C4					; 
		dc.b	$D0					; 
		dc.b	$B1					; 
		dc.b	$68					; h
		dc.b	$2D					; -
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	$59					; Y
		dc.b	$3A					; :
		dc.b	$41					; A
		dc.b	$53					; S
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$4A					; J
		dc.b	$31					; 1
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$BE					; 
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$3A					; :
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$65					; e
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$D3					; 
		dc.b	$D4					; 
		dc.b	$AB					; 
		dc.b	$A7					; 
		dc.b	$AB					; 
		dc.b	$8F					; 
		dc.b	$60					; `
		dc.b	$38					; 8
		dc.b	$33					; 3
		dc.b	$3B					; ;
		dc.b	$33					; 3
		dc.b	$2C					; ,
		dc.b	$34					; 4
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$A4					; 
		dc.b	$DA					; 
		dc.b	$E8					; 
		dc.b	$C2					; 
		dc.b	$B6					; 
		dc.b	$A7					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$2D					; -
		dc.b	$17
		dc.b	$1F
		dc.b	$4F					; O
		dc.b	$83					; 
		dc.b	$B8					; 
		dc.b	$E1					; 
		dc.b	$E7					; 
		dc.b	$DD					; 
		dc.b	$C5					; 
		dc.b	$9B					; 
		dc.b	$6B					; k
		dc.b	$37					; 7
		dc.b	$2A					; *
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$41					; A
		dc.b	$2D					; -
		dc.b	$38					; 8
		dc.b	$60					; `
		dc.b	$99					; 
		dc.b	$C6					; 
		dc.b	$CD					; 
		dc.b	$D3					; 
		dc.b	$DD					; 
		dc.b	$BA					; 
		dc.b	$79					; y
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.b	$48					; H
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$BB					; 
		dc.b	$D0					; 
		dc.b	$BA					; 
		dc.b	$8A					; 
		dc.b	$59					; Y
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$7D					; }
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$43					; C
		dc.b	$49					; I
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$A9					; 
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$AB					; 
		dc.b	$95					; 
		dc.b	$6A					; j
		dc.b	$45					; E
		dc.b	$3B					; ;
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$B4					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$B0					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$42					; B
		dc.b	$3A					; :
		dc.b	$42					; B
		dc.b	$37					; 7
		dc.b	$2E					; .
		dc.b	$3C					; <
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$E1					; 
		dc.b	$EF					; 
		dc.b	$D2					; 
		dc.b	$BE					; 
		dc.b	$B0					; 
		dc.b	$86					; 
		dc.b	$4D					; M
		dc.b	$1C
		dc.b	$1F
		dc.b	$3F					; ?
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$63					; c
		dc.b	$96					; 
		dc.b	$C4					; 
		dc.b	$D0					; 
		dc.b	$D0					; 
		dc.b	$E8					; 
		dc.b	$EE					; 
		dc.b	$B8					; 
		dc.b	$61					; a
		dc.b	$11
		dc.b	1
		dc.b	$E
		dc.b	$17
		dc.b	$3A					; :
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$EB					; 
		dc.b	$DE					; 
		dc.b	$99					; 
		dc.b	$43					; C
		dc.b	$10
		dc.b	1
		dc.b	1
		dc.b	$E
		dc.b	$46					; F
		dc.b	$71					; q
		dc.b	$91					; 
		dc.b	$BB					; 
		dc.b	$D2					; 
		dc.b	$B6					; 
		dc.b	$9C					; 
		dc.b	$B0					; 
		dc.b	$BE					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$6E					; n
		dc.b	$50					; P
		dc.b	$35					; 5
		dc.b	$18
		dc.b	$A
		dc.b	$17
		dc.b	$3B					; ;
		dc.b	$71					; q
		dc.b	$B4					; 
		dc.b	$E7					; 
		dc.b	$E0					; 
		dc.b	$C5					; 
		dc.b	$BB					; 
		dc.b	$A6					; 
		dc.b	$83					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$56					; V
		dc.b	$55					; U
		dc.b	$3F					; ?
		dc.b	$33					; 3
		dc.b	$27					; '
		dc.b	$33					; 3
		dc.b	$5C					; \
		dc.b	$8F					; 
		dc.b	$C8					; 
		dc.b	$D9					; 
		dc.b	$DD					; 
		dc.b	$EB					; 
		dc.b	$C2					; 
		dc.b	$75					; u
		dc.b	$3B					; ;
		dc.b	$2E					; .
		dc.b	$3E					; >
		dc.b	$41					; A
		dc.b	$4A					; J
		dc.b	$52					; R
		dc.b	$4A					; J
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$BD					; 
		dc.b	$DE					; 
		dc.b	$D6					; 
		dc.b	$9D					; 
		dc.b	$4F					; O
		dc.b	$22					; "
		dc.b	$23					; #
		dc.b	$3A					; :
		dc.b	$46					; F
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$91					; 
		dc.b	$B3					; 
		dc.b	$C8					; 
		dc.b	$BE					; 
		dc.b	$BB					; 
		dc.b	$C9					; 
		dc.b	$C4					; 
		dc.b	$8E					; 
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$1F
		dc.b	$1C
		dc.b	$1F
		dc.b	$31					; 1
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$A4					; 
		dc.b	$D2					; 
		dc.b	$E1					; 
		dc.b	$D3					; 
		dc.b	$C6					; 
		dc.b	$CC					; 
		dc.b	$C5					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$29					; )
		dc.b	$23					; #
		dc.b	$1B
		dc.b	$18
		dc.b	$12
		dc.b	$17
		dc.b	$4F					; O
		dc.b	$9B					; 
		dc.b	$C8					; 
		dc.b	$E2					; 
		dc.b	$EB					; 
		dc.b	$E0					; 
		dc.b	$D7					; 
		dc.b	$C9					; 
		dc.b	$9C					; 
		dc.b	$53					; S
		dc.b	$27					; '
		dc.b	$29					; )
		dc.b	$26					; &
		dc.b	$15
		dc.b	$D
		dc.b	$15
		dc.b	$37					; 7
		dc.b	$73					; s
		dc.b	$C2					; 
		dc.b	$F0					; 
		dc.b	$EE					; 
		dc.b	$E9					; 
		dc.b	$EE					; 
		dc.b	$D4					; 
		dc.b	$94					; 
		dc.b	$49					; I
		dc.b	$2C					; ,
		dc.b	$30					; 0
		dc.b	$2C					; ,
		dc.b	$1B
		dc.b	$10
		dc.b	$15
		dc.b	$30					; 0
		dc.b	$63					; c
		dc.b	$99					; 
		dc.b	$B8					; 
		dc.b	$D6					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$C1					; 
		dc.b	$71					; q
		dc.b	$3E					; >
		dc.b	$41					; A
		dc.b	$38					; 8
		dc.b	$19
		dc.b	1
		dc.b	3
		dc.b	$25					; %
		dc.b	$55					; U
		dc.b	$96					; 
		dc.b	$BF					; 
		dc.b	$BF					; 
		dc.b	$CD					; 
		dc.b	$E7					; 
		dc.b	$E8					; 
		dc.b	$B3					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$2C					; ,
		dc.b	$11
		dc.b	$1F
		dc.b	$48					; H
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$AA					; 
		dc.b	$C9					; 
		dc.b	$F2					; 
		dc.b	$E7					; 
		dc.b	$AB					; 
		dc.b	$5A					; Z
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$45					; E
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$88					; 
		dc.b	$9B					; 
		dc.b	$C9					; 
		dc.b	$DA					; 
		dc.b	$BA					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$42					; B
		dc.b	$30					; 0
		dc.b	$29					; )
		dc.b	$35					; 5
		dc.b	$5E					; ^
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$C6					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$4C					; L
		dc.b	$1F
		dc.b	$3A					; :
		dc.b	$71					; q
		dc.b	$AD					; 
		dc.b	$D6					; 
		dc.b	$D0					; 
		dc.b	$C8					; 
		dc.b	$C6					; 
		dc.b	$B4					; 
		dc.b	$78					; x
		dc.b	$19
		dc.b	9
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$6A					; j
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$BE					; 
		dc.b	$D0					; 
		dc.b	$BF					; 
		dc.b	$8F					; 
		dc.b	$3F					; ?
		dc.b	$1C
		dc.b	$26					; &
		dc.b	$27					; '
		dc.b	$2D					; -
		dc.b	$2D					; -
		dc.b	$3C					; <
		dc.b	$71					; q
		dc.b	$B1					; 
		dc.b	$E9					; 
		dc.b	$F0					; 
		dc.b	$C8					; 
		dc.b	$BA					; 
		dc.b	$B4					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$3C					; <
		dc.b	$18
		dc.b	$2E					; .
		dc.b	$42					; B
		dc.b	$35					; 5
		dc.b	$22					; "
		dc.b	$25					; %
		dc.b	$3A					; :
		dc.b	$73					; s
		dc.b	$D9					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$EB					; 
		dc.b	$F0					; 
		dc.b	$CC					; 
		dc.b	$87					; 
		dc.b	$3C					; <
		dc.b	$20
		dc.b	$23					; #
		dc.b	$25					; %
		dc.b	$2A					; *
		dc.b	$1F
		dc.b	$23					; #
		dc.b	$48					; H
		dc.b	$78					; x
		dc.b	$BF					; 
		dc.b	$E5					; 
		dc.b	$D7					; 
		dc.b	$D6					; 
		dc.b	$E2					; 
		dc.b	$D4					; 
		dc.b	$9F					; 
		dc.b	$57					; W
		dc.b	$34					; 4
		dc.b	$2D					; -
		dc.b	$2C					; ,
		dc.b	$29					; )
		dc.b	$11
		dc.b	$11
		dc.b	$38					; 8
		dc.b	$6E					; n
		dc.b	$BF					; 
		dc.b	$F2					; 
		dc.b	$E5					; 
		dc.b	$E7					; 
		dc.b	$F6					; 
		dc.b	$D2					; 
		dc.b	$81					; 
		dc.b	$33					; 3
		dc.b	$1F
		dc.b	$23					; #
		dc.b	$27					; '
		dc.b	$3A					; :
		dc.b	$33					; 3
		dc.b	$27					; '
		dc.b	$3E					; >
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$C2					; 
		dc.b	$BB					; 
		dc.b	$D0					; 
		dc.b	$E1					; 
		dc.b	$DB					; 
		dc.b	$A9					; 
		dc.b	$55					; U
		dc.b	$3E					; >
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$29					; )
		dc.b	$23					; #
		dc.b	$2E					; .
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$C8					; 
		dc.b	$B7					; 
		dc.b	$C4					; 
		dc.b	$D6					; 
		dc.b	$BE					; 
		dc.b	$81					; 
		dc.b	$49					; I
		dc.b	$37					; 7
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$29					; )
		dc.b	$3A					; :
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$D3					; 
		dc.b	$E9					; 
		dc.b	$C8					; 
		dc.b	$C9					; 
		dc.b	$D3					; 
		dc.b	$B3					; 
		dc.b	$75					; u
		dc.b	$2D					; -
		dc.b	$15
		dc.b	$26					; &
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$5E					; ^
		dc.b	$8D					; 
		dc.b	$D2					; 
		dc.b	$EF					; 
		dc.b	$CF					; 
		dc.b	$C5					; 
		dc.b	$D6					; 
		dc.b	$C4					; 
		dc.b	$7C					; |
		dc.b	$2D					; -
		dc.b	$1E
		dc.b	$1B
		dc.b	$1F
		dc.b	$3C					; <
		dc.b	$3C					; <
		dc.b	$3B					; ;
		dc.b	$67					; g
		dc.b	$A6					; 
		dc.b	$DA					; 
		dc.b	$E9					; 
		dc.b	$D7					; 
		dc.b	$CF					; 
		dc.b	$CF					; 
		dc.b	$B3					; 
		dc.b	$73					; s
		dc.b	$29					; )
		dc.b	$A
		dc.b	$10
		dc.b	$10
		dc.b	$B
		dc.b	$18
		dc.b	$43					; C
		dc.b	$6F					; o
		dc.b	$AA					; 
		dc.b	$F6					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$DB					; 
		dc.b	$DE					; 
		dc.b	$C2					; 
		dc.b	$80					; 
		dc.b	$30					; 0
		dc.b	$17
		dc.b	$19
		dc.b	$12
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$37					; 7
		dc.b	$94					; 
		dc.b	$F3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CC					; 
		dc.b	$79					; y
		dc.b	$2A					; *
		dc.b	4
		dc.b	3
		dc.b	$1E
		dc.b	$27					; '
		dc.b	$18
		dc.b	$22					; "
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$BD					; 
		dc.b	$E8					; 
		dc.b	$D9					; 
		dc.b	$E8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$AB					; 
		dc.b	$52					; R
		dc.b	$26					; &
		dc.b	$12
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	2
		dc.b	$3B					; ;
		dc.b	$8A					; 
		dc.b	$D0					; 
		dc.b	$F2					; 
		dc.b	$EE					; 
		dc.b	$E5					; 
		dc.b	$E9					; 
		dc.b	$E4					; 
		dc.b	$B1					; 
		dc.b	$63					; c
		dc.b	$41					; A
		dc.b	$37					; 7
		dc.b	$27					; '
		dc.b	$18
		dc.b	4
		dc.b	9
		dc.b	$27					; '
		dc.b	$57					; W
		dc.b	$A4					; 
		dc.b	$D3					; 
		dc.b	$CF					; 
		dc.b	$E4					; 
		dc.b	$FA					; 
		dc.b	$E8					; 
		dc.b	$B1					; 
		dc.b	$6B					; k
		dc.b	$3F					; ?
		dc.b	$35					; 5
		dc.b	$2C					; ,
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$AD					; 
		dc.b	$CD					; 
		dc.b	$BB					; 
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$BE					; 
		dc.b	$A3					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$3B					; ;
		dc.b	$38					; 8
		dc.b	$1C
		dc.b	$1B
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$D3					; 
		dc.b	$D3					; 
		dc.b	$C5					; 
		dc.b	$CD					; 
		dc.b	$C1					; 
		dc.b	$98					; 
		dc.b	$67					; g
		dc.b	$49					; I
		dc.b	$3B					; ;
		dc.b	$3C					; <
		dc.b	$3E					; >
		dc.b	$34					; 4
		dc.b	$2E					; .
		dc.b	$41					; A
		dc.b	$68					; h
		dc.b	$A0					; 
		dc.b	$C8					; 
		dc.b	$C1					; 
		dc.b	$B8					; 
		dc.b	$B7					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$DB					; 
		dc.b	$E1					; 
		dc.b	$BA					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$81					; 
		dc.b	$5A					; Z
		dc.b	$2E					; .
		dc.b	$23					; #
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$94					; 
		dc.b	$D6					; 
		dc.b	$E4					; 
		dc.b	$C8					; 
		dc.b	$BA					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$43					; C
		dc.b	$26					; &
		dc.b	$1B
		dc.b	$29					; )
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$8B					; 
		dc.b	$AE					; 
		dc.b	$DE					; 
		dc.b	$F3					; 
		dc.b	$CF					; 
		dc.b	$B7					; 
		dc.b	$9D					; 
		dc.b	$72					; r
		dc.b	$4C					; L
		dc.b	$26					; &
		dc.b	$20
		dc.b	$20
		dc.b	$22					; "
		dc.b	$37					; 7
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$83					; 
		dc.b	$A2					; 
		dc.b	$E7					; 
		dc.b	$FE					; 
		dc.b	$F7					; 
		dc.b	$D6					; 
		dc.b	$AE					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$23					; #
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$1C
		dc.b	$34					; 4
		dc.b	$53					; S
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E1					; 
		dc.b	$CC					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$43					; C
		dc.b	$12
		dc.b	1
		dc.b	1
		dc.b	3
		dc.b	$19
		dc.b	$2E					; .
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$CF					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$B8					; 
		dc.b	$87					; 
		dc.b	$49					; I
		dc.b	$15
		dc.b	2
		dc.b	1
		dc.b	$A
		dc.b	$26					; &
		dc.b	$3C					; <
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$B3					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$DB					; 
		dc.b	$BD					; 
		dc.b	$A4					; 
		dc.b	$76					; v
		dc.b	$3A					; :
		dc.b	$17
		dc.b	1
		dc.b	1
		dc.b	4
		dc.b	$1C
		dc.b	$41					; A
		dc.b	$68					; h
		dc.b	$AE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$C5					; 
		dc.b	$9D					; 
		dc.b	$6E					; n
		dc.b	$3A					; :
		dc.b	$14
		dc.b	1
		dc.b	3
		dc.b	$20
		dc.b	$2D					; -
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$A0					; 
		dc.b	$E4					; 
		dc.b	$F0					; 
		dc.b	$E1					; 
		dc.b	$E0					; 
		dc.b	$C4					; 
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$3A					; :
		dc.b	$19
		dc.b	$A
		dc.b	$11
		dc.b	$1B
		dc.b	$3A					; :
		dc.b	$61					; a
		dc.b	$9D					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$F2					; 
		dc.b	$DD					; 
		dc.b	$B3					; 
		dc.b	$8D					; 
		dc.b	$5A					; Z
		dc.b	$3B					; ;
		dc.b	$37					; 7
		dc.b	$27					; '
		dc.b	$25					; %
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.b	$3E					; >
		dc.b	$53					; S
		dc.b	$75					; u
		dc.b	$B8					; 
		dc.b	$D7					; 
		dc.b	$DA					; 
		dc.b	$E0					; 
		dc.b	$D0					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$60					; `
		dc.b	$3C					; <
		dc.b	$22					; "
		dc.b	$E
		dc.b	$1B
		dc.b	$1E
		dc.b	$33					; 3
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$D3					; 
		dc.b	$E9					; 
		dc.b	$DA					; 
		dc.b	$D0					; 
		dc.b	$AA					; 
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$3C					; <
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$A9					; 
		dc.b	$C8					; 
		dc.b	$BD					; 
		dc.b	$C1					; 
		dc.b	$B7					; 
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$46					; F
		dc.b	$3A					; :
		dc.b	$30					; 0
		dc.b	$2A					; *
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$C2					; 
		dc.b	$DE					; 
		dc.b	$CD					; 
		dc.b	$BB					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$3E					; >
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$37					; 7
		dc.b	$4C					; L
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$BF					; 
		dc.b	$D4					; 
		dc.b	$CC					; 
		dc.b	$CB					; 
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$38					; 8
		dc.b	$2A					; *
		dc.b	$1C
		dc.b	$18
		dc.b	$3B					; ;
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$B3					; 
		dc.b	$E0					; 
		dc.b	$EC					; 
		dc.b	$C2					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$2E					; .
		dc.b	$33					; 3
		dc.b	$29					; )
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$D2					; 
		dc.b	$FE					; 
		dc.b	$EE					; 
		dc.b	$D7					; 
		dc.b	$C5					; 
		dc.b	$9F					; 
		dc.b	$64					; d
		dc.b	$2D					; -
		dc.b	$17
		dc.b	$E
		dc.b	9
		dc.b	$15
		dc.b	$26					; &
		dc.b	$42					; B
		dc.b	$7A					; z
		dc.b	$AB					; 
		dc.b	$E0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$5A					; Z
		dc.b	$31					; 1
		dc.b	$1C
		dc.b	$22					; "
		dc.b	$20
		dc.b	$30					; 0
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$D4					; 
		dc.b	$FE					; 
		dc.b	$E4					; 
		dc.b	$D2					; 
		dc.b	$CC					; 
		dc.b	$98					; 
		dc.b	$57					; W
		dc.b	$31					; 1
		dc.b	$29					; )
		dc.b	$19
		dc.b	$12
		dc.b	$20
		dc.b	$30					; 0
		dc.b	$49					; I
		dc.b	$6F					; o
		dc.b	$99					; 
		dc.b	$D4					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$E0					; 
		dc.b	$AB					; 
		dc.b	$6F					; o
		dc.b	$2A					; *
		dc.b	$B
		dc.b	$D
		dc.b	3
		dc.b	$11
		dc.b	$2D					; -
		dc.b	$4A					; J
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$BA					; 
		dc.b	$F6					; 
		dc.b	$F7					; 
		dc.b	$E8					; 
		dc.b	$D7					; 
		dc.b	$B1					; 
		dc.b	$81					; 
		dc.b	$42					; B
		dc.b	$20
		dc.b	$15
		dc.b	$E
		dc.b	$19
		dc.b	$2C					; ,
		dc.b	$3C					; <
		dc.b	$59					; Y
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$EB					; 
		dc.b	$E4					; 
		dc.b	$E0					; 
		dc.b	$D4					; 
		dc.b	$AB					; 
		dc.b	$81					; 
		dc.b	$4F					; O
		dc.b	$2C					; ,
		dc.b	$26					; &
		dc.b	$23					; #
		dc.b	$26					; &
		dc.b	$25					; %
		dc.b	$30					; 0
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$DE					; 
		dc.b	$E5					; 
		dc.b	$D7					; 
		dc.b	$DB					; 
		dc.b	$CD					; 
		dc.b	$A6					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$2C					; ,
		dc.b	4
		dc.b	1
		dc.b	9
		dc.b	$19
		dc.b	$45					; E
		dc.b	$79					; y
		dc.b	$B1					; 
		dc.b	$F0					; 
		dc.b	$F7					; 
		dc.b	$E2					; 
		dc.b	$C6					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$46					; F
		dc.b	$3F					; ?
		dc.b	$3A					; :
		dc.b	$3E					; >
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$AB					; 
		dc.b	$C5					; 
		dc.b	$C2					; 
		dc.b	$C1					; 
		dc.b	$B3					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$6A					; j
		dc.b	$4D					; M
		dc.b	$34					; 4
		dc.b	$1F
		dc.b	$1F
		dc.b	$20
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$D0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$55					; U
		dc.b	$3B					; ;
		dc.b	$27					; '
		dc.b	$18
		dc.b	$2A					; *
		dc.b	$4A					; J
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$A6					; 
		dc.b	$D2					; 
		dc.b	$DB					; 
		dc.b	$CC					; 
		dc.b	$B0					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$3E					; >
		dc.b	$2C					; ,
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$5C					; \
		dc.b	$8D					; 
		dc.b	$CD					; 
		dc.b	$E8					; 
		dc.b	$DE					; 
		dc.b	$B4					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$41					; A
		dc.b	$29					; )
		dc.b	$2A					; *
		dc.b	$50					; P
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$BA					; 
		dc.b	$D9					; 
		dc.b	$C5					; 
		dc.b	$A6					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$56					; V
		dc.b	$38					; 8
		dc.b	$25					; %
		dc.b	$35					; 5
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$BE					; 
		dc.b	$E5					; 
		dc.b	$EE					; 
		dc.b	$BD					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$42					; B
		dc.b	$31					; 1
		dc.b	$53					; S
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$C2					; 
		dc.b	$AB					; 
		dc.b	$76					; v
		dc.b	$5E					; ^
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$3F					; ?
		dc.b	$1E
		dc.b	$2A					; *
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$CF					; 
		dc.b	$E1					; 
		dc.b	$F6					; 
		dc.b	$D7					; 
		dc.b	$8A					; 
		dc.b	$52					; R
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$3A					; :
		dc.b	$12
		dc.b	1
		dc.b	$1C
		dc.b	$4F					; O
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$B8					; 
		dc.b	$B4					; 
		dc.b	$C4					; 
		dc.b	$DB					; 
		dc.b	$CF					; 
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$46					; F
		dc.b	$17
		dc.b	$D
		dc.b	$2A					; *
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$E4					; 
		dc.b	$F9					; 
		dc.b	$E0					; 
		dc.b	$9F					; 
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$1C
		dc.b	3
		dc.b	$B
		dc.b	$38					; 8
		dc.b	$7C					; |
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$B6					; 
		dc.b	$D0					; 
		dc.b	$E8					; 
		dc.b	$D0					; 
		dc.b	$92					; 
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$3E					; >
		dc.b	$A
		dc.b	$17
		dc.b	$50					; P
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$BD					; 
		dc.b	$F9					; 
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$29					; )
		dc.b	$10
		dc.b	4
		dc.b	$14
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$D4					; 
		dc.b	$FE					; 
		dc.b	$E1					; 
		dc.b	$9B					; 
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$4A					; J
		dc.b	$22					; "
		dc.b	$10
		dc.b	$17
		dc.b	$4C					; L
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$9B					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$C2					; 
		dc.b	$BF					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$48					; H
		dc.b	$11
		dc.b	3
		dc.b	$20
		dc.b	$46					; F
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$D2					; 
		dc.b	$EB					; 
		dc.b	$D9					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$45					; E
		dc.b	$19
		dc.b	1
		dc.b	2
		dc.b	$25					; %
		dc.b	$42					; B
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$B8					; 
		dc.b	$A6					; 
		dc.b	$95					; 
		dc.b	$9F					; 
		dc.b	$8A					; 
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$1F
		dc.b	$3B					; ;
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$AE					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$88					; 
		dc.b	$52					; R
		dc.b	$2A					; *
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$4F					; O
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$C8					; 
		dc.b	$A9					; 
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$3C					; <
		dc.b	$34					; 4
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$D7					; 
		dc.b	$E1					; 
		dc.b	$99					; 
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$55					; U
		dc.b	$3C					; <
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$C5					; 
		dc.b	$98					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$59					; Y
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$B8					; 
		dc.b	$C2					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$57					; W
		dc.b	$33					; 3
		dc.b	$3C					; <
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$AA					; 
		dc.b	$C5					; 
		dc.b	$B1					; 
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$57					; W
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$BD					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$50					; P
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$BF					; 
		dc.b	$91					; 
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$35					; 5
		dc.b	$30					; 0
		dc.b	$48					; H
		dc.b	$7D					; }
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$A4					; 
		dc.b	$BD					; 
		dc.b	$99					; 
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$5D					; ]
		dc.b	$3A					; :
		dc.b	$3B					; ;
		dc.b	$49					; I
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$6B					; k
		dc.b	$4C					; L
		dc.b	$29					; )
		dc.b	$27					; '
		dc.b	$43					; C
		dc.b	$7C					; |
		dc.b	$A9					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$CB					; 
		dc.b	$E4					; 
		dc.b	$B1					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$2C					; ,
		dc.b	$22					; "
		dc.b	$35					; 5
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$D4					; 
		dc.b	$D9					; 
		dc.b	$A6					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$2C					; ,
		dc.b	$19
		dc.b	$3A					; :
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$59					; Y
		dc.b	$3C					; <
		dc.b	$26					; &
		dc.b	$30					; 0
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$1E
		dc.b	1
		dc.b	$22					; "
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$B8					; 
		dc.b	$F3					; 
		dc.b	$D6					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$67					; g
		dc.b	$35					; 5
		dc.b	4
		dc.b	4
		dc.b	$46					; F
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$B3					; 
		dc.b	$F3					; 
		dc.b	$D9					; 
		dc.b	$A4					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$46					; F
		dc.b	$A
		dc.b	2
		dc.b	$35					; 5
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$9C					; 
		dc.b	$D6					; 
		dc.b	$DB					; 
		dc.b	$AD					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$64					; d
		dc.b	$27					; '
		dc.b	$1F
		dc.b	$45					; E
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$C1					; 
		dc.b	$CD					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$53					; S
		dc.b	$1E
		dc.b	$15
		dc.b	$43					; C
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$B3					; 
		dc.b	$E5					; 
		dc.b	$C2					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$75					; u
		dc.b	$3F					; ?
		dc.b	$29					; )
		dc.b	$37					; 7
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$9D					; 
		dc.b	$C5					; 
		dc.b	$BF					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$5C					; \
		dc.b	$8B					; 
		dc.b	$C5					; 
		dc.b	$D3					; 
		dc.b	$B4					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$8F					; 
		dc.b	$5E					; ^
		dc.b	$3C					; <
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$3A					; :
		dc.b	$3F					; ?
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$88					; 
		dc.b	$BE					; 
		dc.b	$D6					; 
		dc.b	$BE					; 
		dc.b	$98					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$50					; P
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$AA					; 
		dc.b	$C4					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$9F					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$99					; 
		dc.b	$BD					; 
		dc.b	$BA					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$A6					; 
		dc.b	$C9					; 
		dc.b	$AA					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$45					; E
		dc.b	$3B					; ;
		dc.b	$4C					; L
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$C8					; 
		dc.b	$C6					; 
		dc.b	$7D					; }
		dc.b	$3C					; <
		dc.b	$49					; I
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$3C					; <
		dc.b	$3A					; :
		dc.b	$65					; e
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$CC					; 
		dc.b	$DA					; 
		dc.b	$AB					; 
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$48					; H
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$B3					; 
		dc.b	$E1					; 
		dc.b	$CD					; 
		dc.b	$95					; 
		dc.b	$60					; `
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$C4					; 
		dc.b	$D2					; 
		dc.b	$98					; 
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$37					; 7
		dc.b	$4A					; J
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$A3					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$E5					; 
		dc.b	$E1					; 
		dc.b	$A4					; 
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$1E
		dc.b	7
		dc.b	1
		dc.b	$20
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$E1					; 
		dc.b	$E4					; 
		dc.b	$B7					; 
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$20
		dc.b	$D
		dc.b	$23					; #
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$D2					; 
		dc.b	$CB					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$55					; U
		dc.b	$23					; #
		dc.b	$B
		dc.b	$1F
		dc.b	$43					; C
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$BA					; 
		dc.b	$E5					; 
		dc.b	$F2					; 
		dc.b	$C9					; 
		dc.b	$8B					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$11
		dc.b	$1F
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$A9					; 
		dc.b	$D2					; 
		dc.b	$D6					; 
		dc.b	$C2					; 
		dc.b	$A2					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$42					; B
		dc.b	$2E					; .
		dc.b	$3A					; :
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$9D					; 
		dc.b	$E7					; 
		dc.b	$EE					; 
		dc.b	$C6					; 
		dc.b	$A9					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$45					; E
		dc.b	$19
		dc.b	$12
		dc.b	$27					; '
		dc.b	$3F					; ?
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$D0					; 
		dc.b	$F9					; 
		dc.b	$EF					; 
		dc.b	$B0					; 
		dc.b	$76					; v
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$2C					; ,
		dc.b	$E
		dc.b	$1F
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$86					; 
		dc.b	$BA					; 
		dc.b	$E4					; 
		dc.b	$E5					; 
		dc.b	$B7					; 
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$56					; V
		dc.b	$2A					; *
		dc.b	$20
		dc.b	$31					; 1
		dc.b	$3A					; :
		dc.b	$3A					; :
		dc.b	$46					; F
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$C9					; 
		dc.b	$DA					; 
		dc.b	$D2					; 
		dc.b	$B1					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$2C					; ,
		dc.b	$55					; U
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$9C					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$57					; W
		dc.b	$5D					; ]
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$C6					; 
		dc.b	$B4					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$48					; H
		dc.b	$29					; )
		dc.b	$2E					; .
		dc.b	$46					; F
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$99					; 
		dc.b	$C4					; 
		dc.b	$DD					; 
		dc.b	$B8					; 
		dc.b	$73					; s
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$53					; S
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$AB					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$B1					; 
		dc.b	$9B					; 
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$4A					; J
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$AE					; 
		dc.b	$B6					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$B4					; 
		dc.b	$A3					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$38					; 8
		dc.b	$2C					; ,
		dc.b	$35					; 5
		dc.b	$53					; S
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$99					; 
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$56					; V
		dc.b	$3C					; <
		dc.b	$48					; H
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$3A					; :
		dc.b	$3E					; >
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$CF					; 
		dc.b	$C6					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$6B					; k
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$48					; H
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$B4					; 
		dc.b	$E4					; 
		dc.b	$D2					; 
		dc.b	$8D					; 
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$42					; B
		dc.b	$2E					; .
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$9C					; 
		dc.b	$C5					; 
		dc.b	$F0					; 
		dc.b	$D2					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$38					; 8
		dc.b	$23					; #
		dc.b	$26					; &
		dc.b	$48					; H
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$A9					; 
		dc.b	$B8					; 
		dc.b	$CB					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$3B					; ;
		dc.b	$1C
		dc.b	$20
		dc.b	$3E					; >
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$AB					; 
		dc.b	$C2					; 
		dc.b	$C5					; 
		dc.b	$C6					; 
		dc.b	$C5					; 
		dc.b	$B6					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$57					; W
		dc.b	$3A					; :
		dc.b	$1B
		dc.b	$11
		dc.b	$20
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$A6					; 
		dc.b	$D7					; 
		dc.b	$D0					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$3B					; ;
		dc.b	$3B					; ;
		dc.b	$49					; I
		dc.b	$5C					; \
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$9B					; 
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$59					; Y
		dc.b	$3A					; :
		dc.b	$35					; 5
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$BD					; 
		dc.b	$E9					; 
		dc.b	$DA					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$41					; A
		dc.b	$23					; #
		dc.b	$2C					; ,
		dc.b	$46					; F
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$BA					; 
		dc.b	$B8					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$46					; F
		dc.b	$31					; 1
		dc.b	$46					; F
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$B8					; 
		dc.b	$B0					; 
		dc.b	$A0					; 
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$52					; R
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$AA					; 
		dc.b	$BB					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$63					; c
		dc.b	$3F					; ?
		dc.b	$34					; 4
		dc.b	$48					; H
		dc.b	$6E					; n
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$53					; S
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$79					; y
		dc.b	$AD					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$88					; 
		dc.b	$42					; B
		dc.b	$2A					; *
		dc.b	$42					; B
		dc.b	$59					; Y
		dc.b	$55					; U
		dc.b	$4D					; M
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$AB					; 
		dc.b	$BF					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$7D					; }
		dc.b	$4D					; M
		dc.b	$53					; S
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$59					; Y
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$BF					; 
		dc.b	$D0					; 
		dc.b	$B8					; 
		dc.b	$76					; v
		dc.b	$3A					; :
		dc.b	$37					; 7
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$92					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$5D					; ]
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$3C					; <
		dc.b	$3B					; ;
		dc.b	$5D					; ]
		dc.b	$8E					; 
		dc.b	$AD					; 
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$BE					; 
		dc.b	$B8					; 
		dc.b	$88					; 
		dc.b	$52					; R
		dc.b	$43					; C
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$7A					; z
		dc.b	$A3					; 
		dc.b	$B1					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$31					; 1
		dc.b	$14
		dc.b	$27					; '
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$AD					; 
		dc.b	$B7					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$A7					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$3E					; >
		dc.b	$19
		dc.b	$2E					; .
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$B1					; 
		dc.b	$C9					; 
		dc.b	$CF					; 
		dc.b	$A9					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$48					; H
		dc.b	$23					; #
		dc.b	$33					; 3
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$C2					; 
		dc.b	$9D					; 
		dc.b	$65					; e
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$6F					; o
		dc.b	$3E					; >
		dc.b	$33					; 3
		dc.b	$5C					; \
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$A9					; 
		dc.b	$D4					; 
		dc.b	$C1					; 
		dc.b	$8E					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$4A					; J
		dc.b	$5C					; \
		dc.b	$8D					; 
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$C6					; 
		dc.b	$A6					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$50					; P
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$C8					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$AB					; 
		dc.b	$BF					; 
		dc.b	$96					; 
		dc.b	$4C					; L
		dc.b	$29					; )
		dc.b	$3B					; ;
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$4C					; L
		dc.b	$2D					; -
		dc.b	$3F					; ?
		dc.b	$78					; x
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$37					; 7
		dc.b	$5A					; Z
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$57					; W
		dc.b	$3B					; ;
		dc.b	$5E					; ^
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$53					; S
		dc.b	$6B					; k
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$56					; V
		dc.b	$48					; H
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$81					; 
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$3C					; <
		dc.b	$46					; F
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$7F					; 
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$AD					; 
		dc.b	$BA					; 
		dc.b	$8F					; 
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$A7					; 
		dc.b	$C8					; 
		dc.b	$96					; 
		dc.b	$50					; P
		dc.b	$3E					; >
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$5A					; Z
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$61					; a
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$3F					; ?
		dc.b	$2E					; .
		dc.b	$64					; d
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$AE					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$52					; R
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$23					; #
		dc.b	$4D					; M
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$C4					; 
		dc.b	$BE					; 
		dc.b	$8A					; 
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$52					; R
		dc.b	$4C					; L
		dc.b	$46					; F
		dc.b	$35					; 5
		dc.b	$45					; E
		dc.b	$7A					; z
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$49					; I
		dc.b	$3A					; :
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$B4					; 
		dc.b	$BF					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$4C					; L
		dc.b	$29					; )
		dc.b	$31					; 1
		dc.b	$6C					; l
		dc.b	$86					; 
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$72					; r
		dc.b	$AE					; 
		dc.b	$D3					; 
		dc.b	$BD					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$78					; x
		dc.b	$53					; S
		dc.b	$84					; 
		dc.b	$C4					; 
		dc.b	$AD					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$53					; S
		dc.b	$30					; 0
		dc.b	$3C					; <
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$57					; W
		dc.b	$1C
		dc.b	$B
		dc.b	$29					; )
		dc.b	$52					; R
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$B1					; 
		dc.b	$DE					; 
		dc.b	$C2					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$B0					; 
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$3A					; :
		dc.b	$26					; &
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$53					; S
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$A3					; 
		dc.b	$BE					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$BD					; 
		dc.b	$B7					; 
		dc.b	$7A					; z
		dc.b	$4A					; J
		dc.b	$3E					; >
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$68					; h
		dc.b	$A2					; 
		dc.b	$B7					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$95					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$50					; P
		dc.b	$46					; F
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$6F					; o
		dc.b	$4A					; J
		dc.b	$7C					; |
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$6B					; k
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$53					; S
		dc.b	$2A					; *
		dc.b	$3B					; ;
		dc.b	$88					; 
		dc.b	$D2					; 
		dc.b	$C8					; 
		dc.b	$8A					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$67					; g
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$A7					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$64					; d
		dc.b	$56					; V
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$46					; F
		dc.b	$34					; 4
		dc.b	$5E					; ^
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$A4					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$49					; I
		dc.b	$57					; W
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$6E					; n
		dc.b	$27					; '
		dc.b	$1F
		dc.b	$59					; Y
		dc.b	$9C					; 
		dc.b	$C5					; 
		dc.b	$AB					; 
		dc.b	$59					; Y
		dc.b	$43					; C
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$B7					; 
		dc.b	$C1					; 
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$38					; 8
		dc.b	$41					; A
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$49					; I
		dc.b	$79					; y
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$C1					; 
		dc.b	$C4					; 
		dc.b	$88					; 
		dc.b	$49					; I
		dc.b	$2E					; .
		dc.b	$55					; U
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$A6					; 
		dc.b	$C6					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$AE					; 
		dc.b	$87					; 
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$46					; F
		dc.b	$37					; 7
		dc.b	$59					; Y
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$C2					; 
		dc.b	$B1					; 
		dc.b	$59					; Y
		dc.b	$14
		dc.b	$25					; %
		dc.b	$65					; e
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$B0					; 
		dc.b	$91					; 
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$BA					; 
		dc.b	$C2					; 
		dc.b	$81					; 
		dc.b	$3E					; >
		dc.b	$31					; 1
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$59					; Y
		dc.b	$3C					; <
		dc.b	$41					; A
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$B1					; 
		dc.b	$C9					; 
		dc.b	$D4					; 
		dc.b	$D9					; 
		dc.b	$B8					; 
		dc.b	$7D					; }
		dc.b	$50					; P
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$2D					; -
		dc.b	$2E					; .
		dc.b	$45					; E
		dc.b	$6C					; l
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$A7					; 
		dc.b	$D9					; 
		dc.b	$CD					; 
		dc.b	$99					; 
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$55					; U
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$73					; s
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$C5					; 
		dc.b	$D3					; 
		dc.b	$80					; 
		dc.b	$37					; 7
		dc.b	$49					; I
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$55					; U
		dc.b	$1E
		dc.b	$30					; 0
		dc.b	$64					; d
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$BF					; 
		dc.b	$C4					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$56					; V
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$3E					; >
		dc.b	$2A					; *
		dc.b	$3E					; >
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$AD					; 
		dc.b	$C9					; 
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$45					; E
		dc.b	$3E					; >
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$71					; q
		dc.b	$50					; P
		dc.b	$7C					; |
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$3C					; <
		dc.b	$34					; 4
		dc.b	$7C					; |
		dc.b	$BF					; 
		dc.b	$A6					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$A6					; 
		dc.b	$C1					; 
		dc.b	$9D					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$8E					; 
		dc.b	$B4					; 
		dc.b	$94					; 
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$A2					; 
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$46					; F
		dc.b	$6A					; j
		dc.b	$AE					; 
		dc.b	$C1					; 
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$63					; c
		dc.b	$95					; 
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$52					; R
		dc.b	$3A					; :
		dc.b	$63					; c
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$7F					; 
		dc.b	$38					; 8
		dc.b	$2A					; *
		dc.b	$6A					; j
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$56					; V
		dc.b	$4C					; L
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$50					; P
		dc.b	$31					; 1
		dc.b	$37					; 7
		dc.b	$67					; g
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$3E					; >
		dc.b	$2A					; *
		dc.b	$7F					; 
		dc.b	$DE					; 
		dc.b	$D9					; 
		dc.b	$A2					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$6F					; o
		dc.b	$42					; B
		dc.b	$2D					; -
		dc.b	$42					; B
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$42					; B
		dc.b	$35					; 5
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$7C					; |
		dc.b	$B0					; 
		dc.b	$B7					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$B7					; 
		dc.b	$F6					; 
		dc.b	$C5					; 
		dc.b	$55					; U
		dc.b	$1C
		dc.b	$41					; A
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$3E					; >
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$5C					; \
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$CB					; 
		dc.b	$BB					; 
		dc.b	$61					; a
		dc.b	$29					; )
		dc.b	$4F					; O
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$68					; h
		dc.b	$A7					; 
		dc.b	$B3					; 
		dc.b	$9F					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$B6					; 
		dc.b	$C9					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$56					; V
		dc.b	$19
		dc.b	6
		dc.b	$34					; 4
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$9F					; 
		dc.b	$B6					; 
		dc.b	$C1					; 
		dc.b	$BB					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$5A					; Z
		dc.b	$42					; B
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$4C					; L
		dc.b	$92					; 
		dc.b	$D9					; 
		dc.b	$BA					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$81					; 
		dc.b	$49					; I
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$B4					; 
		dc.b	$D0					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$87					; 
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$46					; F
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$8E					; 
		dc.b	$DD					; 
		dc.b	$B0					; 
		dc.b	$4D					; M
		dc.b	$3B					; ;
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$4F					; O
		dc.b	$33					; 3
		dc.b	$56					; V
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$B0					; 
		dc.b	$AA					; 
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$B0					; 
		dc.b	$BA					; 
		dc.b	$75					; u
		dc.b	$3B					; ;
		dc.b	$59					; Y
		dc.b	$79					; y
		dc.b	$4D					; M
		dc.b	$35					; 5
		dc.b	$73					; s
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$8F					; 
		dc.b	$C5					; 
		dc.b	$9F					; 
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$49					; I
		dc.b	$3E					; >
		dc.b	$7C					; |
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$9C					; 
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$6F					; o
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.b	$4F					; O
		dc.b	$5A					; Z
		dc.b	$38					; 8
		dc.b	$41					; A
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$C4					; 
		dc.b	$AD					; 
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$37					; 7
		dc.b	$4D					; M
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$B8					; 
		dc.b	$C8					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$57					; W
		dc.b	$2C					; ,
		dc.b	$50					; P
		dc.b	$A6					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$71					; q
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$64					; d
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$87					; 
		dc.b	$A9					; 
		dc.b	$86					; 
		dc.b	$4A					; J
		dc.b	$3A					; :
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$A2					; 
		dc.b	$83					; 
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$49					; I
		dc.b	$52					; R
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$A4					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$4C					; L
		dc.b	$4A					; J
		dc.b	$8E					; 
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$42					; B
		dc.b	$7C					; |
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$DB					; 
		dc.b	$E2					; 
		dc.b	$7F					; 
		dc.b	$45					; E
		dc.b	$7D					; }
		dc.b	$B0					; 
		dc.b	$7D					; }
		dc.b	$38					; 8
		dc.b	$3C					; <
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$9F					; 
		dc.b	$BA					; 
		dc.b	$94					; 
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$57					; W
		dc.b	$33					; 3
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$43					; C
		dc.b	$15
		dc.b	$3C					; <
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$AB					; 
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$C4					; 
		dc.b	$A0					; 
		dc.b	$3C					; <
		dc.b	$19
		dc.b	$61					; a
		dc.b	$98					; 
		dc.b	$6E					; n
		dc.b	$37					; 7
		dc.b	$53					; S
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$84					; 
		dc.b	$BE					; 
		dc.b	$B4					; 
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$60					; `
		dc.b	$27					; '
		dc.b	$3E					; >
		dc.b	$68					; h
		dc.b	$61					; a
		dc.b	$55					; U
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$B0					; 
		dc.b	$EE					; 
		dc.b	$CB					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$3A					; :
		dc.b	$2C					; ,
		dc.b	$38					; 8
		dc.b	$41					; A
		dc.b	$45					; E
		dc.b	$75					; u
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$CC					; 
		dc.b	$E1					; 
		dc.b	$9F					; 
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$68					; h
		dc.b	$3A					; :
		dc.b	$1F
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$50					; P
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$A2					; 
		dc.b	$E9					; 
		dc.b	$E9					; 
		dc.b	$BB					; 
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$8A					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$4C					; L
		dc.b	$25					; %
		dc.b	$12
		dc.b	$3F					; ?
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$AA					; 
		dc.b	$DA					; 
		dc.b	$B1					; 
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$A3					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$3E					; >
		dc.b	$41					; A
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$60					; `
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$52					; R
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$A2					; 
		dc.b	$71					; q
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$68					; h
		dc.b	$29					; )
		dc.b	$2A					; *
		dc.b	$6B					; k
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$D0					; 
		dc.b	$C2					; 
		dc.b	$9B					; 
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$3B					; ;
		dc.b	$14
		dc.b	$34					; 4
		dc.b	$5D					; ]
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$BB					; 
		dc.b	$C9					; 
		dc.b	$96					; 
		dc.b	$4F					; O
		dc.b	$3F					; ?
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$4D					; M
		dc.b	$99					; 
		dc.b	$D7					; 
		dc.b	$D2					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$4C					; L
		dc.b	$4A					; J
		dc.b	$88					; 
		dc.b	$A0					; 
		dc.b	$6E					; n
		dc.b	$46					; F
		dc.b	$65					; e
		dc.b	$B8					; 
		dc.b	$E1					; 
		dc.b	$A2					; 
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$3E					; >
		dc.b	$49					; I
		dc.b	$81					; 
		dc.b	$A6					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$6A					; j
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$2D					; -
		dc.b	$3E					; >
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$A6					; 
		dc.b	$BE					; 
		dc.b	$8F					; 
		dc.b	$49					; I
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$91					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$79					; y
		dc.b	$56					; V
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$AE					; 
		dc.b	$C2					; 
		dc.b	$9D					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$49					; I
		dc.b	$43					; C
		dc.b	$3F					; ?
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$95					; 
		dc.b	$CF					; 
		dc.b	$C6					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$BA					; 
		dc.b	$A2					; 
		dc.b	$83					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$3C					; <
		dc.b	$29					; )
		dc.b	$4A					; J
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$CD					; 
		dc.b	$DB					; 
		dc.b	$D7					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$7A					; z
		dc.b	$53					; S
		dc.b	3
		dc.b	1
		dc.b	$18
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$AE					; 
		dc.b	$BE					; 
		dc.b	$C5					; 
		dc.b	$B4					; 
		dc.b	$8F					; 
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$3E					; >
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$B7					; 
		dc.b	$B8					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$4C					; L
		dc.b	$30					; 0
		dc.b	$2D					; -
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$AB					; 
		dc.b	$C9					; 
		dc.b	$BB					; 
		dc.b	$8F					; 
		dc.b	$5E					; ^
		dc.b	$53					; S
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$3C					; <
		dc.b	$3E					; >
		dc.b	$5A					; Z
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$56					; V
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$C9					; 
		dc.b	$D3					; 
		dc.b	$C5					; 
		dc.b	$86					; 
		dc.b	$41					; A
		dc.b	$48					; H
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$A9					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$3E					; >
		dc.b	$2A					; *
		dc.b	$5D					; ]
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$56					; V
		dc.b	$7D					; }
		dc.b	$B1					; 
		dc.b	$D0					; 
		dc.b	$BE					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$5E					; ^
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$4D					; M
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$A4					; 
		dc.b	$CD					; 
		dc.b	$BA					; 
		dc.b	$84					; 
		dc.b	$5E					; ^
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$4A					; J
		dc.b	$45					; E
		dc.b	$7D					; }
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$87					; 
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$3B					; ;
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$DD					; 
		dc.b	$FC					; 
		dc.b	$E5					; 
		dc.b	$A9					; 
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$31					; 1
		dc.b	$23					; #
		dc.b	$3A					; :
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$A3					; 
		dc.b	$BF					; 
		dc.b	$94					; 
		dc.b	$6A					; j
		dc.b	$92					; 
		dc.b	$CC					; 
		dc.b	$D3					; 
		dc.b	$C2					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$52					; R
		dc.b	$1F
		dc.b	6
		dc.b	$29					; )
		dc.b	$4C					; L
		dc.b	$48					; H
		dc.b	$41					; A
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$D9					; 
		dc.b	$FD					; 
		dc.b	$DE					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$1B
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$99					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$5A					; Z
		dc.b	$53					; S
		dc.b	$48					; H
		dc.b	$45					; E
		dc.b	$43					; C
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$A7					; 
		dc.b	$C4					; 
		dc.b	$C9					; 
		dc.b	$C4					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$2D					; -
		dc.b	$3A					; :
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$4A					; J
		dc.b	$6A					; j
		dc.b	$94					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$22					; "
		dc.b	$2E					; .
		dc.b	$71					; q
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$7F					; 
		dc.b	$5D					; ]
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$6A					; j
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$A0					; 
		dc.b	$C8					; 
		dc.b	$AE					; 
		dc.b	$61					; a
		dc.b	$30					; 0
		dc.b	$37					; 7
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$4C					; L
		dc.b	$67					; g
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$8E					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$76					; v
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$BD					; 
		dc.b	$AB					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$57					; W
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$63					; c
		dc.b	$3A					; :
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$C8					; 
		dc.b	$A9					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$7D					; }
		dc.b	$43					; C
		dc.b	$2C					; ,
		dc.b	$3F					; ?
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$A0					; 
		dc.b	$BA					; 
		dc.b	$D4					; 
		dc.b	$C6					; 
		dc.b	$8F					; 
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$48					; H
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$56					; V
		dc.b	$31					; 1
		dc.b	$26					; &
		dc.b	$5A					; Z
		dc.b	$A7					; 
		dc.b	$BE					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$C1					; 
		dc.b	$D0					; 
		dc.b	$A9					; 
		dc.b	$6C					; l
		dc.b	$42					; B
		dc.b	$2C					; ,
		dc.b	$2E					; .
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$81					; 
		dc.b	$AA					; 
		dc.b	$DD					; 
		dc.b	$E0					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$67					; g
		dc.b	$2E					; .
		dc.b	$25					; %
		dc.b	$3E					; >
		dc.b	$5D					; ]
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$8D					; 
		dc.b	$CC					; 
		dc.b	$C4					; 
		dc.b	$98					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$BA					; 
		dc.b	$9D					; 
		dc.b	$5E					; ^
		dc.b	$4D					; M
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$A6					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$30					; 0
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$4A					; J
		dc.b	$84					; 
		dc.b	$B7					; 
		dc.b	$A7					; 
		dc.b	$6E					; n
		dc.b	$3F					; ?
		dc.b	$46					; F
		dc.b	$83					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$6E					; n
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$92					; 
		dc.b	$B0					; 
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$42					; B
		dc.b	$23					; #
		dc.b	$27					; '
		dc.b	$5C					; \
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$BA					; 
		dc.b	$CB					; 
		dc.b	$C9					; 
		dc.b	$9C					; 
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$61					; a
		dc.b	$20
		dc.b	$1E
		dc.b	$2E					; .
		dc.b	$38					; 8
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$AB					; 
		dc.b	$B8					; 
		dc.b	$BE					; 
		dc.b	$D6					; 
		dc.b	$E1					; 
		dc.b	$C1					; 
		dc.b	$84					; 
		dc.b	$46					; F
		dc.b	$34					; 4
		dc.b	$57					; W
		dc.b	$67					; g
		dc.b	$33					; 3
		dc.b	1
		dc.b	1
		dc.b	$42					; B
		dc.b	$81					; 
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$E7					; 
		dc.b	$EB					; 
		dc.b	$AD					; 
		dc.b	$5E					; ^
		dc.b	$3F					; ?
		dc.b	$30					; 0
		dc.b	$27					; '
		dc.b	$37					; 7
		dc.b	$50					; P
		dc.b	$45					; E
		dc.b	$29					; )
		dc.b	$31					; 1
		dc.b	$4A					; J
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$C2					; 
		dc.b	$D7					; 
		dc.b	$D3					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$31					; 1
		dc.b	2
		dc.b	4
		dc.b	$29					; )
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$48					; H
		dc.b	$59					; Y
		dc.b	$9D					; 
		dc.b	$C9					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$59					; Y
		dc.b	$45					; E
		dc.b	$48					; H
		dc.b	$48					; H
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$8E					; 
		dc.b	$4F					; O
		dc.b	$3C					; <
		dc.b	$67					; g
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$B0					; 
		dc.b	$C1					; 
		dc.b	$88					; 
		dc.b	$3F					; ?
		dc.b	$27					; '
		dc.b	$3C					; <
		dc.b	$6C					; l
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$56					; V
		dc.b	$4D					; M
		dc.b	$88					; 
		dc.b	$B3					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$B6					; 
		dc.b	$8E					; 
		dc.b	$3E					; >
		dc.b	$12
		dc.b	$2E					; .
		dc.b	$6C					; l
		dc.b	$9D					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$7A					; z
		dc.b	$A2					; 
		dc.b	$CC					; 
		dc.b	$C2					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$AE					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$6A					; j
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$9F					; 
		dc.b	$B4					; 
		dc.b	$8B					; 
		dc.b	$48					; H
		dc.b	$37					; 7
		dc.b	$61					; a
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$50					; P
		dc.b	$1F
		dc.b	$37					; 7
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$9D					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$CC					; 
		dc.b	$D7					; 
		dc.b	$9D					; 
		dc.b	$42					; B
		dc.b	$A
		dc.b	$23					; #
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$57					; W
		dc.b	$42					; B
		dc.b	$33					; 3
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$BB					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$C6					; 
		dc.b	$D6					; 
		dc.b	$BE					; 
		dc.b	$9C					; 
		dc.b	$6F					; o
		dc.b	$2C					; ,
		dc.b	1
		dc.b	$10
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$84					; 
		dc.b	$BD					; 
		dc.b	$D9					; 
		dc.b	$D3					; 
		dc.b	$96					; 
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$8B					; 
		dc.b	$C8					; 
		dc.b	$CD					; 
		dc.b	$91					; 
		dc.b	$48					; H
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$9F					; 
		dc.b	$CD					; 
		dc.b	$D4					; 
		dc.b	$9D					; 
		dc.b	$60					; `
		dc.b	$4A					; J
		dc.b	$5D					; ]
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$7D					; }
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$59					; Y
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$65					; e
		dc.b	$A7					; 
		dc.b	$BA					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$A0					; 
		dc.b	$64					; d
		dc.b	$49					; I
		dc.b	$5E					; ^
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$49					; I
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$94					; 
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$59					; Y
		dc.b	$37					; 7
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$71					; q
		dc.b	$2A					; *
		dc.b	$49					; I
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$95					; 
		dc.b	$6F					; o
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$5A					; Z
		dc.b	$7F					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$87					; 
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$64					; d
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$6A					; j
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$C4					; 
		dc.b	$A9					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$30					; 0
		dc.b	$18
		dc.b	$45					; E
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$A9					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$50					; P
		dc.b	$35					; 5
		dc.b	$5C					; \
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$55					; U
		dc.b	$3B					; ;
		dc.b	$33					; 3
		dc.b	$5E					; ^
		dc.b	$9B					; 
		dc.b	$BA					; 
		dc.b	$C2					; 
		dc.b	$A2					; 
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$DB					; 
		dc.b	$F5					; 
		dc.b	$A0					; 
		dc.b	$45					; E
		dc.b	$1C
		dc.b	7
		dc.b	$19
		dc.b	$50					; P
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$65					; e
		dc.b	$B1					; 
		dc.b	$E8					; 
		dc.b	$E1					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$60					; `
		dc.b	$2E					; .
		dc.b	$10
		dc.b	$29					; )
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$C5					; 
		dc.b	$F7					; 
		dc.b	$EF					; 
		dc.b	$AB					; 
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$4C					; L
		dc.b	$37					; 7
		dc.b	$29					; )
		dc.b	$17
		dc.b	$2E					; .
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$AA					; 
		dc.b	$A3					; 
		dc.b	$96					; 
		dc.b	$A3					; 
		dc.b	$C2					; 
		dc.b	$C8					; 
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$48					; H
		dc.b	$3C					; <
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$A4					; 
		dc.b	$CF					; 
		dc.b	$CB					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$64					; d
		dc.b	$3F					; ?
		dc.b	$2A					; *
		dc.b	$4A					; J
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$AE					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$63					; c
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$7F					; 
		dc.b	$BD					; 
		dc.b	$DE					; 
		dc.b	$B1					; 
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$3C					; <
		dc.b	$3C					; <
		dc.b	$57					; W
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$59					; Y
		dc.b	$80					; 
		dc.b	$AB					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$50					; P
		dc.b	$33					; 3
		dc.b	$5A					; Z
		dc.b	$A4					; 
		dc.b	$AD					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$68					; h
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$42					; B
		dc.b	$5D					; ]
		dc.b	$94					; 
		dc.b	$AD					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$30					; 0
		dc.b	$3E					; >
		dc.b	$72					; r
		dc.b	$A7					; 
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$73					; s
		dc.b	$41					; A
		dc.b	$56					; V
		dc.b	$AD					; 
		dc.b	$E9					; 
		dc.b	$D9					; 
		dc.b	$8D					; 
		dc.b	$3E					; >
		dc.b	$12
		dc.b	$25					; %
		dc.b	$6A					; j
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$4F					; O
		dc.b	$17
		dc.b	$B
		dc.b	$59					; Y
		dc.b	$BE					; 
		dc.b	$E0					; 
		dc.b	$CD					; 
		dc.b	$B3					; 
		dc.b	$8B					; 
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$71					; q
		dc.b	$31					; 1
		dc.b	4
		dc.b	$18
		dc.b	$60					; `
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$55					; U
		dc.b	$2C					; ,
		dc.b	$60					; `
		dc.b	$C6					; 
		dc.b	$F0					; 
		dc.b	$CC					; 
		dc.b	$88					; 
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$9D					; 
		dc.b	$B3					; 
		dc.b	$94					; 
		dc.b	$5D					; ]
		dc.b	$31					; 1
		dc.b	$1F
		dc.b	$46					; F
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$B3					; 
		dc.b	$A0					; 
		dc.b	$76					; v
		dc.b	$52					; R
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$53					; S
		dc.b	$19
		dc.b	$43					; C
		dc.b	$AE					; 
		dc.b	$D7					; 
		dc.b	$B1					; 
		dc.b	$76					; v
		dc.b	$4D					; M
		dc.b	$4C					; L
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$83					; 
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$6F					; o
		dc.b	$38					; 8
		dc.b	$12
		dc.b	$10
		dc.b	$59					; Y
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$8A					; 
		dc.b	$B4					; 
		dc.b	$BA					; 
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$59					; Y
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$B3					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$3C					; <
		dc.b	$3F					; ?
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$52					; R
		dc.b	$78					; x
		dc.b	$B8					; 
		dc.b	$B4					; 
		dc.b	$5A					; Z
		dc.b	$1B
		dc.b	$23					; #
		dc.b	$4D					; M
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$56					; V
		dc.b	$7C					; |
		dc.b	$B7					; 
		dc.b	$DB					; 
		dc.b	$C8					; 
		dc.b	$8E					; 
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$73					; s
		dc.b	$49					; I
		dc.b	$3E					; >
		dc.b	$49					; I
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$56					; V
		dc.b	$2D					; -
		dc.b	$42					; B
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$B4					; 
		dc.b	$AE					; 
		dc.b	$8B					; 
		dc.b	$63					; c
		dc.b	$42					; B
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$4A					; J
		dc.b	$3E					; >
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$7D					; }
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$41					; A
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$49					; I
		dc.b	$3A					; :
		dc.b	$53					; S
		dc.b	$8D					; 
		dc.b	$BF					; 
		dc.b	$CC					; 
		dc.b	$BF					; 
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$64					; d
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$4C					; L
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$5E					; ^
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$9D					; 
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$9F					; 
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$5D					; ]
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$50					; P
		dc.b	$94					; 
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$7C					; |
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$6E					; n
		dc.b	$53					; S
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$49					; I
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$A6					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$33					; 3
		dc.b	$3F					; ?
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$B7					; 
		dc.b	$CF					; 
		dc.b	$A7					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$4D					; M
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$B4					; 
		dc.b	$BE					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$59					; Y
		dc.b	$49					; I
		dc.b	$3F					; ?
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$AB					; 
		dc.b	$DA					; 
		dc.b	$BF					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$81					; 
		dc.b	$59					; Y
		dc.b	$45					; E
		dc.b	$3E					; >
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$94					; 
		dc.b	$BD					; 
		dc.b	$A3					; 
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$8F					; 
		dc.b	$BE					; 
		dc.b	$BB					; 
		dc.b	$96					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$91					; 
		dc.b	$A4					; 
		dc.b	$83					; 
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$AB					; 
		dc.b	$78					; x
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$A4					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$B6					; 
		dc.b	$A7					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$43					; C
		dc.b	$59					; Y
		dc.b	$8A					; 
		dc.b	$A9					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$9F					; 
		dc.b	$5C					; \
		dc.b	$4C					; L
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$48					; H
		dc.b	$53					; S
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$75					; u
		dc.b	$B3					; 
		dc.b	$B0					; 
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$C2					; 
		dc.b	$AD					; 
		dc.b	$52					; R
		dc.b	$2E					; .
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$57					; W
		dc.b	$3B					; ;
		dc.b	$46					; F
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$4A					; J
		dc.b	$2C					; ,
		dc.b	$4A					; J
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$49					; I
		dc.b	$2D					; -
		dc.b	$45					; E
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$D3					; 
		dc.b	$D9					; 
		dc.b	$B6					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$67					; g
		dc.b	$42					; B
		dc.b	$2D					; -
		dc.b	$27					; '
		dc.b	$2A					; *
		dc.b	$3B					; ;
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$A7					; 
		dc.b	$C1					; 
		dc.b	$D7					; 
		dc.b	$CC					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$29					; )
		dc.b	$38					; 8
		dc.b	$5A					; Z
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$9D					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$76					; v
		dc.b	$A9					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$37					; 7
		dc.b	$2C					; ,
		dc.b	$3F					; ?
		dc.b	$6C					; l
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$76					; v
		dc.b	$A3					; 
		dc.b	$A4					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$95					; 
		dc.b	$AA					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$5A					; Z
		dc.b	$49					; I
		dc.b	$64					; d
		dc.b	$A0					; 
		dc.b	$B8					; 
		dc.b	$8D					; 
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$B6					; 
		dc.b	$C2					; 
		dc.b	$94					; 
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$9B					; 
		dc.b	$C5					; 
		dc.b	$C1					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$A0					; 
		dc.b	$7A					; z
		dc.b	$4D					; M
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$DB					; 
		dc.b	$F0					; 
		dc.b	$B8					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$35					; 5
		dc.b	$27					; '
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$B8					; 
		dc.b	$A7					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$55					; U
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$49					; I
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$AD					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$64					; d
		dc.b	$2E					; .
		dc.b	$30					; 0
		dc.b	$5E					; ^
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$B0					; 
		dc.b	$BA					; 
		dc.b	$C4					; 
		dc.b	$A6					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$AD					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$3A					; :
		dc.b	$2A					; *
		dc.b	$33					; 3
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$53					; S
		dc.b	$4C					; L
		dc.b	$4D					; M
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$48					; H
		dc.b	$38					; 8
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$4C					; L
		dc.b	$4D					; M
		dc.b	$56					; V
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$5A					; Z
		dc.b	$3E					; >
		dc.b	$3B					; ;
		dc.b	$3B					; ;
		dc.b	$42					; B
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$A9					; 
		dc.b	$BD					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$43					; C
		dc.b	$3A					; :
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$A2					; 
		dc.b	$C5					; 
		dc.b	$CC					; 
		dc.b	$BF					; 
		dc.b	$B7					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$3B					; ;
		dc.b	$3C					; <
		dc.b	$3F					; ?
		dc.b	$3B					; ;
		dc.b	$3E					; >
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$53					; S
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$3E					; >
		dc.b	$45					; E
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$5A					; Z
		dc.b	$41					; A
		dc.b	$3A					; :
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$B7					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$56					; V
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$B0					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$29					; )
		dc.b	$18
		dc.b	$30					; 0
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$B1					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$4C					; L
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$33					; 3
		dc.b	$4D					; M
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$C9					; 
		dc.b	$D2					; 
		dc.b	$AB					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$46					; F
		dc.b	$38					; 8
		dc.b	$37					; 7
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$9B					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$97					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$56					; V
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$5B					; [
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$89					; 
		dc.b	$A1					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$51					; Q
		dc.b	$53					; S
		dc.b	$58					; X
		dc.b	$63					; c
		dc.b	$68					; h
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$A0					; 
		dc.b	$A1					; 
		dc.b	$A1					; 
		dc.b	$A2					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$51					; Q
		dc.b	$3A					; :
		dc.b	$3B					; ;
		dc.b	$4B					; K
		dc.b	$52					; R
		dc.b	$51					; Q
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$B3					; 
		dc.b	$BB					; 
		dc.b	$B7					; 
		dc.b	$A8					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$58					; X
		dc.b	$4C					; L
		dc.b	$62					; b
		dc.b	$5F					; _
		dc.b	$58					; X
		dc.b	$51					; Q
		dc.b	$4E					; N
		dc.b	$58					; X
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$AD					; 
		dc.b	$CF					; 
		dc.b	$C1					; 
		dc.b	$94					; 
		dc.b	$82					; 
		dc.b	$73					; s
		dc.b	$5F					; _
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$35					; 5
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$6D					; m
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$A4					; 
		dc.b	$93					; 
		dc.b	$9F					; 
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$62					; b
		dc.b	$5F					; _
		dc.b	$5F					; _
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$58					; X
		dc.b	$71					; q
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$A8					; 
		dc.b	$93					; 
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$97					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$90					; 
		dc.b	$AF					; 
		dc.b	$BC					; 
		dc.b	$98					; 
		dc.b	$6F					; o
		dc.b	$66					; f
		dc.b	$59					; Y
		dc.b	$4E					; N
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$30					; 0
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$AC					; 
		dc.b	$BC					; 
		dc.b	$C8					; 
		dc.b	$B9					; 
		dc.b	$8C					; 
		dc.b	$67					; g
		dc.b	$66					; f
		dc.b	$6C					; l
		dc.b	$62					; b
		dc.b	$46					; F
		dc.b	$2D					; -
		dc.b	$26					; &
		dc.b	$39					; 9
		dc.b	$55					; U
		dc.b	$65					; e
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$BF					; 
		dc.b	$D2					; 
		dc.b	$C6					; 
		dc.b	$AD					; 
		dc.b	$9A					; 
		dc.b	$9F					; 
		dc.b	$95					; 
		dc.b	$6C					; l
		dc.b	$3C					; <
		dc.b	$28					; (
		dc.b	$38					; 8
		dc.b	$48					; H
		dc.b	$54					; T
		dc.b	$53					; S
		dc.b	$47					; G
		dc.b	$5D					; ]
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$93					; 
		dc.b	$96					; 
		dc.b	$A5					; 
		dc.b	$9F					; 
		dc.b	$82					; 
		dc.b	$5D					; ]
		dc.b	$4B					; K
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$93					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$4E					; N
		dc.b	$4F					; O
		dc.b	$69					; i
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$70					; p
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$6C					; l
		dc.b	$54					; T
		dc.b	$46					; F
		dc.b	$51					; Q
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$90					; 
		dc.b	$A8					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$6D					; m
		dc.b	$55					; U
		dc.b	$4B					; K
		dc.b	$4C					; L
		dc.b	$5F					; _
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$67					; g
		dc.b	$66					; f
		dc.b	$69					; i
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$58					; X
		dc.b	$57					; W
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$5E					; ^
		dc.b	$74					; t
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$8C					; 
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$97					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$6D					; m
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$6A					; j
		dc.b	$62					; b
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$85					; 
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$6E					; n
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$70					; p
		dc.b	$89					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$86					; 
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$53					; S
		dc.b	$3C					; <
		dc.b	$4B					; K
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$8C					; 
		dc.b	$AA					; 
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$5B					; [
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$76					; v
		dc.b	$82					; 
		dc.b	$91					; 
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$A4					; 
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$3F					; ?
		dc.b	$49					; I
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$9A					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$6D					; m
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$69					; i
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$69					; i
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$A0					; 
		dc.b	$AC					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$9C					; 
		dc.b	$B9					; 
		dc.b	$B2					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$51					; Q
		dc.b	$43					; C
		dc.b	$44					; D
		dc.b	$54					; T
		dc.b	$66					; f
		dc.b	$56					; V
		dc.b	$45					; E
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$7B					; {
		dc.b	$94					; 
		dc.b	$AC					; 
		dc.b	$B5					; 
		dc.b	$AB					; 
		dc.b	$A2					; 
		dc.b	$A1					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$69					; i
		dc.b	$50					; P
		dc.b	$58					; X
		dc.b	$59					; Y
		dc.b	$50					; P
		dc.b	$54					; T
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$A5					; 
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$62					; b
		dc.b	$66					; f
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$69					; i
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$9B					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$58					; X
		dc.b	$4F					; O
		dc.b	$51					; Q
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$A9					; 
		dc.b	$AC					; 
		dc.b	$98					; 
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$55					; U
		dc.b	$62					; b
		dc.b	$88					; 
		dc.b	$A1					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$9A					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$52					; R
		dc.b	$62					; b
		dc.b	$5C					; \
		dc.b	$51					; Q
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$A6					; 
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$5B					; [
		dc.b	$55					; U
		dc.b	$58					; X
		dc.b	$62					; b
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$A5					; 
		dc.b	$9F					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$69					; i
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$66					; f
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$8C					; 
		dc.b	$9E					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$9A					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$82					; 
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$54					; T
		dc.b	$56					; V
		dc.b	$5F					; _
		dc.b	$5F					; _
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$86					; 
		dc.b	$9A					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$90					; 
		dc.b	$A1					; 
		dc.b	$A5					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$7B					; {
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$58					; X
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$93					; 
		dc.b	$90					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$70					; p
		dc.b	$6A					; j
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$69					; i
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$70					; p
		dc.b	$8A					; 
		dc.b	$9E					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$70					; p
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$9B					; 
		dc.b	$A8					; 
		dc.b	$9E					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$41					; A
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$62					; b
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$5F					; _
		dc.b	$60					; `
		dc.b	$77					; w
		dc.b	$8E					; 
		dc.b	$9A					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$89					; 
		dc.b	$93					; 
		dc.b	$93					; 
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$51					; Q
		dc.b	$5F					; _
		dc.b	$76					; v
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$69					; i
		dc.b	$6D					; m
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$82					; 
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$8C					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$85					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$6D					; m
		dc.b	$66					; f
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$97					; 
		dc.b	$93					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$69					; i
		dc.b	$5F					; _
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$5B					; [
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$71					; q
		dc.b	$56					; V
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$A6					; 
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$9A					; 
		dc.b	$86					; 
		dc.b	$74					; t
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$9A					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$93					; 
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$6D					; m
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$9E					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$62					; b
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$77					; w
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$9E					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$58					; X
		dc.b	$5B					; [
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7B					; {
		dc.b	$6D					; m
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$4C					; L
		dc.b	$5B					; [
		dc.b	$6E					; n
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$AA					; 
		dc.b	$A5					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$66					; f
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$77					; w
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$6D					; m
		dc.b	$71					; q
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$5F					; _
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$8C					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$90					; 
		dc.b	$8E					; 
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$5B					; [
		dc.b	$5B					; [
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$9E					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$6D					; m
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$85					; 
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$78					; x
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$66					; f
		dc.b	$6A					; j
		dc.b	$7B					; {
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$70					; p
		dc.b	$6B					; k
		dc.b	$66					; f
		dc.b	$6F					; o
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$69					; i
		dc.b	$67					; g
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$93					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$5F					; _
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$9A					; 
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$6D					; m
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$66					; f
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$93					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$69					; i
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$93					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$8C					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$60					; `
		dc.b	$58					; X
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$93					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$69					; i
		dc.b	$5B					; [
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$93					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$7E					; ~
		dc.b	$72					; r
		dc.b	$66					; f
		dc.b	$5B					; [
		dc.b	$58					; X
		dc.b	$64					; d
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$66					; f
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$70					; p
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$74					; t
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$69					; i
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$69					; i
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$88					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$90					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$8C					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$97					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$93					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$70					; p
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$90					; 
		dc.b	$97					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$74					; t
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$74					; t
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$7B					; {
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
PCM_CrowdCheer:	dc.b	$80					; 		; DATA XREF: ROM:0001B044o
		dc.b	$81					; 
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$62					; b
		dc.b	$45					; E
		dc.b	$4A					; J
		dc.b	$69					; i
		dc.b	$7B					; {
		dc.b	$89					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$B5					; 
		dc.b	$A8					; 
		dc.b	$89					; 
		dc.b	$66					; f
		dc.b	$5F					; _
		dc.b	$61					; a
		dc.b	$62					; b
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$51					; Q
		dc.b	$63					; c
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$54					; T
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$94					; 
		dc.b	$AF					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$77					; w
		dc.b	$64					; d
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$9E					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$58					; X
		dc.b	$66					; f
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$93					; 
		dc.b	$A5					; 
		dc.b	$8A					; 
		dc.b	$6E					; n
		dc.b	$69					; i
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$B4					; 
		dc.b	$9F					; 
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$69					; i
		dc.b	$84					; 
		dc.b	$89					; 
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$5B					; [
		dc.b	$43					; C
		dc.b	$56					; V
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$9A					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$2C					; ,
		dc.b	$20
		dc.b	$60					; `
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$B9					; 
		dc.b	$A2					; 
		dc.b	$60					; `
		dc.b	$52					; R
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$58					; X
		dc.b	$78					; x
		dc.b	$85					; 
		dc.b	$7A					; z
		dc.b	$A7					; 
		dc.b	$BC					; 
		dc.b	$94					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$A1					; 
		dc.b	$AA					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$34					; 4
		dc.b	$68					; h
		dc.b	$A0					; 
		dc.b	$97					; 
		dc.b	$7B					; {
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$7E					; ~
		dc.b	$B4					; 
		dc.b	$BA					; 
		dc.b	$AD					; 
		dc.b	$AC					; 
		dc.b	$8C					; 
		dc.b	$7B					; {
		dc.b	$91					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$2E					; .
		dc.b	$20
		dc.b	$62					; b
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$BD					; 
		dc.b	$B7					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$49					; I
		dc.b	$2D					; -
		dc.b	$54					; T
		dc.b	$78					; x
		dc.b	$62					; b
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$AF					; 
		dc.b	$86					; 
		dc.b	$54					; T
		dc.b	$52					; R
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$5F					; _
		dc.b	$6D					; m
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$CA					; 
		dc.b	$A8					; 
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$89					; 
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$6B					; k
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$9E					; 
		dc.b	$B8					; 
		dc.b	$8D					; 
		dc.b	$5D					; ]
		dc.b	$69					; i
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$AD					; 
		dc.b	$9B					; 
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$9F					; 
		dc.b	$8D					; 
		dc.b	$E
		dc.b	$41					; A
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$D5					; 
		dc.b	$FC					; 
		dc.b	$9E					; 
		dc.b	$24					; $
		dc.b	$7B					; {
		dc.b	$AC					; 
		dc.b	$32					; 2
		dc.b	$69					; i
		dc.b	$A8					; 
		dc.b	$2F					; /
		dc.b	$25					; %
		dc.b	$BD					; 
		dc.b	$C2					; 
		dc.b	$3F					; ?
		dc.b	$7F					; 
		dc.b	$C8					; 
		dc.b	$A5					; 
		dc.b	$86					; 
		dc.b	$C
		dc.b	$1D
		dc.b	$2B					; +
		dc.b	$55					; U
		dc.b	$E7					; 
		dc.b	$FC					; 
		dc.b	$FA					; 
		dc.b	$65					; e
		dc.b	$18
		dc.b	$51					; Q
		dc.b	$B2					; 
		dc.b	$94					; 
		dc.b	$62					; b
		dc.b	$51					; Q
		dc.b	1
		dc.b	$49					; I
		dc.b	$41					; A
		dc.b	$7F					; 
		dc.b	$FC					; 
		dc.b	$DF					; 
		dc.b	$B0					; 
		dc.b	$40					; @
		dc.b	$25					; %
		dc.b	$89					; 
		dc.b	$9E					; 
		dc.b	$5D					; ]
		dc.b	$A8					; 
		dc.b	$EC					; 
		dc.b	$BF					; 
		dc.b	$C6					; 
		dc.b	$FC					; 
		dc.b	$C2					; 
		dc.b	$53					; S
		dc.b	1
		dc.b	$4E					; N
		dc.b	$14
		dc.b	1
		dc.b	$53					; S
		dc.b	$6C					; l
		dc.b	$1C
		dc.b	$55					; U
		dc.b	$B8					; 
		dc.b	$39					; 9
		dc.b	1
		dc.b	$54					; T
		dc.b	$C3					; 
		dc.b	$E6					; 
		dc.b	$C6					; 
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$C0					; 
		dc.b	$FC					; 
		dc.b	$E7					; 
		dc.b	$30					; 0
		dc.b	$3C					; <
		dc.b	$D0					; 
		dc.b	$DC					; 
		dc.b	$5D					; ]
		dc.b	$83					; 
		dc.b	$48					; H
		dc.b	$AF					; 
		dc.b	$FB					; 
		dc.b	$C7					; 
		dc.b	$43					; C
		dc.b	1
		dc.b	1
		dc.b	$77					; w
		dc.b	$D3					; 
		dc.b	$A2					; 
		dc.b	$1F
		dc.b	1
		dc.b	$5B					; [
		dc.b	$B3					; 
		dc.b	$DD					; 
		dc.b	$A2					; 
		dc.b	$54					; T
		dc.b	$6D					; m
		dc.b	$88					; 
		dc.b	$2F					; /
		dc.b	2
		dc.b	$1F
		dc.b	$5A					; Z
		dc.b	$B1					; 
		dc.b	$FB					; 
		dc.b	$FA					; 
		dc.b	$92					; 
		dc.b	$2F					; /
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$61					; a
		dc.b	$58					; X
		dc.b	$3D					; =
		dc.b	$71					; q
		dc.b	$9A					; 
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$A8					; 
		dc.b	$AB					; 
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$DA					; 
		dc.b	$5F					; _
		dc.b	1
		dc.b	9
		dc.b	$59					; Y
		dc.b	$69					; i
		dc.b	$93					; 
		dc.b	$FC					; 
		dc.b	$E3					; 
		dc.b	$82					; 
		dc.b	$F3					; 
		dc.b	$DB					; 
		dc.b	$66					; f
		dc.b	6
		dc.b	1
		dc.b	$14
		dc.b	$81					; 
		dc.b	$EC					; 
		dc.b	$AB					; 
		dc.b	$19
		dc.b	$15
		dc.b	$9F					; 
		dc.b	$F7					; 
		dc.b	$85					; 
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$55					; U
		dc.b	6
		dc.b	$40					; @
		dc.b	$7A					; z
		dc.b	$9A					; 
		dc.b	$D6					; 
		dc.b	$DF					; 
		dc.b	$C8					; 
		dc.b	$6D					; m
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$8C					; 
		dc.b	$7B					; {
		dc.b	$8C					; 
		dc.b	$AF					; 
		dc.b	$7E					; ~
		dc.b	$58					; X
		dc.b	$4E					; N
		dc.b	$65					; e
		dc.b	$5B					; [
		dc.b	$79					; y
		dc.b	$30					; 0
		dc.b	$10
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$A1					; 
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$42					; B
		dc.b	$21					; !
		dc.b	$A8					; 
		dc.b	$B4					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$74					; t
		dc.b	$93					; 
		dc.b	$7E					; ~
		dc.b	$D8					; 
		dc.b	$E2					; 
		dc.b	$CE					; 
		dc.b	$D2					; 
		dc.b	$A6					; 
		dc.b	$68					; h
		dc.b	$19
		dc.b	$31					; 1
		dc.b	$4B					; K
		dc.b	$1A
		dc.b	$37					; 7
		dc.b	$32					; 2
		dc.b	$1E
		dc.b	$55					; U
		dc.b	$86					; 
		dc.b	$A7					; 
		dc.b	$D5					; 
		dc.b	$CC					; 
		dc.b	$98					; 
		dc.b	$4F					; O
		dc.b	$28					; (
		dc.b	$54					; T
		dc.b	$9B					; 
		dc.b	$AF					; 
		dc.b	$CB					; 
		dc.b	$A8					; 
		dc.b	$A1					; 
		dc.b	$D6					; 
		dc.b	$8C					; 
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$96					; 
		dc.b	$68					; h
		dc.b	$82					; 
		dc.b	$34					; 4
		dc.b	$74					; t
		dc.b	$42					; B
		dc.b	$62					; b
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$77					; w
		dc.b	$64					; d
		dc.b	$44					; D
		dc.b	$5A					; Z
		dc.b	$6C					; l
		dc.b	$9D					; 
		dc.b	$61					; a
		dc.b	$34					; 4
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$A6					; 
		dc.b	$4C					; L
		dc.b	$9A					; 
		dc.b	$36					; 6
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$4E					; N
		dc.b	$4B					; K
		dc.b	$96					; 
		dc.b	$28					; (
		dc.b	$1E
		dc.b	$6D					; m
		dc.b	$A8					; 
		dc.b	$BB					; 
		dc.b	$89					; 
		dc.b	$C5					; 
		dc.b	$E0					; 
		dc.b	$DE					; 
		dc.b	$C4					; 
		dc.b	$70					; p
		dc.b	$51					; Q
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$CB					; 
		dc.b	$A2					; 
		dc.b	$7C					; |
		dc.b	$9A					; 
		dc.b	$98					; 
		dc.b	$DE					; 
		dc.b	$9C					; 
		dc.b	$3F					; ?
		dc.b	$19
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$3E					; >
		dc.b	1
		dc.b	$36					; 6
		dc.b	$81					; 
		dc.b	$DF					; 
		dc.b	$FA					; 
		dc.b	$D7					; 
		dc.b	$57					; W
		dc.b	$2E					; .
		dc.b	$46					; F
		dc.b	$7E					; ~
		dc.b	$94					; 
		dc.b	$3B					; ;
		dc.b	1
		dc.b	1
		dc.b	$3D					; =
		dc.b	$93					; 
		dc.b	$67					; g
		dc.b	$AF					; 
		dc.b	$E9					; 
		dc.b	$8D					; 
		dc.b	$9E					; 
		dc.b	$F3					; 
		dc.b	$FC					; 
		dc.b	$BB					; 
		dc.b	$3E					; >
		dc.b	$22					; "
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$4C					; L
		dc.b	2
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	1
		dc.b	$B
		dc.b	$46					; F
		dc.b	$7E					; ~
		dc.b	$AF					; 
		dc.b	$D5					; 
		dc.b	$F7					; 
		dc.b	$FB					; 
		dc.b	$FA					; 
		dc.b	$E8					; 
		dc.b	$EB					; 
		dc.b	$DB					; 
		dc.b	$94					; 
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$58					; X
		dc.b	$30					; 0
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$1F
		dc.b	$91					; 
		dc.b	$E9					; 
		dc.b	$FC					; 
		dc.b	$E0					; 
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$BB					; 
		dc.b	$B5					; 
		dc.b	$AD					; 
		dc.b	$B8					; 
		dc.b	$A9					; 
		dc.b	$66					; f
		dc.b	$30					; 0
		dc.b	1
		dc.b	1
		dc.b	$1C
		dc.b	$5C					; \
		dc.b	$B6					; 
		dc.b	$FD					; 
		dc.b	$DF					; 
		dc.b	$BC					; 
		dc.b	$FD					; 
		dc.b	$C6					; 
		dc.b	$EC					; 
		dc.b	$8D					; 
		dc.b	$AC					; 
		dc.b	$7B					; {
		dc.b	$30					; 0
		dc.b	1
		dc.b	$20
		dc.b	$6E					; n
		dc.b	$ED					; 
		dc.b	$FC					; 
		dc.b	$F5					; 
		dc.b	$97					; 
		dc.b	1
		dc.b	1
		dc.b	$18
		dc.b	$70					; p
		dc.b	$B6					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$36					; 6
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	1
		dc.b	$41					; A
		dc.b	$7B					; {
		dc.b	$B8					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$B6					; 
		dc.b	$D5					; 
		dc.b	$AA					; 
		dc.b	$69					; i
		dc.b	$34					; 4
		dc.b	$1F
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$21					; !
		dc.b	$5F					; _
		dc.b	$A9					; 
		dc.b	$BC					; 
		dc.b	$F3					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$EA					; 
		dc.b	$C6					; 
		dc.b	$89					; 
		dc.b	$56					; V
		dc.b	$24					; $
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$90					; 
		dc.b	$32					; 2
		dc.b	$2C					; ,
		dc.b	$5B					; [
		dc.b	$93					; 
		dc.b	$F7					; 
		dc.b	$FC					; 
		dc.b	$D1					; 
		dc.b	$89					; 
		dc.b	$CA					; 
		dc.b	$BF					; 
		dc.b	$B4					; 
		dc.b	$E3					; 
		dc.b	$C6					; 
		dc.b	$78					; x
		dc.b	3
		dc.b	4
		dc.b	$23					; #
		dc.b	$3B					; ;
		dc.b	$67					; g
		dc.b	$45					; E
		dc.b	$3A					; :
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$5D					; ]
		dc.b	$8C					; 
		dc.b	$BF					; 
		dc.b	$ED					; 
		dc.b	$CD					; 
		dc.b	$A9					; 
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$72					; r
		dc.b	$5F					; _
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$54					; T
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$9E					; 
		dc.b	$95					; 
		dc.b	$5F					; _
		dc.b	$3A					; :
		dc.b	$3A					; :
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$54					; T
		dc.b	$47					; G
		dc.b	$69					; i
		dc.b	$9E					; 
		dc.b	$C6					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$FC					; 
		dc.b	$DB					; 
		dc.b	$9E					; 
		dc.b	$6E					; n
		dc.b	$26					; &
		dc.b	$26					; &
		dc.b	$3B					; ;
		dc.b	$71					; q
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$82					; 
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$93					; 
		dc.b	$A8					; 
		dc.b	$98					; 
		dc.b	$58					; X
		dc.b	$E
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$24					; $
		dc.b	$3B					; ;
		dc.b	$54					; T
		dc.b	$B5					; 
		dc.b	$CA					; 
		dc.b	$BB					; 
		dc.b	$B4					; 
		dc.b	$B2					; 
		dc.b	$97					; 
		dc.b	$AD					; 
		dc.b	$CB					; 
		dc.b	$BC					; 
		dc.b	$94					; 
		dc.b	$7B					; {
		dc.b	$3C					; <
		dc.b	$1B
		dc.b	$2E					; .
		dc.b	$48					; H
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$82					; 
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$74					; t
		dc.b	$98					; 
		dc.b	$67					; g
		dc.b	$37					; 7
		dc.b	$77					; w
		dc.b	$98					; 
		dc.b	$A1					; 
		dc.b	$B5					; 
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$AC					; 
		dc.b	$CB					; 
		dc.b	$AC					; 
		dc.b	$85					; 
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$A5					; 
		dc.b	$AC					; 
		dc.b	$AF					; 
		dc.b	$9A					; 
		dc.b	$5C					; \
		dc.b	$39					; 9
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$6A					; j
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$5F					; _
		dc.b	$74					; t
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$B0					; 
		dc.b	$AC					; 
		dc.b	$94					; 
		dc.b	$90					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$82					; 
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$49					; I
		dc.b	$80					; 
		dc.b	$9A					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$AC					; 
		dc.b	$C7					; 
		dc.b	$BE					; 
		dc.b	$8B					; 
		dc.b	$77					; w
		dc.b	$5F					; _
		dc.b	$51					; Q
		dc.b	$51					; Q
		dc.b	$70					; p
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$A2					; 
		dc.b	$97					; 
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$48					; H
		dc.b	$63					; c
		dc.b	$74					; t
		dc.b	$6C					; l
		dc.b	$58					; X
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$2C					; ,
		dc.b	$20
		dc.b	$3E					; >
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$A1					; 
		dc.b	$C9					; 
		dc.b	$B0					; 
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$87					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$A5					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$8E					; 
		dc.b	$A5					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$5B					; [
		dc.b	$64					; d
		dc.b	$87					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$8E					; 
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$A6					; 
		dc.b	$B9					; 
		dc.b	$CE					; 
		dc.b	$C9					; 
		dc.b	$D0					; 
		dc.b	$9A					; 
		dc.b	$35					; 5
		dc.b	3
		dc.b	$10
		dc.b	$46					; F
		dc.b	$7C					; |
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$B5					; 
		dc.b	$A9					; 
		dc.b	$93					; 
		dc.b	$94					; 
		dc.b	$61					; a
		dc.b	$29					; )
		dc.b	1
		dc.b	$11
		dc.b	$D
		dc.b	$25					; %
		dc.b	$52					; R
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$96					; 
		dc.b	$B2					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$47					; G
		dc.b	$3E					; >
		dc.b	$62					; b
		dc.b	$86					; 
		dc.b	$A2					; 
		dc.b	$C2					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$CD					; 
		dc.b	$EC					; 
		dc.b	$C3					; 
		dc.b	$91					; 
		dc.b	$85					; 
		dc.b	$68					; h
		dc.b	$3B					; ;
		dc.b	$4C					; L
		dc.b	$34					; 4
		dc.b	$2D					; -
		dc.b	$49					; I
		dc.b	$66					; f
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$D0					; 
		dc.b	$A2					; 
		dc.b	$62					; b
		dc.b	$3C					; <
		dc.b	$26					; &
		dc.b	$18
		dc.b	$52					; R
		dc.b	$A1					; 
		dc.b	$70					; p
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$58					; X
		dc.b	$89					; 
		dc.b	$A9					; 
		dc.b	$A8					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$A8					; 
		dc.b	$83					; 
		dc.b	$15
		dc.b	$2B					; +
		dc.b	$9A					; 
		dc.b	$D0					; 
		dc.b	$B3					; 
		dc.b	$5E					; ^
		dc.b	$1A
		dc.b	$1C
		dc.b	$4F					; O
		dc.b	$CC					; 
		dc.b	$C9					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$B4					; 
		dc.b	$AC					; 
		dc.b	$78					; x
		dc.b	$5F					; _
		dc.b	$4C					; L
		dc.b	$3B					; ;
		dc.b	$40					; @
		dc.b	$77					; w
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$A8					; 
		dc.b	$AC					; 
		dc.b	$DF					; 
		dc.b	$E9					; 
		dc.b	$BF					; 
		dc.b	$95					; 
		dc.b	$6D					; m
		dc.b	$25					; %
		dc.b	$12
		dc.b	$20
		dc.b	$23					; #
		dc.b	$26					; &
		dc.b	$38					; 8
		dc.b	$3E					; >
		dc.b	$4D					; M
		dc.b	$94					; 
		dc.b	$D5					; 
		dc.b	$DB					; 
		dc.b	$AF					; 
		dc.b	$89					; 
		dc.b	$71					; q
		dc.b	$9E					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$36					; 6
		dc.b	$46					; F
		dc.b	$9C					; 
		dc.b	$D6					; 
		dc.b	$C1					; 
		dc.b	$6E					; n
		dc.b	$31					; 1
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$36					; 6
		dc.b	$4B					; K
		dc.b	$88					; 
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$3D					; =
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$A6					; 
		dc.b	$DE					; 
		dc.b	$B9					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$BA					; 
		dc.b	$D3					; 
		dc.b	$99					; 
		dc.b	$4F					; O
		dc.b	$4E					; N
		dc.b	$4C					; L
		dc.b	1
		dc.b	1
		dc.b	$12
		dc.b	$41					; A
		dc.b	$97					; 
		dc.b	$D2					; 
		dc.b	$AC					; 
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$A2					; 
		dc.b	$C6					; 
		dc.b	$BE					; 
		dc.b	$8F					; 
		dc.b	$4F					; O
		dc.b	$26					; &
		dc.b	$54					; T
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$82					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$86					; 
		dc.b	$66					; f
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$AD					; 
		dc.b	$CC					; 
		dc.b	$97					; 
		dc.b	$49					; I
		dc.b	$32					; 2
		dc.b	$39					; 9
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$3B					; ;
		dc.b	3
		dc.b	$A
		dc.b	$44					; D
		dc.b	$8F					; 
		dc.b	$A5					; 
		dc.b	$AF					; 
		dc.b	$93					; 
		dc.b	$61					; a
		dc.b	$70					; p
		dc.b	$A9					; 
		dc.b	$DB					; 
		dc.b	$DD					; 
		dc.b	$B4					; 
		dc.b	$9A					; 
		dc.b	$90					; 
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$9A					; 
		dc.b	$AA					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$4C					; L
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$4B					; K
		dc.b	$50					; P
		dc.b	$54					; T
		dc.b	$69					; i
		dc.b	$88					; 
		dc.b	$AC					; 
		dc.b	$AF					; 
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$C8					; 
		dc.b	$92					; 
		dc.b	$6D					; m
		dc.b	$A5					; 
		dc.b	$C3					; 
		dc.b	$A9					; 
		dc.b	$6F					; o
		dc.b	$4D					; M
		dc.b	$42					; B
		dc.b	$1B
		dc.b	$4F					; O
		dc.b	$7C					; |
		dc.b	$4E					; N
		dc.b	$51					; Q
		dc.b	$6A					; j
		dc.b	$46					; F
		dc.b	$78					; x
		dc.b	$B2					; 
		dc.b	$B0					; 
		dc.b	$B1					; 
		dc.b	$A8					; 
		dc.b	$74					; t
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$55					; U
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$A2					; 
		dc.b	$EA					; 
		dc.b	$FC					; 
		dc.b	$FB					; 
		dc.b	$A8					; 
		dc.b	$6D					; m
		dc.b	$5B					; [
		dc.b	$55					; U
		dc.b	$45					; E
		dc.b	$35					; 5
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$93					; 
		dc.b	$C3					; 
		dc.b	$C9					; 
		dc.b	$82					; 
		dc.b	$3C					; <
		dc.b	$25					; %
		dc.b	$34					; 4
		dc.b	$5B					; [
		dc.b	$A1					; 
		dc.b	$DC					; 
		dc.b	$BF					; 
		dc.b	$94					; 
		dc.b	$58					; X
		dc.b	$20
		dc.b	$42					; B
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$D8					; 
		dc.b	$FC					; 
		dc.b	$B0					; 
		dc.b	$45					; E
		dc.b	1
		dc.b	$3E					; >
		dc.b	$7C					; |
		dc.b	$5F					; _
		dc.b	$66					; f
		dc.b	$AF					; 
		dc.b	$C6					; 
		dc.b	$94					; 
		dc.b	$9A					; 
		dc.b	$A3					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$5D					; ]
		dc.b	$1E
		dc.b	$43					; C
		dc.b	$83					; 
		dc.b	$AA					; 
		dc.b	$B9					; 
		dc.b	$85					; 
		dc.b	$46					; F
		dc.b	$34					; 4
		dc.b	$60					; `
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$A1					; 
		dc.b	$B5					; 
		dc.b	$AC					; 
		dc.b	$B0					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$97					; 
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$70					; p
		dc.b	$34					; 4
		dc.b	1
		dc.b	$16
		dc.b	$62					; b
		dc.b	$77					; w
		dc.b	$62					; b
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$9E					; 
		dc.b	$B1					; 
		dc.b	$AF					; 
		dc.b	$8C					; 
		dc.b	$B4					; 
		dc.b	$ED					; 
		dc.b	$B1					; 
		dc.b	$4C					; L
		dc.b	9
		dc.b	3
		dc.b	$3C					; <
		dc.b	$B0					; 
		dc.b	$FC					; 
		dc.b	$C9					; 
		dc.b	$7A					; z
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$72					; r
		dc.b	$45					; E
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$8C					; 
		dc.b	$66					; f
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$BB					; 
		dc.b	$F0					; 
		dc.b	$AD					; 
		dc.b	$4D					; M
		dc.b	$2D					; -
		dc.b	$3F					; ?
		dc.b	$75					; u
		dc.b	$A9					; 
		dc.b	$90					; 
		dc.b	$3E					; >
		dc.b	$1E
		dc.b	$4A					; J
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$A2					; 
		dc.b	$A8					; 
		dc.b	$B2					; 
		dc.b	$A6					; 
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$89					; 
		dc.b	$A5					; 
		dc.b	$9B					; 
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$64					; d
		dc.b	$35					; 5
		dc.b	$3E					; >
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$C9					; 
		dc.b	$C6					; 
		dc.b	$AF					; 
		dc.b	$90					; 
		dc.b	$63					; c
		dc.b	$48					; H
		dc.b	$29					; )
		dc.b	$3F					; ?
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$5F					; _
		dc.b	$70					; p
		dc.b	$7E					; ~
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$59					; Y
		dc.b	$6D					; m
		dc.b	$B0					; 
		dc.b	$B1					; 
		dc.b	$82					; 
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$C3					; 
		dc.b	$C3					; 
		dc.b	$9B					; 
		dc.b	$78					; x
		dc.b	$52					; R
		dc.b	$42					; B
		dc.b	$4A					; J
		dc.b	$59					; Y
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$9F					; 
		dc.b	$7C					; |
		dc.b	$54					; T
		dc.b	$34					; 4
		dc.b	$29					; )
		dc.b	$41					; A
		dc.b	$6A					; j
		dc.b	$C2					; 
		dc.b	$EE					; 
		dc.b	$D0					; 
		dc.b	$8C					; 
		dc.b	$5B					; [
		dc.b	$77					; w
		dc.b	$A5					; 
		dc.b	$B6					; 
		dc.b	$CB					; 
		dc.b	$B6					; 
		dc.b	$78					; x
		dc.b	$58					; X
		dc.b	$5B					; [
		dc.b	$3F					; ?
		dc.b	$2A					; *
		dc.b	$46					; F
		dc.b	$4F					; O
		dc.b	$27					; '
		dc.b	7
		dc.b	$2E					; .
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$97					; 
		dc.b	$EA					; 
		dc.b	$FC					; 
		dc.b	$D0					; 
		dc.b	$6D					; m
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$B9					; 
		dc.b	$BB					; 
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$BC					; 
		dc.b	$BF					; 
		dc.b	$A9					; 
		dc.b	$97					; 
		dc.b	$77					; w
		dc.b	$5B					; [
		dc.b	$58					; X
		dc.b	$66					; f
		dc.b	$49					; I
		dc.b	$16
		dc.b	3
		dc.b	1
		dc.b	$14
		dc.b	$35					; 5
		dc.b	$64					; d
		dc.b	$A6					; 
		dc.b	$E4					; 
		dc.b	$E8					; 
		dc.b	$98					; 
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$74					; t
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$B5					; 
		dc.b	$AC					; 
		dc.b	$97					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$7E					; ~
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$8F					; 
		dc.b	$56					; V
		dc.b	6
		dc.b	1
		dc.b	9
		dc.b	$4D					; M
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$7F					; 
		dc.b	$77					; w
		dc.b	$9F					; 
		dc.b	$D3					; 
		dc.b	$F7					; 
		dc.b	$F7					; 
		dc.b	$D9					; 
		dc.b	$AF					; 
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$98					; 
		dc.b	$57					; W
		dc.b	$16
		dc.b	1
		dc.b	1
		dc.b	$17
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$DD					; 
		dc.b	$DD					; 
		dc.b	$C2					; 
		dc.b	$AE					; 
		dc.b	$78					; x
		dc.b	$3B					; ;
		dc.b	$3A					; :
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$6D					; m
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$9B					; 
		dc.b	$D5					; 
		dc.b	$DB					; 
		dc.b	$80					; 
		dc.b	$48					; H
		dc.b	$60					; `
		dc.b	$59					; Y
		dc.b	$4A					; J
		dc.b	$6A					; j
		dc.b	$9F					; 
		dc.b	$C1					; 
		dc.b	$B6					; 
		dc.b	$95					; 
		dc.b	$66					; f
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$62					; b
		dc.b	$99					; 
		dc.b	$D0					; 
		dc.b	$DC					; 
		dc.b	$BB					; 
		dc.b	$78					; x
		dc.b	$6D					; m
		dc.b	$62					; b
		dc.b	$5F					; _
		dc.b	$69					; i
		dc.b	$5C					; \
		dc.b	$40					; @
		dc.b	$47					; G
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$4C					; L
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$4C					; L
		dc.b	$65					; e
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$B0					; 
		dc.b	$D8					; 
		dc.b	$E9					; 
		dc.b	$DE					; 
		dc.b	$DB					; 
		dc.b	$CD					; 
		dc.b	$B1					; 
		dc.b	$81					; 
		dc.b	$53					; S
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$48					; H
		dc.b	1
		dc.b	1
		dc.b	$33					; 3
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$52					; R
		dc.b	$54					; T
		dc.b	$65					; e
		dc.b	$8A					; 
		dc.b	$A6					; 
		dc.b	$82					; 
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$A9					; 
		dc.b	$B9					; 
		dc.b	$B2					; 
		dc.b	$B2					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$63					; c
		dc.b	$40					; @
		dc.b	$4E					; N
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$46					; F
		dc.b	$6A					; j
		dc.b	$A2					; 
		dc.b	$A5					; 
		dc.b	$A3					; 
		dc.b	$B7					; 
		dc.b	$C1					; 
		dc.b	$B5					; 
		dc.b	$A3					; 
		dc.b	$A1					; 
		dc.b	$AB					; 
		dc.b	$AE					; 
		dc.b	$A1					; 
		dc.b	$93					; 
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$2C					; ,
		dc.b	3
		dc.b	$F
		dc.b	$32					; 2
		dc.b	$28					; (
		dc.b	$38					; 8
		dc.b	$53					; S
		dc.b	$5B					; [
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$4D					; M
		dc.b	$28					; (
		dc.b	$4A					; J
		dc.b	$A5					; 
		dc.b	$C5					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$BF					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$A9					; 
		dc.b	$D6					; 
		dc.b	$DE					; 
		dc.b	$BE					; 
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$53					; S
		dc.b	$27					; '
		dc.b	$1E
		dc.b	$1B
		dc.b	3
		dc.b	1
		dc.b	5
		dc.b	7
		dc.b	$F
		dc.b	$30					; 0
		dc.b	$62					; b
		dc.b	$A2					; 
		dc.b	$D6					; 
		dc.b	$E8					; 
		dc.b	$DA					; 
		dc.b	$C8					; 
		dc.b	$B5					; 
		dc.b	$AD					; 
		dc.b	$C2					; 
		dc.b	$C8					; 
		dc.b	$B1					; 
		dc.b	$CE					; 
		dc.b	$B9					; 
		dc.b	$71					; q
		dc.b	$48					; H
		dc.b	$62					; b
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$69					; i
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$4B					; K
		dc.b	$50					; P
		dc.b	$1B
		dc.b	$2D					; -
		dc.b	$96					; 
		dc.b	$C8					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$B6					; 
		dc.b	$F6					; 
		dc.b	$E3					; 
		dc.b	$AF					; 
		dc.b	$B9					; 
		dc.b	$9D					; 
		dc.b	$49					; I
		dc.b	7
		dc.b	$2B					; +
		dc.b	$89					; 
		dc.b	$AF					; 
		dc.b	$88					; 
		dc.b	$39					; 9
		dc.b	$27					; '
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$4A					; J
		dc.b	$23					; #
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$45					; E
		dc.b	$E
		dc.b	$13
		dc.b	$4E					; N
		dc.b	$A2					; 
		dc.b	$CE					; 
		dc.b	$B8					; 
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$F3					; 
		dc.b	$F4					; 
		dc.b	$AF					; 
		dc.b	$69					; i
		dc.b	$65					; e
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$52					; R
		dc.b	$6D					; m
		dc.b	$5B					; [
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$3A					; :
		dc.b	$53					; S
		dc.b	$91					; 
		dc.b	$C6					; 
		dc.b	$CB					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$98					; 
		dc.b	$D5					; 
		dc.b	$CB					; 
		dc.b	$B3					; 
		dc.b	$B7					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$30					; 0
		dc.b	1
		dc.b	$19
		dc.b	$30					; 0
		dc.b	$F
		dc.b	$13
		dc.b	$11
		dc.b	$63					; c
		dc.b	$9E					; 
		dc.b	$97					; 
		dc.b	$A1					; 
		dc.b	$CC					; 
		dc.b	$E8					; 
		dc.b	$DE					; 
		dc.b	$D7					; 
		dc.b	$DF					; 
		dc.b	$EC					; 
		dc.b	$97					; 
		dc.b	$2E					; .
		dc.b	$2F					; /
		dc.b	$88					; 
		dc.b	$B5					; 
		dc.b	$85					; 
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$A1					; 
		dc.b	$D1					; 
		dc.b	$AB					; 
		dc.b	$47					; G
		dc.b	5
		dc.b	$27					; '
		dc.b	$51					; Q
		dc.b	$33					; 3
		dc.b	5
		dc.b	1
		dc.b	$2A					; *
		dc.b	$89					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$B4					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$D6					; 
		dc.b	$99					; 
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$7E					; ~
		dc.b	$8F					; 
		dc.b	$48					; H
		dc.b	$D
		dc.b	$27					; '
		dc.b	$3A					; :
		dc.b	$18
		dc.b	$1E
		dc.b	$4A					; J
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$5E					; ^
		dc.b	$2F					; /
		dc.b	$6D					; m
		dc.b	$D1					; 
		dc.b	$C4					; 
		dc.b	$61					; a
		dc.b	$52					; R
		dc.b	$99					; 
		dc.b	$EC					; 
		dc.b	$D9					; 
		dc.b	$9A					; 
		dc.b	$A9					; 
		dc.b	$D2					; 
		dc.b	$99					; 
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$42					; B
		dc.b	$31					; 1
		dc.b	$7C					; |
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$44					; D
		dc.b	$31					; 1
		dc.b	$5A					; Z
		dc.b	$38					; 8
		dc.b	4
		dc.b	1
		dc.b	$18
		dc.b	$41					; A
		dc.b	$40					; @
		dc.b	$44					; D
		dc.b	$75					; u
		dc.b	$AC					; 
		dc.b	$CC					; 
		dc.b	$E5					; 
		dc.b	$E8					; 
		dc.b	$F3					; 
		dc.b	$FD					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$D1					; 
		dc.b	$B7					; 
		dc.b	$8C					; 
		dc.b	$58					; X
		dc.b	$21					; !
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$28					; (
		dc.b	$41					; A
		dc.b	3
		dc.b	1
		dc.b	$2E					; .
		dc.b	$A9					; 
		dc.b	$A5					; 
		dc.b	$6F					; o
		dc.b	$7B					; {
		dc.b	$A7					; 
		dc.b	$C9					; 
		dc.b	$C1					; 
		dc.b	$E6					; 
		dc.b	$FA					; 
		dc.b	$C7					; 
		dc.b	$82					; 
		dc.b	$B0					; 
		dc.b	$D4					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$BE					; 
		dc.b	$B2					; 
		dc.b	$7E					; ~
		dc.b	$55					; U
		dc.b	$42					; B
		dc.b	$2D					; -
		dc.b	$2B					; +
		dc.b	$2C					; ,
		dc.b	$1C
		dc.b	$16
		dc.b	1
		dc.b	$A
		dc.b	$23					; #
		dc.b	$15
		dc.b	$37					; 7
		dc.b	$73					; s
		dc.b	$CF					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$D7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A9					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$4B					; K
		dc.b	$19
		dc.b	$39					; 9
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$33					; 3
		dc.b	1
		dc.b	1
		dc.b	$51					; Q
		dc.b	$77					; w
		dc.b	$46					; F
		dc.b	$D
		dc.b	$2B					; +
		dc.b	$92					; 
		dc.b	$C4					; 
		dc.b	$95					; 
		dc.b	$CB					; 
		dc.b	$FE					; 
		dc.b	$AB					; 
		dc.b	$6F					; o
		dc.b	$C3					; 
		dc.b	$EB					; 
		dc.b	$8F					; 
		dc.b	$68					; h
		dc.b	$93					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$84					; 
		dc.b	$6D					; m
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$40					; @
		dc.b	$34					; 4
		dc.b	$30					; 0
		dc.b	$14
		dc.b	$13
		dc.b	$11
		dc.b	$11
		dc.b	$3D					; =
		dc.b	$37					; 7
		dc.b	$6C					; l
		dc.b	$AB					; 
		dc.b	$D5					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$FA					; 
		dc.b	$F6					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C0					; 
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$69					; i
		dc.b	$35					; 5
		dc.b	$4F					; O
		dc.b	$6F					; o
		dc.b	$7E					; ~
		dc.b	$54					; T
		dc.b	8
		dc.b	8
		dc.b	$3E					; >
		dc.b	$68					; h
		dc.b	$3D					; =
		dc.b	6
		dc.b	$24					; $
		dc.b	$A5					; 
		dc.b	$C7					; 
		dc.b	$9E					; 
		dc.b	$9E					; 
		dc.b	$B4					; 
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$86					; 
		dc.b	$D0					; 
		dc.b	$D6					; 
		dc.b	$C1					; 
		dc.b	$BA					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$37					; 7
		dc.b	$20
		dc.b	$13
		dc.b	$11
		dc.b	$A
		dc.b	5
		dc.b	8
		dc.b	$38					; 8
		dc.b	$7A					; z
		dc.b	$BC					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$F2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A6					; 
		dc.b	$B3					; 
		dc.b	$D6					; 
		dc.b	$AB					; 
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$82					; 
		dc.b	$37					; 7
		dc.b	$13
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	2
		dc.b	$74					; t
		dc.b	$8F					; 
		dc.b	$43					; C
		dc.b	$56					; V
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$48					; H
		dc.b	$94					; 
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$E3					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$87					; 
		dc.b	$58					; X
		dc.b	$30					; 0
		dc.b	$32					; 2
		dc.b	$66					; f
		dc.b	$62					; b
		dc.b	$16
		dc.b	$D
		dc.b	$E
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$50					; P
		dc.b	$4F					; O
		dc.b	$A5					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C5					; 
		dc.b	$91					; 
		dc.b	$AF					; 
		dc.b	$F0					; 
		dc.b	$FE					; 
		dc.b	$EF					; 
		dc.b	$C9					; 
		dc.b	$BB					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$41					; A
		dc.b	$39					; 9
		dc.b	$2D					; -
		dc.b	3
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$D
		dc.b	$52					; R
		dc.b	$8E					; 
		dc.b	$BE					; 
		dc.b	$DA					; 
		dc.b	$F3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$D4					; 
		dc.b	$D9					; 
		dc.b	$D4					; 
		dc.b	$B3					; 
		dc.b	$82					; 
		dc.b	$66					; f
		dc.b	$67					; g
		dc.b	$7A					; z
		dc.b	$5B					; [
		dc.b	8
		dc.b	$F
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$C
		dc.b	1
		dc.b	1
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$47					; G
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$C7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FD					; 
		dc.b	$D6					; 
		dc.b	$8F					; 
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$27					; '
		dc.b	$3D					; =
		dc.b	$58					; X
		dc.b	$69					; i
		dc.b	$B1					; 
		dc.b	$E2					; 
		dc.b	$F5					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E1					; 
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$5A					; Z
		dc.b	$42					; B
		dc.b	$2B					; +
		dc.b	$1A
		dc.b	$1A
		dc.b	$B
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$4A					; J
		dc.b	$6F					; o
		dc.b	$4E					; N
		dc.b	$59					; Y
		dc.b	$AE					; 
		dc.b	$CD					; 
		dc.b	$C6					; 
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EF					; 
		dc.b	$5D					; ]
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$14
		dc.b	$80					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$E7					; 
		dc.b	$E9					; 
		dc.b	$C7					; 
		dc.b	$A8					; 
		dc.b	$BD					; 
		dc.b	$A8					; 
		dc.b	$62					; b
		dc.b	$72					; r
		dc.b	$A8					; 
		dc.b	$94					; 
		dc.b	1
		dc.b	1
		dc.b	$21					; !
		dc.b	$3F					; ?
		dc.b	$31					; 1
		dc.b	$3D					; =
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$56					; V
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$B3					; 
		dc.b	$C7					; 
		dc.b	$C9					; 
		dc.b	$A5					; 
		dc.b	$A6					; 
		dc.b	$C0					; 
		dc.b	$A8					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$A5					; 
		dc.b	$C6					; 
		dc.b	$D8					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$46					; F
		dc.b	$40					; @
		dc.b	$31					; 1
		dc.b	7
		dc.b	1
		dc.b	$37					; 7
		dc.b	$91					; 
		dc.b	$E1					; 
		dc.b	$D2					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$AB					; 
		dc.b	$C7					; 
		dc.b	$A2					; 
		dc.b	$5C					; \
		dc.b	$62					; b
		dc.b	$D2					; 
		dc.b	$FE					; 
		dc.b	$E3					; 
		dc.b	$B9					; 
		dc.b	$D7					; 
		dc.b	$FE					; 
		dc.b	$D1					; 
		dc.b	$39					; 9
		dc.b	$34					; 4
		dc.b	$47					; G
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$4D					; M
		dc.b	$3D					; =
		dc.b	$21					; !
		dc.b	1
		dc.b	1
		dc.b	$31					; 1
		dc.b	$5A					; Z
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$85					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$7F					; 
		dc.b	$BB					; 
		dc.b	$FB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D9					; 
		dc.b	$3E					; >
		dc.b	$40					; @
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$43					; C
		dc.b	8
		dc.b	3
		dc.b	$16
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$22					; "
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$85					; 
		dc.b	$DC					; 
		dc.b	$EC					; 
		dc.b	$EC					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$FC					; 
		dc.b	$89					; 
		dc.b	$65					; e
		dc.b	$50					; P
		dc.b	$3C					; <
		dc.b	$28					; (
		dc.b	7
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$2C					; ,
		dc.b	$79					; y
		dc.b	$8C					; 
		dc.b	$CB					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$60					; `
		dc.b	$28					; (
		dc.b	$12
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$2F					; /
		dc.b	$61					; a
		dc.b	$95					; 
		dc.b	$D3					; 
		dc.b	$E6					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B7					; 
		dc.b	$26					; &
		dc.b	$1C
		dc.b	$C
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	3
		dc.b	$5C					; \
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B0					; 
		dc.b	$27					; '
		dc.b	$B
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$15
		dc.b	$31					; 1
		dc.b	$57					; W
		dc.b	$C4					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$67					; g
		dc.b	$41					; A
		dc.b	$1E
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$15
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$12
		dc.b	$4A					; J
		dc.b	$5B					; [
		dc.b	$A9					; 
		dc.b	$DB					; 
		dc.b	$C4					; 
		dc.b	$CE					; 
		dc.b	$E7					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$DB					; 
		dc.b	$90					; 
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$49					; I
		dc.b	$19
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$1E
		dc.b	$54					; T
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$DB					; 
		dc.b	$F3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F2					; 
		dc.b	$CC					; 
		dc.b	$82					; 
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$4D					; M
		dc.b	$2B					; +
		dc.b	$2D					; -
		dc.b	$1C
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$19
		dc.b	$66					; f
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$74					; t
		dc.b	$79					; y
		dc.b	$93					; 
		dc.b	$9C					; 
		dc.b	$B9					; 
		dc.b	$E3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F7					; 
		dc.b	$B1					; 
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$5E					; ^
		dc.b	$38					; 8
		dc.b	$4E					; N
		dc.b	$67					; g
		dc.b	$40					; @
		dc.b	$12
		dc.b	$A
		dc.b	$2E					; .
		dc.b	$34					; 4
		dc.b	$11
		dc.b	$16
		dc.b	$56					; V
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$9A					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$B9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$AD					; 
		dc.b	$68					; h
		dc.b	$49					; I
		dc.b	$47					; G
		dc.b	$4D					; M
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$3F					; ?
		dc.b	$5F					; _
		dc.b	$90					; 
		dc.b	$B6					; 
		dc.b	$9B					; 
		dc.b	$7B					; {
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$A2					; 
		dc.b	$AF					; 
		dc.b	$C2					; 
		dc.b	$D4					; 
		dc.b	$F1					; 
		dc.b	$F3					; 
		dc.b	$D7					; 
		dc.b	$82					; 
		dc.b	$26					; &
		dc.b	$C
		dc.b	$1D
		dc.b	$21					; !
		dc.b	$F
		dc.b	$13
		dc.b	$34					; 4
		dc.b	$2A					; *
		dc.b	1
		dc.b	1
		dc.b	$15
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$49					; I
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F3					; 
		dc.b	$93					; 
		dc.b	$15
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$37					; 7
		dc.b	$AD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D2					; 
		dc.b	$D7					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$A8					; 
		dc.b	$90					; 
		dc.b	$A8					; 
		dc.b	$86					; 
		dc.b	$3F					; ?
		dc.b	$3D					; =
		dc.b	$23					; #
		dc.b	$F
		dc.b	$35					; 5
		dc.b	$35					; 5
		dc.b	$E
		dc.b	1
		dc.b	$26					; &
		dc.b	$2F					; /
		dc.b	$1D
		dc.b	$32					; 2
		dc.b	$4F					; O
		dc.b	$9E					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E6					; 
		dc.b	$B4					; 
		dc.b	$A1					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$3B					; ;
		dc.b	$2E					; .
		dc.b	$3C					; <
		dc.b	$1C
		dc.b	3
		dc.b	$1A
		dc.b	$2E					; .
		dc.b	$4A					; J
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$71					; q
		dc.b	$1C
		dc.b	$1D
		dc.b	$50					; P
		dc.b	$86					; 
		dc.b	$AF					; 
		dc.b	$BD					; 
		dc.b	$DC					; 
		dc.b	$ED					; 
		dc.b	$E0					; 
		dc.b	$D1					; 
		dc.b	$E7					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$F1					; 
		dc.b	$A1					; 
		dc.b	$A8					; 
		dc.b	$DE					; 
		dc.b	$C3					; 
		dc.b	$3C					; <
		dc.b	1
		dc.b	1
		dc.b	$2E					; .
		dc.b	$1F
		dc.b	$A
		dc.b	$E
		dc.b	$38					; 8
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$81					; 
		dc.b	$49					; I
		dc.b	$89					; 
		dc.b	$DD					; 
		dc.b	$DB					; 
		dc.b	$B6					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$93					; 
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$B0					; 
		dc.b	$BF					; 
		dc.b	$7B					; {
		dc.b	$23					; #
		dc.b	$2A					; *
		dc.b	$2D					; -
		dc.b	$1D
		dc.b	$36					; 6
		dc.b	$6E					; n
		dc.b	$9C					; 
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$40					; @
		dc.b	$6B					; k
		dc.b	$EF					; 
		dc.b	$FE					; 
		dc.b	$DE					; 
		dc.b	$BD					; 
		dc.b	$AF					; 
		dc.b	$A5					; 
		dc.b	$B2					; 
		dc.b	$C3					; 
		dc.b	$C0					; 
		dc.b	$B2					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$9A					; 
		dc.b	$B8					; 
		dc.b	$B5					; 
		dc.b	$5B					; [
		dc.b	$2A					; *
		dc.b	$21					; !
		dc.b	$32					; 2
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$21					; !
		dc.b	$1F
		dc.b	$16
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$4A					; J
		dc.b	$9F					; 
		dc.b	$DB					; 
		dc.b	$D7					; 
		dc.b	$CA					; 
		dc.b	$EC					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$EC					; 
		dc.b	$D4					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$1A
		dc.b	8
		dc.b	8
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$72					; r
		dc.b	$97					; 
		dc.b	$70					; p
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$BF					; 
		dc.b	$DE					; 
		dc.b	$D9					; 
		dc.b	$E5					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$D6					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$2F					; /
		dc.b	$23					; #
		dc.b	$21					; !
		dc.b	$B
		dc.b	1
		dc.b	1
		dc.b	4
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	7
		dc.b	$28					; (
		dc.b	$69					; i
		dc.b	$91					; 
		dc.b	$C0					; 
		dc.b	$CE					; 
		dc.b	$E2					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C9					; 
		dc.b	$5C					; \
		dc.b	$43					; C
		dc.b	$32					; 2
		dc.b	$1F
		dc.b	$E
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$B
		dc.b	$C
		dc.b	$65					; e
		dc.b	$F8					; 
		dc.b	$FE					; 
		dc.b	$DF					; 
		dc.b	$D0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E2					; 
		dc.b	$F7					; 
		dc.b	$F3					; 
		dc.b	$DF					; 
		dc.b	$E7					; 
		dc.b	$F0					; 
		dc.b	$DE					; 
		dc.b	$7A					; z
		dc.b	$24					; $
		dc.b	$F
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$2F					; /
		dc.b	$36					; 6
		dc.b	$47					; G
		dc.b	$56					; V
		dc.b	$6E					; n
		dc.b	$BB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C4					; 
		dc.b	$9A					; 
		dc.b	$DB					; 
		dc.b	$DE					; 
		dc.b	$AC					; 
		dc.b	$90					; 
		dc.b	$91					; 
		dc.b	$2D					; -
		dc.b	1
		dc.b	$11
		dc.b	$24					; $
		dc.b	$18
		dc.b	1
		dc.b	$1F
		dc.b	$16
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$28					; (
		dc.b	$3B					; ;
		dc.b	$27					; '
		dc.b	$54					; T
		dc.b	$A4					; 
		dc.b	$C7					; 
		dc.b	$D2					; 
		dc.b	$DF					; 
		dc.b	$DC					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$DB					; 
		dc.b	$CA					; 
		dc.b	$B0					; 
		dc.b	$9A					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$B6					; 
		dc.b	$D4					; 
		dc.b	$AF					; 
		dc.b	$77					; w
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$6A					; j
		dc.b	$40					; @
		dc.b	$27					; '
		dc.b	$12
		dc.b	$19
		dc.b	$3F					; ?
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$5B					; [
		dc.b	$85					; 
		dc.b	$94					; 
		dc.b	$C0					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$FB					; 
		dc.b	$E2					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$19
		dc.b	$19
		dc.b	$24					; $
		dc.b	$40					; @
		dc.b	$34					; 4
		dc.b	$4B					; K
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$9A					; 
		dc.b	$A3					; 
		dc.b	$B4					; 
		dc.b	$D1					; 
		dc.b	$DF					; 
		dc.b	$ED					; 
		dc.b	$E9					; 
		dc.b	$F4					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$ED					; 
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$26					; &
		dc.b	$15
		dc.b	$35					; 5
		dc.b	$1F
		dc.b	$A
		dc.b	1
		dc.b	3
		dc.b	1
		dc.b	1
		dc.b	$2F					; /
		dc.b	$72					; r
		dc.b	$A9					; 
		dc.b	$F9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$B8					; 
		dc.b	$B4					; 
		dc.b	$D1					; 
		dc.b	$D9					; 
		dc.b	$BB					; 
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$B9					; 
		dc.b	$B1					; 
		dc.b	$C9					; 
		dc.b	$92					; 
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$46					; F
		dc.b	$24					; $
		dc.b	$27					; '
		dc.b	$12
		dc.b	$E
		dc.b	$20
		dc.b	$21					; !
		dc.b	1
		dc.b	$22					; "
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$9F					; 
		dc.b	$BF					; 
		dc.b	$D8					; 
		dc.b	$EC					; 
		dc.b	$F0					; 
		dc.b	$E3					; 
		dc.b	$C7					; 
		dc.b	$D9					; 
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$95					; 
		dc.b	$32					; 2
		dc.b	$16
		dc.b	4
		dc.b	8
		dc.b	$A
		dc.b	1
		dc.b	$B
		dc.b	$1A
		dc.b	$2F					; /
		dc.b	$F
		dc.b	$13
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$8D					; 
		dc.b	$A0					; 
		dc.b	$D6					; 
		dc.b	$E2					; 
		dc.b	$EE					; 
		dc.b	$E7					; 
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	$C6					; 
		dc.b	$B8					; 
		dc.b	$AF					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$A1					; 
		dc.b	$8D					; 
		dc.b	$2D					; -
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	4
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$36					; 6
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$78					; x
		dc.b	$B6					; 
		dc.b	$D9					; 
		dc.b	$E2					; 
		dc.b	$D7					; 
		dc.b	$E5					; 
		dc.b	$E2					; 
		dc.b	$D4					; 
		dc.b	$D7					; 
		dc.b	$E0					; 
		dc.b	$D7					; 
		dc.b	$E5					; 
		dc.b	$F1					; 
		dc.b	$F5					; 
		dc.b	$FA					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$C6					; 
		dc.b	$73					; s
		dc.b	$4D					; M
		dc.b	$21					; !
		dc.b	$F
		dc.b	$16
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	4
		dc.b	$30					; 0
		dc.b	$85					; 
		dc.b	$B0					; 
		dc.b	$A5					; 
		dc.b	$AF					; 
		dc.b	$B6					; 
		dc.b	$C9					; 
		dc.b	$CA					; 
		dc.b	$C9					; 
		dc.b	$E0					; 
		dc.b	$F8					; 
		dc.b	$E3					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$DE					; 
		dc.b	$AA					; 
		dc.b	$29					; )
		dc.b	$12
		dc.b	$16
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$26					; &
		dc.b	$A
		dc.b	$12
		dc.b	$35					; 5
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$CA					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$B4					; 
		dc.b	$A5					; 
		dc.b	$9B					; 
		dc.b	$B9					; 
		dc.b	$E2					; 
		dc.b	$E3					; 
		dc.b	$BF					; 
		dc.b	$C9					; 
		dc.b	$C9					; 
		dc.b	$DC					; 
		dc.b	$E2					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$5C					; \
		dc.b	$40					; @
		dc.b	$18
		dc.b	$19
		dc.b	$31					; 1
		dc.b	$54					; T
		dc.b	$85					; 
		dc.b	$90					; 
		dc.b	$B2					; 
		dc.b	$DF					; 
		dc.b	$BB					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$AB					; 
		dc.b	$AC					; 
		dc.b	$B6					; 
		dc.b	$D4					; 
		dc.b	$D2					; 
		dc.b	$D0					; 
		dc.b	$D9					; 
		dc.b	$F4					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$B1					; 
		dc.b	$32					; 2
		dc.b	$31					; 1
		dc.b	$51					; Q
		dc.b	$64					; d
		dc.b	$4B					; K
		dc.b	$16
		dc.b	1
		dc.b	$11
		dc.b	$15
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$31					; 1
		dc.b	$70					; p
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$9E					; 
		dc.b	$89					; 
		dc.b	$7B					; {
		dc.b	$B9					; 
		dc.b	$E6					; 
		dc.b	$E3					; 
		dc.b	$DC					; 
		dc.b	$F3					; 
		dc.b	$FE					; 
		dc.b	$FB					; 
		dc.b	$F4					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CF					; 
		dc.b	$49					; I
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$4E					; N
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$57					; W
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$3F					; ?
		dc.b	$3E					; >
		dc.b	$6A					; j
		dc.b	$90					; 
		dc.b	$8C					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$AF					; 
		dc.b	$AE					; 
		dc.b	$91					; 
		dc.b	$90					; 
		dc.b	$A9					; 
		dc.b	$CE					; 
		dc.b	$C0					; 
		dc.b	$C0					; 
		dc.b	$D8					; 
		dc.b	$CB					; 
		dc.b	$75					; u
		dc.b	$23					; #
		dc.b	$1F
		dc.b	$2F					; /
		dc.b	$3D					; =
		dc.b	$3D					; =
		dc.b	$47					; G
		dc.b	$4E					; N
		dc.b	$4E					; N
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$49					; I
		dc.b	$4E					; N
		dc.b	$61					; a
		dc.b	$BB					; 
		dc.b	$C6					; 
		dc.b	$C9					; 
		dc.b	$D5					; 
		dc.b	$EC					; 
		dc.b	$F5					; 
		dc.b	$DF					; 
		dc.b	$DE					; 
		dc.b	$EA					; 
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	$D7					; 
		dc.b	$B2					; 
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$BB					; 
		dc.b	$92					; 
		dc.b	$22					; "
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$B
		dc.b	1
		dc.b	$13
		dc.b	$4B					; K
		dc.b	$5D					; ]
		dc.b	$6A					; j
		dc.b	$BD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FC					; 
		dc.b	$F5					; 
		dc.b	$D9					; 
		dc.b	$C0					; 
		dc.b	$B2					; 
		dc.b	$AF					; 
		dc.b	$B2					; 
		dc.b	$9F					; 
		dc.b	$5C					; \
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$C
		dc.b	$23					; #
		dc.b	$15
		dc.b	3
		dc.b	$16
		dc.b	$2D					; -
		dc.b	$3B					; ;
		dc.b	$32					; 2
		dc.b	$57					; W
		dc.b	$B8					; 
		dc.b	$EA					; 
		dc.b	$FA					; 
		dc.b	$F7					; 
		dc.b	$F7					; 
		dc.b	$EA					; 
		dc.b	$E2					; 
		dc.b	$E3					; 
		dc.b	$E5					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CA					; 
		dc.b	$AF					; 
		dc.b	$CA					; 
		dc.b	$F1					; 
		dc.b	$C8					; 
		dc.b	$A5					; 
		dc.b	$B5					; 
		dc.b	$75					; u
		dc.b	8
		dc.b	1
		dc.b	1
		dc.b	$2B					; +
		dc.b	$36					; 6
		dc.b	$18
		dc.b	$2A					; *
		dc.b	$47					; G
		dc.b	$4E					; N
		dc.b	$34					; 4
		dc.b	$11
		dc.b	$C
		dc.b	$14
		dc.b	$44					; D
		dc.b	$B2					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$9B					; 
		dc.b	$CD					; 
		dc.b	$F0					; 
		dc.b	$D1					; 
		dc.b	$DE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D3					; 
		dc.b	$5E					; ^
		dc.b	$21					; !
		dc.b	$F
		dc.b	$21					; !
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$3D					; =
		dc.b	$2E					; .
		dc.b	$15
		dc.b	$1F
		dc.b	$40					; @
		dc.b	$3C					; <
		dc.b	$2B					; +
		dc.b	$66					; f
		dc.b	$C5					; 
		dc.b	$B5					; 
		dc.b	$C9					; 
		dc.b	$CF					; 
		dc.b	$C9					; 
		dc.b	$BC					; 
		dc.b	$BF					; 
		dc.b	$C4					; 
		dc.b	$C4					; 
		dc.b	$C3					; 
		dc.b	$C9					; 
		dc.b	$D2					; 
		dc.b	$B4					; 
		dc.b	$9F					; 
		dc.b	$63					; c
		dc.b	$44					; D
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$27					; '
		dc.b	$38					; 8
		dc.b	$4A					; J
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$2A					; *
		dc.b	$1C
		dc.b	$2C					; ,
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$D1					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$CD					; 
		dc.b	$CB					; 
		dc.b	$B2					; 
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$74					; t
		dc.b	$41					; A
		dc.b	$80					; 
		dc.b	$A8					; 
		dc.b	$AC					; 
		dc.b	$3E					; >
		dc.b	1
		dc.b	$37					; 7
		dc.b	$9F					; 
		dc.b	$85					; 
		dc.b	$36					; 6
		dc.b	$31					; 1
		dc.b	$47					; G
		dc.b	$56					; V
		dc.b	$37					; 7
		dc.b	$38					; 8
		dc.b	$40					; @
		dc.b	$47					; G
		dc.b	$8C					; 
		dc.b	$E4					; 
		dc.b	$E8					; 
		dc.b	$CD					; 
		dc.b	$E3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$CF					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$70					; p
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$BB					; 
		dc.b	$BD					; 
		dc.b	$BE					; 
		dc.b	$80					; 
		dc.b	$49					; I
		dc.b	$55					; U
		dc.b	$97					; 
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$4B					; K
		dc.b	$42					; B
		dc.b	$1A
		dc.b	$A
		dc.b	$1F
		dc.b	$2E					; .
		dc.b	$6C					; l
		dc.b	$AA					; 
		dc.b	$C0					; 
		dc.b	$B2					; 
		dc.b	$B8					; 
		dc.b	$C6					; 
		dc.b	$BD					; 
		dc.b	$B6					; 
		dc.b	$9E					; 
		dc.b	$B1					; 
		dc.b	$CA					; 
		dc.b	$DC					; 
		dc.b	$EE					; 
		dc.b	$F6					; 
		dc.b	$A1					; 
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$C8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$DA					; 
		dc.b	$89					; 
		dc.b	$70					; p
		dc.b	$83					; 
		dc.b	$70					; p
		dc.b	$40					; @
		dc.b	$24					; $
		dc.b	$28					; (
		dc.b	$12
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$53					; S
		dc.b	$96					; 
		dc.b	$B0					; 
		dc.b	$CB					; 
		dc.b	$D2					; 
		dc.b	$D4					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$DE					; 
		dc.b	$CD					; 
		dc.b	$E3					; 
		dc.b	$F3					; 
		dc.b	$DC					; 
		dc.b	$6F					; o
		dc.b	$11
		dc.b	$C
		dc.b	$39					; 9
		dc.b	$1D
		dc.b	1
		dc.b	1
		dc.b	6
		dc.b	$30					; 0
		dc.b	$15
		dc.b	$24					; $
		dc.b	$20
		dc.b	8
		dc.b	$63					; c
		dc.b	$C2					; 
		dc.b	$CE					; 
		dc.b	$B6					; 
		dc.b	$C2					; 
		dc.b	$D7					; 
		dc.b	$EF					; 
		dc.b	$F5					; 
		dc.b	$D8					; 
		dc.b	$E4					; 
		dc.b	$F8					; 
		dc.b	$D8					; 
		dc.b	$C3					; 
		dc.b	$CD					; 
		dc.b	$CA					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$C2					; 
		dc.b	$9D					; 
		dc.b	$3E					; >
		dc.b	$18
		dc.b	3
		dc.b	4
		dc.b	$1D
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$27					; '
		dc.b	$43					; C
		dc.b	$40					; @
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$70					; p
		dc.b	$95					; 
		dc.b	$D8					; 
		dc.b	$EA					; 
		dc.b	$F8					; 
		dc.b	$F8					; 
		dc.b	$EC					; 
		dc.b	$BC					; 
		dc.b	$CA					; 
		dc.b	$D2					; 
		dc.b	$D8					; 
		dc.b	$DD					; 
		dc.b	$E5					; 
		dc.b	$BB					; 
		dc.b	$8C					; 
		dc.b	$7E					; ~
		dc.b	$9D					; 
		dc.b	$B2					; 
		dc.b	$93					; 
		dc.b	$A5					; 
		dc.b	$7B					; {
		dc.b	$14
		dc.b	8
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$13
		dc.b	1
		dc.b	$C
		dc.b	$26					; &
		dc.b	$3B					; ;
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$49					; I
		dc.b	$5A					; Z
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E4					; 
		dc.b	$C5					; 
		dc.b	$D4					; 
		dc.b	$DC					; 
		dc.b	$E4					; 
		dc.b	$CA					; 
		dc.b	$C9					; 
		dc.b	$E2					; 
		dc.b	$C6					; 
		dc.b	$BC					; 
		dc.b	$C4					; 
		dc.b	$CF					; 
		dc.b	$CE					; 
		dc.b	$E3					; 
		dc.b	$FC					; 
		dc.b	$D3					; 
		dc.b	$68					; h
		dc.b	1
		dc.b	$23					; #
		dc.b	$20
		dc.b	1
		dc.b	5
		dc.b	$15
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	3
		dc.b	$18
		dc.b	$12
		dc.b	$6A					; j
		dc.b	$BB					; 
		dc.b	$D2					; 
		dc.b	$E6					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$EA					; 
		dc.b	$F4					; 
		dc.b	$ED					; 
		dc.b	$EE					; 
		dc.b	$E3					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$D0					; 
		dc.b	$B4					; 
		dc.b	$6D					; m
		dc.b	$3B					; ;
		dc.b	$35					; 5
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$33					; 3
		dc.b	$1F
		dc.b	1
		dc.b	1
		dc.b	$20
		dc.b	$38					; 8
		dc.b	$36					; 6
		dc.b	$5F					; _
		dc.b	$85					; 
		dc.b	$5B					; [
		dc.b	$47					; G
		dc.b	$2E					; .
		dc.b	$35					; 5
		dc.b	$79					; y
		dc.b	$C5					; 
		dc.b	$B9					; 
		dc.b	$C0					; 
		dc.b	$BD					; 
		dc.b	$AC					; 
		dc.b	$A1					; 
		dc.b	$9C					; 
		dc.b	$B2					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$AF					; 
		dc.b	$C6					; 
		dc.b	$B8					; 
		dc.b	$9F					; 
		dc.b	$9E					; 
		dc.b	$BD					; 
		dc.b	$D6					; 
		dc.b	$D7					; 
		dc.b	$79					; y
		dc.b	$50					; P
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$4A					; J
		dc.b	$4B					; K
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$4B					; K
		dc.b	$2A					; *
		dc.b	$35					; 5
		dc.b	$30					; 0
		dc.b	$7D					; }
		dc.b	$D2					; 
		dc.b	$DE					; 
		dc.b	$C0					; 
		dc.b	$B0					; 
		dc.b	$AF					; 
		dc.b	$B0					; 
		dc.b	$A8					; 
		dc.b	$8C					; 
		dc.b	$70					; p
		dc.b	$8D					; 
		dc.b	$9A					; 
		dc.b	$82					; 
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$3C					; <
		dc.b	$2F					; /
		dc.b	$69					; i
		dc.b	$C6					; 
		dc.b	$C8					; 
		dc.b	$83					; 
		dc.b	$2B					; +
		dc.b	$26					; &
		dc.b	$1D
		dc.b	$3C					; <
		dc.b	$5A					; Z
		dc.b	$77					; w
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$69					; i
		dc.b	$2B					; +
		dc.b	$12
		dc.b	$38					; 8
		dc.b	$86					; 
		dc.b	$BA					; 
		dc.b	$D9					; 
		dc.b	$E6					; 
		dc.b	$D9					; 
		dc.b	$EA					; 
		dc.b	$E9					; 
		dc.b	$CB					; 
		dc.b	$C3					; 
		dc.b	$A5					; 
		dc.b	$AA					; 
		dc.b	$93					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$20
		dc.b	$F
		dc.b	$A
		dc.b	$10
		dc.b	$17
		dc.b	$5B					; [
		dc.b	$AE					; 
		dc.b	$CB					; 
		dc.b	$91					; 
		dc.b	$41					; A
		dc.b	$47					; G
		dc.b	$59					; Y
		dc.b	$83					; 
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$E5					; 
		dc.b	$B4					; 
		dc.b	$76					; v
		dc.b	$3E					; >
		dc.b	$3D					; =
		dc.b	$6F					; o
		dc.b	$BC					; 
		dc.b	$DD					; 
		dc.b	$D4					; 
		dc.b	$B3					; 
		dc.b	$B3					; 
		dc.b	$AC					; 
		dc.b	$6E					; n
		dc.b	$3C					; <
		dc.b	$28					; (
		dc.b	$56					; V
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$B1					; 
		dc.b	$F6					; 
		dc.b	$DE					; 
		dc.b	$D2					; 
		dc.b	$AE					; 
		dc.b	$40					; @
		dc.b	$38					; 8
		dc.b	$4A					; J
		dc.b	$49					; I
		dc.b	$36					; 6
		dc.b	$4F					; O
		dc.b	$98					; 
		dc.b	$7E					; ~
		dc.b	$1D
		dc.b	$A
		dc.b	$1F
		dc.b	$61					; a
		dc.b	$A5					; 
		dc.b	$B3					; 
		dc.b	$A8					; 
		dc.b	$A8					; 
		dc.b	$C2					; 
		dc.b	$BE					; 
		dc.b	$94					; 
		dc.b	$54					; T
		dc.b	$4E					; N
		dc.b	$76					; v
		dc.b	$B4					; 
		dc.b	$B2					; 
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$61					; a
		dc.b	$85					; 
		dc.b	$D9					; 
		dc.b	$DB					; 
		dc.b	$C7					; 
		dc.b	$C3					; 
		dc.b	$B5					; 
		dc.b	$58					; X
		dc.b	$17
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$2D					; -
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$5D					; ]
		dc.b	$2B					; +
		dc.b	$32					; 2
		dc.b	$3B					; ;
		dc.b	$76					; v
		dc.b	$D8					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$E3					; 
		dc.b	$C7					; 
		dc.b	$74					; t
		dc.b	$2D					; -
		dc.b	$3A					; :
		dc.b	$6C					; l
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$45					; E
		dc.b	$1E
		dc.b	$12
		dc.b	$3B					; ;
		dc.b	$76					; v
		dc.b	$C4					; 
		dc.b	$F3					; 
		dc.b	$F1					; 
		dc.b	$D0					; 
		dc.b	$70					; p
		dc.b	$2A					; *
		dc.b	$15
		dc.b	$25					; %
		dc.b	$40					; @
		dc.b	$72					; r
		dc.b	$B8					; 
		dc.b	$FE					; 
		dc.b	$DB					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$47					; G
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$88					; 
		dc.b	$21					; !
		dc.b	1
		dc.b	$2E					; .
		dc.b	$9A					; 
		dc.b	$E5					; 
		dc.b	$E2					; 
		dc.b	$BE					; 
		dc.b	$A2					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$BC					; 
		dc.b	$B5					; 
		dc.b	$AC					; 
		dc.b	$A9					; 
		dc.b	$47					; G
		dc.b	$51					; Q
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$3C					; <
		dc.b	$6B					; k
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$57					; W
		dc.b	$3B					; ;
		dc.b	$2A					; *
		dc.b	$27					; '
		dc.b	$6F					; o
		dc.b	$C4					; 
		dc.b	$C3					; 
		dc.b	$C3					; 
		dc.b	$C6					; 
		dc.b	$B8					; 
		dc.b	$75					; u
		dc.b	$24					; $
		dc.b	$14
		dc.b	$57					; W
		dc.b	$A8					; 
		dc.b	$B4					; 
		dc.b	$B1					; 
		dc.b	$61					; a
		dc.b	$17
		dc.b	$1F
		dc.b	$1F
		dc.b	$39					; 9
		dc.b	$66					; f
		dc.b	$AD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$D0					; 
		dc.b	$5D					; ]
		dc.b	$38					; 8
		dc.b	$4A					; J
		dc.b	$89					; 
		dc.b	$A3					; 
		dc.b	$B7					; 
		dc.b	$83					; 
		dc.b	$43					; C
		dc.b	$13
		dc.b	$A
		dc.b	7
		dc.b	$D
		dc.b	$50					; P
		dc.b	$A8					; 
		dc.b	$CD					; 
		dc.b	$E5					; 
		dc.b	$DC					; 
		dc.b	$9B					; 
		dc.b	$6D					; m
		dc.b	$49					; I
		dc.b	$3B					; ;
		dc.b	$6E					; n
		dc.b	$9F					; 
		dc.b	$CE					; 
		dc.b	$D3					; 
		dc.b	$91					; 
		dc.b	$34					; 4
		dc.b	$16
		dc.b	$2C					; ,
		dc.b	$5B					; [
		dc.b	$90					; 
		dc.b	$D3					; 
		dc.b	$DB					; 
		dc.b	$C2					; 
		dc.b	$DB					; 
		dc.b	$CD					; 
		dc.b	$54					; T
		dc.b	$44					; D
		dc.b	$7C					; |
		dc.b	$BD					; 
		dc.b	$AC					; 
		dc.b	$9F					; 
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$1B
		dc.b	$12
		dc.b	$21					; !
		dc.b	$3D					; =
		dc.b	$AD					; 
		dc.b	$CA					; 
		dc.b	$D7					; 
		dc.b	$D2					; 
		dc.b	$8F					; 
		dc.b	$6A					; j
		dc.b	$54					; T
		dc.b	$3D					; =
		dc.b	$40					; @
		dc.b	$73					; s
		dc.b	$A1					; 
		dc.b	$CD					; 
		dc.b	$A3					; 
		dc.b	$2F					; /
		dc.b	$2F					; /
		dc.b	$3F					; ?
		dc.b	$44					; D
		dc.b	$89					; 
		dc.b	$DA					; 
		dc.b	$D9					; 
		dc.b	$D5					; 
		dc.b	$E2					; 
		dc.b	$DF					; 
		dc.b	$7E					; ~
		dc.b	1
		dc.b	$B
		dc.b	$7C					; |
		dc.b	$D4					; 
		dc.b	$B7					; 
		dc.b	$6A					; j
		dc.b	$48					; H
		dc.b	$39					; 9
		dc.b	$2B					; +
		dc.b	$21					; !
		dc.b	$20
		dc.b	$4F					; O
		dc.b	$CD					; 
		dc.b	$DC					; 
		dc.b	$E3					; 
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$A6					; 
		dc.b	$77					; w
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$87					; 
		dc.b	$B6					; 
		dc.b	$9F					; 
		dc.b	$4C					; L
		dc.b	$20
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$83					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$D7					; 
		dc.b	$D4					; 
		dc.b	$BB					; 
		dc.b	$C0					; 
		dc.b	$73					; s
		dc.b	$1E
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$2D					; -
		dc.b	4
		dc.b	4
		dc.b	7
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$32					; 2
		dc.b	$D1					; 
		dc.b	$ED					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$CB					; 
		dc.b	$62					; b
		dc.b	$1C
		dc.b	$F
		dc.b	$A
		dc.b	$33					; 3
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$82					; 
		dc.b	$90					; 
		dc.b	$B9					; 
		dc.b	$B1					; 
		dc.b	$C4					; 
		dc.b	$8C					; 
		dc.b	$31					; 1
		dc.b	$44					; D
		dc.b	$AD					; 
		dc.b	$8A					; 
		dc.b	$3B					; ;
		dc.b	$19
		dc.b	$2A					; *
		dc.b	$35					; 5
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$57					; W
		dc.b	$7E					; ~
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EA					; 
		dc.b	$E8					; 
		dc.b	$B6					; 
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$AA					; 
		dc.b	$A5					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	1
		dc.b	$A
		dc.b	$3A					; :
		dc.b	$60					; `
		dc.b	$98					; 
		dc.b	$C3					; 
		dc.b	$E6					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$DF					; 
		dc.b	$E5					; 
		dc.b	$E7					; 
		dc.b	$DC					; 
		dc.b	$7C					; |
		dc.b	$3D					; =
		dc.b	$42					; B
		dc.b	$20
		dc.b	$21					; !
		dc.b	$1D
		dc.b	$20
		dc.b	$5B					; [
		dc.b	$66					; f
		dc.b	$49					; I
		dc.b	$47					; G
		dc.b	$70					; p
		dc.b	$9D					; 
		dc.b	$C0					; 
		dc.b	$BB					; 
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$BB					; 
		dc.b	$C8					; 
		dc.b	$A6					; 
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$45					; E
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$5F					; _
		dc.b	$7E					; ~
		dc.b	$BE					; 
		dc.b	$CB					; 
		dc.b	$B3					; 
		dc.b	$B6					; 
		dc.b	$CE					; 
		dc.b	$B4					; 
		dc.b	$BC					; 
		dc.b	$BD					; 
		dc.b	$C7					; 
		dc.b	$F8					; 
		dc.b	$7B					; {
		dc.b	$20
		dc.b	$19
		dc.b	$1F
		dc.b	$1B
		dc.b	$47					; G
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$44					; D
		dc.b	$5B					; [
		dc.b	$42					; B
		dc.b	$47					; G
		dc.b	$AF					; 
		dc.b	$D9					; 
		dc.b	$CB					; 
		dc.b	$E6					; 
		dc.b	$ED					; 
		dc.b	$85					; 
		dc.b	$19
		dc.b	$1D
		dc.b	$26					; &
		dc.b	$24					; $
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$5D					; ]
		dc.b	$59					; Y
		dc.b	$82					; 
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$B8					; 
		dc.b	$C0					; 
		dc.b	$DC					; 
		dc.b	$CE					; 
		dc.b	$DB					; 
		dc.b	$EA					; 
		dc.b	$FA					; 
		dc.b	$FC					; 
		dc.b	$7C					; |
		dc.b	$51					; Q
		dc.b	$4F					; O
		dc.b	$4B					; K
		dc.b	$47					; G
		dc.b	$3D					; =
		dc.b	$43					; C
		dc.b	$64					; d
		dc.b	$42					; B
		dc.b	$1D
		dc.b	$1C
		dc.b	$66					; f
		dc.b	$E0					; 
		dc.b	$D9					; 
		dc.b	$D8					; 
		dc.b	$B0					; 
		dc.b	$A1					; 
		dc.b	$9C					; 
		dc.b	$8C					; 
		dc.b	6
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$2B					; +
		dc.b	$46					; F
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$85					; 
		dc.b	$8F					; 
		dc.b	$B2					; 
		dc.b	$EA					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$E6					; 
		dc.b	$EA					; 
		dc.b	$D5					; 
		dc.b	$BC					; 
		dc.b	$B2					; 
		dc.b	$AB					; 
		dc.b	$63					; c
		dc.b	$21					; !
		dc.b	$43					; C
		dc.b	$2D					; -
		dc.b	$16
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	$2D					; -
		dc.b	$7A					; z
		dc.b	$C6					; 
		dc.b	$D2					; 
		dc.b	$F3					; 
		dc.b	$FA					; 
		dc.b	$E7					; 
		dc.b	$F3					; 
		dc.b	$E9					; 
		dc.b	$DF					; 
		dc.b	$E5					; 
		dc.b	$DD					; 
		dc.b	$85					; 
		dc.b	$26					; &
		dc.b	$2D					; -
		dc.b	$42					; B
		dc.b	$12
		dc.b	$C
		dc.b	$33					; 3
		dc.b	$2B					; +
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$C7					; 
		dc.b	$C0					; 
		dc.b	$B8					; 
		dc.b	$A5					; 
		dc.b	$A8					; 
		dc.b	$D7					; 
		dc.b	$E6					; 
		dc.b	$E0					; 
		dc.b	$ED					; 
		dc.b	$F7					; 
		dc.b	$83					; 
		dc.b	$32					; 2
		dc.b	$46					; F
		dc.b	$46					; F
		dc.b	$49					; I
		dc.b	$53					; S
		dc.b	$62					; b
		dc.b	$A9					; 
		dc.b	$CE					; 
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$89					; 
		dc.b	$8A					; 
		dc.b	$AF					; 
		dc.b	$A2					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$43					; C
		dc.b	$2B					; +
		dc.b	$31					; 1
		dc.b	$46					; F
		dc.b	$5A					; Z
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$A1					; 
		dc.b	$CF					; 
		dc.b	$D3					; 
		dc.b	$C3					; 
		dc.b	$DF					; 
		dc.b	$FC					; 
		dc.b	$F4					; 
		dc.b	$F8					; 
		dc.b	$DA					; 
		dc.b	$89					; 
		dc.b	$6F					; o
		dc.b	$39					; 9
		dc.b	$28					; (
		dc.b	$C
		dc.b	$16
		dc.b	$2A					; *
		dc.b	$17
		dc.b	$3A					; :
		dc.b	$86					; 
		dc.b	$BE					; 
		dc.b	$E5					; 
		dc.b	$E9					; 
		dc.b	$EB					; 
		dc.b	$CA					; 
		dc.b	$E4					; 
		dc.b	$F0					; 
		dc.b	$F4					; 
		dc.b	$F4					; 
		dc.b	$B1					; 
		dc.b	$39					; 9
		dc.b	1
		dc.b	$16
		dc.b	$2A					; *
		dc.b	$11
		dc.b	3
		dc.b	$4F					; O
		dc.b	$A5					; 
		dc.b	$B6					; 
		dc.b	$AB					; 
		dc.b	$82					; 
		dc.b	$6C					; l
		dc.b	$89					; 
		dc.b	$97					; 
		dc.b	$B9					; 
		dc.b	$CA					; 
		dc.b	$EA					; 
		dc.b	$D4					; 
		dc.b	$7E					; ~
		dc.b	$4C					; L
		dc.b	$3B					; ;
		dc.b	$44					; D
		dc.b	$45					; E
		dc.b	$48					; H
		dc.b	$73					; s
		dc.b	$B5					; 
		dc.b	$AF					; 
		dc.b	$4E					; N
		dc.b	$44					; D
		dc.b	$52					; R
		dc.b	$8A					; 
		dc.b	$D7					; 
		dc.b	$E3					; 
		dc.b	$E0					; 
		dc.b	$CE					; 
		dc.b	$BF					; 
		dc.b	$76					; v
		dc.b	$E
		dc.b	1
		dc.b	$10
		dc.b	$22					; "
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$29					; )
		dc.b	$17
		dc.b	$1D
		dc.b	$21					; !
		dc.b	$4A					; J
		dc.b	$99					; 
		dc.b	$F5					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$F8					; 
		dc.b	$E2					; 
		dc.b	$BB					; 
		dc.b	$77					; w
		dc.b	$9C					; 
		dc.b	$B3					; 
		dc.b	$C0					; 
		dc.b	$98					; 
		dc.b	$60					; `
		dc.b	$42					; B
		dc.b	$24					; $
		dc.b	$1D
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$A5					; 
		dc.b	$C9					; 
		dc.b	$B9					; 
		dc.b	$C2					; 
		dc.b	$C3					; 
		dc.b	$66					; f
		dc.b	8
		dc.b	$E
		dc.b	$4B					; K
		dc.b	$8E					; 
		dc.b	$B0					; 
		dc.b	$CE					; 
		dc.b	$D1					; 
		dc.b	$74					; t
		dc.b	$3F					; ?
		dc.b	$4B					; K
		dc.b	$34					; 4
		dc.b	$3E					; >
		dc.b	$BE					; 
		dc.b	$D2					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$B7					; 
		dc.b	$5D					; ]
		dc.b	$1C
		dc.b	$23					; #
		dc.b	$6F					; o
		dc.b	$CD					; 
		dc.b	$DF					; 
		dc.b	$D2					; 
		dc.b	$D4					; 
		dc.b	$7B					; {
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$7B					; {
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$C7					; 
		dc.b	$BD					; 
		dc.b	$6B					; k
		dc.b	$42					; B
		dc.b	$5E					; ^
		dc.b	$BF					; 
		dc.b	$D8					; 
		dc.b	$CE					; 
		dc.b	$D2					; 
		dc.b	$BA					; 
		dc.b	$5D					; ]
		dc.b	$16
		dc.b	$39					; 9
		dc.b	$3B					; ;
		dc.b	$E
		dc.b	$51					; Q
		dc.b	$BC					; 
		dc.b	$CD					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$3D					; =
		dc.b	$40					; @
		dc.b	$8A					; 
		dc.b	$D5					; 
		dc.b	$E2					; 
		dc.b	$E6					; 
		dc.b	$E5					; 
		dc.b	$C0					; 
		dc.b	$54					; T
		dc.b	$31					; 1
		dc.b	$1B
		dc.b	$57					; W
		dc.b	$B2					; 
		dc.b	$C6					; 
		dc.b	$C0					; 
		dc.b	$A6					; 
		dc.b	$45					; E
		dc.b	$15
		dc.b	$16
		dc.b	$11
		dc.b	$5C					; \
		dc.b	$C1					; 
		dc.b	$CB					; 
		dc.b	$CE					; 
		dc.b	$C1					; 
		dc.b	$7E					; ~
		dc.b	$54					; T
		dc.b	$86					; 
		dc.b	$D7					; 
		dc.b	$D1					; 
		dc.b	$B3					; 
		dc.b	$50					; P
		dc.b	$3F					; ?
		dc.b	$39					; 9
		dc.b	$3B					; ;
		dc.b	$2F					; /
		dc.b	$46					; F
		dc.b	$90					; 
		dc.b	$CA					; 
		dc.b	$DE					; 
		dc.b	$96					; 
		dc.b	$28					; (
		dc.b	$2D					; -
		dc.b	$5B					; [
		dc.b	$93					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$AE					; 
		dc.b	$B6					; 
		dc.b	$C8					; 
		dc.b	$AE					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$9A					; 
		dc.b	$AD					; 
		dc.b	$87					; 
		dc.b	$4D					; M
		dc.b	$4B					; K
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$9C					; 
		dc.b	$90					; 
		dc.b	$63					; c
		dc.b	$62					; b
		dc.b	$90					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$42					; B
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$89					; 
		dc.b	$CD					; 
		dc.b	$DF					; 
		dc.b	$E9					; 
		dc.b	$BD					; 
		dc.b	$38					; 8
		dc.b	$23					; #
		dc.b	$27					; '
		dc.b	$64					; d
		dc.b	$75					; u
		dc.b	$9A					; 
		dc.b	$C2					; 
		dc.b	$C2					; 
		dc.b	$8E					; 
		dc.b	$35					; 5
		dc.b	$2E					; .
		dc.b	$3B					; ;
		dc.b	$8A					; 
		dc.b	$D4					; 
		dc.b	$DC					; 
		dc.b	$CE					; 
		dc.b	$AD					; 
		dc.b	$83					; 
		dc.b	$4D					; M
		dc.b	1
		dc.b	$16
		dc.b	$26					; &
		dc.b	$5A					; Z
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$66					; f
		dc.b	$60					; `
		dc.b	$B7					; 
		dc.b	$E4					; 
		dc.b	$DC					; 
		dc.b	$D4					; 
		dc.b	$D7					; 
		dc.b	$D4					; 
		dc.b	$D1					; 
		dc.b	$9D					; 
		dc.b	$5E					; ^
		dc.b	$74					; t
		dc.b	$AF					; 
		dc.b	$C3					; 
		dc.b	$CB					; 
		dc.b	$C7					; 
		dc.b	$97					; 
		dc.b	$32					; 2
		dc.b	$27					; '
		dc.b	$23					; #
		dc.b	$20
		dc.b	$2D					; -
		dc.b	$75					; u
		dc.b	$D1					; 
		dc.b	$E5					; 
		dc.b	$E9					; 
		dc.b	$C5					; 
		dc.b	$76					; v
		dc.b	$56					; V
		dc.b	$97					; 
		dc.b	$CE					; 
		dc.b	$DB					; 
		dc.b	$E9					; 
		dc.b	$D7					; 
		dc.b	$55					; U
		dc.b	$2A					; *
		dc.b	$23					; #
		dc.b	$16
		dc.b	$53					; S
		dc.b	$C3					; 
		dc.b	$D2					; 
		dc.b	$D0					; 
		dc.b	$CA					; 
		dc.b	$94					; 
		dc.b	$2A					; *
		dc.b	$1F
		dc.b	$33					; 3
		dc.b	$99					; 
		dc.b	$D4					; 
		dc.b	$D9					; 
		dc.b	$DF					; 
		dc.b	$B7					; 
		dc.b	$46					; F
		dc.b	$2E					; .
		dc.b	$2B					; +
		dc.b	$32					; 2
		dc.b	$6E					; n
		dc.b	$C9					; 
		dc.b	$E3					; 
		dc.b	$B1					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$66					; f
		dc.b	$84					; 
		dc.b	$8C					; 
		dc.b	$8B					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$60					; `
		dc.b	$3A					; :
		dc.b	$42					; B
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$69					; i
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$CA					; 
		dc.b	$B3					; 
		dc.b	$6B					; k
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$8D					; 
		dc.b	$CC					; 
		dc.b	$DA					; 
		dc.b	$B6					; 
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$1E
		dc.b	$E
		dc.b	$A
		dc.b	$29					; )
		dc.b	$76					; v
		dc.b	$AB					; 
		dc.b	$B3					; 
		dc.b	$A7					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$AA					; 
		dc.b	$B9					; 
		dc.b	$D3					; 
		dc.b	$C9					; 
		dc.b	$87					; 
		dc.b	$55					; U
		dc.b	$3A					; :
		dc.b	$35					; 5
		dc.b	$67					; g
		dc.b	$C6					; 
		dc.b	$DF					; 
		dc.b	$D7					; 
		dc.b	$C2					; 
		dc.b	$3F					; ?
		dc.b	$24					; $
		dc.b	$26					; &
		dc.b	$2D					; -
		dc.b	$6B					; k
		dc.b	$BD					; 
		dc.b	$D8					; 
		dc.b	$CD					; 
		dc.b	$B7					; 
		dc.b	$38					; 8
		dc.b	$16
		dc.b	$31					; 1
		dc.b	$7A					; z
		dc.b	$CF					; 
		dc.b	$D5					; 
		dc.b	$D4					; 
		dc.b	$D0					; 
		dc.b	$BC					; 
		dc.b	$68					; h
		dc.b	$2E					; .
		dc.b	$2B					; +
		dc.b	$2B					; +
		dc.b	$3A					; :
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$74					; t
		dc.b	$91					; 
		dc.b	$AC					; 
		dc.b	$B6					; 
		dc.b	$C1					; 
		dc.b	$D8					; 
		dc.b	$D7					; 
		dc.b	$D5					; 
		dc.b	$C1					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$5B					; [
		dc.b	$66					; f
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$55					; U
		dc.b	$32					; 2
		dc.b	$34					; 4
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$77					; w
		dc.b	$AC					; 
		dc.b	$BD					; 
		dc.b	$BD					; 
		dc.b	$A9					; 
		dc.b	$68					; h
		dc.b	$2F					; /
		dc.b	$3C					; <
		dc.b	$7B					; {
		dc.b	$AA					; 
		dc.b	$97					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$71					; q
		dc.b	$B4					; 
		dc.b	$D3					; 
		dc.b	$D0					; 
		dc.b	$9E					; 
		dc.b	$54					; T
		dc.b	$2A					; *
		dc.b	$2A					; *
		dc.b	$2D					; -
		dc.b	$4F					; O
		dc.b	$90					; 
		dc.b	$C9					; 
		dc.b	$D7					; 
		dc.b	$CF					; 
		dc.b	$8C					; 
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$8B					; 
		dc.b	$A9					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$4C					; L
		dc.b	$3E					; >
		dc.b	$62					; b
		dc.b	$A1					; 
		dc.b	$C6					; 
		dc.b	$95					; 
		dc.b	$37					; 7
		dc.b	$1C
		dc.b	$31					; 1
		dc.b	$8E					; 
		dc.b	$CD					; 
		dc.b	$DF					; 
		dc.b	$D9					; 
		dc.b	$C4					; 
		dc.b	$9A					; 
		dc.b	$5E					; ^
		dc.b	$37					; 7
		dc.b	$40					; @
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$BC					; 
		dc.b	$9E					; 
		dc.b	$6E					; n
		dc.b	$4B					; K
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$68					; h
		dc.b	$85					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$95					; 
		dc.b	$74					; t
		dc.b	$66					; f
		dc.b	$64					; d
		dc.b	$70					; p
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$69					; i
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$6D					; m
		dc.b	$5F					; _
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$5B					; [
		dc.b	$61					; a
		dc.b	$93					; 
		dc.b	$DA					; 
		dc.b	$D8					; 
		dc.b	$CE					; 
		dc.b	$C0					; 
		dc.b	$37					; 7
		dc.b	$1C
		dc.b	$2A					; *
		dc.b	$4A					; J
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$CA					; 
		dc.b	$DE					; 
		dc.b	$C0					; 
		dc.b	$4F					; O
		dc.b	$28					; (
		dc.b	$2C					; ,
		dc.b	$77					; w
		dc.b	$CB					; 
		dc.b	$DC					; 
		dc.b	$DF					; 
		dc.b	$DF					; 
		dc.b	$D7					; 
		dc.b	$99					; 
		dc.b	$36					; 6
		dc.b	$20
		dc.b	$26					; &
		dc.b	$40					; @
		dc.b	$46					; F
		dc.b	$2C					; ,
		dc.b	$1F
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$85					; 
		dc.b	$A5					; 
		dc.b	$C5					; 
		dc.b	$CA					; 
		dc.b	$C6					; 
		dc.b	$CD					; 
		dc.b	$D0					; 
		dc.b	$BD					; 
		dc.b	$A5					; 
		dc.b	$B5					; 
		dc.b	$B5					; 
		dc.b	$94					; 
		dc.b	$70					; p
		dc.b	$5B					; [
		dc.b	$4B					; K
		dc.b	$38					; 8
		dc.b	$36					; 6
		dc.b	$59					; Y
		dc.b	$AB					; 
		dc.b	$DB					; 
		dc.b	$BB					; 
		dc.b	$48					; H
		dc.b	$26					; &
		dc.b	$45					; E
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$A5					; 
		dc.b	$C8					; 
		dc.b	$C8					; 
		dc.b	$9E					; 
		dc.b	$4D					; M
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$4E					; N
		dc.b	$91					; 
		dc.b	$B4					; 
		dc.b	$BD					; 
		dc.b	$A6					; 
		dc.b	$9A					; 
		dc.b	$8E					; 
		dc.b	$69					; i
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$57					; W
		dc.b	$80					; 
		dc.b	$93					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$BD					; 
		dc.b	$DB					; 
		dc.b	$C4					; 
		dc.b	$8B					; 
		dc.b	$66					; f
		dc.b	$5D					; ]
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$36					; 6
		dc.b	$37					; 7
		dc.b	$52					; R
		dc.b	$86					; 
		dc.b	$B9					; 
		dc.b	$D0					; 
		dc.b	$D2					; 
		dc.b	$B9					; 
		dc.b	$8C					; 
		dc.b	$59					; Y
		dc.b	$39					; 9
		dc.b	$49					; I
		dc.b	$78					; x
		dc.b	$9A					; 
		dc.b	$A3					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$5B					; [
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$9E					; 
		dc.b	$BA					; 
		dc.b	$BD					; 
		dc.b	$9F					; 
		dc.b	$7B					; {
		dc.b	$5D					; ]
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$5F					; _
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$C0					; 
		dc.b	$D3					; 
		dc.b	$D1					; 
		dc.b	$D0					; 
		dc.b	$93					; 
		dc.b	$2B					; +
		dc.b	$28					; (
		dc.b	$2D					; -
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$9E					; 
		dc.b	$C7					; 
		dc.b	$D2					; 
		dc.b	$91					; 
		dc.b	$34					; 4
		dc.b	$24					; $
		dc.b	$44					; D
		dc.b	$9D					; 
		dc.b	$D7					; 
		dc.b	$D8					; 
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	$69					; i
		dc.b	$52					; R
		dc.b	$32					; 2
		dc.b	$32					; 2
		dc.b	$6B					; k
		dc.b	$A6					; 
		dc.b	$A5					; 
		dc.b	$57					; W
		dc.b	$27					; '
		dc.b	$2A					; *
		dc.b	$2F					; /
		dc.b	$56					; V
		dc.b	$8F					; 
		dc.b	$C5					; 
		dc.b	$D2					; 
		dc.b	$D5					; 
		dc.b	$D7					; 
		dc.b	$C2					; 
		dc.b	$AB					; 
		dc.b	$CA					; 
		dc.b	$D8					; 
		dc.b	$C4					; 
		dc.b	$89					; 
		dc.b	$5D					; ]
		dc.b	$44					; D
		dc.b	$30					; 0
		dc.b	$24					; $
		dc.b	$34					; 4
		dc.b	$77					; w
		dc.b	$B5					; 
		dc.b	$C7					; 
		dc.b	$BA					; 
		dc.b	$A4					; 
		dc.b	$77					; w
		dc.b	$38					; 8
		dc.b	$27					; '
		dc.b	$2C					; ,
		dc.b	$68					; h
		dc.b	$B6					; 
		dc.b	$D6					; 
		dc.b	$C0					; 
		dc.b	$8F					; 
		dc.b	$62					; b
		dc.b	$3C					; <
		dc.b	$27					; '
		dc.b	$32					; 2
		dc.b	$6B					; k
		dc.b	$B1					; 
		dc.b	$D5					; 
		dc.b	$D6					; 
		dc.b	$B2					; 
		dc.b	$6A					; j
		dc.b	$38					; 8
		dc.b	$2F					; /
		dc.b	$4A					; J
		dc.b	$78					; x
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$5B					; [
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$66					; f
		dc.b	$9F					; 
		dc.b	$CF					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$A9					; 
		dc.b	$6F					; o
		dc.b	$4B					; K
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$4D					; M
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$5C					; \
		dc.b	$8A					; 
		dc.b	$B2					; 
		dc.b	$C4					; 
		dc.b	$BA					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$A8					; 
		dc.b	$C1					; 
		dc.b	$B6					; 
		dc.b	$76					; v
		dc.b	$31					; 1
		dc.b	$27					; '
		dc.b	$29					; )
		dc.b	$50					; P
		dc.b	$96					; 
		dc.b	$C2					; 
		dc.b	$C1					; 
		dc.b	$9B					; 
		dc.b	$69					; i
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$AB					; 
		dc.b	$B0					; 
		dc.b	$9E					; 
		dc.b	$72					; r
		dc.b	$42					; B
		dc.b	$2C					; ,
		dc.b	$46					; F
		dc.b	$7D					; }
		dc.b	$AC					; 
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$5B					; [
		dc.b	$33					; 3
		dc.b	$2E					; .
		dc.b	$55					; U
		dc.b	$98					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$CA					; 
		dc.b	$A8					; 
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$4E					; N
		dc.b	$51					; Q
		dc.b	$69					; i
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$4E					; N
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$AF					; 
		dc.b	$B1					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$CA					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$BE					; 
		dc.b	$3A					; :
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$2C					; ,
		dc.b	$61					; a
		dc.b	$BC					; 
		dc.b	$D8					; 
		dc.b	$D6					; 
		dc.b	$9F					; 
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$98					; 
		dc.b	$D2					; 
		dc.b	$D8					; 
		dc.b	$C5					; 
		dc.b	$83					; 
		dc.b	$64					; d
		dc.b	$41					; A
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$42					; B
		dc.b	$6E					; n
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$BC					; 
		dc.b	$C6					; 
		dc.b	$CD					; 
		dc.b	$D6					; 
		dc.b	$D6					; 
		dc.b	$BC					; 
		dc.b	$8C					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$58					; X
		dc.b	$40					; @
		dc.b	$39					; 9
		dc.b	$4C					; L
		dc.b	$58					; X
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$B9					; 
		dc.b	$D5					; 
		dc.b	$D8					; 
		dc.b	$D0					; 
		dc.b	$A1					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$51					; Q
		dc.b	$4E					; N
		dc.b	$66					; f
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$8C					; 
		dc.b	$44					; D
		dc.b	$27					; '
		dc.b	$29					; )
		dc.b	$53					; S
		dc.b	$99					; 
		dc.b	$C5					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$D7					; 
		dc.b	$B4					; 
		dc.b	$61					; a
		dc.b	$2C					; ,
		dc.b	$2F					; /
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$B7					; 
		dc.b	$C8					; 
		dc.b	$B6					; 
		dc.b	$88					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$82					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$74					; t
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$90					; 
		dc.b	$9E					; 
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$CB					; 
		dc.b	$5E					; ^
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$4D					; M
		dc.b	$8C					; 
		dc.b	$B1					; 
		dc.b	$D0					; 
		dc.b	$D8					; 
		dc.b	$D4					; 
		dc.b	$9E					; 
		dc.b	$50					; P
		dc.b	$43					; C
		dc.b	$6D					; m
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$66					; f
		dc.b	$3C					; <
		dc.b	$29					; )
		dc.b	$2B					; +
		dc.b	$3A					; :
		dc.b	$5F					; _
		dc.b	$A0					; 
		dc.b	$D2					; 
		dc.b	$D5					; 
		dc.b	$9F					; 
		dc.b	$46					; F
		dc.b	$29					; )
		dc.b	$4D					; M
		dc.b	$92					; 
		dc.b	$BD					; 
		dc.b	$C1					; 
		dc.b	$B2					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$74					; t
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$BA					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$71					; q
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$46					; F
		dc.b	$7D					; }
		dc.b	$AB					; 
		dc.b	$D2					; 
		dc.b	$D8					; 
		dc.b	$D0					; 
		dc.b	$8E					; 
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$58					; X
		dc.b	$94					; 
		dc.b	$B6					; 
		dc.b	$AB					; 
		dc.b	$85					; 
		dc.b	$6A					; j
		dc.b	$55					; U
		dc.b	$3E					; >
		dc.b	$38					; 8
		dc.b	$55					; U
		dc.b	$8B					; 
		dc.b	$B4					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$59					; Y
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$AF					; 
		dc.b	$BB					; 
		dc.b	$B7					; 
		dc.b	$A8					; 
		dc.b	$97					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$51					; Q
		dc.b	$5B					; [
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$A0					; 
		dc.b	$B3					; 
		dc.b	$B0					; 
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$62					; b
		dc.b	$73					; s
		dc.b	$85					; 
		dc.b	$93					; 
		dc.b	$9B					; 
		dc.b	$93					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$70					; p
		dc.b	$47					; G
		dc.b	$2A					; *
		dc.b	$2E					; .
		dc.b	$58					; X
		dc.b	$92					; 
		dc.b	$BF					; 
		dc.b	$CB					; 
		dc.b	$B9					; 
		dc.b	$97					; 
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$6D					; m
		dc.b	$58					; X
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$AD					; 
		dc.b	$BF					; 
		dc.b	$B6					; 
		dc.b	$A0					; 
		dc.b	$89					; 
		dc.b	$78					; x
		dc.b	$6D					; m
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$74					; t
		dc.b	$5E					; ^
		dc.b	$54					; T
		dc.b	$66					; f
		dc.b	$86					; 
		dc.b	$9D					; 
		dc.b	$A1					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$8D					; 
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$B1					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$C6					; 
		dc.b	$5E					; ^
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$28					; (
		dc.b	$57					; W
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$D0					; 
		dc.b	$8E					; 
		dc.b	$43					; C
		dc.b	$31					; 1
		dc.b	$5C					; \
		dc.b	$97					; 
		dc.b	$B6					; 
		dc.b	$B1					; 
		dc.b	$8A					; 
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$57					; W
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$C5					; 
		dc.b	$D2					; 
		dc.b	$9F					; 
		dc.b	$41					; A
		dc.b	$27					; '
		dc.b	$2E					; .
		dc.b	$65					; e
		dc.b	$9E					; 
		dc.b	$C1					; 
		dc.b	$D1					; 
		dc.b	$D6					; 
		dc.b	$C6					; 
		dc.b	$91					; 
		dc.b	$5D					; ]
		dc.b	$53					; S
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$5B					; [
		dc.b	$41					; A
		dc.b	$3F					; ?
		dc.b	$51					; Q
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$C8					; 
		dc.b	$B6					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$90					; 
		dc.b	$CB					; 
		dc.b	$D8					; 
		dc.b	$B1					; 
		dc.b	$39					; 9
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$4A					; J
		dc.b	$90					; 
		dc.b	$B9					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$AA					; 
		dc.b	$3C					; <
		dc.b	$27					; '
		dc.b	$2C					; ,
		dc.b	$71					; q
		dc.b	$B8					; 
		dc.b	$C2					; 
		dc.b	$AE					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$CF					; 
		dc.b	$91					; 
		dc.b	$35					; 5
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$65					; e
		dc.b	$A2					; 
		dc.b	$D2					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$C0					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$69					; i
		dc.b	$4E					; N
		dc.b	$3C					; <
		dc.b	$3C					; <
		dc.b	$4D					; M
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$B5					; 
		dc.b	$C5					; 
		dc.b	$BA					; 
		dc.b	$A9					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$62					; b
		dc.b	$51					; Q
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$A0					; 
		dc.b	$C4					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$77					; w
		dc.b	$2A					; *
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$60					; `
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$B0					; 
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$8F					; 
		dc.b	$B9					; 
		dc.b	$B8					; 
		dc.b	$93					; 
		dc.b	$71					; q
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$4A					; J
		dc.b	$56					; V
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$AC					; 
		dc.b	$93					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$89					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$A8					; 
		dc.b	$9E					; 
		dc.b	$85					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$7C					; |
		dc.b	$9A					; 
		dc.b	$B5					; 
		dc.b	$C5					; 
		dc.b	$D5					; 
		dc.b	$D6					; 
		dc.b	$9B					; 
		dc.b	$33					; 3
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$38					; 8
		dc.b	$94					; 
		dc.b	$CE					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$95					; 
		dc.b	$42					; B
		dc.b	$27					; '
		dc.b	$40					; @
		dc.b	$8F					; 
		dc.b	$C1					; 
		dc.b	$BA					; 
		dc.b	$98					; 
		dc.b	$6C					; l
		dc.b	$4D					; M
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$83					; 
		dc.b	$B6					; 
		dc.b	$BF					; 
		dc.b	$94					; 
		dc.b	$5A					; Z
		dc.b	$32					; 2
		dc.b	$2B					; +
		dc.b	$46					; F
		dc.b	$71					; q
		dc.b	$94					; 
		dc.b	$B6					; 
		dc.b	$C8					; 
		dc.b	$B6					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$7E					; ~
		dc.b	$56					; V
		dc.b	$3C					; <
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$B4					; 
		dc.b	$AF					; 
		dc.b	$9E					; 
		dc.b	$90					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$77					; w
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$70					; p
		dc.b	$81					; 
		dc.b	$90					; 
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$B1					; 
		dc.b	$38					; 8
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$31					; 1
		dc.b	$71					; q
		dc.b	$AA					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$CD					; 
		dc.b	$98					; 
		dc.b	$5A					; Z
		dc.b	$41					; A
		dc.b	$5B					; [
		dc.b	$95					; 
		dc.b	$B2					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$94					; 
		dc.b	$A8					; 
		dc.b	$9D					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$8E					; 
		dc.b	$62					; b
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$3C					; <
		dc.b	$40					; @
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$A3					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$7B					; {
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$82					; 
		dc.b	$9E					; 
		dc.b	$BE					; 
		dc.b	$D6					; 
		dc.b	$D8					; 
		dc.b	$C7					; 
		dc.b	$43					; C
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$4A					; J
		dc.b	$A6					; 
		dc.b	$D5					; 
		dc.b	$D8					; 
		dc.b	$D4					; 
		dc.b	$9A					; 
		dc.b	$45					; E
		dc.b	$2A					; *
		dc.b	$50					; P
		dc.b	$9F					; 
		dc.b	$BC					; 
		dc.b	$A1					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$36					; 6
		dc.b	$27					; '
		dc.b	$33					; 3
		dc.b	$81					; 
		dc.b	$BA					; 
		dc.b	$BA					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$6A					; j
		dc.b	$56					; V
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$9E					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$3F					; ?
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$82					; 
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$BB					; 
		dc.b	$B9					; 
		dc.b	$AF					; 
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$70					; p
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$C4					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$B1					; 
		dc.b	$42					; B
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$31					; 1
		dc.b	$81					; 
		dc.b	$CD					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$9D					; 
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$4A					; J
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$B0					; 
		dc.b	$A1					; 
		dc.b	$88					; 
		dc.b	$6D					; m
		dc.b	$55					; U
		dc.b	$4D					; M
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$A0					; 
		dc.b	$AB					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$90					; 
		dc.b	$A8					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$A8					; 
		dc.b	$BD					; 
		dc.b	$BC					; 
		dc.b	$83					; 
		dc.b	$36					; 6
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$42					; B
		dc.b	$84					; 
		dc.b	$B7					; 
		dc.b	$CF					; 
		dc.b	$CA					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$66					; f
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$66					; f
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$6F					; o
		dc.b	$46					; F
		dc.b	$38					; 8
		dc.b	$48					; H
		dc.b	$61					; a
		dc.b	$7B					; {
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$97					; 
		dc.b	$9C					; 
		dc.b	$A1					; 
		dc.b	$9A					; 
		dc.b	$85					; 
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$74					; t
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$70					; p
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$B0					; 
		dc.b	$CB					; 
		dc.b	$D8					; 
		dc.b	$AA					; 
		dc.b	$35					; 5
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$28					; (
		dc.b	$55					; U
		dc.b	$B0					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$D7					; 
		dc.b	$A7					; 
		dc.b	$5E					; ^
		dc.b	$4D					; M
		dc.b	$6F					; o
		dc.b	$9A					; 
		dc.b	$8F					; 
		dc.b	$5E					; ^
		dc.b	$55					; U
		dc.b	$5D					; ]
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$84					; 
		dc.b	$B8					; 
		dc.b	$C5					; 
		dc.b	$9A					; 
		dc.b	$63					; c
		dc.b	$53					; S
		dc.b	$69					; i
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$A1					; 
		dc.b	$C0					; 
		dc.b	$D3					; 
		dc.b	$C4					; 
		dc.b	$94					; 
		dc.b	$57					; W
		dc.b	$2D					; -
		dc.b	$27					; '
		dc.b	$28					; (
		dc.b	$45					; E
		dc.b	$76					; v
		dc.b	$90					; 
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$77					; w
		dc.b	$5E					; ^
		dc.b	$51					; Q
		dc.b	$50					; P
		dc.b	$58					; X
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$AC					; 
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$AE					; 
		dc.b	$A2					; 
		dc.b	$4A					; J
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$2B					; +
		dc.b	$5B					; [
		dc.b	$A6					; 
		dc.b	$D4					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$D1					; 
		dc.b	$9B					; 
		dc.b	$66					; f
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$44					; D
		dc.b	$55					; U
		dc.b	$88					; 
		dc.b	$BB					; 
		dc.b	$D4					; 
		dc.b	$CD					; 
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$3A					; :
		dc.b	$2A					; *
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$B8					; 
		dc.b	$BC					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$61					; a
		dc.b	$38					; 8
		dc.b	$2F					; /
		dc.b	$3C					; <
		dc.b	$4B					; K
		dc.b	$5F					; _
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$40					; @
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$5A					; Z
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$2E					; .
		dc.b	$66					; f
		dc.b	$AE					; 
		dc.b	$D4					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$D6					; 
		dc.b	$A1					; 
		dc.b	$4F					; O
		dc.b	$2A					; *
		dc.b	$2A					; *
		dc.b	$48					; H
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$B9					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$39					; 9
		dc.b	$27					; '
		dc.b	$2A					; *
		dc.b	$44					; D
		dc.b	$6D					; m
		dc.b	$99					; 
		dc.b	$B8					; 
		dc.b	$C5					; 
		dc.b	$CA					; 
		dc.b	$C1					; 
		dc.b	$A8					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$48					; H
		dc.b	$38					; 8
		dc.b	$36					; 6
		dc.b	$3A					; :
		dc.b	$51					; Q
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$BB					; 
		dc.b	$B8					; 
		dc.b	$A5					; 
		dc.b	$90					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$47					; G
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$70					; p
		dc.b	$89					; 
		dc.b	$A9					; 
		dc.b	$BE					; 
		dc.b	$BF					; 
		dc.b	$B4					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$91					; 
		dc.b	$C6					; 
		dc.b	$C1					; 
		dc.b	$57					; W
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$D7					; 
		dc.b	$D8					; 
		dc.b	$D6					; 
		dc.b	$9A					; 
		dc.b	$3F					; ?
		dc.b	$27					; '
		dc.b	$2B					; +
		dc.b	$57					; W
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$71					; q
		dc.b	$48					; H
		dc.b	$2A					; *
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$5F					; _
		dc.b	$91					; 
		dc.b	$C6					; 
		dc.b	$D8					; 
		dc.b	$D8					; 
		dc.b	$CD					; 
		dc.b	$9E					; 
		dc.b	$85					; 
		dc.b	$6D					; m
		dc.b	$43					; C
		dc.b	$28					; (
		dc.b	$2A					; *
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$90					; 
		dc.b	$BF					; 
		dc.b	$D6					; 
		dc.b	$CA					; 
		dc.b	$A9					; 
		dc.b	$8F					; 
		dc.b	$7B					; {
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$44					; D
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$A4					; 
		dc.b	$BB					; 
		dc.b	$BF					; 
		dc.b	$B2					; 
		dc.b	$98					; 
		dc.b	$73					; s
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$AF					; 
		dc.b	$90					; 
		dc.b	$5F					; _
		dc.b	$37					; 7
		dc.b	$27					; '
		dc.b	$2C					; ,
		dc.b	$47					; G
		dc.b	$77					; w
		dc.b	$AD					; 
		dc.b	$C7					; 
		dc.b	$C1					; 
		dc.b	$B0					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$61					; a
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$93					; 
		dc.b	$9F					; 
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$66					; f
		dc.b	$65					; e
		dc.b	$69					; i
		dc.b	$6B					; k
		dc.b	$74					; t
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$A0					; 
		dc.b	$A5					; 
		dc.b	$98					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$5B					; [
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$9A					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$93					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$89					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6D					; m
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$56					; V
		dc.b	$40					; @
		dc.b	$47					; G
		dc.b	$65					; e
		dc.b	$85					; 
		dc.b	$97					; 
		dc.b	$9F					; 
		dc.b	$9F					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$74					; t
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$6D					; m
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$6F					; o
		dc.b	$70					; p
		dc.b	$7B					; {
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$93					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$62					; b
		dc.b	$5A					; Z
		dc.b	$5F					; _
		dc.b	$70					; p
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$70					; p
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$9E					; 
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$74					; t
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$66					; f
		dc.b	$62					; b
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$70					; p
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$74					; t
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$6D					; m
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$93					; 
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$69					; i
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$82					; 
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$70					; p
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$8C					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$89					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$74					; t
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$77					; w
		dc.b	$71					; q
		dc.b	$6D					; m
		dc.b	$70					; p
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$89					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$89					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$7E					; ~
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$8C					; 
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$6D					; m
		dc.b	$6D					; m
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$77					; w
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
PCM_Oooff:	dc.b	$80					; 		; DATA XREF: ROM:0001B084o
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$71					; q
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$3E					; >
		dc.b	$38					; 8
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$52					; R
		dc.b	$3A					; :
		dc.b	$4A					; J
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$50					; P
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$49					; I
		dc.b	$3A					; :
		dc.b	$4A					; J
		dc.b	$7C					; |
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$A4					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$53					; S
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$43					; C
		dc.b	$65					; e
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$49					; I
		dc.b	$64					; d
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$B6					; 
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$65					; e
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$C1					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$56					; V
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$5D					; ]
		dc.b	$7F					; 
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$CC					; 
		dc.b	$98					; 
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$C2					; 
		dc.b	$C4					; 
		dc.b	$8A					; 
		dc.b	$63					; c
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$5A					; Z
		dc.b	$BA					; 
		dc.b	$E7					; 
		dc.b	$B8					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$AA					; 
		dc.b	$B3					; 
		dc.b	$98					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$B7					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$64					; d
		dc.b	$41					; A
		dc.b	$3C					; <
		dc.b	$63					; c
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$78					; x
		dc.b	$48					; H
		dc.b	$34					; 4
		dc.b	$4F					; O
		dc.b	$87					; 
		dc.b	$B8					; 
		dc.b	$B7					; 
		dc.b	$AA					; 
		dc.b	$88					; 
		dc.b	$48					; H
		dc.b	$1F
		dc.b	$1E
		dc.b	$43					; C
		dc.b	$8F					; 
		dc.b	$CB					; 
		dc.b	$D7					; 
		dc.b	$CF					; 
		dc.b	$B6					; 
		dc.b	$7D					; }
		dc.b	$30					; 0
		dc.b	$17
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$D2					; 
		dc.b	$D9					; 
		dc.b	$B8					; 
		dc.b	$86					; 
		dc.b	$41					; A
		dc.b	$D
		dc.b	$20
		dc.b	$52					; R
		dc.b	$6E					; n
		dc.b	$A0					; 
		dc.b	$C6					; 
		dc.b	$C9					; 
		dc.b	$A3					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$A6					; 
		dc.b	$DB					; 
		dc.b	$C9					; 
		dc.b	$98					; 
		dc.b	$6B					; k
		dc.b	$34					; 4
		dc.b	$17
		dc.b	$37					; 7
		dc.b	$75					; u
		dc.b	$B3					; 
		dc.b	$CD					; 
		dc.b	$BA					; 
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$4D					; M
		dc.b	$27					; '
		dc.b	$2E					; .
		dc.b	$53					; S
		dc.b	$8E					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$80					; 
		dc.b	$3E					; >
		dc.b	$25					; %
		dc.b	$4C					; L
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$CF					; 
		dc.b	$CB					; 
		dc.b	$C1					; 
		dc.b	$AD					; 
		dc.b	$65					; e
		dc.b	$31					; 1
		dc.b	$4A					; J
		dc.b	$6F					; o
		dc.b	$92					; 
		dc.b	$C5					; 
		dc.b	$E4					; 
		dc.b	$B3					; 
		dc.b	$7A					; z
		dc.b	$4A					; J
		dc.b	$27					; '
		dc.b	$34					; 4
		dc.b	$60					; `
		dc.b	$9B					; 
		dc.b	$CB					; 
		dc.b	$DA					; 
		dc.b	$A3					; 
		dc.b	$4C					; L
		dc.b	$38					; 8
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$B8					; 
		dc.b	$95					; 
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$2E					; .
		dc.b	$4F					; O
		dc.b	$9F					; 
		dc.b	$DD					; 
		dc.b	$D2					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$3F					; ?
		dc.b	1
		dc.b	7
		dc.b	$5A					; Z
		dc.b	$A6					; 
		dc.b	$D4					; 
		dc.b	$DE					; 
		dc.b	$C2					; 
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$50					; P
		dc.b	$3C					; <
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$BB					; 
		dc.b	$B3					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$3B					; ;
		dc.b	$42					; B
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$81					; 
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$86					; 
		dc.b	$BB					; 
		dc.b	$CF					; 
		dc.b	$91					; 
		dc.b	$4A					; J
		dc.b	$2D					; -
		dc.b	$46					; F
		dc.b	$73					; s
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$BB					; 
		dc.b	$87					; 
		dc.b	$42					; B
		dc.b	$46					; F
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$A4					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$B6					; 
		dc.b	$AD					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$87					; 
		dc.b	$AE					; 
		dc.b	$AA					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$B4					; 
		dc.b	$CD					; 
		dc.b	$8D					; 
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$4D					; M
		dc.b	$8A					; 
		dc.b	$A9					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$8F					; 
		dc.b	$6C					; l
		dc.b	$42					; B
		dc.b	$3E					; >
		dc.b	$57					; W
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$B1					; 
		dc.b	$E5					; 
		dc.b	$C8					; 
		dc.b	$80					; 
		dc.b	$3A					; :
		dc.b	$E
		dc.b	$20
		dc.b	$49					; I
		dc.b	$6C					; l
		dc.b	$B0					; 
		dc.b	$F0					; 
		dc.b	$E9					; 
		dc.b	$B1					; 
		dc.b	$7A					; z
		dc.b	$45					; E
		dc.b	$2C					; ,
		dc.b	$3A					; :
		dc.b	$67					; g
		dc.b	$A0					; 
		dc.b	$BE					; 
		dc.b	$B3					; 
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$5D					; ]
		dc.b	$41					; A
		dc.b	$35					; 5
		dc.b	$5E					; ^
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$6C					; l
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$55					; U
		dc.b	$83					; 
		dc.b	$AB					; 
		dc.b	$B3					; 
		dc.b	$A4					; 
		dc.b	$8A					; 
		dc.b	$6C					; l
		dc.b	$3E					; >
		dc.b	$1F
		dc.b	$3B					; ;
		dc.b	$71					; q
		dc.b	$A2					; 
		dc.b	$CB					; 
		dc.b	$C2					; 
		dc.b	$88					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$67					; g
		dc.b	$A3					; 
		dc.b	$B3					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$2A					; *
		dc.b	$49					; I
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$AE					; 
		dc.b	$B7					; 
		dc.b	$88					; 
		dc.b	$64					; d
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$6A					; j
		dc.b	$35					; 5
		dc.b	$35					; 5
		dc.b	$68					; h
		dc.b	$9D					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$5E					; ^
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$4D					; M
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$63					; c
		dc.b	$34					; 4
		dc.b	$3E					; >
		dc.b	$7F					; 
		dc.b	$B6					; 
		dc.b	$B6					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$5C					; \
		dc.b	$18
		dc.b	9
		dc.b	$3E					; >
		dc.b	$9B					; 
		dc.b	$D6					; 
		dc.b	$E4					; 
		dc.b	$BD					; 
		dc.b	$8B					; 
		dc.b	$5E					; ^
		dc.b	$34					; 4
		dc.b	$37					; 7
		dc.b	$6F					; o
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$CC					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$3A					; :
		dc.b	$2E					; .
		dc.b	$35					; 5
		dc.b	$60					; `
		dc.b	$91					; 
		dc.b	$BA					; 
		dc.b	$BD					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$6C					; l
		dc.b	$30					; 0
		dc.b	$3B					; ;
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$C2					; 
		dc.b	$C6					; 
		dc.b	$BF					; 
		dc.b	$98					; 
		dc.b	$5A					; Z
		dc.b	$2C					; ,
		dc.b	$2A					; *
		dc.b	$53					; S
		dc.b	$8E					; 
		dc.b	$A3					; 
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$BA					; 
		dc.b	$C4					; 
		dc.b	$AB					; 
		dc.b	$67					; g
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$49					; I
		dc.b	$6C					; l
		dc.b	$AD					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$BF					; 
		dc.b	$A9					; 
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$3F					; ?
		dc.b	$37					; 7
		dc.b	$5A					; Z
		dc.b	$92					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$B6					; 
		dc.b	$C8					; 
		dc.b	$A9					; 
		dc.b	$6A					; j
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$3E					; >
		dc.b	$31					; 1
		dc.b	$67					; g
		dc.b	$9F					; 
		dc.b	$C4					; 
		dc.b	$D2					; 
		dc.b	$B1					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$55					; U
		dc.b	$26					; &
		dc.b	$41					; A
		dc.b	$6C					; l
		dc.b	$9B					; 
		dc.b	$C2					; 
		dc.b	$C8					; 
		dc.b	$A2					; 
		dc.b	$6C					; l
		dc.b	$26					; &
		dc.b	1
		dc.b	$2A					; *
		dc.b	$87					; 
		dc.b	$C8					; 
		dc.b	$C6					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$41					; A
		dc.b	$48					; H
		dc.b	$95					; 
		dc.b	$C9					; 
		dc.b	$A7					; 
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$64					; d
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$68					; h
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$80					; 
		dc.b	$BD					; 
		dc.b	$AB					; 
		dc.b	$4C					; L
		dc.b	$2E					; .
		dc.b	$5A					; Z
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$AB					; 
		dc.b	$BE					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$B7					; 
		dc.b	$D2					; 
		dc.b	$B0					; 
		dc.b	$79					; y
		dc.b	$48					; H
		dc.b	$12
		dc.b	$15
		dc.b	$5D					; ]
		dc.b	$9B					; 
		dc.b	$BB					; 
		dc.b	$C8					; 
		dc.b	$A3					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$8F					; 
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$4A					; J
		dc.b	$E
		dc.b	$25					; %
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$B3					; 
		dc.b	$C6					; 
		dc.b	$CF					; 
		dc.b	$BD					; 
		dc.b	$5E					; ^
		dc.b	$18
		dc.b	$35					; 5
		dc.b	$64					; d
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$37					; 7
		dc.b	$26					; &
		dc.b	$53					; S
		dc.b	$98					; 
		dc.b	$BE					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$9D					; 
		dc.b	$BA					; 
		dc.b	$80					; 
		dc.b	$3E					; >
		dc.b	$5A					; Z
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$43					; C
		dc.b	$7F					; 
		dc.b	$A7					; 
		dc.b	$7C					; |
		dc.b	$23					; #
		dc.b	$1E
		dc.b	$68					; h
		dc.b	$A4					; 
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$B4					; 
		dc.b	$9B					; 
		dc.b	$63					; c
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$59					; Y
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$96					; 
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$6B					; k
		dc.b	$3C					; <
		dc.b	$4F					; O
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$83					; 
		dc.b	$A0					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$76					; v
		dc.b	$B8					; 
		dc.b	$CB					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$7F					; 
		dc.b	$CC					; 
		dc.b	$EB					; 
		dc.b	$B0					; 
		dc.b	$3B					; ;
		dc.b	6
		dc.b	$12
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$B4					; 
		dc.b	$F0					; 
		dc.b	$EF					; 
		dc.b	$AE					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$B1					; 
		dc.b	$EC					; 
		dc.b	$DE					; 
		dc.b	$98					; 
		dc.b	$38					; 8
		dc.b	$12
		dc.b	$11
		dc.b	$38					; 8
		dc.b	$5C					; \
		dc.b	$8F					; 
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$CD					; 
		dc.b	$92					; 
		dc.b	$4F					; O
		dc.b	$22					; "
		dc.b	$38					; 8
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$AE					; 
		dc.b	$D7					; 
		dc.b	$BD					; 
		dc.b	$7A					; z
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$98					; 
		dc.b	$BB					; 
		dc.b	$D6					; 
		dc.b	$C6					; 
		dc.b	$8D					; 
		dc.b	$56					; V
		dc.b	$3F					; ?
		dc.b	$3A					; :
		dc.b	$5E					; ^
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$84					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$76					; v
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$AD					; 
		dc.b	$81					; 
		dc.b	$3A					; :
		dc.b	$35					; 5
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$CC					; 
		dc.b	$C9					; 
		dc.b	$A7					; 
		dc.b	$50					; P
		dc.b	2
		dc.b	$26					; &
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$BD					; 
		dc.b	$BD					; 
		dc.b	$83					; 
		dc.b	$48					; H
		dc.b	$3A					; :
		dc.b	$60					; `
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$D6					; 
		dc.b	$E7					; 
		dc.b	$7F					; 
		dc.b	$12
		dc.b	$D
		dc.b	$45					; E
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$98					; 
		dc.b	$CF					; 
		dc.b	$C5					; 
		dc.b	$7A					; z
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$3C					; <
		dc.b	$3A					; :
		dc.b	$65					; e
		dc.b	$A4					; 
		dc.b	$C9					; 
		dc.b	$D3					; 
		dc.b	$C1					; 
		dc.b	$8B					; 
		dc.b	$56					; V
		dc.b	$29					; )
		dc.b	$34					; 4
		dc.b	$65					; e
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$91					; 
		dc.b	$B3					; 
		dc.b	$CF					; 
		dc.b	$98					; 
		dc.b	$3A					; :
		dc.b	2
		dc.b	$15
		dc.b	$57					; W
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$C9					; 
		dc.b	$CC					; 
		dc.b	$94					; 
		dc.b	$31					; 1
		dc.b	1
		dc.b	$3B					; ;
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$C8					; 
		dc.b	$FD					; 
		dc.b	$D7					; 
		dc.b	$71					; q
		dc.b	$14
		dc.b	$1C
		dc.b	$3A					; :
		dc.b	$33					; 3
		dc.b	$48					; H
		dc.b	$75					; u
		dc.b	$B7					; 
		dc.b	$E0					; 
		dc.b	$CB					; 
		dc.b	$AA					; 
		dc.b	$84					; 
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$94					; 
		dc.b	$CD					; 
		dc.b	$D7					; 
		dc.b	$95					; 
		dc.b	$4C					; L
		dc.b	$29					; )
		dc.b	$30					; 0
		dc.b	$65					; e
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$AD					; 
		dc.b	$9C					; 
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$8E					; 
		dc.b	$A2					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$A2					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$46					; F
		dc.b	$3E					; >
		dc.b	$6C					; l
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$8F					; 
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$A7					; 
		dc.b	$E4					; 
		dc.b	$A4					; 
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$43					; C
		dc.b	$6B					; k
		dc.b	$A9					; 
		dc.b	$AE					; 
		dc.b	$73					; s
		dc.b	$49					; I
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$9B					; 
		dc.b	$E0					; 
		dc.b	$E4					; 
		dc.b	$A0					; 
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$94					; 
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$8F					; 
		dc.b	$BD					; 
		dc.b	$AA					; 
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$C6					; 
		dc.b	$FE					; 
		dc.b	$E2					; 
		dc.b	$8F					; 
		dc.b	$56					; V
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$6F					; o
		dc.b	$BA					; 
		dc.b	$E9					; 
		dc.b	$DE					; 
		dc.b	$9C					; 
		dc.b	$46					; F
		dc.b	$1B
		dc.b	$19
		dc.b	$2E					; .
		dc.b	$4F					; O
		dc.b	$8D					; 
		dc.b	$E8					; 
		dc.b	$FE					; 
		dc.b	$CD					; 
		dc.b	$7F					; 
		dc.b	$4A					; J
		dc.b	$3A					; :
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$49					; I
		dc.b	$7D					; }
		dc.b	$DD					; 
		dc.b	$F3					; 
		dc.b	$B8					; 
		dc.b	$6C					; l
		dc.b	$29					; )
		dc.b	$11
		dc.b	$19
		dc.b	$22					; "
		dc.b	$50					; P
		dc.b	$CC					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$8E					; 
		dc.b	$46					; F
		dc.b	$31					; 1
		dc.b	$4C					; L
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$6C					; l
		dc.b	$C2					; 
		dc.b	$EC					; 
		dc.b	$CD					; 
		dc.b	$86					; 
		dc.b	$4D					; M
		dc.b	$35					; 5
		dc.b	$2D					; -
		dc.b	$1E
		dc.b	$49					; I
		dc.b	$9B					; 
		dc.b	$DD					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$8B					; 
		dc.b	$52					; R
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$43					; C
		dc.b	$6E					; n
		dc.b	$BF					; 
		dc.b	$D2					; 
		dc.b	$8A					; 
		dc.b	$3A					; :
		dc.b	$57					; W
		dc.b	$88					; 
		dc.b	$5C					; \
		dc.b	$25					; %
		dc.b	$38					; 8
		dc.b	$88					; 
		dc.b	$B3					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$B7					; 
		dc.b	$A2					; 
		dc.b	$78					; x
		dc.b	$50					; P
		dc.b	$3C					; <
		dc.b	$56					; V
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$49					; I
		dc.b	$3C					; <
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$7C					; |
		dc.b	$A3					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$3A					; :
		dc.b	$42					; B
		dc.b	$75					; u
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$48					; H
		dc.b	7
		dc.b	$37					; 7
		dc.b	$A9					; 
		dc.b	$CC					; 
		dc.b	$A3					; 
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$BD					; 
		dc.b	$9D					; 
		dc.b	$59					; Y
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$3A					; :
		dc.b	$41					; A
		dc.b	$92					; 
		dc.b	$C4					; 
		dc.b	$B7					; 
		dc.b	$6E					; n
		dc.b	$37					; 7
		dc.b	$3C					; <
		dc.b	$3B					; ;
		dc.b	$42					; B
		dc.b	$87					; 
		dc.b	$DA					; 
		dc.b	$E1					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$60					; `
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$4C					; L
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$41					; A
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$C2					; 
		dc.b	$BB					; 
		dc.b	$B7					; 
		dc.b	$9C					; 
		dc.b	$53					; S
		dc.b	$29					; )
		dc.b	$43					; C
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$CF					; 
		dc.b	$C2					; 
		dc.b	$76					; v
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$A3					; 
		dc.b	$E5					; 
		dc.b	$E4					; 
		dc.b	$B1					; 
		dc.b	$71					; q
		dc.b	$2D					; -
		dc.b	$11
		dc.b	$33					; 3
		dc.b	$63					; c
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$AD					; 
		dc.b	$6E					; n
		dc.b	$27					; '
		dc.b	$D
		dc.b	$37					; 7
		dc.b	$84					; 
		dc.b	$BA					; 
		dc.b	$C5					; 
		dc.b	$B6					; 
		dc.b	$A0					; 
		dc.b	$A2					; 
		dc.b	$91					; 
		dc.b	$3F					; ?
		dc.b	9
		dc.b	$3F					; ?
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$95					; 
		dc.b	$BD					; 
		dc.b	$EE					; 
		dc.b	$B6					; 
		dc.b	$55					; U
		dc.b	2
		dc.b	1
		dc.b	$43					; C
		dc.b	$8B					; 
		dc.b	$B3					; 
		dc.b	$D7					; 
		dc.b	$E1					; 
		dc.b	$9F					; 
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$4C					; L
		dc.b	$41					; A
		dc.b	$8A					; 
		dc.b	$C8					; 
		dc.b	$B8					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$52					; R
		dc.b	$10
		dc.b	$B
		dc.b	$43					; C
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$E5					; 
		dc.b	$F5					; 
		dc.b	$D4					; 
		dc.b	$A9					; 
		dc.b	$57					; W
		dc.b	4
		dc.b	$26					; &
		dc.b	$84					; 
		dc.b	$BB					; 
		dc.b	$CB					; 
		dc.b	$AA					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$1F
		dc.b	1
		dc.b	$26					; &
		dc.b	$79					; y
		dc.b	$A7					; 
		dc.b	$BD					; 
		dc.b	$D4					; 
		dc.b	$D3					; 
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$3A					; :
		dc.b	$4D					; M
		dc.b	$96					; 
		dc.b	$B3					; 
		dc.b	$8A					; 
		dc.b	$4C					; L
		dc.b	$50					; P
		dc.b	$98					; 
		dc.b	$B1					; 
		dc.b	$83					; 
		dc.b	$4C					; L
		dc.b	$45					; E
		dc.b	$37					; 7
		dc.b	$22					; "
		dc.b	$3E					; >
		dc.b	$96					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$C2					; 
		dc.b	$65					; e
		dc.b	$5E					; ^
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$37					; 7
		dc.b	$6A					; j
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$2C					; ,
		dc.b	$2C					; ,
		dc.b	$6A					; j
		dc.b	$B8					; 
		dc.b	$C4					; 
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$7F					; 
		dc.b	$20
		dc.b	$E
		dc.b	$49					; I
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$C1					; 
		dc.b	$BD					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$35					; 5
		dc.b	$10
		dc.b	$43					; C
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$C8					; 
		dc.b	$EB					; 
		dc.b	$D7					; 
		dc.b	$68					; h
		dc.b	$2A					; *
		dc.b	$2D					; -
		dc.b	$43					; C
		dc.b	$3F					; ?
		dc.b	$59					; Y
		dc.b	$AA					; 
		dc.b	$D6					; 
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$6A					; j
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$CC					; 
		dc.b	$B1					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$46					; F
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$AA					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$84					; 
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$AE					; 
		dc.b	$E8					; 
		dc.b	$CD					; 
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$53					; S
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$9B					; 
		dc.b	$D2					; 
		dc.b	$C2					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$61					; a
		dc.b	$2E					; .
		dc.b	$23					; #
		dc.b	$4F					; O
		dc.b	$A6					; 
		dc.b	$E4					; 
		dc.b	$D9					; 
		dc.b	$A3					; 
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$3C					; <
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$C1					; 
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$55					; U
		dc.b	$48					; H
		dc.b	$83					; 
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$B3					; 
		dc.b	$88					; 
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$6A					; j
		dc.b	$8E					; 
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$B4					; 
		dc.b	$CB					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$5E					; ^
		dc.b	$2D					; -
		dc.b	$25					; %
		dc.b	$56					; V
		dc.b	$B7					; 
		dc.b	$D6					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$67					; g
		dc.b	$37					; 7
		dc.b	$27					; '
		dc.b	$73					; s
		dc.b	$B8					; 
		dc.b	$BF					; 
		dc.b	$AA					; 
		dc.b	$B6					; 
		dc.b	$B8					; 
		dc.b	$81					; 
		dc.b	$2E					; .
		dc.b	1
		dc.b	$18
		dc.b	$68					; h
		dc.b	$A6					; 
		dc.b	$C2					; 
		dc.b	$E1					; 
		dc.b	$CB					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$42					; B
		dc.b	$19
		dc.b	$19
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$B6					; 
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$55					; U
		dc.b	$3C					; <
		dc.b	$59					; Y
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$AE					; 
		dc.b	$84					; 
		dc.b	$3A					; :
		dc.b	$43					; C
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$BA					; 
		dc.b	$C6					; 
		dc.b	$C1					; 
		dc.b	$A7					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$33					; 3
		dc.b	7
		dc.b	$33					; 3
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$92					; 
		dc.b	$CC					; 
		dc.b	$D2					; 
		dc.b	$A0					; 
		dc.b	$78					; x
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$CF					; 
		dc.b	$C2					; 
		dc.b	$5E					; ^
		dc.b	$29					; )
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$6B					; k
		dc.b	$9F					; 
		dc.b	$DD					; 
		dc.b	$C8					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$B7					; 
		dc.b	$DE					; 
		dc.b	$BA					; 
		dc.b	$73					; s
		dc.b	$4D					; M
		dc.b	$29					; )
		dc.b	$15
		dc.b	$52					; R
		dc.b	$AD					; 
		dc.b	$D3					; 
		dc.b	$A3					; 
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$99					; 
		dc.b	$C4					; 
		dc.b	$99					; 
		dc.b	$5D					; ]
		dc.b	$2E					; .
		dc.b	$11
		dc.b	$35					; 5
		dc.b	$B1					; 
		dc.b	$FD					; 
		dc.b	$DD					; 
		dc.b	$80					; 
		dc.b	$19
		dc.b	4
		dc.b	$53					; S
		dc.b	$91					; 
		dc.b	$BA					; 
		dc.b	$C8					; 
		dc.b	$86					; 
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$2D					; -
		dc.b	$33					; 3
		dc.b	$8A					; 
		dc.b	$BE					; 
		dc.b	$DE					; 
		dc.b	$92					; 
		dc.b	$2E					; .
		dc.b	$2C					; ,
		dc.b	$6A					; j
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$37					; 7
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$5A					; Z
		dc.b	$BA					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$79					; y
		dc.b	$3B					; ;
		dc.b	$49					; I
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$DB					; 
		dc.b	$D0					; 
		dc.b	$6C					; l
		dc.b	$B
		dc.b	$B
		dc.b	$1B
		dc.b	$5A					; Z
		dc.b	$E7					; 
		dc.b	$FE					; 
		dc.b	$B3					; 
		dc.b	$7A					; z
		dc.b	$3C					; <
		dc.b	$1B
		dc.b	$55					; U
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$9B					; 
		dc.b	$BD					; 
		dc.b	$CD					; 
		dc.b	$DE					; 
		dc.b	$9F					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$1F
		dc.b	$2A					; *
		dc.b	$83					; 
		dc.b	$B4					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$A0					; 
		dc.b	$7C					; |
		dc.b	$26					; &
		dc.b	$35					; 5
		dc.b	$6E					; n
		dc.b	$A2					; 
		dc.b	$A9					; 
		dc.b	$D3					; 
		dc.b	$CC					; 
		dc.b	$7F					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$3F					; ?
		dc.b	$8F					; 
		dc.b	$BB					; 
		dc.b	$8F					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$4A					; J
		dc.b	$46					; F
		dc.b	$63					; c
		dc.b	$8D					; 
		dc.b	$C5					; 
		dc.b	$D7					; 
		dc.b	$8D					; 
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$50					; P
		dc.b	$73					; s
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$68					; h
		dc.b	$96					; 
		dc.b	$BE					; 
		dc.b	$A7					; 
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$BA					; 
		dc.b	$BA					; 
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$6C					; l
		dc.b	$1C
		dc.b	$11
		dc.b	$1E
		dc.b	$81					; 
		dc.b	$F3					; 
		dc.b	$E8					; 
		dc.b	$CF					; 
		dc.b	$BE					; 
		dc.b	$71					; q
		dc.b	$19
		dc.b	$B
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$8D					; 
		dc.b	$12
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$37					; 7
		dc.b	$D2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C2					; 
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$35					; 5
		dc.b	$1E
		dc.b	$2A					; *
		dc.b	$22					; "
		dc.b	$46					; F
		dc.b	$CD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$34					; 4
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$68					; h
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F9					; 
		dc.b	$C4					; 
		dc.b	$B4					; 
		dc.b	$57					; W
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$7C					; |
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FD					; 
		dc.b	$8F					; 
		dc.b	1
		dc.b	1
		dc.b	$3A					; :
		dc.b	$30					; 0
		dc.b	$15
		dc.b	$4F					; O
		dc.b	$AE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CD					; 
		dc.b	$68					; h
		dc.b	$37					; 7
		dc.b	2
		dc.b	7
		dc.b	$2D					; -
		dc.b	$91					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$BF					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$3C					; <
		dc.b	$3F					; ?
		dc.b	$2C					; ,
		dc.b	$3F					; ?
		dc.b	$73					; s
		dc.b	$A0					; 
		dc.b	$D6					; 
		dc.b	$D2					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$41					; A
		dc.b	$18
		dc.b	$1E
		dc.b	$75					; u
		dc.b	$DB					; 
		dc.b	$E4					; 
		dc.b	$C9					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$3E					; >
		dc.b	$57					; W
		dc.b	$80					; 
		dc.b	$AE					; 
		dc.b	$D9					; 
		dc.b	$98					; 
		dc.b	$53					; S
		dc.b	$38					; 8
		dc.b	$3B					; ;
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	7
		dc.b	$33					; 3
		dc.b	$BF					; 
		dc.b	$FE					; 
		dc.b	$C6					; 
		dc.b	$68					; h
		dc.b	$38					; 8
		dc.b	$3B					; ;
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$E8					; 
		dc.b	$C9					; 
		dc.b	$64					; d
		dc.b	$35					; 5
		dc.b	$6C					; l
		dc.b	$B7					; 
		dc.b	$A6					; 
		dc.b	$5A					; Z
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$8F					; 
		dc.b	$3E					; >
		dc.b	$10
		dc.b	$37					; 7
		dc.b	$6B					; k
		dc.b	$9B					; 
		dc.b	$C6					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$4D					; M
		dc.b	$2A					; *
		dc.b	$45					; E
		dc.b	$8D					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$BA					; 
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$B6					; 
		dc.b	$E0					; 
		dc.b	$CF					; 
		dc.b	$5C					; \
		dc.b	3
		dc.b	$10
		dc.b	$50					; P
		dc.b	$83					; 
		dc.b	$D3					; 
		dc.b	$E9					; 
		dc.b	$B0					; 
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$B6					; 
		dc.b	$9D					; 
		dc.b	$78					; x
		dc.b	$3E					; >
		dc.b	9
		dc.b	$6E					; n
		dc.b	$EF					; 
		dc.b	$AD					; 
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$41					; A
		dc.b	$31					; 1
		dc.b	$55					; U
		dc.b	$AB					; 
		dc.b	$E2					; 
		dc.b	$F9					; 
		dc.b	$9B					; 
		dc.b	$3A					; :
		dc.b	$31					; 1
		dc.b	$3C					; <
		dc.b	$91					; 
		dc.b	$B4					; 
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$B1					; 
		dc.b	$CD					; 
		dc.b	$C4					; 
		dc.b	$75					; u
		dc.b	$27					; '
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$BD					; 
		dc.b	$D7					; 
		dc.b	$C9					; 
		dc.b	$7F					; 
		dc.b	$20
		dc.b	$4F					; O
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$49					; I
		dc.b	$48					; H
		dc.b	$8B					; 
		dc.b	$E1					; 
		dc.b	$E1					; 
		dc.b	$C9					; 
		dc.b	$7F					; 
		dc.b	$35					; 5
		dc.b	$2D					; -
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$92					; 
		dc.b	$C1					; 
		dc.b	$9C					; 
		dc.b	$BA					; 
		dc.b	$7C					; |
		dc.b	$10
		dc.b	$65					; e
		dc.b	$A9					; 
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$CB					; 
		dc.b	$FE					; 
		dc.b	$92					; 
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$42					; B
		dc.b	$5D					; ]
		dc.b	$A2					; 
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$B1					; 
		dc.b	$8F					; 
		dc.b	$43					; C
		dc.b	$2A					; *
		dc.b	$D
		dc.b	$38					; 8
		dc.b	$87					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$C1					; 
		dc.b	$81					; 
		dc.b	4
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$BA					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$B4					; 
		dc.b	$E0					; 
		dc.b	$BF					; 
		dc.b	$60					; `
		dc.b	3
		dc.b	1
		dc.b	$2A					; *
		dc.b	$33					; 3
		dc.b	$52					; R
		dc.b	$BF					; 
		dc.b	$F0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$AE					; 
		dc.b	$42					; B
		dc.b	1
		dc.b	1
		dc.b	$34					; 4
		dc.b	$A4					; 
		dc.b	$C2					; 
		dc.b	$C9					; 
		dc.b	$E2					; 
		dc.b	$CC					; 
		dc.b	$73					; s
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$22					; "
		dc.b	6
		dc.b	$2C					; ,
		dc.b	$7F					; 
		dc.b	$E5					; 
		dc.b	$FE					; 
		dc.b	$E0					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$53					; S
		dc.b	$17
		dc.b	$19
		dc.b	$50					; P
		dc.b	$7F					; 
		dc.b	$CB					; 
		dc.b	$FE					; 
		dc.b	$EE					; 
		dc.b	$C5					; 
		dc.b	$9B					; 
		dc.b	$5E					; ^
		dc.b	6
		dc.b	1
		dc.b	1
		dc.b	$57					; W
		dc.b	$EB					; 
		dc.b	$FD					; 
		dc.b	$BE					; 
		dc.b	$AA					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$15
		dc.b	$14
		dc.b	$4F					; O
		dc.b	$81					; 
		dc.b	$BF					; 
		dc.b	$F0					; 
		dc.b	$FC					; 
		dc.b	$F6					; 
		dc.b	$A4					; 
		dc.b	$D
		dc.b	1
		dc.b	1
		dc.b	$3B					; ;
		dc.b	$98					; 
		dc.b	$E9					; 
		dc.b	$E2					; 
		dc.b	$F2					; 
		dc.b	$D4					; 
		dc.b	$53					; S
		dc.b	$10
		dc.b	$1C
		dc.b	$3C					; <
		dc.b	$42					; B
		dc.b	$5D					; ]
		dc.b	$BA					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CF					; 
		dc.b	$50					; P
		dc.b	$1C
		dc.b	$2C					; ,
		dc.b	$3A					; :
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$EF					; 
		dc.b	$FE					; 
		dc.b	$D9					; 
		dc.b	$67					; g
		dc.b	$27					; '
		dc.b	$43					; C
		dc.b	$6E					; n
		dc.b	$42					; B
		dc.b	$A
		dc.b	$2D					; -
		dc.b	$C8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A7					; 
		dc.b	$11
		dc.b	1
		dc.b	7
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$84					; 
		dc.b	$C8					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$37					; 7
		dc.b	$37					; 7
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$68					; h
		dc.b	$95					; 
		dc.b	$D6					; 
		dc.b	$8B					; 
		dc.b	$3F					; ?
		dc.b	$2A					; *
		dc.b	1
		dc.b	$60					; `
		dc.b	$F9					; 
		dc.b	$DB					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$5E					; ^
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$5A					; Z
		dc.b	$46					; F
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$C9					; 
		dc.b	$D4					; 
		dc.b	$D6					; 
		dc.b	$79					; y
		dc.b	1
		dc.b	1
		dc.b	$4A					; J
		dc.b	$CF					; 
		dc.b	$EE					; 
		dc.b	$BE					; 
		dc.b	$D3					; 
		dc.b	$8E					; 
		dc.b	$22					; "
		dc.b	$3A					; :
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$4D					; M
		dc.b	$3A					; :
		dc.b	$D6					; 
		dc.b	$FE					; 
		dc.b	$C6					; 
		dc.b	$3A					; :
		dc.b	$12
		dc.b	$14
		dc.b	$37					; 7
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$B3					; 
		dc.b	$DB					; 
		dc.b	$DE					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$1C
		dc.b	$26					; &
		dc.b	$46					; F
		dc.b	$6C					; l
		dc.b	$F2					; 
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$7A					; z
		dc.b	$2D					; -
		dc.b	1
		dc.b	$18
		dc.b	$94					; 
		dc.b	$7C					; |
		dc.b	$57					; W
		dc.b	$BE					; 
		dc.b	$F3					; 
		dc.b	$E5					; 
		dc.b	$C8					; 
		dc.b	$4F					; O
		dc.b	$18
		dc.b	$11
		dc.b	$E
		dc.b	$53					; S
		dc.b	$91					; 
		dc.b	$C4					; 
		dc.b	$DE					; 
		dc.b	$CB					; 
		dc.b	$BF					; 
		dc.b	$AA					; 
		dc.b	$33					; 3
		dc.b	1
		dc.b	2
		dc.b	$3C					; <
		dc.b	$98					; 
		dc.b	$EE					; 
		dc.b	$D0					; 
		dc.b	$C8					; 
		dc.b	$DB					; 
		dc.b	$64					; d
		dc.b	$E
		dc.b	1
		dc.b	1
		dc.b	$46					; F
		dc.b	$DD					; 
		dc.b	$FE					; 
		dc.b	$FA					; 
		dc.b	$A7					; 
		dc.b	$6B					; k
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$17
		dc.b	$11
		dc.b	$55					; U
		dc.b	$84					; 
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$9D					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$43					; C
		dc.b	$57					; W
		dc.b	$29					; )
		dc.b	$3B					; ;
		dc.b	$C1					; 
		dc.b	$D6					; 
		dc.b	$B0					; 
		dc.b	$AB					; 
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$C5					; 
		dc.b	$B4					; 
		dc.b	$5A					; Z
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$6F					; o
		dc.b	$84					; 
		dc.b	$4D					; M
		dc.b	$3E					; >
		dc.b	$A3					; 
		dc.b	$F0					; 
		dc.b	$AB					; 
		dc.b	$3B					; ;
		dc.b	$35					; 5
		dc.b	$72					; r
		dc.b	$A4					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$4D					; M
		dc.b	$49					; I
		dc.b	$78					; x
		dc.b	$A6					; 
		dc.b	$CF					; 
		dc.b	$C6					; 
		dc.b	$78					; x
		dc.b	$42					; B
		dc.b	$64					; d
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$A4					; 
		dc.b	$A9					; 
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$88					; 
		dc.b	$63					; c
		dc.b	$26					; &
		dc.b	1
		dc.b	$34					; 4
		dc.b	$EF					; 
		dc.b	$FA					; 
		dc.b	$C4					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$31					; 1
		dc.b	$12
		dc.b	$45					; E
		dc.b	$84					; 
		dc.b	$BD					; 
		dc.b	$DA					; 
		dc.b	$CF					; 
		dc.b	$B6					; 
		dc.b	$6A					; j
		dc.b	$12
		dc.b	$46					; F
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$E9					; 
		dc.b	$A4					; 
		dc.b	$5A					; Z
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$8A					; 
		dc.b	$55					; U
		dc.b	$50					; P
		dc.b	$BF					; 
		dc.b	$CB					; 
		dc.b	$75					; u
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$CD					; 
		dc.b	$C1					; 
		dc.b	$1F
		dc.b	$10
		dc.b	$64					; d
		dc.b	$A0					; 
		dc.b	$E4					; 
		dc.b	$AE					; 
		dc.b	$4A					; J
		dc.b	$4C					; L
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$B6					; 
		dc.b	$B8					; 
		dc.b	$79					; y
		dc.b	$31					; 1
		dc.b	$2A					; *
		dc.b	$6B					; k
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$AE					; 
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$4A					; J
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$96					; 
		dc.b	$BB					; 
		dc.b	$99					; 
		dc.b	$50					; P
		dc.b	$30					; 0
		dc.b	$3F					; ?
		dc.b	$80					; 
		dc.b	$C6					; 
		dc.b	$AD					; 
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$B6					; 
		dc.b	$D3					; 
		dc.b	$84					; 
		dc.b	$49					; I
		dc.b	$A
		dc.b	1
		dc.b	$56					; V
		dc.b	$CB					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$CB					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$19
		dc.b	$33					; 3
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$DA					; 
		dc.b	$F2					; 
		dc.b	$94					; 
		dc.b	$3E					; >
		dc.b	$14
		dc.b	$38					; 8
		dc.b	$5A					; Z
		dc.b	$2D					; -
		dc.b	$19
		dc.b	$81					; 
		dc.b	$D3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$7F					; 
		dc.b	$1F
		dc.b	$D
		dc.b	$B
		dc.b	$3B					; ;
		dc.b	$AA					; 
		dc.b	$CD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A2					; 
		dc.b	$41					; A
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$25					; %
		dc.b	$E
		dc.b	$5C					; \
		dc.b	$B3					; 
		dc.b	$BE					; 
		dc.b	$CF					; 
		dc.b	$FA					; 
		dc.b	$DB					; 
		dc.b	$78					; x
		dc.b	1
		dc.b	1
		dc.b	$6A					; j
		dc.b	$E0					; 
		dc.b	$DB					; 
		dc.b	$B7					; 
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$C2					; 
		dc.b	$6E					; n
		dc.b	$20
		dc.b	$53					; S
		dc.b	$3E					; >
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$56					; V
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$38					; 8
		dc.b	$67					; g
		dc.b	$E8					; 
		dc.b	$D3					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$60					; `
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$56					; V
		dc.b	$37					; 7
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$FE					; 
		dc.b	$E8					; 
		dc.b	$76					; v
		dc.b	$3B					; ;
		dc.b	1
		dc.b	$E
		dc.b	$AB					; 
		dc.b	$D2					; 
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$EE					; 
		dc.b	$E0					; 
		dc.b	$A7					; 
		dc.b	$64					; d
		dc.b	$29					; )
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$37					; 7
		dc.b	2
		dc.b	$6E					; n
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$63					; c
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$3A					; :
		dc.b	$3F					; ?
		dc.b	$C5					; 
		dc.b	$DB					; 
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$55					; U
		dc.b	$27					; '
		dc.b	$79					; y
		dc.b	$A3					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$57					; W
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$33					; 3
		dc.b	$15
		dc.b	$5A					; Z
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$AB					; 
		dc.b	$DB					; 
		dc.b	$E7					; 
		dc.b	$83					; 
		dc.b	$22					; "
		dc.b	1
		dc.b	$38					; 8
		dc.b	$9F					; 
		dc.b	$B4					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$A3					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$95					; 
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$88					; 
		dc.b	$A9					; 
		dc.b	$B0					; 
		dc.b	$75					; u
		dc.b	$4D					; M
		dc.b	$8D					; 
		dc.b	$BA					; 
		dc.b	$6E					; n
		dc.b	$2C					; ,
		dc.b	$48					; H
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$73					; s
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$22					; "
		dc.b	$1B
		dc.b	$8F					; 
		dc.b	$F9					; 
		dc.b	$F7					; 
		dc.b	$AD					; 
		dc.b	$5A					; Z
		dc.b	$29					; )
		dc.b	$20
		dc.b	$56					; V
		dc.b	$99					; 
		dc.b	$A9					; 
		dc.b	$C4					; 
		dc.b	$B1					; 
		dc.b	$46					; F
		dc.b	$3C					; <
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$45					; E
		dc.b	$3B					; ;
		dc.b	$50					; P
		dc.b	$7C					; |
		dc.b	$E1					; 
		dc.b	$CD					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$55					; U
		dc.b	$1C
		dc.b	$64					; d
		dc.b	$99					; 
		dc.b	$CB					; 
		dc.b	$E5					; 
		dc.b	$80					; 
		dc.b	$48					; H
		dc.b	$91					; 
		dc.b	$6A					; j
		dc.b	$48					; H
		dc.b	$AE					; 
		dc.b	$7C					; |
		dc.b	$42					; B
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$E2					; 
		dc.b	$8D					; 
		dc.b	$2E					; .
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$49					; I
		dc.b	$2E					; .
		dc.b	$4F					; O
		dc.b	$D6					; 
		dc.b	$F6					; 
		dc.b	$9C					; 
		dc.b	$6A					; j
		dc.b	$3A					; :
		dc.b	$33					; 3
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$78					; x
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$CB					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$26					; &
		dc.b	$5A					; Z
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$37					; 7
		dc.b	$43					; C
		dc.b	$7A					; z
		dc.b	$A2					; 
		dc.b	$B8					; 
		dc.b	$7C					; |
		dc.b	$2E					; .
		dc.b	$86					; 
		dc.b	$EC					; 
		dc.b	$B1					; 
		dc.b	$55					; U
		dc.b	$27					; '
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$99					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A9					; 
		dc.b	$42					; B
		dc.b	1
		dc.b	1
		dc.b	$48					; H
		dc.b	$B4					; 
		dc.b	$C4					; 
		dc.b	$E2					; 
		dc.b	$EB					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$20
		dc.b	$3A					; :
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$AE					; 
		dc.b	$A2					; 
		dc.b	$B4					; 
		dc.b	$CF					; 
		dc.b	$9D					; 
		dc.b	$73					; s
		dc.b	$3B					; ;
		dc.b	$A
		dc.b	$3C					; <
		dc.b	$A0					; 
		dc.b	$AE					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$5C					; \
		dc.b	$20
		dc.b	$72					; r
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$84					; 
		dc.b	$46					; F
		dc.b	$45					; E
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$7A					; z
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$91					; 
		dc.b	$60					; `
		dc.b	$94					; 
		dc.b	$AE					; 
		dc.b	$5D					; ]
		dc.b	$8B					; 
		dc.b	$A7					; 
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$49					; I
		dc.b	$33					; 3
		dc.b	$98					; 
		dc.b	$B3					; 
		dc.b	$3E					; >
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$AE					; 
		dc.b	$DE					; 
		dc.b	$A2					; 
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$4F					; O
		dc.b	$6F					; o
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$55					; U
		dc.b	$A9					; 
		dc.b	$DE					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$3E					; >
		dc.b	$78					; x
		dc.b	$53					; S
		dc.b	$34					; 4
		dc.b	$6E					; n
		dc.b	$CF					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$A9					; 
		dc.b	$25					; %
		dc.b	$2C					; ,
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$BE					; 
		dc.b	$B4					; 
		dc.b	$55					; U
		dc.b	$3F					; ?
		dc.b	$53					; S
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$9C					; 
		dc.b	$BB					; 
		dc.b	$CD					; 
		dc.b	$67					; g
		dc.b	2
		dc.b	$4F					; O
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$C5					; 
		dc.b	$BD					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$A6					; 
		dc.b	$71					; q
		dc.b	$B
		dc.b	$4C					; L
		dc.b	$60					; `
		dc.b	$9B					; 
		dc.b	$EB					; 
		dc.b	$C4					; 
		dc.b	$A3					; 
		dc.b	$65					; e
		dc.b	1
		dc.b	$19
		dc.b	$8D					; 
		dc.b	$B1					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$60					; `
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$98					; 
		dc.b	$BB					; 
		dc.b	$34					; 4
		dc.b	$1C
		dc.b	$76					; v
		dc.b	$A6					; 
		dc.b	$D0					; 
		dc.b	$D4					; 
		dc.b	$79					; y
		dc.b	$50					; P
		dc.b	$63					; c
		dc.b	$96					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$43					; C
		dc.b	$6B					; k
		dc.b	$B6					; 
		dc.b	$BB					; 
		dc.b	$72					; r
		dc.b	$3E					; >
		dc.b	$4F					; O
		dc.b	$96					; 
		dc.b	$D3					; 
		dc.b	$C2					; 
		dc.b	$9C					; 
		dc.b	$5E					; ^
		dc.b	$42					; B
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$3F					; ?
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$B1					; 
		dc.b	$D0					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$1C
		dc.b	$31					; 1
		dc.b	$63					; c
		dc.b	$B7					; 
		dc.b	$FA					; 
		dc.b	$9D					; 
		dc.b	$96					; 
		dc.b	$BD					; 
		dc.b	$C2					; 
		dc.b	$8F					; 
		dc.b	$18
		dc.b	1
		dc.b	9
		dc.b	$99					; 
		dc.b	$BE					; 
		dc.b	$E5					; 
		dc.b	$E1					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$1B
		dc.b	$20
		dc.b	$3C					; <
		dc.b	$35					; 5
		dc.b	$80					; 
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$A7					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$AB					; 
		dc.b	$8F					; 
		dc.b	$3B					; ;
		dc.b	$27					; '
		dc.b	$67					; g
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$2C					; ,
		dc.b	$64					; d
		dc.b	$D9					; 
		dc.b	$8F					; 
		dc.b	$45					; E
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$DB					; 
		dc.b	$B7					; 
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$B3					; 
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	$76					; v
		dc.b	$C4					; 
		dc.b	$5C					; \
		dc.b	$45					; E
		dc.b	$6E					; n
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$AE					; 
		dc.b	$78					; x
		dc.b	$3A					; :
		dc.b	$8E					; 
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$A
		dc.b	$E1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$BD					; 
		dc.b	$23					; #
		dc.b	$2A					; *
		dc.b	$83					; 
		dc.b	$37					; 7
		dc.b	$1C
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$20
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$64					; d
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$B8					; 
		dc.b	$E9					; 
		dc.b	$C1					; 
		dc.b	$17
		dc.b	$18
		dc.b	$55					; U
		dc.b	$5E					; ^
		dc.b	$84					; 
		dc.b	$EC					; 
		dc.b	$BD					; 
		dc.b	$6A					; j
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	1
		dc.b	$2A					; *
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$BF					; 
		dc.b	$E7					; 
		dc.b	$49					; I
		dc.b	$46					; F
		dc.b	$73					; s
		dc.b	$3A					; :
		dc.b	$73					; s
		dc.b	$A7					; 
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$D2					; 
		dc.b	$C4					; 
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$78					; x
		dc.b	$2E					; .
		dc.b	1
		dc.b	9
		dc.b	$80					; 
		dc.b	$CC					; 
		dc.b	$A7					; 
		dc.b	$7D					; }
		dc.b	$C1					; 
		dc.b	$D4					; 
		dc.b	$76					; v
		dc.b	$41					; A
		dc.b	$35					; 5
		dc.b	1
		dc.b	$53					; S
		dc.b	$E8					; 
		dc.b	$B7					; 
		dc.b	$B0					; 
		dc.b	$8D					; 
		dc.b	$57					; W
		dc.b	$D0					; 
		dc.b	$D6					; 
		dc.b	$1F
		dc.b	$E
		dc.b	$5E					; ^
		dc.b	$45					; E
		dc.b	$65					; e
		dc.b	$A2					; 
		dc.b	$8B					; 
		dc.b	$C2					; 
		dc.b	$DA					; 
		dc.b	$81					; 
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$2A					; *
		dc.b	$5E					; ^
		dc.b	$B4					; 
		dc.b	$C6					; 
		dc.b	$D4					; 
		dc.b	$BD					; 
		dc.b	$87					; 
		dc.b	$3A					; :
		dc.b	$20
		dc.b	$2C					; ,
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$A4					; 
		dc.b	$FE					; 
		dc.b	$C5					; 
		dc.b	$38					; 8
		dc.b	$19
		dc.b	$23					; #
		dc.b	$5D					; ]
		dc.b	$B4					; 
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$E0					; 
		dc.b	$AD					; 
		dc.b	$99					; 
		dc.b	$A2					; 
		dc.b	$A
		dc.b	1
		dc.b	$7A					; z
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$8D					; 
		dc.b	$4D					; M
		dc.b	$78					; x
		dc.b	$A2					; 
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$31					; 1
		dc.b	$59					; Y
		dc.b	$FA					; 
		dc.b	$E2					; 
		dc.b	$E7					; 
		dc.b	$8B					; 
		dc.b	1
		dc.b	$23					; #
		dc.b	$30					; 0
		dc.b	$23					; #
		dc.b	$96					; 
		dc.b	$BB					; 
		dc.b	$8F					; 
		dc.b	$C1					; 
		dc.b	$BD					; 
		dc.b	$B4					; 
		dc.b	$B4					; 
		dc.b	$42					; B
		dc.b	$E
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$8A					; 
		dc.b	$B7					; 
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$D3					; 
		dc.b	$AA					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$2D					; -
		dc.b	$57					; W
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$4A					; J
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$D7					; 
		dc.b	$C1					; 
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$B4					; 
		dc.b	$BD					; 
		dc.b	$78					; x
		dc.b	$3E					; >
		dc.b	$23					; #
		dc.b	$2C					; ,
		dc.b	$7C					; |
		dc.b	$DA					; 
		dc.b	$BF					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$6F					; o
		dc.b	$37					; 7
		dc.b	$B6					; 
		dc.b	$F6					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$95					; 
		dc.b	$76					; v
		dc.b	6
		dc.b	$1B
		dc.b	$5D					; ]
		dc.b	$9D					; 
		dc.b	$F0					; 
		dc.b	$CD					; 
		dc.b	$75					; u
		dc.b	$57					; W
		dc.b	$48					; H
		dc.b	$98					; 
		dc.b	$C2					; 
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$CD					; 
		dc.b	$87					; 
		dc.b	1
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$53					; S
		dc.b	$B3					; 
		dc.b	$D9					; 
		dc.b	$DA					; 
		dc.b	$FE					; 
		dc.b	$71					; q
		dc.b	1
		dc.b	$27					; '
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$BA					; 
		dc.b	$87					; 
		dc.b	$A9					; 
		dc.b	$EB					; 
		dc.b	$60					; `
		dc.b	$3C					; <
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$1B
		dc.b	$49					; I
		dc.b	$F2					; 
		dc.b	$E7					; 
		dc.b	$B3					; 
		dc.b	$96					; 
		dc.b	$D
		dc.b	$1B
		dc.b	$31					; 1
		dc.b	$1B
		dc.b	$60					; `
		dc.b	$91					; 
		dc.b	$D6					; 
		dc.b	$EB					; 
		dc.b	$BF					; 
		dc.b	$83					; 
		dc.b	$4F					; O
		dc.b	1
		dc.b	6
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$E4					; 
		dc.b	$B8					; 
		dc.b	$E2					; 
		dc.b	$E2					; 
		dc.b	$30					; 0
		dc.b	$15
		dc.b	$29					; )
		dc.b	$2D					; -
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$6E					; n
		dc.b	$B1					; 
		dc.b	$BE					; 
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$4D					; M
		dc.b	$22					; "
		dc.b	$45					; E
		dc.b	$7C					; |
		dc.b	$CF					; 
		dc.b	$FE					; 
		dc.b	$A0					; 
		dc.b	$6F					; o
		dc.b	$42					; B
		dc.b	$38					; 8
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$A3					; 
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$2E					; .
		dc.b	$31					; 1
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$FE					; 
		dc.b	$B4					; 
		dc.b	$38					; 8
		dc.b	$95					; 
		dc.b	$52					; R
		dc.b	$14
		dc.b	$76					; v
		dc.b	$48					; H
		dc.b	$25					; %
		dc.b	$C6					; 
		dc.b	$D3					; 
		dc.b	$75					; u
		dc.b	$F5					; 
		dc.b	$DE					; 
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$27					; '
		dc.b	1
		dc.b	$A
		dc.b	$3F					; ?
		dc.b	$DB					; 
		dc.b	$F7					; 
		dc.b	$C1					; 
		dc.b	$AD					; 
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$BF					; 
		dc.b	$8D					; 
		dc.b	$42					; B
		dc.b	$9B					; 
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$87					; 
		dc.b	$3E					; >
		dc.b	$45					; E
		dc.b	$A7					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$78					; x
		dc.b	$C4					; 
		dc.b	$A3					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$75					; u
		dc.b	$AE					; 
		dc.b	$60					; `
		dc.b	$35					; 5
		dc.b	$63					; c
		dc.b	$81					; 
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$48					; H
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$64					; d
		dc.b	$99					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$AA					; 
		dc.b	$8B					; 
		dc.b	$A
		dc.b	$3F					; ?
		dc.b	$9B					; 
		dc.b	$2A					; *
		dc.b	$6E					; n
		dc.b	$FE					; 
		dc.b	$BD					; 
		dc.b	$B3					; 
		dc.b	$AD					; 
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$3E					; >
		dc.b	$34					; 4
		dc.b	$3B					; ;
		dc.b	$2D					; -
		dc.b	$9B					; 
		dc.b	$F6					; 
		dc.b	$BB					; 
		dc.b	$B7					; 
		dc.b	$9D					; 
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$6E					; n
		dc.b	$43					; C
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$15
		dc.b	$46					; F
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$E7					; 
		dc.b	$92					; 
		dc.b	$64					; d
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$33					; 3
		dc.b	$6E					; n
		dc.b	$BA					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$EB					; 
		dc.b	$C8					; 
		dc.b	$86					; 
		dc.b	$55					; U
		dc.b	$22					; "
		dc.b	$3B					; ;
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$D4					; 
		dc.b	$E2					; 
		dc.b	$87					; 
		dc.b	$22					; "
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$C5					; 
		dc.b	$B1					; 
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$2E					; .
		dc.b	$71					; q
		dc.b	$B3					; 
		dc.b	$BA					; 
		dc.b	$A3					; 
		dc.b	$31					; 1
		dc.b	$17
		dc.b	$43					; C
		dc.b	$72					; r
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$A2					; 
		dc.b	$6C					; l
		dc.b	$B3					; 
		dc.b	$6C					; l
		dc.b	2
		dc.b	$87					; 
		dc.b	$B4					; 
		dc.b	$7D					; }
		dc.b	$94					; 
		dc.b	$5A					; Z
		dc.b	$2A					; *
		dc.b	$76					; v
		dc.b	$C2					; 
		dc.b	$C1					; 
		dc.b	$B6					; 
		dc.b	$7F					; 
		dc.b	$3B					; ;
		dc.b	$2E					; .
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$9D					; 
		dc.b	$FE					; 
		dc.b	$C6					; 
		dc.b	$76					; v
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$9F					; 
		dc.b	$A2					; 
		dc.b	$1C
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$BD					; 
		dc.b	$FA					; 
		dc.b	$DE					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$56					; V
		dc.b	$33					; 3
		dc.b	$48					; H
		dc.b	$67					; g
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$79					; y
		dc.b	$AD					; 
		dc.b	$78					; x
		dc.b	$2D					; -
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$BD					; 
		dc.b	$D4					; 
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$6B					; k
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$86					; 
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$65					; e
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$94					; 
		dc.b	$EB					; 
		dc.b	$D2					; 
		dc.b	$76					; v
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$E
		dc.b	$2D					; -
		dc.b	$6E					; n
		dc.b	$99					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	1
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$C9					; 
		dc.b	$D3					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$19
		dc.b	$18
		dc.b	$64					; d
		dc.b	$D9					; 
		dc.b	$BA					; 
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$CB					; 
		dc.b	$99					; 
		dc.b	$45					; E
		dc.b	$E9					; 
		dc.b	$B7					; 
		dc.b	$34					; 4
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$3A					; :
		dc.b	$C2					; 
		dc.b	$78					; x
		dc.b	$20
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$B4					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$CC					; 
		dc.b	$94					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	3
		dc.b	$63					; c
		dc.b	$B1					; 
		dc.b	$45					; E
		dc.b	$72					; r
		dc.b	$EF					; 
		dc.b	$DA					; 
		dc.b	$BD					; 
		dc.b	$7A					; z
		dc.b	1
		dc.b	$42					; B
		dc.b	$AE					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$3C					; <
		dc.b	1
		dc.b	$6B					; k
		dc.b	$EC					; 
		dc.b	$E4					; 
		dc.b	$DB					; 
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$37					; 7
		dc.b	2
		dc.b	$52					; R
		dc.b	$C9					; 
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$4C					; L
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$41					; A
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$B1					; 
		dc.b	$E8					; 
		dc.b	$83					; 
		dc.b	1
		dc.b	$37					; 7
		dc.b	$AA					; 
		dc.b	$AD					; 
		dc.b	$BD					; 
		dc.b	$8F					; 
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$8F					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$48					; H
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$BE					; 
		dc.b	$B4					; 
		dc.b	$7C					; |
		dc.b	$B4					; 
		dc.b	$8E					; 
		dc.b	$2E					; .
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$2D					; -
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$98					; 
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$B8					; 
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$F6					; 
		dc.b	$C6					; 
		dc.b	$4F					; O
		dc.b	$46					; F
		dc.b	$31					; 1
		dc.b	$4A					; J
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$99					; 
		dc.b	$E4					; 
		dc.b	$E7					; 
		dc.b	$9B					; 
		dc.b	$49					; I
		dc.b	$D
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$6B					; k
		dc.b	$A3					; 
		dc.b	$E4					; 
		dc.b	$F0					; 
		dc.b	$BA					; 
		dc.b	$63					; c
		dc.b	$25					; %
		dc.b	$6A					; j
		dc.b	$B4					; 
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$8F					; 
		dc.b	$D4					; 
		dc.b	$94					; 
		dc.b	$6E					; n
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$3C					; <
		dc.b	$81					; 
		dc.b	$BE					; 
		dc.b	$4C					; L
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$F6					; 
		dc.b	$DE					; 
		dc.b	$6C					; l
		dc.b	$18
		dc.b	1
		dc.b	$14
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$9D					; 
		dc.b	$59					; Y
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	1
		dc.b	$7C					; |
		dc.b	$FE					; 
		dc.b	$FC					; 
		dc.b	$D0					; 
		dc.b	$94					; 
		dc.b	$B
		dc.b	$56					; V
		dc.b	$59					; Y
		dc.b	1
		dc.b	$2D					; -
		dc.b	$BD					; 
		dc.b	$9F					; 
		dc.b	$DD					; 
		dc.b	$E2					; 
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$59					; Y
		dc.b	$A
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$35					; 5
		dc.b	$C4					; 
		dc.b	$D7					; 
		dc.b	$98					; 
		dc.b	$BF					; 
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$B6					; 
		dc.b	$3A					; :
		dc.b	$59					; Y
		dc.b	$EE					; 
		dc.b	$9C					; 
		dc.b	$AE					; 
		dc.b	$BE					; 
		dc.b	1
		dc.b	9
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$A3					; 
		dc.b	$C6					; 
		dc.b	$61					; a
		dc.b	$C5					; 
		dc.b	$E4					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	6
		dc.b	1
		dc.b	$AD					; 
		dc.b	$D6					; 
		dc.b	$A9					; 
		dc.b	$DA					; 
		dc.b	$CB					; 
		dc.b	$CC					; 
		dc.b	$AE					; 
		dc.b	$B
		dc.b	1
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$A9					; 
		dc.b	$F3					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$6F					; o
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$59					; Y
		dc.b	$9D					; 
		dc.b	$FC					; 
		dc.b	$FE					; 
		dc.b	$96					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$30					; 0
		dc.b	$E
		dc.b	$3F					; ?
		dc.b	$4C					; L
		dc.b	$99					; 
		dc.b	$BE					; 
		dc.b	$87					; 
		dc.b	$F3					; 
		dc.b	$BE					; 
		dc.b	$35					; 5
		dc.b	$6F					; o
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$C8					; 
		dc.b	$50					; P
		dc.b	7
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$FE					; 
		dc.b	$7C					; |
		dc.b	6
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$41					; A
		dc.b	$81					; 
		dc.b	$DE					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$BD					; 
		dc.b	$88					; 
		dc.b	$38					; 8
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$3F					; ?
		dc.b	$E7					; 
		dc.b	$C8					; 
		dc.b	$59					; Y
		dc.b	$99					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$9D					; 
		dc.b	$19
		dc.b	$A
		dc.b	$EE					; 
		dc.b	$A4					; 
		dc.b	$2D					; -
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$A2					; 
		dc.b	$DB					; 
		dc.b	$22					; "
		dc.b	$45					; E
		dc.b	$C4					; 
		dc.b	$8B					; 
		dc.b	$5C					; \
		dc.b	$34					; 4
		dc.b	1
		dc.b	$9D					; 
		dc.b	$D6					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$EB					; 
		dc.b	$73					; s
		dc.b	$33					; 3
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$9D					; 
		dc.b	$C4					; 
		dc.b	$49					; I
		dc.b	$8B					; 
		dc.b	$CB					; 
		dc.b	$91					; 
		dc.b	$9D					; 
		dc.b	$22					; "
		dc.b	6
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$91					; 
		dc.b	$41					; A
		dc.b	$B8					; 
		dc.b	$CD					; 
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$C6					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$5A					; Z
		dc.b	$1B
		dc.b	7
		dc.b	$26					; &
		dc.b	$7C					; |
		dc.b	$EC					; 
		dc.b	$E1					; 
		dc.b	$9F					; 
		dc.b	$42					; B
		dc.b	$53					; S
		dc.b	$7A					; z
		dc.b	$57					; W
		dc.b	$4A					; J
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$C5					; 
		dc.b	$B3					; 
		dc.b	$76					; v
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$7F					; 
		dc.b	$25					; %
		dc.b	1
		dc.b	$3A					; :
		dc.b	$C1					; 
		dc.b	$AE					; 
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$4A					; J
		dc.b	$6B					; k
		dc.b	$8A					; 
		dc.b	$48					; H
		dc.b	$41					; A
		dc.b	$6C					; l
		dc.b	$BF					; 
		dc.b	$F9					; 
		dc.b	$BE					; 
		dc.b	$59					; Y
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$31					; 1
		dc.b	$37					; 7
		dc.b	$72					; r
		dc.b	$B1					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$B4					; 
		dc.b	$6B					; k
		dc.b	$29					; )
		dc.b	$56					; V
		dc.b	$7F					; 
		dc.b	$EF					; 
		dc.b	$FE					; 
		dc.b	$6A					; j
		dc.b	$D
		dc.b	$38					; 8
		dc.b	$5A					; Z
		dc.b	$B8					; 
		dc.b	$C2					; 
		dc.b	$2E					; .
		dc.b	$2C					; ,
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$EE					; 
		dc.b	$DB					; 
		dc.b	$75					; u
		dc.b	$A2					; 
		dc.b	$20
		dc.b	$12
		dc.b	$BA					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$A4					; 
		dc.b	$27					; '
		dc.b	$6F					; o
		dc.b	$E0					; 
		dc.b	$B8					; 
		dc.b	$AE					; 
		dc.b	$27					; '
		dc.b	1
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$BF					; 
		dc.b	$AA					; 
		dc.b	$AB					; 
		dc.b	$CD					; 
		dc.b	$76					; v
		dc.b	$22					; "
		dc.b	$2E					; .
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$A7					; 
		dc.b	$9B					; 
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$A2					; 
		dc.b	$71					; q
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$4A					; J
		dc.b	$61					; a
		dc.b	$CB					; 
		dc.b	$CF					; 
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$CF					; 
		dc.b	$BA					; 
		dc.b	$19
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$9B					; 
		dc.b	$E4					; 
		dc.b	$BD					; 
		dc.b	$34					; 4
		dc.b	$5E					; ^
		dc.b	$68					; h
		dc.b	$2E					; .
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$AB					; 
		dc.b	$D2					; 
		dc.b	$31					; 1
		dc.b	1
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$CB					; 
		dc.b	$F3					; 
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$CF					; 
		dc.b	$4D					; M
		dc.b	$7D					; }
		dc.b	$BB					; 
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$3E					; >
		dc.b	$75					; u
		dc.b	$26					; &
		dc.b	$2D					; -
		dc.b	$B8					; 
		dc.b	$A9					; 
		dc.b	$C1					; 
		dc.b	$BB					; 
		dc.b	$1E
		dc.b	$86					; 
		dc.b	$BD					; 
		dc.b	$10
		dc.b	$49					; I
		dc.b	$A9					; 
		dc.b	$12
		dc.b	$67					; g
		dc.b	$E7					; 
		dc.b	$55					; U
		dc.b	$96					; 
		dc.b	$FE					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$46					; F
		dc.b	1
		dc.b	$73					; s
		dc.b	$DB					; 
		dc.b	$AA					; 
		dc.b	$D6					; 
		dc.b	$A4					; 
		dc.b	$83					; 
		dc.b	$AD					; 
		dc.b	$34					; 4
		dc.b	$30					; 0
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$B3					; 
		dc.b	$99					; 
		dc.b	$26					; &
		dc.b	$3F					; ?
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$E4					; 
		dc.b	$67					; g
		dc.b	7
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$61					; a
		dc.b	$D6					; 
		dc.b	$BB					; 
		dc.b	$55					; U
		dc.b	$95					; 
		dc.b	$37					; 7
		dc.b	6
		dc.b	$C1					; 
		dc.b	$9D					; 
		dc.b	$4C					; L
		dc.b	$C9					; 
		dc.b	$CF					; 
		dc.b	$95					; 
		dc.b	$6F					; o
		dc.b	$D
		dc.b	1
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$A9					; 
		dc.b	$C8					; 
		dc.b	$E4					; 
		dc.b	$B7					; 
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$3E					; >
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$56					; V
		dc.b	$6E					; n
		dc.b	$EC					; 
		dc.b	$8D					; 
		dc.b	$45					; E
		dc.b	$A9					; 
		dc.b	$29					; )
		dc.b	$18
		dc.b	$8E					; 
		dc.b	$42					; B
		dc.b	$80					; 
		dc.b	$BD					; 
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$CC					; 
		dc.b	$DA					; 
		dc.b	$CB					; 
		dc.b	$79					; y
		dc.b	7
		dc.b	1
		dc.b	$2C					; ,
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$AD					; 
		dc.b	$D2					; 
		dc.b	$CB					; 
		dc.b	$C8					; 
		dc.b	$63					; c
		dc.b	1
		dc.b	$27					; '
		dc.b	$5D					; ]
		dc.b	$80					; 
		dc.b	$DD					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$F7					; 
		dc.b	$9D					; 
		dc.b	$31					; 1
		dc.b	$7A					; z
		dc.b	$A4					; 
		dc.b	$61					; a
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$7A					; z
		dc.b	$BE					; 
		dc.b	$BE					; 
		dc.b	$5C					; \
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$1B
		dc.b	$56					; V
		dc.b	$C5					; 
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$B7					; 
		dc.b	$4A					; J
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$33					; 3
		dc.b	$4C					; L
		dc.b	$7C					; |
		dc.b	$E5					; 
		dc.b	$DD					; 
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$18
		dc.b	$22					; "
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$B4					; 
		dc.b	$FE					; 
		dc.b	$D9					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$46					; F
		dc.b	$38					; 8
		dc.b	$3C					; <
		dc.b	$71					; q
		dc.b	$BF					; 
		dc.b	$A2					; 
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$3F					; ?
		dc.b	$7D					; }
		dc.b	$A9					; 
		dc.b	$41					; A
		dc.b	$38					; 8
		dc.b	$EC					; 
		dc.b	$BD					; 
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$BB					; 
		dc.b	$8D					; 
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$2C					; ,
		dc.b	$A6					; 
		dc.b	$E2					; 
		dc.b	$80					; 
		dc.b	$37					; 7
		dc.b	$19
		dc.b	2
		dc.b	$61					; a
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$C9					; 
		dc.b	$E0					; 
		dc.b	$8F					; 
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$9B					; 
		dc.b	$1C
		dc.b	$1C
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$9D					; 
		dc.b	$C5					; 
		dc.b	$83					; 
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$8E					; 
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$33					; 3
		dc.b	$72					; r
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$BD					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$10
		dc.b	$4F					; O
		dc.b	$B7					; 
		dc.b	$94					; 
		dc.b	$D4					; 
		dc.b	$CD					; 
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$26					; &
		dc.b	6
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$5C					; \
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$E1					; 
		dc.b	$95					; 
		dc.b	$64					; d
		dc.b	$3A					; :
		dc.b	$1E
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$95					; 
		dc.b	$F5					; 
		dc.b	$B0					; 
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$4F					; O
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$2C					; ,
		dc.b	$AD					; 
		dc.b	$CB					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$3F					; ?
		dc.b	1
		dc.b	$59					; Y
		dc.b	$88					; 
		dc.b	$C1					; 
		dc.b	$F9					; 
		dc.b	$C2					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$49					; I
		dc.b	$2C					; ,
		dc.b	$11
		dc.b	$50					; P
		dc.b	$8D					; 
		dc.b	$68					; h
		dc.b	$C9					; 
		dc.b	$E5					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$29					; )
		dc.b	$2D					; -
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$D3					; 
		dc.b	$D3					; 
		dc.b	$86					; 
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$A9					; 
		dc.b	$68					; h
		dc.b	$3B					; ;
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$C6					; 
		dc.b	$C4					; 
		dc.b	$72					; r
		dc.b	$50					; P
		dc.b	$55					; U
		dc.b	$35					; 5
		dc.b	$38					; 8
		dc.b	$AA					; 
		dc.b	$E4					; 
		dc.b	$C2					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$A
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	1
		dc.b	$56					; V
		dc.b	$CF					; 
		dc.b	$F9					; 
		dc.b	$FE					; 
		dc.b	$A6					; 
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$14
		dc.b	$11
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$D9					; 
		dc.b	$9B					; 
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$7D					; }
		dc.b	$20
		dc.b	2
		dc.b	2
		dc.b	$55					; U
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$B7					; 
		dc.b	$D0					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$5D					; ]
		dc.b	$49					; I
		dc.b	$52					; R
		dc.b	$3F					; ?
		dc.b	$75					; u
		dc.b	$CD					; 
		dc.b	$BA					; 
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$AA					; 
		dc.b	$45					; E
		dc.b	1
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$FE					; 
		dc.b	$A2					; 
		dc.b	$43					; C
		dc.b	$7A					; z
		dc.b	$42					; B
		dc.b	1
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$C6					; 
		dc.b	$D6					; 
		dc.b	$8F					; 
		dc.b	1
		dc.b	$1C
		dc.b	$27					; '
		dc.b	$64					; d
		dc.b	$FE					; 
		dc.b	$A6					; 
		dc.b	$7A					; z
		dc.b	$D2					; 
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$C1					; 
		dc.b	$10
		dc.b	$23					; #
		dc.b	$55					; U
		dc.b	$33					; 3
		dc.b	$C6					; 
		dc.b	$C9					; 
		dc.b	$72					; r
		dc.b	$FE					; 
		dc.b	$D3					; 
		dc.b	$37					; 7
		dc.b	$6F					; o
		dc.b	7
		dc.b	1
		dc.b	$A7					; 
		dc.b	$C4					; 
		dc.b	$A3					; 
		dc.b	$BD					; 
		dc.b	$75					; u
		dc.b	$A3					; 
		dc.b	$FE					; 
		dc.b	$5D					; ]
		dc.b	$3E					; >
		dc.b	$59					; Y
		dc.b	$1B
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$FC					; 
		dc.b	$C4					; 
		dc.b	$6C					; l
		dc.b	$91					; 
		dc.b	$26					; &
		dc.b	$18
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$B0					; 
		dc.b	$C8					; 
		dc.b	$4C					; L
		dc.b	$2A					; *
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$4C					; L
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$B0					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$A2					; 
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$B6					; 
		dc.b	$55					; U
		dc.b	$38					; 8
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$3F					; ?
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	6
		dc.b	$30					; 0
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$C9					; 
		dc.b	$FE					; 
		dc.b	$45					; E
		dc.b	$7A					; z
		dc.b	$C2					; 
		dc.b	$26					; &
		dc.b	$27					; '
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$99					; 
		dc.b	$CD					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$C1					; 
		dc.b	$BF					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$5C					; \
		dc.b	$4D					; M
		dc.b	$AE					; 
		dc.b	$99					; 
		dc.b	$4F					; O
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$45					; E
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$29					; )
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$3C					; <
		dc.b	$5E					; ^
		dc.b	$E4					; 
		dc.b	$B4					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$59					; Y
		dc.b	$A2					; 
		dc.b	$B7					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$38					; 8
		dc.b	$33					; 3
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$AE					; 
		dc.b	$D3					; 
		dc.b	$92					; 
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$57					; W
		dc.b	$30					; 0
		dc.b	$25					; %
		dc.b	$6A					; j
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$8B					; 
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$B1					; 
		dc.b	$86					; 
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$BD					; 
		dc.b	$8E					; 
		dc.b	$34					; 4
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$34					; 4
		dc.b	$3A					; :
		dc.b	$D3					; 
		dc.b	$DA					; 
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$65					; e
		dc.b	$4C					; L
		dc.b	$5E					; ^
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$EB					; 
		dc.b	$F2					; 
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$3A					; :
		dc.b	$D
		dc.b	$79					; y
		dc.b	$BF					; 
		dc.b	$C4					; 
		dc.b	$DD					; 
		dc.b	$D6					; 
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$B6					; 
		dc.b	$84					; 
		dc.b	$27					; '
		dc.b	7
		dc.b	$43					; C
		dc.b	$67					; g
		dc.b	$8D					; 
		dc.b	$BE					; 
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$B3					; 
		dc.b	$52					; R
		dc.b	$2A					; *
		dc.b	$52					; R
		dc.b	$50					; P
		dc.b	$53					; S
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$6B					; k
		dc.b	$D3					; 
		dc.b	$95					; 
		dc.b	$5D					; ]
		dc.b	$B7					; 
		dc.b	$30					; 0
		dc.b	2
		dc.b	$88					; 
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$D9					; 
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$3B					; ;
		dc.b	$3E					; >
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$59					; Y
		dc.b	$8A					; 
		dc.b	$B8					; 
		dc.b	$E0					; 
		dc.b	$B1					; 
		dc.b	$5A					; Z
		dc.b	$27					; '
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$3A					; :
		dc.b	$5D					; ]
		dc.b	$B3					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$AA					; 
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$46					; F
		dc.b	$48					; H
		dc.b	$95					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$10
		dc.b	$38					; 8
		dc.b	$B1					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$E8					; 
		dc.b	$A7					; 
		dc.b	$48					; H
		dc.b	$7A					; z
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$E7					; 
		dc.b	$6A					; j
		dc.b	$18
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$4A					; J
		dc.b	$BE					; 
		dc.b	$92					; 
		dc.b	$2D					; -
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$25					; %
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$EB					; 
		dc.b	$E2					; 
		dc.b	$49					; I
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$5C					; \
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$49					; I
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	$50					; P
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$B6					; 
		dc.b	$8E					; 
		dc.b	$E
		dc.b	$41					; A
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$FE					; 
		dc.b	$FA					; 
		dc.b	$6E					; n
		dc.b	$41					; A
		dc.b	$41					; A
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$98					; 
		dc.b	$6A					; j
		dc.b	$91					; 
		dc.b	$B1					; 
		dc.b	$3F					; ?
		dc.b	1
		dc.b	$22					; "
		dc.b	$60					; `
		dc.b	$C2					; 
		dc.b	$EF					; 
		dc.b	$C2					; 
		dc.b	$73					; s
		dc.b	$A4					; 
		dc.b	$5A					; Z
		dc.b	1
		dc.b	$42					; B
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$A2					; 
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$E2					; 
		dc.b	$DE					; 
		dc.b	$BD					; 
		dc.b	$5E					; ^
		dc.b	$10
		dc.b	9
		dc.b	$4C					; L
		dc.b	$7D					; }
		dc.b	$57					; W
		dc.b	$9D					; 
		dc.b	$EB					; 
		dc.b	$C1					; 
		dc.b	$BA					; 
		dc.b	$BA					; 
		dc.b	$3A					; :
		dc.b	$15
		dc.b	$6C					; l
		dc.b	$2D					; -
		dc.b	$3E					; >
		dc.b	$C8					; 
		dc.b	$BD					; 
		dc.b	$CC					; 
		dc.b	$F2					; 
		dc.b	$5A					; Z
		dc.b	1
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$A3					; 
		dc.b	$8A					; 
		dc.b	$49					; I
		dc.b	$A3					; 
		dc.b	$CB					; 
		dc.b	$5C					; \
		dc.b	$A7					; 
		dc.b	$7D					; }
		dc.b	$12
		dc.b	$87					; 
		dc.b	$96					; 
		dc.b	$6A					; j
		dc.b	$D7					; 
		dc.b	$BE					; 
		dc.b	$50					; P
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$6A					; j
		dc.b	$92					; 
		dc.b	$79					; y
		dc.b	$19
		dc.b	$49					; I
		dc.b	$99					; 
		dc.b	$BD					; 
		dc.b	$96					; 
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$D9					; 
		dc.b	$CF					; 
		dc.b	$86					; 
		dc.b	$10
		dc.b	$14
		dc.b	$50					; P
		dc.b	$50					; P
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$FE					; 
		dc.b	$D2					; 
		dc.b	$5A					; Z
		dc.b	$38					; 8
		dc.b	$12
		dc.b	$49					; I
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$79					; y
		dc.b	$B3					; 
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$D9					; 
		dc.b	$59					; Y
		dc.b	$35					; 5
		dc.b	$64					; d
		dc.b	1
		dc.b	$5D					; ]
		dc.b	$DB					; 
		dc.b	$65					; e
		dc.b	$95					; 
		dc.b	$FE					; 
		dc.b	$53					; S
		dc.b	$52					; R
		dc.b	$BD					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$1B
		dc.b	$64					; d
		dc.b	$D2					; 
		dc.b	$B4					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$48					; H
		dc.b	$78					; x
		dc.b	$9C					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$3C					; <
		dc.b	$37					; 7
		dc.b	$A6					; 
		dc.b	$B8					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$A7					; 
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$4D					; M
		dc.b	$B4					; 
		dc.b	$9D					; 
		dc.b	$3A					; :
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$43					; C
		dc.b	$AE					; 
		dc.b	$92					; 
		dc.b	$3B					; ;
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$3C					; <
		dc.b	$B0					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$C4					; 
		dc.b	$4F					; O
		dc.b	6
		dc.b	$92					; 
		dc.b	$AA					; 
		dc.b	$B4					; 
		dc.b	$E4					; 
		dc.b	$6C					; l
		dc.b	$2A					; *
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$8F					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$B3					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$3B					; ;
		dc.b	$5C					; \
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$E5					; 
		dc.b	$BA					; 
		dc.b	$22					; "
		dc.b	$8E					; 
		dc.b	$57					; W
		dc.b	1
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$F7					; 
		dc.b	$C8					; 
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$46					; F
		dc.b	$48					; H
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$B6					; 
		dc.b	$6F					; o
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$9C					; 
		dc.b	$BD					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$C8					; 
		dc.b	$98					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$29					; )
		dc.b	$94					; 
		dc.b	$E5					; 
		dc.b	$AB					; 
		dc.b	$42					; B
		dc.b	$29					; )
		dc.b	$49					; I
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$94					; 
		dc.b	$B4					; 
		dc.b	$6B					; k
		dc.b	1
		dc.b	$60					; `
		dc.b	$BB					; 
		dc.b	$E1					; 
		dc.b	$E4					; 
		dc.b	$3B					; ;
		dc.b	1
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$5A					; Z
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$B7					; 
		dc.b	$A0					; 
		dc.b	$65					; e
		dc.b	$30					; 0
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$45					; E
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$CC					; 
		dc.b	$B8					; 
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$49					; I
		dc.b	$67					; g
		dc.b	$A7					; 
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$52					; R
		dc.b	$6B					; k
		dc.b	$C1					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$AD					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$4C					; L
		dc.b	$31					; 1
		dc.b	$B6					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$53					; S
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$88					; 
		dc.b	$CC					; 
		dc.b	$7C					; |
		dc.b	$59					; Y
		dc.b	$72					; r
		dc.b	$59					; Y
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$45					; E
		dc.b	$96					; 
		dc.b	$F9					; 
		dc.b	$B6					; 
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$3A					; :
		dc.b	$29					; )
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	2
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$61					; a
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$CD					; 
		dc.b	$98					; 
		dc.b	$5C					; \
		dc.b	$B4					; 
		dc.b	$88					; 
		dc.b	$18
		dc.b	$43					; C
		dc.b	$3F					; ?
		dc.b	$94					; 
		dc.b	$E2					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$C8					; 
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$96					; 
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$AA					; 
		dc.b	$BE					; 
		dc.b	$88					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$B7					; 
		dc.b	$6C					; l
		dc.b	$42					; B
		dc.b	$68					; h
		dc.b	$56					; V
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$AB					; 
		dc.b	$41					; A
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$22					; "
		dc.b	$76					; v
		dc.b	$B7					; 
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$AE					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$7D					; }
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$A2					; 
		dc.b	$A9					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$A4					; 
		dc.b	$99					; 
		dc.b	$79					; y
		dc.b	$B3					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$D
		dc.b	$64					; d
		dc.b	$B3					; 
		dc.b	$A0					; 
		dc.b	$FE					; 
		dc.b	$B4					; 
		dc.b	$1F
		dc.b	$42					; B
		dc.b	$4C					; L
		dc.b	$8F					; 
		dc.b	$CF					; 
		dc.b	$4F					; O
		dc.b	$D
		dc.b	$BF					; 
		dc.b	$CF					; 
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$43					; C
		dc.b	$2A					; *
		dc.b	$96					; 
		dc.b	$67					; g
		dc.b	$30					; 0
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$F7					; 
		dc.b	$68					; h
		dc.b	$37					; 7
		dc.b	$8F					; 
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$AE					; 
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$EC					; 
		dc.b	$DB					; 
		dc.b	$B6					; 
		dc.b	$4F					; O
		dc.b	1
		dc.b	1
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$AA					; 
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$61					; a
		dc.b	$95					; 
		dc.b	$61					; a
		dc.b	$41					; A
		dc.b	$78					; x
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$AA					; 
		dc.b	$A2					; 
		dc.b	$3A					; :
		dc.b	$1C
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$A7					; 
		dc.b	$B8					; 
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$43					; C
		dc.b	$4C					; L
		dc.b	$96					; 
		dc.b	$B1					; 
		dc.b	$C5					; 
		dc.b	$E1					; 
		dc.b	$4F					; O
		dc.b	2
		dc.b	$3A					; :
		dc.b	$7F					; 
		dc.b	$B8					; 
		dc.b	$B1					; 
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$64					; d
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$A3					; 
		dc.b	$DB					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$20
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$D4					; 
		dc.b	$C6					; 
		dc.b	$65					; e
		dc.b	$C1					; 
		dc.b	$88					; 
		dc.b	$23					; #
		dc.b	$57					; W
		dc.b	$33					; 3
		dc.b	$5D					; ]
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$AB					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$D
		dc.b	1
		dc.b	$86					; 
		dc.b	$AB					; 
		dc.b	$CB					; 
		dc.b	$C4					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$38					; 8
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$80					; 
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$94					; 
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$3A					; :
		dc.b	$75					; u
		dc.b	$BA					; 
		dc.b	$C6					; 
		dc.b	$BB					; 
		dc.b	$4C					; L
		dc.b	$11
		dc.b	$43					; C
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$C6					; 
		dc.b	$6F					; o
		dc.b	$43					; C
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$AA					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$4C					; L
		dc.b	$33					; 3
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$AB					; 
		dc.b	$AA					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$CC					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$48					; H
		dc.b	$41					; A
		dc.b	$F3					; 
		dc.b	$F3					; 
		dc.b	$73					; s
		dc.b	$5A					; Z
		dc.b	$19
		dc.b	$35					; 5
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$60					; `
		dc.b	$3E					; >
		dc.b	$34					; 4
		dc.b	$61					; a
		dc.b	$B6					; 
		dc.b	$B7					; 
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$6C					; l
		dc.b	$3E					; >
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$48					; H
		dc.b	$81					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$65					; e
		dc.b	$3B					; ;
		dc.b	$3C					; <
		dc.b	$6A					; j
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$AA					; 
		dc.b	$72					; r
		dc.b	$3A					; :
		dc.b	$19
		dc.b	$49					; I
		dc.b	$AB					; 
		dc.b	$D0					; 
		dc.b	$B0					; 
		dc.b	$CC					; 
		dc.b	$A6					; 
		dc.b	$D
		dc.b	1
		dc.b	$19
		dc.b	$75					; u
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$4C					; L
		dc.b	$53					; S
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$D2					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$43					; C
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$46					; F
		dc.b	$84					; 
		dc.b	$D0					; 
		dc.b	$B1					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$48					; H
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$C5					; 
		dc.b	$EE					; 
		dc.b	$C6					; 
		dc.b	$61					; a
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$5D					; ]
		dc.b	$AE					; 
		dc.b	$B0					; 
		dc.b	$48					; H
		dc.b	$7F					; 
		dc.b	$CC					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$20
		dc.b	$30					; 0
		dc.b	$CC					; 
		dc.b	$A7					; 
		dc.b	$64					; d
		dc.b	$91					; 
		dc.b	$3F					; ?
		dc.b	$5A					; Z
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$A7					; 
		dc.b	$A4					; 
		dc.b	$75					; u
		dc.b	$12
		dc.b	1
		dc.b	$6C					; l
		dc.b	$C2					; 
		dc.b	$F7					; 
		dc.b	$C9					; 
		dc.b	$38					; 8
		dc.b	$42					; B
		dc.b	$6A					; j
		dc.b	$41					; A
		dc.b	$45					; E
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$C1					; 
		dc.b	$EE					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$20
		dc.b	$59					; Y
		dc.b	$B7					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$C4					; 
		dc.b	$B1					; 
		dc.b	$CC					; 
		dc.b	$AA					; 
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.b	$3F					; ?
		dc.b	$6F					; o
		dc.b	$BA					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$61					; a
		dc.b	1
		dc.b	$B
		dc.b	$5C					; \
		dc.b	$A3					; 
		dc.b	$D6					; 
		dc.b	$B7					; 
		dc.b	$7C					; |
		dc.b	$CF					; 
		dc.b	$9F					; 
		dc.b	$11
		dc.b	$3B					; ;
		dc.b	$7C					; |
		dc.b	$A0					; 
		dc.b	$F6					; 
		dc.b	$A4					; 
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$B6					; 
		dc.b	$BF					; 
		dc.b	$68					; h
		dc.b	$9C					; 
		dc.b	$81					; 
		dc.b	$15
		dc.b	$4C					; L
		dc.b	$57					; W
		dc.b	$3E					; >
		dc.b	$C2					; 
		dc.b	$C2					; 
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$38					; 8
		dc.b	$96					; 
		dc.b	$68					; h
		dc.b	$1C
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$C6					; 
		dc.b	$EB					; 
		dc.b	$5A					; Z
		dc.b	2
		dc.b	$37					; 7
		dc.b	$3E					; >
		dc.b	$A3					; 
		dc.b	$CF					; 
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$35					; 5
		dc.b	$2E					; .
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$C5					; 
		dc.b	$B8					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$3E					; >
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$C4					; 
		dc.b	$C8					; 
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$D7					; 
		dc.b	$B1					; 
		dc.b	$4A					; J
		dc.b	$9F					; 
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$D0					; 
		dc.b	$5D					; ]
		dc.b	$53					; S
		dc.b	$A3					; 
		dc.b	$27					; '
		dc.b	$4D					; M
		dc.b	$B7					; 
		dc.b	$73					; s
		dc.b	$BF					; 
		dc.b	$D3					; 
		dc.b	$1E
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$48					; H
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$5A					; Z
		dc.b	$9C					; 
		dc.b	$B0					; 
		dc.b	$53					; S
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$B3					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$9B					; 
		dc.b	$78					; x
		dc.b	$15
		dc.b	$52					; R
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$BF					; 
		dc.b	$9B					; 
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$A0					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$27					; '
		dc.b	$11
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$B4					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$37					; 7
		dc.b	$4F					; O
		dc.b	$92					; 
		dc.b	$C9					; 
		dc.b	$BF					; 
		dc.b	$9D					; 
		dc.b	$7C					; |
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$38					; 8
		dc.b	$38					; 8
		dc.b	$D0					; 
		dc.b	$D2					; 
		dc.b	$B8					; 
		dc.b	$F2					; 
		dc.b	$8F					; 
		dc.b	$18
		dc.b	$43					; C
		dc.b	$2D					; -
		dc.b	$37					; 7
		dc.b	$AA					; 
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$D4					; 
		dc.b	$6B					; k
		dc.b	$29					; )
		dc.b	$52					; R
		dc.b	$60					; `
		dc.b	$BA					; 
		dc.b	$B3					; 
		dc.b	$43					; C
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$2E					; .
		dc.b	$6A					; j
		dc.b	$BF					; 
		dc.b	$A2					; 
		dc.b	$6A					; j
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$94					; 
		dc.b	$C5					; 
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$63					; c
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$8D					; 
		dc.b	$AE					; 
		dc.b	$D7					; 
		dc.b	$B6					; 
		dc.b	$5E					; ^
		dc.b	$31					; 1
		dc.b	$18
		dc.b	$35					; 5
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$9D					; 
		dc.b	$4F					; O
		dc.b	$1F
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$43					; C
		dc.b	1
		dc.b	1
		dc.b	$81					; 
		dc.b	$CD					; 
		dc.b	$A9					; 
		dc.b	$D7					; 
		dc.b	$80					; 
		dc.b	$48					; H
		dc.b	$7C					; |
		dc.b	$4A					; J
		dc.b	$1B
		dc.b	$7F					; 
		dc.b	$B6					; 
		dc.b	$94					; 
		dc.b	$CC					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$BD					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$4A					; J
		dc.b	$19
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$46					; F
		dc.b	$5A					; Z
		dc.b	$B3					; 
		dc.b	$B4					; 
		dc.b	$84					; 
		dc.b	$6F					; o
		dc.b	$45					; E
		dc.b	$3B					; ;
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$3C					; <
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$9F					; 
		dc.b	$FA					; 
		dc.b	$D0					; 
		dc.b	$27					; '
		dc.b	$41					; A
		dc.b	$6F					; o
		dc.b	$86					; 
		dc.b	$BB					; 
		dc.b	$DE					; 
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$2A					; *
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$BD					; 
		dc.b	$8A					; 
		dc.b	$38					; 8
		dc.b	$50					; P
		dc.b	$31					; 1
		dc.b	$55					; U
		dc.b	$B1					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$49					; I
		dc.b	$84					; 
		dc.b	$D6					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$8F					; 
		dc.b	$43					; C
		dc.b	$A
		dc.b	$81					; 
		dc.b	$D9					; 
		dc.b	$CF					; 
		dc.b	$95					; 
		dc.b	$20
		dc.b	$26					; &
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$91					; 
		dc.b	$61					; a
		dc.b	$AB					; 
		dc.b	$FC					; 
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$73					; s
		dc.b	$2C					; ,
		dc.b	$57					; W
		dc.b	3
		dc.b	$1C
		dc.b	$C6					; 
		dc.b	$DE					; 
		dc.b	$CF					; 
		dc.b	$C5					; 
		dc.b	$2C					; ,
		dc.b	1
		dc.b	$4F					; O
		dc.b	$25					; %
		dc.b	$72					; r
		dc.b	$CF					; 
		dc.b	$8A					; 
		dc.b	$BA					; 
		dc.b	$C9					; 
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$B6					; 
		dc.b	$7A					; z
		dc.b	$A7					; 
		dc.b	$80					; 
		dc.b	1
		dc.b	$18
		dc.b	$AA					; 
		dc.b	$DA					; 
		dc.b	$E4					; 
		dc.b	$8E					; 
		dc.b	$29					; )
		dc.b	$3B					; ;
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$5A					; Z
		dc.b	$46					; F
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$5A					; Z
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$4A					; J
		dc.b	$81					; 
		dc.b	$B8					; 
		dc.b	$F5					; 
		dc.b	$CB					; 
		dc.b	$20
		dc.b	$1C
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$BE					; 
		dc.b	$A7					; 
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$31					; 1
		dc.b	$5A					; Z
		dc.b	$AA					; 
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$49					; I
		dc.b	$11
		dc.b	$41					; A
		dc.b	$9D					; 
		dc.b	$DA					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$59					; Y
		dc.b	$6C					; l
		dc.b	$B6					; 
		dc.b	$76					; v
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$A2					; 
		dc.b	$C2					; 
		dc.b	$76					; v
		dc.b	$22					; "
		dc.b	$2C					; ,
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$8D					; 
		dc.b	$E
		dc.b	1
		dc.b	$57					; W
		dc.b	$DA					; 
		dc.b	$9B					; 
		dc.b	$CB					; 
		dc.b	$B1					; 
		dc.b	$31					; 1
		dc.b	$86					; 
		dc.b	$5A					; Z
		dc.b	$22					; "
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$64					; d
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$A7					; 
		dc.b	$6A					; j
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$D3					; 
		dc.b	$B1					; 
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$6E					; n
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$A6					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$57					; W
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$6B					; k
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$CC					; 
		dc.b	$B3					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$3C					; <
		dc.b	$5D					; ]
		dc.b	$AD					; 
		dc.b	$12
		dc.b	1
		dc.b	$61					; a
		dc.b	$AD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$99					; 
		dc.b	$4D					; M
		dc.b	$2D					; -
		dc.b	$2A					; *
		dc.b	$65					; e
		dc.b	$A0					; 
		dc.b	$D0					; 
		dc.b	$B4					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$60					; `
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$A0					; 
		dc.b	$C6					; 
		dc.b	$AB					; 
		dc.b	$81					; 
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$2D					; -
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$C2					; 
		dc.b	$C2					; 
		dc.b	$AE					; 
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$3F					; ?
		dc.b	$4F					; O
		dc.b	$91					; 
		dc.b	$BB					; 
		dc.b	$A4					; 
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$72					; r
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$B3					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$64					; d
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$59					; Y
		dc.b	$42					; B
		dc.b	$56					; V
		dc.b	$91					; 
		dc.b	$D3					; 
		dc.b	$DB					; 
		dc.b	$9B					; 
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$3B					; ;
		dc.b	$37					; 7
		dc.b	$8B					; 
		dc.b	$D0					; 
		dc.b	$FA					; 
		dc.b	$BD					; 
		dc.b	$4F					; O
		dc.b	$29					; )
		dc.b	$10
		dc.b	$73					; s
		dc.b	$D2					; 
		dc.b	$BD					; 
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$53					; S
		dc.b	$8D					; 
		dc.b	$B6					; 
		dc.b	$99					; 
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$A4					; 
		dc.b	$BB					; 
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$73					; s
		dc.b	$3F					; ?
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$88					; 
		dc.b	$E4					; 
		dc.b	$DB					; 
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$18
		dc.b	$3B					; ;
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$EC					; 
		dc.b	$C6					; 
		dc.b	$61					; a
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$38					; 8
		dc.b	$6A					; j
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$C6					; 
		dc.b	$88					; 
		dc.b	$18
		dc.b	$5C					; \
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$A0					; 
		dc.b	$5A					; Z
		dc.b	$4C					; L
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$C9					; 
		dc.b	$C6					; 
		dc.b	$BF					; 
		dc.b	$84					; 
		dc.b	$2C					; ,
		dc.b	$4C					; L
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$B6					; 
		dc.b	$D7					; 
		dc.b	$A4					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$9B					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$D
		dc.b	$38					; 8
		dc.b	$CF					; 
		dc.b	$E1					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$19
		dc.b	$30					; 0
		dc.b	$C5					; 
		dc.b	$F0					; 
		dc.b	$C1					; 
		dc.b	$7F					; 
		dc.b	$31					; 1
		dc.b	$B
		dc.b	$57					; W
		dc.b	$9D					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$55					; U
		dc.b	$34					; 4
		dc.b	$46					; F
		dc.b	$73					; s
		dc.b	$A3					; 
		dc.b	$BF					; 
		dc.b	$D7					; 
		dc.b	$9F					; 
		dc.b	$7C					; |
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$A6					; 
		dc.b	$BB					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$81					; 
		dc.b	$5A					; Z
		dc.b	$53					; S
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$63					; c
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$4F					; O
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$48					; H
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$AD					; 
		dc.b	$9B					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$83					; 
		dc.b	$B4					; 
		dc.b	$AD					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$AB					; 
		dc.b	$BE					; 
		dc.b	$CC					; 
		dc.b	$6F					; o
		dc.b	$42					; B
		dc.b	$34					; 4
		dc.b	$2C					; ,
		dc.b	$65					; e
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$35					; 5
		dc.b	$6E					; n
		dc.b	$C8					; 
		dc.b	$B8					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$7F					; 
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$41					; A
		dc.b	$65					; e
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$56					; V
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$4A					; J
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$A6					; 
		dc.b	$D9					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$7C					; |
		dc.b	$2E					; .
		dc.b	$67					; g
		dc.b	$F2					; 
		dc.b	$D0					; 
		dc.b	$71					; q
		dc.b	$3C					; <
		dc.b	$10
		dc.b	$45					; E
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$AE					; 
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$45					; E
		dc.b	1
		dc.b	$4F					; O
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$E0					; 
		dc.b	$BA					; 
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$E
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$DA					; 
		dc.b	$D2					; 
		dc.b	$8A					; 
		dc.b	$12
		dc.b	$22					; "
		dc.b	$41					; A
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$CD					; 
		dc.b	$B7					; 
		dc.b	$57					; W
		dc.b	$41					; A
		dc.b	$43					; C
		dc.b	$20
		dc.b	$37					; 7
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$C4					; 
		dc.b	$E7					; 
		dc.b	$DD					; 
		dc.b	$83					; 
		dc.b	$3A					; :
		dc.b	$25					; %
		dc.b	$33					; 3
		dc.b	$68					; h
		dc.b	$96					; 
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$B4					; 
		dc.b	$AA					; 
		dc.b	$D7					; 
		dc.b	$B6					; 
		dc.b	$4D					; M
		dc.b	$27					; '
		dc.b	$26					; &
		dc.b	$17
		dc.b	$95					; 
		dc.b	$D3					; 
		dc.b	$B0					; 
		dc.b	$DE					; 
		dc.b	$87					; 
		dc.b	$2D					; -
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$D7					; 
		dc.b	$E8					; 
		dc.b	$92					; 
		dc.b	$1F
		dc.b	$17
		dc.b	$76					; v
		dc.b	$CB					; 
		dc.b	$CB					; 
		dc.b	$A6					; 
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$84					; 
		dc.b	$B1					; 
		dc.b	$DA					; 
		dc.b	$9F					; 
		dc.b	$29					; )
		dc.b	$41					; A
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$B6					; 
		dc.b	$96					; 
		dc.b	$57					; W
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$68					; h
		dc.b	$8F					; 
		dc.b	$56					; V
		dc.b	$7A					; z
		dc.b	$EB					; 
		dc.b	$95					; 
		dc.b	$56					; V
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$B0					; 
		dc.b	$68					; h
		dc.b	$3E					; >
		dc.b	$4F					; O
		dc.b	$78					; x
		dc.b	$BE					; 
		dc.b	$BB					; 
		dc.b	$A9					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$3C					; <
		dc.b	$4C					; L
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$BA					; 
		dc.b	$BD					; 
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$DA					; 
		dc.b	$DE					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$4A					; J
		dc.b	$1F
		dc.b	$25					; %
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$C5					; 
		dc.b	$B6					; 
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$91					; 
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$AE					; 
		dc.b	$B6					; 
		dc.b	$88					; 
		dc.b	$1E
		dc.b	$52					; R
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$1C
		dc.b	$43					; C
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$57					; W
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$DA					; 
		dc.b	$B4					; 
		dc.b	$3C					; <
		dc.b	$29					; )
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$B4					; 
		dc.b	$F5					; 
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$8D					; 
		dc.b	$56					; V
		dc.b	$41					; A
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$72					; r
		dc.b	$E5					; 
		dc.b	$CC					; 
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$43					; C
		dc.b	$38					; 8
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$B6					; 
		dc.b	$BD					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$60					; `
		dc.b	$AD					; 
		dc.b	$A2					; 
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$AB					; 
		dc.b	$BE					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$2E					; .
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$A9					; 
		dc.b	$BD					; 
		dc.b	$60					; `
		dc.b	$4C					; L
		dc.b	$8B					; 
		dc.b	$52					; R
		dc.b	$8A					; 
		dc.b	$B6					; 
		dc.b	$4F					; O
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$3C					; <
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$45					; E
		dc.b	$4A					; J
		dc.b	$8E					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$94					; 
		dc.b	$5C					; \
		dc.b	$B1					; 
		dc.b	$CF					; 
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$63					; c
		dc.b	$46					; F
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$81					; 
		dc.b	$C9					; 
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$42					; B
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$C2					; 
		dc.b	$CB					; 
		dc.b	$63					; c
		dc.b	$3A					; :
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$6E					; n
		dc.b	$A0					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$C5					; 
		dc.b	$B1					; 
		dc.b	$6C					; l
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$8E					; 
		dc.b	$B3					; 
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$56					; V
		dc.b	$4C					; L
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$95					; 
		dc.b	$A6					; 
		dc.b	$C8					; 
		dc.b	$A2					; 
		dc.b	$73					; s
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$C4					; 
		dc.b	$C8					; 
		dc.b	$6F					; o
		dc.b	$4C					; L
		dc.b	$2C					; ,
		dc.b	$41					; A
		dc.b	$8D					; 
		dc.b	$BA					; 
		dc.b	$C9					; 
		dc.b	$87					; 
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$E2					; 
		dc.b	$BD					; 
		dc.b	$3B					; ;
		dc.b	$48					; H
		dc.b	$59					; Y
		dc.b	$52					; R
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$42					; B
		dc.b	1
		dc.b	$33					; 3
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$D4					; 
		dc.b	$EB					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$3E					; >
		dc.b	$25					; %
		dc.b	$72					; r
		dc.b	$B8					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$C5					; 
		dc.b	$BD					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$30					; 0
		dc.b	$33					; 3
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$B1					; 
		dc.b	$E0					; 
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$4C					; L
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$33					; 3
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$8A					; 
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$BB					; 
		dc.b	$C5					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$3F					; ?
		dc.b	$19
		dc.b	$84					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$60					; `
		dc.b	$3C					; <
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$50					; P
		dc.b	$9F					; 
		dc.b	$A9					; 
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$31					; 1
		dc.b	$2E					; .
		dc.b	$84					; 
		dc.b	$CC					; 
		dc.b	$EF					; 
		dc.b	$9C					; 
		dc.b	$50					; P
		dc.b	$38					; 8
		dc.b	$14
		dc.b	$30					; 0
		dc.b	$B4					; 
		dc.b	$AA					; 
		dc.b	$C2					; 
		dc.b	$E0					; 
		dc.b	$63					; c
		dc.b	$37					; 7
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$81					; 
		dc.b	$35					; 5
		dc.b	$5C					; \
		dc.b	$CF					; 
		dc.b	$CD					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$20
		dc.b	1
		dc.b	$3C					; <
		dc.b	$AE					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$9D					; 
		dc.b	$4F					; O
		dc.b	$14
		dc.b	$60					; `
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$C1					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$22					; "
		dc.b	$46					; F
		dc.b	$A9					; 
		dc.b	$C5					; 
		dc.b	$D7					; 
		dc.b	$8A					; 
		dc.b	$2C					; ,
		dc.b	$11
		dc.b	$5C					; \
		dc.b	$A3					; 
		dc.b	$91					; 
		dc.b	$B6					; 
		dc.b	$87					; 
		dc.b	$45					; E
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$10
		dc.b	$86					; 
		dc.b	$D3					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$3F					; ?
		dc.b	$63					; c
		dc.b	$BD					; 
		dc.b	$C8					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$B8					; 
		dc.b	$A3					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$3C					; <
		dc.b	$26					; &
		dc.b	$71					; q
		dc.b	$99					; 
		dc.b	$D0					; 
		dc.b	$D7					; 
		dc.b	$BA					; 
		dc.b	$60					; `
		dc.b	$3B					; ;
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$BF					; 
		dc.b	$AB					; 
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$4D					; M
		dc.b	$1C
		dc.b	$49					; I
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$CC					; 
		dc.b	$A0					; 
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$CB					; 
		dc.b	$C2					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$D2					; 
		dc.b	$7F					; 
		dc.b	$2E					; .
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$D0					; 
		dc.b	$A6					; 
		dc.b	$4C					; L
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$AA					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$3E					; >
		dc.b	$73					; s
		dc.b	$C1					; 
		dc.b	$F5					; 
		dc.b	$C1					; 
		dc.b	$8A					; 
		dc.b	$4D					; M
		dc.b	$B
		dc.b	$3B					; ;
		dc.b	$38					; 8
		dc.b	$65					; e
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$7C					; |
		dc.b	1
		dc.b	2
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$9D					; 
		dc.b	$73					; s
		dc.b	$B4					; 
		dc.b	$B8					; 
		dc.b	$7C					; |
		dc.b	$BE					; 
		dc.b	$84					; 
		dc.b	$45					; E
		dc.b	$46					; F
		dc.b	2
		dc.b	$4A					; J
		dc.b	$AD					; 
		dc.b	$9B					; 
		dc.b	$C1					; 
		dc.b	$FE					; 
		dc.b	$8B					; 
		dc.b	$49					; I
		dc.b	$4C					; L
		dc.b	1
		dc.b	$5E					; ^
		dc.b	$B6					; 
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$D2					; 
		dc.b	$CC					; 
		dc.b	$65					; e
		dc.b	$45					; E
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$68					; h
		dc.b	$AA					; 
		dc.b	$D4					; 
		dc.b	$C1					; 
		dc.b	$99					; 
		dc.b	$3C					; <
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$2E					; .
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$61					; a
		dc.b	$34					; 4
		dc.b	$59					; Y
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$AE					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$B3					; 
		dc.b	$76					; v
		dc.b	$55					; U
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$5E					; ^
		dc.b	$8D					; 
		dc.b	$BB					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$B7					; 
		dc.b	$73					; s
		dc.b	$35					; 5
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$C8					; 
		dc.b	$AD					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$9F					; 
		dc.b	$DD					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$27					; '
		dc.b	$64					; d
		dc.b	$B0					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$CB					; 
		dc.b	$78					; x
		dc.b	$49					; I
		dc.b	$7D					; }
		dc.b	$5A					; Z
		dc.b	$48					; H
		dc.b	$96					; 
		dc.b	$BB					; 
		dc.b	$63					; c
		dc.b	$92					; 
		dc.b	$AB					; 
		dc.b	$50					; P
		dc.b	$3F					; ?
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$AE					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$4C					; L
		dc.b	$2E					; .
		dc.b	$8F					; 
		dc.b	$B8					; 
		dc.b	$C1					; 
		dc.b	$A6					; 
		dc.b	$5E					; ^
		dc.b	$3B					; ;
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$B4					; 
		dc.b	$B6					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$31					; 1
		dc.b	$3A					; :
		dc.b	$B3					; 
		dc.b	$EE					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$AB					; 
		dc.b	$6A					; j
		dc.b	$4A					; J
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$7C					; |
		dc.b	$43					; C
		dc.b	$64					; d
		dc.b	$BE					; 
		dc.b	$B7					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	1
		dc.b	7
		dc.b	$80					; 
		dc.b	$CD					; 
		dc.b	$E2					; 
		dc.b	$F2					; 
		dc.b	$A7					; 
		dc.b	$57					; W
		dc.b	$37					; 7
		dc.b	$3F					; ?
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$AB					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$5A					; Z
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$B8					; 
		dc.b	$A7					; 
		dc.b	$BF					; 
		dc.b	$8E					; 
		dc.b	$4D					; M
		dc.b	$41					; A
		dc.b	$4A					; J
		dc.b	$46					; F
		dc.b	$5D					; ]
		dc.b	$50					; P
		dc.b	$94					; 
		dc.b	$BA					; 
		dc.b	$A4					; 
		dc.b	$B7					; 
		dc.b	$AE					; 
		dc.b	$6C					; l
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$A6					; 
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$AA					; 
		dc.b	$79					; y
		dc.b	$9F					; 
		dc.b	$76					; v
		dc.b	$2D					; -
		dc.b	$5E					; ^
		dc.b	$A4					; 
		dc.b	$D6					; 
		dc.b	$E5					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$17
		dc.b	$B
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$B0					; 
		dc.b	$34					; 4
		dc.b	1
		dc.b	$31					; 1
		dc.b	$46					; F
		dc.b	$AB					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$4D					; M
		dc.b	$3A					; :
		dc.b	$30					; 0
		dc.b	$88					; 
		dc.b	$D3					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$A6					; 
		dc.b	$C2					; 
		dc.b	$B3					; 
		dc.b	$73					; s
		dc.b	$27					; '
		dc.b	6
		dc.b	$43					; C
		dc.b	$A2					; 
		dc.b	$CD					; 
		dc.b	$DB					; 
		dc.b	$C2					; 
		dc.b	$3A					; :
		dc.b	$18
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$B8					; 
		dc.b	$BF					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$2D					; -
		dc.b	$4A					; J
		dc.b	$78					; x
		dc.b	$AE					; 
		dc.b	$A6					; 
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$B7					; 
		dc.b	$C5					; 
		dc.b	$A7					; 
		dc.b	$75					; u
		dc.b	$3B					; ;
		dc.b	$1F
		dc.b	$5E					; ^
		dc.b	$8A					; 
		dc.b	$E7					; 
		dc.b	$E2					; 
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$B1					; 
		dc.b	$C9					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$38					; 8
		dc.b	$2C					; ,
		dc.b	$56					; V
		dc.b	$6B					; k
		dc.b	$92					; 
		dc.b	$B6					; 
		dc.b	$98					; 
		dc.b	$68					; h
		dc.b	$9B					; 
		dc.b	$6E					; n
		dc.b	$3E					; >
		dc.b	$3F					; ?
		dc.b	$42					; B
		dc.b	$67					; g
		dc.b	$B3					; 
		dc.b	$C2					; 
		dc.b	$B4					; 
		dc.b	$C9					; 
		dc.b	$50					; P
		dc.b	$19
		dc.b	$4C					; L
		dc.b	$64					; d
		dc.b	$A2					; 
		dc.b	$CC					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$AB					; 
		dc.b	$A3					; 
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$57					; W
		dc.b	$23					; #
		dc.b	$4D					; M
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$4C					; L
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$7A					; z
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	$3F					; ?
		dc.b	$53					; S
		dc.b	$9F					; 
		dc.b	$D4					; 
		dc.b	$C1					; 
		dc.b	$8D					; 
		dc.b	$35					; 5
		dc.b	$29					; )
		dc.b	$48					; H
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$60					; `
		dc.b	$38					; 8
		dc.b	$20
		dc.b	$4C					; L
		dc.b	$A0					; 
		dc.b	$B3					; 
		dc.b	$BD					; 
		dc.b	$D2					; 
		dc.b	$9D					; 
		dc.b	$29					; )
		dc.b	9
		dc.b	$27					; '
		dc.b	$6C					; l
		dc.b	$AD					; 
		dc.b	$C4					; 
		dc.b	$BD					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$5A					; Z
		dc.b	$61					; a
		dc.b	$CD					; 
		dc.b	$BD					; 
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$2E					; .
		dc.b	$4F					; O
		dc.b	$9B					; 
		dc.b	$E0					; 
		dc.b	$C6					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$72					; r
		dc.b	$3C					; <
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$95					; 
		dc.b	$E0					; 
		dc.b	$BB					; 
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$43					; C
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$8E					; 
		dc.b	$BF					; 
		dc.b	$C5					; 
		dc.b	$98					; 
		dc.b	$BB					; 
		dc.b	$95					; 
		dc.b	$37					; 7
		dc.b	$19
		dc.b	$17
		dc.b	$50					; P
		dc.b	$88					; 
		dc.b	$C1					; 
		dc.b	$E4					; 
		dc.b	$B3					; 
		dc.b	$B3					; 
		dc.b	$59					; Y
		dc.b	$19
		dc.b	$35					; 5
		dc.b	$4A					; J
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$BA					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$B6					; 
		dc.b	$60					; `
		dc.b	$3B					; ;
		dc.b	$33					; 3
		dc.b	$55					; U
		dc.b	$80					; 
		dc.b	$B8					; 
		dc.b	$DE					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$72					; r
		dc.b	1
		dc.b	$1E
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$D6					; 
		dc.b	$FE					; 
		dc.b	$BA					; 
		dc.b	$61					; a
		dc.b	$27					; '
		dc.b	$12
		dc.b	$5A					; Z
		dc.b	$9D					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$4D					; M
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$99					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$43					; C
		dc.b	$38					; 8
		dc.b	$8E					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$29					; )
		dc.b	$6E					; n
		dc.b	$A3					; 
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$A3					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$B1					; 
		dc.b	$60					; `
		dc.b	$31					; 1
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$A4					; 
		dc.b	$AB					; 
		dc.b	$9D					; 
		dc.b	$4C					; L
		dc.b	$2E					; .
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$B8					; 
		dc.b	$94					; 
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$38					; 8
		dc.b	$17
		dc.b	$8D					; 
		dc.b	$C1					; 
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$75					; u
		dc.b	$34					; 4
		dc.b	$67					; g
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$CC					; 
		dc.b	$C9					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$68					; h
		dc.b	$2D					; -
		dc.b	$29					; )
		dc.b	$5E					; ^
		dc.b	$AD					; 
		dc.b	$DE					; 
		dc.b	$A0					; 
		dc.b	$59					; Y
		dc.b	$30					; 0
		dc.b	$5D					; ]
		dc.b	$BB					; 
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$18
		dc.b	$56					; V
		dc.b	$8E					; 
		dc.b	$C9					; 
		dc.b	$E5					; 
		dc.b	$B1					; 
		dc.b	$3B					; ;
		dc.b	$45					; E
		dc.b	$53					; S
		dc.b	$4D					; M
		dc.b	$9B					; 
		dc.b	$B8					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$8D					; 
		dc.b	$C6					; 
		dc.b	$BE					; 
		dc.b	$9B					; 
		dc.b	$49					; I
		dc.b	$3C					; <
		dc.b	$4D					; M
		dc.b	$68					; h
		dc.b	$B1					; 
		dc.b	$DE					; 
		dc.b	$CD					; 
		dc.b	$AD					; 
		dc.b	$4D					; M
		dc.b	$3B					; ;
		dc.b	$5C					; \
		dc.b	$A0					; 
		dc.b	$B8					; 
		dc.b	$A0					; 
		dc.b	$6E					; n
		dc.b	$2D					; -
		dc.b	$17
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$C5					; 
		dc.b	$C9					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$50					; P
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$78					; x
		dc.b	$C4					; 
		dc.b	$B3					; 
		dc.b	$84					; 
		dc.b	$49					; I
		dc.b	1
		dc.b	$27					; '
		dc.b	$78					; x
		dc.b	$BB					; 
		dc.b	$E0					; 
		dc.b	$75					; u
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$9D					; 
		dc.b	$C2					; 
		dc.b	$99					; 
		dc.b	$42					; B
		dc.b	$4A					; J
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$A9					; 
		dc.b	$B3					; 
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$BD					; 
		dc.b	$A2					; 
		dc.b	$8B					; 
		dc.b	$3C					; <
		dc.b	$3C					; <
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$D9					; 
		dc.b	$A0					; 
		dc.b	$63					; c
		dc.b	$48					; H
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$6F					; o
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$A2					; 
		dc.b	$4F					; O
		dc.b	$2E					; .
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$B6					; 
		dc.b	$B8					; 
		dc.b	$52					; R
		dc.b	$19
		dc.b	$3C					; <
		dc.b	$92					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$AD					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$31					; 1
		dc.b	$12
		dc.b	$29					; )
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$CD					; 
		dc.b	$E7					; 
		dc.b	$91					; 
		dc.b	$3E					; >
		dc.b	$71					; q
		dc.b	$2E					; .
		dc.b	$2D					; -
		dc.b	$88					; 
		dc.b	$C1					; 
		dc.b	$B3					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$37					; 7
		dc.b	$6A					; j
		dc.b	$D0					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$50					; P
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$75					; u
		dc.b	$25					; %
		dc.b	$15
		dc.b	$61					; a
		dc.b	$A7					; 
		dc.b	$FE					; 
		dc.b	$FC					; 
		dc.b	$65					; e
		dc.b	$18
		dc.b	$38					; 8
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$61					; a
		dc.b	$2C					; ,
		dc.b	$5C					; \
		dc.b	$AA					; 
		dc.b	$CC					; 
		dc.b	$A7					; 
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$C6					; 
		dc.b	$B0					; 
		dc.b	$15
		dc.b	1
		dc.b	$7A					; z
		dc.b	$D9					; 
		dc.b	$D9					; 
		dc.b	$B1					; 
		dc.b	$6F					; o
		dc.b	$37					; 7
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	$31					; 1
		dc.b	$11
		dc.b	$65					; e
		dc.b	$CD					; 
		dc.b	$B8					; 
		dc.b	$BD					; 
		dc.b	$AB					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$B8					; 
		dc.b	$AD					; 
		dc.b	$61					; a
		dc.b	$8B					; 
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$BF					; 
		dc.b	$56					; V
		dc.b	$3E					; >
		dc.b	$5D					; ]
		dc.b	$38					; 8
		dc.b	$78					; x
		dc.b	$AD					; 
		dc.b	$AD					; 
		dc.b	$CF					; 
		dc.b	$8F					; 
		dc.b	$43					; C
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$8F					; 
		dc.b	$C5					; 
		dc.b	$BD					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$4C					; L
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$9F					; 
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$D7					; 
		dc.b	$76					; v
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$38					; 8
		dc.b	$64					; d
		dc.b	$BD					; 
		dc.b	$95					; 
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$5E					; ^
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$8D					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$67					; g
		dc.b	$AD					; 
		dc.b	$C6					; 
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$53					; S
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$3B					; ;
		dc.b	$3F					; ?
		dc.b	$83					; 
		dc.b	$A9					; 
		dc.b	$C1					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$42					; B
		dc.b	$7D					; }
		dc.b	$A0					; 
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$42					; B
		dc.b	$68					; h
		dc.b	$AE					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$18
		dc.b	$33					; 3
		dc.b	$BA					; 
		dc.b	$E9					; 
		dc.b	$D0					; 
		dc.b	$98					; 
		dc.b	$38					; 8
		dc.b	$3B					; ;
		dc.b	$5E					; ^
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$AD					; 
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$3C					; <
		dc.b	$29					; )
		dc.b	$42					; B
		dc.b	$AB					; 
		dc.b	$BA					; 
		dc.b	$B7					; 
		dc.b	$78					; x
		dc.b	$15
		dc.b	$31					; 1
		dc.b	$A3					; 
		dc.b	$DD					; 
		dc.b	$CC					; 
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$9B					; 
		dc.b	$61					; a
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$B8					; 
		dc.b	$AE					; 
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$59					; Y
		dc.b	$2D					; -
		dc.b	$6C					; l
		dc.b	$A4					; 
		dc.b	$B3					; 
		dc.b	$B0					; 
		dc.b	$6F					; o
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$3C					; <
		dc.b	$A4					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$9C					; 
		dc.b	$AE					; 
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$73					; s
		dc.b	7
		dc.b	$6C					; l
		dc.b	$BB					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$B7					; 
		dc.b	$BA					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$2C					; ,
		dc.b	$22					; "
		dc.b	$61					; a
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$E7					; 
		dc.b	$83					; 
		dc.b	$3E					; >
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$2A					; *
		dc.b	$BB					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	1
		dc.b	1
		dc.b	$64					; d
		dc.b	$AD					; 
		dc.b	$E4					; 
		dc.b	$B0					; 
		dc.b	$71					; q
		dc.b	$2E					; .
		dc.b	$50					; P
		dc.b	$60					; `
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$A3					; 
		dc.b	$EE					; 
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$53					; S
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$A3					; 
		dc.b	$BF					; 
		dc.b	$B8					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$B8					; 
		dc.b	$72					; r
		dc.b	$1B
		dc.b	$34					; 4
		dc.b	$99					; 
		dc.b	$FE					; 
		dc.b	$BA					; 
		dc.b	$65					; e
		dc.b	$42					; B
		dc.b	$1C
		dc.b	$56					; V
		dc.b	$92					; 
		dc.b	$C8					; 
		dc.b	$C2					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$30					; 0
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$45					; E
		dc.b	$7C					; |
		dc.b	$CC					; 
		dc.b	$BF					; 
		dc.b	$A3					; 
		dc.b	$4C					; L
		dc.b	$3A					; :
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$34					; 4
		dc.b	$4D					; M
		dc.b	$31					; 1
		dc.b	$60					; `
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$D3					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$11
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$CC					; 
		dc.b	$BA					; 
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$2E					; .
		dc.b	$22					; "
		dc.b	$8B					; 
		dc.b	$B3					; 
		dc.b	$DB					; 
		dc.b	$AD					; 
		dc.b	$42					; B
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$A9					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$2C					; ,
		dc.b	$52					; R
		dc.b	$80					; 
		dc.b	$B8					; 
		dc.b	$D0					; 
		dc.b	$8D					; 
		dc.b	$34					; 4
		dc.b	$49					; I
		dc.b	$6B					; k
		dc.b	$B1					; 
		dc.b	$CD					; 
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$3C					; <
		dc.b	$3A					; :
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$CC					; 
		dc.b	$BE					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$50					; P
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$55					; U
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$59					; Y
		dc.b	$52					; R
		dc.b	$8E					; 
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$A0					; 
		dc.b	$CD					; 
		dc.b	$96					; 
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$68					; h
		dc.b	6
		dc.b	$3F					; ?
		dc.b	$C4					; 
		dc.b	$D7					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$25					; %
		dc.b	$52					; R
		dc.b	$A6					; 
		dc.b	$CB					; 
		dc.b	$E9					; 
		dc.b	$AB					; 
		dc.b	$83					; 
		dc.b	$43					; C
		dc.b	$38					; 8
		dc.b	$7D					; }
		dc.b	$56					; V
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$A2					; 
		dc.b	$C8					; 
		dc.b	$BB					; 
		dc.b	$A2					; 
		dc.b	$48					; H
		dc.b	$2E					; .
		dc.b	$35					; 5
		dc.b	$71					; q
		dc.b	$BF					; 
		dc.b	$B3					; 
		dc.b	$95					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$9F					; 
		dc.b	$E4					; 
		dc.b	$B4					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$33					; 3
		dc.b	$2D					; -
		dc.b	$64					; d
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$AA					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$43					; C
		dc.b	$2E					; .
		dc.b	$50					; P
		dc.b	$79					; y
		dc.b	$6C					; l
		dc.b	$A2					; 
		dc.b	$D9					; 
		dc.b	$EC					; 
		dc.b	$96					; 
		dc.b	$19
		dc.b	$61					; a
		dc.b	$F0					; 
		dc.b	$87					; 
		dc.b	$18
		dc.b	$61					; a
		dc.b	$87					; 
		dc.b	$C1					; 
		dc.b	$B1					; 
		dc.b	$A3					; 
		dc.b	$80					; 
		dc.b	$3C					; <
		dc.b	$31					; 1
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$4C					; L
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$41					; A
		dc.b	$46					; F
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$E4					; 
		dc.b	$72					; r
		dc.b	$45					; E
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$4C					; L
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$B6					; 
		dc.b	$81					; 
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$B7					; 
		dc.b	$81					; 
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$A9					; 
		dc.b	$F0					; 
		dc.b	$9B					; 
		dc.b	$7F					; 
		dc.b	$60					; `
		dc.b	$2C					; ,
		dc.b	$33					; 3
		dc.b	$B8					; 
		dc.b	$D0					; 
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$5D					; ]
		dc.b	$B
		dc.b	$B
		dc.b	$84					; 
		dc.b	$DB					; 
		dc.b	$C6					; 
		dc.b	$81					; 
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$56					; V
		dc.b	$2C					; ,
		dc.b	$73					; s
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$29					; )
		dc.b	$5E					; ^
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$38					; 8
		dc.b	$88					; 
		dc.b	$B7					; 
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$B0					; 
		dc.b	$9B					; 
		dc.b	$6A					; j
		dc.b	$25					; %
		dc.b	$30					; 0
		dc.b	$34					; 4
		dc.b	$67					; g
		dc.b	$CC					; 
		dc.b	$D9					; 
		dc.b	$BE					; 
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$9F					; 
		dc.b	$DE					; 
		dc.b	$AB					; 
		dc.b	$6B					; k
		dc.b	$4C					; L
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$4A					; J
		dc.b	$56					; V
		dc.b	$30					; 0
		dc.b	$6E					; n
		dc.b	$A4					; 
		dc.b	$8F					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$A3					; 
		dc.b	$7D					; }
		dc.b	$45					; E
		dc.b	$26					; &
		dc.b	$34					; 4
		dc.b	$5D					; ]
		dc.b	$87					; 
		dc.b	$DD					; 
		dc.b	$EF					; 
		dc.b	$E2					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$2D					; -
		dc.b	$11
		dc.b	$1B
		dc.b	$7C					; |
		dc.b	$B1					; 
		dc.b	$CC					; 
		dc.b	$C6					; 
		dc.b	$9B					; 
		dc.b	$6E					; n
		dc.b	$42					; B
		dc.b	$65					; e
		dc.b	$50					; P
		dc.b	$B
		dc.b	$45					; E
		dc.b	$8E					; 
		dc.b	$C2					; 
		dc.b	$FE					; 
		dc.b	$DB					; 
		dc.b	$A2					; 
		dc.b	$59					; Y
		dc.b	1
		dc.b	$37					; 7
		dc.b	$57					; W
		dc.b	$EC					; 
		dc.b	$D0					; 
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$3F					; ?
		dc.b	$8E					; 
		dc.b	$A4					; 
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$3F					; ?
		dc.b	$49					; I
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$7D					; }
		dc.b	$D2					; 
		dc.b	$F6					; 
		dc.b	$87					; 
		dc.b	$45					; E
		dc.b	$3E					; >
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$A0					; 
		dc.b	$7C					; |
		dc.b	$5E					; ^
		dc.b	$8A					; 
		dc.b	$BF					; 
		dc.b	$B8					; 
		dc.b	$7D					; }
		dc.b	$5D					; ]
		dc.b	$2A					; *
		dc.b	$67					; g
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	$AA					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$BA					; 
		dc.b	$C6					; 
		dc.b	$A0					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$45					; E
		dc.b	$27					; '
		dc.b	$60					; `
		dc.b	$CF					; 
		dc.b	$B7					; 
		dc.b	$A6					; 
		dc.b	$88					; 
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$38					; 8
		dc.b	$71					; q
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$4A					; J
		dc.b	$96					; 
		dc.b	$60					; `
		dc.b	$12
		dc.b	$71					; q
		dc.b	$BF					; 
		dc.b	$B7					; 
		dc.b	$B3					; 
		dc.b	$A3					; 
		dc.b	$37					; 7
		dc.b	$34					; 4
		dc.b	$A4					; 
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$5E					; ^
		dc.b	$49					; I
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$87					; 
		dc.b	$38					; 8
		dc.b	$2C					; ,
		dc.b	$3F					; ?
		dc.b	$87					; 
		dc.b	$A6					; 
		dc.b	$C4					; 
		dc.b	$C5					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$35					; 5
		dc.b	$4D					; M
		dc.b	$C5					; 
		dc.b	$C4					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$92					; 
		dc.b	$A9					; 
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$B0					; 
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$41					; A
		dc.b	$12
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$C2					; 
		dc.b	$CB					; 
		dc.b	$91					; 
		dc.b	$45					; E
		dc.b	$20
		dc.b	$5C					; \
		dc.b	$8B					; 
		dc.b	$B3					; 
		dc.b	$C6					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$52					; R
		dc.b	$91					; 
		dc.b	$E7					; 
		dc.b	$7F					; 
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$12
		dc.b	$5A					; Z
		dc.b	$BB					; 
		dc.b	$AE					; 
		dc.b	$C5					; 
		dc.b	$AD					; 
		dc.b	$55					; U
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$2A					; *
		dc.b	$91					; 
		dc.b	$AA					; 
		dc.b	$76					; v
		dc.b	$4C					; L
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$84					; 
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$8B					; 
		dc.b	$AE					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$46					; F
		dc.b	$55					; U
		dc.b	$C4					; 
		dc.b	$D2					; 
		dc.b	$99					; 
		dc.b	$35					; 5
		dc.b	$2C					; ,
		dc.b	$49					; I
		dc.b	$8E					; 
		dc.b	$E0					; 
		dc.b	$BE					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$6E					; n
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$9C					; 
		dc.b	$CB					; 
		dc.b	$BF					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$D4					; 
		dc.b	$B6					; 
		dc.b	$57					; W
		dc.b	$52					; R
		dc.b	$34					; 4
		dc.b	9
		dc.b	$64					; d
		dc.b	$DE					; 
		dc.b	$E2					; 
		dc.b	$B8					; 
		dc.b	$52					; R
		dc.b	$14
		dc.b	$27					; '
		dc.b	$AB					; 
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$67					; g
		dc.b	$DB					; 
		dc.b	$E7					; 
		dc.b	$CC					; 
		dc.b	$99					; 
		dc.b	1
		dc.b	$2D					; -
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$55					; U
		dc.b	$4A					; J
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$8F					; 
		dc.b	$D4					; 
		dc.b	$B4					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$2E					; .
		dc.b	$27					; '
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$C5					; 
		dc.b	$CD					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$B1					; 
		dc.b	$AB					; 
		dc.b	$86					; 
		dc.b	$3E					; >
		dc.b	$4A					; J
		dc.b	$91					; 
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$38					; 8
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$D2					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$31					; 1
		dc.b	$3A					; :
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$A7					; 
		dc.b	$83					; 
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$26					; &
		dc.b	$3A					; :
		dc.b	$8A					; 
		dc.b	$B0					; 
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$AB					; 
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$5A					; Z
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$94					; 
		dc.b	$61					; a
		dc.b	$3E					; >
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$B8					; 
		dc.b	$5A					; Z
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$79					; y
		dc.b	$C5					; 
		dc.b	$B3					; 
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$42					; B
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$BF					; 
		dc.b	$A3					; 
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$46					; F
		dc.b	$60					; `
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$DD					; 
		dc.b	$87					; 
		dc.b	$5A					; Z
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$3C					; <
		dc.b	$95					; 
		dc.b	$B4					; 
		dc.b	$98					; 
		dc.b	$B8					; 
		dc.b	$86					; 
		dc.b	$37					; 7
		dc.b	$26					; &
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$78					; x
		dc.b	$45					; E
		dc.b	$3F					; ?
		dc.b	$84					; 
		dc.b	$A9					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$B6					; 
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$43					; C
		dc.b	$65					; e
		dc.b	$CF					; 
		dc.b	$CD					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$3B					; ;
		dc.b	$31					; 1
		dc.b	$61					; a
		dc.b	$A6					; 
		dc.b	$B0					; 
		dc.b	$B8					; 
		dc.b	$6C					; l
		dc.b	$1C
		dc.b	$48					; H
		dc.b	$C5					; 
		dc.b	$DD					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$8B					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$C5					; 
		dc.b	$B0					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$1C
		dc.b	$49					; I
		dc.b	$95					; 
		dc.b	$C9					; 
		dc.b	$CD					; 
		dc.b	$88					; 
		dc.b	$57					; W
		dc.b	$3F					; ?
		dc.b	$3A					; :
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$CC					; 
		dc.b	$EB					; 
		dc.b	$92					; 
		dc.b	$26					; &
		dc.b	$37					; 7
		dc.b	$4C					; L
		dc.b	$7D					; }
		dc.b	$BF					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$4D					; M
		dc.b	$3C					; <
		dc.b	$3E					; >
		dc.b	$91					; 
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$42					; B
		dc.b	$2E					; .
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$A7					; 
		dc.b	$BD					; 
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$AB					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$3C					; <
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$A4					; 
		dc.b	$D2					; 
		dc.b	$5D					; ]
		dc.b	$1B
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$CF					; 
		dc.b	$B1					; 
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$A7					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$46					; F
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$34					; 4
		dc.b	$1F
		dc.b	$46					; F
		dc.b	$8E					; 
		dc.b	$E2					; 
		dc.b	$DB					; 
		dc.b	$8D					; 
		dc.b	$1E
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$68					; h
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$45					; E
		dc.b	$30					; 0
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$B6					; 
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$AA					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$88					; 
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$55					; U
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$D3					; 
		dc.b	$DE					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$7F					; 
		dc.b	$4D					; M
		dc.b	$68					; h
		dc.b	$8F					; 
		dc.b	$A2					; 
		dc.b	$E0					; 
		dc.b	$7C					; |
		dc.b	1
		dc.b	$23					; #
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$C8					; 
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$4D					; M
		dc.b	$3B					; ;
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$9F					; 
		dc.b	$BD					; 
		dc.b	$8D					; 
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$84					; 
		dc.b	$52					; R
		dc.b	$37					; 7
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$A3					; 
		dc.b	$D4					; 
		dc.b	$BE					; 
		dc.b	$7A					; z
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$83					; 
		dc.b	$31					; 1
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$61					; a
		dc.b	$8B					; 
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$68					; h
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$59					; Y
		dc.b	$3B					; ;
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$56					; V
		dc.b	$76					; v
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$C2					; 
		dc.b	$A6					; 
		dc.b	$35					; 5
		dc.b	$B
		dc.b	$56					; V
		dc.b	$95					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$B4					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$50					; P
		dc.b	$19
		dc.b	$53					; S
		dc.b	$C2					; 
		dc.b	$C6					; 
		dc.b	$AA					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$41					; A
		dc.b	$3F					; ?
		dc.b	$76					; v
		dc.b	$B3					; 
		dc.b	$DE					; 
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$8B					; 
		dc.b	$D2					; 
		dc.b	$92					; 
		dc.b	$37					; 7
		dc.b	$17
		dc.b	$46					; F
		dc.b	$9B					; 
		dc.b	$D0					; 
		dc.b	$BE					; 
		dc.b	$88					; 
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$34					; 4
		dc.b	$71					; q
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$A0					; 
		dc.b	$C6					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$BB					; 
		dc.b	$BA					; 
		dc.b	$C5					; 
		dc.b	$8B					; 
		dc.b	$23					; #
		dc.b	$35					; 5
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$A0					; 
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$83					; 
		dc.b	$C6					; 
		dc.b	$CC					; 
		dc.b	$7D					; }
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$43					; C
		dc.b	$BD					; 
		dc.b	$B4					; 
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$5C					; \
		dc.b	$4C					; L
		dc.b	$D3					; 
		dc.b	$AE					; 
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$AA					; 
		dc.b	$D9					; 
		dc.b	$BD					; 
		dc.b	$9C					; 
		dc.b	$60					; `
		dc.b	$3A					; :
		dc.b	$43					; C
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$B0					; 
		dc.b	$83					; 
		dc.b	$56					; V
		dc.b	$5A					; Z
		dc.b	$50					; P
		dc.b	$81					; 
		dc.b	$C1					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$6E					; n
		dc.b	$3E					; >
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$35					; 5
		dc.b	$5D					; ]
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$3A					; :
		dc.b	$3E					; >
		dc.b	$A6					; 
		dc.b	$7C					; |
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$88					; 
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$64					; d
		dc.b	$38					; 8
		dc.b	$49					; I
		dc.b	$68					; h
		dc.b	$D7					; 
		dc.b	$F3					; 
		dc.b	$C6					; 
		dc.b	$73					; s
		dc.b	$E
		dc.b	$18
		dc.b	$5E					; ^
		dc.b	$A6					; 
		dc.b	$BD					; 
		dc.b	$B4					; 
		dc.b	$57					; W
		dc.b	$2E					; .
		dc.b	$7A					; z
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$BD					; 
		dc.b	$8F					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$26					; &
		dc.b	$94					; 
		dc.b	$FE					; 
		dc.b	$C9					; 
		dc.b	$B6					; 
		dc.b	$3A					; :
		dc.b	1
		dc.b	$38					; 8
		dc.b	$79					; y
		dc.b	$CD					; 
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$6A					; j
		dc.b	$17
		dc.b	6
		dc.b	$55					; U
		dc.b	$D3					; 
		dc.b	$FE					; 
		dc.b	$B3					; 
		dc.b	$9F					; 
		dc.b	$2E					; .
		dc.b	1
		dc.b	$34					; 4
		dc.b	$73					; s
		dc.b	$BB					; 
		dc.b	$C1					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$49					; I
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$A3					; 
		dc.b	$42					; B
		dc.b	$61					; a
		dc.b	$3C					; <
		dc.b	6
		dc.b	$6B					; k
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$A4					; 
		dc.b	$A9					; 
		dc.b	$4F					; O
		dc.b	$27					; '
		dc.b	$3F					; ?
		dc.b	$84					; 
		dc.b	$C9					; 
		dc.b	$C6					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$B0					; 
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$9F					; 
		dc.b	$BA					; 
		dc.b	$86					; 
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$6F					; o
		dc.b	$59					; Y
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$B7					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$C4					; 
		dc.b	$7C					; |
		dc.b	$4A					; J
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$A9					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$6B					; k
		dc.b	$B3					; 
		dc.b	$AE					; 
		dc.b	$75					; u
		dc.b	$3A					; :
		dc.b	$43					; C
		dc.b	$A7					; 
		dc.b	$C5					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$3E					; >
		dc.b	$6B					; k
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$46					; F
		dc.b	$3A					; :
		dc.b	$71					; q
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$CB					; 
		dc.b	$9C					; 
		dc.b	$4A					; J
		dc.b	$20
		dc.b	$35					; 5
		dc.b	$AE					; 
		dc.b	$CC					; 
		dc.b	$AE					; 
		dc.b	$83					; 
		dc.b	$35					; 5
		dc.b	$34					; 4
		dc.b	$9F					; 
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$C6					; 
		dc.b	$AE					; 
		dc.b	$65					; e
		dc.b	$43					; C
		dc.b	$35					; 5
		dc.b	$5C					; \
		dc.b	$9D					; 
		dc.b	$B7					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$43					; C
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$C1					; 
		dc.b	$CD					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$56					; V
		dc.b	$45					; E
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$A9					; 
		dc.b	$A4					; 
		dc.b	$6A					; j
		dc.b	$3B					; ;
		dc.b	$2D					; -
		dc.b	$59					; Y
		dc.b	$B0					; 
		dc.b	$9B					; 
		dc.b	$C4					; 
		dc.b	$B4					; 
		dc.b	$55					; U
		dc.b	$45					; E
		dc.b	$22					; "
		dc.b	$73					; s
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$52					; R
		dc.b	$A3					; 
		dc.b	$E7					; 
		dc.b	$CC					; 
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$CB					; 
		dc.b	$BD					; 
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$A4					; 
		dc.b	$84					; 
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$48					; H
		dc.b	$42					; B
		dc.b	$B6					; 
		dc.b	$E4					; 
		dc.b	$DB					; 
		dc.b	$A4					; 
		dc.b	$19
		dc.b	1
		dc.b	$34					; 4
		dc.b	$AA					; 
		dc.b	$FC					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$C5					; 
		dc.b	$8D					; 
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$95					; 
		dc.b	$D3					; 
		dc.b	$8E					; 
		dc.b	$6B					; k
		dc.b	$30					; 0
		dc.b	$45					; E
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$B1					; 
		dc.b	$8F					; 
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$8F					; 
		dc.b	$B7					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$5D					; ]
		dc.b	$1B
		dc.b	$4D					; M
		dc.b	$9D					; 
		dc.b	$E8					; 
		dc.b	$BA					; 
		dc.b	$3C					; <
		dc.b	$4A					; J
		dc.b	$46					; F
		dc.b	$6F					; o
		dc.b	$B3					; 
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$C8					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$59					; Y
		dc.b	$4A					; J
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$52					; R
		dc.b	$17
		dc.b	$42					; B
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$3E					; >
		dc.b	$4D					; M
		dc.b	$76					; v
		dc.b	$AB					; 
		dc.b	$DB					; 
		dc.b	$98					; 
		dc.b	$6C					; l
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$9F					; 
		dc.b	$A9					; 
		dc.b	$5A					; Z
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$48					; H
		dc.b	$BB					; 
		dc.b	$CF					; 
		dc.b	$A2					; 
		dc.b	$7F					; 
		dc.b	$3F					; ?
		dc.b	$14
		dc.b	$61					; a
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$B8					; 
		dc.b	$A6					; 
		dc.b	$4C					; L
		dc.b	$52					; R
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$4A					; J
		dc.b	$7F					; 
		dc.b	$D4					; 
		dc.b	$B4					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$79					; y
		dc.b	$B8					; 
		dc.b	$DD					; 
		dc.b	$CF					; 
		dc.b	$52					; R
		dc.b	$10
		dc.b	$3F					; ?
		dc.b	$78					; x
		dc.b	$BD					; 
		dc.b	$9D					; 
		dc.b	$6B					; k
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$BB					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$49					; I
		dc.b	$3F					; ?
		dc.b	$57					; W
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$52					; R
		dc.b	$31					; 1
		dc.b	$68					; h
		dc.b	$8F					; 
		dc.b	$C8					; 
		dc.b	$AA					; 
		dc.b	$7D					; }
		dc.b	$50					; P
		dc.b	$50					; P
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$C8					; 
		dc.b	$AA					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$BA					; 
		dc.b	$BA					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$3E					; >
		dc.b	$38					; 8
		dc.b	$5A					; Z
		dc.b	$6B					; k
		dc.b	$B1					; 
		dc.b	$CD					; 
		dc.b	$AD					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$6E					; n
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$A0					; 
		dc.b	$7A					; z
		dc.b	$9C					; 
		dc.b	$D2					; 
		dc.b	$80					; 
		dc.b	$17
		dc.b	$3E					; >
		dc.b	$78					; x
		dc.b	$AA					; 
		dc.b	$BD					; 
		dc.b	$9C					; 
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$9C					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$34					; 4
		dc.b	$5A					; Z
		dc.b	$CF					; 
		dc.b	$B1					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$9D					; 
		dc.b	$E2					; 
		dc.b	$D2					; 
		dc.b	$48					; H
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$55					; U
		dc.b	$B0					; 
		dc.b	$C8					; 
		dc.b	$A9					; 
		dc.b	$6E					; n
		dc.b	$3F					; ?
		dc.b	$75					; u
		dc.b	$A6					; 
		dc.b	$B6					; 
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$D0					; 
		dc.b	$9F					; 
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$65					; e
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$2D					; -
		dc.b	$12
		dc.b	$3E					; >
		dc.b	$AB					; 
		dc.b	$D6					; 
		dc.b	$AA					; 
		dc.b	$55					; U
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$79					; y
		dc.b	$B1					; 
		dc.b	$D3					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$5A					; Z
		dc.b	$2C					; ,
		dc.b	$4F					; O
		dc.b	$84					; 
		dc.b	$A9					; 
		dc.b	$BB					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$52					; R
		dc.b	$37					; 7
		dc.b	$76					; v
		dc.b	$B0					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$59					; Y
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$AA					; 
		dc.b	$D7					; 
		dc.b	$C9					; 
		dc.b	$AB					; 
		dc.b	$50					; P
		dc.b	$E
		dc.b	$43					; C
		dc.b	$87					; 
		dc.b	$BE					; 
		dc.b	$C6					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$57					; W
		dc.b	$45					; E
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$49					; I
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$9F					; 
		dc.b	$F7					; 
		dc.b	$B7					; 
		dc.b	$56					; V
		dc.b	$46					; F
		dc.b	$2D					; -
		dc.b	$38					; 8
		dc.b	$76					; v
		dc.b	$B4					; 
		dc.b	$CD					; 
		dc.b	$88					; 
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$6F					; o
		dc.b	$AD					; 
		dc.b	$BB					; 
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$5C					; \
		dc.b	$99					; 
		dc.b	$B8					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$43					; C
		dc.b	$4F					; O
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$CB					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$4D					; M
		dc.b	$5D					; ]
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$50					; P
		dc.b	$83					; 
		dc.b	$D6					; 
		dc.b	$A9					; 
		dc.b	$43					; C
		dc.b	$4A					; J
		dc.b	$2C					; ,
		dc.b	$4F					; O
		dc.b	$C2					; 
		dc.b	$B0					; 
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$43					; C
		dc.b	$56					; V
		dc.b	$CF					; 
		dc.b	$BB					; 
		dc.b	$BA					; 
		dc.b	$7A					; z
		dc.b	$20
		dc.b	$55					; U
		dc.b	$B3					; 
		dc.b	$DA					; 
		dc.b	$84					; 
		dc.b	$27					; '
		dc.b	$3C					; <
		dc.b	$57					; W
		dc.b	$C1					; 
		dc.b	$E0					; 
		dc.b	$6C					; l
		dc.b	$4F					; O
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$5D					; ]
		dc.b	$86					; 
		dc.b	$B4					; 
		dc.b	$C2					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$43					; C
		dc.b	$2A					; *
		dc.b	$AD					; 
		dc.b	$DA					; 
		dc.b	$68					; h
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$45					; E
		dc.b	$A2					; 
		dc.b	$BE					; 
		dc.b	$A6					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$AA					; 
		dc.b	$79					; y
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$60					; `
		dc.b	$4C					; L
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$3F					; ?
		dc.b	$63					; c
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$86					; 
		dc.b	$3F					; ?
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$63					; c
		dc.b	$26					; &
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$6F					; o
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$5D					; ]
		dc.b	$A0					; 
		dc.b	$F6					; 
		dc.b	$D6					; 
		dc.b	$5E					; ^
		dc.b	$4D					; M
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$6F					; o
		dc.b	$52					; R
		dc.b	$50					; P
		dc.b	$72					; r
		dc.b	$A3					; 
		dc.b	$BF					; 
		dc.b	$98					; 
		dc.b	$5E					; ^
		dc.b	$38					; 8
		dc.b	$38					; 8
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$AD					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$B4					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$46					; F
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$C5					; 
		dc.b	$8B					; 
		dc.b	$45					; E
		dc.b	$5A					; Z
		dc.b	$92					; 
		dc.b	$AA					; 
		dc.b	$AD					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$49					; I
		dc.b	$3F					; ?
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$C2					; 
		dc.b	$7F					; 
		dc.b	$4F					; O
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$B1					; 
		dc.b	$B8					; 
		dc.b	$83					; 
		dc.b	$55					; U
		dc.b	$59					; Y
		dc.b	$3F					; ?
		dc.b	$56					; V
		dc.b	$CC					; 
		dc.b	$AE					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$A3					; 
		dc.b	$56					; V
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$B3					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$7C					; |
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$80					; 
		dc.b	$A0					; 
		dc.b	$9C					; 
		dc.b	$9C					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$C4					; 
		dc.b	$8B					; 
		dc.b	$5E					; ^
		dc.b	$7D					; }
		dc.b	$67					; g
		dc.b	$8D					; 
		dc.b	$C4					; 
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$1B
		dc.b	$45					; E
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$E2					; 
		dc.b	$E9					; 
		dc.b	$A3					; 
		dc.b	$53					; S
		dc.b	$18
		dc.b	$2C					; ,
		dc.b	$57					; W
		dc.b	$AA					; 
		dc.b	$D4					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$BA					; 
		dc.b	$A3					; 
		dc.b	$53					; S
		dc.b	$3E					; >
		dc.b	$3C					; <
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$45					; E
		dc.b	$4C					; L
		dc.b	$8F					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$AD					; 
		dc.b	$B4					; 
		dc.b	$98					; 
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$4A					; J
		dc.b	$71					; q
		dc.b	$9D					; 
		dc.b	$5A					; Z
		dc.b	$9F					; 
		dc.b	$AA					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$AE					; 
		dc.b	$D7					; 
		dc.b	$8E					; 
		dc.b	$6B					; k
		dc.b	$5C					; \
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$9B					; 
		dc.b	$D2					; 
		dc.b	$A3					; 
		dc.b	$20
		dc.b	$27					; '
		dc.b	$64					; d
		dc.b	$CC					; 
		dc.b	$E9					; 
		dc.b	$B3					; 
		dc.b	$67					; g
		dc.b	$26					; &
		dc.b	$55					; U
		dc.b	$88					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$B6					; 
		dc.b	$AB					; 
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$43					; C
		dc.b	$25					; %
		dc.b	$49					; I
		dc.b	$B0					; 
		dc.b	$AE					; 
		dc.b	$5A					; Z
		dc.b	$35					; 5
		dc.b	$6B					; k
		dc.b	$B0					; 
		dc.b	$DB					; 
		dc.b	$8F					; 
		dc.b	$49					; I
		dc.b	$38					; 8
		dc.b	$3F					; ?
		dc.b	$A4					; 
		dc.b	$D7					; 
		dc.b	$C9					; 
		dc.b	$6E					; n
		dc.b	$31					; 1
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$50					; P
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$D4					; 
		dc.b	$D4					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$55					; U
		dc.b	$3A					; :
		dc.b	$9C					; 
		dc.b	$CF					; 
		dc.b	$9B					; 
		dc.b	$B7					; 
		dc.b	$A6					; 
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$59					; Y
		dc.b	$4D					; M
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$5D					; ]
		dc.b	$14
		dc.b	$1F
		dc.b	$6F					; o
		dc.b	$A0					; 
		dc.b	$D3					; 
		dc.b	$BB					; 
		dc.b	$56					; V
		dc.b	$14
		dc.b	$3B					; ;
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$C4					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$31					; 1
		dc.b	$3A					; :
		dc.b	$78					; x
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$42					; B
		dc.b	$1F
		dc.b	$41					; A
		dc.b	$83					; 
		dc.b	$CB					; 
		dc.b	$C6					; 
		dc.b	$8A					; 
		dc.b	$33					; 3
		dc.b	$3F					; ?
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$B6					; 
		dc.b	$D0					; 
		dc.b	$76					; v
		dc.b	$A
		dc.b	$3E					; >
		dc.b	$52					; R
		dc.b	$8E					; 
		dc.b	$D3					; 
		dc.b	$D6					; 
		dc.b	$B4					; 
		dc.b	$79					; y
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$67					; g
		dc.b	$C6					; 
		dc.b	$A9					; 
		dc.b	$B0					; 
		dc.b	$AA					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$43					; C
		dc.b	$64					; d
		dc.b	$C1					; 
		dc.b	$CF					; 
		dc.b	$F0					; 
		dc.b	$9C					; 
		dc.b	$A
		dc.b	1
		dc.b	$46					; F
		dc.b	$87					; 
		dc.b	$F5					; 
		dc.b	$FE					; 
		dc.b	$A9					; 
		dc.b	$55					; U
		dc.b	2
		dc.b	$19
		dc.b	$5D					; ]
		dc.b	$B1					; 
		dc.b	$DA					; 
		dc.b	$AA					; 
		dc.b	$52					; R
		dc.b	$14
		dc.b	$3F					; ?
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$7A					; z
		dc.b	$29					; )
		dc.b	$43					; C
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$BA					; 
		dc.b	$C9					; 
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$29					; )
		dc.b	$6E					; n
		dc.b	$E9					; 
		dc.b	$9C					; 
		dc.b	$6A					; j
		dc.b	$56					; V
		dc.b	$33					; 3
		dc.b	$6F					; o
		dc.b	$B6					; 
		dc.b	$A9					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$37					; 7
		dc.b	$9C					; 
		dc.b	$BD					; 
		dc.b	$B8					; 
		dc.b	$AB					; 
		dc.b	$5D					; ]
		dc.b	$1F
		dc.b	$34					; 4
		dc.b	$6F					; o
		dc.b	$D6					; 
		dc.b	$F9					; 
		dc.b	$87					; 
		dc.b	$4D					; M
		dc.b	4
		dc.b	$27					; '
		dc.b	$88					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$9F					; 
		dc.b	$52					; R
		dc.b	$2E					; .
		dc.b	$38					; 8
		dc.b	$6E					; n
		dc.b	$CB					; 
		dc.b	$DA					; 
		dc.b	$CF					; 
		dc.b	$87					; 
		dc.b	$18
		dc.b	$26					; &
		dc.b	$63					; c
		dc.b	$A9					; 
		dc.b	$E5					; 
		dc.b	$96					; 
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$C6					; 
		dc.b	$C2					; 
		dc.b	$57					; W
		dc.b	$56					; V
		dc.b	$31					; 1
		dc.b	$56					; V
		dc.b	$C8					; 
		dc.b	$B1					; 
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$31					; 1
		dc.b	$12
		dc.b	$46					; F
		dc.b	$A3					; 
		dc.b	$B6					; 
		dc.b	$C6					; 
		dc.b	$DE					; 
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$EE					; 
		dc.b	$CD					; 
		dc.b	$65					; e
		dc.b	$4C					; L
		dc.b	$4C					; L
		dc.b	$52					; R
		dc.b	$91					; 
		dc.b	$BE					; 
		dc.b	$CB					; 
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$34					; 4
		dc.b	$8B					; 
		dc.b	$C5					; 
		dc.b	$64					; d
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$A2					; 
		dc.b	$CF					; 
		dc.b	$C1					; 
		dc.b	$75					; u
		dc.b	$34					; 4
		dc.b	$55					; U
		dc.b	$37					; 7
		dc.b	$5D					; ]
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$8B					; 
		dc.b	$59					; Y
		dc.b	1
		dc.b	$3C					; <
		dc.b	$C6					; 
		dc.b	$E5					; 
		dc.b	$A4					; 
		dc.b	$64					; d
		dc.b	$34					; 4
		dc.b	$4C					; L
		dc.b	$5E					; ^
		dc.b	$B7					; 
		dc.b	$C2					; 
		dc.b	$6F					; o
		dc.b	$4A					; J
		dc.b	$48					; H
		dc.b	$37					; 7
		dc.b	$72					; r
		dc.b	$CF					; 
		dc.b	$C2					; 
		dc.b	$7A					; z
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$96					; 
		dc.b	$BB					; 
		dc.b	$88					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$20
		dc.b	$99					; 
		dc.b	$EF					; 
		dc.b	$9C					; 
		dc.b	$63					; c
		dc.b	$46					; F
		dc.b	$E
		dc.b	$95					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$B0					; 
		dc.b	$D9					; 
		dc.b	$84					; 
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$48					; H
		dc.b	$7F					; 
		dc.b	$CC					; 
		dc.b	$9F					; 
		dc.b	$6B					; k
		dc.b	$41					; A
		dc.b	$29					; )
		dc.b	$67					; g
		dc.b	$CF					; 
		dc.b	$E2					; 
		dc.b	$A4					; 
		dc.b	$64					; d
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$37					; 7
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$BE					; 
		dc.b	$C5					; 
		dc.b	$84					; 
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$6E					; n
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$C2					; 
		dc.b	$A2					; 
		dc.b	$53					; S
		dc.b	$2D					; -
		dc.b	$57					; W
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$CD					; 
		dc.b	$A0					; 
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$37					; 7
		dc.b	$53					; S
		dc.b	$C1					; 
		dc.b	$F2					; 
		dc.b	$DA					; 
		dc.b	$76					; v
		dc.b	$29					; )
		dc.b	$1F
		dc.b	$59					; Y
		dc.b	$AE					; 
		dc.b	$F2					; 
		dc.b	$C9					; 
		dc.b	$64					; d
		dc.b	$20
		dc.b	$3A					; :
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$D0					; 
		dc.b	$B1					; 
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$2A					; *
		dc.b	$4F					; O
		dc.b	$B8					; 
		dc.b	$B7					; 
		dc.b	$AD					; 
		dc.b	$9D					; 
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$65					; e
		dc.b	$92					; 
		dc.b	$BA					; 
		dc.b	$B8					; 
		dc.b	$7D					; }
		dc.b	$5D					; ]
		dc.b	$4A					; J
		dc.b	$3B					; ;
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$AA					; 
		dc.b	$E0					; 
		dc.b	$73					; s
		dc.b	$3B					; ;
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$59					; Y
		dc.b	$2E					; .
		dc.b	$53					; S
		dc.b	$80					; 
		dc.b	$BF					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$B1					; 
		dc.b	$9F					; 
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$71					; q
		dc.b	$9F					; 
		dc.b	$C1					; 
		dc.b	$84					; 
		dc.b	$33					; 3
		dc.b	$55					; U
		dc.b	$C5					; 
		dc.b	$EC					; 
		dc.b	$A6					; 
		dc.b	$52					; R
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$2E					; .
		dc.b	$4D					; M
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$D0					; 
		dc.b	$9F					; 
		dc.b	$2E					; .
		dc.b	6
		dc.b	$1F
		dc.b	$59					; Y
		dc.b	$B0					; 
		dc.b	$EC					; 
		dc.b	$78					; x
		dc.b	$55					; U
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$61					; a
		dc.b	$3F					; ?
		dc.b	$50					; P
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$57					; W
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$C8					; 
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$50					; P
		dc.b	$8B					; 
		dc.b	$C1					; 
		dc.b	$B0					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$AB					; 
		dc.b	$B0					; 
		dc.b	$83					; 
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$46					; F
		dc.b	$6F					; o
		dc.b	$BE					; 
		dc.b	$B6					; 
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$38					; 8
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$8F					; 
		dc.b	$6E					; n
		dc.b	$2D					; -
		dc.b	$6F					; o
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$5D					; ]
		dc.b	$2E					; .
		dc.b	$38					; 8
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$C8					; 
		dc.b	$94					; 
		dc.b	$3C					; <
		dc.b	$30					; 0
		dc.b	$73					; s
		dc.b	$B6					; 
		dc.b	$B3					; 
		dc.b	$98					; 
		dc.b	$41					; A
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$8E					; 
		dc.b	$D7					; 
		dc.b	$E9					; 
		dc.b	$88					; 
		dc.b	$38					; 8
		dc.b	$2C					; ,
		dc.b	$37					; 7
		dc.b	$81					; 
		dc.b	$C9					; 
		dc.b	$E0					; 
		dc.b	$E4					; 
		dc.b	$7F					; 
		dc.b	$35					; 5
		dc.b	$29					; )
		dc.b	$49					; I
		dc.b	$BE					; 
		dc.b	$DB					; 
		dc.b	$EB					; 
		dc.b	$6A					; j
		dc.b	$17
		dc.b	$61					; a
		dc.b	$A6					; 
		dc.b	$C9					; 
		dc.b	$C4					; 
		dc.b	$73					; s
		dc.b	$41					; A
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$4A					; J
		dc.b	$99					; 
		dc.b	$C9					; 
		dc.b	$A6					; 
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$57					; W
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$BB					; 
		dc.b	$A4					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$35					; 5
		dc.b	$3F					; ?
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$53					; S
		dc.b	$5A					; Z
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$2D					; -
		dc.b	$98					; 
		dc.b	$B1					; 
		dc.b	$42					; B
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$4F					; O
		dc.b	$30					; 0
		dc.b	$5D					; ]
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$BB					; 
		dc.b	$A6					; 
		dc.b	$59					; Y
		dc.b	$72					; r
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$48					; H
		dc.b	$63					; c
		dc.b	$C4					; 
		dc.b	$B3					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$4F					; O
		dc.b	$3E					; >
		dc.b	$55					; U
		dc.b	$8E					; 
		dc.b	$BB					; 
		dc.b	$92					; 
		dc.b	$5A					; Z
		dc.b	$6C					; l
		dc.b	$27					; '
		dc.b	$2A					; *
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$71					; q
		dc.b	$4A					; J
		dc.b	$53					; S
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$C1					; 
		dc.b	$DD					; 
		dc.b	$72					; r
		dc.b	$4A					; J
		dc.b	$6B					; k
		dc.b	$3F					; ?
		dc.b	$81					; 
		dc.b	$C6					; 
		dc.b	$CC					; 
		dc.b	$B1					; 
		dc.b	$3A					; :
		dc.b	1
		dc.b	$A
		dc.b	$84					; 
		dc.b	$DB					; 
		dc.b	$CD					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$3A					; :
		dc.b	$2A					; *
		dc.b	$68					; h
		dc.b	$79					; y
		dc.b	$CF					; 
		dc.b	$FE					; 
		dc.b	$C4					; 
		dc.b	$6F					; o
		dc.b	$3A					; :
		dc.b	$43					; C
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$B0					; 
		dc.b	$CB					; 
		dc.b	$98					; 
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$57					; W
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$AA					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$59					; Y
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$9C					; 
		dc.b	$C1					; 
		dc.b	$95					; 
		dc.b	$49					; I
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$3B					; ;
		dc.b	$57					; W
		dc.b	$AE					; 
		dc.b	$81					; 
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$48					; H
		dc.b	$34					; 4
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$95					; 
		dc.b	$CB					; 
		dc.b	$A4					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$AE					; 
		dc.b	$7C					; |
		dc.b	$50					; P
		dc.b	$4C					; L
		dc.b	$49					; I
		dc.b	$88					; 
		dc.b	$C6					; 
		dc.b	$B4					; 
		dc.b	$5E					; ^
		dc.b	$45					; E
		dc.b	$31					; 1
		dc.b	$63					; c
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$A0					; 
		dc.b	$B3					; 
		dc.b	$A6					; 
		dc.b	$73					; s
		dc.b	$34					; 4
		dc.b	$29					; )
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$33					; 3
		dc.b	$61					; a
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$B3					; 
		dc.b	$C9					; 
		dc.b	$38					; 8
		dc.b	$22					; "
		dc.b	$50					; P
		dc.b	$79					; y
		dc.b	$B6					; 
		dc.b	$B8					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$45					; E
		dc.b	$3B					; ;
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$29					; )
		dc.b	$38					; 8
		dc.b	$6E					; n
		dc.b	$A7					; 
		dc.b	$E1					; 
		dc.b	$D0					; 
		dc.b	$60					; `
		dc.b	$1C
		dc.b	$38					; 8
		dc.b	$61					; a
		dc.b	$B4					; 
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$46					; F
		dc.b	$5D					; ]
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$BA					; 
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$55					; U
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$D3					; 
		dc.b	$83					; 
		dc.b	$65					; e
		dc.b	$49					; I
		dc.b	$2C					; ,
		dc.b	$48					; H
		dc.b	$8D					; 
		dc.b	$BB					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$46					; F
		dc.b	$23					; #
		dc.b	$5A					; Z
		dc.b	$C5					; 
		dc.b	$C2					; 
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$AE					; 
		dc.b	$E8					; 
		dc.b	$AE					; 
		dc.b	$56					; V
		dc.b	$50					; P
		dc.b	$18
		dc.b	$53					; S
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$B6					; 
		dc.b	$BF					; 
		dc.b	$9B					; 
		dc.b	$5E					; ^
		dc.b	$15
		dc.b	$55					; U
		dc.b	$96					; 
		dc.b	$B1					; 
		dc.b	$DB					; 
		dc.b	$83					; 
		dc.b	$4C					; L
		dc.b	$3C					; <
		dc.b	$34					; 4
		dc.b	$A0					; 
		dc.b	$D3					; 
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$43					; C
		dc.b	$5A					; Z
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$5D					; ]
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$30					; 0
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$4C					; L
		dc.b	$5C					; \
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$9C					; 
		dc.b	$C5					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$B0					; 
		dc.b	$96					; 
		dc.b	$67					; g
		dc.b	$50					; P
		dc.b	$4C					; L
		dc.b	$6C					; l
		dc.b	$BF					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$B7					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$4F					; O
		dc.b	$37					; 7
		dc.b	$50					; P
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$B4					; 
		dc.b	$BA					; 
		dc.b	$80					; 
		dc.b	$2E					; .
		dc.b	$20
		dc.b	$49					; I
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$C9					; 
		dc.b	$BD					; 
		dc.b	$68					; h
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$C6					; 
		dc.b	$CC					; 
		dc.b	$95					; 
		dc.b	$22					; "
		dc.b	$42					; B
		dc.b	$84					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$43					; C
		dc.b	$3C					; <
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$9F					; 
		dc.b	$C5					; 
		dc.b	$B8					; 
		dc.b	$6B					; k
		dc.b	$29					; )
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$3A					; :
		dc.b	$4C					; L
		dc.b	$4D					; M
		dc.b	$A7					; 
		dc.b	$D0					; 
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$41					; A
		dc.b	$20
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$B1					; 
		dc.b	$C5					; 
		dc.b	$84					; 
		dc.b	$3A					; :
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$A3					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$9D					; 
		dc.b	$AD					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$4A					; J
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$A9					; 
		dc.b	$B0					; 
		dc.b	$71					; q
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$A9					; 
		dc.b	$BD					; 
		dc.b	$AA					; 
		dc.b	$63					; c
		dc.b	$E
		dc.b	$3A					; :
		dc.b	$6F					; o
		dc.b	$B6					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$6B					; k
		dc.b	$50					; P
		dc.b	$53					; S
		dc.b	$81					; 
		dc.b	$B0					; 
		dc.b	$BF					; 
		dc.b	$94					; 
		dc.b	$3C					; <
		dc.b	$26					; &
		dc.b	$3B					; ;
		dc.b	$8B					; 
		dc.b	$F9					; 
		dc.b	$EB					; 
		dc.b	$72					; r
		dc.b	$17
		dc.b	$18
		dc.b	$41					; A
		dc.b	$96					; 
		dc.b	$D3					; 
		dc.b	$B6					; 
		dc.b	$7A					; z
		dc.b	$1E
		dc.b	$57					; W
		dc.b	$9F					; 
		dc.b	$B3					; 
		dc.b	$A4					; 
		dc.b	$98					; 
		dc.b	$6C					; l
		dc.b	$46					; F
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$92					; 
		dc.b	$B8					; 
		dc.b	$96					; 
		dc.b	$6C					; l
		dc.b	$4F					; O
		dc.b	$2E					; .
		dc.b	$76					; v
		dc.b	$BF					; 
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$48					; H
		dc.b	$3C					; <
		dc.b	$53					; S
		dc.b	$7F					; 
		dc.b	$BB					; 
		dc.b	$AB					; 
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$AE					; 
		dc.b	$94					; 
		dc.b	$59					; Y
		dc.b	$34					; 4
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$C1					; 
		dc.b	$9B					; 
		dc.b	$2D					; -
		dc.b	$35					; 5
		dc.b	$61					; a
		dc.b	$C2					; 
		dc.b	$E0					; 
		dc.b	$AA					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$60					; `
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$B8					; 
		dc.b	$8F					; 
		dc.b	$18
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$D7					; 
		dc.b	$B0					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$2D					; -
		dc.b	$7A					; z
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$5A					; Z
		dc.b	$60					; `
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$48					; H
		dc.b	$73					; s
		dc.b	$9F					; 
		dc.b	$8E					; 
		dc.b	$6C					; l
		dc.b	$43					; C
		dc.b	$18
		dc.b	$5E					; ^
		dc.b	$B1					; 
		dc.b	$D7					; 
		dc.b	$F2					; 
		dc.b	$BA					; 
		dc.b	$4A					; J
		dc.b	$11
		dc.b	$4C					; L
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$81					; 
		dc.b	$B8					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$4D					; M
		dc.b	$1C
		dc.b	$3B					; ;
		dc.b	$98					; 
		dc.b	$DE					; 
		dc.b	$CD					; 
		dc.b	$6B					; k
		dc.b	6
		dc.b	3
		dc.b	$3E					; >
		dc.b	$98					; 
		dc.b	$E0					; 
		dc.b	$F2					; 
		dc.b	$B1					; 
		dc.b	$73					; s
		dc.b	$27					; '
		dc.b	$18
		dc.b	$41					; A
		dc.b	$6F					; o
		dc.b	$C4					; 
		dc.b	$C4					; 
		dc.b	$8B					; 
		dc.b	$4C					; L
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$67					; g
		dc.b	$33					; 3
		dc.b	$59					; Y
		dc.b	$BF					; 
		dc.b	$C8					; 
		dc.b	$C2					; 
		dc.b	$B4					; 
		dc.b	$53					; S
		dc.b	$2C					; ,
		dc.b	$5D					; ]
		dc.b	$8D					; 
		dc.b	$B3					; 
		dc.b	$BB					; 
		dc.b	$9C					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$CB					; 
		dc.b	$C5					; 
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$A0					; 
		dc.b	$AB					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$63					; c
		dc.b	$49					; I
		dc.b	$37					; 7
		dc.b	$50					; P
		dc.b	$CB					; 
		dc.b	$EF					; 
		dc.b	$B1					; 
		dc.b	$8D					; 
		dc.b	$3C					; <
		dc.b	$2E					; .
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$C4					; 
		dc.b	$B8					; 
		dc.b	$71					; q
		dc.b	$33					; 3
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$B0					; 
		dc.b	$BA					; 
		dc.b	$AB					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$80					; 
PCM_Punch:	dc.b	$9D					; 		; DATA XREF: ROM:0001B08Co
		dc.b	$BE					; 
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$31					; 1
		dc.b	$A
		dc.b	$6A					; j
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$99					; 
		dc.b	$88					; 
		dc.b	$4C					; L
		dc.b	$30					; 0
		dc.b	$61					; a
		dc.b	$80					; 
		dc.b	$AA					; 
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$43					; C
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$BD					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$53					; S
		dc.b	$1F
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$83					; 
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$81					; 
		dc.b	$B8					; 
		dc.b	$DD					; 
		dc.b	$63					; c
		dc.b	6
		dc.b	$2C					; ,
		dc.b	$6C					; l
		dc.b	$C9					; 
		dc.b	$FA					; 
		dc.b	$A6					; 
		dc.b	$4C					; L
		dc.b	$2D					; -
		dc.b	$64					; d
		dc.b	$A3					; 
		dc.b	$91					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$B0					; 
		dc.b	$98					; 
		dc.b	$64					; d
		dc.b	6
		dc.b	$B
		dc.b	$3C					; <
		dc.b	$9D					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A3					; 
		dc.b	$14
		dc.b	1
		dc.b	$25					; %
		dc.b	$B8					; 
		dc.b	$EB					; 
		dc.b	$DE					; 
		dc.b	$9B					; 
		dc.b	$33					; 3
		dc.b	7
		dc.b	$61					; a
		dc.b	$AB					; 
		dc.b	$C2					; 
		dc.b	$B7					; 
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$67					; g
		dc.b	$95					; 
		dc.b	$AA					; 
		dc.b	$C5					; 
		dc.b	$B3					; 
		dc.b	$53					; S
		dc.b	$3B					; ;
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$B3					; 
		dc.b	$C5					; 
		dc.b	$B4					; 
		dc.b	$4D					; M
		dc.b	$20
		dc.b	$59					; Y
		dc.b	$8F					; 
		dc.b	$98					; 
		dc.b	$B0					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$43					; C
		dc.b	$45					; E
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$D3					; 
		dc.b	$91					; 
		dc.b	$38					; 8
		dc.b	$3B					; ;
		dc.b	$5A					; Z
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$48					; H
		dc.b	$38					; 8
		dc.b	$A6					; 
		dc.b	$B6					; 
		dc.b	$B4					; 
		dc.b	$8D					; 
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$9F					; 
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$AD					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$2C					; ,
		dc.b	$43					; C
		dc.b	$6C					; l
		dc.b	$BA					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$1F
		dc.b	$20
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$AE					; 
		dc.b	$6B					; k
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$65					; e
		dc.b	$19
		dc.b	$B
		dc.b	$68					; h
		dc.b	$D0					; 
		dc.b	$E0					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$2D					; -
		dc.b	$37					; 7
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$CD					; 
		dc.b	$B3					; 
		dc.b	$3F					; ?
		dc.b	1
		dc.b	$22					; "
		dc.b	$A0					; 
		dc.b	$C6					; 
		dc.b	$E4					; 
		dc.b	$BD					; 
		dc.b	$3F					; ?
		dc.b	$46					; F
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$4F					; O
		dc.b	$1F
		dc.b	$79					; y
		dc.b	$EB					; 
		dc.b	$FC					; 
		dc.b	$B3					; 
		dc.b	$6F					; o
		dc.b	$31					; 1
		dc.b	1
		dc.b	$55					; U
		dc.b	$BE					; 
		dc.b	$BF					; 
		dc.b	$CD					; 
		dc.b	$8F					; 
		dc.b	$2C					; ,
		dc.b	$1C
		dc.b	$75					; u
		dc.b	$D2					; 
		dc.b	$CF					; 
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	4
		dc.b	1
		dc.b	$3C					; <
		dc.b	$92					; 
		dc.b	$D6					; 
		dc.b	$FE					; 
		dc.b	$C2					; 
		dc.b	$2A					; *
		dc.b	1
		dc.b	$2A					; *
		dc.b	$56					; V
		dc.b	$9C					; 
		dc.b	$F3					; 
		dc.b	$AB					; 
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$87					; 
		dc.b	$9F					; 
		dc.b	$71					; q
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$42					; B
		dc.b	$45					; E
		dc.b	$8B					; 
		dc.b	$AB					; 
		dc.b	$C2					; 
		dc.b	$BD					; 
		dc.b	$8E					; 
		dc.b	$57					; W
		dc.b	$1C
		dc.b	$60					; `
		dc.b	$B1					; 
		dc.b	$C1					; 
		dc.b	$A4					; 
		dc.b	$8F					; 
		dc.b	$5A					; Z
		dc.b	$8D					; 
		dc.b	$B0					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$C2					; 
		dc.b	$C1					; 
		dc.b	$92					; 
		dc.b	$45					; E
		dc.b	$34					; 4
		dc.b	$56					; V
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$C2					; 
		dc.b	$D6					; 
		dc.b	$C2					; 
		dc.b	$92					; 
		dc.b	$12
		dc.b	$20
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$A2					; 
		dc.b	$D3					; 
		dc.b	$B6					; 
		dc.b	$79					; y
		dc.b	$2D					; -
		dc.b	$22					; "
		dc.b	$68					; h
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$52					; R
		dc.b	$22					; "
		dc.b	$79					; y
		dc.b	$A3					; 
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$A6					; 
		dc.b	$BB					; 
		dc.b	$BB					; 
		dc.b	$6B					; k
		dc.b	$1E
		dc.b	$12
		dc.b	$4F					; O
		dc.b	$F9					; 
		dc.b	$FC					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$B
		dc.b	$76					; v
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$CB					; 
		dc.b	$AA					; 
		dc.b	$7D					; }
		dc.b	$3A					; :
		dc.b	$45					; E
		dc.b	$9D					; 
		dc.b	$BD					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$57					; W
		dc.b	$4A					; J
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$6C					; l
		dc.b	$AB					; 
		dc.b	$B6					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$42					; B
		dc.b	$8B					; 
		dc.b	$D9					; 
		dc.b	$CC					; 
		dc.b	$8E					; 
		dc.b	$35					; 5
		dc.b	$12
		dc.b	$38					; 8
		dc.b	$92					; 
		dc.b	$CB					; 
		dc.b	$F3					; 
		dc.b	$B7					; 
		dc.b	$49					; I
		dc.b	1
		dc.b	$34					; 4
		dc.b	$A7					; 
		dc.b	$B3					; 
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$5C					; \
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$99					; 
		dc.b	$C9					; 
		dc.b	$9C					; 
		dc.b	$49					; I
		dc.b	$43					; C
		dc.b	1
		dc.b	$4F					; O
		dc.b	$C9					; 
		dc.b	$F6					; 
		dc.b	$F0					; 
		dc.b	$BD					; 
		dc.b	$3C					; <
		dc.b	9
		dc.b	$3A					; :
		dc.b	$78					; x
		dc.b	$E7					; 
		dc.b	$D6					; 
		dc.b	$B3					; 
		dc.b	$6C					; l
		dc.b	$22					; "
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$B7					; 
		dc.b	$C9					; 
		dc.b	$B3					; 
		dc.b	$5C					; \
		dc.b	$2D					; -
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$4C					; L
		dc.b	$23					; #
		dc.b	$92					; 
		dc.b	$D4					; 
		dc.b	$7A					; z
		dc.b	$59					; Y
		dc.b	$3E					; >
		dc.b	$2E					; .
		dc.b	$7F					; 
		dc.b	$B0					; 
		dc.b	$C2					; 
		dc.b	$AB					; 
		dc.b	$76					; v
		dc.b	$2D					; -
		dc.b	$19
		dc.b	$68					; h
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$9C					; 
		dc.b	$67					; g
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$9B					; 
		dc.b	$E1					; 
		dc.b	$CD					; 
		dc.b	$8F					; 
		dc.b	$20
		dc.b	$D
		dc.b	$60					; `
		dc.b	$C8					; 
		dc.b	$D2					; 
		dc.b	$8A					; 
		dc.b	$2D					; -
		dc.b	$11
		dc.b	6
		dc.b	$68					; h
		dc.b	$EB					; 
		dc.b	$FC					; 
		dc.b	$DE					; 
		dc.b	$79					; y
		dc.b	$31					; 1
		dc.b	$1C
		dc.b	$33					; 3
		dc.b	$C8					; 
		dc.b	$DE					; 
		dc.b	$94					; 
		dc.b	$67					; g
		dc.b	$38					; 8
		dc.b	$68					; h
		dc.b	$A7					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$4C					; L
		dc.b	$45					; E
		dc.b	$5E					; ^
		dc.b	$9C					; 
		dc.b	$D2					; 
		dc.b	$C9					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$37					; 7
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$B7					; 
		dc.b	$6E					; n
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$41					; A
		dc.b	$26					; &
		dc.b	$34					; 4
		dc.b	$2E					; .
		dc.b	$BA					; 
		dc.b	$F2					; 
		dc.b	$F3					; 
		dc.b	$B8					; 
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$AD					; 
		dc.b	$AE					; 
		dc.b	$5A					; Z
		dc.b	$20
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$72					; r
		dc.b	$CF					; 
		dc.b	$FC					; 
		dc.b	$99					; 
		dc.b	$30					; 0
		dc.b	$37					; 7
		dc.b	$3C					; <
		dc.b	$6E					; n
		dc.b	$D3					; 
		dc.b	$B6					; 
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$8D					; 
		dc.b	$AD					; 
		dc.b	$8D					; 
		dc.b	$48					; H
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$BF					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$B7					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$53					; S
		dc.b	$52					; R
		dc.b	$6B					; k
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$A2					; 
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$3E					; >
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$3C					; <
		dc.b	7
		dc.b	$5D					; ]
		dc.b	$DE					; 
		dc.b	$F2					; 
		dc.b	$A7					; 
		dc.b	$5A					; Z
		dc.b	$30					; 0
		dc.b	$49					; I
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$B8					; 
		dc.b	$94					; 
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$5D					; ]
		dc.b	$A2					; 
		dc.b	$BA					; 
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$A3					; 
		dc.b	$B8					; 
		dc.b	$87					; 
		dc.b	$60					; `
		dc.b	$33					; 3
		dc.b	$38					; 8
		dc.b	$9B					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$94					; 
		dc.b	$C8					; 
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$27					; '
		dc.b	$3A					; :
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$65					; e
		dc.b	$4D					; M
		dc.b	$8F					; 
		dc.b	$C5					; 
		dc.b	$D0					; 
		dc.b	$79					; y
		dc.b	$2E					; .
		dc.b	$34					; 4
		dc.b	$2A					; *
		dc.b	$63					; c
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$3A					; :
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$AD					; 
		dc.b	$86					; 
		dc.b	$A2					; 
		dc.b	$B4					; 
		dc.b	$5C					; \
		dc.b	$27					; '
		dc.b	$45					; E
		dc.b	$9C					; 
		dc.b	$CD					; 
		dc.b	$D3					; 
		dc.b	$BA					; 
		dc.b	$4C					; L
		dc.b	1
		dc.b	$3A					; :
		dc.b	$96					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$A6					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$9F					; 
		dc.b	$9F					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$55					; U
		dc.b	$95					; 
		dc.b	$B4					; 
		dc.b	$67					; g
		dc.b	$4D					; M
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$42					; B
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$67					; g
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$B7					; 
		dc.b	$94					; 
		dc.b	$A9					; 
		dc.b	$BE					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$48					; H
		dc.b	$57					; W
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$6E					; n
		dc.b	$29					; )
		dc.b	$50					; P
		dc.b	$76					; v
		dc.b	$B4					; 
		dc.b	$FE					; 
		dc.b	$E0					; 
		dc.b	$78					; x
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$55					; U
		dc.b	$67					; g
		dc.b	$B3					; 
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$A2					; 
		dc.b	$A3					; 
		dc.b	$94					; 
		dc.b	$65					; e
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$56					; V
		dc.b	$2C					; ,
		dc.b	$2D					; -
		dc.b	$43					; C
		dc.b	$96					; 
		dc.b	$D0					; 
		dc.b	$C2					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$1B
		dc.b	$49					; I
		dc.b	$A4					; 
		dc.b	$9F					; 
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$3C					; <
		dc.b	$5E					; ^
		dc.b	$33					; 3
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$DE					; 
		dc.b	$BF					; 
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$AA					; 
		dc.b	$D0					; 
		dc.b	$AA					; 
		dc.b	$99					; 
		dc.b	$49					; I
		dc.b	1
		dc.b	3
		dc.b	$A7					; 
		dc.b	$E2					; 
		dc.b	$E1					; 
		dc.b	$E7					; 
		dc.b	$B1					; 
		dc.b	$4F					; O
		dc.b	1
		dc.b	$31					; 1
		dc.b	$7D					; }
		dc.b	$BE					; 
		dc.b	$BB					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$3C					; <
		dc.b	7
		dc.b	$91					; 
		dc.b	$EF					; 
		dc.b	$B6					; 
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$6B					; k
		dc.b	$B0					; 
		dc.b	$AA					; 
		dc.b	$A2					; 
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$4C					; L
		dc.b	$81					; 
		dc.b	$BA					; 
		dc.b	$67					; g
		dc.b	$3E					; >
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$D4					; 
		dc.b	$D3					; 
		dc.b	$52					; R
		dc.b	1
		dc.b	$20
		dc.b	$9F					; 
		dc.b	$E9					; 
		dc.b	$DB					; 
		dc.b	$B8					; 
		dc.b	$48					; H
		dc.b	1
		dc.b	$14
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$9F					; 
		dc.b	$D0					; 
		dc.b	$A7					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$18
		dc.b	$B
		dc.b	$8E					; 
		dc.b	$B6					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$C5					; 
		dc.b	$98					; 
		dc.b	$22					; "
		dc.b	1
		dc.b	$2E					; .
		dc.b	$E0					; 
		dc.b	$FA					; 
		dc.b	$B0					; 
		dc.b	$91					; 
		dc.b	$5D					; ]
		dc.b	$2A					; *
		dc.b	$42					; B
		dc.b	$84					; 
		dc.b	$CC					; 
		dc.b	$D2					; 
		dc.b	$4F					; O
		dc.b	$25					; %
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$C5					; 
		dc.b	$AE					; 
		dc.b	$94					; 
		dc.b	$5A					; Z
		dc.b	9
		dc.b	$25					; %
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$B6					; 
		dc.b	$78					; x
		dc.b	$38					; 8
		dc.b	$4F					; O
		dc.b	$8D					; 
		dc.b	$CB					; 
		dc.b	$DA					; 
		dc.b	$E9					; 
		dc.b	$6B					; k
		dc.b	$22					; "
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$7C					; |
		dc.b	$B1					; 
		dc.b	$D3					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$75					; u
		dc.b	$2D					; -
		dc.b	$6F					; o
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$2E					; .
		dc.b	$75					; u
		dc.b	$C6					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$2E					; .
		dc.b	$4D					; M
		dc.b	$83					; 
		dc.b	$A0					; 
		dc.b	$B4					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$88					; 
		dc.b	$67					; g
		dc.b	$49					; I
		dc.b	$5A					; Z
		dc.b	$9C					; 
		dc.b	$CF					; 
		dc.b	$7C					; |
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$A9					; 
		dc.b	$BF					; 
		dc.b	$A6					; 
		dc.b	$65					; e
		dc.b	$4C					; L
		dc.b	$1B
		dc.b	3
		dc.b	$81					; 
		dc.b	$AD					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$34					; 4
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$A7					; 
		dc.b	$C8					; 
		dc.b	$8E					; 
		dc.b	$5C					; \
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$48					; H
		dc.b	$26					; &
		dc.b	$64					; d
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$B3					; 
		dc.b	$98					; 
		dc.b	$48					; H
		dc.b	$49					; I
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$23					; #
		dc.b	$59					; Y
		dc.b	$FE					; 
		dc.b	$FA					; 
		dc.b	$9B					; 
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$22					; "
		dc.b	4
		dc.b	$4A					; J
		dc.b	$BF					; 
		dc.b	$BF					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$33					; 3
		dc.b	$49					; I
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$B4					; 
		dc.b	$EF					; 
		dc.b	$9F					; 
		dc.b	$31					; 1
		dc.b	$15
		dc.b	$45					; E
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$C5					; 
		dc.b	$A0					; 
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$88					; 
		dc.b	$29					; )
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$A4					; 
		dc.b	$C2					; 
		dc.b	$A9					; 
		dc.b	$50					; P
		dc.b	$35					; 5
		dc.b	$33					; 3
		dc.b	$56					; V
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$64					; d
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$C4					; 
		dc.b	$B7					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$2C					; ,
		dc.b	$43					; C
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$E7					; 
		dc.b	$B4					; 
		dc.b	$30					; 0
		dc.b	1
		dc.b	9
		dc.b	$64					; d
		dc.b	$E7					; 
		dc.b	$FE					; 
		dc.b	$CB					; 
		dc.b	$7C					; |
		dc.b	$26					; &
		dc.b	$1C
		dc.b	$27					; '
		dc.b	$50					; P
		dc.b	$9F					; 
		dc.b	$D4					; 
		dc.b	$CD					; 
		dc.b	$B7					; 
		dc.b	$73					; s
		dc.b	$3F					; ?
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$A7					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$BA					; 
		dc.b	$61					; a
		dc.b	$E
		dc.b	$12
		dc.b	$7A					; z
		dc.b	$C1					; 
		dc.b	$BD					; 
		dc.b	$A6					; 
		dc.b	$6F					; o
		dc.b	$A3					; 
		dc.b	$A4					; 
		dc.b	$60					; `
		dc.b	$46					; F
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$B7					; 
		dc.b	$C8					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$45					; E
		dc.b	$2C					; ,
		dc.b	$67					; g
		dc.b	$B0					; 
		dc.b	$B8					; 
		dc.b	$CC					; 
		dc.b	$C6					; 
		dc.b	$80					; 
		dc.b	$46					; F
		dc.b	$3E					; >
		dc.b	$98					; 
		dc.b	$C4					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$56					; V
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$5A					; Z
		dc.b	1
		dc.b	$31					; 1
		dc.b	$AA					; 
		dc.b	$E9					; 
		dc.b	$D2					; 
		dc.b	$94					; 
		dc.b	$55					; U
		dc.b	1
		dc.b	$1C
		dc.b	$78					; x
		dc.b	$9F					; 
		dc.b	$BD					; 
		dc.b	$E1					; 
		dc.b	$84					; 
		dc.b	$38					; 8
		dc.b	7
		dc.b	$71					; q
		dc.b	$BE					; 
		dc.b	$99					; 
		dc.b	$6C					; l
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$94					; 
		dc.b	$DA					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$91					; 
		dc.b	$63					; c
		dc.b	$17
		dc.b	$30					; 0
		dc.b	$A9					; 
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$B8					; 
		dc.b	$4D					; M
		dc.b	$23					; #
		dc.b	$52					; R
		dc.b	$79					; y
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$65					; e
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$3B					; ;
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$C1					; 
		dc.b	$CC					; 
		dc.b	$96					; 
		dc.b	$56					; V
		dc.b	$3A					; :
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$CF					; 
		dc.b	$EF					; 
		dc.b	$C1					; 
		dc.b	$59					; Y
		dc.b	1
		dc.b	$34					; 4
		dc.b	$8D					; 
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$9C					; 
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$AE					; 
		dc.b	$76					; v
		dc.b	$26					; &
		dc.b	$41					; A
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$A4					; 
		dc.b	$29					; )
		dc.b	$3A					; :
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$78					; x
		dc.b	$E0					; 
		dc.b	$E1					; 
		dc.b	$9C					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$45					; E
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$5D					; ]
		dc.b	$26					; &
		dc.b	$3A					; :
		dc.b	$99					; 
		dc.b	$C8					; 
		dc.b	$9D					; 
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$46					; F
		dc.b	$4F					; O
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$3A					; :
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$A7					; 
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$8A					; 
		dc.b	$2D					; -
		dc.b	$3B					; ;
		dc.b	$A0					; 
		dc.b	$D7					; 
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$53					; S
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$4D					; M
		dc.b	$A7					; 
		dc.b	$BD					; 
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$CD					; 
		dc.b	$63					; c
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$4A					; J
		dc.b	$45					; E
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$42					; B
		dc.b	$53					; S
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$BE					; 
		dc.b	$CD					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$33					; 3
		dc.b	$37					; 7
		dc.b	$92					; 
		dc.b	$D4					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$75					; u
		dc.b	6
		dc.b	1
		dc.b	$64					; d
		dc.b	$D2					; 
		dc.b	$CF					; 
		dc.b	$D2					; 
		dc.b	$C2					; 
		dc.b	$52					; R
		dc.b	$18
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$A4					; 
		dc.b	$87					; 
		dc.b	$4A					; J
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$C8					; 
		dc.b	$8E					; 
		dc.b	$38					; 8
		dc.b	$50					; P
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$CB					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$43					; C
		dc.b	$2D					; -
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$BA					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$61					; a
		dc.b	$B0					; 
		dc.b	$C9					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$6E					; n
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$A9					; 
		dc.b	$96					; 
		dc.b	$78					; x
		dc.b	$38					; 8
		dc.b	$55					; U
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$C2					; 
		dc.b	$B8					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$31					; 1
		dc.b	$5A					; Z
		dc.b	$B6					; 
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$52					; R
		dc.b	$3B					; ;
		dc.b	$81					; 
		dc.b	$A3					; 
		dc.b	$73					; s
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$86					; 
		dc.b	$CC					; 
		dc.b	$B4					; 
		dc.b	$87					; 
		dc.b	$45					; E
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$B8					; 
		dc.b	$B0					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$3F					; ?
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$45					; E
		dc.b	$88					; 
		dc.b	$C8					; 
		dc.b	$C9					; 
		dc.b	$91					; 
		dc.b	$5D					; ]
		dc.b	$29					; )
		dc.b	$48					; H
		dc.b	$84					; 
		dc.b	$AA					; 
		dc.b	$BD					; 
		dc.b	$A4					; 
		dc.b	$B7					; 
		dc.b	$A4					; 
		dc.b	$33					; 3
		dc.b	9
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$5C					; \
		dc.b	$80					; 
		dc.b	$B4					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$9C					; 
		dc.b	$B0					; 
		dc.b	$B6					; 
		dc.b	$3B					; ;
		dc.b	1
		dc.b	$41					; A
		dc.b	$B4					; 
		dc.b	$E0					; 
		dc.b	$AB					; 
		dc.b	$61					; a
		dc.b	$2D					; -
		dc.b	$48					; H
		dc.b	$8D					; 
		dc.b	$B1					; 
		dc.b	$DD					; 
		dc.b	$AB					; 
		dc.b	$38					; 8
		dc.b	$E
		dc.b	$57					; W
		dc.b	$AA					; 
		dc.b	$B7					; 
		dc.b	$BE					; 
		dc.b	$BF					; 
		dc.b	$48					; H
		dc.b	$18
		dc.b	$57					; W
		dc.b	$63					; c
		dc.b	$96					; 
		dc.b	$E1					; 
		dc.b	$B6					; 
		dc.b	$3E					; >
		dc.b	$10
		dc.b	$2C					; ,
		dc.b	$6E					; n
		dc.b	$BD					; 
		dc.b	$CF					; 
		dc.b	$BE					; 
		dc.b	$9D					; 
		dc.b	$2E					; .
		dc.b	$26					; &
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$5C					; \
		dc.b	$43					; C
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$A3					; 
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$43					; C
		dc.b	$4D					; M
		dc.b	$84					; 
		dc.b	$AE					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$61					; a
		dc.b	$2E					; .
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$BA					; 
		dc.b	$D9					; 
		dc.b	$84					; 
		dc.b	$2C					; ,
		dc.b	$20
		dc.b	$7D					; }
		dc.b	$BD					; 
		dc.b	$B6					; 
		dc.b	$C1					; 
		dc.b	$96					; 
		dc.b	$61					; a
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$8F					; 
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$AB					; 
		dc.b	$B1					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$3E					; >
		dc.b	$29					; )
		dc.b	$42					; B
		dc.b	$92					; 
		dc.b	$E0					; 
		dc.b	$DD					; 
		dc.b	$6B					; k
		dc.b	$22					; "
		dc.b	$56					; V
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$D3					; 
		dc.b	$C1					; 
		dc.b	$63					; c
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$81					; 
		dc.b	$BA					; 
		dc.b	$C1					; 
		dc.b	$C9					; 
		dc.b	$60					; `
		dc.b	$25					; %
		dc.b	$4D					; M
		dc.b	$84					; 
		dc.b	$BE					; 
		dc.b	$CF					; 
		dc.b	$9C					; 
		dc.b	$31					; 1
		dc.b	$2A					; *
		dc.b	$72					; r
		dc.b	$B1					; 
		dc.b	$D2					; 
		dc.b	$9D					; 
		dc.b	$4D					; M
		dc.b	$35					; 5
		dc.b	$4F					; O
		dc.b	$A0					; 
		dc.b	$C4					; 
		dc.b	$BD					; 
		dc.b	$8B					; 
		dc.b	$50					; P
		dc.b	$4C					; L
		dc.b	$5E					; ^
		dc.b	$A4					; 
		dc.b	$AA					; 
		dc.b	$B6					; 
		dc.b	$B3					; 
		dc.b	$7A					; z
		dc.b	$D
		dc.b	$2D					; -
		dc.b	$8A					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$42					; B
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$D3					; 
		dc.b	$6E					; n
		dc.b	$18
		dc.b	$B
		dc.b	$48					; H
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$A6					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$80					; 
		dc.b	$BE					; 
		dc.b	$84					; 
		dc.b	$41					; A
		dc.b	$33					; 3
		dc.b	$42					; B
		dc.b	$9C					; 
		dc.b	$DA					; 
		dc.b	$E5					; 
		dc.b	$A7					; 
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$2C					; ,
		dc.b	$98					; 
		dc.b	$E5					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$6A					; j
		dc.b	$37					; 7
		dc.b	$42					; B
		dc.b	$A3					; 
		dc.b	$E9					; 
		dc.b	$B3					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$4C					; L
		dc.b	$75					; u
		dc.b	$D3					; 
		dc.b	$C9					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$4F					; O
		dc.b	$1E
		dc.b	$3C					; <
		dc.b	$C1					; 
		dc.b	$DA					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$46					; F
		dc.b	$5C					; \
		dc.b	$A4					; 
		dc.b	$AB					; 
		dc.b	$63					; c
		dc.b	$2A					; *
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$84					; 
		dc.b	$DA					; 
		dc.b	$C5					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$43					; C
		dc.b	$2D					; -
		dc.b	$3B					; ;
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$C5					; 
		dc.b	$B6					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$AE					; 
		dc.b	$76					; v
		dc.b	$2E					; .
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$63					; c
		dc.b	$BF					; 
		dc.b	$E5					; 
		dc.b	$A7					; 
		dc.b	$34					; 4
		dc.b	$1B
		dc.b	$41					; A
		dc.b	$6E					; n
		dc.b	$BD					; 
		dc.b	$D4					; 
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$59					; Y
		dc.b	$3F					; ?
		dc.b	$68					; h
		dc.b	$9B					; 
		dc.b	$D4					; 
		dc.b	$C9					; 
		dc.b	$A2					; 
		dc.b	$60					; `
		dc.b	$D
		dc.b	$31					; 1
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$AA					; 
		dc.b	$8F					; 
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$C5					; 
		dc.b	$B6					; 
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$A9					; 
		dc.b	$D4					; 
		dc.b	$D9					; 
		dc.b	$AA					; 
		dc.b	$26					; &
		dc.b	$D
		dc.b	$68					; h
		dc.b	$86					; 
		dc.b	$AB					; 
		dc.b	$C9					; 
		dc.b	$A3					; 
		dc.b	$65					; e
		dc.b	$27					; '
		dc.b	$5A					; Z
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$CD					; 
		dc.b	$9C					; 
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$57					; W
		dc.b	$56					; V
		dc.b	$76					; v
		dc.b	$E4					; 
		dc.b	$CB					; 
		dc.b	$42					; B
		dc.b	$3A					; :
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$BD					; 
		dc.b	$B0					; 
		dc.b	$75					; u
		dc.b	$14
		dc.b	$3B					; ;
		dc.b	$A2					; 
		dc.b	$BD					; 
		dc.b	$C2					; 
		dc.b	$8D					; 
		dc.b	$38					; 8
		dc.b	3
		dc.b	$2E					; .
		dc.b	$A6					; 
		dc.b	$D4					; 
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$4D					; M
		dc.b	$2E					; .
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$48					; H
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$35					; 5
		dc.b	$8B					; 
		dc.b	$F6					; 
		dc.b	$C8					; 
		dc.b	$25					; %
		dc.b	$E
		dc.b	$88					; 
		dc.b	$A2					; 
		dc.b	$B8					; 
		dc.b	$EE					; 
		dc.b	$5E					; ^
		dc.b	1
		dc.b	$10
		dc.b	$BE					; 
		dc.b	$E1					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$BE					; 
		dc.b	$BB					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$65					; e
		dc.b	$2A					; *
		dc.b	$46					; F
		dc.b	$6C					; l
		dc.b	$C6					; 
		dc.b	$FE					; 
		dc.b	$D2					; 
		dc.b	$37					; 7
		dc.b	1
		dc.b	$11
		dc.b	$63					; c
		dc.b	$CC					; 
		dc.b	$F9					; 
		dc.b	$AB					; 
		dc.b	$49					; I
		dc.b	$71					; q
		dc.b	$BD					; 
		dc.b	$C9					; 
		dc.b	$B8					; 
		dc.b	$7F					; 
		dc.b	$1B
		dc.b	1
		dc.b	$2E					; .
		dc.b	$AA					; 
		dc.b	$C8					; 
		dc.b	$D0					; 
		dc.b	$AA					; 
		dc.b	$6F					; o
		dc.b	$1F
		dc.b	$D
		dc.b	$6B					; k
		dc.b	$C9					; 
		dc.b	$F2					; 
		dc.b	$C6					; 
		dc.b	$4F					; O
		dc.b	1
		dc.b	$35					; 5
		dc.b	$6E					; n
		dc.b	$D2					; 
		dc.b	$F6					; 
		dc.b	$99					; 
		dc.b	1
		dc.b	1
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$C1					; 
		dc.b	$D7					; 
		dc.b	$BE					; 
		dc.b	$8F					; 
		dc.b	$27					; '
		dc.b	7
		dc.b	$3B					; ;
		dc.b	$80					; 
		dc.b	$D9					; 
		dc.b	$DA					; 
		dc.b	$5E					; ^
		dc.b	$59					; Y
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$A9					; 
		dc.b	$9F					; 
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$B6					; 
		dc.b	$C5					; 
		dc.b	$88					; 
		dc.b	$61					; a
		dc.b	$3F					; ?
		dc.b	1
		dc.b	$75					; u
		dc.b	$F9					; 
		dc.b	$C2					; 
		dc.b	$7A					; z
		dc.b	$3C					; <
		dc.b	$15
		dc.b	$5D					; ]
		dc.b	$9D					; 
		dc.b	$C2					; 
		dc.b	$D7					; 
		dc.b	$E7					; 
		dc.b	$C5					; 
		dc.b	$37					; 7
		dc.b	1
		dc.b	$4F					; O
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$BB					; 
		dc.b	$79					; y
		dc.b	$2D					; -
		dc.b	$52					; R
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$3E					; >
		dc.b	$29					; )
		dc.b	$78					; x
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$65					; e
		dc.b	$1E
		dc.b	$31					; 1
		dc.b	$9B					; 
		dc.b	$DE					; 
		dc.b	$C9					; 
		dc.b	$A6					; 
		dc.b	$50					; P
		dc.b	3
		dc.b	$53					; S
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$76					; v
		dc.b	$4A					; J
		dc.b	$71					; q
		dc.b	$B7					; 
		dc.b	$DD					; 
		dc.b	$98					; 
		dc.b	$15
		dc.b	$2D					; -
		dc.b	$6A					; j
		dc.b	$94					; 
		dc.b	$CB					; 
		dc.b	$C4					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$61					; a
		dc.b	$31					; 1
		dc.b	$78					; x
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$67					; g
		dc.b	$33					; 3
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$A6					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$A7					; 
		dc.b	$49					; I
		dc.b	$2C					; ,
		dc.b	1
		dc.b	$53					; S
		dc.b	$DD					; 
		dc.b	$D4					; 
		dc.b	$A0					; 
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$61					; a
		dc.b	$4A					; J
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$B3					; 
		dc.b	$CD					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$3E					; >
		dc.b	$6C					; l
		dc.b	$B7					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$3E					; >
		dc.b	$2C					; ,
		dc.b	$48					; H
		dc.b	$99					; 
		dc.b	$FE					; 
		dc.b	$BB					; 
		dc.b	$31					; 1
		dc.b	$22					; "
		dc.b	$17
		dc.b	$6A					; j
		dc.b	$DA					; 
		dc.b	$DB					; 
		dc.b	$C9					; 
		dc.b	$7A					; z
		dc.b	$20
		dc.b	$49					; I
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$9C					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$C4					; 
		dc.b	$F7					; 
		dc.b	$B1					; 
		dc.b	$63					; c
		dc.b	$45					; E
		dc.b	$23					; #
		dc.b	$5D					; ]
		dc.b	$B3					; 
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$6B					; k
		dc.b	$23					; #
		dc.b	$46					; F
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$B1					; 
		dc.b	$BF					; 
		dc.b	$6E					; n
		dc.b	$3B					; ;
		dc.b	$6C					; l
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$8A					; 
		dc.b	$3A					; :
		dc.b	$29					; )
		dc.b	$5E					; ^
		dc.b	$9B					; 
		dc.b	$A0					; 
		dc.b	$C1					; 
		dc.b	$98					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$3E					; >
		dc.b	$68					; h
		dc.b	$C4					; 
		dc.b	$C5					; 
		dc.b	$6C					; l
		dc.b	$3E					; >
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.b	$68					; h
		dc.b	$E2					; 
		dc.b	$F3					; 
		dc.b	$A9					; 
		dc.b	$52					; R
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$B1					; 
		dc.b	$D6					; 
		dc.b	$6A					; j
		dc.b	$3F					; ?
		dc.b	$5E					; ^
		dc.b	$8D					; 
		dc.b	$C5					; 
		dc.b	$AB					; 
		dc.b	$61					; a
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$B4					; 
		dc.b	$79					; y
		dc.b	$4F					; O
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$6A					; j
		dc.b	$BF					; 
		dc.b	$81					; 
		dc.b	$23					; #
		dc.b	$4F					; O
		dc.b	$9F					; 
		dc.b	$BF					; 
		dc.b	$AD					; 
		dc.b	$86					; 
		dc.b	$63					; c
		dc.b	$68					; h
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$A9					; 
		dc.b	$A3					; 
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$3A					; :
		dc.b	$50					; P
		dc.b	$A7					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$2A					; *
		dc.b	$33					; 3
		dc.b	$73					; s
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$4C					; L
		dc.b	$49					; I
		dc.b	$34					; 4
		dc.b	$59					; Y
		dc.b	$A3					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$92					; 
		dc.b	$B3					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$4A					; J
		dc.b	$73					; s
		dc.b	$AA					; 
		dc.b	$BD					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$67					; g
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$B1					; 
		dc.b	$C1					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$41					; A
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$91					; 
		dc.b	$C1					; 
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$A3					; 
		dc.b	$98					; 
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$C4					; 
		dc.b	$C8					; 
		dc.b	$B4					; 
		dc.b	$95					; 
		dc.b	$78					; x
		dc.b	$48					; H
		dc.b	$1F
		dc.b	$42					; B
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$6B					; k
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$38					; 8
		dc.b	$59					; Y
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$6E					; n
		dc.b	$38					; 8
		dc.b	$76					; v
		dc.b	$A7					; 
		dc.b	$96					; 
		dc.b	$60					; `
		dc.b	$A9					; 
		dc.b	$A0					; 
		dc.b	$48					; H
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$4F					; O
		dc.b	$8E					; 
		dc.b	$C6					; 
		dc.b	$91					; 
		dc.b	$3A					; :
		dc.b	$19
		dc.b	$52					; R
		dc.b	$A3					; 
		dc.b	$F3					; 
		dc.b	$E4					; 
		dc.b	$55					; U
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$8D					; 
		dc.b	$BF					; 
		dc.b	$C9					; 
		dc.b	$B0					; 
		dc.b	$6B					; k
		dc.b	$D
		dc.b	$20
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$C2					; 
		dc.b	$A4					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$46					; F
		dc.b	$94					; 
		dc.b	$CF					; 
		dc.b	$F2					; 
		dc.b	$8A					; 
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$38					; 8
		dc.b	$A0					; 
		dc.b	$FE					; 
		dc.b	$BE					; 
		dc.b	$1C
		dc.b	1
		dc.b	$3C					; <
		dc.b	$BE					; 
		dc.b	$EC					; 
		dc.b	$BE					; 
		dc.b	$91					; 
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$99					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$87					; 
		dc.b	$C1					; 
		dc.b	$C2					; 
		dc.b	$9C					; 
		dc.b	$BF					; 
		dc.b	$B0					; 
		dc.b	1
		dc.b	1
		dc.b	$4A					; J
		dc.b	$C1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$98					; 
		dc.b	1
		dc.b	1
		dc.b	$23					; #
		dc.b	$CF					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$3E					; >
		dc.b	1
		dc.b	$63					; c
		dc.b	$CD					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$5E					; ^
		dc.b	$A
		dc.b	$33					; 3
		dc.b	$A9					; 
		dc.b	$DE					; 
		dc.b	$BD					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$18
		dc.b	$11
		dc.b	$45					; E
		dc.b	$7F					; 
		dc.b	$F6					; 
		dc.b	$EC					; 
		dc.b	$A7					; 
		dc.b	$37					; 7
		dc.b	1
		dc.b	$B
		dc.b	$79					; y
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B8					; 
		dc.b	$33					; 3
		dc.b	7
		dc.b	$34					; 4
		dc.b	$A6					; 
		dc.b	$F7					; 
		dc.b	$DE					; 
		dc.b	$50					; P
		dc.b	1
		dc.b	$B
		dc.b	$73					; s
		dc.b	$C9					; 
		dc.b	$D6					; 
		dc.b	$BB					; 
		dc.b	$83					; 
		dc.b	$2C					; ,
		dc.b	$2D					; -
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$C2					; 
		dc.b	$8E					; 
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$6F					; o
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$E4					; 
		dc.b	$E9					; 
		dc.b	$9C					; 
		dc.b	$18
		dc.b	$2E					; .
		dc.b	$68					; h
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$AA					; 
		dc.b	$5D					; ]
		dc.b	$3B					; ;
		dc.b	$8F					; 
		dc.b	$CF					; 
		dc.b	$B1					; 
		dc.b	$5D					; ]
		dc.b	$1F
		dc.b	$22					; "
		dc.b	$43					; C
		dc.b	$8A					; 
		dc.b	$C1					; 
		dc.b	$A0					; 
		dc.b	$73					; s
		dc.b	$60					; `
		dc.b	$38					; 8
		dc.b	$42					; B
		dc.b	$99					; 
		dc.b	$CC					; 
		dc.b	$B4					; 
		dc.b	$78					; x
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$25					; %
		dc.b	$95					; 
		dc.b	$FD					; 
		dc.b	$C6					; 
		dc.b	$92					; 
		dc.b	$2D					; -
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$CF					; 
		dc.b	$F0					; 
		dc.b	$B0					; 
		dc.b	$3C					; <
		dc.b	1
		dc.b	$D
		dc.b	$8A					; 
		dc.b	$EF					; 
		dc.b	$E8					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$35					; 5
		dc.b	$15
		dc.b	$65					; e
		dc.b	$B8					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$9B					; 
		dc.b	$5D					; ]
		dc.b	$3C					; <
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$9F					; 
		dc.b	$7C					; |
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$B1					; 
		dc.b	$C2					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$33					; 3
		dc.b	$3A					; :
		dc.b	$BA					; 
		dc.b	$DA					; 
		dc.b	$98					; 
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$9C					; 
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$AA					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$C8					; 
		dc.b	$8A					; 
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$E8					; 
		dc.b	$C2					; 
		dc.b	$5A					; Z
		dc.b	$37					; 7
		dc.b	$19
		dc.b	$6C					; l
		dc.b	$D4					; 
		dc.b	$C5					; 
		dc.b	$A7					; 
		dc.b	$67					; g
		dc.b	$2D					; -
		dc.b	$3C					; <
		dc.b	$63					; c
		dc.b	$AB					; 
		dc.b	$C5					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$60					; `
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$A0					; 
		dc.b	$DE					; 
		dc.b	$EC					; 
		dc.b	$AD					; 
		dc.b	$76					; v
		dc.b	$46					; F
		dc.b	$1F
		dc.b	$30					; 0
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$A0					; 
		dc.b	$83					; 
		dc.b	$42					; B
		dc.b	$52					; R
		dc.b	$7D					; }
		dc.b	$A6					; 
		dc.b	$9C					; 
		dc.b	$5D					; ]
		dc.b	$33					; 3
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$A2					; 
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$84					; 
		dc.b	$AD					; 
		dc.b	$B8					; 
		dc.b	$95					; 
		dc.b	$3F					; ?
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$25					; %
		dc.b	$48					; H
		dc.b	$98					; 
		dc.b	$C1					; 
		dc.b	$BA					; 
		dc.b	$8D					; 
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$5D					; ]
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$88					; 
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$76					; v
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$5D					; ]
		dc.b	$3B					; ;
		dc.b	$59					; Y
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$C6					; 
		dc.b	$B7					; 
		dc.b	$4F					; O
		dc.b	$B
		dc.b	$46					; F
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$92					; 
		dc.b	$60					; `
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$8E					; 
		dc.b	$A6					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$67					; g
		dc.b	$19
		dc.b	$3A					; :
		dc.b	$BB					; 
		dc.b	$E8					; 
		dc.b	$C5					; 
		dc.b	$64					; d
		dc.b	$14
		dc.b	$14
		dc.b	$3B					; ;
		dc.b	$99					; 
		dc.b	$F9					; 
		dc.b	$BB					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$A4					; 
		dc.b	$CD					; 
		dc.b	$A9					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$3C					; <
		dc.b	$38					; 8
		dc.b	$6C					; l
		dc.b	$A9					; 
		dc.b	$B3					; 
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$34					; 4
		dc.b	$26					; &
		dc.b	$9D					; 
		dc.b	$FD					; 
		dc.b	$CD					; 
		dc.b	$71					; q
		dc.b	$26					; &
		dc.b	7
		dc.b	$80					; 
		dc.b	$E1					; 
		dc.b	$D0					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$27					; '
		dc.b	$12
		dc.b	$96					; 
		dc.b	$D4					; 
		dc.b	$B0					; 
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$2D					; -
		dc.b	$12
		dc.b	$64					; d
		dc.b	$DB					; 
		dc.b	$F5					; 
		dc.b	$BF					; 
		dc.b	$65					; e
		dc.b	$15
		dc.b	$15
		dc.b	$60					; `
		dc.b	$9F					; 
		dc.b	$E7					; 
		dc.b	$FA					; 
		dc.b	$9D					; 
		dc.b	7
		dc.b	1
		dc.b	$6E					; n
		dc.b	$B7					; 
		dc.b	$A4					; 
		dc.b	$79					; y
		dc.b	$55					; U
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$C6					; 
		dc.b	$D9					; 
		dc.b	$7D					; }
		dc.b	$3E					; >
		dc.b	$19
		dc.b	$31					; 1
		dc.b	$88					; 
		dc.b	$C6					; 
		dc.b	$B6					; 
		dc.b	$91					; 
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$99					; 
		dc.b	$B8					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$52					; R
		dc.b	$2D					; -
		dc.b	$3F					; ?
		dc.b	$88					; 
		dc.b	$BA					; 
		dc.b	$AD					; 
		dc.b	$71					; q
		dc.b	$2A					; *
		dc.b	$48					; H
		dc.b	$92					; 
		dc.b	$D6					; 
		dc.b	$BA					; 
		dc.b	$4F					; O
		dc.b	$31					; 1
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.b	$6F					; o
		dc.b	$FE					; 
		dc.b	$F2					; 
		dc.b	$67					; g
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$BD					; 
		dc.b	$BF					; 
		dc.b	$5D					; ]
		dc.b	$10
		dc.b	$37					; 7
		dc.b	$94					; 
		dc.b	$DA					; 
		dc.b	$CD					; 
		dc.b	$71					; q
		dc.b	$48					; H
		dc.b	$42					; B
		dc.b	$41					; A
		dc.b	$5E					; ^
		dc.b	$9D					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$57					; W
		dc.b	$46					; F
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$8B					; 
		dc.b	$B6					; 
		dc.b	$91					; 
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$BB					; 
		dc.b	$C6					; 
		dc.b	$76					; v
		dc.b	$3B					; ;
		dc.b	$55					; U
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$B4					; 
		dc.b	$75					; u
		dc.b	$43					; C
		dc.b	$75					; u
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$B6					; 
		dc.b	$A9					; 
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$67					; g
		dc.b	$2D					; -
		dc.b	$4D					; M
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$3F					; ?
		dc.b	1
		dc.b	$4F					; O
		dc.b	$BA					; 
		dc.b	$D9					; 
		dc.b	$CB					; 
		dc.b	$7A					; z
		dc.b	$1F
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$83					; 
		dc.b	$CC					; 
		dc.b	$CC					; 
		dc.b	$83					; 
		dc.b	$2D					; -
		dc.b	$18
		dc.b	$94					; 
		dc.b	$CF					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$48					; H
		dc.b	$7F					; 
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$7C					; |
		dc.b	$34					; 4
		dc.b	$3B					; ;
		dc.b	$60					; `
		dc.b	$A3					; 
		dc.b	$C1					; 
		dc.b	$95					; 
		dc.b	$6C					; l
		dc.b	$46					; F
		dc.b	$2D					; -
		dc.b	$5A					; Z
		dc.b	$B3					; 
		dc.b	$E8					; 
		dc.b	$E9					; 
		dc.b	$9C					; 
		dc.b	2
		dc.b	1
		dc.b	$43					; C
		dc.b	$AA					; 
		dc.b	$EB					; 
		dc.b	$D9					; 
		dc.b	$65					; e
		dc.b	$14
		dc.b	$1F
		dc.b	$98					; 
		dc.b	$F2					; 
		dc.b	$EC					; 
		dc.b	$AE					; 
		dc.b	$3E					; >
		dc.b	9
		dc.b	$53					; S
		dc.b	$B1					; 
		dc.b	$E2					; 
		dc.b	$B8					; 
		dc.b	$75					; u
		dc.b	$35					; 5
		dc.b	$31					; 1
		dc.b	$41					; A
		dc.b	$87					; 
		dc.b	$DE					; 
		dc.b	$F0					; 
		dc.b	$BB					; 
		dc.b	$19
		dc.b	1
		dc.b	$4D					; M
		dc.b	$B7					; 
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$D3					; 
		dc.b	$30					; 0
		dc.b	1
		dc.b	$17
		dc.b	$B6					; 
		dc.b	$FE					; 
		dc.b	$D2					; 
		dc.b	$7A					; z
		dc.b	$3A					; :
		dc.b	$31					; 1
		dc.b	$6B					; k
		dc.b	$92					; 
		dc.b	$C9					; 
		dc.b	$CD					; 
		dc.b	$57					; W
		dc.b	$33					; 3
		dc.b	$5E					; ^
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$AA					; 
		dc.b	$91					; 
		dc.b	$5C					; \
		dc.b	$19
		dc.b	$22					; "
		dc.b	$81					; 
		dc.b	$C9					; 
		dc.b	$D9					; 
		dc.b	$A3					; 
		dc.b	$42					; B
		dc.b	$B
		dc.b	$3C					; <
		dc.b	$81					; 
		dc.b	$CF					; 
		dc.b	$E8					; 
		dc.b	$84					; 
		dc.b	9
		dc.b	1
		dc.b	$31					; 1
		dc.b	$86					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$A3					; 
		dc.b	$1F
		dc.b	1
		dc.b	6
		dc.b	$5D					; ]
		dc.b	$DA					; 
		dc.b	$FE					; 
		dc.b	$8F					; 
		dc.b	$27					; '
		dc.b	$26					; &
		dc.b	$41					; A
		dc.b	$59					; Y
		dc.b	$A4					; 
		dc.b	$FA					; 
		dc.b	$B3					; 
		dc.b	$48					; H
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$B4					; 
		dc.b	$D0					; 
		dc.b	$86					; 
		dc.b	$55					; U
		dc.b	$3E					; >
		dc.b	$41					; A
		dc.b	$8A					; 
		dc.b	$B6					; 
		dc.b	$AB					; 
		dc.b	$B0					; 
		dc.b	$7D					; }
		dc.b	$2D					; -
		dc.b	$2D					; -
		dc.b	$53					; S
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$A7					; 
		dc.b	$71					; q
		dc.b	$1E
		dc.b	$2D					; -
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$AB					; 
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$6C					; l
		dc.b	$27					; '
		dc.b	$5D					; ]
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$A3					; 
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$43					; C
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$98					; 
		dc.b	$DA					; 
		dc.b	$B4					; 
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$56					; V
		dc.b	$49					; I
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$7C					; |
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$49					; I
		dc.b	$75					; u
		dc.b	$BD					; 
		dc.b	$B3					; 
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$4D					; M
		dc.b	$59					; Y
		dc.b	$95					; 
		dc.b	$BB					; 
		dc.b	$9C					; 
		dc.b	$3F					; ?
		dc.b	$2C					; ,
		dc.b	$68					; h
		dc.b	$B7					; 
		dc.b	$A0					; 
		dc.b	$6A					; j
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$61					; a
		dc.b	$49					; I
		dc.b	$5D					; ]
		dc.b	$98					; 
		dc.b	$DA					; 
		dc.b	$D2					; 
		dc.b	$8A					; 
		dc.b	$4A					; J
		dc.b	$55					; U
		dc.b	$50					; P
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$81					; 
		dc.b	$AB					; 
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$5D					; ]
		dc.b	$96					; 
		dc.b	$C5					; 
		dc.b	$A0					; 
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$AD					; 
		dc.b	$D0					; 
		dc.b	$AD					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$B6					; 
		dc.b	$C1					; 
		dc.b	$45					; E
		dc.b	$1F
		dc.b	$46					; F
		dc.b	$8F					; 
		dc.b	$F9					; 
		dc.b	$F9					; 
		dc.b	$9D					; 
		dc.b	$53					; S
		dc.b	1
		dc.b	$B
		dc.b	$95					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$A7					; 
		dc.b	$29					; )
		dc.b	$26					; &
		dc.b	$59					; Y
		dc.b	$8E					; 
		dc.b	$D0					; 
		dc.b	$A6					; 
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$59					; Y
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$D7					; 
		dc.b	$CF					; 
		dc.b	$5D					; ]
		dc.b	1
		dc.b	1
		dc.b	$4F					; O
		dc.b	$D2					; 
		dc.b	$FE					; 
		dc.b	$F6					; 
		dc.b	$6A					; j
		dc.b	7
		dc.b	1
		dc.b	$45					; E
		dc.b	$A2					; 
		dc.b	$D7					; 
		dc.b	$C5					; 
		dc.b	$6F					; o
		dc.b	$3C					; <
		dc.b	$1F
		dc.b	$48					; H
		dc.b	$AB					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$31					; 1
		dc.b	$15
		dc.b	$75					; u
		dc.b	$CC					; 
		dc.b	$AB					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$B6					; 
		dc.b	$E5					; 
		dc.b	$6F					; o
		dc.b	$E
		dc.b	$20
		dc.b	$63					; c
		dc.b	$B3					; 
		dc.b	$DD					; 
		dc.b	$E1					; 
		dc.b	$B0					; 
		dc.b	$5D					; ]
		dc.b	$2A					; *
		dc.b	$45					; E
		dc.b	$84					; 
		dc.b	$B4					; 
		dc.b	$AA					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$5A					; Z
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$A7					; 
		dc.b	$7F					; 
		dc.b	$61					; a
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$3E					; >
		dc.b	$12
		dc.b	$48					; H
		dc.b	$84					; 
		dc.b	$A4					; 
		dc.b	$C2					; 
		dc.b	$D2					; 
		dc.b	$87					; 
		dc.b	1
		dc.b	1
		dc.b	$63					; c
		dc.b	$D4					; 
		dc.b	$D2					; 
		dc.b	$98					; 
		dc.b	$5A					; Z
		dc.b	$31					; 1
		dc.b	$3E					; >
		dc.b	$50					; P
		dc.b	$8A					; 
		dc.b	$BF					; 
		dc.b	$C8					; 
		dc.b	$96					; 
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$AB					; 
		dc.b	$B6					; 
		dc.b	$B0					; 
		dc.b	$75					; u
		dc.b	$27					; '
		dc.b	$59					; Y
		dc.b	$AE					; 
		dc.b	$DB					; 
		dc.b	$DB					; 
		dc.b	$5A					; Z
		dc.b	$29					; )
		dc.b	$46					; F
		dc.b	$61					; a
		dc.b	$AE					; 
		dc.b	$D0					; 
		dc.b	$BE					; 
		dc.b	$9D					; 
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$75					; u
		dc.b	$4C					; L
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$A3					; 
		dc.b	$68					; h
		dc.b	7
		dc.b	$25					; %
		dc.b	$9C					; 
		dc.b	$C9					; 
		dc.b	$BB					; 
		dc.b	$AB					; 
		dc.b	$25					; %
		dc.b	1
		dc.b	6
		dc.b	$80					; 
		dc.b	$EE					; 
		dc.b	$FE					; 
		dc.b	$D9					; 
		dc.b	$19
		dc.b	1
		dc.b	$38					; 8
		dc.b	$68					; h
		dc.b	$91					; 
		dc.b	$CF					; 
		dc.b	$DB					; 
		dc.b	$91					; 
		dc.b	$56					; V
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$31					; 1
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$99					; 
		dc.b	$C2					; 
		dc.b	$CD					; 
		dc.b	$AB					; 
		dc.b	$6A					; j
		dc.b	$26					; &
		dc.b	$1B
		dc.b	$81					; 
		dc.b	$C1					; 
		dc.b	$E4					; 
		dc.b	$E1					; 
		dc.b	$57					; W
		dc.b	1
		dc.b	$31					; 1
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$C4					; 
		dc.b	$BE					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$37					; 7
		dc.b	$5D					; ]
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$68					; h
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$95					; 
		dc.b	$B7					; 
		dc.b	$96					; 
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$71					; q
		dc.b	$B8					; 
		dc.b	$9B					; 
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$C8					; 
		dc.b	$91					; 
		dc.b	$31					; 1
		dc.b	$3F					; ?
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$D3					; 
		dc.b	$B3					; 
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$30					; 0
		dc.b	$55					; U
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$AE					; 
		dc.b	$C2					; 
		dc.b	$8F					; 
		dc.b	$64					; d
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$98					; 
		dc.b	$BE					; 
		dc.b	$A3					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$B0					; 
		dc.b	$AD					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$87					; 
		dc.b	$46					; F
		dc.b	$3A					; :
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$46					; F
		dc.b	$7A					; z
		dc.b	$BF					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$49					; I
		dc.b	$42					; B
		dc.b	$8F					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$48					; H
		dc.b	$35					; 5
		dc.b	$43					; C
		dc.b	$81					; 
		dc.b	$C6					; 
		dc.b	$DB					; 
		dc.b	$99					; 
		dc.b	$65					; e
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$BB					; 
		dc.b	$67					; g
		dc.b	$2C					; ,
		dc.b	$41					; A
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$4F					; O
		dc.b	$3C					; <
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$91					; 
		dc.b	$D3					; 
		dc.b	$BD					; 
		dc.b	$8E					; 
		dc.b	$5D					; ]
		dc.b	$1B
		dc.b	$2E					; .
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$B1					; 
		dc.b	$5D					; ]
		dc.b	$34					; 4
		dc.b	$67					; g
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$A0					; 
		dc.b	$B7					; 
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$A6					; 
		dc.b	$BE					; 
		dc.b	$B0					; 
		dc.b	$7D					; }
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$AB					; 
		dc.b	$AE					; 
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$8A					; 
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$84					; 
		dc.b	$B4					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$4C					; L
		dc.b	$60					; `
		dc.b	$A2					; 
		dc.b	$BA					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$59					; Y
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$84					; 
		dc.b	$BD					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$41					; A
		dc.b	$27					; '
		dc.b	$75					; u
		dc.b	$AB					; 
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$35					; 5
		dc.b	7
		dc.b	$59					; Y
		dc.b	$B7					; 
		dc.b	$DA					; 
		dc.b	$B8					; 
		dc.b	$88					; 
		dc.b	$3A					; :
		dc.b	$2E					; .
		dc.b	$6B					; k
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$48					; H
		dc.b	$3A					; :
		dc.b	$6E					; n
		dc.b	$B0					; 
		dc.b	$CD					; 
		dc.b	$DB					; 
		dc.b	$BD					; 
		dc.b	$3C					; <
		dc.b	1
		dc.b	$15
		dc.b	$8B					; 
		dc.b	$E1					; 
		dc.b	$FC					; 
		dc.b	$C5					; 
		dc.b	$5C					; \
		dc.b	6
		dc.b	9
		dc.b	$6B					; k
		dc.b	$C6					; 
		dc.b	$E9					; 
		dc.b	$8A					; 
		dc.b	1
		dc.b	$23					; #
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$B1					; 
		dc.b	$DD					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$4C					; L
		dc.b	$3B					; ;
		dc.b	$72					; r
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$AB					; 
		dc.b	$C5					; 
		dc.b	$78					; x
		dc.b	$26					; &
		dc.b	$2D					; -
		dc.b	$41					; A
		dc.b	$73					; s
		dc.b	$C6					; 
		dc.b	$D6					; 
		dc.b	$9F					; 
		dc.b	$67					; g
		dc.b	$50					; P
		dc.b	$3A					; :
		dc.b	$92					; 
		dc.b	$EF					; 
		dc.b	$B6					; 
		dc.b	$5D					; ]
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$95					; 
		dc.b	$C4					; 
		dc.b	$CD					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$45					; E
		dc.b	$25					; %
		dc.b	$3B					; ;
		dc.b	$99					; 
		dc.b	$C6					; 
		dc.b	$D6					; 
		dc.b	$D3					; 
		dc.b	$59					; Y
		dc.b	$E
		dc.b	$3A					; :
		dc.b	$67					; g
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$45					; E
		dc.b	$5D					; ]
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$B4					; 
		dc.b	$E1					; 
		dc.b	$BA					; 
		dc.b	$50					; P
		dc.b	$31					; 1
		dc.b	$29					; )
		dc.b	$3E					; >
		dc.b	$7D					; }
		dc.b	$B4					; 
		dc.b	$CF					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$83					; 
		dc.b	$4D					; M
		dc.b	$2A					; *
		dc.b	$55					; U
		dc.b	$94					; 
		dc.b	$CC					; 
		dc.b	$A4					; 
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$4C					; L
		dc.b	$AD					; 
		dc.b	$D0					; 
		dc.b	$C5					; 
		dc.b	$8F					; 
		dc.b	$4F					; O
		dc.b	$25					; %
		dc.b	$2C					; ,
		dc.b	$6A					; j
		dc.b	$BD					; 
		dc.b	$FE					; 
		dc.b	$E1					; 
		dc.b	$98					; 
		dc.b	$3E					; >
		dc.b	1
		dc.b	$29					; )
		dc.b	$AD					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A7					; 
		dc.b	$2D					; -
		dc.b	1
		dc.b	1
		dc.b	$72					; r
		dc.b	$CB					; 
		dc.b	$E1					; 
		dc.b	$C8					; 
		dc.b	$A0					; 
		dc.b	$5E					; ^
		dc.b	$22					; "
		dc.b	$37					; 7
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$B7					; 
		dc.b	$9C					; 
		dc.b	$4A					; J
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$2A					; *
		dc.b	1
		dc.b	$35					; 5
		dc.b	$7C					; |
		dc.b	$A2					; 
		dc.b	$DA					; 
		dc.b	$D4					; 
		dc.b	$9B					; 
		dc.b	$6B					; k
		dc.b	$45					; E
		dc.b	$43					; C
		dc.b	$46					; F
		dc.b	$52					; R
		dc.b	$87					; 
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$48					; H
		dc.b	$23					; #
		dc.b	$53					; S
		dc.b	$B1					; 
		dc.b	$EB					; 
		dc.b	$C9					; 
		dc.b	$A4					; 
		dc.b	$48					; H
		dc.b	1
		dc.b	$18
		dc.b	$6C					; l
		dc.b	$C4					; 
		dc.b	$E0					; 
		dc.b	$D9					; 
		dc.b	$A0					; 
		dc.b	$6E					; n
		dc.b	$50					; P
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$BA					; 
		dc.b	$94					; 
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$AE					; 
		dc.b	$6F					; o
		dc.b	$2D					; -
		dc.b	$19
		dc.b	$4F					; O
		dc.b	$CF					; 
		dc.b	$F2					; 
		dc.b	$C2					; 
		dc.b	$59					; Y
		dc.b	$30					; 0
		dc.b	$48					; H
		dc.b	$5D					; ]
		dc.b	$86					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$75					; u
		dc.b	$42					; B
		dc.b	$37					; 7
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$9D					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$43					; C
		dc.b	$75					; u
		dc.b	$A9					; 
		dc.b	$C2					; 
		dc.b	$96					; 
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$94					; 
		dc.b	$B4					; 
		dc.b	$BD					; 
		dc.b	$B1					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$11
		dc.b	$37					; 7
		dc.b	$A6					; 
		dc.b	$E1					; 
		dc.b	$EC					; 
		dc.b	$DA					; 
		dc.b	$52					; R
		dc.b	1
		dc.b	$2A					; *
		dc.b	$6A					; j
		dc.b	$A7					; 
		dc.b	$EF					; 
		dc.b	$FE					; 
		dc.b	$B8					; 
		dc.b	$27					; '
		dc.b	1
		dc.b	$49					; I
		dc.b	$BE					; 
		dc.b	$B6					; 
		dc.b	$B7					; 
		dc.b	$9D					; 
		dc.b	$34					; 4
		dc.b	1
		dc.b	$3A					; :
		dc.b	$A3					; 
		dc.b	$CD					; 
		dc.b	$C6					; 
		dc.b	$AA					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$BF					; 
		dc.b	$B3					; 
		dc.b	$63					; c
		dc.b	$22					; "
		dc.b	$1B
		dc.b	$5A					; Z
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$79					; y
		dc.b	6
		dc.b	1
		dc.b	1
		dc.b	$49					; I
		dc.b	$CC					; 
		dc.b	$FE					; 
		dc.b	$C2					; 
		dc.b	$5E					; ^
		dc.b	$33					; 3
		dc.b	$2E					; .
		dc.b	$35					; 5
		dc.b	$6E					; n
		dc.b	$DE					; 
		dc.b	$FE					; 
		dc.b	$91					; 
		dc.b	$42					; B
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$3A					; :
		dc.b	$63					; c
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$81					; 
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$4A					; J
		dc.b	$87					; 
		dc.b	$C6					; 
		dc.b	$C5					; 
		dc.b	$7C					; |
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$94					; 
		dc.b	$B4					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$4F					; O
		dc.b	$11
		dc.b	9
		dc.b	$59					; Y
		dc.b	$C1					; 
		dc.b	$FE					; 
		dc.b	$E4					; 
		dc.b	$84					; 
		dc.b	$2E					; .
		dc.b	4
		dc.b	$4A					; J
		dc.b	$91					; 
		dc.b	$E1					; 
		dc.b	$E9					; 
		dc.b	$80					; 
		dc.b	$3E					; >
		dc.b	$2E					; .
		dc.b	$52					; R
		dc.b	$A0					; 
		dc.b	$D0					; 
		dc.b	$AE					; 
		dc.b	$61					; a
		dc.b	$25					; %
		dc.b	$14
		dc.b	$49					; I
		dc.b	$8E					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$31					; 1
		dc.b	6
		dc.b	$3C					; <
		dc.b	$91					; 
		dc.b	$EE					; 
		dc.b	$F6					; 
		dc.b	$EF					; 
		dc.b	$96					; 
		dc.b	1
		dc.b	1
		dc.b	$29					; )
		dc.b	$59					; Y
		dc.b	$9D					; 
		dc.b	$EC					; 
		dc.b	$FA					; 
		dc.b	$D2					; 
		dc.b	$61					; a
		dc.b	$E
		dc.b	$2C					; ,
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$B0					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$9D					; 
		dc.b	$6B					; k
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$BA					; 
		dc.b	$C6					; 
		dc.b	$8A					; 
		dc.b	$46					; F
		dc.b	$4C					; L
		dc.b	$68					; h
		dc.b	$91					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$75					; u
		dc.b	$5D					; ]
		dc.b	$30					; 0
		dc.b	$5E					; ^
		dc.b	$A7					; 
		dc.b	$C1					; 
		dc.b	$B7					; 
		dc.b	$AE					; 
		dc.b	$79					; y
		dc.b	$4C					; L
		dc.b	$31					; 1
		dc.b	$42					; B
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$B0					; 
		dc.b	$76					; v
		dc.b	$3E					; >
		dc.b	$80					; 
		dc.b	$BD					; 
		dc.b	$91					; 
		dc.b	$55					; U
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$B4					; 
		dc.b	$8A					; 
		dc.b	$4F					; O
		dc.b	$1C
		dc.b	$E
		dc.b	$37					; 7
		dc.b	$A3					; 
		dc.b	$F0					; 
		dc.b	$C8					; 
		dc.b	$AB					; 
		dc.b	$83					; 
		dc.b	$35					; 5
		dc.b	$2A					; *
		dc.b	$63					; c
		dc.b	$B7					; 
		dc.b	$B6					; 
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$3B					; ;
		dc.b	$19
		dc.b	$31					; 1
		dc.b	$80					; 
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$DD					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	1
		dc.b	7
		dc.b	$84					; 
		dc.b	$C8					; 
		dc.b	$E1					; 
		dc.b	$C2					; 
		dc.b	$81					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$56					; V
		dc.b	$8E					; 
		dc.b	$D2					; 
		dc.b	$95					; 
		dc.b	$48					; H
		dc.b	$33					; 3
		dc.b	$3C					; <
		dc.b	$55					; U
		dc.b	$A9					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$80					; 
		dc.b	$11
		dc.b	$25					; %
		dc.b	$67					; g
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$BE					; 
		dc.b	$C4					; 
		dc.b	$91					; 
		dc.b	$53					; S
		dc.b	$2A					; *
		dc.b	$D
		dc.b	$22					; "
		dc.b	$6E					; n
		dc.b	$BE					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$AB					; 
		dc.b	$78					; x
		dc.b	$4A					; J
		dc.b	1
		dc.b	1
		dc.b	$7C					; |
		dc.b	$D0					; 
		dc.b	$B3					; 
		dc.b	$48					; H
		dc.b	$64					; d
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	6
		dc.b	1
		dc.b	$4C					; L
		dc.b	$BE					; 
		dc.b	$F7					; 
		dc.b	$BF					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$4C					; L
		dc.b	$25					; %
		dc.b	$3A					; :
		dc.b	$73					; s
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$6C					; l
		dc.b	$29					; )
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$88					; 
		dc.b	$C1					; 
		dc.b	$D4					; 
		dc.b	$AE					; 
		dc.b	$83					; 
		dc.b	$46					; F
		dc.b	$19
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$D9					; 
		dc.b	$CB					; 
		dc.b	$8E					; 
		dc.b	$3F					; ?
		dc.b	$34					; 4
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$A3					; 
		dc.b	$AE					; 
		dc.b	$7C					; |
		dc.b	$19
		dc.b	1
		dc.b	$49					; I
		dc.b	$BE					; 
		dc.b	$FD					; 
		dc.b	$E2					; 
		dc.b	$A6					; 
		dc.b	$5C					; \
		dc.b	7
		dc.b	1
		dc.b	$20
		dc.b	$A4					; 
		dc.b	$FA					; 
		dc.b	$DA					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$7D					; }
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$2E					; .
		dc.b	$B
		dc.b	$3E					; >
		dc.b	$94					; 
		dc.b	$BF					; 
		dc.b	$B0					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$60					; `
		dc.b	$11
		dc.b	$35					; 5
		dc.b	$95					; 
		dc.b	$C5					; 
		dc.b	$D3					; 
		dc.b	$99					; 
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$57					; W
		dc.b	$48					; H
		dc.b	$5D					; ]
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$9F					; 
		dc.b	$3B					; ;
		dc.b	1
		dc.b	$15
		dc.b	$63					; c
		dc.b	$A6					; 
		dc.b	$E2					; 
		dc.b	$BF					; 
		dc.b	$63					; c
		dc.b	$35					; 5
		dc.b	$49					; I
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$B3					; 
		dc.b	$BA					; 
		dc.b	$83					; 
		dc.b	$59					; Y
		dc.b	$38					; 8
		dc.b	$2E					; .
		dc.b	$5C					; \
		dc.b	$95					; 
		dc.b	$B6					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$48					; H
		dc.b	$63					; c
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$E1					; 
		dc.b	$FE					; 
		dc.b	$C1					; 
		dc.b	$2C					; ,
		dc.b	1
		dc.b	6
		dc.b	$60					; `
		dc.b	$CB					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$98					; 
		dc.b	$11
		dc.b	1
		dc.b	$3C					; <
		dc.b	$96					; 
		dc.b	$B0					; 
		dc.b	$BB					; 
		dc.b	$B7					; 
		dc.b	$88					; 
		dc.b	$65					; e
		dc.b	$46					; F
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$49					; I
		dc.b	$34					; 4
		dc.b	$7A					; z
		dc.b	$C1					; 
		dc.b	$E5					; 
		dc.b	$C5					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$5D					; ]
		dc.b	$17
		dc.b	$23					; #
		dc.b	$84					; 
		dc.b	$D3					; 
		dc.b	$B3					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$AB					; 
		dc.b	$9B					; 
		dc.b	$45					; E
		dc.b	$29					; )
		dc.b	$41					; A
		dc.b	$7C					; |
		dc.b	$B3					; 
		dc.b	$EC					; 
		dc.b	$BD					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$48					; H
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$52					; R
		dc.b	$76					; v
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$91					; 
		dc.b	$C4					; 
		dc.b	$B0					; 
		dc.b	$86					; 
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$65					; e
		dc.b	$3A					; :
		dc.b	$41					; A
		dc.b	$60					; `
		dc.b	$96					; 
		dc.b	$B8					; 
		dc.b	$EB					; 
		dc.b	$CC					; 
		dc.b	$6B					; k
		dc.b	$5D					; ]
		dc.b	$2C					; ,
		dc.b	1
		dc.b	$4C					; L
		dc.b	$CC					; 
		dc.b	$F5					; 
		dc.b	$C5					; 
		dc.b	$65					; e
		dc.b	$23					; #
		dc.b	$1F
		dc.b	$20
		dc.b	$53					; S
		dc.b	$D4					; 
		dc.b	$DE					; 
		dc.b	$BA					; 
		dc.b	$95					; 
		dc.b	$41					; A
		dc.b	$20
		dc.b	$3E					; >
		dc.b	$8A					; 
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$C6					; 
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$52					; R
		dc.b	1
		dc.b	$B
		dc.b	$6C					; l
		dc.b	$AA					; 
		dc.b	$D9					; 
		dc.b	$D7					; 
		dc.b	$A6					; 
		dc.b	$48					; H
		dc.b	$20
		dc.b	$19
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$BD					; 
		dc.b	$4D					; M
		dc.b	$42					; B
		dc.b	$25					; %
		dc.b	$50					; P
		dc.b	$A6					; 
		dc.b	$C4					; 
		dc.b	$C6					; 
		dc.b	$92					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$46					; F
		dc.b	$43					; C
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$BD					; 
		dc.b	$C5					; 
		dc.b	$9D					; 
		dc.b	$63					; c
		dc.b	$1B
		dc.b	$17
		dc.b	$76					; v
		dc.b	$E7					; 
		dc.b	$E8					; 
		dc.b	$96					; 
		dc.b	$53					; S
		dc.b	$22					; "
		dc.b	$20
		dc.b	$48					; H
		dc.b	$9F					; 
		dc.b	$DB					; 
		dc.b	$F0					; 
		dc.b	$AB					; 
		dc.b	$4D					; M
		dc.b	$45					; E
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$55					; U
		dc.b	$56					; V
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$9D					; 
		dc.b	$61					; a
		dc.b	$38					; 8
		dc.b	$6E					; n
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$68					; h
		dc.b	$4C					; L
		dc.b	$61					; a
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$91					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$5A					; Z
		dc.b	$37					; 7
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$A4					; 
		dc.b	$C8					; 
		dc.b	$96					; 
		dc.b	$4A					; J
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$A2					; 
		dc.b	$3A					; :
		dc.b	1
		dc.b	$48					; H
		dc.b	$8D					; 
		dc.b	$B7					; 
		dc.b	$D7					; 
		dc.b	$C2					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$20
		dc.b	$2D					; -
		dc.b	$8A					; 
		dc.b	$EC					; 
		dc.b	$E4					; 
		dc.b	$67					; g
		dc.b	$2C					; ,
		dc.b	$4F					; O
		dc.b	$83					; 
		dc.b	$A2					; 
		dc.b	$C1					; 
		dc.b	$96					; 
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$91					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$56					; V
		dc.b	$57					; W
		dc.b	$94					; 
		dc.b	$A2					; 
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$5E					; ^
		dc.b	$35					; 5
		dc.b	$38					; 8
		dc.b	$7A					; z
		dc.b	$D9					; 
		dc.b	$BE					; 
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$2A					; *
		dc.b	$22					; "
		dc.b	$80					; 
		dc.b	$C8					; 
		dc.b	$CD					; 
		dc.b	$E2					; 
		dc.b	$B6					; 
		dc.b	$61					; a
		dc.b	$18
		dc.b	1
		dc.b	$52					; R
		dc.b	$B7					; 
		dc.b	$C2					; 
		dc.b	$C9					; 
		dc.b	$E7					; 
		dc.b	$BF					; 
		dc.b	$8F					; 
		dc.b	$4A					; J
		dc.b	$2C					; ,
		dc.b	$4C					; L
		dc.b	$8B					; 
		dc.b	$B7					; 
		dc.b	$BA					; 
		dc.b	$A2					; 
		dc.b	$56					; V
		dc.b	$2C					; ,
		dc.b	$33					; 3
		dc.b	$72					; r
		dc.b	$B7					; 
		dc.b	$E1					; 
		dc.b	$BE					; 
		dc.b	$76					; v
		dc.b	$60					; `
		dc.b	$49					; I
		dc.b	$33					; 3
		dc.b	$2E					; .
		dc.b	$71					; q
		dc.b	$C9					; 
		dc.b	$D4					; 
		dc.b	$B3					; 
		dc.b	$88					; 
		dc.b	$60					; `
		dc.b	$2D					; -
		dc.b	$2E					; .
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$AB					; 
		dc.b	$BB					; 
		dc.b	$91					; 
		dc.b	$57					; W
		dc.b	$1F
		dc.b	$12
		dc.b	$56					; V
		dc.b	$AA					; 
		dc.b	$C6					; 
		dc.b	$BA					; 
		dc.b	$BF					; 
		dc.b	$95					; 
		dc.b	$50					; P
		dc.b	9
		dc.b	$11
		dc.b	$8B					; 
		dc.b	$C8					; 
		dc.b	$CC					; 
		dc.b	$A7					; 
		dc.b	$64					; d
		dc.b	$2C					; ,
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$BE					; 
		dc.b	$B4					; 
		dc.b	$6B					; k
		dc.b	$2D					; -
		dc.b	$B
		dc.b	$34					; 4
		dc.b	$BB					; 
		dc.b	$FE					; 
		dc.b	$F7					; 
		dc.b	$A3					; 
		dc.b	$6F					; o
		dc.b	$27					; '
		dc.b	$12
		dc.b	$45					; E
		dc.b	$72					; r
		dc.b	$B4					; 
		dc.b	$9D					; 
		dc.b	$A9					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$A9					; 
		dc.b	$DB					; 
		dc.b	$BE					; 
		dc.b	$9B					; 
		dc.b	$73					; s
		dc.b	$26					; &
		dc.b	6
		dc.b	$64					; d
		dc.b	$D2					; 
		dc.b	$EC					; 
		dc.b	$C4					; 
		dc.b	$83					; 
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$4C					; L
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$5D					; ]
		dc.b	$3F					; ?
		dc.b	$48					; H
		dc.b	$3E					; >
		dc.b	$64					; d
		dc.b	$B8					; 
		dc.b	$BF					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$60					; `
		dc.b	$30					; 0
		dc.b	$1E
		dc.b	$5D					; ]
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$AD					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$4F					; O
		dc.b	$3A					; :
		dc.b	$49					; I
		dc.b	$84					; 
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$96					; 
		dc.b	$E1					; 
		dc.b	$D9					; 
		dc.b	$A6					; 
		dc.b	$75					; u
		dc.b	$38					; 8
		dc.b	$2C					; ,
		dc.b	$6F					; o
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$D7					; 
		dc.b	$CC					; 
		dc.b	$6E					; n
		dc.b	$4C					; L
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$80					; 
		dc.b	$B8					; 
		dc.b	$C8					; 
		dc.b	$84					; 
		dc.b	$3F					; ?
		dc.b	$1C
		dc.b	$3C					; <
		dc.b	$8B					; 
		dc.b	$B7					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$78					; x
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$C9					; 
		dc.b	$FE					; 
		dc.b	$BF					; 
		dc.b	$6B					; k
		dc.b	$38					; 8
		dc.b	$45					; E
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$6B					; k
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$B6					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$59					; Y
		dc.b	$20
		dc.b	$26					; &
		dc.b	$6E					; n
		dc.b	$9F					; 
		dc.b	$B0					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$94					; 
		dc.b	$B4					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$38					; 8
		dc.b	$3E					; >
		dc.b	$79					; y
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$AA					; 
		dc.b	$96					; 
		dc.b	$4D					; M
		dc.b	4
		dc.b	$20
		dc.b	$75					; u
		dc.b	$B4					; 
		dc.b	$BD					; 
		dc.b	$BD					; 
		dc.b	$8B					; 
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$3E					; >
		dc.b	$60					; `
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$49					; I
		dc.b	$2E					; .
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$95					; 
		dc.b	$B6					; 
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$52					; R
		dc.b	$A2					; 
		dc.b	$C9					; 
		dc.b	$B0					; 
		dc.b	$76					; v
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$53					; S
		dc.b	$5C					; \
		dc.b	$68					; h
		dc.b	$87					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$4A					; J
		dc.b	$12
		dc.b	$50					; P
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$B3					; 
		dc.b	$B8					; 
		dc.b	$9F					; 
		dc.b	$6A					; j
		dc.b	$29					; )
		dc.b	$2C					; ,
		dc.b	$72					; r
		dc.b	$A3					; 
		dc.b	$B4					; 
		dc.b	$A2					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$B4					; 
		dc.b	$C8					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$34					; 4
		dc.b	$3F					; ?
		dc.b	$71					; q
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$D2					; 
		dc.b	$B8					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$52					; R
		dc.b	$2C					; ,
		dc.b	$35					; 5
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$B4					; 
		dc.b	$A7					; 
		dc.b	$B0					; 
		dc.b	$81					; 
		dc.b	$5D					; ]
		dc.b	$49					; I
		dc.b	$41					; A
		dc.b	$35					; 5
		dc.b	$60					; `
		dc.b	$B0					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$59					; Y
		dc.b	$3F					; ?
		dc.b	$5E					; ^
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$B1					; 
		dc.b	$CB					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$6E					; n
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$B0					; 
		dc.b	$9B					; 
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$B3					; 
		dc.b	$99					; 
		dc.b	$6C					; l
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$99					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$49					; I
		dc.b	$56					; V
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$61					; a
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$A2					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$8F					; 
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$73					; s
		dc.b	$A7					; 
		dc.b	$CC					; 
		dc.b	$9B					; 
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$59					; Y
		dc.b	$9F					; 
		dc.b	$BE					; 
		dc.b	$C2					; 
		dc.b	$87					; 
		dc.b	$4C					; L
		dc.b	$3C					; <
		dc.b	$4C					; L
		dc.b	$84					; 
		dc.b	$B6					; 
		dc.b	$CF					; 
		dc.b	$CD					; 
		dc.b	$79					; y
		dc.b	$33					; 3
		dc.b	$3E					; >
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$33					; 3
		dc.b	$29					; )
		dc.b	$A4					; 
		dc.b	$D0					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$9F					; 
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$64					; d
		dc.b	$37					; 7
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$A4					; 
		dc.b	$D4					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$63					; c
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$3E					; >
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$D9					; 
		dc.b	$C9					; 
		dc.b	$A0					; 
		dc.b	$6C					; l
		dc.b	$2C					; ,
		dc.b	$29					; )
		dc.b	$5E					; ^
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$35					; 5
		dc.b	$5D					; ]
		dc.b	$A7					; 
		dc.b	$D9					; 
		dc.b	$C5					; 
		dc.b	$98					; 
		dc.b	$46					; F
		dc.b	$23					; #
		dc.b	$3A					; :
		dc.b	$52					; R
		dc.b	$AB					; 
		dc.b	$FE					; 
		dc.b	$C4					; 
		dc.b	$41					; A
		dc.b	$11
		dc.b	$41					; A
		dc.b	$8F					; 
		dc.b	$B4					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$A9					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$2C					; ,
		dc.b	$2E					; .
		dc.b	$73					; s
		dc.b	$94					; 
		dc.b	$C2					; 
		dc.b	$D2					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$3C					; <
		dc.b	$2E					; .
		dc.b	$5A					; Z
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$41					; A
		dc.b	$6E					; n
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$6B					; k
		dc.b	$22					; "
		dc.b	$15
		dc.b	$71					; q
		dc.b	$A0					; 
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$96					; 
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$B0					; 
		dc.b	$B1					; 
		dc.b	$72					; r
		dc.b	$2A					; *
		dc.b	$38					; 8
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$94					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$55					; U
		dc.b	$2A					; *
		dc.b	$41					; A
		dc.b	$7A					; z
		dc.b	$AD					; 
		dc.b	$CD					; 
		dc.b	$C6					; 
		dc.b	$91					; 
		dc.b	$45					; E
		dc.b	$45					; E
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$4D					; M
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$84					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$56					; V
		dc.b	$96					; 
		dc.b	$BE					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$49					; I
		dc.b	$2E					; .
		dc.b	$4A					; J
		dc.b	$8A					; 
		dc.b	$BF					; 
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$AD					; 
		dc.b	$61					; a
		dc.b	$3E					; >
		dc.b	$6A					; j
		dc.b	$B0					; 
		dc.b	$A0					; 
		dc.b	$61					; a
		dc.b	$53					; S
		dc.b	$52					; R
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$A3					; 
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$3F					; ?
		dc.b	$34					; 4
		dc.b	$5C					; \
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$E8					; 
		dc.b	$68					; h
		dc.b	$31					; 1
		dc.b	$3A					; :
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$57					; W
		dc.b	$30					; 0
		dc.b	$55					; U
		dc.b	$AE					; 
		dc.b	$B6					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$2A					; *
		dc.b	$2E					; .
		dc.b	$5D					; ]
		dc.b	$A0					; 
		dc.b	$BB					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$83					; 
		dc.b	$61					; a
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$B6					; 
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$4C					; L
		dc.b	$48					; H
		dc.b	$7D					; }
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$8F					; 
		dc.b	$64					; d
		dc.b	$45					; E
		dc.b	$6C					; l
		dc.b	$B6					; 
		dc.b	$A6					; 
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$79					; y
		dc.b	$52					; R
		dc.b	$4C					; L
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$64					; d
		dc.b	$A0					; 
		dc.b	$C4					; 
		dc.b	$CD					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$45					; E
		dc.b	$2C					; ,
		dc.b	$53					; S
		dc.b	$A0					; 
		dc.b	$C4					; 
		dc.b	$A7					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$50					; P
		dc.b	$49					; I
		dc.b	$48					; H
		dc.b	$75					; u
		dc.b	$B0					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$7F					; 
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$A3					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$29					; )
		dc.b	$2E					; .
		dc.b	$6A					; j
		dc.b	$AE					; 
		dc.b	$C5					; 
		dc.b	$BA					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$37					; 7
		dc.b	$33					; 3
		dc.b	$49					; I
		dc.b	$6F					; o
		dc.b	$BB					; 
		dc.b	$D2					; 
		dc.b	$C5					; 
		dc.b	$AB					; 
		dc.b	$7F					; 
		dc.b	$4F					; O
		dc.b	$38					; 8
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$83					; 
		dc.b	$56					; V
		dc.b	$35					; 5
		dc.b	$3E					; >
		dc.b	$86					; 
		dc.b	$C2					; 
		dc.b	$A7					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$9D					; 
		dc.b	$D0					; 
		dc.b	$D6					; 
		dc.b	$87					; 
		dc.b	$59					; Y
		dc.b	$1C
		dc.b	4
		dc.b	$3E					; >
		dc.b	$6F					; o
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A3					; 
		dc.b	$31					; 1
		dc.b	6
		dc.b	$2C					; ,
		dc.b	$35					; 5
		dc.b	$6F					; o
		dc.b	$B1					; 
		dc.b	$BA					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$6F					; o
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$6C					; l
		dc.b	$57					; W
		dc.b	$42					; B
		dc.b	$71					; q
		dc.b	$AD					; 
		dc.b	$AB					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$8B					; 
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$A0					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$4F					; O
		dc.b	$98					; 
		dc.b	$D3					; 
		dc.b	$B4					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$5A					; Z
		dc.b	$43					; C
		dc.b	$4D					; M
		dc.b	$6F					; o
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$76					; v
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$BD					; 
		dc.b	$DE					; 
		dc.b	$81					; 
		dc.b	$42					; B
		dc.b	$4C					; L
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$9D					; 
		dc.b	$D6					; 
		dc.b	$9B					; 
		dc.b	$59					; Y
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$D0					; 
		dc.b	$B4					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$CF					; 
		dc.b	$9D					; 
		dc.b	$38					; 8
		dc.b	$49					; I
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$CF					; 
		dc.b	$DD					; 
		dc.b	$A6					; 
		dc.b	$57					; W
		dc.b	$31					; 1
		dc.b	$56					; V
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$B0					; 
		dc.b	$CD					; 
		dc.b	$80					; 
		dc.b	$3F					; ?
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$52					; R
		dc.b	$57					; W
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$AA					; 
		dc.b	$C4					; 
		dc.b	$AE					; 
		dc.b	$3F					; ?
		dc.b	$72					; r
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$43					; C
		dc.b	$6F					; o
		dc.b	$9D					; 
		dc.b	$A7					; 
		dc.b	$95					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$A4					; 
		dc.b	$52					; R
		dc.b	$1C
		dc.b	$38					; 8
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$A4					; 
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$48					; H
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$C4					; 
		dc.b	$94					; 
		dc.b	$42					; B
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$46					; F
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$B1					; 
		dc.b	$C6					; 
		dc.b	$81					; 
		dc.b	$5A					; Z
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$50					; P
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$B3					; 
		dc.b	$B4					; 
		dc.b	$7F					; 
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$A2					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$48					; H
		dc.b	$43					; C
		dc.b	$73					; s
		dc.b	$C4					; 
		dc.b	$F5					; 
		dc.b	$BD					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$27					; '
		dc.b	$2E					; .
		dc.b	$9D					; 
		dc.b	$EB					; 
		dc.b	$C6					; 
		dc.b	$A6					; 
		dc.b	$7D					; }
		dc.b	$43					; C
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$CF					; 
		dc.b	$94					; 
		dc.b	$5C					; \
		dc.b	$23					; #
		dc.b	9
		dc.b	$4F					; O
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$BF					; 
		dc.b	$AE					; 
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$AB					; 
		dc.b	$95					; 
		dc.b	$34					; 4
		dc.b	$22					; "
		dc.b	$60					; `
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$84					; 
		dc.b	$A7					; 
		dc.b	$A7					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$61					; a
		dc.b	$9B					; 
		dc.b	$C8					; 
		dc.b	$BE					; 
		dc.b	$8A					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$60					; `
		dc.b	$63					; c
		dc.b	$98					; 
		dc.b	$6E					; n
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$50					; P
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$AA					; 
		dc.b	$C2					; 
		dc.b	$9F					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$50					; P
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$4A					; J
		dc.b	$45					; E
		dc.b	$92					; 
		dc.b	$D0					; 
		dc.b	$BB					; 
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$9B					; 
		dc.b	$CB					; 
		dc.b	$B7					; 
		dc.b	$6E					; n
		dc.b	$2E					; .
		dc.b	$5A					; Z
		dc.b	$92					; 
		dc.b	$B0					; 
		dc.b	$91					; 
		dc.b	$4F					; O
		dc.b	$43					; C
		dc.b	$61					; a
		dc.b	$95					; 
		dc.b	$C9					; 
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$AA					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$4D					; M
		dc.b	$41					; A
		dc.b	$86					; 
		dc.b	$BB					; 
		dc.b	$95					; 
		dc.b	$5A					; Z
		dc.b	$94					; 
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$AB					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$61					; a
		dc.b	$98					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$8D					; 
		dc.b	$55					; U
		dc.b	$42					; B
		dc.b	$5A					; Z
		dc.b	$7D					; }
		dc.b	$C6					; 
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$B1					; 
		dc.b	$7C					; |
		dc.b	$4D					; M
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$6E					; n
		dc.b	$AA					; 
		dc.b	$C9					; 
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$48					; H
		dc.b	$34					; 4
		dc.b	$6B					; k
		dc.b	$A0					; 
		dc.b	$AD					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$4F					; O
		dc.b	$43					; C
		dc.b	$57					; W
		dc.b	$71					; q
		dc.b	$BA					; 
		dc.b	$B7					; 
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	$27					; '
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$AD					; 
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$50					; P
		dc.b	$56					; V
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$59					; Y
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$73					; s
		dc.b	$B1					; 
		dc.b	$C4					; 
		dc.b	$A3					; 
		dc.b	$91					; 
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$31					; 1
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$CF					; 
		dc.b	$DA					; 
		dc.b	$AB					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$E
		dc.b	$2C					; ,
		dc.b	$C9					; 
		dc.b	$D9					; 
		dc.b	$B0					; 
		dc.b	$AD					; 
		dc.b	$B8					; 
		dc.b	$80					; 
		dc.b	$1C
		dc.b	$3C					; <
		dc.b	$68					; h
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$5C					; \
		dc.b	$68					; h
		dc.b	$B4					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$53					; S
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$9D					; 
		dc.b	$C2					; 
		dc.b	$B8					; 
		dc.b	$8F					; 
		dc.b	$38					; 8
		dc.b	$45					; E
		dc.b	$8F					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$B3					; 
		dc.b	$7F					; 
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$B3					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$3E					; >
		dc.b	$4C					; L
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$30					; 0
		dc.b	$43					; C
		dc.b	$92					; 
		dc.b	$C6					; 
		dc.b	$E2					; 
		dc.b	$B1					; 
		dc.b	$67					; g
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$B3					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$B6					; 
		dc.b	$5C					; \
		dc.b	$3E					; >
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$33					; 3
		dc.b	$1B
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$C1					; 
		dc.b	$CF					; 
		dc.b	$83					; 
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$53					; S
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$B3					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$B4					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$2D					; -
		dc.b	$3A					; :
		dc.b	$4D					; M
		dc.b	$4D					; M
		dc.b	$79					; y
		dc.b	$BB					; 
		dc.b	$F0					; 
		dc.b	$EB					; 
		dc.b	$7D					; }
		dc.b	$25					; %
		dc.b	$15
		dc.b	$2A					; *
		dc.b	$88					; 
		dc.b	$B1					; 
		dc.b	$B4					; 
		dc.b	$A7					; 
		dc.b	$83					; 
		dc.b	$31					; 1
		dc.b	$37					; 7
		dc.b	$81					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$79					; y
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$3C					; <
		dc.b	$19
		dc.b	$6A					; j
		dc.b	$BE					; 
		dc.b	$CD					; 
		dc.b	$94					; 
		dc.b	$42					; B
		dc.b	4
		dc.b	$33					; 3
		dc.b	$86					; 
		dc.b	$BD					; 
		dc.b	$C9					; 
		dc.b	$D6					; 
		dc.b	$C9					; 
		dc.b	$6A					; j
		dc.b	$19
		dc.b	$2E					; .
		dc.b	$7A					; z
		dc.b	$BA					; 
		dc.b	$A9					; 
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$8D					; 
		dc.b	$B4					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$3E					; >
		dc.b	$29					; )
		dc.b	$38					; 8
		dc.b	$76					; v
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$C1					; 
		dc.b	$49					; I
		dc.b	6
		dc.b	$22					; "
		dc.b	$7C					; |
		dc.b	$EC					; 
		dc.b	$FE					; 
		dc.b	$A7					; 
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$A9					; 
		dc.b	$DB					; 
		dc.b	$B1					; 
		dc.b	$49					; I
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$55					; U
		dc.b	$94					; 
		dc.b	$DD					; 
		dc.b	$EF					; 
		dc.b	$C2					; 
		dc.b	$95					; 
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$4A					; J
		dc.b	$42					; B
		dc.b	$57					; W
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$B8					; 
		dc.b	$BE					; 
		dc.b	$B1					; 
		dc.b	$81					; 
		dc.b	$59					; Y
		dc.b	$10
		dc.b	1
		dc.b	1
		dc.b	$56					; V
		dc.b	$BD					; 
		dc.b	$BB					; 
		dc.b	$C1					; 
		dc.b	$8F					; 
		dc.b	$48					; H
		dc.b	$38					; 8
		dc.b	$34					; 4
		dc.b	$55					; U
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$9D					; 
		dc.b	$CC					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$52					; R
		dc.b	$46					; F
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$C9					; 
		dc.b	$BA					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$2C					; ,
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$A4					; 
		dc.b	$68					; h
		dc.b	$20
		dc.b	$41					; A
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$48					; H
		dc.b	$4D					; M
		dc.b	$59					; Y
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$6B					; k
		dc.b	$38					; 8
		dc.b	$48					; H
		dc.b	$79					; y
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$7A					; z
		dc.b	$A6					; 
		dc.b	$A0					; 
		dc.b	$AA					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$C9					; 
		dc.b	$C6					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$55					; U
		dc.b	$3C					; <
		dc.b	$5A					; Z
		dc.b	$8B					; 
		dc.b	$B8					; 
		dc.b	$C8					; 
		dc.b	$A7					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$3C					; <
		dc.b	$43					; C
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$6B					; k
		dc.b	$4A					; J
		dc.b	$2C					; ,
		dc.b	$41					; A
		dc.b	$53					; S
		dc.b	$7D					; }
		dc.b	$B4					; 
		dc.b	$C5					; 
		dc.b	$A6					; 
		dc.b	$50					; P
		dc.b	$2D					; -
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$65					; e
		dc.b	$A6					; 
		dc.b	$C8					; 
		dc.b	$C4					; 
		dc.b	$A9					; 
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$B1					; 
		dc.b	$75					; u
		dc.b	$4F					; O
		dc.b	$5D					; ]
		dc.b	$6F					; o
		dc.b	$4D					; M
		dc.b	$4C					; L
		dc.b	$88					; 
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$35					; 5
		dc.b	$27					; '
		dc.b	$45					; E
		dc.b	$7D					; }
		dc.b	$D9					; 
		dc.b	$DB					; 
		dc.b	$B1					; 
		dc.b	$72					; r
		dc.b	$3C					; <
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$9D					; 
		dc.b	$D0					; 
		dc.b	$9B					; 
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$7C					; |
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$6C					; l
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$B0					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$46					; F
		dc.b	$50					; P
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$27					; '
		dc.b	$E
		dc.b	$57					; W
		dc.b	$AD					; 
		dc.b	$BE					; 
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$84					; 
		dc.b	$34					; 4
		dc.b	$17
		dc.b	$3C					; <
		dc.b	$64					; d
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$B4					; 
		dc.b	$DD					; 
		dc.b	$BF					; 
		dc.b	$8D					; 
		dc.b	$4A					; J
		dc.b	$61					; a
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$8D					; 
		dc.b	$A7					; 
		dc.b	$76					; v
		dc.b	$37					; 7
		dc.b	$65					; e
		dc.b	$A7					; 
		dc.b	$8E					; 
		dc.b	$5E					; ^
		dc.b	$94					; 
		dc.b	$A9					; 
		dc.b	$53					; S
		dc.b	$14
		dc.b	$29					; )
		dc.b	$45					; E
		dc.b	$88					; 
		dc.b	$B7					; 
		dc.b	$BF					; 
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$79					; y
		dc.b	$1E
		dc.b	$1F
		dc.b	$3B					; ;
		dc.b	$59					; Y
		dc.b	$96					; 
		dc.b	$C4					; 
		dc.b	$D7					; 
		dc.b	$C6					; 
		dc.b	$91					; 
		dc.b	$4A					; J
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$76					; v
		dc.b	$3F					; ?
		dc.b	$57					; W
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$BD					; 
		dc.b	$A0					; 
		dc.b	$68					; h
		dc.b	$49					; I
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$6B					; k
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$6E					; n
		dc.b	$3A					; :
		dc.b	$50					; P
		dc.b	$57					; W
		dc.b	$6F					; o
		dc.b	$B0					; 
		dc.b	$D4					; 
		dc.b	$95					; 
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$53					; S
		dc.b	$42					; B
		dc.b	$4C					; L
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$52					; R
		dc.b	$65					; e
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$4F					; O
		dc.b	$3E					; >
		dc.b	$55					; U
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$4A					; J
		dc.b	$5E					; ^
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$60					; `
		dc.b	$64					; d
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$96					; 
		dc.b	$BF					; 
		dc.b	$AD					; 
		dc.b	$71					; q
		dc.b	$53					; S
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$98					; 
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$53					; S
		dc.b	$53					; S
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$91					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$87					; 
		dc.b	$B7					; 
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$4A					; J
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$84					; 
		dc.b	$C4					; 
		dc.b	$D4					; 
		dc.b	$94					; 
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$3F					; ?
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$87					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$53					; S
		dc.b	$3F					; ?
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$B8					; 
		dc.b	$B1					; 
		dc.b	$92					; 
		dc.b	$6C					; l
		dc.b	$20
		dc.b	$22					; "
		dc.b	$75					; u
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$A2					; 
		dc.b	$BE					; 
		dc.b	$92					; 
		dc.b	$60					; `
		dc.b	$87					; 
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$4C					; L
		dc.b	$5D					; ]
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$C8					; 
		dc.b	$AD					; 
		dc.b	$8F					; 
		dc.b	$5A					; Z
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$3C					; <
		dc.b	$52					; R
		dc.b	$72					; r
		dc.b	$BE					; 
		dc.b	$F3					; 
		dc.b	$CC					; 
		dc.b	$6B					; k
		dc.b	$3A					; :
		dc.b	$30					; 0
		dc.b	$2A					; *
		dc.b	$63					; c
		dc.b	$C1					; 
		dc.b	$FE					; 
		dc.b	$D0					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$37					; 7
		dc.b	$3B					; ;
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$95					; 
		dc.b	$E7					; 
		dc.b	$E4					; 
		dc.b	$A6					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$22					; "
		dc.b	$3C					; <
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$C6					; 
		dc.b	$B1					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$4A					; J
		dc.b	$14
		dc.b	$57					; W
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$AB					; 
		dc.b	$B7					; 
		dc.b	$86					; 
		dc.b	$52					; R
		dc.b	$41					; A
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$B0					; 
		dc.b	$CC					; 
		dc.b	$CF					; 
		dc.b	$8E					; 
		dc.b	$35					; 5
		dc.b	$12
		dc.b	$26					; &
		dc.b	$4C					; L
		dc.b	$92					; 
		dc.b	$BF					; 
		dc.b	$F2					; 
		dc.b	$C2					; 
		dc.b	$59					; Y
		dc.b	$18
		dc.b	$30					; 0
		dc.b	$57					; W
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$C5					; 
		dc.b	$A4					; 
		dc.b	$4A					; J
		dc.b	$38					; 8
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$9C					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$CF					; 
		dc.b	$D6					; 
		dc.b	$BD					; 
		dc.b	$60					; `
		dc.b	9
		dc.b	6
		dc.b	$38					; 8
		dc.b	$8D					; 
		dc.b	$D6					; 
		dc.b	$EE					; 
		dc.b	$E7					; 
		dc.b	$B1					; 
		dc.b	$6B					; k
		dc.b	$3A					; :
		dc.b	$38					; 8
		dc.b	$67					; g
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$A3					; 
		dc.b	$67					; g
		dc.b	$1F
		dc.b	$22					; "
		dc.b	$5A					; Z
		dc.b	$A3					; 
		dc.b	$D3					; 
		dc.b	$BE					; 
		dc.b	$8A					; 
		dc.b	$65					; e
		dc.b	$33					; 3
		dc.b	$30					; 0
		dc.b	$3E					; >
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$9C					; 
		dc.b	$A9					; 
		dc.b	$98					; 
		dc.b	$56					; V
		dc.b	$43					; C
		dc.b	$41					; A
		dc.b	$76					; v
		dc.b	$AD					; 
		dc.b	$BE					; 
		dc.b	$A3					; 
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$5D					; ]
		dc.b	$4C					; L
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$79					; y
		dc.b	$37					; 7
		dc.b	$37					; 7
		dc.b	$37					; 7
		dc.b	$53					; S
		dc.b	$8B					; 
		dc.b	$B3					; 
		dc.b	$BB					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$8E					; 
		dc.b	$56					; V
		dc.b	$52					; R
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$8E					; 
		dc.b	$A6					; 
		dc.b	$A3					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$A4					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$2E					; .
		dc.b	$2D					; -
		dc.b	$50					; P
		dc.b	$A3					; 
		dc.b	$C9					; 
		dc.b	$CF					; 
		dc.b	$A4					; 
		dc.b	$4D					; M
		dc.b	$22					; "
		dc.b	$3F					; ?
		dc.b	$71					; q
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$86					; 
		dc.b	$3F					; ?
		dc.b	$35					; 5
		dc.b	$5D					; ]
		dc.b	$A2					; 
		dc.b	$B3					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$65					; e
		dc.b	$84					; 
		dc.b	$B7					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$9D					; 
		dc.b	$C1					; 
		dc.b	$9D					; 
		dc.b	$67					; g
		dc.b	$48					; H
		dc.b	$2E					; .
		dc.b	$38					; 8
		dc.b	$7D					; }
		dc.b	$B7					; 
		dc.b	$DB					; 
		dc.b	$D9					; 
		dc.b	$B3					; 
		dc.b	$67					; g
		dc.b	$3A					; :
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$88					; 
		dc.b	$AA					; 
		dc.b	$C9					; 
		dc.b	$B4					; 
		dc.b	$71					; q
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$BB					; 
		dc.b	$A9					; 
		dc.b	$75					; u
		dc.b	$57					; W
		dc.b	$4C					; L
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$5C					; \
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$91					; 
		dc.b	$A9					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$78					; x
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$59					; Y
		dc.b	$45					; E
		dc.b	$6F					; o
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$A3					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$45					; E
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$A6					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$A7					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$3A					; :
		dc.b	$17
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$B1					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$4F					; O
		dc.b	$1F
		dc.b	$5E					; ^
		dc.b	$B1					; 
		dc.b	$E1					; 
		dc.b	$D7					; 
		dc.b	$8B					; 
		dc.b	$55					; U
		dc.b	$2A					; *
		dc.b	$34					; 4
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$AB					; 
		dc.b	$D7					; 
		dc.b	$A4					; 
		dc.b	$38					; 8
		dc.b	$D
		dc.b	$3E					; >
		dc.b	$6B					; k
		dc.b	$9C					; 
		dc.b	$E2					; 
		dc.b	$E9					; 
		dc.b	$B3					; 
		dc.b	$63					; c
		dc.b	$2E					; .
		dc.b	$45					; E
		dc.b	$81					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$4A					; J
		dc.b	$49					; I
		dc.b	$72					; r
		dc.b	$8F					; 
		dc.b	$A7					; 
		dc.b	$99					; 
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$9C					; 
		dc.b	$BD					; 
		dc.b	$9D					; 
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$B8					; 
		dc.b	$98					; 
		dc.b	$68					; h
		dc.b	$56					; V
		dc.b	$41					; A
		dc.b	$3A					; :
		dc.b	$6E					; n
		dc.b	$CD					; 
		dc.b	$D7					; 
		dc.b	$8D					; 
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$6B					; k
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$A2					; 
		dc.b	$99					; 
		dc.b	$C2					; 
		dc.b	$B0					; 
		dc.b	$6B					; k
		dc.b	$4D					; M
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$59					; Y
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$5D					; ]
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$57					; W
		dc.b	$3E					; >
		dc.b	$6A					; j
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$A7					; 
		dc.b	$A7					; 
		dc.b	$BB					; 
		dc.b	$A9					; 
		dc.b	$4D					; M
		dc.b	$31					; 1
		dc.b	$6F					; o
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$57					; W
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$A2					; 
		dc.b	$9F					; 
		dc.b	$79					; y
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$B7					; 
		dc.b	$CC					; 
		dc.b	$B6					; 
		dc.b	$8A					; 
		dc.b	$56					; V
		dc.b	$26					; &
		dc.b	$34					; 4
		dc.b	$67					; g
		dc.b	$A7					; 
		dc.b	$B4					; 
		dc.b	$86					; 
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$84					; 
		dc.b	$B4					; 
		dc.b	$CB					; 
		dc.b	$B0					; 
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$3A					; :
		dc.b	$48					; H
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$9F					; 
		dc.b	$95					; 
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$AA					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$AA					; 
		dc.b	$92					; 
		dc.b	$55					; U
		dc.b	$46					; F
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$5C					; \
		dc.b	$34					; 4
		dc.b	$45					; E
		dc.b	$8D					; 
		dc.b	$D4					; 
		dc.b	$DA					; 
		dc.b	$AB					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$41					; A
		dc.b	$79					; y
		dc.b	$BB					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$43					; C
		dc.b	$42					; B
		dc.b	$61					; a
		dc.b	$87					; 
		dc.b	$A4					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$A2					; 
		dc.b	$A6					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$4C					; L
		dc.b	$17
		dc.b	$4D					; M
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$53					; S
		dc.b	$33					; 3
		dc.b	$6C					; l
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$D9					; 
		dc.b	$E1					; 
		dc.b	$B8					; 
		dc.b	$76					; v
		dc.b	$43					; C
		dc.b	$E
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$91					; 
		dc.b	$63					; c
		dc.b	$2C					; ,
		dc.b	$31					; 1
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$BD					; 
		dc.b	$E7					; 
		dc.b	$DB					; 
		dc.b	$83					; 
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$56					; V
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$9F					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$B3					; 
		dc.b	$D9					; 
		dc.b	$A2					; 
		dc.b	$50					; P
		dc.b	$30					; 0
		dc.b	$22					; "
		dc.b	9
		dc.b	$34					; 4
		dc.b	$92					; 
		dc.b	$D0					; 
		dc.b	$E2					; 
		dc.b	$A3					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$AB					; 
		dc.b	$B4					; 
		dc.b	$C1					; 
		dc.b	$A7					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$98					; 
		dc.b	$5D					; ]
		dc.b	$1B
		dc.b	$D
		dc.b	$64					; d
		dc.b	$BA					; 
		dc.b	$CB					; 
		dc.b	$D7					; 
		dc.b	$CB					; 
		dc.b	$98					; 
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$63					; c
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$9F					; 
		dc.b	$68					; h
		dc.b	$3A					; :
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$8A					; 
		dc.b	$BB					; 
		dc.b	$C1					; 
		dc.b	$84					; 
		dc.b	$4D					; M
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$95					; 
		dc.b	$D0					; 
		dc.b	$91					; 
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$56					; V
		dc.b	$1F
		dc.b	$6C					; l
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$6A					; j
		dc.b	$8A					; 
		dc.b	$C2					; 
		dc.b	$E0					; 
		dc.b	$C8					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$46					; F
		dc.b	$26					; &
		dc.b	$5C					; \
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$55					; U
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$B8					; 
		dc.b	$BF					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$41					; A
		dc.b	$3C					; <
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$A3					; 
		dc.b	$D9					; 
		dc.b	$BE					; 
		dc.b	$A6					; 
		dc.b	$78					; x
		dc.b	$2D					; -
		dc.b	$3F					; ?
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$8E					; 
		dc.b	$6A					; j
		dc.b	$43					; C
		dc.b	$29					; )
		dc.b	$55					; U
		dc.b	$AD					; 
		dc.b	$CC					; 
		dc.b	$A2					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$43					; C
		dc.b	$4D					; M
		dc.b	$7D					; }
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$5E					; ^
		dc.b	$59					; Y
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$4C					; L
		dc.b	$35					; 5
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$5A					; Z
		dc.b	$73					; s
		dc.b	$B3					; 
		dc.b	$B7					; 
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$50					; P
		dc.b	$61					; a
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$75					; u
		dc.b	$C2					; 
		dc.b	$A2					; 
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$43					; C
		dc.b	$37					; 7
		dc.b	$63					; c
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$BF					; 
		dc.b	$CC					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$50					; P
		dc.b	$1B
		dc.b	$3E					; >
		dc.b	$8B					; 
		dc.b	$AA					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$91					; 
		dc.b	$B8					; 
		dc.b	$9B					; 
		dc.b	$4F					; O
		dc.b	$4C					; L
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$81					; 
		dc.b	$CB					; 
		dc.b	$B3					; 
		dc.b	$5A					; Z
		dc.b	$2A					; *
		dc.b	$3C					; <
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$CD					; 
		dc.b	$E1					; 
		dc.b	$A0					; 
		dc.b	$46					; F
		dc.b	$42					; B
		dc.b	$6A					; j
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$5A					; Z
		dc.b	$3C					; <
		dc.b	$56					; V
		dc.b	$AB					; 
		dc.b	$BB					; 
		dc.b	$9B					; 
		dc.b	$6B					; k
		dc.b	$38					; 8
		dc.b	$71					; q
		dc.b	$BA					; 
		dc.b	$CD					; 
		dc.b	$AA					; 
		dc.b	$88					; 
		dc.b	$41					; A
		dc.b	$4D					; M
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$92					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$95					; 
		dc.b	$BE					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$7C					; |
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$75					; u
		dc.b	$45					; E
		dc.b	$2A					; *
		dc.b	$3F					; ?
		dc.b	$7F					; 
		dc.b	$B0					; 
		dc.b	$BA					; 
		dc.b	$AA					; 
		dc.b	$6A					; j
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$AB					; 
		dc.b	$8E					; 
		dc.b	$68					; h
		dc.b	$34					; 4
		dc.b	$2A					; *
		dc.b	$84					; 
		dc.b	$D0					; 
		dc.b	$AE					; 
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$48					; H
		dc.b	$B1					; 
		dc.b	$D9					; 
		dc.b	$A9					; 
		dc.b	$7A					; z
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$73					; s
		dc.b	$BA					; 
		dc.b	$9D					; 
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$A9					; 
		dc.b	$C4					; 
		dc.b	$8B					; 
		dc.b	$5D					; ]
		dc.b	$49					; I
		dc.b	$48					; H
		dc.b	$5E					; ^
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$B1					; 
		dc.b	$BB					; 
		dc.b	$A3					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$5D					; ]
		dc.b	$52					; R
		dc.b	$72					; r
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$67					; g
		dc.b	$9F					; 
		dc.b	$C1					; 
		dc.b	$BB					; 
		dc.b	$8D					; 
		dc.b	$63					; c
		dc.b	$3A					; :
		dc.b	$38					; 8
		dc.b	$5E					; ^
		dc.b	$AB					; 
		dc.b	$E9					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$45					; E
		dc.b	$3E					; >
		dc.b	$5A					; Z
		dc.b	$73					; s
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$86					; 
		dc.b	$5A					; Z
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$53					; S
		dc.b	$50					; P
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$AD					; 
		dc.b	$B7					; 
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$7C					; |
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$31					; 1
		dc.b	$34					; 4
		dc.b	$57					; W
		dc.b	$B6					; 
		dc.b	$B7					; 
		dc.b	$A4					; 
		dc.b	$99					; 
		dc.b	$71					; q
		dc.b	$33					; 3
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$95					; 
		dc.b	$D7					; 
		dc.b	$A2					; 
		dc.b	$52					; R
		dc.b	$42					; B
		dc.b	$50					; P
		dc.b	$27					; '
		dc.b	$46					; F
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$C4					; 
		dc.b	$BD					; 
		dc.b	$81					; 
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$4D					; M
		dc.b	$71					; q
		dc.b	$D6					; 
		dc.b	$B4					; 
		dc.b	$53					; S
		dc.b	$4C					; L
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$AA					; 
		dc.b	$94					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$63					; c
		dc.b	$50					; P
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$A9					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$38					; 8
		dc.b	$63					; c
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$D2					; 
		dc.b	$D9					; 
		dc.b	$80					; 
		dc.b	$29					; )
		dc.b	$34					; 4
		dc.b	$5A					; Z
		dc.b	$6B					; k
		dc.b	$95					; 
		dc.b	$BD					; 
		dc.b	$92					; 
		dc.b	$53					; S
		dc.b	$3A					; :
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$8F					; 
		dc.b	$C4					; 
		dc.b	$99					; 
		dc.b	$6E					; n
		dc.b	$4A					; J
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$9F					; 
		dc.b	$C5					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$9F					; 
		dc.b	$76					; v
		dc.b	$42					; B
		dc.b	$29					; )
		dc.b	$4F					; O
		dc.b	$A3					; 
		dc.b	$C5					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$6C					; l
		dc.b	$91					; 
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$7F					; 
		dc.b	$9B					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$53					; S
		dc.b	$43					; C
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$8D					; 
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$99					; 
		dc.b	$A6					; 
		dc.b	$9C					; 
		dc.b	$B0					; 
		dc.b	$7D					; }
		dc.b	$45					; E
		dc.b	$55					; U
		dc.b	$86					; 
		dc.b	$68					; h
		dc.b	$48					; H
		dc.b	$59					; Y
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$91					; 
		dc.b	$AB					; 
		dc.b	$96					; 
		dc.b	$AB					; 
		dc.b	$92					; 
		dc.b	$55					; U
		dc.b	$46					; F
		dc.b	$73					; s
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$8A					; 
		dc.b	$C9					; 
		dc.b	$96					; 
		dc.b	$88					; 
		dc.b	$B4					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$7A					; z
		dc.b	$46					; F
		dc.b	$33					; 3
		dc.b	$49					; I
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$BF					; 
		dc.b	$9F					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$61					; a
		dc.b	$22					; "
		dc.b	$45					; E
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$8D					; 
		dc.b	$BF					; 
		dc.b	$D3					; 
		dc.b	$80					; 
		dc.b	$3C					; <
		dc.b	$4F					; O
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$98					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$8A					; 
		dc.b	$4C					; L
		dc.b	$2D					; -
		dc.b	$4A					; J
		dc.b	$80					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$61					; a
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$9C					; 
		dc.b	$BA					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$59					; Y
		dc.b	$57					; W
		dc.b	$3A					; :
		dc.b	$5A					; Z
		dc.b	$A3					; 
		dc.b	$B3					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$53					; S
		dc.b	$3E					; >
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$80					; 
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$AD					; 
		dc.b	$DA					; 
		dc.b	$B6					; 
		dc.b	$83					; 
		dc.b	$49					; I
		dc.b	$37					; 7
		dc.b	$55					; U
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$3A					; :
		dc.b	$7F					; 
		dc.b	$D0					; 
		dc.b	$BF					; 
		dc.b	$95					; 
		dc.b	$65					; e
		dc.b	$49					; I
		dc.b	$43					; C
		dc.b	$52					; R
		dc.b	$63					; c
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$B4					; 
		dc.b	$A4					; 
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$86					; 
		dc.b	$9D					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$46					; F
		dc.b	$5C					; \
		dc.b	$57					; W
		dc.b	$9D					; 
		dc.b	$BA					; 
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$A4					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$94					; 
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$8B					; 
		dc.b	$C1					; 
		dc.b	$A2					; 
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$7C					; |
		dc.b	$3B					; ;
		dc.b	$2E					; .
		dc.b	$61					; a
		dc.b	$AB					; 
		dc.b	$C2					; 
		dc.b	$BB					; 
		dc.b	$86					; 
		dc.b	$3F					; ?
		dc.b	$38					; 8
		dc.b	$57					; W
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$42					; B
		dc.b	$48					; H
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$A9					; 
		dc.b	$A0					; 
		dc.b	$5D					; ]
		dc.b	$5C					; \
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$A7					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$3A					; :
		dc.b	$55					; U
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$BE					; 
		dc.b	$A4					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$50					; P
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$A6					; 
		dc.b	$C5					; 
		dc.b	$95					; 
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$95					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$B6					; 
		dc.b	$C5					; 
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$4D					; M
		dc.b	$31					; 1
		dc.b	$5E					; ^
		dc.b	$9F					; 
		dc.b	$C9					; 
		dc.b	$B3					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$50					; P
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$3F					; ?
		dc.b	$35					; 5
		dc.b	$60					; `
		dc.b	$55					; U
		dc.b	$48					; H
		dc.b	$96					; 
		dc.b	$D4					; 
		dc.b	$B6					; 
		dc.b	$7D					; }
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$8B					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$49					; I
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$A0					; 
		dc.b	$BE					; 
		dc.b	$9F					; 
		dc.b	$75					; u
		dc.b	$50					; P
		dc.b	$1F
		dc.b	$4F					; O
		dc.b	$AA					; 
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$B3					; 
		dc.b	$C5					; 
		dc.b	$AE					; 
		dc.b	$7D					; }
		dc.b	$5E					; ^
		dc.b	$48					; H
		dc.b	$5A					; Z
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$79					; y
		dc.b	$B3					; 
		dc.b	$BF					; 
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$59					; Y
		dc.b	$2E					; .
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$A9					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$64					; d
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$9B					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$8D					; 
		dc.b	$A4					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$41					; A
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$BE					; 
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$61					; a
		dc.b	$8B					; 
		dc.b	$BB					; 
		dc.b	$9B					; 
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$64					; d
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$AB					; 
		dc.b	$92					; 
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$56					; V
		dc.b	$59					; Y
		dc.b	$6E					; n
		dc.b	$AE					; 
		dc.b	$E0					; 
		dc.b	$BA					; 
		dc.b	$64					; d
		dc.b	$25					; %
		dc.b	$2C					; ,
		dc.b	$6B					; k
		dc.b	$A6					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$A7					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$43					; C
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$8F					; 
		dc.b	$DB					; 
		dc.b	$D0					; 
		dc.b	$7A					; z
		dc.b	$48					; H
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$96					; 
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$5D					; ]
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$75					; u
		dc.b	$48					; H
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$4F					; O
		dc.b	$8A					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$5A					; Z
		dc.b	$41					; A
		dc.b	$86					; 
		dc.b	$AD					; 
		dc.b	$A2					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$2E					; .
		dc.b	$56					; V
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$3F					; ?
		dc.b	$50					; P
		dc.b	$95					; 
		dc.b	$C5					; 
		dc.b	$A0					; 
		dc.b	$6A					; j
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$92					; 
		dc.b	$A7					; 
		dc.b	$96					; 
		dc.b	$80					; 
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$A0					; 
		dc.b	$49					; I
		dc.b	$31					; 1
		dc.b	$7D					; }
		dc.b	$99					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$49					; I
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$96					; 
		dc.b	$A6					; 
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$A3					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$3C					; <
		dc.b	$50					; P
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$C1					; 
		dc.b	$C8					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$50					; P
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$C4					; 
		dc.b	$8F					; 
		dc.b	$60					; `
		dc.b	$4D					; M
		dc.b	$46					; F
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$50					; P
		dc.b	$30					; 0
		dc.b	$27					; '
		dc.b	$64					; d
		dc.b	$AA					; 
		dc.b	$CB					; 
		dc.b	$CD					; 
		dc.b	$BB					; 
		dc.b	$98					; 
		dc.b	$27					; '
		dc.b	$B
		dc.b	$5D					; ]
		dc.b	$A4					; 
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$5A					; Z
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$53					; S
		dc.b	$65					; e
		dc.b	$A2					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$63					; c
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$7D					; }
		dc.b	$B6					; 
		dc.b	$C4					; 
		dc.b	$84					; 
		dc.b	$3E					; >
		dc.b	$3A					; :
		dc.b	$1B
		dc.b	$31					; 1
		dc.b	$9D					; 
		dc.b	$C9					; 
		dc.b	$BA					; 
		dc.b	$B8					; 
		dc.b	$9B					; 
		dc.b	$49					; I
		dc.b	$26					; &
		dc.b	$49					; I
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$A9					; 
		dc.b	$A7					; 
		dc.b	$7C					; |
		dc.b	$4D					; M
		dc.b	$48					; H
		dc.b	$49					; I
		dc.b	$6A					; j
		dc.b	$A0					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$A7					; 
		dc.b	$76					; v
		dc.b	$42					; B
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$67					; g
		dc.b	$46					; F
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$6A					; j
		dc.b	$59					; Y
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$94					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$B8					; 
		dc.b	$96					; 
		dc.b	$60					; `
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$63					; c
		dc.b	$63					; c
		dc.b	$96					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$4F					; O
		dc.b	$1E
		dc.b	$76					; v
		dc.b	$A7					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$A9					; 
		dc.b	$AA					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$49					; I
		dc.b	$53					; S
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$60					; `
		dc.b	$46					; F
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$5C					; \
		dc.b	$75					; u
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$80					; 
		dc.b	$AA					; 
		dc.b	$AA					; 
		dc.b	$9F					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$53					; S
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$4D					; M
		dc.b	$64					; d
		dc.b	$B7					; 
		dc.b	$E1					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$48					; H
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$95					; 
		dc.b	$B7					; 
		dc.b	$E0					; 
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$7C					; |
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$9D					; 
		dc.b	$87					; 
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$45					; E
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$B6					; 
		dc.b	$99					; 
		dc.b	$50					; P
		dc.b	$41					; A
		dc.b	$4C					; L
		dc.b	$56					; V
		dc.b	$87					; 
		dc.b	$B7					; 
		dc.b	$A0					; 
		dc.b	$75					; u
		dc.b	$50					; P
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$52					; R
		dc.b	$46					; F
		dc.b	$5A					; Z
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$88					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$49					; I
		dc.b	$3F					; ?
		dc.b	$60					; `
		dc.b	$91					; 
		dc.b	$B3					; 
		dc.b	$C2					; 
		dc.b	$C1					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$57					; W
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$7F					; 
		dc.b	$C1					; 
		dc.b	$A6					; 
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$98					; 
		dc.b	$AE					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$6C					; l
		dc.b	$9D					; 
		dc.b	$C2					; 
		dc.b	$9D					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$4A					; J
		dc.b	$3A					; :
		dc.b	$6A					; j
		dc.b	$A3					; 
		dc.b	$BD					; 
		dc.b	$C2					; 
		dc.b	$A7					; 
		dc.b	$57					; W
		dc.b	$26					; &
		dc.b	$52					; R
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$B3					; 
		dc.b	$B0					; 
		dc.b	$7C					; |
		dc.b	$65					; e
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$48					; H
		dc.b	$7F					; 
		dc.b	$A6					; 
		dc.b	$87					; 
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$94					; 
		dc.b	$AD					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$7A					; z
		dc.b	$B1					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$71					; q
		dc.b	$4C					; L
		dc.b	$4A					; J
		dc.b	$3F					; ?
		dc.b	$78					; x
		dc.b	$B6					; 
		dc.b	$CB					; 
		dc.b	$9B					; 
		dc.b	$64					; d
		dc.b	$4C					; L
		dc.b	$3A					; :
		dc.b	$65					; e
		dc.b	$9D					; 
		dc.b	$B8					; 
		dc.b	$B3					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$A6					; 
		dc.b	$BD					; 
		dc.b	$8F					; 
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$64					; d
		dc.b	$3C					; <
		dc.b	$33					; 3
		dc.b	$76					; v
		dc.b	$C9					; 
		dc.b	$BA					; 
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$4C					; L
		dc.b	$73					; s
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$3C					; <
		dc.b	$57					; W
		dc.b	$79					; y
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$84					; 
		dc.b	$5D					; ]
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$CB					; 
		dc.b	$C1					; 
		dc.b	$79					; y
		dc.b	$38					; 8
		dc.b	$20
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$BD					; 
		dc.b	$A6					; 
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$50					; P
		dc.b	$30					; 0
		dc.b	$6A					; j
		dc.b	$A9					; 
		dc.b	$8E					; 
		dc.b	$A7					; 
		dc.b	$C2					; 
		dc.b	$84					; 
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$8D					; 
		dc.b	$C1					; 
		dc.b	$80					; 
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$79					; y
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$B8					; 
		dc.b	$B1					; 
		dc.b	$6B					; k
		dc.b	$3F					; ?
		dc.b	$55					; U
		dc.b	$43					; C
		dc.b	$6B					; k
		dc.b	$BF					; 
		dc.b	$C2					; 
		dc.b	$A0					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$2D					; -
		dc.b	$33					; 3
		dc.b	$92					; 
		dc.b	$F5					; 
		dc.b	$E4					; 
		dc.b	$A7					; 
		dc.b	$65					; e
		dc.b	$1F
		dc.b	$1B
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$A4					; 
		dc.b	$B8					; 
		dc.b	$B4					; 
		dc.b	$83					; 
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$A9					; 
		dc.b	$D3					; 
		dc.b	$C1					; 
		dc.b	$72					; r
		dc.b	$3B					; ;
		dc.b	$3F					; ?
		dc.b	$41					; A
		dc.b	$60					; `
		dc.b	$A3					; 
		dc.b	$B1					; 
		dc.b	$9F					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$8B					; 
		dc.b	$59					; Y
		dc.b	$38					; 8
		dc.b	$38					; 8
		dc.b	$55					; U
		dc.b	$86					; 
		dc.b	$BB					; 
		dc.b	$CD					; 
		dc.b	$8B					; 
		dc.b	$61					; a
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$B6					; 
		dc.b	$A9					; 
		dc.b	$79					; y
		dc.b	$4D					; M
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$53					; S
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$A7					; 
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$8D					; 
		dc.b	$B3					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$63					; c
		dc.b	$94					; 
		dc.b	$A4					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$AA					; 
		dc.b	$A6					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$8B					; 
		dc.b	$C2					; 
		dc.b	$AA					; 
		dc.b	$72					; r
		dc.b	$48					; H
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$98					; 
		dc.b	$B4					; 
		dc.b	$AA					; 
		dc.b	$99					; 
		dc.b	$46					; F
		dc.b	$17
		dc.b	$38					; 8
		dc.b	$81					; 
		dc.b	$AB					; 
		dc.b	$C4					; 
		dc.b	$B7					; 
		dc.b	$75					; u
		dc.b	$4A					; J
		dc.b	$50					; P
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$8E					; 
		dc.b	$A9					; 
		dc.b	$BB					; 
		dc.b	$B0					; 
		dc.b	$83					; 
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$8B					; 
		dc.b	$53					; S
		dc.b	$5D					; ]
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$5E					; ^
		dc.b	$53					; S
		dc.b	$72					; r
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$D2					; 
		dc.b	$DA					; 
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	2
		dc.b	$23					; #
		dc.b	$87					; 
		dc.b	$B6					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$5E					; ^
		dc.b	$19
		dc.b	$23					; #
		dc.b	$68					; h
		dc.b	$B4					; 
		dc.b	$CB					; 
		dc.b	$AE					; 
		dc.b	$6E					; n
		dc.b	$49					; I
		dc.b	$49					; I
		dc.b	$63					; c
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$83					; 
		dc.b	$64					; d
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$6F					; o
		dc.b	$AE					; 
		dc.b	$D4					; 
		dc.b	$A2					; 
		dc.b	$52					; R
		dc.b	$43					; C
		dc.b	$5A					; Z
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$B3					; 
		dc.b	$AB					; 
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$A3					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$7A					; z
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$49					; I
		dc.b	$6B					; k
		dc.b	$AE					; 
		dc.b	$D2					; 
		dc.b	$AB					; 
		dc.b	$8A					; 
		dc.b	$6F					; o
		dc.b	$38					; 8
		dc.b	$33					; 3
		dc.b	$5A					; Z
		dc.b	$91					; 
		dc.b	$BB					; 
		dc.b	$BB					; 
		dc.b	$99					; 
		dc.b	$60					; `
		dc.b	$41					; A
		dc.b	$45					; E
		dc.b	$88					; 
		dc.b	$B3					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$A3					; 
		dc.b	$8E					; 
		dc.b	$5E					; ^
		dc.b	$35					; 5
		dc.b	$3F					; ?
		dc.b	$83					; 
		dc.b	$A2					; 
		dc.b	$95					; 
		dc.b	$A9					; 
		dc.b	$BA					; 
		dc.b	$75					; u
		dc.b	$38					; 8
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$6B					; k
		dc.b	$3A					; :
		dc.b	$4A					; J
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$B1					; 
		dc.b	$8D					; 
		dc.b	$4A					; J
		dc.b	$3E					; >
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$BB					; 
		dc.b	$BA					; 
		dc.b	$94					; 
		dc.b	$6B					; k
		dc.b	$49					; I
		dc.b	$3B					; ;
		dc.b	$60					; `
		dc.b	$AB					; 
		dc.b	$D4					; 
		dc.b	$C8					; 
		dc.b	$91					; 
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$3B					; ;
		dc.b	$67					; g
		dc.b	$AE					; 
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$98					; 
		dc.b	$6E					; n
		dc.b	$37					; 7
		dc.b	$3F					; ?
		dc.b	$76					; v
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$5C					; \
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$B0					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$52					; R
		dc.b	$79					; y
		dc.b	$9B					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$72					; r
		dc.b	$56					; V
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$59					; Y
		dc.b	$A9					; 
		dc.b	$BA					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$98					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$AD					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$5D					; ]
		dc.b	$2D					; -
		dc.b	$37					; 7
		dc.b	$7A					; z
		dc.b	$BA					; 
		dc.b	$C6					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$55					; U
		dc.b	$49					; I
		dc.b	$83					; 
		dc.b	$B0					; 
		dc.b	$B3					; 
		dc.b	$83					; 
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$5C					; \
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$A7					; 
		dc.b	$81					; 
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$F0					; 
		dc.b	$C4					; 
		dc.b	$5A					; Z
		dc.b	$3A					; :
		dc.b	$59					; Y
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$4A					; J
		dc.b	$6B					; k
		dc.b	$A6					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$63					; c
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$5A					; Z
		dc.b	$92					; 
		dc.b	$CB					; 
		dc.b	$95					; 
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$49					; I
		dc.b	$3C					; <
		dc.b	$7C					; |
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$AE					; 
		dc.b	$8E					; 
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$A9					; 
		dc.b	$D0					; 
		dc.b	$A2					; 
		dc.b	$59					; Y
		dc.b	$45					; E
		dc.b	$4C					; L
		dc.b	$53					; S
		dc.b	$6C					; l
		dc.b	$98					; 
		dc.b	$B6					; 
		dc.b	$BB					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$48					; H
		dc.b	$38					; 8
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$98					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$4C					; L
		dc.b	$34					; 4
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$AE					; 
		dc.b	$BB					; 
		dc.b	$91					; 
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$99					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$8B					; 
		dc.b	$BA					; 
		dc.b	$84					; 
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$B0					; 
		dc.b	$80					; 
		dc.b	$31					; 1
		dc.b	$4C					; L
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$78					; x
		dc.b	$4A					; J
		dc.b	$48					; H
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$A4					; 
		dc.b	$B3					; 
		dc.b	$9D					; 
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$48					; H
		dc.b	$6B					; k
		dc.b	$B0					; 
		dc.b	$B6					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$52					; R
		dc.b	$2A					; *
		dc.b	$46					; F
		dc.b	$9F					; 
		dc.b	$CC					; 
		dc.b	$A4					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$3B					; ;
		dc.b	$3B					; ;
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$A2					; 
		dc.b	$BB					; 
		dc.b	$92					; 
		dc.b	$57					; W
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$C5					; 
		dc.b	$BF					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$56					; V
		dc.b	$30					; 0
		dc.b	$14
		dc.b	$52					; R
		dc.b	$D3					; 
		dc.b	$FE					; 
		dc.b	$C1					; 
		dc.b	$6B					; k
		dc.b	$4F					; O
		dc.b	$26					; &
		dc.b	$2C					; ,
		dc.b	$6E					; n
		dc.b	$A3					; 
		dc.b	$BF					; 
		dc.b	$D4					; 
		dc.b	$B8					; 
		dc.b	$6A					; j
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$6C					; l
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$5E					; ^
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$61					; a
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$63					; c
		dc.b	$2A					; *
		dc.b	$52					; R
		dc.b	$AA					; 
		dc.b	$BF					; 
		dc.b	$88					; 
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$50					; P
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$7D					; }
		dc.b	$A6					; 
		dc.b	$91					; 
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$73					; s
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$A4					; 
		dc.b	$B8					; 
		dc.b	$A9					; 
		dc.b	$78					; x
		dc.b	$42					; B
		dc.b	$42					; B
		dc.b	$6F					; o
		dc.b	$6C					; l
		dc.b	$96					; 
		dc.b	$DB					; 
		dc.b	$AA					; 
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$6A					; j
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$7C					; |
		dc.b	$AB					; 
		dc.b	$9B					; 
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$60					; `
		dc.b	$6E					; n
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$AD					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$A6					; 
		dc.b	$BF					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$65					; e
		dc.b	$A6					; 
		dc.b	$A6					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$60					; `
		dc.b	$59					; Y
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$9B					; 
		dc.b	$B4					; 
		dc.b	$A2					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$68					; h
		dc.b	$A2					; 
		dc.b	$AD					; 
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$3B					; ;
		dc.b	$43					; C
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$96					; 
		dc.b	$BD					; 
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$5D					; ]
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$BE					; 
		dc.b	$B6					; 
		dc.b	$6E					; n
		dc.b	$3C					; <
		dc.b	$31					; 1
		dc.b	$3F					; ?
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$B4					; 
		dc.b	$B3					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$AD					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$4F					; O
		dc.b	$31					; 1
		dc.b	$4A					; J
		dc.b	$75					; u
		dc.b	$A4					; 
		dc.b	$BF					; 
		dc.b	$A7					; 
		dc.b	$A9					; 
		dc.b	$96					; 
		dc.b	$5A					; Z
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$83					; 
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$75					; u
		dc.b	$55					; U
		dc.b	$7A					; z
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$59					; Y
		dc.b	$3C					; <
		dc.b	$5C					; \
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$43					; C
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$AA					; 
		dc.b	$B7					; 
		dc.b	$6F					; o
		dc.b	$3F					; ?
		dc.b	$4A					; J
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$A7					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$63					; c
		dc.b	$55					; U
		dc.b	$A6					; 
		dc.b	$B1					; 
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$57					; W
		dc.b	$42					; B
		dc.b	$7A					; z
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$B1					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$5D					; ]
		dc.b	$48					; H
		dc.b	$3A					; :
		dc.b	$49					; I
		dc.b	$86					; 
		dc.b	$A0					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$7F					; 
		dc.b	$42					; B
		dc.b	$49					; I
		dc.b	$6E					; n
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$75					; u
		dc.b	$55					; U
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$94					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$56					; V
		dc.b	$45					; E
		dc.b	$59					; Y
		dc.b	$9C					; 
		dc.b	$D3					; 
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$A3					; 
		dc.b	$92					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$41					; A
		dc.b	$75					; u
		dc.b	$C2					; 
		dc.b	$D2					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$5A					; Z
		dc.b	$3B					; ;
		dc.b	$4F					; O
		dc.b	$5A					; Z
		dc.b	$83					; 
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$88					; 
		dc.b	$49					; I
		dc.b	$60					; `
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$8E					; 
		dc.b	$AA					; 
		dc.b	$9F					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$9F					; 
		dc.b	$92					; 
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$92					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$57					; W
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$52					; R
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$B3					; 
		dc.b	$92					; 
		dc.b	$48					; H
		dc.b	$43					; C
		dc.b	$59					; Y
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$86					; 
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$AE					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$6F					; o
		dc.b	$57					; W
		dc.b	$73					; s
		dc.b	$A2					; 
		dc.b	$BE					; 
		dc.b	$AD					; 
		dc.b	$8B					; 
		dc.b	$6A					; j
		dc.b	$3C					; <
		dc.b	$49					; I
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$64					; d
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$98					; 
		dc.b	$79					; y
		dc.b	$57					; W
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$96					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$A7					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$71					; q
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$B8					; 
		dc.b	$8E					; 
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$6B					; k
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$AD					; 
		dc.b	$A6					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$5A					; Z
		dc.b	$53					; S
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$5C					; \
		dc.b	$37					; 7
		dc.b	$48					; H
		dc.b	$7C					; |
		dc.b	$A3					; 
		dc.b	$C1					; 
		dc.b	$92					; 
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$9B					; 
		dc.b	$B3					; 
		dc.b	$AB					; 
		dc.b	$8F					; 
		dc.b	$5A					; Z
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$56					; V
		dc.b	$41					; A
		dc.b	$88					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$5C					; \
		dc.b	$4F					; O
		dc.b	$73					; s
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$5C					; \
		dc.b	$9C					; 
		dc.b	$C5					; 
		dc.b	$AD					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$63					; c
		dc.b	$41					; A
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$8A					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$C6					; 
		dc.b	$B1					; 
		dc.b	$6F					; o
		dc.b	$50					; P
		dc.b	$4D					; M
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$57					; W
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$64					; d
		dc.b	$3C					; <
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$6E					; n
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$B0					; 
		dc.b	$8F					; 
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$3F					; ?
		dc.b	$43					; C
		dc.b	$7F					; 
		dc.b	$AD					; 
		dc.b	$A0					; 
		dc.b	$87					; 
		dc.b	$9F					; 
		dc.b	$A2					; 
		dc.b	$65					; e
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$9B					; 
		dc.b	$55					; U
		dc.b	$41					; A
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$B4					; 
		dc.b	$E5					; 
		dc.b	$A7					; 
		dc.b	$59					; Y
		dc.b	$46					; F
		dc.b	$48					; H
		dc.b	$57					; W
		dc.b	$7F					; 
		dc.b	$A4					; 
		dc.b	$A2					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$78					; x
		dc.b	$46					; F
		dc.b	$48					; H
		dc.b	$6F					; o
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$5D					; ]
		dc.b	$3E					; >
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$A2					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$83					; 
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$9B					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$6C					; l
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$98					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$8B					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$9D					; 
		dc.b	$AE					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$86					; 
		dc.b	$BA					; 
		dc.b	$A3					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$5C					; \
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$B3					; 
		dc.b	$92					; 
		dc.b	$50					; P
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$96					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$5A					; Z
		dc.b	$46					; F
		dc.b	$6A					; j
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$AA					; 
		dc.b	$C9					; 
		dc.b	$B1					; 
		dc.b	$8F					; 
		dc.b	$5C					; \
		dc.b	$2E					; .
		dc.b	$3C					; <
		dc.b	$6C					; l
		dc.b	$8B					; 
		dc.b	$AE					; 
		dc.b	$BA					; 
		dc.b	$A0					; 
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$B7					; 
		dc.b	$A3					; 
		dc.b	$78					; x
		dc.b	$52					; R
		dc.b	$38					; 8
		dc.b	$5A					; Z
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$C2					; 
		dc.b	$92					; 
		dc.b	$57					; W
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$83					; 
		dc.b	$AE					; 
		dc.b	$A2					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$94					; 
		dc.b	$A3					; 
		dc.b	$6B					; k
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$7A					; z
		dc.b	$A4					; 
		dc.b	$AD					; 
		dc.b	$91					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$59					; Y
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$9B					; 
		dc.b	$C9					; 
		dc.b	$C6					; 
		dc.b	$80					; 
		dc.b	$3E					; >
		dc.b	$19
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$91					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$A6					; 
		dc.b	$AE					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$65					; e
		dc.b	$A7					; 
		dc.b	$CC					; 
		dc.b	$A4					; 
		dc.b	$4D					; M
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$A2					; 
		dc.b	$AA					; 
		dc.b	$95					; 
		dc.b	$67					; g
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$5E					; ^
		dc.b	$95					; 
		dc.b	$BB					; 
		dc.b	$B8					; 
		dc.b	$A7					; 
		dc.b	$6B					; k
		dc.b	$30					; 0
		dc.b	$3A					; :
		dc.b	$67					; g
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$9F					; 
		dc.b	$61					; a
		dc.b	$4A					; J
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$C5					; 
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	$43					; C
		dc.b	$1C
		dc.b	$35					; 5
		dc.b	$83					; 
		dc.b	$BD					; 
		dc.b	$C2					; 
		dc.b	$BB					; 
		dc.b	$AE					; 
		dc.b	$7D					; }
		dc.b	$49					; I
		dc.b	$3C					; <
		dc.b	$5C					; \
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$A3					; 
		dc.b	$A6					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$4A					; J
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$75					; u
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$3F					; ?
		dc.b	$42					; B
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$7D					; }
		dc.b	$3C					; <
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$BE					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$63					; c
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$83					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$9D					; 
		dc.b	$9B					; 
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$7C					; |
		dc.b	$55					; U
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$92					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$BB					; 
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$6C					; l
		dc.b	$48					; H
		dc.b	$49					; I
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$9C					; 
		dc.b	$B0					; 
		dc.b	$9D					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$79					; y
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$4F					; O
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$96					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$59					; Y
		dc.b	$53					; S
		dc.b	$71					; q
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$9C					; 
		dc.b	$AD					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$6C					; l
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	$57					; W
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$A6					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$B0					; 
		dc.b	$C4					; 
		dc.b	$8D					; 
		dc.b	$4C					; L
		dc.b	$35					; 5
		dc.b	$53					; S
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$B8					; 
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$64					; d
		dc.b	$4D					; M
		dc.b	$68					; h
		dc.b	$87					; 
		dc.b	$98					; 
		dc.b	$AA					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$63					; c
		dc.b	$8A					; 
		dc.b	$C8					; 
		dc.b	$B0					; 
		dc.b	$81					; 
		dc.b	$59					; Y
		dc.b	$52					; R
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$7F					; 
		dc.b	$59					; Y
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$6F					; o
		dc.b	$56					; V
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$64					; d
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$A3					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$5C					; \
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$78					; x
		dc.b	$A9					; 
		dc.b	$B3					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$AE					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$5C					; \
		dc.b	$63					; c
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$64					; d
		dc.b	$52					; R
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$9C					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$95					; 
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$3A					; :
		dc.b	$60					; `
		dc.b	$8E					; 
		dc.b	$BB					; 
		dc.b	$C9					; 
		dc.b	$BA					; 
		dc.b	$7F					; 
		dc.b	$3E					; >
		dc.b	$3E					; >
		dc.b	$5C					; \
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$B3					; 
		dc.b	$95					; 
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$BB					; 
		dc.b	$AB					; 
		dc.b	$80					; 
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$5C					; \
		dc.b	$52					; R
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$75					; u
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$67					; g
		dc.b	$5D					; ]
		dc.b	$84					; 
		dc.b	$A6					; 
		dc.b	$A3					; 
		dc.b	$7C					; |
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$98					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$5C					; \
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$9D					; 
		dc.b	$B4					; 
		dc.b	$7A					; z
		dc.b	$41					; A
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$60					; `
		dc.b	$8A					; 
		dc.b	$A6					; 
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$91					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$9F					; 
		dc.b	$9D					; 
		dc.b	$94					; 
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$92					; 
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$56					; V
		dc.b	$5D					; ]
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$94					; 
		dc.b	$BE					; 
		dc.b	$B4					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$5A					; Z
		dc.b	$49					; I
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$7A					; z
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$9D					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$9F					; 
		dc.b	$B8					; 
		dc.b	$94					; 
		dc.b	$78					; x
		dc.b	$5C					; \
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$4F					; O
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$67					; g
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$AA					; 
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$5C					; \
		dc.b	$8A					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$91					; 
		dc.b	$60					; `
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$6E					; n
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$4F					; O
		dc.b	$6B					; k
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$5C					; \
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$61					; a
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$83					; 
		dc.b	$6C					; l
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$7F					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$81					; 
		dc.b	$63					; c
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$A0					; 
		dc.b	$A6					; 
		dc.b	$78					; x
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$9C					; 
		dc.b	$87					; 
		dc.b	$56					; V
		dc.b	$60					; `
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$6F					; o
		dc.b	$9B					; 
		dc.b	$B8					; 
		dc.b	$A2					; 
		dc.b	$73					; s
		dc.b	$50					; P
		dc.b	$59					; Y
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$95					; 
		dc.b	$A7					; 
		dc.b	$87					; 
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$61					; a
		dc.b	$5C					; \
		dc.b	$87					; 
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$80					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$75					; u
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$67					; g
		dc.b	$5E					; ^
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$96					; 
		dc.b	$A2					; 
		dc.b	$8D					; 
		dc.b	$6C					; l
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$73					; s
		dc.b	$68					; h
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$96					; 
		dc.b	$AA					; 
		dc.b	$98					; 
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$5D					; ]
		dc.b	$4F					; O
		dc.b	$63					; c
		dc.b	$80					; 
		dc.b	$9C					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$83					; 
		dc.b	$5D					; ]
		dc.b	$46					; F
		dc.b	$4F					; O
		dc.b	$79					; y
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$75					; u
		dc.b	$63					; c
		dc.b	$57					; W
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$AE					; 
		dc.b	$AD					; 
		dc.b	$81					; 
		dc.b	$56					; V
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$86					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$8E					; 
		dc.b	$6C					; l
		dc.b	$59					; Y
		dc.b	$5D					; ]
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$A9					; 
		dc.b	$91					; 
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$6E					; n
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$99					; 
		dc.b	$5A					; Z
		dc.b	$46					; F
		dc.b	$56					; V
		dc.b	$61					; a
		dc.b	$8D					; 
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$8D					; 
		dc.b	$64					; d
		dc.b	$4F					; O
		dc.b	$53					; S
		dc.b	$8A					; 
		dc.b	$AB					; 
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$9F					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$76					; v
		dc.b	$A0					; 
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$65					; e
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$AA					; 
		dc.b	$B4					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$96					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$57					; W
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$80					; 
		dc.b	$A6					; 
		dc.b	$A3					; 
		dc.b	$83					; 
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$A9					; 
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$B0					; 
		dc.b	$99					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$94					; 
		dc.b	$B0					; 
		dc.b	$A9					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$46					; F
		dc.b	$55					; U
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$9D					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$81					; 
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$65					; e
		dc.b	$87					; 
		dc.b	$AA					; 
		dc.b	$A0					; 
		dc.b	$92					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$9F					; 
		dc.b	$83					; 
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$8E					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$4A					; J
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$98					; 
		dc.b	$95					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$76					; v
		dc.b	$53					; S
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$92					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$61					; a
		dc.b	$5D					; ]
		dc.b	$7A					; z
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$9F					; 
		dc.b	$7D					; }
		dc.b	$63					; c
		dc.b	$56					; V
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$61					; a
		dc.b	$57					; W
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$A4					; 
		dc.b	$A7					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$5C					; \
		dc.b	$53					; S
		dc.b	$60					; `
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$86					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$76					; v
		dc.b	$56					; V
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$52					; R
		dc.b	$52					; R
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$59					; Y
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$A7					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$68					; h
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$6B					; k
		dc.b	$8B					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$67					; g
		dc.b	$67					; g
		dc.b	$75					; u
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$6F					; o
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$78					; x
		dc.b	$60					; `
		dc.b	$59					; Y
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$AD					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$61					; a
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$9C					; 
		dc.b	$8F					; 
		dc.b	$7D					; }
		dc.b	$5A					; Z
		dc.b	$4C					; L
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$A6					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$95					; 
		dc.b	$67					; g
		dc.b	$4D					; M
		dc.b	$4F					; O
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$AA					; 
		dc.b	$9D					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$59					; Y
		dc.b	$5D					; ]
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$67					; g
		dc.b	$57					; W
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$6F					; o
		dc.b	$4F					; O
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$6E					; n
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$86					; 
		dc.b	$9D					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$57					; W
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$98					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$4C					; L
		dc.b	$43					; C
		dc.b	$68					; h
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$A2					; 
		dc.b	$8F					; 
		dc.b	$73					; s
		dc.b	$59					; Y
		dc.b	$57					; W
		dc.b	$7A					; z
		dc.b	$94					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$63					; c
		dc.b	$4C					; L
		dc.b	$6E					; n
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$78					; x
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$A3					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$63					; c
		dc.b	$72					; r
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$61					; a
		dc.b	$56					; V
		dc.b	$64					; d
		dc.b	$7A					; z
		dc.b	$9F					; 
		dc.b	$A7					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$99					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$55					; U
		dc.b	$6C					; l
		dc.b	$8D					; 
		dc.b	$A2					; 
		dc.b	$B3					; 
		dc.b	$A9					; 
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$53					; S
		dc.b	$4F					; O
		dc.b	$71					; q
		dc.b	$8F					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$6B					; k
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$9C					; 
		dc.b	$88					; 
		dc.b	$71					; q
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$A6					; 
		dc.b	$98					; 
		dc.b	$68					; h
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$95					; 
		dc.b	$AD					; 
		dc.b	$96					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$68					; h
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$80					; 
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$63					; c
		dc.b	$84					; 
		dc.b	$AD					; 
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$60					; `
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$75					; u
		dc.b	$91					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$9D					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	$57					; W
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$63					; c
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$5A					; Z
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$5D					; ]
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$95					; 
		dc.b	$71					; q
		dc.b	$5E					; ^
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$9B					; 
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$95					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$98					; 
		dc.b	$AD					; 
		dc.b	$A3					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$6B					; k
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$9F					; 
		dc.b	$81					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$5E					; ^
		dc.b	$73					; s
		dc.b	$8B					; 
		dc.b	$9B					; 
		dc.b	$98					; 
		dc.b	$7C					; |
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$94					; 
		dc.b	$9B					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$9B					; 
		dc.b	$8B					; 
		dc.b	$72					; r
		dc.b	$61					; a
		dc.b	$63					; c
		dc.b	$7C					; |
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$9F					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$95					; 
		dc.b	$7A					; z
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$6F					; o
		dc.b	$71					; q
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$83					; 
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$A2					; 
		dc.b	$9B					; 
		dc.b	$73					; s
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$91					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$95					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$68					; h
		dc.b	$63					; c
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$88					; 
		dc.b	$A3					; 
		dc.b	$9F					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$78					; x
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$7D					; }
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$50					; P
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$AA					; 
		dc.b	$9B					; 
		dc.b	$79					; y
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$78					; x
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$81					; 
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$71					; q
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$75					; u
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$7C					; |
		dc.b	$8E					; 
		dc.b	$9D					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$64					; d
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$63					; c
		dc.b	$68					; h
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$9C					; 
		dc.b	$96					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$6F					; o
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$94					; 
		dc.b	$7D					; }
		dc.b	$61					; a
		dc.b	$5E					; ^
		dc.b	$71					; q
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$94					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$92					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$8D					; 
		dc.b	$75					; u
		dc.b	$61					; a
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$9B					; 
		dc.b	$96					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$91					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7F					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$87					; 
		dc.b	$94					; 
		dc.b	$91					; 
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$7A					; z
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$7D					; }
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$6B					; k
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$8F					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7D					; }
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$75					; u
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$73					; s
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$83					; 
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$91					; 
		dc.b	$8F					; 
		dc.b	$83					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$78					; x
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
unk_29FFA:	dc.b	$80					; 		; DATA XREF: ROM:0001B04Co
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$86					; 
		dc.b	$58					; X
		dc.b	$A4					; 
		dc.b	$B7					; 
		dc.b	$86					; 
		dc.b	$54					; T
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$66					; f
		dc.b	$92					; 
		dc.b	$AC					; 
		dc.b	$60					; `
		dc.b	$2F					; /
		dc.b	$5C					; \
		dc.b	$A6					; 
		dc.b	$9A					; 
		dc.b	$3B					; ;
		dc.b	$3F					; ?
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$9E					; 
		dc.b	$49					; I
		dc.b	$4D					; M
		dc.b	$47					; G
		dc.b	$70					; p
		dc.b	$8E					; 
		dc.b	$A0					; 
		dc.b	$90					; 
		dc.b	$C7					; 
		dc.b	$92					; 
		dc.b	$68					; h
		dc.b	$5E					; ^
		dc.b	$3B					; ;
		dc.b	$39					; 9
		dc.b	$52					; R
		dc.b	$8C					; 
		dc.b	$B5					; 
		dc.b	$94					; 
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$66					; f
		dc.b	$23					; #
		dc.b	$3F					; ?
		dc.b	$B0					; 
		dc.b	$AE					; 
		dc.b	$5C					; \
		dc.b	$3B					; ;
		dc.b	$35					; 5
		dc.b	$25					; %
		dc.b	$6A					; j
		dc.b	$C7					; 
		dc.b	$FE					; 
		dc.b	$D1					; 
		dc.b	$A6					; 
		dc.b	$74					; t
		dc.b	$6A					; j
		dc.b	$4B					; K
		dc.b	$3B					; ;
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$4D					; M
		dc.b	$9E					; 
		dc.b	$D5					; 
		dc.b	$C3					; 
		dc.b	$A2					; 
		dc.b	$7E					; ~
		dc.b	$2D					; -
		dc.b	1
		dc.b	$2D					; -
		dc.b	$74					; t
		dc.b	$96					; 
		dc.b	$82					; 
		dc.b	$92					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$B0					; 
		dc.b	$A8					; 
		dc.b	$7E					; ~
		dc.b	$5A					; Z
		dc.b	$58					; X
		dc.b	$C7					; 
		dc.b	$B7					; 
		dc.b	$52					; R
		dc.b	$19
		dc.b	$13
		dc.b	$5C					; \
		dc.b	$B1					; 
		dc.b	$CB					; 
		dc.b	$C5					; 
		dc.b	$82					; 
		dc.b	$31					; 1
		dc.b	$62					; b
		dc.b	$66					; f
		dc.b	$7C					; |
		dc.b	$62					; b
		dc.b	$8A					; 
		dc.b	$E5					; 
		dc.b	$D9					; 
		dc.b	$A4					; 
		dc.b	$5C					; \
		dc.b	$62					; b
		dc.b	$49					; I
		dc.b	1
		dc.b	$21					; !
		dc.b	$86					; 
		dc.b	$E1					; 
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$60					; `
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$60					; `
		dc.b	$8C					; 
		dc.b	$B3					; 
		dc.b	$94					; 
		dc.b	$9E					; 
		dc.b	$B1					; 
		dc.b	$84					; 
		dc.b	$29					; )
		dc.b	$17
		dc.b	$3F					; ?
		dc.b	$84					; 
		dc.b	$A4					; 
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$B1					; 
		dc.b	$AA					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$7C					; |
		dc.b	$A6					; 
		dc.b	$B7					; 
		dc.b	$B5					; 
		dc.b	$3D					; =
		dc.b	1
		dc.b	$6A					; j
		dc.b	$D1					; 
		dc.b	$7A					; z
		dc.b	$3B					; ;
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$AA					; 
		dc.b	$D3					; 
		dc.b	$9E					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$7E					; ~
		dc.b	$90					; 
		dc.b	$82					; 
		dc.b	$9A					; 
		dc.b	$B1					; 
		dc.b	$72					; r
		dc.b	$3D					; =
		dc.b	$5E					; ^
		dc.b	$96					; 
		dc.b	$9A					; 
		dc.b	$5A					; Z
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$8C					; 
		dc.b	$6E					; n
		dc.b	$5C					; \
		dc.b	$AA					; 
		dc.b	$86					; 
		dc.b	$64					; d
		dc.b	$82					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$49					; I
		dc.b	$2F					; /
		dc.b	$6E					; n
		dc.b	$9E					; 
		dc.b	$B0					; 
		dc.b	$B5					; 
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$5E					; ^
		dc.b	$4B					; K
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$66					; f
		dc.b	$7A					; z
		dc.b	$B3					; 
		dc.b	$B3					; 
		dc.b	$72					; r
		dc.b	$52					; R
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$62					; b
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$B5					; 
		dc.b	$DD					; 
		dc.b	$A2					; 
		dc.b	$70					; p
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$A8					; 
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$7E					; ~
		dc.b	$6E					; n
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$45					; E
		dc.b	$74					; t
		dc.b	$72					; r
		dc.b	$58					; X
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$A8					; 
		dc.b	$70					; p
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$9A					; 
		dc.b	$96					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$62					; b
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$94					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$90					; 
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$84					; 
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$39					; 9
		dc.b	$50					; P
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$5A					; Z
		dc.b	$64					; d
		dc.b	$90					; 
		dc.b	$DB					; 
		dc.b	$D5					; 
		dc.b	$9E					; 
		dc.b	$41					; A
		dc.b	$13
		dc.b	$3D					; =
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$D1					; 
		dc.b	$C3					; 
		dc.b	$6A					; j
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$72					; r
		dc.b	$66					; f
		dc.b	$4B					; K
		dc.b	$78					; x
		dc.b	$AC					; 
		dc.b	$C9					; 
		dc.b	$B7					; 
		dc.b	$6A					; j
		dc.b	$4B					; K
		dc.b	$70					; p
		dc.b	$9C					; 
		dc.b	$90					; 
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$6C					; l
		dc.b	$3D					; =
		dc.b	$4F					; O
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$4D					; M
		dc.b	$43					; C
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$9A					; 
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$B5					; 
		dc.b	$78					; x
		dc.b	$58					; X
		dc.b	$3D					; =
		dc.b	$3B					; ;
		dc.b	$88					; 
		dc.b	$A4					; 
		dc.b	$4F					; O
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$CD					; 
		dc.b	$CB					; 
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$47					; G
		dc.b	$D
		dc.b	$D
		dc.b	$2F					; /
		dc.b	$8C					; 
		dc.b	$E7					; 
		dc.b	$C5					; 
		dc.b	$9C					; 
		dc.b	$7A					; z
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$50					; P
		dc.b	$6C					; l
		dc.b	$4B					; K
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$A8					; 
		dc.b	$9E					; 
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$AC					; 
		dc.b	$90					; 
		dc.b	$56					; V
		dc.b	$47					; G
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$5C					; \
		dc.b	$74					; t
		dc.b	$BD					; 
		dc.b	$FE					; 
		dc.b	$E1					; 
		dc.b	$76					; v
		dc.b	$58					; X
		dc.b	$A2					; 
		dc.b	$9A					; 
		dc.b	$86					; 
		dc.b	$8C					; 
		dc.b	$7C					; |
		dc.b	$58					; X
		dc.b	$52					; R
		dc.b	$74					; t
		dc.b	$66					; f
		dc.b	$39					; 9
		dc.b	$F
		dc.b	$1B
		dc.b	$33					; 3
		dc.b	$25					; %
		dc.b	$13
		dc.b	$35					; 5
		dc.b	$50					; P
		dc.b	$7E					; ~
		dc.b	$C7					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$B1					; 
		dc.b	$B1					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$D3					; 
		dc.b	$FE					; 
		dc.b	$F1					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$74					; t
		dc.b	$94					; 
		dc.b	$76					; v
		dc.b	$37					; 7
		dc.b	$41					; A
		dc.b	$21					; !
		dc.b	9
		dc.b	$33					; 3
		dc.b	$3D					; =
		dc.b	$39					; 9
		dc.b	$33					; 3
		dc.b	$2B					; +
		dc.b	$66					; f
		dc.b	$9E					; 
		dc.b	$C9					; 
		dc.b	$C7					; 
		dc.b	$62					; b
		dc.b	$41					; A
		dc.b	$49					; I
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$6C					; l
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$74					; t
		dc.b	$8E					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$45					; E
		dc.b	$33					; 3
		dc.b	$35					; 5
		dc.b	$7A					; z
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$FE					; 
		dc.b	$ED					; 
		dc.b	$8C					; 
		dc.b	$7E					; ~
		dc.b	$AA					; 
		dc.b	$94					; 
		dc.b	$88					; 
		dc.b	$78					; x
		dc.b	$6E					; n
		dc.b	$5E					; ^
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$2D					; -
		dc.b	$3D					; =
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$7A					; z
		dc.b	$62					; b
		dc.b	$4D					; M
		dc.b	$6C					; l
		dc.b	$4B					; K
		dc.b	$60					; `
		dc.b	$84					; 
		dc.b	$74					; t
		dc.b	$6E					; n
		dc.b	$AA					; 
		dc.b	$C1					; 
		dc.b	$5E					; ^
		dc.b	$1F
		dc.b	$2D					; -
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$A2					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$88					; 
		dc.b	$72					; r
		dc.b	$6A					; j
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$64					; d
		dc.b	$58					; X
		dc.b	$64					; d
		dc.b	$5C					; \
		dc.b	$56					; V
		dc.b	$7A					; z
		dc.b	$DF					; 
		dc.b	$CD					; 
		dc.b	$AE					; 
		dc.b	$B1					; 
		dc.b	$9C					; 
		dc.b	$62					; b
		dc.b	$23					; #
		dc.b	$F
		dc.b	$3B					; ;
		dc.b	$8A					; 
		dc.b	$B0					; 
		dc.b	$BF					; 
		dc.b	$96					; 
		dc.b	$C1					; 
		dc.b	$BB					; 
		dc.b	$76					; v
		dc.b	$4D					; M
		dc.b	$52					; R
		dc.b	$31					; 1
		dc.b	$19
		dc.b	$60					; `
		dc.b	$C7					; 
		dc.b	$E9					; 
		dc.b	$90					; 
		dc.b	$76					; v
		dc.b	$8C					; 
		dc.b	$90					; 
		dc.b	$C5					; 
		dc.b	$D1					; 
		dc.b	$6C					; l
		dc.b	$39					; 9
		dc.b	$50					; P
		dc.b	$58					; X
		dc.b	$6A					; j
		dc.b	$60					; `
		dc.b	$60					; `
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$58					; X
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$B9					; 
		dc.b	$82					; 
		dc.b	$2F					; /
		dc.b	$2F					; /
		dc.b	$80					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$A0					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$45					; E
		dc.b	$1B
		dc.b	$68					; h
		dc.b	$7C					; |
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$88					; 
		dc.b	$6A					; j
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$7C					; |
		dc.b	$88					; 
		dc.b	$68					; h
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$8C					; 
		dc.b	$A6					; 
		dc.b	$86					; 
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$7C					; |
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$9C					; 
		dc.b	$AC					; 
		dc.b	$9C					; 
		dc.b	$5C					; \
		dc.b	$58					; X
		dc.b	$78					; x
		dc.b	$82					; 
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$88					; 
		dc.b	$56					; V
		dc.b	$37					; 7
		dc.b	$52					; R
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$B3					; 
		dc.b	$F1					; 
		dc.b	$DB					; 
		dc.b	$A4					; 
		dc.b	$62					; b
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$31					; 1
		dc.b	$64					; d
		dc.b	$B1					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$C5					; 
		dc.b	$9E					; 
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$8C					; 
		dc.b	$80					; 
		dc.b	$39					; 9
		dc.b	$3B					; ;
		dc.b	$56					; V
		dc.b	$9C					; 
		dc.b	$EB					; 
		dc.b	$C5					; 
		dc.b	$80					; 
		dc.b	$3D					; =
		dc.b	$43					; C
		dc.b	$58					; X
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$66					; f
		dc.b	$3F					; ?
		dc.b	$15
		dc.b	$6A					; j
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$58					; X
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$5A					; Z
		dc.b	$3D					; =
		dc.b	$5E					; ^
		dc.b	$58					; X
		dc.b	$6E					; n
		dc.b	$98					; 
		dc.b	$76					; v
		dc.b	$1D
		dc.b	$23					; #
		dc.b	$6E					; n
		dc.b	$90					; 
		dc.b	$CB					; 
		dc.b	$ED					; 
		dc.b	$FB					; 
		dc.b	$DD					; 
		dc.b	$D9					; 
		dc.b	$A8					; 
		dc.b	$50					; P
		dc.b	$25					; %
		dc.b	$3B					; ;
		dc.b	$D
		dc.b	$D
		dc.b	$5A					; Z
		dc.b	$C1					; 
		dc.b	$D9					; 
		dc.b	$B9					; 
		dc.b	$A6					; 
		dc.b	$7E					; ~
		dc.b	$49					; I
		dc.b	$3F					; ?
		dc.b	$6E					; n
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$82					; 
		dc.b	$68					; h
		dc.b	$62					; b
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$9E					; 
		dc.b	$A2					; 
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$56					; V
		dc.b	$3F					; ?
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$5A					; Z
		dc.b	$37					; 7
		dc.b	$70					; p
		dc.b	$B5					; 
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$BF					; 
		dc.b	$A6					; 
		dc.b	$3F					; ?
		dc.b	$2B					; +
		dc.b	$2D					; -
		dc.b	$2F					; /
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$D9					; 
		dc.b	$FE					; 
		dc.b	$DF					; 
		dc.b	$60					; `
		dc.b	$13
		dc.b	$31					; 1
		dc.b	$6C					; l
		dc.b	$A6					; 
		dc.b	$CF					; 
		dc.b	$DB					; 
		dc.b	$AA					; 
		dc.b	$6C					; l
		dc.b	$19
		dc.b	1
		dc.b	1
		dc.b	$A0					; 
		dc.b	$8E					; 
		dc.b	$31					; 1
		dc.b	$C5					; 
		dc.b	$8E					; 
		dc.b	1
		dc.b	$80					; 
		dc.b	$D5					; 
		dc.b	$90					; 
		dc.b	$C9					; 
		dc.b	$FE					; 
		dc.b	$CD					; 
		dc.b	$C3					; 
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$B5					; 
		dc.b	$3F					; ?
		dc.b	$47					; G
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$C3					; 
		dc.b	$8C					; 
		dc.b	$2F					; /
		dc.b	$2B					; +
		dc.b	$2B					; +
		dc.b	$50					; P
		dc.b	$F5					; 
		dc.b	$FE					; 
		dc.b	$E5					; 
		dc.b	$4B					; K
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$54					; T
		dc.b	$9A					; 
		dc.b	$82					; 
		dc.b	$90					; 
		dc.b	$AA					; 
		dc.b	$15
		dc.b	1
		dc.b	$43					; C
		dc.b	$D5					; 
		dc.b	$FE					; 
		dc.b	$EF					; 
		dc.b	$D1					; 
		dc.b	$C9					; 
		dc.b	$50					; P
		dc.b	$4B					; K
		dc.b	$5C					; \
		dc.b	$4D					; M
		dc.b	$21					; !
		dc.b	1
		dc.b	1
		dc.b	$1D
		dc.b	$AE					; 
		dc.b	$C5					; 
		dc.b	$90					; 
		dc.b	$B9					; 
		dc.b	$AC					; 
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D1					; 
		dc.b	$50					; P
		dc.b	1
		dc.b	1
		dc.b	$2F					; /
		dc.b	1
		dc.b	1
		dc.b	$4D					; M
		dc.b	$A6					; 
		dc.b	$D3					; 
		dc.b	$AA					; 
		dc.b	$FD					; 
		dc.b	$FE					; 
		dc.b	$CB					; 
		dc.b	$9A					; 
		dc.b	$C3					; 
		dc.b	$CF					; 
		dc.b	$68					; h
		dc.b	$25					; %
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$3F					; ?
		dc.b	$5C					; \
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$BD					; 
		dc.b	$C1					; 
		dc.b	$B3					; 
		dc.b	$60					; `
		dc.b	$29					; )
		dc.b	7
		dc.b	1
		dc.b	1
		dc.b	$33					; 3
		dc.b	$6A					; j
		dc.b	$74					; t
		dc.b	$72					; r
		dc.b	$CF					; 
		dc.b	$F3					; 
		dc.b	$EF					; 
		dc.b	$C9					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$4B					; K
		dc.b	$3D					; =
		dc.b	$52					; R
		dc.b	7
		dc.b	$2F					; /
		dc.b	$72					; r
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$BD					; 
		dc.b	$C9					; 
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C3					; 
		dc.b	$58					; X
		dc.b	$25					; %
		dc.b	7
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$23					; #
		dc.b	$66					; f
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$58					; X
		dc.b	$58					; X
		dc.b	$86					; 
		dc.b	$AC					; 
		dc.b	$A6					; 
		dc.b	$76					; v
		dc.b	$BF					; 
		dc.b	$FE					; 
		dc.b	$FD					; 
		dc.b	$B9					; 
		dc.b	$96					; 
		dc.b	$64					; d
		dc.b	$43					; C
		dc.b	$35					; 5
		dc.b	$3B					; ;
		dc.b	$11
		dc.b	$35					; 5
		dc.b	$7C					; |
		dc.b	$B9					; 
		dc.b	$CB					; 
		dc.b	$B9					; 
		dc.b	$B7					; 
		dc.b	$B1					; 
		dc.b	$BF					; 
		dc.b	$76					; v
		dc.b	$62					; b
		dc.b	$9A					; 
		dc.b	$84					; 
		dc.b	$4D					; M
		dc.b	$2D					; -
		dc.b	$25					; %
		dc.b	$4D					; M
		dc.b	$56					; V
		dc.b	$62					; b
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$58					; X
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$92					; 
		dc.b	$AC					; 
		dc.b	$CF					; 
		dc.b	$BF					; 
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$B7					; 
		dc.b	$B1					; 
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$94					; 
		dc.b	$78					; x
		dc.b	$9E					; 
		dc.b	$AE					; 
		dc.b	$A0					; 
		dc.b	$B3					; 
		dc.b	$8E					; 
		dc.b	$7C					; |
		dc.b	$9A					; 
		dc.b	$70					; p
		dc.b	$39					; 9
		dc.b	$2F					; /
		dc.b	$2F					; /
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$4B					; K
		dc.b	$F
		dc.b	3
		dc.b	$5C					; \
		dc.b	$7A					; z
		dc.b	$62					; b
		dc.b	$5C					; \
		dc.b	$5A					; Z
		dc.b	$35					; 5
		dc.b	$11
		dc.b	$11
		dc.b	$74					; t
		dc.b	$A8					; 
		dc.b	$98					; 
		dc.b	$62					; b
		dc.b	$6E					; n
		dc.b	$B3					; 
		dc.b	$98					; 
		dc.b	$C3					; 
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$ED					; 
		dc.b	$F7					; 
		dc.b	$E7					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$9E					; 
		dc.b	$98					; 
		dc.b	$66					; f
		dc.b	$4B					; K
		dc.b	$52					; R
		dc.b	$62					; b
		dc.b	$4D					; M
		dc.b	$39					; 9
		dc.b	$90					; 
		dc.b	$F9					; 
		dc.b	$A8					; 
		dc.b	$54					; T
		dc.b	$E7					; 
		dc.b	$FE					; 
		dc.b	$A8					; 
		dc.b	$62					; b
		dc.b	7
		dc.b	$56					; V
		dc.b	$CD					; 
		dc.b	$D9					; 
		dc.b	$9C					; 
		dc.b	$49					; I
		dc.b	$B
		dc.b	$35					; 5
		dc.b	$98					; 
		dc.b	$C5					; 
		dc.b	$CF					; 
		dc.b	$78					; x
		dc.b	$15
		dc.b	$3D					; =
		dc.b	$C9					; 
		dc.b	$94					; 
		dc.b	$47					; G
		dc.b	$52					; R
		dc.b	$33					; 3
		dc.b	$54					; T
		dc.b	$31					; 1
		dc.b	$3F					; ?
		dc.b	$2F					; /
		dc.b	$2B					; +
		dc.b	$56					; V
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$3D					; =
		dc.b	$68					; h
		dc.b	$60					; `
		dc.b	$56					; V
		dc.b	$76					; v
		dc.b	$6C					; l
		dc.b	$43					; C
		dc.b	$88					; 
		dc.b	$AE					; 
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$C7					; 
		dc.b	$EB					; 
		dc.b	$9E					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$A4					; 
		dc.b	$DB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E7					; 
		dc.b	$DB					; 
		dc.b	$AE					; 
		dc.b	$5A					; Z
		dc.b	$6A					; j
		dc.b	$7E					; ~
		dc.b	$62					; b
		dc.b	$6A					; j
		dc.b	$8C					; 
		dc.b	$94					; 
		dc.b	$64					; d
		dc.b	$35					; 5
		dc.b	$5C					; \
		dc.b	$49					; I
		dc.b	$84					; 
		dc.b	$5C					; \
		dc.b	$3B					; ;
		dc.b	$F
		dc.b	$11
		dc.b	$2B					; +
		dc.b	$52					; R
		dc.b	$56					; V
		dc.b	$3F					; ?
		dc.b	$41					; A
		dc.b	$74					; t
		dc.b	$33					; 3
		dc.b	$11
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$A2					; 
		dc.b	$C7					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$CB					; 
		dc.b	$A4					; 
		dc.b	$7E					; ~
		dc.b	$52					; R
		dc.b	$4B					; K
		dc.b	$4F					; O
		dc.b	$2F					; /
		dc.b	$35					; 5
		dc.b	$62					; b
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$A2					; 
		dc.b	$B9					; 
		dc.b	$90					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$4B					; K
		dc.b	$66					; f
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$58					; X
		dc.b	$7A					; z
		dc.b	$AA					; 
		dc.b	$D5					; 
		dc.b	$B7					; 
		dc.b	$9A					; 
		dc.b	$B1					; 
		dc.b	$78					; x
		dc.b	$41					; A
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$49					; I
		dc.b	$39					; 9
		dc.b	$56					; V
		dc.b	$4B					; K
		dc.b	$6A					; j
		dc.b	$62					; b
		dc.b	$2B					; +
		dc.b	$29					; )
		dc.b	$39					; 9
		dc.b	$31					; 1
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$84					; 
		dc.b	$AE					; 
		dc.b	$88					; 
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$6A					; j
		dc.b	$31					; 1
		dc.b	$52					; R
		dc.b	$96					; 
		dc.b	$A4					; 
		dc.b	$9C					; 
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$BB					; 
		dc.b	$C7					; 
		dc.b	$AA					; 
		dc.b	$9A					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$6C					; l
		dc.b	$98					; 
		dc.b	$A4					; 
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$9A					; 
		dc.b	$88					; 
		dc.b	$58					; X
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$4D					; M
		dc.b	$50					; P
		dc.b	$5A					; Z
		dc.b	$66					; f
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$3F					; ?
		dc.b	$47					; G
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$50					; P
		dc.b	$8A					; 
		dc.b	$C5					; 
		dc.b	$9C					; 
		dc.b	$90					; 
		dc.b	$56					; V
		dc.b	$6E					; n
		dc.b	$9A					; 
		dc.b	$47					; G
		dc.b	$7C					; |
		dc.b	$BF					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$74					; t
		dc.b	$5A					; Z
		dc.b	$8C					; 
		dc.b	$AE					; 
		dc.b	$AC					; 
		dc.b	$B0					; 
		dc.b	$7C					; |
		dc.b	$5A					; Z
		dc.b	$7A					; z
		dc.b	$CF					; 
		dc.b	$DB					; 
		dc.b	$9A					; 
		dc.b	$70					; p
		dc.b	$43					; C
		dc.b	$3B					; ;
		dc.b	$56					; V
		dc.b	$72					; r
		dc.b	$98					; 
		dc.b	$BF					; 
		dc.b	$A6					; 
		dc.b	$8A					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$B0					; 
		dc.b	$C1					; 
		dc.b	$A4					; 
		dc.b	$8A					; 
		dc.b	$74					; t
		dc.b	$7A					; z
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$45					; E
		dc.b	$2B					; +
		dc.b	$2B					; +
		dc.b	$54					; T
		dc.b	$74					; t
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$76					; v
		dc.b	$45					; E
		dc.b	$39					; 9
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$21					; !
		dc.b	$13
		dc.b	$3F					; ?
		dc.b	$5A					; Z
		dc.b	$8C					; 
		dc.b	$92					; 
		dc.b	$62					; b
		dc.b	$29					; )
unk_2A4AA:	dc.b	$1D					; DATA XREF: ROM:0001B03Co
		dc.b	$3D					; =
		dc.b	$58					; X
		dc.b	$4D					; M
		dc.b	$62					; b
		dc.b	$AC					; 
		dc.b	$BF					; 
		dc.b	$84					; 
		dc.b	$5A					; Z
		dc.b	$AC					; 
		dc.b	$B9					; 
		dc.b	$68					; h
		dc.b	$58					; X
		dc.b	$9A					; 
		dc.b	$B5					; 
		dc.b	$B3					; 
		dc.b	$AA					; 
		dc.b	$B7					; 
		dc.b	$C9					; 
		dc.b	$BB					; 
		dc.b	$B0					; 
		dc.b	$8A					; 
		dc.b	$5C					; \
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$8A					; 
		dc.b	$A8					; 
		dc.b	$CB					; 
		dc.b	$B3					; 
		dc.b	$94					; 
		dc.b	$86					; 
		dc.b	$94					; 
		dc.b	$A6					; 
		dc.b	$78					; x
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$8A					; 
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$39					; 9
		dc.b	$62					; b
		dc.b	$68					; h
		dc.b	$56					; V
		dc.b	$6C					; l
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$B7					; 
		dc.b	$82					; 
		dc.b	$4B					; K
		dc.b	$2B					; +
		dc.b	$23					; #
		dc.b	$47					; G
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$5E					; ^
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$29					; )
		dc.b	$58					; X
		dc.b	$45					; E
		dc.b	$4D					; M
		dc.b	$5E					; ^
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$52					; R
		dc.b	$4F					; O
		dc.b	$66					; f
		dc.b	$5E					; ^
		dc.b	$86					; 
		dc.b	$9A					; 
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$AE					; 
		dc.b	$86					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$82					; 
		dc.b	$7A					; z
		dc.b	$AC					; 
		dc.b	$D7					; 
		dc.b	$C1					; 
		dc.b	$80					; 
		dc.b	$56					; V
		dc.b	$50					; P
		dc.b	$66					; f
		dc.b	$64					; d
		dc.b	$66					; f
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$47					; G
		dc.b	$60					; `
		dc.b	$78					; x
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$3F					; ?
		dc.b	$5A					; Z
		dc.b	$9C					; 
		dc.b	$BD					; 
		dc.b	$DD					; 
		dc.b	$E3					; 
		dc.b	$BB					; 
		dc.b	$AA					; 
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$68					; h
		dc.b	$8E					; 
		dc.b	$B9					; 
		dc.b	$B0					; 
		dc.b	$A8					; 
		dc.b	$8C					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$D
		dc.b	1
		dc.b	$31					; 1
		dc.b	$47					; G
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$74					; t
		dc.b	$58					; X
		dc.b	$7A					; z
		dc.b	$9E					; 
		dc.b	$AE					; 
		dc.b	$90					; 
		dc.b	$76					; v
		dc.b	$45					; E
		dc.b	$4B					; K
		dc.b	$7E					; ~
		dc.b	$9A					; 
		dc.b	$A8					; 
		dc.b	$8C					; 
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$9C					; 
		dc.b	$A6					; 
		dc.b	$B5					; 
		dc.b	$BB					; 
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$A2					; 
		dc.b	$A2					; 
		dc.b	$60					; `
		dc.b	$50					; P
		dc.b	$A6					; 
		dc.b	$25					; %
		dc.b	1
		dc.b	7
		dc.b	$13
		dc.b	$5E					; ^
		dc.b	$96					; 
		dc.b	$CD					; 
		dc.b	$BF					; 
		dc.b	$9C					; 
		dc.b	$A0					; 
		dc.b	$2B					; +
		dc.b	1
		dc.b	1
		dc.b	$49					; I
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$CD					; 
		dc.b	1
		dc.b	1
		dc.b	$17
		dc.b	$BB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B9					; 
		dc.b	$2F					; /
		dc.b	$D3					; 
		dc.b	$C5					; 
		dc.b	$5C					; \
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$84					; 
		dc.b	$FE					; 
		dc.b	$9A					; 
		dc.b	$5C					; \
		dc.b	$C1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E3					; 
		dc.b	$98					; 
		dc.b	$13
		dc.b	1
		dc.b	$78					; x
		dc.b	5
		dc.b	1
		dc.b	$60					; `
		dc.b	$B5					; 
		dc.b	$DD					; 
		dc.b	$FE					; 
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$F
		dc.b	$5C					; \
		dc.b	$F
		dc.b	1
		dc.b	9
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FD					; 
		dc.b	$13
		dc.b	1
		dc.b	$29					; )
		dc.b	$EB					; 
		dc.b	$FE					; 
		dc.b	$88					; 
		dc.b	$50					; P
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$62					; b
		dc.b	$1D
		dc.b	$3B					; ;
		dc.b	1
		dc.b	1
		dc.b	$4B					; K
		dc.b	$94					; 
		dc.b	$BF					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$94					; 
		dc.b	$1F
		dc.b	1
		dc.b	1
		dc.b	9
		dc.b	7
		dc.b	$2B					; +
		dc.b	$6E					; n
		dc.b	$9A					; 
		dc.b	$C9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B1					; 
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$A0					; 
		dc.b	$D1					; 
		dc.b	$A6					; 
		dc.b	$3D					; =
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$56					; V
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	3
		dc.b	$68					; h
		dc.b	$A4					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$C7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$AC					; 
		dc.b	$84					; 
		dc.b	$74					; t
		dc.b	$74					; t
		dc.b	$5C					; \
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$4D					; M
		dc.b	$3B					; ;
		dc.b	$78					; x
		dc.b	$CF					; 
		dc.b	$F7					; 
		dc.b	$F5					; 
		dc.b	$C5					; 
		dc.b	$78					; x
		dc.b	$58					; X
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$9A					; 
		dc.b	$A4					; 
		dc.b	$A8					; 
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$6C					; l
		dc.b	$5E					; ^
		dc.b	$25					; %
		dc.b	$11
		dc.b	$2D					; -
		dc.b	$50					; P
		dc.b	$60					; `
		dc.b	$35					; 5
		dc.b	$37					; 7
		dc.b	$5A					; Z
		dc.b	$68					; h
		dc.b	$45					; E
		dc.b	$74					; t
		dc.b	$B0					; 
		dc.b	$A6					; 
		dc.b	$68					; h
		dc.b	$3F					; ?
		dc.b	$35					; 5
		dc.b	$50					; P
		dc.b	$90					; 
		dc.b	$92					; 
		dc.b	$9E					; 
		dc.b	$F9					; 
		dc.b	$BB					; 
		dc.b	1
		dc.b	$AC					; 
		dc.b	$FB					; 
		dc.b	$3B					; ;
		dc.b	$9C					; 
		dc.b	$DD					; 
		dc.b	$F7					; 
		dc.b	$FE					; 
		dc.b	$50					; P
		dc.b	1
		dc.b	$AE					; 
		dc.b	$CF					; 
		dc.b	$49					; I
		dc.b	$94					; 
		dc.b	$8A					; 
		dc.b	$E5					; 
		dc.b	$70					; p
		dc.b	1
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$7E					; ~
		dc.b	$4F					; O
		dc.b	$2F					; /
		dc.b	$D3					; 
		dc.b	$E7					; 
		dc.b	$19
		dc.b	$62					; b
		dc.b	$72					; r
		dc.b	$25					; %
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$6C					; l
		dc.b	$FB					; 
		dc.b	$D3					; 
		dc.b	$54					; T
		dc.b	$76					; v
		dc.b	9
		dc.b	$17
		dc.b	$C9					; 
		dc.b	$9C					; 
		dc.b	$68					; h
		dc.b	$B
		dc.b	$35					; 5
		dc.b	$C1					; 
		dc.b	$23					; #
		dc.b	1
		dc.b	$41					; A
		dc.b	$47					; G
		dc.b	$AE					; 
		dc.b	$9A					; 
		dc.b	$6C					; l
		dc.b	$50					; P
		dc.b	$A2					; 
		dc.b	$8C					; 
		dc.b	$EB					; 
		dc.b	$BD					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$66					; f
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$45					; E
		dc.b	$88					; 
		dc.b	$BB					; 
		dc.b	$52					; R
		dc.b	$A0					; 
		dc.b	$BB					; 
		dc.b	$88					; 
		dc.b	$AC					; 
		dc.b	$70					; p
		dc.b	$9C					; 
		dc.b	1
		dc.b	$43					; C
		dc.b	$AC					; 
		dc.b	$8E					; 
		dc.b	$FE					; 
		dc.b	$AC					; 
		dc.b	$13
		dc.b	$35					; 5
		dc.b	$9C					; 
		dc.b	$D1					; 
		dc.b	$CD					; 
		dc.b	$9A					; 
		dc.b	$6E					; n
		dc.b	$CB					; 
		dc.b	$5E					; ^
		dc.b	$54					; T
		dc.b	$A2					; 
		dc.b	$70					; p
		dc.b	$B3					; 
		dc.b	$43					; C
		dc.b	$54					; T
		dc.b	$B7					; 
		dc.b	$33					; 3
		dc.b	$96					; 
		dc.b	$D5					; 
		dc.b	$5C					; \
		dc.b	$25					; %
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$64					; d
		dc.b	$29					; )
		dc.b	$8A					; 
		dc.b	$D5					; 
		dc.b	$9E					; 
		dc.b	$9A					; 
		dc.b	$D7					; 
		dc.b	$AE					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$80					; 
		dc.b	$D9					; 
		dc.b	$DB					; 
		dc.b	$B5					; 
		dc.b	$9A					; 
		dc.b	$54					; T
		dc.b	$5E					; ^
		dc.b	$31					; 1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$25					; %
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$8C					; 
		dc.b	$DD					; 
		dc.b	$FB					; 
		dc.b	$DB					; 
		dc.b	$BB					; 
		dc.b	$96					; 
		dc.b	$8C					; 
		dc.b	$78					; x
		dc.b	$45					; E
		dc.b	$4F					; O
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$62					; b
		dc.b	$62					; b
		dc.b	$84					; 
		dc.b	$C7					; 
		dc.b	$B3					; 
		dc.b	$41					; A
		dc.b	$1D
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$A8					; 
		dc.b	$72					; r
		dc.b	$11
		dc.b	1
		dc.b	9
		dc.b	$74					; t
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A0					; 
		dc.b	$23					; #
		dc.b	1
		dc.b	1
		dc.b	$D
		dc.b	$56					; V
		dc.b	$B1					; 
		dc.b	$4D					; M
		dc.b	$4B					; K
		dc.b	$68					; h
		dc.b	$6E					; n
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$92					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$8C					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$A6					; 
		dc.b	$A4					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$23					; #
		dc.b	$45					; E
		dc.b	$6E					; n
		dc.b	$92					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$9C					; 
		dc.b	$A4					; 
		dc.b	$80					; 
		dc.b	$66					; f
		dc.b	$82					; 
		dc.b	$96					; 
		dc.b	$82					; 
		dc.b	$9E					; 
		dc.b	$B3					; 
		dc.b	$B9					; 
		dc.b	$B9					; 
		dc.b	$B9					; 
		dc.b	$C1					; 
		dc.b	$B7					; 
		dc.b	$BB					; 
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$A2					; 
		dc.b	$8E					; 
		dc.b	$8A					; 
		dc.b	$96					; 
		dc.b	$B0					; 
		dc.b	$C3					; 
		dc.b	$B5					; 
		dc.b	$92					; 
		dc.b	$74					; t
		dc.b	$50					; P
		dc.b	$4B					; K
		dc.b	$74					; t
		dc.b	$92					; 
		dc.b	$66					; f
		dc.b	$41					; A
		dc.b	$5C					; \
		dc.b	$7E					; ~
		dc.b	$A6					; 
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$AC					; 
		dc.b	$7E					; ~
		dc.b	$6A					; j
		dc.b	$9E					; 
		dc.b	$BD					; 
		dc.b	$C5					; 
		dc.b	$8E					; 
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$94					; 
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$8C					; 
		dc.b	$62					; b
		dc.b	$52					; R
		dc.b	$50					; P
		dc.b	$41					; A
		dc.b	$1F
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$21					; !
		dc.b	$21					; !
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$2D					; -
		dc.b	$78					; x
		dc.b	$9A					; 
		dc.b	$C1					; 
		dc.b	$CB					; 
		dc.b	$B5					; 
		dc.b	$B3					; 
		dc.b	$BB					; 
		dc.b	$B0					; 
		dc.b	$7C					; |
		dc.b	$6E					; n
		dc.b	$62					; b
		dc.b	$4B					; K
		dc.b	$56					; V
		dc.b	$27					; '
		dc.b	$35					; 5
		dc.b	$41					; A
		dc.b	$3B					; ;
		dc.b	$66					; f
		dc.b	$B7					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$DD					; 
		dc.b	$C9					; 
		dc.b	$D5					; 
		dc.b	$E3					; 
		dc.b	$EB					; 
		dc.b	$BD					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$B9					; 
		dc.b	$D5					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$69					; i
		dc.b	$69					; i
		dc.b	$5D					; ]
		dc.b	$45					; E
		dc.b	$50					; P
		dc.b	$7E					; ~
		dc.b	$A1					; 
		dc.b	$B0					; 
		dc.b	$C2					; 
		dc.b	$E4					; 
		dc.b	$EA					; 
		dc.b	$BD					; 
		dc.b	$A3					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$77					; w
		dc.b	$5F					; _
		dc.b	$3E					; >
		dc.b	$1D
		dc.b	$34					; 4
		dc.b	$6E					; n
		dc.b	$68					; h
		dc.b	$81					; 
		dc.b	$60					; `
		dc.b	$51					; Q
		dc.b	$B1					; 
		dc.b	$AB					; 
		dc.b	$68					; h
		dc.b	$3A					; :
		dc.b	$8F					; 
		dc.b	$AD					; 
		dc.b	$58					; X
		dc.b	$68					; h
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$D4					; 
		dc.b	$AF					; 
		dc.b	$A2					; 
		dc.b	$A6					; 
		dc.b	$50					; P
		dc.b	$55					; U
		dc.b	$4F					; O
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$74					; t
		dc.b	$92					; 
		dc.b	$D0					; 
		dc.b	$BB					; 
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$4E					; N
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$93					; 
		dc.b	$7F					; 
		dc.b	$8F					; 
		dc.b	$68					; h
		dc.b	$44					; D
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$6D					; m
		dc.b	$57					; W
		dc.b	$F
		dc.b	$28					; (
		dc.b	$65					; e
		dc.b	$AC					; 
		dc.b	$C4					; 
		dc.b	$8E					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$5A					; Z
		dc.b	$88					; 
		dc.b	$9F					; 
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$9A					; 
		dc.b	$59					; Y
		dc.b	$4A					; J
		dc.b	$7A					; z
		dc.b	$B4					; 
		dc.b	$A8					; 
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$93					; 
		dc.b	$96					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$64					; d
		dc.b	$C7					; 
		dc.b	$F1					; 
		dc.b	$C1					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$5E					; ^
		dc.b	$52					; R
		dc.b	$53					; S
		dc.b	$5B					; [
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$6F					; o
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$77					; w
		dc.b	$36					; 6
		dc.b	$19
		dc.b	$49					; I
		dc.b	$89					; 
		dc.b	$C1					; 
		dc.b	$BB					; 
		dc.b	$6F					; o
		dc.b	$60					; `
		dc.b	$65					; e
		dc.b	$81					; 
		dc.b	$C4					; 
		dc.b	$DE					; 
		dc.b	$B4					; 
		dc.b	$66					; f
		dc.b	$23					; #
		dc.b	5
		dc.b	$4B					; K
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$BE					; 
		dc.b	$B2					; 
		dc.b	$78					; x
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$61					; a
		dc.b	$6E					; n
		dc.b	$4F					; O
		dc.b	$1B
		dc.b	$2A					; *
		dc.b	$71					; q
		dc.b	$AC					; 
		dc.b	$E0					; 
		dc.b	$CE					; 
		dc.b	$B3					; 
		dc.b	$6F					; o
		dc.b	$36					; 6
		dc.b	$2B					; +
		dc.b	$8B					; 
		dc.b	$FB					; 
		dc.b	$DD					; 
		dc.b	$92					; 
		dc.b	$43					; C
		dc.b	$A
		dc.b	$F
		dc.b	$7F					; 
		dc.b	$D9					; 
		dc.b	$E1					; 
		dc.b	$B0					; 
		dc.b	$64					; d
		dc.b	$64					; d
		dc.b	$96					; 
		dc.b	$AC					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$96					; 
		dc.b	$6D					; m
		dc.b	$8C					; 
		dc.b	$AD					; 
		dc.b	$86					; 
		dc.b	$6F					; o
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$9B					; 
		dc.b	$C9					; 
		dc.b	$AF					; 
		dc.b	$86					; 
		dc.b	$5F					; _
		dc.b	$1C
		dc.b	$12
		dc.b	$5F					; _
		dc.b	$9D					; 
		dc.b	$90					; 
		dc.b	$6F					; o
		dc.b	$58					; X
		dc.b	$70					; p
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$54					; T
		dc.b	$70					; p
		dc.b	$8B					; 
		dc.b	$B5					; 
		dc.b	$D4					; 
		dc.b	$BF					; 
		dc.b	$7A					; z
		dc.b	$47					; G
		dc.b	$33					; 3
		dc.b	$7D					; }
		dc.b	$B8					; 
		dc.b	$B1					; 
		dc.b	$57					; W
		dc.b	$C
		dc.b	$10
		dc.b	$21					; !
		dc.b	$78					; x
		dc.b	$B5					; 
		dc.b	$9D					; 
		dc.b	$66					; f
		dc.b	$40					; @
		dc.b	$3C					; <
		dc.b	$6C					; l
		dc.b	$B0					; 
		dc.b	$B6					; 
		dc.b	$A8					; 
		dc.b	$C1					; 
		dc.b	$CB					; 
		dc.b	$CE					; 
		dc.b	$C1					; 
		dc.b	$8E					; 
		dc.b	$80					; 
		dc.b	$9D					; 
		dc.b	$CE					; 
		dc.b	$E5					; 
		dc.b	$C4					; 
		dc.b	$58					; X
		dc.b	$40					; @
		dc.b	$31					; 1
		dc.b	$37					; 7
		dc.b	$74					; t
		dc.b	$97					; 
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$4B					; K
		dc.b	$53					; S
		dc.b	$7B					; {
		dc.b	$B7					; 
		dc.b	$DD					; 
		dc.b	$D6					; 
		dc.b	$A8					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$6F					; o
		dc.b	$14
		dc.b	1
		dc.b	1
		dc.b	$18
		dc.b	$65					; e
		dc.b	$B0					; 
		dc.b	$A4					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$63					; c
		dc.b	$70					; p
		dc.b	$CF					; 
		dc.b	$E6					; 
		dc.b	$C5					; 
		dc.b	$B1					; 
		dc.b	$71					; q
		dc.b	$55					; U
		dc.b	$9C					; 
		dc.b	$B7					; 
		dc.b	$DA					; 
		dc.b	$E8					; 
		dc.b	$77					; w
		dc.b	$1C
		dc.b	$28					; (
		dc.b	$26					; &
		dc.b	$53					; S
		dc.b	$9F					; 
		dc.b	$60					; `
		dc.b	$26					; &
		dc.b	$1B
		dc.b	$E
		dc.b	$25					; %
		dc.b	$A5					; 
		dc.b	$E8					; 
		dc.b	$F5					; 
		dc.b	$E8					; 
		dc.b	$D9					; 
		dc.b	$D9					; 
		dc.b	$D7					; 
		dc.b	$CD					; 
		dc.b	$BF					; 
		dc.b	$75					; u
		dc.b	$3C					; <
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$37					; 7
		dc.b	$8E					; 
		dc.b	$B8					; 
		dc.b	$66					; f
		dc.b	$34					; 4
		dc.b	$31					; 1
		dc.b	$72					; r
		dc.b	$E5					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$D1					; 
		dc.b	$32					; 2
		dc.b	1
		dc.b	$51					; Q
		dc.b	$AC					; 
		dc.b	$B7					; 
		dc.b	$4F					; O
		dc.b	9
		dc.b	$16
		dc.b	$24					; $
		dc.b	$8A					; 
		dc.b	$F5					; 
		dc.b	$ED					; 
		dc.b	$AB					; 
		dc.b	$63					; c
		dc.b	$35					; 5
		dc.b	$3D					; =
		dc.b	$B4					; 
		dc.b	$E6					; 
		dc.b	$D0					; 
		dc.b	$C1					; 
		dc.b	$C1					; 
		dc.b	$D1					; 
		dc.b	$C6					; 
		dc.b	$BB					; 
		dc.b	$D7					; 
		dc.b	$DB					; 
		dc.b	$74					; t
		dc.b	$46					; F
		dc.b	$40					; @
		dc.b	$3E					; >
		dc.b	$34					; 4
		dc.b	$17
		dc.b	$1A
		dc.b	$12
		dc.b	4
		dc.b	$58					; X
		dc.b	$BD					; 
		dc.b	$CF					; 
		dc.b	$BB					; 
		dc.b	$B2					; 
		dc.b	$CB					; 
		dc.b	$D0					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$8B					; 
		dc.b	$37					; 7
		dc.b	$21					; !
		dc.b	$2E					; .
		dc.b	$28					; (
		dc.b	$33					; 3
		dc.b	$7B					; {
		dc.b	$E6					; 
		dc.b	$FB					; 
		dc.b	$BC					; 
		dc.b	$30					; 0
		dc.b	1
		dc.b	$11
		dc.b	$BA					; 
		dc.b	$FE					; 
		dc.b	$FB					; 
		dc.b	$D4					; 
		dc.b	$B8					; 
		dc.b	$B6					; 
		dc.b	$D4					; 
		dc.b	$E6					; 
		dc.b	$E4					; 
		dc.b	$A8					; 
		dc.b	$37					; 7
		dc.b	$E
		dc.b	1
		dc.b	1
		dc.b	9
		dc.b	$13
		dc.b	$36					; 6
		dc.b	$32					; 2
		dc.b	$A
		dc.b	$3C					; <
		dc.b	$A2					; 
		dc.b	$E2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$DB					; 
		dc.b	$F5					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$DD					; 
		dc.b	$9B					; 
		dc.b	$3A					; :
		dc.b	$1C
		dc.b	1
		dc.b	$36					; 6
		dc.b	$46					; F
		dc.b	$49					; I
		dc.b	$57					; W
		dc.b	$5C					; \
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$E8					; 
		dc.b	$F2					; 
		dc.b	$D2					; 
		dc.b	$AF					; 
		dc.b	$A6					; 
		dc.b	$BF					; 
		dc.b	$E6					; 
		dc.b	$EB					; 
		dc.b	$D8					; 
		dc.b	$9F					; 
		dc.b	$6A					; j
		dc.b	$25					; %
		dc.b	1
		dc.b	1
		dc.b	$1F
		dc.b	$44					; D
		dc.b	$40					; @
		dc.b	$1D
		dc.b	$1F
		dc.b	$36					; 6
		dc.b	$91					; 
		dc.b	$96					; 
		dc.b	$A7					; 
		dc.b	$E1					; 
		dc.b	$CF					; 
		dc.b	$BF					; 
		dc.b	$CE					; 
		dc.b	$CA					; 
		dc.b	$EC					; 
		dc.b	$A2					; 
		dc.b	$1C
		dc.b	1
		dc.b	$11
		dc.b	$19
		dc.b	$21					; !
		dc.b	$29					; )
		dc.b	$27					; '
		dc.b	$30					; 0
		dc.b	$2A					; *
		dc.b	$5F					; _
		dc.b	$D1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E6					; 
		dc.b	$D6					; 
		dc.b	$DB					; 
		dc.b	$D7					; 
		dc.b	$EC					; 
		dc.b	$7D					; }
		dc.b	$E
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$1A
		dc.b	$34					; 4
		dc.b	$29					; )
		dc.b	$35					; 5
		dc.b	$4F					; O
		dc.b	$62					; b
		dc.b	$82					; 
		dc.b	$93					; 
		dc.b	$AD					; 
		dc.b	$B4					; 
		dc.b	$AF					; 
		dc.b	$A9					; 
		dc.b	$BF					; 
		dc.b	$D6					; 
		dc.b	$BA					; 
		dc.b	$8B					; 
		dc.b	$1F
		dc.b	$E
		dc.b	$29					; )
		dc.b	$4D					; M
		dc.b	$57					; W
		dc.b	$4D					; M
		dc.b	$3F					; ?
		dc.b	$3F					; ?
		dc.b	$33					; 3
		dc.b	$46					; F
		dc.b	$98					; 
		dc.b	$C1					; 
		dc.b	$E4					; 
		dc.b	$FB					; 
		dc.b	$F7					; 
		dc.b	$ED					; 
		dc.b	$CD					; 
		dc.b	$BE					; 
		dc.b	$CF					; 
		dc.b	$CB					; 
		dc.b	$7F					; 
		dc.b	$2D					; -
		dc.b	$E
		dc.b	1
		dc.b	5
		dc.b	$22					; "
		dc.b	$3E					; >
		dc.b	$4B					; K
		dc.b	$44					; D
		dc.b	$1C
		dc.b	$4B					; K
		dc.b	$D5					; 
		dc.b	$FE					; 
		dc.b	$ED					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E8					; 
		dc.b	$D6					; 
		dc.b	$CA					; 
		dc.b	$BF					; 
		dc.b	$CF					; 
		dc.b	$9B					; 
		dc.b	$1F
		dc.b	$C
		dc.b	$14
		dc.b	$32					; 2
		dc.b	$40					; @
		dc.b	$1A
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$15
		dc.b	$95					; 
		dc.b	$E4					; 
		dc.b	$CF					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$CF					; 
		dc.b	$CD					; 
		dc.b	$C4					; 
		dc.b	$CC					; 
		dc.b	$C0					; 
		dc.b	$54					; T
		dc.b	$1A
		dc.b	$F
		dc.b	$21					; !
		dc.b	$3E					; >
		dc.b	$50					; P
		dc.b	$42					; B
		dc.b	$39					; 9
		dc.b	$2E					; .
		dc.b	$41					; A
		dc.b	$9C					; 
		dc.b	$B8					; 
		dc.b	$B8					; 
		dc.b	$CF					; 
		dc.b	$D7					; 
		dc.b	$D5					; 
		dc.b	$C1					; 
		dc.b	$B1					; 
		dc.b	$BD					; 
		dc.b	$C9					; 
		dc.b	$96					; 
		dc.b	$27					; '
		dc.b	$1A
		dc.b	$1B
		dc.b	$F
		dc.b	$15
		dc.b	$23					; #
		dc.b	$31					; 1
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$6F					; o
		dc.b	$CD					; 
		dc.b	$F6					; 
		dc.b	$EB					; 
		dc.b	$F1					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EF					; 
		dc.b	$D8					; 
		dc.b	$D3					; 
		dc.b	$BC					; 
		dc.b	$52					; R
		dc.b	$14
		dc.b	$24					; $
		dc.b	$29					; )
		dc.b	$36					; 6
		dc.b	$30					; 0
		dc.b	$1E
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$70					; p
		dc.b	$EE					; 
		dc.b	$E5					; 
		dc.b	$C6					; 
		dc.b	$BE					; 
		dc.b	$BA					; 
		dc.b	$C3					; 
		dc.b	$CF					; 
		dc.b	$DC					; 
		dc.b	$EE					; 
		dc.b	$F0					; 
		dc.b	$AC					; 
		dc.b	$40					; @
		dc.b	$1F
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$A
		dc.b	$2B					; +
		dc.b	$36					; 6
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$AF					; 
		dc.b	$D0					; 
		dc.b	$C7					; 
		dc.b	$CA					; 
		dc.b	$C9					; 
		dc.b	$C9					; 
		dc.b	$D0					; 
		dc.b	$E1					; 
		dc.b	$F9					; 
		dc.b	$FE					; 
		dc.b	$C9					; 
		dc.b	$47					; G
		dc.b	$2D					; -
		dc.b	$38					; 8
		dc.b	$2B					; +
		dc.b	$10
		dc.b	$12
		dc.b	$24					; $
		dc.b	$33					; 3
		dc.b	$33					; 3
		dc.b	$4D					; M
		dc.b	$BF					; 
		dc.b	$F5					; 
		dc.b	$E8					; 
		dc.b	$CA					; 
		dc.b	$C9					; 
		dc.b	$DB					; 
		dc.b	$E0					; 
		dc.b	$DA					; 
		dc.b	$D5					; 
		dc.b	$E3					; 
		dc.b	$D2					; 
		dc.b	$56					; V
		dc.b	1
		dc.b	5
		dc.b	$13
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	7
		dc.b	$1C
		dc.b	$15
		dc.b	$6F					; o
		dc.b	$EB					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E0					; 
		dc.b	$E1					; 
		dc.b	$D7					; 
		dc.b	$C5					; 
		dc.b	$B4					; 
		dc.b	$B6					; 
		dc.b	$DA					; 
		dc.b	$CD					; 
		dc.b	$5A					; Z
		dc.b	$3C					; <
		dc.b	$2E					; .
		dc.b	$29					; )
		dc.b	$23					; #
		dc.b	$1D
		dc.b	$34					; 4
		dc.b	$51					; Q
		dc.b	$4C					; L
		dc.b	$69					; i
		dc.b	$C9					; 
		dc.b	$E8					; 
		dc.b	$D1					; 
		dc.b	$BA					; 
		dc.b	$C9					; 
		dc.b	$DA					; 
		dc.b	$E0					; 
		dc.b	$D8					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$DB					; 
		dc.b	$53					; S
		dc.b	$12
		dc.b	3
		dc.b	$11
		dc.b	$10
		dc.b	$13
		dc.b	$29					; )
		dc.b	$3A					; :
		dc.b	$40					; @
		dc.b	$49					; I
		dc.b	$95					; 
		dc.b	$CA					; 
		dc.b	$D5					; 
		dc.b	$D2					; 
		dc.b	$CA					; 
		dc.b	$C2					; 
		dc.b	$B3					; 
		dc.b	$B6					; 
		dc.b	$BF					; 
		dc.b	$BF					; 
		dc.b	$C1					; 
		dc.b	$89					; 
		dc.b	$25					; %
		dc.b	$21					; !
		dc.b	$2F					; /
		dc.b	$26					; &
		dc.b	$18
		dc.b	$26					; &
		dc.b	$33					; 3
		dc.b	$38					; 8
		dc.b	$39					; 9
		dc.b	$58					; X
		dc.b	$A5					; 
		dc.b	$AC					; 
		dc.b	$BD					; 
		dc.b	$CB					; 
		dc.b	$C9					; 
		dc.b	$B7					; 
		dc.b	$C6					; 
		dc.b	$EC					; 
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	$E0					; 
		dc.b	$7B					; {
		dc.b	$45					; E
		dc.b	$30					; 0
		dc.b	7
		dc.b	1
		dc.b	1
		dc.b	$E
		dc.b	$25					; %
		dc.b	$4A					; J
		dc.b	$63					; c
		dc.b	$9D					; 
		dc.b	$D8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F6					; 
		dc.b	$F4					; 
		dc.b	$E3					; 
		dc.b	$D1					; 
		dc.b	$D1					; 
		dc.b	$BC					; 
		dc.b	$9B					; 
		dc.b	$76					; v
		dc.b	3
		dc.b	1
		dc.b	$40					; @
		dc.b	$61					; a
		dc.b	$58					; X
		dc.b	$39					; 9
		dc.b	$29					; )
		dc.b	$31					; 1
		dc.b	$48					; H
		dc.b	$6B					; k
		dc.b	$C7					; 
		dc.b	$F1					; 
		dc.b	$F9					; 
		dc.b	$F2					; 
		dc.b	$E6					; 
		dc.b	$CA					; 
		dc.b	$CF					; 
		dc.b	$E5					; 
		dc.b	$EF					; 
		dc.b	$D0					; 
		dc.b	$9B					; 
		dc.b	$32					; 2
		dc.b	$F
		dc.b	$15
		dc.b	1
		dc.b	1
		dc.b	$14
		dc.b	$28					; (
		dc.b	$2A					; *
		dc.b	$2B					; +
		dc.b	$33					; 3
		dc.b	$81					; 
		dc.b	$A1					; 
		dc.b	$A5					; 
		dc.b	$CF					; 
		dc.b	$DA					; 
		dc.b	$EA					; 
		dc.b	$FE					; 
		dc.b	$FC					; 
		dc.b	$D0					; 
		dc.b	$B3					; 
		dc.b	$C6					; 
		dc.b	$CD					; 
		dc.b	$7D					; }
		dc.b	$5D					; ]
		dc.b	$39					; 9
		dc.b	$1F
		dc.b	$16
		dc.b	3
		dc.b	1
		dc.b	1
		dc.b	$1C
		dc.b	$63					; c
		dc.b	$B9					; 
		dc.b	$BC					; 
		dc.b	$B4					; 
		dc.b	$D1					; 
		dc.b	$EA					; 
		dc.b	$E2					; 
		dc.b	$D6					; 
		dc.b	$E3					; 
		dc.b	$FA					; 
		dc.b	$F1					; 
		dc.b	$D1					; 
		dc.b	$75					; u
		dc.b	$C
		dc.b	$13
		dc.b	$32					; 2
		dc.b	$42					; B
		dc.b	$3C					; <
		dc.b	$27					; '
		dc.b	$17
		dc.b	9
		dc.b	1
		dc.b	$3E					; >
		dc.b	$A6					; 
		dc.b	$CB					; 
		dc.b	$F4					; 
		dc.b	$FE					; 
		dc.b	$DE					; 
		dc.b	$C3					; 
		dc.b	$BB					; 
		dc.b	$BC					; 
		dc.b	$DC					; 
		dc.b	$ED					; 
		dc.b	$DC					; 
		dc.b	$93					; 
		dc.b	$39					; 9
		dc.b	$1B
		dc.b	$1C
		dc.b	$1D
		dc.b	$36					; 6
		dc.b	$46					; F
		dc.b	$2D					; -
		dc.b	9
		dc.b	1
		dc.b	$34					; 4
		dc.b	$A9					; 
		dc.b	$ED					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$D2					; 
		dc.b	$AB					; 
		dc.b	$A4					; 
		dc.b	$AA					; 
		dc.b	$C7					; 
		dc.b	$BD					; 
		dc.b	$68					; h
		dc.b	$48					; H
		dc.b	$39					; 9
		dc.b	$1C
		dc.b	1
		dc.b	$1A
		dc.b	$41					; A
		dc.b	$34					; 4
		dc.b	$2F					; /
		dc.b	$46					; F
		dc.b	$84					; 
		dc.b	$AB					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$D2					; 
		dc.b	$EF					; 
		dc.b	$E2					; 
		dc.b	$D8					; 
		dc.b	$D2					; 
		dc.b	$D0					; 
		dc.b	$C8					; 
		dc.b	$6A					; j
		dc.b	$4F					; O
		dc.b	$55					; U
		dc.b	$34					; 4
		dc.b	6
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	$18
		dc.b	$2F					; /
		dc.b	$67					; g
		dc.b	$BD					; 
		dc.b	$D2					; 
		dc.b	$CE					; 
		dc.b	$D1					; 
		dc.b	$D0					; 
		dc.b	$D7					; 
		dc.b	$E7					; 
		dc.b	$EE					; 
		dc.b	$E7					; 
		dc.b	$C8					; 
		dc.b	$CB					; 
		dc.b	$94					; 
		dc.b	$2F					; /
		dc.b	$17
		dc.b	$22					; "
		dc.b	$31					; 1
		dc.b	$1F
		dc.b	$10
		dc.b	$13
		dc.b	$13
		dc.b	$18
		dc.b	$51					; Q
		dc.b	$B5					; 
		dc.b	$FA					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E2					; 
		dc.b	$D3					; 
		dc.b	$D9					; 
		dc.b	$D8					; 
		dc.b	$E0					; 
		dc.b	$D5					; 
		dc.b	$C9					; 
		dc.b	$C7					; 
		dc.b	$68					; h
		dc.b	$32					; 2
		dc.b	$3B					; ;
		dc.b	$54					; T
		dc.b	$5F					; _
		dc.b	$4E					; N
		dc.b	$27					; '
		dc.b	$E
		dc.b	$D
		dc.b	$19
		dc.b	$4D					; M
		dc.b	$AF					; 
		dc.b	$DC					; 
		dc.b	$F2					; 
		dc.b	$F4					; 
		dc.b	$E4					; 
		dc.b	$CF					; 
		dc.b	$C8					; 
		dc.b	$C1					; 
		dc.b	$C3					; 
		dc.b	$CC					; 
		dc.b	$D7					; 
		dc.b	$D8					; 
		dc.b	$69					; i
		dc.b	$17
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	1
		dc.b	8
		dc.b	5
		dc.b	2
		dc.b	1
		dc.b	$38					; 8
		dc.b	$9F					; 
		dc.b	$CB					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$CB					; 
		dc.b	$C9					; 
		dc.b	$E5					; 
		dc.b	$B4					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D8					; 
		dc.b	$A0					; 
		dc.b	$35					; 5
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$2A					; *
		dc.b	$35					; 5
		dc.b	$30					; 0
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$57					; W
		dc.b	$B1					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D4					; 
		dc.b	$83					; 
		dc.b	$2A					; *
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$3B					; ;
		dc.b	$48					; H
		dc.b	$33					; 3
		dc.b	$25					; %
		dc.b	$2E					; .
		dc.b	$67					; g
		dc.b	$C4					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$C5					; 
		dc.b	$58					; X
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$44					; D
		dc.b	$4E					; N
		dc.b	$33					; 3
		dc.b	$26					; &
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$66					; f
		dc.b	$C9					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D7					; 
		dc.b	$93					; 
		dc.b	$2F					; /
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$3F					; ?
		dc.b	$45					; E
		dc.b	$29					; )
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$39					; 9
		dc.b	$A4					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D1					; 
		dc.b	$75					; u
		dc.b	$2A					; *
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$29					; )
		dc.b	$34					; 4
		dc.b	$27					; '
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$5D					; ]
		dc.b	$C7					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$AC					; 
		dc.b	$3C					; <
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$26					; &
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$2B					; +
		dc.b	$8E					; 
		dc.b	$D6					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D0					; 
		dc.b	$70					; p
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$35					; 5
		dc.b	$A3					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D2					; 
		dc.b	$6C					; l
		dc.b	$28					; (
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$27					; '
		dc.b	$27					; '
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$2B					; +
		dc.b	$7F					; 
		dc.b	$D2					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$A7					; 
		dc.b	$39					; 9
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$41					; A
		dc.b	$39					; 9
		dc.b	$26					; &
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$44					; D
		dc.b	$AF					; 
		dc.b	$D9					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D7					; 
		dc.b	$D6					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D9					; 
		dc.b	$A9					; 
		dc.b	$3B					; ;
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$3A					; :
		dc.b	$31					; 1
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$39					; 9
		dc.b	$9D					; 
		dc.b	$D7					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D5					; 
		dc.b	$85					; 
		dc.b	$2E					; .
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$51					; Q
		dc.b	$AB					; 
		dc.b	$D7					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D9					; 
		dc.b	$A9					; 
		dc.b	$4E					; N
		dc.b	$28					; (
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$2B					; +
		dc.b	$4A					; J
		dc.b	$67					; g
		dc.b	$85					; 
		dc.b	$A2					; 
		dc.b	$BF					; 
		dc.b	$D4					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D4					; 
		dc.b	$A2					; 
		dc.b	$70					; p
		dc.b	$5F					; _
		dc.b	$5F					; _
		dc.b	$60					; `
		dc.b	$5F					; _
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$78					; x
		dc.b	$5D					; ]
		dc.b	$41					; A
		dc.b	$30					; 0
		dc.b	$36					; 6
		dc.b	$51					; Q
		dc.b	$7E					; ~
		dc.b	$B0					; 
		dc.b	$D2					; 
		dc.b	$DA					; 
		dc.b	$D7					; 
		dc.b	$C5					; 
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$BF					; 
		dc.b	$D5					; 
		dc.b	$DA					; 
		dc.b	$D9					; 
		dc.b	$C5					; 
		dc.b	$9A					; 
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$72					; r
		dc.b	$86					; 
		dc.b	$93					; 
		dc.b	$86					; 
		dc.b	$62					; b
		dc.b	$34					; 4
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$3B					; ;
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$99					; 
		dc.b	$C5					; 
		dc.b	$D9					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$CD					; 
		dc.b	$B7					; 
		dc.b	$AE					; 
		dc.b	$9E					; 
		dc.b	$80					; 
		dc.b	$5D					; ]
		dc.b	$41					; A
		dc.b	$29					; )
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$27					; '
		dc.b	$30					; 0
		dc.b	$3B					; ;
		dc.b	$41					; A
		dc.b	$44					; D
		dc.b	$49					; I
		dc.b	$5B					; [
		dc.b	$81					; 
		dc.b	$B4					; 
		dc.b	$D6					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D6					; 
		dc.b	$B7					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$74					; t
		dc.b	$58					; X
		dc.b	$3F					; ?
		dc.b	$3C					; <
		dc.b	$58					; X
		dc.b	$85					; 
		dc.b	$A7					; 
		dc.b	$AA					; 
		dc.b	$91					; 
		dc.b	$67					; g
		dc.b	$41					; A
		dc.b	$2F					; /
		dc.b	$3C					; <
		dc.b	$5C					; \
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$72					; r
		dc.b	$55					; U
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$BF					; 
		dc.b	$D8					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$C2					; 
		dc.b	$8A					; 
		dc.b	$6A					; j
		dc.b	$6F					; o
		dc.b	$88					; 
		dc.b	$A0					; 
		dc.b	$9E					; 
		dc.b	$80					; 
		dc.b	$55					; U
		dc.b	$37					; 7
		dc.b	$36					; 6
		dc.b	$59					; Y
		dc.b	$95					; 
		dc.b	$C7					; 
		dc.b	$DA					; 
		dc.b	$D1					; 
		dc.b	$A8					; 
		dc.b	$71					; q
		dc.b	$4F					; O
		dc.b	$50					; P
		dc.b	$70					; p
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$7E					; ~
		dc.b	$4D					; M
		dc.b	$2A					; *
		dc.b	$25					; %
		dc.b	$2B					; +
		dc.b	$55					; U
		dc.b	$8F					; 
		dc.b	$B2					; 
		dc.b	$B5					; 
		dc.b	$98					; 
		dc.b	$73					; s
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$A2					; 
		dc.b	$CC					; 
		dc.b	$D9					; 
		dc.b	$C9					; 
		dc.b	$A4					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$BD					; 
		dc.b	$D7					; 
		dc.b	$DA					; 
		dc.b	$D4					; 
		dc.b	$A5					; 
		dc.b	$5D					; ]
		dc.b	$2E					; .
		dc.b	$25					; %
		dc.b	$26					; &
		dc.b	$31					; 1
		dc.b	$3B					; ;
		dc.b	$39					; 9
		dc.b	$29					; )
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$4D					; M
		dc.b	$95					; 
		dc.b	$CC					; 
		dc.b	$DA					; 
		dc.b	$DA					; 
		dc.b	$D6					; 
		dc.b	$B8					; 
		dc.b	$A0					; 
		dc.b	$A8					; 
		dc.b	$C4					; 
		dc.b	$D6					; 
		dc.b	$D7					; 
		dc.b	$B9					; 
		dc.b	$81					; 
		dc.b	$52					; R
		dc.b	$39					; 9
		dc.b	$45					; E
		dc.b	$66					; f
		dc.b	$89					; 
		dc.b	$A2					; 
		dc.b	$A3					; 
		dc.b	$8B					; 
		dc.b	$65					; e
		dc.b	$46					; F
		dc.b	$40					; @
		dc.b	$4F					; O
		dc.b	$66					; f
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$68					; h
		dc.b	$4F					; O
		dc.b	$40					; @
		dc.b	$4B					; K
		dc.b	$6D					; m
		dc.b	$9A					; 
		dc.b	$C2					; 
		dc.b	$D1					; 
		dc.b	$C0					; 
		dc.b	$9F					; 
		dc.b	$7F					; 
		dc.b	$77					; w
		dc.b	$81					; 
		dc.b	$95					; 
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$70					; p
		dc.b	$4F					; O
		dc.b	$3B					; ;
		dc.b	$3D					; =
		dc.b	$55					; U
		dc.b	$7A					; z
		dc.b	$9F					; 
		dc.b	$B0					; 
		dc.b	$B0					; 
		dc.b	$A7					; 
		dc.b	$A5					; 
		dc.b	$AD					; 
		dc.b	$BB					; 
		dc.b	$CC					; 
		dc.b	$D0					; 
		dc.b	$BA					; 
		dc.b	$8E					; 
		dc.b	$5D					; ]
		dc.b	$39					; 9
		dc.b	$2B					; +
		dc.b	$30					; 0
		dc.b	$3E					; >
		dc.b	$49					; I
		dc.b	$43					; C
		dc.b	$33					; 3
		dc.b	$2B					; +
		dc.b	$34					; 4
		dc.b	$53					; S
		dc.b	$7E					; ~
		dc.b	$A6					; 
		dc.b	$BD					; 
		dc.b	$BD					; 
		dc.b	$AF					; 
		dc.b	$9D					; 
		dc.b	$91					; 
		dc.b	$90					; 
		dc.b	$99					; 
		dc.b	$A3					; 
		dc.b	$A7					; 
		dc.b	$A1					; 
		dc.b	$91					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$8C					; 
		dc.b	$A9					; 
		dc.b	$C2					; 
		dc.b	$CF					; 
		dc.b	$C9					; 
		dc.b	$B2					; 
		dc.b	$93					; 
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$5F					; _
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$29					; )
		dc.b	$25					; %
		dc.b	$2B					; +
		dc.b	$43					; C
		dc.b	$62					; b
		dc.b	$7F					; 
		dc.b	$90					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$9F					; 
		dc.b	$A6					; 
		dc.b	$A9					; 
		dc.b	$A5					; 
		dc.b	$A0					; 
		dc.b	$97					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$A0					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$BA					; 
		dc.b	$B4					; 
		dc.b	$AB					; 
		dc.b	$9C					; 
		dc.b	$8D					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$66					; f
		dc.b	$63					; c
		dc.b	$5B					; [
		dc.b	$4C					; L
		dc.b	$3E					; >
		dc.b	$34					; 4
		dc.b	$2E					; .
		dc.b	$2E					; .
		dc.b	$32					; 2
		dc.b	$36					; 6
		dc.b	$36					; 6
		dc.b	$33					; 3
		dc.b	$34					; 4
		dc.b	$39					; 9
		dc.b	$43					; C
		dc.b	$50					; P
		dc.b	$5D					; ]
		dc.b	$66					; f
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$91					; 
		dc.b	$99					; 
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$B6					; 
		dc.b	$BE					; 
		dc.b	$C1					; 
		dc.b	$B8					; 
		dc.b	$9B					; 
		dc.b	$6E					; n
		dc.b	$47					; G
		dc.b	$31					; 1
		dc.b	$2E					; .
		dc.b	$3B					; ;
		dc.b	$4E					; N
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$A0					; 
		dc.b	$C0					; 
		dc.b	$D5					; 
		dc.b	$D5					; 
		dc.b	$CA					; 
		dc.b	$BA					; 
		dc.b	$AF					; 
		dc.b	$A3					; 
		dc.b	$99					; 
		dc.b	$8D					; 
		dc.b	$7D					; }
		dc.b	$69					; i
		dc.b	$5D					; ]
		dc.b	$5B					; [
		dc.b	$5F					; _
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$9A					; 
		dc.b	$9A					; 
		dc.b	$99					; 
		dc.b	$96					; 
		dc.b	$95					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$9D					; 
		dc.b	$98					; 
		dc.b	$93					; 
		dc.b	$90					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$94					; 
		dc.b	$A1					; 
		dc.b	$B1					; 
		dc.b	$BC					; 
		dc.b	$C2					; 
		dc.b	$BE					; 
		dc.b	$AE					; 
		dc.b	$9D					; 
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$6A					; j
		dc.b	$52					; R
		dc.b	$3C					; <
		dc.b	$31					; 1
		dc.b	$33					; 3
		dc.b	$3E					; >
		dc.b	$4F					; O
		dc.b	$58					; X
		dc.b	$5D					; ]
		dc.b	$61					; a
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$97					; 
		dc.b	$A0					; 
		dc.b	$99					; 
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$49					; I
		dc.b	$3B					; ;
		dc.b	$39					; 9
		dc.b	$41					; A
		dc.b	$55					; U
		dc.b	$76					; v
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$BC					; 
		dc.b	$CA					; 
		dc.b	$D1					; 
		dc.b	$CD					; 
		dc.b	$C0					; 
		dc.b	$A8					; 
		dc.b	$7F					; 
		dc.b	$4E					; N
		dc.b	$2C					; ,
		dc.b	$25					; %
		dc.b	$25					; %
		dc.b	$28					; (
		dc.b	$3B					; ;
		dc.b	$54					; T
		dc.b	$67					; g
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$92					; 
		dc.b	$A3					; 
		dc.b	$B2					; 
		dc.b	$BB					; 
		dc.b	$B9					; 
		dc.b	$AD					; 
		dc.b	$9D					; 
		dc.b	$8C					; 
		dc.b	$7B					; {
		dc.b	$70					; p
		dc.b	$66					; f
		dc.b	$5D					; ]
		dc.b	$5B					; [
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$70					; p
		dc.b	$7E					; ~
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$90					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$90					; 
		dc.b	$9A					; 
		dc.b	$A3					; 
		dc.b	$A9					; 
		dc.b	$A9					; 
		dc.b	$A5					; 
		dc.b	$A0					; 
		dc.b	$9A					; 
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$88					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$69					; i
		dc.b	$66					; f
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$55					; U
		dc.b	$50					; P
		dc.b	$55					; U
		dc.b	$5B					; [
		dc.b	$64					; d
		dc.b	$6B					; k
		dc.b	$69					; i
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$5D					; ]
		dc.b	$67					; g
		dc.b	$7A					; z
		dc.b	$8D					; 
		dc.b	$98					; 
		dc.b	$97					; 
		dc.b	$91					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$9D					; 
		dc.b	$9D					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$9A					; 
		dc.b	$9E					; 
		dc.b	$9C					; 
		dc.b	$98					; 
		dc.b	$98					; 
		dc.b	$99					; 
		dc.b	$93					; 
		dc.b	$85					; 
		dc.b	$73					; s
		dc.b	$62					; b
		dc.b	$58					; X
		dc.b	$54					; T
		dc.b	$53					; S
		dc.b	$55					; U
		dc.b	$55					; U
		dc.b	$58					; X
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$8E					; 
		dc.b	$98					; 
		dc.b	$9C					; 
		dc.b	$9A					; 
		dc.b	$97					; 
		dc.b	$92					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$88					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$8D					; 
		dc.b	$97					; 
		dc.b	$9A					; 
		dc.b	$97					; 
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$95					; 
		dc.b	$97					; 
		dc.b	$94					; 
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$71					; q
		dc.b	$6B					; k
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$6D					; m
		dc.b	$6F					; o
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$6D					; m
		dc.b	$67					; g
		dc.b	$66					; f
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$74					; t
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$81					; 
		dc.b	$82					; 
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$71					; q
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$60					; `
		dc.b	$5D					; ]
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$6C					; l
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$71					; q
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$85					; 
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$76					; v
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$73					; s
		dc.b	$76					; v
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$FF
unk_2AEDA:	dc.b	$80					; 		; DATA XREF: ROM:0001B054o
					; ROM:0001B06Co
		dc.b	$9E					; 
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$67					; g
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$E9					; 
		dc.b	$85					; 
		dc.b	$6E					; n
		dc.b	$E4					; 
		dc.b	$B8					; 
		dc.b	$ED					; 
		dc.b	$EB					; 
		dc.b	$B4					; 
		dc.b	$33					; 3
		dc.b	$77					; w
		dc.b	$85					; 
		dc.b	$59					; Y
		dc.b	$6B					; k
		dc.b	$50					; P
		dc.b	$20
		dc.b	$27					; '
		dc.b	$25					; %
		dc.b	$31					; 1
		dc.b	$54					; T
		dc.b	$69					; i
		dc.b	$6E					; n
		dc.b	$57					; W
		dc.b	$54					; T
		dc.b	$77					; w
		dc.b	$AC					; 
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$82					; 
		dc.b	$75					; u
		dc.b	$35					; 5
		dc.b	$3E					; >
		dc.b	$48					; H
		dc.b	$79					; y
		dc.b	$B6					; 
		dc.b	$D4					; 
		dc.b	$E0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$A0					; 
		dc.b	$DD					; 
		dc.b	$37					; 7
		dc.b	$4A					; J
		dc.b	$41					; A
		dc.b	$10
		dc.b	$25					; %
		dc.b	$15
		dc.b	$3C					; <
		dc.b	$4E					; N
		dc.b	$60					; `
		dc.b	$99					; 
		dc.b	$97					; 
		dc.b	$6B					; k
		dc.b	$52					; R
		dc.b	$3C					; <
		dc.b	$2E					; .
		dc.b	$41					; A
		dc.b	$59					; Y
		dc.b	$43					; C
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$4E					; N
		dc.b	$4C					; L
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$3A					; :
		dc.b	$48					; H
		dc.b	$67					; g
		dc.b	$82					; 
		dc.b	$B4					; 
		dc.b	$C9					; 
		dc.b	$F2					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$E0					; 
		dc.b	$DD					; 
		dc.b	$9E					; 
		dc.b	$79					; y
		dc.b	$5B					; [
		dc.b	$31					; 1
		dc.b	$27					; '
		dc.b	$37					; 7
		dc.b	$2C					; ,
		dc.b	$62					; b
		dc.b	$5B					; [
		dc.b	$5E					; ^
		dc.b	$4A					; J
		dc.b	$2C					; ,
		dc.b	$1E
		dc.b	$10
		dc.b	$2A					; *
		dc.b	$43					; C
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$64					; d
		dc.b	$3E					; >
		dc.b	$1E
		dc.b	$13
		dc.b	5
		dc.b	$1C
		dc.b	$37					; 7
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$85					; 
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$6B					; k
		dc.b	$60					; `
		dc.b	$54					; T
		dc.b	$3E					; >
		dc.b	$50					; P
		dc.b	$41					; A
		dc.b	$4A					; J
		dc.b	$9E					; 
		dc.b	$B6					; 
		dc.b	$E6					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FB					; 
		dc.b	$EB					; 
		dc.b	$F9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$F4					; 
		dc.b	$D6					; 
		dc.b	$CD					; 
		dc.b	$A5					; 
		dc.b	$95					; 
		dc.b	$9E					; 
		dc.b	$89					; 
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$B4					; 
		dc.b	$CF					; 
		dc.b	$C4					; 
		dc.b	$D6					; 
		dc.b	$CD					; 
		dc.b	$E0					; 
		dc.b	$B4					; 
		dc.b	$97					; 
		dc.b	$67					; g
		dc.b	$41					; A
		dc.b	$3E					; >
		dc.b	$31					; 1
		dc.b	$31					; 1
		dc.b	$3C					; <
		dc.b	$3A					; :
		dc.b	$37					; 7
		dc.b	$1E
		dc.b	$E
		dc.b	9
		dc.b	1
		dc.b	3
		dc.b	$1A
		dc.b	$20
		dc.b	$2A					; *
		dc.b	$1E
		dc.b	$33					; 3
		dc.b	$41					; A
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$1C
		dc.b	$1A
		dc.b	$2A					; *
		dc.b	1
		dc.b	$E
		dc.b	$35					; 5
		dc.b	$3A					; :
		dc.b	$45					; E
		dc.b	$35					; 5
		dc.b	$52					; R
		dc.b	$4C					; L
		dc.b	$5B					; [
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$95					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$97					; 
		dc.b	$AE					; 
		dc.b	$A2					; 
		dc.b	$54					; T
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$6E					; n
		dc.b	$17
		dc.b	$5B					; [
		dc.b	$67					; g
		dc.b	$90					; 
		dc.b	$A7					; 
		dc.b	$5B					; [
		dc.b	$E2					; 
		dc.b	$B6					; 
		dc.b	$B8					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$ED					; 
		dc.b	$FE					; 
		dc.b	$BF					; 
		dc.b	$FE					; 
		dc.b	$A2					; 
		dc.b	$AE					; 
		dc.b	$B8					; 
		dc.b	$8B					; 
		dc.b	$BD					; 
		dc.b	$E9					; 
		dc.b	$D9					; 
		dc.b	$E6					; 
		dc.b	$BF					; 
		dc.b	$D4					; 
		dc.b	$A5					; 
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$3C					; <
		dc.b	$52					; R
		dc.b	$60					; `
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$52					; R
		dc.b	$50					; P
		dc.b	$85					; 
		dc.b	$AC					; 
		dc.b	$A2					; 
		dc.b	$B6					; 
		dc.b	$D2					; 
		dc.b	$92					; 
		dc.b	$B8					; 
		dc.b	$90					; 
		dc.b	$A2					; 
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$5B					; [
		dc.b	$41					; A
		dc.b	$45					; E
		dc.b	$43					; C
		dc.b	$25					; %
		dc.b	$48					; H
		dc.b	5
		dc.b	$3E					; >
		dc.b	1
		dc.b	$35					; 5
		dc.b	$2E					; .
		dc.b	$3C					; <
		dc.b	$69					; i
		dc.b	$31					; 1
		dc.b	$90					; 
		dc.b	$2C					; ,
		dc.b	$57					; W
		dc.b	$90					; 
		dc.b	$A2					; 
		dc.b	$A5					; 
		dc.b	$80					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$43					; C
		dc.b	$70					; p
		dc.b	$72					; r
		dc.b	$4C					; L
		dc.b	$C
		dc.b	$57					; W
		dc.b	$50					; P
		dc.b	$13
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	$72					; r
		dc.b	$48					; H
		dc.b	$60					; `
		dc.b	$70					; p
		dc.b	$A7					; 
		dc.b	$90					; 
		dc.b	$6B					; k
		dc.b	$3C					; <
		dc.b	$70					; p
		dc.b	$B0					; 
		dc.b	$64					; d
		dc.b	$80					; 
		dc.b	$D2					; 
		dc.b	$7B					; {
		dc.b	$E0					; 
		dc.b	$C2					; 
		dc.b	$BF					; 
		dc.b	$AE					; 
		dc.b	$C2					; 
		dc.b	$FD					; 
		dc.b	$69					; i
		dc.b	$B6					; 
		dc.b	$9C					; 
		dc.b	$95					; 
		dc.b	$E9					; 
		dc.b	$54					; T
		dc.b	$B4					; 
		dc.b	$B0					; 
		dc.b	$23					; #
		dc.b	$64					; d
		dc.b	$25					; %
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$23					; #
		dc.b	$70					; p
		dc.b	$A9					; 
		dc.b	$A2					; 
		dc.b	$4C					; L
		dc.b	$C6					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$BD					; 
		dc.b	$48					; H
		dc.b	$B6					; 
		dc.b	$52					; R
		dc.b	$7B					; {
		dc.b	$E6					; 
		dc.b	$CB					; 
		dc.b	$92					; 
		dc.b	$6E					; n
		dc.b	$E9					; 
		dc.b	$33					; 3
		dc.b	$AE					; 
		dc.b	$87					; 
		dc.b	$27					; '
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$FE					; 
		dc.b	$F7					; 
		dc.b	$B0					; 
		dc.b	$FE					; 
		dc.b	$7B					; {
		dc.b	$AC					; 
		dc.b	$A2					; 
		dc.b	$A2					; 
		dc.b	$D4					; 
		dc.b	$F4					; 
		dc.b	$87					; 
		dc.b	$CD					; 
		dc.b	$B0					; 
		dc.b	$20
		dc.b	$72					; r
		dc.b	$62					; b
		dc.b	$80					; 
		dc.b	$5E					; ^
		dc.b	$CB					; 
		dc.b	$6B					; k
		dc.b	$3A					; :
		dc.b	$A7					; 
		dc.b	$60					; `
		dc.b	$2E					; .
		dc.b	$77					; w
		dc.b	9
		dc.b	$85					; 
		dc.b	$62					; b
		dc.b	$92					; 
		dc.b	$7B					; {
		dc.b	$5E					; ^
		dc.b	$54					; T
		dc.b	$2A					; *
		dc.b	$50					; P
		dc.b	$52					; R
		dc.b	$5B					; [
		dc.b	$41					; A
		dc.b	$7E					; ~
		dc.b	$23					; #
		dc.b	$23					; #
		dc.b	$37					; 7
		dc.b	$1A
		dc.b	$45					; E
		dc.b	$4E					; N
		dc.b	$31					; 1
		dc.b	$2E					; .
		dc.b	$37					; 7
		dc.b	$10
		dc.b	$1A
		dc.b	$87					; 
		dc.b	$33					; 3
		dc.b	$3E					; >
		dc.b	$4E					; N
		dc.b	$15
		dc.b	$43					; C
		dc.b	$89					; 
		dc.b	$27					; '
		dc.b	$72					; r
		dc.b	$4E					; N
		dc.b	$31					; 1
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$9E					; 
		dc.b	$A9					; 
		dc.b	$92					; 
		dc.b	$E4					; 
		dc.b	$7B					; {
		dc.b	$D4					; 
		dc.b	$89					; 
		dc.b	$AC					; 
		dc.b	$FE					; 
		dc.b	$B8					; 
		dc.b	$9C					; 
		dc.b	$EB					; 
		dc.b	$A2					; 
		dc.b	$89					; 
		dc.b	$D6					; 
		dc.b	$7B					; {
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$AC					; 
		dc.b	$AE					; 
		dc.b	$BF					; 
		dc.b	$7B					; {
		dc.b	$BD					; 
		dc.b	$F4					; 
		dc.b	$99					; 
		dc.b	$B4					; 
		dc.b	$CB					; 
		dc.b	$80					; 
		dc.b	$BB					; 
		dc.b	$B2					; 
		dc.b	$B2					; 
		dc.b	$D2					; 
		dc.b	$B4					; 
		dc.b	$92					; 
		dc.b	$95					; 
		dc.b	$BD					; 
		dc.b	$70					; p
		dc.b	$99					; 
		dc.b	$70					; p
		dc.b	$41					; A
		dc.b	$69					; i
		dc.b	$48					; H
		dc.b	$67					; g
		dc.b	$97					; 
		dc.b	$85					; 
		dc.b	$5B					; [
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$2E					; .
		dc.b	$7E					; ~
		dc.b	$99					; 
		dc.b	$67					; g
		dc.b	$92					; 
		dc.b	$C9					; 
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$54					; T
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$72					; r
		dc.b	$35					; 5
		dc.b	$89					; 
		dc.b	$57					; W
		dc.b	$9C					; 
		dc.b	$9C					; 
		dc.b	$43					; C
		dc.b	$92					; 
		dc.b	$2A					; *
		dc.b	$7E					; ~
		dc.b	$92					; 
		dc.b	$67					; g
		dc.b	$50					; P
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$17
		dc.b	$6B					; k
		dc.b	$90					; 
		dc.b	$3C					; <
		dc.b	$A7					; 
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$70					; p
		dc.b	$69					; i
		dc.b	$33					; 3
		dc.b	$1E
		dc.b	$57					; W
		dc.b	$3A					; :
		dc.b	$92					; 
		dc.b	$1E
		dc.b	$59					; Y
		dc.b	$5E					; ^
		dc.b	$69					; i
		dc.b	$4A					; J
		dc.b	$85					; 
		dc.b	$52					; R
		dc.b	$99					; 
		dc.b	$70					; p
		dc.b	$92					; 
		dc.b	$BD					; 
		dc.b	$BF					; 
		dc.b	$57					; W
		dc.b	$89					; 
		dc.b	$BF					; 
		dc.b	$CD					; 
		dc.b	$B6					; 
		dc.b	$CF					; 
		dc.b	$75					; u
		dc.b	$BF					; 
		dc.b	$D2					; 
		dc.b	$48					; H
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$75					; u
		dc.b	$AC					; 
		dc.b	$62					; b
		dc.b	$7E					; ~
		dc.b	$92					; 
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$A9					; 
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$C4					; 
		dc.b	$90					; 
		dc.b	$5B					; [
		dc.b	$9C					; 
		dc.b	$4E					; N
		dc.b	$52					; R
		dc.b	$9E					; 
		dc.b	$50					; P
		dc.b	$90					; 
		dc.b	$5E					; ^
		dc.b	$92					; 
		dc.b	$5B					; [
		dc.b	$59					; Y
		dc.b	$A5					; 
		dc.b	$60					; `
		dc.b	$D2					; 
		dc.b	$7E					; ~
		dc.b	$B2					; 
		dc.b	$A2					; 
		dc.b	$6E					; n
		dc.b	$C2					; 
		dc.b	$5B					; [
		dc.b	$9E					; 
		dc.b	$59					; Y
		dc.b	$C2					; 
		dc.b	$9C					; 
		dc.b	$7E					; ~
		dc.b	$92					; 
		dc.b	$50					; P
		dc.b	$95					; 
		dc.b	$A0					; 
		dc.b	$64					; d
		dc.b	$BF					; 
		dc.b	$9E					; 
		dc.b	$67					; g
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$A5					; 
		dc.b	$60					; `
		dc.b	$7B					; {
		dc.b	$BB					; 
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$D6					; 
		dc.b	$4E					; N
		dc.b	$57					; W
		dc.b	$A2					; 
		dc.b	$72					; r
		dc.b	$97					; 
		dc.b	$75					; u
		dc.b	7
		dc.b	$52					; R
		dc.b	$7E					; ~
		dc.b	$4A					; J
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$1C
		dc.b	9
		dc.b	$3C					; <
		dc.b	$75					; u
		dc.b	$4E					; N
		dc.b	$A9					; 
		dc.b	$8B					; 
		dc.b	$69					; i
		dc.b	$5E					; ^
		dc.b	$C2					; 
		dc.b	$50					; P
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$33					; 3
		dc.b	$AC					; 
		dc.b	$75					; u
		dc.b	$54					; T
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$3A					; :
		dc.b	$80					; 
		dc.b	$70					; p
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$77					; w
		dc.b	$5E					; ^
		dc.b	$80					; 
		dc.b	$A7					; 
		dc.b	$A7					; 
		dc.b	$AC					; 
		dc.b	$6B					; k
		dc.b	$F7					; 
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$6E					; n
		dc.b	$BF					; 
		dc.b	$6E					; n
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$5E					; ^
		dc.b	$89					; 
		dc.b	$79					; y
		dc.b	$60					; `
		dc.b	$6B					; k
		dc.b	$69					; i
		dc.b	$C4					; 
		dc.b	$6B					; k
		dc.b	$A5					; 
		dc.b	$67					; g
		dc.b	$C4					; 
		dc.b	$A5					; 
		dc.b	$A9					; 
		dc.b	$B2					; 
		dc.b	$54					; T
		dc.b	$52					; R
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$4E					; N
		dc.b	$CF					; 
		dc.b	$52					; R
		dc.b	$75					; u
		dc.b	$89					; 
		dc.b	$45					; E
		dc.b	$48					; H
		dc.b	$41					; A
		dc.b	$8B					; 
		dc.b	$69					; i
		dc.b	$60					; `
		dc.b	$89					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$5E					; ^
		dc.b	$95					; 
		dc.b	$BF					; 
		dc.b	$5B					; [
		dc.b	$87					; 
		dc.b	$97					; 
		dc.b	$BB					; 
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$90					; 
		dc.b	$5B					; [
		dc.b	$7E					; ~
		dc.b	$9C					; 
		dc.b	$77					; w
		dc.b	$95					; 
		dc.b	$77					; w
		dc.b	$37					; 7
		dc.b	$92					; 
		dc.b	$7B					; {
		dc.b	$90					; 
		dc.b	$69					; i
		dc.b	$9E					; 
		dc.b	$59					; Y
		dc.b	$67					; g
		dc.b	$62					; b
		dc.b	$90					; 
		dc.b	$C9					; 
		dc.b	$89					; 
		dc.b	$85					; 
		dc.b	$AE					; 
		dc.b	$67					; g
		dc.b	$9E					; 
		dc.b	$89					; 
		dc.b	$99					; 
		dc.b	$BB					; 
		dc.b	$D2					; 
		dc.b	$AE					; 
		dc.b	$75					; u
		dc.b	$CD					; 
		dc.b	$BB					; 
		dc.b	$95					; 
		dc.b	$8E					; 
		dc.b	$9E					; 
		dc.b	$85					; 
		dc.b	$BD					; 
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$CD					; 
		dc.b	$6B					; k
		dc.b	$60					; `
		dc.b	$85					; 
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$2E					; .
		dc.b	$5E					; ^
		dc.b	$27					; '
		dc.b	$4A					; J
		dc.b	$1A
		dc.b	$45					; E
		dc.b	$7B					; {
		dc.b	$23					; #
		dc.b	$43					; C
		dc.b	$8E					; 
		dc.b	$48					; H
		dc.b	$2C					; ,
		dc.b	$62					; b
		dc.b	$43					; C
		dc.b	$43					; C
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$4C					; L
		dc.b	$69					; i
		dc.b	$52					; R
		dc.b	$3A					; :
		dc.b	$6B					; k
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$89					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$59					; Y
		dc.b	$3A					; :
		dc.b	$92					; 
		dc.b	$69					; i
		dc.b	$69					; i
		dc.b	$C6					; 
		dc.b	$5B					; [
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$B4					; 
		dc.b	$B2					; 
		dc.b	$B0					; 
		dc.b	$4E					; N
		dc.b	$B0					; 
		dc.b	$CD					; 
		dc.b	$90					; 
		dc.b	$B8					; 
		dc.b	$C9					; 
		dc.b	$B6					; 
		dc.b	$97					; 
		dc.b	$B2					; 
		dc.b	$B6					; 
		dc.b	$A7					; 
		dc.b	$AC					; 
		dc.b	$95					; 
		dc.b	$C9					; 
		dc.b	$A9					; 
		dc.b	$C9					; 
		dc.b	$CD					; 
		dc.b	$75					; u
		dc.b	$95					; 
		dc.b	$87					; 
		dc.b	$99					; 
		dc.b	$A0					; 
		dc.b	$A9					; 
		dc.b	$77					; w
		dc.b	$A7					; 
		dc.b	$9C					; 
		dc.b	$70					; p
		dc.b	$8E					; 
		dc.b	$7E					; ~
		dc.b	$97					; 
		dc.b	$4A					; J
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$C
		dc.b	$7E					; ~
		dc.b	$35					; 5
		dc.b	$35					; 5
		dc.b	$77					; w
		dc.b	$4E					; N
		dc.b	$8B					; 
		dc.b	$48					; H
		dc.b	$4E					; N
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$72					; r
		dc.b	$95					; 
		dc.b	$7B					; {
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$7E					; ~
		dc.b	$89					; 
		dc.b	$92					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$77					; w
		dc.b	$B0					; 
		dc.b	$9C					; 
		dc.b	$6B					; k
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$A0					; 
		dc.b	$9E					; 
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$69					; i
		dc.b	$77					; w
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$64					; d
		dc.b	$59					; Y
		dc.b	$79					; y
		dc.b	$5E					; ^
		dc.b	$6B					; k
		dc.b	$5B					; [
		dc.b	$62					; b
		dc.b	$54					; T
		dc.b	$80					; 
		dc.b	$48					; H
		dc.b	$3A					; :
		dc.b	$60					; `
		dc.b	$5E					; ^
		dc.b	$57					; W
		dc.b	$5E					; ^
		dc.b	$89					; 
		dc.b	$4C					; L
		dc.b	$41					; A
		dc.b	$4E					; N
		dc.b	$64					; d
		dc.b	$45					; E
		dc.b	$86					; 
		dc.b	$43					; C
		dc.b	$86					; 
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$61					; a
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$58					; X
		dc.b	$94					; 
		dc.b	$B6					; 
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$86					; 
		dc.b	$9C					; 
		dc.b	$91					; 
		dc.b	$91					; 
		dc.b	$AF					; 
		dc.b	$A7					; 
		dc.b	$77					; w
		dc.b	$A5					; 
		dc.b	$BE					; 
		dc.b	$8D					; 
		dc.b	$C4					; 
		dc.b	$B2					; 
		dc.b	$8D					; 
		dc.b	$AB					; 
		dc.b	$9F					; 
		dc.b	$B0					; 
		dc.b	$9F					; 
		dc.b	$B7					; 
		dc.b	$AB					; 
		dc.b	$A9					; 
		dc.b	$B6					; 
		dc.b	$BC					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$A3					; 
		dc.b	$A3					; 
		dc.b	$CB					; 
		dc.b	$6F					; o
		dc.b	$95					; 
		dc.b	$89					; 
		dc.b	$63					; c
		dc.b	$93					; 
		dc.b	$9E					; 
		dc.b	$73					; s
		dc.b	$9D					; 
		dc.b	$84					; 
		dc.b	$64					; d
		dc.b	$7E					; ~
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$7B					; {
		dc.b	$58					; X
		dc.b	$80					; 
		dc.b	$67					; g
		dc.b	$55					; U
		dc.b	$76					; v
		dc.b	$5F					; _
		dc.b	$6F					; o
		dc.b	$6B					; k
		dc.b	$4C					; L
		dc.b	$89					; 
		dc.b	$65					; e
		dc.b	$5F					; _
		dc.b	$8E					; 
		dc.b	$46					; F
		dc.b	$87					; 
		dc.b	$57					; W
		dc.b	$3A					; :
		dc.b	$72					; r
		dc.b	$58					; X
		dc.b	$4F					; O
		dc.b	$3D					; =
		dc.b	$6A					; j
		dc.b	$66					; f
		dc.b	$28					; (
		dc.b	$4F					; O
		dc.b	$6D					; m
		dc.b	$5E					; ^
		dc.b	$67					; g
		dc.b	$6A					; j
		dc.b	$74					; t
		dc.b	$60					; `
		dc.b	$68					; h
		dc.b	$57					; W
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$54					; T
		dc.b	$9C					; 
		dc.b	$6A					; j
		dc.b	$91					; 
		dc.b	$6B					; k
		dc.b	$6E					; n
		dc.b	$AB					; 
		dc.b	$51					; Q
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$5E					; ^
		dc.b	$7A					; z
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$59					; Y
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$9B					; 
		dc.b	$73					; s
		dc.b	$96					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$97					; 
		dc.b	$81					; 
		dc.b	$7A					; z
		dc.b	$AF					; 
		dc.b	$99					; 
		dc.b	$B2					; 
		dc.b	$77					; w
		dc.b	$B6					; 
		dc.b	$AE					; 
		dc.b	$9C					; 
		dc.b	$93					; 
		dc.b	$9E					; 
		dc.b	$C7					; 
		dc.b	$9A					; 
		dc.b	$C8					; 
		dc.b	$9A					; 
		dc.b	$9C					; 
		dc.b	$BC					; 
		dc.b	$A4					; 
		dc.b	$D7					; 
		dc.b	$B7					; 
		dc.b	$A7					; 
		dc.b	$A3					; 
		dc.b	$AD					; 
		dc.b	$AA					; 
		dc.b	$C4					; 
		dc.b	$AA					; 
		dc.b	$AF					; 
		dc.b	$8F					; 
		dc.b	$91					; 
		dc.b	$7E					; ~
		dc.b	$9E					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$6D					; m
		dc.b	$7A					; z
		dc.b	$5F					; _
		dc.b	$40					; @
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$45					; E
		dc.b	$8F					; 
		dc.b	$69					; i
		dc.b	$68					; h
		dc.b	$39					; 9
		dc.b	$48					; H
		dc.b	$4A					; J
		dc.b	$57					; W
		dc.b	$24					; $
		dc.b	$50					; P
		dc.b	$48					; H
		dc.b	$26					; &
		dc.b	$3B					; ;
		dc.b	$4B					; K
		dc.b	$4D					; M
		dc.b	$38					; 8
		dc.b	$48					; H
		dc.b	$65					; e
		dc.b	$50					; P
		dc.b	$3B					; ;
		dc.b	$5E					; ^
		dc.b	$7B					; {
		dc.b	$55					; U
		dc.b	$6F					; o
		dc.b	$55					; U
		dc.b	$6A					; j
		dc.b	$65					; e
		dc.b	$5D					; ]
		dc.b	$68					; h
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$66					; f
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$89					; 
		dc.b	$92					; 
		dc.b	$92					; 
		dc.b	$9D					; 
		dc.b	$81					; 
		dc.b	$A1					; 
		dc.b	$A8					; 
		dc.b	$92					; 
		dc.b	$A5					; 
		dc.b	$9D					; 
		dc.b	$BD					; 
		dc.b	$AC					; 
		dc.b	$AC					; 
		dc.b	$AF					; 
		dc.b	$A3					; 
		dc.b	$8D					; 
		dc.b	$B3					; 
		dc.b	$93					; 
		dc.b	$B2					; 
		dc.b	$7B					; {
		dc.b	$AD					; 
		dc.b	$97					; 
		dc.b	$85					; 
		dc.b	$91					; 
		dc.b	$9C					; 
		dc.b	$A1					; 
		dc.b	$7E					; ~
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$9C					; 
		dc.b	$A7					; 
		dc.b	$70					; p
		dc.b	$8D					; 
		dc.b	$66					; f
		dc.b	$8D					; 
		dc.b	$67					; g
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$5A					; Z
		dc.b	$6E					; n
		dc.b	$88					; 
		dc.b	$55					; U
		dc.b	$89					; 
		dc.b	$62					; b
		dc.b	$7C					; |
		dc.b	$62					; b
		dc.b	$62					; b
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$64					; d
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$7B					; {
		dc.b	$72					; r
		dc.b	$7A					; z
		dc.b	$92					; 
		dc.b	$71					; q
		dc.b	$6D					; m
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$69					; i
		dc.b	$64					; d
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$65					; e
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$62					; b
		dc.b	$89					; 
		dc.b	$6C					; l
		dc.b	$83					; 
		dc.b	$90					; 
		dc.b	$6F					; o
		dc.b	$98					; 
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$89					; 
		dc.b	$7D					; }
		dc.b	$98					; 
		dc.b	$6E					; n
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$8D					; 
		dc.b	$70					; p
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$74					; t
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$78					; x
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$70					; p
		dc.b	$7B					; {
		dc.b	$85					; 
		dc.b	$7A					; z
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$8F					; 
		dc.b	$80					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$9D					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$93					; 
		dc.b	$82					; 
		dc.b	$9C					; 
		dc.b	$93					; 
		dc.b	$89					; 
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$82					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$6C					; l
		dc.b	$7D					; }
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$84					; 
		dc.b	$5F					; _
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$5F					; _
		dc.b	$74					; t
		dc.b	$7A					; z
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$74					; t
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$69					; i
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$66					; f
		dc.b	$7C					; |
		dc.b	$6D					; m
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$87					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$77					; w
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$7A					; z
		dc.b	$8C					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$90					; 
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$77					; w
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$75					; u
		dc.b	$8A					; 
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$8C					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$7B					; {
		dc.b	$7F					; 
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
unk_2B430:	dc.b	$80					; 		; DATA XREF: ROM:0001B034o
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$7B					; {
		dc.b	$89					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$70					; p
		dc.b	$6A					; j
		dc.b	$6C					; l
		dc.b	$6C					; l
		dc.b	$6E					; n
		dc.b	$7B					; {
		dc.b	$92					; 
		dc.b	$96					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$69					; i
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$7E					; ~
		dc.b	$63					; c
		dc.b	$65					; e
		dc.b	$63					; c
		dc.b	$6A					; j
		dc.b	$90					; 
		dc.b	$A2					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$89					; 
		dc.b	$73					; s
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$7B					; {
		dc.b	$92					; 
		dc.b	$94					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$89					; 
		dc.b	$7E					; ~
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$92					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$75					; u
		dc.b	$77					; w
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$90					; 
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$87					; 
		dc.b	$8D					; 
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$85					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$6C					; l
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	5
		dc.b	$DD					; 
		dc.b	$FE					; 
		dc.b	$4F					; O
		dc.b	1
		dc.b	$E
		dc.b	1
		dc.b	$6C					; l
		dc.b	1
		dc.b	1
		dc.b	$44					; D
		dc.b	$69					; i
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$23					; #
		dc.b	$42					; B
		dc.b	$A6					; 
		dc.b	$C5					; 
		dc.b	$A8					; 
		dc.b	$E6					; 
		dc.b	$FE					; 
		dc.b	$6C					; l
		dc.b	$5C					; \
		dc.b	$94					; 
		dc.b	$75					; u
		dc.b	$E4					; 
		dc.b	$2B					; +
		dc.b	1
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	1
		dc.b	1
		dc.b	$82					; 
		dc.b	$4D					; M
		dc.b	3
		dc.b	$7B					; {
		dc.b	$FE					; 
		dc.b	$99					; 
		dc.b	$35					; 5
		dc.b	$9B					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$99					; 
		dc.b	$23					; #
		dc.b	$3D					; =
		dc.b	$72					; r
		dc.b	$8E					; 
		dc.b	$94					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$85					; 
		dc.b	$A8					; 
		dc.b	$37					; 7
		dc.b	$42					; B
		dc.b	$B1					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$B9					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$55					; U
		dc.b	$44					; D
		dc.b	$44					; D
		dc.b	$65					; e
		dc.b	$6E					; n
		dc.b	$61					; a
		dc.b	$4D					; M
		dc.b	$58					; X
		dc.b	$65					; e
		dc.b	$4F					; O
		dc.b	$42					; B
		dc.b	$55					; U
		dc.b	$53					; S
		dc.b	$73					; s
		dc.b	$C9					; 
		dc.b	$A8					; 
		dc.b	$97					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$73					; s
		dc.b	$99					; 
		dc.b	$85					; 
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$8D					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$79					; y
		dc.b	$4F					; O
		dc.b	$44					; D
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$65					; e
		dc.b	$44					; D
		dc.b	$55					; U
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$9F					; 
		dc.b	$94					; 
		dc.b	$69					; i
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$82					; 
		dc.b	$72					; r
		dc.b	$73					; s
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$99					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$85					; 
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$8B					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$FE					; 
		dc.b	$E6					; 
		dc.b	$A
		dc.b	1
		dc.b	$23					; #
		dc.b	$4F					; O
		dc.b	$30					; 0
		dc.b	$55					; U
		dc.b	$E2					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$11
		dc.b	$85					; 
		dc.b	$D0					; 
		dc.b	$BE					; 
		dc.b	$4F					; O
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$7E					; ~
		dc.b	1
		dc.b	1
		dc.b	$97					; 
		dc.b	$6C					; l
		dc.b	$35					; 5
		dc.b	$8E					; 
		dc.b	$A8					; 
		dc.b	$35					; 5
		dc.b	1
		dc.b	1
		dc.b	$40					; @
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$1C
		dc.b	$37					; 7
		dc.b	$F6					; 
		dc.b	$B9					; 
		dc.b	$9B					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$A8					; 
		dc.b	$A4					; 
		dc.b	$69					; i
		dc.b	$1A
		dc.b	$7E					; ~
		dc.b	$DD					; 
		dc.b	$A6					; 
		dc.b	$6E					; n
		dc.b	$42					; B
		dc.b	$2E					; .
		dc.b	$55					; U
		dc.b	$61					; a
		dc.b	$99					; 
		dc.b	$75					; u
		dc.b	$A0					; 
		dc.b	$8B					; 
		dc.b	1
		dc.b	$27					; '
		dc.b	$A8					; 
		dc.b	$B3					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$72					; r
		dc.b	$49					; I
		dc.b	$35					; 5
		dc.b	$79					; y
		dc.b	$D2					; 
		dc.b	$BC					; 
		dc.b	$9B					; 
		dc.b	$85					; 
		dc.b	$53					; S
		dc.b	$58					; X
		dc.b	$65					; e
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$A8					; 
		dc.b	$9B					; 
		dc.b	$7E					; ~
		dc.b	$4F					; O
		dc.b	$40					; @
		dc.b	$42					; B
		dc.b	$49					; I
		dc.b	$61					; a
		dc.b	$79					; y
		dc.b	$5C					; \
		dc.b	$7B					; {
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$94					; 
		dc.b	$B1					; 
		dc.b	$82					; 
		dc.b	$47					; G
		dc.b	$85					; 
		dc.b	$B1					; 
		dc.b	$B3					; 
		dc.b	$8D					; 
		dc.b	$4D					; M
		dc.b	$53					; S
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$8B					; 
		dc.b	$69					; i
		dc.b	$49					; I
		dc.b	$65					; e
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$75					; u
		dc.b	$79					; y
		dc.b	$4F					; O
		dc.b	$69					; i
		dc.b	$AD					; 
		dc.b	$A6					; 
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$8B					; 
		dc.b	$7B					; {
		dc.b	$6E					; n
		dc.b	$6E					; n
		dc.b	$75					; u
		dc.b	$60					; `
		dc.b	$61					; a
		dc.b	$8B					; 
		dc.b	$BC					; 
		dc.b	$C9					; 
		dc.b	$BE					; 
		dc.b	$99					; 
		dc.b	$B1					; 
		dc.b	$A8					; 
		dc.b	$53					; S
		dc.b	$37					; 7
		dc.b	$42					; B
		dc.b	$5C					; \
		dc.b	$6C					; l
		dc.b	$5A					; Z
		dc.b	$35					; 5
		dc.b	$4F					; O
		dc.b	$82					; 
		dc.b	$79					; y
		dc.b	$53					; S
		dc.b	$8D					; 
		dc.b	$FE					; 
		dc.b	$92					; 
		dc.b	$A8					; 
		dc.b	$99					; 
		dc.b	$E9					; 
		dc.b	$D9					; 
		dc.b	$55					; U
		dc.b	1
		dc.b	$17
		dc.b	$44					; D
		dc.b	$47					; G
		dc.b	$3D					; =
		dc.b	$92					; 
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$85					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$C0					; 
		dc.b	$BC					; 
		dc.b	$B1					; 
		dc.b	$9F					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$49					; I
		dc.b	$67					; g
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$69					; i
		dc.b	$5A					; Z
		dc.b	$97					; 
		dc.b	$4D					; M
		dc.b	$80					; 
		dc.b	$E6					; 
		dc.b	1
		dc.b	1
		dc.b	$AB					; 
		dc.b	1
		dc.b	1
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	1
		dc.b	$CB					; 
		dc.b	$FE					; 
		dc.b	$E9					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$EB					; 
		dc.b	$B7					; 
		dc.b	$1E
		dc.b	1
		dc.b	5
		dc.b	$1A
		dc.b	1
		dc.b	$8D					; 
		dc.b	$9B					; 
		dc.b	$67					; g
		dc.b	$53					; S
		dc.b	1
		dc.b	5
		dc.b	$61					; a
		dc.b	$E6					; 
		dc.b	$FB					; 
		dc.b	$F8					; 
		dc.b	$E6					; 
		dc.b	$AB					; 
		dc.b	$73					; s
		dc.b	$3B					; ;
		dc.b	3
		dc.b	$37					; 7
		dc.b	$30					; 0
		dc.b	$42					; B
		dc.b	$9F					; 
		dc.b	$9B					; 
		dc.b	$5C					; \
		dc.b	$44					; D
		dc.b	$5C					; \
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$99					; 
		dc.b	$B9					; 
		dc.b	$82					; 
		dc.b	$42					; B
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$40					; @
		dc.b	$2B					; +
		dc.b	$6E					; n
		dc.b	$A4					; 
		dc.b	$7B					; {
		dc.b	$73					; s
		dc.b	$53					; S
		dc.b	$65					; e
		dc.b	$75					; u
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$BC					; 
		dc.b	$B9					; 
		dc.b	$94					; 
		dc.b	$7E					; ~
		dc.b	$AB					; 
		dc.b	$B1					; 
		dc.b	$60					; `
		dc.b	$5C					; \
		dc.b	$23					; #
		dc.b	$1A
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$5C					; \
		dc.b	$58					; X
		dc.b	$8D					; 
		dc.b	$99					; 
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$9B					; 
		dc.b	$B9					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$6E					; n
		dc.b	$82					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$79					; y
		dc.b	$58					; X
		dc.b	$34					; 4
		dc.b	$4F					; O
		dc.b	$80					; 
		dc.b	$69					; i
		dc.b	$42					; B
		dc.b	$1E
		dc.b	$6C					; l
		dc.b	$A0					; 
		dc.b	$CC					; 
		dc.b	$A0					; 
		dc.b	$44					; D
		dc.b	$53					; S
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$97					; 
		dc.b	$B9					; 
		dc.b	$8E					; 
		dc.b	$8E					; 
		dc.b	$8B					; 
		dc.b	$4F					; O
		dc.b	$9B					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$49					; I
		dc.b	$5A					; Z
		dc.b	$89					; 
		dc.b	$A4					; 
		dc.b	$8D					; 
		dc.b	$7B					; {
		dc.b	$8D					; 
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$8D					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$5E					; ^
		dc.b	$5E					; ^
		dc.b	$75					; u
		dc.b	$87					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$6A					; j
		dc.b	$42					; B
		dc.b	$4F					; O
		dc.b	$70					; p
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$B1					; 
		dc.b	$AD					; 
		dc.b	$65					; e
		dc.b	$73					; s
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$7C					; |
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$72					; r
		dc.b	$89					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$29					; )
		dc.b	$55					; U
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$E0					; 
		dc.b	$B5					; 
		dc.b	$A4					; 
		dc.b	$A0					; 
		dc.b	$A4					; 
		dc.b	$B1					; 
		dc.b	$8E					; 
		dc.b	$99					; 
		dc.b	$9F					; 
		dc.b	$96					; 
		dc.b	$9D					; 
		dc.b	$A0					; 
		dc.b	$77					; w
		dc.b	$89					; 
		dc.b	$6A					; j
		dc.b	$39					; 9
		dc.b	$30					; 0
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$55					; U
		dc.b	$80					; 
		dc.b	$96					; 
		dc.b	$AD					; 
		dc.b	$BA					; 
		dc.b	$B1					; 
		dc.b	$A9					; 
		dc.b	$8E					; 
		dc.b	$72					; r
		dc.b	$49					; I
		dc.b	$22					; "
		dc.b	$27					; '
		dc.b	$55					; U
		dc.b	$99					; 
		dc.b	$BE					; 
		dc.b	$C2					; 
		dc.b	$C2					; 
		dc.b	$89					; 
		dc.b	$65					; e
		dc.b	$85					; 
		dc.b	$67					; g
		dc.b	$6E					; n
		dc.b	$84					; 
		dc.b	$B1					; 
		dc.b	$DD					; 
		dc.b	$77					; w
		dc.b	$29					; )
		dc.b	$4F					; O
		dc.b	$5E					; ^
		dc.b	$5C					; \
		dc.b	$5E					; ^
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$89					; 
		dc.b	$A0					; 
		dc.b	$8D					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$A0					; 
		dc.b	$96					; 
		dc.b	$7C					; |
		dc.b	$80					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$94					; 
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$49					; I
		dc.b	$4F					; O
		dc.b	$72					; r
		dc.b	$63					; c
		dc.b	$53					; S
		dc.b	$4B					; K
		dc.b	$5E					; ^
		dc.b	$87					; 
		dc.b	$AB					; 
		dc.b	$AB					; 
		dc.b	$99					; 
		dc.b	$7C					; |
		dc.b	$58					; X
		dc.b	$6C					; l
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$75					; u
		dc.b	$73					; s
		dc.b	$5C					; \
		dc.b	$51					; Q
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$6A					; j
		dc.b	$67					; g
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$A2					; 
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$57					; W
		dc.b	$57					; W
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$72					; r
		dc.b	$80					; 
		dc.b	$89					; 
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$89					; 
		dc.b	$A2					; 
		dc.b	$85					; 
		dc.b	$55					; U
		dc.b	$7B					; {
		dc.b	$8E					; 
		dc.b	$97					; 
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$7E					; ~
		dc.b	$57					; W
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$85					; 
		dc.b	$6C					; l
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$51					; Q
		dc.b	$57					; W
		dc.b	$55					; U
		dc.b	$57					; W
		dc.b	$60					; `
		dc.b	$69					; i
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$A8					; 
		dc.b	$A2					; 
		dc.b	$92					; 
		dc.b	$61					; a
		dc.b	$51					; Q
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$58					; X
		dc.b	$58					; X
		dc.b	$87					; 
		dc.b	$92					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$A2					; 
		dc.b	$AB					; 
		dc.b	$B3					; 
		dc.b	$A2					; 
		dc.b	$75					; u
		dc.b	$58					; X
		dc.b	$69					; i
		dc.b	$82					; 
		dc.b	$8E					; 
		dc.b	$67					; g
		dc.b	$4D					; M
		dc.b	$51					; Q
		dc.b	$58					; X
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$6E					; n
		dc.b	$85					; 
		dc.b	$96					; 
		dc.b	$B5					; 
		dc.b	$B1					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$96					; 
		dc.b	$73					; s
		dc.b	$73					; s
		dc.b	$89					; 
		dc.b	$5A					; Z
		dc.b	$8A					; 
		dc.b	$A7					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$81					; 
		dc.b	$9F					; 
		dc.b	$A3					; 
		dc.b	$9C					; 
		dc.b	$94					; 
		dc.b	$8B					; 
		dc.b	$9D					; 
		dc.b	$A6					; 
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$5D					; ]
		dc.b	$5D					; ]
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$73					; s
		dc.b	$52					; R
		dc.b	$3E					; >
		dc.b	$30					; 0
		dc.b	$4F					; O
		dc.b	$6C					; l
		dc.b	$63					; c
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$6E					; n
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$8E					; 
		dc.b	$84					; 
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$9F					; 
		dc.b	$A4					; 
		dc.b	$A3					; 
		dc.b	$84					; 
		dc.b	$6A					; j
		dc.b	$46					; F
		dc.b	$53					; S
		dc.b	$88					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$96					; 
		dc.b	$AA					; 
		dc.b	$AE					; 
		dc.b	$AE					; 
		dc.b	$A4					; 
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$9B					; 
		dc.b	$A3					; 
		dc.b	$A2					; 
		dc.b	$9D					; 
		dc.b	$AA					; 
		dc.b	$B1					; 
		dc.b	$A2					; 
		dc.b	$98					; 
		dc.b	$9D					; 
		dc.b	$8B					; 
		dc.b	$73					; s
		dc.b	$67					; g
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$6B					; k
		dc.b	$4D					; M
		dc.b	$1B
		dc.b	1
		dc.b	1
		dc.b	$1E
		dc.b	$63					; c
		dc.b	$99					; 
		dc.b	$88					; 
		dc.b	$5C					; \
		dc.b	$48					; H
		dc.b	$50					; P
		dc.b	$45					; E
		dc.b	$63					; c
		dc.b	$8B					; 
		dc.b	$A9					; 
		dc.b	$BA					; 
		dc.b	$C8					; 
		dc.b	$BA					; 
		dc.b	$BE					; 
		dc.b	$AB					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$4D					; M
		dc.b	$52					; R
		dc.b	$5C					; \
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$A4					; 
		dc.b	$C8					; 
		dc.b	$FE					; 
		dc.b	$F0					; 
		dc.b	$E4					; 
		dc.b	$B0					; 
		dc.b	$8B					; 
		dc.b	$75					; u
		dc.b	$38					; 8
		dc.b	$46					; F
		dc.b	$25					; %
		dc.b	1
		dc.b	$2E					; .
		dc.b	$75					; u
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$5C					; \
		dc.b	$A6					; 
		dc.b	$8E					; 
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$D
		dc.b	$71					; q
		dc.b	$AB					; 
		dc.b	$A6					; 
		dc.b	$67					; g
		dc.b	$B1					; 
		dc.b	$FE					; 
		dc.b	$D7					; 
		dc.b	$7C					; |
		dc.b	$94					; 
		dc.b	$CC					; 
		dc.b	$98					; 
		dc.b	$A6					; 
		dc.b	$9D					; 
		dc.b	$72					; r
		dc.b	$41					; A
		dc.b	1
		dc.b	$68					; h
		dc.b	$EC					; 
		dc.b	$C4					; 
		dc.b	$5D					; ]
		dc.b	1
		dc.b	1
		dc.b	$30					; 0
		dc.b	$E8					; 
		dc.b	$C4					; 
		dc.b	$15
		dc.b	1
		dc.b	$3C					; <
		dc.b	$63					; c
		dc.b	$D6					; 
		dc.b	$D6					; 
		dc.b	$64					; d
		dc.b	$D0					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$C4					; 
		dc.b	$15
		dc.b	1
		dc.b	$72					; r
		dc.b	$DA					; 
		dc.b	$CD					; 
		dc.b	$B4					; 
		dc.b	$17
		dc.b	1
		dc.b	$29					; )
		dc.b	1
		dc.b	$86					; 
		dc.b	$FE					; 
		dc.b	$95					; 
		dc.b	$37					; 7
		dc.b	$C9					; 
		dc.b	$73					; s
		dc.b	1
		dc.b	$25					; %
		dc.b	$2D					; -
		dc.b	$E4					; 
		dc.b	$D6					; 
		dc.b	$6B					; k
		dc.b	$67					; g
		dc.b	$19
		dc.b	$33					; 3
		dc.b	$A6					; 
		dc.b	$FE					; 
		dc.b	$FE					; 
		dc.b	$B6					; 
		dc.b	$F3					; 
		dc.b	$6F					; o
		dc.b	1
		dc.b	$23					; #
		dc.b	$E5					; 
		dc.b	$C1					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$30					; 0
		dc.b	$14
		dc.b	6
		dc.b	$B
		dc.b	$3F					; ?
		dc.b	$D6					; 
		dc.b	$FE					; 
		dc.b	$EC					; 
		dc.b	$B3					; 
		dc.b	$2E					; .
		dc.b	$23					; #
		dc.b	$DB					; 
		dc.b	$D7					; 
		dc.b	$1E
		dc.b	$2C					; ,
		dc.b	$E2					; 
		dc.b	$BA					; 
		dc.b	$12
		dc.b	$8D					; 
		dc.b	$F3					; 
		dc.b	$4D					; M
		dc.b	1
		dc.b	$B1					; 
		dc.b	$B7					; 
		dc.b	6
		dc.b	$56					; V
		dc.b	$D2					; 
		dc.b	$29					; )
		dc.b	$22					; "
		dc.b	$DE					; 
		dc.b	$98					; 
		dc.b	$17
		dc.b	$B6					; 
		dc.b	$D2					; 
		dc.b	$45					; E
		dc.b	$7C					; |
		dc.b	$D7					; 
		dc.b	$68					; h
		dc.b	1
		dc.b	$49					; I
		dc.b	$CB					; 
		dc.b	$E5					; 
		dc.b	$A2					; 
		dc.b	$34					; 4
		dc.b	1
		dc.b	$6E					; n
		dc.b	$FE					; 
		dc.b	$CF					; 
		dc.b	$35					; 5
		dc.b	$7D					; }
		dc.b	$FE					; 
		dc.b	$E0					; 
		dc.b	$84					; 
		dc.b	1
		dc.b	1
		dc.b	$79					; y
		dc.b	$8F					; 
		dc.b	$38					; 8
		dc.b	$C9					; 
		dc.b	$FE					; 
		dc.b	$81					; 
		dc.b	$3F					; ?
		dc.b	$9B					; 
		dc.b	$CB					; 
		dc.b	$6C					; l
		dc.b	$BB					; 
		dc.b	$42					; B
		dc.b	$83					; 
		dc.b	$B6					; 
		dc.b	$C8					; 
		dc.b	$27					; '
		dc.b	$67					; g
		dc.b	$65					; e
		dc.b	$A7					; 
		dc.b	$B1					; 
		dc.b	$18
		dc.b	1
		dc.b	$64					; d
		dc.b	$E2					; 
		dc.b	$CC					; 
		dc.b	$5A					; Z
		dc.b	$30					; 0
		dc.b	$A6					; 
		dc.b	$60					; `
		dc.b	1
		dc.b	1
		dc.b	$29					; )
		dc.b	$8A					; 
		dc.b	$A0					; 
		dc.b	$95					; 
		dc.b	$8F					; 
		dc.b	$D0					; 
		dc.b	$AD					; 
		dc.b	$DA					; 
		dc.b	$C9					; 
		dc.b	$E1					; 
		dc.b	$AB					; 
		dc.b	$A4					; 
		dc.b	$BA					; 
		dc.b	$F2					; 
		dc.b	$B4					; 
		dc.b	$94					; 
		dc.b	$3E					; >
		dc.b	$2C					; ,
		dc.b	$5E					; ^
		dc.b	$A
		dc.b	1
		dc.b	$4F					; O
		dc.b	$B0					; 
		dc.b	$78					; x
		dc.b	$2E					; .
		dc.b	$D
		dc.b	$12
		dc.b	$42					; B
		dc.b	$8A					; 
		dc.b	$4C					; L
		dc.b	7
		dc.b	$67					; g
		dc.b	$D3					; 
		dc.b	$C9					; 
		dc.b	$E7					; 
		dc.b	$FE					; 
		dc.b	$D2					; 
		dc.b	$3F					; ?
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$6E					; n
		dc.b	$52					; R
		dc.b	$7F					; 
		dc.b	$FE					; 
		dc.b	$F5					; 
		dc.b	$4F					; O
		dc.b	$64					; d
		dc.b	$E7					; 
		dc.b	$D3					; 
		dc.b	$35					; 5
		dc.b	$46					; F
		dc.b	$CF					; 
		dc.b	$B0					; 
		dc.b	$37					; 7
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$7F					; 
		dc.b	$4C					; L
		dc.b	$6B					; k
		dc.b	$8E					; 
		dc.b	$50					; P
		dc.b	$43					; C
		dc.b	$86					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$19
		dc.b	$18
		dc.b	$9C					; 
		dc.b	$71					; q
		dc.b	7
		dc.b	$20
		dc.b	$A0					; 
		dc.b	$CB					; 
		dc.b	$48					; H
		dc.b	$6A					; j
		dc.b	$DA					; 
		dc.b	$F0					; 
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$FA					; 
		dc.b	$CB					; 
		dc.b	$64					; d
		dc.b	$61					; a
		dc.b	$80					; 
		dc.b	$B6					; 
		dc.b	$99					; 
		dc.b	$73					; s
		dc.b	$4F					; O
		dc.b	$6A					; j
		dc.b	$6E					; n
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$3F					; ?
		dc.b	$5E					; ^
		dc.b	$33					; 3
		dc.b	$52					; R
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$4F					; O
		dc.b	$4F					; O
		dc.b	$78					; x
		dc.b	$A3					; 
		dc.b	$E9					; 
		dc.b	$9F					; 
		dc.b	$52					; R
		dc.b	$98					; 
		dc.b	$AE					; 
		dc.b	$8F					; 
		dc.b	$68					; h
		dc.b	$3C					; <
		dc.b	$91					; 
		dc.b	$DA					; 
		dc.b	$B7					; 
		dc.b	$AB					; 
		dc.b	$7A					; z
		dc.b	$C5					; 
		dc.b	$94					; 
		dc.b	$4C					; L
		dc.b	$61					; a
		dc.b	$78					; x
		dc.b	$6C					; l
		dc.b	$3A					; :
		dc.b	$50					; P
		dc.b	$65					; e
		dc.b	$5C					; \
		dc.b	$38					; 8
		dc.b	$3C					; <
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$5E					; ^
		dc.b	$5A					; Z
		dc.b	$AB					; 
		dc.b	$CC					; 
		dc.b	$72					; r
		dc.b	$72					; r
		dc.b	$9F					; 
		dc.b	$7C					; |
		dc.b	$98					; 
		dc.b	$84					; 
		dc.b	$55					; U
		dc.b	$92					; 
		dc.b	$A0					; 
		dc.b	$55					; U
		dc.b	$5A					; Z
		dc.b	$4D					; M
		dc.b	$8A					; 
		dc.b	$C2					; 
		dc.b	$B7					; 
		dc.b	$81					; 
		dc.b	$5D					; ]
		dc.b	$AB					; 
		dc.b	$65					; e
		dc.b	$6A					; j
		dc.b	$6A					; j
		dc.b	$87					; 
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$B0					; 
		dc.b	$9B					; 
		dc.b	$68					; h
		dc.b	$92					; 
		dc.b	$A7					; 
		dc.b	$76					; v
		dc.b	$68					; h
		dc.b	$43					; C
		dc.b	$57					; W
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$60					; `
		dc.b	$3B					; ;
		dc.b	$92					; 
		dc.b	$B4					; 
		dc.b	$72					; r
		dc.b	$50					; P
		dc.b	$67					; g
		dc.b	$87					; 
		dc.b	$8B					; 
		dc.b	$9C					; 
		dc.b	$B4					; 
		dc.b	$B1					; 
		dc.b	$C4					; 
		dc.b	$AD					; 
		dc.b	$7C					; |
		dc.b	$75					; u
		dc.b	$68					; h
		dc.b	$5A					; Z
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$A9					; 
		dc.b	$8A					; 
		dc.b	$D6					; 
		dc.b	$D7					; 
		dc.b	$86					; 
		dc.b	$3E					; >
		dc.b	$42					; B
		dc.b	$4D					; M
		dc.b	$6A					; j
		dc.b	$79					; y
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$A1					; 
		dc.b	$99					; 
		dc.b	$87					; 
		dc.b	$72					; r
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$92					; 
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$80					; 
		dc.b	$8F					; 
		dc.b	$AB					; 
		dc.b	$95					; 
		dc.b	$7E					; ~
		dc.b	$7A					; z
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$5D					; ]
		dc.b	$88					; 
		dc.b	$93					; 
		dc.b	$7A					; z
		dc.b	$5F					; _
		dc.b	$7B					; {
		dc.b	$90					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$5F					; _
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$62					; b
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$61					; a
		dc.b	$60					; `
		dc.b	$4E					; N
		dc.b	$75					; u
		dc.b	$96					; 
		dc.b	$AC					; 
		dc.b	$A6					; 
		dc.b	$73					; s
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$68					; h
		dc.b	$68					; h
		dc.b	$89					; 
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$6C					; l
		dc.b	$91					; 
		dc.b	$BD					; 
		dc.b	$99					; 
		dc.b	$59					; Y
		dc.b	$65					; e
		dc.b	$7A					; z
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$53					; S
		dc.b	$56					; V
		dc.b	$96					; 
		dc.b	$BD					; 
		dc.b	$8D					; 
		dc.b	$4F					; O
		dc.b	$61					; a
		dc.b	$74					; t
		dc.b	$8E					; 
		dc.b	$A8					; 
		dc.b	$8D					; 
		dc.b	$9F					; 
		dc.b	$80					; 
		dc.b	$53					; S
		dc.b	$44					; D
		dc.b	$8A					; 
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$A6					; 
		dc.b	$4A					; J
		dc.b	$61					; a
		dc.b	$6F					; o
		dc.b	$A6					; 
		dc.b	$CA					; 
		dc.b	$A3					; 
		dc.b	$3D					; =
		dc.b	$6F					; o
		dc.b	$A9					; 
		dc.b	$65					; e
		dc.b	$76					; v
		dc.b	$6F					; o
		dc.b	$8E					; 
		dc.b	$BF					; 
		dc.b	$77					; w
		dc.b	$96					; 
		dc.b	$62					; b
		dc.b	$85					; 
		dc.b	$6A					; j
		dc.b	$63					; c
		dc.b	$F3					; 
		dc.b	$B4					; 
		dc.b	$1E
		dc.b	$66					; f
		dc.b	$6E					; n
		dc.b	$58					; X
		dc.b	$8D					; 
		dc.b	$78					; x
		dc.b	$A8					; 
		dc.b	$6C					; l
		dc.b	$85					; 
		dc.b	$AF					; 
		dc.b	$89					; 
		dc.b	$55					; U
		dc.b	$39					; 9
		dc.b	$A6					; 
		dc.b	$72					; r
		dc.b	$6C					; l
		dc.b	$A5					; 
		dc.b	$94					; 
		dc.b	$99					; 
		dc.b	$43					; C
		dc.b	$BE					; 
		dc.b	$AE					; 
		dc.b	$53					; S
		dc.b	$9E					; 
		dc.b	$A3					; 
		dc.b	$63					; c
		dc.b	$76					; v
		dc.b	$6B					; k
		dc.b	$C3					; 
		dc.b	$6C					; l
		dc.b	$A3					; 
		dc.b	$50					; P
		dc.b	$7B					; {
		dc.b	$96					; 
		dc.b	$50					; P
		dc.b	$8A					; 
		dc.b	$6B					; k
		dc.b	$EE					; 
		dc.b	$A2					; 
		dc.b	$35					; 5
		dc.b	$A5					; 
		dc.b	$9E					; 
		dc.b	$65					; e
		dc.b	$83					; 
		dc.b	$69					; i
		dc.b	$62					; b
		dc.b	$7B					; {
		dc.b	$29					; )
		dc.b	$8F					; 
		dc.b	$A9					; 
		dc.b	$67					; g
		dc.b	$63					; c
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$38					; 8
		dc.b	$5C					; \
		dc.b	$BA					; 
		dc.b	$C2					; 
		dc.b	$5D					; ]
		dc.b	$53					; S
		dc.b	$87					; 
		dc.b	$A2					; 
		dc.b	$56					; V
		dc.b	$A8					; 
		dc.b	$87					; 
		dc.b	$6B					; k
		dc.b	$34					; 4
		dc.b	$82					; 
		dc.b	$C6					; 
		dc.b	$46					; F
		dc.b	$2D					; -
		dc.b	$B9					; 
		dc.b	$CE					; 
		dc.b	$65					; e
		dc.b	$43					; C
		dc.b	$54					; T
		dc.b	$AD					; 
		dc.b	$95					; 
		dc.b	$5B					; [
		dc.b	$7A					; z
		dc.b	$9F					; 
		dc.b	$4C					; L
		dc.b	$66					; f
		dc.b	$B9					; 
		dc.b	$97					; 
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$82					; 
		dc.b	$5F					; _
		dc.b	$66					; f
		dc.b	$94					; 
		dc.b	$A9					; 
		dc.b	$40					; @
		dc.b	$7E					; ~
		dc.b	$8D					; 
		dc.b	$66					; f
		dc.b	$6C					; l
		dc.b	$8E					; 
		dc.b	$9B					; 
		dc.b	$7A					; z
		dc.b	$65					; e
		dc.b	$78					; x
		dc.b	$67					; g
		dc.b	$61					; a
		dc.b	$B3					; 
		dc.b	$99					; 
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$B2					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$4A					; J
		dc.b	$AD					; 
		dc.b	$E1					; 
		dc.b	$8E					; 
		dc.b	$5B					; [
		dc.b	$A0					; 
		dc.b	$B7					; 
		dc.b	$59					; Y
		dc.b	$3B					; ;
		dc.b	$7D					; }
		dc.b	$8D					; 
		dc.b	$6F					; o
		dc.b	$82					; 
		dc.b	$AE					; 
		dc.b	$65					; e
		dc.b	$62					; b
		dc.b	$E8					; 
		dc.b	$8C					; 
		dc.b	$17
		dc.b	$97					; 
		dc.b	$FA					; 
		dc.b	$8A					; 
		dc.b	$18
		dc.b	$63					; c
		dc.b	$C8					; 
		dc.b	$75					; u
		dc.b	$39					; 9
		dc.b	$84					; 
		dc.b	$C5					; 
		dc.b	$7E					; ~
		dc.b	$4C					; L
		dc.b	$90					; 
		dc.b	$A8					; 
		dc.b	$67					; g
		dc.b	$48					; H
		dc.b	$71					; q
		dc.b	$8B					; 
		dc.b	$6E					; n
		dc.b	$6D					; m
		dc.b	$C1					; 
		dc.b	$8A					; 
		dc.b	$22					; "
		dc.b	$79					; y
		dc.b	$AF					; 
		dc.b	$6D					; m
		dc.b	$1C
		dc.b	$75					; u
		dc.b	$DD					; 
		dc.b	$C5					; 
		dc.b	$58					; X
		dc.b	$2E					; .
		dc.b	$44					; D
		dc.b	$9D					; 
		dc.b	$ED					; 
		dc.b	$97					; 
		dc.b	$48					; H
		dc.b	$A8					; 
		dc.b	$69					; i
		dc.b	$58					; X
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$B8					; 
		dc.b	$78					; x
		dc.b	$26					; &
		dc.b	$66					; f
		dc.b	$C8					; 
		dc.b	$69					; i
		dc.b	$30					; 0
		dc.b	$A8					; 
		dc.b	$7B					; {
		dc.b	$AB					; 
		dc.b	$91					; 
		dc.b	$87					; 
		dc.b	$A7					; 
		dc.b	$26					; &
		dc.b	$68					; h
		dc.b	$B6					; 
		dc.b	$8A					; 
		dc.b	$98					; 
		dc.b	$91					; 
		dc.b	$53					; S
		dc.b	$64					; d
		dc.b	$C4					; 
		dc.b	$5F					; _
		dc.b	$5F					; _
		dc.b	$97					; 
		dc.b	$AD					; 
		dc.b	$7C					; |
		dc.b	$53					; S
		dc.b	$C5					; 
		dc.b	$97					; 
		dc.b	$22					; "
		dc.b	$58					; X
		dc.b	$D5					; 
		dc.b	$97					; 
		dc.b	$6F					; o
		dc.b	$4E					; N
		dc.b	$8B					; 
		dc.b	$8B					; 
		dc.b	$4E					; N
		dc.b	$CA					; 
		dc.b	$B7					; 
		dc.b	$16
		dc.b	$72					; r
		dc.b	$A6					; 
		dc.b	$3A					; :
		dc.b	$AA					; 
		dc.b	$D3					; 
		dc.b	$31					; 1
		dc.b	$73					; s
		dc.b	$D9					; 
		dc.b	$56					; V
		dc.b	$80					; 
		dc.b	$E1					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$62					; b
		dc.b	$25					; %
		dc.b	$54					; T
		dc.b	$70					; p
		dc.b	$9A					; 
		dc.b	$8A					; 
		dc.b	$50					; P
		dc.b	$71					; q
		dc.b	$B2					; 
		dc.b	$2B					; +
		dc.b	$71					; q
		dc.b	$F9					; 
		dc.b	$95					; 
		dc.b	$1E
		dc.b	$8A					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$77					; w
		dc.b	$61					; a
		dc.b	$D0					; 
		dc.b	$5C					; \
		dc.b	$35					; 5
		dc.b	$A4					; 
		dc.b	$5F					; _
		dc.b	$6D					; m
		dc.b	$F0					; 
		dc.b	$BE					; 
		dc.b	$59					; Y
		dc.b	$9A					; 
		dc.b	$4C					; L
		dc.b	$43					; C
		dc.b	$94					; 
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$AE					; 
		dc.b	$83					; 
		dc.b	$93					; 
		dc.b	$8F					; 
		dc.b	$10
		dc.b	$34					; 4
		dc.b	$BD					; 
		dc.b	$B6					; 
		dc.b	$6E					; n
		dc.b	$56					; V
		dc.b	$AF					; 
		dc.b	$9E					; 
		dc.b	$6C					; l
		dc.b	$3C					; <
		dc.b	$4A					; J
		dc.b	$8C					; 
		dc.b	$D1					; 
		dc.b	$90					; 
		dc.b	$3E					; >
		dc.b	$5E					; ^
		dc.b	$E5					; 
		dc.b	$97					; 
		dc.b	$1B
		dc.b	$90					; 
		dc.b	$CB					; 
		dc.b	$3C					; <
		dc.b	$43					; C
		dc.b	$D4					; 
		dc.b	$9B					; 
		dc.b	$25					; %
		dc.b	$49					; I
		dc.b	$C8					; 
		dc.b	$D7					; 
		dc.b	$A0					; 
		dc.b	$48					; H
		dc.b	$6A					; j
		dc.b	$95					; 
		dc.b	$5B					; [
		dc.b	$5D					; ]
		dc.b	$A4					; 
		dc.b	$9F					; 
		dc.b	$99					; 
		dc.b	$B0					; 
		dc.b	$4D					; M
		dc.b	$5B					; [
		dc.b	$67					; g
		dc.b	$7F					; 
		dc.b	$C0					; 
		dc.b	$BE					; 
		dc.b	$59					; Y
		dc.b	$2A					; *
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$70					; p
		dc.b	$A4					; 
		dc.b	$A6					; 
		dc.b	$61					; a
		dc.b	$76					; v
		dc.b	$5B					; [
		dc.b	$66					; f
		dc.b	$8C					; 
		dc.b	$8C					; 
		dc.b	$9F					; 
		dc.b	$41					; A
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$64					; d
		dc.b	$85					; 
		dc.b	$53					; S
		dc.b	$B0					; 
		dc.b	$CA					; 
		dc.b	$4E					; N
		dc.b	$7E					; ~
		dc.b	$A3					; 
		dc.b	$5B					; [
		dc.b	$3A					; :
		dc.b	$8F					; 
		dc.b	$A8					; 
		dc.b	$92					; 
		dc.b	$76					; v
		dc.b	$46					; F
		dc.b	$A3					; 
		dc.b	$70					; p
		dc.b	$96					; 
		dc.b	$66					; f
		dc.b	$8C					; 
		dc.b	$71					; q
		dc.b	$89					; 
		dc.b	$86					; 
		dc.b	$47					; G
		dc.b	$AC					; 
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$B0					; 
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$90					; 
		dc.b	$71					; q
		dc.b	$85					; 
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$91					; 
		dc.b	$7D					; }
		dc.b	$7D					; }
		dc.b	$79					; y
		dc.b	$5D					; ]
		dc.b	$C2					; 
		dc.b	$B0					; 
		dc.b	$41					; A
		dc.b	$8D					; 
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$53					; S
		dc.b	$94					; 
		dc.b	$B3					; 
		dc.b	$81					; 
		dc.b	$93					; 
		dc.b	$49					; I
		dc.b	$97					; 
		dc.b	$B3					; 
		dc.b	$74					; t
		dc.b	$64					; d
		dc.b	$60					; `
		dc.b	$AE					; 
		dc.b	$98					; 
		dc.b	$4A					; J
		dc.b	$66					; f
		dc.b	$92					; 
		dc.b	$89					; 
		dc.b	$48					; H
		dc.b	$6F					; o
		dc.b	$B6					; 
		dc.b	$6B					; k
		dc.b	$71					; q
		dc.b	$B2					; 
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$A3					; 
		dc.b	$85					; 
		dc.b	$5A					; Z
		dc.b	$8E					; 
		dc.b	$89					; 
		dc.b	$60					; `
		dc.b	$8B					; 
		dc.b	$74					; t
		dc.b	$5B					; [
		dc.b	$51					; Q
		dc.b	$9F					; 
		dc.b	$AC					; 
		dc.b	$57					; W
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$29					; )
		dc.b	$51					; Q
		dc.b	$B6					; 
		dc.b	$76					; v
		dc.b	$72					; r
		dc.b	$B6					; 
		dc.b	$B2					; 
		dc.b	$8F					; 
		dc.b	$4C					; L
		dc.b	$8B					; 
		dc.b	$C6					; 
		dc.b	$7A					; z
		dc.b	$3D					; =
		dc.b	$70					; p
		dc.b	$84					; 
		dc.b	$C0					; 
		dc.b	$9E					; 
		dc.b	$60					; `
		dc.b	$85					; 
		dc.b	$6F					; o
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$53					; S
		dc.b	$4B					; K
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$6C					; l
		dc.b	$6A					; j
		dc.b	$AE					; 
		dc.b	$94					; 
		dc.b	$41					; A
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$AE					; 
		dc.b	$8D					; 
		dc.b	$37					; 7
		dc.b	$83					; 
		dc.b	$C3					; 
		dc.b	$59					; Y
		dc.b	$56					; V
		dc.b	$A8					; 
		dc.b	$8F					; 
		dc.b	$29					; )
		dc.b	$75					; u
		dc.b	$D3					; 
		dc.b	$75					; u
		dc.b	$37					; 7
		dc.b	$93					; 
		dc.b	$C9					; 
		dc.b	$71					; q
		dc.b	$54					; T
		dc.b	$78					; x
		dc.b	$AB					; 
		dc.b	$74					; t
		dc.b	$67					; g
		dc.b	$B3					; 
		dc.b	$82					; 
		dc.b	$55					; U
		dc.b	$67					; g
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$81					; 
		dc.b	$93					; 
		dc.b	$98					; 
		dc.b	$70					; p
		dc.b	$59					; Y
		dc.b	$78					; x
		dc.b	$5B					; [
		dc.b	$A3					; 
		dc.b	$95					; 
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$6F					; o
		dc.b	$AD					; 
		dc.b	$8E					; 
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$9A					; 
		dc.b	$6E					; n
		dc.b	$58					; X
		dc.b	$AE					; 
		dc.b	$A8					; 
		dc.b	$73					; s
		dc.b	$5B					; [
		dc.b	$3C					; <
		dc.b	$9E					; 
		dc.b	$CA					; 
		dc.b	$6B					; k
		dc.b	$97					; 
		dc.b	$68					; h
		dc.b	$7B					; {
		dc.b	$74					; t
		dc.b	$52					; R
		dc.b	$5D					; ]
		dc.b	$8A					; 
		dc.b	$A2					; 
		dc.b	$71					; q
		dc.b	$5F					; _
		dc.b	$A9					; 
		dc.b	$80					; 
		dc.b	$63					; c
		dc.b	$91					; 
		dc.b	$74					; t
		dc.b	$81					; 
		dc.b	$71					; q
		dc.b	$84					; 
		dc.b	$73					; s
		dc.b	$72					; r
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$A3					; 
		dc.b	$64					; d
		dc.b	$7D					; }
		dc.b	$C4					; 
		dc.b	$65					; e
		dc.b	$4D					; M
		dc.b	$9A					; 
		dc.b	$6D					; m
		dc.b	$52					; R
		dc.b	$B1					; 
		dc.b	$B0					; 
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$5C					; \
		dc.b	$3A					; :
		dc.b	$74					; t
		dc.b	$CB					; 
		dc.b	$BB					; 
		dc.b	$65					; e
		dc.b	$8D					; 
		dc.b	$94					; 
		dc.b	$38					; 8
		dc.b	$5E					; ^
		dc.b	$AF					; 
		dc.b	$94					; 
		dc.b	$9A					; 
		dc.b	$91					; 
		dc.b	$9B					; 
		dc.b	$75					; u
		dc.b	$6C					; l
		dc.b	$66					; f
		dc.b	$93					; 
		dc.b	$94					; 
		dc.b	$A1					; 
		dc.b	$B1					; 
		dc.b	$34					; 4
		dc.b	$41					; A
		dc.b	$B1					; 
		dc.b	$7F					; 
		dc.b	$52					; R
		dc.b	$98					; 
		dc.b	$BE					; 
		dc.b	$99					; 
		dc.b	$61					; a
		dc.b	$5A					; Z
		dc.b	$86					; 
		dc.b	$C8					; 
		dc.b	$A3					; 
		dc.b	$4C					; L
		dc.b	$39					; 9
		dc.b	$97					; 
		dc.b	$A7					; 
		dc.b	$71					; q
		dc.b	$46					; F
		dc.b	$97					; 
		dc.b	$98					; 
		dc.b	$5F					; _
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$4D					; M
		dc.b	$85					; 
		dc.b	$7C					; |
		dc.b	$45					; E
		dc.b	$41					; A
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$55					; U
		dc.b	$72					; r
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$A8					; 
		dc.b	$D6					; 
		dc.b	$79					; y
		dc.b	$58					; X
		dc.b	$97					; 
		dc.b	$9C					; 
		dc.b	$99					; 
		dc.b	$93					; 
		dc.b	$AF					; 
		dc.b	$7C					; |
		dc.b	$41					; A
		dc.b	$67					; g
		dc.b	$81					; 
		dc.b	$AC					; 
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$95					; 
		dc.b	$54					; T
		dc.b	$7C					; |
		dc.b	$AF					; 
		dc.b	$86					; 
		dc.b	$76					; v
		dc.b	$78					; x
		dc.b	$5A					; Z
		dc.b	$62					; b
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$58					; X
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$62					; b
		dc.b	$46					; F
		dc.b	$4D					; M
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$A2					; 
		dc.b	$82					; 
		dc.b	$89					; 
		dc.b	$A6					; 
		dc.b	$52					; R
		dc.b	$86					; 
		dc.b	$B4					; 
		dc.b	$6B					; k
		dc.b	$86					; 
		dc.b	$B5					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$AD					; 
		dc.b	$96					; 
		dc.b	$9B					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$85					; 
		dc.b	$9D					; 
		dc.b	$A9					; 
		dc.b	$8D					; 
		dc.b	$A6					; 
		dc.b	$3C					; <
		dc.b	$51					; Q
		dc.b	$74					; t
		dc.b	$5A					; Z
		dc.b	$9E					; 
		dc.b	$79					; y
		dc.b	$59					; Y
		dc.b	$A2					; 
		dc.b	$51					; Q
		dc.b	$2C					; ,
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$78					; x
		dc.b	$8B					; 
		dc.b	$6A					; j
		dc.b	$83					; 
		dc.b	$99					; 
		dc.b	$79					; y
		dc.b	$61					; a
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$46					; F
		dc.b	$84					; 
		dc.b	$A1					; 
		dc.b	$4B					; K
		dc.b	$5F					; _
		dc.b	$9C					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$C7					; 
		dc.b	$99					; 
		dc.b	$60					; `
		dc.b	$B0					; 
		dc.b	$81					; 
		dc.b	$77					; w
		dc.b	$C9					; 
		dc.b	$81					; 
		dc.b	$4B					; K
		dc.b	$9F					; 
		dc.b	$B9					; 
		dc.b	$47					; G
		dc.b	$8D					; 
		dc.b	$CD					; 
		dc.b	$6D					; m
		dc.b	$52					; R
		dc.b	$6F					; o
		dc.b	$85					; 
		dc.b	$5D					; ]
		dc.b	$74					; t
		dc.b	$AE					; 
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$90					; 
		dc.b	$5A					; Z
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$74					; t
		dc.b	$8F					; 
		dc.b	$65					; e
		dc.b	$66					; f
		dc.b	$90					; 
		dc.b	$60					; `
		dc.b	$67					; g
		dc.b	$9E					; 
		dc.b	$A2					; 
		dc.b	$82					; 
		dc.b	$56					; V
		dc.b	$4E					; N
		dc.b	$95					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$AD					; 
		dc.b	$A8					; 
		dc.b	$8A					; 
		dc.b	$57					; W
		dc.b	$6C					; l
		dc.b	$8C					; 
		dc.b	$95					; 
		dc.b	$B3					; 
		dc.b	$A3					; 
		dc.b	$9D					; 
		dc.b	$86					; 
		dc.b	$6C					; l
		dc.b	$53					; S
		dc.b	$A2					; 
		dc.b	$9C					; 
		dc.b	$4C					; L
		dc.b	$78					; x
		dc.b	$84					; 
		dc.b	$66					; f
		dc.b	$8B					; 
		dc.b	$5E					; ^
		dc.b	$5F					; _
		dc.b	$BB					; 
		dc.b	$87					; 
		dc.b	$31					; 1
		dc.b	$7B					; {
		dc.b	$67					; g
		dc.b	$77					; w
		dc.b	$67					; g
		dc.b	$5A					; Z
		dc.b	$8B					; 
		dc.b	$98					; 
		dc.b	$85					; 
		dc.b	$46					; F
		dc.b	$7D					; }
		dc.b	$C1					; 
		dc.b	$84					; 
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$4A					; J
		dc.b	$72					; r
		dc.b	$91					; 
		dc.b	$90					; 
		dc.b	$9C					; 
		dc.b	$73					; s
		dc.b	$8F					; 
		dc.b	$89					; 
		dc.b	$2C					; ,
		dc.b	$58					; X
		dc.b	$AE					; 
		dc.b	$B3					; 
		dc.b	$68					; h
		dc.b	$70					; p
		dc.b	$96					; 
		dc.b	$96					; 
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$96					; 
		dc.b	$C2					; 
		dc.b	$85					; 
		dc.b	$6B					; k
		dc.b	$7E					; ~
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$46					; F
		dc.b	$A5					; 
		dc.b	$9A					; 
		dc.b	$66					; f
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$89					; 
		dc.b	$66					; f
		dc.b	$82					; 
		dc.b	$A1					; 
		dc.b	$5D					; ]
		dc.b	$71					; q
		dc.b	$A6					; 
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$AF					; 
		dc.b	$A5					; 
		dc.b	$71					; q
		dc.b	$7C					; |
		dc.b	$97					; 
		dc.b	$67					; g
		dc.b	$89					; 
		dc.b	$A5					; 
		dc.b	$96					; 
		dc.b	$79					; y
		dc.b	$B6					; 
		dc.b	$92					; 
		dc.b	$52					; R
		dc.b	$99					; 
		dc.b	$A9					; 
		dc.b	$5B					; [
		dc.b	$7F					; 
		dc.b	$90					; 
		dc.b	$6B					; k
		dc.b	$AB					; 
		dc.b	$7B					; {
		dc.b	$31					; 1
		dc.b	$76					; v
		dc.b	$64					; d
		dc.b	$55					; U
		dc.b	$7A					; z
		dc.b	$74					; t
		dc.b	$50					; P
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$9C					; 
		dc.b	$70					; p
		dc.b	$55					; U
		dc.b	$5D					; ]
		dc.b	$74					; t
		dc.b	$88					; 
		dc.b	$6D					; m
		dc.b	$51					; Q
		dc.b	$6B					; k
		dc.b	$6A					; j
		dc.b	$5E					; ^
		dc.b	$8F					; 
		dc.b	$9B					; 
		dc.b	$52					; R
		dc.b	$6A					; j
		dc.b	$6B					; k
		dc.b	$9D					; 
		dc.b	$99					; 
		dc.b	$52					; R
		dc.b	$7A					; z
		dc.b	$76					; v
		dc.b	$91					; 
		dc.b	$B3					; 
		dc.b	$6E					; n
		dc.b	$A7					; 
		dc.b	$6D					; m
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$A9					; 
		dc.b	$9F					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$95					; 
		dc.b	$A5					; 
		dc.b	$94					; 
		dc.b	$64					; d
		dc.b	$68					; h
		dc.b	$8A					; 
		dc.b	$72					; r
		dc.b	$9A					; 
		dc.b	$D4					; 
		dc.b	$93					; 
		dc.b	$5E					; ^
		dc.b	$76					; v
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$78					; x
		dc.b	$59					; Y
		dc.b	$68					; h
		dc.b	$97					; 
		dc.b	$5F					; _
		dc.b	$87					; 
		dc.b	$A3					; 
		dc.b	$3D					; =
		dc.b	$40					; @
		dc.b	$91					; 
		dc.b	$97					; 
		dc.b	$A6					; 
		dc.b	$5B					; [
		dc.b	$41					; A
		dc.b	$8A					; 
		dc.b	$A4					; 
		dc.b	$7C					; |
		dc.b	$78					; x
		dc.b	$5F					; _
		dc.b	$85					; 
		dc.b	$CF					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$54					; T
		dc.b	$89					; 
		dc.b	$AC					; 
		dc.b	$54					; T
		dc.b	$57					; W
		dc.b	$AC					; 
		dc.b	$BF					; 
		dc.b	$6B					; k
		dc.b	$7A					; z
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$76					; v
		dc.b	$48					; H
		dc.b	$A9					; 
		dc.b	$93					; 
		dc.b	$6D					; m
		dc.b	$91					; 
		dc.b	$61					; a
		dc.b	$97					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$60					; `
		dc.b	$7D					; }
		dc.b	$CA					; 
		dc.b	$84					; 
		dc.b	$39					; 9
		dc.b	$6E					; n
		dc.b	$89					; 
		dc.b	$91					; 
		dc.b	$9D					; 
		dc.b	$68					; h
		dc.b	$A2					; 
		dc.b	$7C					; |
		dc.b	$52					; R
		dc.b	$65					; e
		dc.b	$54					; T
		dc.b	$A4					; 
		dc.b	$B8					; 
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$85					; 
		dc.b	$8F					; 
		dc.b	$4D					; M
		dc.b	$37					; 7
		dc.b	$7E					; ~
		dc.b	$8D					; 
		dc.b	$7C					; |
		dc.b	$B0					; 
		dc.b	$87					; 
		dc.b	$55					; U
		dc.b	$A7					; 
		dc.b	$60					; `
		dc.b	$86					; 
		dc.b	$C0					; 
		dc.b	$66					; f
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$45					; E
		dc.b	$57					; W
		dc.b	$92					; 
		dc.b	$AE					; 
		dc.b	$7F					; 
		dc.b	$A9					; 
		dc.b	$96					; 
		dc.b	$4F					; O
		dc.b	$7B					; {
		dc.b	$5D					; ]
		dc.b	$A0					; 
		dc.b	$86					; 
		dc.b	$4E					; N
		dc.b	$8A					; 
		dc.b	$B1					; 
		dc.b	$6D					; m
		dc.b	$79					; y
		dc.b	$95					; 
		dc.b	$70					; p
		dc.b	$6C					; l
		dc.b	$68					; h
		dc.b	$AB					; 
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$95					; 
		dc.b	$8D					; 
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$95					; 
		dc.b	$82					; 
		dc.b	$98					; 
		dc.b	$50					; P
		dc.b	$3E					; >
		dc.b	$7D					; }
		dc.b	$C1					; 
		dc.b	$9A					; 
		dc.b	$61					; a
		dc.b	$50					; P
		dc.b	$84					; 
		dc.b	$95					; 
		dc.b	$57					; W
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$9A					; 
		dc.b	$5E					; ^
		dc.b	$85					; 
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$9F					; 
		dc.b	$79					; y
		dc.b	$92					; 
		dc.b	$58					; X
		dc.b	$A8					; 
		dc.b	$76					; v
		dc.b	$66					; f
		dc.b	$B0					; 
		dc.b	$65					; e
		dc.b	$82					; 
		dc.b	$97					; 
		dc.b	$82					; 
		dc.b	$96					; 
		dc.b	$59					; Y
		dc.b	$8E					; 
		dc.b	$67					; g
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$6F					; o
		dc.b	$AE					; 
		dc.b	$88					; 
		dc.b	$5C					; \
		dc.b	$7C					; |
		dc.b	$6D					; m
		dc.b	$5F					; _
		dc.b	$B2					; 
		dc.b	$8C					; 
		dc.b	$69					; i
		dc.b	$A3					; 
		dc.b	$67					; g
		dc.b	$64					; d
		dc.b	$A2					; 
		dc.b	$70					; p
		dc.b	$5B					; [
		dc.b	$97					; 
		dc.b	$93					; 
		dc.b	$72					; r
		dc.b	$4F					; O
		dc.b	$81					; 
		dc.b	$6B					; k
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$A2					; 
		dc.b	$B0					; 
		dc.b	$70					; p
		dc.b	$51					; Q
		dc.b	$59					; Y
		dc.b	$71					; q
		dc.b	$A1					; 
		dc.b	$9B					; 
		dc.b	$64					; d
		dc.b	$9D					; 
		dc.b	$9F					; 
		dc.b	$72					; r
		dc.b	$47					; G
		dc.b	$93					; 
		dc.b	$8E					; 
		dc.b	$54					; T
		dc.b	$64					; d
		dc.b	$98					; 
		dc.b	$90					; 
		dc.b	$67					; g
		dc.b	$AE					; 
		dc.b	$6C					; l
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$7F					; 
		dc.b	$A3					; 
		dc.b	$79					; y
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$A9					; 
		dc.b	$AE					; 
		dc.b	$7E					; ~
		dc.b	$72					; r
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$66					; f
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$A6					; 
		dc.b	$8C					; 
		dc.b	$56					; V
		dc.b	$65					; e
		dc.b	$8C					; 
		dc.b	$7A					; z
		dc.b	$55					; U
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$3A					; :
		dc.b	$77					; w
		dc.b	$A3					; 
		dc.b	$74					; t
		dc.b	$60					; `
		dc.b	$62					; b
		dc.b	$92					; 
		dc.b	$85					; 
		dc.b	$6E					; n
		dc.b	$AE					; 
		dc.b	$92					; 
		dc.b	$63					; c
		dc.b	$6C					; l
		dc.b	$66					; f
		dc.b	$69					; i
		dc.b	$8B					; 
		dc.b	$A7					; 
		dc.b	$8F					; 
		dc.b	$6C					; l
		dc.b	$72					; r
		dc.b	$AD					; 
		dc.b	$8F					; 
		dc.b	$72					; r
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$5D					; ]
		dc.b	$A4					; 
		dc.b	$9B					; 
		dc.b	$7B					; {
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$63					; c
		dc.b	$4F					; O
		dc.b	$81					; 
		dc.b	$98					; 
		dc.b	$A2					; 
		dc.b	$9B					; 
		dc.b	$66					; f
		dc.b	$74					; t
		dc.b	$84					; 
		dc.b	$88					; 
		dc.b	$89					; 
		dc.b	$A7					; 
		dc.b	$77					; w
		dc.b	$5D					; ]
		dc.b	$73					; s
		dc.b	$8C					; 
		dc.b	$B1					; 
		dc.b	$99					; 
		dc.b	$5B					; [
		dc.b	$5F					; _
		dc.b	$9E					; 
		dc.b	$7F					; 
		dc.b	$5E					; ^
		dc.b	$89					; 
		dc.b	$81					; 
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$77					; w
		dc.b	$6A					; j
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$61					; a
		dc.b	$77					; w
		dc.b	$79					; y
		dc.b	$75					; u
		dc.b	$97					; 
		dc.b	$9B					; 
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$84					; 
		dc.b	$8F					; 
		dc.b	$84					; 
		dc.b	$70					; p
		dc.b	$7D					; }
		dc.b	$9B					; 
		dc.b	$88					; 
		dc.b	$5B					; [
		dc.b	$84					; 
		dc.b	$9B					; 
		dc.b	$71					; q
		dc.b	$70					; p
		dc.b	$80					; 
		dc.b	$9F					; 
		dc.b	$65					; e
		dc.b	$7B					; {
		dc.b	$8D					; 
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$85					; 
		dc.b	$75					; u
		dc.b	$59					; Y
		dc.b	$75					; u
		dc.b	$69					; i
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$5E					; ^
		dc.b	$9C					; 
		dc.b	$86					; 
		dc.b	$66					; f
		dc.b	$8C					; 
		dc.b	$84					; 
		dc.b	$93					; 
		dc.b	$84					; 
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$63					; c
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$98					; 
		dc.b	$65					; e
		dc.b	$5A					; Z
		dc.b	$90					; 
		dc.b	$8A					; 
		dc.b	$68					; h
		dc.b	$9B					; 
		dc.b	$87					; 
		dc.b	$77					; w
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$8B					; 
		dc.b	$84					; 
		dc.b	$6F					; o
		dc.b	$7F					; 
		dc.b	$87					; 
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$65					; e
		dc.b	$89					; 
		dc.b	$91					; 
		dc.b	$8B					; 
		dc.b	$95					; 
		dc.b	$55					; U
		dc.b	$71					; q
		dc.b	$9F					; 
		dc.b	$90					; 
		dc.b	$6A					; j
		dc.b	$8E					; 
		dc.b	$7A					; z
		dc.b	$5F					; _
		dc.b	$86					; 
		dc.b	$5E					; ^
		dc.b	$83					; 
		dc.b	$AF					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$91					; 
		dc.b	$72					; r
		dc.b	$5A					; Z
		dc.b	$89					; 
		dc.b	$98					; 
		dc.b	$6B					; k
		dc.b	$6C					; l
		dc.b	$74					; t
		dc.b	$AF					; 
		dc.b	$9F					; 
		dc.b	$5C					; \
		dc.b	$6D					; m
		dc.b	$9E					; 
		dc.b	$68					; h
		dc.b	$52					; R
		dc.b	$83					; 
		dc.b	$8E					; 
		dc.b	$63					; c
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$78					; x
		dc.b	$8A					; 
		dc.b	$8E					; 
		dc.b	$79					; y
		dc.b	$93					; 
		dc.b	$6F					; o
		dc.b	$61					; a
		dc.b	$7F					; 
		dc.b	$67					; g
		dc.b	$77					; w
		dc.b	$92					; 
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$6E					; n
		dc.b	$7F					; 
		dc.b	$6B					; k
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$7E					; ~
		dc.b	$93					; 
		dc.b	$87					; 
		dc.b	$63					; c
		dc.b	$5A					; Z
		dc.b	$7C					; |
		dc.b	$95					; 
		dc.b	$73					; s
		dc.b	$8C					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$85					; 
		dc.b	$72					; r
		dc.b	$7C					; |
		dc.b	$8C					; 
		dc.b	$A3					; 
		dc.b	$81					; 
		dc.b	$53					; S
		dc.b	$7C					; |
		dc.b	$9C					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$95					; 
		dc.b	$88					; 
		dc.b	$67					; g
		dc.b	$5C					; \
		dc.b	$7F					; 
		dc.b	$8D					; 
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$A0					; 
		dc.b	$A0					; 
		dc.b	$70					; p
		dc.b	$74					; t
		dc.b	$89					; 
		dc.b	$80					; 
		dc.b	$8A					; 
		dc.b	$8F					; 
		dc.b	$8E					; 
		dc.b	$75					; u
		dc.b	$7E					; ~
		dc.b	$8B					; 
		dc.b	$70					; p
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$8A					; 
		dc.b	$60					; `
		dc.b	$73					; s
		dc.b	$93					; 
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$8C					; 
		dc.b	$74					; t
		dc.b	$64					; d
		dc.b	$6F					; o
		dc.b	$78					; x
		dc.b	$98					; 
		dc.b	$A9					; 
		dc.b	$77					; w
		dc.b	$73					; s
		dc.b	$70					; p
		dc.b	$6F					; o
		dc.b	$9F					; 
		dc.b	$91					; 
		dc.b	$65					; e
		dc.b	$60					; `
		dc.b	$8C					; 
		dc.b	$9D					; 
		dc.b	$68					; h
		dc.b	$5D					; ]
		dc.b	$85					; 
		dc.b	$77					; w
		dc.b	$75					; u
		dc.b	$94					; 
		dc.b	$96					; 
		dc.b	$66					; f
		dc.b	$78					; x
		dc.b	$80					; 
		dc.b	$73					; s
		dc.b	$6F					; o
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$65					; e
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$89					; 
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$74					; t
		dc.b	$61					; a
		dc.b	$87					; 
		dc.b	$91					; 
		dc.b	$8D					; 
		dc.b	$74					; t
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$7D					; }
		dc.b	$6D					; m
		dc.b	$81					; 
		dc.b	$86					; 
		dc.b	$93					; 
		dc.b	$8D					; 
		dc.b	$68					; h
		dc.b	$84					; 
		dc.b	$8A					; 
		dc.b	$85					; 
		dc.b	$81					; 
		dc.b	$70					; p
		dc.b	$9D					; 
		dc.b	$90					; 
		dc.b	$71					; q
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$82					; 
		dc.b	$8A					; 
		dc.b	$9D					; 
		dc.b	$7D					; }
		dc.b	$68					; h
		dc.b	$7E					; ~
		dc.b	$97					; 
		dc.b	$68					; h
		dc.b	$77					; w
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$75					; u
		dc.b	$6E					; n
		dc.b	$79					; y
		dc.b	$79					; y
		dc.b	$9D					; 
		dc.b	$90					; 
		dc.b	$6B					; k
		dc.b	$72					; r
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$81					; 
		dc.b	$6A					; j
		dc.b	$76					; v
		dc.b	$8E					; 
		dc.b	$6E					; n
		dc.b	$80					; 
		dc.b	$92					; 
		dc.b	$65					; e
		dc.b	$6B					; k
		dc.b	$8D					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$87					; 
		dc.b	$90					; 
		dc.b	$7F					; 
		dc.b	$71					; q
		dc.b	$78					; x
		dc.b	$86					; 
		dc.b	$71					; q
		dc.b	$87					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$A1					; 
		dc.b	$95					; 
		dc.b	$6D					; m
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$6A					; j
		dc.b	$86					; 
		dc.b	$98					; 
		dc.b	$71					; q
		dc.b	$8A					; 
		dc.b	$A1					; 
		dc.b	$84					; 
		dc.b	$68					; h
		dc.b	$62					; b
		dc.b	$76					; v
		dc.b	$9A					; 
		dc.b	$7C					; |
		dc.b	$56					; V
		dc.b	$8B					; 
		dc.b	$92					; 
		dc.b	$71					; q
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$74					; t
		dc.b	$69					; i
		dc.b	$73					; s
		dc.b	$7C					; |
		dc.b	$66					; f
		dc.b	$72					; r
		dc.b	$93					; 
		dc.b	$82					; 
		dc.b	$83					; 
		dc.b	$9B					; 
		dc.b	$81					; 
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$87					; 
		dc.b	$85					; 
		dc.b	$9C					; 
		dc.b	$7D					; }
		dc.b	$77					; w
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$6E					; n
		dc.b	$76					; v
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$70					; p
		dc.b	$73					; s
		dc.b	$8C					; 
		dc.b	$8A					; 
		dc.b	$94					; 
		dc.b	$79					; y
		dc.b	$6B					; k
		dc.b	$77					; w
		dc.b	$6F					; o
		dc.b	$90					; 
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$77					; w
		dc.b	$5E					; ^
		dc.b	$86					; 
		dc.b	$75					; u
		dc.b	$6A					; j
		dc.b	$8F					; 
		dc.b	$99					; 
		dc.b	$90					; 
		dc.b	$78					; x
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$62					; b
		dc.b	$7D					; }
		dc.b	$9E					; 
		dc.b	$8C					; 
		dc.b	$7C					; |
		dc.b	$62					; b
		dc.b	$83					; 
		dc.b	$88					; 
		dc.b	$71					; q
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$64					; d
		dc.b	$79					; y
		dc.b	$A5					; 
		dc.b	$87					; 
		dc.b	$61					; a
		dc.b	$7D					; }
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$93					; 
		dc.b	$82					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$73					; s
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$8E					; 
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$86					; 
		dc.b	$69					; i
		dc.b	$77					; w
		dc.b	$6D					; m
		dc.b	$65					; e
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$99					; 
		dc.b	$81					; 
		dc.b	$5E					; ^
		dc.b	$62					; b
		dc.b	$7B					; {
		dc.b	$87					; 
		dc.b	$8A					; 
		dc.b	$82					; 
unk_2BFE8:	dc.b	$84					; 		; DATA XREF: ROM:off_1B01Co
		dc.b	$80					; 
		dc.b	$70					; p
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$8F					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$88					; 
		dc.b	$9B					; 
		dc.b	$91					; 
		dc.b	$6C					; l
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$7B					; {
		dc.b	$91					; 
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$7A					; z
		dc.b	$6E					; n
		dc.b	$83					; 
		dc.b	$84					; 
		dc.b	$71					; q
		dc.b	$88					; 
		dc.b	$8D					; 
		dc.b	$8E					; 
		dc.b	$8D					; 
		dc.b	$76					; v
		dc.b	$7C					; |
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$8A					; 
		dc.b	$95					; 
		dc.b	$79					; y
		dc.b	$74					; t
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7E					; ~
		dc.b	$73					; s
		dc.b	$7B					; {
		dc.b	$71					; q
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7B					; {
		dc.b	$90					; 
		dc.b	$89					; 
		dc.b	$7E					; ~
		dc.b	$64					; d
		dc.b	$88					; 
		dc.b	$96					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$61					; a
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$86					; 
		dc.b	$8E					; 
		dc.b	$64					; d
		dc.b	$8B					; 
		dc.b	$85					; 
		dc.b	$6D					; m
		dc.b	$93					; 
		dc.b	$89					; 
		dc.b	$70					; p
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$73					; s
		dc.b	$89					; 
		dc.b	$77					; w
		dc.b	$7A					; z
		dc.b	$9D					; 
		dc.b	$8F					; 
		dc.b	$79					; y
		dc.b	$63					; c
		dc.b	$6E					; n
		dc.b	$9D					; 
		dc.b	$8E					; 
		dc.b	$67					; g
		dc.b	$79					; y
		dc.b	$98					; 
		dc.b	$75					; u
		dc.b	$75					; u
		dc.b	$88					; 
		dc.b	$75					; u
		dc.b	$86					; 
		dc.b	$89					; 
		dc.b	$72					; r
		dc.b	$77					; w
		dc.b	$87					; 
		dc.b	$8F					; 
		dc.b	$8C					; 
		dc.b	$83					; 
		dc.b	$89					; 
		dc.b	$82					; 
		dc.b	$76					; v
		dc.b	$69					; i
		dc.b	$6A					; j
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$98					; 
		dc.b	$80					; 
		dc.b	$68					; h
		dc.b	$73					; s
		dc.b	$7E					; ~
		dc.b	$8A					; 
		dc.b	$9A					; 
		dc.b	$81					; 
		dc.b	$60					; `
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$75					; u
		dc.b	$7E					; ~
		dc.b	$92					; 
		dc.b	$8B					; 
		dc.b	$80					; 
		dc.b	$76					; v
		dc.b	$5F					; _
		dc.b	$71					; q
		dc.b	$90					; 
		dc.b	$78					; x
		dc.b	$6F					; o
		dc.b	$81					; 
		dc.b	$9B					; 
		dc.b	$8D					; 
		dc.b	$68					; h
		dc.b	$6B					; k
		dc.b	$84					; 
		dc.b	$94					; 
		dc.b	$7A					; z
		dc.b	$6F					; o
		dc.b	$8D					; 
		dc.b	$96					; 
		dc.b	$6F					; o
		dc.b	$76					; v
		dc.b	$A2					; 
		dc.b	$85					; 
		dc.b	$69					; i
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$64					; d
		dc.b	$78					; x
		dc.b	$A1					; 
		dc.b	$81					; 
		dc.b	$65					; e
		dc.b	$84					; 
		dc.b	$96					; 
		dc.b	$6E					; n
		dc.b	$67					; g
		dc.b	$8D					; 
		dc.b	$92					; 
		dc.b	$70					; p
		dc.b	$7B					; {
		dc.b	$87					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$86					; 
		dc.b	$8B					; 
		dc.b	$6F					; o
		dc.b	$6F					; o
		dc.b	$87					; 
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$74					; t
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$79					; y
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$82					; 
		dc.b	$6C					; l
		dc.b	$7C					; |
		dc.b	$99					; 
		dc.b	$85					; 
		dc.b	$67					; g
		dc.b	$6F					; o
		dc.b	$80					; 
		dc.b	$99					; 
		dc.b	$8F					; 
		dc.b	$74					; t
		dc.b	$86					; 
		dc.b	$79					; y
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$8D					; 
		dc.b	$85					; 
		dc.b	$7B					; {
		dc.b	$8E					; 
		dc.b	$82					; 
		dc.b	$73					; s
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$82					; 
		dc.b	$5F					; _
		dc.b	$73					; s
		dc.b	$99					; 
		dc.b	$94					; 
		dc.b	$7B					; {
		dc.b	$79					; y
		dc.b	$84					; 
		dc.b	$70					; p
		dc.b	$70					; p
		dc.b	$87					; 
		dc.b	$8C					; 
		dc.b	$87					; 
		dc.b	$7C					; |
		dc.b	$69					; i
		dc.b	$83					; 
		dc.b	$8C					; 
		dc.b	$82					; 
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$76					; v
		dc.b	$7A					; z
		dc.b	$88					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$88					; 
		dc.b	$90					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$78					; x
		dc.b	$7F					; 
		dc.b	$82					; 
		dc.b	$84					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$8C					; 
		dc.b	$88					; 
		dc.b	$6B					; k
		dc.b	$6B					; k
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$7F					; 
		dc.b	$74					; t
		dc.b	$7F					; 
		dc.b	$8A					; 
		dc.b	$70					; p
		dc.b	$7A					; z
		dc.b	$8B					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$6F					; o
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$85					; 
		dc.b	$83					; 
		dc.b	$8C					; 
		dc.b	$91					; 
		dc.b	$82					; 
		dc.b	$77					; w
		dc.b	$7C					; |
		dc.b	$81					; 
		dc.b	$76					; v
		dc.b	$65					; e
		dc.b	$72					; r
		dc.b	$75					; u
		dc.b	$76					; v
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$8B					; 
		dc.b	$90					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$8F					; 
		dc.b	$7E					; ~
		dc.b	$74					; t
		dc.b	$83					; 
		dc.b	$8F					; 
		dc.b	$8D					; 
		dc.b	$89					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$6F					; o
		dc.b	$75					; u
		dc.b	$80					; 
		dc.b	$71					; q
		dc.b	$6C					; l
		dc.b	$7E					; ~
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$89					; 
		dc.b	$88					; 
		dc.b	$91					; 
		dc.b	$8E					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$86					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$78					; x
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$81					; 
		dc.b	$84					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$7B					; {
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$7B					; {
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$77					; w
		dc.b	$76					; v
		dc.b	$7D					; }
		dc.b	$88					; 
		dc.b	$8B					; 
		dc.b	$7E					; ~
		dc.b	$6D					; m
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$82					; 
		dc.b	$92					; 
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$76					; v
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$80					; 
		dc.b	$7B					; {
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$90					; 
		dc.b	$80					; 
		dc.b	$6C					; l
		dc.b	$73					; s
		dc.b	$84					; 
		dc.b	$87					; 
		dc.b	$86					; 
		dc.b	$72					; r
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$74					; t
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$99					; 
		dc.b	$8B					; 
		dc.b	$70					; p
		dc.b	$7F					; 
		dc.b	$86					; 
		dc.b	$6E					; n
		dc.b	$81					; 
		dc.b	$88					; 
		dc.b	$86					; 
		dc.b	$77					; w
		dc.b	$81					; 
		dc.b	$8A					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$88					; 
		dc.b	$8A					; 
		dc.b	$7A					; z
		dc.b	$80					; 
		dc.b	$86					; 
		dc.b	$80					; 
		dc.b	$79					; y
		dc.b	$86					; 
		dc.b	$88					; 
		dc.b	$88					; 
		dc.b	$84					; 
		dc.b	$7D					; }
		dc.b	$6E					; n
		dc.b	$72					; r
		dc.b	$85					; 
		dc.b	$8C					; 
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$93					; 
		dc.b	$89					; 
		dc.b	$7A					; z
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$86					; 
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$83					; 
		dc.b	$8B					; 
		dc.b	$7C					; |
		dc.b	$79					; y
		dc.b	$78					; x
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$81					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$81					; 
		dc.b	$7F					; 
		dc.b	$77					; w
		dc.b	$84					; 
		dc.b	$8D					; 
		dc.b	$79					; y
		dc.b	$70					; p
		dc.b	$82					; 
		dc.b	$90					; 
		dc.b	$78					; x
		dc.b	$7A					; z
		dc.b	$90					; 
		dc.b	$83					; 
		dc.b	$73					; s
		dc.b	$71					; q
		dc.b	$81					; 
		dc.b	$8B					; 
		dc.b	$83					; 
		dc.b	$75					; u
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$83					; 
		dc.b	$7D					; }
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$84					; 
		dc.b	$84					; 
		dc.b	$7F					; 
		dc.b	$83					; 
		dc.b	$7F					; 
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$83					; 
		dc.b	$8A					; 
		dc.b	$80					; 
		dc.b	$74					; t
		dc.b	$76					; v
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$76					; v
		dc.b	$81					; 
		dc.b	$8F					; 
		dc.b	$85					; 
		dc.b	$80					; 
		dc.b	$72					; r
		dc.b	$76					; v
		dc.b	$8F					; 
		dc.b	$8A					; 
		dc.b	$6E					; n
		dc.b	$71					; q
		dc.b	$8D					; 
		dc.b	$8F					; 
		dc.b	$87					; 
		dc.b	$6F					; o
		dc.b	$73					; s
		dc.b	$86					; 
		dc.b	$8D					; 
		dc.b	$84					; 
		dc.b	$7B					; {
		dc.b	$75					; u
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$8E					; 
		dc.b	$87					; 
		dc.b	$7A					; z
		dc.b	$78					; x
		dc.b	$77					; w
		dc.b	$80					; 
		dc.b	$84					; 
		dc.b	$86					; 
		dc.b	$8A					; 
		dc.b	$84					; 
		dc.b	$79					; y
		dc.b	$7C					; |
		dc.b	$85					; 
		dc.b	$88					; 
		dc.b	$83					; 
		dc.b	$87					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$85					; 
		dc.b	$7F					; 
		dc.b	$75					; u
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$7D					; }
		dc.b	$86					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$76					; v
		dc.b	$73					; s
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7B					; {
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$78					; x
		dc.b	$79					; y
		dc.b	$80					; 
		dc.b	$83					; 
		dc.b	$83					; 
		dc.b	$82					; 
		dc.b	$7E					; ~
		dc.b	$83					; 
		dc.b	$7E					; ~
		dc.b	$76					; v
		dc.b	$79					; y
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$7C					; |
		dc.b	$7E					; ~
		dc.b	$85					; 
		dc.b	$7E					; ~
		dc.b	$79					; y
		dc.b	$7F					; 
		dc.b	$81					; 
		dc.b	$7C					; |
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$81					; 
		dc.b	$7D					; }
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$81					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7E					; ~
		dc.b	$7E					; ~
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7D					; }
		dc.b	$80					; 
		dc.b	$82					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7F					; 
		dc.b	$7F					; 
		dc.b	$7E					; ~
		dc.b	$7F					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$80					; 
		dc.b	$7D					; }
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	0
unk_2C248:	include	"Unknown\2C248.asm"
off_2C648:	include	"Unknown\2C648 (Maybe Array Sprites).asm"
unk_3BE12:	include	"Unknown\3BE12 (Maybe Sprites).asm"
unk_44F54:	include	"Unknown\44F54.asm"
unk_4FE04:	include	"Unknown\4FE04.asm"
off_5DE7A:	include	"Unknown\5DE7A.asm"
off_8000A:	include	"Unknown\8000A.asm"
unk_94444:	include	"Unknown\94444.asm"
unk_AB928:	include	"Unknown\AB928.asm"
unk_B5188:	include	"Unknown\B5188.asm"
off_BFD66:	include	"Unknown\BFD66.asm"
unk_C682E:	include	"Unknown\C682E.asm"
unk_CF15A:	include	"Unknown\CF15A.asm"
off_D82B6:	include	"Unknown\D82B6.asm"
unk_DE9FA:	include	"Unknown\DE9FA.asm"
unk_E5B40:	include	"Unknown\E5B40.asm"
off_E9A80:	include	"Unknown\E9A80.asm"
unk_EE7A8:	include	"Unknown\EE7A8.asm"

; =============== S U B	R O U T	I N E =======================================


sub_F66EE:							; CODE XREF: sub_C2F2+240p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFB75E).w,d0
		bpl.w	loc_F66FC
		neg.w	d0

loc_F66FC:							; CODE XREF: sub_F66EE+8j
		sub.w	#$108,d0
		bpl.w	loc_F6706
		neg.w	d0

loc_F6706:							; CODE XREF: sub_F66EE+12j
		swap	d0
		and.l	#$FFFF0000,d0
		move.w	($FFFFB774).w,d1
		beq.w	loc_F6772
		bpl.w	loc_F671C
		neg.w	d1

loc_F671C:							; CODE XREF: sub_F66EE+28j
		move.w	#$11,d2
		tst.w	($FFFFD06E).w
		beq.w	loc_F672C
		move.w	#$16,d2

loc_F672C:							; CODE XREF: sub_F66EE+36j
		divu.w	d1,d0
		and.l	#$FFFF,d0
		divu.w	d2,d0
		tst.w	d0
		bne.w	loc_F6740
		move.w	#1,d0

loc_F6740:							; CODE XREF: sub_F66EE+4Aj
		move.l	#$A0000,d1
		move.w	d0,d3
		mulu.w	d2,d0
		divu.w	d0,d1
		move.w	d2,d4
		add.w	d2,d2
		add.w	d4,d2
		mulu.w	d3,d2
		cmpnl	 $7FFF,d2
		blt.w	loc_F6762
		move.w	#$7FFF,d2

loc_F6762:							; CODE XREF: sub_F66EE+6Cj
		add.w	d2,d1
		tst.w	d1
		bpl.w	loc_F676E
		move.w	#$7FFF,d1

loc_F676E:							; CODE XREF: sub_F66EE+78j
		move.w	d1,($FFFFB776).w

loc_F6772:							; CODE XREF: sub_F66EE+24j
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F66EE


; =============== S U B	R O U T	I N E =======================================


sub_F6778:							; CODE XREF: sub_BC40:loc_BCA4p
		movem.l	d0-a6,-(sp)
		move.w	#$C,d0
		move.w	($FFFFD418).w,d1
		mulu.w	d1,d0
		swap	d0
		and.l	#$FFFF0000,d0
		jsr	sub_110BE
		move.w	d0,($FFFFB776).w
		ext.l	d0
		move.w	#3,d4
		divu.w	d4,d0
		clr.l	d1
		move.w	($FFFFD414).w,d1
		sub.w	($FFFFB74A).w,d1
		swap	d1
		tst.w	d0
		bne.w	loc_F67B6
		move.w	#1,d0

loc_F67B6:							; CODE XREF: sub_F6778+36j
		divs.w	d0,d1
		move.w	d1,($FFFFB772).w
		clr.l	d1
		move.w	($FFFFD416).w,d1
		sub.w	($FFFFB75E).w,d1
		swap	d1
		divs.w	d0,d1
		move.w	d1,($FFFFB774).w
		move.l	#$FFFFB74A,a3
		move.w	($FFFFB776).w,d0
		jsr	sub_102D2
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F6778


; =============== S U B	R O U T	I N E =======================================


sub_F67E4:							; CODE XREF: sub_BC40+48p
		movem.l	d0-a6,-(sp)
		bclr	#2,($FFFFC2F8).w
		tst.w	$34(a3)
		bne.w	loc_F6800
		move.l	#unk_F691E,a0
		bra.w	loc_F6830
; ---------------------------------------------------------------------------

loc_F6800:							; CODE XREF: sub_F67E4+Ej
		move.w	$14(a3),d0
		btst	#7,$62(a3)
		bne.w	loc_F6810
		neg.w	d0

loc_F6810:							; CODE XREF: sub_F67E4+26j
		bset	#2,($FFFFC2F8).w
		move.l	#word_F68DE,a0
		cmpnw	 $58,d0					; 'X'
		blt.w	loc_F6830
		move.l	#unk_F68FE,a0
		bclr	#2,($FFFFC2F8).w

loc_F6830:							; CODE XREF: sub_F67E4+18j
					; sub_F67E4+3Cj
		move.w	$54(a3),d0
		btst	#7,$62(a3)
		bne.w	loc_F6844
		addq.w	#4,d0
		and.w	#7,d0

loc_F6844:							; CODE XREF: sub_F67E4+56j
		asl.w	#2,d0
		move.w	(a0,d0.w),d1
		move.w	2(a0,d0.w),d2
		cmp.l	#unk_F68FE,a0
		bne.w	loc_F687E
		move.w	$14(a3),d0
		bpl.w	loc_F6862
		neg.w	d0

loc_F6862:							; CODE XREF: sub_F67E4+78j
		cmpnw	 $8A,d0					; ''
		blt.w	loc_F687E
		move.w	(a3),d0
		bpl.w	loc_F6872
		neg.w	d0

loc_F6872:							; CODE XREF: sub_F67E4+88j
		cmpnw	 $37,d0					; '7'
		bgt.w	loc_F687E
		move.w	#$103,d2

loc_F687E:							; CODE XREF: sub_F67E4+70j
					; sub_F67E4+82j ...
		btst	#7,$62(a3)
		bne.w	loc_F688C
		neg.w	d1
		neg.w	d2

loc_F688C:							; CODE XREF: sub_F67E4+A0j
		move.w	#$A,d0
		jsr	sub_11086
		add.w	d0,d1
		move.w	#$A,d0
		jsr	sub_11086
		add.w	d0,d2
		move.w	d1,($FFFFD414).w
		tst.w	$34(a3)
		beq.w	loc_F68D4
		sub.w	(a3),d1
		bmi.w	loc_F68C8
		sub.w	#$3C,d1					; '<'
		bpl.w	loc_F68D4
		neg.w	d1
		add.w	d1,($FFFFD414).w
		bra.w	loc_F68D4
; ---------------------------------------------------------------------------

loc_F68C8:							; CODE XREF: sub_F67E4+CEj
		add.w	#$3C,d1					; '<'
		bmi.w	loc_F68D4
		sub.w	d1,($FFFFD414).w

loc_F68D4:							; CODE XREF: sub_F67E4+C8j
					; sub_F67E4+D6j ...
		move.w	d2,($FFFFD416).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F67E4

; ---------------------------------------------------------------------------
word_F68DE:	dc.w	$74					; DATA XREF: sub_F67E4+32o
		dc.w	$C2
		dc.w	$74
		dc.w	$C2
		dc.w	$74
		dc.w	$C2
		dc.w	$74
		dc.w	$C2
		dc.w	$FF8C
		dc.w	$C2
		dc.w	$FF8C
		dc.w	$C2
		dc.w	$FF8C
		dc.w	$C2
		dc.w	$FF8C
		dc.w	$C2
unk_F68FE:	dc.b	$FF					; DATA XREF: sub_F67E4+40o
					; sub_F67E4+6Ao
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	0
		dc.b	$AB					; 
unk_F691E:	dc.b	$FF					; DATA XREF: sub_F67E4+12o
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	0
		dc.b	$32					; 2
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	0
		dc.b	$32					; 2
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	0
		dc.b	$32					; 2
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	$FF
		dc.b	$FB					; 
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	$FF
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	$FF
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$F7					; 
		dc.b	$FF
		dc.b	$CE					; 
		dc.b	$FF
		dc.b	$F7					; 

; =============== S U B	R O U T	I N E =======================================


sub_F693E:							; DATA XREF: ROM:00018E08o
		bclr	#1,$62(a3)
		beq.w	loc_F6A3C
		bset	#3,$64(a3)
		bne.w	loc_F6A3C
		bclr	#0,($FFFFBF76).w
		bclr	#5,($FFFFC2FE).w
		clr.w	($FFFFBF76).w
		clr.w	($FFFFBF6E).w
		st	($FFFFBEE0).w
		movem.l	d0-a6,-(sp)
		move.w	$52(a3),d0
		move.w	d0,($FFFFBF6C).w
		btst	#3,$62(a3)
		bne.w	loc_F6A16
		tst.w	($FFFFDED8).w
		bmi.w	loc_F6A16
		beq.w	loc_F69D2
		cmp.w	#1,($FFFFDEDA).w
		bne.w	loc_F69C8
		move.w	($FFFFC32A).w,-(sp)
		move.w	($FFFFC322).w,-(sp)
		move.w	($FFFFC32E).w,($FFFFC32A).w
		move.w	($FFFFC326).w,($FFFFC322).w
		jsr	sub_C0DA
		move.w	($FFFFC322).w,($FFFFC326).w
		move.w	($FFFFC32A).w,($FFFFC32E).w
		move.w	(sp)+,($FFFFC322).w
		move.w	(sp)+,($FFFFC32A).w
		bra.w	loc_F6A16
; ---------------------------------------------------------------------------

loc_F69C8:							; CODE XREF: sub_F693E+54j
		jsr	sub_C0DA
		bra.w	loc_F6A16
; ---------------------------------------------------------------------------

loc_F69D2:							; CODE XREF: sub_F693E+4Aj
		tst.w	($FFFFDEDA).w
		bne.w	loc_F6A0C
		move.w	($FFFFC328).w,-(sp)
		move.w	($FFFFC320).w,-(sp)
		move.w	($FFFFC32C).w,($FFFFC328).w
		move.w	($FFFFC324).w,($FFFFC320).w
		jsr	sub_C0BC
		move.w	($FFFFC320).w,($FFFFC324).w
		move.w	($FFFFC328).w,($FFFFC32C).w
		move.w	(sp)+,($FFFFC320).w
		move.w	(sp)+,($FFFFC328).w
		bra.w	loc_F6A16
; ---------------------------------------------------------------------------

loc_F6A0C:							; CODE XREF: sub_F693E+98j
		jsr	sub_C0BC
		bra.w	*+4

loc_F6A16:							; CODE XREF: sub_F693E+3Ej
					; sub_F693E+46j ...
		move.w	d0,-(sp)
		bsr.w	sub_F6C0A
		clr.w	$5A(a3)
		jsr	sub_1073A
		bset	#5,$62(a3)
		bset	#1,$63(a3)
		move.w	(sp)+,d0
		movem.l	(sp)+,d0-a6
		bra.w	locret_F6C08
; ---------------------------------------------------------------------------

loc_F6A3C:							; CODE XREF: sub_F693E+6j
					; sub_F693E+10j
		btst	#0,($FFFFBF76).w
		bne.w	loc_F6A96
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		cmpnw	 $3C,d0					; '<'
		bgt.w	loc_F6A60
		cmpnw	 $FFC4,d0
		bgt.w	loc_F6A72

loc_F6A60:							; CODE XREF: sub_F693E+116j
		move.w	($FFFFB772).w,d1
		eor.w	d1,d0
		bmi.w	loc_F6A72

loc_F6A6A:							; CODE XREF: sub_F693E+152j
		movem.w	(sp)+,d0-d1
		bra.w	loc_F6A9E
; ---------------------------------------------------------------------------

loc_F6A72:							; CODE XREF: sub_F693E+11Ej
					; sub_F693E+128j
		move.w	($FFFFB75E).w,d0
		sub.w	$14(a3),d0
		cmpnw	 $3C,d0					; '<'
		bgt.w	loc_F6A8A
		cmpnw	 $FFC4,d0
		bgt.w	loc_F6A92

loc_F6A8A:							; CODE XREF: sub_F693E+140j
		move.w	($FFFFB774).w,d1
		eor.w	d1,d0
		bpl.s	loc_F6A6A

loc_F6A92:							; CODE XREF: sub_F693E+148j
		movem.w	(sp)+,d0-d1

loc_F6A96:							; CODE XREF: sub_F693E+104j
		tst.w	($FFFFB7AA).w
		bmi.w	loc_F6AA8

loc_F6A9E:							; CODE XREF: sub_F693E+130j
		bclr	#7,($FFFFC2FE).w
		bra.w	loc_F6C02
; ---------------------------------------------------------------------------

loc_F6AA8:							; CODE XREF: sub_F693E+15Cj
		btst	#0,($FFFFC2EA).w
		bne.w	loc_F6C02
		movem.l	d0-d1,-(sp)
		cmp.w	#$10,$5A(a3)
		bge.w	loc_F6B58
		btst	#2,($FFFFBF76).w
		bne.w	loc_F6B58
		move.w	(a3),d0
		move.w	($FFFFB74A).w,d1
		sub.w	d1,d0
		bpl.w	loc_F6AD8
		neg.w	d0

loc_F6AD8:							; CODE XREF: sub_F693E+194j
		move.w	$14(a3),d1
		move.w	($FFFFB75E).w,d2
		sub.w	d2,d1
		bpl.w	loc_F6AE8
		neg.w	d1

loc_F6AE8:							; CODE XREF: sub_F693E+1A4j
		move.w	($FFFFB772).w,d2
		beq.w	loc_F6AF6
		cmp.w	d0,d1
		ble.w	loc_F6AFC

loc_F6AF6:							; CODE XREF: sub_F693E+1AEj
		move.w	($FFFFB774).w,d2
		move.w	d1,d0

loc_F6AFC:							; CODE XREF: sub_F693E+1B4j
		swap	d0
		and.l	#$FFFF0000,d0
		tst.w	d2
		bpl.w	loc_F6B0C
		neg.w	d2

loc_F6B0C:							; CODE XREF: sub_F693E+1C8j
		move.w	#$11,d1
		tst.w	($FFFFD06E).w
		beq.w	loc_F6B1C
		move.w	#$16,d1

loc_F6B1C:							; CODE XREF: sub_F693E+1D6j
		tst.w	d2
		bne.w	loc_F6B26
		move.w	#1,d2

loc_F6B26:							; CODE XREF: sub_F693E+1E0j
		divu.w	d2,d0
		and.l	#$FFFF,d0
		divu.w	d1,d0
		move.w	$5A(a3),d2
		lsr.w	#2,d2
		subq.w	#6,d2
		neg.w	d2
		asl.w	#2,d2
		cmp.w	d2,d0
		bgt.w	loc_F6B50
		neg.w	$5A(a3)
		add.w	#$18,$5A(a3)
		bra.w	loc_F6BC8
; ---------------------------------------------------------------------------

loc_F6B50:							; CODE XREF: sub_F693E+200j
		add.w	d7,($FFFFBF6E).w
		bra.w	loc_F6B7E
; ---------------------------------------------------------------------------

loc_F6B58:							; CODE XREF: sub_F693E+17Ej
					; sub_F693E+188j
		bset	#2,($FFFFBF76).w
		btst	#1,($FFFFBF76).w
		bne.w	loc_F6B7E
		cmp.w	#$18,$5A(a3)
		bne.w	loc_F6B7E
		add.w	#$30,$5C(a3)				; '0'
		bset	#1,($FFFFBF76).w

loc_F6B7E:							; CODE XREF: sub_F693E+216j
					; sub_F693E+226j ...
		btst	#1,($FFFFBF76).w
		beq.w	loc_F6BA4
		cmp.w	#$18,$5A(a3)
		ble.w	loc_F6BA4
		btst	#0,($FFFFBF76).w
		bne.w	loc_F6BA4
		movem.l	(sp)+,d0-d1
		bra.w	loc_F6C02
; ---------------------------------------------------------------------------

loc_F6BA4:							; CODE XREF: sub_F693E+246j
					; sub_F693E+250j ...
		btst	#0,($FFFFBF76).w
		beq.w	loc_F6BC8
		cmp.w	#$18,$5A(a3)
		bne.w	loc_F6BC8
		cmp.w	#1,$5C(a3)
		ble.w	loc_F6BC8
		move.w	#1,$5C(a3)

loc_F6BC8:							; CODE XREF: sub_F693E+20Ej
					; sub_F693E+26Cj ...
		btst	#1,$63(a3)
		bne.w	loc_F6BDA
		movem.l	(sp)+,d0-d1
		bra.w	loc_F6BF8
; ---------------------------------------------------------------------------

loc_F6BDA:							; CODE XREF: sub_F693E+290j
		move.w	$5A(a3),d0
		movem.l	(sp)+,d0-d1
		btst	#0,($FFFFBF76).w
		beq.w	loc_F6BF6
		bclr	#5,$62(a3)
		bra.w	loc_F6BF8
; ---------------------------------------------------------------------------

loc_F6BF6:							; CODE XREF: sub_F693E+2AAj
		nop

loc_F6BF8:							; CODE XREF: sub_F693E+298j
					; sub_F693E+2B4j
		btst	#1,$63(a3)
		bne.w	locret_F6C08

loc_F6C02:							; CODE XREF: sub_F693E+166j
					; sub_F693E+170j ...
		jsr	sub_FEFF0

locret_F6C08:							; CODE XREF: sub_F693E+FAj
					; sub_F693E+2C0j
		rts
; End of function sub_F693E


; =============== S U B	R O U T	I N E =======================================


sub_F6C0A:							; CODE XREF: sub_F693E+DAp
		move.w	#$7FC,d1
		movem.w	d0-d1,-(sp)
		move.w	(a3),d0
		neg.w	d0
		move.w	#$108,d1
		btst	#7,$62(a3)
		bne.w	loc_F6C26
		neg.w	d1

loc_F6C26:							; CODE XREF: sub_F6C0A+16j
		sub.w	$14(a3),d1
		jsr	sub_10676
		jsr	sub_F6D22
		movem.w	(sp)+,d0-d1
		beq.w	locret_F6C42
		move.w	#$92E,d1

locret_F6C42:							; CODE XREF: sub_F6C0A+30j
		rts
; End of function sub_F6C0A


; =============== S U B	R O U T	I N E =======================================


sub_F6C44:							; CODE XREF: sub_B0E8+120p
					; sub_B0E8+622p ...
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB75E).w,d0
		btst	#7,$62(a3)
		bne.w	loc_F6C58
		neg.w	d0

loc_F6C58:							; CODE XREF: sub_F6C44+Ej
		tst.w	d0
		bpl.w	loc_F6CB0
		bra.w	loc_F6CA8
; ---------------------------------------------------------------------------
		move.w	($FFFFBEDC).w,d0
		addq.w	#4,d0
		and.w	#7,d0
		move.w	$54(a3),d1
		cmp.w	d0,d1
		bra.w	loc_F6CB0
; ---------------------------------------------------------------------------
		beq.w	loc_F6CB0
		addq.w	#1,d1
		and.w	#7,d1
		cmp.w	d0,d1
		beq.w	loc_F6CB0
		addq.w	#1,d1
		and.w	#7,d1
		cmp.w	d0,d1
		beq.w	loc_F6CB0
		subq.w	#3,d1
		and.w	#7,d1
		cmp.w	d0,d1
		beq.w	loc_F6CB0
		subq.w	#1,d1
		and.w	#7,d1
		beq.w	loc_F6CB0

loc_F6CA8:							; CODE XREF: sub_F6C44+1Aj
		move.w	#0,d0
		bra.w	loc_F6CB4
; ---------------------------------------------------------------------------

loc_F6CB0:							; CODE XREF: sub_F6C44+16j
					; sub_F6C44+2Ej ...
		move.w	#1,d0

loc_F6CB4:							; CODE XREF: sub_F6C44+68j
		movem.w	(sp)+,d0-d1
		rts
; End of function sub_F6C44


; =============== S U B	R O U T	I N E =======================================


sub_F6CBA:							; CODE XREF: sub_150E4+17Ap
		move.w	#4,($FFFFBEDC).w
		jsr	sub_C2F2
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFC6CE,a0
		btst	#6,$62(a3)
		beq.w	loc_F6CDE
		lea	$364(a0),a0

loc_F6CDE:							; CODE XREF: sub_F6CBA+1Cj
		clr.w	d0
		move.b	$66(a3),d0
		move.w	$1A(a0),$1C(a0)
		move.w	$18(a0),$1A(a0)
		move.w	d0,$18(a0)
		bset	#7,($FFFFC2FE).w
		add.w	#$96,($FFFFB89C).w			; ''
		add.w	#$A,($FFFFB8A2).w
		addq.w	#1,$35C(a0)
		movem.l	(sp)+,d0/a0
		bset	#0,($FFFFBF76).w
		bset	#1,$63(a3)
		bset	#1,($FFFFC2F8).w
		rts
; End of function sub_F6CBA


; =============== S U B	R O U T	I N E =======================================


sub_F6D22:							; CODE XREF: sub_F6C0A+26p
		movem.w	d0-d1,-(sp)
		neg.w	d0
		addq.w	#8,d0
		and.w	#7,d0
		move.w	$54(a3),d1
		add.w	d0,d1
		and.w	#7,d1
		cmpnw	 4,d1
		bgt.w	loc_F6D50
		move.w	4(a3),d0
		eor.w	#$FFFF,d0
		and.w	#$800,d0
		bra.w	loc_F6D58
; ---------------------------------------------------------------------------

loc_F6D50:							; CODE XREF: sub_F6D22+1Aj
		move.w	4(a3),d0
		and.w	#$800,d0

loc_F6D58:							; CODE XREF: sub_F6D22+2Aj
		movem.w	(sp)+,d0-d1
		rts
; End of function sub_F6D22


; =============== S U B	R O U T	I N E =======================================


Unk_ControlsSetRelated:						; CODE XREF: ROM:00007700p
		move.w	#0,($A11200).l
		move.b	#$40,($A10009).l			; '@'
		move.b	#$43,($A1000B).l			; 'C'
		nop
		move.b	#$7C,($A10005).l			; '|'
		nop
		move.b	#$7F,($A1000B).l			; ''
		nop
		move.b	#$7C,($A10005).l			; '|'
		nop
		move.b	($A10003).l,d0
		and.b	#3,d0
		cmpnb	 0,d0
		bne.s	loc_F6DAE
		move.w	#1,($FFFFD046).w
		bra.s	loc_F6DBC
; ---------------------------------------------------------------------------

loc_F6DAE:							; CODE XREF: Unk_ControlsSetRelated+46j
		move.w	#0,($FFFFD046).w
		move.b	#$40,($A1000B).l			; '@'

loc_F6DBC:							; CODE XREF: Unk_ControlsSetRelated+4Ej
		move.w	#$100,($A11200).l
		rts
; End of function Unk_ControlsSetRelated

; ---------------------------------------------------------------------------
		move.b	#0,($A10005).l
		move.w	($FFFFBEFE).w,($FFFFBEFA).w
		jsr	sub_11340
		move.w	($FFFFBEFA).w,($FFFFBEFE).w
		rts
; ---------------------------------------------------------------------------
		move.b	#$10,($A10005).l
		move.w	($FFFFBF00).w,($FFFFBEFA).w
		jsr	sub_11340
		move.w	($FFFFBEFA).w,($FFFFBF00).w
		rts
; ---------------------------------------------------------------------------
		move.b	#$20,($A10005).l			; ' '
		move.w	($FFFFBF02).w,($FFFFBEFA).w
		jsr	sub_11340
		move.w	($FFFFBEFA).w,($FFFFBF02).w
		rts
; ---------------------------------------------------------------------------
		move.b	#$30,($A10005).l			; '0'
		move.w	($FFFFBF04).w,($FFFFBEFA).w
		jsr	sub_11340
		move.w	($FFFFBEFA).w,($FFFFBF04).w
		rts
; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


nullsub_2:							; CODE XREF: sub_159A8+28p
		rts						; REMOVED SUBROUTINE!
; End of function nullsub_2

; ---------------------------------------------------------------------------
		cmp.w	#$FFFF,($FFFFBE78).w
		beq.w	locret_F6E88
		movem.l	d0-d2,-(sp)
		move.w	#1,d0
		btst	#6,$62(a3)
		beq.w	loc_F6E5A
		move.w	#2,d0

loc_F6E5A:							; CODE XREF: ROM:000F6E52j
		cmp.w	($FFFFC32C).w,d0
		beq.w	loc_F6E6E
		move.w	#$F,d1
		move.w	#4,d0
		bra.w	loc_F6E76
; ---------------------------------------------------------------------------

loc_F6E6E:							; CODE XREF: ROM:000F6E5Ej
		move.w	#$F0,d1					; ''
		move.w	#0,d0

loc_F6E76:							; CODE XREF: ROM:000F6E6Aj
		and.w	d1,($FFFFBE86).w
		move.w	$52(a3),d1
		asl.w	d0,d1
		or.w	d1,($FFFFBE86).w
		movem.l	(sp)+,d0-d2

locret_F6E88:							; CODE XREF: ROM:000F6E40j
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F6E8A:							; CODE XREF: sub_7CB0-4D8p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFC330).w,d1
		ext.l	d1
		move.l	#$FFFFCF36,a0
		jsr	sub_F9BE2
		move.b	8(a0),d0
		beq.w	loc_F6EB0
		and.w	#$FF,d0
		bra.w	loc_F6EB4
; ---------------------------------------------------------------------------

loc_F6EB0:							; CODE XREF: sub_F6E8A+1Aj
		move.w	#$50,d0					; 'P'

loc_F6EB4:							; CODE XREF: sub_F6E8A+22j
		move.w	d0,($FFFFC310).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F6E8A


; =============== S U B	R O U T	I N E =======================================


sub_F6EBE:							; CODE XREF: sub_15EC0+40p
		btst	#7,($FFFFC2EC).w
		bne.w	locret_F6F8A
		btst	#0,($FFFFC2FA).w
		bne.w	locret_F6F8A
		btst	#0,($FFFFC2F0).w
		bne.w	locret_F6EFA
		btst	#7,($FFFFC2F6).w
		bne.w	loc_F6F24
		tst.w	($FFFFC304).w
		beq.w	loc_F6EFC
		subq.w	#1,($FFFFC304).w
		bpl.w	locret_F6F8A
		clr.w	($FFFFC304).w

locret_F6EFA:							; CODE XREF: sub_F6EBE+1Aj
		rts
; ---------------------------------------------------------------------------

loc_F6EFC:							; CODE XREF: sub_F6EBE+2Cj
		move.w	($FFFFC310).w,d0
		move.w	d0,($FFFFC312).w
		sub.w	#$B,($FFFFC312).w
		sub.w	#$49,d0					; 'I'
		muls.w	d0,d0
		asr.l	#2,d0
		addq.w	#6,d0
		cmp.w	($FFFFB8A2).w,d0
		bgt.w	locret_F6F8A
		bset	#7,($FFFFC2F6).w
		rts
; ---------------------------------------------------------------------------

loc_F6F24:							; CODE XREF: sub_F6EBE+24j
		move.w	($FFFFC312).w,d0
		sub.w	#$41,d0					; 'A'
		muls.w	d0,d0
		asr.l	#2,d0
		cmp.w	($FFFFB8A2).w,d0
		blt.w	loc_F6F3E
		bclr	#7,($FFFFC2F6).w

loc_F6F3E:							; CODE XREF: sub_F6EBE+76j
		sub.w	d7,($FFFFC306).w
		bpl.w	locret_F6F8A
		move.w	#$14,($FFFFC306).w
		addq.w	#1,($FFFFC308).w
		move.w	($FFFFB8A2).w,d0
		ext.l	d0
		asl.w	#2,d0
		jsr	sub_110BE
		add.w	#$41,d0					; 'A'
		move.w	d0,($FFFFC30E).w
		cmp.w	($FFFFC314).w,d0
		ble.w	loc_F6F72
		move.w	d0,($FFFFC314).w

loc_F6F72:							; CODE XREF: sub_F6EBE+ACj
		move.w	($FFFFC30E).w,d0
		cmp.w	($FFFFC310).w,d0
		blt.w	loc_F6F82
		bra.w	*+4

loc_F6F82:							; CODE XREF: sub_F6EBE+BCj
		movem.l	d0-a6,-(sp)
		movem.l	(sp)+,d0-a6

locret_F6F8A:							; CODE XREF: sub_F6EBE+6j
					; sub_F6EBE+10j ...
		rts
; End of function sub_F6EBE


; =============== S U B	R O U T	I N E =======================================


sub_F6F8C:							; CODE XREF: sub_B0E8+5942p
					; sub_B0E8:loc_10A34p ...
		tst.w	$28(a3)
		bpl.w	loc_F6FA2
		add.w	#$7D0,$28(a3)
		bmi.w	loc_F6FB0
		clr.w	$28(a3)

loc_F6FA2:							; CODE XREF: sub_F6F8C+4j
		sub.w	#$7D0,$28(a3)
		bpl.w	loc_F6FB0
		clr.w	$28(a3)

loc_F6FB0:							; CODE XREF: sub_F6F8C+Ej
					; sub_F6F8C+1Cj
		tst.w	$2A(a3)
		bpl.w	loc_F6FC6
		add.w	#$7D0,$2A(a3)
		bmi.w	locret_F6FD4
		clr.w	$2A(a3)

loc_F6FC6:							; CODE XREF: sub_F6F8C+28j
		sub.w	#$7D0,$2A(a3)
		bpl.w	locret_F6FD4
		clr.w	$2A(a3)

locret_F6FD4:							; CODE XREF: sub_F6F8C+32j
					; sub_F6F8C+40j
		rts
; End of function sub_F6F8C


; =============== S U B	R O U T	I N E =======================================


sub_F6FD6:							; CODE XREF: sub_F8B5A+34p
		bclr	#4,$64(a3)
		move.w	#$88,d4					; ''
		sub.w	($FFFFBD22).w,d4
		move.w	#$12A,d5
		sub.w	($FFFFBD24).w,d5
		movem.w	d2-d5,-(sp)
		neg.w	d4
		neg.w	d5
		add.w	#$40,d4					; '@'
		add.w	#$40,d5					; '@'
		cmp.w	d5,d3
		bgt.w	loc_F7014
		cmp.w	d4,d2
		blt.w	loc_F702E
		neg.w	d4
		cmp.w	d4,d2
		bgt.w	loc_F702E
		bra.w	loc_F7062
; ---------------------------------------------------------------------------

loc_F7014:							; CODE XREF: sub_F6FD6+28j
		neg.w	d5
		cmp.w	d5,d3
		blt.w	loc_F7062
		cmp.w	d4,d2
		blt.w	loc_F702E
		neg.w	d4
		cmp.w	d4,d2
		bgt.w	loc_F702E
		bra.w	loc_F7062
; ---------------------------------------------------------------------------

loc_F702E:							; CODE XREF: sub_F6FD6+2Ej
					; sub_F6FD6+36j ...
		sub.w	d4,d2
		sub.w	d5,d3
		move.w	d3,d0
		move.w	d2,d1
		neg.w	d1
		muls.w	d3,d3
		muls.w	d2,d2
		add.l	d2,d3
		cmpnl	 $1000,d3
		bls.w	loc_F7062
		exg	d0,d3
		jsr	sub_110BE
		exg	d0,d3
		ext.l	d0
		asl.l	#8,d0
		divs.w	d3,d0
		ext.l	d1
		asl.l	#8,d1
		divs.w	d3,d1
		bsr.w	sub_F709C

loc_F7062:							; CODE XREF: sub_F6FD6+3Aj
					; sub_F6FD6+42j ...
		movem.w	(sp)+,d2-d5
		move.w	$4E(a3),d0
		or.w	$50(a3),d0
		bne.w	locret_F709A
		move.w	#$100,d0
		clr.w	d1
		cmp.w	d5,d3
		bge.w	sub_F709C
		neg.w	d5
		neg.w	d0
		cmp.w	d5,d3
		ble.w	sub_F709C
		exg	d0,d1
		cmp.w	d4,d2
		bge.w	sub_F709C
		neg.w	d4
		neg.w	d1
		cmp.w	d4,d2
		ble.w	sub_F709C

locret_F709A:							; CODE XREF: sub_F6FD6+98j
		rts
; End of function sub_F6FD6


; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_F709C:							; CODE XREF: sub_F6FD6+88p
					; sub_F6FD6+A4j ...
		jmp	sub_14BC2
; End of function sub_F709C


; =============== S U B	R O U T	I N E =======================================


sub_F70A2:							; CODE XREF: sub_7CB0-50Cp
					; sub_7CB0-500p ...
		movem.l	d0-d7,-(sp)
		move.w	#$19F,d1

loc_F70AA:							; CODE XREF: sub_F70A2+20j
		move.w	#9,d0
		jsr	sub_1107A
		move.l	a0,-(sp)
		add.l	#$1A2,a0
		move.b	d0,(a0,d1.w)
		move.l	(sp)+,a0
		dbf	d1,loc_F70AA
		movem.l	(sp)+,d0-d7
		rts
; End of function sub_F70A2

; ---------------------------------------------------------------------------
		tst.w	d3
		bpl.w	loc_F70D6
		clr.w	d3
		rts
; ---------------------------------------------------------------------------

loc_F70D6:							; CODE XREF: ROM:000F70CEj
		cmpnw	 $64,d3					; 'd'
		ble.w	locret_F70E2
		move.w	#$64,d3					; 'd'

locret_F70E2:							; CODE XREF: ROM:000F70DAj
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F70E4:							; CODE XREF: sub_15AA4+122p
					; sub_15AA4+138p ...
		movem.l	d0-d2/a1,-(sp)
		move.w	($FFFFBF14).w,d1
		move.l	#$FFFFC6CE,a1
		btst	#6,$62(a3)
		beq.w	loc_F7102
		add.l	#$364,a1

loc_F7102:							; CODE XREF: sub_F70E4+14j
		clr.w	d1
		move.b	$66(a3),d1
		asl.w	#4,d1
		add.l	#$1A2,a1
		move.b	(a1,d1.w),d1
		ext.w	d1
		ext.l	d1
		divs.w	#3,d1
		move.w	d3,-(sp)
		asl.w	#2,d3
		add.w	(sp)+,d3
		add.w	d1,d3
		bmi.w	loc_F7138
		cmpnw	 $1E,d3
		blt.w	loc_F713A
		move.w	#$1E,d3
		bra.w	loc_F713A
; ---------------------------------------------------------------------------

loc_F7138:							; CODE XREF: sub_F70E4+40j
		clr.w	d3

loc_F713A:							; CODE XREF: sub_F70E4+48j
					; sub_F70E4+50j
		and.w	#$FF,d3
		movem.l	(sp)+,d0-d2/a1
		rts
; End of function sub_F70E4


; =============== S U B	R O U T	I N E =======================================


sub_F7144:							; CODE XREF: ROM:loc_1792Ep
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFBF5E,a0
		move.w	($FFFFBF56).w,d0
		add.w	d0,d0
		move.w	(a0,d0.w),d1
		cmp.w	#0,($FFFFBF56).w
		bge.w	loc_F7166
		addq.w	#1,($FFFFBF56).w

loc_F7166:							; CODE XREF: sub_F7144+1Aj
		movem.l	(sp)+,d0/a0
		move.l	#$FFFFC6CE,a1
		rts
; End of function sub_F7144


; =============== S U B	R O U T	I N E =======================================


sub_F7172:							; CODE XREF: ROM:loc_17938p
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFBF5C,a0
		move.w	($FFFFBF54).w,d0
		add.w	d0,d0
		move.w	(a0,d0.w),d1
		cmp.w	#0,($FFFFBF54).w
		bge.w	loc_F7194
		addq.w	#1,($FFFFBF54).w

loc_F7194:							; CODE XREF: sub_F7172+1Aj
		movem.l	(sp)+,d0/a0
		move.l	#$FFFFCA32,a1
		rts
; End of function sub_F7172

; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F71A2:							; CODE XREF: ROM:000FCC7Cp
					; ROM:000FCCBAp
		clr.w	($FFFFBF54).w
		clr.w	($FFFFBF56).w
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFC6CE,a0
		bsr.w	sub_F71F0
		move.l	#$FFFFBF5E,a0
		bsr.w	sub_F72DA
		move.l	#$FFFFBF62,a0
		bsr.w	sub_F72FA
		move.l	#$FFFFCA32,a0
		bsr.w	sub_F71F0
		move.l	#$FFFFBF5C,a0
		bsr.w	sub_F72DA
		move.l	#$FFFFBF60,a0
		bsr.w	sub_F72FA
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F71A2


; =============== S U B	R O U T	I N E =======================================


sub_F71F0:							; CODE XREF: sub_F71A2+12p
					; sub_F71A2+30p ...
		movem.l	d0-a6,-(sp)
		move.l	a0,a2
		add.l	#$1A2,a2
		move.l	$1E(a0),a0
		add.w	6(a0),a0
		move.l	#$FFFFBF20,a1
		move.w	#5,d0

loc_F720E:							; CODE XREF: sub_F71F0+54j
		move.b	(a0)+,d1
		subq.b	#1,d1
		move.b	d1,(a1)+
		ext.w	d1
		asl.w	#4,d1
		clr.b	(a1)
		addq.w	#3,d1
		move.w	#3,d7

loc_F7220:							; CODE XREF: sub_F71F0+50j
		clr.w	d2
		move.b	(a2,d1.w),d2
		cmpnb	 9,d7
		beq.w	loc_F7238
		cmpnb	 $D,d7
		beq.w	loc_F7238
		add.b	d2,(a1)

loc_F7238:							; CODE XREF: sub_F71F0+3Aj
					; sub_F71F0+42j
		addq.w	#1,d1
		addq.w	#1,d7
		cmpnb	 $10,d7
		bne.s	loc_F7220
		tst.b	(a1)+
		dbf	d0,loc_F720E

loc_F7248:							; CODE XREF: sub_F71F0+84j
		move.l	#$FFFFBF20,a1
		clr.w	d1
		move.w	#4,d0

loc_F7254:							; CODE XREF: sub_F71F0+7Ej
		move.b	3(a1),d6
		cmp.b	1(a1),d6
		ble.w	loc_F726C
		st	d1
		move.w	2(a1),d2
		move.w	(a1),2(a1)
		move.w	d2,(a1)

loc_F726C:							; CODE XREF: sub_F71F0+6Cj
		tst.w	(a1)+
		dbf	d0,loc_F7254
		tst.w	d1
		bne.s	loc_F7248
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F71F0


; =============== S U B	R O U T	I N E =======================================


sub_F727C:							; CODE XREF: ROM:loc_17942p
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFBF62,a0
		move.w	($FFFFBF5A).w,d0
		add.w	d0,d0
		move.w	(a0,d0.w),d1
		cmp.w	#0,($FFFFBF5A).w
		bge.w	loc_F729E
		addq.w	#1,($FFFFBF5A).w

loc_F729E:							; CODE XREF: sub_F727C+1Aj
		movem.l	(sp)+,d0/a0
		move.l	#$FFFFC6CE,a1
		rts
; End of function sub_F727C


; =============== S U B	R O U T	I N E =======================================


sub_F72AA:							; CODE XREF: ROM:loc_1794Cp
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFBF60,a0
		move.w	($FFFFBF58).w,d0
		add.w	d0,d0
		move.w	(a0,d0.w),d1
		cmp.w	#0,($FFFFBF58).w
		bge.w	loc_F72CC
		addq.w	#1,($FFFFBF58).w

loc_F72CC:							; CODE XREF: sub_F72AA+1Aj
		movem.l	(sp)+,d0/a0
		move.l	#$FFFFCA32,a1
		rts
; End of function sub_F72AA

; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F72DA:							; CODE XREF: sub_F71A2+1Cp
					; sub_F71A2+3Ap
		movem.l	d0-d1/a0-a1,-(sp)
		move.l	#$FFFFBF20,a1
		move.w	#0,d0

loc_F72E8:							; CODE XREF: sub_F72DA+16j
		clr.b	(a0)+
		move.b	(a1),d1
		move.b	d1,(a0)+
		tst.w	(a1)+
		dbf	d0,loc_F72E8
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_F72DA


; =============== S U B	R O U T	I N E =======================================


sub_F72FA:							; CODE XREF: sub_F71A2+26p
					; sub_F71A2+44p
		movem.l	d0/a0-a1,-(sp)
		move.l	#$FFFFBF2A,a1
		move.w	#0,d0

loc_F7308:							; CODE XREF: sub_F72FA+14j
		clr.b	(a0)+
		move.b	(a1),(a0)+
		tst.w	-(a1)
		dbf	d0,loc_F7308
		movem.l	(sp)+,d0/a0-a1
		rts
; End of function sub_F72FA


; =============== S U B	R O U T	I N E =======================================


sub_F7318:							; CODE XREF: ROM:000FCE72p
		movem.l	d1-a6,-(sp)
		move.l	#$FFFFC6CE,a0
		bsr.w	sub_F737E
		move.w	d1,($FFFFBF12).w
		move.l	#$FFFFCA32,a0
		bsr.w	sub_F737E
		move.w	d1,($FFFFBF14).w
		move.w	($FFFFBF12).w,d1
		move.w	($FFFFBF14).w,d2
		sub.w	d1,d2
		move.w	#0,($FFFFBF50).w
		tst.w	d2
		bmi.w	loc_F7354
		move.w	#1,($FFFFBF50).w

loc_F7354:							; CODE XREF: sub_F7318+32j
		tst.w	d2
		bpl.w	loc_F735C
		neg.w	d2

loc_F735C:							; CODE XREF: sub_F7318+3Ej
		move.w	#$FFFF,d0
		cmpnw	 $5E,d2					; '^'
		blt.w	loc_F7378
		move.w	#$22,d0					; '"'
		cmpnw	 $BD,d2					; ''
		blt.w	loc_F7378
		move.w	#$23,d0					; '#'

loc_F7378:							; CODE XREF: sub_F7318+4Cj
					; sub_F7318+58j
		movem.l	(sp)+,d1-a6
		rts
; End of function sub_F7318


; =============== S U B	R O U T	I N E =======================================


sub_F737E:							; CODE XREF: sub_F7318+Ap
					; sub_F7318+18p
		bsr.w	sub_F71F0
		move.w	#6,d0
		move.l	#$FFFFBF20,a0
		clr.w	d1

loc_F738E:							; CODE XREF: sub_F737E+1Aj
		move.b	1(a0),d2
		ext.w	d2
		add.w	d2,d1
		tst.w	(a0)+
		dbf	d0,loc_F738E
		rts
; End of function sub_F737E


; =============== S U B	R O U T	I N E =======================================


sub_F739E:							; CODE XREF: sub_7CB0+F662p

; FUNCTION CHUNK AT 000F8418 SIZE 0000008A BYTES

		jsr	sub_FE660
		move.l	#$FFFFD088,a3
		jsr	sub_1803E
		move.w	($FFFFD072).w,($FFFFD058).w
		move.w	($FFFFD074).w,($FFFFD048).w
		bclr	#1,($FFFFC2FC).w
		jsr	sub_FD73C
		jsr	sub_F9C68
		bsr.w	loc_F8168
		clr.w	($FFFFD422).w
		clr.w	($FFFFD424).w
		bsr.w	sub_F8070
		tst.w	($FFFFD064).w
		bne.w	loc_F73FE
		moveq	#$1C,d0
		jsr	sub_11086
		move.w	d0,($FFFFD04C).w
		moveq	#$1C,d0
		jsr	sub_11086
		move.w	d0,($FFFFD04E).w

loc_F73FE:							; CODE XREF: sub_F739E+44j
		st	($FFFFD064).w
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F7410
		bra.w	loc_F742C
; ---------------------------------------------------------------------------

loc_F7410:							; CODE XREF: sub_F739E+6Aj
		cmp.w	#2,($FFFFD048).w
		blt.w	loc_F741E
		bsr.w	sub_F76C8

loc_F741E:							; CODE XREF: sub_F739E+78j
		cmp.w	#1,($FFFFD048).w
		bne.w	loc_F742C
		bsr.w	sub_F76CE

loc_F742C:							; CODE XREF: sub_F739E+6Ej
					; sub_F739E+86j
		clr.w	d7
		clr.w	d0
		bsr.w	sub_F76D4
		bsr.w	sub_F78A2
		move.w	($FFFFC330).w,($FFFFD42A).w
		move.w	($FFFFC330).w,($FFFFD428).w
		bsr.w	sub_F85B0
		move.w	($FFFFC332).w,($FFFFD42C).w
		move.w	($FFFFC332).w,($FFFFD428).w
		bsr.w	sub_F8608
		move.w	#$18,($FFFFBD26).w
		bclr	#2,($FFFFBF78).w

loc_F7464:							; CODE XREF: sub_F739E+18Aj
					; sub_F739E+20Cj ...
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F7480
		move.w	#1,($FFFFD058).w
		move.w	#0,($FFFFD056).w
		move.w	#1,($FFFFD054).w

loc_F7480:							; CODE XREF: sub_F739E+CCj
		bsr.w	sub_F8304
		tst.w	d1
		bne.w	loc_F748E
		bra.w	loc_F8418
; ---------------------------------------------------------------------------

loc_F748E:							; CODE XREF: sub_F739E+E8j
		btst	#7,d1
		bne.w	loc_F843E
		btst	#1,d1
		beq.w	loc_F752C
		tst.w	($FFFFD04A).w
		bne.w	loc_F74B6
		cmpnw	 8,d7
		bne.w	loc_F74B6
		move.w	#4,d0
		bra.w	loc_F74BA
; ---------------------------------------------------------------------------

loc_F74B6:							; CODE XREF: sub_F739E+104j
					; sub_F739E+10Cj
		move.w	#2,d0

loc_F74BA:							; CODE XREF: sub_F739E+114j
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F74DC
		tst.w	($FFFFD04A).w
		bne.w	loc_F74DC
		cmpnw	 8,d7
		bne.w	loc_F74DC
		move.w	#6,d0
		bra.w	loc_F7520
; ---------------------------------------------------------------------------

loc_F74DC:							; CODE XREF: sub_F739E+122j
					; sub_F739E+12Aj ...
		cmpnw	 6,d7
		bne.w	loc_F74FC
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F74FC
		tst.w	($FFFFD04A).w
		bne.w	loc_F751C
		clr.w	d0
		bra.w	loc_F7520
; ---------------------------------------------------------------------------

loc_F74FC:							; CODE XREF: sub_F739E+142j
					; sub_F739E+14Cj
		cmpnw	 $A,d7
		bne.w	loc_F7520
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F7514
		clr.w	d0
		bra.w	loc_F7520
; ---------------------------------------------------------------------------

loc_F7514:							; CODE XREF: sub_F739E+16Cj
		tst.w	($FFFFD458).w
		bpl.w	loc_F7520

loc_F751C:							; CODE XREF: sub_F739E+154j
		move.w	#4,d0

loc_F7520:							; CODE XREF: sub_F739E+13Aj
					; sub_F739E+15Aj ...
		bsr.w	sub_F76D4
		bsr.w	sub_F78A2
		bra.w	loc_F7464
; ---------------------------------------------------------------------------

loc_F752C:							; CODE XREF: sub_F739E+FCj
		btst	#0,d1
		beq.w	loc_F75AE
		tst.w	($FFFFD04A).w
		bne.w	loc_F754C
		cmpnw	 $C,d7
		bne.w	loc_F754C
		move.w	#$FFFC,d0
		bra.w	loc_F7572
; ---------------------------------------------------------------------------

loc_F754C:							; CODE XREF: sub_F739E+19Aj
					; sub_F739E+1A2j
		move.w	#$FFFE,d0
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F7572
		tst.w	($FFFFD04A).w
		bne.w	loc_F7572
		cmpnw	 $E,d7
		bne.w	loc_F7572
		move.w	#$FFFA,d0
		bra.w	loc_F75A2
; ---------------------------------------------------------------------------

loc_F7572:							; CODE XREF: sub_F739E+1AAj
					; sub_F739E+1B8j ...
		cmpnw	 $A,d7
		bne.w	loc_F7584
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F759E

loc_F7584:							; CODE XREF: sub_F739E+1D8j
		cmpnw	 $E,d7
		bne.w	loc_F75A2
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F759E
		tst.w	($FFFFD458).w
		bpl.w	loc_F75A2

loc_F759E:							; CODE XREF: sub_F739E+1E2j
					; sub_F739E+1F4j
		move.w	#$FFFC,d0

loc_F75A2:							; CODE XREF: sub_F739E+1D0j
					; sub_F739E+1EAj ...
		bsr.w	sub_F76D4
		bsr.w	sub_F78A2
		bra.w	loc_F7464
; ---------------------------------------------------------------------------

loc_F75AE:							; CODE XREF: sub_F739E+192j
		moveq	#1,d2
		btst	#3,d1
		bne.w	loc_F75C2
		btst	#2,d1
		beq.w	loc_F7464
		moveq	#$FFFFFFFF,d2

loc_F75C2:							; CODE XREF: sub_F739E+216j
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F75D2
		jsr	sub_17E42

loc_F75D2:							; CODE XREF: sub_F739E+22Aj
		move.w	($FFFFD04E).w,-(sp)
		move.w	($FFFFD04C).w,-(sp)
		bsr.w	sub_F7942
		move.w	(sp)+,d0
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F75F2
		tst.w	($FFFFD048).w
		bne.w	loc_F75FE

loc_F75F2:							; CODE XREF: sub_F739E+248j
		move.w	($FFFFD04C).w,($FFFFC330).w
		move.w	($FFFFD04E).w,($FFFFC332).w

loc_F75FE:							; CODE XREF: sub_F739E+250j
		move.w	($FFFFC330).w,($FFFFD428).w
		btst	#6,($FFFFD42E).w
		beq.w	loc_F7622
		movem.w	d0,-(sp)
		move.w	($FFFFD42A).w,d0
		cmp.w	($FFFFC330).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_F762C

loc_F7622:							; CODE XREF: sub_F739E+26Cj
		move.w	($FFFFC330).w,($FFFFD42A).w
		bsr.w	sub_F85B0

loc_F762C:							; CODE XREF: sub_F739E+280j
		btst	#7,($FFFFD42E).w
		beq.w	loc_F764A
		movem.w	d0,-(sp)
		move.w	($FFFFD42C).w,d0
		cmp.w	($FFFFC332).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_F7660

loc_F764A:							; CODE XREF: sub_F739E+294j
		move.w	($FFFFC332).w,($FFFFD42C).w
		move.w	($FFFFC332).w,($FFFFD428).w
		bsr.w	sub_F8608
		move.w	($FFFFC330).w,($FFFFD428).w

loc_F7660:							; CODE XREF: sub_F739E+2A8j
		tst.w	(sp)+
		bra.w	loc_F76BE
; ---------------------------------------------------------------------------
		move.w	(sp)+,d0
		cmp.w	($FFFFC332).w,d0
		beq.w	loc_F76BE
		move.w	($FFFFC332).w,($FFFFD428).w
		movem.w	d0,-(sp)
		move.w	($FFFFC332).w,d0
		cmp.w	($FFFFC332).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_F7694
		move.w	($FFFFC332).w,($FFFFD42C).w
		bsr.w	sub_F8608

loc_F7694:							; CODE XREF: sub_F739E+2E8j
		movem.w	d0,-(sp)
		move.w	($FFFFD42A).w,d0
		cmp.w	($FFFFC330).w,d0
		movem.w	(sp)+,d0
		beq.w	loc_F76BE
		move.w	($FFFFC330).w,($FFFFD42A).w
		move.w	($FFFFC330).w,($FFFFD428).w
		bsr.w	sub_F85B0
		move.w	($FFFFC332).w,($FFFFD428).w

loc_F76BE:							; CODE XREF: sub_F739E+2C4j
					; sub_F739E+2CEj ...
		bsr.w	sub_F78A2
		bra.w	loc_F7464
; End of function sub_F739E

; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_F76C8:							; CODE XREF: sub_F739E+7Cp
					; sub_F7942+1BAj ...
		jmp	sub_17D16
; End of function sub_F76C8


; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_F76CE:							; CODE XREF: sub_F739E+8Ap
		jmp	sub_17CA0
; End of function sub_F76CE


; =============== S U B	R O U T	I N E =======================================


sub_F76D4:							; CODE XREF: sub_F739E+92p
					; sub_F739E:loc_F7520p	...
		move.w	d7,d3

loc_F76D6:							; CODE XREF: sub_F76D4+C2j
					; sub_F76D4+DEj
		bclr	#2,($FFFFC2F6).w
		tst.w	d0
		bpl.w	loc_F770A
		move.w	d0,-(sp)
		move.w	d7,d0
		asr.w	#1,d0
		cmp.w	($FFFFD422).w,d0
		bne.w	loc_F7708
		move.w	($FFFFD422).w,($FFFFD424).w
		bset	#2,($FFFFC2F6).w
		subq.w	#1,($FFFFD422).w
		bpl.w	loc_F7708
		clr.w	($FFFFD422).w

loc_F7708:							; CODE XREF: sub_F76D4+18j
					; sub_F76D4+2Cj
		move.w	(sp)+,d0

loc_F770A:							; CODE XREF: sub_F76D4+Aj
		move.w	d7,($FFFFBF12).w
		add.w	d0,d7
		bpl.w	loc_F7716
		clr.w	d7

loc_F7716:							; CODE XREF: sub_F76D4+3Cj
		tst.w	d0
		bmi.w	loc_F7768
		move.w	d0,-(sp)
		move.w	d7,d0
		asr.w	#1,d0
		sub.w	($FFFFD422).w,d0
		cmpnw	 5,d0
		bgt.w	loc_F7732
		bra.w	loc_F7766
; ---------------------------------------------------------------------------

loc_F7732:							; CODE XREF: sub_F76D4+56j
		move.w	($FFFFD422).w,($FFFFD424).w
		bset	#2,($FFFFC2F6).w
		addq.w	#1,($FFFFD422).w
		move.w	d7,d0
		asr.w	#1,d0
		sub.w	($FFFFD422).w,d0
		cmpnw	 5,d0
		ble.w	loc_F7756
		addq.w	#1,($FFFFD422).w

loc_F7756:							; CODE XREF: sub_F76D4+7Aj
		cmp.w	#3,($FFFFD422).w
		ble.w	loc_F7766
		move.w	#3,($FFFFD422).w

loc_F7766:							; CODE XREF: sub_F76D4+5Aj
					; sub_F76D4+88j
		move.w	(sp)+,d0

loc_F7768:							; CODE XREF: sub_F76D4+44j
		btst	#2,($FFFFC2F6).w
		bne.w	loc_F7778
		move.w	($FFFFD422).w,($FFFFD424).w

loc_F7778:							; CODE XREF: sub_F76D4+9Aj
		cmpnw	 $12,d7
		blt.w	loc_F7782
		sub.w	d0,d7

loc_F7782:							; CODE XREF: sub_F76D4+A8j
		cmp.w	#1,($FFFFD048).w
		bne.w	loc_F779A
		tst.w	d7
		beq.w	loc_F779A
		cmpnw	 6,d7
		bls.w	loc_F76D6

loc_F779A:							; CODE XREF: sub_F76D4+B4j
					; sub_F76D4+BAj
		cmp.w	#2,($FFFFD048).w
		blt.w	loc_F77B6
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F77B6
		cmpnw	 6,d7
		beq.w	loc_F76D6

loc_F77B6:							; CODE XREF: sub_F76D4+CCj
					; sub_F76D4+D6j
		bsr.w	loc_F780C
		movem.l	d0-a6,-(sp)
		move.w	($FFFFD426).w,d4
		move.l	#off_BEFB8,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		move.w	($FFFFB028).w,d0
		subq.w	#1,d0
		move.w	d1,d3
		move.w	($FFFFB02A).w,d1
		sub.w	#$E,d1
		moveq	#$17,d2
		moveq	#0,d5
		bset	#0,($FFFFC2F8).w
		jsr	sub_1169A
		bclr	#0,($FFFFC2F8).w
		movem.l	(sp)+,d0-a6
		move.w	d7,d3
		bsr.w	loc_F7864
		move.w	#$6000,($FFFFB02C).w
		jmp	sub_119B8
; End of function sub_F76D4

; ---------------------------------------------------------------------------

loc_F780C:							; CODE XREF: sub_F76D4:loc_F77B6p
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$10
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$17,d0
		move.w	#$D,d1
		rts
; ---------------------------------------------------------------------------
		add.w	d3,($FFFFB02A).w
		move.w	($FFFFD424).w,d0
		add.w	d0,d0
		sub.w	d0,($FFFFB02A).w
		cmp.w	#$18,($FFFFB02A).w
		ble.w	loc_F7840
		move.w	#$18,($FFFFB02A).w

loc_F7840:							; CODE XREF: ROM:000F7836j
		moveq	#$17,d0
		moveq	#3,d1
		cmp.w	#2,($FFFFD048).w
		blt.w	locret_F78A0
		cmp.w	#4,($FFFFD048).w
		beq.w	locret_F78A0
		cmpnw	 4,d3
		bne.w	locret_F78A0
		moveq	#5,d1
		rts
; ---------------------------------------------------------------------------

loc_F7864:							; CODE XREF: sub_F76D4+128p
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$10
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	d3,($FFFFB02A).w
		move.w	($FFFFD422).w,d0
		add.w	d0,d0
		sub.w	d0,($FFFFB02A).w
		moveq	#$17,d0
		moveq	#3,d1
		cmp.w	#2,($FFFFD048).w
		blt.w	locret_F78A0
		cmp.w	#4,($FFFFD048).w
		beq.w	locret_F78A0
		cmpnw	 4,d3
		bne.w	locret_F78A0
		moveq	#5,d1
; START	OF FUNCTION CHUNK FOR sub_F7942

locret_F78A0:							; CODE XREF: ROM:000F784Aj
					; ROM:000F7854j ...
		rts
; END OF FUNCTION CHUNK	FOR sub_F7942

; =============== S U B	R O U T	I N E =======================================


sub_F78A2:							; CODE XREF: sub_F739E+96p
					; sub_F739E+186p ...
		movem.l	d0-a6,-(sp)
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F78B6
		move.w	#1,($FFFFD054).w

loc_F78B6:							; CODE XREF: sub_F78A2+Aj
		tst.w	($FFFFD04A).w
		bne.w	loc_F78C4
		move.w	#1,($FFFFD054).w

loc_F78C4:							; CODE XREF: sub_F78A2+18j
		tst.w	($FFFFD04A).w
		bne.w	loc_F78F8
		move.w	#0,($FFFFD052).w
		move.w	#0,($FFFFD05A).w
		move.w	#0,($FFFFD05C).w
		tst.w	($FFFFD04A).w
		bne.w	loc_F78F8
		move.w	#1,($FFFFD052).w
		move.w	#1,($FFFFD05A).w
		move.w	#1,($FFFFD05C).w

loc_F78F8:							; CODE XREF: sub_F78A2+26j
					; sub_F78A2+40j
		move.w	($FFFFD422).w,d0
		cmp.w	($FFFFD424).w,d0
		beq.w	loc_F7908
		bsr.w	sub_F8070

loc_F7908:							; CODE XREF: sub_F78A2+5Ej
		bset	#3,($FFFFC2F8).w
		bsr.w	sub_F7B20
		bclr	#3,($FFFFC2F8).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F78A2

; ---------------------------------------------------------------------------
word_F791E:	dc.w	5					; DATA XREF: sub_F7942+68o
		dc.w	8
		dc.w	$1C
		dc.w	$1C
		dc.w	3
		dc.w	2
		dc.w	2
		dc.w	3
		dc.w	3
word_F7930:	dc.w	5					; DATA XREF: sub_F7942+76o
		dc.w	$C
		dc.w	$1C
		dc.w	$1C
		dc.w	3
		dc.w	2
		dc.w	2
		dc.w	3
		dc.w	3

; =============== S U B	R O U T	I N E =======================================


sub_F7942:							; CODE XREF: sub_F739E+23Cp
					; sub_F7B20+8p

; FUNCTION CHUNK AT 000F78A0 SIZE 00000002 BYTES

		tst.w	d7
		bne.w	loc_F79A6
		movem.w	d0/d2,-(sp)
		tst.w	d2
		beq.w	loc_F797A
		move.w	($FFFFD048).w,d0
		add.w	d2,d0
		bpl.w	loc_F7960
		move.w	#4,d0

loc_F7960:							; CODE XREF: sub_F7942+16j
		cmpnw	 4,d0
		bne.w	loc_F797A
		move.w	($FFFFD054).w,($FFFFD05E).w
		move.w	($FFFFD056).w,($FFFFD060).w
		move.w	($FFFFD058).w,($FFFFD062).w

loc_F797A:							; CODE XREF: sub_F7942+Cj
					; sub_F7942+22j
		movem.w	(sp)+,d0/d2
		tst.w	d7
		bne.w	loc_F79A6
		tst.w	d2
		beq.w	loc_F79A6
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F79A6
		move.w	($FFFFD05E).w,($FFFFD054).w
		move.w	($FFFFD060).w,($FFFFD056).w
		move.w	($FFFFD062).w,($FFFFD058).w

loc_F79A6:							; CODE XREF: sub_F7942+2j
					; sub_F7942+3Ej ...
		move.w	#$D048,a2
		move.l	#word_F791E,a3
		tst.w	($FFFFD046).w
		beq.w	loc_F79BE
		move.l	#word_F7930,a3

loc_F79BE:							; CODE XREF: sub_F7942+72j
		clr.w	d5
		move.w	(a3,d7.w),d4
		move.w	(a2,d7.w),d3
		cmp.w	#2,($FFFFD048).w
		blt.w	loc_F79E6
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F79E6
		cmpnw	 4,d7
		bne.w	loc_F79E6
		moveq	#$1A,d4

loc_F79E6:							; CODE XREF: sub_F7942+8Cj
					; sub_F7942+96j ...
		cmpnw	 2,d7
		bne.w	loc_F7A78
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F7A48
		movem.w	d2,-(sp)
		add.w	($FFFFD04A).w,d2
		bpl.w	loc_F7A06
		addq.w	#7,d2

loc_F7A06:							; CODE XREF: sub_F7942+BEj
		cmpnw	 4,d2
		movem.w	(sp)+,d2
		ble.w	loc_F7A26
		tst.w	d2
		bpl.w	loc_F7A20
		move.w	#4,d2
		bra.w	loc_F7A48
; ---------------------------------------------------------------------------

loc_F7A20:							; CODE XREF: sub_F7942+D2j
		addq.w	#2,d2
		bra.w	loc_F7A48
; ---------------------------------------------------------------------------

loc_F7A26:							; CODE XREF: sub_F7942+CCj
		movem.w	d2,-(sp)
		add.w	($FFFFD04A).w,d2
		cmpnw	 3,d2
		movem.w	(sp)+,d2
		bne.w	loc_F7A48
		tst.w	d2
		bpl.w	loc_F7A46
		subq.w	#1,d2
		bra.w	loc_F7A48
; ---------------------------------------------------------------------------

loc_F7A46:							; CODE XREF: sub_F7942+FAj
		addq.w	#1,d2

loc_F7A48:							; CODE XREF: sub_F7942+B2j
					; sub_F7942+DAj ...
		moveq	#5,d5
		tst.w	($FFFFD046).w
		beq.w	loc_F7A58
		move.w	#$A,d4
		addq.w	#2,d5

loc_F7A58:							; CODE XREF: sub_F7942+10Cj
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F7A6A
		tst.w	($FFFFD048).w
		bne.w	loc_F7A6E

loc_F7A6A:							; CODE XREF: sub_F7942+11Cj
		clr.w	d5
		moveq	#5,d4

loc_F7A6E:							; CODE XREF: sub_F7942+124j
					; sub_F7942+1A0j
		tst.w	($FFFFD046).w
		beq.w	loc_F7A78
		addq.w	#2,d4

loc_F7A78:							; CODE XREF: sub_F7942+A8j
					; sub_F7942+130j
		bsr.w	sub_F7B0A
		tst.w	($FFFFD046).w
		beq.w	loc_F7ABA
		cmpnw	 2,d7
		bne.w	loc_F7ABA
		tst.w	d2
		bmi.w	loc_F7AA6
		cmpnw	 9,d3
		beq.w	loc_F7AB6
		cmpnw	 $A,d3
		beq.w	loc_F7AB6
		bra.w	loc_F7ABA
; ---------------------------------------------------------------------------

loc_F7AA6:							; CODE XREF: sub_F7942+14Cj
		cmpnw	 $A,d3
		bne.w	loc_F7ABA
		move.w	#8,d3
		bra.w	loc_F7ABA
; ---------------------------------------------------------------------------

loc_F7AB6:							; CODE XREF: sub_F7942+154j
					; sub_F7942+15Cj
		move.w	#$B,d3

loc_F7ABA:							; CODE XREF: sub_F7942+13Ej
					; sub_F7942+146j ...
		move.w	d3,(a2,d7.w)
		tst.w	d2
		beq.w	locret_F78A0
		cmp.w	#1,($FFFFD048).w
		bne.w	loc_F7AE6
		tst.w	d7
		bne.w	loc_F7AE6
		jsr	sub_17CA0
		move.w	($FFFFCEEC).w,d0
		or.w	($FFFFCEEA).w,d0
		beq.s	loc_F7A6E
		rts
; ---------------------------------------------------------------------------

loc_F7AE6:							; CODE XREF: sub_F7942+188j
					; sub_F7942+18Ej
		cmp.w	#2,($FFFFD048).w
		blt.w	locret_F78A0
		cmp.w	#4,($FFFFD048).w
		beq.w	locret_F78A0
		tst.w	d7
		beq.w	sub_F76C8
		cmpnw	 4,d7
		beq.w	sub_F76C8
		rts
; End of function sub_F7942


; =============== S U B	R O U T	I N E =======================================


sub_F7B0A:							; CODE XREF: sub_F7942:loc_F7A78p
		add.w	d2,d3
		cmp.w	d5,d3
		bge.w	loc_F7B16
		move.w	d4,d3
		subq.w	#1,d3

loc_F7B16:							; CODE XREF: sub_F7B0A+4j
		cmp.w	d4,d3
		blt.w	locret_F78A0
		move.w	d5,d3
		rts
; End of function sub_F7B0A


; =============== S U B	R O U T	I N E =======================================


sub_F7B20:							; CODE XREF: sub_F78A2+6Cp
		move.w	d7,d6
		moveq	#$FFFFFFFE,d7

loc_F7B24:							; CODE XREF: sub_F7B20+26j
		addq.w	#2,d7
		clr.w	d2
		bsr.w	sub_F7942
		movem.l	d7,-(sp)
		asr.w	#1,d7
		cmp.w	($FFFFD422).w,d7
		movem.l	(sp)+,d7
		blt.w	loc_F7B42
		bsr.w	sub_F7B66

loc_F7B42:							; CODE XREF: sub_F7B20+1Aj
		cmpnw	 $10,d7
		bne.s	loc_F7B24
		bsr.w	sub_F7BFA
		move.w	d6,d7
		jsr	sub_17AF4
		btst	#0,($FFFFD42E).w
		bne.w	locret_F7B64
		jsr	sub_F87D2

locret_F7B64:							; CODE XREF: sub_F7B20+3Aj
		rts
; End of function sub_F7B20


; =============== S U B	R O U T	I N E =======================================


sub_F7B66:							; CODE XREF: sub_F7B20+1Ep

; FUNCTION CHUNK AT 000F7C5E SIZE 00000054 BYTES

		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$11
		dc.b	$10
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD422).w,d0
		add.w	d0,d0
		sub.w	d0,($FFFFB02A).w
		add.w	d7,($FFFFB02A).w
		cmp.w	#$1B,($FFFFB02A).w
		bgt.w	locret_F78A0
		subq.w	#1,($FFFFB02A).w
		move.w	#$11,($FFFFB028).w
		cmpnw	 4,d7
		beq.w	loc_F7C5E
		cmpnw	 6,d7
		beq.w	loc_F7C5E
		move.l	#$FFFFD048,a0
		move.w	(a0,d7.w),d3
		move.l	#word_F7D04,a0
		add.w	(a0,d7.w),a0
		cmp.l	#word_F7F3C,a0
		bne.w	loc_F7BD8
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F7BD8
		move.l	#word_F7CEE,a1
		jmp	sub_11BA4
; ---------------------------------------------------------------------------

loc_F7BD8:							; CODE XREF: sub_F7B66+58j
					; sub_F7B66+62j
		tst.w	($FFFFD046).w
		beq.w	loc_F7BEC
		cmpnw	 2,d7
		bne.w	loc_F7BEC
		add.w	#$B0,a0					; ''

loc_F7BEC:							; CODE XREF: sub_F7B66+76j
					; sub_F7B66+7Ej ...
		move.l	a0,a1
		add.w	(a0),a0
		dbf	d3,loc_F7BEC
		jmp	sub_11BA4
; End of function sub_F7B66


; =============== S U B	R O U T	I N E =======================================


sub_F7BFA:							; CODE XREF: sub_F7B20+28p
		move.w	#$19,($FFFFB02A).w
		move.l	#off_F7C5A,a1
		cmp.w	#3,($FFFFD422).w
		beq.w	loc_F7C20
		cmp.w	#4,($FFFFD048).w
		beq.w	loc_F7C20
		move.l	#off_F7C56,a1

loc_F7C20:							; CODE XREF: sub_F7BFA+12j
					; sub_F7BFA+1Cj
		move.w	#2,($FFFFB028).w
		jsr	sub_11BA4
		move.l	#off_F7C5A,a1
		move.w	#$F,($FFFFB02A).w
		tst.w	($FFFFD422).w
		beq.w	loc_F7C46
		move.l	#off_F7C52,a1

loc_F7C46:							; CODE XREF: sub_F7BFA+42j
		move.w	#2,($FFFFB028).w
		jmp	sub_11BA4
; End of function sub_F7BFA

; ---------------------------------------------------------------------------
off_F7C52:	dc.l	unk_47B00				; DATA XREF: sub_F7BFA+46o
off_F7C56:	dc.l	unk_47D00				; DATA XREF: sub_F7BFA+20o
off_F7C5A:	dc.l	unk_42000				; DATA XREF: sub_F7BFA+6o
					; sub_F7BFA+32o
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_F7B66

loc_F7C5E:							; CODE XREF: sub_F7B66+32j
					; sub_F7B66+3Aj
		move.w	($FFFFB028).w,-(sp)
		move.l	#word_F7CD8,a1
		jsr	sub_11BA4
		move.w	(sp)+,($FFFFB028).w
		move.l	#$FFFFD048,a1
		move.w	(a1,d7.w),d3
		move.w	#$30E,a1
		asl.w	#2,d3
		move.l	(a1,d3.w),a1
		lsr.w	#2,d3
		add.w	4(a1),a1
		cmpnw	 $E,d3
		bne.w	loc_F7C9E
		move.l	#word_F7CC6,a1
		bra.w	loc_F7CAC
; ---------------------------------------------------------------------------

loc_F7C9E:							; CODE XREF: sub_F7B66+12Aj
		cmpnw	 $D,d3
		bne.w	loc_F7CAC
		move.l	#word_F7CB2,a1

loc_F7CAC:							; CODE XREF: sub_F7B66+134j
					; sub_F7B66+13Cj
		jmp	sub_11BA4
; END OF FUNCTION CHUNK	FOR sub_F7B66
; ---------------------------------------------------------------------------
word_F7CB2:	dc.w	$14					; DATA XREF: sub_F7B66+140o
aNewYorkIslanders: dc.b	'New York Islanders'
word_F7CC6:	dc.w	$12					; DATA XREF: sub_F7B66+12Eo
aNewYorkRangers: dc.b	'New York Rangers'
word_F7CD8:	dc.w	$16					; DATA XREF: sub_F7B66+FCo
		dc.b	'                    '
word_F7CEE:	dc.w	$16					; DATA XREF: sub_F7B66+66o
aNA:		dc.b	'N/A                 '
word_F7D04:	dc.w	$12					; DATA XREF: sub_F7B66+48o
		dc.w	$80
		dc.w	$2FE
		dc.w	$2FE
		dc.w	$238
		dc.w	$290
		dc.w	$2FE
		dc.w	$2BC
		dc.w	$32A
		dc.w	$16
aRegularSeason:	dc.b	'Regular Season      '
		dc.w	$16
aContinuePlayoffs: dc.b	'Continue Playoffs   '
		dc.w	$16
aNewPlayoffs:	dc.b	'New Playoffs        '
		dc.w	$16
aNewPlayoffs7Game: dc.b	'New Playoffs/7 game '
		dc.w	$16
aShootout:	dc.b	'Shootout            '
		dc.w	$16
aDemo:		dc.b	'Demo                '
		dc.w	$16
aOneHome:	dc.b	'One - Home          '
		dc.w	$16
aOneVisitor:	dc.b	'One - Visitor       '
		dc.w	$16
aTwoTeammates:	dc.b	'Two - Teammates     '
		dc.w	$16
aTwoHeadToHead:	dc.b	'Two - Head to Head  '
		dc.w	$16
aTwoHeadToHead_0: dc.b	'Two - Head to Head  '
		dc.w	$16
aTwoTeammates_0: dc.b	'Two - Teammates     '
		dc.w	$16
aOne:		dc.b	'One                 '
		dc.w	$16
aDemo_0:	dc.b	'Demo                '
		dc.w	$16
aOneHome_0:	dc.b	'One - Home          '
		dc.w	$16
aOneVisitor_0:	dc.b	'One - Visitor       '
		dc.w	$16
aTwoTeammates_1: dc.b	'Two - Teammates     '
		dc.w	$16
aTwoHeadToHead_1: dc.b	'Two - Head to Head  '
		dc.w	$16
aThree:		dc.b	'Three               '
		dc.w	$16
aFour:		dc.b	'Four                '
		dc.w	$16
aTwoHeadToHead_2: dc.b	'Two - Head to Head  '
		dc.w	$16
aTwoTeammates_2: dc.b	'Two - Teammates     '
		dc.w	$16
aThree_0:	dc.b	'Three               '
		dc.w	$16
aFour_0:	dc.b	'Four                '
		dc.w	$16
aOne_0:		dc.b	'One                 '
word_F7F3C:	dc.w	$16					; DATA XREF: sub_F7B66+52o
a5Minutes:	dc.b	'5 Minutes           '
		dc.w	$16
a10Minutes:	dc.b	'10 Minutes          '
		dc.w	$16
a20Minutes:	dc.b	'20 Minutes          '
		dc.w	$16
a30Seconds:	dc.b	'30 Seconds          '
		dc.w	$16
aManualControl:	dc.b	'Manual Control      '
		dc.w	$16
aAutoControl:	dc.b	'Auto Control        '
		dc.w	$16
aOff:		dc.b	'Off                 '
		dc.w	$16
aOn:		dc.b	'On                  '
		dc.w	$16
aOnExceptOffSides: dc.b	'On, Except Off-sides'
		dc.w	$16
aOn_0:		dc.b	'On                  '
		dc.w	$16
aOff_0:		dc.b	'Off                 '
		dc.w	$16
aOn_1:		dc.b	'On                  '
		dc.w	$16
aOff_1:		dc.b	'Off                 '
		dc.w	$16
aAuto:		dc.b	'Auto                '

; =============== S U B	R O U T	I N E =======================================


sub_F8070:							; CODE XREF: sub_F739E+3Cp
					; sub_F78A2+62p
		movem.l	d0-d1,-(sp)
		bset	#3,($FFFFC2F8).w
		move.l	#word_F80D4,a1
		move.w	#$F,d0
		move.w	($FFFFD422).w,d1
		add.w	d1,d1
		sub.w	d1,d0
		move.w	d0,($FFFFB02A).w

loc_F8090:							; CODE XREF: sub_F8070+56j
		cmp.w	#$F,($FFFFB02A).w
		bge.w	loc_F80A0
		add.w	(a1),a1
		bra.w	loc_F80C2
; ---------------------------------------------------------------------------

loc_F80A0:							; CODE XREF: sub_F8070+26j
		cmp.w	#$19,($FFFFB02A).w
		bgt.w	loc_F80C8
		move.w	#3,($FFFFB028).w
		move.w	#0,($FFFFB02C).w
		move.w	#0,($FFFFB02E).w
		jsr	sub_11A48

loc_F80C2:							; CODE XREF: sub_F8070+2Cj
		addq.w	#2,($FFFFB02A).w
		bra.s	loc_F8090
; ---------------------------------------------------------------------------

loc_F80C8:							; CODE XREF: sub_F8070+36j
		bclr	#3,($FFFFC2F8).w
		movem.l	(sp)+,d0-d1
		rts
; End of function sub_F8070

; ---------------------------------------------------------------------------
word_F80D4:	dc.w	$10					; DATA XREF: sub_F8070+Ao
aPlayMode:	dc.b	'Play Mode    ',0
		dc.w	$10
aPlayers:	dc.b	'Players      ',0
		dc.w	$10
aTeam1:		dc.b	'Team 1       ',0
		dc.w	$10
aTeam2:		dc.b	'Team 2       ',0
		dc.w	$10
aPer_Length:	dc.b	'Per. Length  ',0
		dc.w	$10
aGoalies_0:	dc.b	'Goalies      ',0
		dc.w	$10
aUserRecords:	dc.b	'User Records ',0
		dc.w	$10
aPenalties:	dc.b	'Penalties    ',0
		dc.w	$10
aLineChanges:	dc.b	'Line Changes ',0
		dc.w	4
		dc.b	0,0
; ---------------------------------------------------------------------------

loc_F8168:							; CODE XREF: sub_F739E+30p
		move.l	#loc_17C42,($FFFFB03A).w
		move	#$2500,sr
		bclr	#0,($FFFFBF78).w
		bset	#2,($FFFFBF78).w
		bclr	#1,($FFFFBF78).w
		move.w	#0,($FFFFB000).w
		move.w	#$B400,($FFFFB002).w
		move.w	#$B800,($FFFFB00C).w
		move.w	#5,($FFFFB00E).w
		move.w	#$C000,($FFFFB008).w
		move.w	#6,($FFFFB00A).w
		move.w	#$E000,($FFFFB004).w
		move.w	#6,($FFFFB006).w
		move.w	#0,d0
		jsr	sub_11594
		jsr	sub_112BC
		jsr	sub_F84D0
		move.w	d4,($FFFFB01E).w
		move.l	#off_BF54A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$356
		dc.w	$7564
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFB012).w
		move.l	#unk_AAC5A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$FC0
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFBF52).w
		move.l	#unk_BE272,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$1C,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$4B7A0,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$28,d2					; '('
		moveq	#$1C,d3
		moveq	#$D,d5
		jsr	sub_1169A
		move.w	d4,($FFFFD432).w
		add.w	#$24,d4					; '$'
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_4DEEE,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$28,d2					; '('
		moveq	#4,d3
		moveq	#0,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	1
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFD426).w
		move.l	#off_BEFB8,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$26,d2					; '&'
		moveq	#$D,d3
		moveq	#0,d5
		jsr	sub_1169A
		move.w	d4,($FFFFD430).w
		add.w	#$24,d4					; '$'
		move.w	d4,($FFFFD436).w
		move.l	#off_BF70A,a2
		jsr	sub_11738
		move.w	d4,($FFFFD438).w
		move.l	#off_BF70A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$341
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	loc_16E82
		move.w	#$28,($FFFFB066).w			; '('
		move.w	#$28,($FFFFB366).w			; '('
		st	($FFFFC6CE).w
		st	($FFFFCA32).w
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F8304:							; CODE XREF: sub_F739E:loc_F7480p
		move.l	#$5460,d6

loc_F830A:							; CODE XREF: sub_F8304+Ej
					; sub_F8304:loc_F8412j
		move.w	($FFFFB03E).w,d1
		sub.w	($FFFFB040).w,d1
		beq.s	loc_F830A
		move.w	($FFFFB03E).w,($FFFFB040).w
		bclr	#5,($FFFFD42E).w
		jsr	nullsub_3
		jsr	sub_11340
		jsr	sub_11318
		tst.w	d1
		beq.w	loc_F8356
		btst	#5,d1
		beq.w	locret_F8416
		btst	#5,d2
		beq.w	loc_F8350
		move.w	#$53,($FFFFD43C).l			; 'S'

loc_F8350:							; CODE XREF: sub_F8304+40j
		bset	#5,($FFFFD42E).w

loc_F8356:							; CODE XREF: sub_F8304+30j
		jsr	sub_11358
		jsr	sub_11318
		tst.w	d1
		beq.w	loc_F8376
		btst	#5,d1
		beq.w	locret_F8416
		bset	#5,($FFFFD42E).w

loc_F8376:							; CODE XREF: sub_F8304+60j
		tst.w	($FFFFD046).w
		beq.w	loc_F83BE
		jsr	sub_11370
		jsr	sub_11318
		tst.w	d1
		beq.w	loc_F839E
		btst	#5,d1
		beq.w	locret_F8416
		bset	#5,($FFFFD42E).w

loc_F839E:							; CODE XREF: sub_F8304+88j
		jsr	sub_11388
		jsr	sub_11318
		tst.w	d1
		beq.w	loc_F83BE
		btst	#5,d1
		beq.w	locret_F8416
		bset	#5,($FFFFD42E).w

loc_F83BE:							; CODE XREF: sub_F8304+76j
					; sub_F8304+A8j
		btst	#6,($FFFFD42E).w
		beq.w	loc_F83DC
		btst	#7,($FFFFD42E).w
		bne.w	loc_F83F0
		move.w	($FFFFD42C).w,($FFFFD428).w
		bsr.w	sub_F8608

loc_F83DC:							; CODE XREF: sub_F8304+C0j
		btst	#7,($FFFFD42E).w
		beq.w	loc_F83F0
		move.w	($FFFFD42A).w,($FFFFD428).w
		bsr.w	sub_F85B0

loc_F83F0:							; CODE XREF: sub_F8304+CAj
					; sub_F8304+DEj
		cmpnw	 $4650,d6
		bgt.w	loc_F840C
		btst	#0,($FFFFD42E).w
		bne.w	loc_F840C
		jsr	sub_F87D2
		bra.w	loc_F8412
; ---------------------------------------------------------------------------

loc_F840C:							; CODE XREF: sub_F8304+F0j
					; sub_F8304+FAj
		jsr	sub_F87EA

loc_F8412:							; CODE XREF: sub_F8304+104j
		dbf	d6,loc_F830A

locret_F8416:							; CODE XREF: sub_F8304+38j
					; sub_F8304+68j ...
		rts
; End of function sub_F8304

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_F739E

loc_F8418:							; CODE XREF: sub_F739E+ECj
		clr.w	($FFFFD064).w
		move.w	($C00008).l,($FFFFD066).w
		move.w	($C00008).l,($FFFFD068).w
		clr.w	($FFFFD048).w
		clr.w	($FFFFD04A).w
		clr.w	($FFFFD058).w
		move.w	#1,($FFFFD056).w

loc_F843E:							; CODE XREF: sub_F739E+F4j
		move.w	($FFFFD052).w,-(sp)
		move.w	(sp),($FFFFD05A).w
		move.w	(sp)+,($FFFFD05C).w
		jsr	sub_17AC8
		move.w	($C00008).l,($FFFFD066).w
		move.w	($C00008).l,($FFFFD068).w
		move.w	($FFFFD058).w,($FFFFD072).w
		cmp.w	#2,($FFFFD058).w
		bne.w	loc_F847A
		clr.w	($FFFFD058).w
		bset	#4,($FFFFC2FC).w

loc_F847A:							; CODE XREF: sub_F739E+10CEj
		move.w	($FFFFD048).w,($FFFFD074).w
		cmp.w	#4,($FFFFD048).w
		bne.w	loc_F849C
		move.w	#0,($FFFFD048).w
		bset	#0,($FFFFC2FA).w
		jsr	sub_FC47C

loc_F849C:							; CODE XREF: sub_F739E+10E8j
		jmp	sub_17D80
; END OF FUNCTION CHUNK	FOR sub_F739E
; ---------------------------------------------------------------------------
		tst.w	($FFFFD04A).w
		beq.w	locret_F84CE
		cmp.w	#5,($FFFFD04A).w
		beq.w	locret_F84CE
		cmp.w	#6,($FFFFD04A).w
		beq.w	locret_F84CE
		cmp.w	#9,($FFFFD04A).w
		beq.w	locret_F84CE
		cmp.w	#$A,($FFFFD04A).w

locret_F84CE:							; CODE XREF: ROM:000F84A6j
					; ROM:000F84B0j ...
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F84D0:							; CODE XREF: ROM:000F81C6p

arg_4E6D	=	$4E71

		moveq	#2,d4
		move.l	#unk_AFE1A,a2
		jsr	sub_11738
		move.w	d4,($FFFFD43A).w
		move.l	#unk_AFE1A,a2
		jsr	sub_1172C
		bchg	d1,d1
		move.l	-(sp),-$7655(a2)
		muls.w	4+arg_4E6D(sp),d6
; End of function sub_F84D0


; =============== S U B	R O U T	I N E =======================================


nullsub_3:							; CODE XREF: sub_F8304+1Cp
		rts
; End of function nullsub_3

; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F84FC:							; CODE XREF: sub_F85B0+4p sub_F8608+4p
		movem.l	d0-d7,-(sp)
		bclr	#0,($FFFFD42E).w
		beq.w	loc_F85AA
		bclr	#6,($FFFFD42E).w
		bclr	#7,($FFFFD42E).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#5,($FFFFB02A).w
		moveq	#8,d0
		moveq	#8,d1
		move.w	#$7FF,d2
		btst	#1,($FFFFD42E).w
		beq.w	loc_F8574
		move.w	#3,($FFFFB028).w
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$B,($FFFFB028).w
		move.w	#5,($FFFFB02A).w
		move.w	#$19,d0
		move.w	#8,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		bra.w	loc_F85AA
; ---------------------------------------------------------------------------

loc_F8574:							; CODE XREF: sub_F84FC+3Aj
		move.w	#$1C,($FFFFB028).w
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#3,($FFFFB028).w
		move.w	#5,($FFFFB02A).w
		move.w	#$19,d0
		move.w	#8,d1
		move.w	#$7FF,d2
		jsr	sub_1197E

loc_F85AA:							; CODE XREF: sub_F84FC+Aj
					; sub_F84FC+74j
		movem.l	(sp)+,d0-d7
		rts
; End of function sub_F84FC


; =============== S U B	R O U T	I N E =======================================


sub_F85B0:							; CODE XREF: sub_F739E+A6p
					; sub_F739E+28Ap ...
		movem.l	d0-a6,-(sp)
		bsr.w	sub_F84FC
		bset	#2,($FFFFD42E).w
		bset	#6,($FFFFD42E).w
		bsr.w	sub_F8770
		bclr	#3,($FFFFD42E).w
		beq.w	loc_F85F2
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,($FFFFB028).w
		move.w	#5,($FFFFB02A).w
		moveq	#8,d0
		moveq	#8,d1
		move.w	#$7FF,d2

loc_F85F2:							; CODE XREF: sub_F85B0+1Ej
		jsr	sub_11B92
; End of function sub_F85B0

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$19,($FFFFB028).w
		bra.w	loc_F865C

; =============== S U B	R O U T	I N E =======================================


sub_F8608:							; CODE XREF: sub_F739E+B6p
					; sub_F739E+2B8p ...
		movem.l	d0-a6,-(sp)
		bsr.w	sub_F84FC
		bset	#3,($FFFFD42E).w
		bset	#7,($FFFFD42E).w
		bsr.w	sub_F878C
		bclr	#2,($FFFFD42E).w
		beq.w	loc_F864A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$18,($FFFFB028).w
		move.w	#5,($FFFFB02A).w
		moveq	#8,d0
		moveq	#8,d1
		move.w	#$7FF,d2

loc_F864A:							; CODE XREF: sub_F8608+1Ej
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$CF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#9,($FFFFB028).w

loc_F865C:							; CODE XREF: ROM:000F8604j
		move.w	#$B4,($FFFFD440).w			; ''
		move.w	#6,($FFFFB02A).w
		move.w	($FFFFD430).w,d4
		btst	#2,($FFFFD42E).w
		bne.w	loc_F867A
		move.w	($FFFFD432).w,d4

loc_F867A:							; CODE XREF: sub_F8608+6Aj
		move.w	($FFFFD428).w,d3
		asl.w	#2,d3
		move.l	#off_F86F2,a0
		move.l	(a0,d3.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d3
		move.l	#unk_FF462,a0
		btst	#2,($FFFFD42E).w
		beq.w	loc_F86AA
		sub.w	#$20,d3					; ' '
		bra.w	loc_F86AE
; ---------------------------------------------------------------------------

loc_F86AA:							; CODE XREF: sub_F8608+96j
		sub.w	#$40,d3					; '@'

loc_F86AE:							; CODE XREF: sub_F8608+9Ej
		add.w	d3,a0
		add.l	(a2)+,a1
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		move.l	($FFFFBD4A).w,-(sp)
		move.l	($FFFFBD4E).w,-(sp)
		move.w	#4,d5
		btst	#2,($FFFFD42E).w
		beq.w	loc_F86D8
		move.w	#2,d5

loc_F86D8:							; CODE XREF: sub_F8608+C8j
		jsr	sub_1169A
		move.l	(sp)+,($FFFFBD4E).w
		move.l	(sp)+,($FFFFBD4A).w
		move.w	#$64,($FFFFBD26).w			; 'd'
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F8608

; ---------------------------------------------------------------------------
off_F86F2:	dc.l	off_BF8D0				; DATA XREF: sub_F8608+78o
					; ROM:000FA69Co ...
		dc.l	off_BFD66
		dc.l	off_C00BC
		dc.l	off_C0412
		dc.l	off_C08A8
		dc.l	off_C2362
		dc.l	off_C0CDE
		dc.l	off_C1034
		dc.l	off_C142A
		dc.l	off_C1900
		dc.l	off_C1FEC
		dc.l	off_C2638
		dc.l	off_C29AE
		dc.l	off_C1B96
		dc.l	off_C2E64
		dc.l	off_C333A
		dc.l	off_C3750
		dc.l	off_C3B06
		dc.l	off_C3E7C
		dc.l	off_C41D2
		dc.l	off_C4608
		dc.l	off_C49DE
		dc.l	off_C4DF4
		dc.l	off_C514A
		dc.l	off_C5560
		dc.l	off_C57D6
		dc.l	off_C5C4C
		dc.l	off_C6022

; =============== S U B	R O U T	I N E =======================================


sub_F8762:							; CODE XREF: ROM:000F88EAp
		move.w	($FFFFD436).w,d4
		move.w	#$1C,($FFFFB028).w
		bra.w	loc_F8796
; End of function sub_F8762


; =============== S U B	R O U T	I N E =======================================


sub_F8770:							; CODE XREF: sub_F85B0+14p
		move.w	($FFFFD436).w,d4
		move.w	#$18,($FFFFB028).w
		bra.w	loc_F8796
; End of function sub_F8770


; =============== S U B	R O U T	I N E =======================================


sub_F877E:							; CODE XREF: ROM:loc_F88F2p
		move.w	($FFFFD438).w,d4
		move.w	#3,($FFFFB028).w
		bra.w	loc_F8796
; End of function sub_F877E


; =============== S U B	R O U T	I N E =======================================


sub_F878C:							; CODE XREF: sub_F8608+14p
		move.w	($FFFFD438).w,d4
		move.w	#8,($FFFFB028).w

loc_F8796:							; CODE XREF: sub_F8762+Aj sub_F8770+Aj ...
		clr.w	($FFFFB02C).w
		move.w	#5,($FFFFB02A).w
		move.l	#off_BF702,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		move.w	#8,d3
		move.w	#8,d2
		clr.w	d0
		clr.w	d1
		move.w	#0,d5
		bset	#0,($FFFFC2F8).w
		jsr	sub_1169A
		bclr	#0,($FFFFC2F8).w
		rts
; End of function sub_F878C


; =============== S U B	R O U T	I N E =======================================


sub_F87D2:							; CODE XREF: sub_F7B20+3Ep
					; sub_F8304+FEp ...
		move.w	#$AA,($FFFFD43C).w			; ''
		btst	#5,($FFFFD42E).w
		beq.w	locret_F87E8
		move.w	#$53,($FFFFD43C).w			; 'S'

locret_F87E8:							; CODE XREF: sub_F87D2+Cj
		rts
; End of function sub_F87D2


; =============== S U B	R O U T	I N E =======================================


sub_F87EA:							; CODE XREF: sub_F8304:loc_F840Cp
		subq.w	#1,($FFFFD440).w
		bpl.w	locret_F8860
		move.w	#$FFFF,($FFFFD440).w
		bset	#0,($FFFFD42E).w
		btst	#0,($FFFFD42E).w
		beq.w	locret_F8860
		btst	#5,($FFFFD42E).w
		beq.w	loc_F8822
		cmp.w	#$52,($FFFFD43C).w			; 'R'
		bge.w	loc_F8822
		sub.w	#$50,($FFFFD43C).w			; 'P'

loc_F8822:							; CODE XREF: sub_F87EA+24j
					; sub_F87EA+2Ej
		subq.w	#1,($FFFFD43C).w
		bmi.w	loc_F8862
		cmp.w	#$52,($FFFFD43C).w			; 'R'
		bne.w	locret_F8860
		bclr	#2,($FFFFD42E).w
		bclr	#3,($FFFFD42E).w
		bchg	#1,($FFFFD42E).w
		st	($FFFFD42A).w
		st	($FFFFD42C).w
		bclr	#6,($FFFFD42E).w
		bclr	#7,($FFFFD42E).w
		jsr	loc_F8868

locret_F8860:							; CODE XREF: sub_F87EA+4j
					; sub_F87EA+1Aj ...
		rts
; ---------------------------------------------------------------------------

loc_F8862:							; CODE XREF: sub_F87EA+3Cj
		jmp	sub_F87D2
; End of function sub_F87EA

; ---------------------------------------------------------------------------

loc_F8868:							; CODE XREF: sub_F87EA+70p
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$1C,($FFFFB028).w
		btst	#1,($FFFFD42E).w
		bne.w	loc_F888E
		move.w	#3,($FFFFB028).w

loc_F888E:							; CODE XREF: ROM:000F8884j
		move.w	#5,($FFFFB02A).w
		move.w	#8,d0
		move.w	#8,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#3,($FFFFB028).w
		btst	#1,($FFFFD42E).w
		bne.w	loc_F88C8
		move.w	#$B,($FFFFB028).w

loc_F88C8:							; CODE XREF: ROM:000F88BEj
		move.w	#5,($FFFFB02A).w
		move.w	#$19,d0
		move.w	#8,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		btst	#1,($FFFFD42E).w
		bne.w	loc_F88F2
		bsr.w	sub_F8762
		bra.w	loc_F88FA
; ---------------------------------------------------------------------------

loc_F88F2:							; CODE XREF: ROM:000F88E6j
		bsr.w	sub_F877E
		addq.w	#1,($FFFFD43E).w

loc_F88FA:							; CODE XREF: ROM:000F88EEj
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		cmp.w	#6,($FFFFD43E).l
		blt.w	loc_F8916
		clr.w	($FFFFD43E).w

loc_F8916:							; CODE XREF: ROM:000F890Ej
		move.w	($FFFFC330).w,d0
		move.w	#$1D,($FFFFB028).w
		btst	#1,($FFFFD42E).w
		beq.w	loc_F8934
		move.w	($FFFFC332).w,d0
		move.w	#4,($FFFFB028).w

loc_F8934:							; CODE XREF: ROM:000F8926j
		move.w	#6,($FFFFB02A).w
		move.w	($FFFFD430).w,d4
		move.w	($FFFFD43E).w,d3
		mulu.w	#6,d3
		asl.w	#2,d0
		move.l	#off_F92F4,a0
		move.l	(a0,d0.w),a0
		move.w	4(a0,d3.w),($FFFFD434).w
		move.l	(a0,d3.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		move.l	#off_C63F8,a0
		add.l	(a0),a0
		tst.l	(a2)
		bne.w	loc_F8980
		move.l	#off_C63F8,a1
		add.l	4(a1),a1
		tst.l	(a2)+
		bra.w	loc_F8982
; ---------------------------------------------------------------------------

loc_F8980:							; CODE XREF: ROM:000F896Cj
		add.l	(a2)+,a1

loc_F8982:							; CODE XREF: ROM:000F897Cj
		bsr.w	sub_FE98A
		move.l	#$FFFFDA1E,a2
		move.w	($FFFFB03E).w,d3

loc_F8990:							; CODE XREF: ROM:000F8994j
		cmp.w	($FFFFB03E).w,d3
		beq.s	loc_F8990
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		move.l	($FFFFBD4A).w,-(sp)
		move.l	($FFFFBD4E).w,-(sp)
		move.w	#2,d5
		jsr	sub_1169A
		move.l	(sp)+,($FFFFBD4E).w
		move.l	(sp)+,($FFFFBD4A).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$6000,($FFFFB02C).w
		move.w	#3,($FFFFB028).w
		move.w	#$19,d0
		btst	#1,($FFFFD42E).w
		beq.w	loc_F89EC
		move.w	#$B,($FFFFB028).w
		move.w	#$19,d0

loc_F89EC:							; CODE XREF: ROM:000F89DEj
		move.w	#5,($FFFFB02A).w
		move.w	#8,d1
		move.w	($FFFFB028).w,($FFFFD442).w
		addq.w	#1,($FFFFD442).w
		move.w	($FFFFB02A).w,($FFFFD444).w
		addq.w	#1,($FFFFD444).w
		jsr	sub_119B8
		move.w	($FFFFD442).w,($FFFFB028).w
		move.w	($FFFFD444).w,($FFFFB02A).w
		move.w	#0,($FFFFB02C).w
		move.w	#0,($FFFFB02E).w
		move.w	($FFFFC330).w,d0
		btst	#1,($FFFFD42E).w
		beq.w	loc_F8A3A
		move.w	($FFFFC332).w,d0

loc_F8A3A:							; CODE XREF: ROM:000F8A32j
		move.w	d0,($FFFFD446).w
		move.l	#$30E,a2
		asl.w	#2,d0
		move.l	(a2,d0.w),a2
		add.w	(a2),a2
		move.w	($FFFFD434).w,d0
		bra.w	loc_F8A58
; ---------------------------------------------------------------------------

loc_F8A54:							; CODE XREF: ROM:loc_F8A58j
		add.w	(a2),a2
		addq.w	#8,a2

loc_F8A58:							; CODE XREF: ROM:000F8A50j
		dbf	d0,loc_F8A54
		move.l	#$FFFFBFA4,a3
		lea	2(a3),a1
		move.w	#6,(a3)
		move.l	a2,-(sp)
		add.w	(a2),a2
		move.b	(a2),d0
		ext.w	d0
		jsr	sub_18BDC
		move.w	#$2000,4(a3)
		move.l	(sp)+,a1
		jsr	sub_11D9E
		move.l	a3,a1
		bset	#3,($FFFFC2F8).w
		jsr	sub_11A48
		bclr	#3,($FFFFC2F8).w
		move.l	#$FFFFBFA4,a1
		move.w	($FFFFD446).w,d0
		move.w	($FFFFD434).w,d1
		tst.w	($FFFFD458).w
		bmi.w	loc_F8B4E
		jsr	sub_F98C6
		cmp.w	#2,(a1)
		beq.w	loc_F8B4E
		move.w	($FFFFD442).w,($FFFFB028).w
		addq.w	#2,($FFFFB02A).w
		bset	#3,($FFFFC2F8).w
		jsr	sub_11A48
		bclr	#3,($FFFFC2F8).w
		move.l	#$FFFFBFA4,a1
		move.w	($FFFFD446).w,d0
		move.w	($FFFFD434).w,d1
		move.l	#$FFFF0000,a0
		jsr	sub_F9A64
		move.w	($FFFFD442).w,($FFFFB028).w
		cmp.w	#2,(a1)
		beq.w	loc_F8B18
		addq.w	#1,($FFFFB02A).w
		bset	#3,($FFFFC2F8).w
		jsr	sub_11A48
		bclr	#3,($FFFFC2F8).w

loc_F8B18:							; CODE XREF: ROM:000F8AFEj
		move.l	#$FFFFBFA4,a1
		move.w	($FFFFD446).w,d0
		move.w	($FFFFD434).w,d1
		move.l	#$FFFF0000,a0
		jsr	sub_F9AAC
		move.w	($FFFFD442).w,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		bset	#3,($FFFFC2F8).w
		jsr	sub_11A48
		bclr	#3,($FFFFC2F8).w

loc_F8B4E:							; CODE XREF: ROM:000F8AACj
					; ROM:000F8ABAj
		move.w	#$64,($FFFFBD26).w			; 'd'
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_F8B5A:							; CODE XREF: sub_13CD8+A6p
					; sub_14062+10Cp
		tst.w	$34(a2)
		beq.w	locret_F8B9E
		movem.l	d0-d7,-(sp)
		move.w	(a2),d2
		move.w	$14(a2),d3
		move.w	$4A(a2),($FFFFBD22).w
		add.w	#$17,($FFFFBD22).l
		move.w	$4C(a2),($FFFFBD24).w
		add.w	#$17,($FFFFBD24).l
		movem.l	a0-a6,-(sp)
		exg	a2,a3
		jsr	sub_F6FD6
		exg	a2,a3
		movem.l	(sp)+,a0-a6
		movem.l	(sp)+,d0-d7

locret_F8B9E:							; CODE XREF: sub_F8B5A+4j
		rts
; End of function sub_F8B5A


; =============== S U B	R O U T	I N E =======================================


sub_F8BA0:							; CODE XREF: sub_794E:loc_798Ap
		bclr	#5,($FFFFC2F8).w
		tst.w	($FFFFB7AA).w
		bmi.w	locret_F8BF2
		movem.l	d0/a0,-(sp)
		move.l	#$FFFFB04A,a0
		move.w	($FFFFB7AA).w,d0
		asl.w	#7,d0
		add.w	d0,a0
		move.w	$14(a0),d0
		btst	#7,$62(a0)
		bne.w	loc_F8BD0
		neg.w	d0

loc_F8BD0:							; CODE XREF: sub_F8BA0+2Aj
		cmpnw	 $58,d0					; 'X'
		blt.w	loc_F8BEE
		cmp.w	#$47,(a0)				; 'G'
		bgt.w	loc_F8BEE
		cmp.w	#$FFB9,(a0)
		blt.w	loc_F8BEE
		bset	#5,($FFFFC2F8).w

loc_F8BEE:							; CODE XREF: sub_F8BA0+34j
					; sub_F8BA0+3Cj ...
		movem.l	(sp)+,d0/a0

locret_F8BF2:							; CODE XREF: sub_F8BA0+Aj
		rts
; End of function sub_F8BA0

; ---------------------------------------------------------------------------
word_F8BF4:	include	"Unknown\F8BF4.asm"

; =============== S U B	R O U T	I N E =======================================


sub_F98C6:							; CODE XREF: ROM:000F8AB0p
		movem.l	d0-d7/a1-a6,-(sp)
		move.l	a1,-(sp)
		move.l	a1,a3
		move.l	#word_F9950,a1
		movem.l	d0-d1,-(sp)
		bsr.w	sub_F997A
		movem.l	(sp)+,d0-d1
		bsr.w	sub_F99F2
		tst.w	d0
		bne.w	loc_F98F4
		move.l	(sp)+,a1
		move.w	#2,(a1)
		bra.w	loc_F994A
; ---------------------------------------------------------------------------

loc_F98F4:							; CODE XREF: sub_F98C6+20j
		move.l	#$FFFFBF20,a1
		move.w	d0,($FFFFBF14).w
		bsr.w	sub_F998E
		move.l	(sp),a3
		move.l	#$FFFFBF20,a1
		jsr	sub_11D9E
		move.w	($FFFFBF14).w,d0
		move.l	#word_F996A,a1
		cmpnw	 1,d0
		bne.w	loc_F9928
		move.l	#word_F9972,a1

loc_F9928:							; CODE XREF: sub_F98C6+58j
		tst.w	d1
		bne.w	loc_F9942
		move.l	#word_F995A,a1
		cmpnw	 1,d0
		bne.w	loc_F9942
		move.l	#word_F9962,a1

loc_F9942:							; CODE XREF: sub_F98C6+64j
					; sub_F98C6+72j
		move.l	(sp)+,a3
		jsr	sub_11D9E

loc_F994A:							; CODE XREF: sub_F98C6+2Aj
		movem.l	(sp)+,d0-d7/a1-a6
		rts
; End of function sub_F98C6

; ---------------------------------------------------------------------------
word_F9950:	dc.w	$A					; DATA XREF: sub_F98C6+8o
aRecord:	dc.b	'Record ',0
word_F995A:	dc.w	8					; DATA XREF: sub_F98C6+68o
aGoals_0:	dc.b	' goals'
word_F9962:	dc.w	8					; DATA XREF: sub_F98C6+76o
aGoal_0:	dc.b	' goal',0
word_F996A:	dc.w	8					; DATA XREF: sub_F98C6+4Eo
aSaves:		dc.b	' saves'
word_F9972:	dc.w	8					; DATA XREF: sub_F98C6+5Co
aSave:		dc.b	' save',0

; =============== S U B	R O U T	I N E =======================================


sub_F997A:							; CODE XREF: sub_F98C6+12p
					; sub_F9A64+12p ...
		movem.l	d0,-(sp)
		move.w	(a1),d0
		subq.w	#1,d0

loc_F9982:							; CODE XREF: sub_F997A+Aj
		move.b	(a1)+,(a3)+
		dbf	d0,loc_F9982
		movem.l	(sp)+,d0
		rts
; End of function sub_F997A


; =============== S U B	R O U T	I N E =======================================


sub_F998E:							; CODE XREF: sub_F98C6+38p
					; sub_FA75C+72p ...
		movem.w	d0-d6/a0,-(sp)
		move.l	a1,-(sp)
		move.w	d0,d6
		addq.l	#2,a1
		clr.w	d1
		ext.l	d0
		divu.w	#$64,d0					; 'd'
		tst.w	d0
		beq.w	loc_F99AE
		add.w	#$30,d0					; '0'
		move.b	d0,(a1)+
		addq.w	#1,d1

loc_F99AE:							; CODE XREF: sub_F998E+14j
		swap	d0
		ext.l	d0
		divu.w	#$A,d0
		add.w	#$30,d0					; '0'
		cmpnb	 $30,d0					; '0'
		bne.w	loc_F99CA
		cmpnw	 $A,d6
		blt.w	loc_F99CE

loc_F99CA:							; CODE XREF: sub_F998E+30j
		move.b	d0,(a1)+
		addq.w	#1,d1

loc_F99CE:							; CODE XREF: sub_F998E+38j
		swap	d0
		add.w	#$30,d0					; '0'
		move.b	d0,(a1)+
		addq.w	#1,d1
		btst	#0,d1
		beq.w	loc_F99E6
		move.b	#0,(a1)+
		addq.w	#1,d1

loc_F99E6:							; CODE XREF: sub_F998E+4Ej
		move.l	(sp)+,a0
		addq.w	#2,d1
		move.w	d1,(a0)
		movem.w	(sp)+,d0-d6/a0
		rts
; End of function sub_F998E


; =============== S U B	R O U T	I N E =======================================


sub_F99F2:							; CODE XREF: sub_F98C6+1Ap
		movem.l	d2-a6,-(sp)
		move.w	d0,-(sp)
		move.l	#$30E,a2
		asl.w	#2,d0
		move.l	(a2,d0.w),a2
		add.w	(a2),a2
		move.w	d1,d0
		bra.w	loc_F9A10
; ---------------------------------------------------------------------------

loc_F9A0C:							; CODE XREF: sub_F99F2:loc_F9A10j
		add.w	(a2),a2
		addq.w	#8,a2

loc_F9A10:							; CODE XREF: sub_F99F2+16j
		dbf	d0,loc_F9A0C
		move.w	d1,d0
		move.w	(sp)+,d1
		move.w	d0,d5
		move.w	d1,d4
		ext.l	d0
		move.l	#$FFFF0000,a0
		bsr.w	sub_F9B94
		move.b	(a0),d0
		ext.w	d0
		move.l	#$30E,a5
		asl.w	#2,d4
		move.l	(a5,d4.w),a5
		add.w	$A(a5),a5
		move.w	(a5),d1
		movem.w	d0,-(sp)
		clr.w	d0

loc_F9A44:							; CODE XREF: sub_F99F2+56j
		addq.w	#1,d0
		asl.w	#4,d1
		bne.s	loc_F9A44
		cmp.w	d5,d0
		movem.w	(sp)+,d0
		bgt.w	loc_F9A5A
		clr.w	d1
		bra.w	loc_F9A5E
; ---------------------------------------------------------------------------

loc_F9A5A:							; CODE XREF: sub_F99F2+5Ej
		move.w	#1,d1

loc_F9A5E:							; CODE XREF: sub_F99F2+64j
		movem.l	(sp)+,d2-a6
		rts
; End of function sub_F99F2


; =============== S U B	R O U T	I N E =======================================


sub_F9A64:							; CODE XREF: ROM:000F8AEEp
		movem.l	d0-d7/a1-a6,-(sp)
		move.l	a1,-(sp)
		move.l	a1,a3
		move.l	#word_F9AA6,a1
		movem.l	d0-d1,-(sp)
		bsr.w	sub_F997A
		movem.l	(sp)+,d0-d1
		move.l	#$FFFFBF20,a1
		bsr.w	sub_F9AE4
		move.l	(sp)+,a3
		cmp.w	#2,(a1)
		bne.w	loc_F9A9A
		move.w	#2,(a3)
		bra.w	loc_F9AA0
; ---------------------------------------------------------------------------

loc_F9A9A:							; CODE XREF: sub_F9A64+2Aj
		jsr	sub_11D9E

loc_F9AA0:							; CODE XREF: sub_F9A64+32j
		movem.l	(sp)+,d0-d7/a1-a6
		rts
; End of function sub_F9A64

; ---------------------------------------------------------------------------
word_F9AA6:	dc.w	6					; DATA XREF: sub_F9A64+8o
aBy:		dc.b	'by ',0

; =============== S U B	R O U T	I N E =======================================


sub_F9AAC:							; CODE XREF: ROM:000F8B2Cp
		movem.l	d0-d7/a1-a6,-(sp)
		move.l	a1,-(sp)
		move.l	a1,a3
		move.l	#word_F9ADE,a1
		movem.l	d0-d1,-(sp)
		bsr.w	sub_F997A
		movem.l	(sp)+,d0-d1
		move.l	#$FFFFBF20,a1
		bsr.w	sub_F9B2A
		move.l	(sp)+,a3
		jsr	sub_11D9E
		movem.l	(sp)+,d0-d7/a1-a6
		rts
; End of function sub_F9AAC

; ---------------------------------------------------------------------------
word_F9ADE:	dc.w	6					; DATA XREF: sub_F9AAC+8o
aVs_:		dc.b	'vs. '

; =============== S U B	R O U T	I N E =======================================


sub_F9AE4:							; CODE XREF: sub_F9A64+20p
		movem.l	d0-d7/a1-a2,-(sp)
		cmp.l	#0,a0
		bne.w	loc_F9B14
		movem.l	d0-d7/a1-a6,-(sp)
		move.w	d0,-(sp)
		move.w	d1,d0
		move.w	(sp)+,d1
		ext.l	d1
		ext.l	d0
		move.l	#$FFFF0000,a0
		bsr.w	sub_F9B94
		move.l	#$FFFF0000,a0
		movem.l	(sp)+,d0-d7/a1-a6

loc_F9B14:							; CODE XREF: sub_F9AE4+Aj
		move.b	1(a0),d2
		ext.w	d2
		bset	#7,($FFFFC2F8).w
		bsr.w	sub_FA014
		movem.l	(sp)+,d0-d7/a1-a2
		rts
; End of function sub_F9AE4


; =============== S U B	R O U T	I N E =======================================


sub_F9B2A:							; CODE XREF: sub_F9AAC+20p
		movem.l	d0-a6,-(sp)
		cmp.l	#0,a0
		bne.w	loc_F9B5A
		movem.l	d0-d7/a1-a6,-(sp)
		move.w	d0,-(sp)
		move.w	d1,d0
		move.w	(sp)+,d1
		ext.l	d1
		ext.l	d0
		move.l	#$FFFF0000,a0
		bsr.w	sub_F9B94
		move.l	#$FFFF0000,a0
		movem.l	(sp)+,d0-d7/a1-a6

loc_F9B5A:							; CODE XREF: sub_F9B2A+Aj
		move.b	3(a0),d2
		ext.w	d2
		movem.l	d0/a1,-(sp)
		bset	#7,($FFFFC2F8).w
		bsr.w	sub_FA014
		move.l	a1,a3
		move.l	#word_F9B90,a1
		jsr	sub_11D9E
		movem.l	(sp)+,d0/a1
		move.b	2(a0),d0
		jsr	sub_FA880
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9B2A

; ---------------------------------------------------------------------------
word_F9B90:	dc.w	4					; DATA XREF: sub_F9B2A+46o
		dc.b	' ',0

; =============== S U B	R O U T	I N E =======================================


sub_F9B94:							; CODE XREF: sub_F99F2+32p
					; sub_F9AE4+22p ...
		bclr	#6,($FFFFC2F8).w

loc_F9B9A:							; CODE XREF: sub_F9BDA+6j
		movem.l	d0-d1/a0-a1,-(sp)
		asl.l	#2,d0
		add.l	#0,d0
		move.l	#word_F9CA4,a1
		add.w	d1,d1
		move.w	(a1,d1.w),d1
		asl.w	#2,d1
		ext.l	d1
		add.l	d1,d0
		moveq	#4,d1
		btst	#6,($FFFFC2F8).w
		beq.w	loc_F9BCE
		jsr	sub_1A1E4
		bra.w	loc_F9BD4
; ---------------------------------------------------------------------------

loc_F9BCE:							; CODE XREF: sub_F9B94+2Cj
		jsr	sub_1A244

loc_F9BD4:							; CODE XREF: sub_F9B94+36j
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_F9B94


; =============== S U B	R O U T	I N E =======================================


sub_F9BDA:							; CODE XREF: sub_F9F50+64p
		bset	#6,($FFFFC2F8).w
		bra.s	loc_F9B9A
; End of function sub_F9BDA


; =============== S U B	R O U T	I N E =======================================


sub_F9BE2:							; CODE XREF: ROM:00009B02p
					; ROM:00009B60p ...
		bclr	#6,($FFFFC2F8).w

loc_F9BE8:							; CODE XREF: sub_F9C18+6j
		movem.l	d0-d1/a0-a1,-(sp)
		move.l	d1,d0
		asl.w	#4,d0
		add.l	#$B60,d0
		moveq	#$10,d1
		btst	#6,($FFFFC2F8).w
		beq.w	loc_F9C0C
		jsr	sub_1A1E4
		bra.w	loc_F9C12
; ---------------------------------------------------------------------------

loc_F9C0C:							; CODE XREF: sub_F9BE2+1Cj
		jsr	sub_1A244

loc_F9C12:							; CODE XREF: sub_F9BE2+26j
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_F9BE2


; =============== S U B	R O U T	I N E =======================================


sub_F9C18:							; CODE XREF: sub_F9EAA+9Cp
		bset	#6,($FFFFC2F8).w
		bra.s	loc_F9BE8
; End of function sub_F9C18


; =============== S U B	R O U T	I N E =======================================


sub_F9C20:							; CODE XREF: sub_F9DDA+4p
		bclr	#6,($FFFFC2F8).w

loc_F9C26:							; CODE XREF: sub_F9C56+6j
		movem.l	d0-d1/a0-a1,-(sp)
		move.l	d1,d0
		asl.w	#4,d0
		add.l	#$D20,d0
		moveq	#$10,d1
		btst	#6,($FFFFC2F8).w
		beq.w	loc_F9C4A
		jsr	sub_1A1E4
		bra.w	loc_F9C50
; ---------------------------------------------------------------------------

loc_F9C4A:							; CODE XREF: sub_F9C20+1Cj
		jsr	sub_1A244

loc_F9C50:							; CODE XREF: sub_F9C20+26j
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_F9C20


; =============== S U B	R O U T	I N E =======================================


sub_F9C56:							; CODE XREF: sub_F9DDA:loc_F9EA0p
		bset	#6,($FFFFC2F8).w
		bra.s	loc_F9C26
; End of function sub_F9C56


; =============== S U B	R O U T	I N E =======================================


sub_F9C5E:							; CODE XREF: sub_FB9D4+68p
		bset	#6,($FFFFC2F8).w
		bra.w	loc_F9C6E
; End of function sub_F9C5E


; =============== S U B	R O U T	I N E =======================================


sub_F9C68:							; CODE XREF: sub_F739E+2Ap
					; ROM:000FBCC8p ...
		bclr	#6,($FFFFC2F8).w

loc_F9C6E:							; CODE XREF: sub_F9C5E+6j
		movem.l	d0-a6,-(sp)
		move.l	#$80,d1					; ''
		move.l	#$DA0,d0
		move.l	#$FFFFD45A,a0
		btst	#6,($FFFFC2F8).w
		beq.w	loc_F9C98
		jsr	sub_1A1E4
		bra.w	loc_F9C9E
; ---------------------------------------------------------------------------

loc_F9C98:							; CODE XREF: sub_F9C68+22j
		jsr	sub_1A244

loc_F9C9E:							; CODE XREF: sub_F9C68+2Cj
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9C68

; ---------------------------------------------------------------------------
word_F9CA4:	dc.w	0					; DATA XREF: sub_F9B94+12o
		dc.w	$1A
		dc.w	$34
		dc.w	$4E
		dc.w	$68
		dc.w	$82
		dc.w	$9C
		dc.w	$B6
		dc.w	$D0
		dc.w	$EA
		dc.w	$104
		dc.w	$11E
		dc.w	$138
		dc.w	$152
		dc.w	$16C
		dc.w	$186
		dc.w	$1A0
		dc.w	$1BA
		dc.w	$1D4
		dc.w	$1EE
		dc.w	$208
		dc.w	$222
		dc.w	$23C
		dc.w	$256
		dc.w	$270
		dc.w	$28A
		dc.w	$2A4
		dc.w	$2BE
		dc.w	$2D8

; =============== S U B	R O U T	I N E =======================================


sub_F9CDE:							; CODE XREF: sub_130E6+10p
		tst.w	($FFFFD458).w
		bmi.w	locret_F9DD8
		tst.w	($FFFFD054).w
		bne.w	locret_F9DD8
		movem.l	d0-a6,-(sp)
		bsr.w	sub_F9FC0
		bsr.w	sub_F9FEA
		move.w	($FFFFC330).w,d1
		ext.l	d1
		move.l	#$FFFFCF36,a0
		move.l	#$FFFFC6CE,a2
		move.w	($FFFFD448).w,d5
		move.w	($FFFFC332).w,d6
		bsr.w	sub_F9F50
		move.w	($FFFFC332).w,d1
		ext.l	d1
		move.l	#$FFFFCA32,a2
		move.w	($FFFFD44A).w,d5
		move.w	($FFFFC330).w,d6
		bsr.w	sub_F9F50
		clr.w	d7
		move.l	#$FFFFC6CE,a1
		move.w	($FFFFD042).w,d4
		move.w	($FFFFD044).w,d5
		move.w	($FFFFC330).w,d1
		move.w	($FFFFC332).w,d2
		move.w	($FFFFD448).w,d6
		ext.l	d1
		move.l	#$FFFFCF36,a0
		bsr.w	sub_F9EAA
		move.l	#$FFFFCA32,a1
		move.w	($FFFFD044).w,d4
		move.w	($FFFFD042).w,d5
		move.w	($FFFFC332).w,d1
		move.w	($FFFFC330).w,d2
		move.w	($FFFFD44A).w,d6
		ext.l	d1
		move.l	#$FFFFCF36,a0
		bsr.w	sub_F9EAA
		move.l	#$FFFFCF36,a0
		move.w	($FFFFD042).w,d1
		ext.l	d1
		move.w	($FFFFD044).w,d2
		move.w	($FFFFC330).w,d3
		move.w	($FFFFC332).w,d4
		move.l	#$FFFFC6CE,a1
		move.l	#$FFFFCA32,a2
		bsr.w	sub_F9DDA
		move.l	#$FFFFCF36,a0
		move.w	($FFFFD044).w,d1
		ext.l	d1
		move.w	($FFFFD042).w,d2
		move.w	($FFFFC332).w,d3
		move.w	($FFFFC330).w,d4
		move.l	#$FFFFCA32,a1
		move.l	#$FFFFC6CE,a2
		bsr.w	sub_F9DDA
		jsr	sub_1A206
		movem.l	(sp)+,d0-a6

locret_F9DD8:							; CODE XREF: sub_F9CDE+4j sub_F9CDE+Cj
		rts
; End of function sub_F9CDE


; =============== S U B	R O U T	I N E =======================================


sub_F9DDA:							; CODE XREF: sub_F9CDE+C4p
					; sub_F9CDE+ECp
		movem.l	d0-a6,-(sp)
		bsr.w	sub_F9C20
		st	d7
		movem.w	d0,-(sp)
		clr.w	d0
		move.b	$A(a0),d0
		lsl.w	#8,d0
		move.b	$B(a0),d0
		cmpnw	 $2328,d0
		bge.w	loc_F9E4C
		addq.w	#1,d0
		move.b	d0,$B(a0)
		lsr.w	#8,d0
		move.b	d0,$A(a0)
		move.w	$C(a1),d0
		cmp.w	$C(a2),d0
		bgt.w	loc_F9E34
		blt.w	loc_F9E4C
		clr.w	d0
		move.b	$C(a0),d0
		lsl.w	#8,d0
		move.b	$D(a0),d0
		addq.w	#1,d0
		move.b	d0,$D(a0)
		lsr.w	#8,d0
		move.b	d0,$C(a0)
		bra.w	loc_F9E4C
; ---------------------------------------------------------------------------

loc_F9E34:							; CODE XREF: sub_F9DDA+36j
		clr.w	d0
		move.b	8(a0),d0
		lsl.w	#8,d0
		move.b	9(a0),d0
		addq.w	#1,d0
		move.b	d0,9(a0)
		lsr.w	#8,d0
		move.b	d0,8(a0)

loc_F9E4C:							; CODE XREF: sub_F9DDA+1Ej
					; sub_F9DDA+3Aj ...
		movem.w	(sp)+,d0
		movem.w	d0,-(sp)
		move.w	$C(a1),d0
		cmp.w	$C(a2),d0
		movem.w	(sp)+,d0
		ble.w	loc_F9EA0
		move.w	$C(a1),d5
		cmp.b	(a0),d5
		ble.w	loc_F9E7E
		st	d7
		move.b	d5,(a0)
		move.b	d3,1(a0)
		move.b	d4,2(a0)
		move.b	d2,3(a0)

loc_F9E7E:							; CODE XREF: sub_F9DDA+90j
		move.w	$C(a2),d5
		move.w	(a2),d6
		sub.w	d5,d6
		cmp.b	4(a0),d6
		ble.w	loc_F9EA0
		st	d7
		move.b	d6,4(a0)
		move.b	d3,5(a0)
		move.b	d4,6(a0)
		move.b	d2,7(a0)

loc_F9EA0:							; CODE XREF: sub_F9DDA+86j
					; sub_F9DDA+B0j
		bsr.w	sub_F9C56
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9DDA


; =============== S U B	R O U T	I N E =======================================


sub_F9EAA:							; CODE XREF: sub_F9CDE+76p
					; sub_F9CDE+9Cp
		movem.l	d0-a6,-(sp)
		bsr.w	sub_F9BE2
		move.w	$C(a1),d3
		cmp.b	(a0),d3
		ble.w	loc_F9ECC
		move.b	d3,(a0)
		move.b	d4,1(a0)
		move.b	d5,3(a0)
		move.b	d2,2(a0)
		st	d7

loc_F9ECC:							; CODE XREF: sub_F9EAA+Ej
		subq.w	#1,d6
		clr.w	d3

loc_F9ED0:							; CODE XREF: sub_F9EAA:loc_F9EF4j
		move.w	d6,d0
		add.w	#$E8,d0					; ''
		move.b	(a1,d0.w),($FFFFBF12).w
		add.w	#-$34,d0
		move.b	(a1,d0.w),d0
		sub.b	d0,($FFFFBF12).w
		cmp.b	($FFFFBF12).w,d3
		bge.w	loc_F9EF4
		move.b	($FFFFBF12).w,d3

loc_F9EF4:							; CODE XREF: sub_F9EAA+42j
		dbf	d6,loc_F9ED0
		cmp.b	4(a0),d3
		ble.w	loc_F9F12
		st	d7
		move.b	d3,4(a0)
		move.b	d4,5(a0)
		move.b	d5,7(a0)
		move.b	d2,6(a0)

loc_F9F12:							; CODE XREF: sub_F9EAA+52j
		cmp.w	($FFFFC330).w,d1
		bne.w	loc_F9F40
		move.b	8(a0),d3
		and.w	#$FF,d3
		cmp.w	($FFFFC314).w,d3
		bge.w	loc_F9F40
		move.w	($FFFFC314).w,d3
		st	d7
		move.b	d3,8(a0)
		move.b	d4,9(a0)
		move.b	d5,$B(a0)
		move.b	d2,$A(a0)

loc_F9F40:							; CODE XREF: sub_F9EAA+6Cj
					; sub_F9EAA+7Cj
		tst.w	d7
		beq.w	loc_F9F4A
		bsr.w	sub_F9C18

loc_F9F4A:							; CODE XREF: sub_F9EAA+98j
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9EAA


; =============== S U B	R O U T	I N E =======================================


sub_F9F50:							; CODE XREF: sub_F9CDE+36p
					; sub_F9CDE+4Ep
		clr.l	d0
		move.w	#$19,d2

loc_F9F56:							; CODE XREF: sub_F9F50+6Aj
		bsr.w	sub_F9B94
		cmp.w	d5,d0
		blt.w	loc_F9F70
		move.w	d0,-(sp)
		add.w	#$B4,d0					; ''
		move.b	(a2,d0.w),d4
		move.w	(sp)+,d0
		bra.w	loc_F9F84
; ---------------------------------------------------------------------------

loc_F9F70:							; CODE XREF: sub_F9F50+Cj
		move.w	d0,-(sp)
		add.w	#$E8,d0					; ''
		move.b	(a2,d0.w),d4
		add.w	#-$34,d0
		sub.b	(a2,d0.w),d4
		move.w	(sp)+,d0

loc_F9F84:							; CODE XREF: sub_F9F50+1Cj
		move.b	(a0),d3
		cmp.b	d3,d4
		ble.w	loc_F9FB8
		move.b	d4,(a0)
		move.b	($FFFFD043).w,1(a0)
		move.b	($FFFFD045).w,3(a0)
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_F9FB0
		move.b	($FFFFD045).w,1(a0)
		move.b	($FFFFD043).w,3(a0)

loc_F9FB0:							; CODE XREF: sub_F9F50+50j
		move.b	d6,2(a0)
		bsr.w	sub_F9BDA

loc_F9FB8:							; CODE XREF: sub_F9F50+38j
		addq.w	#1,d0
		dbf	d2,loc_F9F56
		rts
; End of function sub_F9F50


; =============== S U B	R O U T	I N E =======================================


sub_F9FC0:							; CODE XREF: ROM:000188F2p
					; sub_F9CDE+14p ...
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFC6CE,a2
		jsr	sub_9F40
		move.w	d0,($FFFFD448).w
		move.l	#$FFFFCA32,a2
		jsr	sub_9F40
		move.w	d0,($FFFFD44A).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9FC0


; =============== S U B	R O U T	I N E =======================================


sub_F9FEA:							; CODE XREF: sub_F9CDE+18p
					; ROM:000FA0E0p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFC6CE,a2
		jsr	sub_9F9A
		move.w	d0,($FFFFD44C).w
		move.l	#$FFFFCA32,a2
		jsr	sub_9F9A
		move.w	d0,($FFFFD44E).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_F9FEA


; =============== S U B	R O U T	I N E =======================================


sub_FA014:							; CODE XREF: sub_F9AE4+3Cp
					; sub_F9B2A+40p ...
		movem.l	d0-d3/a0-a3,-(sp)
		move.l	a1,a2
		tst.w	d2
		beq.w	loc_FA060
		move.w	#$B,d0
		clr.w	d3
		move.l	#$FFFFD45A,a0
		mulu.w	#$C,d2
		add.l	d2,a0

loc_FA032:							; CODE XREF: sub_FA014+32j
		move.b	(a0,d3.w),d1
		bne.w	loc_FA03E
		move.b	#$20,d1					; ' '

loc_FA03E:							; CODE XREF: sub_FA014+22j
		move.b	d1,2(a1)
		tst.b	(a1)+
		addq.w	#1,d3
		dbf	d0,loc_FA032
		move.w	#$E,(a2)
		btst	#7,($FFFFC2F8).w
		beq.w	loc_FA05C
		bsr.w	sub_FAF66

loc_FA05C:							; CODE XREF: sub_FA014+40j
		bra.w	loc_FA076
; ---------------------------------------------------------------------------

loc_FA060:							; CODE XREF: sub_FA014+8j
		move.l	a1,a3
		move.l	a3,-(sp)
		move.l	#word_FA07C,a1
		jsr	sub_F997A
		move.l	(sp)+,a2
		bsr.w	sub_FAF66

loc_FA076:							; CODE XREF: sub_FA014:loc_FA05Cj
		movem.l	(sp)+,d0-d3/a0-a3
		rts
; End of function sub_FA014

; ---------------------------------------------------------------------------
word_FA07C:	dc.w	2					; DATA XREF: sub_FA014+50o
; ---------------------------------------------------------------------------

loc_FA07E:							; DATA XREF: ROM:000196C8o
					; ROM:0001983Co ...
		tst.w	($FFFFD458).w
		bpl.w	loc_FA088
		rts
; ---------------------------------------------------------------------------

loc_FA088:							; CODE XREF: ROM:000FA082j
		movem.l	d0-a6,-(sp)
		jsr	sub_10F32
		bclr	#0,($FFFFBF78).w
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.w	($FFFFB002).w,d0
		jsr	VDP_SetAddress
		move.l	#0,(a0)
		move.w	#$8C81,4(a0)
		move.w	#6,($FFFFB00E).w
		move.w	#$8D00,4(a0)
		clr.w	d0
		jsr	VDP_SetAddress
		move.l	#0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		bclr	#1,($FFFFBF78).w
		bsr.w	sub_F9FC0
		bsr.w	sub_F9FEA
		move.w	($FFFFB012).w,d4
		move.l	#unk_AAC5A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$523
		dc.w	$4167
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_E6632,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#$28,d2					; '('
		move.w	#$1C,d3
		moveq	#$D,d5
		move.w	#1,d4
		jsr	sub_1169A
		move.w	d4,($FFFFD450).w
		add.w	#$24,d4					; '$'
		move.w	d4,($FFFFD452).w
		add.w	#$24,d4					; '$'
		move.w	d4,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$40,d0					; '@'
		moveq	#$20,d1					; ' '
		move.w	#$7FF,d2
		jsr	sub_1197E
		move.w	(sp)+,d4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_E9A80,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#8,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$21
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_E9ED6,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#0,d5
		jsr	sub_1169A
		bsr.w	sub_FAFBA
		move.w	#$18,($FFFFBD26).w
		clr.w	($FFFFD4F0).w
		clr.w	($FFFFD4F2).w
		tst.w	($FFFFD046).w
		beq.w	loc_FA1DA
		cmp.w	#3,($FFFFC316).w
		beq.w	loc_FA1EC
		bra.w	loc_FA1E4
; ---------------------------------------------------------------------------

loc_FA1DA:							; CODE XREF: ROM:000FA1C8j
		btst	#1,($FFFFC2EC).w
		beq.w	loc_FA1EC

loc_FA1E4:							; CODE XREF: ROM:000FA1D6j
		move.w	($FFFFC32A).w,d0
		bra.w	loc_FA1F0
; ---------------------------------------------------------------------------

loc_FA1EC:							; CODE XREF: ROM:000FA1D2j
					; ROM:000FA1E0j
		move.w	($FFFFC328).w,d0

loc_FA1F0:							; CODE XREF: ROM:000FA1E8j
		subq.w	#1,d0
		bpl.w	loc_FA1F8
		clr.w	d0

loc_FA1F8:							; CODE XREF: ROM:000FA1F2j
		and.w	#1,d0
		move.l	#$FFFFD4F2,a0
		move.w	d0,($FFFFD4F6).w
		bne.w	loc_FA210
		move.l	#$FFFFD4F0,a0

loc_FA210:							; CODE XREF: ROM:000FA206j
		bsr.w	sub_FA2C0

loc_FA214:							; CODE XREF: ROM:000FA268j
					; ROM:000FA294j
		jsr	sub_80BA
		jsr	sub_A41E
		jsr	sub_11318
		btst	#7,d3
		bne.w	loc_FA298
		btst	#6,d1
		beq.w	loc_FA242
		eor.w	#1,($FFFFD4F6).w
		clr.w	d0
		bra.w	loc_FA26A
; ---------------------------------------------------------------------------

loc_FA242:							; CODE XREF: ROM:000FA232j
		move.w	#1,d0
		btst	#5,d1
		bne.w	loc_FA26A
		moveq	#1,d0
		jsr	sub_112F0
		btst	#3,d1
		bne.w	loc_FA26A
		neg.w	d0
		btst	#2,d1
		bne.w	loc_FA26A
		bra.s	loc_FA214
; ---------------------------------------------------------------------------

loc_FA26A:							; CODE XREF: ROM:000FA23Ej
					; ROM:000FA24Aj ...
		move.l	#$FFFFD4F0,a0
		move.w	($FFFFD44C).w,($FFFFD4F8).w
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA28A
		move.l	#$FFFFD4F2,a0
		move.w	($FFFFD44E).w,($FFFFD4F8).w

loc_FA28A:							; CODE XREF: ROM:000FA27Aj
		add.w	d0,(a0)
		bsr.w	sub_FA2A2
		bsr.w	sub_FA2C0
		bra.w	loc_FA214
; ---------------------------------------------------------------------------

loc_FA298:							; CODE XREF: ROM:000FA22Aj
		movem.l	(sp)+,d0-a6
		jmp	loc_9CD8

; =============== S U B	R O U T	I N E =======================================


sub_FA2A2:							; CODE XREF: ROM:000FA28Cp
		move.w	d0,-(sp)
		move.w	($FFFFD4F8).w,d0
		addq.w	#1,d0
		cmp.w	(a0),d0
		bne.w	loc_FA2B2
		clr.w	(a0)

loc_FA2B2:							; CODE XREF: sub_FA2A2+Aj
		tst.w	(a0)
		bpl.w	loc_FA2BC
		subq.w	#1,d0
		move.w	d0,(a0)

loc_FA2BC:							; CODE XREF: sub_FA2A2+12j
		move.w	(sp)+,d0
		rts
; End of function sub_FA2A2


; =============== S U B	R O U T	I N E =======================================


sub_FA2C0:							; CODE XREF: ROM:loc_FA210p
					; ROM:000FA290p
		movem.l	d0-d7/a1-a6,-(sp)
		move.l	a0,-(sp)
		move.w	($FFFFC330).w,d0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA2D6
		move.w	($FFFFC332).w,d0

loc_FA2D6:							; CODE XREF: sub_FA2C0+Ej
		asl.w	#6,d0
		move.l	#word_F8BF4,a6
		move.l	$26(a6,d0.w),($FFFFBD4A).w
		move.w	#$64,($FFFFBD26).w			; 'd'
		jsr	sub_11B92
; End of function sub_FA2C0

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$28,d0					; '('
		move.w	#$15,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	(sp),a0
		tst.w	(a0)
		bne.w	loc_FA324
		bsr.w	loc_FA550
		bra.w	loc_FA328
; ---------------------------------------------------------------------------

loc_FA324:							; CODE XREF: ROM:000FA318j
		bsr.w	loc_FA330

loc_FA328:							; CODE XREF: ROM:000FA320j
		move.l	(sp)+,a0
		movem.l	(sp)+,d0-d7/a1-a6
		rts
; ---------------------------------------------------------------------------

loc_FA330:							; CODE XREF: ROM:loc_FA324p
		movem.l	d0-a6,-(sp)
		move.w	(a0),($FFFFD4F4).w
		bsr.w	loc_FAEBC
		move.w	#$21,d0					; '!'
		move.w	#8,d1
		jsr	sub_FA66E
		move.w	#$64,($FFFFBD26).w			; 'd'
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA36C
		move.w	($FFFFC332).w,d0

loc_FA36C:							; CODE XREF: ROM:000FA364j
		move.w	d0,($FFFFBF14).w
		move.l	#$30E,a2
		asl.w	#2,d0
		move.l	(a2,d0.w),a2
		add.w	(a2),a2
		bsr.w	sub_FAF8C
		bra.w	loc_FA38A
; ---------------------------------------------------------------------------

loc_FA386:							; CODE XREF: ROM:loc_FA38Aj
		add.w	(a2),a2
		addq.w	#8,a2

loc_FA38A:							; CODE XREF: ROM:000FA382j
		dbf	d0,loc_FA386
		clr.w	d7
		move.l	#$FFFFCF36,a3
		move.l	a2,a1
		bsr.w	sub_F997A
		move.l	#$FFFFCF38,a3

loc_FA3A2:							; CODE XREF: ROM:000FA3A8j
		addq.w	#1,d7
		cmp.b	#$20,(a3)+				; ' '
		bne.s	loc_FA3A2
		subq.w	#1,d7
		move.b	#0,-(a3)
		move.l	#$FFFFCF36,a3
		addq.w	#1,d7
		and.w	#$FE,d7					; ''
		addq.w	#2,d7
		move.w	d7,(a3)
		move.l	a3,a1
		move.w	#8,($FFFFB028).l
		move.w	#$B,($FFFFB02A).l
		jsr	sub_11A48
		bsr.w	sub_FAF8C
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA3F0
		move.l	#$FFFFCA32,a2

loc_FA3F0:							; CODE XREF: ROM:000FA3E6j
		jsr	sub_18B5E
		move.w	#8,($FFFFB028).l
		move.w	#$C,($FFFFB02A).l
		jsr	sub_11A48
		move.w	#8,d1
		jsr	sub_FAC3E
		move.w	#$15,($FFFFB028).l
		move.w	#$C,($FFFFB02A).l
		move.w	($FFFFB028).w,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aRecords:	dc.b	'  Records',0
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	'-----------',0
; ---------------------------------------------------------------------------
		move.w	#$15,($FFFFB028).l
		move.w	#$12,($FFFFB02A).l
		move.w	($FFFFB028).w,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
aThisGame:	dc.b	' This Game ',0
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	'-----------',0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d1
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA4AC
		move.w	($FFFFC332).w,d1

loc_FA4AC:							; CODE XREF: ROM:000FA4A4j
		ext.l	d1
		bsr.w	sub_FAF8C
		ext.l	d0
		move.l	#$FFFFCF36,a0
		bsr.w	sub_F9B94
		move.l	#$FFFFCF36,a0
		move.b	(a0),d2
		move.b	1(a0),d3
		move.b	3(a0),d4
		move.b	2(a0),d5
		move.w	#$E,d1
		move.w	($FFFFD448).w,d0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA4E6
		move.w	($FFFFD44A).w,d0

loc_FA4E6:							; CODE XREF: ROM:000FA4DEj
		cmp.w	($FFFFD52E).w,d0
		ble.w	loc_FA50E
		move.w	#$15,d0
		bsr.w	sub_FA708
		bsr.w	sub_FAF8C
		move.w	#$15,d1
		move.w	#$14,d2
		bsr.w	sub_FAD84
		st	($FFFFD598).w
		bra.w	loc_FA52A
; ---------------------------------------------------------------------------

loc_FA50E:							; CODE XREF: ROM:000FA4EAj
		move.w	#$15,d0
		bsr.w	sub_FA75C
		bsr.w	sub_FAF8C
		move.w	#$15,d1
		move.w	#$14,d2
		bsr.w	loc_FACF0
		clr.w	($FFFFD598).w

loc_FA52A:							; CODE XREF: ROM:000FA50Aj
		move.w	#3,d1
		move.w	#$11,d2
		bsr.w	sub_FAF8C
		bsr.w	sub_FA8AC
		move.w	#3,d1
		move.w	#$F,d2
		bsr.w	sub_FAF8C
		bsr.w	loc_FAB4C
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_FA550:							; CODE XREF: ROM:000FA31Cp
		movem.l	d0-a6,-(sp)
		move.w	#$21,d0					; '!'
		move.w	#8,d1
		jsr	sub_FA66E
		move.w	#1,d0
		move.w	#8,d1
		bset	#0,($FFFFC2F8).w
		jsr	sub_FA66E
		bclr	#0,($FFFFC2F8).w
		move.w	#$64,($FFFFBD26).w			; 'd'
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d1
		jsr	sub_FAC3E
		move.w	#3,d0
		move.w	#$F,d1
		jsr	loc_FA936
		move.w	#8,d0
		move.w	#$B,d1
		jsr	sub_FABDC
		move.w	#$15,($FFFFB028).w
		move.w	#$C,($FFFFB02A).w
		move.w	($FFFFB028).w,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aRecords_0:	dc.b	'  Records',0
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	'-----------',0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d1
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA602
		move.w	($FFFFC332).w,d1

loc_FA602:							; CODE XREF: ROM:000FA5FAj
		ext.l	d1
		move.l	#$FFFFCF36,a0
		move.l	a0,-(sp)
		jsr	sub_F9BE2
		move.w	#$15,d0
		move.w	#$E,d1
		move.b	(a0),d2
		move.b	1(a0),d3
		move.b	3(a0),d4
		move.b	2(a0),d5
		bsr.w	sub_FA75C
		move.l	(sp),a0
		move.w	#$15,d0
		move.w	#$12,d1
		move.b	4(a0),d2
		move.b	5(a0),d3
		move.b	7(a0),d4
		move.b	6(a0),d5
		bsr.w	sub_FA708
		move.l	(sp)+,a0
		move.w	#$15,d0
		move.w	#$16,d1
		move.b	8(a0),d2
		move.b	9(a0),d3
		move.b	$B(a0),d4
		move.b	$A(a0),d5
		bsr.w	sub_FA6E8
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FA66E:							; CODE XREF: ROM:000FA344p
					; ROM:000FA55Cp ...
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; End of function sub_FA66E

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	d0,($FFFFB028).w
		move.w	d1,($FFFFB02A).w
		move.w	($FFFFD450).w,d4
		move.w	($FFFFC330).w,d3
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA69A
		move.w	($FFFFC332).w,d3

loc_FA69A:							; CODE XREF: ROM:000FA692j
		asl.w	#2,d3
		move.l	#off_F86F2,a0
		move.l	(a0,d3.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d3
		move.l	#unk_FF462,a0
		sub.w	#$20,d3					; ' '
		add.w	d3,a0
		add.l	(a2)+,a1
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		move.l	($FFFFBD4A).w,-(sp)
		move.l	($FFFFBD4E).w,-(sp)
		move.w	#2,d5
		jsr	sub_1169A
		move.l	(sp)+,($FFFFBD4E).w
		move.l	(sp)+,($FFFFBD4A).w
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FA6E8:							; CODE XREF: ROM:000FA664p
		move.l	#word_FA6F6,($FFFFBF18).l
		bra.w	loc_FA7B0
; End of function sub_FA6E8

; ---------------------------------------------------------------------------
word_FA6F6:	dc.w	$12					; DATA XREF: sub_FA6E8o
aDbCrowdLevel:	dc.b	' dB Crowd Level',0

; =============== S U B	R O U T	I N E =======================================


sub_FA708:							; CODE XREF: ROM:000FA4F2p
					; ROM:000FA646p
		movem.l	a1-a3,-(sp)
		move.l	#word_FA734,a1
		tst.w	d3
		beq.w	loc_FA71E
		move.l	#word_FA746,a1

loc_FA71E:							; CODE XREF: sub_FA708+Cj
		cmpnb	 1,d2
		bne.w	loc_FA728
		add.w	(a1),a1

loc_FA728:							; CODE XREF: sub_FA708+1Aj
		move.l	a1,($FFFFBF18).w
		movem.l	(sp)+,a1-a3
		bra.w	loc_FA7B0
; End of function sub_FA708

; ---------------------------------------------------------------------------
word_FA734:	dc.w	$A					; DATA XREF: sub_FA708+4o
aSaves_0:	dc.b	' Saves ',0
		dc.w	8
aSave_0:	dc.b	' Save '
word_FA746:	dc.w	$C					; DATA XREF: sub_FA708+10o
aSavesBy:	dc.b	' Saves by',0
		dc.w	$A
aSaveBy:	dc.b	' Save by'

; =============== S U B	R O U T	I N E =======================================


sub_FA75C:							; CODE XREF: ROM:000FA512p
					; ROM:000FA628p
		movem.l	a1-a3,-(sp)
		move.l	#word_FA788,a1
		tst.w	d3
		beq.w	loc_FA772
		move.l	#word_FA79A,a1

loc_FA772:							; CODE XREF: sub_FA75C+Cj
		cmpnb	 1,d2
		bne.w	loc_FA77C
		add.w	(a1),a1

loc_FA77C:							; CODE XREF: sub_FA75C+1Aj
		move.l	a1,($FFFFBF18).w
		movem.l	(sp)+,a1-a3
		bra.w	loc_FA7B0
; ---------------------------------------------------------------------------
word_FA788:	dc.w	$A					; DATA XREF: sub_FA75C+4o
aGoals_1:	dc.b	' Goals ',0
		dc.w	8
aGoal_1:	dc.b	' Goal '
word_FA79A:	dc.w	$C					; DATA XREF: sub_FA75C+10o
aGoalsBy:	dc.b	' Goals by',0
		dc.w	$A
aGoalBy:	dc.b	' Goal by'
; ---------------------------------------------------------------------------

loc_FA7B0:							; CODE XREF: sub_FA6E8+Aj
					; sub_FA708+28j ...
		tst.b	d2
		bne.w	loc_FA7B8
		rts
; ---------------------------------------------------------------------------

loc_FA7B8:							; CODE XREF: sub_FA75C+56j
		move.w	d0,($FFFFB028).w
		move.w	d1,($FFFFB02A).w
		move.w	d2,d0
		and.w	#$FF,d0
		move.l	#$FFFFBFA4,a1
		move.l	a1,-(sp)
		jsr	sub_F998E
		move.l	(sp)+,a1
		move.l	a1,a3
		move.l	($FFFFBF18).w,a1
		jsr	sub_11D9E
		move.w	($FFFFB028).w,-(sp)
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		move.l	#$FFFFBFA4,a1
		move.w	d3,d2
		ext.w	d2
		bset	#7,($FFFFC2F8).w
		jsr	sub_FA014
		move.l	a1,a3
		move.l	#word_FA876,a1
		jsr	sub_11D9E
		move.w	($FFFFB028).w,-(sp)
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		move.w	d4,d2
		ext.w	d2
		move.l	#$FFFFBFA4,a1
		bset	#7,($FFFFC2F8).w
		jsr	sub_FA014
		move.l	a1,a3
		move.l	#word_FA87C,a1
		jsr	sub_11D9E
		move.w	d5,d0
		move.l	#$FFFFBFA4,a1
		jsr	sub_FA880
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		rts
; End of function sub_FA75C

; ---------------------------------------------------------------------------
word_FA876:	dc.w	6					; DATA XREF: sub_FA75C+B6o
aVs__0:		dc.b	' vs.'
word_FA87C:	dc.w	4					; DATA XREF: sub_FA75C+F2o
		dc.b	' ',0

; =============== S U B	R O U T	I N E =======================================


sub_FA880:							; CODE XREF: sub_F9B2A+5Ap
					; sub_FA75C+106p ...
		movem.l	d0/a0-a3,-(sp)
		ext.w	d0
		asl.w	#2,d0
		move.l	#$30E,a0
		move.l	(a0,d0.w),a0
		move.w	4(a0),d0
		ext.l	d0
		add.l	d0,a0
		add.w	(a0),a0
		move.l	a1,a3
		move.l	a0,a1
		jsr	sub_11D9E
		movem.l	(sp)+,d0/a0-a3
		rts
; End of function sub_FA880


; =============== S U B	R O U T	I N E =======================================


sub_FA8AC:							; CODE XREF: ROM:000FA536p
		movem.l	d0-a6,-(sp)
		move.w	d1,($FFFFB028).w
		move.w	d2,($FFFFB02A).w
		move.w	d1,-(sp)
		jsr	sub_11A36
; End of function sub_FA8AC

; ---------------------------------------------------------------------------
		dc.w	$A
aOverall_1:	dc.b	'Overall',0
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aRating_0:	dc.b	'Rating   ',0
; ---------------------------------------------------------------------------
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD4F6).w
		beq.w	loc_FA8F8
		move.l	#$FFFFCA32,a2

loc_FA8F8:							; CODE XREF: ROM:000FA8EEj
		move.l	(dword_19420).l,d4
		tst.w	($FFFFD598).w
		beq.w	loc_FA90C
		move.l	(dword_19582).l,d4

loc_FA90C:							; CODE XREF: ROM:000FA902j
		bsr.w	sub_FA9F8
		mulu.w	#$64,d0					; 'd'
		divu.w	d1,d0
		bsr.w	sub_FEF7C
		move.l	#$FFFFBFA4,a1
		bsr.w	sub_F998E
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		movem.l	(sp)+,d0-a6

; =============== S U B	R O U T	I N E =======================================


nullsub_6:
		rts
; End of function nullsub_6

; ---------------------------------------------------------------------------

loc_FA936:							; CODE XREF: ROM:000FA5A0p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFC330).w,d2
		bsr.w	sub_FEE4A
		tst.w	($FFFFD4F6).w

loc_FA946:
		beq.w	loc_FA952
		move.w	($FFFFC332).w,d2
		bsr.w	sub_FEE4A

loc_FA952:							; CODE XREF: ROM:loc_FA946j
		move.l	#$30E,a0
		asl.w	#2,d2
		ext.l	d2
		add.l	d2,a0
		move.l	(a0),a0
		move.w	6(a0),d2
		ext.l	d2
		add.l	d2,a0
		move.w	d0,($FFFFB028).w
		move.w	d1,($FFFFB02A).w
		move.w	d0,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$10
aStartingLine:	dc.b	'Starting Line',0
; ---------------------------------------------------------------------------
		addq.w	#2,($FFFFB02A).w
		move.w	(sp),($FFFFB028).w
		move.l	#word_FA9D2,a1

loc_FA996:							; CODE XREF: ROM:000FA9C8j
		cmp.w	#2,(a1)
		beq.w	loc_FA9CA
		move.w	(sp),($FFFFB028).w
		move.l	a1,-(sp)
		jsr	sub_11A48
		move.b	(a0)+,d0
		ext.w	d0
		subq.w	#1,d0
		jsr	sub_18B6E
		jsr	sub_11A48
		move.l	(sp)+,a1
		move.w	(a1),d0
		ext.l	d0
		add.l	d0,a1
		addq.w	#1,($FFFFB02A).w
		bra.s	loc_FA996
; ---------------------------------------------------------------------------

loc_FA9CA:							; CODE XREF: ROM:000FA99Aj
		move.w	(sp)+,d0
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------
word_FA9D2:	dc.w	6					; DATA XREF: ROM:000FA990o
aG_0:		dc.b	'G   '
		dc.w	6
aLd_0:		dc.b	'LD  '
		dc.w	6
aRd_0:		dc.b	'RD  '
		dc.w	6
aLw_0:		dc.b	'LW  '
		dc.w	6
		dc.b	'C   '
		dc.w	6
aRw_0:		dc.b	'RW  '
		dc.w	2

; =============== S U B	R O U T	I N E =======================================


sub_FA9F8:							; CODE XREF: sub_8D4E+22p
					; ROM:loc_FA90Cp ...
		move.l	a6,-(sp)
		clr.w	($FFFFD6CE).w
		clr.w	($FFFFD6D0).w
		move.l	#unk_FAB3C,a6
		cmp.l	(dword_19420).l,d4
		bne.w	loc_FAA18
		move.l	#unk_FAB1C,a6

loc_FAA18:							; CODE XREF: sub_FA9F8+16j
		cmp.l	(dword_19582).l,d4
		bne.w	loc_FAA28
		move.l	#unk_FAB2C,a6

loc_FAA28:							; CODE XREF: sub_FA9F8+26j
		move.l	$1E(a2),a0
		lea	$1A2(a2),a4
		clr.l	d1
		move.w	d0,d1
		asl.w	#4,d1
		add.l	d1,a4
		add.l	#$10,a4
		add.w	(a0),a0

loc_FAA40:							; CODE XREF: sub_FA9F8+4Cj
		add.w	(a0),a0
		addq.w	#8,a0
		dbf	d0,loc_FAA40
		clr.w	d0
		clr.w	d1
		moveq	#$F,d2
		swap	d4

loc_FAA50:							; CODE XREF: sub_FA9F8:loc_FAAE6j
		btst	d2,d4
		beq.w	loc_FAAE6
		move.w	d2,d3
		lsr.w	#1,d3
		neg.w	d3
		move.b	-1(a0,d3.w),d3
		btst	#0,d2
		beq.w	loc_FAA6A
		lsr.w	#4,d3

loc_FAA6A:							; CODE XREF: sub_FA9F8+6Cj
		and.w	#$F,d3
		cmpnw	 $D,d2
		bne.w	loc_FAA7A
		bra.w	loc_FAAE0
; ---------------------------------------------------------------------------

loc_FAA7A:							; CODE XREF: sub_FA9F8+7Aj
		cmpnw	 6,d2
		beq.w	loc_FAAE0
		movem.l	d5-d7,-(sp)
		move.b	(a6,d2.w),d5
		ext.w	d5
		cmpnw	 2,d5
		beq.w	loc_FAAA2
		move.w	d3,-(sp)
		subq.w	#2,d5

loc_FAA98:							; CODE XREF: sub_FA9F8+A2j
		add.w	(sp),d3
		dbf	d5,loc_FAA98
		asr.w	#1,d3
		tst.w	(sp)+

loc_FAAA2:							; CODE XREF: sub_FA9F8+98j
		movem.l	(sp)+,d5-d7
		cmp.l	#unk_FAB3C,a6
		beq.w	loc_FAACA
		neg.w	d2
		move.w	d7,-(sp)
		move.b	-1(a4,d2.w),d7
		ext.w	d7
		add.w	d7,($FFFFD6CE).w
		addq.w	#1,($FFFFD6D0).w
		move.w	(sp)+,d7
		neg.w	d2
		bra.w	loc_FAAE0
; ---------------------------------------------------------------------------

loc_FAACA:							; CODE XREF: sub_FA9F8+B4j
		move.w	d3,-(sp)
		asl.w	#4,d3
		add.w	(sp),d3
		add.w	(sp)+,d3
		neg.w	d2
		add.b	-1(a4,d2.w),d3
		bpl.w	loc_FAADE
		clr.b	d3

loc_FAADE:							; CODE XREF: sub_FA9F8+E0j
		neg.w	d2

loc_FAAE0:							; CODE XREF: sub_FA9F8+7Ej
					; sub_FA9F8+86j ...
		add.w	d3,d0
		add.w	#$64,d1					; 'd'

loc_FAAE6:							; CODE XREF: sub_FA9F8+5Aj
		dbf	d2,loc_FAA50
		cmp.l	#unk_FAB3C,a6
		beq.w	loc_FAB0E
		move.w	#$64,d1					; 'd'
		movem.l	d6-d7,-(sp)
		move.w	($FFFFD6CE).w,d6
		ext.l	d6
		move.w	($FFFFD6D0).w,d7
		divs.w	d7,d6
		add.w	d6,d0
		movem.l	(sp)+,d6-d7

loc_FAB0E:							; CODE XREF: sub_FA9F8+F8j
		cmp.w	d1,d0
		blt.w	loc_FAB18
		move.w	d0,d1
		subq.w	#1,d0

loc_FAB18:							; CODE XREF: sub_FA9F8+118j
		move.l	(sp)+,a6
		rts
; End of function sub_FA9F8

; ---------------------------------------------------------------------------
unk_FAB1C:	dc.b	2					; DATA XREF: sub_FA9F8+1Ao
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	4
		dc.b	6
		dc.b	2
		dc.b	4
		dc.b	2
		dc.b	4
		dc.b	6
		dc.b	6
		dc.b	4
		dc.b	2
		dc.b	2
		dc.b	2
unk_FAB2C:	dc.b	2					; DATA XREF: sub_FA9F8+2Ao
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	9
		dc.b	9
		dc.b	2
		dc.b	2
		dc.b	9
		dc.b	2
		dc.b	2
		dc.b	2
unk_FAB3C:	dc.b	2					; DATA XREF: sub_FA9F8+Ao
					; sub_FA9F8+AEo ...
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
		dc.b	2
; ---------------------------------------------------------------------------

loc_FAB4C:							; CODE XREF: ROM:000FA546p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFBFA4,a1
		bsr.w	sub_FAC90
		move.w	d1,($FFFFB028).w
		move.w	d2,($FFFFB02A).w
		jsr	sub_11A48
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD4F6).w
		beq.w	loc_FAB7C
		move.l	#$FFFFCA32,a2

loc_FAB7C:							; CODE XREF: ROM:000FAB72j
		move.w	d0,-(sp)
		jsr	sub_9F5C
		cmp.w	(sp),d0
		ble.w	loc_FABC0
		move.w	(sp),d0
		jsr	sub_9F40
		cmp.w	(sp),d0
		ble.w	loc_FABAC
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
aGoalie:	dc.b	' Goalie',0
; ---------------------------------------------------------------------------
		bra.w	loc_FABD4
; ---------------------------------------------------------------------------

loc_FABAC:							; CODE XREF: ROM:000FAB94j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
aForward:	dc.b	' Forward'
; ---------------------------------------------------------------------------
		bra.w	loc_FABD4
; ---------------------------------------------------------------------------

loc_FABC0:							; CODE XREF: ROM:000FAB86j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
aDefenseman:	dc.b	' Defenseman',0
; ---------------------------------------------------------------------------

loc_FABD4:							; CODE XREF: ROM:000FABA8j
					; ROM:000FABBCj
		tst.w	(sp)+
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FABDC:							; CODE XREF: ROM:000FA5AEp
		movem.l	d0-a6,-(sp)
		move.w	d0,($FFFFB028).w
		move.w	d1,($FFFFB02A).w
		move.w	d0,-(sp)
		jsr	sub_11A36
; End of function sub_FABDC

; ---------------------------------------------------------------------------
		dc.w	6
aTeam:		dc.b	'Team'
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
aRating_1:	dc.b	'Rating  '
; ---------------------------------------------------------------------------
		move.l	#$FFFFC6CE,a0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FAC22
		move.l	#$FFFFCA32,a0

loc_FAC22:							; CODE XREF: ROM:000FAC18j
		move.l	a0,a2
		bsr.w	sub_FE172
		move.w	#2,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FAC3E:							; CODE XREF: ROM:000FA410p
					; ROM:000FA592p
		movem.l	d0-a6,-(sp)
		move.w	d1,($FFFFB02A).w
		move.l	#$FFFFC6CE,a0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FAC5A
		move.l	#$FFFFCA32,a0

loc_FAC5A:							; CODE XREF: sub_FAC3E+12j
		move.l	$1E(a0),a0
		add.w	4(a0),a0

loc_FAC62:
		bsr.w	sub_FAC78
		add.w	(a0),a0
		add.w	(a0),a0
		addq.w	#1,($FFFFB02A).w
		bsr.w	sub_FAC78

loc_FAC72:
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FAC3E


; =============== S U B	R O U T	I N E =======================================


sub_FAC78:							; CODE XREF: sub_FAC3E:loc_FAC62p
					; sub_FAC3E+30p
		move.w	#$14,($FFFFB028).w
		move.w	(a0),d0
		subq.w	#2,d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		move.l	a0,a1
		jmp	sub_11A48
; End of function sub_FAC78


; =============== S U B	R O U T	I N E =======================================


sub_FAC90:							; CODE XREF: ROM:000FAB56p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFC6CE,a0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FACA8
		move.l	#$FFFFCA32,a0

loc_FACA8:							; CODE XREF: sub_FAC90+Ej
		move.l	$1E(a0),a0
		move.w	(a0),d1
		ext.l	d1
		add.l	d1,a0

loc_FACB2:							; CODE XREF: sub_FAC90+2Aj
		move.w	(a0),d1
		ext.l	d1
		add.l	d1,a0
		addq.l	#8,a0
		dbf	d0,loc_FACB2
		subq.l	#8,a0
		move.w	#4,(a1)+
		move.b	(a0),d0
		lsr.w	#4,d0
		and.w	#$F,d0
		bne.w	loc_FACD8
		move.b	#$20,d0					; ' '
		bra.w	loc_FACDC
; ---------------------------------------------------------------------------

loc_FACD8:							; CODE XREF: sub_FAC90+3Cj
		add.b	#$30,d0					; '0'

loc_FACDC:							; CODE XREF: sub_FAC90+44j
		move.b	d0,(a1)+
		move.b	(a0),d0
		and.w	#$F,d0
		add.b	#$30,d0					; '0'
		move.b	d0,(a1)+
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FAC90

; ---------------------------------------------------------------------------

loc_FACF0:							; CODE XREF: ROM:000FA522p
		movem.l	d0-a6,-(sp)
		bsr.w	sub_FAF50
		move.l	a0,-(sp)
		add.l	#$B4,a0					; ''
		move.b	(a0,d0.w),d3
		ext.w	d3
		move.l	(sp)+,a0
		add.l	#$CE,a0					; ''
		move.b	(a0,d0.w),d4
		ext.w	d4
		move.w	d1,($FFFFB028).w
		move.w	d2,($FFFFB02A).w
		move.w	d1,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aGoals_2:	dc.b	'Goals    ',0
; ---------------------------------------------------------------------------
		move.w	d0,-(sp)
		move.w	d3,d0
		move.l	#$FFFFBFA4,a1
		bsr.w	sub_F998E
		move.w	(sp)+,d0
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aAssists_0:	dc.b	'Assists  ',0
; ---------------------------------------------------------------------------
		move.w	d4,d0
		move.l	#$FFFFBFA4,a1
		bsr.w	sub_F998E
		move.l	#$FFFFBFA4,a1
		jsr	sub_11A48
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FAD84:							; CODE XREF: ROM:000FA502p
		movem.l	d0-a6,-(sp)
		bsr.w	sub_FAF50
		move.l	a0,-(sp)
		add.l	#$B4,a0					; ''

loc_FAD94:
		move.b	(a0,d0.w),d3
; End of function sub_FAD84

		ext.w	d3
		move.l	(sp)+,a0
		add.l	#$E8,a0					; ''
		move.b	(a0,d0.w),d4
		ext.w	d4
		move.w	d4,d7
		sub.w	d3,d7
		move.w	d1,($FFFFB028).w
		move.w	d2,($FFFFB02A).w
		move.w	d1,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aSaves_1:	dc.b	'Saves    ',0
; ---------------------------------------------------------------------------
		move.w	d7,d0
		move.l	#$FFFFBFA4,a1
		move.l	a1,-(sp)
		bsr.w	sub_F998E
		move.l	(sp)+,a1
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
aSave_1:	dc.b	'Save %   ',0
; ---------------------------------------------------------------------------
		move.w	d7,d0
		mulu.w	#$64,d0					; 'd'
		tst.w	d4
		bne.w	loc_FAE0A
		clr.w	d0
		bra.w	loc_FAE0C
; ---------------------------------------------------------------------------

loc_FAE0A:							; CODE XREF: ROM:000FAE00j
		divu.w	d4,d0

loc_FAE0C:							; CODE XREF: ROM:000FAE06j
		move.l	#$FFFFBFA4,a1
		move.l	a1,-(sp)
		bsr.w	sub_F998E
		move.l	(sp)+,a1
		jsr	sub_11A48
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FAE26:							; CODE XREF: ROM:loc_FAED4p
					; sub_FD084+6Ap ...
		movem.l	d0-d7/a1-a6,-(sp)
		move.l	#off_F92F4,a0
		move.w	d1,d2
		asl.w	#2,d2
		move.l	(a0,d2.w),a0
		move.w	#$FFFF,d2
		bra.w	loc_FAE50
; ---------------------------------------------------------------------------

loc_FAE40:							; CODE XREF: sub_FAE26+2Ej
		cmp.w	4(a0),d0
		bne.w	loc_FAE4E
		move.l	(a0),a0
		bra.w	loc_FAEB6
; ---------------------------------------------------------------------------

loc_FAE4E:							; CODE XREF: sub_FAE26+1Ej
		addq.l	#6,a0

loc_FAE50:							; CODE XREF: sub_FAE26+16j
		addq.w	#1,d2
		tst.l	(a0)
		bne.s	loc_FAE40
		move.l	#$30E,a0
		asl.w	#2,d1
		move.l	(a0,d1.w),a0
		move.l	a0,a6
		move.w	d0,d6
		add.w	(a6),a6
		bra.w	loc_FAE70
; ---------------------------------------------------------------------------

loc_FAE6C:							; CODE XREF: sub_FAE26:loc_FAE70j
		add.w	(a6),a6
		addq.w	#8,a6

loc_FAE70:							; CODE XREF: sub_FAE26+42j
		dbf	d6,loc_FAE6C
		add.w	(a6),a6
		add.w	$A(a0),a0
		move.w	(a0),d1
		clr.w	d3

loc_FAE7E:							; CODE XREF: sub_FAE26+5Cj
		addq.w	#1,d3
		asl.w	#4,d1
		bne.s	loc_FAE7E
		move.l	#unk_C6F02,a0
		btst	#0,4(a6)
		bne.w	loc_FAE9A
		move.l	#unk_C726C,a0

loc_FAE9A:							; CODE XREF: sub_FAE26+6Aj
		cmp.w	d3,d0
		blt.w	loc_FAEB6
		move.l	#unk_C6B98,a0
		btst	#0,4(a6)
		bne.w	loc_FAEB6
		move.l	#unk_C682E,a0

loc_FAEB6:							; CODE XREF: sub_FAE26+24j
					; sub_FAE26+76j ...
		movem.l	(sp)+,d0-d7/a1-a6
		rts
; End of function sub_FAE26

; ---------------------------------------------------------------------------

loc_FAEBC:							; CODE XREF: ROM:000FA338p
		movem.l	d0-a6,-(sp)
		bsr.w	sub_FAF8C
		move.w	($FFFFC330).w,d1
		tst.w	($FFFFD4F6).w
		beq.w	loc_FAED4
		move.w	($FFFFC332).w,d1

loc_FAED4:							; CODE XREF: ROM:000FAECCj
		bsr.w	sub_FAE26
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$8E
		dc.b	1
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD452).w,d4
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		move.l	#off_C63F8,a0
		add.l	(a0),a0
		tst.l	(a2)
		bne.w	loc_FAF0C
		move.l	#off_C63F8,a1
		add.l	4(a1),a1
		tst.l	(a2)+
		bra.w	loc_FAF0E
; ---------------------------------------------------------------------------

loc_FAF0C:							; CODE XREF: ROM:000FAEF8j
		add.l	(a2)+,a1

loc_FAF0E:							; CODE XREF: ROM:000FAF08j
		bsr.w	sub_FE98A
		move.l	#$FFFFDA1E,a2
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		movem.l	d0/a0-a1,-(sp)
		add.w	#$20,a0					; ' '
		move.l	#$FFFFBD68,a1
		move.w	#7,d0

loc_FAF36:							; CODE XREF: ROM:000FAF38j
		move.l	(a0)+,(a1)+
		dbf	d0,loc_FAF36
		movem.l	(sp)+,d0/a0-a1
		move.w	#0,d5
		jsr	sub_1169A
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FAF50:							; CODE XREF: ROM:000FACF4p
					; sub_FAD84+4p
		move.l	#$FFFFC6CE,a0
		tst.w	($FFFFD4F6).w
		beq.w	locret_FAF64
		move.l	#$FFFFCA32,a0

locret_FAF64:							; CODE XREF: sub_FAF50+Aj
		rts
; End of function sub_FAF50


; =============== S U B	R O U T	I N E =======================================


sub_FAF66:							; CODE XREF: sub_FA014+44p
					; sub_FA014+5Ep
		movem.l	a3,-(sp)
		move.l	a2,a3
		add.w	(a2),a3

loc_FAF6E:							; CODE XREF: sub_FAF66+18j
		move.b	-(a3),d0
		cmpnb	 $20,d0					; ' '
		bne.w	loc_FAF80
		move.b	#0,(a3)
		subq.w	#1,(a2)
		bra.s	loc_FAF6E
; ---------------------------------------------------------------------------

loc_FAF80:							; CODE XREF: sub_FAF66+Ej
		addq.w	#1,(a2)
		and.w	#$FE,(a2)				; ''
		movem.l	(sp)+,a3
		rts
; End of function sub_FAF66


; =============== S U B	R O U T	I N E =======================================


sub_FAF8C:							; CODE XREF: ROM:000FA37Ep
					; ROM:000FA3D8p ...
		movem.l	d1/a0,-(sp)
		move.l	#$FFFFD4FA,a0
		tst.w	($FFFFD4F6).w
		beq.w	loc_FAFA4
		move.l	#$FFFFD514,a0

loc_FAFA4:							; CODE XREF: sub_FAF8C+Ej
		move.w	($FFFFD4F4).w,d1
		subq.w	#1,d1
		move.b	(a0,d1.w),d0
		ext.w	d0
		move.w	d0,($FFFFD52E).w
		movem.l	(sp)+,d1/a0
		rts
; End of function sub_FAF8C


; =============== S U B	R O U T	I N E =======================================


sub_FAFBA:							; CODE XREF: ROM:000FA1B2p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFD4FA,a1
		move.l	#$FFFFC6CE,a0
		bsr.w	sub_FAFE4
		move.l	#$FFFFD514,a1
		move.l	#$FFFFCA32,a0
		bsr.w	sub_FAFE4
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FAFBA


; =============== S U B	R O U T	I N E =======================================


sub_FAFE4:							; CODE XREF: sub_FAFBA+10p
					; sub_FAFBA+20p
		move.l	$1E(a0),a0
		add.w	6(a0),a0
		move.l	a1,a2
		addq.w	#6,a2
		clr.w	d0

loc_FAFF2:							; CODE XREF: sub_FAFE4+30j
		move.w	#5,d1

loc_FAFF6:							; CODE XREF: sub_FAFE4+1Ej
		move.b	(a0,d1.w),d3
		subq.b	#1,d3
		cmp.b	d3,d0
		beq.w	loc_FB00C
		dbf	d1,loc_FAFF6
		move.b	d0,(a2)+
		bra.w	loc_FB00E
; ---------------------------------------------------------------------------

loc_FB00C:							; CODE XREF: sub_FAFE4+1Aj
		move.b	d0,(a1)+

loc_FB00E:							; CODE XREF: sub_FAFE4+24j
		addq.w	#1,d0
		cmpnw	 $1A,d0
		blt.s	loc_FAFF2
		rts
; End of function sub_FAFE4

; ---------------------------------------------------------------------------

loc_FB018:							; CODE XREF: sub_FBB88+22p
					; sub_FBB88+4Ep
		movem.l	d0-a6,-(sp)
		move	#$2700,sr
		move.w	#2,d4
		move.l	#loc_17C42,($FFFFB03A).w
		bclr	#0,($FFFFBF78).w
		bset	#2,($FFFFBF78).w
		bclr	#1,($FFFFBF78).w
		move.w	#0,($FFFFB000).w
		move.w	#$B400,($FFFFB002).w
		move.w	#$B800,($FFFFB00C).w
		move.w	#5,($FFFFB00E).w
		move.w	#$C000,($FFFFB008).w
		move.w	#6,($FFFFB00A).w
		move.w	#$E000,($FFFFB004).w
		move.w	#6,($FFFFB006).w
		move.w	#0,d0
		jsr	sub_11594
		move	#$2500,sr
		jsr	sub_112BC
		move.w	d4,($FFFFB01E).w
		move.l	#unk_55B86,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$9123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFD530).w
		move.l	#unk_55B86,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$9123
		dc.w	$4560
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.w	d4,($FFFFB012).w
		jsr	sub_11F04
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$1C,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		jsr	sub_16C96
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$28,d2					; '('
		moveq	#$1C,d3
		moveq	#$D,d5
		jsr	sub_1169A
		move.w	d4,($FFFFD430).w
		add.w	#$24,d4					; '$'
		move.w	d4,($FFFFD436).w
		move.l	#off_BF70A,a2
		jsr	sub_11738
		move.w	d4,($FFFFB010).w
		move.l	#unk_A9A18,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$7123
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$1F
		dc.b	6
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d0
		tst.w	($FFFFD4EE).w
		beq.w	loc_FB152
		move.w	($FFFFC332).w,d0

loc_FB152:							; CODE XREF: ROM:000FB14Aj
		move.l	#$30E,a1
		asl.w	#2,d0
		move.l	(a1,d0.w),a1
		add.w	4(a1),a1
		move.w	(a1),d0
		subq.w	#2,d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$1B
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#8,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$8F
		dc.b	$1C
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFC330).w,d3
		tst.w	($FFFFD4EE).w
		beq.w	loc_FB1A8
		move.w	($FFFFC332).w,d3

loc_FB1A8:							; CODE XREF: ROM:000FB1A0j
		asl.w	#2,d3
		move.l	#off_F86F2,a0
		move.l	(a0,d3.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d3
		move.l	#unk_FF462,a0
		sub.w	#$40,d3					; '@'
		add.w	d3,a0
		add.l	(a2)+,a1
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		move.w	#4,d5
		move.w	($FFFFD430).w,d4
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	2
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$24,d0					; '$'
		move.w	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$10
		dc.b	$BF
		dc.b	$A
		dc.b	2
aNameEntrya:	dc.b	'NAME  ENTRYa',0
; ---------------------------------------------------------------------------
		roxr.w	#5,d2
		clr.w	($FFFFD4EC).w
		move.w	#1,($FFFFD4EA).w
		clr.w	d0
		bsr.w	sub_FBBDE
		move.l	#$FFFFD4DA,a1
		bsr.w	sub_FB992
		bsr.w	loc_FB822
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	$BF
		dc.b	7
		dc.b	6
aNameLog:	dc.b	'Name Log',0
; ---------------------------------------------------------------------------
		move.w	#$18,($FFFFBD26).w
		bclr	#2,($FFFFBF78).w
		clr.w	d4
		bsr.w	sub_FB722
		bsr.w	sub_FB77A
		bsr.w	sub_FB7CA
		move.w	($FFFFB028).w,-(sp)
		move.w	($FFFFB02A).w,-(sp)
		jsr	sub_FB97C
		move.w	(sp)+,($FFFFB02A).w
		addq.w	#1,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_FB7A0
		move.w	#1,($FFFFD6C4).w
		bsr.w	loc_FB4C4
		clr.w	d0
		bra.w	loc_FB410
; ---------------------------------------------------------------------------

loc_FB296:							; CODE XREF: ROM:000FB324j
					; ROM:000FB370j ...
		bsr.w	sub_FB8EC
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	5
		dc.b	$12
		dc.b	0
; ---------------------------------------------------------------------------
		tst.w	($FFFFD6C4).w
		beq.w	loc_FB2BC
		move.l	#$FFFFD4DA,a1
		bsr.w	sub_FB91A
		bsr.w	sub_FB6F0

loc_FB2BC:							; CODE XREF: ROM:000FB2AAj
		bsr.w	sub_FB874

loc_FB2C0:							; CODE XREF: ROM:000FB2E8j
					; ROM:000FB2F6j
		move.w	($FFFFB03E).w,d0

loc_FB2C4:							; CODE XREF: ROM:000FB2C8j
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_FB2C4
		movem.w	d1,-(sp)
		move.w	($FFFFC328).w,d1
		subq.w	#1,d1
		cmp.w	($FFFFD4EE).w,d1
		movem.w	(sp)+,d1
		beq.w	loc_FB2EE
		jsr	sub_11358
		tst.w	d1
		beq.s	loc_FB2C0
		bra.w	loc_FB2F8
; ---------------------------------------------------------------------------

loc_FB2EE:							; CODE XREF: ROM:000FB2DCj
		jsr	sub_11340
		tst.w	d1
		beq.s	loc_FB2C0

loc_FB2F8:							; CODE XREF: ROM:000FB2EAj
		btst	#7,d1
		bne.w	loc_FB4BA
		tst.w	($FFFFD6C4).w
		bne.w	loc_FB332
		move.w	#1,d0
		btst	#1,d1
		bne.w	loc_FB37C
		move.w	#$FFFF,d0
		btst	#0,d1
		bne.w	loc_FB37C
		btst	#4,d1
		beq.w	loc_FB296
		bsr.w	loc_FB4C4
		clr.w	d0
		bra.w	loc_FB37C
; ---------------------------------------------------------------------------

loc_FB332:							; CODE XREF: ROM:000FB304j
		moveq	#$FFFFFFFF,d0
		btst	#6,d1
		bne.w	loc_FB410
		neg.w	d0
		btst	#5,d1
		bne.w	loc_FB410
		btst	#3,d1
		bne.w	loc_FB43E
		neg.w	d0
		btst	#2,d1
		bne.w	loc_FB43E
		moveq	#6,d0
		btst	#1,d1
		bne.w	loc_FB43E
		neg.w	d0
		btst	#0,d1
		bne.w	loc_FB43E
		btst	#4,d1
		beq.w	loc_FB296
		bsr.w	loc_FB4C4
		bra.w	loc_FB296
; ---------------------------------------------------------------------------

loc_FB37C:							; CODE XREF: ROM:000FB310j
					; ROM:000FB31Cj ...
		add.w	d0,($FFFFD4EA).w

loc_FB380:							; CODE XREF: ROM:000FB3A2j
		cmp.w	#7,($FFFFD4EA).w
		ble.w	loc_FB390
		move.w	#1,($FFFFD4EA).w

loc_FB390:							; CODE XREF: ROM:000FB386j
		tst.w	($FFFFD4EA).w
		bne.w	loc_FB39E
		move.w	#7,($FFFFD4EA).w

loc_FB39E:							; CODE XREF: ROM:000FB394j
		bsr.w	sub_FBBDE
		beq.s	loc_FB380
		move.l	#$FFFFD4DA,a1
		bsr.w	sub_FB992
		tst.w	($FFFFD6C4).w
		bne.w	loc_FB3C0
		jsr	loc_FB634
		bra.w	loc_FB296
; ---------------------------------------------------------------------------

loc_FB3C0:							; CODE XREF: ROM:000FB3B2j
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	4
		dc.b	$13
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	d4,($FFFFB028).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	'   ',0
; ---------------------------------------------------------------------------
		bsr.w	sub_FB7CA
		moveq	#1,d2
		move.w	($FFFFB028).w,-(sp)
		move.w	($FFFFB02A).w,-(sp)
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w
		addq.w	#1,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_FB7A0
		bsr.w	sub_FB722
		move.w	d4,d0
		neg.w	d0
		bra.w	*+4

loc_FB410:							; CODE XREF: ROM:000FB292j
					; ROM:000FB338j ...
		add.w	d4,d0
		cmpnw	 $B,d0
		bhi.w	loc_FB296
		move.w	d0,d4
		bsr.w	sub_FB77A
		move.l	#$FFFFD4DA,a0
		clr.w	d0
		cmp.b	#$2D,(a0,d4.w)				; '-'
		beq.w	loc_FB43E
		move.b	(a0,d4.w),d0
		ext.w	d0
		bsr.w	sub_FB750
		sub.w	d5,d0

loc_FB43E:							; CODE XREF: ROM:000FB34Aj
					; ROM:000FB354j ...
		add.w	d5,d0
		cmpnw	 $1D,d0
		bhi.w	loc_FB296
		move.w	d0,-(sp)
		bsr.w	sub_FB7CA
		moveq	#1,d2
		move.w	($FFFFB028).w,-(sp)
		move.w	($FFFFB02A).w,-(sp)
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w
		addq.w	#1,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_FB7A0
		move.w	(sp)+,d5
		bsr.w	sub_FB7CA
		move.w	($FFFFB028).w,-(sp)
		move.w	($FFFFB02A).w,-(sp)
		tst.w	($FFFFD6C4).w
		beq.w	loc_FB48E
		jsr	sub_FB97C

loc_FB48E:							; CODE XREF: ROM:000FB484j
		move.w	(sp)+,($FFFFB02A).w
		addq.w	#1,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB028).w
		bsr.w	sub_FB7A0
		move.l	#$FFFFD4DA,a0
		move.l	#aAbcdefghijklmnopqrstu,a1		; "ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -"
		move.b	(a1,d5.w),d0
		move.b	d0,(a0,d4.w)
		bra.w	loc_FB296
; ---------------------------------------------------------------------------

loc_FB4BA:							; CODE XREF: ROM:000FB2FCj
		bsr.w	sub_FB9D4
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_FB4C4:							; CODE XREF: ROM:000FB28Cp
					; ROM:000FB328p ...
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	1
		dc.b	$14
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$16,d0
		move.w	#7,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		bchg	#0,($FFFFD6C5).w
		bne.w	loc_FB5F0
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$19
		dc.b	$11
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#aAbcdefghijklmnopqrstu,a0		; "ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -"
		moveq	#4,d0

loc_FB504:							; CODE XREF: ROM:000FB52Ej
		moveq	#5,d1
		move.l	#$FFFFBFA4,a1
		move.w	#$E,(a1)+

loc_FB510:							; CODE XREF: ROM:000FB516j
		move.b	(a0)+,(a1)+
		move.b	#$20,(a1)+				; ' '
		dbf	d1,loc_FB510
		move.w	#$BFA4,a1
		jsr	sub_11BA4
		addq.w	#2,($FFFFB02A).w
		sub.w	#$C,($FFFFB028).w
		dbf	d0,loc_FB504
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	1
		dc.b	$14
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$16,d0
		moveq	#7,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BF
		dc.b	2
		dc.b	$15
aDPadToALetter_: dc.b	'D-Pad to a letter.',0
; ---------------------------------------------------------------------------
		move.w	#2,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		move.w	($FFFFB028).w,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
aCToSelectLetter_: dc.b	'C to select letter.',0
; ---------------------------------------------------------------------------
		move.w	(sp),($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$10
aAToGoBack_:	dc.b	'A to go back.',0
; ---------------------------------------------------------------------------
		move.w	(sp),($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$E
aBToCancel_:	dc.b	'B to cancel.'
; ---------------------------------------------------------------------------
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$12
aStartWhenDone_: dc.b	'START when done.'
; ---------------------------------------------------------------------------

loc_FB5EA:							; CODE XREF: ROM:000FB632j
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_FB5F0:							; CODE XREF: ROM:000FB4ECj
		move.l	#$FFFFD4DA,a1

loc_FB5F6:
		bsr.w	sub_FB992
		move.w	#0,($FFFFB028).w
		move.w	#$F,($FFFFB02A).w
		move.w	#$28,d0					; '('

loc_FB60A:
		move.w	#$D,d1
		move.w	#$7FF,d2

loc_FB612:
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	1
		dc.b	$14
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$16,d0
		moveq	#7,d1
		jsr	sub_119B8
		bsr.w	loc_FB634
		bra.s	loc_FB5EA
; ---------------------------------------------------------------------------

loc_FB634:							; CODE XREF: ROM:000FB3B6p
					; ROM:000FB62Ep
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1A
		dc.b	$BF
		dc.b	2
		dc.b	$15
aDPadUpDownTo:	dc.b	'D-Pad up/down to    ',0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1A
		dc.b	$BF
		dc.b	2
		dc.b	$16
aMoveArrows_:	dc.b	'move arrows.        ',0
; ---------------------------------------------------------------------------
		move.w	#2,($FFFFB028).w
		move.w	($FFFFB028).w,-(sp)
		addq.w	#1,($FFFFB02A).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
aPressStartTo:	dc.b	'Press START to      '
; ---------------------------------------------------------------------------
		addq.w	#1,($FFFFB02A).w
		move.w	(sp),($FFFFB028).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
aSelectName_:	dc.b	'select name.        '
; ---------------------------------------------------------------------------
		addq.w	#1,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
aPressBToEdit_:	dc.b	'Press B to edit.    '
; ---------------------------------------------------------------------------
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FB6F0:							; CODE XREF: ROM:000FB2B8p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFD4EA).w,d0
		mulu.w	#$C,d0
		move.l	#$FFFFD45A,a0

loc_FB702:
		move.l	#word_FB7F2,a1
		tst.b	(a0,d0.w)
		beq.w	loc_FB716
		move.l	#word_FB80A,a1

loc_FB716:							; CODE XREF: sub_FB6F0+1Cj
		jsr	sub_11BA4
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB6F0


; =============== S U B	R O U T	I N E =======================================


sub_FB722:							; CODE XREF: ROM:000FB258p
					; ROM:000FB404p
		movem.l	d0-d4/a0-a6,-(sp)
		move.b	($FFFFD4DA).w,d0

loc_FB72A:
		bsr.w	sub_FB750
		cmpnw	 $1E,d0
		blt.w	loc_FB73C
		clr.w	d5
		bra.w	loc_FB74A
; ---------------------------------------------------------------------------

loc_FB73C:							; CODE XREF: sub_FB722+10j
		move.w	d0,d5
		move.l	#aAbcdefghijklmnopqrstu,a0		; "ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -"
		move.b	(a0,d5.w),($FFFFD4DA).w

loc_FB74A:							; CODE XREF: sub_FB722+16j
		movem.l	(sp)+,d0-d4/a0-a6
		rts
; End of function sub_FB722


; =============== S U B	R O U T	I N E =======================================


sub_FB750:							; CODE XREF: ROM:000FB438p
					; sub_FB722:loc_FB72Ap
		movem.l	d1-d3/a0-a6,-(sp)
		move.l	#aAbcdefghijklmnopqrstu,a0		; "ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -"
		move.b	d0,d1
		clr.w	d0
		move.w	#$1E,d3

loc_FB762:							; CODE XREF: sub_FB750+1Cj
		cmp.b	(a0,d0.w),d1
		beq.w	loc_FB774
		addq.w	#1,d0
		dbf	d3,loc_FB762
		move.w	#$1E,d0

loc_FB774:							; CODE XREF: sub_FB750+16j
		movem.l	(sp)+,d1-d3/a0-a6
		rts
; End of function sub_FB750


; =============== S U B	R O U T	I N E =======================================


sub_FB77A:							; CODE XREF: ROM:000FB25Cp
					; ROM:000FB41Cp
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	4
		dc.b	$13
		dc.b	0
; ---------------------------------------------------------------------------
		add.w	d4,($FFFFB028).w
		tst.w	($FFFFD6C4).w
		beq.w	locret_FB79E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$20
		dc.b	$3C
		dc.b	$20
		dc.b	0
; ---------------------------------------------------------------------------

locret_FB79E:							; CODE XREF: sub_FB77A+14j
					; sub_FB7A0+4j
		rts
; End of function sub_FB77A


; =============== S U B	R O U T	I N E =======================================


sub_FB7A0:							; CODE XREF: ROM:000FB282p
					; ROM:000FB400p ...
		tst.w	($FFFFD6C4).w
		beq.s	locret_FB79E
		move.l	#$FFFFCF36,a1
		move.w	#4,(a1)
		move.b	#0,3(a1)
		move.l	#aAbcdefghijklmnopqrstu,a0		; "ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -"
		move.b	(a0,d5.w),d0
		move.b	d0,2(a1)
		jmp	sub_11BA4
; End of function sub_FB7A0


; =============== S U B	R O U T	I N E =======================================


sub_FB7CA:							; CODE XREF: ROM:000FB260p
					; ROM:000FB3DCp ...
		jsr	sub_11B92
; End of function sub_FB7CA

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	$18
		dc.b	$10
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	d5,d0
		ext.l	d0
		divu.w	#6,d0
		asl.w	#1,d0
		add.w	d0,($FFFFB02A).w
		swap	d0
		asl.w	#1,d0
		add.w	d0,($FFFFB028).w
		moveq	#3,d0
		moveq	#3,d1
		rts
; ---------------------------------------------------------------------------
word_FB7F2:	dc.w	$18					; DATA XREF: sub_FB6F0:loc_FB702o
		dc.b	$BF
		dc.b	2
		dc.b	$10
aEnterNewName:	dc.b	'  Enter new name: ',0
word_FB80A:	dc.w	$18					; DATA XREF: sub_FB6F0+20o
		dc.b	$BF
		dc.b	2
		dc.b	$10
aSelectOrReplace: dc.b	'Select or replace:',0
; ---------------------------------------------------------------------------

loc_FB822:							; CODE XREF: ROM:000FB232p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFD4E8).w,-(sp)
		move.w	($FFFFD4EA).w,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	5
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#6,d7
		move.w	#1,($FFFFD4EA).w
		move.l	#$FFFFBF20,a1

loc_FB84A:							; CODE XREF: ROM:000FB862j
		bsr.w	sub_FB992
		move.w	($FFFFB028).w,-(sp)
		bsr.w	sub_FB91A
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		addq.w	#1,($FFFFD4EA).w
		dbf	d7,loc_FB84A
		move.w	(sp)+,($FFFFD4EA).w
		move.w	(sp)+,($FFFFD4E8).w
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FB874:							; CODE XREF: ROM:loc_FB2BCp
		movem.l	d0-a6,-(sp)
		move.w	($FFFFD4EA).w,d0
		cmp.w	($FFFFD4EC).w,d0
		beq.w	loc_FB898
		move.l	#word_FB8E4,a1
		bsr.w	loc_FB89E
		move.l	#word_FB8DC,a1
		bsr.w	loc_FB89E

loc_FB898:							; CODE XREF: sub_FB874+Cj
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB874

; ---------------------------------------------------------------------------

loc_FB89E:							; CODE XREF: sub_FB874+16p
					; sub_FB874+20p
		tst.w	($FFFFD4EC).w
		beq.w	loc_FB8D4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	3
		dc.b	8
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	($FFFFD4EC).w,d0
		subq.w	#1,d0
		add.w	d0,($FFFFB02A).w
		move.l	a1,-(sp)
		jsr	sub_11BA4
		move.l	(sp)+,a1
		add.w	(a1),a1
		move.w	#$12,($FFFFB028).w
		jsr	sub_11BA4

loc_FB8D4:							; CODE XREF: ROM:000FB8A2j
		move.w	($FFFFD4EA).w,($FFFFD4EC).w
		rts
; ---------------------------------------------------------------------------
word_FB8DC:	dc.w	4					; DATA XREF: sub_FB874+1Ao
		dc.b	']',0
		dc.w	4
		dc.b	'[',0
word_FB8E4:	dc.w	4					; DATA XREF: sub_FB874+10o
		dc.b	' ',0
		dc.w	4
		dc.b	' ',0

; =============== S U B	R O U T	I N E =======================================


sub_FB8EC:							; CODE XREF: ROM:loc_FB296p
		movem.l	d0-a6,-(sp)
		move.w	#$B,d3
		move.l	#$FFFFD4DA,a0
		clr.w	($FFFFD4E8).w

loc_FB8FE:							; CODE XREF: sub_FB8EC+24j
		cmp.b	#$2D,(a0)				; '-'
		beq.w	loc_FB914
		tst.b	(a0)+
		beq.w	loc_FB914
		addq.w	#1,($FFFFD4E8).w
		dbf	d3,loc_FB8FE

loc_FB914:							; CODE XREF: sub_FB8EC+16j
					; sub_FB8EC+1Cj
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB8EC


; =============== S U B	R O U T	I N E =======================================


sub_FB91A:							; CODE XREF: ROM:000FB2B4p
					; ROM:000FB852p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFCF36,a0
		move.w	($FFFFD4E8).w,d0
		move.w	#0,d3
		move.w	#$E,(a0)+

loc_FB930:							; CODE XREF: sub_FB91A+2Ej
		move.b	(a1,d3.w),d1
		cmp.w	($FFFFD4E8).w,d3
		blt.w	loc_FB940
		move.b	#$2D,d1					; '-'

loc_FB940:							; CODE XREF: sub_FB91A+1Ej
		move.b	d1,(a0)+
		addq.w	#1,d3
		cmpnw	 $C,d3
		blt.s	loc_FB930
		move.l	#$FFFFCF36,a1
		jsr	sub_11BA4
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB91A

; ---------------------------------------------------------------------------
aAbcdefghijklmnopqrstu: dc.b 'ABCDEFGHIJKLMNOPQRSTUVWXYZ.12 -',0	; DATA XREF: ROM:000FB4A8o
					; ROM:000FB4FCo ...

; =============== S U B	R O U T	I N E =======================================


sub_FB97C:							; CODE XREF: ROM:000FB26Cp
					; ROM:000FB488p
		move.w	($FFFFB01E).w,-(sp)
		move.w	($FFFFD530).w,($FFFFB01E).w
		jsr	sub_119B8
		move.w	(sp)+,($FFFFB01E).w
		rts
; End of function sub_FB97C


; =============== S U B	R O U T	I N E =======================================


sub_FB992:							; CODE XREF: ROM:000FB22Ep
					; ROM:000FB3AAp ...
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFD45A,a0
		move.w	($FFFFD4EA).w,d2
		mulu.w	#$C,d2
		move.w	#0,($FFFFD4E8).w
		move.w	#$B,d3

loc_FB9AE:							; CODE XREF: sub_FB992+38j
		move.b	(a0,d2.w),d0
		beq.w	loc_FB9BA
		bra.w	loc_FB9C2
; ---------------------------------------------------------------------------

loc_FB9BA:							; CODE XREF: sub_FB992+20j
		move.b	#$2D,d0					; '-'
		subq.w	#1,($FFFFD4E8).w

loc_FB9C2:							; CODE XREF: sub_FB992+24j
		move.b	d0,(a1)+
		addq.w	#1,($FFFFD4E8).w
		addq.w	#1,d2
		dbf	d3,loc_FB9AE
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB992


; =============== S U B	R O U T	I N E =======================================


sub_FB9D4:							; CODE XREF: ROM:loc_FB4BAp
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFD45A,a0
		move.w	($FFFFD4EA).w,d0
		mulu.w	#$C,d0
		add.w	d0,a0
		move.l	#$FFFFD4DA,a1
		move.w	#$B,d4

loc_FB9F2:							; CODE XREF: sub_FB9D4+32j
		cmp.b	#$2D,(a1)				; '-'
		bne.w	loc_FB9FE
		move.b	#0,(a1)

loc_FB9FE:							; CODE XREF: sub_FB9D4+22j
		move.b	(a1)+,d1
		cmp.b	(a0)+,d1
		bne.w	loc_FBA0E
		dbf	d4,loc_FB9F2
		bra.w	loc_FBA4C
; ---------------------------------------------------------------------------

loc_FBA0E:							; CODE XREF: sub_FB9D4+2Ej
		move.l	#$FFFFD45A,a0
		move.w	($FFFFD4EA).w,d0
		mulu.w	#$C,d0
		add.w	d0,a0
		move.l	#$FFFFD4DA,a1
		move.w	#$B,d4

loc_FBA28:							; CODE XREF: sub_FB9D4+64j
		move.b	(a1)+,d1
		cmpnb	 $2D,d1					; '-'
		bne.w	loc_FBA36
		move.b	#0,d1

loc_FBA36:							; CODE XREF: sub_FB9D4+5Aj
		move.b	d1,(a0)+
		dbf	d4,loc_FBA28
		bsr.w	sub_F9C5E
		jsr	sub_FBA76
		jsr	sub_1A206

loc_FBA4C:							; CODE XREF: sub_FB9D4+36j
		move.l	#$FFFFD042,a5
		tst.w	($FFFFD4EE).w
		beq.w	loc_FBA60
		move.l	#$FFFFD044,a5

loc_FBA60:							; CODE XREF: sub_FB9D4+82j
		tst.b	($FFFFD4DA).w
		bne.w	loc_FBA6C
		clr.w	($FFFFD4EA).w

loc_FBA6C:							; CODE XREF: sub_FB9D4+90j
		move.w	($FFFFD4EA).w,(a5)
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FB9D4


; =============== S U B	R O U T	I N E =======================================


sub_FBA76:							; CODE XREF: sub_FB9D4+6Cp
		movem.l	d0-a6,-(sp)
		move.w	#$2D7,d7
		moveq	#0,d0
		moveq	#4,d1
		move.l	#$FFFFCF36,a0
		move.w	($FFFFD4EA).w,d6

loc_FBA8C:							; CODE XREF: sub_FBA76+48j
		clr.w	d5
		jsr	sub_1A244
		cmp.b	1(a0),d6
		bne.w	loc_FBAA2
		clr.b	1(a0)
		st	d5

loc_FBAA2:							; CODE XREF: sub_FBA76+22j
		cmp.b	3(a0),d6
		bne.w	loc_FBAB0
		clr.b	3(a0)
		st	d5

loc_FBAB0:							; CODE XREF: sub_FBA76+30j
		tst.w	d5
		beq.w	loc_FBABC
		jsr	sub_1A1E4

loc_FBABC:							; CODE XREF: sub_FBA76+3Cj
		addq.l	#4,d0
		dbf	d7,loc_FBA8C
		move.w	#$1B,d7
		move.l	#$B60,d0
		moveq	#$10,d1
		move.l	#$FFFFCF36,a0
		move.w	($FFFFD4EA).w,d6

loc_FBAD8:							; CODE XREF: sub_FBA76+D0j
		clr.w	d5
		jsr	sub_1A244
		cmp.b	1(a0),d6
		bne.w	loc_FBAEE
		clr.b	1(a0)
		st	d5

loc_FBAEE:							; CODE XREF: sub_FBA76+6Ej
		cmp.b	3(a0),d6
		bne.w	loc_FBAFC
		clr.b	3(a0)
		st	d5

loc_FBAFC:							; CODE XREF: sub_FBA76+7Cj
		cmp.b	5(a0),d6
		bne.w	loc_FBB0A
		clr.b	5(a0)
		st	d5

loc_FBB0A:							; CODE XREF: sub_FBA76+8Aj
		cmp.b	7(a0),d6
		bne.w	loc_FBB18
		clr.b	7(a0)
		st	d5

loc_FBB18:							; CODE XREF: sub_FBA76+98j
		cmp.b	9(a0),d6
		bne.w	loc_FBB26
		clr.b	9(a0)
		st	d5

loc_FBB26:							; CODE XREF: sub_FBA76+A6j
		cmp.b	$B(a0),d6
		bne.w	loc_FBB34
		clr.b	$B(a0)
		st	d5

loc_FBB34:							; CODE XREF: sub_FBA76+B4j
		tst.w	d5
		beq.w	loc_FBB40
		jsr	sub_1A1E4

loc_FBB40:							; CODE XREF: sub_FBA76+C0j
		add.l	#$10,d0
		dbf	d7,loc_FBAD8
		move.l	#$D20,d0
		move.w	($FFFFD4EA).w,d3
		asl.w	#4,d3
		ext.l	d3
		add.l	d3,d0
		moveq	#$10,d1
		move.l	#$FFFFCF36,a0
		jsr	sub_1A244
		move.l	a0,-(sp)
		move.w	#$F,d3

loc_FBB6E:							; CODE XREF: sub_FBA76+FAj
		clr.b	(a0)+
		dbf	d3,loc_FBB6E
		move.l	(sp)+,a0
		jsr	sub_1A1E4
		jsr	sub_1A206
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FBA76


; =============== S U B	R O U T	I N E =======================================


sub_FBB88:							; CODE XREF: sub_7CB0+F668p
		tst.w	($FFFFD054).w
		bne.w	locret_FBBDC
		clr.w	($FFFFD042).w
		clr.w	($FFFFD044).w
		tst.w	($FFFFC328).w
		beq.w	loc_FBBB0
		move.w	($FFFFC328).w,($FFFFD4EE).w
		subq.w	#1,($FFFFD4EE).w
		jsr	loc_FB018

loc_FBBB0:							; CODE XREF: sub_FBB88+14j
		tst.w	($FFFFC32A).w
		beq.w	locret_FBBDC
		movem.w	d0,-(sp)
		move.w	($FFFFC328).w,d0
		cmp.w	($FFFFC32A).w,d0
		movem.w	(sp)+,d0
		beq.w	locret_FBBDC
		move.w	($FFFFC32A).w,($FFFFD4EE).w
		subq.w	#1,($FFFFD4EE).w
		jsr	loc_FB018

locret_FBBDC:							; CODE XREF: sub_FBB88+4j
					; sub_FBB88+2Cj ...
		rts
; End of function sub_FBB88


; =============== S U B	R O U T	I N E =======================================


sub_FBBDE:							; CODE XREF: ROM:000FB224p
					; ROM:loc_FB39Ep
		movem.w	d0,-(sp)
		move.w	($FFFFD044).w,d0
		tst.w	($FFFFD4EE).w
		beq.w	loc_FBBF2
		move.w	($FFFFD042).w,d0

loc_FBBF2:							; CODE XREF: sub_FBBDE+Cj
		cmp.w	($FFFFD4EA).w,d0
		bne.w	loc_FBC0E
		move.w	#1,d0
		tst.w	(sp)
		bpl.w	loc_FBC08
		move.w	#$FFFF,d0

loc_FBC08:							; CODE XREF: sub_FBBDE+22j
		add.w	d0,($FFFFD4EA).w
		clr.w	d0

loc_FBC0E:							; CODE XREF: sub_FBBDE+18j
		movem.w	(sp)+,d0
		rts
; End of function sub_FBBDE

; ---------------------------------------------------------------------------

loc_FBC14:							; DATA XREF: ROM:000196E0o
					; ROM:00019854o ...
		move.w	#0,d0
		move.w	#$1A,d1
		jsr	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_E9A80,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#8,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$21
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_E9ED6,a0
		move.l	a0,a1

loc_FBC62:
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#0,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	$B
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$12,d0
		move.w	#7,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	$E
		dc.b	2
aRecord_0:	dc.b	'Record'
		dc.b	$BD
		dc.b	$D
		dc.b	5
aHolders:	dc.b	'Holders'
		dc.b	$BD
		dc.b	1
		dc.b	9
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$10
		dc.b	$F8
		dc.b	6
		dc.b	3
		dc.b	5
		dc.b	8
		dc.b	$F9
		dc.b	1
aName:		dc.b	'Name',0,0
; ---------------------------------------------------------------------------
		bsr.w	sub_F9C68
		clr.w	($FFFFBF12).w
		bclr	#6,($FFFFC2F8).w
		bsr.w	sub_FC282
		bset	#6,($FFFFC2F8).w
		bsr.w	sub_FC282
		bsr.w	sub_FC184
		bsr.w	sub_FBD86
		bsr.w	loc_FBEA4

loc_FBCF0:							; CODE XREF: ROM:000FBD34j
					; ROM:000FBD44j ...
		jsr	sub_80BA
		jsr	sub_A41E
		jsr	sub_11318
		btst	#7,d3
		beq.w	loc_FBD10
		jmp	loc_9CD8
; ---------------------------------------------------------------------------

loc_FBD10:							; CODE XREF: ROM:000FBD06j
		tst.w	($FFFFBF12).w
		bne.w	loc_FBD36
		btst	#6,d3
		beq.w	loc_FBD36
		btst	#5,d3
		beq.w	loc_FBD36
		bsr.w	sub_FE6D2
		bsr.w	sub_FC184
		bsr.w	loc_FBEA4
		bra.s	loc_FBCF0
; ---------------------------------------------------------------------------

loc_FBD36:							; CODE XREF: ROM:000FBD14j
					; ROM:000FBD1Cj ...
		btst	#3,d1
		beq.w	loc_FBD58
		cmp.w	#2,($FFFFBF12).w
		beq.s	loc_FBCF0
		bsr.w	sub_FBFA0
		addq.w	#1,($FFFFBF12).w
		bsr.w	sub_FBD86
		bsr.w	sub_FBD7A
		bra.s	loc_FBCF0
; ---------------------------------------------------------------------------

loc_FBD58:							; CODE XREF: ROM:000FBD3Aj
		btst	#2,d1
		beq.s	loc_FBCF0
		tst.w	($FFFFBF12).w
		beq.s	loc_FBCF0
		bsr.w	sub_FBFA0
		subq.w	#1,($FFFFBF12).w
		bsr.w	sub_FBD86
		bsr.w	sub_FBD7A
		bra.w	loc_FBCF0
; ---------------------------------------------------------------------------
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FBD7A:							; CODE XREF: ROM:000FBD52p
					; ROM:000FBD70p

; FUNCTION CHUNK AT 000FBFDE SIZE 0000001E BYTES

		tst.w	($FFFFBF12).w
		beq.w	loc_FBEA4
		bra.w	loc_FBFDE
; End of function sub_FBD7A


; =============== S U B	R O U T	I N E =======================================


sub_FBD86:							; CODE XREF: ROM:000FBCE8p
					; ROM:000FBD4Ep ...
		move.l	#word_FBDB2,a1
		tst.w	($FFFFBF12).w
		beq.w	loc_FBDAA
		move.l	#word_FBE54,a1
		cmp.w	#2,($FFFFBF12).w
		beq.w	loc_FBDAA
		move.l	#word_FBE04,a1

loc_FBDAA:							; CODE XREF: sub_FBD86+Aj
					; sub_FBD86+1Aj
		jsr	sub_11A48
		rts
; End of function sub_FBD86

; ---------------------------------------------------------------------------
word_FBDB2:	dc.w	$52					; DATA XREF: sub_FBD86o
aWinWinLossTieUseACToC: dc.b '',6,3,$10,8,'',1,'   Win %   Win-Loss-Tie',4,'',$19,'Use A+C to clear '
		dc.b	'ALL win records',$10,'',$1A,'  More ]',0
word_FBE04:	dc.w	$50					; DATA XREF: sub_FBD86+1Eo
aGoalsTeamsTeamMustWin: dc.b '',6,3,$12,8,'',1,'Goals       Teams    ',4,'',$19,'    TEAM MUST WIN T'
		dc.b	'O QUALIFY    ',$10,'',$1A,'[ More ]',0
word_FBE54:	dc.w	$50					; DATA XREF: sub_FBD86+Eo
aSavesTeamsTeamMustWin: dc.b '',6,3,$12,8,'',1,'Saves       Teams    ',4,'',$19,'    TEAM MUST WIN T'
		dc.b	'O QUALIFY    ',$10,'',$1A,'[ More  ',0
; ---------------------------------------------------------------------------

loc_FBEA4:							; CODE XREF: ROM:000FBCECp
					; ROM:000FBD30p ...
		movem.l	d0-a6,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	2
		dc.b	$A
		dc.b	$F9
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#6,d7
		move.l	#$FFFFD542,a0
		move.l	#$FFFFD54A,a2
		move.l	#$FFFFCF36,a5

loc_FBECE:							; CODE XREF: ROM:000FBF96j
		move.w	#2,($FFFFB028).w
		move.b	(a0)+,d0
		ext.w	d0
		move.b	(a2,d0.w),d5
		asl.w	#4,d0
		move.b	$A(a5,d0.w),d4
		lsl.w	#8,d4
		move.b	$B(a5,d0.w),d4
		move.w	d4,-(sp)
		move.b	$C(a5,d0.w),d4
		lsl.w	#8,d4
		move.b	$D(a5,d0.w),d4
		move.w	d4,($FFFFBF4A).w
		move.b	8(a5,d0.w),d4
		lsl.w	#8,d4
		move.b	9(a5,d0.w),d4
		move.w	d4,($FFFFBF4E).w
		add.w	($FFFFBF4A).w,d4
		sub.w	(sp),d4
		neg.w	d4
		move.w	d4,($FFFFBF4C).w
		move.w	(sp)+,d4
		tst.w	d4
		beq.w	loc_FBF86
		bsr.w	sub_FC136
		move.w	#$13,($FFFFB028).w
		move.w	d0,-(sp)
		move.w	d5,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		move.w	#$1A,($FFFFB028).w
		move.w	($FFFFBF4E).w,d0
		move.w	#4,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		move.w	#$1F,($FFFFB028).w
		move.w	($FFFFBF4C).w,d0
		move.w	#4,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		move.w	#$24,($FFFFB028).w			; '$'
		move.w	($FFFFBF4A).w,d0
		move.w	#3,d1
		jsr	sub_11D3A
		move.w	(sp)+,d0
		bra.w	loc_FBF8C
; ---------------------------------------------------------------------------

loc_FBF86:							; CODE XREF: ROM:000FBF16j
		move.l	#word_FC100,a1

loc_FBF8C:							; CODE XREF: ROM:000FBF82j
		jsr	sub_11A48
		addq.w	#2,($FFFFB02A).w
		dbf	d7,loc_FBECE
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FBFA0:							; CODE XREF: ROM:000FBD46p
					; ROM:000FBD64p
		movem.l	d0-a6,-(sp)
		move.w	($FFFFB028).w,-(sp)
		move.w	($FFFFB02A).w,-(sp)
		move.w	($FFFFB02E).w,-(sp)
		jsr	sub_11A36
; End of function sub_FBFA0

; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	0
		dc.b	$A
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$28,d0					; '('
		move.w	#$12,d1
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02E).w
		move.w	(sp)+,($FFFFB02A).w
		move.w	(sp)+,($FFFFB028).w
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_FBD7A

loc_FBFDE:							; CODE XREF: sub_FBD7A+8j
		movem.l	d0-a6,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	2
		dc.b	$A
		dc.b	$F9
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#6,d7
		move.l	#$FFFFD532,a0
; END OF FUNCTION CHUNK	FOR sub_FBD7A
		cmp.w	#1,($FFFFBF12).w
		beq.w	loc_FC00C
		move.l	#$FFFFD53A,a0

loc_FC00C:							; CODE XREF: ROM:000FC002j
		move.l	#$FFFFCF36,a2

loc_FC012:							; CODE XREF: ROM:000FC0F6j
		move.w	#2,($FFFFB028).w
		move.b	(a0)+,d0
		ext.w	d0
		asl.w	#4,d0
		move.b	(a2,d0.w),d5
		cmp.w	#1,($FFFFBF12).w
		beq.w	loc_FC030
		move.b	4(a2,d0.w),d5

loc_FC030:							; CODE XREF: ROM:000FC028j
		tst.b	d5
		beq.w	loc_FC0E6
		bsr.w	sub_FC136
		move.w	#$12,($FFFFB028).w
		move.w	d0,-(sp)
		cmp.w	#1,($FFFFBF12).w
		bne.w	loc_FC054
		move.b	(a2,d0.w),d0
		bra.w	loc_FC058
; ---------------------------------------------------------------------------

loc_FC054:							; CODE XREF: ROM:000FC048j
		move.b	4(a2,d0.w),d0

loc_FC058:							; CODE XREF: ROM:000FC050j
		and.w	#$FF,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		move.w	(sp)+,d0
		move.w	#$19,($FFFFB028).w
		move.l	#word_FC128,a1
		move.l	#$FFFFBFA4,a3
		bsr.w	sub_F997A
		move.w	d0,-(sp)
		cmp.w	#1,($FFFFBF12).w
		bne.w	loc_FC098
		move.b	1(a2,d0.w),d0
		bra.w	loc_FC09C
; ---------------------------------------------------------------------------

loc_FC098:							; CODE XREF: ROM:000FC08Cj
		move.b	5(a2,d0.w),d0

loc_FC09C:							; CODE XREF: ROM:000FC094j
		and.w	#$FF,d0
		move.l	#$FFFFBFA4,a1
		bsr.w	sub_FA880
		move.l	#word_FC12E,a1
		move.l	#$FFFFBFA4,a3
		jsr	sub_11D9E
		move.l	#$FFFFBFA4,a1
		move.w	(sp)+,d0
		cmp.w	#1,($FFFFBF12).w
		bne.w	loc_FC0D6
		move.b	2(a2,d0.w),d0
		bra.w	loc_FC0DA
; ---------------------------------------------------------------------------

loc_FC0D6:							; CODE XREF: ROM:000FC0CAj
		move.b	6(a2,d0.w),d0

loc_FC0DA:							; CODE XREF: ROM:000FC0D2j
		and.w	#$FF,d0
		bsr.w	sub_FA880
		bra.w	loc_FC0EC
; ---------------------------------------------------------------------------

loc_FC0E6:							; CODE XREF: ROM:000FC032j
		move.l	#word_FC100,a1

loc_FC0EC:							; CODE XREF: ROM:000FC0E2j
		jsr	sub_11A48
		addq.w	#2,($FFFFB02A).w
		dbf	d7,loc_FC012
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------
word_FC100:	dc.w	$28					; DATA XREF: ROM:loc_FBF86o
					; ROM:loc_FC0E6o
		dc.b	'                                      '
word_FC128:	dc.w	6					; DATA XREF: ROM:000FC074o
aBy_1:		dc.b	'by ',0
word_FC12E:	dc.w	8					; DATA XREF: ROM:000FC0AAo
aVs__1:		dc.b	' vs. ',0

; =============== S U B	R O U T	I N E =======================================


sub_FC136:							; CODE XREF: ROM:000FBF1Ap
					; ROM:000FC036p
		move.w	d7,-(sp)
		neg.w	d7
		addq.w	#7,d7
		add.w	#$30,d7					; '0'
		move.l	#$FFFFBFA4,a1
		move.w	#6,(a1)
		move.b	d7,2(a1)
		move.b	#$2E,3(a1)				; '.'
		move.b	#$20,4(a1)				; ' '
		move.b	#0,5(a1)
		jsr	sub_11A48
		move.w	(sp)+,d7
		move.b	-1(a0),d2
		ext.w	d2
		move.l	#$FFFFBFA4,a1
		bclr	#7,($FFFFC2F8).w
		bsr.w	sub_FA014
		jmp	sub_11A48
; End of function sub_FC136


; =============== S U B	R O U T	I N E =======================================


sub_FC184:							; CODE XREF: ROM:000FBCE4p
					; ROM:000FBD2Cp
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFCF36,a0
		move.l	#$FFFFD54A,a1
		move.l	#$FFFFD552,a6
		move.l	#$FFFFD562,a5
		move.w	#7,d7

loc_FC1A4:							; CODE XREF: sub_FC184+56j
		move.b	8(a0),d0
		lsl.w	#8,d0
		move.b	9(a0),d0
		move.b	$A(a0),d1
		lsl.w	#8,d1
		move.b	$B(a0),d1
		move.w	d1,(a5)+
		tst.w	d1
		bne.w	loc_FC1C6
		clr.w	d0
		bra.w	loc_FC1CC
; ---------------------------------------------------------------------------

loc_FC1C6:							; CODE XREF: sub_FC184+38j
		mulu.w	#$64,d0					; 'd'
		divu.w	d1,d0

loc_FC1CC:							; CODE XREF: sub_FC184+3Ej
		move.b	d0,(a1)+
		move.b	$C(a0),(a6)+
		move.b	$D(a0),(a6)+
		add.w	#$10,a0
		dbf	d7,loc_FC1A4
		move.l	#$FFFFD542,a1
		move.l	a1,-(sp)
		move.w	#1,d0
		move.w	#6,d7

loc_FC1EE:							; CODE XREF: sub_FC184+6Ej
		move.b	d0,(a1)+
		addq.w	#1,d0
		dbf	d7,loc_FC1EE
		move.l	(sp),a1
		move.l	#$FFFFD54A,a0
		move.l	#$FFFFD552,a6
		move.l	#$FFFFD562,a5

loc_FC20A:							; CODE XREF: sub_FC184+F4j
		move.l	(sp),a1
		move.w	#5,d7
		clr.w	d6

loc_FC212:							; CODE XREF: sub_FC184:loc_FC272j
		move.b	(a1)+,d1
		ext.w	d1
		move.b	(a1),d2
		ext.w	d2
		move.b	(a0,d1.w),d0
		move.b	(a0,d2.w),d3
		cmp.b	d3,d0
		bgt.w	loc_FC272
		blt.w	loc_FC264
		movem.l	d1-d3,-(sp)
		add.w	d1,d1
		add.w	d2,d2
		move.w	(a6,d1.w),d0
		move.w	(a6,d2.w),d3
		cmp.w	d3,d0
		movem.l	(sp)+,d1-d3
		bgt.w	loc_FC272
		blt.w	loc_FC264
		movem.l	d1-d3,-(sp)
		add.w	d1,d1
		add.w	d2,d2
		move.w	(a5,d1.w),d0
		move.w	(a5,d2.w),d3
		cmp.w	d3,d0
		movem.l	(sp)+,d1-d3
		bge.w	loc_FC272

loc_FC264:							; CODE XREF: sub_FC184+A4j
					; sub_FC184+C2j
		move.b	(a1),d0
		move.b	-1(a1),d1
		move.b	d0,-1(a1)
		move.b	d1,(a1)
		st	d6

loc_FC272:							; CODE XREF: sub_FC184+A0j
					; sub_FC184+BEj ...
		dbf	d7,loc_FC212
		tst.w	d6
		bne.s	loc_FC20A
		move.l	(sp)+,a1
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FC184


; =============== S U B	R O U T	I N E =======================================


sub_FC282:							; CODE XREF: ROM:000FBCD6p
					; ROM:000FBCE0p
		movem.l	d0-a6,-(sp)
		move.l	#$D20,d0
		move.l	#$80,d1					; ''
		move.l	#$FFFFCF36,a0
		jsr	sub_1A244
		move.l	#$FFFFD532,a1
		btst	#6,($FFFFC2F8).w
		beq.w	loc_FC2B4
		move.l	#$FFFFD53A,a1

loc_FC2B4:							; CODE XREF: sub_FC282+28j
		move.l	a1,-(sp)
		move.w	#1,d0
		move.w	#6,d7

loc_FC2BE:							; CODE XREF: sub_FC282+40j
		move.b	d0,(a1)+
		addq.w	#1,d0
		dbf	d7,loc_FC2BE
		move.l	(sp),a1
		move.l	#$FFFFCF36,a0

loc_FC2CE:							; CODE XREF: sub_FC282+94j
		move.l	(sp),a1
		move.w	#5,d7
		clr.w	d6

loc_FC2D6:							; CODE XREF: sub_FC282:loc_FC310j
		move.b	(a1)+,d1
		ext.w	d1
		move.b	(a1),d2
		ext.w	d2
		asl.w	#4,d1
		asl.w	#4,d2
		move.b	(a0,d1.w),d0
		move.b	(a0,d2.w),d3
		btst	#6,($FFFFC2F8).w
		beq.w	loc_FC2FC
		move.b	4(a0,d1.w),d0
		move.b	4(a0,d2.w),d3

loc_FC2FC:							; CODE XREF: sub_FC282+6Ej
		cmp.b	d3,d0
		bge.w	loc_FC310
		move.b	(a1),d0
		move.b	-1(a1),d1
		move.b	d0,-1(a1)
		move.b	d1,(a1)
		st	d6

loc_FC310:							; CODE XREF: sub_FC282+7Cj
		dbf	d7,loc_FC2D6
		tst.w	d6
		bne.s	loc_FC2CE
		move.l	(sp)+,a1
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FC282

; ---------------------------------------------------------------------------

loc_FC320:							; CODE XREF: ROM:000187C8j
		movem.l	d0-a6,-(sp)
		move.w	#$FFFF,d0
		jsr	sub_1277A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	3
		dc.b	2
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1B,d0
		moveq	#$C,d1
		jsr	sub_119B8
		lea	word_FC466(pc),a1
		jsr	sub_11DF4
		move.w	($FFFFD40A).w,d0
		move.l	#$FFFFC6CE,a2
		tst.w	($FFFFD404).w
		beq.w	loc_FC366
		move.l	#$FFFFCA32,a2

loc_FC366:							; CODE XREF: ROM:000FC35Cj
		jsr	sub_18AE8
		move.w	($FFFFB028).w,-(sp)
		jsr	sub_11BA4
		move.w	(sp)+,($FFFFB028).w
		addq.w	#1,($FFFFB02A).w
		addq.w	#6,($FFFFB028).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$A
aVs__2:		dc.b	' vs.',4,7,0
; ---------------------------------------------------------------------------
		move.w	($FFFFD408).w,d0
		asl.w	#7,d0
		move.l	#$FFFFB04A,a2
		add.w	d0,a2
		clr.w	d0
		move.b	$66(a2),d0
		move.l	#$FFFFCA32,a2
		tst.w	($FFFFD404).w
		beq.w	loc_FC3BA
		move.l	#$FFFFC6CE,a2

loc_FC3BA:							; CODE XREF: ROM:000FC3B0j
		jsr	sub_18AE8
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	4
		dc.b	9
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$FFFFC6CE,a1
		move.l	$1E(a1),a1
		add.w	4(a1),a1
		jsr	sub_11BA4
		move.w	#$15,($FFFFB028).w
		move.w	($FFFFD574).w,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	4
		dc.b	$A
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$FFFFCA32,a1
		move.l	$1E(a1),a1
		add.w	4(a1),a1
		jsr	sub_11BA4
		move.w	#$15,($FFFFB028).w
		move.w	($FFFFD576).w,d0
		move.w	#3,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$C
		dc.b	$BF
		dc.b	4
		dc.b	$C
aRound:		dc.b	'Round ',0
; ---------------------------------------------------------------------------
		move.w	($FFFFD578).w,d0
		move.w	#2,d1
		jsr	sub_11D3A
		jsr	sub_11BA4
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------
word_FC466:	dc.w	$16					; DATA XREF: ROM:000FC344o
		dc.b	$BF
		dc.b	4
		dc.b	3
aShootoutMode:	dc.b	'SHOOTOUT MODE'
		dc.b	$BF
		dc.b	4
		dc.b	5
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_FC47C:							; CODE XREF: sub_F739E+10F8p
		move.l	#$FFFFB04A,a0
		move.w	#$3FF,d0

loc_FC486:							; CODE XREF: sub_FC47C+Cj
		clr.w	(a0)+
		dbf	d0,loc_FC486
		clr.w	($FFFFDED0).w
		move.w	#1,($FFFFD578).w
		bclr	#3,($FFFFC2FA).w
		clr.w	($FFFFD574).w
		clr.w	($FFFFD576).w
		clr.w	($FFFFD586).w
		clr.w	($FFFFD594).w

loc_FC4AC:
		bsr.w	sub_FC570
		move.w	#1,($FFFFD594).w
		bsr.w	sub_FC570
		clr.w	($FFFFD594).w
		rts
; End of function sub_FC47C


; =============== S U B	R O U T	I N E =======================================


sub_FC4C0:							; CODE XREF: ROM:0000ECE2p
		btst	#3,($FFFFC2FA).w
		beq.w	loc_FC4D0
		jmp	loc_172E4
; ---------------------------------------------------------------------------

loc_FC4D0:							; CODE XREF: sub_FC4C0+6j
		movem.l	d0-a6,-(sp)
		bsr.w	sub_FE756
		move.w	($FFFFD594).w,($FFFFD404).w
		move.w	#$B,($FFFFD408).w
		move.l	#$FFFFD57A,a0
		tst.w	($FFFFD594).w
		beq.w	loc_FC4FE
		move.w	#5,($FFFFD408).w
		move.l	#$FFFFD588,a0

loc_FC4FE:							; CODE XREF: sub_FC4C0+2Ej
		move.w	($FFFFD586).w,d0
		add.w	d0,d0
		move.w	(a0,d0.w),($FFFFD40A).w
		bclr	#2,($FFFFB7AC).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FC4C0


; =============== S U B	R O U T	I N E =======================================


sub_FC516:							; CODE XREF: sub_F37C+60p
		movem.l	d0-a6,-(sp)
		btst	#0,($FFFFD595).w
		beq.w	loc_FC54A
		cmp.w	#5,($FFFFD578).w
		blt.w	loc_FC54A
		move.w	($FFFFD574).w,d0
		cmp.w	($FFFFD576).w,d0
		beq.w	loc_FC54A
		bset	#3,($FFFFC2FA).w
		jsr	sub_FD618
		bra.w	loc_FC56A
; ---------------------------------------------------------------------------

loc_FC54A:							; CODE XREF: sub_FC516+Aj
					; sub_FC516+14j ...
		eor.w	#1,($FFFFD594).w
		bne.w	loc_FC56A
		addq.w	#1,($FFFFD586).w
		cmp.w	#5,($FFFFD586).w
		blt.w	loc_FC566
		clr.w	($FFFFD586).w

loc_FC566:							; CODE XREF: sub_FC516+48j
		addq.w	#1,($FFFFD578).w

loc_FC56A:							; CODE XREF: sub_FC516+30j
					; sub_FC516+3Aj
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FC516


; =============== S U B	R O U T	I N E =======================================


sub_FC570:							; CODE XREF: sub_FC47C:loc_FC4ACp
					; sub_FC47C+3Ap
		move.l	#$FFFFD586,a0
		move.w	($FFFFC330).w,d0
		tst.w	($FFFFD594).w
		beq.w	loc_FC58C
		move.w	($FFFFC332).w,d0
		move.l	#$FFFFD594,a0

loc_FC58C:							; CODE XREF: sub_FC570+Ej
		move.l	#$30E,a1
		asl.w	#2,d0
		move.l	(a1,d0.w),a1
		add.w	6(a1),a1
		move.w	#5,d0

loc_FC5A0:							; CODE XREF: sub_FC570+38j
		clr.w	d1
		move.b	(a1)+,d1
		subq.w	#1,d1
		move.w	d1,-(a0)
		dbf	d0,loc_FC5A0
		rts
; End of function sub_FC570

; ---------------------------------------------------------------------------

loc_FC5AE:							; CODE XREF: sub_12638+36p
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BF
		dc.b	1
		dc.b	$D
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#6,d1
		move.w	#$1E,d0
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$BF
		dc.b	2
		dc.b	$E
aShootoutWonBy:	dc.b	'SHOOTOUT WON BY'
; ---------------------------------------------------------------------------
		addq.w	#2,($FFFFB02A).w
		move.w	($FFFFC330).w,d1
		move.w	($FFFFD574).w,d0
		cmp.w	($FFFFD576).w,d0
		bgt.w	loc_FC5FE
		move.w	($FFFFC332).w,d1

loc_FC5FE:							; CODE XREF: ROM:000FC5F6j
		asl.w	#2,d1
		move.l	#$30E,a1
		move.l	(a1,d1.w),a1
		add.w	4(a1),a1
		move.w	#2,($FFFFB028).w
		jsr	sub_11DF4
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------

loc_FC620:							; DATA XREF: ROM:00019A34o
		btst	#0,($FFFFC2FA).w
		beq.w	loc_FC6C6
		moveq	#0,d0
		moveq	#$1C,d1
		jsr	loc_9BD8
		clr.w	($FFFFD5AC).w
		clr.w	($FFFFD5B0).w
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FF
		dc.b	2
		dc.b	$FD
		dc.b	0
		dc.b	$FC
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$1C,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		bsr.w	sub_FCA1E
		bsr.w	loc_FCA28
		move.w	#0,($FFFFD5B8).w

loc_FC666:							; CODE XREF: ROM:000FC7ACj
					; ROM:000FC84Cj
		bsr.w	loc_FC986
		bsr.w	loc_FCB1A
		bsr.w	loc_FCAB8

loc_FC672:							; CODE XREF: ROM:000FC6C4j
					; ROM:000FC94Aj ...
		jsr	sub_80BA
		jsr	sub_A41E
		jsr	sub_11318
		btst	#7,d1
		bne.w	loc_FC6C6
		move.w	#1,d0
		btst	#1,d1
		bne.w	loc_FC952
		move.w	#$FFFF,d0
		btst	#0,d1
		bne.w	loc_FC952
		move.w	#0,d0
		btst	#2,d1
		bne.w	loc_FC946
		move.w	#5,d0
		btst	#3,d1
		bne.w	loc_FC946
		btst	#5,d1
		bne.w	loc_FC6CC
		bra.s	loc_FC672
; ---------------------------------------------------------------------------

loc_FC6C6:							; CODE XREF: ROM:000FC626j
					; ROM:000FC688j
		jmp	loc_9CD8
; ---------------------------------------------------------------------------

loc_FC6CC:							; CODE XREF: ROM:000FC6C0j
		move.w	#$C000,d7
		move.l	#$FFFFD57A,a0
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_FC6E6
		move.l	#$FFFFD588,a0

loc_FC6E6:							; CODE XREF: ROM:000FC6DCj
		move.w	($FFFFD5B8).w,d0
		asl.w	#1,d0
		move.w	(a0,d0.w),d6
		jsr	sub_9F40
		move.w	d0,d1
		jsr	sub_9F9A
		sub.w	d1,d0
		subq.w	#1,d0
		cmp.w	#5,($FFFFD5B8).w
		bne.w	loc_FC712
		move.w	d1,d0
		subq.w	#1,d0
		clr.w	d1

loc_FC712:							; CODE XREF: ROM:000FC708j
		move.w	d0,($FFFFD5B6).w
		clr.w	($FFFFD5B0).w
		clr.w	($FFFFD5AE).w
		move.w	#$C018,a0
		clr.w	d2

loc_FC724:							; CODE XREF: ROM:000FC736j
		move.b	d1,(a0,d2.w)
		cmp.w	d1,d6
		bne.w	loc_FC732
		move.w	d2,($FFFFD5AE).w

loc_FC732:							; CODE XREF: ROM:000FC72Aj
		addq.w	#1,d1
		addq.w	#1,d2
		dbf	d0,loc_FC724
		bsr.w	sub_FCA04
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$F8
		dc.b	0
		dc.b	3
		dc.b	1
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$12,d0
		move.w	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	$15
		dc.b	$FC
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$12,d0
		move.w	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	2
		dc.b	1
aSelectPlayer_0: dc.b	'{Select Player}'
; ---------------------------------------------------------------------------
		clr.w	d0
		bra.w	loc_FC7EE
; ---------------------------------------------------------------------------

loc_FC796:							; CODE XREF: ROM:000FC7DCj
					; ROM:000FC7E2j ...
		jsr	sub_80BA
		jsr	sub_A41E
		jsr	sub_11318
		btst	#7,d1
		bne.w	loc_FC666
		btst	#5,d1
		bne.w	loc_FC820
		moveq	#1,d0
		btst	#1,d1
		bne.w	loc_FC7EE
		btst	#3,d1
		bne.w	loc_FC7DE
		moveq	#$FFFFFFFF,d0
		btst	#0,d1
		bne.w	loc_FC7EE
		btst	#2,d1
		bne.w	loc_FC7DE
		bra.s	loc_FC796
; ---------------------------------------------------------------------------

loc_FC7DE:							; CODE XREF: ROM:000FC7C6j
					; ROM:000FC7D8j
		add.w	($FFFFD5AC).w,d0
		bmi.s	loc_FC796
		move.w	d0,($FFFFD5AC).w
		bsr.w	loc_FC850
		bra.s	loc_FC796
; ---------------------------------------------------------------------------

loc_FC7EE:							; CODE XREF: ROM:000FC792j
					; ROM:000FC7BEj ...
		add.w	($FFFFD5AE).w,d0
		bmi.s	loc_FC796
		cmp.w	($FFFFD5B6).w,d0
		bgt.s	loc_FC796
		move.w	d0,($FFFFD5AE).w
		cmp.w	($FFFFD5B0).w,d0
		bgt.w	loc_FC80A
		move.w	d0,($FFFFD5B0).w

loc_FC80A:							; CODE XREF: ROM:000FC802j
		subq.w	#5,d0
		cmp.w	($FFFFD5B0).w,d0
		ble.w	loc_FC818
		move.w	d0,($FFFFD5B0).w

loc_FC818:							; CODE XREF: ROM:000FC810j
		bsr.w	loc_FC850
		bra.w	loc_FC796
; ---------------------------------------------------------------------------

loc_FC820:							; CODE XREF: ROM:000FC7B4j
		move.w	#$C018,a3
		add.w	($FFFFD5AE).w,a3
		clr.w	d0
		move.b	(a3),d0
		move.w	($FFFFD5B8).w,d2
		move.l	#$FFFFD57A,a0
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_FC846
		move.l	#$FFFFD588,a0

loc_FC846:							; CODE XREF: ROM:000FC83Cj
		asl.w	#1,d2
		move.w	d0,(a0,d2.w)
		bra.w	loc_FC666
; ---------------------------------------------------------------------------

loc_FC850:							; CODE XREF: ROM:000FC7E8p
					; ROM:loc_FC818p
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$16
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------

loc_FC85C:							; CODE XREF: ROM:000FC88Cj
		move.l	#word_1940E,a1
		cmp.w	#5,($FFFFD5B8).w
		bne.w	loc_FC872
		move.l	#word_19570,a1

loc_FC872:							; CODE XREF: ROM:000FC868j
		move.w	($FFFFD5AC).w,d0
		bra.w	loc_FC87E
; ---------------------------------------------------------------------------

loc_FC87A:							; CODE XREF: ROM:000FC880j
		add.w	(a1),a1
		addq.w	#4,a1

loc_FC87E:							; CODE XREF: ROM:000FC876j
		tst.w	(a1)
		dbmi	d0,loc_FC87A
		bpl.w	loc_FC88E
		subq.w	#1,($FFFFD5AC).w
		bra.s	loc_FC85C
; ---------------------------------------------------------------------------

loc_FC88E:							; CODE XREF: ROM:000FC884j
		cmp.l	#word_1940E,a1
		bne.w	loc_FC89E
		move.l	#word_FC91A,a1

loc_FC89E:							; CODE XREF: ROM:000FC894j
		cmp.l	#word_19570,a1
		bne.w	loc_FC8AE
		move.l	#word_FC930,a1

loc_FC8AE:							; CODE XREF: ROM:000FC8A4j
		jsr	sub_11BA4
		move.l	(a1),d4
		move.w	#$C018,a3
		move.w	($FFFFD5B0).w,d2
		move.w	($FFFFD5B6).w,d1
		sub.w	d2,d1
		cmpnw	 5,d1
		bls.w	loc_FC8CE
		moveq	#5,d1

loc_FC8CE:							; CODE XREF: ROM:000FC8C8j
		move.w	#2,($FFFFB02A).w

loc_FC8D4:							; CODE XREF: ROM:000FC914j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$20
		dc.b	$FE
		dc.b	4
		dc.b	$FD
		dc.b	5
		dc.b	$FA
		dc.b	1
		dc.b	'                      '
		dc.b	$FD
		dc.b	5
; ---------------------------------------------------------------------------
		cmp.w	($FFFFD5AE).w,d2
		bne.w	loc_FC906
		move.w	d7,($FFFFB02C).w

loc_FC906:							; CODE XREF: ROM:000FC8FEj
		clr.w	d0
		move.b	(a3,d2.w),d0
		jsr	sub_8D4E
		addq.w	#1,d2
		dbf	d1,loc_FC8D4
		rts
; ---------------------------------------------------------------------------
word_FC91A:	dc.w	$12					; DATA XREF: ROM:000FC898o
aOverall_2:	dc.b	'    Overall    ]'
		dc.b	$1F
		dc.b	$3A
		dc.b	0
		dc.b	$A
word_FC930:	dc.w	$12					; DATA XREF: ROM:000FC8A8o
aOverall_3:	dc.b	'    Overall    ]'
		dc.b	$1B
		dc.b	$F
		dc.b	0
		dc.b	$A
; ---------------------------------------------------------------------------

loc_FC946:							; CODE XREF: ROM:000FC6ACj
					; ROM:000FC6B8j
		cmp.w	($FFFFD5B8).w,d0
		beq.w	loc_FC672
		bra.w	loc_FC976
; ---------------------------------------------------------------------------

loc_FC952:							; CODE XREF: ROM:000FC694j
					; ROM:000FC6A0j
		cmp.w	#5,($FFFFD5B8).w
		beq.w	loc_FC672
		add.w	($FFFFD5B8).w,d0
		bmi.w	loc_FC972
		cmpnw	 5,d0
		blt.w	loc_FC976
		clr.w	d0
		bra.w	loc_FC976
; ---------------------------------------------------------------------------

loc_FC972:							; CODE XREF: ROM:000FC960j
		move.w	#4,d0

loc_FC976:							; CODE XREF: ROM:000FC94Ej
					; ROM:000FC968j ...
		move.w	d0,($FFFFD5B8).w
		bsr.w	loc_FCB1A
		bsr.w	loc_FCAB8
		bra.w	loc_FC672
; ---------------------------------------------------------------------------

loc_FC986:							; CODE XREF: ROM:loc_FC666p
		bsr.w	sub_FCA04
		movem.l	d0-d5/a0-a2,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a1
		add.l	4(a1),a1
		move.w	#$30A,a2
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#1,d4
		moveq	#0,d5
		jsr	sub_1169A
		movem.l	(sp)+,d0-d5/a0-a2
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	7
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$1A,d0
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$14
		dc.b	$BE
		dc.b	$A
		dc.b	4
aShootout_0:	dc.b	'  Shootout  '
		dc.b	$BE
		dc.b	$F
		dc.b	1
; ---------------------------------------------------------------------------
		clr.w	d0
		cmp.w	#$C6CE,a2
		beq.w	loc_FC9FE
		move.w	#$2C,d0					; ','

loc_FC9FE:							; CODE XREF: ROM:000FC9F6j
		jmp	sub_8078

; =============== S U B	R O U T	I N E =======================================


sub_FCA04:							; CODE XREF: ROM:000FC73Ap
					; ROM:loc_FC986p
		jsr	sub_11B92
; End of function sub_FCA04

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$28,d0					; '('
		moveq	#$A,d1
		move.w	#$7FF,d2
		jmp	sub_1197E

; =============== S U B	R O U T	I N E =======================================


sub_FCA1E:							; CODE XREF: ROM:000FC658p
		clr.w	($FFFFBD82).w
		clr.w	($FFFFBDA2).w
		rts
; End of function sub_FCA1E

; ---------------------------------------------------------------------------

loc_FCA28:							; CODE XREF: ROM:000FC65Cp
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FF
		dc.b	2
		dc.b	$FD
		dc.b	0
		dc.b	$FC
		dc.b	$A
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'',4
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	$FD
		dc.b	7
		dc.b	$FC
		dc.b	$C
aShooters:	dc.b	'Shooters'
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$FD
		dc.b	3
		dc.b	$FC
		dc.b	$E
a1_:		dc.b	'1. ',0
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$FD
		dc.b	3
		dc.b	$FC
		dc.b	$10
a2_:		dc.b	'2. ',0
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$FD
		dc.b	3
		dc.b	$FC
		dc.b	$12
a3_:		dc.b	'3. ',0
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$FD
		dc.b	3
		dc.b	$FC
		dc.b	$14
a4_:		dc.b	'4. ',0
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$FD
		dc.b	3
		dc.b	$FC
		dc.b	$16
a5_:		dc.b	'5. ',0
; ---------------------------------------------------------------------------
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$C
		dc.b	$FD
		dc.b	$1A
		dc.b	$FC
		dc.b	$C
aGoalie_0:	dc.b	'Goalie'
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_FCAB8:							; CODE XREF: ROM:000FC66Ep
					; ROM:000FC97Ep
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	2
		dc.b	8
		dc.b	7
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$18,d0
		moveq	#3,d1
		jsr	sub_119B8
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FD
		dc.b	$15
		dc.b	$FC
		dc.b	8
		dc.b	$FE
		dc.b	6
; ---------------------------------------------------------------------------
		move.l	#$FFFFD57A,a1
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_FCAF6
		move.l	#$FFFFD588,a1

loc_FCAF6:							; CODE XREF: ROM:000FCAECj
		move.w	($FFFFD5B8).w,d0
		asl.w	#1,d0
		move.w	(a1,d0.w),d0
		jsr	sub_18A90
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		jsr	sub_11A48
		clr.w	($FFFFB030).w
		rts
; ---------------------------------------------------------------------------

loc_FCB1A:							; CODE XREF: ROM:000FC66Ap
					; ROM:000FC97Ap
		movem.l	d0-a6,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	8
		dc.b	$FD
		dc.b	6
		dc.b	$FC
		dc.b	$E
		dc.b	$FE
		dc.b	6
; ---------------------------------------------------------------------------
		move.w	#0,d1
		move.l	#$FFFFD57A,a0
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_FCB46
		move.l	#$FFFFD588,a0

loc_FCB46:							; CODE XREF: ROM:000FCB3Cj
					; ROM:000FCB92j
		move.w	d1,d0
		asl.w	#1,d0
		move.w	(a0,d0.w),d0
		jsr	sub_18B6E
		move.w	($FFFFB028).w,-(sp)
		cmp.w	($FFFFD5B8).w,d1
		bne.w	loc_FCB66
		move.w	#2,($FFFFB030).w

loc_FCB66:							; CODE XREF: ROM:000FCB5Cj
		cmpnw	 5,d1
		bne.w	loc_FCB7A
		move.w	#$19,($FFFFB028).w
		move.w	#$E,($FFFFB02A).w

loc_FCB7A:							; CODE XREF: ROM:000FCB6Aj
		jsr	sub_11A48
		clr.w	($FFFFB030).w
		move.w	(sp)+,($FFFFB028).w
		addq.w	#2,($FFFFB02A).w
		addq.w	#1,d1
		cmpnw	 6,d1
		blt.s	loc_FCB46
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FCB9A:							; CODE XREF: ROM:loc_178C0p
		movem.l	d0-a0/a2-a6,-(sp)
		move.w	($FFFFC330).w,d1
		ext.l	d1
		move.l	#$FFFFCFFE,a0
		move.l	a0,-(sp)
		jsr	sub_F9BE2
		move.l	(sp)+,a0
		move.l	#$FFFFBF20,a1
		clr.w	d0
		move.b	8(a0),d0
		bne.w	loc_FCBC8
		move.b	#$50,d0					; 'P'

loc_FCBC8:							; CODE XREF: sub_FCB9A+26j
		move.l	a1,-(sp)
		jsr	sub_F998E
		move.l	(sp)+,a1
		movem.l	(sp)+,d0-a0/a2-a6
		rts
; End of function sub_FCB9A

; ---------------------------------------------------------------------------
		movem.l	d0-a0/a2-a6,-(sp)
		move.w	($FFFFC330).w,d1
		ext.l	d1
		move.l	#$FFFFCFFE,a0
		move.l	a0,-(sp)
		jsr	sub_F9BE2
		move.l	(sp)+,a0
		tst.b	9(a0)
		beq.w	loc_FCC3E
		move.l	#$FFFFCF36,a3
		move.l	#word_FCC4C,a1
		move.l	a3,-(sp)
		jsr	sub_F997A
		move.l	(sp)+,a3
		clr.w	d2
		move.b	9(a0),d2
		move.l	#$FFFFBF20,a1
		bset	#7,($FFFFC2F8).w
		bsr.w	sub_FA014
		jsr	sub_11D9E
		move.l	#word_FCC52,a1
		jsr	sub_11D9E
		move.l	a3,a1
		bra.w	loc_FCC44
; ---------------------------------------------------------------------------

loc_FCC3E:							; CODE XREF: ROM:000FCBF6j
		move.l	#word_FCC4A,a1

loc_FCC44:							; CODE XREF: ROM:000FCC3Aj
		movem.l	(sp)+,d0-a0/a2-a6
		rts
; ---------------------------------------------------------------------------
word_FCC4A:	dc.w	2					; DATA XREF: ROM:loc_FCC3Eo
word_FCC4C:	dc.w	6					; DATA XREF: ROM:000FCC00o
aBy_0:		dc.b	'by ',0
word_FCC52:	dc.w	4					; DATA XREF: ROM:000FCC2Co
		dc.b	' ',0
; ---------------------------------------------------------------------------
		movem.l	d0-a0/a2-a6,-(sp)
		clr.w	d2
		move.b	$A(a0),d2
		asl.w	#2,d2
		move.l	#$30E,a1
		move.l	(a1,d2.w),a1
		add.w	4(a1),a1
		movem.l	(sp)+,d0-a0/a2-a6
		rts
; ---------------------------------------------------------------------------

loc_FCC76:							; CODE XREF: sub_7CB0+F67Cp
		jsr	sub_F9C68
		jsr	sub_F71A2
		clr.w	($FFFFD5A4).w
		move.w	#$7A,-(sp)				; 'z'
		jsr	sub_11156
		move.l	#sub_15E4C,($FFFFB03A).w
		jsr	sub_17102
		move.l	a0,-(sp)
		move.l	#$FFFFC6CE,a0
		jsr	sub_F70A2
		move.l	#$FFFFCA32,a0
		jsr	sub_F70A2
		move.l	(sp)+,a0
		jsr	sub_F71A2
		bclr	#1,($FFFFBF78).w
		move.w	#6,($FFFFB006).w
		move.w	#6,($FFFFB00A).w
		move.w	#0,d0
		jsr	sub_11594
		move.w	#2,d4
		move.w	d4,($FFFFD430).w
		add.w	#$24,d4					; '$'
		move.w	d4,($FFFFD432).w
		add.w	#$24,d4					; '$'
		move.w	d4,($FFFFB010).w
		move.l	#unk_A9A18,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$543
		dc.w	$1567
		dc.w	$79AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11F04
		move.w	d4,($FFFFB014).w
		move.l	#unk_AAC5A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$A23
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		move.l	#unk_55B86,a2
		move.w	d4,($FFFFB01E).w
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$542
		dc.w	$2561
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_54E24,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		moveq	#$28,d2					; '('
		moveq	#$1C,d3
		moveq	#3,d5
		jsr	sub_1169A
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$E
		dc.b	$DF
		dc.b	$C
		dc.b	$D
aMatchups:	dc.b	'MATCHUPS',0
; ---------------------------------------------------------------------------
		move.w	#0,($FFFFB030).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$10
		dc.b	$FF
		dc.b	$10
		dc.b	$11
aAdvantage:	dc.b	'ADVANTAGE:',0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$DE
		dc.b	6
		dc.b	3
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$21,d0					; '!'
		moveq	#9,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$DE
		dc.b	1
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#8,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$DE
		dc.b	$1F
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#8,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	1
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_B389C,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#$C,d5
		jsr	sub_1169A
		move.l	#unk_F5AFE,a2
		move.w	d4,($FFFFDEE4).w
		jsr	sub_11738
		move.l	#unk_F5D24,a2
		move.w	d4,($FFFFDEE6).w
		jsr	sub_11738
		clr.w	($FFFFD598).w
		clr.w	($FFFFD59A).w
		move.w	#$10D,($FFFFD5A8).l
		bsr.w	sub_FD46C
		bsr.w	sub_FD084
		bsr.w	loc_FD1FE
		move.w	#$18,($FFFFBD26).w
		move.l	#sub_15E4C,($FFFFB03A).w
		move	#$2500,sr
		jsr	sub_17718
		move.w	#$1B,(a0)+
		move.w	#$1D,(a0)+
		move.w	#$1A,(a0)+
		move.w	#$1D,(a0)+
		jsr	sub_F7318
		tst.w	d0
		bmi.w	loc_FCE84
		move.w	d0,(a0)+
		move.w	#$1D,(a0)+

loc_FCE84:							; CODE XREF: ROM:000FCE7Aj
		move.w	($FFFFC330).w,d0
		move.w	#$1E,(a0)+
		move.w	#$1D,(a0)+
		cmp.w	($FFFFC332).w,d0
		beq.w	loc_FCE9C
		move.w	#$1F,(a0)+

loc_FCE9C:							; CODE XREF: ROM:000FCE94j
		move.w	#$1D,(a0)+
		move.w	#$20,(a0)+				; ' '
		move.w	#$1D,(a0)+
		move.w	#$21,(a0)+				; '!'
		move.w	#$FFFF,(a0)
		clr.w	($FFFFB044).w
		move.w	#$7FFF,($FFFFD5B6).w
		move.l	#$8CA0,($FFFFDEEC).w

loc_FCEC2:							; CODE XREF: ROM:000FCF64j
					; ROM:000FCF7Aj
		moveq	#0,d0
		jsr	sub_11176
		subq.w	#1,($FFFFD5A8).w
		bpl.w	loc_FCEDA
		move.w	#$10E,($FFFFD5A8).l

loc_FCEDA:							; CODE XREF: ROM:000FCECEj
		bsr.w	sub_FD4CA
		btst	#7,d1
		bne.w	loc_FCF7E
		cmp.w	#$10E,($FFFFD5A8).w
		beq.w	loc_FCF06
		btst	#5,d1
		beq.w	loc_FCF16
		move.w	#$7FFF,($FFFFD5B6).w
		move.l	#$8CA0,($FFFFDEEC).w

loc_FCF06:							; CODE XREF: ROM:000FCEECj
		bsr.w	sub_FD4C0
		move.w	#1,d0
		bsr.w	sub_FCF86
		bra.w	loc_FCF4C
; ---------------------------------------------------------------------------

loc_FCF16:							; CODE XREF: ROM:000FCEF4j
		btst	#6,d1
		beq.w	loc_FCF3C
		move.w	#$7FFF,($FFFFD5B6).w
		move.l	#$8CA0,($FFFFDEEC).w
		bsr.w	sub_FD4C0
		move.w	#$FFFF,d0
		bsr.w	sub_FCF86
		bra.w	loc_FCF4C
; ---------------------------------------------------------------------------

loc_FCF3C:							; CODE XREF: ROM:000FCF1Aj
		clr.w	($FFFFB044).w
		btst	#1,d1
		beq.w	loc_FCF4C
		st	($FFFFB044).w

loc_FCF4C:							; CODE XREF: ROM:000FCF12j
					; ROM:000FCF38j ...
		btst	#1,($FFFFDED5).w
		beq.w	loc_FCF5A
		st	($FFFFB044).w

loc_FCF5A:							; CODE XREF: ROM:000FCF52j
		jsr	loc_17730
		subq.w	#1,($FFFFD5B6).w
		bpl.w	loc_FCEC2
		tst.w	($FFFFD04A).w
		beq.w	loc_FCF7E
		move.w	#$FFFF,($FFFFD5B6).w
		subq.l	#1,($FFFFDEEC).w
		bpl.w	loc_FCEC2

loc_FCF7E:							; CODE XREF: ROM:000FCEE2j
					; ROM:000FCF6Cj
		move.w	#0,($FFFFB030).w
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FCF86:							; CODE XREF: ROM:000FCF0Ep
					; ROM:000FCF34p
		move.w	#$10E,($FFFFD5A8).w
		add.w	($FFFFD598).w,d0
		bmi.w	loc_FCFA2
		cmpnw	 7,d0
		blt.w	loc_FCFA6
		clr.w	d0
		bra.w	loc_FCFA6
; ---------------------------------------------------------------------------

loc_FCFA2:							; CODE XREF: sub_FCF86+Aj
		move.w	#6,d0

loc_FCFA6:							; CODE XREF: sub_FCF86+12j
					; sub_FCF86+18j
		move.w	d0,($FFFFD598).w
		bsr.w	sub_FD46C
		bsr.w	sub_FD084
		bsr.w	loc_FD1FE
		rts
; End of function sub_FCF86


; =============== S U B	R O U T	I N E =======================================


sub_FCFB8:							; CODE XREF: ROM:000FD2AAp
					; ROM:000FD2C6p
		bsr.w	sub_FCFDC
		move.l	a2,-(sp)
		move.l	a0,a2
		bsr.w	sub_FE172
		move.l	(sp)+,a2
		move.l	d0,-(sp)
		move.w	#2,d1
		jsr	sub_11D3A
		jsr	sub_11A48
		move.l	(sp)+,d0
		rts
; End of function sub_FCFB8


; =============== S U B	R O U T	I N E =======================================


sub_FCFDC:							; CODE XREF: sub_FCFB8p ROM:000FD394p	...
		move.l	a1,-(sp)
		move.w	($FFFFB028).w,-(sp)
		move.l	#word_FCFF6,a1
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB028).w
		move.l	(sp)+,a1
		rts
; End of function sub_FCFDC

; ---------------------------------------------------------------------------
word_FCFF6:	dc.w	4					; DATA XREF: sub_FCFDC+6o
		dc.b	'  '
word_FCFFA:	dc.w	$A					; DATA XREF: ROM:000FD216o
aShooting:	dc.b	'Shooting'
		dc.w	$A
aSkating:	dc.b	'Skating',0
		dc.w	$A
aPassing:	dc.b	'Passing',0
		dc.w	$A
aDefense:	dc.b	'Defense',0
		dc.w	$A
aChecking_1:	dc.b	'Checking'
		dc.w	$A
aFighting_1:	dc.b	'Fighting'
		dc.w	$E
aGoalkeeping:	dc.b	'Goalkeeping',0
		dc.w	$12
aPowerPlayAdv_:	dc.b	'Power Play Adv.',0
		dc.w	$1A
aHomeTeamAdv_:	dc.b	'-    Home Team Adv.    ',0
		dc.w	$12
aOverall_4:	dc.b	'    Overall     '
		dc.w	$FFFF

; =============== S U B	R O U T	I N E =======================================


sub_FD084:							; CODE XREF: ROM:000FCE42p
					; sub_FCF86+28p
		movem.l	d0-a6,-(sp)
		tst.w	($FFFFD598).w
		bne.w	loc_FD0E6
		move.w	($FFFFC332).w,d3
		bsr.w	sub_FD1F0
		move.w	#2,($FFFFB028).w
		move.w	#$F,($FFFFB02A).w
		move.w	($FFFFD432).w,d4
		move.w	#4,d5
		move.w	#$6000,($FFFFB02C).w
		bsr.w	sub_FD17E
		move.w	($FFFFC330).w,d3
		bsr.w	sub_FD1F0
		move.w	#$20,($FFFFB028).w			; ' '
		move.w	#$F,($FFFFB02A).w
		move.w	($FFFFD430).w,d4
		move.w	#0,($FFFFB02C).w
		move.w	#2,d5
		bsr.w	sub_FD17E
		move.w	#$64,($FFFFBD26).w			; 'd'
		bra.w	loc_FD144
; ---------------------------------------------------------------------------

loc_FD0E6:							; CODE XREF: sub_FD084+8j
		move.w	($FFFFC332).w,d1
		move.w	($FFFFD59E).w,d0
		jsr	sub_FAE26
		move.w	#2,($FFFFB028).w
		move.w	#$F,($FFFFB02A).w
		move.w	($FFFFD432).w,d4
		move.w	#2,d5
		move.w	#0,($FFFFB02C).w
		bsr.w	sub_FD14A
		move.w	($FFFFC330).w,d1
		move.w	($FFFFD59C).w,d0
		jsr	sub_FAE26
		move.w	#$20,($FFFFB028).w			; ' '
		move.w	#$F,($FFFFB02A).w
		move.w	($FFFFD430).w,d4
		move.w	#0,d5
		move.w	#0,($FFFFB02C).w
		bsr.w	sub_FD14A
		move.w	#$64,($FFFFBD26).w			; 'd'

loc_FD144:							; CODE XREF: sub_FD084+5Ej
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FD084


; =============== S U B	R O U T	I N E =======================================


sub_FD14A:							; CODE XREF: sub_FD084+8Ap
					; sub_FD084+B6p
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		move.l	#off_C63F8,a0
		add.l	(a0),a0
		tst.l	(a2)
		bne.w	loc_FD16E
		move.l	#off_C63F8,a1
		add.l	4(a1),a1
		tst.l	(a2)+
		bra.w	loc_FD170
; ---------------------------------------------------------------------------

loc_FD16E:							; CODE XREF: sub_FD14A+10j
		add.l	(a2)+,a1

loc_FD170:							; CODE XREF: sub_FD14A+20j
		bsr.w	sub_FE98A
		move.l	#$FFFFDA1E,a2
		bra.w	loc_FD1A4
; End of function sub_FD14A


; =============== S U B	R O U T	I N E =======================================


sub_FD17E:							; CODE XREF: sub_FD084+2Ep
					; sub_FD084+54p
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d3
		cmpnw	 4,d5
		beq.w	loc_FD196
		sub.w	#$20,d3					; ' '
		bra.w	loc_FD19A
; ---------------------------------------------------------------------------

loc_FD196:							; CODE XREF: sub_FD17E+Cj
		sub.w	#$40,d3					; '@'

loc_FD19A:							; CODE XREF: sub_FD17E+14j
		move.l	#unk_FF462,a0
		add.w	d3,a0
		add.l	(a2)+,a1

loc_FD1A4:							; CODE XREF: sub_FD14A+30j
		move.w	#6,d3
		move.w	#6,d2
		clr.w	d0
		clr.w	d1
		jsr	sub_1169A
		cmp.w	#$14,($FFFFB028).w
		bgt.w	loc_FD1CC
		move.w	#2,($FFFFB028).l
		bra.w	loc_FD1D4
; ---------------------------------------------------------------------------

loc_FD1CC:							; CODE XREF: sub_FD17E+3Ej
		move.w	#$20,($FFFFB028).l			; ' '

loc_FD1D4:							; CODE XREF: sub_FD17E+4Aj
		move.w	#$19,($FFFFB02A).l
		move.w	#8,d0
		move.w	#3,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		rts
; End of function sub_FD17E


; =============== S U B	R O U T	I N E =======================================


sub_FD1F0:							; CODE XREF: sub_FD084+10p
					; sub_FD084+36p
		asl.w	#2,d3
		move.l	#off_F86F2,a0
		move.l	(a0,d3.w),a0
		rts
; End of function sub_FD1F0

; ---------------------------------------------------------------------------

loc_FD1FE:							; CODE XREF: ROM:000FCE46p
					; sub_FCF86+2Cp
		tst.w	($FFFFD598).w
		bne.w	loc_FD2D2
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	0
		dc.b	1
		dc.b	$16
		dc.b	$16
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		lea	word_FCFFA(pc),a1
		move.w	#0,d7

loc_FD21E:							; CODE XREF: ROM:000FD256j
		move.w	(a1),d0
		asr.w	#1,d0
		neg.w	d0
		add.w	#$16,d0
		move.w	d0,($FFFFB028).w
		cmpnw	 9,d7
		beq.w	loc_FD23A
		add.w	(a1),a1
		bra.w	loc_FD252
; ---------------------------------------------------------------------------

loc_FD23A:							; CODE XREF: ROM:000FD230j
		move.w	#2,($FFFFB030).w
		move.w	($FFFFB02A).w,-(sp)
		subq.w	#3,($FFFFB02A).w
		jsr	sub_11A48
		move.w	(sp)+,($FFFFB02A).w

loc_FD252:							; CODE XREF: ROM:000FD236j
		addq.w	#1,d7
		tst.w	(a1)
		bpl.s	loc_FD21E
		move.w	#2,($FFFFB030).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$2E
		dc.b	$FF
		dc.b	0
		dc.b	$17
		dc.b	'                                        ',0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$22
		dc.b	$16
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$FFFFC6CE,a0
		move.l	#$FFFFCA32,a2
		bsr.w	sub_FCFB8
		move.w	d0,($FFFFD5A0).w
		move.w	#2,($FFFFB030).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	4
		dc.b	$16
		dc.b	0
; ---------------------------------------------------------------------------
		exg	a0,a2
		bsr.w	sub_FCFB8
		move.w	d0,($FFFFD5A2).w
		bra.w	locret_FD3FE
; ---------------------------------------------------------------------------

loc_FD2D2:							; CODE XREF: ROM:000FD202j
		move.w	($FFFFD598).w,d0
		subq.w	#1,d0
		move.l	#word_FD400,a1
		bra.w	loc_FD2E4
; ---------------------------------------------------------------------------

loc_FD2E2:							; CODE XREF: ROM:loc_FD2E4j
		add.w	(a1),a1

loc_FD2E4:							; CODE XREF: ROM:000FD2DEj
		dbf	d0,loc_FD2E2
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	0
		dc.b	1
		dc.b	$D
		dc.b	$13
		dc.b	$F9
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		jsr	sub_11A48
		move.w	#$16,($FFFFB02A).w
		move.w	#2,($FFFFB030).w
		move.l	#$FFFFC6CE,a2
		move.l	(dword_19420).l,d4
		tst.w	($FFFFD5AA).w
		bne.w	loc_FD324
		move.l	(dword_19582).l,d4

loc_FD324:							; CODE XREF: ROM:000FD31Aj
		move.w	($FFFFD59C).w,d0
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$2E
		dc.b	$FF
		dc.b	0
		dc.b	$17
		dc.b	'                                        ',0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$23
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_FD89A
		bsr.w	sub_FA9F8
		mulu.w	#$64,d0					; 'd'
		divu.w	d1,d0
		bsr.w	sub_FEF7C
		move.w	#2,d1
		move.w	d0,($FFFFD5A0).w
		jsr	sub_11D3A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$22
		dc.b	$16
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_FCFDC
		jsr	sub_11A48
		move.l	#$FFFFCA32,a2
		move.l	(dword_19420).l,d4
		tst.w	($FFFFD5AA).w
		bne.w	loc_FD3B8
		move.l	(dword_19582).l,d4

loc_FD3B8:							; CODE XREF: ROM:000FD3AEj
		move.w	($FFFFD59E).w,d0
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_FD89A
		bsr.w	sub_FA9F8
		mulu.w	#$64,d0					; 'd'
		divu.w	d1,d0
		bsr.w	sub_FEF7C
		move.w	d0,($FFFFD5A2).w
		move.w	#2,d1
		jsr	sub_11D3A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	4
		dc.b	$16
		dc.b	0
; ---------------------------------------------------------------------------
		bsr.w	sub_FCFDC
		jsr	sub_11A48

locret_FD3FE:							; CODE XREF: ROM:000FD2CEj
		rts
; ---------------------------------------------------------------------------
word_FD400:	dc.w	$12					; DATA XREF: ROM:000FD2D8o
aCenter:	dc.b	'     center     '
		dc.w	$12
aLeftForward:	dc.b	' left forward   '
		dc.w	$12
aRightForward:	dc.b	' right forward  '
		dc.w	$12
aLeftDefenseman: dc.b	'left defenseman '
		dc.w	$12
aRightDefenseman: dc.b	'right defenseman'
		dc.w	$12
aGoalie_1:	dc.b	'     goalie     '

; =============== S U B	R O U T	I N E =======================================


sub_FD46C:							; CODE XREF: ROM:000FCE3Ep
					; sub_FCF86+24p
		movem.l	d0/a0-a1,-(sp)
		move.l	#$FFFFC6CE,a0
		bsr.w	sub_FD492
		move.w	d0,($FFFFD59C).w
		move.l	#$FFFFCA32,a0
		bsr.w	sub_FD492
		move.w	d0,($FFFFD59E).w
		movem.l	(sp)+,d0/a0-a1
		rts
; End of function sub_FD46C


; =============== S U B	R O U T	I N E =======================================


sub_FD492:							; CODE XREF: sub_FD46C+Ap
					; sub_FD46C+18p
		move.l	$1E(a0),a0
		add.w	6(a0),a0
		move.w	($FFFFD598).w,d0
		move.l	#byte_FD4B8,a1
		move.b	(a1,d0.w),d0
		ext.w	d0
		move.w	d0,($FFFFD5AA).w
		move.b	(a0,d0.w),d0
		ext.w	d0
		subq.w	#1,d0
		rts
; End of function sub_FD492

; ---------------------------------------------------------------------------
byte_FD4B8:	dc.b	0					; DATA XREF: sub_FD492+Co
		dc.b	4
		dc.b	3
		dc.b	5
		dc.b	1
		dc.b	2
		dc.b	0
		dc.b	0

; =============== S U B	R O U T	I N E =======================================


sub_FD4C0:							; CODE XREF: ROM:loc_FCF06p
					; ROM:000FCF2Cp
		clr.w	($FFFFD5A6).w
		clr.w	($FFFFD5A4).w
		rts
; End of function sub_FD4C0


; =============== S U B	R O U T	I N E =======================================


sub_FD4CA:							; CODE XREF: ROM:loc_FCEDAp
		movem.l	d0-a6,-(sp)
		addq.w	#1,($FFFFD5A6).w
		cmp.w	#7,($FFFFD5A6).w
		blt.w	loc_FD4F4
		clr.w	($FFFFD5A6).w
		addq.w	#1,($FFFFD5A4).w
		cmp.w	#6,($FFFFD5A4).w
		blt.w	loc_FD4F4
		move.w	#5,($FFFFD5A4).w

loc_FD4F4:							; CODE XREF: sub_FD4CA+Ej
					; sub_FD4CA+20j
		move.w	($FFFFD5A4).w,d1
		asl.w	#2,d1
		move.l	#off_FD548,a1
		move.w	($FFFFD5A0).w,d0
		cmp.w	($FFFFD5A2).w,d0
		bgt.w	loc_FD516
		beq.w	loc_FD51E
		move.l	#off_FD560,a1

loc_FD516:							; CODE XREF: sub_FD4CA+3Ej
		move.l	(a1,d1.w),a1
		bra.w	loc_FD524
; ---------------------------------------------------------------------------

loc_FD51E:							; CODE XREF: sub_FD4CA+42j
		move.l	#word_FD5A8,a1

loc_FD524:							; CODE XREF: sub_FD4CA+50j
		move.w	#2,($FFFFB030).w
		jsr	sub_11B92
; End of function sub_FD4CA

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	$13
		dc.b	$16
		dc.b	0
; ---------------------------------------------------------------------------
		jsr	sub_11A48
		move.w	#0,($FFFFB030).w
		movem.l	(sp)+,d0-a6
		rts
; ---------------------------------------------------------------------------
off_FD548:	dc.l	word_FD578				; DATA XREF: sub_FD4CA+30o
		dc.l	word_FD57E
		dc.l	word_FD584
		dc.l	word_FD58A
		dc.l	word_FD58A
		dc.l	word_FD58A
off_FD560:	dc.l	word_FD590				; DATA XREF: sub_FD4CA+46o
		dc.l	word_FD596
		dc.l	word_FD59C
		dc.l	word_FD5A2
		dc.l	word_FD5A2
		dc.l	word_FD5A2
word_FD578:	dc.w	6					; DATA XREF: ROM:off_FD548o
		dc.b	'   ',0
word_FD57E:	dc.w	6					; DATA XREF: ROM:000FD54Co
		dc.b	']  ',0
word_FD584:	dc.w	6					; DATA XREF: ROM:000FD550o
		dc.b	']] ',0
word_FD58A:	dc.w	6					; DATA XREF: ROM:000FD554o
					; ROM:000FD558o ...
asc_FD58C:	dc.b	']]]',0
word_FD590:	dc.w	6					; DATA XREF: ROM:off_FD560o
		dc.b	'   ',0
word_FD596:	dc.w	6					; DATA XREF: ROM:000FD564o
		dc.b	'  [',0
word_FD59C:	dc.w	6					; DATA XREF: ROM:000FD568o
		dc.b	' [[',0
word_FD5A2:	dc.w	6					; DATA XREF: ROM:000FD56Co
					; ROM:000FD570o ...
		dc.b	'[[[',0
word_FD5A8:	dc.w	6					; DATA XREF: sub_FD4CA:loc_FD51Eo
		dc.b	'   ',0

; =============== S U B	R O U T	I N E =======================================


sub_FD5AE:							; CODE XREF: ROM:000178E6p
					; ROM:000178FAp
		movem.l	d0/a0/a2,-(sp)
		move.l	a2,a1
		move.l	$1E(a1),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		add.w	(a1),a1
		cmp.w	#2,(a1)
		bne.w	loc_FD5DE
		move.l	#word_FD5E4,a1
		cmp.l	#$FFFFC6CE,a2
		beq.w	loc_FD5DE
		move.l	#word_FD5EA,a1

loc_FD5DE:							; CODE XREF: sub_FD5AE+16j
					; sub_FD5AE+26j
		movem.l	(sp)+,d0/a0/a2
		rts
; End of function sub_FD5AE

; ---------------------------------------------------------------------------
word_FD5E4:	dc.w	6					; DATA XREF: sub_FD5AE+1Ao
aHome:		dc.b	'Home'
word_FD5EA:	dc.w	$A					; DATA XREF: sub_FD5AE+2Ao
aVisitors:	dc.b	'Visitors'

; =============== S U B	R O U T	I N E =======================================


sub_FD5F4:							; CODE XREF: ROM:0001790Ep
					; ROM:00017922p
		movem.l	d0/a0/a2,-(sp)
		move.l	a2,a1
		move.l	$1E(a1),a1
		add.w	4(a1),a1
		add.w	(a1),a1
		add.w	(a1),a1
		add.w	(a1),a1
		movem.l	(sp)+,d0/a0/a2
		rts
; End of function sub_FD5F4

; ---------------------------------------------------------------------------
		movem.l	d0-a6,-(sp)
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FD618:							; CODE XREF: sub_FC516+2Ap
		movem.l	d0-a6,-(sp)
		bset	#2,($FFFFC2EE).w
		jsr	sub_AFB6
		move.w	#$60,($FFFFDED0).w			; '`'
		move.l	#$FFFFB04A,a3
		move.w	#0,d0
		move.w	($FFFFD574).w,d2
		cmp.w	($FFFFD576).w,d2
		bgt.w	loc_FD648
		move.w	#6,d0

loc_FD648:							; CODE XREF: sub_FD618+28j
		asl.w	#7,d0
		add.w	d0,a3
		move.w	#4,d2
		move.w	#6,d3
		bra.w	loc_FD6A4
; ---------------------------------------------------------------------------

loc_FD658:							; CODE XREF: sub_FD618+92j
		movem.w	d2-d3,-(sp)
		jsr	sub_15AA4
		move.w	#2,$34(a3)
		move.w	#$F0,d0					; ''
		tst.w	($FFFFBD18).w
		bmi.w	loc_FD678
		move.w	#$FF10,d0

loc_FD678:							; CODE XREF: sub_FD618+58j
		add.w	($FFFFBD18).w,d0
		move.w	d0,$14(a3)
		move.w	#0,(a3)
		move.w	#7,d0
		jsr	sub_10658
		bclr	#5,$62(a3)
		bclr	#1,$63(a3)
		bclr	#2,$62(a3)
		movem.w	(sp)+,d2-d3

loc_FD6A4:							; CODE XREF: sub_FD618+3Cj
		add.l	#$80,a3					; ''
		dbf	d2,loc_FD658
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FD618

; ---------------------------------------------------------------------------
		movem.l	d0-a0,-(sp)
		move.w	$34(a3),d7
		subq.w	#1,d7
		move.l	#word_FD71C,a0
		btst	#7,$62(a3)
		beq.w	loc_FD6D4
		move.l	#word_FD72C,a0

loc_FD6D4:							; CODE XREF: ROM:000FD6CAj
		add.w	d7,d7
		move.w	(a0,d7.w),d0
		cmp.w	$44(a3),d0
		blt.w	loc_FD6E6
		move.w	d0,$44(a3)

loc_FD6E6:							; CODE XREF: ROM:000FD6DEj
		move.w	2(a0,d7.w),d0
		cmp.w	$44(a3),d0
		bgt.w	loc_FD6F6
		move.w	d0,$44(a3)

loc_FD6F6:							; CODE XREF: ROM:000FD6EEj
		move.w	4(a0,d7.w),d0
		cmp.w	$46(a3),d0
		blt.w	loc_FD706
		move.w	d0,$46(a3)

loc_FD706:							; CODE XREF: ROM:000FD6FEj
		move.w	6(a0,d7.w),d0
		cmp.w	$46(a3),d0
		bgt.w	loc_FD716
		move.w	d0,$46(a3)

loc_FD716:							; CODE XREF: ROM:000FD70Ej
		movem.l	(sp)+,d0-a0
		rts
; ---------------------------------------------------------------------------
word_FD71C:	dc.w	$FFB5					; DATA XREF: ROM:000FD6BEo
		dc.w	0
		dc.w	0
		dc.w	$108
		dc.w	0
		dc.w	$4B
		dc.w	0
		dc.w	$108
word_FD72C:	dc.w	$FFB5					; DATA XREF: ROM:000FD6CEo
		dc.w	0
		dc.w	$FEF8
		dc.w	0
		dc.w	0
		dc.w	$4B
		dc.w	$FEF8
		dc.w	0

; =============== S U B	R O U T	I N E =======================================


sub_FD73C:							; CODE XREF: sub_F739E+24p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFD572,a0
		moveq	#$12,d0

loc_FD748:							; CODE XREF: sub_FD73C+Ej
		clr.w	(a0)+
		dbf	d0,loc_FD748
		move.l	#$FFFFC6CE,a0
		move.w	#$363,d0

loc_FD758:							; CODE XREF: sub_FD73C+1Ej
		clr.w	(a0)+
		dbf	d0,loc_FD758
		move.l	#$FFFFC3A4,a0
		moveq	#$24,d0					; '$'

loc_FD766:							; CODE XREF: sub_FD73C+2Cj
		clr.w	(a0)+
		dbf	d0,loc_FD766
		clr.w	($FFFFC2F2).w
		clr.w	($FFFFC2F4).w
		clr.w	($FFFFC2F8).w
		clr.w	($FFFFC2FA).w
		clr.w	($FFFFD42E).w
		clr.w	($FFFFD43E).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FD73C


; =============== S U B	R O U T	I N E =======================================


sub_FD78A:							; CODE XREF: sub_15EC0+32p
		tst.w	($FFFFD6B4).w
		bmi.w	locret_FD898
		tst.l	($FFFFD6BA).w
		beq.w	locret_FD898
		btst	#4,($FFFFC2F4).w
		bne.w	locret_FD898
		btst	#3,($FFFFC2EE).w
		bne.w	locret_FD898
		btst	#7,($FFFFC2EC).w
		bne.w	locret_FD898
		move.l	($FFFFD6BA).w,a1
		add.l	4(a1),a1
		move.w	($FFFFBD1C).w,d4
		move.w	($FFFFBD18).w,d5
		move.w	($FFFFD6AE).w,d0
		bra.w	loc_FD7F0
; ---------------------------------------------------------------------------
		and.w	#$F,d2
		cmpnw	 3,d2
		bls.w	loc_FD7DE
		moveq	#3,d2

loc_FD7DE:							; CODE XREF: sub_FD78A+4Ej
		bra.w	loc_FD7EA
; ---------------------------------------------------------------------------

loc_FD7E2:							; CODE XREF: sub_FD78A:loc_FD7EAj
		move.w	d3,d0
		add.w	d2,d0
		bsr.w	loc_FD7F0

loc_FD7EA:							; CODE XREF: sub_FD78A:loc_FD7DEj
		dbf	d2,loc_FD7E2
		rts
; ---------------------------------------------------------------------------

loc_FD7F0:							; CODE XREF: sub_FD78A+42j
					; sub_FD78A+5Cp
		ext.w	d0
		beq.w	locret_FD898
		cmpnw	 $40,d6					; '@'
		bge.w	locret_FD898
		movem.l	d0-d5,-(sp)
		add.w	d0,d0
		move.l	a1,a0
		move.w	2(a0,d0.w),d1
		sub.w	(a0,d0.w),d1
		lsr.w	#3,d1
		subq.w	#1,d1
		move.w	d1,-(sp)
		add.w	(a0,d0.w),a0
		move.w	d4,d0
		add.w	#$C0,d0					; ''
		move.w	d0,d1
		sub.w	#$90,d0					; ''
		add.w	#$80,d1					; ''
		move.w	#$170,d2
		sub.w	d5,d2
		move.w	d2,d3
		sub.w	#$80,d2					; ''
		add.w	#$70,d3					; 'p'
		move.w	(sp)+,d4

loc_FD83A:							; CODE XREF: sub_FD78A+106j
		cmp.w	2(a0),d2
		bgt.w	loc_FD88E
		cmp.w	2(a0),d3
		blt.w	loc_FD88E
		cmp.w	(a0),d0
		bgt.w	loc_FD88E
		cmp.w	(a0),d1
		blt.w	loc_FD88E
		move.w	2(a0),d5
		add.w	#$70,d5					; 'p'
		sub.w	d2,d5
		move.w	d5,(a6)+
		move.b	7(a0),(a6)+
		move.b	d6,(a6)+
		move.w	6(a0),d5
		and.w	#$F800,d5
		add.w	4(a0),d5
		add.w	($FFFFD6AC).w,d5
		move.w	d5,(a6)+
		move.w	(a0),d5
		add.w	#$70,d5					; 'p'
		sub.w	d0,d5
		move.w	d5,(a6)+
		addq.w	#1,d6
		cmpnw	 $40,d6					; '@'
		beq.w	loc_FD894

loc_FD88E:							; CODE XREF: sub_FD78A+B4j
					; sub_FD78A+BCj ...
		addq.w	#8,a0
		dbf	d4,loc_FD83A

loc_FD894:							; CODE XREF: sub_FD78A+100j
		movem.l	(sp)+,d0-d5

locret_FD898:							; CODE XREF: sub_FD78A+4j sub_FD78A+Cj ...
		rts
; End of function sub_FD78A


; =============== S U B	R O U T	I N E =======================================


sub_FD89A:							; CODE XREF: ROM:000FD368p
					; ROM:000FD3C8p
		movem.l	d0-a6,-(sp)
		move.w	d0,($FFFFDEE8).w
		move.l	a2,-(sp)
		jsr	sub_18B26
		move.l	a1,a0
		move.w	(a0),d0
		move.w	d0,d5
		tst.b	-1(a0,d5.w)
		bne.w	loc_FD8C4
		subq.w	#1,d0
		tst.b	-2(a0,d5.w)
		bne.w	loc_FD8C4
		subq.w	#1,d0

loc_FD8C4:							; CODE XREF: sub_FD89A+1Aj
					; sub_FD89A+24j
		add.w	($FFFFB028).w,d0
		sub.w	#$29,d0					; ')'
		bmi.w	loc_FD8D6
		neg.w	d0
		add.w	d0,($FFFFB028).w

loc_FD8D6:							; CODE XREF: sub_FD89A+32j
		move.l	a0,a1
		jsr	sub_11BA4
		move.l	(sp)+,a2
		jsr	sub_FF8DE
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FD89A

; ---------------------------------------------------------------------------
		movem.l	d0-a0/a2-a3,-(sp)
		jsr	sub_18AE8
		move.w	(a1),d0
		subq.w	#2,d0
		move.w	d0,2(a1)
		addq.w	#2,a1
		movem.l	(sp)+,d0-a0/a2-a3
		rts
; ---------------------------------------------------------------------------
		dc.w	6
a_:		dc.b	' . ',0
; ---------------------------------------------------------------------------

loc_FD90C:							; DATA XREF: ROM:0001986Co
					; ROM:000199E0o ...
		movem.l	d0-a6,-(sp)
		clr.w	($FFFFD598).w
		moveq	#0,d0
		moveq	#$1C,d1
		jsr	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	$B
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#8,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	$13
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#8,d0
		move.w	#8,d1
		jsr	sub_119B8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$9E
		dc.b	2
		dc.b	$C
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_F86F2,a0
		move.w	($FFFFC332).w,d0
		asl.w	#2,d0
		move.l	(a0,d0.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d0
		move.l	#unk_FF462,a0
		sub.w	#$40,d0					; '@'
		add.w	d0,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#4,d5
		jsr	sub_1169A
		move.l	#$FFFFBD68,a0
		move.l	#$FFFFBD88,a1
		move.w	#7,d0

loc_FD9A6:							; CODE XREF: ROM:000FD9A8j
		move.l	(a0)+,(a1)+
		dbf	d0,loc_FD9A6
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$8E
		dc.b	2
		dc.b	$14
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_F86F2,a0
		move.w	($FFFFC330).w,d0
		asl.w	#2,d0
		move.l	(a0,d0.w),a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		asl.w	#3,d0
		sub.w	#$40,d0					; '@'
		move.l	#unk_FF462,a0
		add.w	d0,a0
		add.l	(a2)+,a1
		clr.w	d1
		clr.w	d0
		moveq	#6,d2
		move.w	#6,d3
		moveq	#4,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	4
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	6
		dc.b	3
aPeriodStats_4:	dc.b	'  Period Stats  '
		dc.b	$BE
		dc.b	6
		dc.b	1
; ---------------------------------------------------------------------------
		btst	#1,($FFFFC2FC).w
		beq.w	loc_FDA76
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1C
		dc.b	$BE
		dc.b	$C
		dc.b	9
		dc.b	'1',$11
		dc.b	9
		dc.b	'2',$16
		dc.b	9
		dc.b	'3',$1A
		dc.b	9
aOt_1:		dc.b	'OT '
		dc.b	9
aTotal:		dc.b	'Total',0
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$1C
		dc.b	$BE
		dc.b	$C
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$11
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$16
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$1A
		dc.b	$A
		dc.b	$3C
		dc.b	$3C
		dc.b	$BE
		dc.b	$20
		dc.b	$A
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
		dc.b	0
; ---------------------------------------------------------------------------
		bra.w	loc_FDAAE
; ---------------------------------------------------------------------------

loc_FDA76:							; CODE XREF: ROM:000FDA2Aj
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$BE
		dc.b	$C
		dc.b	9
		dc.b	$31
		dc.b	$BE
		dc.b	$11
		dc.b	9
		dc.b	'2'
		dc.b	$16
		dc.b	9
		dc.b	$33
		dc.b	$BE
		dc.b	$20
		dc.b	9
aTotal_0:	dc.b	'Total'
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$16
		dc.b	$BE
		dc.b	$C
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$11
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$16
		dc.b	$A
		dc.b	$3C
		dc.b	$BE
		dc.b	$20
		dc.b	$A
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
		dc.b	$3C
; ---------------------------------------------------------------------------

loc_FDAAE:							; CODE XREF: ROM:000FDA72j
		bsr.w	loc_FDB0E
		bsr.w	loc_FDB56
		move.w	#$18,($FFFFBD26).w

loc_FDABC:							; CODE XREF: ROM:000FDADAj
					; ROM:000FDAE0j ...
		jsr	sub_9FB8
		btst	#7,d1
		bne.w	loc_FDB04
		btst	#2,d1
		bne.w	loc_FDAF0
		btst	#3,d1
		bne.w	loc_FDADC
		bra.s	loc_FDABC
; ---------------------------------------------------------------------------

loc_FDADC:							; CODE XREF: ROM:000FDAD6j
		tst.w	($FFFFD598).w
		bne.s	loc_FDABC
		st	($FFFFD598).w
		bsr.w	loc_FDB0E
		bsr.w	loc_FDB56
		bra.s	loc_FDABC
; ---------------------------------------------------------------------------

loc_FDAF0:							; CODE XREF: ROM:000FDACEj
		tst.w	($FFFFD598).w
		beq.s	loc_FDABC
		clr.w	($FFFFD598).w
		bsr.w	loc_FDB0E
		bsr.w	loc_FDB56
		bra.s	loc_FDABC
; ---------------------------------------------------------------------------

loc_FDB04:							; CODE XREF: ROM:000FDAC6j
		movem.l	(sp)+,d0-a6
		jmp	loc_9CD8
; ---------------------------------------------------------------------------

loc_FDB0E:							; CODE XREF: ROM:loc_FDAAEp
					; ROM:000FDAE6p ...
		tst.w	($FFFFD598).w
		beq.w	loc_FDB36
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BE
		dc.b	$12
		dc.b	7
aShots:		dc.b	'Shots'
		dc.b	$BE
		dc.b	$F
		dc.b	$1A
aForGoals:	dc.b	'[ For Goals'
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_FDB36:							; CODE XREF: ROM:000FDB12j
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BE
		dc.b	$12
		dc.b	7
aGoals_3:	dc.b	'Goals'
		dc.b	$BE
		dc.b	$F
		dc.b	$1A
aForShots:	dc.b	'For Shots ]'
; ---------------------------------------------------------------------------
		rts
; ---------------------------------------------------------------------------

loc_FDB56:							; CODE XREF: ROM:000FDAB2p
					; ROM:000FDAEAp ...
		movem.l	d0-a6,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$FFFFCA32,a2
		move.w	#$E,($FFFFB02A).w
		bsr.w	sub_FDB8C
		move.l	#$FFFFC6CE,a2
		move.w	#$16,($FFFFB02A).w
		bsr.w	sub_FDB8C
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FDB8C:							; CODE XREF: ROM:000FDB72p
					; ROM:000FDB82p
		lea	$342(a2),a0
		tst.w	($FFFFD598).w
		beq.w	loc_FDB9C
		lea	$34A(a2),a0

loc_FDB9C:							; CODE XREF: sub_FDB8C+8j
		clr.w	($FFFFD5A8).w
		move.w	#$B,($FFFFB028).w
		bsr.w	sub_FDC28
		cmp.w	#1,($FFFFC466).w
		blt.w	loc_FDBF0
		move.w	#$10,($FFFFB028).w
		bsr.w	sub_FDC28
		cmp.w	#2,($FFFFC466).w
		blt.w	loc_FDBF0
		move.w	#$15,($FFFFB028).w
		bsr.w	sub_FDC28
		cmp.w	#3,($FFFFC466).w
		blt.w	loc_FDBF0
		btst	#1,($FFFFC2FC).w
		beq.w	loc_FDBF0
		move.w	#$1A,($FFFFB028).w
		bsr.w	sub_FDC28

loc_FDBF0:							; CODE XREF: sub_FDB8C+24j
					; sub_FDB8C+38j ...
		move.w	#$21,($FFFFB028).w			; '!'
		jsr	sub_11B92
; End of function sub_FDB8C

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	'   ',0
; ---------------------------------------------------------------------------
		subq.w	#3,($FFFFB028).w
		move.w	($FFFFD5A8).w,d0
		move.w	#2,d1
		cmpnw	 $64,d0					; 'd'
		blt.w	loc_FDC1A
		move.w	#3,d1

loc_FDC1A:							; CODE XREF: ROM:000FDC12j
		jsr	sub_11D3A
		jsr	sub_11BA4
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FDC28:							; CODE XREF: sub_FDB8C+1Ap
					; sub_FDB8C+2Ep ...
		move.w	(a0)+,d0
		add.w	d0,($FFFFD5A8).w
		jsr	sub_11B92
; End of function sub_FDC28

; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	'   ',0
; ---------------------------------------------------------------------------
		subq.w	#3,($FFFFB028).w
		move.w	#2,d1
		cmpnw	 $64,d0					; 'd'
		blt.w	loc_FDC4E
		move.w	#3,d1

loc_FDC4E:							; CODE XREF: ROM:000FDC46j
		jsr	sub_11D3A
		jmp	sub_11BA4
; ---------------------------------------------------------------------------

loc_FDC5A:							; DATA XREF: ROM:0001977Co
					; ROM:00019908o ...
		moveq	#9,d0
		moveq	#$1A,d1
		jsr	loc_9BD8
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	4
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$20,d0					; ' '
		moveq	#6,d1
		jsr	sub_119B8
		jsr	sub_11DE2
; ---------------------------------------------------------------------------
		dc.w	$18
		dc.b	$BD
		dc.b	6
		dc.b	4
aGameStatistics: dc.b	'Game  Statistics'
		dc.b	$BD
		dc.b	7
		dc.b	1
; ---------------------------------------------------------------------------
		move.w	#$2C,d0					; ','
		jsr	sub_8078
		addq.w	#4,($FFFFB028).w
		clr.w	d0
		jsr	sub_8078
		clr.w	($FFFFD5AC).w
		clr.w	($FFFFD5AE).w
		clr.w	($FFFFD5B0).w
		move.w	#$70,($FFFFD59E).w			; 'p'
		tst.w	($FFFFD056).w
		bne.w	loc_FDCCE
		move.w	#$30,($FFFFD59E).w			; '0'

loc_FDCCE:							; CODE XREF: ROM:000FDCC4j
		bsr.w	sub_FDE24
		bsr.w	loc_FDDAA

loc_FDCD6:							; CODE XREF: ROM:000FDD18j
		jsr	sub_80BA
		jsr	sub_A41E
		btst	#7,d3
		beq.w	loc_FDCF0
		jmp	loc_9CD8
; ---------------------------------------------------------------------------

loc_FDCF0:							; CODE XREF: ROM:000FDCE6j
		jsr	sub_112F0
		move.w	#$FFFE,d0
		btst	#0,d3
		bne.w	loc_FDD0C
		neg.w	d0
		btst	#1,d3
		beq.w	loc_FDD10

loc_FDD0C:							; CODE XREF: ROM:000FDCFEj
		move.w	d0,($FFFFD5B0).w

loc_FDD10:							; CODE XREF: ROM:000FDD08j
		bsr.w	sub_FDD1A
		bsr.w	loc_FDDAA
		bra.s	loc_FDCD6

; =============== S U B	R O U T	I N E =======================================


sub_FDD1A:							; CODE XREF: ROM:loc_FDD10p

; FUNCTION CHUNK AT 000FE14A SIZE 00000002 BYTES

		move.w	($FFFFD5B0).w,d0
		beq.w	locret_FE14A
		add.w	($FFFFD5AE).w,d0
		bmi.w	locret_FE14A
		cmp.w	($FFFFD59E).w,d0
		bgt.w	locret_FE14A
		move.w	($FFFFD5AE).w,d1
		move.w	d0,($FFFFD5AE).w
		and.w	#$F,d0
		bne.w	loc_FDD46
		clr.w	($FFFFD5B0).w

loc_FDD46:							; CODE XREF: sub_FDD1A+24j
		and.w	#$F,d1
		bne.w	sub_FDD6A
		move.w	d0,-(sp)
		cmpnw	 $E,d0
		bne.w	loc_FDD5C
		bsr.w	sub_FDD92

loc_FDD5C:							; CODE XREF: sub_FDD1A+3Aj
		move.w	(sp)+,d0
		cmpnw	 2,d0
		bne.w	sub_FDD6A
		bsr.w	sub_FDD9C
; End of function sub_FDD1A


; =============== S U B	R O U T	I N E =======================================


sub_FDD6A:							; CODE XREF: sub_FDD1A+30j
					; sub_FDD1A+48j ...
		move.w	($FFFFBF78).w,-(sp)
		bset	#2,($FFFFBF78).w
		move.l	#$C00000,a0
		move.l	#$40020010,4(a0)
		move.w	#$FFB0,d0
		add.w	($FFFFD5AE).w,d0
		move.w	d0,(a0)
		move.w	(sp)+,($FFFFBF78).w
		rts
; End of function sub_FDD6A


; =============== S U B	R O U T	I N E =======================================


sub_FDD92:							; CODE XREF: sub_FDD1A+3Ep
		move.w	($FFFFD5AE).w,d3
		lsr.w	#4,d3
		bra.w	locret_FDDA8
; End of function sub_FDD92


; =============== S U B	R O U T	I N E =======================================


sub_FDD9C:							; CODE XREF: sub_FDD1A+4Cp
		move.w	($FFFFD5AE).w,d3
		lsr.w	#4,d3
		addq.w	#6,d3
		bra.w	*+4

locret_FDDA8:							; CODE XREF: sub_FDD92+6j
		rts
; End of function sub_FDD9C

; ---------------------------------------------------------------------------

loc_FDDAA:							; CODE XREF: ROM:000FDCD2p
					; ROM:000FDD14p
		movem.l	d0-a6,-(sp)
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	$F9
		dc.b	1
; ---------------------------------------------------------------------------
		tst.w	($FFFFD5AE).w
		beq.w	loc_FDDD4
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	$26
		dc.b	8
		dc.b	$F9
		dc.b	1
		dc.b	$7B
; ---------------------------------------------------------------------------
		bra.w	loc_FDDE4
; ---------------------------------------------------------------------------

loc_FDDD4:							; CODE XREF: ROM:000FDDBCj
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	$26
		dc.b	8
		dc.b	$F9
		dc.b	1
		dc.b	$20
; ---------------------------------------------------------------------------

loc_FDDE4:							; CODE XREF: ROM:000FDDD0j
		move.w	($FFFFD59E).w,d0
		cmp.w	($FFFFD5AE).w,d0
		beq.w	loc_FDE04
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	$26
		dc.b	$1A
		dc.b	$F9
		dc.b	1
		dc.b	$7D
; ---------------------------------------------------------------------------
		bra.w	loc_FDE14
; ---------------------------------------------------------------------------

loc_FDE04:							; CODE XREF: ROM:000FDDECj
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	$A
		dc.b	$F8
		dc.b	4
		dc.b	3
		dc.b	$26
		dc.b	$1A
		dc.b	$F9
		dc.b	1
		dc.b	$20
; ---------------------------------------------------------------------------

loc_FDE14:							; CODE XREF: ROM:000FDE00j
		jsr	sub_11A36
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'',0
; ---------------------------------------------------------------------------
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FDE24:							; CODE XREF: ROM:loc_FDCCEp
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	2
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		move.w	#$2C,d0					; ','
		jsr	sub_8078
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BD
		dc.b	$1A
		dc.b	7
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#0,d0
		jsr	sub_8078
		bsr.w	sub_FDD6A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		moveq	#$E,d6
		lea	word_FDFAA(pc),a1
		tst.w	($FFFFD056).w
		bne.w	loc_FDE74
		move.w	#$A,d6
		lea	word_FE098(pc),a1

loc_FDE74:							; CODE XREF: sub_FDE24+44j
					; sub_FDE24+8Aj
		move.w	(a1),d0
		lsr.w	#1,d0
		neg.w	d0
		add.w	#$15,d0
		move.w	d0,($FFFFB028).w
		jsr	sub_11BA4
		move.l	a1,a0
		move.w	#$22,($FFFFB028).w			; '"'
		move.w	#$C6CE,a2
		bsr.w	sub_FDEB4
		move.w	#9,($FFFFB028).w
		lea	$364(a2),a2
		bsr.w	sub_FDEB4
		lea	4(a0),a1
		addq.w	#2,($FFFFB02A).w
		dbf	d6,loc_FDE74
		rts
; End of function sub_FDE24


; =============== S U B	R O U T	I N E =======================================


sub_FDEB4:							; CODE XREF: sub_FDE24+70p
					; sub_FDE24+7Ep
		move.w	#$BFA4,a3
		move.w	#2,(a3)
		cmp.w	#$FFFF,(a0)
		bne.w	loc_FDEF8
		move.w	#$C,d0
		move.w	(a2,d0.w),d0
		mulu.w	#$64,d0					; 'd'
		move.w	#0,d1
		move.w	(a2,d1.w),d1
		beq.w	loc_FDEDE
		divu.w	d1,d0

loc_FDEDE:							; CODE XREF: sub_FDEB4+24j
		jsr	sub_11D06
		jsr	sub_11D9E
		jsr	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'%',0
; ---------------------------------------------------------------------------
		bra.w	loc_FDF8E
; ---------------------------------------------------------------------------

loc_FDEF8:							; CODE XREF: sub_FDEB4+Cj
		move.w	(a0),d0
		move.w	(a2,d0.w),d0
		cmp.w	#$352,(a0)
		bne.w	loc_FDF0E
		bsr.w	sub_FDFA4
		bra.w	loc_FDF26
; ---------------------------------------------------------------------------

loc_FDF0E:							; CODE XREF: sub_FDEB4+4Ej
		cmp.w	#$A,(a0)
		beq.w	loc_FDF1A
		bsr.w	sub_FDF9E

loc_FDF1A:							; CODE XREF: sub_FDEB4+5Ej
		cmp.w	#$A,(a0)
		bne.w	loc_FDF26
		bsr.w	sub_FDFA4

loc_FDF26:							; CODE XREF: sub_FDEB4+56j
					; sub_FDEB4+6Aj
		jsr	sub_11D9E
		move.w	2(a0),d0
		bmi.w	loc_FDF8E
		jsr	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'/',0
; ---------------------------------------------------------------------------
		move.w	(a2,d0.w),d0
		jsr	sub_11D06
		jsr	sub_11D9E
		cmp.w	#$14,(a0)
		bne.w	loc_FDF8E
		jsr	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	' ('
; ---------------------------------------------------------------------------
		move.w	(a0),d0
		move.w	(a2,d0.w),d0
		mulu.w	#$64,d0					; 'd'
		move.w	2(a0),d1
		move.w	(a2,d1.w),d1
		beq.w	loc_FDF78
		divu.w	d1,d0

loc_FDF78:							; CODE XREF: sub_FDEB4+BEj
		jsr	sub_11D06
		jsr	sub_11D9E
		jsr	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	'%)'
; ---------------------------------------------------------------------------

loc_FDF8E:							; CODE XREF: sub_FDEB4+40j
					; sub_FDEB4+7Cj ...
		move.w	a3,a1
		move.w	(a1),d0
		lsr.w	#1,d0
		sub.w	d0,($FFFFB028).w
		jmp	sub_11BA4
; End of function sub_FDEB4


; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_FDF9E:							; CODE XREF: sub_FDEB4+62p
		jmp	sub_11D06
; End of function sub_FDF9E


; =============== S U B	R O U T	I N E =======================================

; Attributes: thunk

sub_FDFA4:							; CODE XREF: sub_FDEB4+52p
					; sub_FDEB4+6Ep
		jmp	sub_11CA2
; End of function sub_FDFA4

; ---------------------------------------------------------------------------
word_FDFAA:	dc.w	8					; DATA XREF: sub_FDE24+3Co
aScore:		dc.b	'Score',0
		dc.w	$C
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	8
aShots_0:	dc.b	'Shots',0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.w	$E
aShootingPct:	dc.b	'Shooting Pct'
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.w	$C
aPowerPlay:	dc.b	'Power Play'
		dc.w	2
		dc.w	4
		dc.w	$C
aPpMinutes:	dc.b	'PP Minutes'
		dc.b	3
		dc.b	$52
		dc.b	$FF
		dc.b	$FF
		dc.w	$A
aPpShots:	dc.b	'PP Shots'
		dc.b	3
		dc.b	$54
		dc.b	$FF
		dc.b	$FF
		dc.w	$A
aShGoals:	dc.b	'SH Goals'
		dc.b	3
		dc.b	$56
		dc.b	$FF
		dc.b	$FF
		dc.w	$C
aBreakaways:	dc.b	'Breakaways'
		dc.b	3
		dc.b	$5A
		dc.b	3
		dc.b	$58
		dc.w	$C
aOneTimers:	dc.b	'One-Timers'
		dc.b	3
		dc.b	$5E
		dc.b	3
		dc.b	$5C
		dc.w	$10
aPenaltyShots:	dc.b	'Penalty Shots',0
		dc.b	3
		dc.b	$62
		dc.b	3
		dc.b	$60
		dc.w	$E
aFaceoffsWon:	dc.b	'Faceoffs Won'
		dc.b	0
		dc.b	$E
		dc.b	$FF
		dc.b	$FF
		dc.w	$E
aBodyChecks:	dc.b	'Body Checks',0
		dc.b	0
		dc.b	$10
		dc.b	$FF
		dc.b	$FF
		dc.w	$C
aPenalties_0:	dc.b	'Penalties',0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.w	$E
aAttackZone:	dc.b	'Attack Zone',0
		dc.b	0
		dc.b	$A
		dc.b	$FF
		dc.b	$FF
		dc.w	$A
aPassing_0:	dc.b	'Passing',0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$12
word_FE098:	dc.w	8					; DATA XREF: sub_FDE24+4Co
aScore_0:	dc.b	'Score',0
		dc.b	0
		dc.b	$C
		dc.b	$FF
		dc.b	$FF
		dc.w	8
aShots_1:	dc.b	'Shots',0
		dc.b	0
		dc.b	0
		dc.b	$FF
		dc.b	$FF
		dc.w	$E
aShootingPct_0:	dc.b	'Shooting Pct'
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.b	$FF
		dc.w	$C
aBreakaways_0:	dc.b	'Breakaways'
		dc.b	3
		dc.b	$5A
		dc.b	3
		dc.b	$58
		dc.w	$C
aOneTimers_0:	dc.b	'One-Timers'
		dc.b	3
		dc.b	$5E
		dc.b	3
		dc.b	$5C
		dc.w	$10
aPenaltyShots_0: dc.b	'Penalty Shots',0
		dc.b	3
		dc.b	$62
		dc.b	3
		dc.b	$60
		dc.w	$E
aFaceoffsWon_0:	dc.b	'Faceoffs Won'
		dc.b	0
		dc.b	$E
		dc.b	$FF
		dc.b	$FF
		dc.w	$E
aBodyChecks_0:	dc.b	'Body Checks',0
		dc.b	0
		dc.b	$10
		dc.b	$FF
		dc.b	$FF
		dc.w	$C
aPenalties_1:	dc.b	'Penalties',0
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	8
		dc.w	$E
aAttackZone_0:	dc.b	'Attack Zone',0
		dc.w	$A
		dc.b	$FF
		dc.b	$FF
		dc.w	$A
aPassing_1:	dc.b	'Passing',0
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$12
; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_FDD1A

locret_FE14A:							; CODE XREF: sub_FDD1A+4j sub_FDD1A+Cj ...
		rts
; END OF FUNCTION CHUNK	FOR sub_FDD1A

; =============== S U B	R O U T	I N E =======================================


sub_FE14C:							; CODE XREF: sub_128B2+28p
		btst	#5,($FFFFC2EE).w
		beq.w	locret_FE170
		move.l	#$FFFFC6CE,a2
		btst	#6,($FFFFC2EE).w
		beq.w	loc_FE16C
		move.l	#$FFFFCA32,a2

loc_FE16C:							; CODE XREF: sub_FE14C+16j
		addq.w	#1,$352(a2)

locret_FE170:							; CODE XREF: sub_FE14C+6j
		rts
; End of function sub_FE14C


; =============== S U B	R O U T	I N E =======================================


sub_FE172:							; CODE XREF: ROM:000FAC24p
					; sub_FCFB8+8p
		movem.l	d1-a6,-(sp)
		move.w	$28(a2),d0
		move.l	#unk_FE18E,a0
		clr.w	d1
		move.b	(a0,d0.w),d1
		move.w	d1,d0
		movem.l	(sp)+,d1-a6
		rts
; End of function sub_FE172

; ---------------------------------------------------------------------------
unk_FE18E:	dc.b	$33					; 3		; DATA XREF: sub_FE172+8o
		dc.b	$4C					; L
		dc.b	$49					; I
		dc.b	$4B					; K
		dc.b	$4E					; N
		dc.b	$43					; C
		dc.b	$4B					; K
		dc.b	$43					; C
		dc.b	$34					; 4
		dc.b	$42					; B
		dc.b	$4A					; J
		dc.b	$49					; I
		dc.b	$44					; D
		dc.b	$42					; B
		dc.b	$4A					; J
		dc.b	$37					; 7
		dc.b	$45					; E
		dc.b	$4B					; K
		dc.b	$47					; G
		dc.b	$38					; 8
		dc.b	$45					; E
		dc.b	$38					; 8
		dc.b	$48					; H
		dc.b	$47					; G
		dc.b	$48					; H
		dc.b	$46					; F
		dc.b	$5B					; [
		dc.b	$59					; Y

; =============== S U B	R O U T	I N E =======================================


sub_FE1AA:							; CODE XREF: sub_B0E8+6Ap
		movem.l	d0,-(sp)
		tst.w	($FFFFC32A).w
		beq.w	loc_FE1C2
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		bne.w	loc_FE1D2

loc_FE1C2:							; CODE XREF: sub_FE1AA+8j
		cmp.w	#2,($FFFFC31A).w
		ble.w	loc_FE1D2
		move.w	#2,($FFFFC31A).w

loc_FE1D2:							; CODE XREF: sub_FE1AA+14j
					; sub_FE1AA+1Ej
		movem.l	(sp)+,d0
		rts
; End of function sub_FE1AA


; =============== S U B	R O U T	I N E =======================================


sub_FE1D8:							; DATA XREF: ROM:000196F8o
					; ROM:00019884o ...
		movem.l	d0-a6,-(sp)
		tst.w	($FFFFD04A).w
		beq.w	loc_FE214
		move.w	($FFFFC328).w,d0
		cmp.w	($FFFFC32A).w,d0
		bne.w	loc_FE1FA
		eor.w	#1,($FFFFD05C).w
		bra.w	loc_FE20E
; ---------------------------------------------------------------------------

loc_FE1FA:							; CODE XREF: sub_FE1D8+14j
		btst	#1,($FFFFC2EC).w
		beq.w	loc_FE20E
		eor.w	#1,($FFFFD05C).w
		bra.w	loc_FE214
; ---------------------------------------------------------------------------

loc_FE20E:							; CODE XREF: sub_FE1D8+1Ej
					; sub_FE1D8+28j
		eor.w	#1,($FFFFD05A).w

loc_FE214:							; CODE XREF: sub_FE1D8+8j
					; sub_FE1D8+32j
		btst	#0,($FFFFC2FA).w
		bne.w	loc_FE228
		btst	#2,($FFFFC2F2).w
		beq.w	loc_FE2C2

loc_FE228:							; CODE XREF: sub_FE1D8+42j
		btst	#1,($FFFFC2EC).w
		bne.w	loc_FE28A
		cmp.w	#2,($FFFFD04A).w
		bne.w	loc_FE24A
		cmp.w	#5,($FFFFD406).w
		bgt.w	loc_FE2C2
		bra.w	loc_FE254
; ---------------------------------------------------------------------------

loc_FE24A:							; CODE XREF: sub_FE1D8+60j
		cmp.w	#5,($FFFFD406).w
		ble.w	loc_FE2C2

loc_FE254:							; CODE XREF: sub_FE1D8+6Ej
		move.w	#0,d0
		cmp.w	#2,($FFFFD04A).w
		bne.w	loc_FE266
		move.w	#6,d0

loc_FE266:							; CODE XREF: sub_FE1D8+86j
		tst.w	($FFFFD05A).w
		bne.w	loc_FE280
		move.w	#5,d0
		cmp.w	#2,($FFFFD04A).w
		bne.w	loc_FE280
		move.w	#$B,d0

loc_FE280:							; CODE XREF: sub_FE1D8+92j
					; sub_FE1D8+A0j
		jsr	sub_C0BC
		bra.w	loc_FE2C2
; ---------------------------------------------------------------------------

loc_FE28A:							; CODE XREF: sub_FE1D8+56j
		cmp.w	#2,($FFFFD04A).w
		bne.w	loc_FE2A2
		cmp.w	#5,($FFFFD406).w
		ble.w	loc_FE2C2
		bra.w	loc_FE2AC
; ---------------------------------------------------------------------------

loc_FE2A2:							; CODE XREF: sub_FE1D8+B8j
		cmp.w	#5,($FFFFD406).w
		bgt.w	loc_FE2C2

loc_FE2AC:							; CODE XREF: sub_FE1D8+C6j
		move.w	#6,d0
		tst.w	($FFFFD05C).w
		bne.w	loc_FE2BC
		move.w	#$B,d0

loc_FE2BC:							; CODE XREF: sub_FE1D8+DCj
		jsr	sub_C0DA

loc_FE2C2:							; CODE XREF: sub_FE1D8+4Cj
					; sub_FE1D8+6Aj ...
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE1D8


; =============== S U B	R O U T	I N E =======================================


sub_FE2C8:							; CODE XREF: sub_799E+Ap sub_111D0+1Ep
		tst.w	($FFFFD6B4).w
		bmi.w	locret_FE324
		movem.l	d0-a6,-(sp)
		tst.l	($FFFFD6BA).w
		bne.w	loc_FE310
		move.w	($FFFFD6B4).w,d0
		asl.w	#3,d0
		move.l	#off_FE364,a0
		move.l	(a0,d0.w),($FFFFD6B6).w
		move.l	4(a0,d0.w),($FFFFD6BA).w
		move.l	4(a0,d0.w),a2
		addq.w	#8,a2
		move.w	($FFFFD6AC).w,d4
		jsr	sub_11738
		clr.w	($FFFFD6B0).w
		bsr.w	sub_FE326
		bra.w	loc_FE320
; ---------------------------------------------------------------------------

loc_FE310:							; CODE XREF: sub_FE2C8+10j
		sub.w	d7,($FFFFD6B2).w
		bpl.w	loc_FE320
		addq.w	#1,($FFFFD6B0).w
		bsr.w	sub_FE326

loc_FE320:							; CODE XREF: sub_FE2C8+44j
					; sub_FE2C8+4Cj
		movem.l	(sp)+,d0-a6

locret_FE324:							; CODE XREF: sub_FE2C8+4j
		rts
; End of function sub_FE2C8


; =============== S U B	R O U T	I N E =======================================


sub_FE326:							; CODE XREF: sub_FE2C8+40p
					; sub_FE2C8+54p ...
		move.l	($FFFFD6B6).w,a0
		move.w	($FFFFD6B0).w,d0
		add.w	d0,d0
		move.b	(a0,d0.w),d1
		ext.w	d1
		move.w	d1,($FFFFD6AE).w
		move.b	1(a0,d0.w),d1
		ext.w	d1
		move.w	d1,($FFFFD6B2).w
		cmp.w	#$FFFF,($FFFFD6AE).w
		bne.w	loc_FE354
		clr.w	($FFFFD6B0).w
		bra.s	sub_FE326
; ---------------------------------------------------------------------------

loc_FE354:							; CODE XREF: sub_FE326+24j
		cmp.w	#$FFFE,($FFFFD6AE).w
		bne.w	locret_FE362
		bsr.w	sub_FE548

locret_FE362:							; CODE XREF: sub_FE326+34j
		rts
; End of function sub_FE326

; ---------------------------------------------------------------------------
off_FE364:	dc.l	unk_FE4DC				; DATA XREF: sub_FE2C8+1Ao
		dc.l	off_EA2EC
		dc.l	unk_FE490
		dc.l	off_EAB10
		dc.l	unk_FE4B6
		dc.l	off_EAB10
		dc.l	unk_FE47E
		dc.l	off_EB816
		dc.l	word_FE428
		dc.l	off_EC284
		dc.l	word_FE3F8
		dc.l	off_EC284
		dc.l	word_FE3E2
		dc.l	off_ED1B0
		dc.l	word_FE3A4
		dc.l	off_ED1B0
word_FE3A4:	dc.w	$10A					; DATA XREF: ROM:000FE39Co
		dc.w	$20A
		dc.w	$30A
		dc.w	$40A
		dc.w	$50A
		dc.w	$60A
		dc.w	$70A
		dc.w	$80A
		dc.w	$90A
		dc.w	$A0A
		dc.w	$10A
		dc.w	$20A
		dc.w	$30A
		dc.w	$40A
		dc.w	$50A
		dc.w	$60A
		dc.w	$70A
		dc.w	$80A
		dc.w	$90A
		dc.w	$A0A
		dc.w	$10A
		dc.w	$20A
		dc.w	$30A
		dc.w	$40A
		dc.w	$50A
		dc.w	$60A
		dc.w	$70A
		dc.w	$80A
		dc.w	$90A
		dc.w	$A0A
		dc.w	$FE00
word_FE3E2:	dc.w	$10A					; DATA XREF: ROM:000FE394o
		dc.w	$20A
		dc.w	$30A
		dc.w	$40A
		dc.w	$50A
		dc.w	$60A
		dc.w	$70A
		dc.w	$80A
		dc.w	$90A
		dc.w	$A0A
		dc.w	$FF00
word_FE3F8:	dc.w	$10A					; DATA XREF: ROM:000FE38Co
		dc.w	$20A
		dc.w	$30A
		dc.w	$40A
		dc.w	$50A
		dc.w	$60A
		dc.w	$70A
		dc.w	$80A
		dc.w	$90A
		dc.w	$A0A
		dc.w	$B0A
		dc.w	$C0A
		dc.w	$D0A
		dc.w	$E0A
		dc.w	$F0A
		dc.w	$100A
		dc.w	$110A
		dc.w	$120A
		dc.w	$130A
		dc.w	$140A
		dc.w	$150A
		dc.w	$160A
		dc.w	$170A
		dc.w	$FE00
word_FE428:	dc.w	$108					; DATA XREF: ROM:000FE384o
		dc.w	$208
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$33C
		dc.w	$43C
		dc.w	$208
		dc.w	$108
		dc.w	$FF00
unk_FE47E:	dc.b	1					; DATA XREF: ROM:000FE37Co
		dc.b	8
		dc.b	2
		dc.b	8
		dc.b	3
		dc.b	8
		dc.b	4
		dc.b	8
		dc.b	5
		dc.b	8
		dc.b	6
		dc.b	8
		dc.b	7
		dc.b	8
		dc.b	8
		dc.b	8
		dc.b	$FF
		dc.b	0
unk_FE490:	dc.b	1					; DATA XREF: ROM:000FE36Co
		dc.b	8
		dc.b	2
		dc.b	8
		dc.b	3
		dc.b	8
		dc.b	4
		dc.b	$1E
		dc.b	5
		dc.b	8
		dc.b	6
		dc.b	8
		dc.b	7
		dc.b	8
		dc.b	8
		dc.b	$1E
		dc.b	5
		dc.b	8
		dc.b	6
		dc.b	8
		dc.b	7
		dc.b	8
		dc.b	8
		dc.b	$1E
		dc.b	9
		dc.b	8
		dc.b	$A
		dc.b	8
		dc.b	$B
		dc.b	8
		dc.b	$C
		dc.b	8
		dc.b	$D
		dc.b	8
		dc.b	$E
		dc.b	8
		dc.b	$FE					; 
		dc.b	0
unk_FE4B6:	dc.b	$F					; DATA XREF: ROM:000FE374o
		dc.b	8
		dc.b	$10
		dc.b	8
		dc.b	$11
		dc.b	8
		dc.b	$12
		dc.b	$1E
		dc.b	$13
		dc.b	8
		dc.b	$14
		dc.b	8
		dc.b	$15
		dc.b	8
		dc.b	$16
		dc.b	$1E
		dc.b	$13
		dc.b	8
		dc.b	$14
		dc.b	8
		dc.b	$15
		dc.b	8
		dc.b	$16
		dc.b	$1E
		dc.b	$17
		dc.b	8
		dc.b	$18
		dc.b	8
		dc.b	$19
		dc.b	8
		dc.b	$1A
		dc.b	8
		dc.b	$1B
		dc.b	8
		dc.b	$1C
		dc.b	8
		dc.b	$FE					; 
		dc.b	0
unk_FE4DC:	dc.b	1					; DATA XREF: ROM:off_FE364o
		dc.b	$A
		dc.b	2
		dc.b	$A
		dc.b	3
		dc.b	$A
		dc.b	4
		dc.b	$A
		dc.b	5
		dc.b	$A
		dc.b	6
		dc.b	$A
		dc.b	7
		dc.b	$A
		dc.b	8
		dc.b	$A
		dc.b	9
		dc.b	$A
		dc.b	$A
		dc.b	$A
		dc.b	$B
		dc.b	$A
		dc.w	$C0A
		dc.w	$D0A
		dc.w	$E0A
		dc.w	$F7F
		dc.w	$FE00
; ---------------------------------------------------------------------------
		bset	#6,($FFFFC2EC).w
		move.w	#0,($FFFFBF8E).w
		move.w	#$160,($FFFFBF90).w
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FE510:							; CODE XREF: ROM:0000FB4Ep
					; sub_14062+274p ...
		movem.l	d0-a6,-(sp)
		move.w	d0,($FFFFD6B4).w
		move.l	#0,($FFFFD6B6).w
		move.l	#0,($FFFFD6BA).w
		move.w	#1,($FFFFD6AE).w
		clr.w	($FFFFD6B0).w
		st	($FFFFD6BE).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE510


; =============== S U B	R O U T	I N E =======================================


sub_FE53C:							; CODE XREF: sub_F3E4:loc_F460p
		bset	#0,($FFFFC2FC).w
		move.w	d0,($FFFFD6BE).w
		rts
; End of function sub_FE53C


; =============== S U B	R O U T	I N E =======================================


sub_FE548:							; CODE XREF: ROM:0000FCE4p
					; sub_FE326+38p
		move.w	#$FFFF,($FFFFD6B4).w
		bclr	#0,($FFFFC2FC).w
		rts
; End of function sub_FE548


; =============== S U B	R O U T	I N E =======================================


sub_FE556:							; CODE XREF: sub_7CB0-398p
					; sub_7A34+30p	...
		movem.l	d0-d1/a0-a1,-(sp)
		btst	#4,($FFFFC2EA).w
		beq.w	loc_FE56C
		move.w	#$FFFF,d1
		bra.w	loc_FE5A6
; ---------------------------------------------------------------------------

loc_FE56C:							; CODE XREF: sub_FE556+Aj
		bclr	#6,($FFFFC2FE).w
		move.l	#unk_FE5B0,a0
		move.l	#unk_FE658,a1
		move.w	($FFFFD6CA).w,d0
		mulu.w	#6,d0
		add.w	($FFFFD6CC).w,d0
		clr.w	d1
		move.b	(a0,d0.w),d1
		bne.w	loc_FE5A6
		move.w	#8,d0
		jsr	sub_11086
		and.w	#7,d0
		move.b	(a1,d0.w),d1

loc_FE5A6:							; CODE XREF: sub_FE556+12j
					; sub_FE556+3Aj
		move.w	d1,($FFFFD6C8).w
		movem.l	(sp)+,d0-d1/a0-a1
		rts
; End of function sub_FE556

; ---------------------------------------------------------------------------
unk_FE5B0:	include	"Unknown\FE5B0.asm"

unk_FE658:	dc.b	$34					; 4		; DATA XREF: sub_FE556+22o
		dc.b	$40					; @
		dc.b	$42					; B
		dc.b	$49					; I
		dc.b	$58					; X
		dc.b	$5C					; \
		dc.b	$65					; e
		dc.b	$66					; f

; =============== S U B	R O U T	I N E =======================================


sub_FE660:							; CODE XREF: ROM:00007718p sub_F739Ep
		movem.l	d0-d1/a0,-(sp)
		move.l	#$100,d1
		move.l	#$1EF6,d0
		move.l	#$FFFFD076,a0
		jsr	sub_1A244
		bclr	#4,($FFFFC2EC).w
		move.w	#$FFFF,($FFFFC016).w
		move.l	#$FFFF0000,($FFFFB036).w
		movem.l	(sp)+,d0-d1/a0
		rts
; End of function sub_FE660


; =============== S U B	R O U T	I N E =======================================


sub_FE696:							; CODE XREF: sub_8928:loc_897Ep
					; sub_180FC+48p ...
		movem.l	d0-d1/a0,-(sp)
		move.l	#$100,d1
		move.l	#$1EF6,d0
		move.l	#$FFFFD076,a0
		jsr	sub_1A1E4
		jsr	sub_1A206
		bclr	#4,($FFFFC2EC).w
		move.w	#$FFFF,($FFFFC016).w
		move.l	#$FFFF0000,($FFFFB036).w
		movem.l	(sp)+,d0-d1/a0
		rts
; End of function sub_FE696


; =============== S U B	R O U T	I N E =======================================


sub_FE6D2:							; CODE XREF: ROM:000FBD28p
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFCF36,a0
		move.w	#7,d7

loc_FE6E0:							; CODE XREF: sub_FE6D2+22j
		clr.b	8(a0)
		clr.b	9(a0)
		clr.b	$A(a0)
		clr.b	$B(a0)
		add.w	#$10,a0
		dbf	d7,loc_FE6E0
		move.l	#$D20,d0
		move.l	#$80,d1					; ''
		move.l	#$FFFFCF36,a0
		jsr	sub_1A1E4
		jsr	sub_1A206
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE6D2


; =============== S U B	R O U T	I N E =======================================


sub_FE71C:							; CODE XREF: sub_C566:loc_C63Cp
		movem.l	d0-a6,-(sp)
		move.w	($FFFFDA1A).w,d0
		btst	#7,$62(a3)
		bne.w	loc_FE73A
		move.l	#unk_FE744,a0
		add.w	d0,d0
		move.w	(a0,d0.w),d0

loc_FE73A:							; CODE XREF: sub_FE71C+Ej
		move.w	d0,($FFFFBEDC).w
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE71C

; ---------------------------------------------------------------------------
unk_FE744:	dc.b	0					; DATA XREF: sub_FE71C+12o
		dc.b	0
		dc.b	0
		dc.b	7
		dc.b	0
		dc.b	6
		dc.b	0
		dc.b	5
		dc.b	0
		dc.b	4
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	8

; =============== S U B	R O U T	I N E =======================================


sub_FE756:							; CODE XREF: ROM:loc_ECECp
					; sub_FC4C0+14p
		movem.l	d0-a6,-(sp)

loc_FE75A:							; CODE XREF: sub_FE756+12j
		move.w	#7,d0
		jsr	sub_11086
		cmpnw	 6,d0
		bgt.s	loc_FE75A
		move.w	d0,($FFFFDA12).w
		bclr	#3,($FFFFC2FC).w
		move.w	#$FFFF,($FFFFDA14).w
		clr.w	($FFFFDA16).w
		bsr.w	sub_FE7FC
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE756

; ---------------------------------------------------------------------------
off_FE788:	dc.l	unk_FE7A4				; DATA XREF: sub_FE7FC+18o
		dc.l	unk_FE7B0
		dc.l	unk_FE7C0
		dc.l	unk_FE7CC
		dc.l	unk_FE7D8
		dc.l	unk_FE7E4
		dc.l	unk_FE7F0
unk_FE7A4:	dc.b	0					; DATA XREF: ROM:off_FE788o
		dc.b	$10
		dc.b	0
		dc.b	$E2					; 
		dc.b	0
		dc.b	$10
		dc.b	0
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	0
		dc.b	5
unk_FE7B0:	dc.b	$FF					; DATA XREF: ROM:000FE78Co
		dc.b	$C9					; 
		dc.b	0
		dc.b	$A0					; 
		dc.b	$FF
		dc.b	$FF
		dc.b	0
		dc.b	$C8					; 
		dc.b	$FF
		dc.b	$E0					; 
		dc.b	0
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	0
		dc.b	5
unk_FE7C0:	dc.b	$FF					; DATA XREF: ROM:000FE790o
		dc.b	$B0					; 
		dc.b	0
		dc.b	$40					; @
		dc.b	0
		dc.b	$1A
		dc.b	0
		dc.b	$BC					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	0
		dc.b	5
unk_FE7CC:	dc.b	$FF					; DATA XREF: ROM:000FE794o
		dc.b	$CE					; 
		dc.b	0
		dc.b	$58					; X
		dc.b	0
		dc.b	$14
		dc.b	0
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$2C					; ,
		dc.b	0
		dc.b	5
unk_FE7D8:	dc.b	$FF					; DATA XREF: ROM:000FE798o
		dc.b	$CE					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$20
		dc.b	0
		dc.b	$D0					; 
		dc.b	$80					; 
		dc.b	$28					; (
		dc.b	0
		dc.b	6
unk_FE7E4:	dc.b	0					; DATA XREF: ROM:000FE79Co
		dc.b	$1C
		dc.b	0
		dc.b	$F4					; 
		dc.b	0
		dc.b	8
		dc.b	0
		dc.b	$E0					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	0
		dc.b	6
unk_FE7F0:	dc.b	$FF					; DATA XREF: ROM:000FE7A0o
		dc.b	$E6					; 
		dc.b	0
		dc.b	$F8					; 
		dc.b	0
		dc.b	0
		dc.b	0
		dc.b	$E0					; 
		dc.b	$80					; 
		dc.b	$20
		dc.b	0
		dc.b	2

; =============== S U B	R O U T	I N E =======================================


sub_FE7FC:							; CODE XREF: sub_FE756+28p
					; sub_FE864:loc_FE8DAp
		movem.l	d0-a6,-(sp)
		cmp.b	#$80,($FFFFDA16).w
		beq.w	loc_FE85E
		addq.w	#1,($FFFFDA14).w
		move.w	($FFFFDA14).w,d0
		asl.w	#2,d0
		move.l	#off_FE788,a0
		move.w	($FFFFDA12).w,d1
		asl.w	#2,d1
		move.l	(a0,d1.w),a0
		move.w	(a0,d0.w),($FFFFDA16).w
		btst	#0,$76(a3)
		beq.w	loc_FE838
		neg.w	($FFFFDA16).w

loc_FE838:							; CODE XREF: sub_FE7FC+34j
		move.w	2(a0,d0.w),($FFFFDA18).w
		cmp.b	#$80,4(a0,d0.w)
		bne.w	loc_FE85E
		bset	#3,($FFFFC2FC).w
		clr.w	($FFFFDA1C).w
		move.b	5(a0,d0.w),($FFFFDA1D).w
		move.w	6(a0,d0.w),($FFFFDA1A).w

loc_FE85E:							; CODE XREF: sub_FE7FC+Aj
					; sub_FE7FC+48j
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE7FC


; =============== S U B	R O U T	I N E =======================================


sub_FE864:							; CODE XREF: sub_DEEE:loc_E0EEp
		movem.l	d2-a6,-(sp)
		cmp.b	#$80,($FFFFDA16).w
		beq.w	loc_FE8E6
		move.w	($FFFFDA16).w,d0
		move.w	($FFFFDA18).w,d1
		btst	#7,$62(a3)
		bne.w	loc_FE888
		neg.w	d0
		neg.w	d1

loc_FE888:							; CODE XREF: sub_FE864+1Cj
		sub.w	(a3),d0
		bpl.w	loc_FE890
		neg.w	d0

loc_FE890:							; CODE XREF: sub_FE864+26j
		tst.w	$28(a3)
		bne.w	loc_FE8A8
		tst.w	$2A(a3)
		bne.w	loc_FE8A8
		cmpnw	 $12,d0
		ble.w	loc_FE8B0

loc_FE8A8:							; CODE XREF: sub_FE864+30j
					; sub_FE864+38j
		cmpnw	 $A,d0
		bgt.w	loc_FE8DE

loc_FE8B0:							; CODE XREF: sub_FE864+40j
		sub.w	$14(a3),d1
		bpl.w	loc_FE8BA
		neg.w	d1

loc_FE8BA:							; CODE XREF: sub_FE864+50j
		tst.w	$28(a3)
		bne.w	loc_FE8D2
		tst.w	$2A(a3)
		bne.w	loc_FE8D2
		cmpnw	 $12,d1
		ble.w	loc_FE8DA

loc_FE8D2:							; CODE XREF: sub_FE864+5Aj
					; sub_FE864+62j
		cmpnw	 $A,d1
		bgt.w	loc_FE8DE

loc_FE8DA:							; CODE XREF: sub_FE864+6Aj
		bsr.w	sub_FE7FC

loc_FE8DE:							; CODE XREF: sub_FE864+48j
					; sub_FE864+72j
		move.w	($FFFFDA16).w,d0
		move.w	($FFFFDA18).w,d1

loc_FE8E6:							; CODE XREF: sub_FE864+Aj
		movem.l	(sp)+,d2-a6
		rts
; End of function sub_FE864


; =============== S U B	R O U T	I N E =======================================


sub_FE8EC:							; CODE XREF: sub_DEEE:loc_E0A4p
		movem.l	d0-a6,-(sp)
		btst	#1,($FFFFC2FA).w
		beq.w	loc_FE980
		move.w	$52(a3),d0
		cmp.w	($FFFFB7AA).w,d0
		bne.w	loc_FE980
		cmp.w	#3,($FFFFD454).w
		ble.w	loc_FE97A
		cmp.b	#$80,($FFFFDA16).w
		beq.w	loc_FE97A
		btst	#3,($FFFFC2FC).w
		bne.w	loc_FE942
		tst.w	($FFFFB772).w
		bne.w	loc_FE980
		tst.w	($FFFFB774).w
		bne.w	loc_FE980
		cmp.w	#$F,($FFFFD454).w
		blt.w	loc_FE97A
		bra.w	loc_FE980
; ---------------------------------------------------------------------------

loc_FE942:							; CODE XREF: sub_FE8EC+34j
		move.w	($FFFFDA16).w,d0
		move.w	($FFFFDA18).w,d1
		btst	#7,$62(a3)
		bne.w	loc_FE958
		neg.w	d0
		neg.w	d1

loc_FE958:							; CODE XREF: sub_FE8EC+64j
		sub.w	(a3),d0
		bpl.w	loc_FE960
		neg.w	d0

loc_FE960:							; CODE XREF: sub_FE8EC+6Ej
		sub.w	$14(a3),d1
		bpl.w	loc_FE96A
		neg.w	d1

loc_FE96A:							; CODE XREF: sub_FE8EC+78j
		cmp.w	($FFFFDA1C).w,d0
		bgt.w	loc_FE980
		cmp.w	($FFFFDA1C).w,d1
		bgt.w	loc_FE980

loc_FE97A:							; CODE XREF: sub_FE8EC+20j
					; sub_FE8EC+2Aj ...
		clr.w	d0
		bra.w	loc_FE984
; ---------------------------------------------------------------------------

loc_FE980:							; CODE XREF: sub_FE8EC+Aj
					; sub_FE8EC+16j ...
		move.w	#1,d0

loc_FE984:							; CODE XREF: sub_FE8EC+90j
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE8EC


; =============== S U B	R O U T	I N E =======================================


sub_FE98A:							; CODE XREF: ROM:loc_F8982p
					; ROM:loc_FAF0Ep ...
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFDA1E,a0
		move.l	#$FFFFDEA0,a1
		move.w	(a2)+,d0
		move.w	d0,(a0)+
		asl.w	#3,d0
		subq.w	#1,d0

loc_FE9A2:							; CODE XREF: sub_FE98A+BEj
		clr.l	(a0)
		clr.l	(a1)
		move.b	(a2)+,(a1)
		move.b	(a2)+,1(a1)
		move.b	(a2)+,2(a1)
		move.l	#0,-(sp)
		move.l	(a1),d2
		and.l	#$E0000000,d2
		lsr.l	#1,d2
		add.l	#$50000000,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#$1C000000,d2
		lsr.l	#2,d2
		add.l	#$5000000,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#$3800000,d2
		lsr.l	#3,d2
		add.l	#$500000,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#$700000,d2
		lsr.l	#4,d2
		add.l	#unk_50000,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#unk_E0000,d2
		lsr.l	#5,d2
		add.l	#$5000,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#unk_1C000,d2
		lsr.l	#6,d2
		add.l	#$500,d2
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#$3800,d2
		lsr.l	#7,d2
		add.l	#$50,d2					; 'P'
		or.l	d2,(sp)
		move.l	(a1),d2
		and.l	#$700,d2
		move.w	#8,d5
		lsr.l	d5,d2
		addq.l	#5,d2
		or.l	d2,(sp)
		move.l	(sp)+,(a0)+
		dbf	d0,loc_FE9A2
		movem.l	(sp)+,d0-a6
		rts
; End of function sub_FE98A


; =============== S U B	R O U T	I N E =======================================


sub_FEA52:							; CODE XREF: ROM:00009D36p
					; sub_12B40+20p ...
		move.w	d0,-(sp)
		sub.w	#$18,d4
		move.l	#off_FEA74,a2
		move.w	($FFFFC330).w,d0
		asl.w	#2,d0
		move.l	(a2,d0.w),a2
		addq.w	#8,a2
		jsr	sub_11738
		move.w	(sp)+,d0
		rts
; End of function sub_FEA52

; ---------------------------------------------------------------------------
off_FEA74:	dc.l	unk_EDE8A				; DATA XREF: sub_FEA52+6o
		dc.l	unk_F2A84
		dc.l	unk_EE194
		dc.l	unk_EE49E
		dc.l	unk_EE7A8
		dc.l	unk_EEAB2
		dc.l	unk_EEDBC
		dc.l	unk_EF0C6
		dc.l	unk_EF3D0
		dc.l	unk_EF6DA
		dc.l	unk_EF9E4
		dc.l	unk_EFCEE
		dc.l	unk_EFFF8
		dc.l	unk_F0302
		dc.l	unk_F060C
		dc.l	unk_F0916
		dc.l	unk_F0C20
		dc.l	unk_F0F2A
		dc.l	unk_F1234
		dc.l	unk_F153E
		dc.l	unk_F1848
		dc.l	unk_F1B52
		dc.l	unk_F1E5C
		dc.l	unk_F2166
		dc.l	unk_F2470
		dc.l	unk_F277A
		dc.l	unk_F2D8E
		dc.l	unk_F2D8E

; =============== S U B	R O U T	I N E =======================================


sub_FEAE4:							; CODE XREF: ROM:000189DCp
					; ROM:00018A02p
		movem.l	d0/a2,-(sp)
		jsr	sub_11B92
; End of function sub_FEAE4

; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	' ('
; ---------------------------------------------------------------------------
		add.w	#$CE,a2					; ''
		bra.w	loc_FEB0C

; =============== S U B	R O U T	I N E =======================================


sub_FEAFA:							; CODE XREF: ROM:000189A0p
		movem.l	d0/a2,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	' ('
; ---------------------------------------------------------------------------
		add.w	#$B4,a2					; ''

loc_FEB0C:							; CODE XREF: ROM:000FEAF6j
		move.b	(a2,d0.w),d0
		ext.w	d0
		move.w	#1,d1
		cmpnw	 9,d0
		ble.w	loc_FEB2E
		move.w	#2,d1
		cmpnw	 $63,d0					; 'c'
		ble.w	loc_FEB2E
		move.w	#3,d1

loc_FEB2E:							; CODE XREF: sub_FEAFA+20j
					; sub_FEAFA+2Cj
		jsr	sub_11D3A
		jsr	sub_11BA4
		jsr	sub_11B92
; End of function sub_FEAFA

; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	')',0
; ---------------------------------------------------------------------------
		addq.w	#1,($FFFFB02A).w
		move.w	#$E,($FFFFB028).w
		movem.l	(sp)+,d0/a2
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FEB54:							; CODE XREF: sub_A9D6+CAp
		cmp.w	#$193E,$58(a3)
		beq.w	loc_FEC44
		cmp.w	#$1A00,$58(a3)
		beq.w	loc_FEC0C
		cmp.w	#$18CC,$58(a3)
		beq.w	loc_FEC28
		cmp.w	#$17E8,$58(a3)
		beq.w	loc_FEBF0
		cmp.w	#$1776,$58(a3)
		beq.w	loc_FEB94
		cmp.w	#$185A,$58(a3)
		beq.w	loc_FEBC2
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEB94:							; CODE XREF: sub_FEB54+2Ej
		move.w	#$124,d0
		cmp.w	#$56,($FFFFD41C).w			; 'V'
		bgt.w	loc_FEBAC
		cmp.w	#$FFAA,($FFFFD41C).w
		bgt.w	loc_FEBB0

loc_FEBAC:							; CODE XREF: sub_FEB54+4Aj
		move.w	#$116,d0

loc_FEBB0:							; CODE XREF: sub_FEB54+54j
		sub.w	$14(a3),d0
		bmi.w	locret_FEC5C
		asr.w	#1,d0
		add.w	d0,$14(a3)
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEBC2:							; CODE XREF: sub_FEB54+38j
		move.w	#$FEDC,d0
		cmp.w	#$56,($FFFFD41C).w			; 'V'
		bgt.w	loc_FEBDA
		cmp.w	#$FFAA,($FFFFD41C).w
		bgt.w	loc_FEBDE

loc_FEBDA:							; CODE XREF: sub_FEB54+78j
		move.w	#$FEEA,d0

loc_FEBDE:							; CODE XREF: sub_FEB54+82j
		sub.w	$14(a3),d0
		bpl.w	locret_FEC5C
		asr.w	#1,d0
		add.w	d0,$14(a3)
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEBF0:							; CODE XREF: sub_FEB54+24j
		move.w	#$82,d0					; ''
		btst	#3,4(a3)
		beq.w	loc_FEC02
		move.w	#$FF7E,d0

loc_FEC02:							; CODE XREF: sub_FEB54+A6j
		sub.w	(a3),d0
		asr.w	#1,d0
		add.w	d0,(a3)
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEC0C:							; CODE XREF: sub_FEB54+10j
		move.w	#$88,d0					; ''
		btst	#3,4(a3)
		beq.w	loc_FEC1E
		move.w	#$FF78,d0

loc_FEC1E:							; CODE XREF: sub_FEB54+C2j
		sub.w	(a3),d0
		asr.w	#1,d0
		add.w	d0,(a3)
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEC28:							; CODE XREF: sub_FEB54+1Aj
		move.w	#$FF7E,d0
		btst	#3,4(a3)
		beq.w	loc_FEC3A
		move.w	#$82,d0					; ''

loc_FEC3A:							; CODE XREF: sub_FEB54+DEj
		sub.w	(a3),d0
		asr.w	#1,d0
		add.w	d0,(a3)
		bra.w	locret_FEC5C
; ---------------------------------------------------------------------------

loc_FEC44:							; CODE XREF: sub_FEB54+6j
		move.w	#$FF78,d0
		btst	#3,4(a3)
		beq.w	loc_FEC56
		move.w	#$88,d0					; ''

loc_FEC56:							; CODE XREF: sub_FEB54+FAj
		sub.w	(a3),d0
		asr.w	#1,d0
		add.w	d0,(a3)

locret_FEC5C:							; CODE XREF: sub_FEB54+3Cj
					; sub_FEB54+60j ...
		rts
; End of function sub_FEB54


; =============== S U B	R O U T	I N E =======================================


sub_FEC5E:							; CODE XREF: ROM:00009B50p
		movem.l	d1-a6,-(sp)
		move.w	#$1B,d1
		ext.l	d1
		clr.w	d7
		move.l	#$FFFFCF36,a0

loc_FEC70:							; CODE XREF: sub_FEC5E:loc_FEC8Ej
		jsr	sub_F9BE2
		clr.w	d2
		move.b	8(a0),d2
		bne.w	loc_FEC84
		move.w	#$50,d2					; 'P'

loc_FEC84:							; CODE XREF: sub_FEC5E+1Ej
		cmp.w	d7,d2
		ble.w	loc_FEC8E
		move.w	d2,d7
		move.w	d1,d0

loc_FEC8E:							; CODE XREF: sub_FEC5E+28j
		dbf	d1,loc_FEC70
		movem.l	(sp)+,d1-a6
		rts
; End of function sub_FEC5E


; =============== S U B	R O U T	I N E =======================================


sub_FEC98:							; CODE XREF: sub_144AC+52p
		clr.w	d0
		move.b	$74(a2),d0
		lsr.w	#2,d0
		rts
; End of function sub_FEC98


; =============== S U B	R O U T	I N E =======================================


sub_FECA2:							; CODE XREF: sub_144AC+84p
		btst	#1,$74(a2)
		rts
; End of function sub_FECA2


; =============== S U B	R O U T	I N E =======================================


sub_FECAA:							; CODE XREF: sub_F3E4:loc_F454p
		movem.l	a0,-(sp)
		move.l	#unk_FECE6,a0
		move.w	#$64,d0					; 'd'
		jsr	sub_11086
		and.w	#7,d0
		add.w	d0,d0
		move.w	(a0,d0.w),d0
		cmp.w	#$FFC0,($FFFFBF94).w
		bgt.w	loc_FECE0
		cmp.w	#$C0,($FFFFBF96).w			; ''
		bgt.w	loc_FECE0
		move.w	#6,d0

loc_FECE0:							; CODE XREF: sub_FECAA+24j
					; sub_FECAA+2Ej
		movem.l	(sp)+,a0
		rts
; End of function sub_FECAA

; ---------------------------------------------------------------------------
unk_FECE6:	dc.b	0					; DATA XREF: sub_FECAA+4o
		dc.b	3
		dc.b	0
		dc.b	5
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	1
		dc.b	0
		dc.b	5
		dc.b	0
		dc.b	2
		dc.b	0
		dc.b	3
		dc.b	0
		dc.b	3
		dc.b	$FF
		dc.b	$FF

; =============== S U B	R O U T	I N E =======================================


sub_FECF8:							; CODE XREF: sub_F3E4+5Ap
		movem.l	d0-d2/a0-a2,-(sp)
		move.l	#$FFFFC6CE,a0
		move.l	#$FFFFCA32,a1
		move.w	$24(a0),d0
		sub.w	$24(a1),d0
		beq.w	loc_FED24
		bpl.w	loc_FED1A
		exg	a0,a1

loc_FED1A:							; CODE XREF: sub_FECF8+1Cj
		cmp.w	#$13,$28(a0)
		bne.w	*+4

loc_FED24:							; CODE XREF: sub_FECF8+18j
		movem.l	(sp)+,d0-d2/a0-a2
		rts
; End of function sub_FECF8


; =============== S U B	R O U T	I N E =======================================


sub_FED2A:							; CODE XREF: ROM:00017536p
		move.w	#$C018,a6
		moveq	#1,d6
		move.l	#unk_F3098,a0
		move.w	($FFFFB016).w,d3
		or.w	#$8000,d3
		move.w	($FFFFB8AE).w,d0
		move.w	($FFFFB8B0).w,d1
		move.w	#1,d2
		jsr	sub_16178
		cmp.w	#$C018,a6
		bne.w	loc_FED5C
		clr.l	(a6)+
		clr.l	(a6)+

loc_FED5C:							; CODE XREF: sub_FED2A+2Aj
		clr.b	-5(a6)
		move.l	a6,d0
		sub.l	#$FFFFC018,d0
		lsr.w	#1,d0
		move.w	d0,($FFFFC2E8).w
		rts
; End of function sub_FED2A


; =============== S U B	R O U T	I N E =======================================


sub_FED70:							; CODE XREF: ROM:0000771Ep
		move.l	#sub_15E4C,($FFFFB03A).w
		bclr	#1,($FFFFBF78).w
		move.w	#5,($FFFFB00E).w
		move.w	#$A000,($FFFFB008).w
		move.w	#7,($FFFFB00A).w
		move.w	#$C000,($FFFFB004).w
		move.w	#7,($FFFFB006).w
		move.w	#$F000,($FFFFB00C).w
		move.w	#$F800,($FFFFB002).w
		move.w	#$FC00,($FFFFB000).w
		jsr	sub_10F32
		move.w	#$BD28,a0
		moveq	#$1F,d1

loc_FEDBA:							; CODE XREF: sub_FED70+4Cj
		clr.l	(a0)+
		dbf	d1,loc_FEDBA
		jsr	VDP_InitialSetup
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$E
		dc.b	3
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#unk_F4378,a2
		move.l	a2,a0
		move.l	a2,a1
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		clr.w	d4
		moveq	#$F,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$10
		dc.b	$10
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#unk_F5338,a2
		move.l	a2,a0
		move.l	a2,a1
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#0,d5
		jsr	sub_1169A
		move.w	#$18,($FFFFBD26).w
		move	#$2500,sr
		move.w	#$50,($FFFFD066).w			; 'P'

loc_FEE30:							; CODE XREF: sub_FED70+D2j
		moveq	#4,d0
		jsr	sub_11176
		tst.w	d1
		bne.w	loc_FEE44
		subq.w	#1,($FFFFD066).w
		bpl.s	loc_FEE30

loc_FEE44:							; CODE XREF: sub_FED70+CAj
		move	#$2700,sr
		rts
; End of function sub_FED70


; =============== S U B	R O U T	I N E =======================================


sub_FEE4A:							; CODE XREF: ROM:000FA93Ep
					; ROM:000FA94Ep
		move.l	#$FFFFC6CE,a2
		cmp.w	$28(a2),d2
		beq.w	locret_FEE5E
		move.l	#$FFFFCA32,a2

locret_FEE5E:							; CODE XREF: sub_FEE4A+Aj
		rts
; End of function sub_FEE4A


; =============== S U B	R O U T	I N E =======================================


sub_FEE60:							; CODE XREF: sub_B0E8+678p
					; sub_B0E8+728p
		movem.l	d1-a0,-(sp)
		btst	#0,($FFFFC2EA).w
		bne.w	loc_FEE9A
		move.w	($FFFFBF12).w,d1
		move.l	#$FFFFDEA4,a0
		tst.w	d4
		beq.w	loc_FEE84
		move.l	#$FFFFDEAC,a0

loc_FEE84:							; CODE XREF: sub_FEE60+1Aj
		addq.l	#1,(a0)
		cmpnw	 8,d1
		beq.w	loc_FEE9A
		cmp.w	$54(a3),d1
		beq.w	loc_FEE9A
		addq.l	#1,4(a0)

loc_FEE9A:							; CODE XREF: sub_FEE60+Aj
					; sub_FEE60+2Aj ...
		movem.l	(sp)+,d1-a0
		rts
; End of function sub_FEE60

; ---------------------------------------------------------------------------
		movem.l	d0-a6,-(sp)
		move.l	#$FFFFDEBA,a1
		bsr.w	sub_FB992
		move.w	($FFFFD4E8).w,d0
		move.b	#0,(a1,d0.w)
		addq.w	#1,d0
		and.w	#$FFFE,d0
		move.w	d0,-2(a1)
		movem.l	(sp)+,d0-a6
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FEEC8:							; CODE XREF: ROM:loc_178CAp
		movem.l	d0-a0/a2-a6,-(sp)
		move.w	($FFFFD042).w,d0

loc_FEED0:							; CODE XREF: sub_FEF5A+8j
		tst.w	d0
		beq.w	loc_FEF52
		move.l	#$FFFFD4DC,a1
		move.w	d0,($FFFFD4EA).w
		bsr.w	sub_FB992
		add.w	($FFFFD4E8).w,a1
		move.b	#0,(a1)
		addq.w	#1,($FFFFD4E8).w
		and.w	#$FFFE,($FFFFD4E8).w
		addq.w	#2,($FFFFD4E8).w
		move.w	($FFFFD4E8).w,($FFFFD4DA).w
		move.l	#$FFFFD4DA,a1
		move.l	a1,-(sp)
		move.l	#$FFFFBF20,a3
		move.l	#word_FEF42,a1
		bsr.w	sub_F997A
		move.l	(sp)+,a1
		move.l	#$FFFFBF20,a3
		jsr	sub_11D9E
		move.l	#$FFFFBF20,a3
		jsr	sub_11D96
; ---------------------------------------------------------------------------
		dc.w	4
		dc.b	')',0
; ---------------------------------------------------------------------------
		move.l	#$FFFFBF20,a1

loc_FEF3C:							; CODE XREF: sub_FEEC8+90j
		movem.l	(sp)+,d0-a0/a2-a6
		rts
; ---------------------------------------------------------------------------
word_FEF42:	dc.w	$E					; DATA XREF: sub_FEEC8+46o
aPlayedBy:	dc.b	'(played by ',0
word_FEF50:	dc.w	2					; DATA XREF: sub_FEEC8:loc_FEF52o
; ---------------------------------------------------------------------------

loc_FEF52:							; CODE XREF: sub_FEEC8+Aj
		move.l	#word_FEF50,a1
		bra.s	loc_FEF3C
; End of function sub_FEEC8


; =============== S U B	R O U T	I N E =======================================


sub_FEF5A:							; CODE XREF: ROM:loc_178D4p
		movem.l	d0-a0/a2-a6,-(sp)
		move.w	($FFFFD044).w,d0
		bra.w	loc_FEED0
; End of function sub_FEF5A


; =============== S U B	R O U T	I N E =======================================


sub_FEF66:							; CODE XREF: sub_A9D6:loc_AA3Ap
		bra.w	loc_FEF74
; ---------------------------------------------------------------------------
		bclr	#1,($FFFFC2FE).w
		bra.w	locret_FEF7A
; ---------------------------------------------------------------------------

loc_FEF74:							; CODE XREF: sub_FEF66j
		bset	#1,($FFFFC2FE).w

locret_FEF7A:							; CODE XREF: sub_FEF66+Aj
		rts
; End of function sub_FEF66


; =============== S U B	R O U T	I N E =======================================


sub_FEF7C:							; CODE XREF: ROM:00008E46p
					; ROM:000FA916p ...
		cmpnw	 $32,d0					; '2'
		bge.w	locret_FEF8A
		asr.w	#1,d0
		add.w	#$19,d0

locret_FEF8A:							; CODE XREF: sub_FEF7C+4j
		rts
; End of function sub_FEF7C

; ---------------------------------------------------------------------------
		movem.w	d0-d1,-(sp)
		move.w	($FFFFB74A).w,d0
		sub.w	(a3),d0
		cmpnw	 $1E,d0
		bgt.w	loc_FEFC6
		move.w	($FFFFB772).w,d1
		eor.w	d1,d0
		and.w	#$8000,d0
		beq.w	loc_FEFC6
		move.w	($FFFFB75E).w,d0
		sub.w	$14(a3),d0
		cmpnw	 $1E,d0
		bgt.w	loc_FEFC6
		move.w	($FFFFB774).w,d1
		eor.w	d1,d0
		and.w	#$8000,d0

loc_FEFC6:							; CODE XREF: ROM:000FEF9Aj
					; ROM:000FEFA8j ...
		movem.w	(sp)+,d0-d1
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FEFCC:							; CODE XREF: sub_C2F2+12Ap
					; ROM:0000CD40p ...
		movem.l	a1,-(sp)
		move.l	#$FFFFC6CE,a1
		btst	#6,$62(a3)
		beq.w	loc_FEFE6
		move.l	#$FFFFCA32,a1

loc_FEFE6:							; CODE XREF: sub_FEFCC+10j
		tst.w	$26(a1)
		movem.l	(sp)+,a1
		rts
; End of function sub_FEFCC


; =============== S U B	R O U T	I N E =======================================


sub_FEFF0:							; CODE XREF: sub_14A54+2Ap
					; sub_F693E:loc_F6C02p
		movem.l	d0/a0,-(sp)
		bclr	#3,$64(a3)
		bclr	#5,$62(a3)
		bclr	#1,$63(a3)
		clr.w	($FFFFBF76).w
		st	($FFFFBF6C).w
		move.w	d1,-(sp)
		move.w	#$50C,d1
		jsr	sub_1073A
		tst.w	$34(a3)
		bpl.w	loc_FF02C
		jsr	sub_10646
		bra.w	loc_FF032
; ---------------------------------------------------------------------------

loc_FF02C:							; CODE XREF: sub_FEFF0+2Ej
		jsr	sub_15A88

loc_FF032:							; CODE XREF: sub_FEFF0+38j
		clr.w	$5A(a3)
		st	$5C(a3)
		move.w	(sp)+,d1
		movem.l	(sp)+,d0/a0
		rts
; End of function sub_FEFF0

; ---------------------------------------------------------------------------

loc_FF042:							; CODE XREF: ROM:000172F4p
		move	#$2700,sr
		move.w	($C00008).l,($FFFFD066).w
		move.w	($C00008).l,($FFFFD068).w
		move.l	#loc_FF348,($FFFFB03A).l
		bset	#1,($FFFFBF78).w
		move.w	#5,($FFFFB00E).w
		move.w	#$A000,($FFFFB008).w
		move.w	#7,($FFFFB00A).w
		move.w	#$C000,($FFFFB004).w
		move.w	#7,($FFFFB006).w
		move.w	#$F000,($FFFFB00C).w
		move.w	#$F800,($FFFFB002).w
		move.w	#$FC00,($FFFFB000).w
		move.w	#0,d0
		jsr	sub_11594
		move.l	#$C00000,a0
		move.w	#$9100,4(a0)
		move.w	#$9217,4(a0)
		move.w	#$8B03,4(a0)
		clr.w	($FFFFBD20).w
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$80,d0					; ''
		moveq	#$20,d1					; ' '
		move.l	#$7FF,d2
		jsr	sub_1197E
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$80,d0					; ''
		moveq	#$20,d1					; ' '
		move.l	#$7FF,d2
		jsr	sub_1197E
		clr.w	d4
		move.w	d4,-(sp)
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FD
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_4E45C,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	#$17,d3
		moveq	#$F,d5
		jsr	sub_1169A
		move.w	d4,($FFFFBF12).w
		move.w	(sp)+,d4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$FE
		dc.b	0
		dc.b	$17
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_4E45C,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		move.w	#$17,d1
		move.w	(a1),d2
		move.w	#5,d3
		moveq	#0,d5
		bset	#0,($FFFFC2F8).w
		jsr	sub_1169A
		bclr	#0,($FFFFC2F8).w
		move.w	($FFFFBF12).w,d4
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_52DAA,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#8,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	$19
		dc.b	1
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_5338C,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#0,d5
		jsr	sub_1169A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$BE
		dc.b	1
		dc.b	$E
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#off_5394E,a0
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#8,d5
		jsr	sub_1169A
		clr.w	($FFFFBD28).w
		clr.l	($FFFFBDA8).w
		clr.w	($FFFFBDAC).w
		move.w	d4,($FFFFB012).w
		move.l	#unk_AAC5A,a2
		jsr	sub_1172C
; ---------------------------------------------------------------------------
		dc.w	$D10
		dc.w	$4567
		dc.w	$89AB
		dc.w	$CDEF
; ---------------------------------------------------------------------------
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$5776,a1
		jsr	sub_FF2C8
		add.w	#$20,($FFFFBD20).w			; ' '
		move.w	#$104,($FFFFB8B2).w
		move.w	#$120,($FFFFB8AE).w
		move.w	#$D0,($FFFFB8B0).w			; ''
		clr.w	($FFFFB044).w
		move.w	#$FFCE,($FFFFD5AC).w
		move.w	#$20,($FFFFBD26).w			; ' '
		move.w	#$78,-(sp)				; 'x'
		jsr	sub_11156
		bsr.w	sub_FF3B0
		move	#$2500,sr

loc_FF26A:							; CODE XREF: ROM:000FF274j
		jsr	sub_FF318
		tst.w	($FFFFB8B2).w
		bne.s	loc_FF26A
		move.w	#$3C,d3					; '<'

loc_FF27A:							; CODE XREF: ROM:000FF280j
		jsr	sub_FF318
		dbf	d3,loc_FF27A
		jsr	sub_11B92
; ---------------------------------------------------------------------------
		dc.w	6
		dc.b	$EF
		dc.b	0
		dc.b	0
		dc.b	0
; ---------------------------------------------------------------------------
		move.l	#$57B8,a1

loc_FF296:							; CODE XREF: ROM:000FF2C4j
		jsr	sub_FF2C8
		add.w	(a1),a1
		moveq	#$27,d4					; '''

loc_FF2A0:							; CODE XREF: ROM:000FF2B0j
		jsr	sub_FF318
		jsr	sub_FF318
		addq.w	#1,($FFFFBD20).w
		dbf	d4,loc_FF2A0
		moveq	#$78,d4					; 'x'

loc_FF2B6:							; CODE XREF: ROM:000FF2BCj
		jsr	sub_FF318
		dbf	d4,loc_FF2B6
		tst.w	2(a1)
		bpl.s	loc_FF296
		rts

; =============== S U B	R O U T	I N E =======================================


sub_FF2C8:							; CODE XREF: ROM:000FF22Ap
					; ROM:loc_FF296p
		move.w	($FFFFBD20).w,d0
		asr.w	#3,d0
		add.w	#$1C,d0
		and.w	#$1F,d0
		move.w	d0,($FFFFB02A).w
		move.w	d0,-(sp)
		clr.w	($FFFFB028).w
		moveq	#$20,d0					; ' '
		moveq	#6,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		move.w	(sp)+,($FFFFB02A).w

loc_FF2F2:							; CODE XREF: sub_FF2C8+4Cj
		move.w	(a1),d0
		asr.w	#1,d0
		neg.w	d0
		add.w	#$11,d0
		move.w	d0,($FFFFB028).w
		jsr	sub_11BA4
		addq.w	#1,($FFFFB02A).w
		and.w	#$1F,($FFFFB02A).w
		tst.w	2(a1)
		bpl.s	loc_FF2F2
		rts
; End of function sub_FF2C8


; =============== S U B	R O U T	I N E =======================================


sub_FF318:							; CODE XREF: ROM:loc_FF26Ap
					; ROM:loc_FF27Ap ...
		movem.l	d0-a6,-(sp)
		move.w	($FFFFB03E).w,d0

loc_FF320:							; CODE XREF: sub_FF318+Cj
		cmp.w	($FFFFB03E).w,d0
		beq.s	loc_FF320
		subq.w	#2,($FFFFB8B2).w
		bpl.w	loc_FF332
		clr.w	($FFFFB8B2).w

loc_FF332:							; CODE XREF: sub_FF318+12j
		jsr	sub_112BC
; End of function sub_FF318

		btst	#7,d1
		movem.l	(sp)+,d0-a6
		beq.w	locret_FF346
		addq.w	#4,sp

locret_FF346:							; CODE XREF: ROM:000FF340j
		rts
; ---------------------------------------------------------------------------

loc_FF348:							; DATA XREF: ROM:000FF056o
		movem.l	d0-a6,-(sp)
		btst	#2,($FFFFBF78).w
		bne.w	loc_FF39A
		move.w	#$B04A,a0
		move.w	($FFFFB000).w,d1
		move.w	#$1C0,d0
		jsr	sub_113E4
		move.l	#$C00000,a0
		move.l	#$40000010,4(a0)
		move.w	($FFFFBD20).w,(a0)
		move.w	#$C018,a0
		move.w	($FFFFC2E8).w,d0
		beq.w	loc_FF394
		clr.w	($FFFFC2E8).w
		move.w	($FFFFB002).w,d1
		jsr	sub_113E4

loc_FF394:							; CODE XREF: ROM:000FF382j
		jsr	sub_10FB6

loc_FF39A:							; CODE XREF: ROM:000FF352j
		addq.w	#1,($FFFFB03E).w
		jsr	sub_FF3F8
		jsr	sub_1A50A
		movem.l	(sp)+,d0-a6
		rte

; =============== S U B	R O U T	I N E =======================================


sub_FF3B0:							; CODE XREF: ROM:000FF262p
		movem.l	d0-d1,-(sp)
		move.w	#$14,($FFFFBD1E).w
		move.w	#$50,($FFFFBF14).w			; 'P'
		move.w	#$50,d0					; 'P'
		move.l	#$FFFFB04A,a0
		move.l	#$FD80,d1

loc_FF3D0:							; CODE XREF: sub_FF3B0:loc_FF3DCj
		move.l	d1,(a0)+
		btst	#0,d0
		bne.w	loc_FF3DC
		subq.w	#1,d1

loc_FF3DC:							; CODE XREF: sub_FF3B0+26j
		dbf	d0,loc_FF3D0
		move.w	#$60,d0					; '`'
		move.l	#$100,d1

loc_FF3EA:							; CODE XREF: sub_FF3B0+3Ej
		move.l	d1,(a0)+
		addq.l	#4,d1
		dbf	d0,loc_FF3EA
		movem.l	(sp)+,d0-d1
		rts
; End of function sub_FF3B0


; =============== S U B	R O U T	I N E =======================================


sub_FF3F8:							; CODE XREF: ROM:000FF39Ep
		subq.w	#1,($FFFFBF14).w
		bmi.w	loc_FF402
		rts
; ---------------------------------------------------------------------------

loc_FF402:							; CODE XREF: sub_FF3F8+4j
		clr.w	($FFFFBF14).w
		movem.l	d0-d2/a0,-(sp)
		move.w	#$DF,d0					; ''
		move.w	#$B04A,a0

loc_FF412:							; CODE XREF: sub_FF3F8:loc_FF458j
		tst.l	(a0)+
		cmpnw	 $28,d0					; '('
		ble.w	loc_FF458
		move.w	($FFFFBD1E).w,d1
		cmpnw	 $8F,d0					; ''
		blt.w	loc_FF44C
		move.w	-2(a0),d2
		beq.w	loc_FF458
		add.w	d1,d2
		move.w	d2,-2(a0)
		and.w	#$FC00,d2
		cmpnw	 $FC00,d2
		beq.w	loc_FF458
		move.w	#0,-2(a0)
		bra.w	loc_FF458
; ---------------------------------------------------------------------------

loc_FF44C:							; CODE XREF: sub_FF3F8+2Cj
		sub.w	d1,-2(a0)
		bpl.w	loc_FF458
		clr.w	-2(a0)

loc_FF458:							; CODE XREF: sub_FF3F8+20j
					; sub_FF3F8+34j ...
		dbf	d0,loc_FF412
		movem.l	(sp)+,d0-d2/a0
		rts
; End of function sub_FF3F8

; ---------------------------------------------------------------------------
unk_FF462:	include	"Unknown\FF462.asm"

; =============== S U B	R O U T	I N E =======================================


sub_FF7E2:							; CODE XREF: ROM:0000FCA6p

; FUNCTION CHUNK AT 000FF884 SIZE 0000000A BYTES

		movem.l	d0/a0-a3,-(sp)
		move.w	($FFFFC2EE).w,-(sp)
		btst	#1,($FFFFC2FA).w
		bne.w	loc_FF884
		move.w	#$C6CE,a2
		lea	$364(a2),a3
		move.w	$24(a2),d0
		sub.w	$24(a3),d0
		beq.w	loc_FF884
		bpl.w	loc_FF824
		btst	#6,($FFFFC2EE).w
		bne.w	loc_FF83A
		bsr.w	sub_FF87C
		bset	#6,($FFFFC2EE).w
		bra.w	loc_FF83A
; ---------------------------------------------------------------------------

loc_FF824:							; CODE XREF: sub_FF7E2+26j
		exg	a2,a3
		btst	#6,($FFFFC2EE).w
		beq.w	loc_FF83A
		bsr.w	sub_FF87C
		bclr	#6,($FFFFC2EE).w

loc_FF83A:							; CODE XREF: sub_FF7E2+30j
					; sub_FF7E2+3Ej ...
		bset	#5,($FFFFC2EE).w
		bne.w	loc_FF884
		cmp.w	#$C6CE,a3
		bne.w	loc_FF868
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#1,($FFFFD6CC).w
		jsr	sub_FE556

loc_FF85E:							; CODE XREF: sub_FF7E2+98j
		bset	#6,($FFFFC2FE).w
		bra.w	loc_FF884
; ---------------------------------------------------------------------------

loc_FF868:							; CODE XREF: sub_FF7E2+66j
		move.w	($FFFFC330).w,($FFFFD6CA).w
		move.w	#4,($FFFFD6CC).w
		jsr	sub_FE556
		bra.s	loc_FF85E
; End of function sub_FF7E2


; =============== S U B	R O U T	I N E =======================================


sub_FF87C:							; CODE XREF: sub_FF7E2+34p
					; sub_FF7E2+4Ep
		bclr	#5,($FFFFC2EE).w
		rts
; End of function sub_FF87C

; ---------------------------------------------------------------------------
; START	OF FUNCTION CHUNK FOR sub_FF7E2

loc_FF884:							; CODE XREF: sub_FF7E2+Ej
					; sub_FF7E2+22j ...
		move.w	(sp)+,($FFFFC2EE).w
		movem.l	(sp)+,d0/a0-a3
		rts
; END OF FUNCTION CHUNK	FOR sub_FF7E2

; =============== S U B	R O U T	I N E =======================================


sub_FF88E:							; CODE XREF: sub_7B30+FEp
		movem.l	d0/a0,-(sp)
		clr.w	($FFFFC3EA).w
		clr.w	($FFFFC3E8).w
		clr.w	($FFFFC3E6).w
		move.w	#$10,d0
		move.l	#$FFFFC3A4,a0

loc_FF8A8:							; CODE XREF: sub_FF88E+1Cj
		clr.l	(a0)+
		dbf	d0,loc_FF8A8
		move.l	#$FFFFC6CE,a0
		bsr.w	sub_FF8C8
		move.l	#$FFFFCA32,a0
		bsr.w	sub_FF8C8
		movem.l	(sp)+,d0/a0
		rts
; End of function sub_FF88E


; =============== S U B	R O U T	I N E =======================================


sub_FF8C8:							; CODE XREF: sub_FF88E+26p
					; sub_FF88E+30p
		move.w	#$19,d0
		add.w	#$66,a0					; 'f'

loc_FF8D0:							; CODE XREF: sub_FF8C8+Cj
		move.w	#$FFFE,(a0)+
		dbf	d0,loc_FF8D0
		move.w	#$FFFF,(a0)
		rts
; End of function sub_FF8C8


; =============== S U B	R O U T	I N E =======================================


sub_FF8DE:							; CODE XREF: sub_FD89A+46p
		movem.l	d0/a0-a1,-(sp)
		move.l	#$FFFFBF5C,a0
		move.w	#2,($FFFFDEEA).w
		cmp.l	#$FFFFC6CE,a2
		bne.w	loc_FF904
		move.l	#$FFFFBF5E,a0
		move.w	#$20,($FFFFDEEA).w			; ' '

loc_FF904:							; CODE XREF: sub_FF8DE+16j
		move.w	#0,d0

loc_FF908:							; CODE XREF: sub_FF8DE+34j
		move.w	(a0)+,d1
		cmp.w	($FFFFDEE8).w,d1
		beq.w	loc_FF960
		dbf	d0,loc_FF908
		move.l	#$FFFFBF60,a0
		cmp.l	#$FFFFC6CE,a2
		bne.w	loc_FF92C
		move.l	#$FFFFBF62,a0

loc_FF92C:							; CODE XREF: sub_FF8DE+44j
		move.w	#0,d0

loc_FF930:							; CODE XREF: sub_FF8DE+5Cj
		move.w	(a0)+,d1
		cmp.w	($FFFFDEE8).w,d1
		beq.w	loc_FF96E
		dbf	d0,loc_FF930
		move.w	($FFFFDEEA).w,($FFFFB028).w
		move.w	#$19,($FFFFB02A).w
		move.w	#8,d0
		move.w	#3,d1
		move.w	#$7FF,d2
		jsr	sub_1197E
		bra.w	loc_FF9A2
; ---------------------------------------------------------------------------

loc_FF960:							; CODE XREF: sub_FF8DE+30j
		move.w	($FFFFDEE4).w,d4
		move.l	#unk_F5AF6,a0
		bra.w	loc_FF978
; ---------------------------------------------------------------------------

loc_FF96E:							; CODE XREF: sub_FF8DE+58j
		move.w	($FFFFDEE6).w,d4
		move.l	#unk_F5D1C,a0

loc_FF978:							; CODE XREF: sub_FF8DE+8Cj
		move.w	($FFFFDEEA).w,($FFFFB028).w
		move.w	#$19,($FFFFB02A).w
		move.l	a0,a1
		move.l	a0,a2
		add.l	(a2)+,a0
		add.l	(a2)+,a1
		move.w	#$30A,a2
		clr.w	d0
		clr.w	d1
		move.w	(a1),d2
		move.w	2(a1),d3
		moveq	#0,d5
		jsr	sub_1169A

loc_FF9A2:							; CODE XREF: sub_FF8DE+7Ej
		movem.l	(sp)+,d0/a0-a1
		rts
; End of function sub_FF8DE


; =============== S U B	R O U T	I N E =======================================


sub_FF9A8:							; CODE XREF: sub_BFBCj

; FUNCTION CHUNK AT 0000C0AE SIZE 0000000E BYTES

		btst	#6,($FFFFC2F6).w
		bne.w	locret_FFAAC
		movem.l	d0-d6/a0-a1,-(sp)
		move.w	($FFFFB772).w,d0
		asr.w	#8,d0
		add.w	($FFFFB74A).w,d0
		move.w	($FFFFB774).w,d1
		asr.w	#8,d1
		add.w	($FFFFB75E).w,d1
		movem.w	d0-d1,-(sp)
		moveq	#5,d2
		move.w	d4,d3
		eor.w	#2,d3
		moveq	#$FFFFFFFF,d5
		move.w	#$B04A,a0
		move.w	#$C328,a1
		cmp.w	#1,(a1,d4.w)
		beq.w	loc_FF9EE
		add.w	#$300,a0

loc_FF9EE:							; CODE XREF: sub_FF9A8+3Ej
		move.w	#$C320,a1

loc_FF9F2:							; CODE XREF: sub_FF9A8+E0j
		tst.w	$34(a0)
		ble.w	loc_FFA84
		btst	#2,$63(a0)
		bne.w	loc_FFA84
		movem.w	d1,-(sp)
		move.w	$52(a0),d1
		cmp.w	(a1,d4.w),d1
		movem.w	(sp)+,d1
		beq.w	loc_FFA22
		btst	#3,$62(a0)
		bne.w	loc_FFA84

loc_FFA22:							; CODE XREF: sub_FF9A8+6Cj
		btst	#2,($FFFFC2F2).w
		beq.w	loc_FFA54
		movem.l	d0,-(sp)
		move.w	($FFFFD406).w,d0
		cmp.w	$52(a0),d0
		movem.l	(sp)+,d0
		beq.w	loc_FFA54
		movem.l	d0,-(sp)
		move.w	($FFFFD408).w,d0
		cmp.w	$52(a0),d0
		movem.l	(sp)+,d0
		bne.w	loc_FFA84

loc_FFA54:							; CODE XREF: sub_FF9A8+80j
					; sub_FF9A8+94j
		btst	#5,$62(a0)
		bne.w	loc_FFA84
		movem.w	(sp),d0-d1
		sub.w	(a0),d0
		muls.w	d0,d0
		sub.w	$14(a0),d1
		muls.w	d1,d1
		add.l	d1,d0
		cmp.l	d5,d0
		bhi.w	loc_FFA84
		move.w	$52(a0),d1
		cmp.w	(a1,d3.w),d1
		beq.w	loc_FFA84
		move.l	d0,d5
		move.w	d1,d6

loc_FFA84:							; CODE XREF: sub_FF9A8+4Ej
					; sub_FF9A8+58j ...
		add.w	#$80,a0					; ''
		dbf	d2,loc_FF9F2
		addq.w	#4,sp
		pea	(loc_FFAA8).l
		cmp.w	(a1,d4.w),d6
		beq.w	loc_FFAAE
		move.w	d6,d0
		tst.w	d4
		beq.w	loc_FFAB4
		bra.w	loc_FFABA
; ---------------------------------------------------------------------------

loc_FFAA8:							; DATA XREF: sub_FF9A8+E6o
		movem.l	(sp)+,d0-d6/a0-a1

locret_FFAAC:							; CODE XREF: sub_FF9A8+6j
		rts
; ---------------------------------------------------------------------------

loc_FFAAE:							; CODE XREF: sub_FF9A8+F0j
		jmp	loc_C0AE
; ---------------------------------------------------------------------------

loc_FFAB4:							; CODE XREF: sub_FF9A8+F8j
		jmp	sub_C0BC
; ---------------------------------------------------------------------------

loc_FFABA:							; CODE XREF: sub_FF9A8+FCj
		jmp	sub_C0DA
; End of function sub_FF9A8


; =============== S U B	R O U T	I N E =======================================

CalcChecksum:
		align	$100000,$00
FinishRom:
		END
